"""csv plugin for translating Financial Information eXchange (FIX) messages.

https://github.com/markuskimius/csvmagic
"""

from csvmagic.libcsv import Cell

__copyright__ = "Copyright 2018-2024 Mark Kim"
__license__ = "Apache 2.0"
__version__ = '2.2.2'
__author__ = "Mark Kim"


class Translator(object):
    def __init__(self):
        self.__fixversion = None
        self.__has_header = False

    def __call__(self, cell):
        tagnum = cell.colname()
        value = cell.value()

        # Are we expected to extract our own header or provide them?
        if tagnum is not None:
            self.__has_header = True

        # Extract the header from the value if required
        if not self.__has_header and value.find('=') >= 0:
            tagnum, value = value.split("=", 1)

        # Save the FIX version for more precise translation
        if tagnum == '8':
            self.__fixversion = value

        # Translate the tag name and value
        tagname = self.__get_tag(tagnum, value)
        value = self.__get_value(tagnum, value)

        # We don't want tagname to be None because if it gets set to None then the
        # read command displays the column number which looks confusing because it
        # looks like a tag number but they're not the same thing!
        if tagname is None: tagname = ''

        return Cell(cell.rownum(), tagname, cell.colnum(), value)

    def __get_tag(self, tagnum, value):
        tagname = Translator.__TAG_MAP.get((self.__fixversion, tagnum),
                  Translator.__TAG_MAP.get(("FIX.5.0", tagnum),
                  Translator.__TAG_MAP.get(("FIX.4.4", tagnum),
                  Translator.__TAG_MAP.get(("FIX.4.2", tagnum),
                  Translator.__TAG_MAP.get(("FIX.4.0", tagnum), tagnum)))))

        return tagname

    def __get_value(self, tagnum, value):
        if(tagnum in ("30", "100", "207")):
            value = Translator.__MARKET_MAP.get(("MIC", value),
                    Translator.__MARKET_MAP.get(("RIC", value), value))
        else:
            value = Translator.__VALUE_MAP.get((self.__fixversion, tagnum, value),
                    Translator.__VALUE_MAP.get(("FIX.5.0", tagnum, value),
                    Translator.__VALUE_MAP.get(("FIX.4.4", tagnum, value),
                    Translator.__VALUE_MAP.get(("FIX.4.2", tagnum, value),
                    Translator.__VALUE_MAP.get(("FIX.4.0", tagnum, value), value)))))

        return value

    __TAG_MAP = {
        ("FIX.4.0", "1"): "1    Account",
        ("FIX.4.0", "2"): "2    AdvId",
        ("FIX.4.0", "3"): "3    AdvRefID",
        ("FIX.4.0", "4"): "4    AdvSide",
        ("FIX.4.0", "5"): "5    AdvTransType",
        ("FIX.4.0", "6"): "6    AvgPx",
        ("FIX.4.0", "7"): "7    BeginSeqNo",
        ("FIX.4.0", "8"): "8    BeginString",
        ("FIX.4.0", "9"): "9    BodyLength",
        ("FIX.4.0", "10"): "10   CheckSum",
        ("FIX.4.0", "11"): "11   ClOrdID",
        ("FIX.4.0", "12"): "12   Commission",
        ("FIX.4.0", "13"): "13   CommType",
        ("FIX.4.0", "14"): "14   CumQty",
        ("FIX.4.0", "15"): "15   Currency",
        ("FIX.4.0", "16"): "16   EndSeqNo",
        ("FIX.4.0", "17"): "17   ExecID",
        ("FIX.4.0", "18"): "18   ExecInst",
        ("FIX.4.0", "19"): "19   ExecRefID",
        ("FIX.4.0", "20"): "20   ExecTransType",
        ("FIX.4.0", "21"): "21   HandlInst",
        ("FIX.4.0", "22"): "22   IDSource",
        ("FIX.4.0", "23"): "23   IOIid",
        ("FIX.4.0", "24"): "24   IOIOthSvc",
        ("FIX.4.0", "25"): "25   IOIQltyInd",
        ("FIX.4.0", "26"): "26   IOIRefID",
        ("FIX.4.0", "27"): "27   IOIShares",
        ("FIX.4.0", "28"): "28   IOITransType",
        ("FIX.4.0", "29"): "29   LastCapacity",
        ("FIX.4.0", "30"): "30   LastMkt",
        ("FIX.4.0", "31"): "31   LastPx",
        ("FIX.4.0", "32"): "32   LastShares",
        ("FIX.4.0", "33"): "33   LinesOfText",
        ("FIX.4.0", "34"): "34   MsgSeqNum",
        ("FIX.4.0", "35"): "35   MsgType",
        ("FIX.4.0", "36"): "36   NewSeqNo",
        ("FIX.4.0", "37"): "37   OrderID",
        ("FIX.4.0", "38"): "38   OrderQty",
        ("FIX.4.0", "39"): "39   OrdStatus",
        ("FIX.4.0", "40"): "40   OrdType",
        ("FIX.4.0", "41"): "41   OrigClOrdID",
        ("FIX.4.0", "42"): "42   OrigTime",
        ("FIX.4.0", "43"): "43   PossDupFlag",
        ("FIX.4.0", "44"): "44   Price",
        ("FIX.4.0", "45"): "45   RefSeqNum",
        ("FIX.4.0", "46"): "46   RelatdSym",
        ("FIX.4.0", "47"): "47   Rule80A",
        ("FIX.4.0", "48"): "48   SecurityID",
        ("FIX.4.0", "49"): "49   SenderCompID",
        ("FIX.4.0", "50"): "50   SenderSubID",
        ("FIX.4.0", "52"): "52   SendingTime",
        ("FIX.4.0", "53"): "53   Shares",
        ("FIX.4.0", "54"): "54   Side",
        ("FIX.4.0", "55"): "55   Symbol",
        ("FIX.4.0", "56"): "56   TargetCompID",
        ("FIX.4.0", "57"): "57   TargetSubID",
        ("FIX.4.0", "58"): "58   Text",
        ("FIX.4.0", "59"): "59   TimeInForce",
        ("FIX.4.0", "60"): "60   TransactTime",
        ("FIX.4.0", "61"): "61   Urgency",
        ("FIX.4.0", "62"): "62   ValidUntilTime",
        ("FIX.4.0", "63"): "63   SettlmntTyp",
        ("FIX.4.0", "64"): "64   FutSettDate",
        ("FIX.4.0", "65"): "65   SymbolSfx",
        ("FIX.4.0", "66"): "66   ListID",
        ("FIX.4.0", "67"): "67   ListSeqNo",
        ("FIX.4.0", "68"): "68   ListNoOrds",
        ("FIX.4.0", "69"): "69   ListExecInst",
        ("FIX.4.0", "70"): "70   AllocID",
        ("FIX.4.0", "71"): "71   AllocTransType",
        ("FIX.4.0", "72"): "72   RefAllocID",
        ("FIX.4.0", "73"): "73   NoOrders",
        ("FIX.4.0", "74"): "74   AvgPrxPrecision",
        ("FIX.4.0", "75"): "75   TradeDate",
        ("FIX.4.0", "76"): "76   ExecBroker",
        ("FIX.4.0", "77"): "77   OpenClose",
        ("FIX.4.0", "78"): "78   NoAllocs",
        ("FIX.4.0", "79"): "79   AllocAccount",
        ("FIX.4.0", "80"): "80   AllocShares",
        ("FIX.4.0", "81"): "81   ProcessCode",
        ("FIX.4.0", "82"): "82   NoRpts",
        ("FIX.4.0", "83"): "83   RptSeq",
        ("FIX.4.0", "84"): "84   CxlQty",
        ("FIX.4.0", "85"): "85   NoDlvyInst",
        ("FIX.4.0", "86"): "86   DlvyInst",
        ("FIX.4.0", "87"): "87   AllocStatus",
        ("FIX.4.0", "88"): "88   AllocRejCode",
        ("FIX.4.0", "89"): "89   Signature",
        ("FIX.4.0", "90"): "90   SecureDataLen",
        ("FIX.4.0", "91"): "91   SecureData",
        ("FIX.4.0", "92"): "92   BrokerOfCredit",
        ("FIX.4.0", "93"): "93   SignatureLength",
        ("FIX.4.0", "94"): "94   EmailType",
        ("FIX.4.0", "95"): "95   RawDataLength",
        ("FIX.4.0", "96"): "96   RawData",
        ("FIX.4.0", "97"): "97   PossResend",
        ("FIX.4.0", "98"): "98   EncryptMethod",
        ("FIX.4.0", "99"): "99   StopPx",
        ("FIX.4.0", "100"): "100  ExDestination",
        ("FIX.4.0", "102"): "102  CxlRejReason",
        ("FIX.4.0", "103"): "103  OrdRejReason",
        ("FIX.4.0", "104"): "104  IOIQualifier",
        ("FIX.4.0", "105"): "105  WaveNo",
        ("FIX.4.0", "106"): "106  Issuer",
        ("FIX.4.0", "107"): "107  SecurityDesc",
        ("FIX.4.0", "108"): "108  HeartBtInt",
        ("FIX.4.0", "109"): "109  ClientID",
        ("FIX.4.0", "110"): "110  MinQty",
        ("FIX.4.0", "111"): "111  MaxFloor",
        ("FIX.4.0", "112"): "112  TestReqID",
        ("FIX.4.0", "113"): "113  ReportToExch",
        ("FIX.4.0", "114"): "114  LocateReqd",
        ("FIX.4.0", "115"): "115  OnBehalfOfCompID",
        ("FIX.4.0", "116"): "116  OnBehalfOfSubID",
        ("FIX.4.0", "117"): "117  QuoteID",
        ("FIX.4.0", "118"): "118  NetMoney",
        ("FIX.4.0", "119"): "119  SettlCurrAmt",
        ("FIX.4.0", "120"): "120  SettlCurrency",
        ("FIX.4.0", "121"): "121  ForexReq",
        ("FIX.4.0", "122"): "122  OrigSendingTime",
        ("FIX.4.0", "123"): "123  GapFillFlag",
        ("FIX.4.0", "124"): "124  NoExecs",
        ("FIX.4.0", "125"): "125  CxlType",
        ("FIX.4.0", "126"): "126  ExpireTime",
        ("FIX.4.0", "127"): "127  DKReason",
        ("FIX.4.0", "128"): "128  DeliverToCompID",
        ("FIX.4.0", "129"): "129  DeliverToSubID",
        ("FIX.4.0", "130"): "130  IOINaturalFlag",
        ("FIX.4.0", "131"): "131  QuoteReqID",
        ("FIX.4.0", "132"): "132  BidPx",
        ("FIX.4.0", "133"): "133  OfferPx",
        ("FIX.4.0", "134"): "134  BidSize",
        ("FIX.4.0", "135"): "135  OfferSize",
        ("FIX.4.0", "136"): "136  NoMiscFees",
        ("FIX.4.0", "137"): "137  MiscFeeAmt",
        ("FIX.4.0", "138"): "138  MiscFeeCurr",
        ("FIX.4.0", "139"): "139  MiscFeeType",
        ("FIX.4.0", "140"): "140  PrevClosePx",
        ("FIX.4.2", "1"): "1    Account",
        ("FIX.4.2", "2"): "2    AdvId",
        ("FIX.4.2", "3"): "3    AdvRefID",
        ("FIX.4.2", "4"): "4    AdvSide",
        ("FIX.4.2", "5"): "5    AdvTransType",
        ("FIX.4.2", "6"): "6    AvgPx",
        ("FIX.4.2", "7"): "7    BeginSeqNo",
        ("FIX.4.2", "8"): "8    BeginString",
        ("FIX.4.2", "9"): "9    BodyLength",
        ("FIX.4.2", "10"): "10   CheckSum",
        ("FIX.4.2", "11"): "11   ClOrdID",
        ("FIX.4.2", "12"): "12   Commission",
        ("FIX.4.2", "13"): "13   CommType",
        ("FIX.4.2", "14"): "14   CumQty",
        ("FIX.4.2", "15"): "15   Currency",
        ("FIX.4.2", "16"): "16   EndSeqNo",
        ("FIX.4.2", "17"): "17   ExecID",
        ("FIX.4.2", "18"): "18   ExecInst",
        ("FIX.4.2", "19"): "19   ExecRefID",
        ("FIX.4.2", "20"): "20   ExecTransType",
        ("FIX.4.2", "21"): "21   HandlInst",
        ("FIX.4.2", "22"): "22   IDSource",
        ("FIX.4.2", "23"): "23   IOIid",
        ("FIX.4.2", "24"): "24   IOIOthSvc",
        ("FIX.4.2", "25"): "25   IOIQltyInd",
        ("FIX.4.2", "26"): "26   IOIRefID",
        ("FIX.4.2", "27"): "27   IOIShares",
        ("FIX.4.2", "28"): "28   IOITransType",
        ("FIX.4.2", "29"): "29   LastCapacity",
        ("FIX.4.2", "30"): "30   LastMkt",
        ("FIX.4.2", "31"): "31   LastPx",
        ("FIX.4.2", "32"): "32   LastShares",
        ("FIX.4.2", "33"): "33   LinesOfText",
        ("FIX.4.2", "34"): "34   MsgSeqNum",
        ("FIX.4.2", "35"): "35   MsgType",
        ("FIX.4.2", "36"): "36   NewSeqNo",
        ("FIX.4.2", "37"): "37   OrderID",
        ("FIX.4.2", "38"): "38   OrderQty",
        ("FIX.4.2", "39"): "39   OrdStatus",
        ("FIX.4.2", "40"): "40   OrdType",
        ("FIX.4.2", "41"): "41   OrigClOrdID",
        ("FIX.4.2", "42"): "42   OrigTime",
        ("FIX.4.2", "43"): "43   PossDupFlag",
        ("FIX.4.2", "44"): "44   Price",
        ("FIX.4.2", "45"): "45   RefSeqNum",
        ("FIX.4.2", "46"): "46   RelatdSym",
        ("FIX.4.2", "47"): "47   Rule80A",
        ("FIX.4.2", "48"): "48   SecurityID",
        ("FIX.4.2", "49"): "49   SenderCompID",
        ("FIX.4.2", "50"): "50   SenderSubID",
        ("FIX.4.2", "51"): "51   SendingDate",
        ("FIX.4.2", "52"): "52   SendingTime",
        ("FIX.4.2", "53"): "53   Shares",
        ("FIX.4.2", "54"): "54   Side",
        ("FIX.4.2", "55"): "55   Symbol",
        ("FIX.4.2", "56"): "56   TargetCompID",
        ("FIX.4.2", "57"): "57   TargetSubID",
        ("FIX.4.2", "58"): "58   Text",
        ("FIX.4.2", "59"): "59   TimeInForce",
        ("FIX.4.2", "60"): "60   TransactTime",
        ("FIX.4.2", "61"): "61   Urgency",
        ("FIX.4.2", "62"): "62   ValidUntilTime",
        ("FIX.4.2", "63"): "63   SettlmntTyp",
        ("FIX.4.2", "64"): "64   FutSettDate",
        ("FIX.4.2", "65"): "65   SymbolSfx",
        ("FIX.4.2", "66"): "66   ListID",
        ("FIX.4.2", "67"): "67   ListSeqNo",
        ("FIX.4.2", "68"): "68   TotNoOrders",
        ("FIX.4.2", "69"): "69   ListExecInst",
        ("FIX.4.2", "70"): "70   AllocID",
        ("FIX.4.2", "71"): "71   AllocTransType",
        ("FIX.4.2", "72"): "72   RefAllocID",
        ("FIX.4.2", "73"): "73   NoOrders",
        ("FIX.4.2", "74"): "74   AvgPrxPrecision",
        ("FIX.4.2", "75"): "75   TradeDate",
        ("FIX.4.2", "76"): "76   ExecBroker",
        ("FIX.4.2", "77"): "77   OpenClose",
        ("FIX.4.2", "78"): "78   NoAllocs",
        ("FIX.4.2", "79"): "79   AllocAccount",
        ("FIX.4.2", "80"): "80   AllocShares",
        ("FIX.4.2", "81"): "81   ProcessCode",
        ("FIX.4.2", "82"): "82   NoRpts",
        ("FIX.4.2", "83"): "83   RptSeq",
        ("FIX.4.2", "84"): "84   CxlQty",
        ("FIX.4.2", "85"): "85   NoDlvyInst",
        ("FIX.4.2", "86"): "86   DlvyInst",
        ("FIX.4.2", "87"): "87   AllocStatus",
        ("FIX.4.2", "88"): "88   AllocRejCode",
        ("FIX.4.2", "89"): "89   Signature",
        ("FIX.4.2", "90"): "90   SecureDataLen",
        ("FIX.4.2", "91"): "91   SecureData",
        ("FIX.4.2", "92"): "92   BrokerOfCredit",
        ("FIX.4.2", "93"): "93   SignatureLength",
        ("FIX.4.2", "94"): "94   EmailType",
        ("FIX.4.2", "95"): "95   RawDataLength",
        ("FIX.4.2", "96"): "96   RawData",
        ("FIX.4.2", "97"): "97   PossResend",
        ("FIX.4.2", "98"): "98   EncryptMethod",
        ("FIX.4.2", "99"): "99   StopPx",
        ("FIX.4.2", "100"): "100  ExDestination",
        ("FIX.4.2", "102"): "102  CxlRejReason",
        ("FIX.4.2", "103"): "103  OrdRejReason",
        ("FIX.4.2", "104"): "104  IOIQualifier",
        ("FIX.4.2", "105"): "105  WaveNo",
        ("FIX.4.2", "106"): "106  Issuer",
        ("FIX.4.2", "107"): "107  SecurityDesc",
        ("FIX.4.2", "108"): "108  HeartBtInt",
        ("FIX.4.2", "109"): "109  ClientID",
        ("FIX.4.2", "110"): "110  MinQty",
        ("FIX.4.2", "111"): "111  MaxFloor",
        ("FIX.4.2", "112"): "112  TestReqID",
        ("FIX.4.2", "113"): "113  ReportToExch",
        ("FIX.4.2", "114"): "114  LocateReqd",
        ("FIX.4.2", "115"): "115  OnBehalfOfCompID",
        ("FIX.4.2", "116"): "116  OnBehalfOfSubID",
        ("FIX.4.2", "117"): "117  QuoteID",
        ("FIX.4.2", "118"): "118  NetMoney",
        ("FIX.4.2", "119"): "119  SettlCurrAmt",
        ("FIX.4.2", "120"): "120  SettlCurrency",
        ("FIX.4.2", "121"): "121  ForexReq",
        ("FIX.4.2", "122"): "122  OrigSendingTime",
        ("FIX.4.2", "123"): "123  GapFillFlag",
        ("FIX.4.2", "124"): "124  NoExecs",
        ("FIX.4.2", "125"): "125  CxlType",
        ("FIX.4.2", "126"): "126  ExpireTime",
        ("FIX.4.2", "127"): "127  DKReason",
        ("FIX.4.2", "128"): "128  DeliverToCompID",
        ("FIX.4.2", "129"): "129  DeliverToSubID",
        ("FIX.4.2", "130"): "130  IOINaturalFlag",
        ("FIX.4.2", "131"): "131  QuoteReqID",
        ("FIX.4.2", "132"): "132  BidPx",
        ("FIX.4.2", "133"): "133  OfferPx",
        ("FIX.4.2", "134"): "134  BidSize",
        ("FIX.4.2", "135"): "135  OfferSize",
        ("FIX.4.2", "136"): "136  NoMiscFees",
        ("FIX.4.2", "137"): "137  MiscFeeAmt",
        ("FIX.4.2", "138"): "138  MiscFeeCurr",
        ("FIX.4.2", "139"): "139  MiscFeeType",
        ("FIX.4.2", "140"): "140  PrevClosePx",
        ("FIX.4.2", "141"): "141  ResetSeqNumFlag",
        ("FIX.4.2", "142"): "142  SenderLocationID",
        ("FIX.4.2", "143"): "143  TargetLocationID",
        ("FIX.4.2", "144"): "144  OnBehalfOfLocationID",
        ("FIX.4.2", "145"): "145  DeliverToLocationID",
        ("FIX.4.2", "146"): "146  NoRelatedSym",
        ("FIX.4.2", "147"): "147  Subject",
        ("FIX.4.2", "148"): "148  Headline",
        ("FIX.4.2", "149"): "149  URLLink",
        ("FIX.4.2", "150"): "150  ExecType",
        ("FIX.4.2", "151"): "151  LeavesQty",
        ("FIX.4.2", "152"): "152  CashOrderQty",
        ("FIX.4.2", "153"): "153  AllocAvgPx",
        ("FIX.4.2", "154"): "154  AllocNetMoney",
        ("FIX.4.2", "155"): "155  SettlCurrFxRate",
        ("FIX.4.2", "156"): "156  SettlCurrFxRateCalc",
        ("FIX.4.2", "157"): "157  NumDaysInterest",
        ("FIX.4.2", "158"): "158  AccruedInterestRate",
        ("FIX.4.2", "159"): "159  AccruedInterestAmt",
        ("FIX.4.2", "160"): "160  SettlInstMode",
        ("FIX.4.2", "161"): "161  AllocText",
        ("FIX.4.2", "162"): "162  SettlInstID",
        ("FIX.4.2", "163"): "163  SettlInstTransType",
        ("FIX.4.2", "164"): "164  EmailThreadID",
        ("FIX.4.2", "165"): "165  SettlInstSource",
        ("FIX.4.2", "166"): "166  SettlLocation",
        ("FIX.4.2", "167"): "167  SecurityType",
        ("FIX.4.2", "168"): "168  EffectiveTime",
        ("FIX.4.2", "169"): "169  StandInstDbType",
        ("FIX.4.2", "170"): "170  StandInstDbName",
        ("FIX.4.2", "171"): "171  StandInstDbID",
        ("FIX.4.2", "172"): "172  SettlDeliveryType",
        ("FIX.4.2", "173"): "173  SettlDepositoryCode",
        ("FIX.4.2", "174"): "174  SettlBrkrCode",
        ("FIX.4.2", "175"): "175  SettlInstCode",
        ("FIX.4.2", "176"): "176  SecuritySettlAgentName",
        ("FIX.4.2", "177"): "177  SecuritySettlAgentCode",
        ("FIX.4.2", "178"): "178  SecuritySettlAgentAcctNum",
        ("FIX.4.2", "179"): "179  SecuritySettlAgentAcctName",
        ("FIX.4.2", "180"): "180  SecuritySettlAgentContactName",
        ("FIX.4.2", "181"): "181  SecuritySettlAgentContactPhone",
        ("FIX.4.2", "182"): "182  CashSettlAgentName",
        ("FIX.4.2", "183"): "183  CashSettlAgentCode",
        ("FIX.4.2", "184"): "184  CashSettlAgentAcctNum",
        ("FIX.4.2", "185"): "185  CashSettlAgentAcctName",
        ("FIX.4.2", "186"): "186  CashSettlAgentContactName",
        ("FIX.4.2", "187"): "187  CashSettlAgentContactPhone",
        ("FIX.4.2", "188"): "188  BidSpotRate",
        ("FIX.4.2", "189"): "189  BidForwardPoints",
        ("FIX.4.2", "190"): "190  OfferSpotRate",
        ("FIX.4.2", "191"): "191  OfferForwardPoints",
        ("FIX.4.2", "192"): "192  OrderQty2",
        ("FIX.4.2", "193"): "193  FutSettDate2",
        ("FIX.4.2", "194"): "194  LastSpotRate",
        ("FIX.4.2", "195"): "195  LastForwardPoints",
        ("FIX.4.2", "196"): "196  AllocLinkID",
        ("FIX.4.2", "197"): "197  AllocLinkType",
        ("FIX.4.2", "198"): "198  SecondaryOrderID",
        ("FIX.4.2", "199"): "199  NoIOIQualifiers",
        ("FIX.4.2", "200"): "200  MaturityMonthYear",
        ("FIX.4.2", "201"): "201  PutOrCall",
        ("FIX.4.2", "202"): "202  StrikePrice",
        ("FIX.4.2", "203"): "203  CoveredOrUncovered",
        ("FIX.4.2", "204"): "204  CustomerOrFirm",
        ("FIX.4.2", "205"): "205  MaturityDay",
        ("FIX.4.2", "206"): "206  OptAttribute",
        ("FIX.4.2", "207"): "207  SecurityExchange",
        ("FIX.4.2", "208"): "208  NotifyBrokerOfCredit",
        ("FIX.4.2", "209"): "209  AllocHandlInst",
        ("FIX.4.2", "210"): "210  MaxShow",
        ("FIX.4.2", "211"): "211  PegDifference",
        ("FIX.4.2", "212"): "212  XmlDataLen",
        ("FIX.4.2", "213"): "213  XmlData",
        ("FIX.4.2", "214"): "214  SettlInstRefID",
        ("FIX.4.2", "215"): "215  NoRoutingIDs",
        ("FIX.4.2", "216"): "216  RoutingType",
        ("FIX.4.2", "217"): "217  RoutingID",
        ("FIX.4.2", "218"): "218  SpreadToBenchmark",
        ("FIX.4.2", "219"): "219  Benchmark",
        ("FIX.4.2", "223"): "223  CouponRate",
        ("FIX.4.2", "231"): "231  ContractMultiplier",
        ("FIX.4.2", "262"): "262  MDReqID",
        ("FIX.4.2", "263"): "263  SubscriptionRequestType",
        ("FIX.4.2", "264"): "264  MarketDepth",
        ("FIX.4.2", "265"): "265  MDUpdateType",
        ("FIX.4.2", "266"): "266  AggregatedBook",
        ("FIX.4.2", "267"): "267  NoMDEntryTypes",
        ("FIX.4.2", "268"): "268  NoMDEntries",
        ("FIX.4.2", "269"): "269  MDEntryType",
        ("FIX.4.2", "270"): "270  MDEntryPx",
        ("FIX.4.2", "271"): "271  MDEntrySize",
        ("FIX.4.2", "272"): "272  MDEntryDate",
        ("FIX.4.2", "273"): "273  MDEntryTime",
        ("FIX.4.2", "274"): "274  TickDirection",
        ("FIX.4.2", "275"): "275  MDMkt",
        ("FIX.4.2", "276"): "276  QuoteCondition",
        ("FIX.4.2", "277"): "277  TradeCondition",
        ("FIX.4.2", "278"): "278  MDEntryID",
        ("FIX.4.2", "279"): "279  MDUpdateAction",
        ("FIX.4.2", "280"): "280  MDEntryRefID",
        ("FIX.4.2", "281"): "281  MDReqRejReason",
        ("FIX.4.2", "282"): "282  MDEntryOriginator",
        ("FIX.4.2", "283"): "283  LocationID",
        ("FIX.4.2", "284"): "284  DeskID",
        ("FIX.4.2", "285"): "285  DeleteReason",
        ("FIX.4.2", "286"): "286  OpenCloseSettleFlag",
        ("FIX.4.2", "287"): "287  SellerDays",
        ("FIX.4.2", "288"): "288  MDEntryBuyer",
        ("FIX.4.2", "289"): "289  MDEntrySeller",
        ("FIX.4.2", "290"): "290  MDEntryPositionNo",
        ("FIX.4.2", "291"): "291  FinancialStatus",
        ("FIX.4.2", "292"): "292  CorporateAction",
        ("FIX.4.2", "293"): "293  DefBidSize",
        ("FIX.4.2", "294"): "294  DefOfferSize",
        ("FIX.4.2", "295"): "295  NoQuoteEntries",
        ("FIX.4.2", "296"): "296  NoQuoteSets",
        ("FIX.4.2", "297"): "297  QuoteAckStatus",
        ("FIX.4.2", "298"): "298  QuoteCancelType",
        ("FIX.4.2", "299"): "299  QuoteEntryID",
        ("FIX.4.2", "300"): "300  QuoteRejectReason",
        ("FIX.4.2", "301"): "301  QuoteResponseLevel",
        ("FIX.4.2", "302"): "302  QuoteSetID",
        ("FIX.4.2", "303"): "303  QuoteRequestType",
        ("FIX.4.2", "304"): "304  TotQuoteEntries",
        ("FIX.4.2", "305"): "305  UnderlyingIDSource",
        ("FIX.4.2", "306"): "306  UnderlyingIssuer",
        ("FIX.4.2", "307"): "307  UnderlyingSecurityDesc",
        ("FIX.4.2", "308"): "308  UnderlyingSecurityExchange",
        ("FIX.4.2", "309"): "309  UnderlyingSecurityID",
        ("FIX.4.2", "310"): "310  UnderlyingSecurityType",
        ("FIX.4.2", "311"): "311  UnderlyingSymbol",
        ("FIX.4.2", "312"): "312  UnderlyingSymbolSfx",
        ("FIX.4.2", "313"): "313  UnderlyingMaturityMonthYear",
        ("FIX.4.2", "314"): "314  UnderlyingMaturityDay",
        ("FIX.4.2", "315"): "315  UnderlyingPutOrCall",
        ("FIX.4.2", "316"): "316  UnderlyingStrikePrice",
        ("FIX.4.2", "317"): "317  UnderlyingOptAttribute",
        ("FIX.4.2", "318"): "318  UnderlyingCurrency",
        ("FIX.4.2", "319"): "319  RatioQty",
        ("FIX.4.2", "320"): "320  SecurityReqID",
        ("FIX.4.2", "321"): "321  SecurityRequestType",
        ("FIX.4.2", "322"): "322  SecurityResponseID",
        ("FIX.4.2", "323"): "323  SecurityResponseType",
        ("FIX.4.2", "324"): "324  SecurityStatusReqID",
        ("FIX.4.2", "325"): "325  UnsolicitedIndicator",
        ("FIX.4.2", "326"): "326  SecurityTradingStatus",
        ("FIX.4.2", "327"): "327  HaltReason",
        ("FIX.4.2", "328"): "328  InViewOfCommon",
        ("FIX.4.2", "329"): "329  DueToRelated",
        ("FIX.4.2", "330"): "330  BuyVolume",
        ("FIX.4.2", "331"): "331  SellVolume",
        ("FIX.4.2", "332"): "332  HighPx",
        ("FIX.4.2", "333"): "333  LowPx",
        ("FIX.4.2", "334"): "334  Adjustment",
        ("FIX.4.2", "335"): "335  TradSesReqID",
        ("FIX.4.2", "336"): "336  TradingSessionID",
        ("FIX.4.2", "337"): "337  ContraTrader",
        ("FIX.4.2", "338"): "338  TradSesMethod",
        ("FIX.4.2", "339"): "339  TradSesMode",
        ("FIX.4.2", "340"): "340  TradSesStatus",
        ("FIX.4.2", "341"): "341  TradSesStartTime",
        ("FIX.4.2", "342"): "342  TradSesOpenTime",
        ("FIX.4.2", "343"): "343  TradSesPreCloseTime",
        ("FIX.4.2", "344"): "344  TradSesCloseTime",
        ("FIX.4.2", "345"): "345  TradSesEndTime",
        ("FIX.4.2", "346"): "346  NumberOfOrders",
        ("FIX.4.2", "347"): "347  MessageEncoding",
        ("FIX.4.2", "348"): "348  EncodedIssuerLen",
        ("FIX.4.2", "349"): "349  EncodedIssuer",
        ("FIX.4.2", "350"): "350  EncodedSecurityDescLen",
        ("FIX.4.2", "351"): "351  EncodedSecurityDesc",
        ("FIX.4.2", "352"): "352  EncodedListExecInstLen",
        ("FIX.4.2", "353"): "353  EncodedListExecInst",
        ("FIX.4.2", "354"): "354  EncodedTextLen",
        ("FIX.4.2", "355"): "355  EncodedText",
        ("FIX.4.2", "356"): "356  EncodedSubjectLen",
        ("FIX.4.2", "357"): "357  EncodedSubject",
        ("FIX.4.2", "358"): "358  EncodedHeadlineLen",
        ("FIX.4.2", "359"): "359  EncodedHeadline",
        ("FIX.4.2", "360"): "360  EncodedAllocTextLen",
        ("FIX.4.2", "361"): "361  EncodedAllocText",
        ("FIX.4.2", "362"): "362  EncodedUnderlyingIssuerLen",
        ("FIX.4.2", "363"): "363  EncodedUnderlyingIssuer",
        ("FIX.4.2", "364"): "364  EncodedUnderlyingSecurityDescLen",
        ("FIX.4.2", "365"): "365  EncodedUnderlyingSecurityDesc",
        ("FIX.4.2", "366"): "366  AllocPrice",
        ("FIX.4.2", "367"): "367  QuoteSetValidUntilTime",
        ("FIX.4.2", "368"): "368  QuoteEntryRejectReason",
        ("FIX.4.2", "369"): "369  LastMsgSeqNumProcessed",
        ("FIX.4.2", "370"): "370  OnBehalfOfSendingTime",
        ("FIX.4.2", "371"): "371  RefTagID",
        ("FIX.4.2", "372"): "372  RefMsgType",
        ("FIX.4.2", "373"): "373  SessionRejectReason",
        ("FIX.4.2", "374"): "374  BidRequestTransType",
        ("FIX.4.2", "375"): "375  ContraBroker",
        ("FIX.4.2", "376"): "376  ComplianceID",
        ("FIX.4.2", "377"): "377  SolicitedFlag",
        ("FIX.4.2", "378"): "378  ExecRestatementReason",
        ("FIX.4.2", "379"): "379  BusinessRejectRefID",
        ("FIX.4.2", "380"): "380  BusinessRejectReason",
        ("FIX.4.2", "381"): "381  GrossTradeAmt",
        ("FIX.4.2", "382"): "382  NoContraBrokers",
        ("FIX.4.2", "383"): "383  MaxMessageSize",
        ("FIX.4.2", "384"): "384  NoMsgTypes",
        ("FIX.4.2", "385"): "385  MsgDirection",
        ("FIX.4.2", "386"): "386  NoTradingSessions",
        ("FIX.4.2", "387"): "387  TotalVolumeTraded",
        ("FIX.4.2", "388"): "388  DiscretionInst",
        ("FIX.4.2", "389"): "389  DiscretionOffset",
        ("FIX.4.2", "390"): "390  BidID",
        ("FIX.4.2", "391"): "391  ClientBidID",
        ("FIX.4.2", "392"): "392  ListName",
        ("FIX.4.2", "393"): "393  TotalNumSecurities",
        ("FIX.4.2", "394"): "394  BidType",
        ("FIX.4.2", "395"): "395  NumTickets",
        ("FIX.4.2", "396"): "396  SideValue1",
        ("FIX.4.2", "397"): "397  SideValue2",
        ("FIX.4.2", "398"): "398  NoBidDescriptors",
        ("FIX.4.2", "399"): "399  BidDescriptorType",
        ("FIX.4.2", "400"): "400  BidDescriptor",
        ("FIX.4.2", "401"): "401  SideValueInd",
        ("FIX.4.2", "402"): "402  LiquidityPctLow",
        ("FIX.4.2", "403"): "403  LiquidityPctHigh",
        ("FIX.4.2", "404"): "404  LiquidityValue",
        ("FIX.4.2", "405"): "405  EFPTrackingError",
        ("FIX.4.2", "406"): "406  FairValue",
        ("FIX.4.2", "407"): "407  OutsideIndexPct",
        ("FIX.4.2", "408"): "408  ValueOfFutures",
        ("FIX.4.2", "409"): "409  LiquidityIndType",
        ("FIX.4.2", "410"): "410  WtAverageLiquidity",
        ("FIX.4.2", "411"): "411  ExchangeForPhysical",
        ("FIX.4.2", "412"): "412  OutMainCntryUIndex",
        ("FIX.4.2", "413"): "413  CrossPercent",
        ("FIX.4.2", "414"): "414  ProgRptReqs",
        ("FIX.4.2", "415"): "415  ProgPeriodInterval",
        ("FIX.4.2", "416"): "416  IncTaxInd",
        ("FIX.4.2", "417"): "417  NumBidders",
        ("FIX.4.2", "418"): "418  TradeType",
        ("FIX.4.2", "419"): "419  BasisPxType",
        ("FIX.4.2", "420"): "420  NoBidComponents",
        ("FIX.4.2", "421"): "421  Country",
        ("FIX.4.2", "422"): "422  TotNoStrikes",
        ("FIX.4.2", "423"): "423  PriceType",
        ("FIX.4.2", "424"): "424  DayOrderQty",
        ("FIX.4.2", "425"): "425  DayCumQty",
        ("FIX.4.2", "426"): "426  DayAvgPx",
        ("FIX.4.2", "427"): "427  GTBookingInst",
        ("FIX.4.2", "428"): "428  NoStrikes",
        ("FIX.4.2", "429"): "429  ListStatusType",
        ("FIX.4.2", "430"): "430  NetGrossInd",
        ("FIX.4.2", "431"): "431  ListOrderStatus",
        ("FIX.4.2", "432"): "432  ExpireDate",
        ("FIX.4.2", "433"): "433  ListExecInstType",
        ("FIX.4.2", "434"): "434  CxlRejResponseTo",
        ("FIX.4.2", "435"): "435  UnderlyingCouponRate",
        ("FIX.4.2", "436"): "436  UnderlyingContractMultiplier",
        ("FIX.4.2", "437"): "437  ContraTradeQty",
        ("FIX.4.2", "438"): "438  ContraTradeTime",
        ("FIX.4.2", "439"): "439  ClearingFirm",
        ("FIX.4.2", "440"): "440  ClearingAccount",
        ("FIX.4.2", "441"): "441  LiquidityNumSecurities",
        ("FIX.4.2", "442"): "442  MultiLegReportingType",
        ("FIX.4.2", "443"): "443  StrikeTime",
        ("FIX.4.2", "444"): "444  ListStatusText",
        ("FIX.4.2", "445"): "445  EncodedListStatusTextLen",
        ("FIX.4.2", "446"): "446  EncodedListStatusText",
        ("FIX.4.4", "1"): "1    Account",
        ("FIX.4.4", "2"): "2    AdvId",
        ("FIX.4.4", "3"): "3    AdvRefID",
        ("FIX.4.4", "4"): "4    AdvSide",
        ("FIX.4.4", "5"): "5    AdvTransType",
        ("FIX.4.4", "6"): "6    AvgPx",
        ("FIX.4.4", "7"): "7    BeginSeqNo",
        ("FIX.4.4", "8"): "8    BeginString",
        ("FIX.4.4", "9"): "9    BodyLength",
        ("FIX.4.4", "10"): "10   CheckSum",
        ("FIX.4.4", "11"): "11   ClOrdID",
        ("FIX.4.4", "12"): "12   Commission",
        ("FIX.4.4", "13"): "13   CommType",
        ("FIX.4.4", "14"): "14   CumQty",
        ("FIX.4.4", "15"): "15   Currency",
        ("FIX.4.4", "16"): "16   EndSeqNo",
        ("FIX.4.4", "17"): "17   ExecID",
        ("FIX.4.4", "18"): "18   ExecInst",
        ("FIX.4.4", "19"): "19   ExecRefID",
        ("FIX.4.4", "21"): "21   HandlInst",
        ("FIX.4.4", "22"): "22   SecurityIDSource",
        ("FIX.4.4", "23"): "23   IOIID",
        ("FIX.4.4", "25"): "25   IOIQltyInd",
        ("FIX.4.4", "26"): "26   IOIRefID",
        ("FIX.4.4", "27"): "27   IOIQty",
        ("FIX.4.4", "28"): "28   IOITransType",
        ("FIX.4.4", "29"): "29   LastCapacity",
        ("FIX.4.4", "30"): "30   LastMkt",
        ("FIX.4.4", "31"): "31   LastPx",
        ("FIX.4.4", "32"): "32   LastQty",
        ("FIX.4.4", "33"): "33   NoLinesOfText",
        ("FIX.4.4", "34"): "34   MsgSeqNum",
        ("FIX.4.4", "35"): "35   MsgType",
        ("FIX.4.4", "36"): "36   NewSeqNo",
        ("FIX.4.4", "37"): "37   OrderID",
        ("FIX.4.4", "38"): "38   OrderQty",
        ("FIX.4.4", "39"): "39   OrdStatus",
        ("FIX.4.4", "40"): "40   OrdType",
        ("FIX.4.4", "41"): "41   OrigClOrdID",
        ("FIX.4.4", "42"): "42   OrigTime",
        ("FIX.4.4", "43"): "43   PossDupFlag",
        ("FIX.4.4", "44"): "44   Price",
        ("FIX.4.4", "45"): "45   RefSeqNum",
        ("FIX.4.4", "48"): "48   SecurityID",
        ("FIX.4.4", "49"): "49   SenderCompID",
        ("FIX.4.4", "50"): "50   SenderSubID",
        ("FIX.4.4", "52"): "52   SendingTime",
        ("FIX.4.4", "53"): "53   Quantity",
        ("FIX.4.4", "54"): "54   Side",
        ("FIX.4.4", "55"): "55   Symbol",
        ("FIX.4.4", "56"): "56   TargetCompID",
        ("FIX.4.4", "57"): "57   TargetSubID",
        ("FIX.4.4", "58"): "58   Text",
        ("FIX.4.4", "59"): "59   TimeInForce",
        ("FIX.4.4", "60"): "60   TransactTime",
        ("FIX.4.4", "61"): "61   Urgency",
        ("FIX.4.4", "62"): "62   ValidUntilTime",
        ("FIX.4.4", "63"): "63   SettlType",
        ("FIX.4.4", "64"): "64   SettlDate",
        ("FIX.4.4", "65"): "65   SymbolSfx",
        ("FIX.4.4", "66"): "66   ListID",
        ("FIX.4.4", "67"): "67   ListSeqNo",
        ("FIX.4.4", "68"): "68   TotNoOrders",
        ("FIX.4.4", "69"): "69   ListExecInst",
        ("FIX.4.4", "70"): "70   AllocID",
        ("FIX.4.4", "71"): "71   AllocTransType",
        ("FIX.4.4", "72"): "72   RefAllocID",
        ("FIX.4.4", "73"): "73   NoOrders",
        ("FIX.4.4", "74"): "74   AvgPxPrecision",
        ("FIX.4.4", "75"): "75   TradeDate",
        ("FIX.4.4", "77"): "77   PositionEffect",
        ("FIX.4.4", "78"): "78   NoAllocs",
        ("FIX.4.4", "79"): "79   AllocAccount",
        ("FIX.4.4", "80"): "80   AllocQty",
        ("FIX.4.4", "81"): "81   ProcessCode",
        ("FIX.4.4", "82"): "82   NoRpts",
        ("FIX.4.4", "83"): "83   RptSeq",
        ("FIX.4.4", "84"): "84   CxlQty",
        ("FIX.4.4", "85"): "85   NoDlvyInst",
        ("FIX.4.4", "87"): "87   AllocStatus",
        ("FIX.4.4", "88"): "88   AllocRejCode",
        ("FIX.4.4", "89"): "89   Signature",
        ("FIX.4.4", "90"): "90   SecureDataLen",
        ("FIX.4.4", "91"): "91   SecureData",
        ("FIX.4.4", "93"): "93   SignatureLength",
        ("FIX.4.4", "94"): "94   EmailType",
        ("FIX.4.4", "95"): "95   RawDataLength",
        ("FIX.4.4", "96"): "96   RawData",
        ("FIX.4.4", "97"): "97   PossResend",
        ("FIX.4.4", "98"): "98   EncryptMethod",
        ("FIX.4.4", "99"): "99   StopPx",
        ("FIX.4.4", "100"): "100  ExDestination",
        ("FIX.4.4", "102"): "102  CxlRejReason",
        ("FIX.4.4", "103"): "103  OrdRejReason",
        ("FIX.4.4", "104"): "104  IOIQualifier",
        ("FIX.4.4", "106"): "106  Issuer",
        ("FIX.4.4", "107"): "107  SecurityDesc",
        ("FIX.4.4", "108"): "108  HeartBtInt",
        ("FIX.4.4", "110"): "110  MinQty",
        ("FIX.4.4", "111"): "111  MaxFloor",
        ("FIX.4.4", "112"): "112  TestReqID",
        ("FIX.4.4", "113"): "113  ReportToExch",
        ("FIX.4.4", "114"): "114  LocateReqd",
        ("FIX.4.4", "115"): "115  OnBehalfOfCompID",
        ("FIX.4.4", "116"): "116  OnBehalfOfSubID",
        ("FIX.4.4", "117"): "117  QuoteID",
        ("FIX.4.4", "118"): "118  NetMoney",
        ("FIX.4.4", "119"): "119  SettlCurrAmt",
        ("FIX.4.4", "120"): "120  SettlCurrency",
        ("FIX.4.4", "121"): "121  ForexReq",
        ("FIX.4.4", "122"): "122  OrigSendingTime",
        ("FIX.4.4", "123"): "123  GapFillFlag",
        ("FIX.4.4", "124"): "124  NoExecs",
        ("FIX.4.4", "126"): "126  ExpireTime",
        ("FIX.4.4", "127"): "127  DKReason",
        ("FIX.4.4", "128"): "128  DeliverToCompID",
        ("FIX.4.4", "129"): "129  DeliverToSubID",
        ("FIX.4.4", "130"): "130  IOINaturalFlag",
        ("FIX.4.4", "131"): "131  QuoteReqID",
        ("FIX.4.4", "132"): "132  BidPx",
        ("FIX.4.4", "133"): "133  OfferPx",
        ("FIX.4.4", "134"): "134  BidSize",
        ("FIX.4.4", "135"): "135  OfferSize",
        ("FIX.4.4", "136"): "136  NoMiscFees",
        ("FIX.4.4", "137"): "137  MiscFeeAmt",
        ("FIX.4.4", "138"): "138  MiscFeeCurr",
        ("FIX.4.4", "139"): "139  MiscFeeType",
        ("FIX.4.4", "140"): "140  PrevClosePx",
        ("FIX.4.4", "141"): "141  ResetSeqNumFlag",
        ("FIX.4.4", "142"): "142  SenderLocationID",
        ("FIX.4.4", "143"): "143  TargetLocationID",
        ("FIX.4.4", "144"): "144  OnBehalfOfLocationID",
        ("FIX.4.4", "145"): "145  DeliverToLocationID",
        ("FIX.4.4", "146"): "146  NoRelatedSym",
        ("FIX.4.4", "147"): "147  Subject",
        ("FIX.4.4", "148"): "148  Headline",
        ("FIX.4.4", "149"): "149  URLLink",
        ("FIX.4.4", "150"): "150  ExecType",
        ("FIX.4.4", "151"): "151  LeavesQty",
        ("FIX.4.4", "152"): "152  CashOrderQty",
        ("FIX.4.4", "153"): "153  AllocAvgPx",
        ("FIX.4.4", "154"): "154  AllocNetMoney",
        ("FIX.4.4", "155"): "155  SettlCurrFxRate",
        ("FIX.4.4", "156"): "156  SettlCurrFxRateCalc",
        ("FIX.4.4", "157"): "157  NumDaysInterest",
        ("FIX.4.4", "158"): "158  AccruedInterestRate",
        ("FIX.4.4", "159"): "159  AccruedInterestAmt",
        ("FIX.4.4", "160"): "160  SettlInstMode",
        ("FIX.4.4", "161"): "161  AllocText",
        ("FIX.4.4", "162"): "162  SettlInstID",
        ("FIX.4.4", "163"): "163  SettlInstTransType",
        ("FIX.4.4", "164"): "164  EmailThreadID",
        ("FIX.4.4", "165"): "165  SettlInstSource",
        ("FIX.4.4", "167"): "167  SecurityType",
        ("FIX.4.4", "168"): "168  EffectiveTime",
        ("FIX.4.4", "169"): "169  StandInstDbType",
        ("FIX.4.4", "170"): "170  StandInstDbName",
        ("FIX.4.4", "171"): "171  StandInstDbID",
        ("FIX.4.4", "172"): "172  SettlDeliveryType",
        ("FIX.4.4", "188"): "188  BidSpotRate",
        ("FIX.4.4", "189"): "189  BidForwardPoints",
        ("FIX.4.4", "190"): "190  OfferSpotRate",
        ("FIX.4.4", "191"): "191  OfferForwardPoints",
        ("FIX.4.4", "192"): "192  OrderQty2",
        ("FIX.4.4", "193"): "193  SettlDate2",
        ("FIX.4.4", "194"): "194  LastSpotRate",
        ("FIX.4.4", "195"): "195  LastForwardPoints",
        ("FIX.4.4", "196"): "196  AllocLinkID",
        ("FIX.4.4", "197"): "197  AllocLinkType",
        ("FIX.4.4", "198"): "198  SecondaryOrderID",
        ("FIX.4.4", "199"): "199  NoIOIQualifiers",
        ("FIX.4.4", "200"): "200  MaturityMonthYear",
        ("FIX.4.4", "201"): "201  PutOrCall",
        ("FIX.4.4", "202"): "202  StrikePrice",
        ("FIX.4.4", "203"): "203  CoveredOrUncovered",
        ("FIX.4.4", "206"): "206  OptAttribute",
        ("FIX.4.4", "207"): "207  SecurityExchange",
        ("FIX.4.4", "208"): "208  NotifyBrokerOfCredit",
        ("FIX.4.4", "209"): "209  AllocHandlInst",
        ("FIX.4.4", "210"): "210  MaxShow",
        ("FIX.4.4", "211"): "211  PegOffsetValue",
        ("FIX.4.4", "212"): "212  XmlDataLen",
        ("FIX.4.4", "213"): "213  XmlData",
        ("FIX.4.4", "214"): "214  SettlInstRefID",
        ("FIX.4.4", "215"): "215  NoRoutingIDs",
        ("FIX.4.4", "216"): "216  RoutingType",
        ("FIX.4.4", "217"): "217  RoutingID",
        ("FIX.4.4", "218"): "218  Spread",
        ("FIX.4.4", "220"): "220  BenchmarkCurveCurrency",
        ("FIX.4.4", "221"): "221  BenchmarkCurveName",
        ("FIX.4.4", "222"): "222  BenchmarkCurvePoint",
        ("FIX.4.4", "223"): "223  CouponRate",
        ("FIX.4.4", "224"): "224  CouponPaymentDate",
        ("FIX.4.4", "225"): "225  IssueDate",
        ("FIX.4.4", "226"): "226  RepurchaseTerm",
        ("FIX.4.4", "227"): "227  RepurchaseRate",
        ("FIX.4.4", "228"): "228  Factor",
        ("FIX.4.4", "229"): "229  TradeOriginationDate",
        ("FIX.4.4", "230"): "230  ExDate",
        ("FIX.4.4", "231"): "231  ContractMultiplier",
        ("FIX.4.4", "232"): "232  NoStipulations",
        ("FIX.4.4", "233"): "233  StipulationType",
        ("FIX.4.4", "234"): "234  StipulationValue",
        ("FIX.4.4", "235"): "235  YieldType",
        ("FIX.4.4", "236"): "236  Yield",
        ("FIX.4.4", "237"): "237  TotalTakedown",
        ("FIX.4.4", "238"): "238  Concession",
        ("FIX.4.4", "239"): "239  RepoCollateralSecurityType",
        ("FIX.4.4", "240"): "240  RedemptionDate",
        ("FIX.4.4", "241"): "241  UnderlyingCouponPaymentDate",
        ("FIX.4.4", "242"): "242  UnderlyingIssueDate",
        ("FIX.4.4", "243"): "243  UnderlyingRepoCollateralSecurityType",
        ("FIX.4.4", "244"): "244  UnderlyingRepurchaseTerm",
        ("FIX.4.4", "245"): "245  UnderlyingRepurchaseRate",
        ("FIX.4.4", "246"): "246  UnderlyingFactor",
        ("FIX.4.4", "247"): "247  UnderlyingRedemptionDate",
        ("FIX.4.4", "248"): "248  LegCouponPaymentDate",
        ("FIX.4.4", "249"): "249  LegIssueDate",
        ("FIX.4.4", "250"): "250  LegRepoCollateralSecurityType",
        ("FIX.4.4", "251"): "251  LegRepurchaseTerm",
        ("FIX.4.4", "252"): "252  LegRepurchaseRate",
        ("FIX.4.4", "253"): "253  LegFactor",
        ("FIX.4.4", "254"): "254  LegRedemptionDate",
        ("FIX.4.4", "255"): "255  CreditRating",
        ("FIX.4.4", "256"): "256  UnderlyingCreditRating",
        ("FIX.4.4", "257"): "257  LegCreditRating",
        ("FIX.4.4", "258"): "258  TradedFlatSwitch",
        ("FIX.4.4", "259"): "259  BasisFeatureDate",
        ("FIX.4.4", "260"): "260  BasisFeaturePrice",
        ("FIX.4.4", "262"): "262  MDReqID",
        ("FIX.4.4", "263"): "263  SubscriptionRequestType",
        ("FIX.4.4", "264"): "264  MarketDepth",
        ("FIX.4.4", "265"): "265  MDUpdateType",
        ("FIX.4.4", "266"): "266  AggregatedBook",
        ("FIX.4.4", "267"): "267  NoMDEntryTypes",
        ("FIX.4.4", "268"): "268  NoMDEntries",
        ("FIX.4.4", "269"): "269  MDEntryType",
        ("FIX.4.4", "270"): "270  MDEntryPx",
        ("FIX.4.4", "271"): "271  MDEntrySize",
        ("FIX.4.4", "272"): "272  MDEntryDate",
        ("FIX.4.4", "273"): "273  MDEntryTime",
        ("FIX.4.4", "274"): "274  TickDirection",
        ("FIX.4.4", "275"): "275  MDMkt",
        ("FIX.4.4", "276"): "276  QuoteCondition",
        ("FIX.4.4", "277"): "277  TradeCondition",
        ("FIX.4.4", "278"): "278  MDEntryID",
        ("FIX.4.4", "279"): "279  MDUpdateAction",
        ("FIX.4.4", "280"): "280  MDEntryRefID",
        ("FIX.4.4", "281"): "281  MDReqRejReason",
        ("FIX.4.4", "282"): "282  MDEntryOriginator",
        ("FIX.4.4", "283"): "283  LocationID",
        ("FIX.4.4", "284"): "284  DeskID",
        ("FIX.4.4", "285"): "285  DeleteReason",
        ("FIX.4.4", "286"): "286  OpenCloseSettlFlag",
        ("FIX.4.4", "287"): "287  SellerDays",
        ("FIX.4.4", "288"): "288  MDEntryBuyer",
        ("FIX.4.4", "289"): "289  MDEntrySeller",
        ("FIX.4.4", "290"): "290  MDEntryPositionNo",
        ("FIX.4.4", "291"): "291  FinancialStatus",
        ("FIX.4.4", "292"): "292  CorporateAction",
        ("FIX.4.4", "293"): "293  DefBidSize",
        ("FIX.4.4", "294"): "294  DefOfferSize",
        ("FIX.4.4", "295"): "295  NoQuoteEntries",
        ("FIX.4.4", "296"): "296  NoQuoteSets",
        ("FIX.4.4", "297"): "297  QuoteStatus",
        ("FIX.4.4", "298"): "298  QuoteCancelType",
        ("FIX.4.4", "299"): "299  QuoteEntryID",
        ("FIX.4.4", "300"): "300  QuoteRejectReason",
        ("FIX.4.4", "301"): "301  QuoteResponseLevel",
        ("FIX.4.4", "302"): "302  QuoteSetID",
        ("FIX.4.4", "303"): "303  QuoteRequestType",
        ("FIX.4.4", "304"): "304  TotNoQuoteEntries",
        ("FIX.4.4", "305"): "305  UnderlyingSecurityIDSource",
        ("FIX.4.4", "306"): "306  UnderlyingIssuer",
        ("FIX.4.4", "307"): "307  UnderlyingSecurityDesc",
        ("FIX.4.4", "308"): "308  UnderlyingSecurityExchange",
        ("FIX.4.4", "309"): "309  UnderlyingSecurityID",
        ("FIX.4.4", "310"): "310  UnderlyingSecurityType",
        ("FIX.4.4", "311"): "311  UnderlyingSymbol",
        ("FIX.4.4", "312"): "312  UnderlyingSymbolSfx",
        ("FIX.4.4", "313"): "313  UnderlyingMaturityMonthYear",
        ("FIX.4.4", "315"): "315  UnderlyingPutOrCall",
        ("FIX.4.4", "316"): "316  UnderlyingStrikePrice",
        ("FIX.4.4", "317"): "317  UnderlyingOptAttribute",
        ("FIX.4.4", "318"): "318  UnderlyingCurrency",
        ("FIX.4.4", "320"): "320  SecurityReqID",
        ("FIX.4.4", "321"): "321  SecurityRequestType",
        ("FIX.4.4", "322"): "322  SecurityResponseID",
        ("FIX.4.4", "323"): "323  SecurityResponseType",
        ("FIX.4.4", "324"): "324  SecurityStatusReqID",
        ("FIX.4.4", "325"): "325  UnsolicitedIndicator",
        ("FIX.4.4", "326"): "326  SecurityTradingStatus",
        ("FIX.4.4", "327"): "327  HaltReason",
        ("FIX.4.4", "328"): "328  InViewOfCommon",
        ("FIX.4.4", "329"): "329  DueToRelated",
        ("FIX.4.4", "330"): "330  BuyVolume",
        ("FIX.4.4", "331"): "331  SellVolume",
        ("FIX.4.4", "332"): "332  HighPx",
        ("FIX.4.4", "333"): "333  LowPx",
        ("FIX.4.4", "334"): "334  Adjustment",
        ("FIX.4.4", "335"): "335  TradSesReqID",
        ("FIX.4.4", "336"): "336  TradingSessionID",
        ("FIX.4.4", "337"): "337  ContraTrader",
        ("FIX.4.4", "338"): "338  TradSesMethod",
        ("FIX.4.4", "339"): "339  TradSesMode",
        ("FIX.4.4", "340"): "340  TradSesStatus",
        ("FIX.4.4", "341"): "341  TradSesStartTime",
        ("FIX.4.4", "342"): "342  TradSesOpenTime",
        ("FIX.4.4", "343"): "343  TradSesPreCloseTime",
        ("FIX.4.4", "344"): "344  TradSesCloseTime",
        ("FIX.4.4", "345"): "345  TradSesEndTime",
        ("FIX.4.4", "346"): "346  NumberOfOrders",
        ("FIX.4.4", "347"): "347  MessageEncoding",
        ("FIX.4.4", "348"): "348  EncodedIssuerLen",
        ("FIX.4.4", "349"): "349  EncodedIssuer",
        ("FIX.4.4", "350"): "350  EncodedSecurityDescLen",
        ("FIX.4.4", "351"): "351  EncodedSecurityDesc",
        ("FIX.4.4", "352"): "352  EncodedListExecInstLen",
        ("FIX.4.4", "353"): "353  EncodedListExecInst",
        ("FIX.4.4", "354"): "354  EncodedTextLen",
        ("FIX.4.4", "355"): "355  EncodedText",
        ("FIX.4.4", "356"): "356  EncodedSubjectLen",
        ("FIX.4.4", "357"): "357  EncodedSubject",
        ("FIX.4.4", "358"): "358  EncodedHeadlineLen",
        ("FIX.4.4", "359"): "359  EncodedHeadline",
        ("FIX.4.4", "360"): "360  EncodedAllocTextLen",
        ("FIX.4.4", "361"): "361  EncodedAllocText",
        ("FIX.4.4", "362"): "362  EncodedUnderlyingIssuerLen",
        ("FIX.4.4", "363"): "363  EncodedUnderlyingIssuer",
        ("FIX.4.4", "364"): "364  EncodedUnderlyingSecurityDescLen",
        ("FIX.4.4", "365"): "365  EncodedUnderlyingSecurityDesc",
        ("FIX.4.4", "366"): "366  AllocPrice",
        ("FIX.4.4", "367"): "367  QuoteSetValidUntilTime",
        ("FIX.4.4", "368"): "368  QuoteEntryRejectReason",
        ("FIX.4.4", "369"): "369  LastMsgSeqNumProcessed",
        ("FIX.4.4", "371"): "371  RefTagID",
        ("FIX.4.4", "372"): "372  RefMsgType",
        ("FIX.4.4", "373"): "373  SessionRejectReason",
        ("FIX.4.4", "374"): "374  BidRequestTransType",
        ("FIX.4.4", "375"): "375  ContraBroker",
        ("FIX.4.4", "376"): "376  ComplianceID",
        ("FIX.4.4", "377"): "377  SolicitedFlag",
        ("FIX.4.4", "378"): "378  ExecRestatementReason",
        ("FIX.4.4", "379"): "379  BusinessRejectRefID",
        ("FIX.4.4", "380"): "380  BusinessRejectReason",
        ("FIX.4.4", "381"): "381  GrossTradeAmt",
        ("FIX.4.4", "382"): "382  NoContraBrokers",
        ("FIX.4.4", "383"): "383  MaxMessageSize",
        ("FIX.4.4", "384"): "384  NoMsgTypes",
        ("FIX.4.4", "385"): "385  MsgDirection",
        ("FIX.4.4", "386"): "386  NoTradingSessions",
        ("FIX.4.4", "387"): "387  TotalVolumeTraded",
        ("FIX.4.4", "388"): "388  DiscretionInst",
        ("FIX.4.4", "389"): "389  DiscretionOffsetValue",
        ("FIX.4.4", "390"): "390  BidID",
        ("FIX.4.4", "391"): "391  ClientBidID",
        ("FIX.4.4", "392"): "392  ListName",
        ("FIX.4.4", "393"): "393  TotNoRelatedSym",
        ("FIX.4.4", "394"): "394  BidType",
        ("FIX.4.4", "395"): "395  NumTickets",
        ("FIX.4.4", "396"): "396  SideValue1",
        ("FIX.4.4", "397"): "397  SideValue2",
        ("FIX.4.4", "398"): "398  NoBidDescriptors",
        ("FIX.4.4", "399"): "399  BidDescriptorType",
        ("FIX.4.4", "400"): "400  BidDescriptor",
        ("FIX.4.4", "401"): "401  SideValueInd",
        ("FIX.4.4", "402"): "402  LiquidityPctLow",
        ("FIX.4.4", "403"): "403  LiquidityPctHigh",
        ("FIX.4.4", "404"): "404  LiquidityValue",
        ("FIX.4.4", "405"): "405  EFPTrackingError",
        ("FIX.4.4", "406"): "406  FairValue",
        ("FIX.4.4", "407"): "407  OutsideIndexPct",
        ("FIX.4.4", "408"): "408  ValueOfFutures",
        ("FIX.4.4", "409"): "409  LiquidityIndType",
        ("FIX.4.4", "410"): "410  WtAverageLiquidity",
        ("FIX.4.4", "411"): "411  ExchangeForPhysical",
        ("FIX.4.4", "412"): "412  OutMainCntryUIndex",
        ("FIX.4.4", "413"): "413  CrossPercent",
        ("FIX.4.4", "414"): "414  ProgRptReqs",
        ("FIX.4.4", "415"): "415  ProgPeriodInterval",
        ("FIX.4.4", "416"): "416  IncTaxInd",
        ("FIX.4.4", "417"): "417  NumBidders",
        ("FIX.4.4", "418"): "418  BidTradeType",
        ("FIX.4.4", "419"): "419  BasisPxType",
        ("FIX.4.4", "420"): "420  NoBidComponents",
        ("FIX.4.4", "421"): "421  Country",
        ("FIX.4.4", "422"): "422  TotNoStrikes",
        ("FIX.4.4", "423"): "423  PriceType",
        ("FIX.4.4", "424"): "424  DayOrderQty",
        ("FIX.4.4", "425"): "425  DayCumQty",
        ("FIX.4.4", "426"): "426  DayAvgPx",
        ("FIX.4.4", "427"): "427  GTBookingInst",
        ("FIX.4.4", "428"): "428  NoStrikes",
        ("FIX.4.4", "429"): "429  ListStatusType",
        ("FIX.4.4", "430"): "430  NetGrossInd",
        ("FIX.4.4", "431"): "431  ListOrderStatus",
        ("FIX.4.4", "432"): "432  ExpireDate",
        ("FIX.4.4", "433"): "433  ListExecInstType",
        ("FIX.4.4", "434"): "434  CxlRejResponseTo",
        ("FIX.4.4", "435"): "435  UnderlyingCouponRate",
        ("FIX.4.4", "436"): "436  UnderlyingContractMultiplier",
        ("FIX.4.4", "437"): "437  ContraTradeQty",
        ("FIX.4.4", "438"): "438  ContraTradeTime",
        ("FIX.4.4", "441"): "441  LiquidityNumSecurities",
        ("FIX.4.4", "442"): "442  MultiLegReportingType",
        ("FIX.4.4", "443"): "443  StrikeTime",
        ("FIX.4.4", "444"): "444  ListStatusText",
        ("FIX.4.4", "445"): "445  EncodedListStatusTextLen",
        ("FIX.4.4", "446"): "446  EncodedListStatusText",
        ("FIX.4.4", "447"): "447  PartyIDSource",
        ("FIX.4.4", "448"): "448  PartyID",
        ("FIX.4.4", "451"): "451  NetChgPrevDay",
        ("FIX.4.4", "452"): "452  PartyRole",
        ("FIX.4.4", "453"): "453  NoPartyIDs",
        ("FIX.4.4", "454"): "454  NoSecurityAltID",
        ("FIX.4.4", "455"): "455  SecurityAltID",
        ("FIX.4.4", "456"): "456  SecurityAltIDSource",
        ("FIX.4.4", "457"): "457  NoUnderlyingSecurityAltID",
        ("FIX.4.4", "458"): "458  UnderlyingSecurityAltID",
        ("FIX.4.4", "459"): "459  UnderlyingSecurityAltIDSource",
        ("FIX.4.4", "460"): "460  Product",
        ("FIX.4.4", "461"): "461  CFICode",
        ("FIX.4.4", "462"): "462  UnderlyingProduct",
        ("FIX.4.4", "463"): "463  UnderlyingCFICode",
        ("FIX.4.4", "464"): "464  TestMessageIndicator",
        ("FIX.4.4", "466"): "466  BookingRefID",
        ("FIX.4.4", "467"): "467  IndividualAllocID",
        ("FIX.4.4", "468"): "468  RoundingDirection",
        ("FIX.4.4", "469"): "469  RoundingModulus",
        ("FIX.4.4", "470"): "470  CountryOfIssue",
        ("FIX.4.4", "471"): "471  StateOrProvinceOfIssue",
        ("FIX.4.4", "472"): "472  LocaleOfIssue",
        ("FIX.4.4", "473"): "473  NoRegistDtls",
        ("FIX.4.4", "474"): "474  MailingDtls",
        ("FIX.4.4", "475"): "475  InvestorCountryOfResidence",
        ("FIX.4.4", "476"): "476  PaymentRef",
        ("FIX.4.4", "477"): "477  DistribPaymentMethod",
        ("FIX.4.4", "478"): "478  CashDistribCurr",
        ("FIX.4.4", "479"): "479  CommCurrency",
        ("FIX.4.4", "480"): "480  CancellationRights",
        ("FIX.4.4", "481"): "481  MoneyLaunderingStatus",
        ("FIX.4.4", "482"): "482  MailingInst",
        ("FIX.4.4", "483"): "483  TransBkdTime",
        ("FIX.4.4", "484"): "484  ExecPriceType",
        ("FIX.4.4", "485"): "485  ExecPriceAdjustment",
        ("FIX.4.4", "486"): "486  DateOfBirth",
        ("FIX.4.4", "487"): "487  TradeReportTransType",
        ("FIX.4.4", "488"): "488  CardHolderName",
        ("FIX.4.4", "489"): "489  CardNumber",
        ("FIX.4.4", "490"): "490  CardExpDate",
        ("FIX.4.4", "491"): "491  CardIssNum",
        ("FIX.4.4", "492"): "492  PaymentMethod",
        ("FIX.4.4", "493"): "493  RegistAcctType",
        ("FIX.4.4", "494"): "494  Designation",
        ("FIX.4.4", "495"): "495  TaxAdvantageType",
        ("FIX.4.4", "496"): "496  RegistRejReasonText",
        ("FIX.4.4", "497"): "497  FundRenewWaiv",
        ("FIX.4.4", "498"): "498  CashDistribAgentName",
        ("FIX.4.4", "499"): "499  CashDistribAgentCode",
        ("FIX.4.4", "500"): "500  CashDistribAgentAcctNumber",
        ("FIX.4.4", "501"): "501  CashDistribPayRef",
        ("FIX.4.4", "502"): "502  CashDistribAgentAcctName",
        ("FIX.4.4", "503"): "503  CardStartDate",
        ("FIX.4.4", "504"): "504  PaymentDate",
        ("FIX.4.4", "505"): "505  PaymentRemitterID",
        ("FIX.4.4", "506"): "506  RegistStatus",
        ("FIX.4.4", "507"): "507  RegistRejReasonCode",
        ("FIX.4.4", "508"): "508  RegistRefID",
        ("FIX.4.4", "509"): "509  RegistDtls",
        ("FIX.4.4", "510"): "510  NoDistribInsts",
        ("FIX.4.4", "511"): "511  RegistEmail",
        ("FIX.4.4", "512"): "512  DistribPercentage",
        ("FIX.4.4", "513"): "513  RegistID",
        ("FIX.4.4", "514"): "514  RegistTransType",
        ("FIX.4.4", "515"): "515  ExecValuationPoint",
        ("FIX.4.4", "516"): "516  OrderPercent",
        ("FIX.4.4", "517"): "517  OwnershipType",
        ("FIX.4.4", "518"): "518  NoContAmts",
        ("FIX.4.4", "519"): "519  ContAmtType",
        ("FIX.4.4", "520"): "520  ContAmtValue",
        ("FIX.4.4", "521"): "521  ContAmtCurr",
        ("FIX.4.4", "522"): "522  OwnerType",
        ("FIX.4.4", "523"): "523  PartySubID",
        ("FIX.4.4", "524"): "524  NestedPartyID",
        ("FIX.4.4", "525"): "525  NestedPartyIDSource",
        ("FIX.4.4", "526"): "526  SecondaryClOrdID",
        ("FIX.4.4", "527"): "527  SecondaryExecID",
        ("FIX.4.4", "528"): "528  OrderCapacity",
        ("FIX.4.4", "529"): "529  OrderRestrictions",
        ("FIX.4.4", "530"): "530  MassCancelRequestType",
        ("FIX.4.4", "531"): "531  MassCancelResponse",
        ("FIX.4.4", "532"): "532  MassCancelRejectReason",
        ("FIX.4.4", "533"): "533  TotalAffectedOrders",
        ("FIX.4.4", "534"): "534  NoAffectedOrders",
        ("FIX.4.4", "535"): "535  AffectedOrderID",
        ("FIX.4.4", "536"): "536  AffectedSecondaryOrderID",
        ("FIX.4.4", "537"): "537  QuoteType",
        ("FIX.4.4", "538"): "538  NestedPartyRole",
        ("FIX.4.4", "539"): "539  NoNestedPartyIDs",
        ("FIX.4.4", "540"): "540  TotalAccruedInterestAmt",
        ("FIX.4.4", "541"): "541  MaturityDate",
        ("FIX.4.4", "542"): "542  UnderlyingMaturityDate",
        ("FIX.4.4", "543"): "543  InstrRegistry",
        ("FIX.4.4", "544"): "544  CashMargin",
        ("FIX.4.4", "545"): "545  NestedPartySubID",
        ("FIX.4.4", "546"): "546  Scope",
        ("FIX.4.4", "547"): "547  MDImplicitDelete",
        ("FIX.4.4", "548"): "548  CrossID",
        ("FIX.4.4", "549"): "549  CrossType",
        ("FIX.4.4", "550"): "550  CrossPrioritization",
        ("FIX.4.4", "551"): "551  OrigCrossID",
        ("FIX.4.4", "552"): "552  NoSides",
        ("FIX.4.4", "553"): "553  Username",
        ("FIX.4.4", "554"): "554  Password",
        ("FIX.4.4", "555"): "555  NoLegs",
        ("FIX.4.4", "556"): "556  LegCurrency",
        ("FIX.4.4", "557"): "557  TotNoSecurityTypes",
        ("FIX.4.4", "558"): "558  NoSecurityTypes",
        ("FIX.4.4", "559"): "559  SecurityListRequestType",
        ("FIX.4.4", "560"): "560  SecurityRequestResult",
        ("FIX.4.4", "561"): "561  RoundLot",
        ("FIX.4.4", "562"): "562  MinTradeVol",
        ("FIX.4.4", "563"): "563  MultiLegRptTypeReq",
        ("FIX.4.4", "564"): "564  LegPositionEffect",
        ("FIX.4.4", "565"): "565  LegCoveredOrUncovered",
        ("FIX.4.4", "566"): "566  LegPrice",
        ("FIX.4.4", "567"): "567  TradSesStatusRejReason",
        ("FIX.4.4", "568"): "568  TradeRequestID",
        ("FIX.4.4", "569"): "569  TradeRequestType",
        ("FIX.4.4", "570"): "570  PreviouslyReported",
        ("FIX.4.4", "571"): "571  TradeReportID",
        ("FIX.4.4", "572"): "572  TradeReportRefID",
        ("FIX.4.4", "573"): "573  MatchStatus",
        ("FIX.4.4", "574"): "574  MatchType",
        ("FIX.4.4", "575"): "575  OddLot",
        ("FIX.4.4", "576"): "576  NoClearingInstructions",
        ("FIX.4.4", "577"): "577  ClearingInstruction",
        ("FIX.4.4", "578"): "578  TradeInputSource",
        ("FIX.4.4", "579"): "579  TradeInputDevice",
        ("FIX.4.4", "580"): "580  NoDates",
        ("FIX.4.4", "581"): "581  AccountType",
        ("FIX.4.4", "582"): "582  CustOrderCapacity",
        ("FIX.4.4", "583"): "583  ClOrdLinkID",
        ("FIX.4.4", "584"): "584  MassStatusReqID",
        ("FIX.4.4", "585"): "585  MassStatusReqType",
        ("FIX.4.4", "586"): "586  OrigOrdModTime",
        ("FIX.4.4", "587"): "587  LegSettlType",
        ("FIX.4.4", "588"): "588  LegSettlDate",
        ("FIX.4.4", "589"): "589  DayBookingInst",
        ("FIX.4.4", "590"): "590  BookingUnit",
        ("FIX.4.4", "591"): "591  PreallocMethod",
        ("FIX.4.4", "592"): "592  UnderlyingCountryOfIssue",
        ("FIX.4.4", "593"): "593  UnderlyingStateOrProvinceOfIssue",
        ("FIX.4.4", "594"): "594  UnderlyingLocaleOfIssue",
        ("FIX.4.4", "595"): "595  UnderlyingInstrRegistry",
        ("FIX.4.4", "596"): "596  LegCountryOfIssue",
        ("FIX.4.4", "597"): "597  LegStateOrProvinceOfIssue",
        ("FIX.4.4", "598"): "598  LegLocaleOfIssue",
        ("FIX.4.4", "599"): "599  LegInstrRegistry",
        ("FIX.4.4", "600"): "600  LegSymbol",
        ("FIX.4.4", "601"): "601  LegSymbolSfx",
        ("FIX.4.4", "602"): "602  LegSecurityID",
        ("FIX.4.4", "603"): "603  LegSecurityIDSource",
        ("FIX.4.4", "604"): "604  NoLegSecurityAltID",
        ("FIX.4.4", "605"): "605  LegSecurityAltID",
        ("FIX.4.4", "606"): "606  LegSecurityAltIDSource",
        ("FIX.4.4", "607"): "607  LegProduct",
        ("FIX.4.4", "608"): "608  LegCFICode",
        ("FIX.4.4", "609"): "609  LegSecurityType",
        ("FIX.4.4", "610"): "610  LegMaturityMonthYear",
        ("FIX.4.4", "611"): "611  LegMaturityDate",
        ("FIX.4.4", "612"): "612  LegStrikePrice",
        ("FIX.4.4", "613"): "613  LegOptAttribute",
        ("FIX.4.4", "614"): "614  LegContractMultiplier",
        ("FIX.4.4", "615"): "615  LegCouponRate",
        ("FIX.4.4", "616"): "616  LegSecurityExchange",
        ("FIX.4.4", "617"): "617  LegIssuer",
        ("FIX.4.4", "618"): "618  EncodedLegIssuerLen",
        ("FIX.4.4", "619"): "619  EncodedLegIssuer",
        ("FIX.4.4", "620"): "620  LegSecurityDesc",
        ("FIX.4.4", "621"): "621  EncodedLegSecurityDescLen",
        ("FIX.4.4", "622"): "622  EncodedLegSecurityDesc",
        ("FIX.4.4", "623"): "623  LegRatioQty",
        ("FIX.4.4", "624"): "624  LegSide",
        ("FIX.4.4", "625"): "625  TradingSessionSubID",
        ("FIX.4.4", "626"): "626  AllocType",
        ("FIX.4.4", "627"): "627  NoHops",
        ("FIX.4.4", "628"): "628  HopCompID",
        ("FIX.4.4", "629"): "629  HopSendingTime",
        ("FIX.4.4", "630"): "630  HopRefID",
        ("FIX.4.4", "631"): "631  MidPx",
        ("FIX.4.4", "632"): "632  BidYield",
        ("FIX.4.4", "633"): "633  MidYield",
        ("FIX.4.4", "634"): "634  OfferYield",
        ("FIX.4.4", "635"): "635  ClearingFeeIndicator",
        ("FIX.4.4", "636"): "636  WorkingIndicator",
        ("FIX.4.4", "637"): "637  LegLastPx",
        ("FIX.4.4", "638"): "638  PriorityIndicator",
        ("FIX.4.4", "639"): "639  PriceImprovement",
        ("FIX.4.4", "640"): "640  Price2",
        ("FIX.4.4", "641"): "641  LastForwardPoints2",
        ("FIX.4.4", "642"): "642  BidForwardPoints2",
        ("FIX.4.4", "643"): "643  OfferForwardPoints2",
        ("FIX.4.4", "644"): "644  RFQReqID",
        ("FIX.4.4", "645"): "645  MktBidPx",
        ("FIX.4.4", "646"): "646  MktOfferPx",
        ("FIX.4.4", "647"): "647  MinBidSize",
        ("FIX.4.4", "648"): "648  MinOfferSize",
        ("FIX.4.4", "649"): "649  QuoteStatusReqID",
        ("FIX.4.4", "650"): "650  LegalConfirm",
        ("FIX.4.4", "651"): "651  UnderlyingLastPx",
        ("FIX.4.4", "652"): "652  UnderlyingLastQty",
        ("FIX.4.4", "654"): "654  LegRefID",
        ("FIX.4.4", "655"): "655  ContraLegRefID",
        ("FIX.4.4", "656"): "656  SettlCurrBidFxRate",
        ("FIX.4.4", "657"): "657  SettlCurrOfferFxRate",
        ("FIX.4.4", "658"): "658  QuoteRequestRejectReason",
        ("FIX.4.4", "659"): "659  SideComplianceID",
        ("FIX.4.4", "660"): "660  AcctIDSource",
        ("FIX.4.4", "661"): "661  AllocAcctIDSource",
        ("FIX.4.4", "662"): "662  BenchmarkPrice",
        ("FIX.4.4", "663"): "663  BenchmarkPriceType",
        ("FIX.4.4", "664"): "664  ConfirmID",
        ("FIX.4.4", "665"): "665  ConfirmStatus",
        ("FIX.4.4", "666"): "666  ConfirmTransType",
        ("FIX.4.4", "667"): "667  ContractSettlMonth",
        ("FIX.4.4", "668"): "668  DeliveryForm",
        ("FIX.4.4", "669"): "669  LastParPx",
        ("FIX.4.4", "670"): "670  NoLegAllocs",
        ("FIX.4.4", "671"): "671  LegAllocAccount",
        ("FIX.4.4", "672"): "672  LegIndividualAllocID",
        ("FIX.4.4", "673"): "673  LegAllocQty",
        ("FIX.4.4", "674"): "674  LegAllocAcctIDSource",
        ("FIX.4.4", "675"): "675  LegSettlCurrency",
        ("FIX.4.4", "676"): "676  LegBenchmarkCurveCurrency",
        ("FIX.4.4", "677"): "677  LegBenchmarkCurveName",
        ("FIX.4.4", "678"): "678  LegBenchmarkCurvePoint",
        ("FIX.4.4", "679"): "679  LegBenchmarkPrice",
        ("FIX.4.4", "680"): "680  LegBenchmarkPriceType",
        ("FIX.4.4", "681"): "681  LegBidPx",
        ("FIX.4.4", "682"): "682  LegIOIQty",
        ("FIX.4.4", "683"): "683  NoLegStipulations",
        ("FIX.4.4", "684"): "684  LegOfferPx",
        ("FIX.4.4", "686"): "686  LegPriceType",
        ("FIX.4.4", "687"): "687  LegQty",
        ("FIX.4.4", "688"): "688  LegStipulationType",
        ("FIX.4.4", "689"): "689  LegStipulationValue",
        ("FIX.4.4", "690"): "690  LegSwapType",
        ("FIX.4.4", "691"): "691  Pool",
        ("FIX.4.4", "692"): "692  QuotePriceType",
        ("FIX.4.4", "693"): "693  QuoteRespID",
        ("FIX.4.4", "694"): "694  QuoteRespType",
        ("FIX.4.4", "695"): "695  QuoteQualifier",
        ("FIX.4.4", "696"): "696  YieldRedemptionDate",
        ("FIX.4.4", "697"): "697  YieldRedemptionPrice",
        ("FIX.4.4", "698"): "698  YieldRedemptionPriceType",
        ("FIX.4.4", "699"): "699  BenchmarkSecurityID",
        ("FIX.4.4", "700"): "700  ReversalIndicator",
        ("FIX.4.4", "701"): "701  YieldCalcDate",
        ("FIX.4.4", "702"): "702  NoPositions",
        ("FIX.4.4", "703"): "703  PosType",
        ("FIX.4.4", "704"): "704  LongQty",
        ("FIX.4.4", "705"): "705  ShortQty",
        ("FIX.4.4", "706"): "706  PosQtyStatus",
        ("FIX.4.4", "707"): "707  PosAmtType",
        ("FIX.4.4", "708"): "708  PosAmt",
        ("FIX.4.4", "709"): "709  PosTransType",
        ("FIX.4.4", "710"): "710  PosReqID",
        ("FIX.4.4", "711"): "711  NoUnderlyings",
        ("FIX.4.4", "712"): "712  PosMaintAction",
        ("FIX.4.4", "713"): "713  OrigPosReqRefID",
        ("FIX.4.4", "714"): "714  PosMaintRptRefID",
        ("FIX.4.4", "715"): "715  ClearingBusinessDate",
        ("FIX.4.4", "716"): "716  SettlSessID",
        ("FIX.4.4", "717"): "717  SettlSessSubID",
        ("FIX.4.4", "718"): "718  AdjustmentType",
        ("FIX.4.4", "719"): "719  ContraryInstructionIndicator",
        ("FIX.4.4", "720"): "720  PriorSpreadIndicator",
        ("FIX.4.4", "721"): "721  PosMaintRptID",
        ("FIX.4.4", "722"): "722  PosMaintStatus",
        ("FIX.4.4", "723"): "723  PosMaintResult",
        ("FIX.4.4", "724"): "724  PosReqType",
        ("FIX.4.4", "725"): "725  ResponseTransportType",
        ("FIX.4.4", "726"): "726  ResponseDestination",
        ("FIX.4.4", "727"): "727  TotalNumPosReports",
        ("FIX.4.4", "728"): "728  PosReqResult",
        ("FIX.4.4", "729"): "729  PosReqStatus",
        ("FIX.4.4", "730"): "730  SettlPrice",
        ("FIX.4.4", "731"): "731  SettlPriceType",
        ("FIX.4.4", "732"): "732  UnderlyingSettlPrice",
        ("FIX.4.4", "733"): "733  UnderlyingSettlPriceType",
        ("FIX.4.4", "734"): "734  PriorSettlPrice",
        ("FIX.4.4", "735"): "735  NoQuoteQualifiers",
        ("FIX.4.4", "736"): "736  AllocSettlCurrency",
        ("FIX.4.4", "737"): "737  AllocSettlCurrAmt",
        ("FIX.4.4", "738"): "738  InterestAtMaturity",
        ("FIX.4.4", "739"): "739  LegDatedDate",
        ("FIX.4.4", "740"): "740  LegPool",
        ("FIX.4.4", "741"): "741  AllocInterestAtMaturity",
        ("FIX.4.4", "742"): "742  AllocAccruedInterestAmt",
        ("FIX.4.4", "743"): "743  DeliveryDate",
        ("FIX.4.4", "744"): "744  AssignmentMethod",
        ("FIX.4.4", "745"): "745  AssignmentUnit",
        ("FIX.4.4", "746"): "746  OpenInterest",
        ("FIX.4.4", "747"): "747  ExerciseMethod",
        ("FIX.4.4", "748"): "748  TotNumTradeReports",
        ("FIX.4.4", "749"): "749  TradeRequestResult",
        ("FIX.4.4", "750"): "750  TradeRequestStatus",
        ("FIX.4.4", "751"): "751  TradeReportRejectReason",
        ("FIX.4.4", "752"): "752  SideMultiLegReportingType",
        ("FIX.4.4", "753"): "753  NoPosAmt",
        ("FIX.4.4", "754"): "754  AutoAcceptIndicator",
        ("FIX.4.4", "755"): "755  AllocReportID",
        ("FIX.4.4", "756"): "756  NoNested2PartyIDs",
        ("FIX.4.4", "757"): "757  Nested2PartyID",
        ("FIX.4.4", "758"): "758  Nested2PartyIDSource",
        ("FIX.4.4", "759"): "759  Nested2PartyRole",
        ("FIX.4.4", "760"): "760  Nested2PartySubID",
        ("FIX.4.4", "761"): "761  BenchmarkSecurityIDSource",
        ("FIX.4.4", "762"): "762  SecuritySubType",
        ("FIX.4.4", "763"): "763  UnderlyingSecuritySubType",
        ("FIX.4.4", "764"): "764  LegSecuritySubType",
        ("FIX.4.4", "765"): "765  AllowableOneSidednessPct",
        ("FIX.4.4", "766"): "766  AllowableOneSidednessValue",
        ("FIX.4.4", "767"): "767  AllowableOneSidednessCurr",
        ("FIX.4.4", "768"): "768  NoTrdRegTimestamps",
        ("FIX.4.4", "769"): "769  TrdRegTimestamp",
        ("FIX.4.4", "770"): "770  TrdRegTimestampType",
        ("FIX.4.4", "771"): "771  TrdRegTimestampOrigin",
        ("FIX.4.4", "772"): "772  ConfirmRefID",
        ("FIX.4.4", "773"): "773  ConfirmType",
        ("FIX.4.4", "774"): "774  ConfirmRejReason",
        ("FIX.4.4", "775"): "775  BookingType",
        ("FIX.4.4", "776"): "776  IndividualAllocRejCode",
        ("FIX.4.4", "777"): "777  SettlInstMsgID",
        ("FIX.4.4", "778"): "778  NoSettlInst",
        ("FIX.4.4", "779"): "779  LastUpdateTime",
        ("FIX.4.4", "780"): "780  AllocSettlInstType",
        ("FIX.4.4", "781"): "781  NoSettlPartyIDs",
        ("FIX.4.4", "782"): "782  SettlPartyID",
        ("FIX.4.4", "783"): "783  SettlPartyIDSource",
        ("FIX.4.4", "784"): "784  SettlPartyRole",
        ("FIX.4.4", "785"): "785  SettlPartySubID",
        ("FIX.4.4", "786"): "786  SettlPartySubIDType",
        ("FIX.4.4", "787"): "787  DlvyInstType",
        ("FIX.4.4", "788"): "788  TerminationType",
        ("FIX.4.4", "789"): "789  NextExpectedMsgSeqNum",
        ("FIX.4.4", "790"): "790  OrdStatusReqID",
        ("FIX.4.4", "791"): "791  SettlInstReqID",
        ("FIX.4.4", "792"): "792  SettlInstReqRejCode",
        ("FIX.4.4", "793"): "793  SecondaryAllocID",
        ("FIX.4.4", "794"): "794  AllocReportType",
        ("FIX.4.4", "795"): "795  AllocReportRefID",
        ("FIX.4.4", "796"): "796  AllocCancReplaceReason",
        ("FIX.4.4", "797"): "797  CopyMsgIndicator",
        ("FIX.4.4", "798"): "798  AllocAccountType",
        ("FIX.4.4", "799"): "799  OrderAvgPx",
        ("FIX.4.4", "800"): "800  OrderBookingQty",
        ("FIX.4.4", "801"): "801  NoSettlPartySubIDs",
        ("FIX.4.4", "802"): "802  NoPartySubIDs",
        ("FIX.4.4", "803"): "803  PartySubIDType",
        ("FIX.4.4", "804"): "804  NoNestedPartySubIDs",
        ("FIX.4.4", "805"): "805  NestedPartySubIDType",
        ("FIX.4.4", "806"): "806  NoNested2PartySubIDs",
        ("FIX.4.4", "807"): "807  Nested2PartySubIDType",
        ("FIX.4.4", "808"): "808  AllocIntermedReqType",
        ("FIX.4.4", "810"): "810  UnderlyingPx",
        ("FIX.4.4", "811"): "811  PriceDelta",
        ("FIX.4.4", "812"): "812  ApplQueueMax",
        ("FIX.4.4", "813"): "813  ApplQueueDepth",
        ("FIX.4.4", "814"): "814  ApplQueueResolution",
        ("FIX.4.4", "815"): "815  ApplQueueAction",
        ("FIX.4.4", "816"): "816  NoAltMDSource",
        ("FIX.4.4", "817"): "817  AltMDSourceID",
        ("FIX.4.4", "818"): "818  SecondaryTradeReportID",
        ("FIX.4.4", "819"): "819  AvgPxIndicator",
        ("FIX.4.4", "820"): "820  TradeLinkID",
        ("FIX.4.4", "821"): "821  OrderInputDevice",
        ("FIX.4.4", "822"): "822  UnderlyingTradingSessionID",
        ("FIX.4.4", "823"): "823  UnderlyingTradingSessionSubID",
        ("FIX.4.4", "824"): "824  TradeLegRefID",
        ("FIX.4.4", "825"): "825  ExchangeRule",
        ("FIX.4.4", "826"): "826  TradeAllocIndicator",
        ("FIX.4.4", "827"): "827  ExpirationCycle",
        ("FIX.4.4", "828"): "828  TrdType",
        ("FIX.4.4", "829"): "829  TrdSubType",
        ("FIX.4.4", "830"): "830  TransferReason",
        ("FIX.4.4", "832"): "832  TotNumAssignmentReports",
        ("FIX.4.4", "833"): "833  AsgnRptID",
        ("FIX.4.4", "834"): "834  ThresholdAmount",
        ("FIX.4.4", "835"): "835  PegMoveType",
        ("FIX.4.4", "836"): "836  PegOffsetType",
        ("FIX.4.4", "837"): "837  PegLimitType",
        ("FIX.4.4", "838"): "838  PegRoundDirection",
        ("FIX.4.4", "839"): "839  PeggedPrice",
        ("FIX.4.4", "840"): "840  PegScope",
        ("FIX.4.4", "841"): "841  DiscretionMoveType",
        ("FIX.4.4", "842"): "842  DiscretionOffsetType",
        ("FIX.4.4", "843"): "843  DiscretionLimitType",
        ("FIX.4.4", "844"): "844  DiscretionRoundDirection",
        ("FIX.4.4", "845"): "845  DiscretionPrice",
        ("FIX.4.4", "846"): "846  DiscretionScope",
        ("FIX.4.4", "847"): "847  TargetStrategy",
        ("FIX.4.4", "848"): "848  TargetStrategyParameters",
        ("FIX.4.4", "849"): "849  ParticipationRate",
        ("FIX.4.4", "850"): "850  TargetStrategyPerformance",
        ("FIX.4.4", "851"): "851  LastLiquidityInd",
        ("FIX.4.4", "852"): "852  PublishTrdIndicator",
        ("FIX.4.4", "853"): "853  ShortSaleReason",
        ("FIX.4.4", "854"): "854  QtyType",
        ("FIX.4.4", "855"): "855  SecondaryTrdType",
        ("FIX.4.4", "856"): "856  TradeReportType",
        ("FIX.4.4", "857"): "857  AllocNoOrdersType",
        ("FIX.4.4", "858"): "858  SharedCommission",
        ("FIX.4.4", "859"): "859  ConfirmReqID",
        ("FIX.4.4", "860"): "860  AvgParPx",
        ("FIX.4.4", "861"): "861  ReportedPx",
        ("FIX.4.4", "862"): "862  NoCapacities",
        ("FIX.4.4", "863"): "863  OrderCapacityQty",
        ("FIX.4.4", "864"): "864  NoEvents",
        ("FIX.4.4", "865"): "865  EventType",
        ("FIX.4.4", "866"): "866  EventDate",
        ("FIX.4.4", "867"): "867  EventPx",
        ("FIX.4.4", "868"): "868  EventText",
        ("FIX.4.4", "869"): "869  PctAtRisk",
        ("FIX.4.4", "870"): "870  NoInstrAttrib",
        ("FIX.4.4", "871"): "871  InstrAttribType",
        ("FIX.4.4", "872"): "872  InstrAttribValue",
        ("FIX.4.4", "873"): "873  DatedDate",
        ("FIX.4.4", "874"): "874  InterestAccrualDate",
        ("FIX.4.4", "875"): "875  CPProgram",
        ("FIX.4.4", "876"): "876  CPRegType",
        ("FIX.4.4", "877"): "877  UnderlyingCPProgram",
        ("FIX.4.4", "878"): "878  UnderlyingCPRegType",
        ("FIX.4.4", "879"): "879  UnderlyingQty",
        ("FIX.4.4", "880"): "880  TrdMatchID",
        ("FIX.4.4", "881"): "881  SecondaryTradeReportRefID",
        ("FIX.4.4", "882"): "882  UnderlyingDirtyPrice",
        ("FIX.4.4", "883"): "883  UnderlyingEndPrice",
        ("FIX.4.4", "884"): "884  UnderlyingStartValue",
        ("FIX.4.4", "885"): "885  UnderlyingCurrentValue",
        ("FIX.4.4", "886"): "886  UnderlyingEndValue",
        ("FIX.4.4", "887"): "887  NoUnderlyingStips",
        ("FIX.4.4", "888"): "888  UnderlyingStipType",
        ("FIX.4.4", "889"): "889  UnderlyingStipValue",
        ("FIX.4.4", "890"): "890  MaturityNetMoney",
        ("FIX.4.4", "891"): "891  MiscFeeBasis",
        ("FIX.4.4", "892"): "892  TotNoAllocs",
        ("FIX.4.4", "893"): "893  LastFragment",
        ("FIX.4.4", "894"): "894  CollReqID",
        ("FIX.4.4", "895"): "895  CollAsgnReason",
        ("FIX.4.4", "896"): "896  CollInquiryQualifier",
        ("FIX.4.4", "897"): "897  NoTrades",
        ("FIX.4.4", "898"): "898  MarginRatio",
        ("FIX.4.4", "899"): "899  MarginExcess",
        ("FIX.4.4", "900"): "900  TotalNetValue",
        ("FIX.4.4", "901"): "901  CashOutstanding",
        ("FIX.4.4", "902"): "902  CollAsgnID",
        ("FIX.4.4", "903"): "903  CollAsgnTransType",
        ("FIX.4.4", "904"): "904  CollRespID",
        ("FIX.4.4", "905"): "905  CollAsgnRespType",
        ("FIX.4.4", "906"): "906  CollAsgnRejectReason",
        ("FIX.4.4", "907"): "907  CollAsgnRefID",
        ("FIX.4.4", "908"): "908  CollRptID",
        ("FIX.4.4", "909"): "909  CollInquiryID",
        ("FIX.4.4", "910"): "910  CollStatus",
        ("FIX.4.4", "911"): "911  TotNumReports",
        ("FIX.4.4", "912"): "912  LastRptRequested",
        ("FIX.4.4", "913"): "913  AgreementDesc",
        ("FIX.4.4", "914"): "914  AgreementID",
        ("FIX.4.4", "915"): "915  AgreementDate",
        ("FIX.4.4", "916"): "916  StartDate",
        ("FIX.4.4", "917"): "917  EndDate",
        ("FIX.4.4", "918"): "918  AgreementCurrency",
        ("FIX.4.4", "919"): "919  DeliveryType",
        ("FIX.4.4", "920"): "920  EndAccruedInterestAmt",
        ("FIX.4.4", "921"): "921  StartCash",
        ("FIX.4.4", "922"): "922  EndCash",
        ("FIX.4.4", "923"): "923  UserRequestID",
        ("FIX.4.4", "924"): "924  UserRequestType",
        ("FIX.4.4", "925"): "925  NewPassword",
        ("FIX.4.4", "926"): "926  UserStatus",
        ("FIX.4.4", "927"): "927  UserStatusText",
        ("FIX.4.4", "928"): "928  StatusValue",
        ("FIX.4.4", "929"): "929  StatusText",
        ("FIX.4.4", "930"): "930  RefCompID",
        ("FIX.4.4", "931"): "931  RefSubID",
        ("FIX.4.4", "932"): "932  NetworkResponseID",
        ("FIX.4.4", "933"): "933  NetworkRequestID",
        ("FIX.4.4", "934"): "934  LastNetworkResponseID",
        ("FIX.4.4", "935"): "935  NetworkRequestType",
        ("FIX.4.4", "936"): "936  NoCompIDs",
        ("FIX.4.4", "937"): "937  NetworkStatusResponseType",
        ("FIX.4.4", "938"): "938  NoCollInquiryQualifier",
        ("FIX.4.4", "939"): "939  TrdRptStatus",
        ("FIX.4.4", "940"): "940  AffirmStatus",
        ("FIX.4.4", "941"): "941  UnderlyingStrikeCurrency",
        ("FIX.4.4", "942"): "942  LegStrikeCurrency",
        ("FIX.4.4", "943"): "943  TimeBracket",
        ("FIX.4.4", "944"): "944  CollAction",
        ("FIX.4.4", "945"): "945  CollInquiryStatus",
        ("FIX.4.4", "946"): "946  CollInquiryResult",
        ("FIX.4.4", "947"): "947  StrikeCurrency",
        ("FIX.4.4", "948"): "948  NoNested3PartyIDs",
        ("FIX.4.4", "949"): "949  Nested3PartyID",
        ("FIX.4.4", "950"): "950  Nested3PartyIDSource",
        ("FIX.4.4", "951"): "951  Nested3PartyRole",
        ("FIX.4.4", "952"): "952  NoNested3PartySubIDs",
        ("FIX.4.4", "953"): "953  Nested3PartySubID",
        ("FIX.4.4", "954"): "954  Nested3PartySubIDType",
        ("FIX.4.4", "955"): "955  LegContractSettlMonth",
        ("FIX.4.4", "956"): "956  LegInterestAccrualDate",
        ("FIX.5.0", "1"): "1    Account",
        ("FIX.5.0", "2"): "2    AdvId",
        ("FIX.5.0", "3"): "3    AdvRefID",
        ("FIX.5.0", "4"): "4    AdvSide",
        ("FIX.5.0", "5"): "5    AdvTransType",
        ("FIX.5.0", "6"): "6    AvgPx",
        ("FIX.5.0", "7"): "7    BeginSeqNo",
        ("FIX.5.0", "8"): "8    BeginString",
        ("FIX.5.0", "9"): "9    BodyLength",
        ("FIX.5.0", "10"): "10   CheckSum",
        ("FIX.5.0", "11"): "11   ClOrdID",
        ("FIX.5.0", "12"): "12   Commission",
        ("FIX.5.0", "13"): "13   CommType",
        ("FIX.5.0", "14"): "14   CumQty",
        ("FIX.5.0", "15"): "15   Currency",
        ("FIX.5.0", "16"): "16   EndSeqNo",
        ("FIX.5.0", "17"): "17   ExecID",
        ("FIX.5.0", "18"): "18   ExecInst",
        ("FIX.5.0", "19"): "19   ExecRefID",
        ("FIX.5.0", "21"): "21   HandlInst",
        ("FIX.5.0", "22"): "22   SecurityIDSource",
        ("FIX.5.0", "23"): "23   IOIID",
        ("FIX.5.0", "25"): "25   IOIQltyInd",
        ("FIX.5.0", "26"): "26   IOIRefID",
        ("FIX.5.0", "27"): "27   IOIQty",
        ("FIX.5.0", "28"): "28   IOITransType",
        ("FIX.5.0", "29"): "29   LastCapacity",
        ("FIX.5.0", "30"): "30   LastMkt",
        ("FIX.5.0", "31"): "31   LastPx",
        ("FIX.5.0", "32"): "32   LastQty",
        ("FIX.5.0", "33"): "33   NoLinesOfText",
        ("FIX.5.0", "34"): "34   MsgSeqNum",
        ("FIX.5.0", "35"): "35   MsgType",
        ("FIX.5.0", "36"): "36   NewSeqNo",
        ("FIX.5.0", "37"): "37   OrderID",
        ("FIX.5.0", "38"): "38   OrderQty",
        ("FIX.5.0", "39"): "39   OrdStatus",
        ("FIX.5.0", "40"): "40   OrdType",
        ("FIX.5.0", "41"): "41   OrigClOrdID",
        ("FIX.5.0", "42"): "42   OrigTime",
        ("FIX.5.0", "43"): "43   PossDupFlag",
        ("FIX.5.0", "44"): "44   Price",
        ("FIX.5.0", "45"): "45   RefSeqNum",
        ("FIX.5.0", "48"): "48   SecurityID",
        ("FIX.5.0", "49"): "49   SenderCompID",
        ("FIX.5.0", "50"): "50   SenderSubID",
        ("FIX.5.0", "52"): "52   SendingTime",
        ("FIX.5.0", "53"): "53   Quantity",
        ("FIX.5.0", "54"): "54   Side",
        ("FIX.5.0", "55"): "55   Symbol",
        ("FIX.5.0", "56"): "56   TargetCompID",
        ("FIX.5.0", "57"): "57   TargetSubID",
        ("FIX.5.0", "58"): "58   Text",
        ("FIX.5.0", "59"): "59   TimeInForce",
        ("FIX.5.0", "60"): "60   TransactTime",
        ("FIX.5.0", "61"): "61   Urgency",
        ("FIX.5.0", "62"): "62   ValidUntilTime",
        ("FIX.5.0", "63"): "63   SettlType",
        ("FIX.5.0", "64"): "64   SettlDate",
        ("FIX.5.0", "65"): "65   SymbolSfx",
        ("FIX.5.0", "66"): "66   ListID",
        ("FIX.5.0", "67"): "67   ListSeqNo",
        ("FIX.5.0", "68"): "68   TotNoOrders",
        ("FIX.5.0", "69"): "69   ListExecInst",
        ("FIX.5.0", "70"): "70   AllocID",
        ("FIX.5.0", "71"): "71   AllocTransType",
        ("FIX.5.0", "72"): "72   RefAllocID",
        ("FIX.5.0", "73"): "73   NoOrders",
        ("FIX.5.0", "74"): "74   AvgPxPrecision",
        ("FIX.5.0", "75"): "75   TradeDate",
        ("FIX.5.0", "77"): "77   PositionEffect",
        ("FIX.5.0", "78"): "78   NoAllocs",
        ("FIX.5.0", "79"): "79   AllocAccount",
        ("FIX.5.0", "80"): "80   AllocQty",
        ("FIX.5.0", "81"): "81   ProcessCode",
        ("FIX.5.0", "82"): "82   NoRpts",
        ("FIX.5.0", "83"): "83   RptSeq",
        ("FIX.5.0", "84"): "84   CxlQty",
        ("FIX.5.0", "85"): "85   NoDlvyInst",
        ("FIX.5.0", "87"): "87   AllocStatus",
        ("FIX.5.0", "88"): "88   AllocRejCode",
        ("FIX.5.0", "89"): "89   Signature",
        ("FIX.5.0", "90"): "90   SecureDataLen",
        ("FIX.5.0", "91"): "91   SecureData",
        ("FIX.5.0", "93"): "93   SignatureLength",
        ("FIX.5.0", "94"): "94   EmailType",
        ("FIX.5.0", "95"): "95   RawDataLength",
        ("FIX.5.0", "96"): "96   RawData",
        ("FIX.5.0", "97"): "97   PossResend",
        ("FIX.5.0", "98"): "98   EncryptMethod",
        ("FIX.5.0", "99"): "99   StopPx",
        ("FIX.5.0", "100"): "100  ExDestination",
        ("FIX.5.0", "102"): "102  CxlRejReason",
        ("FIX.5.0", "103"): "103  OrdRejReason",
        ("FIX.5.0", "104"): "104  IOIQualifier",
        ("FIX.5.0", "106"): "106  Issuer",
        ("FIX.5.0", "107"): "107  SecurityDesc",
        ("FIX.5.0", "108"): "108  HeartBtInt",
        ("FIX.5.0", "110"): "110  MinQty",
        ("FIX.5.0", "111"): "111  MaxFloor",
        ("FIX.5.0", "112"): "112  TestReqID",
        ("FIX.5.0", "113"): "113  ReportToExch",
        ("FIX.5.0", "114"): "114  LocateReqd",
        ("FIX.5.0", "115"): "115  OnBehalfOfCompID",
        ("FIX.5.0", "116"): "116  OnBehalfOfSubID",
        ("FIX.5.0", "117"): "117  QuoteID",
        ("FIX.5.0", "118"): "118  NetMoney",
        ("FIX.5.0", "119"): "119  SettlCurrAmt",
        ("FIX.5.0", "120"): "120  SettlCurrency",
        ("FIX.5.0", "121"): "121  ForexReq",
        ("FIX.5.0", "122"): "122  OrigSendingTime",
        ("FIX.5.0", "123"): "123  GapFillFlag",
        ("FIX.5.0", "124"): "124  NoExecs",
        ("FIX.5.0", "126"): "126  ExpireTime",
        ("FIX.5.0", "127"): "127  DKReason",
        ("FIX.5.0", "128"): "128  DeliverToCompID",
        ("FIX.5.0", "129"): "129  DeliverToSubID",
        ("FIX.5.0", "130"): "130  IOINaturalFlag",
        ("FIX.5.0", "131"): "131  QuoteReqID",
        ("FIX.5.0", "132"): "132  BidPx",
        ("FIX.5.0", "133"): "133  OfferPx",
        ("FIX.5.0", "134"): "134  BidSize",
        ("FIX.5.0", "135"): "135  OfferSize",
        ("FIX.5.0", "136"): "136  NoMiscFees",
        ("FIX.5.0", "137"): "137  MiscFeeAmt",
        ("FIX.5.0", "138"): "138  MiscFeeCurr",
        ("FIX.5.0", "139"): "139  MiscFeeType",
        ("FIX.5.0", "140"): "140  PrevClosePx",
        ("FIX.5.0", "141"): "141  ResetSeqNumFlag",
        ("FIX.5.0", "142"): "142  SenderLocationID",
        ("FIX.5.0", "143"): "143  TargetLocationID",
        ("FIX.5.0", "144"): "144  OnBehalfOfLocationID",
        ("FIX.5.0", "145"): "145  DeliverToLocationID",
        ("FIX.5.0", "146"): "146  NoRelatedSym",
        ("FIX.5.0", "147"): "147  Subject",
        ("FIX.5.0", "148"): "148  Headline",
        ("FIX.5.0", "149"): "149  URLLink",
        ("FIX.5.0", "150"): "150  ExecType",
        ("FIX.5.0", "151"): "151  LeavesQty",
        ("FIX.5.0", "152"): "152  CashOrderQty",
        ("FIX.5.0", "153"): "153  AllocAvgPx",
        ("FIX.5.0", "154"): "154  AllocNetMoney",
        ("FIX.5.0", "155"): "155  SettlCurrFxRate",
        ("FIX.5.0", "156"): "156  SettlCurrFxRateCalc",
        ("FIX.5.0", "157"): "157  NumDaysInterest",
        ("FIX.5.0", "158"): "158  AccruedInterestRate",
        ("FIX.5.0", "159"): "159  AccruedInterestAmt",
        ("FIX.5.0", "160"): "160  SettlInstMode",
        ("FIX.5.0", "161"): "161  AllocText",
        ("FIX.5.0", "162"): "162  SettlInstID",
        ("FIX.5.0", "163"): "163  SettlInstTransType",
        ("FIX.5.0", "164"): "164  EmailThreadID",
        ("FIX.5.0", "165"): "165  SettlInstSource",
        ("FIX.5.0", "167"): "167  SecurityType",
        ("FIX.5.0", "168"): "168  EffectiveTime",
        ("FIX.5.0", "169"): "169  StandInstDbType",
        ("FIX.5.0", "170"): "170  StandInstDbName",
        ("FIX.5.0", "171"): "171  StandInstDbID",
        ("FIX.5.0", "172"): "172  SettlDeliveryType",
        ("FIX.5.0", "188"): "188  BidSpotRate",
        ("FIX.5.0", "189"): "189  BidForwardPoints",
        ("FIX.5.0", "190"): "190  OfferSpotRate",
        ("FIX.5.0", "191"): "191  OfferForwardPoints",
        ("FIX.5.0", "192"): "192  OrderQty2",
        ("FIX.5.0", "193"): "193  SettlDate2",
        ("FIX.5.0", "194"): "194  LastSpotRate",
        ("FIX.5.0", "195"): "195  LastForwardPoints",
        ("FIX.5.0", "196"): "196  AllocLinkID",
        ("FIX.5.0", "197"): "197  AllocLinkType",
        ("FIX.5.0", "198"): "198  SecondaryOrderID",
        ("FIX.5.0", "199"): "199  NoIOIQualifiers",
        ("FIX.5.0", "200"): "200  MaturityMonthYear",
        ("FIX.5.0", "201"): "201  PutOrCall",
        ("FIX.5.0", "202"): "202  StrikePrice",
        ("FIX.5.0", "203"): "203  CoveredOrUncovered",
        ("FIX.5.0", "206"): "206  OptAttribute",
        ("FIX.5.0", "207"): "207  SecurityExchange",
        ("FIX.5.0", "208"): "208  NotifyBrokerOfCredit",
        ("FIX.5.0", "209"): "209  AllocHandlInst",
        ("FIX.5.0", "210"): "210  MaxShow",
        ("FIX.5.0", "211"): "211  PegOffsetValue",
        ("FIX.5.0", "212"): "212  XmlDataLen",
        ("FIX.5.0", "213"): "213  XmlData",
        ("FIX.5.0", "214"): "214  SettlInstRefID",
        ("FIX.5.0", "215"): "215  NoRoutingIDs",
        ("FIX.5.0", "216"): "216  RoutingType",
        ("FIX.5.0", "217"): "217  RoutingID",
        ("FIX.5.0", "218"): "218  Spread",
        ("FIX.5.0", "220"): "220  BenchmarkCurveCurrency",
        ("FIX.5.0", "221"): "221  BenchmarkCurveName",
        ("FIX.5.0", "222"): "222  BenchmarkCurvePoint",
        ("FIX.5.0", "223"): "223  CouponRate",
        ("FIX.5.0", "224"): "224  CouponPaymentDate",
        ("FIX.5.0", "225"): "225  IssueDate",
        ("FIX.5.0", "226"): "226  RepurchaseTerm",
        ("FIX.5.0", "227"): "227  RepurchaseRate",
        ("FIX.5.0", "228"): "228  Factor",
        ("FIX.5.0", "229"): "229  TradeOriginationDate",
        ("FIX.5.0", "230"): "230  ExDate",
        ("FIX.5.0", "231"): "231  ContractMultiplier",
        ("FIX.5.0", "232"): "232  NoStipulations",
        ("FIX.5.0", "233"): "233  StipulationType",
        ("FIX.5.0", "234"): "234  StipulationValue",
        ("FIX.5.0", "235"): "235  YieldType",
        ("FIX.5.0", "236"): "236  Yield",
        ("FIX.5.0", "237"): "237  TotalTakedown",
        ("FIX.5.0", "238"): "238  Concession",
        ("FIX.5.0", "239"): "239  RepoCollateralSecurityType",
        ("FIX.5.0", "240"): "240  RedemptionDate",
        ("FIX.5.0", "241"): "241  UnderlyingCouponPaymentDate",
        ("FIX.5.0", "242"): "242  UnderlyingIssueDate",
        ("FIX.5.0", "243"): "243  UnderlyingRepoCollateralSecurityType",
        ("FIX.5.0", "244"): "244  UnderlyingRepurchaseTerm",
        ("FIX.5.0", "245"): "245  UnderlyingRepurchaseRate",
        ("FIX.5.0", "246"): "246  UnderlyingFactor",
        ("FIX.5.0", "247"): "247  UnderlyingRedemptionDate",
        ("FIX.5.0", "248"): "248  LegCouponPaymentDate",
        ("FIX.5.0", "249"): "249  LegIssueDate",
        ("FIX.5.0", "250"): "250  LegRepoCollateralSecurityType",
        ("FIX.5.0", "251"): "251  LegRepurchaseTerm",
        ("FIX.5.0", "252"): "252  LegRepurchaseRate",
        ("FIX.5.0", "253"): "253  LegFactor",
        ("FIX.5.0", "254"): "254  LegRedemptionDate",
        ("FIX.5.0", "255"): "255  CreditRating",
        ("FIX.5.0", "256"): "256  UnderlyingCreditRating",
        ("FIX.5.0", "257"): "257  LegCreditRating",
        ("FIX.5.0", "258"): "258  TradedFlatSwitch",
        ("FIX.5.0", "259"): "259  BasisFeatureDate",
        ("FIX.5.0", "260"): "260  BasisFeaturePrice",
        ("FIX.5.0", "262"): "262  MDReqID",
        ("FIX.5.0", "263"): "263  SubscriptionRequestType",
        ("FIX.5.0", "264"): "264  MarketDepth",
        ("FIX.5.0", "265"): "265  MDUpdateType",
        ("FIX.5.0", "266"): "266  AggregatedBook",
        ("FIX.5.0", "267"): "267  NoMDEntryTypes",
        ("FIX.5.0", "268"): "268  NoMDEntries",
        ("FIX.5.0", "269"): "269  MDEntryType",
        ("FIX.5.0", "270"): "270  MDEntryPx",
        ("FIX.5.0", "271"): "271  MDEntrySize",
        ("FIX.5.0", "272"): "272  MDEntryDate",
        ("FIX.5.0", "273"): "273  MDEntryTime",
        ("FIX.5.0", "274"): "274  TickDirection",
        ("FIX.5.0", "275"): "275  MDMkt",
        ("FIX.5.0", "276"): "276  QuoteCondition",
        ("FIX.5.0", "277"): "277  TradeCondition",
        ("FIX.5.0", "278"): "278  MDEntryID",
        ("FIX.5.0", "279"): "279  MDUpdateAction",
        ("FIX.5.0", "280"): "280  MDEntryRefID",
        ("FIX.5.0", "281"): "281  MDReqRejReason",
        ("FIX.5.0", "282"): "282  MDEntryOriginator",
        ("FIX.5.0", "283"): "283  LocationID",
        ("FIX.5.0", "284"): "284  DeskID",
        ("FIX.5.0", "285"): "285  DeleteReason",
        ("FIX.5.0", "286"): "286  OpenCloseSettlFlag",
        ("FIX.5.0", "287"): "287  SellerDays",
        ("FIX.5.0", "288"): "288  MDEntryBuyer",
        ("FIX.5.0", "289"): "289  MDEntrySeller",
        ("FIX.5.0", "290"): "290  MDEntryPositionNo",
        ("FIX.5.0", "291"): "291  FinancialStatus",
        ("FIX.5.0", "292"): "292  CorporateAction",
        ("FIX.5.0", "293"): "293  DefBidSize",
        ("FIX.5.0", "294"): "294  DefOfferSize",
        ("FIX.5.0", "295"): "295  NoQuoteEntries",
        ("FIX.5.0", "296"): "296  NoQuoteSets",
        ("FIX.5.0", "297"): "297  QuoteStatus",
        ("FIX.5.0", "298"): "298  QuoteCancelType",
        ("FIX.5.0", "299"): "299  QuoteEntryID",
        ("FIX.5.0", "300"): "300  QuoteRejectReason",
        ("FIX.5.0", "301"): "301  QuoteResponseLevel",
        ("FIX.5.0", "302"): "302  QuoteSetID",
        ("FIX.5.0", "303"): "303  QuoteRequestType",
        ("FIX.5.0", "304"): "304  TotNoQuoteEntries",
        ("FIX.5.0", "305"): "305  UnderlyingSecurityIDSource",
        ("FIX.5.0", "306"): "306  UnderlyingIssuer",
        ("FIX.5.0", "307"): "307  UnderlyingSecurityDesc",
        ("FIX.5.0", "308"): "308  UnderlyingSecurityExchange",
        ("FIX.5.0", "309"): "309  UnderlyingSecurityID",
        ("FIX.5.0", "310"): "310  UnderlyingSecurityType",
        ("FIX.5.0", "311"): "311  UnderlyingSymbol",
        ("FIX.5.0", "312"): "312  UnderlyingSymbolSfx",
        ("FIX.5.0", "313"): "313  UnderlyingMaturityMonthYear",
        ("FIX.5.0", "315"): "315  UnderlyingPutOrCall",
        ("FIX.5.0", "316"): "316  UnderlyingStrikePrice",
        ("FIX.5.0", "317"): "317  UnderlyingOptAttribute",
        ("FIX.5.0", "318"): "318  UnderlyingCurrency",
        ("FIX.5.0", "320"): "320  SecurityReqID",
        ("FIX.5.0", "321"): "321  SecurityRequestType",
        ("FIX.5.0", "322"): "322  SecurityResponseID",
        ("FIX.5.0", "323"): "323  SecurityResponseType",
        ("FIX.5.0", "324"): "324  SecurityStatusReqID",
        ("FIX.5.0", "325"): "325  UnsolicitedIndicator",
        ("FIX.5.0", "326"): "326  SecurityTradingStatus",
        ("FIX.5.0", "327"): "327  HaltReason",
        ("FIX.5.0", "328"): "328  InViewOfCommon",
        ("FIX.5.0", "329"): "329  DueToRelated",
        ("FIX.5.0", "330"): "330  BuyVolume",
        ("FIX.5.0", "331"): "331  SellVolume",
        ("FIX.5.0", "332"): "332  HighPx",
        ("FIX.5.0", "333"): "333  LowPx",
        ("FIX.5.0", "334"): "334  Adjustment",
        ("FIX.5.0", "335"): "335  TradSesReqID",
        ("FIX.5.0", "336"): "336  TradingSessionID",
        ("FIX.5.0", "337"): "337  ContraTrader",
        ("FIX.5.0", "338"): "338  TradSesMethod",
        ("FIX.5.0", "339"): "339  TradSesMode",
        ("FIX.5.0", "340"): "340  TradSesStatus",
        ("FIX.5.0", "341"): "341  TradSesStartTime",
        ("FIX.5.0", "342"): "342  TradSesOpenTime",
        ("FIX.5.0", "343"): "343  TradSesPreCloseTime",
        ("FIX.5.0", "344"): "344  TradSesCloseTime",
        ("FIX.5.0", "345"): "345  TradSesEndTime",
        ("FIX.5.0", "346"): "346  NumberOfOrders",
        ("FIX.5.0", "347"): "347  MessageEncoding",
        ("FIX.5.0", "348"): "348  EncodedIssuerLen",
        ("FIX.5.0", "349"): "349  EncodedIssuer",
        ("FIX.5.0", "350"): "350  EncodedSecurityDescLen",
        ("FIX.5.0", "351"): "351  EncodedSecurityDesc",
        ("FIX.5.0", "352"): "352  EncodedListExecInstLen",
        ("FIX.5.0", "353"): "353  EncodedListExecInst",
        ("FIX.5.0", "354"): "354  EncodedTextLen",
        ("FIX.5.0", "355"): "355  EncodedText",
        ("FIX.5.0", "356"): "356  EncodedSubjectLen",
        ("FIX.5.0", "357"): "357  EncodedSubject",
        ("FIX.5.0", "358"): "358  EncodedHeadlineLen",
        ("FIX.5.0", "359"): "359  EncodedHeadline",
        ("FIX.5.0", "360"): "360  EncodedAllocTextLen",
        ("FIX.5.0", "361"): "361  EncodedAllocText",
        ("FIX.5.0", "362"): "362  EncodedUnderlyingIssuerLen",
        ("FIX.5.0", "363"): "363  EncodedUnderlyingIssuer",
        ("FIX.5.0", "364"): "364  EncodedUnderlyingSecurityDescLen",
        ("FIX.5.0", "365"): "365  EncodedUnderlyingSecurityDesc",
        ("FIX.5.0", "366"): "366  AllocPrice",
        ("FIX.5.0", "367"): "367  QuoteSetValidUntilTime",
        ("FIX.5.0", "368"): "368  QuoteEntryRejectReason",
        ("FIX.5.0", "369"): "369  LastMsgSeqNumProcessed",
        ("FIX.5.0", "371"): "371  RefTagID",
        ("FIX.5.0", "372"): "372  RefMsgType",
        ("FIX.5.0", "373"): "373  SessionRejectReason",
        ("FIX.5.0", "374"): "374  BidRequestTransType",
        ("FIX.5.0", "375"): "375  ContraBroker",
        ("FIX.5.0", "376"): "376  ComplianceID",
        ("FIX.5.0", "377"): "377  SolicitedFlag",
        ("FIX.5.0", "378"): "378  ExecRestatementReason",
        ("FIX.5.0", "379"): "379  BusinessRejectRefID",
        ("FIX.5.0", "380"): "380  BusinessRejectReason",
        ("FIX.5.0", "381"): "381  GrossTradeAmt",
        ("FIX.5.0", "382"): "382  NoContraBrokers",
        ("FIX.5.0", "383"): "383  MaxMessageSize",
        ("FIX.5.0", "384"): "384  NoMsgTypes",
        ("FIX.5.0", "385"): "385  MsgDirection",
        ("FIX.5.0", "386"): "386  NoTradingSessions",
        ("FIX.5.0", "387"): "387  TotalVolumeTraded",
        ("FIX.5.0", "388"): "388  DiscretionInst",
        ("FIX.5.0", "389"): "389  DiscretionOffsetValue",
        ("FIX.5.0", "390"): "390  BidID",
        ("FIX.5.0", "391"): "391  ClientBidID",
        ("FIX.5.0", "392"): "392  ListName",
        ("FIX.5.0", "393"): "393  TotNoRelatedSym",
        ("FIX.5.0", "394"): "394  BidType",
        ("FIX.5.0", "395"): "395  NumTickets",
        ("FIX.5.0", "396"): "396  SideValue1",
        ("FIX.5.0", "397"): "397  SideValue2",
        ("FIX.5.0", "398"): "398  NoBidDescriptors",
        ("FIX.5.0", "399"): "399  BidDescriptorType",
        ("FIX.5.0", "400"): "400  BidDescriptor",
        ("FIX.5.0", "401"): "401  SideValueInd",
        ("FIX.5.0", "402"): "402  LiquidityPctLow",
        ("FIX.5.0", "403"): "403  LiquidityPctHigh",
        ("FIX.5.0", "404"): "404  LiquidityValue",
        ("FIX.5.0", "405"): "405  EFPTrackingError",
        ("FIX.5.0", "406"): "406  FairValue",
        ("FIX.5.0", "407"): "407  OutsideIndexPct",
        ("FIX.5.0", "408"): "408  ValueOfFutures",
        ("FIX.5.0", "409"): "409  LiquidityIndType",
        ("FIX.5.0", "410"): "410  WtAverageLiquidity",
        ("FIX.5.0", "411"): "411  ExchangeForPhysical",
        ("FIX.5.0", "412"): "412  OutMainCntryUIndex",
        ("FIX.5.0", "413"): "413  CrossPercent",
        ("FIX.5.0", "414"): "414  ProgRptReqs",
        ("FIX.5.0", "415"): "415  ProgPeriodInterval",
        ("FIX.5.0", "416"): "416  IncTaxInd",
        ("FIX.5.0", "417"): "417  NumBidders",
        ("FIX.5.0", "418"): "418  BidTradeType",
        ("FIX.5.0", "419"): "419  BasisPxType",
        ("FIX.5.0", "420"): "420  NoBidComponents",
        ("FIX.5.0", "421"): "421  Country",
        ("FIX.5.0", "422"): "422  TotNoStrikes",
        ("FIX.5.0", "423"): "423  PriceType",
        ("FIX.5.0", "424"): "424  DayOrderQty",
        ("FIX.5.0", "425"): "425  DayCumQty",
        ("FIX.5.0", "426"): "426  DayAvgPx",
        ("FIX.5.0", "427"): "427  GTBookingInst",
        ("FIX.5.0", "428"): "428  NoStrikes",
        ("FIX.5.0", "429"): "429  ListStatusType",
        ("FIX.5.0", "430"): "430  NetGrossInd",
        ("FIX.5.0", "431"): "431  ListOrderStatus",
        ("FIX.5.0", "432"): "432  ExpireDate",
        ("FIX.5.0", "433"): "433  ListExecInstType",
        ("FIX.5.0", "434"): "434  CxlRejResponseTo",
        ("FIX.5.0", "435"): "435  UnderlyingCouponRate",
        ("FIX.5.0", "436"): "436  UnderlyingContractMultiplier",
        ("FIX.5.0", "437"): "437  ContraTradeQty",
        ("FIX.5.0", "438"): "438  ContraTradeTime",
        ("FIX.5.0", "441"): "441  LiquidityNumSecurities",
        ("FIX.5.0", "442"): "442  MultiLegReportingType",
        ("FIX.5.0", "443"): "443  StrikeTime",
        ("FIX.5.0", "444"): "444  ListStatusText",
        ("FIX.5.0", "445"): "445  EncodedListStatusTextLen",
        ("FIX.5.0", "446"): "446  EncodedListStatusText",
        ("FIX.5.0", "447"): "447  PartyIDSource",
        ("FIX.5.0", "448"): "448  PartyID",
        ("FIX.5.0", "451"): "451  NetChgPrevDay",
        ("FIX.5.0", "452"): "452  PartyRole",
        ("FIX.5.0", "453"): "453  NoPartyIDs",
        ("FIX.5.0", "454"): "454  NoSecurityAltID",
        ("FIX.5.0", "455"): "455  SecurityAltID",
        ("FIX.5.0", "456"): "456  SecurityAltIDSource",
        ("FIX.5.0", "457"): "457  NoUnderlyingSecurityAltID",
        ("FIX.5.0", "458"): "458  UnderlyingSecurityAltID",
        ("FIX.5.0", "459"): "459  UnderlyingSecurityAltIDSource",
        ("FIX.5.0", "460"): "460  Product",
        ("FIX.5.0", "461"): "461  CFICode",
        ("FIX.5.0", "462"): "462  UnderlyingProduct",
        ("FIX.5.0", "463"): "463  UnderlyingCFICode",
        ("FIX.5.0", "464"): "464  TestMessageIndicator",
        ("FIX.5.0", "466"): "466  BookingRefID",
        ("FIX.5.0", "467"): "467  IndividualAllocID",
        ("FIX.5.0", "468"): "468  RoundingDirection",
        ("FIX.5.0", "469"): "469  RoundingModulus",
        ("FIX.5.0", "470"): "470  CountryOfIssue",
        ("FIX.5.0", "471"): "471  StateOrProvinceOfIssue",
        ("FIX.5.0", "472"): "472  LocaleOfIssue",
        ("FIX.5.0", "473"): "473  NoRegistDtls",
        ("FIX.5.0", "474"): "474  MailingDtls",
        ("FIX.5.0", "475"): "475  InvestorCountryOfResidence",
        ("FIX.5.0", "476"): "476  PaymentRef",
        ("FIX.5.0", "477"): "477  DistribPaymentMethod",
        ("FIX.5.0", "478"): "478  CashDistribCurr",
        ("FIX.5.0", "479"): "479  CommCurrency",
        ("FIX.5.0", "480"): "480  CancellationRights",
        ("FIX.5.0", "481"): "481  MoneyLaunderingStatus",
        ("FIX.5.0", "482"): "482  MailingInst",
        ("FIX.5.0", "483"): "483  TransBkdTime",
        ("FIX.5.0", "484"): "484  ExecPriceType",
        ("FIX.5.0", "485"): "485  ExecPriceAdjustment",
        ("FIX.5.0", "486"): "486  DateOfBirth",
        ("FIX.5.0", "487"): "487  TradeReportTransType",
        ("FIX.5.0", "488"): "488  CardHolderName",
        ("FIX.5.0", "489"): "489  CardNumber",
        ("FIX.5.0", "490"): "490  CardExpDate",
        ("FIX.5.0", "491"): "491  CardIssNum",
        ("FIX.5.0", "492"): "492  PaymentMethod",
        ("FIX.5.0", "493"): "493  RegistAcctType",
        ("FIX.5.0", "494"): "494  Designation",
        ("FIX.5.0", "495"): "495  TaxAdvantageType",
        ("FIX.5.0", "496"): "496  RegistRejReasonText",
        ("FIX.5.0", "497"): "497  FundRenewWaiv",
        ("FIX.5.0", "498"): "498  CashDistribAgentName",
        ("FIX.5.0", "499"): "499  CashDistribAgentCode",
        ("FIX.5.0", "500"): "500  CashDistribAgentAcctNumber",
        ("FIX.5.0", "501"): "501  CashDistribPayRef",
        ("FIX.5.0", "502"): "502  CashDistribAgentAcctName",
        ("FIX.5.0", "503"): "503  CardStartDate",
        ("FIX.5.0", "504"): "504  PaymentDate",
        ("FIX.5.0", "505"): "505  PaymentRemitterID",
        ("FIX.5.0", "506"): "506  RegistStatus",
        ("FIX.5.0", "507"): "507  RegistRejReasonCode",
        ("FIX.5.0", "508"): "508  RegistRefID",
        ("FIX.5.0", "509"): "509  RegistDtls",
        ("FIX.5.0", "510"): "510  NoDistribInsts",
        ("FIX.5.0", "511"): "511  RegistEmail",
        ("FIX.5.0", "512"): "512  DistribPercentage",
        ("FIX.5.0", "513"): "513  RegistID",
        ("FIX.5.0", "514"): "514  RegistTransType",
        ("FIX.5.0", "515"): "515  ExecValuationPoint",
        ("FIX.5.0", "516"): "516  OrderPercent",
        ("FIX.5.0", "517"): "517  OwnershipType",
        ("FIX.5.0", "518"): "518  NoContAmts",
        ("FIX.5.0", "519"): "519  ContAmtType",
        ("FIX.5.0", "520"): "520  ContAmtValue",
        ("FIX.5.0", "521"): "521  ContAmtCurr",
        ("FIX.5.0", "522"): "522  OwnerType",
        ("FIX.5.0", "523"): "523  PartySubID",
        ("FIX.5.0", "524"): "524  NestedPartyID",
        ("FIX.5.0", "525"): "525  NestedPartyIDSource",
        ("FIX.5.0", "526"): "526  SecondaryClOrdID",
        ("FIX.5.0", "527"): "527  SecondaryExecID",
        ("FIX.5.0", "528"): "528  OrderCapacity",
        ("FIX.5.0", "529"): "529  OrderRestrictions",
        ("FIX.5.0", "530"): "530  MassCancelRequestType",
        ("FIX.5.0", "531"): "531  MassCancelResponse",
        ("FIX.5.0", "532"): "532  MassCancelRejectReason",
        ("FIX.5.0", "533"): "533  TotalAffectedOrders",
        ("FIX.5.0", "534"): "534  NoAffectedOrders",
        ("FIX.5.0", "535"): "535  AffectedOrderID",
        ("FIX.5.0", "536"): "536  AffectedSecondaryOrderID",
        ("FIX.5.0", "537"): "537  QuoteType",
        ("FIX.5.0", "538"): "538  NestedPartyRole",
        ("FIX.5.0", "539"): "539  NoNestedPartyIDs",
        ("FIX.5.0", "540"): "540  TotalAccruedInterestAmt",
        ("FIX.5.0", "541"): "541  MaturityDate",
        ("FIX.5.0", "542"): "542  UnderlyingMaturityDate",
        ("FIX.5.0", "543"): "543  InstrRegistry",
        ("FIX.5.0", "544"): "544  CashMargin",
        ("FIX.5.0", "545"): "545  NestedPartySubID",
        ("FIX.5.0", "546"): "546  Scope",
        ("FIX.5.0", "547"): "547  MDImplicitDelete",
        ("FIX.5.0", "548"): "548  CrossID",
        ("FIX.5.0", "549"): "549  CrossType",
        ("FIX.5.0", "550"): "550  CrossPrioritization",
        ("FIX.5.0", "551"): "551  OrigCrossID",
        ("FIX.5.0", "552"): "552  NoSides",
        ("FIX.5.0", "553"): "553  Username",
        ("FIX.5.0", "554"): "554  Password",
        ("FIX.5.0", "555"): "555  NoLegs",
        ("FIX.5.0", "556"): "556  LegCurrency",
        ("FIX.5.0", "557"): "557  TotNoSecurityTypes",
        ("FIX.5.0", "558"): "558  NoSecurityTypes",
        ("FIX.5.0", "559"): "559  SecurityListRequestType",
        ("FIX.5.0", "560"): "560  SecurityRequestResult",
        ("FIX.5.0", "561"): "561  RoundLot",
        ("FIX.5.0", "562"): "562  MinTradeVol",
        ("FIX.5.0", "563"): "563  MultiLegRptTypeReq",
        ("FIX.5.0", "564"): "564  LegPositionEffect",
        ("FIX.5.0", "565"): "565  LegCoveredOrUncovered",
        ("FIX.5.0", "566"): "566  LegPrice",
        ("FIX.5.0", "567"): "567  TradSesStatusRejReason",
        ("FIX.5.0", "568"): "568  TradeRequestID",
        ("FIX.5.0", "569"): "569  TradeRequestType",
        ("FIX.5.0", "570"): "570  PreviouslyReported",
        ("FIX.5.0", "571"): "571  TradeReportID",
        ("FIX.5.0", "572"): "572  TradeReportRefID",
        ("FIX.5.0", "573"): "573  MatchStatus",
        ("FIX.5.0", "574"): "574  MatchType",
        ("FIX.5.0", "575"): "575  OddLot",
        ("FIX.5.0", "576"): "576  NoClearingInstructions",
        ("FIX.5.0", "577"): "577  ClearingInstruction",
        ("FIX.5.0", "578"): "578  TradeInputSource",
        ("FIX.5.0", "579"): "579  TradeInputDevice",
        ("FIX.5.0", "580"): "580  NoDates",
        ("FIX.5.0", "581"): "581  AccountType",
        ("FIX.5.0", "582"): "582  CustOrderCapacity",
        ("FIX.5.0", "583"): "583  ClOrdLinkID",
        ("FIX.5.0", "584"): "584  MassStatusReqID",
        ("FIX.5.0", "585"): "585  MassStatusReqType",
        ("FIX.5.0", "586"): "586  OrigOrdModTime",
        ("FIX.5.0", "587"): "587  LegSettlType",
        ("FIX.5.0", "588"): "588  LegSettlDate",
        ("FIX.5.0", "589"): "589  DayBookingInst",
        ("FIX.5.0", "590"): "590  BookingUnit",
        ("FIX.5.0", "591"): "591  PreallocMethod",
        ("FIX.5.0", "592"): "592  UnderlyingCountryOfIssue",
        ("FIX.5.0", "593"): "593  UnderlyingStateOrProvinceOfIssue",
        ("FIX.5.0", "594"): "594  UnderlyingLocaleOfIssue",
        ("FIX.5.0", "595"): "595  UnderlyingInstrRegistry",
        ("FIX.5.0", "596"): "596  LegCountryOfIssue",
        ("FIX.5.0", "597"): "597  LegStateOrProvinceOfIssue",
        ("FIX.5.0", "598"): "598  LegLocaleOfIssue",
        ("FIX.5.0", "599"): "599  LegInstrRegistry",
        ("FIX.5.0", "600"): "600  LegSymbol",
        ("FIX.5.0", "601"): "601  LegSymbolSfx",
        ("FIX.5.0", "602"): "602  LegSecurityID",
        ("FIX.5.0", "603"): "603  LegSecurityIDSource",
        ("FIX.5.0", "604"): "604  NoLegSecurityAltID",
        ("FIX.5.0", "605"): "605  LegSecurityAltID",
        ("FIX.5.0", "606"): "606  LegSecurityAltIDSource",
        ("FIX.5.0", "607"): "607  LegProduct",
        ("FIX.5.0", "608"): "608  LegCFICode",
        ("FIX.5.0", "609"): "609  LegSecurityType",
        ("FIX.5.0", "610"): "610  LegMaturityMonthYear",
        ("FIX.5.0", "611"): "611  LegMaturityDate",
        ("FIX.5.0", "612"): "612  LegStrikePrice",
        ("FIX.5.0", "613"): "613  LegOptAttribute",
        ("FIX.5.0", "614"): "614  LegContractMultiplier",
        ("FIX.5.0", "615"): "615  LegCouponRate",
        ("FIX.5.0", "616"): "616  LegSecurityExchange",
        ("FIX.5.0", "617"): "617  LegIssuer",
        ("FIX.5.0", "618"): "618  EncodedLegIssuerLen",
        ("FIX.5.0", "619"): "619  EncodedLegIssuer",
        ("FIX.5.0", "620"): "620  LegSecurityDesc",
        ("FIX.5.0", "621"): "621  EncodedLegSecurityDescLen",
        ("FIX.5.0", "622"): "622  EncodedLegSecurityDesc",
        ("FIX.5.0", "623"): "623  LegRatioQty",
        ("FIX.5.0", "624"): "624  LegSide",
        ("FIX.5.0", "625"): "625  TradingSessionSubID",
        ("FIX.5.0", "626"): "626  AllocType",
        ("FIX.5.0", "627"): "627  NoHops",
        ("FIX.5.0", "628"): "628  HopCompID",
        ("FIX.5.0", "629"): "629  HopSendingTime",
        ("FIX.5.0", "630"): "630  HopRefID",
        ("FIX.5.0", "631"): "631  MidPx",
        ("FIX.5.0", "632"): "632  BidYield",
        ("FIX.5.0", "633"): "633  MidYield",
        ("FIX.5.0", "634"): "634  OfferYield",
        ("FIX.5.0", "635"): "635  ClearingFeeIndicator",
        ("FIX.5.0", "636"): "636  WorkingIndicator",
        ("FIX.5.0", "637"): "637  LegLastPx",
        ("FIX.5.0", "638"): "638  PriorityIndicator",
        ("FIX.5.0", "639"): "639  PriceImprovement",
        ("FIX.5.0", "640"): "640  Price2",
        ("FIX.5.0", "641"): "641  LastForwardPoints2",
        ("FIX.5.0", "642"): "642  BidForwardPoints2",
        ("FIX.5.0", "643"): "643  OfferForwardPoints2",
        ("FIX.5.0", "644"): "644  RFQReqID",
        ("FIX.5.0", "645"): "645  MktBidPx",
        ("FIX.5.0", "646"): "646  MktOfferPx",
        ("FIX.5.0", "647"): "647  MinBidSize",
        ("FIX.5.0", "648"): "648  MinOfferSize",
        ("FIX.5.0", "649"): "649  QuoteStatusReqID",
        ("FIX.5.0", "650"): "650  LegalConfirm",
        ("FIX.5.0", "651"): "651  UnderlyingLastPx",
        ("FIX.5.0", "652"): "652  UnderlyingLastQty",
        ("FIX.5.0", "654"): "654  LegRefID",
        ("FIX.5.0", "655"): "655  ContraLegRefID",
        ("FIX.5.0", "656"): "656  SettlCurrBidFxRate",
        ("FIX.5.0", "657"): "657  SettlCurrOfferFxRate",
        ("FIX.5.0", "658"): "658  QuoteRequestRejectReason",
        ("FIX.5.0", "659"): "659  SideComplianceID",
        ("FIX.5.0", "660"): "660  AcctIDSource",
        ("FIX.5.0", "661"): "661  AllocAcctIDSource",
        ("FIX.5.0", "662"): "662  BenchmarkPrice",
        ("FIX.5.0", "663"): "663  BenchmarkPriceType",
        ("FIX.5.0", "664"): "664  ConfirmID",
        ("FIX.5.0", "665"): "665  ConfirmStatus",
        ("FIX.5.0", "666"): "666  ConfirmTransType",
        ("FIX.5.0", "667"): "667  ContractSettlMonth",
        ("FIX.5.0", "668"): "668  DeliveryForm",
        ("FIX.5.0", "669"): "669  LastParPx",
        ("FIX.5.0", "670"): "670  NoLegAllocs",
        ("FIX.5.0", "671"): "671  LegAllocAccount",
        ("FIX.5.0", "672"): "672  LegIndividualAllocID",
        ("FIX.5.0", "673"): "673  LegAllocQty",
        ("FIX.5.0", "674"): "674  LegAllocAcctIDSource",
        ("FIX.5.0", "675"): "675  LegSettlCurrency",
        ("FIX.5.0", "676"): "676  LegBenchmarkCurveCurrency",
        ("FIX.5.0", "677"): "677  LegBenchmarkCurveName",
        ("FIX.5.0", "678"): "678  LegBenchmarkCurvePoint",
        ("FIX.5.0", "679"): "679  LegBenchmarkPrice",
        ("FIX.5.0", "680"): "680  LegBenchmarkPriceType",
        ("FIX.5.0", "681"): "681  LegBidPx",
        ("FIX.5.0", "682"): "682  LegIOIQty",
        ("FIX.5.0", "683"): "683  NoLegStipulations",
        ("FIX.5.0", "684"): "684  LegOfferPx",
        ("FIX.5.0", "685"): "685  LegOrderQty",
        ("FIX.5.0", "686"): "686  LegPriceType",
        ("FIX.5.0", "687"): "687  LegQty",
        ("FIX.5.0", "688"): "688  LegStipulationType",
        ("FIX.5.0", "689"): "689  LegStipulationValue",
        ("FIX.5.0", "690"): "690  LegSwapType",
        ("FIX.5.0", "691"): "691  Pool",
        ("FIX.5.0", "692"): "692  QuotePriceType",
        ("FIX.5.0", "693"): "693  QuoteRespID",
        ("FIX.5.0", "694"): "694  QuoteRespType",
        ("FIX.5.0", "695"): "695  QuoteQualifier",
        ("FIX.5.0", "696"): "696  YieldRedemptionDate",
        ("FIX.5.0", "697"): "697  YieldRedemptionPrice",
        ("FIX.5.0", "698"): "698  YieldRedemptionPriceType",
        ("FIX.5.0", "699"): "699  BenchmarkSecurityID",
        ("FIX.5.0", "700"): "700  ReversalIndicator",
        ("FIX.5.0", "701"): "701  YieldCalcDate",
        ("FIX.5.0", "702"): "702  NoPositions",
        ("FIX.5.0", "703"): "703  PosType",
        ("FIX.5.0", "704"): "704  LongQty",
        ("FIX.5.0", "705"): "705  ShortQty",
        ("FIX.5.0", "706"): "706  PosQtyStatus",
        ("FIX.5.0", "707"): "707  PosAmtType",
        ("FIX.5.0", "708"): "708  PosAmt",
        ("FIX.5.0", "709"): "709  PosTransType",
        ("FIX.5.0", "710"): "710  PosReqID",
        ("FIX.5.0", "711"): "711  NoUnderlyings",
        ("FIX.5.0", "712"): "712  PosMaintAction",
        ("FIX.5.0", "713"): "713  OrigPosReqRefID",
        ("FIX.5.0", "714"): "714  PosMaintRptRefID",
        ("FIX.5.0", "715"): "715  ClearingBusinessDate",
        ("FIX.5.0", "716"): "716  SettlSessID",
        ("FIX.5.0", "717"): "717  SettlSessSubID",
        ("FIX.5.0", "718"): "718  AdjustmentType",
        ("FIX.5.0", "719"): "719  ContraryInstructionIndicator",
        ("FIX.5.0", "720"): "720  PriorSpreadIndicator",
        ("FIX.5.0", "721"): "721  PosMaintRptID",
        ("FIX.5.0", "722"): "722  PosMaintStatus",
        ("FIX.5.0", "723"): "723  PosMaintResult",
        ("FIX.5.0", "724"): "724  PosReqType",
        ("FIX.5.0", "725"): "725  ResponseTransportType",
        ("FIX.5.0", "726"): "726  ResponseDestination",
        ("FIX.5.0", "727"): "727  TotalNumPosReports",
        ("FIX.5.0", "728"): "728  PosReqResult",
        ("FIX.5.0", "729"): "729  PosReqStatus",
        ("FIX.5.0", "730"): "730  SettlPrice",
        ("FIX.5.0", "731"): "731  SettlPriceType",
        ("FIX.5.0", "732"): "732  UnderlyingSettlPrice",
        ("FIX.5.0", "733"): "733  UnderlyingSettlPriceType",
        ("FIX.5.0", "734"): "734  PriorSettlPrice",
        ("FIX.5.0", "735"): "735  NoQuoteQualifiers",
        ("FIX.5.0", "736"): "736  AllocSettlCurrency",
        ("FIX.5.0", "737"): "737  AllocSettlCurrAmt",
        ("FIX.5.0", "738"): "738  InterestAtMaturity",
        ("FIX.5.0", "739"): "739  LegDatedDate",
        ("FIX.5.0", "740"): "740  LegPool",
        ("FIX.5.0", "741"): "741  AllocInterestAtMaturity",
        ("FIX.5.0", "742"): "742  AllocAccruedInterestAmt",
        ("FIX.5.0", "743"): "743  DeliveryDate",
        ("FIX.5.0", "744"): "744  AssignmentMethod",
        ("FIX.5.0", "745"): "745  AssignmentUnit",
        ("FIX.5.0", "746"): "746  OpenInterest",
        ("FIX.5.0", "747"): "747  ExerciseMethod",
        ("FIX.5.0", "748"): "748  TotNumTradeReports",
        ("FIX.5.0", "749"): "749  TradeRequestResult",
        ("FIX.5.0", "750"): "750  TradeRequestStatus",
        ("FIX.5.0", "751"): "751  TradeReportRejectReason",
        ("FIX.5.0", "752"): "752  SideMultiLegReportingType",
        ("FIX.5.0", "753"): "753  NoPosAmt",
        ("FIX.5.0", "754"): "754  AutoAcceptIndicator",
        ("FIX.5.0", "755"): "755  AllocReportID",
        ("FIX.5.0", "756"): "756  NoNested2PartyIDs",
        ("FIX.5.0", "757"): "757  Nested2PartyID",
        ("FIX.5.0", "758"): "758  Nested2PartyIDSource",
        ("FIX.5.0", "759"): "759  Nested2PartyRole",
        ("FIX.5.0", "760"): "760  Nested2PartySubID",
        ("FIX.5.0", "761"): "761  BenchmarkSecurityIDSource",
        ("FIX.5.0", "762"): "762  SecuritySubType",
        ("FIX.5.0", "763"): "763  UnderlyingSecuritySubType",
        ("FIX.5.0", "764"): "764  LegSecuritySubType",
        ("FIX.5.0", "765"): "765  AllowableOneSidednessPct",
        ("FIX.5.0", "766"): "766  AllowableOneSidednessValue",
        ("FIX.5.0", "767"): "767  AllowableOneSidednessCurr",
        ("FIX.5.0", "768"): "768  NoTrdRegTimestamps",
        ("FIX.5.0", "769"): "769  TrdRegTimestamp",
        ("FIX.5.0", "770"): "770  TrdRegTimestampType",
        ("FIX.5.0", "771"): "771  TrdRegTimestampOrigin",
        ("FIX.5.0", "772"): "772  ConfirmRefID",
        ("FIX.5.0", "773"): "773  ConfirmType",
        ("FIX.5.0", "774"): "774  ConfirmRejReason",
        ("FIX.5.0", "775"): "775  BookingType",
        ("FIX.5.0", "776"): "776  IndividualAllocRejCode",
        ("FIX.5.0", "777"): "777  SettlInstMsgID",
        ("FIX.5.0", "778"): "778  NoSettlInst",
        ("FIX.5.0", "779"): "779  LastUpdateTime",
        ("FIX.5.0", "780"): "780  AllocSettlInstType",
        ("FIX.5.0", "781"): "781  NoSettlPartyIDs",
        ("FIX.5.0", "782"): "782  SettlPartyID",
        ("FIX.5.0", "783"): "783  SettlPartyIDSource",
        ("FIX.5.0", "784"): "784  SettlPartyRole",
        ("FIX.5.0", "785"): "785  SettlPartySubID",
        ("FIX.5.0", "786"): "786  SettlPartySubIDType",
        ("FIX.5.0", "787"): "787  DlvyInstType",
        ("FIX.5.0", "788"): "788  TerminationType",
        ("FIX.5.0", "789"): "789  NextExpectedMsgSeqNum",
        ("FIX.5.0", "790"): "790  OrdStatusReqID",
        ("FIX.5.0", "791"): "791  SettlInstReqID",
        ("FIX.5.0", "792"): "792  SettlInstReqRejCode",
        ("FIX.5.0", "793"): "793  SecondaryAllocID",
        ("FIX.5.0", "794"): "794  AllocReportType",
        ("FIX.5.0", "795"): "795  AllocReportRefID",
        ("FIX.5.0", "796"): "796  AllocCancReplaceReason",
        ("FIX.5.0", "797"): "797  CopyMsgIndicator",
        ("FIX.5.0", "798"): "798  AllocAccountType",
        ("FIX.5.0", "799"): "799  OrderAvgPx",
        ("FIX.5.0", "800"): "800  OrderBookingQty",
        ("FIX.5.0", "801"): "801  NoSettlPartySubIDs",
        ("FIX.5.0", "802"): "802  NoPartySubIDs",
        ("FIX.5.0", "803"): "803  PartySubIDType",
        ("FIX.5.0", "804"): "804  NoNestedPartySubIDs",
        ("FIX.5.0", "805"): "805  NestedPartySubIDType",
        ("FIX.5.0", "806"): "806  NoNested2PartySubIDs",
        ("FIX.5.0", "807"): "807  Nested2PartySubIDType",
        ("FIX.5.0", "808"): "808  AllocIntermedReqType",
        ("FIX.5.0", "809"): "809  NoUsernames",
        ("FIX.5.0", "810"): "810  UnderlyingPx",
        ("FIX.5.0", "811"): "811  PriceDelta",
        ("FIX.5.0", "812"): "812  ApplQueueMax",
        ("FIX.5.0", "813"): "813  ApplQueueDepth",
        ("FIX.5.0", "814"): "814  ApplQueueResolution",
        ("FIX.5.0", "815"): "815  ApplQueueAction",
        ("FIX.5.0", "816"): "816  NoAltMDSource",
        ("FIX.5.0", "817"): "817  AltMDSourceID",
        ("FIX.5.0", "818"): "818  SecondaryTradeReportID",
        ("FIX.5.0", "819"): "819  AvgPxIndicator",
        ("FIX.5.0", "820"): "820  TradeLinkID",
        ("FIX.5.0", "821"): "821  OrderInputDevice",
        ("FIX.5.0", "822"): "822  UnderlyingTradingSessionID",
        ("FIX.5.0", "823"): "823  UnderlyingTradingSessionSubID",
        ("FIX.5.0", "824"): "824  TradeLegRefID",
        ("FIX.5.0", "825"): "825  ExchangeRule",
        ("FIX.5.0", "826"): "826  TradeAllocIndicator",
        ("FIX.5.0", "827"): "827  ExpirationCycle",
        ("FIX.5.0", "828"): "828  TrdType",
        ("FIX.5.0", "829"): "829  TrdSubType",
        ("FIX.5.0", "830"): "830  TransferReason",
        ("FIX.5.0", "832"): "832  TotNumAssignmentReports",
        ("FIX.5.0", "833"): "833  AsgnRptID",
        ("FIX.5.0", "834"): "834  ThresholdAmount",
        ("FIX.5.0", "835"): "835  PegMoveType",
        ("FIX.5.0", "836"): "836  PegOffsetType",
        ("FIX.5.0", "837"): "837  PegLimitType",
        ("FIX.5.0", "838"): "838  PegRoundDirection",
        ("FIX.5.0", "839"): "839  PeggedPrice",
        ("FIX.5.0", "840"): "840  PegScope",
        ("FIX.5.0", "841"): "841  DiscretionMoveType",
        ("FIX.5.0", "842"): "842  DiscretionOffsetType",
        ("FIX.5.0", "843"): "843  DiscretionLimitType",
        ("FIX.5.0", "844"): "844  DiscretionRoundDirection",
        ("FIX.5.0", "845"): "845  DiscretionPrice",
        ("FIX.5.0", "846"): "846  DiscretionScope",
        ("FIX.5.0", "847"): "847  TargetStrategy",
        ("FIX.5.0", "848"): "848  TargetStrategyParameters",
        ("FIX.5.0", "849"): "849  ParticipationRate",
        ("FIX.5.0", "850"): "850  TargetStrategyPerformance",
        ("FIX.5.0", "851"): "851  LastLiquidityInd",
        ("FIX.5.0", "852"): "852  PublishTrdIndicator",
        ("FIX.5.0", "853"): "853  ShortSaleReason",
        ("FIX.5.0", "854"): "854  QtyType",
        ("FIX.5.0", "855"): "855  SecondaryTrdType",
        ("FIX.5.0", "856"): "856  TradeReportType",
        ("FIX.5.0", "857"): "857  AllocNoOrdersType",
        ("FIX.5.0", "858"): "858  SharedCommission",
        ("FIX.5.0", "859"): "859  ConfirmReqID",
        ("FIX.5.0", "860"): "860  AvgParPx",
        ("FIX.5.0", "861"): "861  ReportedPx",
        ("FIX.5.0", "862"): "862  NoCapacities",
        ("FIX.5.0", "863"): "863  OrderCapacityQty",
        ("FIX.5.0", "864"): "864  NoEvents",
        ("FIX.5.0", "865"): "865  EventType",
        ("FIX.5.0", "866"): "866  EventDate",
        ("FIX.5.0", "867"): "867  EventPx",
        ("FIX.5.0", "868"): "868  EventText",
        ("FIX.5.0", "869"): "869  PctAtRisk",
        ("FIX.5.0", "870"): "870  NoInstrAttrib",
        ("FIX.5.0", "871"): "871  InstrAttribType",
        ("FIX.5.0", "872"): "872  InstrAttribValue",
        ("FIX.5.0", "873"): "873  DatedDate",
        ("FIX.5.0", "874"): "874  InterestAccrualDate",
        ("FIX.5.0", "875"): "875  CPProgram",
        ("FIX.5.0", "876"): "876  CPRegType",
        ("FIX.5.0", "877"): "877  UnderlyingCPProgram",
        ("FIX.5.0", "878"): "878  UnderlyingCPRegType",
        ("FIX.5.0", "879"): "879  UnderlyingQty",
        ("FIX.5.0", "880"): "880  TrdMatchID",
        ("FIX.5.0", "881"): "881  SecondaryTradeReportRefID",
        ("FIX.5.0", "882"): "882  UnderlyingDirtyPrice",
        ("FIX.5.0", "883"): "883  UnderlyingEndPrice",
        ("FIX.5.0", "884"): "884  UnderlyingStartValue",
        ("FIX.5.0", "885"): "885  UnderlyingCurrentValue",
        ("FIX.5.0", "886"): "886  UnderlyingEndValue",
        ("FIX.5.0", "887"): "887  NoUnderlyingStips",
        ("FIX.5.0", "888"): "888  UnderlyingStipType",
        ("FIX.5.0", "889"): "889  UnderlyingStipValue",
        ("FIX.5.0", "890"): "890  MaturityNetMoney",
        ("FIX.5.0", "891"): "891  MiscFeeBasis",
        ("FIX.5.0", "892"): "892  TotNoAllocs",
        ("FIX.5.0", "893"): "893  LastFragment",
        ("FIX.5.0", "894"): "894  CollReqID",
        ("FIX.5.0", "895"): "895  CollAsgnReason",
        ("FIX.5.0", "896"): "896  CollInquiryQualifier",
        ("FIX.5.0", "897"): "897  NoTrades",
        ("FIX.5.0", "898"): "898  MarginRatio",
        ("FIX.5.0", "899"): "899  MarginExcess",
        ("FIX.5.0", "900"): "900  TotalNetValue",
        ("FIX.5.0", "901"): "901  CashOutstanding",
        ("FIX.5.0", "902"): "902  CollAsgnID",
        ("FIX.5.0", "903"): "903  CollAsgnTransType",
        ("FIX.5.0", "904"): "904  CollRespID",
        ("FIX.5.0", "905"): "905  CollAsgnRespType",
        ("FIX.5.0", "906"): "906  CollAsgnRejectReason",
        ("FIX.5.0", "907"): "907  CollAsgnRefID",
        ("FIX.5.0", "908"): "908  CollRptID",
        ("FIX.5.0", "909"): "909  CollInquiryID",
        ("FIX.5.0", "910"): "910  CollStatus",
        ("FIX.5.0", "911"): "911  TotNumReports",
        ("FIX.5.0", "912"): "912  LastRptRequested",
        ("FIX.5.0", "913"): "913  AgreementDesc",
        ("FIX.5.0", "914"): "914  AgreementID",
        ("FIX.5.0", "915"): "915  AgreementDate",
        ("FIX.5.0", "916"): "916  StartDate",
        ("FIX.5.0", "917"): "917  EndDate",
        ("FIX.5.0", "918"): "918  AgreementCurrency",
        ("FIX.5.0", "919"): "919  DeliveryType",
        ("FIX.5.0", "920"): "920  EndAccruedInterestAmt",
        ("FIX.5.0", "921"): "921  StartCash",
        ("FIX.5.0", "922"): "922  EndCash",
        ("FIX.5.0", "923"): "923  UserRequestID",
        ("FIX.5.0", "924"): "924  UserRequestType",
        ("FIX.5.0", "925"): "925  NewPassword",
        ("FIX.5.0", "926"): "926  UserStatus",
        ("FIX.5.0", "927"): "927  UserStatusText",
        ("FIX.5.0", "928"): "928  StatusValue",
        ("FIX.5.0", "929"): "929  StatusText",
        ("FIX.5.0", "930"): "930  RefCompID",
        ("FIX.5.0", "931"): "931  RefSubID",
        ("FIX.5.0", "932"): "932  NetworkResponseID",
        ("FIX.5.0", "933"): "933  NetworkRequestID",
        ("FIX.5.0", "934"): "934  LastNetworkResponseID",
        ("FIX.5.0", "935"): "935  NetworkRequestType",
        ("FIX.5.0", "936"): "936  NoCompIDs",
        ("FIX.5.0", "937"): "937  NetworkStatusResponseType",
        ("FIX.5.0", "938"): "938  NoCollInquiryQualifier",
        ("FIX.5.0", "939"): "939  TrdRptStatus",
        ("FIX.5.0", "940"): "940  AffirmStatus",
        ("FIX.5.0", "941"): "941  UnderlyingStrikeCurrency",
        ("FIX.5.0", "942"): "942  LegStrikeCurrency",
        ("FIX.5.0", "943"): "943  TimeBracket",
        ("FIX.5.0", "944"): "944  CollAction",
        ("FIX.5.0", "945"): "945  CollInquiryStatus",
        ("FIX.5.0", "946"): "946  CollInquiryResult",
        ("FIX.5.0", "947"): "947  StrikeCurrency",
        ("FIX.5.0", "948"): "948  NoNested3PartyIDs",
        ("FIX.5.0", "949"): "949  Nested3PartyID",
        ("FIX.5.0", "950"): "950  Nested3PartyIDSource",
        ("FIX.5.0", "951"): "951  Nested3PartyRole",
        ("FIX.5.0", "952"): "952  NoNested3PartySubIDs",
        ("FIX.5.0", "953"): "953  Nested3PartySubID",
        ("FIX.5.0", "954"): "954  Nested3PartySubIDType",
        ("FIX.5.0", "955"): "955  LegContractSettlMonth",
        ("FIX.5.0", "956"): "956  LegInterestAccrualDate",
        ("FIX.5.0", "957"): "957  NoStrategyParameters",
        ("FIX.5.0", "958"): "958  StrategyParameterName",
        ("FIX.5.0", "959"): "959  StrategyParameterType",
        ("FIX.5.0", "960"): "960  StrategyParameterValue",
        ("FIX.5.0", "961"): "961  HostCrossID",
        ("FIX.5.0", "962"): "962  SideTimeInForce",
        ("FIX.5.0", "963"): "963  MDReportID",
        ("FIX.5.0", "964"): "964  SecurityReportID",
        ("FIX.5.0", "965"): "965  SecurityStatus",
        ("FIX.5.0", "966"): "966  SettleOnOpenFlag",
        ("FIX.5.0", "967"): "967  StrikeMultiplier",
        ("FIX.5.0", "968"): "968  StrikeValue",
        ("FIX.5.0", "969"): "969  MinPriceIncrement",
        ("FIX.5.0", "970"): "970  PositionLimit",
        ("FIX.5.0", "971"): "971  NTPositionLimit",
        ("FIX.5.0", "972"): "972  UnderlyingAllocationPercent",
        ("FIX.5.0", "973"): "973  UnderlyingCashAmount",
        ("FIX.5.0", "974"): "974  UnderlyingCashType",
        ("FIX.5.0", "975"): "975  UnderlyingSettlementType",
        ("FIX.5.0", "976"): "976  QuantityDate",
        ("FIX.5.0", "977"): "977  ContIntRptID",
        ("FIX.5.0", "978"): "978  LateIndicator",
        ("FIX.5.0", "979"): "979  InputSource",
        ("FIX.5.0", "980"): "980  SecurityUpdateAction",
        ("FIX.5.0", "981"): "981  NoExpiration",
        ("FIX.5.0", "982"): "982  ExpirationQtyType",
        ("FIX.5.0", "983"): "983  ExpQty",
        ("FIX.5.0", "984"): "984  NoUnderlyingAmounts",
        ("FIX.5.0", "985"): "985  UnderlyingPayAmount",
        ("FIX.5.0", "986"): "986  UnderlyingCollectAmount",
        ("FIX.5.0", "987"): "987  UnderlyingSettlementDate",
        ("FIX.5.0", "988"): "988  UnderlyingSettlementStatus",
        ("FIX.5.0", "989"): "989  SecondaryIndividualAllocID",
        ("FIX.5.0", "990"): "990  LegReportID",
        ("FIX.5.0", "991"): "991  RndPx",
        ("FIX.5.0", "992"): "992  IndividualAllocType",
        ("FIX.5.0", "993"): "993  AllocCustomerCapacity",
        ("FIX.5.0", "994"): "994  TierCode",
        ("FIX.5.0", "996"): "996  UnitOfMeasure",
        ("FIX.5.0", "997"): "997  TimeUnit",
        ("FIX.5.0", "998"): "998  UnderlyingUnitOfMeasure",
        ("FIX.5.0", "999"): "999  LegUnitOfMeasure",
        ("FIX.5.0", "1000"): "1000 UnderlyingTimeUnit",
        ("FIX.5.0", "1001"): "1001 LegTimeUnit",
        ("FIX.5.0", "1002"): "1002 AllocMethod",
        ("FIX.5.0", "1003"): "1003 TradeID",
        ("FIX.5.0", "1005"): "1005 SideTradeReportID",
        ("FIX.5.0", "1006"): "1006 SideFillStationCd",
        ("FIX.5.0", "1007"): "1007 SideReasonCd",
        ("FIX.5.0", "1008"): "1008 SideTrdSubTyp",
        ("FIX.5.0", "1009"): "1009 SideLastQty",
        ("FIX.5.0", "1011"): "1011 MessageEventSource",
        ("FIX.5.0", "1012"): "1012 SideTrdRegTimestamp",
        ("FIX.5.0", "1013"): "1013 SideTrdRegTimestampType",
        ("FIX.5.0", "1014"): "1014 SideTrdRegTimestampSrc",
        ("FIX.5.0", "1015"): "1015 AsOfIndicator",
        ("FIX.5.0", "1016"): "1016 NoSideTrdRegTS",
        ("FIX.5.0", "1017"): "1017 LegOptionRatio",
        ("FIX.5.0", "1018"): "1018 NoInstrumentParties",
        ("FIX.5.0", "1019"): "1019 InstrumentPartyID",
        ("FIX.5.0", "1020"): "1020 TradeVolume",
        ("FIX.5.0", "1021"): "1021 MDBookType",
        ("FIX.5.0", "1022"): "1022 MDFeedType",
        ("FIX.5.0", "1023"): "1023 MDPriceLevel",
        ("FIX.5.0", "1024"): "1024 MDOriginType",
        ("FIX.5.0", "1025"): "1025 FirstPx",
        ("FIX.5.0", "1026"): "1026 MDEntrySpotRate",
        ("FIX.5.0", "1027"): "1027 MDEntryForwardPoints",
        ("FIX.5.0", "1028"): "1028 ManualOrderIndicator",
        ("FIX.5.0", "1029"): "1029 CustDirectedOrder",
        ("FIX.5.0", "1030"): "1030 ReceivedDeptID",
        ("FIX.5.0", "1031"): "1031 CustOrderHandlingInst",
        ("FIX.5.0", "1032"): "1032 OrderHandlingInstSource",
        ("FIX.5.0", "1033"): "1033 DeskType",
        ("FIX.5.0", "1034"): "1034 DeskTypeSource",
        ("FIX.5.0", "1035"): "1035 DeskOrderHandlingInst",
        ("FIX.5.0", "1036"): "1036 ExecAckStatus",
        ("FIX.5.0", "1037"): "1037 UnderlyingDeliveryAmount",
        ("FIX.5.0", "1038"): "1038 UnderlyingCapValue",
        ("FIX.5.0", "1039"): "1039 UnderlyingSettlMethod",
        ("FIX.5.0", "1040"): "1040 SecondaryTradeID",
        ("FIX.5.0", "1041"): "1041 FirmTradeID",
        ("FIX.5.0", "1042"): "1042 SecondaryFirmTradeID",
        ("FIX.5.0", "1043"): "1043 CollApplType",
        ("FIX.5.0", "1044"): "1044 UnderlyingAdjustedQuantity",
        ("FIX.5.0", "1045"): "1045 UnderlyingFXRate",
        ("FIX.5.0", "1046"): "1046 UnderlyingFXRateCalc",
        ("FIX.5.0", "1047"): "1047 AllocPositionEffect",
        ("FIX.5.0", "1048"): "1048 DealingCapacity",
        ("FIX.5.0", "1049"): "1049 InstrmtAssignmentMethod",
        ("FIX.5.0", "1050"): "1050 InstrumentPartyIDSource",
        ("FIX.5.0", "1051"): "1051 InstrumentPartyRole",
        ("FIX.5.0", "1052"): "1052 NoInstrumentPartySubIDs",
        ("FIX.5.0", "1053"): "1053 InstrumentPartySubID",
        ("FIX.5.0", "1054"): "1054 InstrumentPartySubIDType",
        ("FIX.5.0", "1055"): "1055 PositionCurrency",
        ("FIX.5.0", "1056"): "1056 CalculatedCcyLastQty",
        ("FIX.5.0", "1057"): "1057 AggressorIndicator",
        ("FIX.5.0", "1058"): "1058 NoUndlyInstrumentParties",
        ("FIX.5.0", "1059"): "1059 UnderlyingInstrumentPartyID",
        ("FIX.5.0", "1060"): "1060 UnderlyingInstrumentPartyIDSource",
        ("FIX.5.0", "1061"): "1061 UnderlyingInstrumentPartyRole",
        ("FIX.5.0", "1062"): "1062 NoUndlyInstrumentPartySubIDs",
        ("FIX.5.0", "1063"): "1063 UnderlyingInstrumentPartySubID",
        ("FIX.5.0", "1064"): "1064 UnderlyingInstrumentPartySubIDType",
        ("FIX.5.0", "1065"): "1065 BidSwapPoints",
        ("FIX.5.0", "1066"): "1066 OfferSwapPoints",
        ("FIX.5.0", "1067"): "1067 LegBidForwardPoints",
        ("FIX.5.0", "1068"): "1068 LegOfferForwardPoints",
        ("FIX.5.0", "1069"): "1069 SwapPoints",
        ("FIX.5.0", "1070"): "1070 MDQuoteType",
        ("FIX.5.0", "1071"): "1071 LastSwapPoints",
        ("FIX.5.0", "1072"): "1072 SideGrossTradeAmt",
        ("FIX.5.0", "1073"): "1073 LegLastForwardPoints",
        ("FIX.5.0", "1074"): "1074 LegCalculatedCcyLastQty",
        ("FIX.5.0", "1075"): "1075 LegGrossTradeAmt",
        ("FIX.5.0", "1079"): "1079 MaturityTime",
        ("FIX.5.0", "1080"): "1080 RefOrderID",
        ("FIX.5.0", "1081"): "1081 RefOrderIDSource",
        ("FIX.5.0", "1082"): "1082 SecondaryDisplayQty",
        ("FIX.5.0", "1083"): "1083 DisplayWhen",
        ("FIX.5.0", "1084"): "1084 DisplayMethod",
        ("FIX.5.0", "1085"): "1085 DisplayLowQty",
        ("FIX.5.0", "1086"): "1086 DisplayHighQty",
        ("FIX.5.0", "1087"): "1087 DisplayMinIncr",
        ("FIX.5.0", "1088"): "1088 RefreshQty",
        ("FIX.5.0", "1089"): "1089 MatchIncrement",
        ("FIX.5.0", "1090"): "1090 MaxPriceLevels",
        ("FIX.5.0", "1091"): "1091 PreTradeAnonymity",
        ("FIX.5.0", "1092"): "1092 PriceProtectionScope",
        ("FIX.5.0", "1093"): "1093 LotType",
        ("FIX.5.0", "1094"): "1094 PegPriceType",
        ("FIX.5.0", "1095"): "1095 PeggedRefPrice",
        ("FIX.5.0", "1096"): "1096 PegSecurityIDSource",
        ("FIX.5.0", "1097"): "1097 PegSecurityID",
        ("FIX.5.0", "1098"): "1098 PegSymbol",
        ("FIX.5.0", "1099"): "1099 PegSecurityDesc",
        ("FIX.5.0", "1100"): "1100 TriggerType",
        ("FIX.5.0", "1101"): "1101 TriggerAction",
        ("FIX.5.0", "1102"): "1102 TriggerPrice",
        ("FIX.5.0", "1103"): "1103 TriggerSymbol",
        ("FIX.5.0", "1104"): "1104 TriggerSecurityID",
        ("FIX.5.0", "1105"): "1105 TriggerSecurityIDSource",
        ("FIX.5.0", "1106"): "1106 TriggerSecurityDesc",
        ("FIX.5.0", "1107"): "1107 TriggerPriceType",
        ("FIX.5.0", "1108"): "1108 TriggerPriceTypeScope",
        ("FIX.5.0", "1109"): "1109 TriggerPriceDirection",
        ("FIX.5.0", "1110"): "1110 TriggerNewPrice",
        ("FIX.5.0", "1111"): "1111 TriggerOrderType",
        ("FIX.5.0", "1112"): "1112 TriggerNewQty",
        ("FIX.5.0", "1113"): "1113 TriggerTradingSessionID",
        ("FIX.5.0", "1114"): "1114 TriggerTradingSessionSubID",
        ("FIX.5.0", "1115"): "1115 OrderCategory",
        ("FIX.5.0", "1116"): "1116 NoRootPartyIDs",
        ("FIX.5.0", "1117"): "1117 RootPartyID",
        ("FIX.5.0", "1118"): "1118 RootPartyIDSource",
        ("FIX.5.0", "1119"): "1119 RootPartyRole",
        ("FIX.5.0", "1120"): "1120 NoRootPartySubIDs",
        ("FIX.5.0", "1121"): "1121 RootPartySubID",
        ("FIX.5.0", "1122"): "1122 RootPartySubIDType",
        ("FIX.5.0", "1123"): "1123 TradeHandlingInstr",
        ("FIX.5.0", "1124"): "1124 OrigTradeHandlingInstr",
        ("FIX.5.0", "1125"): "1125 OrigTradeDate",
        ("FIX.5.0", "1126"): "1126 OrigTradeID",
        ("FIX.5.0", "1127"): "1127 OrigSecondaryTradeID",
        ("FIX.5.0", "1128"): "1128 ApplVerID",
        ("FIX.5.0", "1129"): "1129 CstmApplVerID",
        ("FIX.5.0", "1130"): "1130 RefApplVerID",
        ("FIX.5.0", "1131"): "1131 RefCstmApplVerID",
        ("FIX.5.0", "1132"): "1132 TZTransactTime",
        ("FIX.5.0", "1133"): "1133 ExDestinationIDSource",
        ("FIX.5.0", "1134"): "1134 ReportedPxDiff",
        ("FIX.5.0", "1135"): "1135 RptSys",
        ("FIX.5.0", "1136"): "1136 AllocClearingFeeIndicator",
        ("FIX.5.0", "1137"): "1137 DefaultApplVerID",
        ("FIX.5.0", "1138"): "1138 DisplayQty",
        ("FIX.5.0", "1139"): "1139 ExchangeSpecialInstructions",
        ("FIX.5.0", "1140"): "1140 MaxTradeVol",
        ("FIX.5.0", "1141"): "1141 NoMDFeedTypes",
        ("FIX.5.0", "1142"): "1142 MatchAlgorithm",
        ("FIX.5.0", "1143"): "1143 MaxPriceVariation",
        ("FIX.5.0", "1144"): "1144 ImpliedMarketIndicator",
        ("FIX.5.0", "1145"): "1145 EventTime",
        ("FIX.5.0", "1146"): "1146 MinPriceIncrementAmount",
        ("FIX.5.0", "1147"): "1147 UnitOfMeasureQty",
        ("FIX.5.0", "1148"): "1148 LowLimitPrice",
        ("FIX.5.0", "1149"): "1149 HighLimitPrice",
        ("FIX.5.0", "1150"): "1150 TradingReferencePrice",
        ("FIX.5.0", "1151"): "1151 SecurityGroup",
        ("FIX.5.0", "1152"): "1152 LegNumber",
        ("FIX.5.0", "1153"): "1153 SettlementCycleNo",
        ("FIX.5.0", "1154"): "1154 SideCurrency",
        ("FIX.5.0", "1155"): "1155 SideSettlCurrency",
        ("FIX.5.0", "1156"): "1156 ApplExtID",
        ("FIX.5.0", "1157"): "1157 CcyAmt",
        ("FIX.5.0", "1158"): "1158 NoSettlDetails",
        ("FIX.5.0", "1159"): "1159 SettlObligMode",
        ("FIX.5.0", "1160"): "1160 SettlObligMsgID",
        ("FIX.5.0", "1161"): "1161 SettlObligID",
        ("FIX.5.0", "1162"): "1162 SettlObligTransType",
        ("FIX.5.0", "1163"): "1163 SettlObligRefID",
        ("FIX.5.0", "1164"): "1164 SettlObligSource",
        ("FIX.5.0", "1165"): "1165 NoSettlOblig",
        ("FIX.5.0", "1166"): "1166 QuoteMsgID",
        ("FIX.5.0", "1167"): "1167 QuoteEntryStatus",
        ("FIX.5.0", "1168"): "1168 TotNoCxldQuotes",
        ("FIX.5.0", "1169"): "1169 TotNoAccQuotes",
        ("FIX.5.0", "1170"): "1170 TotNoRejQuotes",
        ("FIX.5.0", "1171"): "1171 PrivateQuote",
        ("FIX.5.0", "1172"): "1172 RespondentType",
        ("FIX.5.0", "1173"): "1173 MDSubBookType",
        ("FIX.5.0", "1174"): "1174 SecurityTradingEvent",
        ("FIX.5.0", "1175"): "1175 NoStatsIndicators",
        ("FIX.5.0", "1176"): "1176 StatsType",
        ("FIX.5.0", "1177"): "1177 NoOfSecSizes",
        ("FIX.5.0", "1178"): "1178 MDSecSizeType",
        ("FIX.5.0", "1179"): "1179 MDSecSize",
        ("FIX.5.0", "1180"): "1180 ApplID",
        ("FIX.5.0", "1181"): "1181 ApplSeqNum",
        ("FIX.5.0", "1182"): "1182 ApplBegSeqNum",
        ("FIX.5.0", "1183"): "1183 ApplEndSeqNum",
        ("FIX.5.0", "1184"): "1184 SecurityXMLLen",
        ("FIX.5.0", "1185"): "1185 SecurityXML",
        ("FIX.5.0", "1186"): "1186 SecurityXMLSchema",
        ("FIX.5.0", "1187"): "1187 RefreshIndicator",
        ("FIX.5.0", "1188"): "1188 Volatility",
        ("FIX.5.0", "1189"): "1189 TimeToExpiration",
        ("FIX.5.0", "1190"): "1190 RiskFreeRate",
        ("FIX.5.0", "1191"): "1191 PriceUnitOfMeasure",
        ("FIX.5.0", "1192"): "1192 PriceUnitOfMeasureQty",
        ("FIX.5.0", "1193"): "1193 SettlMethod",
        ("FIX.5.0", "1194"): "1194 ExerciseStyle",
        ("FIX.5.0", "1195"): "1195 OptPayoutAmount",
        ("FIX.5.0", "1196"): "1196 PriceQuoteMethod",
        ("FIX.5.0", "1197"): "1197 ValuationMethod",
        ("FIX.5.0", "1198"): "1198 ListMethod",
        ("FIX.5.0", "1199"): "1199 CapPrice",
        ("FIX.5.0", "1200"): "1200 FloorPrice",
        ("FIX.5.0", "1201"): "1201 NoStrikeRules",
        ("FIX.5.0", "1202"): "1202 StartStrikePxRange",
        ("FIX.5.0", "1203"): "1203 EndStrikePxRange",
        ("FIX.5.0", "1204"): "1204 StrikeIncrement",
        ("FIX.5.0", "1205"): "1205 NoTickRules",
        ("FIX.5.0", "1206"): "1206 StartTickPriceRange",
        ("FIX.5.0", "1207"): "1207 EndTickPriceRange",
        ("FIX.5.0", "1208"): "1208 TickIncrement",
        ("FIX.5.0", "1209"): "1209 TickRuleType",
        ("FIX.5.0", "1210"): "1210 NestedInstrAttribType",
        ("FIX.5.0", "1211"): "1211 NestedInstrAttribValue",
        ("FIX.5.0", "1212"): "1212 LegMaturityTime",
        ("FIX.5.0", "1213"): "1213 UnderlyingMaturityTime",
        ("FIX.5.0", "1214"): "1214 DerivativeSymbol",
        ("FIX.5.0", "1215"): "1215 DerivativeSymbolSfx",
        ("FIX.5.0", "1216"): "1216 DerivativeSecurityID",
        ("FIX.5.0", "1217"): "1217 DerivativeSecurityIDSource",
        ("FIX.5.0", "1218"): "1218 NoDerivativeSecurityAltID",
        ("FIX.5.0", "1219"): "1219 DerivativeSecurityAltID",
        ("FIX.5.0", "1220"): "1220 DerivativeSecurityAltIDSource",
        ("FIX.5.0", "1221"): "1221 SecondaryLowLimitPrice",
        ("FIX.5.0", "1222"): "1222 MaturityRuleID",
        ("FIX.5.0", "1223"): "1223 StrikeRuleID",
        ("FIX.5.0", "1224"): "1224 LegUnitOfMeasureQty",
        ("FIX.5.0", "1225"): "1225 DerivativeOptPayAmount",
        ("FIX.5.0", "1226"): "1226 EndMaturityMonthYear",
        ("FIX.5.0", "1227"): "1227 ProductComplex",
        ("FIX.5.0", "1228"): "1228 DerivativeProductComplex",
        ("FIX.5.0", "1229"): "1229 MaturityMonthYearIncrement",
        ("FIX.5.0", "1230"): "1230 SecondaryHighLimitPrice",
        ("FIX.5.0", "1231"): "1231 MinLotSize",
        ("FIX.5.0", "1232"): "1232 NoExecInstRules",
        ("FIX.5.0", "1233"): "1233 CommRate",
        ("FIX.5.0", "1234"): "1234 NoLotTypeRules",
        ("FIX.5.0", "1235"): "1235 NoMatchRules",
        ("FIX.5.0", "1236"): "1236 NoMaturityRules",
        ("FIX.5.0", "1237"): "1237 NoOrdTypeRules",
        ("FIX.5.0", "1238"): "1238 CommUnitOfMeasure",
        ("FIX.5.0", "1239"): "1239 NoTimeInForceRules",
        ("FIX.5.0", "1240"): "1240 SecondaryTradingReferencePrice",
        ("FIX.5.0", "1241"): "1241 StartMaturityMonthYear",
        ("FIX.5.0", "1242"): "1242 FlexProductEligibilityIndicator",
        ("FIX.5.0", "1243"): "1243 DerivFlexProductEligibilityIndicator",
        ("FIX.5.0", "1244"): "1244 FlexibleIndicator",
        ("FIX.5.0", "1245"): "1245 TradingCurrency",
        ("FIX.5.0", "1246"): "1246 DerivativeProduct",
        ("FIX.5.0", "1247"): "1247 DerivativeSecurityGroup",
        ("FIX.5.0", "1248"): "1248 DerivativeCFICode",
        ("FIX.5.0", "1249"): "1249 DerivativeSecurityType",
        ("FIX.5.0", "1250"): "1250 DerivativeSecuritySubType",
        ("FIX.5.0", "1251"): "1251 DerivativeMaturityMonthYear",
        ("FIX.5.0", "1252"): "1252 DerivativeMaturityDate",
        ("FIX.5.0", "1253"): "1253 DerivativeMaturityTime",
        ("FIX.5.0", "1254"): "1254 DerivativeSettleOnOpenFlag",
        ("FIX.5.0", "1255"): "1255 DerivativeInstrmtAssignmentMethod",
        ("FIX.5.0", "1256"): "1256 DerivativeSecurityStatus",
        ("FIX.5.0", "1257"): "1257 DerivativeInstrRegistry",
        ("FIX.5.0", "1258"): "1258 DerivativeCountryOfIssue",
        ("FIX.5.0", "1259"): "1259 DerivativeStateOrProvinceOfIssue",
        ("FIX.5.0", "1260"): "1260 DerivativeLocaleOfIssue",
        ("FIX.5.0", "1261"): "1261 DerivativeStrikePrice",
        ("FIX.5.0", "1262"): "1262 DerivativeStrikeCurrency",
        ("FIX.5.0", "1263"): "1263 DerivativeStrikeMultiplier",
        ("FIX.5.0", "1264"): "1264 DerivativeStrikeValue",
        ("FIX.5.0", "1265"): "1265 DerivativeOptAttribute",
        ("FIX.5.0", "1266"): "1266 DerivativeContractMultiplier",
        ("FIX.5.0", "1267"): "1267 DerivativeMinPriceIncrement",
        ("FIX.5.0", "1268"): "1268 DerivativeMinPriceIncrementAmount",
        ("FIX.5.0", "1269"): "1269 DerivativeUnitOfMeasure",
        ("FIX.5.0", "1270"): "1270 DerivativeUnitOfMeasureQty",
        ("FIX.5.0", "1271"): "1271 DerivativeTimeUnit",
        ("FIX.5.0", "1272"): "1272 DerivativeSecurityExchange",
        ("FIX.5.0", "1273"): "1273 DerivativePositionLimit",
        ("FIX.5.0", "1274"): "1274 DerivativeNTPositionLimit",
        ("FIX.5.0", "1275"): "1275 DerivativeIssuer",
        ("FIX.5.0", "1276"): "1276 DerivativeIssueDate",
        ("FIX.5.0", "1277"): "1277 DerivativeEncodedIssuerLen",
        ("FIX.5.0", "1278"): "1278 DerivativeEncodedIssuer",
        ("FIX.5.0", "1279"): "1279 DerivativeSecurityDesc",
        ("FIX.5.0", "1280"): "1280 DerivativeEncodedSecurityDescLen",
        ("FIX.5.0", "1281"): "1281 DerivativeEncodedSecurityDesc",
        ("FIX.5.0", "1282"): "1282 DerivativeSecurityXMLLen",
        ("FIX.5.0", "1283"): "1283 DerivativeSecurityXML",
        ("FIX.5.0", "1284"): "1284 DerivativeSecurityXMLSchema",
        ("FIX.5.0", "1285"): "1285 DerivativeContractSettlMonth",
        ("FIX.5.0", "1286"): "1286 NoDerivativeEvents",
        ("FIX.5.0", "1287"): "1287 DerivativeEventType",
        ("FIX.5.0", "1288"): "1288 DerivativeEventDate",
        ("FIX.5.0", "1289"): "1289 DerivativeEventTime",
        ("FIX.5.0", "1290"): "1290 DerivativeEventPx",
        ("FIX.5.0", "1291"): "1291 DerivativeEventText",
        ("FIX.5.0", "1292"): "1292 NoDerivativeInstrumentParties",
        ("FIX.5.0", "1293"): "1293 DerivativeInstrumentPartyID",
        ("FIX.5.0", "1294"): "1294 DerivativeInstrumentPartyIDSource",
        ("FIX.5.0", "1295"): "1295 DerivativeInstrumentPartyRole",
        ("FIX.5.0", "1296"): "1296 NoDerivativeInstrumentPartySubIDs",
        ("FIX.5.0", "1297"): "1297 DerivativeInstrumentPartySubID",
        ("FIX.5.0", "1298"): "1298 DerivativeInstrumentPartySubIDType",
        ("FIX.5.0", "1299"): "1299 DerivativeExerciseStyle",
        ("FIX.5.0", "1300"): "1300 MarketSegmentID",
        ("FIX.5.0", "1301"): "1301 MarketID",
        ("FIX.5.0", "1302"): "1302 MaturityMonthYearIncrementUnits",
        ("FIX.5.0", "1303"): "1303 MaturityMonthYearFormat",
        ("FIX.5.0", "1304"): "1304 StrikeExerciseStyle",
        ("FIX.5.0", "1305"): "1305 SecondaryPriceLimitType",
        ("FIX.5.0", "1306"): "1306 PriceLimitType",
        ("FIX.5.0", "1308"): "1308 ExecInstValue",
        ("FIX.5.0", "1309"): "1309 NoTradingSessionRules",
        ("FIX.5.0", "1310"): "1310 NoMarketSegments",
        ("FIX.5.0", "1311"): "1311 NoDerivativeInstrAttrib",
        ("FIX.5.0", "1312"): "1312 NoNestedInstrAttrib",
        ("FIX.5.0", "1313"): "1313 DerivativeInstrAttribType",
        ("FIX.5.0", "1314"): "1314 DerivativeInstrAttribValue",
        ("FIX.5.0", "1315"): "1315 DerivativePriceUnitOfMeasure",
        ("FIX.5.0", "1316"): "1316 DerivativePriceUnitOfMeasureQty",
        ("FIX.5.0", "1317"): "1317 DerivativeSettlMethod",
        ("FIX.5.0", "1318"): "1318 DerivativePriceQuoteMethod",
        ("FIX.5.0", "1319"): "1319 DerivativeValuationMethod",
        ("FIX.5.0", "1320"): "1320 DerivativeListMethod",
        ("FIX.5.0", "1321"): "1321 DerivativeCapPrice",
        ("FIX.5.0", "1322"): "1322 DerivativeFloorPrice",
        ("FIX.5.0", "1323"): "1323 DerivativePutOrCall",
        ("FIX.5.0", "1324"): "1324 ListUpdateAction",
        ("FIX.5.0", "1325"): "1325 ParentMktSegmID",
        ("FIX.5.0", "1326"): "1326 TradingSessionDesc",
        ("FIX.5.0", "1327"): "1327 TradSesUpdateAction",
        ("FIX.5.0", "1328"): "1328 RejectText",
        ("FIX.5.0", "1329"): "1329 FeeMultiplier",
        ("FIX.5.0", "1330"): "1330 UnderlyingLegSymbol",
        ("FIX.5.0", "1331"): "1331 UnderlyingLegSymbolSfx",
        ("FIX.5.0", "1332"): "1332 UnderlyingLegSecurityID",
        ("FIX.5.0", "1333"): "1333 UnderlyingLegSecurityIDSource",
        ("FIX.5.0", "1334"): "1334 NoUnderlyingLegSecurityAltID",
        ("FIX.5.0", "1335"): "1335 UnderlyingLegSecurityAltID",
        ("FIX.5.0", "1336"): "1336 UnderlyingLegSecurityAltIDSource",
        ("FIX.5.0", "1337"): "1337 UnderlyingLegSecurityType",
        ("FIX.5.0", "1338"): "1338 UnderlyingLegSecuritySubType",
        ("FIX.5.0", "1339"): "1339 UnderlyingLegMaturityMonthYear",
        ("FIX.5.0", "1340"): "1340 UnderlyingLegStrikePrice",
        ("FIX.5.0", "1341"): "1341 UnderlyingLegSecurityExchange",
        ("FIX.5.0", "1342"): "1342 NoOfLegUnderlyings",
        ("FIX.5.0", "1343"): "1343 UnderlyingLegPutOrCall",
        ("FIX.5.0", "1344"): "1344 UnderlyingLegCFICode",
        ("FIX.5.0", "1345"): "1345 UnderlyingLegMaturityDate",
        ("FIX.5.0", "1346"): "1346 ApplReqID",
        ("FIX.5.0", "1347"): "1347 ApplReqType",
        ("FIX.5.0", "1348"): "1348 ApplResponseType",
        ("FIX.5.0", "1349"): "1349 ApplTotalMessageCount",
        ("FIX.5.0", "1350"): "1350 ApplLastSeqNum",
        ("FIX.5.0", "1351"): "1351 NoApplIDs",
        ("FIX.5.0", "1352"): "1352 ApplResendFlag",
        ("FIX.5.0", "1353"): "1353 ApplResponseID",
        ("FIX.5.0", "1354"): "1354 ApplResponseError",
        ("FIX.5.0", "1355"): "1355 RefApplID",
        ("FIX.5.0", "1356"): "1356 ApplReportID",
        ("FIX.5.0", "1357"): "1357 RefApplLastSeqNum",
        ("FIX.5.0", "1358"): "1358 LegPutOrCall",
        ("FIX.5.0", "1361"): "1361 TotNoFills",
        ("FIX.5.0", "1362"): "1362 NoFills",
        ("FIX.5.0", "1363"): "1363 FillExecID",
        ("FIX.5.0", "1364"): "1364 FillPx",
        ("FIX.5.0", "1365"): "1365 FillQty",
        ("FIX.5.0", "1366"): "1366 LegAllocID",
        ("FIX.5.0", "1367"): "1367 LegAllocSettlCurrency",
        ("FIX.5.0", "1368"): "1368 TradSesEvent",
        ("FIX.5.0", "1369"): "1369 MassActionReportID",
        ("FIX.5.0", "1370"): "1370 NoNotAffectedOrders",
        ("FIX.5.0", "1371"): "1371 NotAffectedOrderID",
        ("FIX.5.0", "1372"): "1372 NotAffOrigClOrdID",
        ("FIX.5.0", "1373"): "1373 MassActionType",
        ("FIX.5.0", "1374"): "1374 MassActionScope",
        ("FIX.5.0", "1375"): "1375 MassActionResponse",
        ("FIX.5.0", "1376"): "1376 MassActionRejectReason",
        ("FIX.5.0", "1377"): "1377 MultilegModel",
        ("FIX.5.0", "1378"): "1378 MultilegPriceMethod",
        ("FIX.5.0", "1379"): "1379 LegVolatility",
        ("FIX.5.0", "1380"): "1380 DividendYield",
        ("FIX.5.0", "1381"): "1381 LegDividendYield",
        ("FIX.5.0", "1382"): "1382 CurrencyRatio",
        ("FIX.5.0", "1383"): "1383 LegCurrencyRatio",
        ("FIX.5.0", "1384"): "1384 LegExecInst",
        ("FIX.5.0", "1385"): "1385 ContingencyType",
        ("FIX.5.0", "1386"): "1386 ListRejectReason",
        ("FIX.5.0", "1387"): "1387 NoTrdRepIndicators",
        ("FIX.5.0", "1388"): "1388 TrdRepPartyRole",
        ("FIX.5.0", "1389"): "1389 TrdRepIndicator",
        ("FIX.5.0", "1390"): "1390 TradePublishIndicator",
        ("FIX.5.0", "1391"): "1391 UnderlyingLegOptAttribute",
        ("FIX.5.0", "1392"): "1392 UnderlyingLegSecurityDesc",
        ("FIX.5.0", "1393"): "1393 MarketReqID",
        ("FIX.5.0", "1394"): "1394 MarketReportID",
        ("FIX.5.0", "1395"): "1395 MarketUpdateAction",
        ("FIX.5.0", "1396"): "1396 MarketSegmentDesc",
        ("FIX.5.0", "1397"): "1397 EncodedMktSegmDescLen",
        ("FIX.5.0", "1398"): "1398 EncodedMktSegmDesc",
        ("FIX.5.0", "1399"): "1399 ApplNewSeqNum",
        ("FIX.5.0", "1400"): "1400 EncryptedPasswordMethod",
        ("FIX.5.0", "1401"): "1401 EncryptedPasswordLen",
        ("FIX.5.0", "1402"): "1402 EncryptedPassword",
        ("FIX.5.0", "1403"): "1403 EncryptedNewPasswordLen",
        ("FIX.5.0", "1404"): "1404 EncryptedNewPassword",
        ("FIX.5.0", "1405"): "1405 UnderlyingLegMaturityTime",
        ("FIX.5.0", "1406"): "1406 RefApplExtID",
        ("FIX.5.0", "1407"): "1407 DefaultApplExtID",
        ("FIX.5.0", "1408"): "1408 DefaultCstmApplVerID",
        ("FIX.5.0", "1409"): "1409 SessionStatus",
        ("FIX.5.0", "1410"): "1410 DefaultVerIndicator",
        ("FIX.5.0", "1411"): "1411 Nested4PartySubIDType",
        ("FIX.5.0", "1412"): "1412 Nested4PartySubID",
        ("FIX.5.0", "1413"): "1413 NoNested4PartySubIDs",
        ("FIX.5.0", "1414"): "1414 NoNested4PartyIDs",
        ("FIX.5.0", "1415"): "1415 Nested4PartyID",
        ("FIX.5.0", "1416"): "1416 Nested4PartyIDSource",
        ("FIX.5.0", "1417"): "1417 Nested4PartyRole",
        ("FIX.5.0", "1418"): "1418 LegLastQty",
        ("FIX.5.0", "1419"): "1419 UnderlyingExerciseStyle",
        ("FIX.5.0", "1420"): "1420 LegExerciseStyle",
        ("FIX.5.0", "1421"): "1421 LegPriceUnitOfMeasure",
        ("FIX.5.0", "1422"): "1422 LegPriceUnitOfMeasureQty",
        ("FIX.5.0", "1423"): "1423 UnderlyingUnitOfMeasureQty",
        ("FIX.5.0", "1424"): "1424 UnderlyingPriceUnitOfMeasure",
        ("FIX.5.0", "1425"): "1425 UnderlyingPriceUnitOfMeasureQty",
        ("FIX.5.0", "1426"): "1426 ApplReportType",
        ("FIX.5.0", "1427"): "1427 SideExecID",
        ("FIX.5.0", "1428"): "1428 OrderDelay",
        ("FIX.5.0", "1429"): "1429 OrderDelayUnit",
        ("FIX.5.0", "1430"): "1430 VenueType",
        ("FIX.5.0", "1431"): "1431 RefOrdIDReason",
        ("FIX.5.0", "1432"): "1432 OrigCustOrderCapacity",
        ("FIX.5.0", "1433"): "1433 RefApplReqID",
        ("FIX.5.0", "1434"): "1434 ModelType",
        ("FIX.5.0", "1435"): "1435 ContractMultiplierUnit",
        ("FIX.5.0", "1436"): "1436 LegContractMultiplierUnit",
        ("FIX.5.0", "1437"): "1437 UnderlyingContractMultiplierUnit",
        ("FIX.5.0", "1438"): "1438 DerivativeContractMultiplierUnit",
        ("FIX.5.0", "1439"): "1439 FlowScheduleType",
        ("FIX.5.0", "1440"): "1440 LegFlowScheduleType",
        ("FIX.5.0", "1441"): "1441 UnderlyingFlowScheduleType",
        ("FIX.5.0", "1442"): "1442 DerivativeFlowScheduleType",
        ("FIX.5.0", "1443"): "1443 FillLiquidityInd",
        ("FIX.5.0", "1444"): "1444 SideLiquidityInd",
        ("FIX.5.0", "1445"): "1445 NoRateSources",
        ("FIX.5.0", "1446"): "1446 RateSource",
        ("FIX.5.0", "1447"): "1447 RateSourceType",
        ("FIX.5.0", "1448"): "1448 ReferencePage",
        ("FIX.5.0", "1449"): "1449 RestructuringType",
        ("FIX.5.0", "1450"): "1450 Seniority",
        ("FIX.5.0", "1451"): "1451 NotionalPercentageOutstanding",
        ("FIX.5.0", "1452"): "1452 OriginalNotionalPercentageOutstanding",
        ("FIX.5.0", "1453"): "1453 UnderlyingRestructuringType",
        ("FIX.5.0", "1454"): "1454 UnderlyingSeniority",
        ("FIX.5.0", "1455"): "1455 UnderlyingNotionalPercentageOutstanding",
        ("FIX.5.0", "1456"): "1456 UnderlyingOriginalNotionalPercentageOutstanding",
        ("FIX.5.0", "1457"): "1457 AttachmentPoint",
        ("FIX.5.0", "1458"): "1458 DetachmentPoint",
        ("FIX.5.0", "1459"): "1459 UnderlyingAttachmentPoint",
        ("FIX.5.0", "1460"): "1460 UnderlyingDetachmentPoint",
        ("FIX.5.0", "1461"): "1461 NoTargetPartyIDs",
        ("FIX.5.0", "1462"): "1462 TargetPartyID",
        ("FIX.5.0", "1463"): "1463 TargetPartyIDSource",
        ("FIX.5.0", "1464"): "1464 TargetPartyRole",
        ("FIX.5.0", "1465"): "1465 SecurityListID",
        ("FIX.5.0", "1466"): "1466 SecurityListRefID",
        ("FIX.5.0", "1467"): "1467 SecurityListDesc",
        ("FIX.5.0", "1468"): "1468 EncodedSecurityListDescLen",
        ("FIX.5.0", "1469"): "1469 EncodedSecurityListDesc",
        ("FIX.5.0", "1470"): "1470 SecurityListType",
        ("FIX.5.0", "1471"): "1471 SecurityListTypeSource",
        ("FIX.5.0", "1472"): "1472 NewsID",
        ("FIX.5.0", "1473"): "1473 NewsCategory",
        ("FIX.5.0", "1474"): "1474 LanguageCode",
        ("FIX.5.0", "1475"): "1475 NoNewsRefIDs",
        ("FIX.5.0", "1476"): "1476 NewsRefID",
        ("FIX.5.0", "1477"): "1477 NewsRefType",
        ("FIX.5.0", "1478"): "1478 StrikePriceDeterminationMethod",
        ("FIX.5.0", "1479"): "1479 StrikePriceBoundaryMethod",
        ("FIX.5.0", "1480"): "1480 StrikePriceBoundaryPrecision",
        ("FIX.5.0", "1481"): "1481 UnderlyingPriceDeterminationMethod",
        ("FIX.5.0", "1482"): "1482 OptPayoutType",
        ("FIX.5.0", "1483"): "1483 NoComplexEvents",
        ("FIX.5.0", "1484"): "1484 ComplexEventType",
        ("FIX.5.0", "1485"): "1485 ComplexOptPayoutAmount",
        ("FIX.5.0", "1486"): "1486 ComplexEventPrice",
        ("FIX.5.0", "1487"): "1487 ComplexEventPriceBoundaryMethod",
        ("FIX.5.0", "1488"): "1488 ComplexEventPriceBoundaryPrecision",
        ("FIX.5.0", "1489"): "1489 ComplexEventPriceTimeType",
        ("FIX.5.0", "1490"): "1490 ComplexEventCondition",
        ("FIX.5.0", "1491"): "1491 NoComplexEventDates",
        ("FIX.5.0", "1492"): "1492 ComplexEventStartDate",
        ("FIX.5.0", "1493"): "1493 ComplexEventEndDate",
        ("FIX.5.0", "1494"): "1494 NoComplexEventTimes",
        ("FIX.5.0", "1495"): "1495 ComplexEventStartTime",
        ("FIX.5.0", "1496"): "1496 ComplexEventEndTime",
        ("FIX.5.0", "1497"): "1497 StreamAsgnReqID",
        ("FIX.5.0", "1498"): "1498 StreamAsgnReqType",
        ("FIX.5.0", "1499"): "1499 NoAsgnReqs",
        ("FIX.5.0", "1500"): "1500 MDStreamID",
        ("FIX.5.0", "1501"): "1501 StreamAsgnRptID",
        ("FIX.5.0", "1502"): "1502 StreamAsgnRejReason",
        ("FIX.5.0", "1503"): "1503 StreamAsgnAckType",
        ("FIX.5.0", "1504"): "1504 RelSymTransactTime",
        ("FIX.5.0", "1505"): "1505 PartyDetailsListRequestID",
        ("FIX.5.0", "1506"): "1506 SideTradeID",
        ("FIX.5.0", "1507"): "1507 SideOrigTradeID",
        ("FIX.5.0", "1508"): "1508 NoRequestedPartyRoles",
        ("FIX.5.0", "1509"): "1509 RequestedPartyRole",
        ("FIX.5.0", "1510"): "1510 PartyDetailsListReportID",
        ("FIX.5.0", "1511"): "1511 RequestResult",
        ("FIX.5.0", "1512"): "1512 TotNoParties",
        ("FIX.5.0", "1513"): "1513 DocumentationText",
        ("FIX.5.0", "1514"): "1514 NoPartyRelationships",
        ("FIX.5.0", "1515"): "1515 PartyRelationship",
        ("FIX.5.0", "1516"): "1516 NoPartyDetailAltID",
        ("FIX.5.0", "1517"): "1517 PartyDetailAltID",
        ("FIX.5.0", "1518"): "1518 PartyDetailAltIDSource",
        ("FIX.5.0", "1519"): "1519 NoPartyDetailAltSubIDs",
        ("FIX.5.0", "1520"): "1520 PartyDetailAltSubID",
        ("FIX.5.0", "1521"): "1521 PartyDetailAltSubIDType",
        ("FIX.5.0", "1522"): "1522 DifferentialPrice",
        ("FIX.5.0", "1523"): "1523 TrdAckStatus",
        ("FIX.5.0", "1524"): "1524 PriceQuoteCurrency",
        ("FIX.5.0", "1525"): "1525 EncodedDocumentationTextLen",
        ("FIX.5.0", "1526"): "1526 UnderlyingPriceQuoteCurrency",
        ("FIX.5.0", "1527"): "1527 EncodedDocumentationText",
        ("FIX.5.0", "1528"): "1528 LegPriceQuoteCurrency",
        ("FIX.5.0", "1529"): "1529 NoRiskLimitTypes",
        ("FIX.5.0", "1530"): "1530 RiskLimitType",
        ("FIX.5.0", "1531"): "1531 RiskLimitAmount",
        ("FIX.5.0", "1532"): "1532 RiskLimitCurrency",
        ("FIX.5.0", "1533"): "1533 RiskLimitPlatform",
        ("FIX.5.0", "1534"): "1534 NoRiskInstrumentScopes",
        ("FIX.5.0", "1535"): "1535 InstrumentScopeOperator",
        ("FIX.5.0", "1536"): "1536 InstrumentScopeSymbol",
        ("FIX.5.0", "1537"): "1537 InstrumentScopeSymbolSfx",
        ("FIX.5.0", "1538"): "1538 InstrumentScopeSecurityID",
        ("FIX.5.0", "1539"): "1539 InstrumentScopeSecurityIDSource",
        ("FIX.5.0", "1540"): "1540 NoInstrumentScopeSecurityAltID",
        ("FIX.5.0", "1541"): "1541 InstrumentScopeSecurityAltID",
        ("FIX.5.0", "1542"): "1542 InstrumentScopeSecurityAltIDSource",
        ("FIX.5.0", "1543"): "1543 InstrumentScopeProduct",
        ("FIX.5.0", "1544"): "1544 InstrumentScopeProductComplex",
        ("FIX.5.0", "1545"): "1545 InstrumentScopeSecurityGroup",
        ("FIX.5.0", "1546"): "1546 InstrumentScopeCFICode",
        ("FIX.5.0", "1547"): "1547 InstrumentScopeSecurityType",
        ("FIX.5.0", "1548"): "1548 InstrumentScopeSecuritySubType",
        ("FIX.5.0", "1549"): "1549 InstrumentScopeMaturityMonthYear",
        ("FIX.5.0", "1550"): "1550 InstrumentScopeMaturityTime",
        ("FIX.5.0", "1551"): "1551 InstrumentScopeRestructuringType",
        ("FIX.5.0", "1552"): "1552 InstrumentScopeSeniority",
        ("FIX.5.0", "1553"): "1553 InstrumentScopePutOrCall",
        ("FIX.5.0", "1554"): "1554 InstrumentScopeFlexibleIndicator",
        ("FIX.5.0", "1555"): "1555 InstrumentScopeCouponRate",
        ("FIX.5.0", "1556"): "1556 InstrumentScopeSecurityDesc",
        ("FIX.5.0", "1557"): "1557 InstrumentScopeSettlType",
        ("FIX.5.0", "1558"): "1558 RiskInstrumentMultiplier",
        ("FIX.5.0", "1559"): "1559 NoRiskWarningLevels",
        ("FIX.5.0", "1560"): "1560 RiskWarningLevelPercent",
        ("FIX.5.0", "1561"): "1561 RiskWarningLevelName",
        ("FIX.5.0", "1562"): "1562 NoRelatedPartyDetailID",
        ("FIX.5.0", "1563"): "1563 RelatedPartyDetailID",
        ("FIX.5.0", "1564"): "1564 RelatedPartyDetailIDSource",
        ("FIX.5.0", "1565"): "1565 RelatedPartyDetailRole",
        ("FIX.5.0", "1566"): "1566 NoRelatedPartyDetailSubIDs",
        ("FIX.5.0", "1567"): "1567 RelatedPartyDetailSubID",
        ("FIX.5.0", "1568"): "1568 RelatedPartyDetailSubIDType",
        ("FIX.5.0", "1569"): "1569 NoRelatedPartyDetailAltID",
        ("FIX.5.0", "1570"): "1570 RelatedPartyDetailAltID",
        ("FIX.5.0", "1571"): "1571 RelatedPartyDetailAltIDSource",
        ("FIX.5.0", "1572"): "1572 NoRelatedPartyDetailAltSubIDs",
        ("FIX.5.0", "1573"): "1573 RelatedPartyDetailAltSubID",
        ("FIX.5.0", "1574"): "1574 RelatedPartyDetailAltSubIDType",
        ("FIX.5.0", "1575"): "1575 SwapSubClass",
        ("FIX.5.0", "1576"): "1576 DerivativePriceQuoteCurrency",
        ("FIX.5.0", "1577"): "1577 SettlRateIndex",
        ("FIX.5.0", "1578"): "1578 EncodedEventTextLen",
        ("FIX.5.0", "1579"): "1579 EncodedEventText",
        ("FIX.5.0", "1580"): "1580 SettlRateIndexLocation",
        ("FIX.5.0", "1581"): "1581 OptionExpirationDesc",
        ("FIX.5.0", "1582"): "1582 NoSecurityClassifications",
        ("FIX.5.0", "1583"): "1583 SecurityClassificationReason",
        ("FIX.5.0", "1584"): "1584 SecurityClassificationValue",
        ("FIX.5.0", "1585"): "1585 PosAmtReason",
        ("FIX.5.0", "1586"): "1586 NoLegPosAmt",
        ("FIX.5.0", "1587"): "1587 LegPosAmt",
        ("FIX.5.0", "1588"): "1588 LegPosAmtType",
        ("FIX.5.0", "1589"): "1589 LegPosCurrency",
        ("FIX.5.0", "1590"): "1590 LegPosAmtReason",
        ("FIX.5.0", "1591"): "1591 LegQtyType",
        ("FIX.5.0", "1592"): "1592 DiscountFactor",
        ("FIX.5.0", "1593"): "1593 ParentAllocID",
        ("FIX.5.0", "1594"): "1594 LegSecurityGroup",
        ("FIX.5.0", "1595"): "1595 PositionContingentPrice",
        ("FIX.5.0", "1596"): "1596 ClearingTradePrice",
        ("FIX.5.0", "1597"): "1597 SideClearingTradePrice",
        ("FIX.5.0", "1598"): "1598 SideClearingTradePriceType",
        ("FIX.5.0", "1599"): "1599 SidePriceDifferential",
        ("FIX.5.0", "1600"): "1600 FIXEngineName",
        ("FIX.5.0", "1601"): "1601 FIXEngineVersion",
        ("FIX.5.0", "1602"): "1602 FIXEngineVendor",
        ("FIX.5.0", "1603"): "1603 ApplicationSystemName",
        ("FIX.5.0", "1604"): "1604 ApplicationSystemVersion",
        ("FIX.5.0", "1605"): "1605 ApplicationSystemVendor",
        ("FIX.5.0", "1606"): "1606 NumOfSimpleInstruments",
        ("FIX.5.0", "1607"): "1607 SecurityRejectReason",
        ("FIX.5.0", "1608"): "1608 InitialDisplayQty",
        ("FIX.5.0", "1609"): "1609 ThrottleStatus",
        ("FIX.5.0", "1610"): "1610 NoThrottles",
        ("FIX.5.0", "1611"): "1611 ThrottleAction",
        ("FIX.5.0", "1612"): "1612 ThrottleType",
        ("FIX.5.0", "1613"): "1613 ThrottleNoMsgs",
        ("FIX.5.0", "1614"): "1614 ThrottleTimeInterval",
        ("FIX.5.0", "1615"): "1615 ThrottleTimeUnit",
        ("FIX.5.0", "1616"): "1616 InstrumentScopeSecurityExchange",
        ("FIX.5.0", "1617"): "1617 StreamAsgnType",
        ("FIX.5.0", "1618"): "1618 NoThrottleMsgType",
        ("FIX.5.0", "1619"): "1619 ThrottleMsgType",
        ("FIX.5.0", "1620"): "1620 InstrumentScopeEncodedSecurityDescLen",
        ("FIX.5.0", "1621"): "1621 InstrumentScopeEncodedSecurityDesc",
        ("FIX.5.0", "1622"): "1622 FillYieldType",
        ("FIX.5.0", "1623"): "1623 FillYield",
        ("FIX.5.0", "1624"): "1624 NoMatchInst",
        ("FIX.5.0", "1625"): "1625 MatchInst",
        ("FIX.5.0", "1626"): "1626 MatchAttribTagID",
        ("FIX.5.0", "1627"): "1627 MatchAttribValue",
        ("FIX.5.0", "1628"): "1628 TriggerScope",
        ("FIX.5.0", "1629"): "1629 ExposureDuration",
        ("FIX.5.0", "1630"): "1630 NoLimitAmts",
        ("FIX.5.0", "1631"): "1631 LimitAmtType",
        ("FIX.5.0", "1632"): "1632 LastLimitAmt",
        ("FIX.5.0", "1633"): "1633 LimitAmtRemaining",
        ("FIX.5.0", "1634"): "1634 LimitAmtCurrency",
        ("FIX.5.0", "1635"): "1635 MarginReqmtInqID",
        ("FIX.5.0", "1636"): "1636 NoMarginReqmtInqQualifier",
        ("FIX.5.0", "1637"): "1637 MarginReqmtInqQualifier",
        ("FIX.5.0", "1638"): "1638 MarginReqmtRptType",
        ("FIX.5.0", "1639"): "1639 MarginClass",
        ("FIX.5.0", "1640"): "1640 MarginReqmtInqStatus",
        ("FIX.5.0", "1641"): "1641 MarginReqmtInqResult",
        ("FIX.5.0", "1642"): "1642 MarginReqmtRptID",
        ("FIX.5.0", "1643"): "1643 NoMarginAmt",
        ("FIX.5.0", "1644"): "1644 MarginAmtType",
        ("FIX.5.0", "1645"): "1645 MarginAmt",
        ("FIX.5.0", "1646"): "1646 MarginAmtCcy",
        ("FIX.5.0", "1647"): "1647 NoRelatedInstruments",
        ("FIX.5.0", "1648"): "1648 RelatedInstrumentType",
        ("FIX.5.0", "1649"): "1649 RelatedSymbol",
        ("FIX.5.0", "1650"): "1650 RelatedSecurityID",
        ("FIX.5.0", "1651"): "1651 RelatedSecurityIDSource",
        ("FIX.5.0", "1652"): "1652 RelatedSecurityType",
        ("FIX.5.0", "1653"): "1653 RelatedMaturityMonthYear",
        ("FIX.5.0", "1654"): "1654 CoveredQty",
        ("FIX.5.0", "1655"): "1655 MarketMakerActivity",
        ("FIX.5.0", "1656"): "1656 NoInstrumentScopes",
        ("FIX.5.0", "1657"): "1657 NoRequestingPartyIDs",
        ("FIX.5.0", "1658"): "1658 RequestingPartyID",
        ("FIX.5.0", "1659"): "1659 RequestingPartyIDSource",
        ("FIX.5.0", "1660"): "1660 RequestingPartyRole",
        ("FIX.5.0", "1661"): "1661 NoRequestingPartySubIDs",
        ("FIX.5.0", "1662"): "1662 RequestingPartySubID",
        ("FIX.5.0", "1663"): "1663 RequestingPartySubIDType",
        ("FIX.5.0", "1664"): "1664 EncodedRejectTextLen",
        ("FIX.5.0", "1665"): "1665 EncodedRejectText",
        ("FIX.5.0", "1666"): "1666 RiskLimitRequestID",
        ("FIX.5.0", "1667"): "1667 RiskLimitReportID",
        ("FIX.5.0", "1668"): "1668 NoRequestedRiskLimitType",
        ("FIX.5.0", "1669"): "1669 NoRiskLimits",
        ("FIX.5.0", "1670"): "1670 RiskLimitID",
        ("FIX.5.0", "1671"): "1671 NoPartyDetails",
        ("FIX.5.0", "1672"): "1672 PartyDetailStatus",
        ("FIX.5.0", "1673"): "1673 MatchInstMarketID",
        ("FIX.5.0", "1674"): "1674 PartyDetailRoleQualifier",
        ("FIX.5.0", "1675"): "1675 RelatedPartyDetailRoleQualifier",
        ("FIX.5.0", "1676"): "1676 NoPartyUpdates",
        ("FIX.5.0", "1677"): "1677 NoPartyRiskLimits",
        ("FIX.5.0", "1678"): "1678 EncodedOptionExpirationDescLen",
        ("FIX.5.0", "1679"): "1679 SecurityMassTradingStatus",
        ("FIX.5.0", "1680"): "1680 SecurityMassTradingEvent",
        ("FIX.5.0", "1681"): "1681 MassHaltReason",
        ("FIX.5.0", "1682"): "1682 MDSecurityTradingStatus",
        ("FIX.5.0", "1683"): "1683 MDSubFeedType",
        ("FIX.5.0", "1684"): "1684 MDHaltReason",
        ("FIX.5.0", "1685"): "1685 ThrottleInst",
        ("FIX.5.0", "1686"): "1686 ThrottleCountIndicator",
        ("FIX.5.0", "1687"): "1687 ShortSaleRestriction",
        ("FIX.5.0", "1688"): "1688 ShortSaleExemptionReason",
        ("FIX.5.0", "1689"): "1689 LegShortSaleExemptionReason",
        ("FIX.5.0", "1690"): "1690 SideShortSaleExemptionReason",
        ("FIX.5.0", "1691"): "1691 PartyDetailID",
        ("FIX.5.0", "1692"): "1692 PartyDetailIDSource",
        ("FIX.5.0", "1693"): "1693 PartyDetailRole",
        ("FIX.5.0", "1694"): "1694 NoPartyDetailSubIDs",
        ("FIX.5.0", "1695"): "1695 PartyDetailSubID",
        ("FIX.5.0", "1696"): "1696 PartyDetailSubIDType",
        ("FIX.5.0", "1697"): "1697 EncodedOptionExpirationDesc",
        ("FIX.5.0", "1698"): "1698 StrikeUnitOfMeasure",
        ("FIX.5.0", "1699"): "1699 AccountSummaryReportID",
        ("FIX.5.0", "1700"): "1700 NoSettlementAmounts",
        ("FIX.5.0", "1701"): "1701 SettlementAmount",
        ("FIX.5.0", "1702"): "1702 SettlementAmountCurrency",
        ("FIX.5.0", "1703"): "1703 NoCollateralAmounts",
        ("FIX.5.0", "1704"): "1704 CurrentCollateralAmount",
        ("FIX.5.0", "1705"): "1705 CollateralCurrency",
        ("FIX.5.0", "1706"): "1706 CollateralType",
        ("FIX.5.0", "1707"): "1707 NoPayCollects",
        ("FIX.5.0", "1708"): "1708 PayCollectType",
        ("FIX.5.0", "1709"): "1709 PayCollectCurrency",
        ("FIX.5.0", "1710"): "1710 PayAmount",
        ("FIX.5.0", "1711"): "1711 CollectAmount",
        ("FIX.5.0", "1712"): "1712 PayCollectMarketSegmentID",
        ("FIX.5.0", "1713"): "1713 PayCollectMarketID",
        ("FIX.5.0", "1714"): "1714 MarginAmountMarketSegmentID",
        ("FIX.5.0", "1715"): "1715 MarginAmountMarketID",
        ("FIX.5.0", "1716"): "1716 UnitOfMeasureCurrency",
        ("FIX.5.0", "1717"): "1717 PriceUnitOfMeasureCurrency",
        ("FIX.5.0", "1718"): "1718 UnderlyingUnitOfMeasureCurrency",
        ("FIX.5.0", "1719"): "1719 UnderlyingPriceUnitOfMeasureCurrency",
        ("FIX.5.0", "1720"): "1720 LegUnitOfMeasureCurrency",
        ("FIX.5.0", "1721"): "1721 LegPriceUnitOfMeasureCurrency",
        ("FIX.5.0", "1722"): "1722 DerivativeUnitOfMeasureCurrency",
        ("FIX.5.0", "1723"): "1723 DerivativePriceUnitOfMeasureCurrency",
        ("FIX.5.0", "1724"): "1724 OrderOrigination",
        ("FIX.5.0", "1725"): "1725 OriginatingDeptID",
        ("FIX.5.0", "1726"): "1726 ReceivingDeptID",
        ("FIX.5.0", "1727"): "1727 InformationBarrierID",
        ("FIX.5.0", "1728"): "1728 FirmGroupID",
        ("FIX.5.0", "1729"): "1729 FirmMnemonic",
        ("FIX.5.0", "1730"): "1730 AllocGroupID",
        ("FIX.5.0", "1731"): "1731 AvgPxGroupID",
        ("FIX.5.0", "1732"): "1732 FirmAllocText",
        ("FIX.5.0", "1733"): "1733 EncodedFirmAllocTextLen",
        ("FIX.5.0", "1734"): "1734 EncodedFirmAllocText",
        ("FIX.5.0", "1735"): "1735 AllocationRollupInstruction",
        ("FIX.5.0", "1736"): "1736 AllocGroupQuantity",
        ("FIX.5.0", "1737"): "1737 AllocGroupRemainingQuantity",
        ("FIX.5.0", "1738"): "1738 AllocReversalStatus",
        ("FIX.5.0", "1739"): "1739 ObligationType",
        ("FIX.5.0", "1740"): "1740 TradePriceNegotiationMethod",
        ("FIX.5.0", "1741"): "1741 UpfrontPriceType",
        ("FIX.5.0", "1742"): "1742 UpfrontPrice",
        ("FIX.5.0", "1743"): "1743 LastUpfrontPrice",
        ("FIX.5.0", "1744"): "1744 ApplLevelRecoveryIndicator",
        ("FIX.5.0", "1745"): "1745 BidMDEntryID",
        ("FIX.5.0", "1746"): "1746 OfferMDEntryID",
        ("FIX.5.0", "1747"): "1747 BidQuoteID",
        ("FIX.5.0", "1748"): "1748 OfferQuoteID",
        ("FIX.5.0", "1749"): "1749 TotalBidSize",
        ("FIX.5.0", "1750"): "1750 TotalOfferSize",
        ("FIX.5.0", "1751"): "1751 SecondaryQuoteID",
        ("FIX.5.0", "1752"): "1752 CustodialLotID",
        ("FIX.5.0", "1753"): "1753 VersusPurchaseDate",
        ("FIX.5.0", "1754"): "1754 VersusPurchasePrice",
        ("FIX.5.0", "1755"): "1755 CurrentCostBasis",
        ("FIX.5.0", "1756"): "1756 LegCustodialLotID",
        ("FIX.5.0", "1757"): "1757 LegVersusPurchaseDate",
        ("FIX.5.0", "1758"): "1758 LegVersusPurchasePrice",
        ("FIX.5.0", "1759"): "1759 LegCurrentCostBasis",
        ("FIX.5.0", "1760"): "1760 RiskLimitRequestType",
        ("FIX.5.0", "1761"): "1761 RiskLimitRequestResult",
        ("FIX.5.0", "1762"): "1762 RiskLimitRequestStatus",
        ("FIX.5.0", "1763"): "1763 RiskLimitStatus",
        ("FIX.5.0", "1764"): "1764 RiskLimitResult",
        ("FIX.5.0", "1765"): "1765 RiskLimitUtilizationPercent",
        ("FIX.5.0", "1766"): "1766 RiskLimitUtilizationAmount",
        ("FIX.5.0", "1767"): "1767 RiskLimitAction",
        ("FIX.5.0", "1768"): "1768 RiskWarningLevelAmount",
        ("FIX.5.0", "1769"): "1769 RiskWarningLevelAction",
        ("FIX.5.0", "1770"): "1770 EntitlementRequestID",
        ("FIX.5.0", "1771"): "1771 EntitlementReportID",
        ("FIX.5.0", "1772"): "1772 NoPartyEntitlements",
        ("FIX.5.0", "1773"): "1773 NoEntitlements",
        ("FIX.5.0", "1774"): "1774 EntitlementIndicator",
        ("FIX.5.0", "1775"): "1775 EntitlementType",
        ("FIX.5.0", "1776"): "1776 EntitlementID",
        ("FIX.5.0", "1777"): "1777 NoEntitlementAttrib",
        ("FIX.5.0", "1778"): "1778 EntitlementAttribType",
        ("FIX.5.0", "1779"): "1779 EntitlementAttribDatatype",
        ("FIX.5.0", "1780"): "1780 EntitlementAttribValue",
        ("FIX.5.0", "1781"): "1781 EntitlementAttribCurrency",
        ("FIX.5.0", "1782"): "1782 EntitlementStartDate",
        ("FIX.5.0", "1783"): "1783 EntitlementEndDate",
        ("FIX.5.0", "1784"): "1784 EntitlementPlatform",
        ("FIX.5.0", "1785"): "1785 TradSesControl",
        ("FIX.5.0", "1786"): "1786 TradeVolType",
        ("FIX.5.0", "1787"): "1787 RefTickTableID",
        ("FIX.5.0", "1788"): "1788 LegID",
        ("FIX.5.0", "1789"): "1789 NoTargetMarketSegments",
        ("FIX.5.0", "1790"): "1790 TargetMarketSegmentID",
        ("FIX.5.0", "1791"): "1791 NoAffectedMarketSegments",
        ("FIX.5.0", "1792"): "1792 AffectedMarketSegmentID",
        ("FIX.5.0", "1793"): "1793 NoNotAffectedMarketSegments",
        ("FIX.5.0", "1794"): "1794 NotAffectedMarketSegmentID",
        ("FIX.5.0", "1795"): "1795 NoOrderEvents",
        ("FIX.5.0", "1796"): "1796 OrderEventType",
        ("FIX.5.0", "1797"): "1797 OrderEventExecID",
        ("FIX.5.0", "1798"): "1798 OrderEventReason",
        ("FIX.5.0", "1799"): "1799 OrderEventPx",
        ("FIX.5.0", "1800"): "1800 OrderEventQty",
        ("FIX.5.0", "1801"): "1801 OrderEventLiquidityIndicator",
        ("FIX.5.0", "1802"): "1802 OrderEventText",
        ("FIX.5.0", "1803"): "1803 AuctionType",
        ("FIX.5.0", "1804"): "1804 AuctionAllocationPct",
        ("FIX.5.0", "1805"): "1805 AuctionInstruction",
        ("FIX.5.0", "1806"): "1806 RefClOrdID",
        ("FIX.5.0", "1807"): "1807 LockType",
        ("FIX.5.0", "1808"): "1808 LockedQty",
        ("FIX.5.0", "1809"): "1809 SecondaryLockedQty",
        ("FIX.5.0", "1810"): "1810 ReleaseInstruction",
        ("FIX.5.0", "1811"): "1811 ReleaseQty",
        ("FIX.5.0", "1812"): "1812 NoDisclosureInstructions",
        ("FIX.5.0", "1813"): "1813 DisclosureType",
        ("FIX.5.0", "1814"): "1814 DisclosureInstruction",
        ("FIX.5.0", "1815"): "1815 TradingCapacity",
        ("FIX.5.0", "1816"): "1816 ClearingAccountType",
        ("FIX.5.0", "1817"): "1817 LegClearingAccountType",
        ("FIX.5.0", "1818"): "1818 TargetPartyRoleQualifier",
        ("FIX.5.0", "1819"): "1819 RelatedHighPrice",
        ("FIX.5.0", "1820"): "1820 RelatedLowPrice",
        ("FIX.5.0", "1821"): "1821 RelatedPriceSource",
        ("FIX.5.0", "1822"): "1822 MinQtyMethod",
        ("FIX.5.0", "1823"): "1823 Triggered",
        ("FIX.5.0", "1824"): "1824 AffectedOrigClOrdID",
        ("FIX.5.0", "1825"): "1825 NotAffSecondaryOrderID",
        ("FIX.5.0", "1826"): "1826 EventTimePeriod",
        ("FIX.5.0", "1827"): "1827 EventTimeUnit",
        ("FIX.5.0", "1828"): "1828 LastQtyVariance",
        ("FIX.5.0", "1829"): "1829 NoCrossLegs",
        ("FIX.5.0", "1830"): "1830 SettlPriceIncrement",
        ("FIX.5.0", "1831"): "1831 SettlPriceSecondaryIncrement",
        ("FIX.5.0", "1832"): "1832 ClearedIndicator",
        ("FIX.5.0", "1833"): "1833 ContractRefPosType",
        ("FIX.5.0", "1834"): "1834 PositionCapacity",
        ("FIX.5.0", "1835"): "1835 PosQtyUnitOfMeasureCurrency",
        ("FIX.5.0", "1836"): "1836 PosQtyUnitOfMeasure",
        ("FIX.5.0", "1837"): "1837 UnderlyingContractPriceRefMonth",
        ("FIX.5.0", "1838"): "1838 NoTradePriceConditions",
        ("FIX.5.0", "1839"): "1839 TradePriceCondition",
        ("FIX.5.0", "1840"): "1840 TradeAllocStatus",
        ("FIX.5.0", "1841"): "1841 NoTradeQtys",
        ("FIX.5.0", "1842"): "1842 TradeQtyType",
        ("FIX.5.0", "1843"): "1843 TradeQty",
        ("FIX.5.0", "1844"): "1844 NoTradeAllocAmts",
        ("FIX.5.0", "1845"): "1845 TradeAllocAmtType",
        ("FIX.5.0", "1846"): "1846 TradeAllocAmt",
        ("FIX.5.0", "1847"): "1847 TradeAllocCurrency",
        ("FIX.5.0", "1848"): "1848 TradeAllocGroupInstruction",
        ("FIX.5.0", "1849"): "1849 OffsetInstruction",
        ("FIX.5.0", "1850"): "1850 TradeAllocAmtReason",
        ("FIX.5.0", "1851"): "1851 StrategyLinkID",
        ("FIX.5.0", "1852"): "1852 SideAvgPx",
        ("FIX.5.0", "1853"): "1853 SideAvgPxIndicator",
        ("FIX.5.0", "1854"): "1854 SideAvgPxGroupID",
        ("FIX.5.0", "1855"): "1855 NoRelatedTrades",
        ("FIX.5.0", "1856"): "1856 RelatedTradeID",
        ("FIX.5.0", "1857"): "1857 RelatedTradeIDSource",
        ("FIX.5.0", "1858"): "1858 RelatedTradeDate",
        ("FIX.5.0", "1859"): "1859 RelatedTradeMarketID",
        ("FIX.5.0", "1860"): "1860 RelatedTradeQuantity",
        ("FIX.5.0", "1861"): "1861 NoRelatedPositions",
        ("FIX.5.0", "1862"): "1862 RelatedPositionID",
        ("FIX.5.0", "1863"): "1863 RelatedPositionIDSource",
        ("FIX.5.0", "1864"): "1864 RelatedPositionDate",
        ("FIX.5.0", "1865"): "1865 QuoteAckStatus",
        ("FIX.5.0", "1866"): "1866 StrikeIndex",
        ("FIX.5.0", "1867"): "1867 OfferID",
        ("FIX.5.0", "1868"): "1868 NoValueChecks",
        ("FIX.5.0", "1869"): "1869 ValueCheckType",
        ("FIX.5.0", "1870"): "1870 ValueCheckAction",
        ("FIX.5.0", "1871"): "1871 LegSecurityXMLLen",
        ("FIX.5.0", "1872"): "1872 LegSecurityXML",
        ("FIX.5.0", "1873"): "1873 LegSecurityXMLSchema",
        ("FIX.5.0", "1874"): "1874 UnderlyingSecurityXMLLen",
        ("FIX.5.0", "1875"): "1875 UnderlyingSecurityXML",
        ("FIX.5.0", "1876"): "1876 UnderlyingSecurityXMLSchema",
        ("FIX.5.0", "1877"): "1877 PartyDetailRequestResult",
        ("FIX.5.0", "1878"): "1878 PartyDetailRequestStatus",
        ("FIX.5.0", "1879"): "1879 PartyDetailDefinitionStatus",
        ("FIX.5.0", "1880"): "1880 PartyDetailDefinitionResult",
        ("FIX.5.0", "1881"): "1881 EntitlementRequestResult",
        ("FIX.5.0", "1882"): "1882 EntitlementRequestStatus",
        ("FIX.5.0", "1883"): "1883 EntitlementStatus",
        ("FIX.5.0", "1884"): "1884 EntitlementResult",
        ("FIX.5.0", "1885"): "1885 EntitlementRefID",
        ("FIX.5.0", "1886"): "1886 SettlPriceUnitOfMeasure",
        ("FIX.5.0", "1887"): "1887 SettlPriceUnitOfMeasureCurrency",
        ("FIX.5.0", "1888"): "1888 TradeMatchTimestamp",
        ("FIX.5.0", "1889"): "1889 NoInstrmtMatchSides",
        ("FIX.5.0", "1890"): "1890 NoTrdMatchSides",
        ("FIX.5.0", "1891"): "1891 TrdMatchSubID",
        ("FIX.5.0", "1892"): "1892 NoLegExecs",
        ("FIX.5.0", "1893"): "1893 LegExecID",
        ("FIX.5.0", "1894"): "1894 LegTradeID",
        ("FIX.5.0", "1895"): "1895 LegTradeReportID",
        ("FIX.5.0", "1896"): "1896 TradeMatchAckStatus",
        ("FIX.5.0", "1897"): "1897 TradeMatchRejectReason",
        ("FIX.5.0", "1898"): "1898 SideMarketSegmentID",
        ("FIX.5.0", "1899"): "1899 SideVenueType",
        ("FIX.5.0", "1900"): "1900 SideExecRefID",
        ("FIX.5.0", "1901"): "1901 LegExecRefID",
        ("FIX.5.0", "1902"): "1902 HaircutIndicator",
        ("FIX.5.0", "1903"): "1903 RegulatoryTradeID",
        ("FIX.5.0", "1904"): "1904 RegulatoryTradeIDEvent",
        ("FIX.5.0", "1905"): "1905 RegulatoryTradeIDSource",
        ("FIX.5.0", "1906"): "1906 RegulatoryTradeIDType",
        ("FIX.5.0", "1907"): "1907 NoRegulatoryTradeIDs",
        ("FIX.5.0", "1908"): "1908 NoAllocRegulatoryTradeIDs",
        ("FIX.5.0", "1909"): "1909 AllocRegulatoryTradeID",
        ("FIX.5.0", "1910"): "1910 AllocRegulatoryTradeIDSource",
        ("FIX.5.0", "1911"): "1911 AllocRegulatoryTradeIDEvent",
        ("FIX.5.0", "1912"): "1912 AllocRegulatoryTradeIDType",
        ("FIX.5.0", "1913"): "1913 NumOfCompetitors",
        ("FIX.5.0", "1914"): "1914 ResponseTime",
        ("FIX.5.0", "1915"): "1915 QuoteDisplayTime",
        ("FIX.5.0", "1916"): "1916 ExposureDurationUnit",
        ("FIX.5.0", "1917"): "1917 CoverPrice",
        ("FIX.5.0", "1918"): "1918 NoClearingAccountTypes",
        ("FIX.5.0", "1919"): "1919 NoPriceMovements",
        ("FIX.5.0", "1920"): "1920 NoPriceMovementValues",
        ("FIX.5.0", "1921"): "1921 PriceMovementValue",
        ("FIX.5.0", "1922"): "1922 PriceMovementPoint",
        ("FIX.5.0", "1923"): "1923 PriceMovementType",
        ("FIX.5.0", "1924"): "1924 ClearingIntention",
        ("FIX.5.0", "1925"): "1925 TradeClearingInstruction",
        ("FIX.5.0", "1926"): "1926 BackloadedTradeIndicator",
        ("FIX.5.0", "1927"): "1927 ConfirmationMethod",
        ("FIX.5.0", "1928"): "1928 MandatoryClearingIndicator",
        ("FIX.5.0", "1929"): "1929 MixedSwapIndicator",
        ("FIX.5.0", "1930"): "1930 OffMarketPriceIndicator",
        ("FIX.5.0", "1931"): "1931 VerificationMethod",
        ("FIX.5.0", "1932"): "1932 ClearingRequirementException",
        ("FIX.5.0", "1933"): "1933 IRSDirection",
        ("FIX.5.0", "1934"): "1934 RegulatoryReportType",
        ("FIX.5.0", "1935"): "1935 VoluntaryRegulatoryReport",
        ("FIX.5.0", "1936"): "1936 TradeCollateralization",
        ("FIX.5.0", "1937"): "1937 TradeContinuation",
        ("FIX.5.0", "1938"): "1938 AssetClass",
        ("FIX.5.0", "1939"): "1939 AssetSubClass",
        ("FIX.5.0", "1940"): "1940 AssetType",
        ("FIX.5.0", "1941"): "1941 SwapClass",
        ("FIX.5.0", "1942"): "1942 NthToDefault",
        ("FIX.5.0", "1943"): "1943 MthToDefault",
        ("FIX.5.0", "1944"): "1944 SettledEntityMatrixSource",
        ("FIX.5.0", "1945"): "1945 SettledEntityMatrixPublicationDate",
        ("FIX.5.0", "1946"): "1946 CouponType",
        ("FIX.5.0", "1947"): "1947 TotalIssuedAmount",
        ("FIX.5.0", "1948"): "1948 CouponFrequencyPeriod",
        ("FIX.5.0", "1949"): "1949 CouponFrequencyUnit",
        ("FIX.5.0", "1950"): "1950 CouponDayCount",
        ("FIX.5.0", "1951"): "1951 ConvertibleBondEquityID",
        ("FIX.5.0", "1952"): "1952 ConvertibleBondEquityIDSource",
        ("FIX.5.0", "1953"): "1953 ContractPriceRefMonth",
        ("FIX.5.0", "1954"): "1954 LienSeniority",
        ("FIX.5.0", "1955"): "1955 LoanFacility",
        ("FIX.5.0", "1956"): "1956 ReferenceEntityType",
        ("FIX.5.0", "1957"): "1957 IndexSeries",
        ("FIX.5.0", "1958"): "1958 IndexAnnexVersion",
        ("FIX.5.0", "1959"): "1959 IndexAnnexDate",
        ("FIX.5.0", "1960"): "1960 IndexAnnexSource",
        ("FIX.5.0", "1961"): "1961 AgreementVersion",
        ("FIX.5.0", "1962"): "1962 MasterConfirmationDesc",
        ("FIX.5.0", "1963"): "1963 MasterConfirmationDate",
        ("FIX.5.0", "1964"): "1964 MasterConfirmationAnnexDesc",
        ("FIX.5.0", "1965"): "1965 MasterConfirmationAnnexDate",
        ("FIX.5.0", "1966"): "1966 BrokerConfirmationDesc",
        ("FIX.5.0", "1967"): "1967 CreditSupportAgreementDesc",
        ("FIX.5.0", "1968"): "1968 CreditSupportAgreementDate",
        ("FIX.5.0", "1969"): "1969 CreditSupportAgreementID",
        ("FIX.5.0", "1970"): "1970 GoverningLaw",
        ("FIX.5.0", "1971"): "1971 NoSideRegulatoryTradeIDs",
        ("FIX.5.0", "1972"): "1972 SideRegulatoryTradeID",
        ("FIX.5.0", "1973"): "1973 SideRegulatoryTradeIDSource",
        ("FIX.5.0", "1974"): "1974 SideRegulatoryTradeIDEvent",
        ("FIX.5.0", "1975"): "1975 SideRegulatoryTradeIDType",
        ("FIX.5.0", "1976"): "1976 NoSecondaryAssetClasses",
        ("FIX.5.0", "1977"): "1977 SecondaryAssetClass",
        ("FIX.5.0", "1978"): "1978 SecondaryAssetSubClass",
        ("FIX.5.0", "1979"): "1979 SecondaryAssetType",
        ("FIX.5.0", "1980"): "1980 BlockTrdAllocIndicator",
        ("FIX.5.0", "1981"): "1981 NoUnderlyingEvents",
        ("FIX.5.0", "1982"): "1982 UnderlyingEventType",
        ("FIX.5.0", "1983"): "1983 UnderlyingEventDate",
        ("FIX.5.0", "1984"): "1984 UnderlyingEventTime",
        ("FIX.5.0", "1985"): "1985 UnderlyingEventTimeUnit",
        ("FIX.5.0", "1986"): "1986 UnderlyingEventTimePeriod",
        ("FIX.5.0", "1987"): "1987 UnderlyingEventPx",
        ("FIX.5.0", "1988"): "1988 UnderlyingConstituentWeight",
        ("FIX.5.0", "1989"): "1989 UnderlyingCouponType",
        ("FIX.5.0", "1990"): "1990 UnderlyingTotalIssuedAmount",
        ("FIX.5.0", "1991"): "1991 UnderlyingCouponFrequencyPeriod",
        ("FIX.5.0", "1992"): "1992 UnderlyingCouponFrequencyUnit",
        ("FIX.5.0", "1993"): "1993 UnderlyingCouponDayCount",
        ("FIX.5.0", "1994"): "1994 UnderlyingObligationID",
        ("FIX.5.0", "1995"): "1995 UnderlyingObligationIDSource",
        ("FIX.5.0", "1996"): "1996 UnderlyingEquityID",
        ("FIX.5.0", "1997"): "1997 UnderlyingEquityIDSource",
        ("FIX.5.0", "1998"): "1998 UnderlyingLienSeniority",
        ("FIX.5.0", "1999"): "1999 UnderlyingLoanFacility",
        ("FIX.5.0", "2000"): "2000 UnderlyingReferenceEntityType",
        ("FIX.5.0", "2001"): "2001 StrikeIndexSpread",
        ("FIX.5.0", "2002"): "2002 ValuationSource",
        ("FIX.5.0", "2003"): "2003 UnderlyingIndexSeries",
        ("FIX.5.0", "2004"): "2004 UnderlyingIndexAnnexVersion",
        ("FIX.5.0", "2005"): "2005 UnderlyingIndexAnnexDate",
        ("FIX.5.0", "2006"): "2006 UnderlyingIndexAnnexSource",
        ("FIX.5.0", "2007"): "2007 UnderlyingProductComplex",
        ("FIX.5.0", "2008"): "2008 UnderlyingSecurityGroup",
        ("FIX.5.0", "2009"): "2009 UnderlyingSettleOnOpenFlag",
        ("FIX.5.0", "2010"): "2010 UnderlyingAssignmentMethod",
        ("FIX.5.0", "2011"): "2011 UnderlyingSecurityStatus",
        ("FIX.5.0", "2012"): "2012 UnderlyingObligationType",
        ("FIX.5.0", "2013"): "2013 UnderlyingAssetClass",
        ("FIX.5.0", "2014"): "2014 UnderlyingAssetSubClass",
        ("FIX.5.0", "2015"): "2015 UnderlyingAssetType",
        ("FIX.5.0", "2016"): "2016 UnderlyingSwapClass",
        ("FIX.5.0", "2017"): "2017 UnderlyingNthToDefault",
        ("FIX.5.0", "2018"): "2018 UnderlyingMthToDefault",
        ("FIX.5.0", "2019"): "2019 UnderlyingSettledEntityMatrixSource",
        ("FIX.5.0", "2020"): "2020 UnderlyingSettledEntityMatrixPublicationDate",
        ("FIX.5.0", "2021"): "2021 UnderlyingStrikeMultiplier",
        ("FIX.5.0", "2022"): "2022 UnderlyingStrikeValue",
        ("FIX.5.0", "2023"): "2023 UnderlyingStrikePriceDeterminationMethod",
        ("FIX.5.0", "2024"): "2024 UnderlyingStrikePriceBoundaryMethod",
        ("FIX.5.0", "2025"): "2025 UnderlyingStrikePriceBoundaryPrecision",
        ("FIX.5.0", "2026"): "2026 UnderlyingMinPriceIncrement",
        ("FIX.5.0", "2027"): "2027 UnderlyingMinPriceIncrementAmount",
        ("FIX.5.0", "2028"): "2028 UnderlyingOptPayoutType",
        ("FIX.5.0", "2029"): "2029 UnderlyingOptPayoutAmount",
        ("FIX.5.0", "2030"): "2030 UnderlyingPriceQuoteMethod",
        ("FIX.5.0", "2031"): "2031 UnderlyingValuationMethod",
        ("FIX.5.0", "2032"): "2032 UnderlyingListMethod",
        ("FIX.5.0", "2033"): "2033 UnderlyingCapPrice",
        ("FIX.5.0", "2034"): "2034 UnderlyingFloorPrice",
        ("FIX.5.0", "2035"): "2035 UnderlyingFlexibleIndicator",
        ("FIX.5.0", "2036"): "2036 UnderlyingFlexProductEligibilityIndicator",
        ("FIX.5.0", "2037"): "2037 UnderlyingPositionLimit",
        ("FIX.5.0", "2038"): "2038 UnderlyingNTPositionLimit",
        ("FIX.5.0", "2039"): "2039 UnderlyingPool",
        ("FIX.5.0", "2040"): "2040 UnderlyingContractSettlMonth",
        ("FIX.5.0", "2041"): "2041 UnderlyingDatedDate",
        ("FIX.5.0", "2042"): "2042 UnderlyingInterestAccrualDate",
        ("FIX.5.0", "2043"): "2043 UnderlyingShortSaleRestriction",
        ("FIX.5.0", "2044"): "2044 UnderlyingRefTickTableID",
        ("FIX.5.0", "2045"): "2045 NoUnderlyingComplexEvents",
        ("FIX.5.0", "2046"): "2046 UnderlyingComplexEventType",
        ("FIX.5.0", "2047"): "2047 UnderlyingComplexOptPayoutAmount",
        ("FIX.5.0", "2048"): "2048 UnderlyingComplexEventPrice",
        ("FIX.5.0", "2049"): "2049 UnderlyingComplexEventPriceBoundaryMethod",
        ("FIX.5.0", "2050"): "2050 UnderlyingComplexEventPriceBoundaryPrecision",
        ("FIX.5.0", "2051"): "2051 UnderlyingComplexEventPriceTimeType",
        ("FIX.5.0", "2052"): "2052 UnderlyingComplexEventCondition",
        ("FIX.5.0", "2053"): "2053 NoUnderlyingComplexEventDates",
        ("FIX.5.0", "2054"): "2054 UnderlyingComplexEventStartDate",
        ("FIX.5.0", "2055"): "2055 UnderlyingComplexEventEndDate",
        ("FIX.5.0", "2056"): "2056 NoUnderlyingComplexEventTimes",
        ("FIX.5.0", "2057"): "2057 UnderlyingComplexEventStartTime",
        ("FIX.5.0", "2058"): "2058 UnderlyingComplexEventEndTime",
        ("FIX.5.0", "2059"): "2059 NoLegEvents",
        ("FIX.5.0", "2060"): "2060 LegEventType",
        ("FIX.5.0", "2061"): "2061 LegEventDate",
        ("FIX.5.0", "2062"): "2062 LegEventTime",
        ("FIX.5.0", "2063"): "2063 LegEventTimeUnit",
        ("FIX.5.0", "2064"): "2064 LegEventTimePeriod",
        ("FIX.5.0", "2065"): "2065 LegEventPx",
        ("FIX.5.0", "2066"): "2066 LegEventText",
        ("FIX.5.0", "2067"): "2067 LegAssetClass",
        ("FIX.5.0", "2068"): "2068 LegAssetSubClass",
        ("FIX.5.0", "2069"): "2069 LegAssetType",
        ("FIX.5.0", "2070"): "2070 LegSwapClass",
        ("FIX.5.0", "2071"): "2071 UnderlyingEventText",
        ("FIX.5.0", "2072"): "2072 EncodedUnderlyingEventTextLen",
        ("FIX.5.0", "2073"): "2073 EncodedUnderlyingEventText",
        ("FIX.5.0", "2074"): "2074 EncodedLegEventTextLen",
        ("FIX.5.0", "2075"): "2075 EncodedLegEventText",
        ("FIX.5.0", "2076"): "2076 NoLegSecondaryAssetClasses",
        ("FIX.5.0", "2077"): "2077 LegSecondaryAssetClass",
        ("FIX.5.0", "2078"): "2078 LegSecondaryAssetSubClass",
        ("FIX.5.0", "2079"): "2079 LegSecondaryAssetType",
        ("FIX.5.0", "2080"): "2080 NoUnderlyingSecondaryAssetClasses",
        ("FIX.5.0", "2081"): "2081 UnderlyingSecondaryAssetClass",
        ("FIX.5.0", "2082"): "2082 UnderlyingSecondaryAssetSubClass",
        ("FIX.5.0", "2083"): "2083 UnderlyingSecondaryAssetType",
        ("FIX.5.0", "2084"): "2084 PreviousClearingBusinessDate",
        ("FIX.5.0", "2085"): "2085 ValuationDate",
        ("FIX.5.0", "2086"): "2086 ValuationTime",
        ("FIX.5.0", "2087"): "2087 ValuationBusinessCenter",
        ("FIX.5.0", "2088"): "2088 MarginAmtFXRate",
        ("FIX.5.0", "2089"): "2089 MarginAmtFXRateCalc",
        ("FIX.5.0", "2090"): "2090 CollateralFXRate",
        ("FIX.5.0", "2091"): "2091 CollateralFXRateCalc",
        ("FIX.5.0", "2092"): "2092 CollateralAmountMarketSegmentID",
        ("FIX.5.0", "2093"): "2093 CollateralAmountMarketID",
        ("FIX.5.0", "2094"): "2094 PayCollectFXRate",
        ("FIX.5.0", "2095"): "2095 PayCollectFXRateCalc",
        ("FIX.5.0", "2096"): "2096 PosAmtStreamDesc",
        ("FIX.5.0", "2097"): "2097 PositionFXRate",
        ("FIX.5.0", "2098"): "2098 PositionFXRateCalc",
        ("FIX.5.0", "2099"): "2099 PosAmtMarketSegmentID",
        ("FIX.5.0", "2100"): "2100 PosAmtMarketID",
        ("FIX.5.0", "2101"): "2101 TerminatedIndicator",
        ("FIX.5.0", "2102"): "2102 ShortMarkingExemptIndicator",
        ("FIX.5.0", "2103"): "2103 RelatedRegulatoryTradeIDSource",
        ("FIX.5.0", "2104"): "2104 NoAttachments",
        ("FIX.5.0", "2105"): "2105 AttachmentName",
        ("FIX.5.0", "2106"): "2106 AttachmentMediaType",
        ("FIX.5.0", "2107"): "2107 AttachmentClassification",
        ("FIX.5.0", "2108"): "2108 AttachmentExternalURL",
        ("FIX.5.0", "2109"): "2109 AttachmentEncodingType",
        ("FIX.5.0", "2110"): "2110 UnencodedAttachmentLen",
        ("FIX.5.0", "2111"): "2111 EncodedAttachmentLen",
        ("FIX.5.0", "2112"): "2112 EncodedAttachment",
        ("FIX.5.0", "2113"): "2113 NoAttachmentKeywords",
        ("FIX.5.0", "2114"): "2114 AttachmentKeyword",
        ("FIX.5.0", "2115"): "2115 NegotiationMethod",
        ("FIX.5.0", "2116"): "2116 NextAuctionTime",
        ("FIX.5.0", "2117"): "2117 ComplexOptPayoutPaySide",
        ("FIX.5.0", "2118"): "2118 ComplexOptPayoutReceiveSide",
        ("FIX.5.0", "2119"): "2119 ComplexOptPayoutUnderlier",
        ("FIX.5.0", "2120"): "2120 ComplexOptPayoutPercentage",
        ("FIX.5.0", "2121"): "2121 ComplexOptPayoutTime",
        ("FIX.5.0", "2122"): "2122 ComplexOptPayoutCurrency",
        ("FIX.5.0", "2123"): "2123 ComplexEventPricePercentage",
        ("FIX.5.0", "2124"): "2124 ComplexEventCurrencyOne",
        ("FIX.5.0", "2125"): "2125 ComplexEventCurrencyTwo",
        ("FIX.5.0", "2126"): "2126 ComplexEventQuoteBasis",
        ("FIX.5.0", "2127"): "2127 ComplexEventFixedFXRate",
        ("FIX.5.0", "2128"): "2128 ComplexEventDeterminationMethod",
        ("FIX.5.0", "2129"): "2129 ComplexEventCalculationAgent",
        ("FIX.5.0", "2130"): "2130 ComplexEventStrikePrice",
        ("FIX.5.0", "2131"): "2131 ComplexEventStrikeFactor",
        ("FIX.5.0", "2132"): "2132 ComplexEventStrikeNumberOfOptions",
        ("FIX.5.0", "2133"): "2133 ComplexEventCreditEventsXIDRef",
        ("FIX.5.0", "2134"): "2134 ComplexEventCreditEventNotifyingParty",
        ("FIX.5.0", "2135"): "2135 ComplexEventCreditEventBusinessCenter",
        ("FIX.5.0", "2136"): "2136 ComplexEventCreditEventStandardSources",
        ("FIX.5.0", "2137"): "2137 ComplexEventCreditEventMinimumSources",
        ("FIX.5.0", "2138"): "2138 ComplexEventXID",
        ("FIX.5.0", "2139"): "2139 ComplexEventXIDRef",
        ("FIX.5.0", "2140"): "2140 ValuationReferenceModel",
        ("FIX.5.0", "2141"): "2141 StrategyType",
        ("FIX.5.0", "2142"): "2142 CommonPricingIndicator",
        ("FIX.5.0", "2143"): "2143 SettlDisruptionProvision",
        ("FIX.5.0", "2144"): "2144 InstrumentRoundingDirection",
        ("FIX.5.0", "2145"): "2145 InstrumentRoundingPrecision",
        ("FIX.5.0", "2146"): "2146 LegSettleOnOpenFlag",
        ("FIX.5.0", "2147"): "2147 LegInstrmtAssignmentMethod",
        ("FIX.5.0", "2148"): "2148 LegSecurityStatus",
        ("FIX.5.0", "2149"): "2149 LegRestructuringType",
        ("FIX.5.0", "2150"): "2150 LegSeniority",
        ("FIX.5.0", "2151"): "2151 LegNotionalPercentageOutstanding",
        ("FIX.5.0", "2152"): "2152 LegOriginalNotionalPercentageOutstanding",
        ("FIX.5.0", "2153"): "2153 LegAttachmentPoint",
        ("FIX.5.0", "2154"): "2154 LegDetachmentPoint",
        ("FIX.5.0", "2155"): "2155 LegObligationType",
        ("FIX.5.0", "2156"): "2156 LegSwapSubClass",
        ("FIX.5.0", "2157"): "2157 LegNthToDefault",
        ("FIX.5.0", "2158"): "2158 LegMthToDefault",
        ("FIX.5.0", "2159"): "2159 LegSettledEntityMatrixSource",
        ("FIX.5.0", "2160"): "2160 LegSettledEntityMatrixPublicationDate",
        ("FIX.5.0", "2161"): "2161 LegCouponType",
        ("FIX.5.0", "2162"): "2162 LegTotalIssuedAmount",
        ("FIX.5.0", "2163"): "2163 LegCouponFrequencyPeriod",
        ("FIX.5.0", "2164"): "2164 LegCouponFrequencyUnit",
        ("FIX.5.0", "2165"): "2165 LegCouponDayCount",
        ("FIX.5.0", "2166"): "2166 LegConvertibleBondEquityID",
        ("FIX.5.0", "2167"): "2167 LegConvertibleBondEquityIDSource",
        ("FIX.5.0", "2168"): "2168 LegContractPriceRefMonth",
        ("FIX.5.0", "2169"): "2169 LegLienSeniority",
        ("FIX.5.0", "2170"): "2170 LegLoanFacility",
        ("FIX.5.0", "2171"): "2171 LegReferenceEntityType",
        ("FIX.5.0", "2172"): "2172 LegIndexSeries",
        ("FIX.5.0", "2173"): "2173 LegIndexAnnexVersion",
        ("FIX.5.0", "2174"): "2174 LegIndexAnnexDate",
        ("FIX.5.0", "2175"): "2175 LegIndexAnnexSource",
        ("FIX.5.0", "2176"): "2176 LegSettlRateIndex",
        ("FIX.5.0", "2177"): "2177 LegSettlRateIndexLocation",
        ("FIX.5.0", "2178"): "2178 LegOptionExpirationDesc",
        ("FIX.5.0", "2179"): "2179 EncodedLegOptionExpirationDescLen",
        ("FIX.5.0", "2180"): "2180 EncodedLegOptionExpirationDesc",
        ("FIX.5.0", "2181"): "2181 LegStrikeMultiplier",
        ("FIX.5.0", "2182"): "2182 LegStrikeValue",
        ("FIX.5.0", "2183"): "2183 LegStrikeUnitOfMeasure",
        ("FIX.5.0", "2184"): "2184 LegStrikeIndex",
        ("FIX.5.0", "2185"): "2185 LegStrikeIndexSpread",
        ("FIX.5.0", "2186"): "2186 LegStrikePriceDeterminationMethod",
        ("FIX.5.0", "2187"): "2187 LegStrikePriceBoundaryMethod",
        ("FIX.5.0", "2188"): "2188 LegStrikePriceBoundaryPrecision",
        ("FIX.5.0", "2189"): "2189 LegUnderlyingPriceDeterminationMethod",
        ("FIX.5.0", "2190"): "2190 LegMinPriceIncrement",
        ("FIX.5.0", "2191"): "2191 LegMinPriceIncrementAmount",
        ("FIX.5.0", "2192"): "2192 LegSettlMethod",
        ("FIX.5.0", "2193"): "2193 LegOptPayoutType",
        ("FIX.5.0", "2194"): "2194 LegOptPayoutAmount",
        ("FIX.5.0", "2195"): "2195 LegPriceQuoteMethod",
        ("FIX.5.0", "2196"): "2196 LegValuationMethod",
        ("FIX.5.0", "2197"): "2197 LegValuationSource",
        ("FIX.5.0", "2198"): "2198 LegValuationReferenceModel",
        ("FIX.5.0", "2199"): "2199 LegListMethod",
        ("FIX.5.0", "2200"): "2200 LegCapPrice",
        ("FIX.5.0", "2201"): "2201 LegFloorPrice",
        ("FIX.5.0", "2202"): "2202 LegFlexibleIndicator",
        ("FIX.5.0", "2203"): "2203 LegFlexProductEligibilityIndicator",
        ("FIX.5.0", "2204"): "2204 LegComplexEventStartTime",
        ("FIX.5.0", "2205"): "2205 LegPositionLimit",
        ("FIX.5.0", "2206"): "2206 LegNTPositionLimit",
        ("FIX.5.0", "2207"): "2207 LegCPProgram",
        ("FIX.5.0", "2208"): "2208 LegCPRegType",
        ("FIX.5.0", "2209"): "2209 LegShortSaleRestriction",
        ("FIX.5.0", "2210"): "2210 AssetGroup",
        ("FIX.5.0", "2211"): "2211 LegStrategyType",
        ("FIX.5.0", "2212"): "2212 LegCommonPricingIndicator",
        ("FIX.5.0", "2213"): "2213 LegSettlDisruptionProvision",
        ("FIX.5.0", "2214"): "2214 LegInstrumentRoundingDirection",
        ("FIX.5.0", "2215"): "2215 LegInstrumentRoundingPrecision",
        ("FIX.5.0", "2216"): "2216 MiscFeeRate",
        ("FIX.5.0", "2217"): "2217 MiscFeeAmountDue",
        ("FIX.5.0", "2218"): "2218 NoLegComplexEvents",
        ("FIX.5.0", "2219"): "2219 LegComplexEventType",
        ("FIX.5.0", "2220"): "2220 LegComplexOptPayoutPaySide",
        ("FIX.5.0", "2221"): "2221 LegComplexOptPayoutReceiveSide",
        ("FIX.5.0", "2222"): "2222 LegComplexOptPayoutUnderlier",
        ("FIX.5.0", "2223"): "2223 LegComplexOptPayoutAmount",
        ("FIX.5.0", "2224"): "2224 LegComplexOptPayoutPercentage",
        ("FIX.5.0", "2225"): "2225 LegComplexOptPayoutTime",
        ("FIX.5.0", "2226"): "2226 LegComplexOptPayoutCurrency",
        ("FIX.5.0", "2227"): "2227 LegComplexEventPrice",
        ("FIX.5.0", "2228"): "2228 LegComplexEventPricePercentage",
        ("FIX.5.0", "2229"): "2229 LegComplexEventPriceBoundaryMethod",
        ("FIX.5.0", "2230"): "2230 LegComplexEventPriceBoundaryPrecision",
        ("FIX.5.0", "2231"): "2231 LegComplexEventPriceTimeType",
        ("FIX.5.0", "2232"): "2232 LegComplexEventCondition",
        ("FIX.5.0", "2233"): "2233 LegComplexEventCurrencyOne",
        ("FIX.5.0", "2234"): "2234 LegComplexEventCurrencyTwo",
        ("FIX.5.0", "2235"): "2235 LegComplexEventQuoteBasis",
        ("FIX.5.0", "2236"): "2236 LegComplexEventFixedFXRate",
        ("FIX.5.0", "2237"): "2237 LegComplexEventDeterminationMethod",
        ("FIX.5.0", "2238"): "2238 LegComplexEventCalculationAgent",
        ("FIX.5.0", "2239"): "2239 LegComplexEventStrikePrice",
        ("FIX.5.0", "2240"): "2240 LegComplexEventStrikeFactor",
        ("FIX.5.0", "2241"): "2241 LegComplexEventStrikeNumberOfOptions",
        ("FIX.5.0", "2242"): "2242 LegComplexEventCreditEventsXIDRef",
        ("FIX.5.0", "2243"): "2243 LegComplexEventCreditEventNotifyingParty",
        ("FIX.5.0", "2244"): "2244 LegComplexEventCreditEventBusinessCenter",
        ("FIX.5.0", "2245"): "2245 LegComplexEventCreditEventStandardSources",
        ("FIX.5.0", "2246"): "2246 LegComplexEventCreditEventMinimumSources",
        ("FIX.5.0", "2247"): "2247 LegComplexEventEndTime",
        ("FIX.5.0", "2248"): "2248 LegComplexEventXID",
        ("FIX.5.0", "2249"): "2249 LegComplexEventXIDRef",
        ("FIX.5.0", "2250"): "2250 NoLegComplexEventDates",
        ("FIX.5.0", "2251"): "2251 LegComplexEventStartDate",
        ("FIX.5.0", "2252"): "2252 LegComplexEventEndDate",
        ("FIX.5.0", "2253"): "2253 NoLegComplexEventTimes",
        ("FIX.5.0", "2254"): "2254 NoLegInstrumentParties",
        ("FIX.5.0", "2255"): "2255 LegInstrumentPartyID",
        ("FIX.5.0", "2256"): "2256 LegInstrumentPartyIDSource",
        ("FIX.5.0", "2257"): "2257 LegInstrumentPartyRole",
        ("FIX.5.0", "2258"): "2258 NoLegInstrumentPartySubIDs",
        ("FIX.5.0", "2259"): "2259 LegInstrumentPartySubID",
        ("FIX.5.0", "2260"): "2260 LegInstrumentPartySubIDType",
        ("FIX.5.0", "2261"): "2261 UnderlyingComplexOptPayoutPaySide",
        ("FIX.5.0", "2262"): "2262 UnderlyingComplexOptPayoutReceiveSide",
        ("FIX.5.0", "2263"): "2263 UnderlyingComplexOptPayoutUnderlier",
        ("FIX.5.0", "2264"): "2264 UnderlyingComplexOptPayoutPercentage",
        ("FIX.5.0", "2265"): "2265 UnderlyingComplexOptPayoutTime",
        ("FIX.5.0", "2266"): "2266 UnderlyingComplexOptPayoutCurrency",
        ("FIX.5.0", "2267"): "2267 UnderlyingComplexEventPricePercentage",
        ("FIX.5.0", "2268"): "2268 UnderlyingComplexEventCurrencyOne",
        ("FIX.5.0", "2269"): "2269 UnderlyingComplexEventCurrencyTwo",
        ("FIX.5.0", "2270"): "2270 UnderlyingComplexEventQuoteBasis",
        ("FIX.5.0", "2271"): "2271 UnderlyingComplexEventFixedFXRate",
        ("FIX.5.0", "2272"): "2272 UnderlyingComplexEventDeterminationMethod",
        ("FIX.5.0", "2273"): "2273 UnderlyingComplexEventCalculationAgent",
        ("FIX.5.0", "2274"): "2274 UnderlyingComplexEventStrikePrice",
        ("FIX.5.0", "2275"): "2275 UnderlyingComplexEventStrikeFactor",
        ("FIX.5.0", "2276"): "2276 UnderlyingComplexEventStrikeNumberOfOptions",
        ("FIX.5.0", "2277"): "2277 UnderlyingComplexEventCreditEventsXIDRef",
        ("FIX.5.0", "2278"): "2278 UnderlyingComplexEventCreditEventNotifyingParty",
        ("FIX.5.0", "2279"): "2279 UnderlyingComplexEventCreditEventBusinessCenter",
        ("FIX.5.0", "2280"): "2280 UnderlyingComplexEventCreditEventStandardSources",
        ("FIX.5.0", "2281"): "2281 UnderlyingComplexEventCreditEventMinimumSources",
        ("FIX.5.0", "2282"): "2282 UnderlyingComplexEventXID",
        ("FIX.5.0", "2283"): "2283 UnderlyingComplexEventXIDRef",
        ("FIX.5.0", "2284"): "2284 UnderlyingSettlRateIndex",
        ("FIX.5.0", "2285"): "2285 UnderlyingSettlRateIndexLocation",
        ("FIX.5.0", "2286"): "2286 UnderlyingOptionExpirationDesc",
        ("FIX.5.0", "2287"): "2287 EncodedUnderlyingOptionExpirationDescLen",
        ("FIX.5.0", "2288"): "2288 EncodedUnderlyingOptionExpirationDesc",
        ("FIX.5.0", "2289"): "2289 UnderlyingSwapSubClass",
        ("FIX.5.0", "2290"): "2290 UnderlyingStrikeUnitOfMeasure",
        ("FIX.5.0", "2291"): "2291 UnderlyingStrikeIndex",
        ("FIX.5.0", "2292"): "2292 UnderlyingStrikeIndexSpread",
        ("FIX.5.0", "2293"): "2293 UnderlyingValuationSource",
        ("FIX.5.0", "2294"): "2294 UnderlyingValuationReferenceModel",
        ("FIX.5.0", "2295"): "2295 UnderlyingStrategyType",
        ("FIX.5.0", "2296"): "2296 UnderlyingCommonPricingIndicator",
        ("FIX.5.0", "2297"): "2297 UnderlyingSettlDisruptionProvision",
        ("FIX.5.0", "2298"): "2298 UnderlyingInstrumentRoundingDirection",
        ("FIX.5.0", "2299"): "2299 UnderlyingInstrumentRoundingPrecision",
        ("FIX.5.0", "2300"): "2300 AllocGrossTradeAmt",
        ("FIX.5.0", "2301"): "2301 LastQtyChanged",
        ("FIX.5.0", "2302"): "2302 TradeVersion",
        ("FIX.5.0", "2303"): "2303 HistoricalReportIndicator",
        ("FIX.5.0", "2304"): "2304 NoAssetAttributes",
        ("FIX.5.0", "2305"): "2305 AssetAttributeType",
        ("FIX.5.0", "2306"): "2306 AssetAttributeValue",
        ("FIX.5.0", "2307"): "2307 AssetAttributeLimit",
        ("FIX.5.0", "2308"): "2308 NoLegAssetAttributes",
        ("FIX.5.0", "2309"): "2309 LegAssetAttributeType",
        ("FIX.5.0", "2310"): "2310 LegAssetAttributeValue",
        ("FIX.5.0", "2311"): "2311 LegAssetAttributeLimit",
        ("FIX.5.0", "2312"): "2312 NoUnderlyingAssetAttributes",
        ("FIX.5.0", "2313"): "2313 UnderlyingAssetAttributeType",
        ("FIX.5.0", "2314"): "2314 UnderlyingAssetAttributeValue",
        ("FIX.5.0", "2315"): "2315 UnderlyingAssetAttributeLimit",
        ("FIX.5.0", "2316"): "2316 RiskLimitReportStatus",
        ("FIX.5.0", "2317"): "2317 RiskLimitReportRejectReason",
        ("FIX.5.0", "2318"): "2318 RiskLimitCheckRequestID",
        ("FIX.5.0", "2319"): "2319 RiskLimitCheckID",
        ("FIX.5.0", "2320"): "2320 RiskLimitCheckTransType",
        ("FIX.5.0", "2321"): "2321 RiskLimitCheckType",
        ("FIX.5.0", "2322"): "2322 RiskLimitCheckRequestRefID",
        ("FIX.5.0", "2323"): "2323 RiskLimitCheckRequestType",
        ("FIX.5.0", "2324"): "2324 RiskLimitCheckAmount",
        ("FIX.5.0", "2325"): "2325 RiskLimitCheckRequestStatus",
        ("FIX.5.0", "2326"): "2326 RiskLimitCheckRequestResult",
        ("FIX.5.0", "2327"): "2327 RiskLimitApprovedAmount",
        ("FIX.5.0", "2328"): "2328 PartyActionRequestID",
        ("FIX.5.0", "2329"): "2329 PartyActionType",
        ("FIX.5.0", "2330"): "2330 ApplTestMessageIndicator",
        ("FIX.5.0", "2331"): "2331 PartyActionReportID",
        ("FIX.5.0", "2332"): "2332 PartyActionResponse",
        ("FIX.5.0", "2333"): "2333 PartyActionRejectReason",
        ("FIX.5.0", "2334"): "2334 RefRiskLimitCheckID",
        ("FIX.5.0", "2335"): "2335 RefRiskLimitCheckIDType",
        ("FIX.5.0", "2336"): "2336 RiskLimitVelocityPeriod",
        ("FIX.5.0", "2337"): "2337 RiskLimitVelocityUnit",
        ("FIX.5.0", "2338"): "2338 RequestingPartyRoleQualifier",
        ("FIX.5.0", "2339"): "2339 RiskLimitCheckModelType",
        ("FIX.5.0", "2340"): "2340 EventMonthYear",
        ("FIX.5.0", "2341"): "2341 LegEventMonthYear",
        ("FIX.5.0", "2342"): "2342 UnderlyingEventMonthYear",
        ("FIX.5.0", "2343"): "2343 RiskLimitCheckStatus",
        ("FIX.5.0", "2344"): "2344 SideRiskLimitCheckStatus",
        ("FIX.5.0", "2345"): "2345 NoEntitlementTypes",
        ("FIX.5.0", "2346"): "2346 LegMidPx",
        ("FIX.5.0", "2347"): "2347 RegulatoryTransactionType",
        ("FIX.5.0", "2348"): "2348 LegAssetGroup",
        ("FIX.5.0", "2349"): "2349 PricePrecision",
        ("FIX.5.0", "2350"): "2350 CollateralPortfolioID",
        ("FIX.5.0", "2351"): "2351 EncodedComplianceTextLen",
        ("FIX.5.0", "2352"): "2352 EncodedComplianceText",
        ("FIX.5.0", "2353"): "2353 TradingUnitPeriodMultiplier",
        ("FIX.5.0", "2354"): "2354 LegTradingUnitPeriodMultiplier",
        ("FIX.5.0", "2355"): "2355 PartyRiskLimitStatus",
        ("FIX.5.0", "2356"): "2356 RemunerationIndicator",
        ("FIX.5.0", "2357"): "2357 LegTotalTradeQty",
        ("FIX.5.0", "2358"): "2358 LegLastMultipliedQty",
        ("FIX.5.0", "2359"): "2359 LegTotalGrossTradeAmt",
        ("FIX.5.0", "2360"): "2360 LegTotalTradeMultipliedQty",
        ("FIX.5.0", "2361"): "2361 CompressionGroupID",
        ("FIX.5.0", "2362"): "2362 SelfMatchPreventionID",
        ("FIX.5.0", "2363"): "2363 UnderlyingTradingUnitPeriodMultiplier",
        ("FIX.5.0", "2364"): "2364 PosReportAction",
        ("FIX.5.0", "2365"): "2365 SettlForwardPoints",
        ("FIX.5.0", "2366"): "2366 SettlPriceFxRateCalc",
        ("FIX.5.0", "2367"): "2367 TotalTradeQty",
        ("FIX.5.0", "2368"): "2368 LastMultipliedQty",
        ("FIX.5.0", "2369"): "2369 TotalGrossTradeAmt",
        ("FIX.5.0", "2370"): "2370 TotalTradeMultipliedQty",
        ("FIX.5.0", "2371"): "2371 EncodedTradeContinuationText",
        ("FIX.5.0", "2372"): "2372 EncodedTradeContinuationTextLen",
        ("FIX.5.0", "2373"): "2373 IntraFirmTradeIndicator",
        ("FIX.5.0", "2374"): "2374 TradeContinuationText",
        ("FIX.5.0", "2375"): "2375 TaxonomyType",
        ("FIX.5.0", "2376"): "2376 PartyRoleQualifier",
        ("FIX.5.0", "2377"): "2377 DerivativeInstrumentPartyRoleQualifier",
        ("FIX.5.0", "2378"): "2378 InstrumentPartyRoleQualifier",
        ("FIX.5.0", "2379"): "2379 LegInstrumentPartyRoleQualifier",
        ("FIX.5.0", "2380"): "2380 LegProvisionPartyRoleQualifier",
        ("FIX.5.0", "2381"): "2381 Nested2PartyRoleQualifier",
        ("FIX.5.0", "2382"): "2382 Nested3PartyRoleQualifier",
        ("FIX.5.0", "2383"): "2383 Nested4PartyRoleQualifier",
        ("FIX.5.0", "2384"): "2384 NestedPartyRoleQualifier",
        ("FIX.5.0", "2385"): "2385 ProvisionPartyRoleQualifier",
        ("FIX.5.0", "2386"): "2386 RequestedPartyRoleQualifier",
        ("FIX.5.0", "2387"): "2387 TradeContingency",
        ("FIX.5.0", "2388"): "2388 RootPartyRoleQualifier",
        ("FIX.5.0", "2389"): "2389 SettlPartyRoleQualifier",
        ("FIX.5.0", "2390"): "2390 TradeConfirmationReferenceID",
        ("FIX.5.0", "2391"): "2391 UnderlyingInstrumentPartyRoleQualifier",
        ("FIX.5.0", "2392"): "2392 AllocRefRiskLimitCheckID",
        ("FIX.5.0", "2393"): "2393 AllocRefRiskLimitCheckIDType",
        ("FIX.5.0", "2394"): "2394 LimitUtilizationAmt",
        ("FIX.5.0", "2395"): "2395 LimitAmt",
        ("FIX.5.0", "2396"): "2396 LimitRole",
        ("FIX.5.0", "2397"): "2397 RegulatoryTradeIDScope",
        ("FIX.5.0", "2398"): "2398 SideRegulatoryTradeIDScope",
        ("FIX.5.0", "2399"): "2399 AllocRegulatoryTradeIDScope",
        ("FIX.5.0", "2400"): "2400 EffectiveBusinessDate",
        ("FIX.5.0", "2401"): "2401 ListManualOrderIndicator",
        ("FIX.5.0", "2402"): "2402 EntitlementSubType",
        ("FIX.5.0", "2403"): "2403 QuoteModelType",
        ("FIX.5.0", "2404"): "2404 ComplianceText",
        ("FIX.5.0", "2405"): "2405 ExecMethod",
        ("FIX.5.0", "2406"): "2406 AllocRegulatoryLegRefID",
        ("FIX.5.0", "2407"): "2407 ComplexEventSpotRate",
        ("FIX.5.0", "2408"): "2408 ComplexEventForwardPoints",
        ("FIX.5.0", "2409"): "2409 LegComplexEventSpotRate",
        ("FIX.5.0", "2410"): "2410 LegComplexEventForwardPoints",
        ("FIX.5.0", "2411"): "2411 RegulatoryLegRefID",
        ("FIX.5.0", "2412"): "2412 RateSourceReferemcePageHeading",
        ("FIX.5.0", "2413"): "2413 RelatedToSecurityID",
        ("FIX.5.0", "2414"): "2414 RelatedToSecurityIDSource",
        ("FIX.5.0", "2415"): "2415 RelatedToStreamXIDRef",
        ("FIX.5.0", "2416"): "2416 SideRegulatoryLegRefID",
        ("FIX.5.0", "2417"): "2417 RelatedToDividendPeriodXIDRef",
        ("FIX.5.0", "2418"): "2418 FirmTradeEventID",
        ("FIX.5.0", "2419"): "2419 UnderlyingComplexEventSpotRate",
        ("FIX.5.0", "2420"): "2420 UnderlyingComplexEventForwardPoints",
        ("FIX.5.0", "2421"): "2421 FillRefID",
        ("FIX.5.0", "2422"): "2422 OrderRequestID",
        ("FIX.5.0", "2423"): "2423 MassOrderRequestID",
        ("FIX.5.0", "2424"): "2424 MassOrderReportID",
        ("FIX.5.0", "2425"): "2425 MassOrderRequestStatus",
        ("FIX.5.0", "2426"): "2426 MassOrderRequestResult",
        ("FIX.5.0", "2427"): "2427 OrderResponseLevel",
        ("FIX.5.0", "2428"): "2428 NoOrderEntries",
        ("FIX.5.0", "2429"): "2429 OrderEntryAction",
        ("FIX.5.0", "2430"): "2430 OrderEntryID",
        ("FIX.5.0", "2431"): "2431 ExecTypeReason",
        ("FIX.5.0", "2432"): "2432 TotNoOrderEntries",
        ("FIX.5.0", "2433"): "2433 NoTargetPartySubIDs",
        ("FIX.5.0", "2434"): "2434 TargetPartySubID",
        ("FIX.5.0", "2435"): "2435 TargetPartySubIDType",
        ("FIX.5.0", "2436"): "2436 TransferInstructionID",
        ("FIX.5.0", "2437"): "2437 TransferID",
        ("FIX.5.0", "2438"): "2438 TransferReportID",
        ("FIX.5.0", "2439"): "2439 TransferTransType",
        ("FIX.5.0", "2440"): "2440 TransferType",
        ("FIX.5.0", "2441"): "2441 TransferScope",
        ("FIX.5.0", "2442"): "2442 TransferStatus",
        ("FIX.5.0", "2443"): "2443 TransferRejectReason",
        ("FIX.5.0", "2444"): "2444 TransferReportType",
        ("FIX.5.0", "2445"): "2445 AggressorTime",
        ("FIX.5.0", "2446"): "2446 AggressorSide",
        ("FIX.5.0", "2447"): "2447 FastMarketIndicator",
        ("FIX.5.0", "2448"): "2448 LinkageHandlingIndicator",
        ("FIX.5.0", "2449"): "2449 NumberOfBuyOrders",
        ("FIX.5.0", "2450"): "2450 NumberOfSellOrders",
        ("FIX.5.0", "2451"): "2451 SettlPriceDeterminationMethod",
        ("FIX.5.0", "2452"): "2452 MDStatisticReqID",
        ("FIX.5.0", "2453"): "2453 MDStatisticRptID",
        ("FIX.5.0", "2454"): "2454 MDStatisticName",
        ("FIX.5.0", "2455"): "2455 MDStatisticDesc",
        ("FIX.5.0", "2456"): "2456 MDStatisticType",
        ("FIX.5.0", "2457"): "2457 MDStatisticScope",
        ("FIX.5.0", "2458"): "2458 MDStatisticSubScope",
        ("FIX.5.0", "2459"): "2459 MDStatisticScopeType",
        ("FIX.5.0", "2460"): "2460 MDStatisticFrequencyPeriod",
        ("FIX.5.0", "2461"): "2461 MDStatisticFrequencyUnit",
        ("FIX.5.0", "2462"): "2462 MDStatisticDelayPeriod",
        ("FIX.5.0", "2463"): "2463 MDStatisticDelayUnit",
        ("FIX.5.0", "2464"): "2464 MDStatisticIntervalType",
        ("FIX.5.0", "2465"): "2465 MDStatisticIntervalTypeUnit",
        ("FIX.5.0", "2466"): "2466 MDStatisticIntervalPeriod",
        ("FIX.5.0", "2467"): "2467 MDStatisticIntervalUnit",
        ("FIX.5.0", "2468"): "2468 MDStatisticStartDate",
        ("FIX.5.0", "2469"): "2469 MDStatisticEndDate",
        ("FIX.5.0", "2470"): "2470 MDStatisticStartTime",
        ("FIX.5.0", "2471"): "2471 MDStatisticEndTime",
        ("FIX.5.0", "2472"): "2472 MDStatisticRatioType",
        ("FIX.5.0", "2473"): "2473 MDStatisticRequestResult",
        ("FIX.5.0", "2474"): "2474 NoMDStatistics",
        ("FIX.5.0", "2475"): "2475 MDStatisticID",
        ("FIX.5.0", "2476"): "2476 MDStatisticTime",
        ("FIX.5.0", "2477"): "2477 MDStatisticStatus",
        ("FIX.5.0", "2478"): "2478 MDStatisticValue",
        ("FIX.5.0", "2479"): "2479 MDStatisticValueType",
        ("FIX.5.0", "2480"): "2480 MDStatisticValueUnit",
        ("FIX.5.0", "2481"): "2481 EncodedMDStatisticDescLen",
        ("FIX.5.0", "2482"): "2482 EncodedMDStatisticDesc",
        ("FIX.5.0", "2483"): "2483 AllocRiskLimitCheckStatus",
        ("FIX.5.0", "2484"): "2484 FirmTransactionID",
        ("FIX.5.0", "2485"): "2485 TransactionID",
        ("FIX.5.0", "2486"): "2486 WireReference",
        ("FIX.5.0", "2487"): "2487 CollRptRejectReason",
        ("FIX.5.0", "2488"): "2488 CollRptStatus",
        ("FIX.5.0", "2489"): "2489 PackageID",
        ("FIX.5.0", "2490"): "2490 TradeNumber",
        ("FIX.5.0", "2491"): "2491 UnderlyingAssetGroup",
        ("FIX.5.0", "2492"): "2492 LegDifferentialPrice",
        ("FIX.5.0", "2493"): "2493 EncodedLegDocumentationText",
        ("FIX.5.0", "2494"): "2494 EncodedLegDocumentationTextLen",
        ("FIX.5.0", "2495"): "2495 LegAgreementCurrency",
        ("FIX.5.0", "2496"): "2496 LegAgreementDate",
        ("FIX.5.0", "2497"): "2497 LegAgreementDesc",
        ("FIX.5.0", "2498"): "2498 LegAgreementID",
        ("FIX.5.0", "2499"): "2499 LegAgreementVersion",
        ("FIX.5.0", "2500"): "2500 LegBrokerConfirmationDesc",
        ("FIX.5.0", "2501"): "2501 LegCreditSupportAgreementDate",
        ("FIX.5.0", "2502"): "2502 LegCreditSupportAgreementDesc",
        ("FIX.5.0", "2503"): "2503 LegCreditSupportAgreementID",
        ("FIX.5.0", "2504"): "2504 LegDeliveryType",
        ("FIX.5.0", "2505"): "2505 LegDocumentationText",
        ("FIX.5.0", "2506"): "2506 LegEndDate",
        ("FIX.5.0", "2507"): "2507 LegGoverningLaw",
        ("FIX.5.0", "2508"): "2508 LegMarginRatio",
        ("FIX.5.0", "2509"): "2509 LegMasterConfirmationAnnexDate",
        ("FIX.5.0", "2510"): "2510 LegMasterConfirmationDate",
        ("FIX.5.0", "2511"): "2511 LegMasterConfirmationDesc",
        ("FIX.5.0", "2512"): "2512 LegMasterConfirmationAnnexDesc",
        ("FIX.5.0", "2513"): "2513 LegStartDate",
        ("FIX.5.0", "2514"): "2514 LegTerminationType",
        ("FIX.5.0", "2515"): "2515 AllocCalculatedCcyQty",
        ("FIX.5.0", "2516"): "2516 CollateralRequestInstruction",
        ("FIX.5.0", "2517"): "2517 CollateralRequestLinkID",
        ("FIX.5.0", "2518"): "2518 CollateralRequestNumber",
        ("FIX.5.0", "2519"): "2519 TotNumCollateralRequests",
        ("FIX.5.0", "2520"): "2520 WarningText",
        ("FIX.5.0", "2521"): "2521 EncodedWarningText",
        ("FIX.5.0", "2522"): "2522 EncodedWarningTextLen",
        ("FIX.5.0", "2523"): "2523 CrossedIndicator",
        ("FIX.5.0", "2524"): "2524 TradeReportingIndicator",
        ("FIX.5.0", "2525"): "2525 AffiliatedFirmsTradeIndicator",
        ("FIX.5.0", "2526"): "2526 InternationalSwapIndicator",
        ("FIX.5.0", "2527"): "2527 MultiAssetSwapIndicator",
        ("FIX.5.0", "2528"): "2528 ClearingSettlPrice",
        ("FIX.5.0", "2529"): "2529 NoRelativeValues",
        ("FIX.5.0", "2530"): "2530 RelativeValueType",
        ("FIX.5.0", "2531"): "2531 RelativeValue",
        ("FIX.5.0", "2532"): "2532 RelativeValueSide",
        ("FIX.5.0", "2533"): "2533 BidSpread",
        ("FIX.5.0", "2534"): "2534 OfferSpread",
        ("FIX.5.0", "2535"): "2535 MDReportEvent",
        ("FIX.5.0", "2536"): "2536 MDReportCount",
        ("FIX.5.0", "2537"): "2537 TotNoMarketSegmentReports",
        ("FIX.5.0", "2538"): "2538 TotNoInstrumentReports",
        ("FIX.5.0", "2539"): "2539 TotNoPartyDetailReports",
        ("FIX.5.0", "2540"): "2540 TotNoEntitlementReports",
        ("FIX.5.0", "2541"): "2541 TotNoRiskLimitReports",
        ("FIX.5.0", "2542"): "2542 MarketSegmentStatus",
        ("FIX.5.0", "2543"): "2543 MarketSegmentType",
        ("FIX.5.0", "2544"): "2544 MarketSegmentSubType",
        ("FIX.5.0", "2545"): "2545 NoRelatedMarketSegments",
        ("FIX.5.0", "2546"): "2546 RelatedMarketSegmentID",
        ("FIX.5.0", "2547"): "2547 MarketSegmentRelationship",
        ("FIX.5.0", "2548"): "2548 NoAuctionTypeRules",
        ("FIX.5.0", "2549"): "2549 AuctionTypeProductComplex",
        ("FIX.5.0", "2550"): "2550 NoPriceRangeRules",
        ("FIX.5.0", "2551"): "2551 StartPriceRange",
        ("FIX.5.0", "2552"): "2552 EndPriceRange",
        ("FIX.5.0", "2553"): "2553 PriceRangeValue",
        ("FIX.5.0", "2554"): "2554 PriceRangePercentage",
        ("FIX.5.0", "2555"): "2555 PriceRangeProductComplex",
        ("FIX.5.0", "2556"): "2556 PriceRangeRuleID",
        ("FIX.5.0", "2557"): "2557 FastMarketPercentage",
        ("FIX.5.0", "2558"): "2558 NoQuoteSizeRules",
        ("FIX.5.0", "2559"): "2559 QuoteSideIndicator",
        ("FIX.5.0", "2560"): "2560 NoFlexProductEligibilities",
        ("FIX.5.0", "2561"): "2561 FlexProductEligibilityComplex",
        ("FIX.5.0", "2562"): "2562 NumOfComplexInstruments",
        ("FIX.5.0", "2563"): "2563 MarketDepthTimeInterval",
        ("FIX.5.0", "2564"): "2564 MarketDepthTimeIntervalUnit",
        ("FIX.5.0", "2565"): "2565 MDRecoveryTimeInterval",
        ("FIX.5.0", "2566"): "2566 MDRecoveryTimeIntervalUnit",
        ("FIX.5.0", "2567"): "2567 PrimaryServiceLocationID",
        ("FIX.5.0", "2568"): "2568 SecondaryServiceLocationID",
        ("FIX.5.0", "2569"): "2569 MatchRuleProductComplex",
        ("FIX.5.0", "2570"): "2570 CustomerPriority",
        ("FIX.5.0", "2571"): "2571 TickRuleProductComplex",
        ("FIX.5.0", "2572"): "2572 PreviousAdjustedOpenInterest",
        ("FIX.5.0", "2573"): "2573 PreviousUnadjustedOpenInterest",
        ("FIX.5.0", "2574"): "2574 LowExercisePriceOptionIndicator",
        ("FIX.5.0", "2575"): "2575 BlockTradeEligibilityIndicator",
        ("FIX.5.0", "2576"): "2576 InstrumentPricePrecision",
        ("FIX.5.0", "2577"): "2577 StrikePricePrecision",
        ("FIX.5.0", "2578"): "2578 OrigStrikePrice",
        ("FIX.5.0", "2579"): "2579 SettlSubMethod",
        ("FIX.5.0", "2580"): "2580 NoClearingPriceParameters",
        ("FIX.5.0", "2581"): "2581 BusinessDayType",
        ("FIX.5.0", "2582"): "2582 ClearingPriceOffset",
        ("FIX.5.0", "2583"): "2583 VegaMultiplier",
        ("FIX.5.0", "2584"): "2584 AnnualTradingBusinessDays",
        ("FIX.5.0", "2585"): "2585 TotalTradingBusinessDays",
        ("FIX.5.0", "2586"): "2586 TradingBusinessDays",
        ("FIX.5.0", "2587"): "2587 RealizedVariance",
        ("FIX.5.0", "2588"): "2588 StandardVariance",
        ("FIX.5.0", "2589"): "2589 RelatedClosePrice",
        ("FIX.5.0", "2590"): "2590 OvernightInterestRate",
        ("FIX.5.0", "2591"): "2591 AccumulatedReturnModifiedVariationMargin",
        ("FIX.5.0", "2592"): "2592 CalculationMethod",
        ("FIX.5.0", "2593"): "2593 NoOrderAttributes",
        ("FIX.5.0", "2594"): "2594 OrderAttributeType",
        ("FIX.5.0", "2595"): "2595 OrderAttributeValue",
        ("FIX.5.0", "2596"): "2596 DeltaCrossed",
        ("FIX.5.0", "2597"): "2597 ComplexEventFuturesPriceValuation",
        ("FIX.5.0", "2598"): "2598 ComplexEventOptionsPriceValuation",
        ("FIX.5.0", "2599"): "2599 ComplexEventPVFinalPriceElectionFallback",
        ("FIX.5.0", "2600"): "2600 StrikeIndexCurvePoint",
        ("FIX.5.0", "2601"): "2601 StrikeIndexQuote",
        ("FIX.5.0", "2602"): "2602 ExtraordinaryEventAdjustmentMethod",
        ("FIX.5.0", "2603"): "2603 ExchangeLookAlike",
        ("FIX.5.0", "2604"): "2604 LegStrikeIndexCurvePoint",
        ("FIX.5.0", "2605"): "2605 LegStrikeIndexQuote",
        ("FIX.5.0", "2606"): "2606 LegExtraordinaryEventAdjustmentMethod",
        ("FIX.5.0", "2607"): "2607 LegExchangeLookAlike",
        ("FIX.5.0", "2608"): "2608 LegComplexEventFuturesPriceValuation",
        ("FIX.5.0", "2609"): "2609 LegComplexEventOptionsPriceValuation",
        ("FIX.5.0", "2610"): "2610 LegComplexEventPVFinalPriceElectionFallback",
        ("FIX.5.0", "2611"): "2611 UnderlyingComplexEventFuturesPriceValuation",
        ("FIX.5.0", "2612"): "2612 UnderlyingComplexEventOptionsPriceValuation",
        ("FIX.5.0", "2613"): "2613 UnderlyingComplexEventPVFinalPriceElectionFallback",
        ("FIX.5.0", "2614"): "2614 UnderlyingNotional",
        ("FIX.5.0", "2615"): "2615 UnderlyingNotionalCurrency",
        ("FIX.5.0", "2616"): "2616 UnderlyingNotionalDeterminationMethod",
        ("FIX.5.0", "2617"): "2617 UnderlyingNotionalAdjustments",
        ("FIX.5.0", "2618"): "2618 PositionID",
        ("FIX.5.0", "2619"): "2619 UnderlyingNotionalXIDRef",
        ("FIX.5.0", "2620"): "2620 UnderlyingFutureID",
        ("FIX.5.0", "2621"): "2621 UnderlyingFutureIDSource",
        ("FIX.5.0", "2622"): "2622 UnderlyingStrikeIndexCurvePoint",
        ("FIX.5.0", "2623"): "2623 UnderlyingStrikeIndexQuote",
        ("FIX.5.0", "2624"): "2624 UnderlyingExtraordinaryEventAdjustmentMethod",
        ("FIX.5.0", "2625"): "2625 UnderlyingExchangeLookAlike",
        ("FIX.5.0", "2626"): "2626 UnderlyingAverageVolumeLimitationPercentage",
        ("FIX.5.0", "2627"): "2627 UnderlyingAverageVolumeLimitationPeriodDays",
        ("FIX.5.0", "2628"): "2628 UnderlyingDepositoryReceiptIndicator",
        ("FIX.5.0", "2629"): "2629 UnderlyingOpenUnits",
        ("FIX.5.0", "2630"): "2630 UnderlyingBasketDivisor",
        ("FIX.5.0", "2631"): "2631 UnderlyingInstrumentXID",
        ("FIX.5.0", "2632"): "2632 CollateralAmountType",
        ("FIX.5.0", "2633"): "2633 NoMiscFeeSubTypes",
        ("FIX.5.0", "2634"): "2634 MiscFeeSubType",
        ("FIX.5.0", "2635"): "2635 MiscFeeSubTypeAmt",
        ("FIX.5.0", "2636"): "2636 MiscFeeSubTypeDesc",
        ("FIX.5.0", "2637"): "2637 EncodedMiscFeeSubTypeDescLen",
        ("FIX.5.0", "2638"): "2638 EncodedMiscFeeSubTypeDesc",
        ("FIX.5.0", "2639"): "2639 NoCommissions",
        ("FIX.5.0", "2640"): "2640 CommissionAmount",
        ("FIX.5.0", "2641"): "2641 CommissionAmountType",
        ("FIX.5.0", "2642"): "2642 CommissionBasis",
        ("FIX.5.0", "2643"): "2643 CommissionCurrency",
        ("FIX.5.0", "2644"): "2644 CommissionUnitOfMeasure",
        ("FIX.5.0", "2645"): "2645 CommissionUnitOfMeasureCurrency",
        ("FIX.5.0", "2646"): "2646 CommissionRate",
        ("FIX.5.0", "2647"): "2647 CommissionSharedIndicator",
        ("FIX.5.0", "2648"): "2648 CommissionAmountShared",
        ("FIX.5.0", "2649"): "2649 CommissionLegRefID",
        ("FIX.5.0", "2650"): "2650 CommissionDesc",
        ("FIX.5.0", "2651"): "2651 EncodedCommissionDescLen",
        ("FIX.5.0", "2652"): "2652 EncodedCommissionDesc",
        ("FIX.5.0", "2653"): "2653 NoAllocCommissions",
        ("FIX.5.0", "2654"): "2654 AllocCommissionAmount",
        ("FIX.5.0", "2655"): "2655 AllocCommissionAmountType",
        ("FIX.5.0", "2656"): "2656 AllocCommissionBasis",
        ("FIX.5.0", "2657"): "2657 AllocCommissionCurrency",
        ("FIX.5.0", "2658"): "2658 AllocCommissionUnitOfMeasure",
        ("FIX.5.0", "2659"): "2659 AllocCommissionUnitOfMeasureCurrency",
        ("FIX.5.0", "2660"): "2660 AllocCommissionRate",
        ("FIX.5.0", "2661"): "2661 AllocCommissionSharedIndicator",
        ("FIX.5.0", "2662"): "2662 AllocCommissionAmountShared",
        ("FIX.5.0", "2663"): "2663 AllocCommissionLegRefID",
        ("FIX.5.0", "2664"): "2664 AllocCommissionDesc",
        ("FIX.5.0", "2665"): "2665 EncodedAllocCommissionDescLen",
        ("FIX.5.0", "2666"): "2666 EncodedAllocCommissionDesc",
        ("FIX.5.0", "2667"): "2667 AlgorithmicTradeIndicator",
        ("FIX.5.0", "2668"): "2668 NoTrdRegPublications",
        ("FIX.5.0", "2669"): "2669 TrdRegPublicationType",
        ("FIX.5.0", "2670"): "2670 TrdRegPublicationReason",
        ("FIX.5.0", "2671"): "2671 SideTradeReportingIndicator",
        ("FIX.5.0", "2672"): "2672 CrossRequestID",
        ("FIX.5.0", "2673"): "2673 FillMatchID",
        ("FIX.5.0", "2674"): "2674 FillMatchSubID",
        ("FIX.5.0", "2675"): "2675 MassActionReason",
        ("FIX.5.0", "2676"): "2676 MaximumPricePercentage",
        ("FIX.5.0", "2677"): "2677 NotAffectedReason",
        ("FIX.5.0", "2678"): "2678 TotalNotAffectedOrders",
        ("FIX.5.0", "2679"): "2679 OrderOwnershipIndicator",
        ("FIX.5.0", "2680"): "2680 LegAccount",
        ("FIX.5.0", "2681"): "2681 InTheMoneyCondition",
        ("FIX.5.0", "2682"): "2682 LegInTheMoneyCondition",
        ("FIX.5.0", "2683"): "2683 UnderlyingInTheMoneyCondition",
        ("FIX.5.0", "2684"): "2684 DerivativeInTheMoneyCondition",
        ("FIX.5.0", "2685"): "2685 ContraryInstructionEligibilityIndicator",
        ("FIX.5.0", "2686"): "2686 LegContraryInstructionEligibilityIndicator",
        ("FIX.5.0", "2687"): "2687 UnderlyingContraryInstructionEligibilityIndicator",
        ("FIX.5.0", "2688"): "2688 DerivativeContraryInstructionEligibilityIndicator",
        ("FIX.5.0", "2689"): "2689 CollateralMarketPrice",
        ("FIX.5.0", "2690"): "2690 CollateralPercentOverage",
        ("FIX.5.0", "2691"): "2691 NoSideCollateralAmounts",
        ("FIX.5.0", "2692"): "2692 SideCollateralAmountMarketID",
        ("FIX.5.0", "2693"): "2693 SideCollateralAmountMarketSegmentID",
        ("FIX.5.0", "2694"): "2694 SideCollateralAmountType",
        ("FIX.5.0", "2695"): "2695 SideCollateralCurrency",
        ("FIX.5.0", "2696"): "2696 SideCollateralFXRate",
        ("FIX.5.0", "2697"): "2697 SideCollateralFXRateCalc",
        ("FIX.5.0", "2698"): "2698 SideCollateralMarketPrice",
        ("FIX.5.0", "2699"): "2699 SideCollateralPercentOverage",
        ("FIX.5.0", "2700"): "2700 SideCollateralPortfolioID",
        ("FIX.5.0", "2701"): "2701 SideCollateralType",
        ("FIX.5.0", "2702"): "2702 SideCurrentCollateralAmount",
        ("FIX.5.0", "2703"): "2703 SideHaircutIndicator",
        ("FIX.5.0", "2704"): "2704 ExDestinationType",
        ("FIX.5.0", "2705"): "2705 MarketCondition",
        ("FIX.5.0", "2706"): "2706 NoQuoteAttributes",
        ("FIX.5.0", "2707"): "2707 QuoteAttributeType",
        ("FIX.5.0", "2708"): "2708 QuoteAttributeValue",
        ("FIX.5.0", "2709"): "2709 NoPriceQualifiers",
        ("FIX.5.0", "2710"): "2710 PriceQualifier",
        ("FIX.5.0", "2711"): "2711 MDValueTier",
        ("FIX.5.0", "2712"): "2712 MiscFeeQualifier",
        ("FIX.5.0", "2713"): "2713 MiscFeeDesc",
        ("FIX.5.0", "2714"): "2714 FinancialInstrumentFullName",
        ("FIX.5.0", "2715"): "2715 EncodedFinancialInstrumentFullNameLen",
        ("FIX.5.0", "2716"): "2716 EncodedFinancialInstrumentFullName",
        ("FIX.5.0", "2717"): "2717 LegFinancialInstrumentFullName",
        ("FIX.5.0", "2718"): "2718 EncodedLegFinancialInstrumentFullNameLen",
        ("FIX.5.0", "2719"): "2719 EncodedLegFinancialInstrumentFullName",
        ("FIX.5.0", "2720"): "2720 UnderlyingFinancialInstrumentFullName",
        ("FIX.5.0", "2721"): "2721 EncodedUnderlyingFinancialInstrumentFullNameLen",
        ("FIX.5.0", "2722"): "2722 EncodedUnderlyingFinancialInstrumentFullName",
        ("FIX.5.0", "2723"): "2723 UnderlyingIndexCurveUnit",
        ("FIX.5.0", "2724"): "2724 UnderlyingIndexCurvePeriod",
        ("FIX.5.0", "2725"): "2725 CommissionAmountSubType",
        ("FIX.5.0", "2726"): "2726 AllocCommissionAmountSubType",
        ("FIX.5.0", "2727"): "2727 AllocLegRefID",
        ("FIX.5.0", "2728"): "2728 FloatingRateIndexCurvePeriod",
        ("FIX.5.0", "2729"): "2729 FloatingRateIndexCurveSpread",
        ("FIX.5.0", "2730"): "2730 FloatingRateIndexCurveUnit",
        ("FIX.5.0", "2731"): "2731 FloatingRateIndexID",
        ("FIX.5.0", "2732"): "2732 FloatingRateIndexIDSource",
        ("FIX.5.0", "2733"): "2733 IndexRollMonth",
        ("FIX.5.0", "2734"): "2734 NoIndexRollMonths",
        ("FIX.5.0", "2735"): "2735 AssetSubType",
        ("FIX.5.0", "2736"): "2736 CommodityFinalPriceType",
        ("FIX.5.0", "2737"): "2737 FinancialInstrumentShortName",
        ("FIX.5.0", "2738"): "2738 NextIndexRollDate",
        ("FIX.5.0", "2739"): "2739 LegAssetSubType",
        ("FIX.5.0", "2740"): "2740 LegFinancialInstrumentShortName",
        ("FIX.5.0", "2741"): "2741 SecondaryAssetSubType",
        ("FIX.5.0", "2742"): "2742 UnderlyingFinancialInstrumentShortName",
        ("FIX.5.0", "2743"): "2743 LegSecondaryAssetSubType",
        ("FIX.5.0", "2744"): "2744 UnderlyingAssetSubType",
        ("FIX.5.0", "2745"): "2745 UnderlyingSecondaryAssetSubType",
        ("FIX.5.0", "2746"): "2746 NoReferenceDataDates",
        ("FIX.5.0", "2747"): "2747 ReferenceDataDate",
        ("FIX.5.0", "2748"): "2748 ReferenceDataDateType",
        ("FIX.5.0", "2749"): "2749 ExecutionTimestamp",
        ("FIX.5.0", "2750"): "2750 ReportingPx",
        ("FIX.5.0", "2751"): "2751 ReportingQty",
        ("FIX.5.0", "2752"): "2752 DeliveryRouteOrCharter",
        ("FIX.5.0", "2753"): "2753 ReturnTrigger",
        ("FIX.5.0", "2754"): "2754 LegDeliveryRouteOrCharter",
        ("FIX.5.0", "2755"): "2755 LegReturnTrigger",
        ("FIX.5.0", "2756"): "2756 UnderlyingDeliveryRouteOrCharter",
        ("FIX.5.0", "2757"): "2757 UnderlyingReturnTrigger",
        ("FIX.5.0", "2758"): "2758 AllocRequestID",
        ("FIX.5.0", "2759"): "2759 GroupAmount",
        ("FIX.5.0", "2760"): "2760 GroupRemainingAmount",
        ("FIX.5.0", "2761"): "2761 AllocGroupAmount",
        ("FIX.5.0", "2762"): "2762 PriceMarkup",
        ("FIX.5.0", "2763"): "2763 AveragePriceType",
        ("FIX.5.0", "2764"): "2764 AveragePriceStartTime",
        ("FIX.5.0", "2765"): "2765 AveragePriceEndTime",
        ("FIX.5.0", "2766"): "2766 OrderPercentOfTotalVolume",
        ("FIX.5.0", "2767"): "2767 AllocGroupStatus",
        ("FIX.5.0", "40000"): "40000 NoAdditionalTermBondRefs",
        ("FIX.5.0", "40001"): "40001 AdditionalTermBondSecurityID",
        ("FIX.5.0", "40002"): "40002 AdditionalTermBondSecurityIDSource",
        ("FIX.5.0", "40003"): "40003 AdditionalTermBondDesc",
        ("FIX.5.0", "40004"): "40004 EncodedAdditionalTermBondDescLen",
        ("FIX.5.0", "40005"): "40005 EncodedAdditionalTermBondDesc",
        ("FIX.5.0", "40006"): "40006 AdditionalTermBondCurrency",
        ("FIX.5.0", "40007"): "40007 AdditionalTermBondIssuer",
        ("FIX.5.0", "40008"): "40008 EncodedAdditionalTermBondIssuerLen",
        ("FIX.5.0", "40009"): "40009 EncodedAdditionalTermBondIssuer",
        ("FIX.5.0", "40010"): "40010 AdditionalTermBondSeniority",
        ("FIX.5.0", "40011"): "40011 AdditionalTermBondCouponType",
        ("FIX.5.0", "40012"): "40012 AdditionalTermBondCouponRate",
        ("FIX.5.0", "40013"): "40013 AdditionalTermBondMaturityDate",
        ("FIX.5.0", "40014"): "40014 AdditionalTermBondParValue",
        ("FIX.5.0", "40015"): "40015 AdditionalTermBondCurrentTotalIssuedAmount",
        ("FIX.5.0", "40016"): "40016 AdditionalTermBondCouponFrequencyPeriod",
        ("FIX.5.0", "40017"): "40017 AdditionalTermBondCouponFrequencyUnit",
        ("FIX.5.0", "40018"): "40018 AdditionalTermBondDayCount",
        ("FIX.5.0", "40019"): "40019 NoAdditionalTerms",
        ("FIX.5.0", "40020"): "40020 AdditionalTermConditionPrecedentBondIndicator",
        ("FIX.5.0", "40021"): "40021 AdditionalTermDiscrepancyClauseIndicator",
        ("FIX.5.0", "40022"): "40022 NoCashSettlTerms",
        ("FIX.5.0", "40023"): "40023 CashSettlCurrency",
        ("FIX.5.0", "40024"): "40024 CashSettlValuationFirstBusinessDayOffset",
        ("FIX.5.0", "40025"): "40025 CashSettlValuationTime",
        ("FIX.5.0", "40026"): "40026 CashSettlBusinessCenter",
        ("FIX.5.0", "40027"): "40027 CashSettlQuoteMethod",
        ("FIX.5.0", "40028"): "40028 CashSettlQuoteAmount",
        ("FIX.5.0", "40029"): "40029 CashSettlQuoteCurrency",
        ("FIX.5.0", "40030"): "40030 CashSettlMinimumQuoteAmount",
        ("FIX.5.0", "40031"): "40031 CashSettlMinimumQuoteCurrency",
        ("FIX.5.0", "40032"): "40032 CashSettlDealer",
        ("FIX.5.0", "40033"): "40033 CashSettlBusinessDays",
        ("FIX.5.0", "40034"): "40034 CashSettlAmount",
        ("FIX.5.0", "40035"): "40035 CashSettlRecoveryFactor",
        ("FIX.5.0", "40036"): "40036 CashSettlFixedTermIndicator",
        ("FIX.5.0", "40037"): "40037 CashSettlAccruedInterestIndicator",
        ("FIX.5.0", "40038"): "40038 CashSettlValuationMethod",
        ("FIX.5.0", "40039"): "40039 CashSettlTermXID",
        ("FIX.5.0", "40040"): "40040 NoContractualDefinitions",
        ("FIX.5.0", "40041"): "40041 ContractualDefinition",
        ("FIX.5.0", "40042"): "40042 NoContractualMatrices",
        ("FIX.5.0", "40043"): "40043 ContractualMatrixSource",
        ("FIX.5.0", "40044"): "40044 ContractualMatrixDate",
        ("FIX.5.0", "40045"): "40045 ContractualMatrixTerm",
        ("FIX.5.0", "40046"): "40046 NoFinancingTermSupplements",
        ("FIX.5.0", "40047"): "40047 FinancingTermSupplementDesc",
        ("FIX.5.0", "40048"): "40048 FinancingTermSupplementDate",
        ("FIX.5.0", "40049"): "40049 NoStreams",
        ("FIX.5.0", "40050"): "40050 StreamType",
        ("FIX.5.0", "40051"): "40051 StreamDesc",
        ("FIX.5.0", "40052"): "40052 StreamPaySide",
        ("FIX.5.0", "40053"): "40053 StreamReceiveSide",
        ("FIX.5.0", "40054"): "40054 StreamNotional",
        ("FIX.5.0", "40055"): "40055 StreamCurrency",
        ("FIX.5.0", "40056"): "40056 StreamText",
        ("FIX.5.0", "40057"): "40057 UnderlyingStreamEffectiveDateUnadjusted",
        ("FIX.5.0", "40058"): "40058 UnderlyingStreamEffectiveDateBusinessDayConvention",
        ("FIX.5.0", "40059"): "40059 UnderlyingStreamEffectiveDateBusinessCenter",
        ("FIX.5.0", "40060"): "40060 UnderlyingStreamEffectiveDateRelativeTo",
        ("FIX.5.0", "40061"): "40061 UnderlyingStreamEffectiveDateOffsetPeriod",
        ("FIX.5.0", "40062"): "40062 UnderlyingStreamEffectiveDateOffsetUnit",
        ("FIX.5.0", "40063"): "40063 UnderlyingStreamEffectiveDateOffsetDayType",
        ("FIX.5.0", "40064"): "40064 UnderlyingStreamEffectiveDateAdjusted",
        ("FIX.5.0", "40065"): "40065 StreamTerminationDateUnadjusted",
        ("FIX.5.0", "40066"): "40066 StreamTerminationDateBusinessDayConvention",
        ("FIX.5.0", "40067"): "40067 StreamTerminationDateBusinessCenter",
        ("FIX.5.0", "40068"): "40068 StreamTerminationDateRelativeTo",
        ("FIX.5.0", "40069"): "40069 StreamTerminationDateOffsetPeriod",
        ("FIX.5.0", "40070"): "40070 StreamTerminationDateOffsetUnit",
        ("FIX.5.0", "40071"): "40071 StreamTerminationDateOffsetDayType",
        ("FIX.5.0", "40072"): "40072 StreamTerminationDateAdjusted",
        ("FIX.5.0", "40073"): "40073 StreamCalculationPeriodBusinessDayConvention",
        ("FIX.5.0", "40074"): "40074 StreamCalculationPeriodBusinessCenter",
        ("FIX.5.0", "40075"): "40075 StreamFirstPeriodStartDateUnadjusted",
        ("FIX.5.0", "40076"): "40076 StreamFirstPeriodStartDateBusinessDayConvention",
        ("FIX.5.0", "40077"): "40077 StreamFirstPeriodStartDateBusinessCenter",
        ("FIX.5.0", "40078"): "40078 StreamFirstPeriodStartDateAdjusted",
        ("FIX.5.0", "40079"): "40079 StreamFirstRegularPeriodStartDateUnadjusted",
        ("FIX.5.0", "40080"): "40080 StreamFirstCompoundingPeriodEndDateUnadjusted",
        ("FIX.5.0", "40081"): "40081 StreamLastRegularPeriodEndDateUnadjusted",
        ("FIX.5.0", "40082"): "40082 StreamCalculationFrequencyPeriod",
        ("FIX.5.0", "40083"): "40083 StreamCalculationFrequencyUnit",
        ("FIX.5.0", "40084"): "40084 StreamCalculationRollConvention",
        ("FIX.5.0", "40085"): "40085 NoSettlRateFallbacks",
        ("FIX.5.0", "40086"): "40086 SettlRatePostponementMaximumDays",
        ("FIX.5.0", "40087"): "40087 LegPaymentStreamNonDeliverableSettlRateSource",
        ("FIX.5.0", "40088"): "40088 SettlRatePostponementSurvey",
        ("FIX.5.0", "40089"): "40089 SettlRatePostponementCalculationAgent",
        ("FIX.5.0", "40090"): "40090 NoProvisions",
        ("FIX.5.0", "40091"): "40091 ProvisionType",
        ("FIX.5.0", "40092"): "40092 ProvisionDateUnadjusted",
        ("FIX.5.0", "40093"): "40093 ProvisionDateBusinessDayConvention",
        ("FIX.5.0", "40094"): "40094 ProvisionDateBusinessCenter",
        ("FIX.5.0", "40095"): "40095 ProvisionDateAdjusted",
        ("FIX.5.0", "40096"): "40096 ProvisionDateTenorPeriod",
        ("FIX.5.0", "40097"): "40097 ProvisionDateTenorUnit",
        ("FIX.5.0", "40098"): "40098 ProvisionCalculationAgent",
        ("FIX.5.0", "40099"): "40099 ProvisionOptionSinglePartyBuyerSide",
        ("FIX.5.0", "40100"): "40100 ProvisionOptionSinglePartySellerSide",
        ("FIX.5.0", "40101"): "40101 ProvisionOptionExerciseStyle",
        ("FIX.5.0", "40102"): "40102 ProvisionOptionExerciseMultipleNotional",
        ("FIX.5.0", "40103"): "40103 ProvisionOptionExerciseMinimumNotional",
        ("FIX.5.0", "40104"): "40104 ProvisionOptionExerciseMaximumNotional",
        ("FIX.5.0", "40105"): "40105 ProvisionOptionMinimumNumber",
        ("FIX.5.0", "40106"): "40106 ProvisionOptionMaximumNumber",
        ("FIX.5.0", "40107"): "40107 ProvisionOptionExerciseConfirmation",
        ("FIX.5.0", "40108"): "40108 ProvisionCashSettlMethod",
        ("FIX.5.0", "40109"): "40109 ProvisionCashSettlCurrency",
        ("FIX.5.0", "40110"): "40110 ProvisionCashSettlCurrency2",
        ("FIX.5.0", "40111"): "40111 ProvisionCashSettlQuoteType",
        ("FIX.5.0", "40112"): "40112 ProvisionCashSettlQuoteSource",
        ("FIX.5.0", "40113"): "40113 ProvisionText",
        ("FIX.5.0", "40114"): "40114 ProvisionCashSettlValueTime",
        ("FIX.5.0", "40115"): "40115 ProvisionCashSettlValueTimeBusinessCenter",
        ("FIX.5.0", "40116"): "40116 ProvisionCashSettlValueDateBusinessDayConvention",
        ("FIX.5.0", "40117"): "40117 ProvisionCashSettlValueDateBusinessCenter",
        ("FIX.5.0", "40118"): "40118 ProvisionCashSettlValueDateRelativeTo",
        ("FIX.5.0", "40119"): "40119 ProvisionCashSettlValueDateOffsetPeriod",
        ("FIX.5.0", "40120"): "40120 ProvisionCashSettlValueDateOffsetUnit",
        ("FIX.5.0", "40121"): "40121 ProvisionCashSettlValueDateOffsetDayType",
        ("FIX.5.0", "40122"): "40122 ProvisionCashSettlValueDateAdjusted",
        ("FIX.5.0", "40123"): "40123 ProvisionOptionExerciseBusinessDayConvention",
        ("FIX.5.0", "40124"): "40124 ProvisionOptionExerciseBusinessCenter",
        ("FIX.5.0", "40125"): "40125 ProvisionOptionExerciseEarliestDateOffsetPeriod",
        ("FIX.5.0", "40126"): "40126 ProvisionOptionExerciseEarliestDateOffsetUnit",
        ("FIX.5.0", "40127"): "40127 ProvisionOptionExerciseFrequencyPeriod",
        ("FIX.5.0", "40128"): "40128 ProvisionOptionExerciseFrequencyUnit",
        ("FIX.5.0", "40129"): "40129 ProvisionOptionExerciseStartDateUnadjusted",
        ("FIX.5.0", "40130"): "40130 ProvisionOptionExerciseStartDateRelativeTo",
        ("FIX.5.0", "40131"): "40131 ProvisionOptionExerciseStartDateOffsetPeriod",
        ("FIX.5.0", "40132"): "40132 ProvisionOptionExerciseStartDateOffsetUnit",
        ("FIX.5.0", "40133"): "40133 ProvisionOptionExerciseStartDateOffsetDayType",
        ("FIX.5.0", "40134"): "40134 ProvisionOptionExerciseStartDateAdjusted",
        ("FIX.5.0", "40135"): "40135 ProvisionOptionExercisePeriodSkip",
        ("FIX.5.0", "40136"): "40136 ProvisionOptionExerciseBoundsFirstDateUnadjusted",
        ("FIX.5.0", "40137"): "40137 ProvisionOptionExerciseBoundsLastDateUnadjusted",
        ("FIX.5.0", "40138"): "40138 ProvisionOptionExerciseEarliestTime",
        ("FIX.5.0", "40139"): "40139 ProvisionOptionExerciseEarliestTimeBusinessCenter",
        ("FIX.5.0", "40140"): "40140 ProvisionOptionExerciseLatestTime",
        ("FIX.5.0", "40141"): "40141 ProvisionOptionExerciseLatestTimeBusinessCenter",
        ("FIX.5.0", "40142"): "40142 NoProvisionOptionExerciseFixedDates",
        ("FIX.5.0", "40143"): "40143 ProvisionOptionExerciseFixedDate",
        ("FIX.5.0", "40144"): "40144 ProvisionOptionExerciseFixedDateType",
        ("FIX.5.0", "40145"): "40145 ProvisionOptionExpirationDateUnadjusted",
        ("FIX.5.0", "40146"): "40146 ProvisionOptionExpirationDateBusinessDayConvention",
        ("FIX.5.0", "40147"): "40147 ProvisionOptionExpirationDateBusinessCenter",
        ("FIX.5.0", "40148"): "40148 ProvisionOptionExpirationDateRelativeTo",
        ("FIX.5.0", "40149"): "40149 ProvisionOptionExpirationDateOffsetPeriod",
        ("FIX.5.0", "40150"): "40150 ProvisionOptionExpirationDateOffsetUnit",
        ("FIX.5.0", "40151"): "40151 ProvisionOptionExpirationDateOffsetDayType",
        ("FIX.5.0", "40152"): "40152 ProvisionOptionExpirationDateAdjusted",
        ("FIX.5.0", "40153"): "40153 ProvisionOptionExpirationTime",
        ("FIX.5.0", "40154"): "40154 ProvisionOptionExpirationTimeBusinessCenter",
        ("FIX.5.0", "40155"): "40155 ProvisionOptionRelevantUnderlyingDateUnadjusted",
        ("FIX.5.0", "40156"): "40156 ProvisionOptionRelevantUnderlyingDateBusinessDayConvention",
        ("FIX.5.0", "40157"): "40157 ProvisionOptionRelevantUnderlyingDateBusinessCenter",
        ("FIX.5.0", "40158"): "40158 ProvisionOptionRelevantUnderlyingDateRelativeTo",
        ("FIX.5.0", "40159"): "40159 ProvisionOptionRelevantUnderlyingDateOffsetPeriod",
        ("FIX.5.0", "40160"): "40160 ProvisionOptionRelevantUnderlyingDateOffsetUnit",
        ("FIX.5.0", "40161"): "40161 ProvisionOptionRelevantUnderlyingDateOffsetDayType",
        ("FIX.5.0", "40162"): "40162 ProvisionOptionRelevantUnderlyingDateAdjusted",
        ("FIX.5.0", "40163"): "40163 ProvisionCashSettlPaymentDateBusinessDayConvention",
        ("FIX.5.0", "40164"): "40164 ProvisionCashSettlPaymentDateBusinessCenter",
        ("FIX.5.0", "40165"): "40165 ProvisionCashSettlPaymentDateRelativeTo",
        ("FIX.5.0", "40166"): "40166 ProvisionCashSettlPaymentDateOffsetPeriod",
        ("FIX.5.0", "40167"): "40167 ProvisionCashSettlPaymentDateOffsetUnit",
        ("FIX.5.0", "40168"): "40168 ProvisionCashSettlPaymentDateOffsetDayType",
        ("FIX.5.0", "40169"): "40169 ProvisionCashSettlPaymentDateRangeFirst",
        ("FIX.5.0", "40170"): "40170 ProvisionCashSettlPaymentDateRangeLast",
        ("FIX.5.0", "40171"): "40171 NoProvisionCashSettlPaymentDates",
        ("FIX.5.0", "40172"): "40172 ProvisionCashSettlPaymentDate",
        ("FIX.5.0", "40173"): "40173 ProvisionCashSettlPaymentDateType",
        ("FIX.5.0", "40174"): "40174 NoProvisionPartyIDs",
        ("FIX.5.0", "40175"): "40175 ProvisionPartyID",
        ("FIX.5.0", "40176"): "40176 ProvisionPartyIDSource",
        ("FIX.5.0", "40177"): "40177 ProvisionPartyRole",
        ("FIX.5.0", "40178"): "40178 NoProvisionPartySubIDs",
        ("FIX.5.0", "40179"): "40179 ProvisionPartySubID",
        ("FIX.5.0", "40180"): "40180 ProvisionPartySubIDType",
        ("FIX.5.0", "40181"): "40181 NoProtectionTerms",
        ("FIX.5.0", "40182"): "40182 ProtectionTermNotional",
        ("FIX.5.0", "40183"): "40183 ProtectionTermCurrency",
        ("FIX.5.0", "40184"): "40184 ProtectionTermSellerNotifies",
        ("FIX.5.0", "40185"): "40185 ProtectionTermBuyerNotifies",
        ("FIX.5.0", "40186"): "40186 ProtectionTermEventBusinessCenter",
        ("FIX.5.0", "40187"): "40187 ProtectionTermStandardSources",
        ("FIX.5.0", "40188"): "40188 ProtectionTermEventMinimumSources",
        ("FIX.5.0", "40189"): "40189 ProtectionTermEventNewsSource",
        ("FIX.5.0", "40190"): "40190 ProtectionTermXID",
        ("FIX.5.0", "40191"): "40191 NoProtectionTermEvents",
        ("FIX.5.0", "40192"): "40192 ProtectionTermEventType",
        ("FIX.5.0", "40193"): "40193 ProtectionTermEventValue",
        ("FIX.5.0", "40194"): "40194 ProtectionTermEventCurrency",
        ("FIX.5.0", "40195"): "40195 ProtectionTermEventPeriod",
        ("FIX.5.0", "40196"): "40196 ProtectionTermEventUnit",
        ("FIX.5.0", "40197"): "40197 ProtectionTermEventDayType",
        ("FIX.5.0", "40198"): "40198 ProtectionTermEventRateSource",
        ("FIX.5.0", "40199"): "40199 NoProtectionTermEventQualifiers",
        ("FIX.5.0", "40200"): "40200 ProtectionTermEventQualifier",
        ("FIX.5.0", "40201"): "40201 NoProtectionTermObligations",
        ("FIX.5.0", "40202"): "40202 ProtectionTermObligationType",
        ("FIX.5.0", "40203"): "40203 ProtectionTermObligationValue",
        ("FIX.5.0", "40204"): "40204 NoPhysicalSettlTerms",
        ("FIX.5.0", "40205"): "40205 PhysicalSettlCurrency",
        ("FIX.5.0", "40206"): "40206 PhysicalSettlBusinessDays",
        ("FIX.5.0", "40207"): "40207 PhysicalSettlMaximumBusinessDays",
        ("FIX.5.0", "40208"): "40208 PhysicalSettlTermXID",
        ("FIX.5.0", "40209"): "40209 NoPhysicalSettlDeliverableObligations",
        ("FIX.5.0", "40210"): "40210 PhysicalSettlDeliverableObligationType",
        ("FIX.5.0", "40211"): "40211 PhysicalSettlDeliverableObligationValue",
        ("FIX.5.0", "40212"): "40212 NoPayments",
        ("FIX.5.0", "40213"): "40213 PaymentType",
        ("FIX.5.0", "40214"): "40214 PaymentPaySide",
        ("FIX.5.0", "40215"): "40215 PaymentReceiveSide",
        ("FIX.5.0", "40216"): "40216 PaymentCurrency",
        ("FIX.5.0", "40217"): "40217 PaymentAmount",
        ("FIX.5.0", "40218"): "40218 PaymentPrice",
        ("FIX.5.0", "40219"): "40219 PaymentDateUnadjusted",
        ("FIX.5.0", "40220"): "40220 PaymentBusinessDayConvention",
        ("FIX.5.0", "40221"): "40221 PaymentBusinessCenter",
        ("FIX.5.0", "40222"): "40222 PaymentDateAdjusted",
        ("FIX.5.0", "40223"): "40223 LegMarketDisruptionValue",
        ("FIX.5.0", "40224"): "40224 PaymentDiscountFactor",
        ("FIX.5.0", "40225"): "40225 PaymentPresentValueAmount",
        ("FIX.5.0", "40226"): "40226 PaymentPresentValueCurrency",
        ("FIX.5.0", "40227"): "40227 PaymentSettlStyle",
        ("FIX.5.0", "40228"): "40228 LegPaymentStreamNonDeliverableSettlReferencePage",
        ("FIX.5.0", "40229"): "40229 PaymentText",
        ("FIX.5.0", "40230"): "40230 NoPaymentSettls",
        ("FIX.5.0", "40231"): "40231 PaymentSettlAmount",
        ("FIX.5.0", "40232"): "40232 PaymentSettlCurrency",
        ("FIX.5.0", "40233"): "40233 NoPaymentSettlPartyIDs",
        ("FIX.5.0", "40234"): "40234 PaymentSettlPartyID",
        ("FIX.5.0", "40235"): "40235 PaymentSettlPartyIDSource",
        ("FIX.5.0", "40236"): "40236 PaymentSettlPartyRole",
        ("FIX.5.0", "40237"): "40237 PaymentSettlPartyRoleQualifier",
        ("FIX.5.0", "40238"): "40238 NoPaymentSettlPartySubIDs",
        ("FIX.5.0", "40239"): "40239 PaymentSettlPartySubID",
        ("FIX.5.0", "40240"): "40240 PaymentSettlPartySubIDType",
        ("FIX.5.0", "40241"): "40241 NoLegStreams",
        ("FIX.5.0", "40242"): "40242 LegStreamType",
        ("FIX.5.0", "40243"): "40243 LegStreamDesc",
        ("FIX.5.0", "40244"): "40244 LegStreamPaySide",
        ("FIX.5.0", "40245"): "40245 LegStreamReceiveSide",
        ("FIX.5.0", "40246"): "40246 LegStreamNotional",
        ("FIX.5.0", "40247"): "40247 LegStreamCurrency",
        ("FIX.5.0", "40248"): "40248 LegStreamText",
        ("FIX.5.0", "40249"): "40249 LegStreamEffectiveDateUnadjusted",
        ("FIX.5.0", "40250"): "40250 LegStreamEffectiveDateBusinessDayConvention",
        ("FIX.5.0", "40251"): "40251 LegStreamEffectiveDateBusinessCenter",
        ("FIX.5.0", "40252"): "40252 LegStreamEffectiveDateRelativeTo",
        ("FIX.5.0", "40253"): "40253 LegStreamEffectiveDateOffsetPeriod",
        ("FIX.5.0", "40254"): "40254 LegStreamEffectiveDateOffsetUnit",
        ("FIX.5.0", "40255"): "40255 LegStreamEffectiveDateOffsetDayType",
        ("FIX.5.0", "40256"): "40256 LegStreamEffectiveDateAdjusted",
        ("FIX.5.0", "40257"): "40257 LegStreamTerminationDateUnadjusted",
        ("FIX.5.0", "40258"): "40258 LegStreamTerminationDateBusinessDayConvention",
        ("FIX.5.0", "40259"): "40259 LegStreamTerminationDateBusinessCenter",
        ("FIX.5.0", "40260"): "40260 LegStreamTerminationDateRelativeTo",
        ("FIX.5.0", "40261"): "40261 LegStreamTerminationDateOffsetPeriod",
        ("FIX.5.0", "40262"): "40262 LegStreamTerminationDateOffsetUnit",
        ("FIX.5.0", "40263"): "40263 LegStreamTerminationDateOffsetDayType",
        ("FIX.5.0", "40264"): "40264 LegStreamTerminationDateAdjusted",
        ("FIX.5.0", "40265"): "40265 LegStreamCalculationPeriodBusinessDayConvention",
        ("FIX.5.0", "40266"): "40266 LegStreamCalculationPeriodBusinessCenter",
        ("FIX.5.0", "40267"): "40267 LegStreamFirstPeriodStartDateUnadjusted",
        ("FIX.5.0", "40268"): "40268 LegStreamFirstPeriodStartDateBusinessDayConvention",
        ("FIX.5.0", "40269"): "40269 LegStreamFirstPeriodStartDateBusinessCenter",
        ("FIX.5.0", "40270"): "40270 LegStreamFirstPeriodStartDateAdjusted",
        ("FIX.5.0", "40271"): "40271 LegStreamFirstRegularPeriodStartDateUnadjusted",
        ("FIX.5.0", "40272"): "40272 LegStreamFirstCompoundingPeriodEndDateUnadjusted",
        ("FIX.5.0", "40273"): "40273 LegStreamLastRegularPeriodEndDateUnadjusted",
        ("FIX.5.0", "40274"): "40274 LegStreamCalculationFrequencyPeriod",
        ("FIX.5.0", "40275"): "40275 LegStreamCalculationFrequencyUnit",
        ("FIX.5.0", "40276"): "40276 LegStreamCalculationRollConvention",
        ("FIX.5.0", "40277"): "40277 NoCashSettlDealers",
        ("FIX.5.0", "40278"): "40278 NoBusinessCenters",
        ("FIX.5.0", "40279"): "40279 LegPaymentStreamType",
        ("FIX.5.0", "40280"): "40280 LegPaymentStreamMarketRate",
        ("FIX.5.0", "40281"): "40281 LegPaymentStreamDelayIndicator",
        ("FIX.5.0", "40282"): "40282 LegPaymentStreamSettlCurrency",
        ("FIX.5.0", "40283"): "40283 LegPaymentStreamDayCount",
        ("FIX.5.0", "40284"): "40284 LegPaymentStreamAccrualDays",
        ("FIX.5.0", "40285"): "40285 LegPaymentStreamDiscountType",
        ("FIX.5.0", "40286"): "40286 LegPaymentStreamDiscountRate",
        ("FIX.5.0", "40287"): "40287 LegPaymentStreamDiscountRateDayCount",
        ("FIX.5.0", "40288"): "40288 LegPaymentStreamCompoundingMethod",
        ("FIX.5.0", "40289"): "40289 LegPaymentStreamInitialPrincipalExchangeIndicator",
        ("FIX.5.0", "40290"): "40290 LegPaymentStreamInterimPrincipalExchangeIndicator",
        ("FIX.5.0", "40291"): "40291 LegPaymentStreamFinalPrincipalExchangeIndicator",
        ("FIX.5.0", "40292"): "40292 LegPaymentStreamPaymentDateBusinessDayConvention",
        ("FIX.5.0", "40293"): "40293 LegPaymentStreamPaymentDateBusinessCenter",
        ("FIX.5.0", "40294"): "40294 LegPaymentStreamPaymentFrequencyPeriod",
        ("FIX.5.0", "40295"): "40295 LegPaymentStreamPaymentFrequencyUnit",
        ("FIX.5.0", "40296"): "40296 LegPaymentStreamPaymentRollConvention",
        ("FIX.5.0", "40297"): "40297 LegPaymentStreamFirstPaymentDateUnadjusted",
        ("FIX.5.0", "40298"): "40298 LegPaymentStreamLastRegularPaymentDateUnadjusted",
        ("FIX.5.0", "40299"): "40299 LegPaymentStreamPaymentDateRelativeTo",
        ("FIX.5.0", "40300"): "40300 LegPaymentStreamPaymentDateOffsetPeriod",
        ("FIX.5.0", "40301"): "40301 LegPaymentStreamPaymentDateOffsetUnit",
        ("FIX.5.0", "40302"): "40302 LegPaymentStreamPaymentDateOffsetDayType",
        ("FIX.5.0", "40303"): "40303 LegPaymentStreamResetDateRelativeTo",
        ("FIX.5.0", "40304"): "40304 LegPaymentStreamResetDateBusinessDayConvention",
        ("FIX.5.0", "40305"): "40305 LegPaymentStreamResetDateBusinessCenter",
        ("FIX.5.0", "40306"): "40306 LegPaymentStreamResetFrequencyPeriod",
        ("FIX.5.0", "40307"): "40307 LegPaymentStreamResetFrequencyUnit",
        ("FIX.5.0", "40308"): "40308 LegPaymentStreamResetWeeklyRollConvention",
        ("FIX.5.0", "40309"): "40309 LegPaymentStreamInitialFixingDateRelativeTo",
        ("FIX.5.0", "40310"): "40310 LegPaymentStreamInitialFixingDateBusinessDayConvention",
        ("FIX.5.0", "40311"): "40311 LegPaymentStreamInitialFixingDateBusinessCenter",
        ("FIX.5.0", "40312"): "40312 LegPaymentStreamInitialFixingDateOffsetPeriod",
        ("FIX.5.0", "40313"): "40313 LegPaymentStreamInitialFixingDateOffsetUnit",
        ("FIX.5.0", "40314"): "40314 LegPaymentStreamInitialFixingDateOffsetDayType",
        ("FIX.5.0", "40315"): "40315 LegPaymentStreamInitialFixingDateAdjusted",
        ("FIX.5.0", "40316"): "40316 LegPaymentStreamFixingDateRelativeTo",
        ("FIX.5.0", "40317"): "40317 LegPaymentStreamFixingDateBusinessDayConvention",
        ("FIX.5.0", "40318"): "40318 LegPaymentStreamFixingDateBusinessCenter",
        ("FIX.5.0", "40319"): "40319 LegPaymentStreamFixingDateOffsetPeriod",
        ("FIX.5.0", "40320"): "40320 LegPaymentStreamFixingDateOffsetUnit",
        ("FIX.5.0", "40321"): "40321 LegPaymentStreamFixingDateOffsetDayType",
        ("FIX.5.0", "40322"): "40322 LegPaymentStreamFixingDateAdjusted",
        ("FIX.5.0", "40323"): "40323 LegPaymentStreamRateCutoffDateOffsetPeriod",
        ("FIX.5.0", "40324"): "40324 LegPaymentStreamRateCutoffDateOffsetUnit",
        ("FIX.5.0", "40325"): "40325 LegPaymentStreamRateCutoffDateOffsetDayType",
        ("FIX.5.0", "40326"): "40326 LegPaymentStreamRate",
        ("FIX.5.0", "40327"): "40327 LegPaymentStreamFixedAmount",
        ("FIX.5.0", "40328"): "40328 LegPaymentStreamRateOrAmountCurrency",
        ("FIX.5.0", "40329"): "40329 LegPaymentStreamFutureValueNotional",
        ("FIX.5.0", "40330"): "40330 LegPaymentStreamFutureValueDateAdjusted",
        ("FIX.5.0", "40331"): "40331 LegPaymentStreamRateIndex",
        ("FIX.5.0", "40332"): "40332 LegPaymentStreamRateIndexSource",
        ("FIX.5.0", "40333"): "40333 LegPaymentStreamRateIndexCurveUnit",
        ("FIX.5.0", "40334"): "40334 LegPaymentStreamRateIndexCurvePeriod",
        ("FIX.5.0", "40335"): "40335 LegPaymentStreamRateMultiplier",
        ("FIX.5.0", "40336"): "40336 LegPaymentStreamRateSpread",
        ("FIX.5.0", "40337"): "40337 LegPaymentStreamRateSpreadPositionType",
        ("FIX.5.0", "40338"): "40338 LegPaymentStreamRateTreatment",
        ("FIX.5.0", "40339"): "40339 LegPaymentStreamCapRate",
        ("FIX.5.0", "40340"): "40340 LegPaymentStreamCapRateBuySide",
        ("FIX.5.0", "40341"): "40341 LegPaymentStreamCapRateSellSide",
        ("FIX.5.0", "40342"): "40342 LegPaymentStreamFloorRate",
        ("FIX.5.0", "40343"): "40343 LegPaymentStreamFloorRateBuySide",
        ("FIX.5.0", "40344"): "40344 LegPaymentStreamFloorRateSellSide",
        ("FIX.5.0", "40345"): "40345 LegPaymentStreamInitialRate",
        ("FIX.5.0", "40346"): "40346 LegPaymentStreamFinalRateRoundingDirection",
        ("FIX.5.0", "40347"): "40347 LegPaymentStreamFinalRatePrecision",
        ("FIX.5.0", "40348"): "40348 LegPaymentStreamAveragingMethod",
        ("FIX.5.0", "40349"): "40349 LegPaymentStreamNegativeRateTreatment",
        ("FIX.5.0", "40350"): "40350 LegPaymentStreamInflationLagPeriod",
        ("FIX.5.0", "40351"): "40351 LegPaymentStreamInflationLagUnit",
        ("FIX.5.0", "40352"): "40352 LegPaymentStreamInflationLagDayType",
        ("FIX.5.0", "40353"): "40353 LegPaymentStreamInflationInterpolationMethod",
        ("FIX.5.0", "40354"): "40354 LegPaymentStreamInflationIndexSource",
        ("FIX.5.0", "40355"): "40355 LegPaymentStreamInflationPublicationSource",
        ("FIX.5.0", "40356"): "40356 LegPaymentStreamInflationInitialIndexLevel",
        ("FIX.5.0", "40357"): "40357 LegPaymentStreamInflationFallbackBondApplicable",
        ("FIX.5.0", "40358"): "40358 LegPaymentStreamFRADiscounting",
        ("FIX.5.0", "40359"): "40359 LegPaymentStreamNonDeliverableRefCurrency",
        ("FIX.5.0", "40360"): "40360 LegPaymentStreamNonDeliverableFixingDatesBusinessDayConvention",
        ("FIX.5.0", "40361"): "40361 LegPaymentStreamNonDeliverableFixingDatesBusinessCenter",
        ("FIX.5.0", "40362"): "40362 LegPaymentStreamNonDeliverableFixingDatesRelativeTo",
        ("FIX.5.0", "40363"): "40363 LegPaymentStreamNonDeliverableFixingDatesOffsetPeriod",
        ("FIX.5.0", "40364"): "40364 LegPaymentStreamNonDeliverableFixingDatesOffsetUnit",
        ("FIX.5.0", "40365"): "40365 LegPaymentStreamNonDeliverableFixingDatesOffsetDayType",
        ("FIX.5.0", "40366"): "40366 LegSettlRateFallbackRateSource",
        ("FIX.5.0", "40367"): "40367 NoLegNonDeliverableFixingDates",
        ("FIX.5.0", "40368"): "40368 LegNonDeliverableFixingDate",
        ("FIX.5.0", "40369"): "40369 LegNonDeliverableFixingDateType",
        ("FIX.5.0", "40370"): "40370 LegSettlRateFallbackReferencePage",
        ("FIX.5.0", "40371"): "40371 PaymentStreamNonDeliverableSettlRateSource",
        ("FIX.5.0", "40372"): "40372 PaymentStreamNonDeliverableSettlReferencePage",
        ("FIX.5.0", "40373"): "40373 SettlRateFallbackRateSource",
        ("FIX.5.0", "40374"): "40374 NoLegPaymentSchedules",
        ("FIX.5.0", "40375"): "40375 LegPaymentScheduleType",
        ("FIX.5.0", "40376"): "40376 LegPaymentScheduleStubType",
        ("FIX.5.0", "40377"): "40377 LegPaymentScheduleStartDateUnadjusted",
        ("FIX.5.0", "40378"): "40378 LegPaymentScheduleEndDateUnadjusted",
        ("FIX.5.0", "40379"): "40379 LegPaymentSchedulePaySide",
        ("FIX.5.0", "40380"): "40380 LegPaymentScheduleReceiveSide",
        ("FIX.5.0", "40381"): "40381 LegPaymentScheduleNotional",
        ("FIX.5.0", "40382"): "40382 LegPaymentScheduleCurrency",
        ("FIX.5.0", "40383"): "40383 LegPaymentScheduleRate",
        ("FIX.5.0", "40384"): "40384 LegPaymentScheduleRateMultiplier",
        ("FIX.5.0", "40385"): "40385 LegPaymentScheduleRateSpread",
        ("FIX.5.0", "40386"): "40386 LegPaymentScheduleRateSpreadPositionType",
        ("FIX.5.0", "40387"): "40387 LegPaymentScheduleRateTreatment",
        ("FIX.5.0", "40388"): "40388 LegPaymentScheduleFixedAmount",
        ("FIX.5.0", "40389"): "40389 LegPaymentScheduleFixedCurrency",
        ("FIX.5.0", "40390"): "40390 LegPaymentScheduleStepFrequencyPeriod",
        ("FIX.5.0", "40391"): "40391 LegPaymentScheduleStepFrequencyUnit",
        ("FIX.5.0", "40392"): "40392 LegPaymentScheduleStepOffsetValue",
        ("FIX.5.0", "40393"): "40393 LegPaymentScheduleStepRate",
        ("FIX.5.0", "40394"): "40394 LegPaymentScheduleStepOffsetRate",
        ("FIX.5.0", "40395"): "40395 LegPaymentScheduleStepRelativeTo",
        ("FIX.5.0", "40396"): "40396 LegPaymentScheduleFixingDateUnadjusted",
        ("FIX.5.0", "40397"): "40397 LegPaymentScheduleWeight",
        ("FIX.5.0", "40398"): "40398 LegPaymentScheduleFixingDateRelativeTo",
        ("FIX.5.0", "40399"): "40399 LegPaymentScheduleFixingDateBusinessDayConvention",
        ("FIX.5.0", "40400"): "40400 LegPaymentScheduleFixingDateBusinessCenter",
        ("FIX.5.0", "40401"): "40401 LegPaymentScheduleFixingDateOffsetPeriod",
        ("FIX.5.0", "40402"): "40402 LegPaymentScheduleFixingDateOffsetUnit",
        ("FIX.5.0", "40403"): "40403 LegPaymentScheduleFixingDateOffsetDayType",
        ("FIX.5.0", "40404"): "40404 LegPaymentScheduleFixingDateAdjusted",
        ("FIX.5.0", "40405"): "40405 LegPaymentScheduleFixingTime",
        ("FIX.5.0", "40406"): "40406 LegPaymentScheduleFixingTimeBusinessCenter",
        ("FIX.5.0", "40407"): "40407 LegPaymentScheduleInterimExchangePaymentDateRelativeTo",
        ("FIX.5.0", "40408"): "40408 LegPaymentScheduleInterimExchangeDatesBusinessDayConvention",
        ("FIX.5.0", "40409"): "40409 LegPaymentScheduleInterimExchangeDatesBusinessCenter",
        ("FIX.5.0", "40410"): "40410 LegPaymentScheduleInterimExchangeDatesOffsetPeriod",
        ("FIX.5.0", "40411"): "40411 LegPaymentScheduleInterimExchangeDatesOffsetUnit",
        ("FIX.5.0", "40412"): "40412 LegPaymentScheduleInterimExchangeDatesOffsetDayType",
        ("FIX.5.0", "40413"): "40413 LegPaymentScheduleInterimExchangeDateAdjusted",
        ("FIX.5.0", "40414"): "40414 NoLegPaymentScheduleRateSources",
        ("FIX.5.0", "40415"): "40415 LegPaymentScheduleRateSource",
        ("FIX.5.0", "40416"): "40416 LegPaymentScheduleRateSourceType",
        ("FIX.5.0", "40417"): "40417 LegPaymentScheduleReferencePage",
        ("FIX.5.0", "40418"): "40418 NoLegPaymentStubs",
        ("FIX.5.0", "40419"): "40419 LegPaymentStubType",
        ("FIX.5.0", "40420"): "40420 LegPaymentStubLength",
        ("FIX.5.0", "40421"): "40421 LegPaymentStubRate",
        ("FIX.5.0", "40422"): "40422 LegPaymentStubFixedAmount",
        ("FIX.5.0", "40423"): "40423 LegPaymentStubFixedCurrency",
        ("FIX.5.0", "40424"): "40424 LegPaymentStubIndex",
        ("FIX.5.0", "40425"): "40425 LegPaymentStubIndexSource",
        ("FIX.5.0", "40426"): "40426 LegPaymentStubIndexCurvePeriod",
        ("FIX.5.0", "40427"): "40427 LegPaymentStubIndexCurveUnit",
        ("FIX.5.0", "40428"): "40428 LegPaymentStubIndexRateMultiplier",
        ("FIX.5.0", "40429"): "40429 LegPaymentStubIndexRateSpread",
        ("FIX.5.0", "40430"): "40430 LegPaymentStubIndexRateSpreadPositionType",
        ("FIX.5.0", "40431"): "40431 LegPaymentStubIndexRateTreatment",
        ("FIX.5.0", "40432"): "40432 LegPaymentStubIndexCapRate",
        ("FIX.5.0", "40433"): "40433 LegPaymentStubIndexCapRateBuySide",
        ("FIX.5.0", "40434"): "40434 LegPaymentStubIndexCapRateSellSide",
        ("FIX.5.0", "40435"): "40435 LegPaymentStubIndexFloorRate",
        ("FIX.5.0", "40436"): "40436 LegPaymentStubIndexFloorRateBuySide",
        ("FIX.5.0", "40437"): "40437 LegPaymentStubIndexFloorRateSellSide",
        ("FIX.5.0", "40438"): "40438 LegPaymentStubIndex2",
        ("FIX.5.0", "40439"): "40439 LegPaymentStubIndex2Source",
        ("FIX.5.0", "40440"): "40440 LegPaymentStubIndex2CurvePeriod",
        ("FIX.5.0", "40441"): "40441 LegPaymentStubIndex2CurveUnit",
        ("FIX.5.0", "40442"): "40442 LegPaymentStubIndex2RateMultiplier",
        ("FIX.5.0", "40443"): "40443 LegPaymentStubIndex2RateSpread",
        ("FIX.5.0", "40444"): "40444 LegPaymentStubIndex2RateSpreadPositionType",
        ("FIX.5.0", "40445"): "40445 LegPaymentStubIndex2RateTreatment",
        ("FIX.5.0", "40446"): "40446 LegPaymentStubIndex2CapRate",
        ("FIX.5.0", "40447"): "40447 LegPaymentStubIndex2FloorRate",
        ("FIX.5.0", "40448"): "40448 NoLegProvisions",
        ("FIX.5.0", "40449"): "40449 LegProvisionType",
        ("FIX.5.0", "40450"): "40450 LegProvisionDateUnadjusted",
        ("FIX.5.0", "40451"): "40451 LegProvisionDateBusinessDayConvention",
        ("FIX.5.0", "40452"): "40452 LegProvisionDateBusinessCenter",
        ("FIX.5.0", "40453"): "40453 LegProvisionDateAdjusted",
        ("FIX.5.0", "40454"): "40454 LegProvisionDateTenorPeriod",
        ("FIX.5.0", "40455"): "40455 LegProvisionDateTenorUnit",
        ("FIX.5.0", "40456"): "40456 LegProvisionCalculationAgent",
        ("FIX.5.0", "40457"): "40457 LegProvisionOptionSinglePartyBuyerSide",
        ("FIX.5.0", "40458"): "40458 LegProvisionOptionSinglePartySellerSide",
        ("FIX.5.0", "40459"): "40459 LegProvisionOptionExerciseStyle",
        ("FIX.5.0", "40460"): "40460 LegProvisionOptionExerciseMultipleNotional",
        ("FIX.5.0", "40461"): "40461 LegProvisionOptionExerciseMinimumNotional",
        ("FIX.5.0", "40462"): "40462 LegProvisionOptionExerciseMaximumNotional",
        ("FIX.5.0", "40463"): "40463 LegProvisionOptionMinimumNumber",
        ("FIX.5.0", "40464"): "40464 LegProvisionOptionMaximumNumber",
        ("FIX.5.0", "40465"): "40465 LegProvisionOptionExerciseConfirmation",
        ("FIX.5.0", "40466"): "40466 LegProvisionCashSettlMethod",
        ("FIX.5.0", "40467"): "40467 LegProvisionCashSettlCurrency",
        ("FIX.5.0", "40468"): "40468 LegProvisionCashSettlCurrency2",
        ("FIX.5.0", "40469"): "40469 LegProvisionCashSettlQuoteType",
        ("FIX.5.0", "40470"): "40470 LegProvisionCashSettlQuoteSource",
        ("FIX.5.0", "40471"): "40471 BusinessCenter",
        ("FIX.5.0", "40472"): "40472 LegProvisionText",
        ("FIX.5.0", "40473"): "40473 NoLegProvisionCashSettlPaymentDates",
        ("FIX.5.0", "40474"): "40474 LegProvisionCashSettlPaymentDate",
        ("FIX.5.0", "40475"): "40475 LegProvisionCashSettlPaymentDateType",
        ("FIX.5.0", "40476"): "40476 LegProvisionOptionExerciseBusinessDayConvention",
        ("FIX.5.0", "40477"): "40477 LegProvisionOptionExerciseBusinessCenter",
        ("FIX.5.0", "40478"): "40478 LegProvisionOptionExerciseEarliestDateOffsetPeriod",
        ("FIX.5.0", "40479"): "40479 LegProvisionOptionExerciseEarliestDateOffsetUnit",
        ("FIX.5.0", "40480"): "40480 LegProvisionOptionExerciseFrequencyPeriod",
        ("FIX.5.0", "40481"): "40481 LegProvisionOptionExerciseFrequencyUnit",
        ("FIX.5.0", "40482"): "40482 LegProvisionOptionExerciseStartDateUnadjusted",
        ("FIX.5.0", "40483"): "40483 LegProvisionOptionExerciseStartDateRelativeTo",
        ("FIX.5.0", "40484"): "40484 LegProvisionOptionExerciseStartDateOffsetPeriod",
        ("FIX.5.0", "40485"): "40485 LegProvisionOptionExerciseStartDateOffsetUnit",
        ("FIX.5.0", "40486"): "40486 LegProvisionOptionExerciseStartDateOffsetDayType",
        ("FIX.5.0", "40487"): "40487 LegProvisionOptionExerciseStartDateAdjusted",
        ("FIX.5.0", "40488"): "40488 LegProvisionOptionExercisePeriodSkip",
        ("FIX.5.0", "40489"): "40489 LegProvisionOptionExerciseBoundsFirstDateUnadjusted",
        ("FIX.5.0", "40490"): "40490 LegProvisionOptionExerciseBoundsLastDateUnadjusted",
        ("FIX.5.0", "40491"): "40491 LegProvisionOptionExerciseEarliestTime",
        ("FIX.5.0", "40492"): "40492 LegProvisionOptionExerciseEarliestTimeBusinessCenter",
        ("FIX.5.0", "40493"): "40493 LegProvisionOptionExerciseLatestTime",
        ("FIX.5.0", "40494"): "40494 LegProvisionOptionExerciseLatestTimeBusinessCenter",
        ("FIX.5.0", "40495"): "40495 NoLegProvisionOptionExerciseFixedDates",
        ("FIX.5.0", "40496"): "40496 LegProvisionOptionExerciseFixedDate",
        ("FIX.5.0", "40497"): "40497 LegProvisionOptionExerciseFixedDateType",
        ("FIX.5.0", "40498"): "40498 LegProvisionOptionExpirationDateUnadjusted",
        ("FIX.5.0", "40499"): "40499 LegProvisionOptionExpirationDateBusinessDayConvention",
        ("FIX.5.0", "40500"): "40500 LegProvisionOptionExpirationDateBusinessCenter",
        ("FIX.5.0", "40501"): "40501 LegProvisionOptionExpirationDateRelativeTo",
        ("FIX.5.0", "40502"): "40502 LegProvisionOptionExpirationDateOffsetPeriod",
        ("FIX.5.0", "40503"): "40503 LegProvisionOptionExpirationDateOffsetUnit",
        ("FIX.5.0", "40504"): "40504 LegProvisionOptionExpirationDateOffsetDayType",
        ("FIX.5.0", "40505"): "40505 LegProvisionOptionExpirationDateAdjusted",
        ("FIX.5.0", "40506"): "40506 LegProvisionOptionExpirationTime",
        ("FIX.5.0", "40507"): "40507 LegProvisionOptionExpirationTimeBusinessCenter",
        ("FIX.5.0", "40508"): "40508 LegProvisionOptionRelevantUnderlyingDateUnadjusted",
        ("FIX.5.0", "40509"): "40509 LegProvisionOptionRelevantUnderlyingDateBusinessDayConvention",
        ("FIX.5.0", "40510"): "40510 LegProvisionOptionRelevantUnderlyingDateBusinessCenter",
        ("FIX.5.0", "40511"): "40511 LegProvisionOptionRelevantUnderlyingDateRelativeTo",
        ("FIX.5.0", "40512"): "40512 LegProvisionOptionRelevantUnderlyingDateOffsetPeriod",
        ("FIX.5.0", "40513"): "40513 LegProvisionOptionRelevantUnderlyingDateOffsetUnit",
        ("FIX.5.0", "40514"): "40514 LegProvisionOptionRelevantUnderlyingDateOffsetDayType",
        ("FIX.5.0", "40515"): "40515 LegProvisionOptionRelevantUnderlyingDateAdjusted",
        ("FIX.5.0", "40516"): "40516 LegProvisionCashSettlPaymentDateBusinessDayConvention",
        ("FIX.5.0", "40517"): "40517 LegProvisionCashSettlPaymentDateBusinessCenter",
        ("FIX.5.0", "40518"): "40518 LegProvisionCashSettlPaymentDateRelativeTo",
        ("FIX.5.0", "40519"): "40519 LegProvisionCashSettlPaymentDateOffsetPeriod",
        ("FIX.5.0", "40520"): "40520 LegProvisionCashSettlPaymentDateOffsetUnit",
        ("FIX.5.0", "40521"): "40521 LegProvisionCashSettlPaymentDateOffsetDayType",
        ("FIX.5.0", "40522"): "40522 LegProvisionCashSettlPaymentDateRangeFirst",
        ("FIX.5.0", "40523"): "40523 LegProvisionCashSettlPaymentDateRangeLast",
        ("FIX.5.0", "40524"): "40524 LegProvisionCashSettlValueTime",
        ("FIX.5.0", "40525"): "40525 LegProvisionCashSettlValueTimeBusinessCenter",
        ("FIX.5.0", "40526"): "40526 LegProvisionCashSettlValueDateBusinessDayConvention",
        ("FIX.5.0", "40527"): "40527 LegProvisionCashSettlValueDateBusinessCenter",
        ("FIX.5.0", "40528"): "40528 LegProvisionCashSettlValueDateRelativeTo",
        ("FIX.5.0", "40529"): "40529 LegProvisionCashSettlValueDateOffsetPeriod",
        ("FIX.5.0", "40530"): "40530 LegProvisionCashSettlValueDateOffsetUnit",
        ("FIX.5.0", "40531"): "40531 LegProvisionCashSettlValueDateOffsetDayType",
        ("FIX.5.0", "40532"): "40532 LegProvisionCashSettlValueDateAdjusted",
        ("FIX.5.0", "40533"): "40533 NoLegProvisionPartyIDs",
        ("FIX.5.0", "40534"): "40534 LegProvisionPartyID",
        ("FIX.5.0", "40535"): "40535 LegProvisionPartyIDSource",
        ("FIX.5.0", "40536"): "40536 LegProvisionPartyRole",
        ("FIX.5.0", "40537"): "40537 NoLegProvisionPartySubIDs",
        ("FIX.5.0", "40538"): "40538 LegProvisionPartySubID",
        ("FIX.5.0", "40539"): "40539 LegProvisionPartySubIDType",
        ("FIX.5.0", "40540"): "40540 NoUnderlyingStreams",
        ("FIX.5.0", "40541"): "40541 UnderlyingStreamType",
        ("FIX.5.0", "40542"): "40542 UnderlyingStreamDesc",
        ("FIX.5.0", "40543"): "40543 UnderlyingStreamPaySide",
        ("FIX.5.0", "40544"): "40544 UnderlyingStreamReceiveSide",
        ("FIX.5.0", "40545"): "40545 UnderlyingStreamNotional",
        ("FIX.5.0", "40546"): "40546 UnderlyingStreamCurrency",
        ("FIX.5.0", "40547"): "40547 UnderlyingStreamText",
        ("FIX.5.0", "40548"): "40548 UnderlyingStreamTerminationDateUnadjusted",
        ("FIX.5.0", "40549"): "40549 UnderlyingStreamTerminationDateBusinessDayConvention",
        ("FIX.5.0", "40550"): "40550 UnderlyingStreamTerminationDateBusinessCenter",
        ("FIX.5.0", "40551"): "40551 UnderlyingStreamTerminationDateRelativeTo",
        ("FIX.5.0", "40552"): "40552 UnderlyingStreamTerminationDateOffsetPeriod",
        ("FIX.5.0", "40553"): "40553 UnderlyingStreamTerminationDateOffsetUnit",
        ("FIX.5.0", "40554"): "40554 UnderlyingStreamTerminationDateOffsetDayType",
        ("FIX.5.0", "40555"): "40555 UnderlyingStreamTerminationDateAdjusted",
        ("FIX.5.0", "40556"): "40556 UnderlyingStreamCalculationPeriodBusinessDayConvention",
        ("FIX.5.0", "40557"): "40557 UnderlyingStreamCalculationPeriodBusinessCenter",
        ("FIX.5.0", "40558"): "40558 UnderlyingStreamFirstPeriodStartDateUnadjusted",
        ("FIX.5.0", "40559"): "40559 UnderlyingStreamFirstPeriodStartDateBusinessDayConvention",
        ("FIX.5.0", "40560"): "40560 UnderlyingStreamFirstPeriodStartDateBusinessCenter",
        ("FIX.5.0", "40561"): "40561 UnderlyingStreamFirstPeriodStartDateAdjusted",
        ("FIX.5.0", "40562"): "40562 UnderlyingStreamFirstRegularPeriodStartDateUnadjusted",
        ("FIX.5.0", "40563"): "40563 UnderlyingStreamFirstCompoundingPeriodEndDateUnadjusted",
        ("FIX.5.0", "40564"): "40564 UnderlyingStreamLastRegularPeriodEndDateUnadjusted",
        ("FIX.5.0", "40565"): "40565 UnderlyingStreamCalculationFrequencyPeriod",
        ("FIX.5.0", "40566"): "40566 UnderlyingStreamCalculationFrequencyUnit",
        ("FIX.5.0", "40567"): "40567 UnderlyingStreamCalculationRollConvention",
        ("FIX.5.0", "40568"): "40568 UnderlyingPaymentStreamType",
        ("FIX.5.0", "40569"): "40569 UnderlyingPaymentStreamMarketRate",
        ("FIX.5.0", "40570"): "40570 UnderlyingPaymentStreamDelayIndicator",
        ("FIX.5.0", "40571"): "40571 UnderlyingPaymentStreamSettlCurrency",
        ("FIX.5.0", "40572"): "40572 UnderlyingPaymentStreamDayCount",
        ("FIX.5.0", "40573"): "40573 UnderlyingPaymentStreamAccrualDays",
        ("FIX.5.0", "40574"): "40574 UnderlyingPaymentStreamDiscountType",
        ("FIX.5.0", "40575"): "40575 UnderlyingPaymentStreamDiscountRate",
        ("FIX.5.0", "40576"): "40576 UnderlyingPaymentStreamDiscountRateDayCount",
        ("FIX.5.0", "40577"): "40577 UnderlyingPaymentStreamCompoundingMethod",
        ("FIX.5.0", "40578"): "40578 UnderlyingPaymentStreamInitialPrincipalExchangeIndicator",
        ("FIX.5.0", "40579"): "40579 UnderlyingPaymentStreamInterimPrincipalExchangeIndicator",
        ("FIX.5.0", "40580"): "40580 UnderlyingPaymentStreamFinalPrincipalExchangeIndicator",
        ("FIX.5.0", "40581"): "40581 UnderlyingPaymentStreamPaymentDateBusinessDayConvention",
        ("FIX.5.0", "40582"): "40582 UnderlyingPaymentStreamPaymentDateBusinessCenter",
        ("FIX.5.0", "40583"): "40583 UnderlyingPaymentStreamPaymentFrequencyPeriod",
        ("FIX.5.0", "40584"): "40584 UnderlyingPaymentStreamPaymentFrequencyUnit",
        ("FIX.5.0", "40585"): "40585 UnderlyingPaymentStreamPaymentRollConvention",
        ("FIX.5.0", "40586"): "40586 UnderlyingPaymentStreamFirstPaymentDateUnadjusted",
        ("FIX.5.0", "40587"): "40587 UnderlyingPaymentStreamLastRegularPaymentDateUnadjusted",
        ("FIX.5.0", "40588"): "40588 UnderlyingPaymentStreamPaymentDateRelativeTo",
        ("FIX.5.0", "40589"): "40589 UnderlyingPaymentStreamPaymentDateOffsetPeriod",
        ("FIX.5.0", "40590"): "40590 UnderlyingPaymentStreamPaymentDateOffsetUnit",
        ("FIX.5.0", "40591"): "40591 UnderlyingPaymentStreamPaymentDateOffsetDayType",
        ("FIX.5.0", "40592"): "40592 UnderlyingPaymentStreamResetDateRelativeTo",
        ("FIX.5.0", "40593"): "40593 UnderlyingPaymentStreamResetDateBusinessDayConvention",
        ("FIX.5.0", "40594"): "40594 UnderlyingPaymentStreamResetDateBusinessCenter",
        ("FIX.5.0", "40595"): "40595 UnderlyingPaymentStreamResetFrequencyPeriod",
        ("FIX.5.0", "40596"): "40596 UnderlyingPaymentStreamResetFrequencyUnit",
        ("FIX.5.0", "40597"): "40597 UnderlyingPaymentStreamResetWeeklyRollConvention",
        ("FIX.5.0", "40598"): "40598 UnderlyingPaymentStreamInitialFixingDateRelativeTo",
        ("FIX.5.0", "40599"): "40599 UnderlyingPaymentStreamInitialFixingDateBusinessDayConvention",
        ("FIX.5.0", "40600"): "40600 UnderlyingPaymentStreamInitialFixingDateBusinessCenter",
        ("FIX.5.0", "40601"): "40601 UnderlyingPaymentStreamInitialFixingDateOffsetPeriod",
        ("FIX.5.0", "40602"): "40602 UnderlyingPaymentStreamInitialFixingDateOffsetUnit",
        ("FIX.5.0", "40603"): "40603 UnderlyingPaymentStreamInitialFixingDateOffsetDayType",
        ("FIX.5.0", "40604"): "40604 UnderlyingPaymentStreamInitialFixingDateAdjusted",
        ("FIX.5.0", "40605"): "40605 UnderlyingPaymentStreamFixingDateRelativeTo",
        ("FIX.5.0", "40606"): "40606 UnderlyingPaymentStreamFixingDateBusinessDayConvention",
        ("FIX.5.0", "40607"): "40607 UnderlyingPaymentStreamFixingDateBusinessCenter",
        ("FIX.5.0", "40608"): "40608 UnderlyingPaymentStreamFixingDateOffsetPeriod",
        ("FIX.5.0", "40609"): "40609 UnderlyingPaymentStreamFixingDateOffsetUnit",
        ("FIX.5.0", "40610"): "40610 UnderlyingPaymentStreamFixingDateOffsetDayType",
        ("FIX.5.0", "40611"): "40611 UnderlyingPaymentStreamFixingDateAdjusted",
        ("FIX.5.0", "40612"): "40612 UnderlyingPaymentStreamRateCutoffDateOffsetPeriod",
        ("FIX.5.0", "40613"): "40613 UnderlyingPaymentStreamRateCutoffDateOffsetUnit",
        ("FIX.5.0", "40614"): "40614 UnderlyingPaymentStreamRateCutoffDateOffsetDayType",
        ("FIX.5.0", "40615"): "40615 UnderlyingPaymentStreamRate",
        ("FIX.5.0", "40616"): "40616 UnderlyingPaymentStreamFixedAmount",
        ("FIX.5.0", "40617"): "40617 UnderlyingPaymentStreamRateOrAmountCurrency",
        ("FIX.5.0", "40618"): "40618 UnderlyingPaymentStreamFutureValueNotional",
        ("FIX.5.0", "40619"): "40619 UnderlyingPaymentStreamFutureValueDateAdjusted",
        ("FIX.5.0", "40620"): "40620 UnderlyingPaymentStreamRateIndex",
        ("FIX.5.0", "40621"): "40621 UnderlyingPaymentStreamRateIndexSource",
        ("FIX.5.0", "40622"): "40622 UnderlyingPaymentStreamRateIndexCurveUnit",
        ("FIX.5.0", "40623"): "40623 UnderlyingPaymentStreamRateIndexCurvePeriod",
        ("FIX.5.0", "40624"): "40624 UnderlyingPaymentStreamRateMultiplier",
        ("FIX.5.0", "40625"): "40625 UnderlyingPaymentStreamRateSpread",
        ("FIX.5.0", "40626"): "40626 UnderlyingPaymentStreamRateSpreadPositionType",
        ("FIX.5.0", "40627"): "40627 UnderlyingPaymentStreamRateTreatment",
        ("FIX.5.0", "40628"): "40628 UnderlyingPaymentStreamCapRate",
        ("FIX.5.0", "40629"): "40629 UnderlyingPaymentStreamCapRateBuySide",
        ("FIX.5.0", "40630"): "40630 UnderlyingPaymentStreamCapRateSellSide",
        ("FIX.5.0", "40631"): "40631 UnderlyingPaymentStreamFloorRate",
        ("FIX.5.0", "40632"): "40632 UnderlyingPaymentStreamFloorRateBuySide",
        ("FIX.5.0", "40633"): "40633 UnderlyingPaymentStreamFloorRateSellSide",
        ("FIX.5.0", "40634"): "40634 UnderlyingPaymentStreamInitialRate",
        ("FIX.5.0", "40635"): "40635 UnderlyingPaymentStreamFinalRateRoundingDirection",
        ("FIX.5.0", "40636"): "40636 UnderlyingPaymentStreamFinalRatePrecision",
        ("FIX.5.0", "40637"): "40637 UnderlyingPaymentStreamAveragingMethod",
        ("FIX.5.0", "40638"): "40638 UnderlyingPaymentStreamNegativeRateTreatment",
        ("FIX.5.0", "40639"): "40639 UnderlyingPaymentStreamInflationLagPeriod",
        ("FIX.5.0", "40640"): "40640 UnderlyingPaymentStreamInflationLagUnit",
        ("FIX.5.0", "40641"): "40641 UnderlyingPaymentStreamInflationLagDayType",
        ("FIX.5.0", "40642"): "40642 UnderlyingPaymentStreamInflationInterpolationMethod",
        ("FIX.5.0", "40643"): "40643 UnderlyingPaymentStreamInflationIndexSource",
        ("FIX.5.0", "40644"): "40644 UnderlyingPaymentStreamInflationPublicationSource",
        ("FIX.5.0", "40645"): "40645 UnderlyingPaymentStreamInflationInitialIndexLevel",
        ("FIX.5.0", "40646"): "40646 UnderlyingPaymentStreamInflationFallbackBondApplicable",
        ("FIX.5.0", "40647"): "40647 UnderlyingPaymentStreamFRADiscounting",
        ("FIX.5.0", "40648"): "40648 UnderlyingPaymentStreamNonDeliverableRefCurrency",
        ("FIX.5.0", "40649"): "40649 UnderlyingPaymentStreamNonDeliverableFixingDatesBusinessDayConvention",
        ("FIX.5.0", "40650"): "40650 UnderlyingPaymentStreamNonDeliverableFixingDatesBusinessCenter",
        ("FIX.5.0", "40651"): "40651 UnderlyingPaymentStreamNonDeliverableFixingDatesRelativeTo",
        ("FIX.5.0", "40652"): "40652 UnderlyingPaymentStreamNonDeliverableFixingDatesOffsetPeriod",
        ("FIX.5.0", "40653"): "40653 UnderlyingPaymentStreamNonDeliverableFixingDatesOffsetUnit",
        ("FIX.5.0", "40654"): "40654 UnderlyingPaymentStreamNonDeliverableFixingDatesOffsetDayType",
        ("FIX.5.0", "40655"): "40655 SettlRateFallbackReferencePage",
        ("FIX.5.0", "40656"): "40656 NoUnderlyingNonDeliverableFixingDates",
        ("FIX.5.0", "40657"): "40657 UnderlyingNonDeliverableFixingDate",
        ("FIX.5.0", "40658"): "40658 UnderlyingNonDeliverableFixingDateType",
        ("FIX.5.0", "40659"): "40659 NoUnderlyingSettlRateFallbacks",
        ("FIX.5.0", "40660"): "40660 UnderlyingSettlRatePostponementMaximumDays",
        ("FIX.5.0", "40661"): "40661 UnderlyingPaymentStreamNonDeliverableSettlRateSource",
        ("FIX.5.0", "40662"): "40662 UnderlyingSettlRatePostponementSurvey",
        ("FIX.5.0", "40663"): "40663 UnderlyingSettlRatePostponementCalculationAgent",
        ("FIX.5.0", "40664"): "40664 NoUnderlyingPaymentSchedules",
        ("FIX.5.0", "40665"): "40665 UnderlyingPaymentScheduleType",
        ("FIX.5.0", "40666"): "40666 UnderlyingPaymentScheduleStubType",
        ("FIX.5.0", "40667"): "40667 UnderlyingPaymentScheduleStartDateUnadjusted",
        ("FIX.5.0", "40668"): "40668 UnderlyingPaymentScheduleEndDateUnadjusted",
        ("FIX.5.0", "40669"): "40669 UnderlyingPaymentSchedulePaySide",
        ("FIX.5.0", "40670"): "40670 UnderlyingPaymentScheduleReceiveSide",
        ("FIX.5.0", "40671"): "40671 UnderlyingPaymentScheduleNotional",
        ("FIX.5.0", "40672"): "40672 UnderlyingPaymentScheduleCurrency",
        ("FIX.5.0", "40673"): "40673 UnderlyingPaymentScheduleRate",
        ("FIX.5.0", "40674"): "40674 UnderlyingPaymentScheduleRateMultiplier",
        ("FIX.5.0", "40675"): "40675 UnderlyingPaymentScheduleRateSpread",
        ("FIX.5.0", "40676"): "40676 UnderlyingPaymentScheduleRateSpreadPositionType",
        ("FIX.5.0", "40677"): "40677 UnderlyingPaymentScheduleRateTreatment",
        ("FIX.5.0", "40678"): "40678 UnderlyingPaymentScheduleFixedAmount",
        ("FIX.5.0", "40679"): "40679 UnderlyingPaymentScheduleFixedCurrency",
        ("FIX.5.0", "40680"): "40680 UnderlyingPaymentScheduleStepFrequencyPeriod",
        ("FIX.5.0", "40681"): "40681 UnderlyingPaymentScheduleStepFrequencyUnit",
        ("FIX.5.0", "40682"): "40682 UnderlyingPaymentScheduleStepOffsetValue",
        ("FIX.5.0", "40683"): "40683 UnderlyingPaymentScheduleStepRate",
        ("FIX.5.0", "40684"): "40684 UnderlyingPaymentScheduleStepOffsetRate",
        ("FIX.5.0", "40685"): "40685 UnderlyingPaymentScheduleStepRelativeTo",
        ("FIX.5.0", "40686"): "40686 UnderlyingPaymentScheduleFixingDateUnadjusted",
        ("FIX.5.0", "40687"): "40687 UnderlyingPaymentScheduleWeight",
        ("FIX.5.0", "40688"): "40688 UnderlyingPaymentScheduleFixingDateRelativeTo",
        ("FIX.5.0", "40689"): "40689 UnderlyingPaymentScheduleFixingDateBusinessDayCnvtn",
        ("FIX.5.0", "40690"): "40690 UnderlyingPaymentScheduleFixingDateBusinessCenter",
        ("FIX.5.0", "40691"): "40691 UnderlyingPaymentScheduleFixingDateOffsetPeriod",
        ("FIX.5.0", "40692"): "40692 UnderlyingPaymentScheduleFixingDateOffsetUnit",
        ("FIX.5.0", "40693"): "40693 UnderlyingPaymentScheduleFixingDateOffsetDayType",
        ("FIX.5.0", "40694"): "40694 UnderlyingPaymentScheduleFixingDateAdjusted",
        ("FIX.5.0", "40695"): "40695 UnderlyingPaymentScheduleFixingTime",
        ("FIX.5.0", "40696"): "40696 UnderlyingPaymentScheduleFixingTimeBusinessCenter",
        ("FIX.5.0", "40697"): "40697 UnderlyingPaymentScheduleInterimExchangePaymentDateRelativeTo",
        ("FIX.5.0", "40698"): "40698 UnderlyingPaymentScheduleInterimExchangeDatesBusinessDayConvention",
        ("FIX.5.0", "40699"): "40699 UnderlyingPaymentScheduleInterimExchangeDatesBusinessCenter",
        ("FIX.5.0", "40700"): "40700 UnderlyingPaymentScheduleInterimExchangeDatesOffsetPeriod",
        ("FIX.5.0", "40701"): "40701 UnderlyingPaymentScheduleInterimExchangeDatesOffsetUnit",
        ("FIX.5.0", "40702"): "40702 UnderlyingPaymentScheduleInterimExchangeDatesOffsetDayType",
        ("FIX.5.0", "40703"): "40703 UnderlyingPaymentScheduleInterimExchangeDateAdjusted",
        ("FIX.5.0", "40704"): "40704 NoUnderlyingPaymentScheduleRateSources",
        ("FIX.5.0", "40705"): "40705 UnderlyingPaymentScheduleRateSource",
        ("FIX.5.0", "40706"): "40706 UnderlyingPaymentScheduleRateSourceType",
        ("FIX.5.0", "40707"): "40707 UnderlyingPaymentScheduleReferencePage",
        ("FIX.5.0", "40708"): "40708 NoUnderlyingPaymentStubs",
        ("FIX.5.0", "40709"): "40709 UnderlyingPaymentStubType",
        ("FIX.5.0", "40710"): "40710 UnderlyingPaymentStubLength",
        ("FIX.5.0", "40711"): "40711 UnderlyingPaymentStubRate",
        ("FIX.5.0", "40712"): "40712 UnderlyingPaymentStubFixedAmount",
        ("FIX.5.0", "40713"): "40713 UnderlyingPaymentStubFixedCurrency",
        ("FIX.5.0", "40714"): "40714 UnderlyingPaymentStubIndex",
        ("FIX.5.0", "40715"): "40715 UnderlyingPaymentStubIndexSource",
        ("FIX.5.0", "40716"): "40716 UnderlyingPaymentStubIndexCurvePeriod",
        ("FIX.5.0", "40717"): "40717 UnderlyingPaymentStubIndexCurveUnit",
        ("FIX.5.0", "40718"): "40718 UnderlyingPaymentStubIndexRateMultiplier",
        ("FIX.5.0", "40719"): "40719 UnderlyingPaymentStubIndexRateSpread",
        ("FIX.5.0", "40720"): "40720 UnderlyingPaymentStubIndexRateSpreadPositionType",
        ("FIX.5.0", "40721"): "40721 UnderlyingPaymentStubIndexRateTreatment",
        ("FIX.5.0", "40722"): "40722 UnderlyingPaymentStubIndexCapRate",
        ("FIX.5.0", "40723"): "40723 UnderlyingPaymentStubIndexCapRateBuySide",
        ("FIX.5.0", "40724"): "40724 UnderlyingPaymentStubIndexCapRateSellSide",
        ("FIX.5.0", "40725"): "40725 UnderlyingPaymentStubIndexFloorRate",
        ("FIX.5.0", "40726"): "40726 UnderlyingPaymentStubIndexFloorRateBuySide",
        ("FIX.5.0", "40727"): "40727 UnderlyingPaymentStubIndexFloorRateSellSide",
        ("FIX.5.0", "40728"): "40728 UnderlyingPaymentStubIndex2",
        ("FIX.5.0", "40729"): "40729 UnderlyingPaymentStubIndex2Source",
        ("FIX.5.0", "40730"): "40730 UnderlyingPaymentStubIndex2CurvePeriod",
        ("FIX.5.0", "40731"): "40731 UnderlyingPaymentStubIndex2CurveUnit",
        ("FIX.5.0", "40732"): "40732 UnderlyingPaymentStubIndex2RateMultiplier",
        ("FIX.5.0", "40733"): "40733 UnderlyingPaymentStubIndex2RateSpread",
        ("FIX.5.0", "40734"): "40734 UnderlyingPaymentStubIndex2RateSpreadPositionType",
        ("FIX.5.0", "40735"): "40735 UnderlyingPaymentStubIndex2RateTreatment",
        ("FIX.5.0", "40736"): "40736 UnderlyingPaymentStubIndex2CapRate",
        ("FIX.5.0", "40737"): "40737 UnderlyingPaymentStubIndex2FloorRate",
        ("FIX.5.0", "40738"): "40738 PaymentStreamType",
        ("FIX.5.0", "40739"): "40739 PaymentStreamMarketRate",
        ("FIX.5.0", "40740"): "40740 PaymentStreamDelayIndicator",
        ("FIX.5.0", "40741"): "40741 PaymentStreamSettlCurrency",
        ("FIX.5.0", "40742"): "40742 PaymentStreamDayCount",
        ("FIX.5.0", "40743"): "40743 PaymentStreamAccrualDays",
        ("FIX.5.0", "40744"): "40744 PaymentStreamDiscountType",
        ("FIX.5.0", "40745"): "40745 PaymentStreamDiscountRate",
        ("FIX.5.0", "40746"): "40746 PaymentStreamDiscountRateDayCount",
        ("FIX.5.0", "40747"): "40747 PaymentStreamCompoundingMethod",
        ("FIX.5.0", "40748"): "40748 PaymentStreamInitialPrincipalExchangeIndicator",
        ("FIX.5.0", "40749"): "40749 PaymentStreamInterimPrincipalExchangeIndicator",
        ("FIX.5.0", "40750"): "40750 PaymentStreamFinalPrincipalExchangeIndicator",
        ("FIX.5.0", "40751"): "40751 PaymentStreamPaymentDateBusinessDayConvention",
        ("FIX.5.0", "40752"): "40752 PaymentStreamPaymentDateBusinessCenter",
        ("FIX.5.0", "40753"): "40753 PaymentStreamPaymentFrequencyPeriod",
        ("FIX.5.0", "40754"): "40754 PaymentStreamPaymentFrequencyUnit",
        ("FIX.5.0", "40755"): "40755 PaymentStreamPaymentRollConvention",
        ("FIX.5.0", "40756"): "40756 PaymentStreamFirstPaymentDateUnadjusted",
        ("FIX.5.0", "40757"): "40757 PaymentStreamLastRegularPaymentDateUnadjusted",
        ("FIX.5.0", "40758"): "40758 PaymentStreamPaymentDateRelativeTo",
        ("FIX.5.0", "40759"): "40759 PaymentStreamPaymentDateOffsetPeriod",
        ("FIX.5.0", "40760"): "40760 PaymentStreamPaymentDateOffsetUnit",
        ("FIX.5.0", "40761"): "40761 PaymentStreamResetDateRelativeTo",
        ("FIX.5.0", "40762"): "40762 PaymentStreamResetDateBusinessDayConvention",
        ("FIX.5.0", "40763"): "40763 PaymentStreamResetDateBusinessCenter",
        ("FIX.5.0", "40764"): "40764 PaymentStreamResetFrequencyPeriod",
        ("FIX.5.0", "40765"): "40765 PaymentStreamResetFrequencyUnit",
        ("FIX.5.0", "40766"): "40766 PaymentStreamResetWeeklyRollConvention",
        ("FIX.5.0", "40767"): "40767 PaymentStreamInitialFixingDateRelativeTo",
        ("FIX.5.0", "40768"): "40768 PaymentStreamInitialFixingDateBusinessDayConvention",
        ("FIX.5.0", "40769"): "40769 PaymentStreamInitialFixingDateBusinessCenter",
        ("FIX.5.0", "40770"): "40770 PaymentStreamInitialFixingDateOffsetPeriod",
        ("FIX.5.0", "40771"): "40771 PaymentStreamInitialFixingDateOffsetUnit",
        ("FIX.5.0", "40772"): "40772 PaymentStreamInitialFixingDateOffsetDayType",
        ("FIX.5.0", "40773"): "40773 PaymentStreamInitialFixingDateAdjusted",
        ("FIX.5.0", "40774"): "40774 PaymentStreamFixingDateRelativeTo",
        ("FIX.5.0", "40775"): "40775 PaymentStreamFixingDateBusinessDayConvention",
        ("FIX.5.0", "40776"): "40776 PaymentStreamFixingDateBusinessCenter",
        ("FIX.5.0", "40777"): "40777 PaymentStreamFixingDateOffsetPeriod",
        ("FIX.5.0", "40778"): "40778 PaymentStreamFixingDateOffsetUnit",
        ("FIX.5.0", "40779"): "40779 PaymentStreamFixingDateOffsetDayType",
        ("FIX.5.0", "40780"): "40780 PaymentStreamFixingDateAdjusted",
        ("FIX.5.0", "40781"): "40781 PaymentStreamRateCutoffDateOffsetPeriod",
        ("FIX.5.0", "40782"): "40782 PaymentStreamRateCutoffDateOffsetUnit",
        ("FIX.5.0", "40783"): "40783 PaymentStreamRateCutoffDateOffsetDayType",
        ("FIX.5.0", "40784"): "40784 PaymentStreamRate",
        ("FIX.5.0", "40785"): "40785 PaymentStreamFixedAmount",
        ("FIX.5.0", "40786"): "40786 PaymentStreamRateOrAmountCurrency",
        ("FIX.5.0", "40787"): "40787 PaymentStreamFutureValueNotional",
        ("FIX.5.0", "40788"): "40788 PaymentStreamFutureValueDateAdjusted",
        ("FIX.5.0", "40789"): "40789 PaymentStreamRateIndex",
        ("FIX.5.0", "40790"): "40790 PaymentStreamRateIndexSource",
        ("FIX.5.0", "40791"): "40791 PaymentStreamRateIndexCurveUnit",
        ("FIX.5.0", "40792"): "40792 PaymentStreamRateIndexCurvePeriod",
        ("FIX.5.0", "40793"): "40793 PaymentStreamRateMultiplier",
        ("FIX.5.0", "40794"): "40794 PaymentStreamRateSpread",
        ("FIX.5.0", "40795"): "40795 PaymentStreamRateSpreadPositionType",
        ("FIX.5.0", "40796"): "40796 PaymentStreamRateTreatment",
        ("FIX.5.0", "40797"): "40797 PaymentStreamCapRate",
        ("FIX.5.0", "40798"): "40798 PaymentStreamCapRateBuySide",
        ("FIX.5.0", "40799"): "40799 PaymentStreamCapRateSellSide",
        ("FIX.5.0", "40800"): "40800 PaymentStreamFloorRate",
        ("FIX.5.0", "40801"): "40801 PaymentStreamFloorRateBuySide",
        ("FIX.5.0", "40802"): "40802 PaymentStreamFloorRateSellSide",
        ("FIX.5.0", "40803"): "40803 PaymentStreamInitialRate",
        ("FIX.5.0", "40804"): "40804 PaymentStreamFinalRateRoundingDirection",
        ("FIX.5.0", "40805"): "40805 PaymentStreamFinalRatePrecision",
        ("FIX.5.0", "40806"): "40806 PaymentStreamAveragingMethod",
        ("FIX.5.0", "40807"): "40807 PaymentStreamNegativeRateTreatment",
        ("FIX.5.0", "40808"): "40808 PaymentStreamInflationLagPeriod",
        ("FIX.5.0", "40809"): "40809 PaymentStreamInflationLagUnit",
        ("FIX.5.0", "40810"): "40810 PaymentStreamInflationLagDayType",
        ("FIX.5.0", "40811"): "40811 PaymentStreamInflationInterpolationMethod",
        ("FIX.5.0", "40812"): "40812 PaymentStreamInflationIndexSource",
        ("FIX.5.0", "40813"): "40813 PaymentStreamInflationPublicationSource",
        ("FIX.5.0", "40814"): "40814 PaymentStreamInflationInitialIndexLevel",
        ("FIX.5.0", "40815"): "40815 PaymentStreamInflationFallbackBondApplicable",
        ("FIX.5.0", "40816"): "40816 PaymentStreamFRADiscounting",
        ("FIX.5.0", "40817"): "40817 PaymentStreamNonDeliverableRefCurrency",
        ("FIX.5.0", "40818"): "40818 PaymentStreamNonDeliverableFixingDatesBusinessDayConvention",
        ("FIX.5.0", "40819"): "40819 PaymentStreamNonDeliverableFixingDatesBusinessCenter",
        ("FIX.5.0", "40820"): "40820 PaymentStreamNonDeliverableFixingDatesRelativeTo",
        ("FIX.5.0", "40821"): "40821 PaymentStreamNonDeliverableFixingDatesOffsetPeriod",
        ("FIX.5.0", "40822"): "40822 PaymentStreamNonDeliverableFixingDatesOffsetUnit",
        ("FIX.5.0", "40823"): "40823 PaymentStreamNonDeliverableFixingDatesOffsetDayType",
        ("FIX.5.0", "40824"): "40824 UnderlyingPaymentStreamNonDeliverableSettlReferencePage",
        ("FIX.5.0", "40825"): "40825 NoNonDeliverableFixingDates",
        ("FIX.5.0", "40826"): "40826 NonDeliverableFixingDate",
        ("FIX.5.0", "40827"): "40827 NonDeliverableFixingDateType",
        ("FIX.5.0", "40828"): "40828 NoPaymentSchedules",
        ("FIX.5.0", "40829"): "40829 PaymentScheduleType",
        ("FIX.5.0", "40830"): "40830 PaymentScheduleStubType",
        ("FIX.5.0", "40831"): "40831 PaymentScheduleStartDateUnadjusted",
        ("FIX.5.0", "40832"): "40832 PaymentScheduleEndDateUnadjusted",
        ("FIX.5.0", "40833"): "40833 PaymentSchedulePaySide",
        ("FIX.5.0", "40834"): "40834 PaymentScheduleReceiveSide",
        ("FIX.5.0", "40835"): "40835 PaymentScheduleNotional",
        ("FIX.5.0", "40836"): "40836 PaymentScheduleCurrency",
        ("FIX.5.0", "40837"): "40837 PaymentScheduleRate",
        ("FIX.5.0", "40838"): "40838 PaymentScheduleRateMultiplier",
        ("FIX.5.0", "40839"): "40839 PaymentScheduleRateSpread",
        ("FIX.5.0", "40840"): "40840 PaymentScheduleRateSpreadPositionType",
        ("FIX.5.0", "40841"): "40841 PaymentScheduleRateTreatment",
        ("FIX.5.0", "40842"): "40842 PaymentScheduleFixedAmount",
        ("FIX.5.0", "40843"): "40843 PaymentScheduleFixedCurrency",
        ("FIX.5.0", "40844"): "40844 PaymentScheduleStepFrequencyPeriod",
        ("FIX.5.0", "40845"): "40845 PaymentScheduleStepFrequencyUnit",
        ("FIX.5.0", "40846"): "40846 PaymentScheduleStepOffsetValue",
        ("FIX.5.0", "40847"): "40847 PaymentScheduleStepRate",
        ("FIX.5.0", "40848"): "40848 PaymentScheduleStepOffsetRate",
        ("FIX.5.0", "40849"): "40849 PaymentScheduleStepRelativeTo",
        ("FIX.5.0", "40850"): "40850 PaymentScheduleFixingDateUnadjusted",
        ("FIX.5.0", "40851"): "40851 PaymentScheduleWeight",
        ("FIX.5.0", "40852"): "40852 PaymentScheduleFixingDateRelativeTo",
        ("FIX.5.0", "40853"): "40853 PaymentScheduleFixingDateBusinessDayConvention",
        ("FIX.5.0", "40854"): "40854 PaymentScheduleFixingDateBusinessCenter",
        ("FIX.5.0", "40855"): "40855 PaymentScheduleFixingDateOffsetPeriod",
        ("FIX.5.0", "40856"): "40856 PaymentScheduleFixingDateOffsetUnit",
        ("FIX.5.0", "40857"): "40857 PaymentScheduleFixingDateOffsetDayType",
        ("FIX.5.0", "40858"): "40858 PaymentScheduleFixingDateAdjusted",
        ("FIX.5.0", "40859"): "40859 PaymentScheduleFixingTime",
        ("FIX.5.0", "40860"): "40860 PaymentScheduleFixingTimeBusinessCenter",
        ("FIX.5.0", "40861"): "40861 PaymentScheduleInterimExchangePaymentDateRelativeTo",
        ("FIX.5.0", "40862"): "40862 PaymentScheduleInterimExchangeDatesBusinessDayConvention",
        ("FIX.5.0", "40863"): "40863 PaymentScheduleInterimExchangeDatesBusinessCenter",
        ("FIX.5.0", "40864"): "40864 PaymentScheduleInterimExchangeDatesOffsetPeriod",
        ("FIX.5.0", "40865"): "40865 PaymentScheduleInterimExchangeDatesOffsetUnit",
        ("FIX.5.0", "40866"): "40866 PaymentScheduleInterimExchangeDatesOffsetDayType",
        ("FIX.5.0", "40867"): "40867 PaymentScheduleInterimExchangeDateAdjusted",
        ("FIX.5.0", "40868"): "40868 NoPaymentScheduleRateSources",
        ("FIX.5.0", "40869"): "40869 PaymentScheduleRateSource",
        ("FIX.5.0", "40870"): "40870 PaymentScheduleRateSourceType",
        ("FIX.5.0", "40871"): "40871 PaymentScheduleReferencePage",
        ("FIX.5.0", "40872"): "40872 NoPaymentStubs",
        ("FIX.5.0", "40873"): "40873 PaymentStubType",
        ("FIX.5.0", "40874"): "40874 PaymentStubLength",
        ("FIX.5.0", "40875"): "40875 PaymentStubRate",
        ("FIX.5.0", "40876"): "40876 PaymentStubFixedAmount",
        ("FIX.5.0", "40877"): "40877 PaymentStubFixedCurrency",
        ("FIX.5.0", "40878"): "40878 PaymentStubIndex",
        ("FIX.5.0", "40879"): "40879 PaymentStubIndexSource",
        ("FIX.5.0", "40880"): "40880 PaymentStubIndexCurvePeriod",
        ("FIX.5.0", "40881"): "40881 PaymentStubIndexCurveUnit",
        ("FIX.5.0", "40882"): "40882 PaymentStubIndexRateMultiplier",
        ("FIX.5.0", "40883"): "40883 PaymentStubIndexRateSpread",
        ("FIX.5.0", "40884"): "40884 PaymentStubIndexRateSpreadPositionType",
        ("FIX.5.0", "40885"): "40885 PaymentStubIndexRateTreatment",
        ("FIX.5.0", "40886"): "40886 PaymentStubIndexCapRate",
        ("FIX.5.0", "40887"): "40887 PaymentStubIndexCapRateBuySide",
        ("FIX.5.0", "40888"): "40888 PaymentStubIndexCapRateSellSide",
        ("FIX.5.0", "40889"): "40889 PaymentStubIndexFloorRate",
        ("FIX.5.0", "40890"): "40890 PaymentStubIndexFloorRateBuySide",
        ("FIX.5.0", "40891"): "40891 PaymentStubIndexFloorRateSellSide",
        ("FIX.5.0", "40892"): "40892 PaymentStubIndex2",
        ("FIX.5.0", "40893"): "40893 PaymentStubIndex2Source",
        ("FIX.5.0", "40894"): "40894 PaymentStubIndex2CurvePeriod",
        ("FIX.5.0", "40895"): "40895 PaymentStubIndex2CurveUnit",
        ("FIX.5.0", "40896"): "40896 PaymentStubIndex2RateMultiplier",
        ("FIX.5.0", "40897"): "40897 PaymentStubIndex2RateSpread",
        ("FIX.5.0", "40898"): "40898 PaymentStubIndex2RateSpreadPositionType",
        ("FIX.5.0", "40899"): "40899 PaymentStubIndex2RateTreatment",
        ("FIX.5.0", "40900"): "40900 PaymentStubIndex2CapRate",
        ("FIX.5.0", "40901"): "40901 PaymentStubIndex2FloorRate",
        ("FIX.5.0", "40902"): "40902 NoLegSettlRateFallbacks",
        ("FIX.5.0", "40903"): "40903 LegSettlRatePostponementMaximumDays",
        ("FIX.5.0", "40904"): "40904 UnderlyingSettlRateFallbackRateSource",
        ("FIX.5.0", "40905"): "40905 LegSettlRatePostponementSurvey",
        ("FIX.5.0", "40906"): "40906 LegSettlRatePostponementCalculationAgent",
        ("FIX.5.0", "40907"): "40907 StreamEffectiveDateUnadjusted",
        ("FIX.5.0", "40908"): "40908 StreamEffectiveDateBusinessDayConvention",
        ("FIX.5.0", "40909"): "40909 StreamEffectiveDateBusinessCenter",
        ("FIX.5.0", "40910"): "40910 StreamEffectiveDateRelativeTo",
        ("FIX.5.0", "40911"): "40911 StreamEffectiveDateOffsetPeriod",
        ("FIX.5.0", "40912"): "40912 StreamEffectiveDateOffsetUnit",
        ("FIX.5.0", "40913"): "40913 StreamEffectiveDateOffsetDayType",
        ("FIX.5.0", "40914"): "40914 StreamEffectiveDateAdjusted",
        ("FIX.5.0", "40915"): "40915 UnderlyingSettlRateFallbackReferencePage",
        ("FIX.5.0", "40916"): "40916 CashSettlValuationSubsequentBusinessDaysOffset",
        ("FIX.5.0", "40917"): "40917 CashSettlNumOfValuationDates",
        ("FIX.5.0", "40918"): "40918 UnderlyingProvisionPartyRoleQualifier",
        ("FIX.5.0", "40919"): "40919 PaymentPriceType",
        ("FIX.5.0", "40920"): "40920 PaymentStreamPaymentDateOffsetDayType",
        ("FIX.5.0", "40921"): "40921 BusinessDayConvention",
        ("FIX.5.0", "40922"): "40922 DateRollConvention",
        ("FIX.5.0", "40923"): "40923 NoLegBusinessCenters",
        ("FIX.5.0", "40924"): "40924 LegBusinessCenter",
        ("FIX.5.0", "40925"): "40925 LegBusinessDayConvention",
        ("FIX.5.0", "40926"): "40926 LegDateRollConvention",
        ("FIX.5.0", "40927"): "40927 NoLegPaymentScheduleFixingDateBusinessCenters",
        ("FIX.5.0", "40928"): "40928 NoLegPaymentScheduleInterimExchangeDateBusinessCenters",
        ("FIX.5.0", "40929"): "40929 NoLegPaymentStreamNonDeliverableFixingDateBusinessCenters",
        ("FIX.5.0", "40930"): "40930 NoLegPaymentStreamPaymentDateBusinessCenters",
        ("FIX.5.0", "40931"): "40931 NoLegPaymentStreamResetDateBusinessCenters",
        ("FIX.5.0", "40932"): "40932 NoLegPaymentStreamInitialFixingDateBusinessCenters",
        ("FIX.5.0", "40933"): "40933 NoLegPaymentStreamFixingDateBusinessCenters",
        ("FIX.5.0", "40934"): "40934 NoLegProvisionCashSettlPaymentDateBusinessCenters",
        ("FIX.5.0", "40935"): "40935 NoLegProvisionCashSettlValueDateBusinessCenters",
        ("FIX.5.0", "40936"): "40936 NoLegProvisionOptionExerciseBusinessCenters",
        ("FIX.5.0", "40937"): "40937 NoLegProvisionOptionExpirationDateBusinessCenters",
        ("FIX.5.0", "40938"): "40938 NoLegProvisionOptionRelevantUnderlyingDateBusinessCenters",
        ("FIX.5.0", "40939"): "40939 NoLegProvisionDateBusinessCenters",
        ("FIX.5.0", "40940"): "40940 NoLegStreamCalculationPeriodBusinessCenters",
        ("FIX.5.0", "40941"): "40941 NoLegStreamFirstPeriodStartDateBusinessCenters",
        ("FIX.5.0", "40942"): "40942 NoLegStreamEffectiveDateBusinessCenters",
        ("FIX.5.0", "40943"): "40943 NoLegStreamTerminationDateBusinessCenters",
        ("FIX.5.0", "40944"): "40944 NoPaymentBusinessCenters",
        ("FIX.5.0", "40945"): "40945 NoPaymentScheduleInterimExchangeDateBusinessCenters",
        ("FIX.5.0", "40946"): "40946 NoPaymentStreamNonDeliverableFixingDatesBusinessCenters",
        ("FIX.5.0", "40947"): "40947 NoPaymentStreamPaymentDateBusinessCenters",
        ("FIX.5.0", "40948"): "40948 NoPaymentStreamResetDateBusinessCenters",
        ("FIX.5.0", "40949"): "40949 NoPaymentStreamInitialFixingDateBusinessCenters",
        ("FIX.5.0", "40950"): "40950 NoPaymentStreamFixingDateBusinessCenters",
        ("FIX.5.0", "40951"): "40951 NoProtectionTermEventNewsSources",
        ("FIX.5.0", "40952"): "40952 NoProvisionCashSettlPaymentDateBusinessCenters",
        ("FIX.5.0", "40953"): "40953 NoProvisionCashSettlValueDateBusinessCenters",
        ("FIX.5.0", "40954"): "40954 NoProvisionOptionExerciseBusinessCenters",
        ("FIX.5.0", "40955"): "40955 NoProvisionOptionExpirationDateBusinessCenters",
        ("FIX.5.0", "40956"): "40956 NoProvisionOptionRelevantUnderlyingDateBusinessCenters",
        ("FIX.5.0", "40957"): "40957 NoProvisionDateBusinessCenters",
        ("FIX.5.0", "40958"): "40958 NoStreamCalculationPeriodBusinessCenters",
        ("FIX.5.0", "40959"): "40959 NoStreamFirstPeriodStartDateBusinessCenters",
        ("FIX.5.0", "40960"): "40960 NoStreamEffectiveBusinessCenters",
        ("FIX.5.0", "40961"): "40961 NoStreamTerminationDateBusinessCenters",
        ("FIX.5.0", "40962"): "40962 NoUnderlyingBusinessCenters",
        ("FIX.5.0", "40963"): "40963 UnderlyingBusinessCenter",
        ("FIX.5.0", "40964"): "40964 UnderlyingBusinessDayConvention",
        ("FIX.5.0", "40965"): "40965 UnderlyingDateRollConvention",
        ("FIX.5.0", "40966"): "40966 NoUnderlyingPaymentScheduleFixingDateBusinessCenters",
        ("FIX.5.0", "40967"): "40967 NoUnderlyingPaymentScheduleInterimExchangeDateBusinessCenters",
        ("FIX.5.0", "40968"): "40968 NoUnderlyingPaymentStreamNonDeliverableFixingDatesBusinessCenters",
        ("FIX.5.0", "40969"): "40969 NoUnderlyingPaymentStreamPaymentDateBusinessCenters",
        ("FIX.5.0", "40970"): "40970 NoUnderlyingPaymentStreamResetDateBusinessCenters",
        ("FIX.5.0", "40971"): "40971 NoUnderlyingPaymentStreamInitialFixingDateBusinessCenters",
        ("FIX.5.0", "40972"): "40972 NoUnderlyingPaymentStreamFixingDateBusinessCenters",
        ("FIX.5.0", "40973"): "40973 NoUnderlyingStreamCalculationPeriodBusinessCenters",
        ("FIX.5.0", "40974"): "40974 NoUnderlyingStreamFirstPeriodStartDateBusinessCenters",
        ("FIX.5.0", "40975"): "40975 NoUnderlyingStreamEffectiveDateBusinessCenters",
        ("FIX.5.0", "40976"): "40976 NoUnderlyingStreamTerminationDateBusinessCenters",
        ("FIX.5.0", "40977"): "40977 NoPaymentScheduleFixingDateBusinessCenters",
        ("FIX.5.0", "40978"): "40978 EncodedLegStreamTextLen",
        ("FIX.5.0", "40979"): "40979 EncodedLegStreamText",
        ("FIX.5.0", "40980"): "40980 EncodedLegProvisionTextLen",
        ("FIX.5.0", "40981"): "40981 EncodedLegProvisionText",
        ("FIX.5.0", "40982"): "40982 EncodedStreamTextLen",
        ("FIX.5.0", "40983"): "40983 EncodedStreamText",
        ("FIX.5.0", "40984"): "40984 EncodedPaymentTextLen",
        ("FIX.5.0", "40985"): "40985 EncodedPaymentText",
        ("FIX.5.0", "40986"): "40986 EncodedProvisionTextLen",
        ("FIX.5.0", "40987"): "40987 EncodedProvisionText",
        ("FIX.5.0", "40988"): "40988 EncodedUnderlyingStreamTextLen",
        ("FIX.5.0", "40989"): "40989 EncodedUnderlyingStreamText",
        ("FIX.5.0", "40990"): "40990 LegMarketDisruptionFallbackValue",
        ("FIX.5.0", "40991"): "40991 MarketDisruptionValue",
        ("FIX.5.0", "40992"): "40992 MarketDisruptionFallbackValue",
        ("FIX.5.0", "40993"): "40993 PaymentSubType",
        ("FIX.5.0", "40994"): "40994 NoComplexEventAveragingObservations",
        ("FIX.5.0", "40995"): "40995 ComplexEventAveragingObservationNumber",
        ("FIX.5.0", "40996"): "40996 ComplexEventAveragingWeight",
        ("FIX.5.0", "40997"): "40997 NoComplexEventCreditEvents",
        ("FIX.5.0", "40998"): "40998 ComplexEventCreditEventType",
        ("FIX.5.0", "40999"): "40999 ComplexEventCreditEventValue",
        ("FIX.5.0", "41000"): "41000 ComplexEventCreditEventCurrency",
        ("FIX.5.0", "41001"): "41001 ComplexEventCreditEventPeriod",
        ("FIX.5.0", "41002"): "41002 ComplexEventCreditEventUnit",
        ("FIX.5.0", "41003"): "41003 ComplexEventCreditEventDayType",
        ("FIX.5.0", "41004"): "41004 ComplexEventCreditEventRateSource",
        ("FIX.5.0", "41005"): "41005 NoComplexEventCreditEventQualifiers",
        ("FIX.5.0", "41006"): "41006 ComplexEventCreditEventQualifier",
        ("FIX.5.0", "41007"): "41007 NoComplexEventPeriodDateTimes",
        ("FIX.5.0", "41008"): "41008 ComplexEventPeriodDate",
        ("FIX.5.0", "41009"): "41009 ComplexEventPeriodTime",
        ("FIX.5.0", "41010"): "41010 NoComplexEventPeriods",
        ("FIX.5.0", "41011"): "41011 ComplexEventPeriodType",
        ("FIX.5.0", "41012"): "41012 ComplexEventBusinessCenter",
        ("FIX.5.0", "41013"): "41013 NoComplexEventRateSources",
        ("FIX.5.0", "41014"): "41014 ComplexEventRateSource",
        ("FIX.5.0", "41015"): "41015 ComplexEventRateSourceType",
        ("FIX.5.0", "41016"): "41016 ComplexEventReferencePage",
        ("FIX.5.0", "41017"): "41017 ComplexEventReferencePageHeading",
        ("FIX.5.0", "41018"): "41018 NoComplexEventDateBusinessCenters",
        ("FIX.5.0", "41019"): "41019 ComplexEventDateBusinessCenter",
        ("FIX.5.0", "41020"): "41020 ComplexEventDateUnadjusted",
        ("FIX.5.0", "41021"): "41021 ComplexEventDateRelativeTo",
        ("FIX.5.0", "41022"): "41022 ComplexEventDateOffsetPeriod",
        ("FIX.5.0", "41023"): "41023 ComplexEventDateOffsetUnit",
        ("FIX.5.0", "41024"): "41024 ComplexEventDateOffsetDayType",
        ("FIX.5.0", "41025"): "41025 ComplexEventDateBusinessDayConvention",
        ("FIX.5.0", "41026"): "41026 ComplexEventDateAdjusted",
        ("FIX.5.0", "41027"): "41027 ComplexEventFixingTime",
        ("FIX.5.0", "41028"): "41028 ComplexEventFixingTimeBusinessCenter",
        ("FIX.5.0", "41029"): "41029 NoComplexEventCreditEventSources",
        ("FIX.5.0", "41030"): "41030 ComplexEventCreditEventSource",
        ("FIX.5.0", "41031"): "41031 NoComplexEventSchedules",
        ("FIX.5.0", "41032"): "41032 ComplexEventScheduleStartDate",
        ("FIX.5.0", "41033"): "41033 ComplexEventScheduleEndDate",
        ("FIX.5.0", "41034"): "41034 ComplexEventScheduleFrequencyPeriod",
        ("FIX.5.0", "41035"): "41035 ComplexEventScheduleFrequencyUnit",
        ("FIX.5.0", "41036"): "41036 ComplexEventScheduleRollConvention",
        ("FIX.5.0", "41037"): "41037 NoDeliverySchedules",
        ("FIX.5.0", "41038"): "41038 DeliveryScheduleType",
        ("FIX.5.0", "41039"): "41039 DeliveryScheduleXID",
        ("FIX.5.0", "41040"): "41040 DeliveryScheduleNotional",
        ("FIX.5.0", "41041"): "41041 DeliveryScheduleNotionalUnitOfMeasure",
        ("FIX.5.0", "41042"): "41042 DeliveryScheduleNotionalCommodityFrequency",
        ("FIX.5.0", "41043"): "41043 DeliveryScheduleNegativeTolerance",
        ("FIX.5.0", "41044"): "41044 DeliverySchedulePositiveTolerance",
        ("FIX.5.0", "41045"): "41045 DeliveryScheduleToleranceUnitOfMeasure",
        ("FIX.5.0", "41046"): "41046 DeliveryScheduleToleranceType",
        ("FIX.5.0", "41047"): "41047 DeliveryScheduleSettlCountry",
        ("FIX.5.0", "41048"): "41048 DeliveryScheduleSettlTimeZone",
        ("FIX.5.0", "41049"): "41049 DeliveryScheduleSettlFlowType",
        ("FIX.5.0", "41050"): "41050 DeliveryScheduleSettlHolidaysProcessingInstruction",
        ("FIX.5.0", "41051"): "41051 NoDeliveryScheduleSettlDays",
        ("FIX.5.0", "41052"): "41052 DeliveryScheduleSettlDay",
        ("FIX.5.0", "41053"): "41053 DeliveryScheduleSettlTotalHours",
        ("FIX.5.0", "41054"): "41054 NoDeliveryScheduleSettlTimes",
        ("FIX.5.0", "41055"): "41055 DeliveryScheduleSettlStart",
        ("FIX.5.0", "41056"): "41056 DeliveryScheduleSettlEnd",
        ("FIX.5.0", "41057"): "41057 DeliveryScheduleSettlTimeType",
        ("FIX.5.0", "41058"): "41058 DeliveryStreamType",
        ("FIX.5.0", "41059"): "41059 DeliveryStreamPipeline",
        ("FIX.5.0", "41060"): "41060 DeliveryStreamEntryPoint",
        ("FIX.5.0", "41061"): "41061 DeliveryStreamWithdrawalPoint",
        ("FIX.5.0", "41062"): "41062 DeliveryStreamDeliveryPoint",
        ("FIX.5.0", "41063"): "41063 DeliveryStreamDeliveryRestriction",
        ("FIX.5.0", "41064"): "41064 DeliveryStreamDeliveryContingency",
        ("FIX.5.0", "41065"): "41065 DeliveryStreamDeliveryContingentPartySide",
        ("FIX.5.0", "41066"): "41066 DeliveryStreamDeliverAtSourceIndicator",
        ("FIX.5.0", "41067"): "41067 DeliveryStreamRiskApportionment",
        ("FIX.5.0", "41068"): "41068 DeliveryStreamTitleTransferLocation",
        ("FIX.5.0", "41069"): "41069 DeliveryStreamTitleTransferCondition",
        ("FIX.5.0", "41070"): "41070 DeliveryStreamImporterOfRecord",
        ("FIX.5.0", "41071"): "41071 DeliveryStreamNegativeTolerance",
        ("FIX.5.0", "41072"): "41072 DeliveryStreamPositiveTolerance",
        ("FIX.5.0", "41073"): "41073 DeliveryStreamToleranceUnitOfMeasure",
        ("FIX.5.0", "41074"): "41074 DeliveryStreamToleranceType",
        ("FIX.5.0", "41075"): "41075 DeliveryStreamToleranceOptionSide",
        ("FIX.5.0", "41076"): "41076 DeliveryStreamTotalPositiveTolerance",
        ("FIX.5.0", "41077"): "41077 DeliveryStreamTotalNegativeTolerance",
        ("FIX.5.0", "41078"): "41078 DeliveryStreamNotionalConversionFactor",
        ("FIX.5.0", "41079"): "41079 DeliveryStreamTransportEquipment",
        ("FIX.5.0", "41080"): "41080 DeliveryStreamElectingPartySide",
        ("FIX.5.0", "41081"): "41081 NoDeliveryStreamCycles",
        ("FIX.5.0", "41082"): "41082 DeliveryStreamCycleDesc",
        ("FIX.5.0", "41083"): "41083 EncodedDeliveryStreamCycleDescLen",
        ("FIX.5.0", "41084"): "41084 EncodedDeliveryStreamCycleDesc",
        ("FIX.5.0", "41085"): "41085 NoDeliveryStreamCommoditySources",
        ("FIX.5.0", "41086"): "41086 DeliveryStreamCommoditySource",
        ("FIX.5.0", "41087"): "41087 MarketDisruptionProvision",
        ("FIX.5.0", "41088"): "41088 MarketDisruptionFallbackProvision",
        ("FIX.5.0", "41089"): "41089 MarketDisruptionMaximumDays",
        ("FIX.5.0", "41090"): "41090 MarketDisruptionMaterialityPercentage",
        ("FIX.5.0", "41091"): "41091 MarketDisruptionMinimumFuturesContracts",
        ("FIX.5.0", "41092"): "41092 NoMarketDisruptionEvents",
        ("FIX.5.0", "41093"): "41093 MarketDisruptionEvent",
        ("FIX.5.0", "41094"): "41094 NoMarketDisruptionFallbacks",
        ("FIX.5.0", "41095"): "41095 MarketDisruptionFallbackType",
        ("FIX.5.0", "41096"): "41096 NoMarketDisruptionFallbackReferencePrices",
        ("FIX.5.0", "41097"): "41097 MarketDisruptionFallbackUnderlierType",
        ("FIX.5.0", "41098"): "41098 MarketDisruptionFallbackUnderlierSecurityID",
        ("FIX.5.0", "41099"): "41099 MarketDisruptionFallbackUnderlierSecurityIDSource",
        ("FIX.5.0", "41100"): "41100 MarketDisruptionFallbackUnderlierSecurityDesc",
        ("FIX.5.0", "41101"): "41101 EncodedMarketDisruptionFallbackUnderlierSecurityDescLen",
        ("FIX.5.0", "41102"): "41102 EncodedMarketDisruptionFallbackUnderlierSecurityDesc",
        ("FIX.5.0", "41103"): "41103 MarketDisruptionFallbackOpenUnits",
        ("FIX.5.0", "41104"): "41104 MarketDisruptionFallbackBasketCurrency",
        ("FIX.5.0", "41105"): "41105 MarketDisruptionFallbackBasketDivisor",
        ("FIX.5.0", "41106"): "41106 ExerciseDesc",
        ("FIX.5.0", "41107"): "41107 EncodedExerciseDescLen",
        ("FIX.5.0", "41108"): "41108 EncodedExerciseDesc",
        ("FIX.5.0", "41109"): "41109 AutomaticExerciseIndicator",
        ("FIX.5.0", "41110"): "41110 AutomaticExerciseThresholdRate",
        ("FIX.5.0", "41111"): "41111 ExerciseConfirmationMethod",
        ("FIX.5.0", "41112"): "41112 ManualNoticeBusinessCenter",
        ("FIX.5.0", "41113"): "41113 FallbackExerciseIndicator",
        ("FIX.5.0", "41114"): "41114 LimitedRightToConfirmIndicator",
        ("FIX.5.0", "41115"): "41115 ExerciseSplitTicketIndicator",
        ("FIX.5.0", "41116"): "41116 NoOptionExerciseBusinessCenters",
        ("FIX.5.0", "41117"): "41117 OptionExerciseBusinessCenter",
        ("FIX.5.0", "41118"): "41118 OptionExerciseBusinessDayConvention",
        ("FIX.5.0", "41119"): "41119 OptionExerciseEarliestDateOffsetDayType",
        ("FIX.5.0", "41120"): "41120 OptionExerciseEarliestDateOffsetPeriod",
        ("FIX.5.0", "41121"): "41121 OptionExerciseEarliestDateOffsetUnit",
        ("FIX.5.0", "41122"): "41122 OptionExerciseFrequencyPeriod",
        ("FIX.5.0", "41123"): "41123 OptionExerciseFrequencyUnit",
        ("FIX.5.0", "41124"): "41124 OptionExerciseStartDateUnadjusted",
        ("FIX.5.0", "41125"): "41125 OptionExerciseStartDateRelativeTo",
        ("FIX.5.0", "41126"): "41126 OptionExerciseStartDateOffsetPeriod",
        ("FIX.5.0", "41127"): "41127 OptionExerciseStartDateOffsetUnit",
        ("FIX.5.0", "41128"): "41128 OptionExerciseStartDateOffsetDayType",
        ("FIX.5.0", "41129"): "41129 OptionExerciseStartDateAdjusted",
        ("FIX.5.0", "41130"): "41130 OptionExerciseSkip",
        ("FIX.5.0", "41131"): "41131 OptionExerciseNominationDeadline",
        ("FIX.5.0", "41132"): "41132 OptionExerciseFirstDateUnadjusted",
        ("FIX.5.0", "41133"): "41133 OptionExerciseLastDateUnadjusted",
        ("FIX.5.0", "41134"): "41134 OptionExerciseEarliestTime",
        ("FIX.5.0", "41135"): "41135 OptionExerciseLatestTime",
        ("FIX.5.0", "41136"): "41136 OptionExerciseTimeBusinessCenter",
        ("FIX.5.0", "41137"): "41137 NoOptionExerciseDates",
        ("FIX.5.0", "41138"): "41138 OptionExerciseDate",
        ("FIX.5.0", "41139"): "41139 OptionExerciseDateType",
        ("FIX.5.0", "41140"): "41140 NoOptionExerciseExpirationDateBusinessCenters",
        ("FIX.5.0", "41141"): "41141 OptionExerciseExpirationDateBusinessCenter",
        ("FIX.5.0", "41142"): "41142 OptionExerciseExpirationDateBusinessDayConvention",
        ("FIX.5.0", "41143"): "41143 OptionExerciseExpirationDateRelativeTo",
        ("FIX.5.0", "41144"): "41144 OptionExerciseExpirationDateOffsetPeriod",
        ("FIX.5.0", "41145"): "41145 OptionExerciseExpirationDateOffsetUnit",
        ("FIX.5.0", "41146"): "41146 OptionExerciseExpirationFrequencyPeriod",
        ("FIX.5.0", "41147"): "41147 OptionExerciseExpirationFrequencyUnit",
        ("FIX.5.0", "41148"): "41148 OptionExerciseExpirationRollConvention",
        ("FIX.5.0", "41149"): "41149 OptionExerciseExpirationDateOffsetDayType",
        ("FIX.5.0", "41150"): "41150 OptionExerciseExpirationTime",
        ("FIX.5.0", "41151"): "41151 OptionExerciseExpirationTimeBusinessCenter",
        ("FIX.5.0", "41152"): "41152 NoOptionExerciseExpirationDates",
        ("FIX.5.0", "41153"): "41153 OptionExerciseExpirationDate",
        ("FIX.5.0", "41154"): "41154 OptionExerciseExpirationDateType",
        ("FIX.5.0", "41155"): "41155 PaymentUnitOfMeasure",
        ("FIX.5.0", "41156"): "41156 PaymentDateRelativeTo",
        ("FIX.5.0", "41157"): "41157 PaymentDateOffsetPeriod",
        ("FIX.5.0", "41158"): "41158 PaymentDateOffsetUnit",
        ("FIX.5.0", "41159"): "41159 PaymentDateOffsetDayType",
        ("FIX.5.0", "41160"): "41160 PaymentForwardStartType",
        ("FIX.5.0", "41161"): "41161 NoPaymentScheduleFixingDays",
        ("FIX.5.0", "41162"): "41162 PaymentScheduleFixingDayOfWeek",
        ("FIX.5.0", "41163"): "41163 PaymentScheduleFixingDayNumber",
        ("FIX.5.0", "41164"): "41164 PaymentScheduleXID",
        ("FIX.5.0", "41165"): "41165 PaymentScheduleXIDRef",
        ("FIX.5.0", "41166"): "41166 PaymentScheduleRateCurrency",
        ("FIX.5.0", "41167"): "41167 PaymentScheduleRateUnitOfMeasure",
        ("FIX.5.0", "41168"): "41168 PaymentScheduleRateConversionFactor",
        ("FIX.5.0", "41169"): "41169 PaymentScheduleRateSpreadType",
        ("FIX.5.0", "41170"): "41170 PaymentScheduleSettlPeriodPrice",
        ("FIX.5.0", "41171"): "41171 PaymentScheduleSettlPeriodPriceCurrency",
        ("FIX.5.0", "41172"): "41172 PaymentScheduleSettlPeriodPriceUnitOfMeasure",
        ("FIX.5.0", "41173"): "41173 PaymentScheduleStepUnitOfMeasure",
        ("FIX.5.0", "41174"): "41174 PaymentScheduleFixingDayDistribution",
        ("FIX.5.0", "41175"): "41175 PaymentScheduleFixingDayCount",
        ("FIX.5.0", "41176"): "41176 PaymentScheduleFixingLagPeriod",
        ("FIX.5.0", "41177"): "41177 PaymentScheduleFixingLagUnit",
        ("FIX.5.0", "41178"): "41178 PaymentScheduleFixingFirstObservationDateOffsetPeriod",
        ("FIX.5.0", "41179"): "41179 PaymentScheduleFixingFirstObservationDateOffsetUnit",
        ("FIX.5.0", "41180"): "41180 PaymentStreamFlatRateIndicator",
        ("FIX.5.0", "41181"): "41181 PaymentStreamFlatRateAmount",
        ("FIX.5.0", "41182"): "41182 PaymentStreamFlatRateCurrency",
        ("FIX.5.0", "41183"): "41183 PaymentStreamMaximumPaymentAmount",
        ("FIX.5.0", "41184"): "41184 PaymentStreamMaximumPaymentCurrency",
        ("FIX.5.0", "41185"): "41185 PaymentStreamMaximumTransactionAmount",
        ("FIX.5.0", "41186"): "41186 PaymentStreamMaximumTransactionCurrency",
        ("FIX.5.0", "41187"): "41187 PaymentStreamFixedAmountUnitOfMeasure",
        ("FIX.5.0", "41188"): "41188 PaymentStreamTotalFixedAmount",
        ("FIX.5.0", "41189"): "41189 PaymentStreamWorldScaleRate",
        ("FIX.5.0", "41190"): "41190 PaymentStreamContractPrice",
        ("FIX.5.0", "41191"): "41191 PaymentStreamContractPriceCurrency",
        ("FIX.5.0", "41192"): "41192 NoPaymentStreamPricingBusinessCenters",
        ("FIX.5.0", "41193"): "41193 PaymentStreamPricingBusinessCenter",
        ("FIX.5.0", "41194"): "41194 PaymentStreamRateIndex2CurvePeriod",
        ("FIX.5.0", "41195"): "41195 PaymentStreamRateIndex2CurveUnit",
        ("FIX.5.0", "41196"): "41196 PaymentStreamRateIndexLocation",
        ("FIX.5.0", "41197"): "41197 PaymentStreamRateIndexLevel",
        ("FIX.5.0", "41198"): "41198 PaymentStreamRateIndexUnitOfMeasure",
        ("FIX.5.0", "41199"): "41199 PaymentStreamSettlLevel",
        ("FIX.5.0", "41200"): "41200 PaymentStreamReferenceLevel",
        ("FIX.5.0", "41201"): "41201 PaymentStreamReferenceLevelUnitOfMeasure",
        ("FIX.5.0", "41202"): "41202 PaymentStreamReferenceLevelEqualsZeroIndicator",
        ("FIX.5.0", "41203"): "41203 PaymentStreamRateSpreadCurrency",
        ("FIX.5.0", "41204"): "41204 PaymentStreamRateSpreadUnitOfMeasure",
        ("FIX.5.0", "41205"): "41205 PaymentStreamRateConversionFactor",
        ("FIX.5.0", "41206"): "41206 PaymentStreamRateSpreadType",
        ("FIX.5.0", "41207"): "41207 PaymentStreamLastResetRate",
        ("FIX.5.0", "41208"): "41208 PaymentStreamFinalRate",
        ("FIX.5.0", "41209"): "41209 PaymentStreamCalculationLagPeriod",
        ("FIX.5.0", "41210"): "41210 PaymentStreamCalculationLagUnit",
        ("FIX.5.0", "41211"): "41211 PaymentStreamFirstObservationDateOffsetPeriod",
        ("FIX.5.0", "41212"): "41212 PaymentStreamFirstObservationDateOffsetUnit",
        ("FIX.5.0", "41213"): "41213 PaymentStreamPricingDayType",
        ("FIX.5.0", "41214"): "41214 PaymentStreamPricingDayDistribution",
        ("FIX.5.0", "41215"): "41215 PaymentStreamPricingDayCount",
        ("FIX.5.0", "41216"): "41216 PaymentStreamPricingBusinessCalendar",
        ("FIX.5.0", "41217"): "41217 PaymentStreamPricingBusinessDayConvention",
        ("FIX.5.0", "41218"): "41218 DeliveryStreamRiskApportionmentSource",
        ("FIX.5.0", "41219"): "41219 LegDeliveryStreamRiskApportionmentSource",
        ("FIX.5.0", "41220"): "41220 NoPaymentStreamPaymentDates",
        ("FIX.5.0", "41221"): "41221 PaymentStreamPaymentDate",
        ("FIX.5.0", "41222"): "41222 PaymentStreamPaymentDateType",
        ("FIX.5.0", "41223"): "41223 PaymentStreamMasterAgreementPaymentDatesIndicator",
        ("FIX.5.0", "41224"): "41224 NoPaymentStreamPricingDates",
        ("FIX.5.0", "41225"): "41225 PaymentStreamPricingDate",
        ("FIX.5.0", "41226"): "41226 PaymentStreamPricingDateType",
        ("FIX.5.0", "41227"): "41227 NoPaymentStreamPricingDays",
        ("FIX.5.0", "41228"): "41228 PaymentStreamPricingDayOfWeek",
        ("FIX.5.0", "41229"): "41229 PaymentStreamPricingDayNumber",
        ("FIX.5.0", "41230"): "41230 NoPricingDateBusinessCenters",
        ("FIX.5.0", "41231"): "41231 PricingDateBusinessCenter",
        ("FIX.5.0", "41232"): "41232 PricingDateUnadjusted",
        ("FIX.5.0", "41233"): "41233 PricingDateBusinessDayConvention",
        ("FIX.5.0", "41234"): "41234 PricingDateAdjusted",
        ("FIX.5.0", "41235"): "41235 PricingTime",
        ("FIX.5.0", "41236"): "41236 PricingTimeBusinessCenter",
        ("FIX.5.0", "41237"): "41237 NoStreamAssetAttributes",
        ("FIX.5.0", "41238"): "41238 StreamAssetAttributeType",
        ("FIX.5.0", "41239"): "41239 StreamAssetAttributeValue",
        ("FIX.5.0", "41240"): "41240 StreamAssetAttributeLimit",
        ("FIX.5.0", "41241"): "41241 NoStreamCalculationPeriodDates",
        ("FIX.5.0", "41242"): "41242 StreamCalculationPeriodDate",
        ("FIX.5.0", "41243"): "41243 StreamCalculationPeriodDateType",
        ("FIX.5.0", "41244"): "41244 StreamCalculationPeriodDatesXID",
        ("FIX.5.0", "41245"): "41245 StreamCalculationPeriodDatesXIDRef",
        ("FIX.5.0", "41246"): "41246 StreamCalculationBalanceOfFirstPeriod",
        ("FIX.5.0", "41247"): "41247 StreamCalculationCorrectionPeriod",
        ("FIX.5.0", "41248"): "41248 StreamCalculationCorrectionUnit",
        ("FIX.5.0", "41249"): "41249 NoStreamCommoditySettlBusinessCenters",
        ("FIX.5.0", "41250"): "41250 StreamCommoditySettlBusinessCenter",
        ("FIX.5.0", "41251"): "41251 StreamCommodityBase",
        ("FIX.5.0", "41252"): "41252 StreamCommodityType",
        ("FIX.5.0", "41253"): "41253 StreamCommoditySecurityID",
        ("FIX.5.0", "41254"): "41254 StreamCommoditySecurityIDSource",
        ("FIX.5.0", "41255"): "41255 StreamCommodityDesc",
        ("FIX.5.0", "41256"): "41256 EncodedStreamCommodityDescLen",
        ("FIX.5.0", "41257"): "41257 EncodedStreamCommodityDesc",
        ("FIX.5.0", "41258"): "41258 StreamCommodityUnitOfMeasure",
        ("FIX.5.0", "41259"): "41259 StreamCommodityCurrency",
        ("FIX.5.0", "41260"): "41260 StreamCommodityExchange",
        ("FIX.5.0", "41261"): "41261 StreamCommodityRateSource",
        ("FIX.5.0", "41262"): "41262 StreamCommodityRateReferencePage",
        ("FIX.5.0", "41263"): "41263 StreamCommodityRateReferencePageHeading",
        ("FIX.5.0", "41264"): "41264 StreamDataProvider",
        ("FIX.5.0", "41265"): "41265 StreamCommodityPricingType",
        ("FIX.5.0", "41266"): "41266 StreamCommodityNearbySettlDayPeriod",
        ("FIX.5.0", "41267"): "41267 StreamCommodityNearbySettlDayUnit",
        ("FIX.5.0", "41268"): "41268 StreamCommoditySettlDateUnadjusted",
        ("FIX.5.0", "41269"): "41269 StreamCommoditySettlDateBusinessDayConvention",
        ("FIX.5.0", "41270"): "41270 StreamCommoditySettlDateAdjusted",
        ("FIX.5.0", "41271"): "41271 StreamCommoditySettlMonth",
        ("FIX.5.0", "41272"): "41272 StreamCommoditySettlDateRollPeriod",
        ("FIX.5.0", "41273"): "41273 StreamCommoditySettlDateRollUnit",
        ("FIX.5.0", "41274"): "41274 StreamCommoditySettlDayType",
        ("FIX.5.0", "41275"): "41275 StreamCommodityXID",
        ("FIX.5.0", "41276"): "41276 StreamCommodityXIDRef",
        ("FIX.5.0", "41277"): "41277 NoStreamCommodityAltIDs",
        ("FIX.5.0", "41278"): "41278 StreamCommodityAltID",
        ("FIX.5.0", "41279"): "41279 StreamCommodityAltIDSource",
        ("FIX.5.0", "41280"): "41280 NoStreamCommodityDataSources",
        ("FIX.5.0", "41281"): "41281 StreamCommodityDataSourceID",
        ("FIX.5.0", "41282"): "41282 StreamCommodityDataSourceIDType",
        ("FIX.5.0", "41283"): "41283 NoStreamCommoditySettlDays",
        ("FIX.5.0", "41284"): "41284 StreamCommoditySettlDay",
        ("FIX.5.0", "41285"): "41285 StreamCommoditySettlTotalHours",
        ("FIX.5.0", "41286"): "41286 NoStreamCommoditySettlTimes",
        ("FIX.5.0", "41287"): "41287 StreamCommoditySettlStart",
        ("FIX.5.0", "41288"): "41288 StreamCommoditySettlEnd",
        ("FIX.5.0", "41289"): "41289 NoStreamCommoditySettlPeriods",
        ("FIX.5.0", "41290"): "41290 StreamCommoditySettlCountry",
        ("FIX.5.0", "41291"): "41291 StreamCommoditySettlTimeZone",
        ("FIX.5.0", "41292"): "41292 StreamCommoditySettlFlowType",
        ("FIX.5.0", "41293"): "41293 StreamCommoditySettlPeriodNotional",
        ("FIX.5.0", "41294"): "41294 StreamCommoditySettlPeriodNotionalUnitOfMeasure",
        ("FIX.5.0", "41295"): "41295 StreamCommoditySettlPeriodFrequencyPeriod",
        ("FIX.5.0", "41296"): "41296 StreamCommoditySettlPeriodFrequencyUnit",
        ("FIX.5.0", "41297"): "41297 StreamCommoditySettlPeriodPrice",
        ("FIX.5.0", "41298"): "41298 StreamCommoditySettlPeriodPriceUnitOfMeasure",
        ("FIX.5.0", "41299"): "41299 StreamCommoditySettlPeriodPriceCurrency",
        ("FIX.5.0", "41300"): "41300 StreamCommoditySettlHolidaysProcessingInstruction",
        ("FIX.5.0", "41301"): "41301 StreamCommoditySettlPeriodXID",
        ("FIX.5.0", "41302"): "41302 StreamCommoditySettlPeriodXIDRef",
        ("FIX.5.0", "41303"): "41303 StreamXID",
        ("FIX.5.0", "41304"): "41304 PaymentLegRefID",
        ("FIX.5.0", "41305"): "41305 StreamNotionalXIDRef",
        ("FIX.5.0", "41306"): "41306 StreamNotionalFrequencyPeriod",
        ("FIX.5.0", "41307"): "41307 StreamNotionalFrequencyUnit",
        ("FIX.5.0", "41308"): "41308 StreamNotionalCommodityFrequency",
        ("FIX.5.0", "41309"): "41309 StreamNotionalUnitOfMeasure",
        ("FIX.5.0", "41310"): "41310 StreamTotalNotional",
        ("FIX.5.0", "41311"): "41311 StreamTotalNotionalUnitOfMeasure",
        ("FIX.5.0", "41312"): "41312 NoMandatoryClearingJurisdictions",
        ("FIX.5.0", "41313"): "41313 MandatoryClearingJurisdiction",
        ("FIX.5.0", "41314"): "41314 UnderlyingProtectionTermXIDRef",
        ("FIX.5.0", "41315"): "41315 UnderlyingSettlTermXIDRef",
        ("FIX.5.0", "41316"): "41316 NoLegAdditionalTermBondRefs",
        ("FIX.5.0", "41317"): "41317 LegAdditionalTermBondSecurityID",
        ("FIX.5.0", "41318"): "41318 LegAdditionalTermBondSecurityIDSource",
        ("FIX.5.0", "41319"): "41319 LegAdditionalTermBondDesc",
        ("FIX.5.0", "41320"): "41320 EncodedLegAdditionalTermBondDescLen",
        ("FIX.5.0", "41321"): "41321 EncodedLegAdditionalTermBondDesc",
        ("FIX.5.0", "41322"): "41322 LegAdditionalTermBondCurrency",
        ("FIX.5.0", "41323"): "41323 LegAdditionalTermBondIssuer",
        ("FIX.5.0", "41324"): "41324 EncodedLegAdditionalTermBondIssuerLen",
        ("FIX.5.0", "41325"): "41325 EncodedLegAdditionalTermBondIssuer",
        ("FIX.5.0", "41326"): "41326 LegAdditionalTermBondSeniority",
        ("FIX.5.0", "41327"): "41327 LegAdditionalTermBondCouponType",
        ("FIX.5.0", "41328"): "41328 LegAdditionalTermBondCouponRate",
        ("FIX.5.0", "41329"): "41329 LegAdditionalTermBondMaturityDate",
        ("FIX.5.0", "41330"): "41330 LegAdditionalTermBondParValue",
        ("FIX.5.0", "41331"): "41331 LegAdditionalTermBondCurrentTotalIssuedAmount",
        ("FIX.5.0", "41332"): "41332 LegAdditionalTermBondCouponFrequencyPeriod",
        ("FIX.5.0", "41333"): "41333 LegAdditionalTermBondCouponFrequencyUnit",
        ("FIX.5.0", "41334"): "41334 LegAdditionalTermBondDayCount",
        ("FIX.5.0", "41335"): "41335 NoLegAdditionalTerms",
        ("FIX.5.0", "41336"): "41336 LegAdditionalTermConditionPrecedentBondIndicator",
        ("FIX.5.0", "41337"): "41337 LegAdditionalTermDiscrepancyClauseIndicator",
        ("FIX.5.0", "41338"): "41338 UnderlyingMarketDisruptionValue",
        ("FIX.5.0", "41339"): "41339 UnderlyingMarketDisruptionFallbackValue",
        ("FIX.5.0", "41340"): "41340 NoUnderlyingAdditionalTermBondRefs",
        ("FIX.5.0", "41341"): "41341 UnderlyingAdditionalTermBondSecurityID",
        ("FIX.5.0", "41342"): "41342 NoLegCashSettlDealers",
        ("FIX.5.0", "41343"): "41343 LegCashSettlDealer",
        ("FIX.5.0", "41344"): "41344 NoLegCashSettlTerms",
        ("FIX.5.0", "41345"): "41345 LegCashSettlCurrency",
        ("FIX.5.0", "41346"): "41346 LegCasSettlValuationFirstBusinessDayOffset",
        ("FIX.5.0", "41347"): "41347 LegCashSettlValuationSubsequentBusinessDaysOffset",
        ("FIX.5.0", "41348"): "41348 LegCashSettlNumOfValuationDates",
        ("FIX.5.0", "41349"): "41349 LegCashSettlValuationTime",
        ("FIX.5.0", "41350"): "41350 LegCashSettlBusinessCenter",
        ("FIX.5.0", "41351"): "41351 LegCashSettlQuoteMethod",
        ("FIX.5.0", "41352"): "41352 LegCashSettlQuoteAmount",
        ("FIX.5.0", "41353"): "41353 LegCashSettlQuoteCurrency",
        ("FIX.5.0", "41354"): "41354 LegCashSettlMinimumQuoteAmount",
        ("FIX.5.0", "41355"): "41355 LegCashSettlMinimumQuoteCurrency",
        ("FIX.5.0", "41356"): "41356 LegCashSettlBusinessDays",
        ("FIX.5.0", "41357"): "41357 LegCashSettlAmount",
        ("FIX.5.0", "41358"): "41358 LegCashSettlRecoveryFactor",
        ("FIX.5.0", "41359"): "41359 LegCashSettlFixedTermIndicator",
        ("FIX.5.0", "41360"): "41360 LegCashSettlAccruedInterestIndicator",
        ("FIX.5.0", "41361"): "41361 LegCashSettlValuationMethod",
        ("FIX.5.0", "41362"): "41362 LegCashSettlTermXID",
        ("FIX.5.0", "41363"): "41363 NoLegComplexEventAveragingObservations",
        ("FIX.5.0", "41364"): "41364 LegComplexEventAveragingObservationNumber",
        ("FIX.5.0", "41365"): "41365 LegComplexEventAveragingWeight",
        ("FIX.5.0", "41366"): "41366 NoLegComplexEventCreditEvents",
        ("FIX.5.0", "41367"): "41367 LegComplexEventCreditEventType",
        ("FIX.5.0", "41368"): "41368 LegComplexEventCreditEventValue",
        ("FIX.5.0", "41369"): "41369 LegComplexEventCreditEventCurrency",
        ("FIX.5.0", "41370"): "41370 LegComplexEventCreditEventPeriod",
        ("FIX.5.0", "41371"): "41371 LegComplexEventCreditEventUnit",
        ("FIX.5.0", "41372"): "41372 LegComplexEventCreditEventDayType",
        ("FIX.5.0", "41373"): "41373 LegComplexEventCreditEventRateSource",
        ("FIX.5.0", "41374"): "41374 NoLegComplexEventCreditEventQualifiers",
        ("FIX.5.0", "41375"): "41375 LegComplexEventCreditEventQualifier",
        ("FIX.5.0", "41376"): "41376 NoLegComplexEventPeriodDateTimes",
        ("FIX.5.0", "41377"): "41377 LegComplexEventPeriodDate",
        ("FIX.5.0", "41378"): "41378 LegComplexEventPeriodTime",
        ("FIX.5.0", "41379"): "41379 NoLegComplexEventPeriods",
        ("FIX.5.0", "41380"): "41380 LegComplexEventPeriodType",
        ("FIX.5.0", "41381"): "41381 LegComplexEventBusinessCenter",
        ("FIX.5.0", "41382"): "41382 NoLegComplexEventRateSources",
        ("FIX.5.0", "41383"): "41383 LegComplexEventRateSource",
        ("FIX.5.0", "41384"): "41384 LegComplexEventRateSourceType",
        ("FIX.5.0", "41385"): "41385 LegComplexEventReferencePage",
        ("FIX.5.0", "41386"): "41386 LegComplexEvenReferencePageHeading",
        ("FIX.5.0", "41387"): "41387 NoLegComplexEventDateBusinessCenters",
        ("FIX.5.0", "41388"): "41388 LegComplexEventDateBusinessCenter",
        ("FIX.5.0", "41389"): "41389 LegComplexEventDateUnadjusted",
        ("FIX.5.0", "41390"): "41390 LegComplexEventDateRelativeTo",
        ("FIX.5.0", "41391"): "41391 LegComplexEventDateOffsetPeriod",
        ("FIX.5.0", "41392"): "41392 LegComplexEventDateOffsetUnit",
        ("FIX.5.0", "41393"): "41393 LegComplexEventDateOffsetDayType",
        ("FIX.5.0", "41394"): "41394 LegComplexEventDateBusinessDayConvention",
        ("FIX.5.0", "41395"): "41395 LegComplexEventDateAdjusted",
        ("FIX.5.0", "41396"): "41396 LegComplexEventFixingTime",
        ("FIX.5.0", "41397"): "41397 LegComplexEventFixingTimeBusinessCenter",
        ("FIX.5.0", "41398"): "41398 NoLegComplexEventCreditEventSources",
        ("FIX.5.0", "41399"): "41399 LegComplexEventCreditEventSource",
        ("FIX.5.0", "41400"): "41400 NoLegComplexEventSchedules",
        ("FIX.5.0", "41401"): "41401 LegComplexEventScheduleStartDate",
        ("FIX.5.0", "41402"): "41402 LegComplexEventScheduleEndDate",
        ("FIX.5.0", "41403"): "41403 LegComplexEventScheduleFrequencyPeriod",
        ("FIX.5.0", "41404"): "41404 LegComplexEventScheduleFrequencyUnit",
        ("FIX.5.0", "41405"): "41405 LegComplexEventScheduleRollConvention",
        ("FIX.5.0", "41406"): "41406 ProvisionCashSettlQuoteReferencePage",
        ("FIX.5.0", "41407"): "41407 LegProvisionCashSettlQuoteReferencePage",
        ("FIX.5.0", "41408"): "41408 NoLegDeliverySchedules",
        ("FIX.5.0", "41409"): "41409 LegDeliveryScheduleType",
        ("FIX.5.0", "41410"): "41410 LegDeliveryScheduleXID",
        ("FIX.5.0", "41411"): "41411 LegDeliveryScheduleNotional",
        ("FIX.5.0", "41412"): "41412 LegDeliveryScheduleNotionalUnitOfMeasure",
        ("FIX.5.0", "41413"): "41413 LegDeliveryScheduleNotionalCommodityFrequency",
        ("FIX.5.0", "41414"): "41414 LegDeliveryScheduleNegativeTolerance",
        ("FIX.5.0", "41415"): "41415 LegDeliverySchedulePositiveTolerance",
        ("FIX.5.0", "41416"): "41416 LegDeliveryScheduleToleranceUnitOfMeasure",
        ("FIX.5.0", "41417"): "41417 LegDeliveryScheduleToleranceType",
        ("FIX.5.0", "41418"): "41418 LegDeliveryScheduleSettlCountry",
        ("FIX.5.0", "41419"): "41419 LegDeliveryScheduleSettlTimeZone",
        ("FIX.5.0", "41420"): "41420 LegDeliveryScheduleSettlFlowType",
        ("FIX.5.0", "41421"): "41421 LegDeliveryScheduleSettlHolidaysProcessingInstruction",
        ("FIX.5.0", "41422"): "41422 NoLegDeliveryScheduleSettlDays",
        ("FIX.5.0", "41423"): "41423 LegDeliveryScheduleSettlDay",
        ("FIX.5.0", "41424"): "41424 LegDeliveryScheduleSettlTotalHours",
        ("FIX.5.0", "41425"): "41425 NoLegDeliveryScheduleSettlTimes",
        ("FIX.5.0", "41426"): "41426 LegDeliveryScheduleSettlStart",
        ("FIX.5.0", "41427"): "41427 LegDeliveryScheduleSettlEnd",
        ("FIX.5.0", "41428"): "41428 LegDeliveryScheduleSettlTimeType",
        ("FIX.5.0", "41429"): "41429 LegDeliveryStreamType",
        ("FIX.5.0", "41430"): "41430 LegDeliveryStreamPipeline",
        ("FIX.5.0", "41431"): "41431 LegDeliveryStreamEntryPoint",
        ("FIX.5.0", "41432"): "41432 LegDeliveryStreamWithdrawalPoint",
        ("FIX.5.0", "41433"): "41433 LegDeliveryStreamDeliveryPoint",
        ("FIX.5.0", "41434"): "41434 LegDeliveryStreamDeliveryRestriction",
        ("FIX.5.0", "41435"): "41435 LegDeliveryStreamDeliveryContingency",
        ("FIX.5.0", "41436"): "41436 LegDeliveryStreamDeliveryContingentPartySide",
        ("FIX.5.0", "41437"): "41437 LegDeliveryStreamDeliverAtSourceIndicator",
        ("FIX.5.0", "41438"): "41438 LegDeliveryStreamRiskApportionment",
        ("FIX.5.0", "41439"): "41439 LegDeliveryStreamTitleTransferLocation",
        ("FIX.5.0", "41440"): "41440 LegDeliveryStreamTitleTransferCondition",
        ("FIX.5.0", "41441"): "41441 LegDeliveryStreamImporterOfRecord",
        ("FIX.5.0", "41442"): "41442 LegDeliveryStreamNegativeTolerance",
        ("FIX.5.0", "41443"): "41443 LegDeliveryStreamPositiveTolerance",
        ("FIX.5.0", "41444"): "41444 LegDeliveryStreamToleranceUnitOfMeasure",
        ("FIX.5.0", "41445"): "41445 LegDeliveryStreamToleranceType",
        ("FIX.5.0", "41446"): "41446 LegDeliveryStreamToleranceOptionSide",
        ("FIX.5.0", "41447"): "41447 LegDeliveryStreamTotalPositiveTolerance",
        ("FIX.5.0", "41448"): "41448 LegDeliveryStreamTotalNegativeTolerance",
        ("FIX.5.0", "41449"): "41449 LegDeliveryStreamNotionalConversionFactor",
        ("FIX.5.0", "41450"): "41450 LegDeliveryStreamTransportEquipment",
        ("FIX.5.0", "41451"): "41451 LegDeliveryStreamElectingPartySide",
        ("FIX.5.0", "41452"): "41452 NoLegStreamAssetAttributes",
        ("FIX.5.0", "41453"): "41453 LegStreamAssetAttributeType",
        ("FIX.5.0", "41454"): "41454 LegStreamAssetAttributeValue",
        ("FIX.5.0", "41455"): "41455 LegStreamAssetAttributeLimit",
        ("FIX.5.0", "41456"): "41456 NoLegDeliveryStreamCycles",
        ("FIX.5.0", "41457"): "41457 LegDeliveryStreamCycleDesc",
        ("FIX.5.0", "41458"): "41458 EncodedLegDeliveryStreamCycleDescLen",
        ("FIX.5.0", "41459"): "41459 EncodedLegDeliveryStreamCycleDesc",
        ("FIX.5.0", "41460"): "41460 NoLegDeliveryStreamCommoditySources",
        ("FIX.5.0", "41461"): "41461 LegDeliveryStreamCommoditySource",
        ("FIX.5.0", "41462"): "41462 LegMarketDisruptionProvision",
        ("FIX.5.0", "41463"): "41463 LegMarketDisruptionFallbackProvision",
        ("FIX.5.0", "41464"): "41464 LegMarketDisruptionMaximumDays",
        ("FIX.5.0", "41465"): "41465 LegMarketDisruptionMaterialityPercentage",
        ("FIX.5.0", "41466"): "41466 LegMarketDisruptionMinimumFuturesContracts",
        ("FIX.5.0", "41467"): "41467 NoLegMarketDisruptionEvents",
        ("FIX.5.0", "41468"): "41468 LegMarketDisruptionEvent",
        ("FIX.5.0", "41469"): "41469 NoLegMarketDisruptionFallbacks",
        ("FIX.5.0", "41470"): "41470 LegMarketDisruptionFallbackType",
        ("FIX.5.0", "41471"): "41471 NoLegMarketDisruptionFallbackReferencePrices",
        ("FIX.5.0", "41472"): "41472 LegMarketDisruptionFallbackUnderlierType",
        ("FIX.5.0", "41473"): "41473 LegMarketDisruptionFallbackUnderlierSecurityID",
        ("FIX.5.0", "41474"): "41474 LegMarketDisruptionFallbackUnderlierSecurityIDSource",
        ("FIX.5.0", "41475"): "41475 LegMarketDisruptionFallbackUnderlierSecurityDesc",
        ("FIX.5.0", "41476"): "41476 EncodedLegMarketDisruptionFallbackUnderlierSecurityDescLen",
        ("FIX.5.0", "41477"): "41477 EncodedLegMarketDisruptionFallbackUnderlierSecurityDesc",
        ("FIX.5.0", "41478"): "41478 LegMarketDisruptionFallbackOpenUnits",
        ("FIX.5.0", "41479"): "41479 LegMarketDisruptionFallbackBasketCurrency",
        ("FIX.5.0", "41480"): "41480 LegMarketDisruptionFallbackBasketDivisor",
        ("FIX.5.0", "41481"): "41481 LegExerciseDesc",
        ("FIX.5.0", "41482"): "41482 EncodedLegExerciseDescLen",
        ("FIX.5.0", "41483"): "41483 EncodedLegExerciseDesc",
        ("FIX.5.0", "41484"): "41484 LegAutomaticExerciseIndicator",
        ("FIX.5.0", "41485"): "41485 LegAutomaticExerciseThresholdRate",
        ("FIX.5.0", "41486"): "41486 LegExerciseConfirmationMethod",
        ("FIX.5.0", "41487"): "41487 LegManualNoticeBusinessCenter",
        ("FIX.5.0", "41488"): "41488 LegFallbackExerciseIndicator",
        ("FIX.5.0", "41489"): "41489 LegLimitRightToConfirmIndicator",
        ("FIX.5.0", "41490"): "41490 LegExerciseSplitTicketIndicator",
        ("FIX.5.0", "41491"): "41491 NoLegOptionExerciseBusinessCenters",
        ("FIX.5.0", "41492"): "41492 LegOptionExerciseBusinessCenter",
        ("FIX.5.0", "41493"): "41493 LegOptionExerciseBusinessDayConvention",
        ("FIX.5.0", "41494"): "41494 LegOptionExerciseEarliestDateOffsetDayType",
        ("FIX.5.0", "41495"): "41495 LegOptionExerciseEarliestDateOffsetPeriod",
        ("FIX.5.0", "41496"): "41496 LegOptionExerciseEarliestDateOffsetUnit",
        ("FIX.5.0", "41497"): "41497 LegOptionExerciseFrequencyPeriod",
        ("FIX.5.0", "41498"): "41498 LegOptionExerciseFrequencyUnit",
        ("FIX.5.0", "41499"): "41499 LegOptionExerciseStartDateUnadjusted",
        ("FIX.5.0", "41500"): "41500 LegOptionExerciseStartDateRelativeTo",
        ("FIX.5.0", "41501"): "41501 LegOptionExerciseStartDateOffsetPeriod",
        ("FIX.5.0", "41502"): "41502 LegOptionExerciseStartDateOffsetUnit",
        ("FIX.5.0", "41503"): "41503 LegOptionExerciseStartDateOffsetDayType",
        ("FIX.5.0", "41504"): "41504 LegOptionExerciseStartDateAdjusted",
        ("FIX.5.0", "41505"): "41505 LegOptionExerciseSkip",
        ("FIX.5.0", "41506"): "41506 LegOptionExerciseNominationDeadline",
        ("FIX.5.0", "41507"): "41507 LegOptionExerciseFirstDateUnadjusted",
        ("FIX.5.0", "41508"): "41508 LegOptionExerciseLastDateUnadjusted",
        ("FIX.5.0", "41509"): "41509 LegOptionExerciseEarliestTime",
        ("FIX.5.0", "41510"): "41510 LegOptionExerciseLatestTime",
        ("FIX.5.0", "41511"): "41511 LegOptionExerciseTimeBusinessCenter",
        ("FIX.5.0", "41512"): "41512 NoLegOptionExerciseDates",
        ("FIX.5.0", "41513"): "41513 LegOptionExerciseDate",
        ("FIX.5.0", "41514"): "41514 LegOptionExerciseDateType",
        ("FIX.5.0", "41515"): "41515 NoLegOptionExerciseExpirationDateBusinessCenters",
        ("FIX.5.0", "41516"): "41516 LegOptionExerciseExpirationDateBusinessCenter",
        ("FIX.5.0", "41517"): "41517 LegOptionExerciseExpirationDateBusinessDayConvention",
        ("FIX.5.0", "41518"): "41518 LegOptionExerciseExpirationDateRelativeTo",
        ("FIX.5.0", "41519"): "41519 LegOptionExerciseExpirationDateOffsetPeriod",
        ("FIX.5.0", "41520"): "41520 LegOptionExerciseExpirationDateOffsetUnit",
        ("FIX.5.0", "41521"): "41521 LegOptionExerciseExpirationFrequencyPeriod",
        ("FIX.5.0", "41522"): "41522 LegOptionExerciseExpirationFrequencyUnit",
        ("FIX.5.0", "41523"): "41523 LegOptionExerciseExpirationRollConvention",
        ("FIX.5.0", "41524"): "41524 LegOptionExerciseExpirationDateOffsetDayType",
        ("FIX.5.0", "41525"): "41525 LegOptionExerciseExpirationTime",
        ("FIX.5.0", "41526"): "41526 LegOptionExerciseExpirationTimeBusinessCenter",
        ("FIX.5.0", "41527"): "41527 NoLegOptionExerciseExpirationDates",
        ("FIX.5.0", "41528"): "41528 LegOptionExerciseExpirationDate",
        ("FIX.5.0", "41529"): "41529 LegOptionExerciseExpirationDateType",
        ("FIX.5.0", "41530"): "41530 NoLegPaymentScheduleFixingDays",
        ("FIX.5.0", "41531"): "41531 LegPaymentScheduleFixingDayOfWeek",
        ("FIX.5.0", "41532"): "41532 LegPaymentScheduleFixingDayNumber",
        ("FIX.5.0", "41533"): "41533 LegPaymentScheduleXID",
        ("FIX.5.0", "41534"): "41534 LegPaymentScheduleXIDRef",
        ("FIX.5.0", "41535"): "41535 LegPaymentScheduleRateCurrency",
        ("FIX.5.0", "41536"): "41536 LegPaymentScheduleRateUnitOfMeasure",
        ("FIX.5.0", "41537"): "41537 LegPaymentScheduleRateConversionFactor",
        ("FIX.5.0", "41538"): "41538 LegPaymentScheduleRateSpreadType",
        ("FIX.5.0", "41539"): "41539 LegPaymentScheduleSettlPeriodPrice",
        ("FIX.5.0", "41540"): "41540 LegPaymentScheduleSettlPeriodPriceCurrency",
        ("FIX.5.0", "41541"): "41541 LegPaymentScheduleSettlPeriodPriceUnitOfMeasure",
        ("FIX.5.0", "41542"): "41542 LegPaymentScheduleStepUnitOfMeasure",
        ("FIX.5.0", "41543"): "41543 LegPaymentScheduleFixingDayDistribution",
        ("FIX.5.0", "41544"): "41544 LegPaymentScheduleFixingDayCount",
        ("FIX.5.0", "41545"): "41545 LegPaymentScheduleFixingLagPeriod",
        ("FIX.5.0", "41546"): "41546 LegPaymentScheduleFixingLagUnit",
        ("FIX.5.0", "41547"): "41547 LegPaymentScheduleFixingFirstObservationDateOffsetPeriod",
        ("FIX.5.0", "41548"): "41548 LegPaymentScheduleFixingFirstObservationDateOffsetUnit",
        ("FIX.5.0", "41549"): "41549 LegPaymentStreamFlatRateIndicator",
        ("FIX.5.0", "41550"): "41550 LegPaymentStreamFlatRateAmount",
        ("FIX.5.0", "41551"): "41551 LegPaymentStreamFlatRateCurrency",
        ("FIX.5.0", "41552"): "41552 LegStreamMaximumPaymentAmount",
        ("FIX.5.0", "41553"): "41553 LegStreamMaximumPaymentCurrency",
        ("FIX.5.0", "41554"): "41554 LegStreamMaximumTransactionAmount",
        ("FIX.5.0", "41555"): "41555 LegStreamMaximumTransactionCurrency",
        ("FIX.5.0", "41556"): "41556 LegPaymentStreamFixedAmountUnitOfMeasure",
        ("FIX.5.0", "41557"): "41557 LegPaymentStreamTotalFixedAmount",
        ("FIX.5.0", "41558"): "41558 LegPaymentStreamWorldScaleRate",
        ("FIX.5.0", "41559"): "41559 LegPaymentStreamContractPrice",
        ("FIX.5.0", "41560"): "41560 LegPaymentStreamContractPriceCurrency",
        ("FIX.5.0", "41561"): "41561 NoLegPaymentStreamPricingBusinessCenters",
        ("FIX.5.0", "41562"): "41562 LegPaymentStreamPricingBusinessCenter",
        ("FIX.5.0", "41563"): "41563 LegPaymentStreamRateIndex2CurveUnit",
        ("FIX.5.0", "41564"): "41564 LegPaymentStreamRateIndex2CurvePeriod",
        ("FIX.5.0", "41565"): "41565 LegPaymentStreamRateIndexLocation",
        ("FIX.5.0", "41566"): "41566 LegPaymentStreamRateIndexLevel",
        ("FIX.5.0", "41567"): "41567 LegPaymentStreamRateIndexUnitOfMeasure",
        ("FIX.5.0", "41568"): "41568 LegPaymentStreamSettlLevel",
        ("FIX.5.0", "41569"): "41569 LegPaymentStreamReferenceLevel",
        ("FIX.5.0", "41570"): "41570 LegPaymentStreamReferenceLevelUnitOfMeasure",
        ("FIX.5.0", "41571"): "41571 LegPaymentStreamReferenceLevelEqualsZeroIndicator",
        ("FIX.5.0", "41572"): "41572 LegPaymentStreamRateSpreadCurrency",
        ("FIX.5.0", "41573"): "41573 LegPaymentStreamRateSpreadUnitOfMeasure",
        ("FIX.5.0", "41574"): "41574 LegPaymentStreamRateConversionFactor",
        ("FIX.5.0", "41575"): "41575 LegPaymentStreamRateSpreadType",
        ("FIX.5.0", "41576"): "41576 LegPaymentStreamLastResetRate",
        ("FIX.5.0", "41577"): "41577 LegPaymentStreamFinalRate",
        ("FIX.5.0", "41578"): "41578 LegPaymentStreamCalculationLagPeriod",
        ("FIX.5.0", "41579"): "41579 LegPaymentStreamCalculationLagUnit",
        ("FIX.5.0", "41580"): "41580 LegPaymentStreamFirstObservationDateOffsetPeriod",
        ("FIX.5.0", "41581"): "41581 LegPaymentStreamFirstObservationDateOffsetUnit",
        ("FIX.5.0", "41582"): "41582 LegPaymentStreamPricingDayType",
        ("FIX.5.0", "41583"): "41583 LegPaymentStreamPricingDayDistribution",
        ("FIX.5.0", "41584"): "41584 LegPaymentStreamPricingDayCount",
        ("FIX.5.0", "41585"): "41585 LegPaymentStreamPricingBusinessCalendar",
        ("FIX.5.0", "41586"): "41586 LegPaymentStreamPricingBusinessDayConvention",
        ("FIX.5.0", "41587"): "41587 UnderlyingDeliveryStreamRiskApportionmentSource",
        ("FIX.5.0", "41588"): "41588 StreamCommoditySettlTimeType",
        ("FIX.5.0", "41589"): "41589 NoLegPaymentStreamPaymentDates",
        ("FIX.5.0", "41590"): "41590 LegPaymentStreamPaymentDate",
        ("FIX.5.0", "41591"): "41591 LegPaymentStreamPaymentDateType",
        ("FIX.5.0", "41592"): "41592 LegPaymentStreamMasterAgreementPaymentDatesIndicator",
        ("FIX.5.0", "41593"): "41593 NoLegPaymentStreamPricingDates",
        ("FIX.5.0", "41594"): "41594 LegPaymentStreamPricingDate",
        ("FIX.5.0", "41595"): "41595 LegPaymentStreamPricingDateType",
        ("FIX.5.0", "41596"): "41596 NoLegPaymentStreamPricingDays",
        ("FIX.5.0", "41597"): "41597 LegPaymentStreamPricingDayOfWeek",
        ("FIX.5.0", "41598"): "41598 LegPaymentStreamPricingDayNumber",
        ("FIX.5.0", "41599"): "41599 NoLegPhysicalSettlTerms",
        ("FIX.5.0", "41600"): "41600 LegPhysicalSettlTermXID",
        ("FIX.5.0", "41601"): "41601 LegPhysicalSettlCurency",
        ("FIX.5.0", "41602"): "41602 LegPhysicalSettlBusinessDays",
        ("FIX.5.0", "41603"): "41603 LegPhysicalSettlMaximumBusinessDays",
        ("FIX.5.0", "41604"): "41604 NoLegPhysicalSettlDeliverableObligations",
        ("FIX.5.0", "41605"): "41605 LegPhysicalSettlDeliverableObligationType",
        ("FIX.5.0", "41606"): "41606 LegPhysicalSettlDeliverableObligationValue",
        ("FIX.5.0", "41607"): "41607 NoLegPricingDateBusinessCenters",
        ("FIX.5.0", "41608"): "41608 LegPricingDateBusinessCenter",
        ("FIX.5.0", "41609"): "41609 LegPricingDateUnadjusted",
        ("FIX.5.0", "41610"): "41610 LegPricingDateBusinessDayConvention",
        ("FIX.5.0", "41611"): "41611 LegPricingDateAdjusted",
        ("FIX.5.0", "41612"): "41612 LegPricingTime",
        ("FIX.5.0", "41613"): "41613 LegPricingTimeBusinessCenter",
        ("FIX.5.0", "41614"): "41614 NoLegProtectionTermEventNewsSources",
        ("FIX.5.0", "41615"): "41615 LegProtectionTermEventNewsSource",
        ("FIX.5.0", "41616"): "41616 NoLegProtectionTerms",
        ("FIX.5.0", "41617"): "41617 LegProtectionTermXID",
        ("FIX.5.0", "41618"): "41618 LegProtectionTermNotional",
        ("FIX.5.0", "41619"): "41619 LegProtectionTermCurrency",
        ("FIX.5.0", "41620"): "41620 LegProtectionTermSellerNotifies",
        ("FIX.5.0", "41621"): "41621 LegProtectionTermBuyerNotifies",
        ("FIX.5.0", "41622"): "41622 LegProtectionTermEventBusinessCenter",
        ("FIX.5.0", "41623"): "41623 LegProtectionTermStandardSources",
        ("FIX.5.0", "41624"): "41624 LegProtectionTermEventMinimumSources",
        ("FIX.5.0", "41625"): "41625 NoLegProtectionTermEvents",
        ("FIX.5.0", "41626"): "41626 LegProtectionTermEventType",
        ("FIX.5.0", "41627"): "41627 LegProtectionTermEventValue",
        ("FIX.5.0", "41628"): "41628 LegProtectionTermEventCurrency",
        ("FIX.5.0", "41629"): "41629 LegProtectionTermEventPeriod",
        ("FIX.5.0", "41630"): "41630 LegProtectionTermEventUnit",
        ("FIX.5.0", "41631"): "41631 LegProtectionTermEventDayType",
        ("FIX.5.0", "41632"): "41632 LegProtectionTermEventRateSource",
        ("FIX.5.0", "41633"): "41633 NoLegProtectionTermEventQualifiers",
        ("FIX.5.0", "41634"): "41634 LegProtectionTermEventQualifier",
        ("FIX.5.0", "41635"): "41635 NoLegProtectionTermObligations",
        ("FIX.5.0", "41636"): "41636 LegProtectionTermObligationType",
        ("FIX.5.0", "41637"): "41637 LegProtectionTermObligationValue",
        ("FIX.5.0", "41638"): "41638 NoLegStreamCalculationPeriodDates",
        ("FIX.5.0", "41639"): "41639 LegStreamCalculationPeriodDate",
        ("FIX.5.0", "41640"): "41640 LegStreamCalculationPeriodDateType",
        ("FIX.5.0", "41641"): "41641 LegStreamCalculationPeriodDatesXID",
        ("FIX.5.0", "41642"): "41642 LegStreamCalculationPeriodDatesXIDRef",
        ("FIX.5.0", "41643"): "41643 LegStreamCalculationBalanceOfFirstPeriod",
        ("FIX.5.0", "41644"): "41644 LegStreamCalculationCorrectionPeriod",
        ("FIX.5.0", "41645"): "41645 LegStreamCalculationCorrectionUnit",
        ("FIX.5.0", "41646"): "41646 NoLegStreamCommoditySettlBusinessCenters",
        ("FIX.5.0", "41647"): "41647 LegStreamCommoditySettlBusinessCenter",
        ("FIX.5.0", "41648"): "41648 LegStreamCommodityBase",
        ("FIX.5.0", "41649"): "41649 LegStreamCommodityType",
        ("FIX.5.0", "41650"): "41650 LegStreamCommoditySecurityID",
        ("FIX.5.0", "41651"): "41651 LegStreamCommoditySecurityIDSource",
        ("FIX.5.0", "41652"): "41652 LegStreamCommodityDesc",
        ("FIX.5.0", "41653"): "41653 EncodedLegStreamCommodityDescLen",
        ("FIX.5.0", "41654"): "41654 EncodedLegStreamCommodityDesc",
        ("FIX.5.0", "41655"): "41655 LegStreamCommodityUnitOfMeasure",
        ("FIX.5.0", "41656"): "41656 LegStreamCommodityCurrency",
        ("FIX.5.0", "41657"): "41657 LegStreamCommodityExchange",
        ("FIX.5.0", "41658"): "41658 LegStreamCommodityRateSource",
        ("FIX.5.0", "41659"): "41659 LegStreamCommodityRateReferencePage",
        ("FIX.5.0", "41660"): "41660 LegStreamCommodityRateReferencePageHeading",
        ("FIX.5.0", "41661"): "41661 LegStreamDataProvider",
        ("FIX.5.0", "41662"): "41662 LegStreamCommodityPricingType",
        ("FIX.5.0", "41663"): "41663 LegStreamCommodityNearbySettlDayPeriod",
        ("FIX.5.0", "41664"): "41664 LegStreamCommodityNearbySettlDayUnit",
        ("FIX.5.0", "41665"): "41665 LegStreamCommoditySettlDateUnadjusted",
        ("FIX.5.0", "41666"): "41666 LegStreamCommoditySettlDateBusinessDayConvention",
        ("FIX.5.0", "41667"): "41667 LegStreamCommoditySettlDateAdjusted",
        ("FIX.5.0", "41668"): "41668 LegStreamCommoditySettlMonth",
        ("FIX.5.0", "41669"): "41669 LegStreamCommoditySettlDateRollPeriod",
        ("FIX.5.0", "41670"): "41670 LegStreamCommoditySettlDateRollUnit",
        ("FIX.5.0", "41671"): "41671 LegStreamCommoditySettlDayType",
        ("FIX.5.0", "41672"): "41672 LegStreamCommodityXID",
        ("FIX.5.0", "41673"): "41673 LegStreamCommodityXIDRef",
        ("FIX.5.0", "41674"): "41674 NoLegStreamCommodityAltIDs",
        ("FIX.5.0", "41675"): "41675 LegStreamCommodityAltID",
        ("FIX.5.0", "41676"): "41676 LegStreamCommodityAltIDSource",
        ("FIX.5.0", "41677"): "41677 NoLegStreamCommodityDataSources",
        ("FIX.5.0", "41678"): "41678 LegStreamCommodityDataSourceID",
        ("FIX.5.0", "41679"): "41679 LegStreamCommodityDataSourceIDType",
        ("FIX.5.0", "41680"): "41680 NoLegStreamCommoditySettlDays",
        ("FIX.5.0", "41681"): "41681 LegStreamCommoditySettlDay",
        ("FIX.5.0", "41682"): "41682 LegStreamCommoditySettlTotalHours",
        ("FIX.5.0", "41683"): "41683 NoLegStreamCommoditySettlTimes",
        ("FIX.5.0", "41684"): "41684 LegStreamCommoditySettlStart",
        ("FIX.5.0", "41685"): "41685 LegStreamCommoditySettlEnd",
        ("FIX.5.0", "41686"): "41686 NoLegStreamCommoditySettlPeriods",
        ("FIX.5.0", "41687"): "41687 LegStreamCommoditySettlCountry",
        ("FIX.5.0", "41688"): "41688 LegStreamCommoditySettlTimeZone",
        ("FIX.5.0", "41689"): "41689 LegStreamCommoditySettlFlowType",
        ("FIX.5.0", "41690"): "41690 LegStreamCommoditySettlPeriodNotional",
        ("FIX.5.0", "41691"): "41691 LegStreamCommoditySettlPeriodNotionalUnitOfMeasure",
        ("FIX.5.0", "41692"): "41692 LegStreamCommoditySettlPeriodFrequencyPeriod",
        ("FIX.5.0", "41693"): "41693 LegStreamCommoditySettlPeriodFrequencyUnit",
        ("FIX.5.0", "41694"): "41694 LegStreamCommoditySettlPeriodPrice",
        ("FIX.5.0", "41695"): "41695 LegStreamCommoditySettlPeriodPriceUnitOfMeasure",
        ("FIX.5.0", "41696"): "41696 LegStreamCommoditySettlPeriodPriceCurrency",
        ("FIX.5.0", "41697"): "41697 LegStreamCommoditySettlHolidaysProcessingInstruction",
        ("FIX.5.0", "41698"): "41698 LegStreamCommoditySettlPeriodXID",
        ("FIX.5.0", "41699"): "41699 LegStreamCommoditySettlPeriodXIDRef",
        ("FIX.5.0", "41700"): "41700 LegStreamXID",
        ("FIX.5.0", "41701"): "41701 UnderlyingAdditionalTermBondSecurityIDSource",
        ("FIX.5.0", "41702"): "41702 LegStreamNotionalXIDRef",
        ("FIX.5.0", "41703"): "41703 LegStreamNotionalFrequencyPeriod",
        ("FIX.5.0", "41704"): "41704 LegStreamNotionalFrequencyUnit",
        ("FIX.5.0", "41705"): "41705 LegStreamNotionalCommodityFrequency",
        ("FIX.5.0", "41706"): "41706 LegStreamNotionalUnitOfMeasure",
        ("FIX.5.0", "41707"): "41707 LegStreamTotalNotional",
        ("FIX.5.0", "41708"): "41708 LegStreamTotalNotionalUnitOfMeasure",
        ("FIX.5.0", "41709"): "41709 UnderlyingAdditionalTermBondDesc",
        ("FIX.5.0", "41710"): "41710 EncodedUnderlyingAdditionalTermBondDescLen",
        ("FIX.5.0", "41711"): "41711 EncodedUnderlyingAdditionalTermBondDesc",
        ("FIX.5.0", "41712"): "41712 UnderlyingAdditionalTermBondCurrency",
        ("FIX.5.0", "41713"): "41713 NoUnderlyingComplexEventAveragingObservations",
        ("FIX.5.0", "41714"): "41714 UnderlyingComplexEventAveragingObservationNumber",
        ("FIX.5.0", "41715"): "41715 UnderlyingComplexEventAveragingWeight",
        ("FIX.5.0", "41716"): "41716 NoUnderlyingComplexEventCreditEvents",
        ("FIX.5.0", "41717"): "41717 UnderlyingComplexEventCreditEventType",
        ("FIX.5.0", "41718"): "41718 UnderlyingComplexEventCreditEventValue",
        ("FIX.5.0", "41719"): "41719 UnderlyingComplexEventCreditEventCurrency",
        ("FIX.5.0", "41720"): "41720 UnderlyingComplexEventCreditEventPeriod",
        ("FIX.5.0", "41721"): "41721 UnderlyingComplexEventCreditEventUnit",
        ("FIX.5.0", "41722"): "41722 UnderlyingComplexEventCreditEventDayType",
        ("FIX.5.0", "41723"): "41723 UnderlyingComplexEventCreditEventRateSource",
        ("FIX.5.0", "41724"): "41724 NoUnderlyingComplexEventCreditEventQualifiers",
        ("FIX.5.0", "41725"): "41725 UnderlyingComplexEventCreditEventQualifier",
        ("FIX.5.0", "41726"): "41726 NoUnderlyingComplexEventPeriodDateTimes",
        ("FIX.5.0", "41727"): "41727 UnderlyingComplexEventPeriodDate",
        ("FIX.5.0", "41728"): "41728 UnderlyingComplexEventPeriodTime",
        ("FIX.5.0", "41729"): "41729 NoUnderlyingComplexEventPeriods",
        ("FIX.5.0", "41730"): "41730 UnderlyingComplexEventPeriodType",
        ("FIX.5.0", "41731"): "41731 UnderlyingComplexEventBusinessCenter",
        ("FIX.5.0", "41732"): "41732 NoUnderlyingComplexEventRateSources",
        ("FIX.5.0", "41733"): "41733 UnderlyingComplexEventRateSource",
        ("FIX.5.0", "41734"): "41734 UnderlyingComplexEventRateSourceType",
        ("FIX.5.0", "41735"): "41735 UnderlyingComplexEventReferencePage",
        ("FIX.5.0", "41736"): "41736 UnderlyingComplexEventReferencePageHeading",
        ("FIX.5.0", "41737"): "41737 NoUnderlyingComplexEventDateBusinessCenters",
        ("FIX.5.0", "41738"): "41738 UnderlyingComplexEventDateBusinessCenter",
        ("FIX.5.0", "41739"): "41739 UnderlyingComplexEventDateUnadjusted",
        ("FIX.5.0", "41740"): "41740 UnderlyingComplexEventDateRelativeTo",
        ("FIX.5.0", "41741"): "41741 UnderlyingComplexEventDateOffsetPeriod",
        ("FIX.5.0", "41742"): "41742 UnderlyingComplexEventDateOffsetUnit",
        ("FIX.5.0", "41743"): "41743 UnderlyingComplexEventDateOffsetDayType",
        ("FIX.5.0", "41744"): "41744 UnderlyingComplexEventDateBusinessDayConvention",
        ("FIX.5.0", "41745"): "41745 UnderlyingComplexEventDateAdjusted",
        ("FIX.5.0", "41746"): "41746 UnderlyingComplexEventFixingTime",
        ("FIX.5.0", "41747"): "41747 UnderlyingComplexEventFixingTimeBusinessCenter",
        ("FIX.5.0", "41748"): "41748 NoUnderlyingComplexEventCreditEventSources",
        ("FIX.5.0", "41749"): "41749 UnderlyingComplexEventCreditEventSource",
        ("FIX.5.0", "41750"): "41750 NoUnderlyingComplexEventSchedules",
        ("FIX.5.0", "41751"): "41751 UnderlyingComplexEventScheduleStartDate",
        ("FIX.5.0", "41752"): "41752 UnderlyingComplexEventScheduleEndDate",
        ("FIX.5.0", "41753"): "41753 UnderlyingComplexEventScheduleFrequencyPeriod",
        ("FIX.5.0", "41754"): "41754 UnderlyingComplexEventScheduleFrequencyUnit",
        ("FIX.5.0", "41755"): "41755 UnderlyingComplexEventScheduleRollConvention",
        ("FIX.5.0", "41756"): "41756 NoUnderlyingDeliverySchedules",
        ("FIX.5.0", "41757"): "41757 UnderlyingDeliveryScheduleType",
        ("FIX.5.0", "41758"): "41758 UnderlyingDeliveryScheduleXID",
        ("FIX.5.0", "41759"): "41759 UnderlyingDeliveryScheduleNotional",
        ("FIX.5.0", "41760"): "41760 UnderlyingDeliveryScheduleNotionalUnitOfMeasure",
        ("FIX.5.0", "41761"): "41761 UnderlyingDeliveryScheduleNotionalCommodityFrequency",
        ("FIX.5.0", "41762"): "41762 UnderlyingDeliveryScheduleNegativeTolerance",
        ("FIX.5.0", "41763"): "41763 UnderlyingDeliverySchedulePositiveTolerance",
        ("FIX.5.0", "41764"): "41764 UnderlyingDeliveryScheduleToleranceUnitOfMeasure",
        ("FIX.5.0", "41765"): "41765 UnderlyingDeliveryScheduleToleranceType",
        ("FIX.5.0", "41766"): "41766 UnderlyingDeliveryScheduleSettlCountry",
        ("FIX.5.0", "41767"): "41767 UnderlyingDeliveryScheduleSettlTimeZone",
        ("FIX.5.0", "41768"): "41768 UnderlyingDeliveryScheduleSettlFlowType",
        ("FIX.5.0", "41769"): "41769 UnderlyingDeliveryScheduleSettlHolidaysProcessingInstruction",
        ("FIX.5.0", "41770"): "41770 NoUnderlyingDeliveryScheduleSettlDays",
        ("FIX.5.0", "41771"): "41771 UnderlyingDeliveryScheduleSettlDay",
        ("FIX.5.0", "41772"): "41772 UnderlyingDeliveryScheduleSettlTotalHours",
        ("FIX.5.0", "41773"): "41773 NoUnderlyingDeliveryScheduleSettlTimes",
        ("FIX.5.0", "41774"): "41774 UnderlyingDeliveryScheduleSettlStart",
        ("FIX.5.0", "41775"): "41775 UnderlyingDeliveryScheduleSettlEnd",
        ("FIX.5.0", "41776"): "41776 UnderlyingDeliveryScheduleSettlTimeType",
        ("FIX.5.0", "41777"): "41777 UnderlyingDeliveryStreamType",
        ("FIX.5.0", "41778"): "41778 UnderlyingDeliveryStreamPipeline",
        ("FIX.5.0", "41779"): "41779 UnderlyingDeliveryStreamEntryPoint",
        ("FIX.5.0", "41780"): "41780 UnderlyingDeliveryStreamWithdrawalPoint",
        ("FIX.5.0", "41781"): "41781 UnderlyingDeliveryStreamDeliveryPoint",
        ("FIX.5.0", "41782"): "41782 UnderlyingDeliveryStreamDeliveryRestriction",
        ("FIX.5.0", "41783"): "41783 UnderlyingDeliveryStreamDeliveryContingency",
        ("FIX.5.0", "41784"): "41784 UnderlyingDeliveryStreamDeliveryContingentPartySide",
        ("FIX.5.0", "41785"): "41785 UnderlyingDeliveryStreamDeliverAtSourceIndicator",
        ("FIX.5.0", "41786"): "41786 UnderlyingDeliveryStreamRiskApportionment",
        ("FIX.5.0", "41787"): "41787 UnderlyingDeliveryStreamTitleTransferLocation",
        ("FIX.5.0", "41788"): "41788 UnderlyingDeliveryStreamTitleTransferCondition",
        ("FIX.5.0", "41789"): "41789 UnderlyingDeliveryStreamImporterOfRecord",
        ("FIX.5.0", "41790"): "41790 UnderlyingDeliveryStreamNegativeTolerance",
        ("FIX.5.0", "41791"): "41791 UnderlyingDeliveryStreamPositiveTolerance",
        ("FIX.5.0", "41792"): "41792 UnderlyingDeliveryStreamToleranceUnitOfMeasure",
        ("FIX.5.0", "41793"): "41793 UnderlyingDeliveryStreamToleranceType",
        ("FIX.5.0", "41794"): "41794 UnderlyingDeliveryStreamToleranceOptionSide",
        ("FIX.5.0", "41795"): "41795 UnderlyingDeliveryStreamTotalPositiveTolerance",
        ("FIX.5.0", "41796"): "41796 UnderlyingDeliveryStreamTotalNegativeTolerance",
        ("FIX.5.0", "41797"): "41797 UnderlyingDeliveryStreamNotionalConversionFactor",
        ("FIX.5.0", "41798"): "41798 UnderlyingDeliveryStreamTransportEquipment",
        ("FIX.5.0", "41799"): "41799 UnderlyingDeliveryStreamElectingPartySide",
        ("FIX.5.0", "41800"): "41800 NoUnderlyingStreamAssetAttributes",
        ("FIX.5.0", "41801"): "41801 UnderlyingStreamAssetAttributeType",
        ("FIX.5.0", "41802"): "41802 UnderlyingStreamAssetAttributeValue",
        ("FIX.5.0", "41803"): "41803 UnderlyingStreamAssetAttributeLimit",
        ("FIX.5.0", "41804"): "41804 NoUnderlyingDeliveryStreamCycles",
        ("FIX.5.0", "41805"): "41805 UnderlyingDeliveryStreamCycleDesc",
        ("FIX.5.0", "41806"): "41806 EncodedUnderlyingDeliveryStreamCycleDescLen",
        ("FIX.5.0", "41807"): "41807 EncodedUnderlyingDeliveryStreamCycleDesc",
        ("FIX.5.0", "41808"): "41808 NoUnderlyingDeliveryStreamCommoditySources",
        ("FIX.5.0", "41809"): "41809 UnderlyingDeliveryStreamCommoditySource",
        ("FIX.5.0", "41810"): "41810 UnderlyingExerciseDesc",
        ("FIX.5.0", "41811"): "41811 EncodedUnderlyingExerciseDescLen",
        ("FIX.5.0", "41812"): "41812 EncodedUnderlyingExerciseDesc",
        ("FIX.5.0", "41813"): "41813 UnderlyingAutomaticExerciseIndicator",
        ("FIX.5.0", "41814"): "41814 UnderlyingAutomaticExerciseThresholdRate",
        ("FIX.5.0", "41815"): "41815 UnderlyingExerciseConfirmationMethod",
        ("FIX.5.0", "41816"): "41816 UnderlyingManualNoticeBusinessCenter",
        ("FIX.5.0", "41817"): "41817 UnderlyingFallbackExerciseIndicator",
        ("FIX.5.0", "41818"): "41818 UnderlyingLimitedRightToConfirmIndicator",
        ("FIX.5.0", "41819"): "41819 UnderlyingExerciseSplitTicketIndicator",
        ("FIX.5.0", "41820"): "41820 NoUnderlyingOptionExerciseBusinessCenters",
        ("FIX.5.0", "41821"): "41821 UnderlyingOptionExerciseBusinessCenter",
        ("FIX.5.0", "41822"): "41822 UnderlyingOptionExerciseBusinessDayConvention",
        ("FIX.5.0", "41823"): "41823 UnderlyingOptionExerciseEarliestDateOffsetDayType",
        ("FIX.5.0", "41824"): "41824 UnderlyingOptionExerciseEarliestDateOffsetPeriod",
        ("FIX.5.0", "41825"): "41825 UnderlyingOptionExerciseEarliestDateOffsetUnit",
        ("FIX.5.0", "41826"): "41826 UnderlyingOptionExerciseFrequencyPeriod",
        ("FIX.5.0", "41827"): "41827 UnderlyingOptionExerciseFrequencyUnit",
        ("FIX.5.0", "41828"): "41828 UnderlyingOptionExerciseStartDateUnadjusted",
        ("FIX.5.0", "41829"): "41829 UnderlyingOptionExerciseStartDateRelativeTo",
        ("FIX.5.0", "41830"): "41830 UnderlyingOptionExerciseStartDateOffsetPeriod",
        ("FIX.5.0", "41831"): "41831 UnderlyingOptionExerciseStartDateOffsetUnit",
        ("FIX.5.0", "41832"): "41832 UnderlyingOptionExerciseStartDateOffsetDayType",
        ("FIX.5.0", "41833"): "41833 UnderlyingOptionExerciseStartDateAdjusted",
        ("FIX.5.0", "41834"): "41834 UnderlyingOptionExerciseSkip",
        ("FIX.5.0", "41835"): "41835 UnderlyingOptionExerciseNominationDeadline",
        ("FIX.5.0", "41836"): "41836 UnderlyingOptionExerciseFirstDateUnadjusted",
        ("FIX.5.0", "41837"): "41837 UnderlyingOptionExerciseLastDateUnadjusted",
        ("FIX.5.0", "41838"): "41838 UnderlyingOptionExerciseEarliestTime",
        ("FIX.5.0", "41839"): "41839 UnderlyingOptionExerciseLatestTime",
        ("FIX.5.0", "41840"): "41840 UnderlyingOptionExerciseTimeBusinessCenter",
        ("FIX.5.0", "41841"): "41841 NoUnderlyingOptionExerciseDates",
        ("FIX.5.0", "41842"): "41842 UnderlyingOptionExerciseDate",
        ("FIX.5.0", "41843"): "41843 UnderlyingOptionExerciseDateType",
        ("FIX.5.0", "41844"): "41844 NoUnderlyingOptionExerciseExpirationDateBusinessCenters",
        ("FIX.5.0", "41845"): "41845 UnderlyingOptionExerciseExpirationDateBusinessCenter",
        ("FIX.5.0", "41846"): "41846 UnderlyingOptionExerciseExpirationDateBusinessDayConvention",
        ("FIX.5.0", "41847"): "41847 UnderlyingOptionExerciseExpirationDateRelativeTo",
        ("FIX.5.0", "41848"): "41848 UnderlyingOptionExerciseExpirationDateOffsetPeriod",
        ("FIX.5.0", "41849"): "41849 UnderlyingOptionExerciseExpirationDateOffsetUnit",
        ("FIX.5.0", "41850"): "41850 UnderlyingOptionExerciseExpirationFrequencyPeriod",
        ("FIX.5.0", "41851"): "41851 UnderlyingOptionExerciseExpirationFrequencyUnit",
        ("FIX.5.0", "41852"): "41852 UnderlyingOptionExerciseExpirationRollConvention",
        ("FIX.5.0", "41853"): "41853 UnderlyingOptionExerciseExpirationDateOffsetDayType",
        ("FIX.5.0", "41854"): "41854 UnderlyingOptionExerciseExpirationTime",
        ("FIX.5.0", "41855"): "41855 UnderlyingOptionExerciseExpirationTimeBusinessCenter",
        ("FIX.5.0", "41856"): "41856 NoUnderlyingOptionExerciseExpirationDates",
        ("FIX.5.0", "41857"): "41857 UnderlyingOptionExerciseExpirationDate",
        ("FIX.5.0", "41858"): "41858 UnderlyingOptionExerciseExpirationDateType",
        ("FIX.5.0", "41859"): "41859 UnderlyingMarketDisruptionProvision",
        ("FIX.5.0", "41860"): "41860 UnderlyingMarketDisruptionFallbackProvision",
        ("FIX.5.0", "41861"): "41861 UnderlyingMarketDisruptionMaximumDays",
        ("FIX.5.0", "41862"): "41862 UnderlyingMarketDisruptionMaterialityPercentage",
        ("FIX.5.0", "41863"): "41863 UnderlyingMarketDisruptionMinimumFuturesContracts",
        ("FIX.5.0", "41864"): "41864 NoUnderlyingMarketDisruptionEvents",
        ("FIX.5.0", "41865"): "41865 UnderlyingMarketDisruptionEvent",
        ("FIX.5.0", "41866"): "41866 NoUnderlyingMarketDisruptionFallbacks",
        ("FIX.5.0", "41867"): "41867 UnderlyingMarketDisruptionFallbackType",
        ("FIX.5.0", "41868"): "41868 NoUnderlyingMarketDisruptionFallbackReferencePrices",
        ("FIX.5.0", "41869"): "41869 UnderlyingMarketDisruptionFallbackUnderlierType",
        ("FIX.5.0", "41870"): "41870 UnderlyingMarketDisruptionFallbackUnderlierSecurityID",
        ("FIX.5.0", "41871"): "41871 UnderlyingMarketDisruptionFallbackUnderlierSecurityIDSource",
        ("FIX.5.0", "41872"): "41872 UnderlyingMarketDisruptionFallbackUnderlierSecurityDesc",
        ("FIX.5.0", "41873"): "41873 EncodedUnderlyingMarketDisruptionFallbackUnderlierSecurityDescLen",
        ("FIX.5.0", "41874"): "41874 EncodedUnderlyingMarketDisruptionFallbackUnderlierSecurityDesc",
        ("FIX.5.0", "41875"): "41875 UnderlyingMarketDisruptionFallbackOpenUnits",
        ("FIX.5.0", "41876"): "41876 UnderlyingMarketDisruptionFallbackBasketCurrency",
        ("FIX.5.0", "41877"): "41877 UnderlyingMarketDisruptionFallbackBasketDivisor",
        ("FIX.5.0", "41878"): "41878 NoUnderlyingPaymentScheduleFixingDays",
        ("FIX.5.0", "41879"): "41879 UnderlyingPaymentScheduleFixingDayOfWeek",
        ("FIX.5.0", "41880"): "41880 UnderlyingPaymentScheduleFixingDayNumber",
        ("FIX.5.0", "41881"): "41881 UnderlyingPaymentScheduleXID",
        ("FIX.5.0", "41882"): "41882 UnderlyingPaymentScheduleXIDRef",
        ("FIX.5.0", "41883"): "41883 UnderlyingPaymentScheduleRateCurrency",
        ("FIX.5.0", "41884"): "41884 UnderlyingPaymentScheduleRateUnitOfMeasure",
        ("FIX.5.0", "41885"): "41885 UnderlyingPaymentScheduleRateConversionFactor",
        ("FIX.5.0", "41886"): "41886 UnderlyingPaymentScheduleRateSpreadType",
        ("FIX.5.0", "41887"): "41887 UnderlyingPaymentScheduleSettlPeriodPrice",
        ("FIX.5.0", "41888"): "41888 UnderlyingPaymentScheduleSettlPeriodPriceCurrency",
        ("FIX.5.0", "41889"): "41889 UnderlyingPaymentScheduleSettlPeriodPriceUnitOfMeasure",
        ("FIX.5.0", "41890"): "41890 UnderlyingPaymentScheduleStepUnitOfMeasure",
        ("FIX.5.0", "41891"): "41891 UnderlyingPaymentScheduleFixingDayDistribution",
        ("FIX.5.0", "41892"): "41892 UnderlyingPaymentScheduleFixingDayCount",
        ("FIX.5.0", "41893"): "41893 UnderlyingPaymentScheduleFixingLagPeriod",
        ("FIX.5.0", "41894"): "41894 UnderlyingPaymentScheduleFixingLagUnit",
        ("FIX.5.0", "41895"): "41895 UnderlyingPaymentScheduleFixingFirstObservationDateOffsetPeriod",
        ("FIX.5.0", "41896"): "41896 UnderlyingPaymentScheduleFixingFirstObservationDateOffsetUnit",
        ("FIX.5.0", "41897"): "41897 UnderlyingPaymentStreamFlatRateIndicator",
        ("FIX.5.0", "41898"): "41898 UnderlyingPaymentStreamFlatRateAmount",
        ("FIX.5.0", "41899"): "41899 UnderlyingPaymentStreamFlatRateCurrency",
        ("FIX.5.0", "41900"): "41900 UnderlyingPaymentStreamMaximumPaymentAmount",
        ("FIX.5.0", "41901"): "41901 UnderlyingPaymentStreamMaximumPaymentCurrency",
        ("FIX.5.0", "41902"): "41902 UnderlyingPaymentStreamMaximumTransactionAmount",
        ("FIX.5.0", "41903"): "41903 UnderlyingPaymentStreamMaximumTransactionCurrency",
        ("FIX.5.0", "41904"): "41904 UnderlyingPaymentStreamFixedAmountUnitOfMeasure",
        ("FIX.5.0", "41905"): "41905 UnderlyingPaymentStreamTotalFixedAmount",
        ("FIX.5.0", "41906"): "41906 UnderlyingPaymentStreamWorldScaleRate",
        ("FIX.5.0", "41907"): "41907 UnderlyingPaymentStreamContractPrice",
        ("FIX.5.0", "41908"): "41908 UnderlyingPaymentStreamContractPriceCurrency",
        ("FIX.5.0", "41909"): "41909 NoUnderlyingPaymentStreamPricingBusinessCenters",
        ("FIX.5.0", "41910"): "41910 UnderlyingPaymentStreamPricingBusinessCenter",
        ("FIX.5.0", "41911"): "41911 UnderlyingPaymentStreamRateIndex2CurveUnit",
        ("FIX.5.0", "41912"): "41912 UnderlyingPaymentStreamRateIndex2CurvePeriod",
        ("FIX.5.0", "41913"): "41913 UnderlyingPaymentStreamRateIndexLocation",
        ("FIX.5.0", "41914"): "41914 UnderlyingPaymentStreamRateIndexLevel",
        ("FIX.5.0", "41915"): "41915 UnderlyingPaymentStreamRateIndexUnitOfMeasure",
        ("FIX.5.0", "41916"): "41916 UnderlyingPaymentStreamSettlLevel",
        ("FIX.5.0", "41917"): "41917 UnderlyingPaymentStreamReferenceLevel",
        ("FIX.5.0", "41918"): "41918 UnderlyingPaymentStreamReferenceLevelUnitOfMeasure",
        ("FIX.5.0", "41919"): "41919 UnderlyingPaymentStreamReferenceLevelEqualsZeroIndicator",
        ("FIX.5.0", "41920"): "41920 UnderlyingPaymentStreamRateSpreadCurrency",
        ("FIX.5.0", "41921"): "41921 UnderlyingPaymentStreamRateSpreadUnitOfMeasure",
        ("FIX.5.0", "41922"): "41922 UnderlyingPaymentStreamRateConversionFactor",
        ("FIX.5.0", "41923"): "41923 UnderlyingPaymentStreamRateSpreadType",
        ("FIX.5.0", "41924"): "41924 UnderlyingPaymentStreamLastResetRate",
        ("FIX.5.0", "41925"): "41925 UnderlyingPaymentStreamFinalRate",
        ("FIX.5.0", "41926"): "41926 UnderlyingPaymentStreamCalculationLagPeriod",
        ("FIX.5.0", "41927"): "41927 UnderlyingPaymentStreamCalculationLagUnit",
        ("FIX.5.0", "41928"): "41928 UnderlyingPaymentStreamFirstObservationDateOffsetPeriod",
        ("FIX.5.0", "41929"): "41929 UnderlyingPaymentStreamFirstObservationDateOffsetUnit",
        ("FIX.5.0", "41930"): "41930 UnderlyingPaymentStreamPricingDayType",
        ("FIX.5.0", "41931"): "41931 UnderlyingPaymentStreamPricingDayDistribution",
        ("FIX.5.0", "41932"): "41932 UnderlyingPaymentStreamPricingDayCount",
        ("FIX.5.0", "41933"): "41933 UnderlyingPaymentStreamPricingBusinessCalendar",
        ("FIX.5.0", "41934"): "41934 UnderlyingPaymentStreamPricingBusinessDayConvention",
        ("FIX.5.0", "41935"): "41935 LegStreamCommoditySettlTimeType",
        ("FIX.5.0", "41936"): "41936 UnderlyingStreamCommoditySettlTimeType",
        ("FIX.5.0", "41937"): "41937 NoUnderlyingPaymentStreamPaymentDates",
        ("FIX.5.0", "41938"): "41938 UnderlyingPaymentStreamPaymentDate",
        ("FIX.5.0", "41939"): "41939 UnderlyingPaymentStreamPaymentDateType",
        ("FIX.5.0", "41940"): "41940 UnderlyingPaymentStreamMasterAgreementPaymentDatesIndicator",
        ("FIX.5.0", "41941"): "41941 NoUnderlyingPaymentStreamPricingDates",
        ("FIX.5.0", "41942"): "41942 UnderlyingPaymentStreamPricingDate",
        ("FIX.5.0", "41943"): "41943 UnderlyingPaymentStreamPricingDateType",
        ("FIX.5.0", "41944"): "41944 NoUnderlyingPaymentStreamPricingDays",
        ("FIX.5.0", "41945"): "41945 UnderlyingPaymentStreamPricingDayOfWeek",
        ("FIX.5.0", "41946"): "41946 UnderlyingPaymentStreamPricingDayNumber",
        ("FIX.5.0", "41947"): "41947 NoUnderlyingPricingDateBusinessCenters",
        ("FIX.5.0", "41948"): "41948 UnderlyingPricingDateBusinessCenter",
        ("FIX.5.0", "41949"): "41949 UnderlyingPricingDateUnadjusted",
        ("FIX.5.0", "41950"): "41950 UnderlyingPricingDateBusinessDayConvention",
        ("FIX.5.0", "41951"): "41951 UnderlyingPricingDateAdjusted",
        ("FIX.5.0", "41952"): "41952 UnderlyingPricingTime",
        ("FIX.5.0", "41953"): "41953 UnderlyingPricingTimeBusinessCenter",
        ("FIX.5.0", "41954"): "41954 NoUnderlyingStreamCalculationPeriodDates",
        ("FIX.5.0", "41955"): "41955 UnderlyingStreamCalculationPeriodDate",
        ("FIX.5.0", "41956"): "41956 UnderlyingStreamCalculationPeriodDateType",
        ("FIX.5.0", "41957"): "41957 UnderlyingStreamCalculationPeriodDatesXID",
        ("FIX.5.0", "41958"): "41958 UnderlyingStreamCalculationPeriodDatesXIDRef",
        ("FIX.5.0", "41959"): "41959 UnderlyingStreamCalculationBalanceOfFirstPeriod",
        ("FIX.5.0", "41960"): "41960 UnderlyingStreamCalculationCorrectionPeriod",
        ("FIX.5.0", "41961"): "41961 UnderlyingStreamCalculationCorrectionUnit",
        ("FIX.5.0", "41962"): "41962 NoUnderlyingStreamCommoditySettlBusinessCenters",
        ("FIX.5.0", "41963"): "41963 UnderlyingStreamCommoditySettlBusinessCenter",
        ("FIX.5.0", "41964"): "41964 UnderlyingStreamCommodityBase",
        ("FIX.5.0", "41965"): "41965 UnderlyingStreamCommodityType",
        ("FIX.5.0", "41966"): "41966 UnderlyingStreamCommoditySecurityID",
        ("FIX.5.0", "41967"): "41967 UnderlyingStreamCommoditySecurityIDSource",
        ("FIX.5.0", "41968"): "41968 UnderlyingStreamCommodityDesc",
        ("FIX.5.0", "41969"): "41969 EncodedUnderlyingStreamCommodityDescLen",
        ("FIX.5.0", "41970"): "41970 EncodedUnderlyingStreamCommodityDesc",
        ("FIX.5.0", "41971"): "41971 UnderlyingStreamCommodityUnitOfMeasure",
        ("FIX.5.0", "41972"): "41972 UnderlyingStreamCommodityCurrency",
        ("FIX.5.0", "41973"): "41973 UnderlyingStreamCommodityExchange",
        ("FIX.5.0", "41974"): "41974 UnderlyingStreamCommodityRateSource",
        ("FIX.5.0", "41975"): "41975 UnderlyingStreamCommodityRateReferencePage",
        ("FIX.5.0", "41976"): "41976 UnderlyingStreamCommodityRateReferencePageHeading",
        ("FIX.5.0", "41977"): "41977 UnderlyingStreamDataProvider",
        ("FIX.5.0", "41978"): "41978 UnderlyingStreamCommodityPricingType",
        ("FIX.5.0", "41979"): "41979 UnderlyingStreamCommodityNearbySettlDayPeriod",
        ("FIX.5.0", "41980"): "41980 UnderlyingStreamCommodityNearbySettlDayUnit",
        ("FIX.5.0", "41981"): "41981 UnderlyingStreamCommoditySettlDateUnadjusted",
        ("FIX.5.0", "41982"): "41982 UnderlyingStreamCommoditySettlDateBusinessDayConvention",
        ("FIX.5.0", "41983"): "41983 UnderlyingStreamCommoditySettlDateAdjusted",
        ("FIX.5.0", "41984"): "41984 UnderlyingStreamCommoditySettlMonth",
        ("FIX.5.0", "41985"): "41985 UnderlyingStreamCommoditySettlDateRollPeriod",
        ("FIX.5.0", "41986"): "41986 UnderlyingStreamCommoditySettlDateRollUnit",
        ("FIX.5.0", "41987"): "41987 UnderlyingStreamCommoditySettlDayType",
        ("FIX.5.0", "41988"): "41988 UnderlyingStreamCommodityXID",
        ("FIX.5.0", "41989"): "41989 UnderlyingStreamCommodityXIDRef",
        ("FIX.5.0", "41990"): "41990 NoUnderlyingStreamCommodityAltIDs",
        ("FIX.5.0", "41991"): "41991 UnderlyingStreamCommodityAltID",
        ("FIX.5.0", "41992"): "41992 UnderlyingStreamCommodityAltIDSource",
        ("FIX.5.0", "41993"): "41993 NoUnderlyingStreamCommodityDataSources",
        ("FIX.5.0", "41994"): "41994 UnderlyingStreamCommodityDataSourceID",
        ("FIX.5.0", "41995"): "41995 UnderlyingStreamCommodityDataSourceIDType",
        ("FIX.5.0", "41996"): "41996 NoUnderlyingStreamCommoditySettlDays",
        ("FIX.5.0", "41997"): "41997 UnderlyingStreamCommoditySettlDay",
        ("FIX.5.0", "41998"): "41998 UnderlyingStreamCommoditySettlTotalHours",
        ("FIX.5.0", "41999"): "41999 NoUnderlyingStreamCommoditySettlTimes",
        ("FIX.5.0", "42000"): "42000 UnderlyingStreamCommoditySettlStart",
        ("FIX.5.0", "42001"): "42001 UnderlyingStreamCommoditySettlEnd",
        ("FIX.5.0", "42002"): "42002 NoUnderlyingStreamCommoditySettlPeriods",
        ("FIX.5.0", "42003"): "42003 UnderlyingStreamCommoditySettlCountry",
        ("FIX.5.0", "42004"): "42004 UnderlyingStreamCommoditySettlTimeZone",
        ("FIX.5.0", "42005"): "42005 UnderlyingStreamCommoditySettlFlowType",
        ("FIX.5.0", "42006"): "42006 UnderlyingStreamCommoditySettlPeriodNotional",
        ("FIX.5.0", "42007"): "42007 UnderlyingStreamCommoditySettlPeriodNotionalUnitOfMeasure",
        ("FIX.5.0", "42008"): "42008 UnderlyingStreamCommoditySettlPeriodFrequencyPeriod",
        ("FIX.5.0", "42009"): "42009 UnderlyingStreamCommoditySettlPeriodFrequencyUnit",
        ("FIX.5.0", "42010"): "42010 UnderlyingStreamCommoditySettlPeriodPrice",
        ("FIX.5.0", "42011"): "42011 UnderlyingStreamCommoditySettlPeriodPriceUnitOfMeasure",
        ("FIX.5.0", "42012"): "42012 UnderlyingStreamCommoditySettlPeriodPriceCurrency",
        ("FIX.5.0", "42013"): "42013 UnderlyingStreamCommoditySettlHolidaysProcessingInstruction",
        ("FIX.5.0", "42014"): "42014 UnderlyingStreamCommoditySettlPeriodXID",
        ("FIX.5.0", "42015"): "42015 UnderlyingStreamCommoditySettlPeriodXIDRef",
        ("FIX.5.0", "42016"): "42016 UnderlyingStreamXID",
        ("FIX.5.0", "42017"): "42017 UnderlyingAdditionalTermBondIssuer",
        ("FIX.5.0", "42018"): "42018 UnderlyingStreamNotionalXIDRef",
        ("FIX.5.0", "42019"): "42019 UnderlyingStreamNotionalFrequencyPeriod",
        ("FIX.5.0", "42020"): "42020 UnderlyingStreamNotionalFrequencyUnit",
        ("FIX.5.0", "42021"): "42021 UnderlyingStreamNotionalCommodityFrequency",
        ("FIX.5.0", "42022"): "42022 UnderlyingStreamNotionalUnitOfMeasure",
        ("FIX.5.0", "42023"): "42023 UnderlyingStreamTotalNotional",
        ("FIX.5.0", "42024"): "42024 UnderlyingStreamTotalNotionalUnitOfMeasure",
        ("FIX.5.0", "42025"): "42025 EncodedUnderlyingAdditionalTermBondIssuerLen",
        ("FIX.5.0", "42026"): "42026 EncodedUnderlyingAdditionalTermBondIssuer",
        ("FIX.5.0", "42027"): "42027 UnderlyingAdditionalTermBondSeniority",
        ("FIX.5.0", "42028"): "42028 UnderlyingAdditionalTermBondCouponType",
        ("FIX.5.0", "42029"): "42029 UnderlyingAdditionalTermBondCouponRate",
        ("FIX.5.0", "42030"): "42030 UnderlyingAdditionalTermBondMaturityDate",
        ("FIX.5.0", "42031"): "42031 UnderlyingAdditionalTermBondParValue",
        ("FIX.5.0", "42032"): "42032 UnderlyingAdditionalTermBondCurrentTotalIssuedAmount",
        ("FIX.5.0", "42033"): "42033 UnderlyingAdditionalTermBondCouponFrequencyPeriod",
        ("FIX.5.0", "42034"): "42034 UnderlyingAdditionalTermBondCouponFrequencyUnit",
        ("FIX.5.0", "42035"): "42035 UnderlyingAdditionalTermBondDayCount",
        ("FIX.5.0", "42036"): "42036 NoUnderlyingAdditionalTerms",
        ("FIX.5.0", "42037"): "42037 UnderlyingAdditionalTermConditionPrecedentBondIndicator",
        ("FIX.5.0", "42038"): "42038 UnderlyingAdditionalTermDiscrepancyClauseIndicator",
        ("FIX.5.0", "42039"): "42039 NoUnderlyingCashSettlDealers",
        ("FIX.5.0", "42040"): "42040 UnderlyingCashSettlDealer",
        ("FIX.5.0", "42041"): "42041 NoUnderlyingCashSettlTerms",
        ("FIX.5.0", "42042"): "42042 UnderlyingCashSettlCurrency",
        ("FIX.5.0", "42043"): "42043 UnderlyingCashSettlValuationFirstBusinessDayOffset",
        ("FIX.5.0", "42044"): "42044 UnderlyingCashSettlValuationSubsequentBusinessDaysOffset",
        ("FIX.5.0", "42045"): "42045 UnderlyingCashSettlNumOfValuationDates",
        ("FIX.5.0", "42046"): "42046 UnderlyingCashSettlValuationTime",
        ("FIX.5.0", "42047"): "42047 UnderlyingCashSettlBusinessCenter",
        ("FIX.5.0", "42048"): "42048 UnderlyingCashSettlQuoteMethod",
        ("FIX.5.0", "42049"): "42049 UnderlyingCashSettlQuoteAmount",
        ("FIX.5.0", "42050"): "42050 UnderlyingCashSettlQuoteCurrency",
        ("FIX.5.0", "42051"): "42051 UnderlyingCashSettlMinimumQuoteAmount",
        ("FIX.5.0", "42052"): "42052 UnderlyingCashSettlMinimumQuoteCurrency",
        ("FIX.5.0", "42053"): "42053 UnderlyingCashSettlBusinessDays",
        ("FIX.5.0", "42054"): "42054 UnderlyingCashSettlAmount",
        ("FIX.5.0", "42055"): "42055 UnderlyingCashSettlRecoveryFactor",
        ("FIX.5.0", "42056"): "42056 UnderlyingCashSettlFixedTermIndicator",
        ("FIX.5.0", "42057"): "42057 UnderlyingCashSettlAccruedInterestIndicator",
        ("FIX.5.0", "42058"): "42058 UnderlyingCashSettlValuationMethod",
        ("FIX.5.0", "42059"): "42059 UnderlyingCashSettlTermXID",
        ("FIX.5.0", "42060"): "42060 NoUnderlyingPhysicalSettlTerms",
        ("FIX.5.0", "42061"): "42061 UnderlyingPhysicalSettlCurrency",
        ("FIX.5.0", "42062"): "42062 UnderlyingPhysicalSettlBusinessDays",
        ("FIX.5.0", "42063"): "42063 UnderlyingPhysicalSettlMaximumBusinessDays",
        ("FIX.5.0", "42064"): "42064 UnderlyingPhysicalSettlTermXID",
        ("FIX.5.0", "42065"): "42065 NoUnderlyingPhysicalSettlDeliverableObligations",
        ("FIX.5.0", "42066"): "42066 UnderlyingPhysicalSettlDeliverableObligationType",
        ("FIX.5.0", "42067"): "42067 UnderlyingPhysicalSettlDeliverableObligationValue",
        ("FIX.5.0", "42068"): "42068 NoUnderlyingProtectionTerms",
        ("FIX.5.0", "42069"): "42069 UnderlyingProtectionTermNotional",
        ("FIX.5.0", "42070"): "42070 UnderlyingProtectionTermCurrency",
        ("FIX.5.0", "42071"): "42071 UnderlyingProtectionTermSellerNotifies",
        ("FIX.5.0", "42072"): "42072 UnderlyingProtectionTermBuyerNotifies",
        ("FIX.5.0", "42073"): "42073 UnderlyingProtectionTermEventBusinessCenter",
        ("FIX.5.0", "42074"): "42074 UnderlyingProtectionTermStandardSources",
        ("FIX.5.0", "42075"): "42075 UnderlyingProtectionTermEventMinimumSources",
        ("FIX.5.0", "42076"): "42076 UnderlyingProtectionTermXID",
        ("FIX.5.0", "42077"): "42077 NoUnderlyingProtectionTermEvents",
        ("FIX.5.0", "42078"): "42078 UnderlyingProtectionTermEventType",
        ("FIX.5.0", "42079"): "42079 UnderlyingProtectionTermEventValue",
        ("FIX.5.0", "42080"): "42080 UnderlyingProtectionTermEventCurrency",
        ("FIX.5.0", "42081"): "42081 UnderlyingProtectionTermEventPeriod",
        ("FIX.5.0", "42082"): "42082 UnderlyingProtectionTermEventUnit",
        ("FIX.5.0", "42083"): "42083 UnderlyingProtectionTermEventDayType",
        ("FIX.5.0", "42084"): "42084 UnderlyingProtectionTermEventRateSource",
        ("FIX.5.0", "42085"): "42085 NoUnderlyingProtectionTermEventQualifiers",
        ("FIX.5.0", "42086"): "42086 UnderlyingProtectionTermEventQualifier",
        ("FIX.5.0", "42087"): "42087 NoUnderlyingProtectionTermObligations",
        ("FIX.5.0", "42088"): "42088 UnderlyingProtectionTermObligationType",
        ("FIX.5.0", "42089"): "42089 UnderlyingProtectionTermObligationValue",
        ("FIX.5.0", "42090"): "42090 NoUnderlyingProtectionTermEventNewsSources",
        ("FIX.5.0", "42091"): "42091 UnderlyingProtectionTermEventNewsSource",
        ("FIX.5.0", "42092"): "42092 UnderlyingProvisionCashSettlPaymentDateBusinessDayConvention",
        ("FIX.5.0", "42093"): "42093 UnderlyingProvisionCashSettlPaymentDateRelativeTo",
        ("FIX.5.0", "42094"): "42094 UnderlyingProvisionCashSettlPaymentDateOffsetPeriod",
        ("FIX.5.0", "42095"): "42095 UnderlyingProvisionCashSettlPaymentDateOffsetUnit",
        ("FIX.5.0", "42096"): "42096 UnderlyingProvisionCashSettlPaymentDateOffsetDayType",
        ("FIX.5.0", "42097"): "42097 UnderlyingProvisionCashSettlPaymentDateRangeFirst",
        ("FIX.5.0", "42098"): "42098 UnderlyingProvisionCashSettlPaymentDateRangeLast",
        ("FIX.5.0", "42099"): "42099 NoUnderlyingProvisionCashSettlPaymentDates",
        ("FIX.5.0", "42100"): "42100 UnderlyingProvisionCashSettlPaymentDate",
        ("FIX.5.0", "42101"): "42101 UnderlyingProvisionCashSettlPaymentDateType",
        ("FIX.5.0", "42102"): "42102 UnderlyingProvisionCashSettlQuoteSource",
        ("FIX.5.0", "42103"): "42103 UnderlyingProvisionCashSettlQuoteReferencePage",
        ("FIX.5.0", "42104"): "42104 UnderlyingProvisionCashSettlValueTime",
        ("FIX.5.0", "42105"): "42105 UnderlyingProvisionCashSettlValueTimeBusinessCenter",
        ("FIX.5.0", "42106"): "42106 UnderlyingProvisionCashSettlValueDateBusinessDayConvention",
        ("FIX.5.0", "42107"): "42107 UnderlyingProvisionCashSettlValueDateRelativeTo",
        ("FIX.5.0", "42108"): "42108 UnderlyingProvisionCashSettlValueDateOffsetPeriod",
        ("FIX.5.0", "42109"): "42109 UnderlyingProvisionCashSettlValueDateOffsetUnit",
        ("FIX.5.0", "42110"): "42110 UnderlyingProvisionCashSettlValueDateOffsetDayType",
        ("FIX.5.0", "42111"): "42111 UnderlyingProvisionCashSettlValueDateAdjusted",
        ("FIX.5.0", "42112"): "42112 NoUnderlyingProvisionOptionExerciseFixedDates",
        ("FIX.5.0", "42113"): "42113 UnderlyingProvisionOptionExerciseFixedDate",
        ("FIX.5.0", "42114"): "42114 UnderlyingProvisionOptionExerciseFixedDateType",
        ("FIX.5.0", "42115"): "42115 UnderlyingProvisionOptionExerciseBusinessDayConvention",
        ("FIX.5.0", "42116"): "42116 UnderlyingProvisionOptionExerciseEarliestDateOffsetPeriod",
        ("FIX.5.0", "42117"): "42117 UnderlyingProvisionOptionExerciseEarliestDateOffsetUnit",
        ("FIX.5.0", "42118"): "42118 UnderlyingProvisionOptionExerciseFrequencyPeriod",
        ("FIX.5.0", "42119"): "42119 UnderlyingProvisionOptionExerciseFrequencyUnit",
        ("FIX.5.0", "42120"): "42120 UnderlyingProvisionOptionExerciseStartDateUnadjusted",
        ("FIX.5.0", "42121"): "42121 UnderlyingProvisionOptionExerciseStartDateRelativeTo",
        ("FIX.5.0", "42122"): "42122 UnderlyingProvisionOptionExerciseStartDateOffsetPeriod",
        ("FIX.5.0", "42123"): "42123 UnderlyingProvisionOptionExerciseStartDateOffsetUnit",
        ("FIX.5.0", "42124"): "42124 UnderlyingProvisionOptionExerciseStartDateOffsetDayType",
        ("FIX.5.0", "42125"): "42125 UnderlyingProvisionOptionExerciseStartDateAdjusted",
        ("FIX.5.0", "42126"): "42126 UnderlyingProvisionOptionExercisePeriodSkip",
        ("FIX.5.0", "42127"): "42127 UnderlyingProvisionOptionExerciseBoundsFirstDateUnadjusted",
        ("FIX.5.0", "42128"): "42128 UnderlyingProvisionOptionExerciseBoundsLastDateUnadjusted",
        ("FIX.5.0", "42129"): "42129 UnderlyingProvisionOptionExerciseEarliestTime",
        ("FIX.5.0", "42130"): "42130 UnderlyingProvisionOptionExerciseEarliestTimeBusinessCenter",
        ("FIX.5.0", "42131"): "42131 UnderlyingProvisionOptionExerciseLatestTime",
        ("FIX.5.0", "42132"): "42132 UnderlyingProvisionOptionExerciseLatestTimeBusinessCenter",
        ("FIX.5.0", "42133"): "42133 UnderlyingProvisionOptionExpirationDateUnadjusted",
        ("FIX.5.0", "42134"): "42134 UnderlyingProvisionOptionExpirationDateBusinessDayConvention",
        ("FIX.5.0", "42135"): "42135 UnderlyingProvisionOptionExpirationDateRelativeTo",
        ("FIX.5.0", "42136"): "42136 UnderlyingProvisionOptionExpirationDateOffsetPeriod",
        ("FIX.5.0", "42137"): "42137 UnderlyingProvisionOptionExpirationDateOffsetUnit",
        ("FIX.5.0", "42138"): "42138 UnderlyingProvisionOptionExpirationDateOffsetDayType",
        ("FIX.5.0", "42139"): "42139 UnderlyingProvisionOptionExpirationDateAdjusted",
        ("FIX.5.0", "42140"): "42140 UnderlyingProvisionOptionExpirationTime",
        ("FIX.5.0", "42141"): "42141 UnderlyingProvisionOptionExpirationTimeBusinessCenter",
        ("FIX.5.0", "42142"): "42142 UnderlyingProvisionOptionRelevantUnderlyingDateUnadjusted",
        ("FIX.5.0", "42143"): "42143 UnderlyingProvisionOptionRelevantUnderlyingDateBusinessDayConvention",
        ("FIX.5.0", "42144"): "42144 UnderlyingProvisionOptionRelevantUnderlyingDateRelativeTo",
        ("FIX.5.0", "42145"): "42145 UnderlyingProvisionOptionRelevantUnderlyingDateOffsetPeriod",
        ("FIX.5.0", "42146"): "42146 UnderlyingProvisionOptionRelevantUnderlyingDateOffsetUnit",
        ("FIX.5.0", "42147"): "42147 UnderlyingProvisionOptionRelevantUnderlyingDateOffsetDayType",
        ("FIX.5.0", "42148"): "42148 UnderlyingProvisionOptionRelevantUnderlyingDateAdjusted",
        ("FIX.5.0", "42149"): "42149 NoUnderlyingProvisions",
        ("FIX.5.0", "42150"): "42150 UnderlyingProvisionType",
        ("FIX.5.0", "42151"): "42151 UnderlyingProvisionDateUnadjusted",
        ("FIX.5.0", "42152"): "42152 UnderlyingProvisionDateBusinessDayConvention",
        ("FIX.5.0", "42153"): "42153 UnderlyingProvisionDateAdjusted",
        ("FIX.5.0", "42154"): "42154 UnderlyingProvisionDateTenorPeriod",
        ("FIX.5.0", "42155"): "42155 UnderlyingProvisionDateTenorUnit",
        ("FIX.5.0", "42156"): "42156 UnderlyingProvisionCalculationAgent",
        ("FIX.5.0", "42157"): "42157 UnderlyingProvisionOptionSinglePartyBuyerSide",
        ("FIX.5.0", "42158"): "42158 UnderlyingProvisionOptionSinglePartySellerSide",
        ("FIX.5.0", "42159"): "42159 UnderlyingProvisionOptionExerciseStyle",
        ("FIX.5.0", "42160"): "42160 UnderlyingProvisionOptionExerciseMultipleNotional",
        ("FIX.5.0", "42161"): "42161 UnderlyingProvisionOptionExerciseMinimumNotional",
        ("FIX.5.0", "42162"): "42162 UnderlyingProvisionOptionExerciseMaximumNotional",
        ("FIX.5.0", "42163"): "42163 UnderlyingProvisionOptionMinimumNumber",
        ("FIX.5.0", "42164"): "42164 UnderlyingProvisionOptionMaximumNumber",
        ("FIX.5.0", "42165"): "42165 UnderlyingProvisionOptionExerciseConfirmation",
        ("FIX.5.0", "42166"): "42166 UnderlyingProvisionCashSettlMethod",
        ("FIX.5.0", "42167"): "42167 UnderlyingProvisionCashSettlCurrency",
        ("FIX.5.0", "42168"): "42168 UnderlyingProvisionCashSettlCurrency2",
        ("FIX.5.0", "42169"): "42169 UnderlyingProvisionCashSettlQuoteType",
        ("FIX.5.0", "42170"): "42170 UnderlyingProvisionText",
        ("FIX.5.0", "42171"): "42171 EncodedUnderlyingProvisionTextLen",
        ("FIX.5.0", "42172"): "42172 EncodedUnderlyingProvisionText",
        ("FIX.5.0", "42173"): "42173 NoUnderlyingProvisionPartyIDs",
        ("FIX.5.0", "42174"): "42174 UnderlyingProvisionPartyID",
        ("FIX.5.0", "42175"): "42175 UnderlyingProvisionPartyIDSource",
        ("FIX.5.0", "42176"): "42176 UnderlyingProvisionPartyRole",
        ("FIX.5.0", "42177"): "42177 NoUnderlyingProvisionPartySubIDs",
        ("FIX.5.0", "42178"): "42178 UnderlyingProvisionPartySubID",
        ("FIX.5.0", "42179"): "42179 UnderlyingProvisionPartySubIDType",
        ("FIX.5.0", "42180"): "42180 NoUnderlyingProvisionCashSettlPaymentDateBusinessCenters",
        ("FIX.5.0", "42181"): "42181 UnderlyingProvisionCashSettlPaymentDateBusinessCenter",
        ("FIX.5.0", "42182"): "42182 NoUnderlyingProvisionCashSettlValueDateBusinessCenters",
        ("FIX.5.0", "42183"): "42183 UnderlyingProvisionCashSettlValueDateBusinessCenter",
        ("FIX.5.0", "42184"): "42184 NoUnderlyingProvisionOptionExerciseBusinessCenters",
        ("FIX.5.0", "42185"): "42185 UnderlyingProvisionOptionExerciseBusinessCenter",
        ("FIX.5.0", "42186"): "42186 NoUnderlyingProvisionOptionExpirationDateBusinessCenters",
        ("FIX.5.0", "42187"): "42187 UnderlyingProvisionOptionExpirationDateBusinessCenter",
        ("FIX.5.0", "42188"): "42188 NoUnderlyingProvisionOptionRelevantUnderlyingDateBusinessCenters",
        ("FIX.5.0", "42189"): "42189 UnderlyingProvisionOptionRelevantUnderlyingDateBusinessCenter",
        ("FIX.5.0", "42190"): "42190 NoUnderlyingProvisionDateBusinessCenters",
        ("FIX.5.0", "42191"): "42191 UnderlyingProvisionDateBusinessCenter",
        ("FIX.5.0", "42192"): "42192 DeliveryStreamDeliveryPointSource",
        ("FIX.5.0", "42193"): "42193 DeliveryStreamDeliveryPointDesc",
        ("FIX.5.0", "42194"): "42194 LegDeliveryStreamDeliveryPointSource",
        ("FIX.5.0", "42195"): "42195 LegDeliveryStreamDeliveryPointDesc",
        ("FIX.5.0", "42196"): "42196 UnderlyingDeliveryStreamDeliveryPointSource",
        ("FIX.5.0", "42197"): "42197 UnderlyingDeliveryStreamDeliveryPointDesc",
        ("FIX.5.0", "42198"): "42198 NoLegContractualDefinitions",
        ("FIX.5.0", "42199"): "42199 LegContractualDefinition",
        ("FIX.5.0", "42200"): "42200 NoLegFinancingTermSupplements",
        ("FIX.5.0", "42201"): "42201 LegFinancingTermSupplementDesc",
        ("FIX.5.0", "42202"): "42202 LegFinancingTermSupplementDate",
        ("FIX.5.0", "42203"): "42203 NoLegContractualMatrices",
        ("FIX.5.0", "42204"): "42204 LegContractualMatrixSource",
        ("FIX.5.0", "42205"): "42205 LegContractualMatrixDate",
        ("FIX.5.0", "42206"): "42206 LegContractualMatrixTerm",
        ("FIX.5.0", "42207"): "42207 CashSettlDateUnadjusted",
        ("FIX.5.0", "42208"): "42208 CashSettlDateBusinessDayConvention",
        ("FIX.5.0", "42209"): "42209 CashSettlDateRelativeTo",
        ("FIX.5.0", "42210"): "42210 CashSettlDateOffsetPeriod",
        ("FIX.5.0", "42211"): "42211 CashSettlDateOffsetUnit",
        ("FIX.5.0", "42212"): "42212 CashSettlDateOffsetDayType",
        ("FIX.5.0", "42213"): "42213 CashSettlDateAdjusted",
        ("FIX.5.0", "42214"): "42214 NoCashSettlDateBusinessCenters",
        ("FIX.5.0", "42215"): "42215 CashSettlDateBusinessCenter",
        ("FIX.5.0", "42216"): "42216 CashSettlPriceSource",
        ("FIX.5.0", "42217"): "42217 CashSettlPriceDefault",
        ("FIX.5.0", "42218"): "42218 DividendFloatingRateIndex",
        ("FIX.5.0", "42219"): "42219 DividendFloatingRateIndexCurvePeriod",
        ("FIX.5.0", "42220"): "42220 DividendFloatingRateIndexCurveUnit",
        ("FIX.5.0", "42221"): "42221 DividendFloatingRateMultiplier",
        ("FIX.5.0", "42222"): "42222 DividendFloatingRateSpread",
        ("FIX.5.0", "42223"): "42223 DividendFloatingRateSpreadPositionType",
        ("FIX.5.0", "42224"): "42224 DividendFloatingRateTreatment",
        ("FIX.5.0", "42225"): "42225 DividendCapRate",
        ("FIX.5.0", "42226"): "42226 DividendCapRateBuySide",
        ("FIX.5.0", "42227"): "42227 DividendCapRateSellSide",
        ("FIX.5.0", "42228"): "42228 DividendFloorRate",
        ("FIX.5.0", "42229"): "42229 DividendFloorRateBuySide",
        ("FIX.5.0", "42230"): "42230 DividendFloorRateSellSide",
        ("FIX.5.0", "42231"): "42231 DividendInitialRate",
        ("FIX.5.0", "42232"): "42232 DividendFinalRateRoundingDirection",
        ("FIX.5.0", "42233"): "42233 DividendFinalRatePrecision",
        ("FIX.5.0", "42234"): "42234 DividendAveragingMethod",
        ("FIX.5.0", "42235"): "42235 DividendNegativeRateTreatment",
        ("FIX.5.0", "42236"): "42236 NoDividendAccrualPaymentDateBusinessCenters",
        ("FIX.5.0", "42237"): "42237 DividendAccrualPaymentDateBusinessCenter",
        ("FIX.5.0", "42238"): "42238 DividendAccrualPaymentDateRelativeTo",
        ("FIX.5.0", "42239"): "42239 DividendAccrualPaymentDateOffsetPeriod",
        ("FIX.5.0", "42240"): "42240 DividendAccrualPaymentDateOffsetUnit",
        ("FIX.5.0", "42241"): "42241 DividendAccrualPaymentDateOffsetDayType",
        ("FIX.5.0", "42242"): "42242 DividendAccrualPaymentDateUnadjusted",
        ("FIX.5.0", "42243"): "42243 DividendAccrualPaymeentDateBusinessDayConvention",
        ("FIX.5.0", "42244"): "42244 DividendAccrualPaymentDateAdjusted",
        ("FIX.5.0", "42245"): "42245 DividendReinvestmentIndicator",
        ("FIX.5.0", "42246"): "42246 DividendEntitlementEvent",
        ("FIX.5.0", "42247"): "42247 DividendAmountType",
        ("FIX.5.0", "42248"): "42248 DividendUnderlierRefID",
        ("FIX.5.0", "42249"): "42249 ExtraordinaryDividendPartySide",
        ("FIX.5.0", "42250"): "42250 ExtraordinaryDividendAmountType",
        ("FIX.5.0", "42251"): "42251 ExtraordinaryDividendCurrency",
        ("FIX.5.0", "42252"): "42252 ExtraordinaryDividendDeterminationMethod",
        ("FIX.5.0", "42253"): "42253 DividendAccrualFixedRate",
        ("FIX.5.0", "42254"): "42254 DividendCompoundingMethod",
        ("FIX.5.0", "42255"): "42255 DividendNumOfIndexUnits",
        ("FIX.5.0", "42256"): "42256 DividendCashPercentage",
        ("FIX.5.0", "42257"): "42257 DividendCashEquivalentPercentage",
        ("FIX.5.0", "42258"): "42258 NonCashDividendTreatment",
        ("FIX.5.0", "42259"): "42259 DividendComposition",
        ("FIX.5.0", "42260"): "42260 SpecialDividendsIndicator",
        ("FIX.5.0", "42261"): "42261 MaterialDividendsIndicator",
        ("FIX.5.0", "42262"): "42262 OptionsExchangeDividendsIndicator",
        ("FIX.5.0", "42263"): "42263 AdditionalDividendsIndicator",
        ("FIX.5.0", "42264"): "42264 AllDividendsIndicator",
        ("FIX.5.0", "42265"): "42265 DividendFXTriggerDateRelativeTo",
        ("FIX.5.0", "42266"): "42266 DividendFXTriggerDateOffsetPeriod",
        ("FIX.5.0", "42267"): "42267 DividendFXTriggerDateOffsetUnit",
        ("FIX.5.0", "42268"): "42268 DividendFXTriggerDateOffsetDayType",
        ("FIX.5.0", "42269"): "42269 DividendFXTriggerDateUnadjusted",
        ("FIX.5.0", "42270"): "42270 DividendFXTriggerDateBusinessDayConvention",
        ("FIX.5.0", "42271"): "42271 DividendFXTriggerDateAdjusted",
        ("FIX.5.0", "42272"): "42272 NoDividendFXTriggerDateBusinessCenters",
        ("FIX.5.0", "42273"): "42273 DividendFXTriggerDateBusinessCenter",
        ("FIX.5.0", "42274"): "42274 NoDividendPeriods",
        ("FIX.5.0", "42275"): "42275 DividendPeriodSequence",
        ("FIX.5.0", "42276"): "42276 DividendPeriodStartDateUnadjusted",
        ("FIX.5.0", "42277"): "42277 DividendPeriodEndDateUnadjusted",
        ("FIX.5.0", "42278"): "42278 DividendPeriodUnderlierRefID",
        ("FIX.5.0", "42279"): "42279 DividendPeriodStrikePrice",
        ("FIX.5.0", "42280"): "42280 DividendPeriodBusinessDayConvention",
        ("FIX.5.0", "42281"): "42281 DividendPeriodValuationDateUnadjusted",
        ("FIX.5.0", "42282"): "42282 DividendPeriodValuationDateRelativeTo",
        ("FIX.5.0", "42283"): "42283 DividendPeriodValuationDateOffsetPeriod",
        ("FIX.5.0", "42284"): "42284 DividendPeriodValuationDateOffsetUnit",
        ("FIX.5.0", "42285"): "42285 DividendPeriodValuationDateOffsetDayType",
        ("FIX.5.0", "42286"): "42286 DividendPeriodValuationDateAdjusted",
        ("FIX.5.0", "42287"): "42287 DividendPeriodPaymentDateUnadjusted",
        ("FIX.5.0", "42288"): "42288 DividendPeriodPaymentDateRelativeTo",
        ("FIX.5.0", "42289"): "42289 DividendPeriodPaymentDateOffsetPeriod",
        ("FIX.5.0", "42290"): "42290 DividendPeriodPaymentDateOffsetUnit",
        ("FIX.5.0", "42291"): "42291 DividendPeriodPaymentDateOffsetDayType",
        ("FIX.5.0", "42292"): "42292 DividendPeriodPaymentDateAdjusted",
        ("FIX.5.0", "42293"): "42293 DividendPeriodXID",
        ("FIX.5.0", "42294"): "42294 NoDividendPeriodBusinessCenters",
        ("FIX.5.0", "42295"): "42295 DividendPeriodBusinessCenter",
        ("FIX.5.0", "42296"): "42296 NoExtraordinaryEvents",
        ("FIX.5.0", "42297"): "42297 ExtraordinaryEventType",
        ("FIX.5.0", "42298"): "42298 ExtraordinaryEventValue",
        ("FIX.5.0", "42299"): "42299 LegCashSettlDateUnadjusted",
        ("FIX.5.0", "42300"): "42300 LegCashSettlDateBusinessDayConvention",
        ("FIX.5.0", "42301"): "42301 LegCashSettlDateRelativeTo",
        ("FIX.5.0", "42302"): "42302 LegCashSettlDateOffsetPeriod",
        ("FIX.5.0", "42303"): "42303 LegCashSettlDateOffsetUnit",
        ("FIX.5.0", "42304"): "42304 LegCashSettlDateOffsetDayType",
        ("FIX.5.0", "42305"): "42305 LegCashSettlDateAdjusted",
        ("FIX.5.0", "42306"): "42306 NoLegCashSettlDateBusinessCenters",
        ("FIX.5.0", "42307"): "42307 LegCashSettlDateBusinessCenter",
        ("FIX.5.0", "42308"): "42308 LegCashSettlPriceSource",
        ("FIX.5.0", "42309"): "42309 LegCashSettlPriceDefault",
        ("FIX.5.0", "42310"): "42310 NoLegDividendAccrualPaymentDateBusinessCenters",
        ("FIX.5.0", "42311"): "42311 LegDividendAccrualPaymentDateBusinessCenter",
        ("FIX.5.0", "42312"): "42312 LegDividendFloatingRateIndex",
        ("FIX.5.0", "42313"): "42313 LegDividendFloatingRateIndexCurvePeriod",
        ("FIX.5.0", "42314"): "42314 LegDividendFloatingRateIndexCurveUnit",
        ("FIX.5.0", "42315"): "42315 LegDividendFloatingRateMultiplier",
        ("FIX.5.0", "42316"): "42316 LegDividendFloatingRateSpread",
        ("FIX.5.0", "42317"): "42317 LegDividendFloatingRateSpreadPositionType",
        ("FIX.5.0", "42318"): "42318 LegDividendFloatingRateTreatment",
        ("FIX.5.0", "42319"): "42319 LegDividendCapRate",
        ("FIX.5.0", "42320"): "42320 LegDividendCapRateBuySide",
        ("FIX.5.0", "42321"): "42321 LegDividendCapRateSellSide",
        ("FIX.5.0", "42322"): "42322 LegDividendFloorRate",
        ("FIX.5.0", "42323"): "42323 LegDividendFloorRateBuySide",
        ("FIX.5.0", "42324"): "42324 LegDividendFloorRateSellSide",
        ("FIX.5.0", "42325"): "42325 LegDividendInitialRate",
        ("FIX.5.0", "42326"): "42326 LegDividendFinalRateRoundingDirection",
        ("FIX.5.0", "42327"): "42327 LegDividendFinalRatePrecision",
        ("FIX.5.0", "42328"): "42328 LegDividendAveragingMethod",
        ("FIX.5.0", "42329"): "42329 LegDividendNegativeRateTreatment",
        ("FIX.5.0", "42330"): "42330 LegDividendAccrualPaymentDateRelativeTo",
        ("FIX.5.0", "42331"): "42331 LegDividendAccrualPaymentDateOffsetPeriod",
        ("FIX.5.0", "42332"): "42332 LegDividendAccrualPaymentDateOffsetUnit",
        ("FIX.5.0", "42333"): "42333 LegDividendAccrualPaymentDateOffsetDayType",
        ("FIX.5.0", "42334"): "42334 LegDividendAccrualPaymentDateUnadjusted",
        ("FIX.5.0", "42335"): "42335 LegDividendAccrualPaymentDateBusinessDayConvention",
        ("FIX.5.0", "42336"): "42336 LegDividendAccrualPaymentDateAdjusted",
        ("FIX.5.0", "42337"): "42337 LegDividendReinvestmentIndicator",
        ("FIX.5.0", "42338"): "42338 LegDividendEntitlementEvent",
        ("FIX.5.0", "42339"): "42339 LegDividendAmountType",
        ("FIX.5.0", "42340"): "42340 LegDividendUnderlierRefID",
        ("FIX.5.0", "42341"): "42341 LegExtraordinaryDividendPartySide",
        ("FIX.5.0", "42342"): "42342 LegExtraordinaryDividendAmountType",
        ("FIX.5.0", "42343"): "42343 LegExtraordinaryDividendCurrency",
        ("FIX.5.0", "42344"): "42344 LegExtraordinaryDividendDeterminationMethod",
        ("FIX.5.0", "42345"): "42345 LegDividendAccrualFixedRate",
        ("FIX.5.0", "42346"): "42346 LegDividendCompoundingMethod",
        ("FIX.5.0", "42347"): "42347 LegDividendNumOfIndexUnits",
        ("FIX.5.0", "42348"): "42348 LegDividendCashPercentage",
        ("FIX.5.0", "42349"): "42349 LegDividendCashEquivalentPercentage",
        ("FIX.5.0", "42350"): "42350 LegNonCashDividendTreatment",
        ("FIX.5.0", "42351"): "42351 LegDividendComposition",
        ("FIX.5.0", "42352"): "42352 LegSpecialDividendsIndicator",
        ("FIX.5.0", "42353"): "42353 LegMaterialDividendsIndicator",
        ("FIX.5.0", "42354"): "42354 LegOptionsExchangeDividendsIndicator",
        ("FIX.5.0", "42355"): "42355 LegAdditionalDividendsIndicator",
        ("FIX.5.0", "42356"): "42356 LegAllDividendsIndicator",
        ("FIX.5.0", "42357"): "42357 LegDividendFXTriggerDateRelativeTo",
        ("FIX.5.0", "42358"): "42358 LegDividendFXTriggerDateOffsetPeriod",
        ("FIX.5.0", "42359"): "42359 LegDividendFXTriggerDateOffsetUnit",
        ("FIX.5.0", "42360"): "42360 LegDividendFXTriggerDateOffsetDayType",
        ("FIX.5.0", "42361"): "42361 LegDividendFXTriggerDateUnadjusted",
        ("FIX.5.0", "42362"): "42362 LegDividendFXTriggerDateBusinessDayConvention",
        ("FIX.5.0", "42363"): "42363 LegDividendFXTriggerDateAdjusted",
        ("FIX.5.0", "42364"): "42364 NoLegDividendFXTriggerDateBusinessCenters",
        ("FIX.5.0", "42365"): "42365 LegDividendFXTriggerDateBusinessCenter",
        ("FIX.5.0", "42366"): "42366 NoLegDividendPeriods",
        ("FIX.5.0", "42367"): "42367 LegDividendPeriodSequence",
        ("FIX.5.0", "42368"): "42368 LegDividendPeriodStartDateUnadjusted",
        ("FIX.5.0", "42369"): "42369 LegDividendPeriodEndDateUnadjusted",
        ("FIX.5.0", "42370"): "42370 LegDividendPeriodUnderlierRefID",
        ("FIX.5.0", "42371"): "42371 LegDividendPeriodStrikePrice",
        ("FIX.5.0", "42372"): "42372 LegDividendPeriodBusinessDayConvention",
        ("FIX.5.0", "42373"): "42373 LegDividendPeriodValuationDateUnadjusted",
        ("FIX.5.0", "42374"): "42374 LegDividendPeriodValuationDateRelativeTo",
        ("FIX.5.0", "42375"): "42375 LegDividendPeriodValuationDateOffsetPeriod",
        ("FIX.5.0", "42376"): "42376 LegDividendPeriodValuationDateOffsetUnit",
        ("FIX.5.0", "42377"): "42377 LegDividendPeriodValuationDateOffsetDayType",
        ("FIX.5.0", "42378"): "42378 LegDividendPeriodValuationDateAdjusted",
        ("FIX.5.0", "42379"): "42379 LegDividendPeriodPaymentDateUnadjusted",
        ("FIX.5.0", "42380"): "42380 LegDividendPeriodPaymentDateRelativeTo",
        ("FIX.5.0", "42381"): "42381 LegDividendPeriodPaymentDateOffsetPeriod",
        ("FIX.5.0", "42382"): "42382 LegDividendPeriodPaymentDateOffsetUnit",
        ("FIX.5.0", "42383"): "42383 LegDividendPeriodPaymentDateOffsetDayType",
        ("FIX.5.0", "42384"): "42384 LegDividendPeriodPaymentDateAdjusted",
        ("FIX.5.0", "42385"): "42385 LegDividendPeriodXID",
        ("FIX.5.0", "42386"): "42386 NoLegDividendPeriodBusinessCenters",
        ("FIX.5.0", "42387"): "42387 LegDividendPeriodBusinessCenter",
        ("FIX.5.0", "42388"): "42388 NoLegExtraordinaryEvents",
        ("FIX.5.0", "42389"): "42389 LegExtraordinaryEventType",
        ("FIX.5.0", "42390"): "42390 LegExtraordinaryEventValue",
        ("FIX.5.0", "42391"): "42391 LegSettlMethodElectingPartySide",
        ("FIX.5.0", "42392"): "42392 LegMakeWholeDate",
        ("FIX.5.0", "42393"): "42393 LegMakeWholeAmount",
        ("FIX.5.0", "42394"): "42394 LegMakeWholeBenchmarkCurveName",
        ("FIX.5.0", "42395"): "42395 LegMakeWholeBenchmarkCurvePoint",
        ("FIX.5.0", "42396"): "42396 LegMakeWholeRecallSpread",
        ("FIX.5.0", "42397"): "42397 LegMakeWholeBenchmarkQuote",
        ("FIX.5.0", "42398"): "42398 LegMakeWholeInterpolationMethod",
        ("FIX.5.0", "42399"): "42399 LegPaymentStreamCashSettlIndicator",
        ("FIX.5.0", "42400"): "42400 LegPaymentStreamCompoundingXIDRef",
        ("FIX.5.0", "42401"): "42401 LegPaymentStreamCompoundingSpread",
        ("FIX.5.0", "42402"): "42402 LegPaymentStreamInterpolationMethod",
        ("FIX.5.0", "42403"): "42403 LegPaymentStreamInterpolationPeriod",
        ("FIX.5.0", "42404"): "42404 LegPaymentStreamCompoundingFixedRate",
        ("FIX.5.0", "42405"): "42405 NoLegPaymentStreamCompoundingDates",
        ("FIX.5.0", "42406"): "42406 LegPaymentStreamCompoundingDate",
        ("FIX.5.0", "42407"): "42407 LegPaymentStreamCompoundingDateType",
        ("FIX.5.0", "42408"): "42408 LegPaymentStreamCompoundingDatesBusinessDayConvention",
        ("FIX.5.0", "42409"): "42409 LegPaymentStreamCompoundingDatesRelativeTo",
        ("FIX.5.0", "42410"): "42410 LegPaymentStreamCompoundingDatesOffsetPeriod",
        ("FIX.5.0", "42411"): "42411 LegPaymentStreamCompoundingDatesOffsetUnit",
        ("FIX.5.0", "42412"): "42412 LegPaymentStreamCompoundingDatesOffsetDayType",
        ("FIX.5.0", "42413"): "42413 LegPaymentStreamCompoundingPeriodSkip",
        ("FIX.5.0", "42414"): "42414 LegPaymentStreamCompoundingFrequencyPeriod",
        ("FIX.5.0", "42415"): "42415 LegPaymentStreamCompoundingFrequencyUnit",
        ("FIX.5.0", "42416"): "42416 LegPaymentStreamCompoundingRollConvention",
        ("FIX.5.0", "42417"): "42417 LegPaymentStreamBoundsFirstDateUnadjusted",
        ("FIX.5.0", "42418"): "42418 LegPaymentStreamBoundsLastDateUnadjusted",
        ("FIX.5.0", "42419"): "42419 NoLegPaymentStreamCompoundingDatesBusinessCenters",
        ("FIX.5.0", "42420"): "42420 LegPaymentStreamCompoundingDatesBusinessCenter",
        ("FIX.5.0", "42421"): "42421 LegPaymentStreamCompoundingEndDateUnadjusted",
        ("FIX.5.0", "42422"): "42422 LegPaymentStreamCompoundingEndDateRelativeTo",
        ("FIX.5.0", "42423"): "42423 LegPaymentStreamCompoundingEndDateOffsetPeriod",
        ("FIX.5.0", "42424"): "42424 LegPaymentStreamCompoundingEndDateOffsetUnit",
        ("FIX.5.0", "42425"): "42425 LegPaymentStreamCompoundingEndDateOffsetDayType",
        ("FIX.5.0", "42426"): "42426 LegPaymentStreamCompoundingEndDateAdjusted",
        ("FIX.5.0", "42427"): "42427 LegPaymentStreamCompoundingRateIndex",
        ("FIX.5.0", "42428"): "42428 LegPaymentStreamCompoundingRateIndexCurvePeriod",
        ("FIX.5.0", "42429"): "42429 LegPaymentStreamCompoundingRateIndexCurveUnit",
        ("FIX.5.0", "42430"): "42430 LegPaymentStreamCompoundingRateMultiplier",
        ("FIX.5.0", "42431"): "42431 LegPaymentStreamCompoundingRateSpread",
        ("FIX.5.0", "42432"): "42432 LegPaymentStreamCompoundingRateSpreadPositionType",
        ("FIX.5.0", "42433"): "42433 LegPaymentStreamCompoundingRateTreatment",
        ("FIX.5.0", "42434"): "42434 LegPaymentStreamCompoundingCapRate",
        ("FIX.5.0", "42435"): "42435 LegPaymentStreamCompoundingCapRateBuySide",
        ("FIX.5.0", "42436"): "42436 LegPaymentStreamCompoundingCapRateSellSide",
        ("FIX.5.0", "42437"): "42437 LegPaymentStreamCompoundingFloorRate",
        ("FIX.5.0", "42438"): "42438 LegPaymentStreamCompoundingFloorRateBuySide",
        ("FIX.5.0", "42439"): "42439 LegPaymentStreamCompoundingFloorRateSellSide",
        ("FIX.5.0", "42440"): "42440 LegPaymentStreamCompoundingInitialRate",
        ("FIX.5.0", "42441"): "42441 LegPaymentStreamCompoundingFinalRateRoundingDirection",
        ("FIX.5.0", "42442"): "42442 LegPaymentStreamCompoundingFinalRatePrecision",
        ("FIX.5.0", "42443"): "42443 LegPaymentStreamCompoundingAveragingMethod",
        ("FIX.5.0", "42444"): "42444 LegPaymentStreamCompoundingNegativeRateTreatment",
        ("FIX.5.0", "42445"): "42445 LegPaymentStreamCompoundingStartDateUnadjusted",
        ("FIX.5.0", "42446"): "42446 LegPaymentStreamCompoundingStartDateRelativeTo",
        ("FIX.5.0", "42447"): "42447 LegPaymentStreamCompoundingStartDateOffsetPeriod",
        ("FIX.5.0", "42448"): "42448 LegPaymentStreamCompoundingStartDateOffsetUnit",
        ("FIX.5.0", "42449"): "42449 LegPaymentStreamCompoundingStartDateOffsetDayType",
        ("FIX.5.0", "42450"): "42450 LegPaymentStreamCompoundingStartDateAdjusted",
        ("FIX.5.0", "42451"): "42451 LegPaymentStreamFormulaImageLength",
        ("FIX.5.0", "42452"): "42452 LegPaymentStreamFormulaImage",
        ("FIX.5.0", "42453"): "42453 LegPaymentStreamFinalPricePaymentDateUnadjusted",
        ("FIX.5.0", "42454"): "42454 LegPaymentStreamFinalPricePaymentDateRelativeTo",
        ("FIX.5.0", "42455"): "42455 LegPaymentStreamFinalPricePaymentDateOffsetPeriod",
        ("FIX.5.0", "42456"): "42456 LegPaymentStreamFinalPricePaymentDateOffsetUnit",
        ("FIX.5.0", "42457"): "42457 LegPaymentStreamFinalPricePaymentDateOffsetDayType",
        ("FIX.5.0", "42458"): "42458 LegPaymentStreamFinalPricePaymentDateAdjusted",
        ("FIX.5.0", "42459"): "42459 NoLegPaymentStreamFixingDates",
        ("FIX.5.0", "42460"): "42460 LegPaymentStreamFixingDate",
        ("FIX.5.0", "42461"): "42461 LegPaymentStreamFixingDateType",
        ("FIX.5.0", "42462"): "42462 LegPaymentStreamFirstObservationDateUnadjusted",
        ("FIX.5.0", "42463"): "42463 LegPaymentStreamFirstObservationDateRelativeTo",
        ("FIX.5.0", "42464"): "42464 LegPaymentStreamFirstObservationDateOffsetDayType",
        ("FIX.5.0", "42465"): "42465 LegPaymentStreamFirstObservationDateAdjusted",
        ("FIX.5.0", "42466"): "42466 LegPaymentStreamUnderlierRefID",
        ("FIX.5.0", "42467"): "42467 LegReturnRateNotionalReset",
        ("FIX.5.0", "42468"): "42468 LegPaymentStreamLinkInitialLevel",
        ("FIX.5.0", "42469"): "42469 LegPaymentStreamLinkClosingLevelIndicator",
        ("FIX.5.0", "42470"): "42470 LegPaymentStreamLinkExpiringLevelIndicator",
        ("FIX.5.0", "42471"): "42471 LegPaymentStreamLinkEstimatedTradingDays",
        ("FIX.5.0", "42472"): "42472 LegPaymentStreamLinkStrikePrice",
        ("FIX.5.0", "42473"): "42473 LegPaymentStreamLinkStrikePriceType",
        ("FIX.5.0", "42474"): "42474 LegPaymentStreamLinkMaximumBoundary",
        ("FIX.5.0", "42475"): "42475 LegPaymentStreamLinkMinimumBoundary",
        ("FIX.5.0", "42476"): "42476 LegPaymentStreamLinkNumberOfDataSeries",
        ("FIX.5.0", "42477"): "42477 LegPaymentStreamVarianceUnadjustedCap",
        ("FIX.5.0", "42478"): "42478 LegPaymentStreamRealizedVarianceMethod",
        ("FIX.5.0", "42479"): "42479 LegPaymentStreamDaysAdjustmentIndicator",
        ("FIX.5.0", "42480"): "42480 LegPaymentStreamNearestExchangeContractRefID",
        ("FIX.5.0", "42481"): "42481 LegPaymentStreamVegaNotionalAmount",
        ("FIX.5.0", "42482"): "42482 LegPaymentStreamFormulaCurrency",
        ("FIX.5.0", "42483"): "42483 LegPaymentStreamFormulaCurrencyDeterminationMethod",
        ("FIX.5.0", "42484"): "42484 LegPaymentStreamFormulaReferenceAmount",
        ("FIX.5.0", "42485"): "42485 NoLegPaymentStreamFormulas",
        ("FIX.5.0", "42486"): "42486 LegPaymentStreamFormula",
        ("FIX.5.0", "42487"): "42487 LegPaymentStreamFormulaDesc",
        ("FIX.5.0", "42488"): "42488 LegPaymentStubEndDateUnadjusted",
        ("FIX.5.0", "42489"): "42489 LegPaymentStubEndDateBusinessDayConvention",
        ("FIX.5.0", "42490"): "42490 LegPaymentStubEndDateRelativeTo",
        ("FIX.5.0", "42491"): "42491 LegPaymentStubEndDateOffsetPeriod",
        ("FIX.5.0", "42492"): "42492 LegPaymentStubEndDateOffsetUnit",
        ("FIX.5.0", "42493"): "42493 LegPaymentStubEndDateOffsetDayType",
        ("FIX.5.0", "42494"): "42494 LegPaymentStubEndDateAdjusted",
        ("FIX.5.0", "42495"): "42495 NoLegPaymentStubEndDateBusinessCenters",
        ("FIX.5.0", "42496"): "42496 LegPaymentStubEndDateBusinessCenter",
        ("FIX.5.0", "42497"): "42497 LegPaymentStubStartDateUnadjusted",
        ("FIX.5.0", "42498"): "42498 LegPaymentStubStartDateBusinessDayConvention",
        ("FIX.5.0", "42499"): "42499 LegPaymentStubStartDateRelativeTo",
        ("FIX.5.0", "42500"): "42500 LegPaymentStubStartDateOffsetPeriod",
        ("FIX.5.0", "42501"): "42501 LegPaymentStubStartDateOffsetUnit",
        ("FIX.5.0", "42502"): "42502 LegPaymentStubStartDateOffsetDayType",
        ("FIX.5.0", "42503"): "42503 LegPaymentStubStartDateAdjusted",
        ("FIX.5.0", "42504"): "42504 NoLegPaymentStubStartDateBusinessCenters",
        ("FIX.5.0", "42505"): "42505 LegPaymentStubStartDateBusinessCenter",
        ("FIX.5.0", "42506"): "42506 LegProvisionBreakFeeElection",
        ("FIX.5.0", "42507"): "42507 LegProvisionBreakFeeRate",
        ("FIX.5.0", "42508"): "42508 NoLegReturnRateDates",
        ("FIX.5.0", "42509"): "42509 LegReturnRateDateMode",
        ("FIX.5.0", "42510"): "42510 LegReturnRateValuationDateRelativeTo",
        ("FIX.5.0", "42511"): "42511 LegReturnRateValuationDateOffsetPeriod",
        ("FIX.5.0", "42512"): "42512 LegReturnRateValuationDateOffsetUnit",
        ("FIX.5.0", "42513"): "42513 LegReturnRateValuationDateOffsetDayType",
        ("FIX.5.0", "42514"): "42514 LegReturnRateValuationStartDateUnadjusted",
        ("FIX.5.0", "42515"): "42515 LegReturnRateValuationStartDateRelativeTo",
        ("FIX.5.0", "42516"): "42516 LegReturnRateValuationStartDateOffsetPeriod",
        ("FIX.5.0", "42517"): "42517 LegReturnRateValuationStartDateOffsetUnit",
        ("FIX.5.0", "42518"): "42518 LegReturnRateValuationStartDateOffsetDayType",
        ("FIX.5.0", "42519"): "42519 LegReturnRateValuationStartDateAdjusted",
        ("FIX.5.0", "42520"): "42520 LegReturnRateValuationEndDateUnadjusted",
        ("FIX.5.0", "42521"): "42521 LegReturnRateValuationEndDateRelativeTo",
        ("FIX.5.0", "42522"): "42522 LegReturnRateValuationEndDateOffsetPeriod",
        ("FIX.5.0", "42523"): "42523 LegReturnRateValuationEndDateOffsetUnit",
        ("FIX.5.0", "42524"): "42524 LegReturnRateValuationEndDateOffsetDayType",
        ("FIX.5.0", "42525"): "42525 LegReturnRateValuationEndDateAdjusted",
        ("FIX.5.0", "42526"): "42526 LegReturnRateValuationFrequencyPeriod",
        ("FIX.5.0", "42527"): "42527 LegReturnRateValuationFrequencyUnit",
        ("FIX.5.0", "42528"): "42528 LegReturnRateValuationFrequencyRollConvention",
        ("FIX.5.0", "42529"): "42529 LegReturnRateValuationDateBusinessDayConvention",
        ("FIX.5.0", "42530"): "42530 NoLegReturnRateFXConversions",
        ("FIX.5.0", "42531"): "42531 LegReturnRateFXCurrencySymbol",
        ("FIX.5.0", "42532"): "42532 LegReturnRateFXRate",
        ("FIX.5.0", "42533"): "42533 LegReturnRateFXRateCalc",
        ("FIX.5.0", "42534"): "42534 NoLegReturnRates",
        ("FIX.5.0", "42535"): "42535 LegReturnRatePriceSequence",
        ("FIX.5.0", "42536"): "42536 LegReturnRateCommissionBasis",
        ("FIX.5.0", "42537"): "42537 LegReturnRateCommissionAmount",
        ("FIX.5.0", "42538"): "42538 LegReturnRateCommissionCurrency",
        ("FIX.5.0", "42539"): "42539 LegReturnRateTotalCommissionPerTrade",
        ("FIX.5.0", "42540"): "42540 LegReturnRateDeterminationMethod",
        ("FIX.5.0", "42541"): "42541 LegReturnRateAmountRelativeTo",
        ("FIX.5.0", "42542"): "42542 LegReturnRateQuoteMeasureType",
        ("FIX.5.0", "42543"): "42543 LegReturnRateQuoteUnits",
        ("FIX.5.0", "42544"): "42544 LegReturnRateQuoteMethod",
        ("FIX.5.0", "42545"): "42545 LegReturnRateQuoteCurrency",
        ("FIX.5.0", "42546"): "42546 LegReturnRateQuoteCurrencyType",
        ("FIX.5.0", "42547"): "42547 LegReturnRateQuoteTimeType",
        ("FIX.5.0", "42548"): "42548 LegReturnRateQuoteTime",
        ("FIX.5.0", "42549"): "42549 LegReturnRateQuoteDate",
        ("FIX.5.0", "42550"): "42550 LegReturnRateQuoteExpirationTime",
        ("FIX.5.0", "42551"): "42551 LegReturnRateQuoteBusinessCenter",
        ("FIX.5.0", "42552"): "42552 LegReturnRateQuoteExchange",
        ("FIX.5.0", "42553"): "42553 LegReturnRateQuotePricingModel",
        ("FIX.5.0", "42554"): "42554 LegReturnRateCashFlowType",
        ("FIX.5.0", "42555"): "42555 LegReturnRateValuationTimeType",
        ("FIX.5.0", "42556"): "42556 LegReturnRateValuationTime",
        ("FIX.5.0", "42557"): "42557 LegReturnRateValuationTimeBusinessCenter",
        ("FIX.5.0", "42558"): "42558 LegReturnRateValuationPriceOption",
        ("FIX.5.0", "42559"): "42559 LegReturnRateFinalPriceFallback",
        ("FIX.5.0", "42560"): "42560 NoLegReturnRateInformationSources",
        ("FIX.5.0", "42561"): "42561 LegReturnRateInformationSource",
        ("FIX.5.0", "42562"): "42562 LegReturnRateReferencePage",
        ("FIX.5.0", "42563"): "42563 LegReturnRateReferencePageHeading",
        ("FIX.5.0", "42564"): "42564 NoLegReturnRatePrices",
        ("FIX.5.0", "42565"): "42565 LegReturnRatePriceBasis",
        ("FIX.5.0", "42566"): "42566 LegReturnRatePrice",
        ("FIX.5.0", "42567"): "42567 LegReturnRatePriceCurrency",
        ("FIX.5.0", "42568"): "42568 LegReturnRatePriceType",
        ("FIX.5.0", "42569"): "42569 NoLegReturnRateValuationDateBusinessCenters",
        ("FIX.5.0", "42570"): "42570 LegReturnRateValuationDateBusinessCenter",
        ("FIX.5.0", "42571"): "42571 NoLegReturnRateValuationDates",
        ("FIX.5.0", "42572"): "42572 LegReturnRateValuationDate",
        ("FIX.5.0", "42573"): "42573 LegReturnRateValuationDateType",
        ("FIX.5.0", "42574"): "42574 LegSettlMethodElectionDateUnadjusted",
        ("FIX.5.0", "42575"): "42575 LegSettlMethodElectionDateBusinessDayConvention",
        ("FIX.5.0", "42576"): "42576 LegSettlMethodElectionDateRelativeTo",
        ("FIX.5.0", "42577"): "42577 LegSettlMethodElectionDateOffsetPeriod",
        ("FIX.5.0", "42578"): "42578 LegSettlMethodElectionDateOffsetUnit",
        ("FIX.5.0", "42579"): "42579 LegSettlMethodElectionDateOffsetDayType",
        ("FIX.5.0", "42580"): "42580 LegSettlMethodElectionDateAdjusted",
        ("FIX.5.0", "42581"): "42581 NoLegSettlMethodElectionDateBusinessCenters",
        ("FIX.5.0", "42582"): "42582 LegSettlMethodElectionDateBusinessCenter",
        ("FIX.5.0", "42583"): "42583 LegStreamVersion",
        ("FIX.5.0", "42584"): "42584 LegStreamVersionEffectiveDate",
        ("FIX.5.0", "42585"): "42585 LegStreamNotionalDeterminationMethod",
        ("FIX.5.0", "42586"): "42586 LegStreamNotionalAdjustments",
        ("FIX.5.0", "42587"): "42587 StreamCommodityDeliveryPricingRegion",
        ("FIX.5.0", "42588"): "42588 LegStreamCommodityDeliveryPricingRegion",
        ("FIX.5.0", "42589"): "42589 UnderlyingStreamCommodityDeliveryPricingRegion",
        ("FIX.5.0", "42590"): "42590 SettlMethodElectingPartySide",
        ("FIX.5.0", "42591"): "42591 MakeWholeDate",
        ("FIX.5.0", "42592"): "42592 MakeWholeAmount",
        ("FIX.5.0", "42593"): "42593 MakeWholeBenchmarkCurveName",
        ("FIX.5.0", "42594"): "42594 MakeWholeBenchmarkCurvePoint",
        ("FIX.5.0", "42595"): "42595 MakeWholeRecallSpread",
        ("FIX.5.0", "42596"): "42596 MakeWholeBenchmarkQuote",
        ("FIX.5.0", "42597"): "42597 MakeWholeInterpolationMethod",
        ("FIX.5.0", "42598"): "42598 PaymentAmountRelativeTo",
        ("FIX.5.0", "42599"): "42599 PaymentAmountDeterminationMethod",
        ("FIX.5.0", "42600"): "42600 PaymentStreamCashSettlIndicator",
        ("FIX.5.0", "42601"): "42601 PaymentStreamCompoundingXIDRef",
        ("FIX.5.0", "42602"): "42602 PaymentStreamCompoundingSpread",
        ("FIX.5.0", "42603"): "42603 PaymentStreamInterpolationMethod",
        ("FIX.5.0", "42604"): "42604 PaymentStreamInterpolationPeriod",
        ("FIX.5.0", "42605"): "42605 PaymentStreamCompoundingFixedRate",
        ("FIX.5.0", "42606"): "42606 NoPaymentStreamCompoundingDates",
        ("FIX.5.0", "42607"): "42607 PaymentStreamCompoundingDate",
        ("FIX.5.0", "42608"): "42608 PaymentStreamCompoundingDateType",
        ("FIX.5.0", "42609"): "42609 PaymentStreamCompoundingDatesBusinessDayConvention",
        ("FIX.5.0", "42610"): "42610 PaymentStreamCompoundingDatesRelativeTo",
        ("FIX.5.0", "42611"): "42611 PaymentStreamCompoundingDatesOffsetPeriod",
        ("FIX.5.0", "42612"): "42612 PaymentStreamCompoundingDatesOffsetUnit",
        ("FIX.5.0", "42613"): "42613 PaymentStreamCompoundingDatesOffsetDayType",
        ("FIX.5.0", "42614"): "42614 PaymentStreamCompoundingPeriodSkip",
        ("FIX.5.0", "42615"): "42615 PaymentStreamCompoundingFrequencyPeriod",
        ("FIX.5.0", "42616"): "42616 PaymentStreamCompoundingFrequencyUnit",
        ("FIX.5.0", "42617"): "42617 PaymentStreamCompoundingRollConvention",
        ("FIX.5.0", "42618"): "42618 PaymentStreamBoundsFirstDateUnadjusted",
        ("FIX.5.0", "42619"): "42619 PaymentStreamBoundsLastDateUnadjusted",
        ("FIX.5.0", "42620"): "42620 NoPaymentStreamCompoundingDatesBusinessCenters",
        ("FIX.5.0", "42621"): "42621 PaymentStreamCompoundingDatesBusinessCenter",
        ("FIX.5.0", "42622"): "42622 PaymentStreamCompoundingEndDateUnadjusted",
        ("FIX.5.0", "42623"): "42623 PaymentStreamCompoundingEndDateRelativeTo",
        ("FIX.5.0", "42624"): "42624 PaymentStreamCompoundingEndDateOffsetPeriod",
        ("FIX.5.0", "42625"): "42625 PaymentStreamCompoundingEndDateOffsetUnit",
        ("FIX.5.0", "42626"): "42626 PaymentStreamCompoundingEndDateOffsetDayType",
        ("FIX.5.0", "42627"): "42627 PaymentStreamCompoundingEndDateAdjusted",
        ("FIX.5.0", "42628"): "42628 PaymentStreamCompoundingRateIndex",
        ("FIX.5.0", "42629"): "42629 PaymentStreamCompoundingRateIndexCurvePeriod",
        ("FIX.5.0", "42630"): "42630 PaymentStreamCompoundingRateIndexCurveUnit",
        ("FIX.5.0", "42631"): "42631 PaymentStreamCompoundingRateMultiplier",
        ("FIX.5.0", "42632"): "42632 PaymentStreamCompoundingRateSpread",
        ("FIX.5.0", "42633"): "42633 PaymentStreamCompoundingRateSpreadPositionType",
        ("FIX.5.0", "42634"): "42634 PaymentStreamCompoundingRateTreatment",
        ("FIX.5.0", "42635"): "42635 PaymentStreamCompoundingCapRate",
        ("FIX.5.0", "42636"): "42636 PaymentStreamCompoundingCapRateBuySide",
        ("FIX.5.0", "42637"): "42637 PaymentStreamCompoundingCapRateSellSide",
        ("FIX.5.0", "42638"): "42638 PaymentStreamCompoundingFloorRate",
        ("FIX.5.0", "42639"): "42639 PaymentStreamCompoundingFloorRateBuySide",
        ("FIX.5.0", "42640"): "42640 PaymentStreamCompoundingFloorRateSellSide",
        ("FIX.5.0", "42641"): "42641 PaymentStreamCompoundingInitialRate",
        ("FIX.5.0", "42642"): "42642 PaymentStreamCompoundingFinalRateRoundingDirection",
        ("FIX.5.0", "42643"): "42643 PaymentStreamCompoundingFinalRatePrecision",
        ("FIX.5.0", "42644"): "42644 PaymentStreamCompoundingAveragingMethod",
        ("FIX.5.0", "42645"): "42645 PaymentStreamCompoundingNegativeRateTreatment",
        ("FIX.5.0", "42646"): "42646 PaymentStreamCompoundingStartDateUnadjusted",
        ("FIX.5.0", "42647"): "42647 PaymentStreamCompoundingStartDateRelativeTo",
        ("FIX.5.0", "42648"): "42648 PaymentStreamCompoundingStartDateOffsetPeriod",
        ("FIX.5.0", "42649"): "42649 PaymentStreamCompoundingStartDateOffsetUnit",
        ("FIX.5.0", "42650"): "42650 PaymentStreamCompoundingStartDateOffsetDayType",
        ("FIX.5.0", "42651"): "42651 PaymentStreamCompoundingStartDateAdjusted",
        ("FIX.5.0", "42652"): "42652 PaymentStreamFormulaImageLength",
        ("FIX.5.0", "42653"): "42653 PaymentStreamFormulaImage",
        ("FIX.5.0", "42654"): "42654 PaymentStreamFinalPricePaymentDateUnadjusted",
        ("FIX.5.0", "42655"): "42655 PaymentStreamFinalPricePaymentDateRelativeTo",
        ("FIX.5.0", "42656"): "42656 PaymentStreamFinalPricePaymentDateOffsetfPeriod",
        ("FIX.5.0", "42657"): "42657 PaymentStreamFinalPricePaymentDateOffsetUnit",
        ("FIX.5.0", "42658"): "42658 PaymentStreamFinalPricePaymentDateOffsetDayType",
        ("FIX.5.0", "42659"): "42659 PaymentStreamFinalPricePaymentDateAdjusted",
        ("FIX.5.0", "42660"): "42660 NoPaymentStreamFixingDates",
        ("FIX.5.0", "42661"): "42661 PaymentStreamFixingDate",
        ("FIX.5.0", "42662"): "42662 PaymentStreamFixingDateType",
        ("FIX.5.0", "42663"): "42663 PaymentStreamFirstObservationDateUnadjusted",
        ("FIX.5.0", "42664"): "42664 PaymentStreamFirstObservationDateRelativeTo",
        ("FIX.5.0", "42665"): "42665 PaymentStreamFirstObservationDateOffsetDayType",
        ("FIX.5.0", "42666"): "42666 PaymentStreamFirstObservationDateAdjusted",
        ("FIX.5.0", "42667"): "42667 PaymentStreamUnderlierRefID",
        ("FIX.5.0", "42668"): "42668 ReturnRateNotionalReset",
        ("FIX.5.0", "42669"): "42669 PaymentStreamLinkInitialLevel",
        ("FIX.5.0", "42670"): "42670 PaymentStreamLinkClosingLevelIndicator",
        ("FIX.5.0", "42671"): "42671 PaymentStreamLinkExpiringLevelIndicator",
        ("FIX.5.0", "42672"): "42672 PaymentStreamLinkEstimatedTradingDays",
        ("FIX.5.0", "42673"): "42673 PaymentStreamLinkStrikePrice",
        ("FIX.5.0", "42674"): "42674 PaymentStreamLinkStrikePriceType",
        ("FIX.5.0", "42675"): "42675 PaymentStreamLinkMaximumBoundary",
        ("FIX.5.0", "42676"): "42676 PaymentStreamLinkMinimumBoundary",
        ("FIX.5.0", "42677"): "42677 PaymentStreamLinkNumberOfDataSeries",
        ("FIX.5.0", "42678"): "42678 PaymentStreamVarianceUnadjustedCap",
        ("FIX.5.0", "42679"): "42679 PaymentStreamRealizedVarianceMethod",
        ("FIX.5.0", "42680"): "42680 PaymentStreamDaysAdjustmentIndicator",
        ("FIX.5.0", "42681"): "42681 PaymentStreamNearestExchangeContractRefID",
        ("FIX.5.0", "42682"): "42682 PaymentStreamVegaNotionalAmount",
        ("FIX.5.0", "42683"): "42683 NoPaymentStreamFormulas",
        ("FIX.5.0", "42684"): "42684 PaymentStreamFormula",
        ("FIX.5.0", "42685"): "42685 PaymentStreamFormulaDesc",
        ("FIX.5.0", "42686"): "42686 PaymentStreamFormulaCurrency",
        ("FIX.5.0", "42687"): "42687 PaymentStreamFormulaCurrencyDeterminationMethod",
        ("FIX.5.0", "42688"): "42688 PaymentStreamFormulaReferenceAmount",
        ("FIX.5.0", "42689"): "42689 PaymentStubEndDateUnadjusted",
        ("FIX.5.0", "42690"): "42690 PaymentStubEndDateBusinessDayConvention",
        ("FIX.5.0", "42691"): "42691 PaymentStubEndDateRelativeTo",
        ("FIX.5.0", "42692"): "42692 PaymentStubEndDateOffsetPeriod",
        ("FIX.5.0", "42693"): "42693 PaymentStubEndDateOffsetUnit",
        ("FIX.5.0", "42694"): "42694 PaymentStubEndDateOffsetDayType",
        ("FIX.5.0", "42695"): "42695 PaymentStubEndDateAdjusted",
        ("FIX.5.0", "42696"): "42696 NoPaymentStubEndDateBusinessCenters",
        ("FIX.5.0", "42697"): "42697 PaymentStubEndDateBusinessCenter",
        ("FIX.5.0", "42698"): "42698 PaymentStubStartDateUnadjusted",
        ("FIX.5.0", "42699"): "42699 PaymentStubStartDateBusinessDayConvention",
        ("FIX.5.0", "42700"): "42700 PaymentStubStartDateRelativeTo",
        ("FIX.5.0", "42701"): "42701 PaymentStubStartDateOffsetPeriod",
        ("FIX.5.0", "42702"): "42702 PaymentStubStartDateOffsetUnit",
        ("FIX.5.0", "42703"): "42703 PaymentStubStartDateOffsetDayType",
        ("FIX.5.0", "42704"): "42704 PaymentStubStartDateAdjusted",
        ("FIX.5.0", "42705"): "42705 NoPaymentStubStartDateBusinessCenters",
        ("FIX.5.0", "42706"): "42706 PaymentStubStartDateBusinessCenter",
        ("FIX.5.0", "42707"): "42707 ProvisionBreakFeeElection",
        ("FIX.5.0", "42708"): "42708 ProvisionBreakFeeRate",
        ("FIX.5.0", "42709"): "42709 NoReturnRateDates",
        ("FIX.5.0", "42710"): "42710 ReturnRateDateMode",
        ("FIX.5.0", "42711"): "42711 ReturnRateValuationDateRelativeTo",
        ("FIX.5.0", "42712"): "42712 ReturnRateValuationDateOffsetPeriod",
        ("FIX.5.0", "42713"): "42713 ReturnRateValuationDateOffsetUnit",
        ("FIX.5.0", "42714"): "42714 ReturnRateValuationDateOffsetDayType",
        ("FIX.5.0", "42715"): "42715 ReturnRateValuationStartDateUnadjusted",
        ("FIX.5.0", "42716"): "42716 ReturnRateValuationStartDateRelativeTo",
        ("FIX.5.0", "42717"): "42717 ReturnRateValuationStartDateOffsetPeriod",
        ("FIX.5.0", "42718"): "42718 ReturnRateValuationStartDateOffsetUnit",
        ("FIX.5.0", "42719"): "42719 ReturnRateValuationStartDateOffsetDayType",
        ("FIX.5.0", "42720"): "42720 ReturnRateValuationStartDateAdjusted",
        ("FIX.5.0", "42721"): "42721 ReturnRateValuationEndDateUnadjusted",
        ("FIX.5.0", "42722"): "42722 ReturnRateValuationEndDateRelativeTo",
        ("FIX.5.0", "42723"): "42723 ReturnRateValuationEndDateOffsetPeriod",
        ("FIX.5.0", "42724"): "42724 ReturnRateValuationEndDateOffsetUnit",
        ("FIX.5.0", "42725"): "42725 ReturnRateValuationEndDateOffsetDayType",
        ("FIX.5.0", "42726"): "42726 ReturnRateValuationEndDateAdjusted",
        ("FIX.5.0", "42727"): "42727 ReturnRateValuationFrequencyPeriod",
        ("FIX.5.0", "42728"): "42728 ReturnRateValuationFrequencyUnit",
        ("FIX.5.0", "42729"): "42729 ReturnRateValuationFrequencyRollConvention",
        ("FIX.5.0", "42730"): "42730 ReturnRateValuationDateBusinessDayConvention",
        ("FIX.5.0", "42731"): "42731 NoReturnRateFXConversions",
        ("FIX.5.0", "42732"): "42732 ReturnRateFXCurrencySymbol",
        ("FIX.5.0", "42733"): "42733 ReturnRateFXRate",
        ("FIX.5.0", "42734"): "42734 ReturnRateFXRateCalc",
        ("FIX.5.0", "42735"): "42735 NoReturnRates",
        ("FIX.5.0", "42736"): "42736 ReturnRatePriceSequence",
        ("FIX.5.0", "42737"): "42737 ReturnRateCommissionBasis",
        ("FIX.5.0", "42738"): "42738 ReturnRateCommissionAmount",
        ("FIX.5.0", "42739"): "42739 ReturnRateCommissionCurrency",
        ("FIX.5.0", "42740"): "42740 ReturnRateTotalCommissionPerTrade",
        ("FIX.5.0", "42741"): "42741 ReturnRateDeterminationMethod",
        ("FIX.5.0", "42742"): "42742 ReturnRateAmountRelativeTo",
        ("FIX.5.0", "42743"): "42743 ReturnRateQuoteMeasureType",
        ("FIX.5.0", "42744"): "42744 ReturnRateQuoteUnits",
        ("FIX.5.0", "42745"): "42745 ReturnRateQuoteMethod",
        ("FIX.5.0", "42746"): "42746 ReturnRateQuoteCurrency",
        ("FIX.5.0", "42747"): "42747 ReturnRateQuoteCurrencyType",
        ("FIX.5.0", "42748"): "42748 ReturnRateQuoteTimeType",
        ("FIX.5.0", "42749"): "42749 ReturnRateQuoteTime",
        ("FIX.5.0", "42750"): "42750 ReturnRateQuoteDate",
        ("FIX.5.0", "42751"): "42751 ReturnRateQuoteExpirationTime",
        ("FIX.5.0", "42752"): "42752 ReturnRateQuoteBusinessCenter",
        ("FIX.5.0", "42753"): "42753 ReturnRateQuoteExchange",
        ("FIX.5.0", "42754"): "42754 ReturnRateQuotePricingModel",
        ("FIX.5.0", "42755"): "42755 ReturnRateCashFlowType",
        ("FIX.5.0", "42756"): "42756 ReturnRateValuationTimeType",
        ("FIX.5.0", "42757"): "42757 ReturnRateValuationTime",
        ("FIX.5.0", "42758"): "42758 ReturnRateValuationTimeBusinessCenter",
        ("FIX.5.0", "42759"): "42759 ReturnRateValuationPriceOption",
        ("FIX.5.0", "42760"): "42760 ReturnRateFinalPriceFallback",
        ("FIX.5.0", "42761"): "42761 NoReturnRateInformationSources",
        ("FIX.5.0", "42762"): "42762 ReturnRateInformationSource",
        ("FIX.5.0", "42763"): "42763 ReturnRateReferencePage",
        ("FIX.5.0", "42764"): "42764 ReturnRateReferencePageHeading",
        ("FIX.5.0", "42765"): "42765 NoReturnRatePrices",
        ("FIX.5.0", "42766"): "42766 ReturnRatePriceBasis",
        ("FIX.5.0", "42767"): "42767 ReturnRatePrice",
        ("FIX.5.0", "42768"): "42768 ReturnRatePriceCurrency",
        ("FIX.5.0", "42769"): "42769 ReturnRatePriceType",
        ("FIX.5.0", "42770"): "42770 NoReturnRateValuationDateBusinessCenters",
        ("FIX.5.0", "42771"): "42771 ReturnRateValuationDateBusinessCenter",
        ("FIX.5.0", "42772"): "42772 NoReturnRateValuationDates",
        ("FIX.5.0", "42773"): "42773 ReturnRateValuationDate",
        ("FIX.5.0", "42774"): "42774 ReturnRateValuationDateType",
        ("FIX.5.0", "42775"): "42775 NoSettlMethodElectionDateBusinessCenters",
        ("FIX.5.0", "42776"): "42776 SettlMethodElectionDateBusinessCenter",
        ("FIX.5.0", "42777"): "42777 SettlMethodElectionDateUnadjusted",
        ("FIX.5.0", "42778"): "42778 SettlMethodElectionDateBusinessDayConvention",
        ("FIX.5.0", "42779"): "42779 SettlMethodElectionDateRelativeTo",
        ("FIX.5.0", "42780"): "42780 SettlMethodElectionDateOffsetPeriod",
        ("FIX.5.0", "42781"): "42781 SettlMethodElectionDateOffsetUnit",
        ("FIX.5.0", "42782"): "42782 SettlMethodElectionDateOffsetDayType",
        ("FIX.5.0", "42783"): "42783 SettlMethodElectionDateAdjusted",
        ("FIX.5.0", "42784"): "42784 StreamVersion",
        ("FIX.5.0", "42785"): "42785 StreamVersionEffectiveDate",
        ("FIX.5.0", "42786"): "42786 StreamNotionalDeterminationMethod",
        ("FIX.5.0", "42787"): "42787 StreamNotionalAdjustments",
        ("FIX.5.0", "42788"): "42788 NoUnderlyingCashSettlDateBusinessCenters",
        ("FIX.5.0", "42789"): "42789 UnderlyingCashSettlDateBusinessCenter",
        ("FIX.5.0", "42790"): "42790 UnderlyingCashSettlDateUnadjusted",
        ("FIX.5.0", "42791"): "42791 UnderlyingCashSettlDateBusinessDayConvention",
        ("FIX.5.0", "42792"): "42792 UnderlyingCashSettlDateRelativeTo",
        ("FIX.5.0", "42793"): "42793 UnderlyingCashSettlDateOffsetPeriod",
        ("FIX.5.0", "42794"): "42794 UnderlyingCashSettlDateOffsetUnit",
        ("FIX.5.0", "42795"): "42795 UnderlyingCashSettlDateOffsetDayType",
        ("FIX.5.0", "42796"): "42796 UnderlyingCashSettlDateAdjusted",
        ("FIX.5.0", "42797"): "42797 UnderlyingCashSettlPriceSource",
        ("FIX.5.0", "42798"): "42798 UnderlyingCashSettlPriceDefault",
        ("FIX.5.0", "42799"): "42799 NoUnderlyingDividendAccrualPaymentDateBusinessCenters",
        ("FIX.5.0", "42800"): "42800 UnderlyingDividendAccrualPaymentDateBusinessCenter",
        ("FIX.5.0", "42801"): "42801 UnderlyingDividendFloatingRateIndex",
        ("FIX.5.0", "42802"): "42802 UnderlyingDividendFloatingRateIndexCurvePeriod",
        ("FIX.5.0", "42803"): "42803 UnderlyingDividendFloatingRateIndexCurveUnit",
        ("FIX.5.0", "42804"): "42804 UnderlyingDividendFloatingRateMultiplier",
        ("FIX.5.0", "42805"): "42805 UnderlyingDividendFloatingRateSpread",
        ("FIX.5.0", "42806"): "42806 UnderlyingDividendFloatingRateSpreadPositionType",
        ("FIX.5.0", "42807"): "42807 UnderlyingDividendFloatingRateTreatment",
        ("FIX.5.0", "42808"): "42808 UnderlyingDividendCapRate",
        ("FIX.5.0", "42809"): "42809 UnderlyingDividendCapRateBuySide",
        ("FIX.5.0", "42810"): "42810 UnderlyingDividendCapRateSellSide",
        ("FIX.5.0", "42811"): "42811 UnderlyingDividendFloorRate",
        ("FIX.5.0", "42812"): "42812 UnderlyingDividendFloorRateBuySide",
        ("FIX.5.0", "42813"): "42813 UnderlyingDividendFloorRateSellSide",
        ("FIX.5.0", "42814"): "42814 UnderlyingDividendInitialRate",
        ("FIX.5.0", "42815"): "42815 UnderlyingDividendFinalRateRoundingDirection",
        ("FIX.5.0", "42816"): "42816 UnderlyingDividendFinalRatePrecision",
        ("FIX.5.0", "42817"): "42817 UnderlyingDividendAveragingMethod",
        ("FIX.5.0", "42818"): "42818 UnderlyingDividendNegativeRateTreatment",
        ("FIX.5.0", "42819"): "42819 UnderlyingDividendAccrualPaymentDateRelativeTo",
        ("FIX.5.0", "42820"): "42820 UnderlyingDividendAccrualPaymentDateOffsetPeriod",
        ("FIX.5.0", "42821"): "42821 UnderlyingDividendAccrualPaymentDateOffsetUnit",
        ("FIX.5.0", "42822"): "42822 UnderlyingDividendAccrualPaymentDateOffsetDayType",
        ("FIX.5.0", "42823"): "42823 UnderlyingDividendAccrualPaymentDateUnadjusted",
        ("FIX.5.0", "42824"): "42824 UnderlyingDividendAccrualPaymentDateBusinessDayConvention",
        ("FIX.5.0", "42825"): "42825 UnderlyingDividendAccrualPaymentDateAdjusted",
        ("FIX.5.0", "42826"): "42826 UnderlyingDividendReinvestmentIndicator",
        ("FIX.5.0", "42827"): "42827 UnderlyingDividendEntitlementEvent",
        ("FIX.5.0", "42828"): "42828 UnderlyingDividendAmountType",
        ("FIX.5.0", "42829"): "42829 UnderlyingDividendUnderlierRefID",
        ("FIX.5.0", "42830"): "42830 UnderlyingExtraordinaryDividendPartySide",
        ("FIX.5.0", "42831"): "42831 UnderlyingExtraordinaryDividendAmountType",
        ("FIX.5.0", "42832"): "42832 UnderlyingExtraordinaryDividendCurrency",
        ("FIX.5.0", "42833"): "42833 UnderlyingExtraordinaryDividendDeterminationMethod",
        ("FIX.5.0", "42834"): "42834 UnderlyingDividendAccrualFixedRate",
        ("FIX.5.0", "42835"): "42835 UnderlyingDividendCompoundingMethod",
        ("FIX.5.0", "42836"): "42836 UnderlyingDividendNumOfIndexUnits",
        ("FIX.5.0", "42837"): "42837 UnderlyingDividendCashPercentage",
        ("FIX.5.0", "42838"): "42838 UnderlyingDividendCashEquivalentPercentage",
        ("FIX.5.0", "42839"): "42839 UnderlyingNonCashDividendTreatment",
        ("FIX.5.0", "42840"): "42840 UnderlyingDividendComposition",
        ("FIX.5.0", "42841"): "42841 UnderlyingSpecialDividendsIndicator",
        ("FIX.5.0", "42842"): "42842 UnderlyingMaterialDividendsIndicator",
        ("FIX.5.0", "42843"): "42843 UnderlyingOptionsExchangeDividendsIndicator",
        ("FIX.5.0", "42844"): "42844 UnderlyingAdditionalDividendsIndicator",
        ("FIX.5.0", "42845"): "42845 UnderlyingAllDividendsIndicator",
        ("FIX.5.0", "42846"): "42846 UnderlyingDividendFXTriggerDateRelativeTo",
        ("FIX.5.0", "42847"): "42847 UnderlyingDividendFXTriggerDateOffsetPeriod",
        ("FIX.5.0", "42848"): "42848 UnderlyingDividendFXTriggerDateOffsetUnit",
        ("FIX.5.0", "42849"): "42849 UnderlyingDividendFXTriggerDateOffsetDayType",
        ("FIX.5.0", "42850"): "42850 UnderlyingDividendFXTriggerDateUnadjusted",
        ("FIX.5.0", "42851"): "42851 UnderlyingDividendFXTriggerDateBusinessDayConvention",
        ("FIX.5.0", "42852"): "42852 UnderlyingDividendFXTriggerDateAdjusted",
        ("FIX.5.0", "42853"): "42853 NoUnderlyingDividendFXTriggerDateBusinessCenters",
        ("FIX.5.0", "42854"): "42854 UnderlyingDividendFXTriggerDateBusinessCenter",
        ("FIX.5.0", "42855"): "42855 NoUnderlyingDividendPayments",
        ("FIX.5.0", "42856"): "42856 UnderlyingDividendPaymentDate",
        ("FIX.5.0", "42857"): "42857 UnderlyingDividendPaymentAmount",
        ("FIX.5.0", "42858"): "42858 UnderlyingDividendPaymentCurrency",
        ("FIX.5.0", "42859"): "42859 UnderlyingDividendAccruedInterest",
        ("FIX.5.0", "42860"): "42860 UnderlyingDividendPayoutRatio",
        ("FIX.5.0", "42861"): "42861 UnderlyingDividendPayoutConditions",
        ("FIX.5.0", "42862"): "42862 NoUnderlyingDividendPeriods",
        ("FIX.5.0", "42863"): "42863 UnderlyingDividendPeriodSequence",
        ("FIX.5.0", "42864"): "42864 UnderlyingDividendPeriodStartDateUnadjusted",
        ("FIX.5.0", "42865"): "42865 UnderlyingDividendPeriodEndDateUnadjusted",
        ("FIX.5.0", "42866"): "42866 UnderlyingDividendPeriodUnderlierRefID",
        ("FIX.5.0", "42867"): "42867 UnderlyingDividendPeriodStrikePrice",
        ("FIX.5.0", "42868"): "42868 UnderlyingDividendPeriodBusinessDayConvention",
        ("FIX.5.0", "42869"): "42869 UnderlyingDividendPeriodValuationDateUnadjusted",
        ("FIX.5.0", "42870"): "42870 UnderlyingDividendPeriodValuationDateRelativeTo",
        ("FIX.5.0", "42871"): "42871 UnderlyingDividendPeriodValuationDateOffsetPeriod",
        ("FIX.5.0", "42872"): "42872 UnderlyingDividendPeriodValuationDateOffsetUnit",
        ("FIX.5.0", "42873"): "42873 UnderlyingDividendPeriodValuationDateOffsetDayType",
        ("FIX.5.0", "42874"): "42874 UnderlyingDividendPeriodValuationDateAdjusted",
        ("FIX.5.0", "42875"): "42875 UnderlyingDividendPeriodPaymentDateUnadjusted",
        ("FIX.5.0", "42876"): "42876 UnderlyingDividendPeriodPaymentDateRelativeTo",
        ("FIX.5.0", "42877"): "42877 UnderlyingDividendPeriodPaymentDateOffsetPeriod",
        ("FIX.5.0", "42878"): "42878 UnderlyingDividendPeriodPaymentDateOffsetUnit",
        ("FIX.5.0", "42879"): "42879 UnderlyingDividendPeriodPaymentDateOffsetDayType",
        ("FIX.5.0", "42880"): "42880 UnderlyingDividendPeriodPaymentDateAdjusted",
        ("FIX.5.0", "42881"): "42881 UnderlyingDividendPeriodXID",
        ("FIX.5.0", "42882"): "42882 NoUnderlyingDividendPeriodBusinessCenters",
        ("FIX.5.0", "42883"): "42883 UnderlyingDividendPeriodBusinessCenter",
        ("FIX.5.0", "42884"): "42884 NoUnderlyingExtraordinaryEvents",
        ("FIX.5.0", "42885"): "42885 UnderlyingExtraordinaryEventType",
        ("FIX.5.0", "42886"): "42886 UnderlyingExtraordinaryEventValue",
        ("FIX.5.0", "42887"): "42887 UnderlyingSettlMethodElectingPartySide",
        ("FIX.5.0", "42888"): "42888 UnderlyingMakeWholeDate",
        ("FIX.5.0", "42889"): "42889 UnderlyingMakeWholeAmount",
        ("FIX.5.0", "42890"): "42890 UnderlyingMakeWholeBenchmarkCurveName",
        ("FIX.5.0", "42891"): "42891 UnderlyingMakeWholeBenchmarkCurvePoint",
        ("FIX.5.0", "42892"): "42892 UnderlyingMakeWholeRecallSpread",
        ("FIX.5.0", "42893"): "42893 UnderlyingMakeWholeBenchmarkQuote",
        ("FIX.5.0", "42894"): "42894 UnderlyingMakeWholeInterpolationMethod",
        ("FIX.5.0", "42895"): "42895 UnderlyingPaymentStreamCashSettlIndicator",
        ("FIX.5.0", "42896"): "42896 UnderlyingPaymentStreamCompoundingXIDRef",
        ("FIX.5.0", "42897"): "42897 UnderlyingPaymentStreamCompoundingSpread",
        ("FIX.5.0", "42898"): "42898 UnderlyingPaymentStreamInterpolationMethod",
        ("FIX.5.0", "42899"): "42899 UnderlyingPaymentStreamInterpolationPeriod",
        ("FIX.5.0", "42900"): "42900 UnderlyingPaymentStreamCompoundingFixedRate",
        ("FIX.5.0", "42901"): "42901 NoUnderlyingPaymentStreamCompoundingDates",
        ("FIX.5.0", "42902"): "42902 UnderlyingPaymentStreamCompoundingDate",
        ("FIX.5.0", "42903"): "42903 UnderlyingPaymentStreamCompoundingDateType",
        ("FIX.5.0", "42904"): "42904 UnderlyingPaymentStreamCompoundingDatesBusinessDayConvention",
        ("FIX.5.0", "42905"): "42905 UnderlyingPaymentStreamCompoundingDatesRelativeTo",
        ("FIX.5.0", "42906"): "42906 UnderlyingPaymentStreamCompoundingDatesOffsetPeriod",
        ("FIX.5.0", "42907"): "42907 UnderlyingPaymentStreamCompoundingDatesOffsetUnit",
        ("FIX.5.0", "42908"): "42908 UnderlyingPaymentStreamCompoundingDatesOffsetDayType",
        ("FIX.5.0", "42909"): "42909 UnderlyingPaymentStreamCompoundingPeriodSkip",
        ("FIX.5.0", "42910"): "42910 UnderlyingPaymentStreamCompoundingFrequencyPeriod",
        ("FIX.5.0", "42911"): "42911 UnderlyingPaymentStreamCompoundingFrequencyUnit",
        ("FIX.5.0", "42912"): "42912 UnderlyingPaymentStreamCompoundingRollConvention",
        ("FIX.5.0", "42913"): "42913 UnderlyingPaymentStreamBoundsFirstDateUnadjusted",
        ("FIX.5.0", "42914"): "42914 UnderlyingPaymentStreamBoundsLastDateUnadjusted",
        ("FIX.5.0", "42915"): "42915 NoUnderlyingPaymentStreamCompoundingDatesBusinessCenters",
        ("FIX.5.0", "42916"): "42916 UnderlyingPaymentStreamCompoundingDatesBusinessCenter",
        ("FIX.5.0", "42917"): "42917 UnderlyingPaymentStreamCompoundingEndDateUnadjusted",
        ("FIX.5.0", "42918"): "42918 UnderlyingPaymentStreamCompoundingEndDateRelativeTo",
        ("FIX.5.0", "42919"): "42919 UnderlyingPaymentStreamCompoundingEndDateOffsetPeriod",
        ("FIX.5.0", "42920"): "42920 UnderlyingPaymentStreamCompoundingEndDateOffsetUnit",
        ("FIX.5.0", "42921"): "42921 UnderlyingPaymentStreamCompoundingEndDateOffsetDayType",
        ("FIX.5.0", "42922"): "42922 UnderlyingPaymentStreamCompoundingEndDateAdjusted",
        ("FIX.5.0", "42923"): "42923 UnderlyingPaymentStreamCompoundingRateIndex",
        ("FIX.5.0", "42924"): "42924 UnderlyingPaymentStreamCompoundingRateIndexCurvePeriod",
        ("FIX.5.0", "42925"): "42925 UnderlyingPaymentStreamCompoundingRateIndexCurveUnit",
        ("FIX.5.0", "42926"): "42926 UnderlyingPaymentStreamCompoundingRateMultiplier",
        ("FIX.5.0", "42927"): "42927 UnderlyingPaymentStreamCompoundingRateSpread",
        ("FIX.5.0", "42928"): "42928 UnderlyingPaymentStreamCompoundingRateSpreadPositionType",
        ("FIX.5.0", "42929"): "42929 UnderlyingPaymentStreamCompoundingRateTreatment",
        ("FIX.5.0", "42930"): "42930 UnderlyingPaymentStreamCompoundingCapRate",
        ("FIX.5.0", "42931"): "42931 UnderlyingPaymentStreamCompoundingCapRateBuySide",
        ("FIX.5.0", "42932"): "42932 UnderlyingPaymentStreamCompoundingCapRateSellSide",
        ("FIX.5.0", "42933"): "42933 UnderlyingPaymentStreamCompoundingFloorRate",
        ("FIX.5.0", "42934"): "42934 UnderlyingPaymentStreamCompoundingFloorRateBuySide",
        ("FIX.5.0", "42935"): "42935 UnderlyingPaymentStreamCompoundingFloorRateSellSide",
        ("FIX.5.0", "42936"): "42936 UnderlyingPaymentStreamCompoundingInitialRate",
        ("FIX.5.0", "42937"): "42937 UnderlyingPaymentStreamCompoundingFinalRateRoundingDirection",
        ("FIX.5.0", "42938"): "42938 UnderlyingPaymentStreamCompoundingFinalRatePrecision",
        ("FIX.5.0", "42939"): "42939 UnderlyingPaymentStreamCompoundingAveragingMethod",
        ("FIX.5.0", "42940"): "42940 UnderlyingPaymentStreamCompoundingNegativeRateTreatment",
        ("FIX.5.0", "42941"): "42941 UnderlyingPaymentStreamCompoundingStartDateUnadjusted",
        ("FIX.5.0", "42942"): "42942 UnderlyingPaymentStreamCompoundingStartDateRelativeTo",
        ("FIX.5.0", "42943"): "42943 UnderlyingPaymentStreamCompoundingStartDateOffsetPeriod",
        ("FIX.5.0", "42944"): "42944 UnderlyingPaymentStreamCompoundingStartDateOffsetUnit",
        ("FIX.5.0", "42945"): "42945 UnderlyingPaymentStreamCompoundingStartDateOffsetDayType",
        ("FIX.5.0", "42946"): "42946 UnderlyingPaymentStreamCompoundingStartDateAdjusted",
        ("FIX.5.0", "42947"): "42947 UnderlyingPaymentStreamFormulaImageLength",
        ("FIX.5.0", "42948"): "42948 UnderlyingPaymentStreamFormulaImage",
        ("FIX.5.0", "42949"): "42949 UnderlyingPaymentStreamFinalPricePaymentDateUnadjusted",
        ("FIX.5.0", "42950"): "42950 UnderlyingPaymentStreamFinalPricePaymentDateRelativeTo",
        ("FIX.5.0", "42951"): "42951 UnderlyingPaymentStreamFinalPricePaymentDateOffsetPeriod",
        ("FIX.5.0", "42952"): "42952 UnderlyingPaymentStreamFinalPricePaymentDateOffsetUnit",
        ("FIX.5.0", "42953"): "42953 UnderlyingPaymentStreamFinalPricePaymentDateOffsetDayType",
        ("FIX.5.0", "42954"): "42954 UnderlyingPaymentStreamFinalPricePaymentDateAdjusted",
        ("FIX.5.0", "42955"): "42955 NoUnderlyingPaymentStreamFixingDates",
        ("FIX.5.0", "42956"): "42956 UnderlyingPaymentStreamFixingDate",
        ("FIX.5.0", "42957"): "42957 UnderlyingPaymentStreamFixingDateType",
        ("FIX.5.0", "42958"): "42958 UnderlyingPaymentStreamFirstObservationDateUnadjusted",
        ("FIX.5.0", "42959"): "42959 UnderlyingPaymentStreamFirstObservationDateRelativeTo",
        ("FIX.5.0", "42960"): "42960 UnderlyingPaymentStreamFirstObservationDateOffsetDayType",
        ("FIX.5.0", "42961"): "42961 UnderlyingPaymentStreamFirstObservationDateAdjusted",
        ("FIX.5.0", "42962"): "42962 UnderlyingPaymentStreamUnderlierRefID",
        ("FIX.5.0", "42963"): "42963 UnderlyingReturnRateNotionalReset",
        ("FIX.5.0", "42964"): "42964 UnderlyingPaymentStreamLinkInitialLevel",
        ("FIX.5.0", "42965"): "42965 UnderlyingPaymentStreamLinkClosingLevelIndicator",
        ("FIX.5.0", "42966"): "42966 UnderlyingPaymentStreamLinkExpiringLevelIndicator",
        ("FIX.5.0", "42967"): "42967 UnderlyingPaymentStreamLinkEstimatedTradingDays",
        ("FIX.5.0", "42968"): "42968 UnderlyingPaymentStreamLinkStrikePrice",
        ("FIX.5.0", "42969"): "42969 UnderlyingPaymentStreamLinkStrikePriceType",
        ("FIX.5.0", "42970"): "42970 UnderlyingPaymentStreamLinkMaximumBoundary",
        ("FIX.5.0", "42971"): "42971 UnderlyingPaymentStreamLinkMinimumBoundary",
        ("FIX.5.0", "42972"): "42972 UnderlyingPaymentStreamLinkNumberOfDataSeries",
        ("FIX.5.0", "42973"): "42973 UnderlyingPaymentStreamVarianceUnadjustedCap",
        ("FIX.5.0", "42974"): "42974 UnderlyingPaymentStreamRealizedVarianceMethod",
        ("FIX.5.0", "42975"): "42975 UnderlyingPaymentStreamDaysAdjustmentIndicator",
        ("FIX.5.0", "42976"): "42976 UnderlyingPaymentStreamNearestExchangeContractRefID",
        ("FIX.5.0", "42977"): "42977 UnderlyingPaymentStreamVegaNotionalAmount",
        ("FIX.5.0", "42978"): "42978 UnderlyingPaymentStreamFormulaCurrency",
        ("FIX.5.0", "42979"): "42979 UnderlyingPaymentStreamFormulaCurrencyDeterminationMethod",
        ("FIX.5.0", "42980"): "42980 UnderlyingPaymentStreamFormulaReferenceAmount",
        ("FIX.5.0", "42981"): "42981 NoUnderlyingPaymentStreamFormulas",
        ("FIX.5.0", "42982"): "42982 UnderlyingPaymentStreamFormula",
        ("FIX.5.0", "42983"): "42983 UnderlyingPaymentStreamFormulaDesc",
        ("FIX.5.0", "42984"): "42984 UnderlyingPaymentStubEndDateUnadjusted",
        ("FIX.5.0", "42985"): "42985 UnderlyingPaymentStubEndDateBusinessDayConvention",
        ("FIX.5.0", "42986"): "42986 UnderlyingPaymentStubEndDateRelativeTo",
        ("FIX.5.0", "42987"): "42987 UnderlyingPaymentStubEndDateOffsetPeriod",
        ("FIX.5.0", "42988"): "42988 UnderlyingPaymentStubEndDateOffsetUnit",
        ("FIX.5.0", "42989"): "42989 UnderlyingPaymentStubEndDateOffsetDayType",
        ("FIX.5.0", "42990"): "42990 UnderlyingPaymentStubEndDateAdjusted",
        ("FIX.5.0", "42991"): "42991 NoUnderlyingPaymentStubEndDateBusinessCenters",
        ("FIX.5.0", "42992"): "42992 UnderlyingPaymentStubEndDateBusinessCenter",
        ("FIX.5.0", "42993"): "42993 UnderlyingPaymentStubStartDateUnadjusted",
        ("FIX.5.0", "42994"): "42994 UnderlyingPaymentStubStartDateBusinessDayConvention",
        ("FIX.5.0", "42995"): "42995 UnderlyingPaymentStubStartDateRelativeTo",
        ("FIX.5.0", "42996"): "42996 UnderlyingPaymentStubStartDateOffsetPeriod",
        ("FIX.5.0", "42997"): "42997 UnderlyingPaymentStubStartDateOffsetUnit",
        ("FIX.5.0", "42998"): "42998 UnderlyingPaymentStubStartDateOffsetDayType",
        ("FIX.5.0", "42999"): "42999 UnderlyingPaymentStubStartDateAdjusted",
        ("FIX.5.0", "43000"): "43000 NoUnderlyingPaymentStubStartDateBusinessCenters",
        ("FIX.5.0", "43001"): "43001 UnderlyingPaymentStubStartDateBusinessCenter",
        ("FIX.5.0", "43002"): "43002 UnderlyingProvisionBreakFeeElection",
        ("FIX.5.0", "43003"): "43003 UnderlyingProvisionBreakFeeRate",
        ("FIX.5.0", "43004"): "43004 UnderlyingRateSpreadInitialValue",
        ("FIX.5.0", "43005"): "43005 NoUnderlyingRateSpreadSteps",
        ("FIX.5.0", "43006"): "43006 UnderlyingRateSpreadStepDate",
        ("FIX.5.0", "43007"): "43007 UnderlyingRateSpreadStepValue",
        ("FIX.5.0", "43008"): "43008 NoUnderlyingReturnRateDates",
        ("FIX.5.0", "43009"): "43009 UnderlyingReturnRateDateMode",
        ("FIX.5.0", "43010"): "43010 UnderlyingReturnRateValuationDateRelativeTo",
        ("FIX.5.0", "43011"): "43011 UnderlyingReturnRateValuationDateOffsetPeriod",
        ("FIX.5.0", "43012"): "43012 UnderlyingReturnRateValuationDateOffsetUnit",
        ("FIX.5.0", "43013"): "43013 UnderlyingReturnRateValuationDateOffsetDayType",
        ("FIX.5.0", "43014"): "43014 UnderlyingReturnRateValuationStartDateUnadjusted",
        ("FIX.5.0", "43015"): "43015 UnderlyingReturnRateValuationStartDateRelativeTo",
        ("FIX.5.0", "43016"): "43016 UnderlyingReturnRateValuationStartDateOffsetPeriod",
        ("FIX.5.0", "43017"): "43017 UnderlyingReturnRateValuationStartDateOffsetUnit",
        ("FIX.5.0", "43018"): "43018 UnderlyingReturnRateValuationStartDateOffsetDayType",
        ("FIX.5.0", "43019"): "43019 UnderlyingReturnRateValuationStartDateAdjusted",
        ("FIX.5.0", "43020"): "43020 UnderlyingReturnRateValuationEndDateUnadjusted",
        ("FIX.5.0", "43021"): "43021 UnderlyingReturnRateValuationEndDateRelativeTo",
        ("FIX.5.0", "43022"): "43022 UnderlyingReturnRateValuationEndDateOffsetPeriod",
        ("FIX.5.0", "43023"): "43023 UnderlyingReturnRateValuationEndDateOffsetUnit",
        ("FIX.5.0", "43024"): "43024 UnderlyingReturnRateValuationEndDateOffsetDayType",
        ("FIX.5.0", "43025"): "43025 UnderlyingReturnRateValuationEndDateAdjusted",
        ("FIX.5.0", "43026"): "43026 UnderlyingReturnRateValuationFrequencyPeriod",
        ("FIX.5.0", "43027"): "43027 UnderlyingReturnRateValuationFrequencyUnit",
        ("FIX.5.0", "43028"): "43028 UnderlyingReturnRateValuationFrequencyRollConvention",
        ("FIX.5.0", "43029"): "43029 UnderlyingReturnRateValuationDateBusinessDayConvention",
        ("FIX.5.0", "43030"): "43030 NoUnderlyingReturnRateFXConversions",
        ("FIX.5.0", "43031"): "43031 UnderlyingReturnRateFXCurrencySymbol",
        ("FIX.5.0", "43032"): "43032 UnderlyingReturnRateFXRate",
        ("FIX.5.0", "43033"): "43033 UnderlyingReturnRateFXRateCalc",
        ("FIX.5.0", "43034"): "43034 NoUnderlyingReturnRates",
        ("FIX.5.0", "43035"): "43035 UnderlyingReturnRatePriceSequence",
        ("FIX.5.0", "43036"): "43036 UnderlyingReturnRateCommissionBasis",
        ("FIX.5.0", "43037"): "43037 UnderlyingReturnRateCommissionAmount",
        ("FIX.5.0", "43038"): "43038 UnderlyingReturnRateCommissionCurrency",
        ("FIX.5.0", "43039"): "43039 UnderlyingReturnRateTotalCommissionPerTrade",
        ("FIX.5.0", "43040"): "43040 UnderlyingReturnRateDeterminationMethod",
        ("FIX.5.0", "43041"): "43041 UnderlyingReturnRateAmountRelativeTo",
        ("FIX.5.0", "43042"): "43042 UnderlyingReturnRateQuoteMeasureType",
        ("FIX.5.0", "43043"): "43043 UnderlyingReturnRateQuoteUnits",
        ("FIX.5.0", "43044"): "43044 UnderlyingReturnRateQuoteMethod",
        ("FIX.5.0", "43045"): "43045 UnderlyingReturnRateQuoteCurrency",
        ("FIX.5.0", "43046"): "43046 UnderlyingReturnRateQuoteCurrencyType",
        ("FIX.5.0", "43047"): "43047 UnderlyingReturnRateQuoteTimeType",
        ("FIX.5.0", "43048"): "43048 UnderlyingReturnRateQuoteTime",
        ("FIX.5.0", "43049"): "43049 UnderlyingReturnRateQuoteDate",
        ("FIX.5.0", "43050"): "43050 UnderlyingReturnRateQuoteExpirationTime",
        ("FIX.5.0", "43051"): "43051 UnderlyingReturnRateQuoteBusinessCenter",
        ("FIX.5.0", "43052"): "43052 UnderlyingReturnRateQuoteExchange",
        ("FIX.5.0", "43053"): "43053 UnderlyingReturnRateQuotePricingModel",
        ("FIX.5.0", "43054"): "43054 UnderlyingReturnRateCashFlowType",
        ("FIX.5.0", "43055"): "43055 UnderlyingReturnRateValuationTimeType",
        ("FIX.5.0", "43056"): "43056 UnderlyingReturnRateValuationTime",
        ("FIX.5.0", "43057"): "43057 UnderlyingReturnRateValuationTimeBusinessCenter",
        ("FIX.5.0", "43058"): "43058 UnderlyingReturnRateValuationPriceOption",
        ("FIX.5.0", "43059"): "43059 UnderlyingReturnRateFinalPriceFallback",
        ("FIX.5.0", "43060"): "43060 NoUnderlyingReturnRateInformationSources",
        ("FIX.5.0", "43061"): "43061 UnderlyingReturnRateInformationSource",
        ("FIX.5.0", "43062"): "43062 UnderlyingReturnRateReferencePage",
        ("FIX.5.0", "43063"): "43063 UnderlyingReturnRateReferencePageHeading",
        ("FIX.5.0", "43064"): "43064 NoUnderlyingReturnRatePrices",
        ("FIX.5.0", "43065"): "43065 UnderlyingReturnRatePriceBasis",
        ("FIX.5.0", "43066"): "43066 UnderlyingReturnRatePrice",
        ("FIX.5.0", "43067"): "43067 UnderlyingReturnRatePriceCurrency",
        ("FIX.5.0", "43068"): "43068 UnderlyingReturnRatePriceType",
        ("FIX.5.0", "43069"): "43069 NoUnderlyingReturnRateValuationDateBusinessCenters",
        ("FIX.5.0", "43070"): "43070 UnderlyingReturnRateValuationDateBusinessCenter",
        ("FIX.5.0", "43071"): "43071 NoUnderlyingReturnRateValuationDates",
        ("FIX.5.0", "43072"): "43072 UnderlyingReturnRateValuationDate",
        ("FIX.5.0", "43073"): "43073 UnderlyingReturnRateValuationDateType",
        ("FIX.5.0", "43074"): "43074 NoUnderlyingSettlMethodElectionDateBusinessCenters",
        ("FIX.5.0", "43075"): "43075 UnderlyingSettlMethodElectionDateBusinessCenter",
        ("FIX.5.0", "43076"): "43076 UnderlyingSettlMethodElectionDateUnadjusted",
        ("FIX.5.0", "43077"): "43077 UnderlyingSettlMethodElectionDateBusinessDayConvention",
        ("FIX.5.0", "43078"): "43078 UnderlyingSettlMethodElectionDateRelativeTo",
        ("FIX.5.0", "43079"): "43079 UnderlyingSettlMethodElectionDateOffsetPeriod",
        ("FIX.5.0", "43080"): "43080 UnderlyingSettlMethodElectionDateOffsetUnit",
        ("FIX.5.0", "43081"): "43081 UnderlyingSettlMethodElectionDateOffsetDayType",
        ("FIX.5.0", "43082"): "43082 UnderlyingSettlMethodElectionDateAdjusted",
        ("FIX.5.0", "43083"): "43083 UnderlyingStreamVersion",
        ("FIX.5.0", "43084"): "43084 UnderlyingStreamVersionEffectiveDate",
        ("FIX.5.0", "43085"): "43085 UnderlyingStreamNotionalDeterminationMethod",
        ("FIX.5.0", "43086"): "43086 UnderlyingStreamNotionalAdjustments",
        ("FIX.5.0", "43087"): "43087 PaymentDesc",
        ("FIX.5.0", "43088"): "43088 LegPaymentStreamRateIndexID",
        ("FIX.5.0", "43089"): "43089 LegPaymentStreamRateIndexIDSource",
        ("FIX.5.0", "43090"): "43090 PaymentStreamRateIndexID",
        ("FIX.5.0", "43091"): "43091 PaymentStreamRateIndexIDSource",
        ("FIX.5.0", "43092"): "43092 UnderlyingPaymentStreamRateIndexID",
        ("FIX.5.0", "43093"): "43093 UnderlyingPaymentStreamRateIndexIDSource",
        ("FIX.5.0", "43094"): "43094 DeliveryStreamRouteOrCharter",
        ("FIX.5.0", "43095"): "43095 LegDeliveryStreamRouteOrCharter",
        ("FIX.5.0", "43096"): "43096 UnderlyingDeliveryStreamRouteOrCharter",
        ("FIX.5.0", "50000"): "50000 BatchID",
        ("FIX.5.0", "50001"): "50001 BatchTotalMessages",
        ("FIX.5.0", "50002"): "50002 BatchProcessMode",
    }

    __VALUE_MAP = {
        ("FIX.4.0", "4", "B"): "B .. Buy",
        ("FIX.4.0", "4", "S"): "S .. Sell",
        ("FIX.4.0", "4", "T"): "T .. Trade",
        ("FIX.4.0", "4", "X"): "X .. Cross",
        ("FIX.4.0", "5", "C"): "C .. Cancel",
        ("FIX.4.0", "5", "N"): "N .. New",
        ("FIX.4.0", "5", "R"): "R .. Replace",
        ("FIX.4.0", "13", "1"): "1 .. PerUnit",
        ("FIX.4.0", "13", "2"): "2 .. Percent",
        ("FIX.4.0", "13", "3"): "3 .. Absolute",
        ("FIX.4.0", "18", "0"): "0 .. StayOnOfferSide",
        ("FIX.4.0", "18", "1"): "1 .. NotHeld",
        ("FIX.4.0", "18", "2"): "2 .. Work",
        ("FIX.4.0", "18", "3"): "3 .. GoAlong",
        ("FIX.4.0", "18", "4"): "4 .. OverTheDay",
        ("FIX.4.0", "18", "5"): "5 .. Held",
        ("FIX.4.0", "18", "6"): "6 .. ParticipateDoNotInitiate",
        ("FIX.4.0", "18", "7"): "7 .. StrictScale",
        ("FIX.4.0", "18", "8"): "8 .. TryToScale",
        ("FIX.4.0", "18", "9"): "9 .. StayOnBidSide",
        ("FIX.4.0", "18", "A"): "A .. NoCross",
        ("FIX.4.0", "18", "B"): "B .. OKToCross",
        ("FIX.4.0", "18", "C"): "C .. CallFirst",
        ("FIX.4.0", "18", "D"): "D .. PercentOfVolume",
        ("FIX.4.0", "18", "E"): "E .. DoNotIncrease",
        ("FIX.4.0", "18", "F"): "F .. DoNotReduce",
        ("FIX.4.0", "18", "G"): "G .. AllOrNone",
        ("FIX.4.0", "18", "I"): "I .. InstitutionsOnly",
        ("FIX.4.0", "18", "L"): "L .. LastPeg",
        ("FIX.4.0", "18", "M"): "M .. MidPricePeg",
        ("FIX.4.0", "18", "N"): "N .. NonNegotiable",
        ("FIX.4.0", "18", "O"): "O .. OpeningPeg",
        ("FIX.4.0", "18", "P"): "P .. MarketPeg",
        ("FIX.4.0", "18", "R"): "R .. PrimaryPeg",
        ("FIX.4.0", "18", "S"): "S .. Suspend",
        ("FIX.4.0", "20", "0"): "0 .. New",
        ("FIX.4.0", "20", "1"): "1 .. Cancel",
        ("FIX.4.0", "20", "2"): "2 .. Correct",
        ("FIX.4.0", "20", "3"): "3 .. Status",
        ("FIX.4.0", "21", "1"): "1 .. AutomatedExecutionNoIntervention",
        ("FIX.4.0", "21", "2"): "2 .. AutomatedExecutionInterventionOK",
        ("FIX.4.0", "21", "3"): "3 .. ManualOrder",
        ("FIX.4.0", "22", "1"): "1 .. CUSIP",
        ("FIX.4.0", "22", "2"): "2 .. SEDOL",
        ("FIX.4.0", "22", "3"): "3 .. QUIK",
        ("FIX.4.0", "22", "4"): "4 .. ISINNumber",
        ("FIX.4.0", "22", "5"): "5 .. RICCode",
        ("FIX.4.0", "24", "A"): "A .. Autex",
        ("FIX.4.0", "24", "B"): "B .. Bridge",
        ("FIX.4.0", "25", "H"): "H .. High",
        ("FIX.4.0", "25", "L"): "L .. Low",
        ("FIX.4.0", "25", "M"): "M .. Medium",
        ("FIX.4.0", "27", "L"): "L .. Large",
        ("FIX.4.0", "27", "M"): "M .. Medium",
        ("FIX.4.0", "27", "S"): "S .. Small",
        ("FIX.4.0", "28", "C"): "C .. Cancel",
        ("FIX.4.0", "28", "N"): "N .. New",
        ("FIX.4.0", "28", "R"): "R .. Replace",
        ("FIX.4.0", "29", "1"): "1 .. Agent",
        ("FIX.4.0", "29", "2"): "2 .. CrossAsAgent",
        ("FIX.4.0", "29", "3"): "3 .. CrossAsPrincipal",
        ("FIX.4.0", "29", "4"): "4 .. Principal",
        ("FIX.4.0", "35", "0"): "0 .. Heartbeat",
        ("FIX.4.0", "35", "1"): "1 .. TestRequest",
        ("FIX.4.0", "35", "2"): "2 .. ResendRequest",
        ("FIX.4.0", "35", "3"): "3 .. Reject",
        ("FIX.4.0", "35", "4"): "4 .. SequenceReset",
        ("FIX.4.0", "35", "5"): "5 .. Logout",
        ("FIX.4.0", "35", "6"): "6 .. IOI",
        ("FIX.4.0", "35", "7"): "7 .. Advertisement",
        ("FIX.4.0", "35", "8"): "8 .. ExecutionReport",
        ("FIX.4.0", "35", "9"): "9 .. OrderCancelReject",
        ("FIX.4.0", "35", "A"): "A .. Logon",
        ("FIX.4.0", "35", "B"): "B .. News",
        ("FIX.4.0", "35", "C"): "C .. Email",
        ("FIX.4.0", "35", "D"): "D .. NewOrderSingle",
        ("FIX.4.0", "35", "E"): "E .. NewOrderList",
        ("FIX.4.0", "35", "F"): "F .. OrderCancelRequest",
        ("FIX.4.0", "35", "G"): "G .. OrderCancelReplaceRequest",
        ("FIX.4.0", "35", "H"): "H .. OrderStatusRequest",
        ("FIX.4.0", "35", "J"): "J .. AllocationInstruction",
        ("FIX.4.0", "35", "K"): "K .. ListCancelRequest",
        ("FIX.4.0", "35", "L"): "L .. ListExecute",
        ("FIX.4.0", "35", "M"): "M .. ListStatusRequest",
        ("FIX.4.0", "35", "N"): "N .. ListStatus",
        ("FIX.4.0", "35", "P"): "P .. AllocationInstructionAck",
        ("FIX.4.0", "35", "Q"): "Q .. DontKnowTrade",
        ("FIX.4.0", "35", "R"): "R .. QuoteRequest",
        ("FIX.4.0", "35", "S"): "S .. Quote",
        ("FIX.4.0", "39", "0"): "0 .. New",
        ("FIX.4.0", "39", "1"): "1 .. PartiallyFilled",
        ("FIX.4.0", "39", "2"): "2 .. Filled",
        ("FIX.4.0", "39", "3"): "3 .. DoneForDay",
        ("FIX.4.0", "39", "4"): "4 .. Canceled",
        ("FIX.4.0", "39", "5"): "5 .. Replaced",
        ("FIX.4.0", "39", "6"): "6 .. PendingCancel",
        ("FIX.4.0", "39", "7"): "7 .. Stopped",
        ("FIX.4.0", "39", "8"): "8 .. Rejected",
        ("FIX.4.0", "39", "9"): "9 .. Suspended",
        ("FIX.4.0", "39", "A"): "A .. PendingNew",
        ("FIX.4.0", "39", "B"): "B .. Calculated",
        ("FIX.4.0", "39", "C"): "C .. Expired",
        ("FIX.4.0", "40", "1"): "1 .. Market",
        ("FIX.4.0", "40", "2"): "2 .. Limit",
        ("FIX.4.0", "40", "3"): "3 .. Stop",
        ("FIX.4.0", "40", "4"): "4 .. StopLimit",
        ("FIX.4.0", "40", "5"): "5 .. MarketOnClose",
        ("FIX.4.0", "40", "6"): "6 .. WithOrWithout",
        ("FIX.4.0", "40", "7"): "7 .. LimitOrBetter",
        ("FIX.4.0", "40", "8"): "8 .. LimitWithOrWithout",
        ("FIX.4.0", "40", "9"): "9 .. OnBasis",
        ("FIX.4.0", "40", "A"): "A .. OnClose",
        ("FIX.4.0", "40", "B"): "B .. LimitOnClose",
        ("FIX.4.0", "40", "C"): "C .. ForexMarket",
        ("FIX.4.0", "40", "D"): "D .. PreviouslyQuoted",
        ("FIX.4.0", "40", "E"): "E .. PreviouslyIndicated",
        ("FIX.4.0", "40", "P"): "P .. Pegged",
        ("FIX.4.0", "43", "N"): "N .. OriginalTransmission",
        ("FIX.4.0", "43", "Y"): "Y .. PossibleDuplicate",
        ("FIX.4.0", "47", "A"): "A .. AgencySingleOrder",
        ("FIX.4.0", "47", "C"): "C .. ProprietaryNonAlgo",
        ("FIX.4.0", "47", "D"): "D .. ProgramOrderMember",
        ("FIX.4.0", "47", "I"): "I .. IndividualInvestor",
        ("FIX.4.0", "47", "J"): "J .. ProprietaryAlgo",
        ("FIX.4.0", "47", "K"): "K .. AgencyAlgo",
        ("FIX.4.0", "47", "M"): "M .. ProgramOrderOtherMember",
        ("FIX.4.0", "47", "N"): "N .. AgentForOtherMember",
        ("FIX.4.0", "47", "U"): "U .. AgencyIndexArb",
        ("FIX.4.0", "47", "W"): "W .. AllOtherOrdersAsAgentForOtherMember",
        ("FIX.4.0", "47", "Y"): "Y .. AgencyNonAlgo",
        ("FIX.4.0", "54", "1"): "1 .. Buy",
        ("FIX.4.0", "54", "2"): "2 .. Sell",
        ("FIX.4.0", "54", "3"): "3 .. BuyMinus",
        ("FIX.4.0", "54", "4"): "4 .. SellPlus",
        ("FIX.4.0", "54", "5"): "5 .. SellShort",
        ("FIX.4.0", "54", "6"): "6 .. SellShortExempt",
        ("FIX.4.0", "59", "0"): "0 .. Day",
        ("FIX.4.0", "59", "1"): "1 .. GoodTillCancel",
        ("FIX.4.0", "59", "2"): "2 .. AtTheOpening",
        ("FIX.4.0", "59", "3"): "3 .. ImmediateOrCancel",
        ("FIX.4.0", "59", "4"): "4 .. FillOrKill",
        ("FIX.4.0", "59", "5"): "5 .. GoodTillCrossing",
        ("FIX.4.0", "59", "6"): "6 .. GoodTillDate",
        ("FIX.4.0", "61", "0"): "0 .. Normal",
        ("FIX.4.0", "61", "1"): "1 .. Flash",
        ("FIX.4.0", "61", "2"): "2 .. Background",
        ("FIX.4.0", "63", "0"): "0 .. Regular",
        ("FIX.4.0", "63", "1"): "1 .. Cash",
        ("FIX.4.0", "63", "2"): "2 .. NextDay",
        ("FIX.4.0", "63", "3"): "3 .. TPlus2",
        ("FIX.4.0", "63", "4"): "4 .. TPlus3",
        ("FIX.4.0", "63", "5"): "5 .. TPlus4",
        ("FIX.4.0", "63", "6"): "6 .. Future",
        ("FIX.4.0", "63", "7"): "7 .. WhenAndIfIssued",
        ("FIX.4.0", "63", "8"): "8 .. SellersOption",
        ("FIX.4.0", "63", "9"): "9 .. TPlus5",
        ("FIX.4.0", "71", "0"): "0 .. New",
        ("FIX.4.0", "71", "1"): "1 .. Replace",
        ("FIX.4.0", "71", "2"): "2 .. Cancel",
        ("FIX.4.0", "81", "0"): "0 .. Regular",
        ("FIX.4.0", "81", "1"): "1 .. SoftDollar",
        ("FIX.4.0", "81", "2"): "2 .. StepIn",
        ("FIX.4.0", "81", "3"): "3 .. StepOut",
        ("FIX.4.0", "81", "4"): "4 .. SoftDollarStepIn",
        ("FIX.4.0", "81", "5"): "5 .. SoftDollarStepOut",
        ("FIX.4.0", "81", "6"): "6 .. PlanSponsor",
        ("FIX.4.0", "87", "0"): "0 .. Accepted",
        ("FIX.4.0", "87", "1"): "1 .. BlockLevelReject",
        ("FIX.4.0", "87", "2"): "2 .. AccountLevelReject",
        ("FIX.4.0", "87", "3"): "3 .. Received",
        ("FIX.4.0", "88", "0"): "0 .. UnknownAccount",
        ("FIX.4.0", "88", "1"): "1 .. IncorrectQuantity",
        ("FIX.4.0", "88", "2"): "2 .. IncorrectAveragegPrice",
        ("FIX.4.0", "88", "3"): "3 .. UnknownExecutingBrokerMnemonic",
        ("FIX.4.0", "88", "4"): "4 .. CommissionDifference",
        ("FIX.4.0", "88", "5"): "5 .. UnknownOrderID",
        ("FIX.4.0", "88", "6"): "6 .. UnknownListID",
        ("FIX.4.0", "88", "7"): "7 .. OtherSeeText",
        ("FIX.4.0", "94", "0"): "0 .. New",
        ("FIX.4.0", "94", "1"): "1 .. Reply",
        ("FIX.4.0", "94", "2"): "2 .. AdminReply",
        ("FIX.4.0", "98", "0"): "0 .. None",
        ("FIX.4.0", "98", "1"): "1 .. PKCS",
        ("FIX.4.0", "98", "2"): "2 .. DES",
        ("FIX.4.0", "98", "3"): "3 .. PKCSDES",
        ("FIX.4.0", "98", "4"): "4 .. PGPDES",
        ("FIX.4.0", "98", "5"): "5 .. PGPDESMD5",
        ("FIX.4.0", "98", "6"): "6 .. PEM",
        ("FIX.4.0", "100", "0"): "0 .. None",
        ("FIX.4.0", "100", "4"): "4 .. POSIT",
        ("FIX.4.0", "102", "0"): "0 .. TooLateToCancel",
        ("FIX.4.0", "102", "1"): "1 .. UnknownOrder",
        ("FIX.4.0", "103", "0"): "0 .. BrokerCredit",
        ("FIX.4.0", "103", "1"): "1 .. UnknownSymbol",
        ("FIX.4.0", "103", "2"): "2 .. ExchangeClosed",
        ("FIX.4.0", "103", "3"): "3 .. OrderExceedsLimit",
        ("FIX.4.0", "103", "4"): "4 .. TooLateToEnter",
        ("FIX.4.0", "104", "A"): "A .. AllOrNone",
        ("FIX.4.0", "104", "C"): "C .. AtTheClose",
        ("FIX.4.0", "104", "I"): "I .. InTouchWith",
        ("FIX.4.0", "104", "L"): "L .. Limit",
        ("FIX.4.0", "104", "M"): "M .. MoreBehind",
        ("FIX.4.0", "104", "O"): "O .. AtTheOpen",
        ("FIX.4.0", "104", "P"): "P .. TakingAPosition",
        ("FIX.4.0", "104", "Q"): "Q .. AtTheMarket",
        ("FIX.4.0", "104", "S"): "S .. PortfolioShown",
        ("FIX.4.0", "104", "T"): "T .. ThroughTheDay",
        ("FIX.4.0", "104", "V"): "V .. Versus",
        ("FIX.4.0", "104", "W"): "W .. Indication",
        ("FIX.4.0", "104", "X"): "X .. CrossingOpportunity",
        ("FIX.4.0", "113", "N"): "N .. SenderReports",
        ("FIX.4.0", "113", "Y"): "Y .. ReceiverReports",
        ("FIX.4.0", "114", "N"): "N .. No",
        ("FIX.4.0", "114", "Y"): "Y .. Yes",
        ("FIX.4.0", "121", "N"): "N .. DoNotExecuteForexAfterSecurityTrade",
        ("FIX.4.0", "121", "Y"): "Y .. ExecuteForexAfterSecurityTrade",
        ("FIX.4.0", "123", "N"): "N .. SequenceReset",
        ("FIX.4.0", "123", "Y"): "Y .. GapFillMessage",
        ("FIX.4.0", "125", "F"): "F .. FullRemainingQuantity",
        ("FIX.4.0", "125", "P"): "P .. PartialCancel",
        ("FIX.4.0", "127", "A"): "A .. UnknownSymbol",
        ("FIX.4.0", "127", "B"): "B .. WrongSide",
        ("FIX.4.0", "127", "C"): "C .. QuantityExceedsOrder",
        ("FIX.4.0", "127", "D"): "D .. NoMatchingOrder",
        ("FIX.4.0", "127", "E"): "E .. PriceExceedsLimit",
        ("FIX.4.0", "127", "Z"): "Z .. Other",
        ("FIX.4.0", "130", "N"): "N .. NotNatural",
        ("FIX.4.0", "130", "Y"): "Y .. Natural",
        ("FIX.4.0", "139", "1"): "1 .. Regulatory",
        ("FIX.4.0", "139", "2"): "2 .. Tax",
        ("FIX.4.0", "139", "3"): "3 .. LocalCommission",
        ("FIX.4.0", "139", "4"): "4 .. ExchangeFees",
        ("FIX.4.0", "139", "5"): "5 .. Stamp",
        ("FIX.4.0", "139", "6"): "6 .. Levy",
        ("FIX.4.0", "139", "7"): "7 .. Other",
        ("FIX.4.2", "4", "B"): "B .. Buy",
        ("FIX.4.2", "4", "S"): "S .. Sell",
        ("FIX.4.2", "4", "T"): "T .. Trade",
        ("FIX.4.2", "4", "X"): "X .. Cross",
        ("FIX.4.2", "5", "C"): "C .. Cancel",
        ("FIX.4.2", "5", "N"): "N .. New",
        ("FIX.4.2", "5", "R"): "R .. Replace",
        ("FIX.4.2", "13", "1"): "1 .. PerUnit",
        ("FIX.4.2", "13", "2"): "2 .. Percent",
        ("FIX.4.2", "13", "3"): "3 .. Absolute",
        ("FIX.4.2", "18", "0"): "0 .. StayOnOfferSide",
        ("FIX.4.2", "18", "1"): "1 .. NotHeld",
        ("FIX.4.2", "18", "2"): "2 .. Work",
        ("FIX.4.2", "18", "3"): "3 .. GoAlong",
        ("FIX.4.2", "18", "4"): "4 .. OverTheDay",
        ("FIX.4.2", "18", "5"): "5 .. Held",
        ("FIX.4.2", "18", "6"): "6 .. ParticipateDoNotInitiate",
        ("FIX.4.2", "18", "7"): "7 .. StrictScale",
        ("FIX.4.2", "18", "8"): "8 .. TryToScale",
        ("FIX.4.2", "18", "9"): "9 .. StayOnBidSide",
        ("FIX.4.2", "18", "A"): "A .. NoCross",
        ("FIX.4.2", "18", "B"): "B .. OKToCross",
        ("FIX.4.2", "18", "C"): "C .. CallFirst",
        ("FIX.4.2", "18", "D"): "D .. PercentOfVolume",
        ("FIX.4.2", "18", "E"): "E .. DoNotIncrease",
        ("FIX.4.2", "18", "F"): "F .. DoNotReduce",
        ("FIX.4.2", "18", "G"): "G .. AllOrNone",
        ("FIX.4.2", "18", "I"): "I .. InstitutionsOnly",
        ("FIX.4.2", "18", "L"): "L .. LastPeg",
        ("FIX.4.2", "18", "M"): "M .. MidPricePeg",
        ("FIX.4.2", "18", "N"): "N .. NonNegotiable",
        ("FIX.4.2", "18", "O"): "O .. OpeningPeg",
        ("FIX.4.2", "18", "P"): "P .. MarketPeg",
        ("FIX.4.2", "18", "R"): "R .. PrimaryPeg",
        ("FIX.4.2", "18", "S"): "S .. Suspend",
        ("FIX.4.2", "18", "T"): "T .. FixedPegToLocalBestBidOrOfferAtTimeOfOrder",
        ("FIX.4.2", "18", "U"): "U .. CustomerDisplayInstruction",
        ("FIX.4.2", "18", "V"): "V .. Netting",
        ("FIX.4.2", "18", "W"): "W .. PegToVWAP",
        ("FIX.4.2", "20", "0"): "0 .. New",
        ("FIX.4.2", "20", "1"): "1 .. Cancel",
        ("FIX.4.2", "20", "2"): "2 .. Correct",
        ("FIX.4.2", "20", "3"): "3 .. Status",
        ("FIX.4.2", "21", "1"): "1 .. AutomatedExecutionNoIntervention",
        ("FIX.4.2", "21", "2"): "2 .. AutomatedExecutionInterventionOK",
        ("FIX.4.2", "21", "3"): "3 .. ManualOrder",
        ("FIX.4.2", "22", "1"): "1 .. CUSIP",
        ("FIX.4.2", "22", "2"): "2 .. SEDOL",
        ("FIX.4.2", "22", "3"): "3 .. QUIK",
        ("FIX.4.2", "22", "4"): "4 .. ISINNumber",
        ("FIX.4.2", "22", "5"): "5 .. RICCode",
        ("FIX.4.2", "22", "6"): "6 .. ISOCurrencyCode",
        ("FIX.4.2", "22", "7"): "7 .. ISOCountryCode",
        ("FIX.4.2", "22", "8"): "8 .. ExchangeSymbol",
        ("FIX.4.2", "22", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.4.2", "25", "H"): "H .. High",
        ("FIX.4.2", "25", "L"): "L .. Low",
        ("FIX.4.2", "25", "M"): "M .. Medium",
        ("FIX.4.2", "27", "L"): "L .. Large",
        ("FIX.4.2", "27", "M"): "M .. Medium",
        ("FIX.4.2", "27", "S"): "S .. Small",
        ("FIX.4.2", "28", "C"): "C .. Cancel",
        ("FIX.4.2", "28", "N"): "N .. New",
        ("FIX.4.2", "28", "R"): "R .. Replace",
        ("FIX.4.2", "29", "1"): "1 .. Agent",
        ("FIX.4.2", "29", "2"): "2 .. CrossAsAgent",
        ("FIX.4.2", "29", "3"): "3 .. CrossAsPrincipal",
        ("FIX.4.2", "29", "4"): "4 .. Principal",
        ("FIX.4.2", "35", "0"): "0 .. Heartbeat",
        ("FIX.4.2", "35", "1"): "1 .. TestRequest",
        ("FIX.4.2", "35", "2"): "2 .. ResendRequest",
        ("FIX.4.2", "35", "3"): "3 .. Reject",
        ("FIX.4.2", "35", "4"): "4 .. SequenceReset",
        ("FIX.4.2", "35", "5"): "5 .. Logout",
        ("FIX.4.2", "35", "6"): "6 .. IOI",
        ("FIX.4.2", "35", "7"): "7 .. Advertisement",
        ("FIX.4.2", "35", "8"): "8 .. ExecutionReport",
        ("FIX.4.2", "35", "9"): "9 .. OrderCancelReject",
        ("FIX.4.2", "35", "A"): "A .. Logon",
        ("FIX.4.2", "35", "B"): "B .. News",
        ("FIX.4.2", "35", "C"): "C .. Email",
        ("FIX.4.2", "35", "D"): "D .. NewOrderSingle",
        ("FIX.4.2", "35", "E"): "E .. NewOrderList",
        ("FIX.4.2", "35", "F"): "F .. OrderCancelRequest",
        ("FIX.4.2", "35", "G"): "G .. OrderCancelReplaceRequest",
        ("FIX.4.2", "35", "H"): "H .. OrderStatusRequest",
        ("FIX.4.2", "35", "J"): "J .. AllocationInstruction",
        ("FIX.4.2", "35", "K"): "K .. ListCancelRequest",
        ("FIX.4.2", "35", "L"): "L .. ListExecute",
        ("FIX.4.2", "35", "M"): "M .. ListStatusRequest",
        ("FIX.4.2", "35", "N"): "N .. ListStatus",
        ("FIX.4.2", "35", "P"): "P .. AllocationInstructionAck",
        ("FIX.4.2", "35", "Q"): "Q .. DontKnowTrade",
        ("FIX.4.2", "35", "R"): "R .. QuoteRequest",
        ("FIX.4.2", "35", "S"): "S .. Quote",
        ("FIX.4.2", "35", "T"): "T .. SettlementInstructions",
        ("FIX.4.2", "35", "V"): "V .. MarketDataRequest",
        ("FIX.4.2", "35", "W"): "W .. MarketDataSnapshotFullRefresh",
        ("FIX.4.2", "35", "X"): "X .. MarketDataIncrementalRefresh",
        ("FIX.4.2", "35", "Y"): "Y .. MarketDataRequestReject",
        ("FIX.4.2", "35", "Z"): "Z .. QuoteCancel",
        ("FIX.4.2", "35", "a"): "a .. QuoteStatusRequest",
        ("FIX.4.2", "35", "b"): "b .. MassQuoteAcknowledgement",
        ("FIX.4.2", "35", "c"): "c .. SecurityDefinitionRequest",
        ("FIX.4.2", "35", "d"): "d .. SecurityDefinition",
        ("FIX.4.2", "35", "e"): "e .. SecurityStatusRequest",
        ("FIX.4.2", "35", "f"): "f .. SecurityStatus",
        ("FIX.4.2", "35", "g"): "g .. TradingSessionStatusRequest",
        ("FIX.4.2", "35", "h"): "h .. TradingSessionStatus",
        ("FIX.4.2", "35", "i"): "i .. MassQuote",
        ("FIX.4.2", "35", "j"): "j .. BusinessMessageReject",
        ("FIX.4.2", "35", "k"): "k .. BidRequest",
        ("FIX.4.2", "35", "l"): "l .. BidResponse",
        ("FIX.4.2", "35", "m"): "m .. ListStrikePrice",
        ("FIX.4.2", "39", "0"): "0 .. New",
        ("FIX.4.2", "39", "1"): "1 .. PartiallyFilled",
        ("FIX.4.2", "39", "2"): "2 .. Filled",
        ("FIX.4.2", "39", "3"): "3 .. DoneForDay",
        ("FIX.4.2", "39", "4"): "4 .. Canceled",
        ("FIX.4.2", "39", "5"): "5 .. Replaced",
        ("FIX.4.2", "39", "6"): "6 .. PendingCancel",
        ("FIX.4.2", "39", "7"): "7 .. Stopped",
        ("FIX.4.2", "39", "8"): "8 .. Rejected",
        ("FIX.4.2", "39", "9"): "9 .. Suspended",
        ("FIX.4.2", "39", "A"): "A .. PendingNew",
        ("FIX.4.2", "39", "B"): "B .. Calculated",
        ("FIX.4.2", "39", "C"): "C .. Expired",
        ("FIX.4.2", "39", "D"): "D .. AcceptedForBidding",
        ("FIX.4.2", "39", "E"): "E .. PendingReplace",
        ("FIX.4.2", "40", "1"): "1 .. Market",
        ("FIX.4.2", "40", "2"): "2 .. Limit",
        ("FIX.4.2", "40", "3"): "3 .. Stop",
        ("FIX.4.2", "40", "4"): "4 .. StopLimit",
        ("FIX.4.2", "40", "5"): "5 .. MarketOnClose",
        ("FIX.4.2", "40", "6"): "6 .. WithOrWithout",
        ("FIX.4.2", "40", "7"): "7 .. LimitOrBetter",
        ("FIX.4.2", "40", "8"): "8 .. LimitWithOrWithout",
        ("FIX.4.2", "40", "9"): "9 .. OnBasis",
        ("FIX.4.2", "40", "A"): "A .. OnClose",
        ("FIX.4.2", "40", "B"): "B .. LimitOnClose",
        ("FIX.4.2", "40", "C"): "C .. ForexMarket",
        ("FIX.4.2", "40", "D"): "D .. PreviouslyQuoted",
        ("FIX.4.2", "40", "E"): "E .. PreviouslyIndicated",
        ("FIX.4.2", "40", "F"): "F .. ForexLimit",
        ("FIX.4.2", "40", "G"): "G .. ForexSwap",
        ("FIX.4.2", "40", "H"): "H .. ForexPreviouslyQuoted",
        ("FIX.4.2", "40", "I"): "I .. Funari",
        ("FIX.4.2", "40", "P"): "P .. Pegged",
        ("FIX.4.2", "43", "N"): "N .. OriginalTransmission",
        ("FIX.4.2", "43", "Y"): "Y .. PossibleDuplicate",
        ("FIX.4.2", "47", "A"): "A .. AgencySingleOrder",
        ("FIX.4.2", "47", "B"): "B .. ShortExemptTransactionAType",
        ("FIX.4.2", "47", "C"): "C .. ProprietaryNonAlgo",
        ("FIX.4.2", "47", "D"): "D .. ProgramOrderMember",
        ("FIX.4.2", "47", "E"): "E .. ShortExemptTransactionForPrincipal",
        ("FIX.4.2", "47", "F"): "F .. ShortExemptTransactionWType",
        ("FIX.4.2", "47", "H"): "H .. ShortExemptTransactionIType",
        ("FIX.4.2", "47", "I"): "I .. IndividualInvestor",
        ("FIX.4.2", "47", "J"): "J .. ProprietaryAlgo",
        ("FIX.4.2", "47", "K"): "K .. AgencyAlgo",
        ("FIX.4.2", "47", "L"): "L .. ShortExemptTransactionMemberAffliated",
        ("FIX.4.2", "47", "M"): "M .. ProgramOrderOtherMember",
        ("FIX.4.2", "47", "N"): "N .. AgentForOtherMember",
        ("FIX.4.2", "47", "O"): "O .. ProprietaryTransactionAffiliated",
        ("FIX.4.2", "47", "P"): "P .. Principal",
        ("FIX.4.2", "47", "R"): "R .. TransactionNonMember",
        ("FIX.4.2", "47", "S"): "S .. SpecialistTrades",
        ("FIX.4.2", "47", "T"): "T .. TransactionUnaffiliatedMember",
        ("FIX.4.2", "47", "U"): "U .. AgencyIndexArb",
        ("FIX.4.2", "47", "W"): "W .. AllOtherOrdersAsAgentForOtherMember",
        ("FIX.4.2", "47", "X"): "X .. ShortExemptTransactionMemberNotAffliated",
        ("FIX.4.2", "47", "Y"): "Y .. AgencyNonAlgo",
        ("FIX.4.2", "47", "Z"): "Z .. ShortExemptTransactionNonMember",
        ("FIX.4.2", "54", "1"): "1 .. Buy",
        ("FIX.4.2", "54", "2"): "2 .. Sell",
        ("FIX.4.2", "54", "3"): "3 .. BuyMinus",
        ("FIX.4.2", "54", "4"): "4 .. SellPlus",
        ("FIX.4.2", "54", "5"): "5 .. SellShort",
        ("FIX.4.2", "54", "6"): "6 .. SellShortExempt",
        ("FIX.4.2", "54", "7"): "7 .. Undisclosed",
        ("FIX.4.2", "54", "8"): "8 .. Cross",
        ("FIX.4.2", "54", "9"): "9 .. CrossShort",
        ("FIX.4.2", "59", "0"): "0 .. Day",
        ("FIX.4.2", "59", "1"): "1 .. GoodTillCancel",
        ("FIX.4.2", "59", "2"): "2 .. AtTheOpening",
        ("FIX.4.2", "59", "3"): "3 .. ImmediateOrCancel",
        ("FIX.4.2", "59", "4"): "4 .. FillOrKill",
        ("FIX.4.2", "59", "5"): "5 .. GoodTillCrossing",
        ("FIX.4.2", "59", "6"): "6 .. GoodTillDate",
        ("FIX.4.2", "61", "0"): "0 .. Normal",
        ("FIX.4.2", "61", "1"): "1 .. Flash",
        ("FIX.4.2", "61", "2"): "2 .. Background",
        ("FIX.4.2", "63", "0"): "0 .. Regular",
        ("FIX.4.2", "63", "1"): "1 .. Cash",
        ("FIX.4.2", "63", "2"): "2 .. NextDay",
        ("FIX.4.2", "63", "3"): "3 .. TPlus2",
        ("FIX.4.2", "63", "4"): "4 .. TPlus3",
        ("FIX.4.2", "63", "5"): "5 .. TPlus4",
        ("FIX.4.2", "63", "6"): "6 .. Future",
        ("FIX.4.2", "63", "7"): "7 .. WhenAndIfIssued",
        ("FIX.4.2", "63", "8"): "8 .. SellersOption",
        ("FIX.4.2", "63", "9"): "9 .. TPlus5",
        ("FIX.4.2", "71", "0"): "0 .. New",
        ("FIX.4.2", "71", "1"): "1 .. Replace",
        ("FIX.4.2", "71", "2"): "2 .. Cancel",
        ("FIX.4.2", "71", "3"): "3 .. Preliminary",
        ("FIX.4.2", "71", "4"): "4 .. Calculated",
        ("FIX.4.2", "71", "5"): "5 .. CalculatedWithoutPreliminary",
        ("FIX.4.2", "77", "C"): "C .. Close",
        ("FIX.4.2", "77", "O"): "O .. Open",
        ("FIX.4.2", "81", "0"): "0 .. Regular",
        ("FIX.4.2", "81", "1"): "1 .. SoftDollar",
        ("FIX.4.2", "81", "2"): "2 .. StepIn",
        ("FIX.4.2", "81", "3"): "3 .. StepOut",
        ("FIX.4.2", "81", "4"): "4 .. SoftDollarStepIn",
        ("FIX.4.2", "81", "5"): "5 .. SoftDollarStepOut",
        ("FIX.4.2", "81", "6"): "6 .. PlanSponsor",
        ("FIX.4.2", "87", "0"): "0 .. Accepted",
        ("FIX.4.2", "87", "1"): "1 .. BlockLevelReject",
        ("FIX.4.2", "87", "2"): "2 .. AccountLevelReject",
        ("FIX.4.2", "87", "3"): "3 .. Received",
        ("FIX.4.2", "88", "0"): "0 .. UnknownAccount",
        ("FIX.4.2", "88", "1"): "1 .. IncorrectQuantity",
        ("FIX.4.2", "88", "2"): "2 .. IncorrectAveragegPrice",
        ("FIX.4.2", "88", "3"): "3 .. UnknownExecutingBrokerMnemonic",
        ("FIX.4.2", "88", "4"): "4 .. CommissionDifference",
        ("FIX.4.2", "88", "5"): "5 .. UnknownOrderID",
        ("FIX.4.2", "88", "6"): "6 .. UnknownListID",
        ("FIX.4.2", "88", "7"): "7 .. OtherSeeText",
        ("FIX.4.2", "94", "0"): "0 .. New",
        ("FIX.4.2", "94", "1"): "1 .. Reply",
        ("FIX.4.2", "94", "2"): "2 .. AdminReply",
        ("FIX.4.2", "97", "N"): "N .. OriginalTransmission",
        ("FIX.4.2", "97", "Y"): "Y .. PossibleResend",
        ("FIX.4.2", "98", "0"): "0 .. None",
        ("FIX.4.2", "98", "1"): "1 .. PKCS",
        ("FIX.4.2", "98", "2"): "2 .. DES",
        ("FIX.4.2", "98", "3"): "3 .. PKCSDES",
        ("FIX.4.2", "98", "4"): "4 .. PGPDES",
        ("FIX.4.2", "98", "5"): "5 .. PGPDESMD5",
        ("FIX.4.2", "98", "6"): "6 .. PEM",
        ("FIX.4.2", "102", "0"): "0 .. TooLateToCancel",
        ("FIX.4.2", "102", "1"): "1 .. UnknownOrder",
        ("FIX.4.2", "102", "2"): "2 .. BrokerCredit",
        ("FIX.4.2", "102", "3"): "3 .. OrderAlreadyInPendingStatus",
        ("FIX.4.2", "103", "0"): "0 .. BrokerCredit",
        ("FIX.4.2", "103", "1"): "1 .. UnknownSymbol",
        ("FIX.4.2", "103", "2"): "2 .. ExchangeClosed",
        ("FIX.4.2", "103", "3"): "3 .. OrderExceedsLimit",
        ("FIX.4.2", "103", "4"): "4 .. TooLateToEnter",
        ("FIX.4.2", "103", "5"): "5 .. UnknownOrder",
        ("FIX.4.2", "103", "6"): "6 .. DuplicateOrder",
        ("FIX.4.2", "103", "7"): "7 .. DuplicateOfAVerballyCommunicatedOrder",
        ("FIX.4.2", "103", "8"): "8 .. StaleOrder",
        ("FIX.4.2", "104", "A"): "A .. AllOrNone",
        ("FIX.4.2", "104", "C"): "C .. AtTheClose",
        ("FIX.4.2", "104", "I"): "I .. InTouchWith",
        ("FIX.4.2", "104", "L"): "L .. Limit",
        ("FIX.4.2", "104", "M"): "M .. MoreBehind",
        ("FIX.4.2", "104", "O"): "O .. AtTheOpen",
        ("FIX.4.2", "104", "P"): "P .. TakingAPosition",
        ("FIX.4.2", "104", "Q"): "Q .. AtTheMarket",
        ("FIX.4.2", "104", "R"): "R .. ReadyToTrade",
        ("FIX.4.2", "104", "S"): "S .. PortfolioShown",
        ("FIX.4.2", "104", "T"): "T .. ThroughTheDay",
        ("FIX.4.2", "104", "V"): "V .. Versus",
        ("FIX.4.2", "104", "W"): "W .. Indication",
        ("FIX.4.2", "104", "X"): "X .. CrossingOpportunity",
        ("FIX.4.2", "104", "Y"): "Y .. AtTheMidpoint",
        ("FIX.4.2", "104", "Z"): "Z .. PreOpen",
        ("FIX.4.2", "113", "N"): "N .. SenderReports",
        ("FIX.4.2", "113", "Y"): "Y .. ReceiverReports",
        ("FIX.4.2", "114", "N"): "N .. No",
        ("FIX.4.2", "114", "Y"): "Y .. Yes",
        ("FIX.4.2", "121", "N"): "N .. DoNotExecuteForexAfterSecurityTrade",
        ("FIX.4.2", "121", "Y"): "Y .. ExecuteForexAfterSecurityTrade",
        ("FIX.4.2", "123", "N"): "N .. SequenceReset",
        ("FIX.4.2", "123", "Y"): "Y .. GapFillMessage",
        ("FIX.4.2", "127", "A"): "A .. UnknownSymbol",
        ("FIX.4.2", "127", "B"): "B .. WrongSide",
        ("FIX.4.2", "127", "C"): "C .. QuantityExceedsOrder",
        ("FIX.4.2", "127", "D"): "D .. NoMatchingOrder",
        ("FIX.4.2", "127", "E"): "E .. PriceExceedsLimit",
        ("FIX.4.2", "127", "Z"): "Z .. Other",
        ("FIX.4.2", "130", "N"): "N .. NotNatural",
        ("FIX.4.2", "130", "Y"): "Y .. Natural",
        ("FIX.4.2", "139", "1"): "1 .. Regulatory",
        ("FIX.4.2", "139", "2"): "2 .. Tax",
        ("FIX.4.2", "139", "3"): "3 .. LocalCommission",
        ("FIX.4.2", "139", "4"): "4 .. ExchangeFees",
        ("FIX.4.2", "139", "5"): "5 .. Stamp",
        ("FIX.4.2", "139", "6"): "6 .. Levy",
        ("FIX.4.2", "139", "7"): "7 .. Other",
        ("FIX.4.2", "139", "8"): "8 .. Markup",
        ("FIX.4.2", "139", "9"): "9 .. ConsumptionTax",
        ("FIX.4.2", "141", "N"): "N .. No",
        ("FIX.4.2", "141", "Y"): "Y .. Yes",
        ("FIX.4.2", "150", "0"): "0 .. New",
        ("FIX.4.2", "150", "1"): "1 .. PartialFill",
        ("FIX.4.2", "150", "2"): "2 .. Fill",
        ("FIX.4.2", "150", "3"): "3 .. DoneForDay",
        ("FIX.4.2", "150", "4"): "4 .. Canceled",
        ("FIX.4.2", "150", "5"): "5 .. Replaced",
        ("FIX.4.2", "150", "6"): "6 .. PendingCancel",
        ("FIX.4.2", "150", "7"): "7 .. Stopped",
        ("FIX.4.2", "150", "8"): "8 .. Rejected",
        ("FIX.4.2", "150", "9"): "9 .. Suspended",
        ("FIX.4.2", "150", "A"): "A .. PendingNew",
        ("FIX.4.2", "150", "B"): "B .. Calculated",
        ("FIX.4.2", "150", "C"): "C .. Expired",
        ("FIX.4.2", "150", "D"): "D .. Restated",
        ("FIX.4.2", "150", "E"): "E .. PendingReplace",
        ("FIX.4.2", "160", "0"): "0 .. Default",
        ("FIX.4.2", "160", "1"): "1 .. StandingInstructionsProvided",
        ("FIX.4.2", "160", "2"): "2 .. SpecificAllocationAccountOverriding",
        ("FIX.4.2", "160", "3"): "3 .. SpecificAllocationAccountStanding",
        ("FIX.4.2", "163", "C"): "C .. Cancel",
        ("FIX.4.2", "163", "N"): "N .. New",
        ("FIX.4.2", "163", "R"): "R .. Replace",
        ("FIX.4.2", "165", "1"): "1 .. BrokerCredit",
        ("FIX.4.2", "165", "2"): "2 .. Institution",
        ("FIX.4.2", "166", "CED"): "CED .. CEDEL",
        ("FIX.4.2", "166", "DTC"): "DTC .. DepositoryTrustCompany",
        ("FIX.4.2", "166", "EUR"): "EUR .. EuroClear",
        ("FIX.4.2", "166", "FED"): "FED .. FederalBookEntry",
        ("FIX.4.2", "166", "ISO Country Code"): "ISO Country Code .. LocalMarketSettleLocation",
        ("FIX.4.2", "166", "PNY"): "PNY .. Physical",
        ("FIX.4.2", "166", "PTC"): "PTC .. ParticipantTrustCompany",
        ("FIX.4.2", "167", "?"): "? .. Wildcard",
        ("FIX.4.2", "167", "BA"): "BA .. BankersAcceptance",
        ("FIX.4.2", "167", "CB"): "CB .. ConvertibleBond",
        ("FIX.4.2", "167", "CD"): "CD .. CertificateOfDeposit",
        ("FIX.4.2", "167", "CMO"): "CMO .. CollateralizedMortgageObligation",
        ("FIX.4.2", "167", "CORP"): "CORP .. CorporateBond",
        ("FIX.4.2", "167", "CP"): "CP .. CommercialPaper",
        ("FIX.4.2", "167", "CPP"): "CPP .. CorporatePrivatePlacement",
        ("FIX.4.2", "167", "CS"): "CS .. CommonStock",
        ("FIX.4.2", "167", "FHA"): "FHA .. FederalHousingAuthority",
        ("FIX.4.2", "167", "FHL"): "FHL .. FederalHomeLoan",
        ("FIX.4.2", "167", "FN"): "FN .. FederalNationalMortgageAssociation",
        ("FIX.4.2", "167", "FOR"): "FOR .. ForeignExchangeContract",
        ("FIX.4.2", "167", "FUT"): "FUT .. Future",
        ("FIX.4.2", "167", "GN"): "GN .. GovernmentNationalMortgageAssociation",
        ("FIX.4.2", "167", "GOVT"): "GOVT .. TreasuriesAgencyDebenture",
        ("FIX.4.2", "167", "IET"): "IET .. IOETTEMortgage",
        ("FIX.4.2", "167", "MF"): "MF .. MutualFund",
        ("FIX.4.2", "167", "MIO"): "MIO .. MortgageInterestOnly",
        ("FIX.4.2", "167", "MPO"): "MPO .. MortgagePrincipalOnly",
        ("FIX.4.2", "167", "MPP"): "MPP .. MortgagePrivatePlacement",
        ("FIX.4.2", "167", "MPT"): "MPT .. MiscellaneousPassThrough",
        ("FIX.4.2", "167", "MUNI"): "MUNI .. MunicipalBond",
        ("FIX.4.2", "167", "NONE"): "NONE .. NoSecurityType",
        ("FIX.4.2", "167", "OPT"): "OPT .. Option",
        ("FIX.4.2", "167", "PS"): "PS .. PreferredStock",
        ("FIX.4.2", "167", "RP"): "RP .. RepurchaseAgreement",
        ("FIX.4.2", "167", "RVRP"): "RVRP .. ReverseRepurchaseAgreement",
        ("FIX.4.2", "167", "SL"): "SL .. StudentLoanMarketingAssociation",
        ("FIX.4.2", "167", "TD"): "TD .. TimeDeposit",
        ("FIX.4.2", "167", "USTB"): "USTB .. USTreasuryBillOld",
        ("FIX.4.2", "167", "WAR"): "WAR .. Warrant",
        ("FIX.4.2", "167", "ZOO"): "ZOO .. CatsTigersAndLions",
        ("FIX.4.2", "169", "0"): "0 .. Other",
        ("FIX.4.2", "169", "1"): "1 .. DTCSID",
        ("FIX.4.2", "169", "2"): "2 .. ThomsonALERT",
        ("FIX.4.2", "169", "3"): "3 .. AGlobalCustodian",
        ("FIX.4.2", "197", "0"): "0 .. FXNetting",
        ("FIX.4.2", "197", "1"): "1 .. FXSwap",
        ("FIX.4.2", "201", "0"): "0 .. Put",
        ("FIX.4.2", "201", "1"): "1 .. Call",
        ("FIX.4.2", "203", "0"): "0 .. Covered",
        ("FIX.4.2", "203", "1"): "1 .. Uncovered",
        ("FIX.4.2", "204", "0"): "0 .. Customer",
        ("FIX.4.2", "204", "1"): "1 .. Firm",
        ("FIX.4.2", "208", "N"): "N .. DetailsShouldNotBeCommunicated",
        ("FIX.4.2", "208", "Y"): "Y .. DetailsShouldBeCommunicated",
        ("FIX.4.2", "209", "1"): "1 .. Match",
        ("FIX.4.2", "209", "2"): "2 .. Forward",
        ("FIX.4.2", "209", "3"): "3 .. ForwardAndMatch",
        ("FIX.4.2", "216", "1"): "1 .. TargetFirm",
        ("FIX.4.2", "216", "2"): "2 .. TargetList",
        ("FIX.4.2", "216", "3"): "3 .. BlockFirm",
        ("FIX.4.2", "216", "4"): "4 .. BlockList",
        ("FIX.4.2", "219", "1"): "1 .. CURVE",
        ("FIX.4.2", "219", "2"): "2 .. FiveYR",
        ("FIX.4.2", "219", "3"): "3 .. OLD5",
        ("FIX.4.2", "219", "4"): "4 .. TenYR",
        ("FIX.4.2", "219", "5"): "5 .. OLD10",
        ("FIX.4.2", "219", "6"): "6 .. ThirtyYR",
        ("FIX.4.2", "219", "7"): "7 .. OLD30",
        ("FIX.4.2", "219", "8"): "8 .. ThreeMOLIBOR",
        ("FIX.4.2", "219", "9"): "9 .. SixMOLIBOR",
        ("FIX.4.2", "263", "0"): "0 .. Snapshot",
        ("FIX.4.2", "263", "1"): "1 .. SnapshotAndUpdates",
        ("FIX.4.2", "263", "2"): "2 .. DisablePreviousSnapshot",
        ("FIX.4.2", "265", "0"): "0 .. FullRefresh",
        ("FIX.4.2", "265", "1"): "1 .. IncrementalRefresh",
        ("FIX.4.2", "266", "N"): "N .. BookEntriesShouldNotBeAggregated",
        ("FIX.4.2", "266", "Y"): "Y .. BookEntriesToBeAggregated",
        ("FIX.4.2", "269", "0"): "0 .. Bid",
        ("FIX.4.2", "269", "1"): "1 .. Offer",
        ("FIX.4.2", "269", "2"): "2 .. Trade",
        ("FIX.4.2", "269", "3"): "3 .. IndexValue",
        ("FIX.4.2", "269", "4"): "4 .. OpeningPrice",
        ("FIX.4.2", "269", "5"): "5 .. ClosingPrice",
        ("FIX.4.2", "269", "6"): "6 .. SettlementPrice",
        ("FIX.4.2", "269", "7"): "7 .. TradingSessionHighPrice",
        ("FIX.4.2", "269", "8"): "8 .. TradingSessionLowPrice",
        ("FIX.4.2", "269", "9"): "9 .. TradingSessionVWAPPrice",
        ("FIX.4.2", "274", "0"): "0 .. PlusTick",
        ("FIX.4.2", "274", "1"): "1 .. ZeroPlusTick",
        ("FIX.4.2", "274", "2"): "2 .. MinusTick",
        ("FIX.4.2", "274", "3"): "3 .. ZeroMinusTick",
        ("FIX.4.2", "276", "A"): "A .. Open",
        ("FIX.4.2", "276", "B"): "B .. Closed",
        ("FIX.4.2", "276", "C"): "C .. ExchangeBest",
        ("FIX.4.2", "276", "D"): "D .. ConsolidatedBest",
        ("FIX.4.2", "276", "E"): "E .. Locked",
        ("FIX.4.2", "276", "F"): "F .. Crossed",
        ("FIX.4.2", "276", "G"): "G .. Depth",
        ("FIX.4.2", "276", "H"): "H .. FastTrading",
        ("FIX.4.2", "276", "I"): "I .. NonFirm",
        ("FIX.4.2", "277", "A"): "A .. Cash",
        ("FIX.4.2", "277", "B"): "B .. AveragePriceTrade",
        ("FIX.4.2", "277", "C"): "C .. CashTrade",
        ("FIX.4.2", "277", "D"): "D .. NextDay",
        ("FIX.4.2", "277", "E"): "E .. Opening",
        ("FIX.4.2", "277", "F"): "F .. IntradayTradeDetail",
        ("FIX.4.2", "277", "G"): "G .. Rule127Trade",
        ("FIX.4.2", "277", "H"): "H .. Rule155Trade",
        ("FIX.4.2", "277", "I"): "I .. SoldLast",
        ("FIX.4.2", "277", "J"): "J .. NextDayTrade",
        ("FIX.4.2", "277", "K"): "K .. Opened",
        ("FIX.4.2", "277", "L"): "L .. Seller",
        ("FIX.4.2", "277", "M"): "M .. Sold",
        ("FIX.4.2", "277", "N"): "N .. StoppedStock",
        ("FIX.4.2", "279", "0"): "0 .. New",
        ("FIX.4.2", "279", "1"): "1 .. Change",
        ("FIX.4.2", "279", "2"): "2 .. Delete",
        ("FIX.4.2", "281", "0"): "0 .. UnknownSymbol",
        ("FIX.4.2", "281", "1"): "1 .. DuplicateMDReqID",
        ("FIX.4.2", "281", "2"): "2 .. InsufficientBandwidth",
        ("FIX.4.2", "281", "3"): "3 .. InsufficientPermissions",
        ("FIX.4.2", "281", "4"): "4 .. UnsupportedSubscriptionRequestType",
        ("FIX.4.2", "281", "5"): "5 .. UnsupportedMarketDepth",
        ("FIX.4.2", "281", "6"): "6 .. UnsupportedMDUpdateType",
        ("FIX.4.2", "281", "7"): "7 .. UnsupportedAggregatedBook",
        ("FIX.4.2", "281", "8"): "8 .. UnsupportedMDEntryType",
        ("FIX.4.2", "285", "0"): "0 .. Cancellation",
        ("FIX.4.2", "285", "1"): "1 .. Error",
        ("FIX.4.2", "286", "0"): "0 .. DailyOpen",
        ("FIX.4.2", "286", "1"): "1 .. SessionOpen",
        ("FIX.4.2", "286", "2"): "2 .. DeliverySettlementEntry",
        ("FIX.4.2", "291", "1"): "1 .. Bankrupt",
        ("FIX.4.2", "292", "A"): "A .. ExDividend",
        ("FIX.4.2", "292", "B"): "B .. ExDistribution",
        ("FIX.4.2", "292", "C"): "C .. ExRights",
        ("FIX.4.2", "292", "D"): "D .. New",
        ("FIX.4.2", "292", "E"): "E .. ExInterest",
        ("FIX.4.2", "297", "0"): "0 .. Accepted",
        ("FIX.4.2", "297", "1"): "1 .. CancelForSymbol",
        ("FIX.4.2", "297", "2"): "2 .. CanceledForSecurityType",
        ("FIX.4.2", "297", "3"): "3 .. CanceledForUnderlying",
        ("FIX.4.2", "297", "4"): "4 .. CanceledAll",
        ("FIX.4.2", "297", "5"): "5 .. Rejected",
        ("FIX.4.2", "298", "1"): "1 .. CancelForOneOrMoreSecurities",
        ("FIX.4.2", "298", "2"): "2 .. CancelForSecurityType",
        ("FIX.4.2", "298", "3"): "3 .. CancelForUnderlyingSecurity",
        ("FIX.4.2", "298", "4"): "4 .. CancelAllQuotes",
        ("FIX.4.2", "300", "1"): "1 .. UnknownSymbol",
        ("FIX.4.2", "300", "2"): "2 .. Exchange",
        ("FIX.4.2", "300", "3"): "3 .. QuoteRequestExceedsLimit",
        ("FIX.4.2", "300", "4"): "4 .. TooLateToEnter",
        ("FIX.4.2", "300", "5"): "5 .. UnknownQuote",
        ("FIX.4.2", "300", "6"): "6 .. DuplicateQuote",
        ("FIX.4.2", "300", "7"): "7 .. InvalidBid",
        ("FIX.4.2", "300", "8"): "8 .. InvalidPrice",
        ("FIX.4.2", "300", "9"): "9 .. NotAuthorizedToQuoteSecurity",
        ("FIX.4.2", "301", "0"): "0 .. NoAcknowledgement",
        ("FIX.4.2", "301", "1"): "1 .. AcknowledgeOnlyNegativeOrErroneousQuotes",
        ("FIX.4.2", "301", "2"): "2 .. AcknowledgeEachQuoteMessage",
        ("FIX.4.2", "303", "1"): "1 .. Manual",
        ("FIX.4.2", "303", "2"): "2 .. Automatic",
        ("FIX.4.2", "321", "0"): "0 .. RequestSecurityIdentityAndSpecifications",
        ("FIX.4.2", "321", "1"): "1 .. RequestSecurityIdentityForSpecifications",
        ("FIX.4.2", "321", "2"): "2 .. RequestListSecurityTypes",
        ("FIX.4.2", "321", "3"): "3 .. RequestListSecurities",
        ("FIX.4.2", "323", "1"): "1 .. AcceptAsIs",
        ("FIX.4.2", "323", "2"): "2 .. AcceptWithRevisions",
        ("FIX.4.2", "323", "3"): "3 .. ListOfSecurityTypesReturnedPerRequest",
        ("FIX.4.2", "323", "4"): "4 .. ListOfSecuritiesReturnedPerRequest",
        ("FIX.4.2", "323", "5"): "5 .. RejectSecurityProposal",
        ("FIX.4.2", "323", "6"): "6 .. CannotMatchSelectionCriteria",
        ("FIX.4.2", "325", "N"): "N .. MessageIsBeingSentAsAResultOfAPriorRequest",
        ("FIX.4.2", "325", "Y"): "Y .. MessageIsBeingSentUnsolicited",
        ("FIX.4.2", "326", "1"): "1 .. OpeningDelay",
        ("FIX.4.2", "326", "10"): "10 .. MarketOnCloseImbalanceSell",
        ("FIX.4.2", "326", "12"): "12 .. NoMarketImbalance",
        ("FIX.4.2", "326", "13"): "13 .. NoMarketOnCloseImbalance",
        ("FIX.4.2", "326", "14"): "14 .. ITSPreOpening",
        ("FIX.4.2", "326", "15"): "15 .. NewPriceIndication",
        ("FIX.4.2", "326", "16"): "16 .. TradeDisseminationTime",
        ("FIX.4.2", "326", "17"): "17 .. ReadyToTrade",
        ("FIX.4.2", "326", "18"): "18 .. NotAvailableForTrading",
        ("FIX.4.2", "326", "19"): "19 .. NotTradedOnThisMarket",
        ("FIX.4.2", "326", "2"): "2 .. TradingHalt",
        ("FIX.4.2", "326", "20"): "20 .. UnknownOrInvalid",
        ("FIX.4.2", "326", "3"): "3 .. Resume",
        ("FIX.4.2", "326", "4"): "4 .. NoOpen",
        ("FIX.4.2", "326", "5"): "5 .. PriceIndication",
        ("FIX.4.2", "326", "6"): "6 .. TradingRangeIndication",
        ("FIX.4.2", "326", "7"): "7 .. MarketImbalanceBuy",
        ("FIX.4.2", "326", "8"): "8 .. MarketImbalanceSell",
        ("FIX.4.2", "326", "9"): "9 .. MarketOnCloseImbalanceBuy",
        ("FIX.4.2", "327", "D"): "D .. NewsDissemination",
        ("FIX.4.2", "327", "E"): "E .. OrderInflux",
        ("FIX.4.2", "327", "I"): "I .. OrderImbalance",
        ("FIX.4.2", "327", "M"): "M .. AdditionalInformation",
        ("FIX.4.2", "327", "P"): "P .. NewsPending",
        ("FIX.4.2", "327", "X"): "X .. EquipmentChangeover",
        ("FIX.4.2", "328", "N"): "N .. HaltWasNotRelatedToAHaltOfTheCommonStock",
        ("FIX.4.2", "328", "Y"): "Y .. HaltWasDueToCommonStockBeingHalted",
        ("FIX.4.2", "329", "N"): "N .. NotRelatedToSecurityHalt",
        ("FIX.4.2", "329", "Y"): "Y .. RelatedToSecurityHalt",
        ("FIX.4.2", "334", "1"): "1 .. Cancel",
        ("FIX.4.2", "334", "2"): "2 .. Error",
        ("FIX.4.2", "334", "3"): "3 .. Correction",
        ("FIX.4.2", "338", "1"): "1 .. Electronic",
        ("FIX.4.2", "338", "2"): "2 .. OpenOutcry",
        ("FIX.4.2", "338", "3"): "3 .. TwoParty",
        ("FIX.4.2", "339", "1"): "1 .. Testing",
        ("FIX.4.2", "339", "2"): "2 .. Simulated",
        ("FIX.4.2", "339", "3"): "3 .. Production",
        ("FIX.4.2", "340", "1"): "1 .. Halted",
        ("FIX.4.2", "340", "2"): "2 .. Open",
        ("FIX.4.2", "340", "3"): "3 .. Closed",
        ("FIX.4.2", "340", "4"): "4 .. PreOpen",
        ("FIX.4.2", "340", "5"): "5 .. PreClose",
        ("FIX.4.2", "347", "EUC-JP"): "EUC-JP .. EUCJP",
        ("FIX.4.2", "347", "ISO-2022-JP"): "ISO-2022-JP .. ISO2022JP",
        ("FIX.4.2", "347", "Shift_JIS"): "Shift_JIS .. ShiftJIS",
        ("FIX.4.2", "347", "UTF-8"): "UTF-8 .. UTF8",
        ("FIX.4.2", "368", "1"): "1 .. UnknownSymbol",
        ("FIX.4.2", "368", "2"): "2 .. Exchange",
        ("FIX.4.2", "368", "3"): "3 .. QuoteExceedsLimit",
        ("FIX.4.2", "368", "4"): "4 .. TooLateToEnter",
        ("FIX.4.2", "368", "5"): "5 .. UnknownQuote",
        ("FIX.4.2", "368", "6"): "6 .. DuplicateQuote",
        ("FIX.4.2", "368", "7"): "7 .. InvalidBidAskSpread",
        ("FIX.4.2", "368", "8"): "8 .. InvalidPrice",
        ("FIX.4.2", "368", "9"): "9 .. NotAuthorizedToQuoteSecurity",
        ("FIX.4.2", "373", "0"): "0 .. InvalidTagNumber",
        ("FIX.4.2", "373", "1"): "1 .. RequiredTagMissing",
        ("FIX.4.2", "373", "10"): "10 .. SendingTimeAccuracyProblem",
        ("FIX.4.2", "373", "11"): "11 .. InvalidMsgType",
        ("FIX.4.2", "373", "2"): "2 .. TagNotDefinedForThisMessageType",
        ("FIX.4.2", "373", "3"): "3 .. UndefinedTag",
        ("FIX.4.2", "373", "4"): "4 .. TagSpecifiedWithoutAValue",
        ("FIX.4.2", "373", "5"): "5 .. ValueIsIncorrect",
        ("FIX.4.2", "373", "6"): "6 .. IncorrectDataFormatForValue",
        ("FIX.4.2", "373", "7"): "7 .. DecryptionProblem",
        ("FIX.4.2", "373", "8"): "8 .. SignatureProblem",
        ("FIX.4.2", "373", "9"): "9 .. CompIDProblem",
        ("FIX.4.2", "374", "C"): "C .. Cancel",
        ("FIX.4.2", "374", "N"): "N .. New",
        ("FIX.4.2", "377", "N"): "N .. WasNotSolicited",
        ("FIX.4.2", "377", "Y"): "Y .. WasSolicited",
        ("FIX.4.2", "378", "0"): "0 .. GTCorporateAction",
        ("FIX.4.2", "378", "1"): "1 .. GTRenewal",
        ("FIX.4.2", "378", "2"): "2 .. VerbalChange",
        ("FIX.4.2", "378", "3"): "3 .. RepricingOfOrder",
        ("FIX.4.2", "378", "4"): "4 .. BrokerOption",
        ("FIX.4.2", "378", "5"): "5 .. PartialDeclineOfOrderQty",
        ("FIX.4.2", "380", "0"): "0 .. Other",
        ("FIX.4.2", "380", "1"): "1 .. UnknownID",
        ("FIX.4.2", "380", "2"): "2 .. UnknownSecurity",
        ("FIX.4.2", "380", "3"): "3 .. UnsupportedMessageType",
        ("FIX.4.2", "380", "4"): "4 .. ApplicationNotAvailable",
        ("FIX.4.2", "380", "5"): "5 .. ConditionallyRequiredFieldMissing",
        ("FIX.4.2", "385", "R"): "R .. Receive",
        ("FIX.4.2", "385", "S"): "S .. Send",
        ("FIX.4.2", "388", "0"): "0 .. RelatedToDisplayedPrice",
        ("FIX.4.2", "388", "1"): "1 .. RelatedToMarketPrice",
        ("FIX.4.2", "388", "2"): "2 .. RelatedToPrimaryPrice",
        ("FIX.4.2", "388", "3"): "3 .. RelatedToLocalPrimaryPrice",
        ("FIX.4.2", "388", "4"): "4 .. RelatedToMidpointPrice",
        ("FIX.4.2", "388", "5"): "5 .. RelatedToLastTradePrice",
        ("FIX.4.2", "409", "1"): "1 .. FiveDayMovingAverage",
        ("FIX.4.2", "409", "2"): "2 .. TwentyDayMovingAverage",
        ("FIX.4.2", "409", "3"): "3 .. NormalMarketSize",
        ("FIX.4.2", "409", "4"): "4 .. Other",
        ("FIX.4.2", "411", "N"): "N .. False",
        ("FIX.4.2", "411", "Y"): "Y .. True",
        ("FIX.4.2", "414", "1"): "1 .. BuySideRequests",
        ("FIX.4.2", "414", "2"): "2 .. SellSideSends",
        ("FIX.4.2", "414", "3"): "3 .. RealTimeExecutionReports",
        ("FIX.4.2", "416", "1"): "1 .. Net",
        ("FIX.4.2", "416", "2"): "2 .. Gross",
        ("FIX.4.2", "418", "A"): "A .. Agency",
        ("FIX.4.2", "418", "G"): "G .. VWAPGuarantee",
        ("FIX.4.2", "418", "J"): "J .. GuaranteedClose",
        ("FIX.4.2", "418", "R"): "R .. RiskTrade",
        ("FIX.4.2", "419", "2"): "2 .. ClosingPriceAtMorningSession",
        ("FIX.4.2", "419", "3"): "3 .. ClosingPrice",
        ("FIX.4.2", "419", "4"): "4 .. CurrentPrice",
        ("FIX.4.2", "419", "5"): "5 .. SQ",
        ("FIX.4.2", "419", "6"): "6 .. VWAPThroughADay",
        ("FIX.4.2", "419", "7"): "7 .. VWAPThroughAMorningSession",
        ("FIX.4.2", "419", "8"): "8 .. VWAPThroughAnAfternoonSession",
        ("FIX.4.2", "419", "9"): "9 .. VWAPThroughADayExcept",
        ("FIX.4.2", "419", "A"): "A .. VWAPThroughAMorningSessionExcept",
        ("FIX.4.2", "419", "B"): "B .. VWAPThroughAnAfternoonSessionExcept",
        ("FIX.4.2", "419", "C"): "C .. Strike",
        ("FIX.4.2", "419", "D"): "D .. Open",
        ("FIX.4.2", "419", "Z"): "Z .. Others",
        ("FIX.4.2", "423", "1"): "1 .. Percentage",
        ("FIX.4.2", "423", "2"): "2 .. PerUnit",
        ("FIX.4.2", "423", "3"): "3 .. FixedAmount",
        ("FIX.4.2", "427", "0"): "0 .. BookOutAllTradesOnDayOfExecution",
        ("FIX.4.2", "427", "1"): "1 .. AccumulateUntilFilledOrExpired",
        ("FIX.4.2", "427", "2"): "2 .. AccumulateUntilVerballlyNotifiedOtherwise",
        ("FIX.4.2", "430", "1"): "1 .. Net",
        ("FIX.4.2", "430", "2"): "2 .. Gross",
        ("FIX.4.2", "433", "1"): "1 .. Immediate",
        ("FIX.4.2", "433", "2"): "2 .. WaitForInstruction",
        ("FIX.4.2", "434", "1"): "1 .. OrderCancelRequest",
        ("FIX.4.2", "434", "2"): "2 .. OrderCancel",
        ("FIX.4.2", "442", "1"): "1 .. SingleSecurity",
        ("FIX.4.2", "442", "2"): "2 .. IndividualLegOfAMultiLegSecurity",
        ("FIX.4.2", "442", "3"): "3 .. MultiLegSecurity",
        ("FIX.4.4", "4", "B"): "B .. Buy",
        ("FIX.4.4", "4", "S"): "S .. Sell",
        ("FIX.4.4", "4", "X"): "X .. Cross",
        ("FIX.4.4", "4", "T"): "T .. Trade",
        ("FIX.4.4", "5", "N"): "N .. New",
        ("FIX.4.4", "5", "C"): "C .. Cancel",
        ("FIX.4.4", "5", "R"): "R .. Replace",
        ("FIX.4.4", "13", "1"): "1 .. PerUnit",
        ("FIX.4.4", "13", "2"): "2 .. Percent",
        ("FIX.4.4", "13", "3"): "3 .. Absolute",
        ("FIX.4.4", "13", "4"): "4 .. PercentageWaivedCashDiscount",
        ("FIX.4.4", "13", "5"): "5 .. PercentageWaivedEnhancedUnits",
        ("FIX.4.4", "13", "6"): "6 .. PointsPerBondOrContract",
        ("FIX.4.4", "18", "1"): "1 .. NotHeld",
        ("FIX.4.4", "18", "2"): "2 .. Work",
        ("FIX.4.4", "18", "3"): "3 .. GoAlong",
        ("FIX.4.4", "18", "4"): "4 .. OverTheDay",
        ("FIX.4.4", "18", "5"): "5 .. Held",
        ("FIX.4.4", "18", "6"): "6 .. ParticipateDoNotInitiate",
        ("FIX.4.4", "18", "7"): "7 .. StrictScale",
        ("FIX.4.4", "18", "8"): "8 .. TryToScale",
        ("FIX.4.4", "18", "9"): "9 .. StayOnBidSide",
        ("FIX.4.4", "18", "0"): "0 .. StayOnOfferSide",
        ("FIX.4.4", "18", "A"): "A .. NoCross",
        ("FIX.4.4", "18", "B"): "B .. OKToCross",
        ("FIX.4.4", "18", "C"): "C .. CallFirst",
        ("FIX.4.4", "18", "D"): "D .. PercentOfVolume",
        ("FIX.4.4", "18", "E"): "E .. DoNotIncrease",
        ("FIX.4.4", "18", "F"): "F .. DoNotReduce",
        ("FIX.4.4", "18", "G"): "G .. AllOrNone",
        ("FIX.4.4", "18", "H"): "H .. ReinstateOnSystemFailure",
        ("FIX.4.4", "18", "I"): "I .. InstitutionsOnly",
        ("FIX.4.4", "18", "J"): "J .. ReinstateOnTradingHalt",
        ("FIX.4.4", "18", "K"): "K .. CancelOnTradingHalt",
        ("FIX.4.4", "18", "L"): "L .. LastPeg",
        ("FIX.4.4", "18", "M"): "M .. MidPricePeg",
        ("FIX.4.4", "18", "N"): "N .. NonNegotiable",
        ("FIX.4.4", "18", "O"): "O .. OpeningPeg",
        ("FIX.4.4", "18", "P"): "P .. MarketPeg",
        ("FIX.4.4", "18", "Q"): "Q .. CancelOnSystemFailure",
        ("FIX.4.4", "18", "R"): "R .. PrimaryPeg",
        ("FIX.4.4", "18", "S"): "S .. Suspend",
        ("FIX.4.4", "18", "U"): "U .. CustomerDisplayInstruction",
        ("FIX.4.4", "18", "V"): "V .. Netting",
        ("FIX.4.4", "18", "W"): "W .. PegToVWAP",
        ("FIX.4.4", "18", "X"): "X .. TradeAlong",
        ("FIX.4.4", "18", "Y"): "Y .. TryToStop",
        ("FIX.4.4", "18", "Z"): "Z .. CancelIfNotBest",
        ("FIX.4.4", "18", "a"): "a .. TrailingStopPeg",
        ("FIX.4.4", "18", "b"): "b .. StrictLimit",
        ("FIX.4.4", "18", "c"): "c .. IgnorePriceValidityChecks",
        ("FIX.4.4", "18", "d"): "d .. PegToLimitPrice",
        ("FIX.4.4", "18", "e"): "e .. WorkToTargetStrategy",
        ("FIX.4.4", "21", "1"): "1 .. AutomatedExecutionNoIntervention",
        ("FIX.4.4", "21", "2"): "2 .. AutomatedExecutionInterventionOK",
        ("FIX.4.4", "21", "3"): "3 .. ManualOrder",
        ("FIX.4.4", "22", "1"): "1 .. CUSIP",
        ("FIX.4.4", "22", "2"): "2 .. SEDOL",
        ("FIX.4.4", "22", "3"): "3 .. QUIK",
        ("FIX.4.4", "22", "4"): "4 .. ISINNumber",
        ("FIX.4.4", "22", "5"): "5 .. RICCode",
        ("FIX.4.4", "22", "6"): "6 .. ISOCurrencyCode",
        ("FIX.4.4", "22", "7"): "7 .. ISOCountryCode",
        ("FIX.4.4", "22", "8"): "8 .. ExchangeSymbol",
        ("FIX.4.4", "22", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.4.4", "22", "A"): "A .. BloombergSymbol",
        ("FIX.4.4", "22", "B"): "B .. Wertpapier",
        ("FIX.4.4", "22", "C"): "C .. Dutch",
        ("FIX.4.4", "22", "D"): "D .. Valoren",
        ("FIX.4.4", "22", "E"): "E .. Sicovam",
        ("FIX.4.4", "22", "F"): "F .. Belgian",
        ("FIX.4.4", "22", "G"): "G .. Common",
        ("FIX.4.4", "22", "H"): "H .. ClearingHouse",
        ("FIX.4.4", "22", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.4.4", "22", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.4.4", "25", "L"): "L .. Low",
        ("FIX.4.4", "25", "M"): "M .. Medium",
        ("FIX.4.4", "25", "H"): "H .. High",
        ("FIX.4.4", "27", "S"): "S .. Small",
        ("FIX.4.4", "27", "M"): "M .. Medium",
        ("FIX.4.4", "27", "L"): "L .. Large",
        ("FIX.4.4", "28", "N"): "N .. New",
        ("FIX.4.4", "28", "C"): "C .. Cancel",
        ("FIX.4.4", "28", "R"): "R .. Replace",
        ("FIX.4.4", "29", "1"): "1 .. Agent",
        ("FIX.4.4", "29", "2"): "2 .. CrossAsAgent",
        ("FIX.4.4", "29", "3"): "3 .. CrossAsPrincipal",
        ("FIX.4.4", "29", "4"): "4 .. Principal",
        ("FIX.4.4", "35", "0"): "0 .. Heartbeat",
        ("FIX.4.4", "35", "1"): "1 .. TestRequest",
        ("FIX.4.4", "35", "2"): "2 .. ResendRequest",
        ("FIX.4.4", "35", "3"): "3 .. Reject",
        ("FIX.4.4", "35", "4"): "4 .. SequenceReset",
        ("FIX.4.4", "35", "5"): "5 .. Logout",
        ("FIX.4.4", "35", "6"): "6 .. IOI",
        ("FIX.4.4", "35", "7"): "7 .. Advertisement",
        ("FIX.4.4", "35", "8"): "8 .. ExecutionReport",
        ("FIX.4.4", "35", "9"): "9 .. OrderCancelReject",
        ("FIX.4.4", "35", "A"): "A .. Logon",
        ("FIX.4.4", "35", "B"): "B .. News",
        ("FIX.4.4", "35", "C"): "C .. Email",
        ("FIX.4.4", "35", "D"): "D .. NewOrderSingle",
        ("FIX.4.4", "35", "E"): "E .. NewOrderList",
        ("FIX.4.4", "35", "F"): "F .. OrderCancelRequest",
        ("FIX.4.4", "35", "G"): "G .. OrderCancelReplaceRequest",
        ("FIX.4.4", "35", "H"): "H .. OrderStatusRequest",
        ("FIX.4.4", "35", "J"): "J .. AllocationInstruction",
        ("FIX.4.4", "35", "K"): "K .. ListCancelRequest",
        ("FIX.4.4", "35", "L"): "L .. ListExecute",
        ("FIX.4.4", "35", "M"): "M .. ListStatusRequest",
        ("FIX.4.4", "35", "N"): "N .. ListStatus",
        ("FIX.4.4", "35", "P"): "P .. AllocationInstructionAck",
        ("FIX.4.4", "35", "Q"): "Q .. DontKnowTrade",
        ("FIX.4.4", "35", "R"): "R .. QuoteRequest",
        ("FIX.4.4", "35", "S"): "S .. Quote",
        ("FIX.4.4", "35", "T"): "T .. SettlementInstructions",
        ("FIX.4.4", "35", "V"): "V .. MarketDataRequest",
        ("FIX.4.4", "35", "W"): "W .. MarketDataSnapshotFullRefresh",
        ("FIX.4.4", "35", "X"): "X .. MarketDataIncrementalRefresh",
        ("FIX.4.4", "35", "Y"): "Y .. MarketDataRequestReject",
        ("FIX.4.4", "35", "Z"): "Z .. QuoteCancel",
        ("FIX.4.4", "35", "a"): "a .. QuoteStatusRequest",
        ("FIX.4.4", "35", "b"): "b .. MassQuoteAcknowledgement",
        ("FIX.4.4", "35", "c"): "c .. SecurityDefinitionRequest",
        ("FIX.4.4", "35", "d"): "d .. SecurityDefinition",
        ("FIX.4.4", "35", "e"): "e .. SecurityStatusRequest",
        ("FIX.4.4", "35", "f"): "f .. SecurityStatus",
        ("FIX.4.4", "35", "g"): "g .. TradingSessionStatusRequest",
        ("FIX.4.4", "35", "h"): "h .. TradingSessionStatus",
        ("FIX.4.4", "35", "i"): "i .. MassQuote",
        ("FIX.4.4", "35", "j"): "j .. BusinessMessageReject",
        ("FIX.4.4", "35", "k"): "k .. BidRequest",
        ("FIX.4.4", "35", "l"): "l .. BidResponse",
        ("FIX.4.4", "35", "m"): "m .. ListStrikePrice",
        ("FIX.4.4", "35", "n"): "n .. XMLNonFIX",
        ("FIX.4.4", "35", "o"): "o .. RegistrationInstructions",
        ("FIX.4.4", "35", "p"): "p .. RegistrationInstructionsResponse",
        ("FIX.4.4", "35", "q"): "q .. OrderMassCancelRequest",
        ("FIX.4.4", "35", "r"): "r .. OrderMassCancelReport",
        ("FIX.4.4", "35", "s"): "s .. NewOrderCross",
        ("FIX.4.4", "35", "t"): "t .. CrossOrderCancelReplaceRequest",
        ("FIX.4.4", "35", "u"): "u .. CrossOrderCancelRequest",
        ("FIX.4.4", "35", "v"): "v .. SecurityTypeRequest",
        ("FIX.4.4", "35", "w"): "w .. SecurityTypes",
        ("FIX.4.4", "35", "x"): "x .. SecurityListRequest",
        ("FIX.4.4", "35", "y"): "y .. SecurityList",
        ("FIX.4.4", "35", "z"): "z .. DerivativeSecurityListRequest",
        ("FIX.4.4", "35", "AA"): "AA .. DerivativeSecurityList",
        ("FIX.4.4", "35", "AB"): "AB .. NewOrderMultileg",
        ("FIX.4.4", "35", "AC"): "AC .. MultilegOrderCancelReplace",
        ("FIX.4.4", "35", "AD"): "AD .. TradeCaptureReportRequest",
        ("FIX.4.4", "35", "AE"): "AE .. TradeCaptureReport",
        ("FIX.4.4", "35", "AF"): "AF .. OrderMassStatusRequest",
        ("FIX.4.4", "35", "AG"): "AG .. QuoteRequestReject",
        ("FIX.4.4", "35", "AH"): "AH .. RFQRequest",
        ("FIX.4.4", "35", "AI"): "AI .. QuoteStatusReport",
        ("FIX.4.4", "35", "AJ"): "AJ .. QuoteResponse",
        ("FIX.4.4", "35", "AK"): "AK .. Confirmation",
        ("FIX.4.4", "35", "AL"): "AL .. PositionMaintenanceRequest",
        ("FIX.4.4", "35", "AM"): "AM .. PositionMaintenanceReport",
        ("FIX.4.4", "35", "AN"): "AN .. RequestForPositions",
        ("FIX.4.4", "35", "AO"): "AO .. RequestForPositionsAck",
        ("FIX.4.4", "35", "AP"): "AP .. PositionReport",
        ("FIX.4.4", "35", "AQ"): "AQ .. TradeCaptureReportRequestAck",
        ("FIX.4.4", "35", "AR"): "AR .. TradeCaptureReportAck",
        ("FIX.4.4", "35", "AS"): "AS .. AllocationReport",
        ("FIX.4.4", "35", "AT"): "AT .. AllocationReportAck",
        ("FIX.4.4", "35", "AU"): "AU .. ConfirmationAck",
        ("FIX.4.4", "35", "AV"): "AV .. SettlementInstructionRequest",
        ("FIX.4.4", "35", "AW"): "AW .. AssignmentReport",
        ("FIX.4.4", "35", "AX"): "AX .. CollateralRequest",
        ("FIX.4.4", "35", "AY"): "AY .. CollateralAssignment",
        ("FIX.4.4", "35", "AZ"): "AZ .. CollateralResponse",
        ("FIX.4.4", "35", "BA"): "BA .. CollateralReport",
        ("FIX.4.4", "35", "BB"): "BB .. CollateralInquiry",
        ("FIX.4.4", "35", "BC"): "BC .. NetworkCounterpartySystemStatusRequest",
        ("FIX.4.4", "35", "BD"): "BD .. NetworkCounterpartySystemStatusResponse",
        ("FIX.4.4", "35", "BE"): "BE .. UserRequest",
        ("FIX.4.4", "35", "BF"): "BF .. UserResponse",
        ("FIX.4.4", "35", "BG"): "BG .. CollateralInquiryAck",
        ("FIX.4.4", "35", "BH"): "BH .. ConfirmationRequest",
        ("FIX.4.4", "39", "0"): "0 .. New",
        ("FIX.4.4", "39", "1"): "1 .. PartiallyFilled",
        ("FIX.4.4", "39", "2"): "2 .. Filled",
        ("FIX.4.4", "39", "3"): "3 .. DoneForDay",
        ("FIX.4.4", "39", "4"): "4 .. Canceled",
        ("FIX.4.4", "39", "6"): "6 .. PendingCancel",
        ("FIX.4.4", "39", "7"): "7 .. Stopped",
        ("FIX.4.4", "39", "8"): "8 .. Rejected",
        ("FIX.4.4", "39", "9"): "9 .. Suspended",
        ("FIX.4.4", "39", "A"): "A .. PendingNew",
        ("FIX.4.4", "39", "B"): "B .. Calculated",
        ("FIX.4.4", "39", "C"): "C .. Expired",
        ("FIX.4.4", "39", "D"): "D .. AcceptedForBidding",
        ("FIX.4.4", "39", "E"): "E .. PendingReplace",
        ("FIX.4.4", "40", "1"): "1 .. Market",
        ("FIX.4.4", "40", "2"): "2 .. Limit",
        ("FIX.4.4", "40", "3"): "3 .. Stop",
        ("FIX.4.4", "40", "4"): "4 .. StopLimit",
        ("FIX.4.4", "40", "6"): "6 .. WithOrWithout",
        ("FIX.4.4", "40", "7"): "7 .. LimitOrBetter",
        ("FIX.4.4", "40", "8"): "8 .. LimitWithOrWithout",
        ("FIX.4.4", "40", "9"): "9 .. OnBasis",
        ("FIX.4.4", "40", "D"): "D .. PreviouslyQuoted",
        ("FIX.4.4", "40", "E"): "E .. PreviouslyIndicated",
        ("FIX.4.4", "40", "G"): "G .. ForexSwap",
        ("FIX.4.4", "40", "I"): "I .. Funari",
        ("FIX.4.4", "40", "J"): "J .. MarketIfTouched",
        ("FIX.4.4", "40", "K"): "K .. MarketWithLeftOverAsLimit",
        ("FIX.4.4", "40", "L"): "L .. PreviousFundValuationPoint",
        ("FIX.4.4", "40", "M"): "M .. NextFundValuationPoint",
        ("FIX.4.4", "40", "P"): "P .. Pegged",
        ("FIX.4.4", "43", "Y"): "Y .. PossibleDuplicate",
        ("FIX.4.4", "43", "N"): "N .. OriginalTransmission",
        ("FIX.4.4", "54", "1"): "1 .. Buy",
        ("FIX.4.4", "54", "2"): "2 .. Sell",
        ("FIX.4.4", "54", "3"): "3 .. BuyMinus",
        ("FIX.4.4", "54", "4"): "4 .. SellPlus",
        ("FIX.4.4", "54", "5"): "5 .. SellShort",
        ("FIX.4.4", "54", "6"): "6 .. SellShortExempt",
        ("FIX.4.4", "54", "7"): "7 .. Undisclosed",
        ("FIX.4.4", "54", "8"): "8 .. Cross",
        ("FIX.4.4", "54", "9"): "9 .. CrossShort",
        ("FIX.4.4", "54", "A"): "A .. CrossShortExempt",
        ("FIX.4.4", "54", "B"): "B .. AsDefined",
        ("FIX.4.4", "54", "C"): "C .. Opposite",
        ("FIX.4.4", "54", "D"): "D .. Subscribe",
        ("FIX.4.4", "54", "E"): "E .. Redeem",
        ("FIX.4.4", "54", "F"): "F .. Lend",
        ("FIX.4.4", "54", "G"): "G .. Borrow",
        ("FIX.4.4", "59", "0"): "0 .. Day",
        ("FIX.4.4", "59", "1"): "1 .. GoodTillCancel",
        ("FIX.4.4", "59", "2"): "2 .. AtTheOpening",
        ("FIX.4.4", "59", "3"): "3 .. ImmediateOrCancel",
        ("FIX.4.4", "59", "4"): "4 .. FillOrKill",
        ("FIX.4.4", "59", "5"): "5 .. GoodTillCrossing",
        ("FIX.4.4", "59", "6"): "6 .. GoodTillDate",
        ("FIX.4.4", "59", "7"): "7 .. AtTheClose",
        ("FIX.4.4", "61", "0"): "0 .. Normal",
        ("FIX.4.4", "61", "1"): "1 .. Flash",
        ("FIX.4.4", "61", "2"): "2 .. Background",
        ("FIX.4.4", "63", "0"): "0 .. Regular",
        ("FIX.4.4", "63", "1"): "1 .. Cash",
        ("FIX.4.4", "63", "2"): "2 .. NextDay",
        ("FIX.4.4", "63", "3"): "3 .. TPlus2",
        ("FIX.4.4", "63", "4"): "4 .. TPlus3",
        ("FIX.4.4", "63", "5"): "5 .. TPlus4",
        ("FIX.4.4", "63", "6"): "6 .. Future",
        ("FIX.4.4", "63", "7"): "7 .. WhenAndIfIssued",
        ("FIX.4.4", "63", "8"): "8 .. SellersOption",
        ("FIX.4.4", "63", "9"): "9 .. TPlus5",
        ("FIX.4.4", "71", "0"): "0 .. New",
        ("FIX.4.4", "71", "1"): "1 .. Replace",
        ("FIX.4.4", "71", "2"): "2 .. Cancel",
        ("FIX.4.4", "77", "O"): "O .. Open",
        ("FIX.4.4", "77", "C"): "C .. Close",
        ("FIX.4.4", "77", "R"): "R .. Rolled",
        ("FIX.4.4", "77", "F"): "F .. FIFO",
        ("FIX.4.4", "81", "0"): "0 .. Regular",
        ("FIX.4.4", "81", "1"): "1 .. SoftDollar",
        ("FIX.4.4", "81", "2"): "2 .. StepIn",
        ("FIX.4.4", "81", "3"): "3 .. StepOut",
        ("FIX.4.4", "81", "4"): "4 .. SoftDollarStepIn",
        ("FIX.4.4", "81", "5"): "5 .. SoftDollarStepOut",
        ("FIX.4.4", "81", "6"): "6 .. PlanSponsor",
        ("FIX.4.4", "87", "0"): "0 .. Accepted",
        ("FIX.4.4", "87", "1"): "1 .. BlockLevelReject",
        ("FIX.4.4", "87", "2"): "2 .. AccountLevelReject",
        ("FIX.4.4", "87", "3"): "3 .. Received",
        ("FIX.4.4", "87", "4"): "4 .. Incomplete",
        ("FIX.4.4", "87", "5"): "5 .. RejectedByIntermediary",
        ("FIX.4.4", "88", "0"): "0 .. UnknownAccount",
        ("FIX.4.4", "88", "1"): "1 .. IncorrectQuantity",
        ("FIX.4.4", "88", "2"): "2 .. IncorrectAveragegPrice",
        ("FIX.4.4", "88", "3"): "3 .. UnknownExecutingBrokerMnemonic",
        ("FIX.4.4", "88", "4"): "4 .. CommissionDifference",
        ("FIX.4.4", "88", "5"): "5 .. UnknownOrderID",
        ("FIX.4.4", "88", "6"): "6 .. UnknownListID",
        ("FIX.4.4", "88", "7"): "7 .. OtherSeeText",
        ("FIX.4.4", "88", "8"): "8 .. IncorrectAllocatedQuantity",
        ("FIX.4.4", "88", "9"): "9 .. CalculationDifference",
        ("FIX.4.4", "88", "10"): "10 .. UnknownOrStaleExecID",
        ("FIX.4.4", "88", "11"): "11 .. MismatchedData",
        ("FIX.4.4", "88", "12"): "12 .. UnknownClOrdID",
        ("FIX.4.4", "88", "13"): "13 .. WarehouseRequestRejected",
        ("FIX.4.4", "94", "0"): "0 .. New",
        ("FIX.4.4", "94", "1"): "1 .. Reply",
        ("FIX.4.4", "94", "2"): "2 .. AdminReply",
        ("FIX.4.4", "97", "Y"): "Y .. PossibleResend",
        ("FIX.4.4", "97", "N"): "N .. OriginalTransmission",
        ("FIX.4.4", "98", "0"): "0 .. None",
        ("FIX.4.4", "98", "1"): "1 .. PKCS",
        ("FIX.4.4", "98", "2"): "2 .. DES",
        ("FIX.4.4", "98", "3"): "3 .. PKCSDES",
        ("FIX.4.4", "98", "4"): "4 .. PGPDES",
        ("FIX.4.4", "98", "5"): "5 .. PGPDESMD5",
        ("FIX.4.4", "98", "6"): "6 .. PEM",
        ("FIX.4.4", "102", "0"): "0 .. TooLateToCancel",
        ("FIX.4.4", "102", "1"): "1 .. UnknownOrder",
        ("FIX.4.4", "102", "2"): "2 .. BrokerCredit",
        ("FIX.4.4", "102", "3"): "3 .. OrderAlreadyInPendingStatus",
        ("FIX.4.4", "102", "4"): "4 .. UnableToProcessOrderMassCancelRequest",
        ("FIX.4.4", "102", "5"): "5 .. OrigOrdModTime",
        ("FIX.4.4", "102", "6"): "6 .. DuplicateClOrdID",
        ("FIX.4.4", "102", "99"): "99 .. Other",
        ("FIX.4.4", "103", "0"): "0 .. BrokerCredit",
        ("FIX.4.4", "103", "1"): "1 .. UnknownSymbol",
        ("FIX.4.4", "103", "2"): "2 .. ExchangeClosed",
        ("FIX.4.4", "103", "3"): "3 .. OrderExceedsLimit",
        ("FIX.4.4", "103", "4"): "4 .. TooLateToEnter",
        ("FIX.4.4", "103", "5"): "5 .. UnknownOrder",
        ("FIX.4.4", "103", "6"): "6 .. DuplicateOrder",
        ("FIX.4.4", "103", "7"): "7 .. DuplicateOfAVerballyCommunicatedOrder",
        ("FIX.4.4", "103", "8"): "8 .. StaleOrder",
        ("FIX.4.4", "103", "9"): "9 .. TradeAlongRequired",
        ("FIX.4.4", "103", "10"): "10 .. InvalidInvestorID",
        ("FIX.4.4", "103", "11"): "11 .. UnsupportedOrderCharacteristic",
        ("FIX.4.4", "103", "13"): "13 .. IncorrectQuantity",
        ("FIX.4.4", "103", "14"): "14 .. IncorrectAllocatedQuantity",
        ("FIX.4.4", "103", "15"): "15 .. UnknownAccount",
        ("FIX.4.4", "103", "99"): "99 .. Other",
        ("FIX.4.4", "104", "A"): "A .. AllOrNone",
        ("FIX.4.4", "104", "B"): "B .. MarketOnClose",
        ("FIX.4.4", "104", "C"): "C .. AtTheClose",
        ("FIX.4.4", "104", "D"): "D .. VWAP",
        ("FIX.4.4", "104", "I"): "I .. InTouchWith",
        ("FIX.4.4", "104", "L"): "L .. Limit",
        ("FIX.4.4", "104", "M"): "M .. MoreBehind",
        ("FIX.4.4", "104", "O"): "O .. AtTheOpen",
        ("FIX.4.4", "104", "P"): "P .. TakingAPosition",
        ("FIX.4.4", "104", "Q"): "Q .. AtTheMarket",
        ("FIX.4.4", "104", "R"): "R .. ReadyToTrade",
        ("FIX.4.4", "104", "S"): "S .. PortfolioShown",
        ("FIX.4.4", "104", "T"): "T .. ThroughTheDay",
        ("FIX.4.4", "104", "V"): "V .. Versus",
        ("FIX.4.4", "104", "W"): "W .. Indication",
        ("FIX.4.4", "104", "X"): "X .. CrossingOpportunity",
        ("FIX.4.4", "104", "Y"): "Y .. AtTheMidpoint",
        ("FIX.4.4", "104", "Z"): "Z .. PreOpen",
        ("FIX.4.4", "113", "Y"): "Y .. ReceiverReports",
        ("FIX.4.4", "113", "N"): "N .. SenderReports",
        ("FIX.4.4", "114", "Y"): "Y .. Yes",
        ("FIX.4.4", "114", "N"): "N .. No",
        ("FIX.4.4", "121", "Y"): "Y .. ExecuteForexAfterSecurityTrade",
        ("FIX.4.4", "121", "N"): "N .. DoNotExecuteForexAfterSecurityTrade",
        ("FIX.4.4", "123", "Y"): "Y .. GapFillMessage",
        ("FIX.4.4", "123", "N"): "N .. SequenceReset",
        ("FIX.4.4", "127", "A"): "A .. UnknownSymbol",
        ("FIX.4.4", "127", "B"): "B .. WrongSide",
        ("FIX.4.4", "127", "C"): "C .. QuantityExceedsOrder",
        ("FIX.4.4", "127", "D"): "D .. NoMatchingOrder",
        ("FIX.4.4", "127", "E"): "E .. PriceExceedsLimit",
        ("FIX.4.4", "127", "F"): "F .. CalculationDifference",
        ("FIX.4.4", "127", "Z"): "Z .. Other",
        ("FIX.4.4", "130", "Y"): "Y .. Natural",
        ("FIX.4.4", "130", "N"): "N .. NotNatural",
        ("FIX.4.4", "139", "1"): "1 .. Regulatory",
        ("FIX.4.4", "139", "2"): "2 .. Tax",
        ("FIX.4.4", "139", "3"): "3 .. LocalCommission",
        ("FIX.4.4", "139", "4"): "4 .. ExchangeFees",
        ("FIX.4.4", "139", "5"): "5 .. Stamp",
        ("FIX.4.4", "139", "6"): "6 .. Levy",
        ("FIX.4.4", "139", "7"): "7 .. Other",
        ("FIX.4.4", "139", "8"): "8 .. Markup",
        ("FIX.4.4", "139", "9"): "9 .. ConsumptionTax",
        ("FIX.4.4", "139", "10"): "10 .. PerTransaction",
        ("FIX.4.4", "139", "11"): "11 .. Conversion",
        ("FIX.4.4", "139", "12"): "12 .. Agent",
        ("FIX.4.4", "141", "Y"): "Y .. Yes",
        ("FIX.4.4", "141", "N"): "N .. No",
        ("FIX.4.4", "150", "0"): "0 .. New",
        ("FIX.4.4", "150", "3"): "3 .. DoneForDay",
        ("FIX.4.4", "150", "4"): "4 .. Canceled",
        ("FIX.4.4", "150", "5"): "5 .. Replaced",
        ("FIX.4.4", "150", "6"): "6 .. PendingCancel",
        ("FIX.4.4", "150", "7"): "7 .. Stopped",
        ("FIX.4.4", "150", "8"): "8 .. Rejected",
        ("FIX.4.4", "150", "9"): "9 .. Suspended",
        ("FIX.4.4", "150", "A"): "A .. PendingNew",
        ("FIX.4.4", "150", "B"): "B .. Calculated",
        ("FIX.4.4", "150", "C"): "C .. Expired",
        ("FIX.4.4", "150", "D"): "D .. Restated",
        ("FIX.4.4", "150", "E"): "E .. PendingReplace",
        ("FIX.4.4", "150", "F"): "F .. Trade",
        ("FIX.4.4", "150", "G"): "G .. TradeCorrect",
        ("FIX.4.4", "150", "H"): "H .. TradeCancel",
        ("FIX.4.4", "150", "I"): "I .. OrderStatus",
        ("FIX.4.4", "156", "M"): "M .. Multiply",
        ("FIX.4.4", "156", "D"): "D .. Divide",
        ("FIX.4.4", "160", "1"): "1 .. StandingInstructionsProvided",
        ("FIX.4.4", "160", "4"): "4 .. SpecificOrderForASingleAccount",
        ("FIX.4.4", "160", "5"): "5 .. RequestReject",
        ("FIX.4.4", "163", "N"): "N .. New",
        ("FIX.4.4", "163", "C"): "C .. Cancel",
        ("FIX.4.4", "163", "R"): "R .. Replace",
        ("FIX.4.4", "163", "T"): "T .. Restate",
        ("FIX.4.4", "165", "1"): "1 .. BrokerCredit",
        ("FIX.4.4", "165", "2"): "2 .. Institution",
        ("FIX.4.4", "165", "3"): "3 .. Investor",
        ("FIX.4.4", "167", "EUSUPRA"): "EUSUPRA .. EuroSupranationalCoupons",
        ("FIX.4.4", "167", "FAC"): "FAC .. FederalAgencyCoupon",
        ("FIX.4.4", "167", "FADN"): "FADN .. FederalAgencyDiscountNote",
        ("FIX.4.4", "167", "PEF"): "PEF .. PrivateExportFunding",
        ("FIX.4.4", "167", "SUPRA"): "SUPRA .. USDSupranationalCoupons",
        ("FIX.4.4", "167", "CORP"): "CORP .. CorporateBond",
        ("FIX.4.4", "167", "CPP"): "CPP .. CorporatePrivatePlacement",
        ("FIX.4.4", "167", "CB"): "CB .. ConvertibleBond",
        ("FIX.4.4", "167", "DUAL"): "DUAL .. DualCurrency",
        ("FIX.4.4", "167", "EUCORP"): "EUCORP .. EuroCorporateBond",
        ("FIX.4.4", "167", "XLINKD"): "XLINKD .. IndexedLinked",
        ("FIX.4.4", "167", "STRUCT"): "STRUCT .. StructuredNotes",
        ("FIX.4.4", "167", "YANK"): "YANK .. YankeeCorporateBond",
        ("FIX.4.4", "167", "FOR"): "FOR .. ForeignExchangeContract",
        ("FIX.4.4", "167", "CS"): "CS .. CommonStock",
        ("FIX.4.4", "167", "PS"): "PS .. PreferredStock",
        ("FIX.4.4", "167", "BRADY"): "BRADY .. BradyBond",
        ("FIX.4.4", "167", "EUSOV"): "EUSOV .. EuroSovereigns",
        ("FIX.4.4", "167", "TBOND"): "TBOND .. USTreasuryBond",
        ("FIX.4.4", "167", "TINT"): "TINT .. InterestStripFromAnyBondOrNote",
        ("FIX.4.4", "167", "TIPS"): "TIPS .. TreasuryInflationProtectedSecurities",
        ("FIX.4.4", "167", "TCAL"): "TCAL .. PrincipalStripOfACallableBondOrNote",
        ("FIX.4.4", "167", "TPRN"): "TPRN .. PrincipalStripFromANonCallableBondOrNote",
        ("FIX.4.4", "167", "UST"): "UST .. USTreasuryNoteOld",
        ("FIX.4.4", "167", "USTB"): "USTB .. USTreasuryBillOld",
        ("FIX.4.4", "167", "TNOTE"): "TNOTE .. USTreasuryNote",
        ("FIX.4.4", "167", "TBILL"): "TBILL .. USTreasuryBill",
        ("FIX.4.4", "167", "REPO"): "REPO .. Repurchase",
        ("FIX.4.4", "167", "FORWARD"): "FORWARD .. Forward",
        ("FIX.4.4", "167", "BUYSELL"): "BUYSELL .. BuySellback",
        ("FIX.4.4", "167", "SECLOAN"): "SECLOAN .. SecuritiesLoan",
        ("FIX.4.4", "167", "SECPLEDGE"): "SECPLEDGE .. SecuritiesPledge",
        ("FIX.4.4", "167", "TERM"): "TERM .. TermLoan",
        ("FIX.4.4", "167", "RVLV"): "RVLV .. RevolverLoan",
        ("FIX.4.4", "167", "RVLVTRM"): "RVLVTRM .. Revolver",
        ("FIX.4.4", "167", "BRIDGE"): "BRIDGE .. BridgeLoan",
        ("FIX.4.4", "167", "LOFC"): "LOFC .. LetterOfCredit",
        ("FIX.4.4", "167", "SWING"): "SWING .. SwingLineFacility",
        ("FIX.4.4", "167", "DINP"): "DINP .. DebtorInPossession",
        ("FIX.4.4", "167", "DEFLTED"): "DEFLTED .. Defaulted",
        ("FIX.4.4", "167", "WITHDRN"): "WITHDRN .. Withdrawn",
        ("FIX.4.4", "167", "REPLACD"): "REPLACD .. Replaced",
        ("FIX.4.4", "167", "MATURED"): "MATURED .. Matured",
        ("FIX.4.4", "167", "AMENDED"): "AMENDED .. Amended",
        ("FIX.4.4", "167", "RETIRED"): "RETIRED .. Retired",
        ("FIX.4.4", "167", "BA"): "BA .. BankersAcceptance",
        ("FIX.4.4", "167", "BN"): "BN .. BankNotes",
        ("FIX.4.4", "167", "BOX"): "BOX .. BillOfExchanges",
        ("FIX.4.4", "167", "CD"): "CD .. CertificateOfDeposit",
        ("FIX.4.4", "167", "CL"): "CL .. CallLoans",
        ("FIX.4.4", "167", "CP"): "CP .. CommercialPaper",
        ("FIX.4.4", "167", "DN"): "DN .. DepositNotes",
        ("FIX.4.4", "167", "EUCD"): "EUCD .. EuroCertificateOfDeposit",
        ("FIX.4.4", "167", "EUCP"): "EUCP .. EuroCommercialPaper",
        ("FIX.4.4", "167", "LQN"): "LQN .. LiquidityNote",
        ("FIX.4.4", "167", "MTN"): "MTN .. MediumTermNotes",
        ("FIX.4.4", "167", "ONITE"): "ONITE .. Overnight",
        ("FIX.4.4", "167", "PN"): "PN .. PromissoryNote",
        ("FIX.4.4", "167", "PZFJ"): "PZFJ .. PlazosFijos",
        ("FIX.4.4", "167", "STN"): "STN .. ShortTermLoanNote",
        ("FIX.4.4", "167", "TD"): "TD .. TimeDeposit",
        ("FIX.4.4", "167", "XCN"): "XCN .. ExtendedCommNote",
        ("FIX.4.4", "167", "YCD"): "YCD .. YankeeCertificateOfDeposit",
        ("FIX.4.4", "167", "ABS"): "ABS .. AssetBackedSecurities",
        ("FIX.4.4", "167", "CMBS"): "CMBS .. Corp",
        ("FIX.4.4", "167", "CMO"): "CMO .. CollateralizedMortgageObligation",
        ("FIX.4.4", "167", "IET"): "IET .. IOETTEMortgage",
        ("FIX.4.4", "167", "MBS"): "MBS .. MortgageBackedSecurities",
        ("FIX.4.4", "167", "MIO"): "MIO .. MortgageInterestOnly",
        ("FIX.4.4", "167", "MPO"): "MPO .. MortgagePrincipalOnly",
        ("FIX.4.4", "167", "MPP"): "MPP .. MortgagePrivatePlacement",
        ("FIX.4.4", "167", "MPT"): "MPT .. MiscellaneousPassThrough",
        ("FIX.4.4", "167", "PFAND"): "PFAND .. Pfandbriefe",
        ("FIX.4.4", "167", "TBA"): "TBA .. ToBeAnnounced",
        ("FIX.4.4", "167", "AN"): "AN .. OtherAnticipationNotes",
        ("FIX.4.4", "167", "COFO"): "COFO .. CertificateOfObligation",
        ("FIX.4.4", "167", "COFP"): "COFP .. CertificateOfParticipation",
        ("FIX.4.4", "167", "GO"): "GO .. GeneralObligationBonds",
        ("FIX.4.4", "167", "MT"): "MT .. MandatoryTender",
        ("FIX.4.4", "167", "RAN"): "RAN .. RevenueAnticipationNote",
        ("FIX.4.4", "167", "REV"): "REV .. RevenueBonds",
        ("FIX.4.4", "167", "SPCLA"): "SPCLA .. SpecialAssessment",
        ("FIX.4.4", "167", "SPCLO"): "SPCLO .. SpecialObligation",
        ("FIX.4.4", "167", "SPCLT"): "SPCLT .. SpecialTax",
        ("FIX.4.4", "167", "TAN"): "TAN .. TaxAnticipationNote",
        ("FIX.4.4", "167", "TAXA"): "TAXA .. TaxAllocation",
        ("FIX.4.4", "167", "TECP"): "TECP .. TaxExemptCommercialPaper",
        ("FIX.4.4", "167", "TRAN"): "TRAN .. TaxRevenueAnticipationNote",
        ("FIX.4.4", "167", "VRDN"): "VRDN .. VariableRateDemandNote",
        ("FIX.4.4", "167", "WAR"): "WAR .. Warrant",
        ("FIX.4.4", "167", "MF"): "MF .. MutualFund",
        ("FIX.4.4", "167", "MLEG"): "MLEG .. MultilegInstrument",
        ("FIX.4.4", "167", "NONE"): "NONE .. NoSecurityType",
        ("FIX.4.4", "167", "FUT"): "FUT .. Future",
        ("FIX.4.4", "167", "OPT"): "OPT .. Option",
        ("FIX.4.4", "169", "0"): "0 .. Other",
        ("FIX.4.4", "169", "1"): "1 .. DTCSID",
        ("FIX.4.4", "169", "2"): "2 .. ThomsonALERT",
        ("FIX.4.4", "169", "3"): "3 .. AGlobalCustodian",
        ("FIX.4.4", "169", "4"): "4 .. AccountNet",
        ("FIX.4.4", "172", "0"): "0 .. Versus",
        ("FIX.4.4", "172", "1"): "1 .. Free",
        ("FIX.4.4", "172", "2"): "2 .. TriParty",
        ("FIX.4.4", "172", "3"): "3 .. HoldInCustody",
        ("FIX.4.4", "197", "0"): "0 .. FXNetting",
        ("FIX.4.4", "197", "1"): "1 .. FXSwap",
        ("FIX.4.4", "201", "0"): "0 .. Put",
        ("FIX.4.4", "201", "1"): "1 .. Call",
        ("FIX.4.4", "203", "0"): "0 .. Covered",
        ("FIX.4.4", "203", "1"): "1 .. Uncovered",
        ("FIX.4.4", "208", "Y"): "Y .. DetailsShouldBeCommunicated",
        ("FIX.4.4", "208", "N"): "N .. DetailsShouldNotBeCommunicated",
        ("FIX.4.4", "209", "1"): "1 .. Match",
        ("FIX.4.4", "209", "2"): "2 .. Forward",
        ("FIX.4.4", "209", "3"): "3 .. ForwardAndMatch",
        ("FIX.4.4", "216", "1"): "1 .. TargetFirm",
        ("FIX.4.4", "216", "2"): "2 .. TargetList",
        ("FIX.4.4", "216", "3"): "3 .. BlockFirm",
        ("FIX.4.4", "216", "4"): "4 .. BlockList",
        ("FIX.4.4", "233", "AMT"): "AMT .. AlternativeMinimumTax",
        ("FIX.4.4", "233", "AUTOREINV"): "AUTOREINV .. AutoReinvestment",
        ("FIX.4.4", "233", "BANKQUAL"): "BANKQUAL .. BankQualified",
        ("FIX.4.4", "233", "BGNCON"): "BGNCON .. BargainConditions",
        ("FIX.4.4", "233", "COUPON"): "COUPON .. CouponRange",
        ("FIX.4.4", "233", "CURRENCY"): "CURRENCY .. ISOCurrencyCode",
        ("FIX.4.4", "233", "CUSTOMDATE"): "CUSTOMDATE .. CustomStart",
        ("FIX.4.4", "233", "GEOG"): "GEOG .. Geographics",
        ("FIX.4.4", "233", "HAIRCUT"): "HAIRCUT .. ValuationDiscount",
        ("FIX.4.4", "233", "INSURED"): "INSURED .. Insured",
        ("FIX.4.4", "233", "ISSUE"): "ISSUE .. IssueDate",
        ("FIX.4.4", "233", "ISSUER"): "ISSUER .. Issuer",
        ("FIX.4.4", "233", "ISSUESIZE"): "ISSUESIZE .. IssueSizeRange",
        ("FIX.4.4", "233", "LOOKBACK"): "LOOKBACK .. LookbackDays",
        ("FIX.4.4", "233", "LOT"): "LOT .. ExplicitLotIdentifier",
        ("FIX.4.4", "233", "LOTVAR"): "LOTVAR .. LotVariance",
        ("FIX.4.4", "233", "MAT"): "MAT .. MaturityYearAndMonth",
        ("FIX.4.4", "233", "MATURITY"): "MATURITY .. MaturityRange",
        ("FIX.4.4", "233", "MAXSUBS"): "MAXSUBS .. MaximumSubstitutions",
        ("FIX.4.4", "233", "MINQTY"): "MINQTY .. MinimumQuantity",
        ("FIX.4.4", "233", "MININCR"): "MININCR .. MinimumIncrement",
        ("FIX.4.4", "233", "MINDNOM"): "MINDNOM .. MinimumDenomination",
        ("FIX.4.4", "233", "PAYFREQ"): "PAYFREQ .. PaymentFrequency",
        ("FIX.4.4", "233", "PIECES"): "PIECES .. NumberOfPieces",
        ("FIX.4.4", "233", "PMAX"): "PMAX .. PoolsMaximum",
        ("FIX.4.4", "233", "PPM"): "PPM .. PoolsPerMillion",
        ("FIX.4.4", "233", "PPL"): "PPL .. PoolsPerLot",
        ("FIX.4.4", "233", "PPT"): "PPT .. PoolsPerTrade",
        ("FIX.4.4", "233", "PRICE"): "PRICE .. PriceRange",
        ("FIX.4.4", "233", "PRICEFREQ"): "PRICEFREQ .. PricingFrequency",
        ("FIX.4.4", "233", "PROD"): "PROD .. ProductionYear",
        ("FIX.4.4", "233", "PROTECT"): "PROTECT .. CallProtection",
        ("FIX.4.4", "233", "PURPOSE"): "PURPOSE .. Purpose",
        ("FIX.4.4", "233", "PXSOURCE"): "PXSOURCE .. BenchmarkPriceSource",
        ("FIX.4.4", "233", "RATING"): "RATING .. RatingSourceAndRange",
        ("FIX.4.4", "233", "REDEMPTION"): "REDEMPTION .. TypeOfRedemption",
        ("FIX.4.4", "233", "RESTRICTED"): "RESTRICTED .. Restricted",
        ("FIX.4.4", "233", "SECTOR"): "SECTOR .. MarketSector",
        ("FIX.4.4", "233", "SECTYPE"): "SECTYPE .. SecurityTypeIncludedOrExcluded",
        ("FIX.4.4", "233", "STRUCT"): "STRUCT .. Structure",
        ("FIX.4.4", "233", "SUBSFREQ"): "SUBSFREQ .. SubstitutionsFrequency",
        ("FIX.4.4", "233", "SUBSLEFT"): "SUBSLEFT .. SubstitutionsLeft",
        ("FIX.4.4", "233", "TEXT"): "TEXT .. FreeformText",
        ("FIX.4.4", "233", "TRDVAR"): "TRDVAR .. TradeVariance",
        ("FIX.4.4", "233", "WAC"): "WAC .. WeightedAverageCoupon",
        ("FIX.4.4", "233", "WAL"): "WAL .. WeightedAverageLifeCoupon",
        ("FIX.4.4", "233", "WALA"): "WALA .. WeightedAverageLoanAge",
        ("FIX.4.4", "233", "WAM"): "WAM .. WeightedAverageMaturity",
        ("FIX.4.4", "233", "WHOLE"): "WHOLE .. WholePool",
        ("FIX.4.4", "233", "YIELD"): "YIELD .. YieldRange",
        ("FIX.4.4", "235", "AFTERTAX"): "AFTERTAX .. AfterTaxYield",
        ("FIX.4.4", "235", "ANNUAL"): "ANNUAL .. AnnualYield",
        ("FIX.4.4", "235", "ATISSUE"): "ATISSUE .. YieldAtIssue",
        ("FIX.4.4", "235", "AVGMATURITY"): "AVGMATURITY .. YieldToAverageMaturity",
        ("FIX.4.4", "235", "BOOK"): "BOOK .. BookYield",
        ("FIX.4.4", "235", "CALL"): "CALL .. YieldToNextCall",
        ("FIX.4.4", "235", "CHANGE"): "CHANGE .. YieldChangeSinceClose",
        ("FIX.4.4", "235", "CLOSE"): "CLOSE .. ClosingYield",
        ("FIX.4.4", "235", "COMPOUND"): "COMPOUND .. CompoundYield",
        ("FIX.4.4", "235", "CURRENT"): "CURRENT .. CurrentYield",
        ("FIX.4.4", "235", "GROSS"): "GROSS .. TrueGrossYield",
        ("FIX.4.4", "235", "GOVTEQUIV"): "GOVTEQUIV .. GvntEquivalentYield",
        ("FIX.4.4", "235", "INFLATION"): "INFLATION .. YieldWithInflationAssumption",
        ("FIX.4.4", "235", "INVERSEFLOATER"): "INVERSEFLOATER .. InverseFloaterBondYield",
        ("FIX.4.4", "235", "LASTCLOSE"): "LASTCLOSE .. MostRecentClosingYield",
        ("FIX.4.4", "235", "LASTMONTH"): "LASTMONTH .. ClosingYieldMostRecentMonth",
        ("FIX.4.4", "235", "LASTQUARTER"): "LASTQUARTER .. ClosingYieldMostRecentQuarter",
        ("FIX.4.4", "235", "LASTYEAR"): "LASTYEAR .. ClosingYieldMostRecentYear",
        ("FIX.4.4", "235", "LONGAVGLIFE"): "LONGAVGLIFE .. YieldToLongestAverageLife",
        ("FIX.4.4", "235", "MARK"): "MARK .. MarkToMarketYield",
        ("FIX.4.4", "235", "MATURITY"): "MATURITY .. YieldToMaturity",
        ("FIX.4.4", "235", "NEXTREFUND"): "NEXTREFUND .. YieldToNextRefund",
        ("FIX.4.4", "235", "OPENAVG"): "OPENAVG .. OpenAverageYield",
        ("FIX.4.4", "235", "PUT"): "PUT .. YieldToNextPut",
        ("FIX.4.4", "235", "PREVCLOSE"): "PREVCLOSE .. PreviousCloseYield",
        ("FIX.4.4", "235", "PROCEEDS"): "PROCEEDS .. ProceedsYield",
        ("FIX.4.4", "235", "SEMIANNUAL"): "SEMIANNUAL .. SemiAnnualYield",
        ("FIX.4.4", "235", "SHORTAVGLIFE"): "SHORTAVGLIFE .. YieldToShortestAverageLife",
        ("FIX.4.4", "235", "SIMPLE"): "SIMPLE .. SimpleYield",
        ("FIX.4.4", "235", "TAXEQUIV"): "TAXEQUIV .. TaxEquivalentYield",
        ("FIX.4.4", "235", "TENDER"): "TENDER .. YieldToTenderDate",
        ("FIX.4.4", "235", "TRUE"): "TRUE .. TrueYield",
        ("FIX.4.4", "235", "VALUE1/32"): "VALUE1/32 .. YieldValueOf132",
        ("FIX.4.4", "235", "WORST"): "WORST .. YieldToWorst",
        ("FIX.4.4", "258", "Y"): "Y .. TradedFlat",
        ("FIX.4.4", "258", "N"): "N .. NotTradedFlat",
        ("FIX.4.4", "263", "0"): "0 .. Snapshot",
        ("FIX.4.4", "263", "1"): "1 .. SnapshotAndUpdates",
        ("FIX.4.4", "263", "2"): "2 .. DisablePreviousSnapshot",
        ("FIX.4.4", "265", "0"): "0 .. FullRefresh",
        ("FIX.4.4", "265", "1"): "1 .. IncrementalRefresh",
        ("FIX.4.4", "266", "Y"): "Y .. BookEntriesToBeAggregated",
        ("FIX.4.4", "266", "N"): "N .. BookEntriesShouldNotBeAggregated",
        ("FIX.4.4", "269", "0"): "0 .. Bid",
        ("FIX.4.4", "269", "1"): "1 .. Offer",
        ("FIX.4.4", "269", "2"): "2 .. Trade",
        ("FIX.4.4", "269", "3"): "3 .. IndexValue",
        ("FIX.4.4", "269", "4"): "4 .. OpeningPrice",
        ("FIX.4.4", "269", "5"): "5 .. ClosingPrice",
        ("FIX.4.4", "269", "6"): "6 .. SettlementPrice",
        ("FIX.4.4", "269", "7"): "7 .. TradingSessionHighPrice",
        ("FIX.4.4", "269", "8"): "8 .. TradingSessionLowPrice",
        ("FIX.4.4", "269", "9"): "9 .. TradingSessionVWAPPrice",
        ("FIX.4.4", "269", "A"): "A .. Imbalance",
        ("FIX.4.4", "269", "B"): "B .. TradeVolume",
        ("FIX.4.4", "269", "C"): "C .. OpenInterest",
        ("FIX.4.4", "274", "0"): "0 .. PlusTick",
        ("FIX.4.4", "274", "1"): "1 .. ZeroPlusTick",
        ("FIX.4.4", "274", "2"): "2 .. MinusTick",
        ("FIX.4.4", "274", "3"): "3 .. ZeroMinusTick",
        ("FIX.4.4", "276", "A"): "A .. Open",
        ("FIX.4.4", "276", "B"): "B .. Closed",
        ("FIX.4.4", "276", "C"): "C .. ExchangeBest",
        ("FIX.4.4", "276", "D"): "D .. ConsolidatedBest",
        ("FIX.4.4", "276", "E"): "E .. Locked",
        ("FIX.4.4", "276", "F"): "F .. Crossed",
        ("FIX.4.4", "276", "G"): "G .. Depth",
        ("FIX.4.4", "276", "H"): "H .. FastTrading",
        ("FIX.4.4", "276", "I"): "I .. NonFirm",
        ("FIX.4.4", "277", "A"): "A .. Cash",
        ("FIX.4.4", "277", "B"): "B .. AveragePriceTrade",
        ("FIX.4.4", "277", "C"): "C .. CashTrade",
        ("FIX.4.4", "277", "D"): "D .. NextDay",
        ("FIX.4.4", "277", "E"): "E .. Opening",
        ("FIX.4.4", "277", "F"): "F .. IntradayTradeDetail",
        ("FIX.4.4", "277", "G"): "G .. Rule127Trade",
        ("FIX.4.4", "277", "H"): "H .. Rule155Trade",
        ("FIX.4.4", "277", "I"): "I .. SoldLast",
        ("FIX.4.4", "277", "J"): "J .. NextDayTrade",
        ("FIX.4.4", "277", "K"): "K .. Opened",
        ("FIX.4.4", "277", "L"): "L .. Seller",
        ("FIX.4.4", "277", "M"): "M .. Sold",
        ("FIX.4.4", "277", "N"): "N .. StoppedStock",
        ("FIX.4.4", "277", "P"): "P .. ImbalanceMoreBuyers",
        ("FIX.4.4", "277", "Q"): "Q .. ImbalanceMoreSellers",
        ("FIX.4.4", "277", "R"): "R .. OpeningPrice",
        ("FIX.4.4", "279", "0"): "0 .. New",
        ("FIX.4.4", "279", "1"): "1 .. Change",
        ("FIX.4.4", "279", "2"): "2 .. Delete",
        ("FIX.4.4", "281", "0"): "0 .. UnknownSymbol",
        ("FIX.4.4", "281", "1"): "1 .. DuplicateMDReqID",
        ("FIX.4.4", "281", "2"): "2 .. InsufficientBandwidth",
        ("FIX.4.4", "281", "3"): "3 .. InsufficientPermissions",
        ("FIX.4.4", "281", "4"): "4 .. UnsupportedSubscriptionRequestType",
        ("FIX.4.4", "281", "5"): "5 .. UnsupportedMarketDepth",
        ("FIX.4.4", "281", "6"): "6 .. UnsupportedMDUpdateType",
        ("FIX.4.4", "281", "7"): "7 .. UnsupportedAggregatedBook",
        ("FIX.4.4", "281", "8"): "8 .. UnsupportedMDEntryType",
        ("FIX.4.4", "281", "9"): "9 .. UnsupportedTradingSessionID",
        ("FIX.4.4", "281", "A"): "A .. UnsupportedScope",
        ("FIX.4.4", "281", "B"): "B .. UnsupportedOpenCloseSettleFlag",
        ("FIX.4.4", "281", "C"): "C .. UnsupportedMDImplicitDelete",
        ("FIX.4.4", "285", "0"): "0 .. Cancellation",
        ("FIX.4.4", "285", "1"): "1 .. Error",
        ("FIX.4.4", "286", "0"): "0 .. DailyOpen",
        ("FIX.4.4", "286", "1"): "1 .. SessionOpen",
        ("FIX.4.4", "286", "2"): "2 .. DeliverySettlementEntry",
        ("FIX.4.4", "286", "3"): "3 .. ExpectedEntry",
        ("FIX.4.4", "286", "4"): "4 .. EntryFromPreviousBusinessDay",
        ("FIX.4.4", "286", "5"): "5 .. TheoreticalPriceValue",
        ("FIX.4.4", "291", "1"): "1 .. Bankrupt",
        ("FIX.4.4", "291", "2"): "2 .. PendingDelisting",
        ("FIX.4.4", "292", "A"): "A .. ExDividend",
        ("FIX.4.4", "292", "B"): "B .. ExDistribution",
        ("FIX.4.4", "292", "C"): "C .. ExRights",
        ("FIX.4.4", "292", "D"): "D .. New",
        ("FIX.4.4", "292", "E"): "E .. ExInterest",
        ("FIX.4.4", "297", "0"): "0 .. Accepted",
        ("FIX.4.4", "297", "1"): "1 .. CancelForSymbol",
        ("FIX.4.4", "297", "2"): "2 .. CanceledForSecurityType",
        ("FIX.4.4", "297", "3"): "3 .. CanceledForUnderlying",
        ("FIX.4.4", "297", "4"): "4 .. CanceledAll",
        ("FIX.4.4", "297", "5"): "5 .. Rejected",
        ("FIX.4.4", "297", "6"): "6 .. RemovedFromMarket",
        ("FIX.4.4", "297", "7"): "7 .. Expired",
        ("FIX.4.4", "297", "8"): "8 .. Query",
        ("FIX.4.4", "297", "9"): "9 .. QuoteNotFound",
        ("FIX.4.4", "297", "10"): "10 .. Pending",
        ("FIX.4.4", "297", "11"): "11 .. Pass",
        ("FIX.4.4", "297", "12"): "12 .. LockedMarketWarning",
        ("FIX.4.4", "297", "13"): "13 .. CrossMarketWarning",
        ("FIX.4.4", "297", "14"): "14 .. CanceledDueToLockMarket",
        ("FIX.4.4", "297", "15"): "15 .. CanceledDueToCrossMarket",
        ("FIX.4.4", "298", "1"): "1 .. CancelForOneOrMoreSecurities",
        ("FIX.4.4", "298", "2"): "2 .. CancelForSecurityType",
        ("FIX.4.4", "298", "3"): "3 .. CancelForUnderlyingSecurity",
        ("FIX.4.4", "298", "4"): "4 .. CancelAllQuotes",
        ("FIX.4.4", "300", "1"): "1 .. UnknownSymbol",
        ("FIX.4.4", "300", "2"): "2 .. Exchange",
        ("FIX.4.4", "300", "3"): "3 .. QuoteRequestExceedsLimit",
        ("FIX.4.4", "300", "4"): "4 .. TooLateToEnter",
        ("FIX.4.4", "300", "5"): "5 .. UnknownQuote",
        ("FIX.4.4", "300", "6"): "6 .. DuplicateQuote",
        ("FIX.4.4", "300", "7"): "7 .. InvalidBid",
        ("FIX.4.4", "300", "8"): "8 .. InvalidPrice",
        ("FIX.4.4", "300", "9"): "9 .. NotAuthorizedToQuoteSecurity",
        ("FIX.4.4", "300", "99"): "99 .. Other",
        ("FIX.4.4", "301", "0"): "0 .. NoAcknowledgement",
        ("FIX.4.4", "301", "1"): "1 .. AcknowledgeOnlyNegativeOrErroneousQuotes",
        ("FIX.4.4", "301", "2"): "2 .. AcknowledgeEachQuoteMessage",
        ("FIX.4.4", "303", "1"): "1 .. Manual",
        ("FIX.4.4", "303", "2"): "2 .. Automatic",
        ("FIX.4.4", "321", "0"): "0 .. RequestSecurityIdentityAndSpecifications",
        ("FIX.4.4", "321", "1"): "1 .. RequestSecurityIdentityForSpecifications",
        ("FIX.4.4", "321", "2"): "2 .. RequestListSecurityTypes",
        ("FIX.4.4", "321", "3"): "3 .. RequestListSecurities",
        ("FIX.4.4", "323", "1"): "1 .. AcceptAsIs",
        ("FIX.4.4", "323", "2"): "2 .. AcceptWithRevisions",
        ("FIX.4.4", "323", "5"): "5 .. RejectSecurityProposal",
        ("FIX.4.4", "323", "6"): "6 .. CannotMatchSelectionCriteria",
        ("FIX.4.4", "325", "Y"): "Y .. MessageIsBeingSentUnsolicited",
        ("FIX.4.4", "325", "N"): "N .. MessageIsBeingSentAsAResultOfAPriorRequest",
        ("FIX.4.4", "326", "1"): "1 .. OpeningDelay",
        ("FIX.4.4", "326", "2"): "2 .. TradingHalt",
        ("FIX.4.4", "326", "3"): "3 .. Resume",
        ("FIX.4.4", "326", "4"): "4 .. NoOpen",
        ("FIX.4.4", "326", "5"): "5 .. PriceIndication",
        ("FIX.4.4", "326", "6"): "6 .. TradingRangeIndication",
        ("FIX.4.4", "326", "7"): "7 .. MarketImbalanceBuy",
        ("FIX.4.4", "326", "8"): "8 .. MarketImbalanceSell",
        ("FIX.4.4", "326", "9"): "9 .. MarketOnCloseImbalanceBuy",
        ("FIX.4.4", "326", "10"): "10 .. MarketOnCloseImbalanceSell",
        ("FIX.4.4", "326", "12"): "12 .. NoMarketImbalance",
        ("FIX.4.4", "326", "13"): "13 .. NoMarketOnCloseImbalance",
        ("FIX.4.4", "326", "14"): "14 .. ITSPreOpening",
        ("FIX.4.4", "326", "15"): "15 .. NewPriceIndication",
        ("FIX.4.4", "326", "16"): "16 .. TradeDisseminationTime",
        ("FIX.4.4", "326", "17"): "17 .. ReadyToTrade",
        ("FIX.4.4", "326", "18"): "18 .. NotAvailableForTrading",
        ("FIX.4.4", "326", "19"): "19 .. NotTradedOnThisMarket",
        ("FIX.4.4", "326", "20"): "20 .. UnknownOrInvalid",
        ("FIX.4.4", "326", "21"): "21 .. PreOpen",
        ("FIX.4.4", "326", "22"): "22 .. OpeningRotation",
        ("FIX.4.4", "326", "23"): "23 .. FastMarket",
        ("FIX.4.4", "327", "I"): "I .. OrderImbalance",
        ("FIX.4.4", "327", "X"): "X .. EquipmentChangeover",
        ("FIX.4.4", "327", "P"): "P .. NewsPending",
        ("FIX.4.4", "327", "D"): "D .. NewsDissemination",
        ("FIX.4.4", "327", "E"): "E .. OrderInflux",
        ("FIX.4.4", "327", "M"): "M .. AdditionalInformation",
        ("FIX.4.4", "328", "Y"): "Y .. HaltWasDueToCommonStockBeingHalted",
        ("FIX.4.4", "328", "N"): "N .. HaltWasNotRelatedToAHaltOfTheCommonStock",
        ("FIX.4.4", "329", "Y"): "Y .. RelatedToSecurityHalt",
        ("FIX.4.4", "329", "N"): "N .. NotRelatedToSecurityHalt",
        ("FIX.4.4", "334", "1"): "1 .. Cancel",
        ("FIX.4.4", "334", "2"): "2 .. Error",
        ("FIX.4.4", "334", "3"): "3 .. Correction",
        ("FIX.4.4", "338", "1"): "1 .. Electronic",
        ("FIX.4.4", "338", "2"): "2 .. OpenOutcry",
        ("FIX.4.4", "338", "3"): "3 .. TwoParty",
        ("FIX.4.4", "339", "1"): "1 .. Testing",
        ("FIX.4.4", "339", "2"): "2 .. Simulated",
        ("FIX.4.4", "339", "3"): "3 .. Production",
        ("FIX.4.4", "340", "0"): "0 .. Unknown",
        ("FIX.4.4", "340", "1"): "1 .. Halted",
        ("FIX.4.4", "340", "2"): "2 .. Open",
        ("FIX.4.4", "340", "3"): "3 .. Closed",
        ("FIX.4.4", "340", "4"): "4 .. PreOpen",
        ("FIX.4.4", "340", "5"): "5 .. PreClose",
        ("FIX.4.4", "340", "6"): "6 .. RequestRejected",
        ("FIX.4.4", "347", "ISO-2022-JP"): "ISO-2022-JP .. ISO2022JP",
        ("FIX.4.4", "347", "EUC-JP"): "EUC-JP .. EUCJP",
        ("FIX.4.4", "347", "Shift_JIS"): "Shift_JIS .. ShiftJIS",
        ("FIX.4.4", "347", "UTF-8"): "UTF-8 .. UTF8",
        ("FIX.4.4", "373", "0"): "0 .. InvalidTagNumber",
        ("FIX.4.4", "373", "1"): "1 .. RequiredTagMissing",
        ("FIX.4.4", "373", "2"): "2 .. TagNotDefinedForThisMessageType",
        ("FIX.4.4", "373", "3"): "3 .. UndefinedTag",
        ("FIX.4.4", "373", "4"): "4 .. TagSpecifiedWithoutAValue",
        ("FIX.4.4", "373", "5"): "5 .. ValueIsIncorrect",
        ("FIX.4.4", "373", "6"): "6 .. IncorrectDataFormatForValue",
        ("FIX.4.4", "373", "7"): "7 .. DecryptionProblem",
        ("FIX.4.4", "373", "8"): "8 .. SignatureProblem",
        ("FIX.4.4", "373", "9"): "9 .. CompIDProblem",
        ("FIX.4.4", "373", "10"): "10 .. SendingTimeAccuracyProblem",
        ("FIX.4.4", "373", "11"): "11 .. InvalidMsgType",
        ("FIX.4.4", "373", "12"): "12 .. XMLValidationError",
        ("FIX.4.4", "373", "13"): "13 .. TagAppearsMoreThanOnce",
        ("FIX.4.4", "373", "14"): "14 .. TagSpecifiedOutOfRequiredOrder",
        ("FIX.4.4", "373", "15"): "15 .. RepeatingGroupFieldsOutOfOrder",
        ("FIX.4.4", "373", "16"): "16 .. IncorrectNumInGroupCountForRepeatingGroup",
        ("FIX.4.4", "373", "17"): "17 .. Non",
        ("FIX.4.4", "373", "99"): "99 .. Other",
        ("FIX.4.4", "374", "N"): "N .. New",
        ("FIX.4.4", "374", "C"): "C .. Cancel",
        ("FIX.4.4", "377", "Y"): "Y .. WasSolicited",
        ("FIX.4.4", "377", "N"): "N .. WasNotSolicited",
        ("FIX.4.4", "378", "0"): "0 .. GTCorporateAction",
        ("FIX.4.4", "378", "1"): "1 .. GTRenewal",
        ("FIX.4.4", "378", "2"): "2 .. VerbalChange",
        ("FIX.4.4", "378", "3"): "3 .. RepricingOfOrder",
        ("FIX.4.4", "378", "4"): "4 .. BrokerOption",
        ("FIX.4.4", "378", "5"): "5 .. PartialDeclineOfOrderQty",
        ("FIX.4.4", "378", "6"): "6 .. CancelOnTradingHalt",
        ("FIX.4.4", "378", "7"): "7 .. CancelOnSystemFailure",
        ("FIX.4.4", "378", "8"): "8 .. Market",
        ("FIX.4.4", "378", "9"): "9 .. Canceled",
        ("FIX.4.4", "378", "10"): "10 .. WarehouseRecap",
        ("FIX.4.4", "378", "99"): "99 .. Other",
        ("FIX.4.4", "380", "0"): "0 .. Other",
        ("FIX.4.4", "380", "1"): "1 .. UnknownID",
        ("FIX.4.4", "380", "2"): "2 .. UnknownSecurity",
        ("FIX.4.4", "380", "3"): "3 .. UnsupportedMessageType",
        ("FIX.4.4", "380", "4"): "4 .. ApplicationNotAvailable",
        ("FIX.4.4", "380", "5"): "5 .. ConditionallyRequiredFieldMissing",
        ("FIX.4.4", "380", "6"): "6 .. NotAuthorized",
        ("FIX.4.4", "380", "7"): "7 .. DeliverToFirmNotAvailableAtThisTime",
        ("FIX.4.4", "385", "S"): "S .. Send",
        ("FIX.4.4", "385", "R"): "R .. Receive",
        ("FIX.4.4", "388", "0"): "0 .. RelatedToDisplayedPrice",
        ("FIX.4.4", "388", "1"): "1 .. RelatedToMarketPrice",
        ("FIX.4.4", "388", "2"): "2 .. RelatedToPrimaryPrice",
        ("FIX.4.4", "388", "3"): "3 .. RelatedToLocalPrimaryPrice",
        ("FIX.4.4", "388", "4"): "4 .. RelatedToMidpointPrice",
        ("FIX.4.4", "388", "5"): "5 .. RelatedToLastTradePrice",
        ("FIX.4.4", "388", "6"): "6 .. RelatedToVWAP",
        ("FIX.4.4", "394", "1"): "1 .. NonDisclosed",
        ("FIX.4.4", "394", "2"): "2 .. Disclosed",
        ("FIX.4.4", "394", "3"): "3 .. NoBiddingProcess",
        ("FIX.4.4", "399", "1"): "1 .. Sector",
        ("FIX.4.4", "399", "2"): "2 .. Country",
        ("FIX.4.4", "399", "3"): "3 .. Index",
        ("FIX.4.4", "401", "1"): "1 .. SideValue1",
        ("FIX.4.4", "401", "2"): "2 .. SideValue2",
        ("FIX.4.4", "409", "1"): "1 .. FiveDayMovingAverage",
        ("FIX.4.4", "409", "2"): "2 .. TwentyDayMovingAverage",
        ("FIX.4.4", "409", "3"): "3 .. NormalMarketSize",
        ("FIX.4.4", "409", "4"): "4 .. Other",
        ("FIX.4.4", "411", "Y"): "Y .. True",
        ("FIX.4.4", "411", "N"): "N .. False",
        ("FIX.4.4", "414", "1"): "1 .. BuySideRequests",
        ("FIX.4.4", "414", "2"): "2 .. SellSideSends",
        ("FIX.4.4", "414", "3"): "3 .. RealTimeExecutionReports",
        ("FIX.4.4", "416", "1"): "1 .. Net",
        ("FIX.4.4", "416", "2"): "2 .. Gross",
        ("FIX.4.4", "418", "R"): "R .. RiskTrade",
        ("FIX.4.4", "418", "G"): "G .. VWAPGuarantee",
        ("FIX.4.4", "418", "A"): "A .. Agency",
        ("FIX.4.4", "418", "J"): "J .. GuaranteedClose",
        ("FIX.4.4", "419", "2"): "2 .. ClosingPriceAtMorningSession",
        ("FIX.4.4", "419", "3"): "3 .. ClosingPrice",
        ("FIX.4.4", "419", "4"): "4 .. CurrentPrice",
        ("FIX.4.4", "419", "5"): "5 .. SQ",
        ("FIX.4.4", "419", "6"): "6 .. VWAPThroughADay",
        ("FIX.4.4", "419", "7"): "7 .. VWAPThroughAMorningSession",
        ("FIX.4.4", "419", "8"): "8 .. VWAPThroughAnAfternoonSession",
        ("FIX.4.4", "419", "9"): "9 .. VWAPThroughADayExcept",
        ("FIX.4.4", "419", "A"): "A .. VWAPThroughAMorningSessionExcept",
        ("FIX.4.4", "419", "B"): "B .. VWAPThroughAnAfternoonSessionExcept",
        ("FIX.4.4", "419", "C"): "C .. Strike",
        ("FIX.4.4", "419", "D"): "D .. Open",
        ("FIX.4.4", "419", "Z"): "Z .. Others",
        ("FIX.4.4", "423", "1"): "1 .. Percentage",
        ("FIX.4.4", "423", "2"): "2 .. PerUnit",
        ("FIX.4.4", "423", "3"): "3 .. FixedAmount",
        ("FIX.4.4", "423", "4"): "4 .. Discount",
        ("FIX.4.4", "423", "5"): "5 .. Premium",
        ("FIX.4.4", "423", "6"): "6 .. Spread",
        ("FIX.4.4", "423", "7"): "7 .. TEDPrice",
        ("FIX.4.4", "423", "8"): "8 .. TEDYield",
        ("FIX.4.4", "423", "9"): "9 .. Yield",
        ("FIX.4.4", "423", "10"): "10 .. FixedCabinetTradePrice",
        ("FIX.4.4", "423", "11"): "11 .. VariableCabinetTradePrice",
        ("FIX.4.4", "427", "0"): "0 .. BookOutAllTradesOnDayOfExecution",
        ("FIX.4.4", "427", "1"): "1 .. AccumulateUntilFilledOrExpired",
        ("FIX.4.4", "427", "2"): "2 .. AccumulateUntilVerballlyNotifiedOtherwise",
        ("FIX.4.4", "429", "1"): "1 .. Ack",
        ("FIX.4.4", "429", "2"): "2 .. Response",
        ("FIX.4.4", "429", "3"): "3 .. Timed",
        ("FIX.4.4", "429", "4"): "4 .. ExecStarted",
        ("FIX.4.4", "429", "5"): "5 .. AllDone",
        ("FIX.4.4", "429", "6"): "6 .. Alert",
        ("FIX.4.4", "430", "1"): "1 .. Net",
        ("FIX.4.4", "430", "2"): "2 .. Gross",
        ("FIX.4.4", "431", "1"): "1 .. InBiddingProcess",
        ("FIX.4.4", "431", "2"): "2 .. ReceivedForExecution",
        ("FIX.4.4", "431", "3"): "3 .. Executing",
        ("FIX.4.4", "431", "4"): "4 .. Cancelling",
        ("FIX.4.4", "431", "5"): "5 .. Alert",
        ("FIX.4.4", "431", "6"): "6 .. AllDone",
        ("FIX.4.4", "431", "7"): "7 .. Reject",
        ("FIX.4.4", "433", "1"): "1 .. Immediate",
        ("FIX.4.4", "433", "2"): "2 .. WaitForInstruction",
        ("FIX.4.4", "433", "3"): "3 .. SellDriven",
        ("FIX.4.4", "433", "4"): "4 .. BuyDrivenCashTopUp",
        ("FIX.4.4", "433", "5"): "5 .. BuyDrivenCashWithdraw",
        ("FIX.4.4", "434", "1"): "1 .. OrderCancelRequest",
        ("FIX.4.4", "434", "2"): "2 .. OrderCancel",
        ("FIX.4.4", "442", "1"): "1 .. SingleSecurity",
        ("FIX.4.4", "442", "2"): "2 .. IndividualLegOfAMultiLegSecurity",
        ("FIX.4.4", "442", "3"): "3 .. MultiLegSecurity",
        ("FIX.4.4", "447", "B"): "B .. BIC",
        ("FIX.4.4", "447", "C"): "C .. GeneralIdentifier",
        ("FIX.4.4", "447", "D"): "D .. Proprietary",
        ("FIX.4.4", "447", "E"): "E .. ISOCountryCode",
        ("FIX.4.4", "447", "F"): "F .. SettlementEntityLocation",
        ("FIX.4.4", "447", "G"): "G .. MIC",
        ("FIX.4.4", "447", "H"): "H .. CSDParticipant",
        ("FIX.4.4", "447", "1"): "1 .. KoreanInvestorID",
        ("FIX.4.4", "447", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.4.4", "447", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.4.4", "447", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.4.4", "447", "5"): "5 .. ChineseInvestorID",
        ("FIX.4.4", "447", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.4.4", "447", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.4.4", "447", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.4.4", "447", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.4.4", "447", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.4.4", "447", "I"): "I .. ISITCAcronym",
        ("FIX.4.4", "452", "1"): "1 .. ExecutingFirm",
        ("FIX.4.4", "452", "2"): "2 .. BrokerOfCredit",
        ("FIX.4.4", "452", "3"): "3 .. ClientID",
        ("FIX.4.4", "452", "4"): "4 .. ClearingFirm",
        ("FIX.4.4", "452", "5"): "5 .. InvestorID",
        ("FIX.4.4", "452", "6"): "6 .. IntroducingFirm",
        ("FIX.4.4", "452", "7"): "7 .. EnteringFirm",
        ("FIX.4.4", "452", "8"): "8 .. Locate",
        ("FIX.4.4", "452", "9"): "9 .. FundManagerClientID",
        ("FIX.4.4", "452", "10"): "10 .. SettlementLocation",
        ("FIX.4.4", "452", "11"): "11 .. OrderOriginationTrader",
        ("FIX.4.4", "452", "12"): "12 .. ExecutingTrader",
        ("FIX.4.4", "452", "13"): "13 .. OrderOriginationFirm",
        ("FIX.4.4", "452", "14"): "14 .. GiveupClearingFirm",
        ("FIX.4.4", "452", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.4.4", "452", "16"): "16 .. ExecutingSystem",
        ("FIX.4.4", "452", "17"): "17 .. ContraFirm",
        ("FIX.4.4", "452", "18"): "18 .. ContraClearingFirm",
        ("FIX.4.4", "452", "19"): "19 .. SponsoringFirm",
        ("FIX.4.4", "452", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.4.4", "452", "21"): "21 .. ClearingOrganization",
        ("FIX.4.4", "452", "22"): "22 .. Exchange",
        ("FIX.4.4", "452", "24"): "24 .. CustomerAccount",
        ("FIX.4.4", "452", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.4.4", "452", "26"): "26 .. CorrespondentBroker",
        ("FIX.4.4", "452", "27"): "27 .. Buyer",
        ("FIX.4.4", "452", "28"): "28 .. Custodian",
        ("FIX.4.4", "452", "29"): "29 .. Intermediary",
        ("FIX.4.4", "452", "30"): "30 .. Agent",
        ("FIX.4.4", "452", "31"): "31 .. SubCustodian",
        ("FIX.4.4", "452", "32"): "32 .. Beneficiary",
        ("FIX.4.4", "452", "33"): "33 .. InterestedParty",
        ("FIX.4.4", "452", "34"): "34 .. RegulatoryBody",
        ("FIX.4.4", "452", "35"): "35 .. LiquidityProvider",
        ("FIX.4.4", "452", "36"): "36 .. EnteringTrader",
        ("FIX.4.4", "452", "37"): "37 .. ContraTrader",
        ("FIX.4.4", "452", "38"): "38 .. PositionAccount",
        ("FIX.4.4", "460", "1"): "1 .. AGENCY",
        ("FIX.4.4", "460", "2"): "2 .. COMMODITY",
        ("FIX.4.4", "460", "3"): "3 .. CORPORATE",
        ("FIX.4.4", "460", "4"): "4 .. CURRENCY",
        ("FIX.4.4", "460", "5"): "5 .. EQUITY",
        ("FIX.4.4", "460", "6"): "6 .. GOVERNMENT",
        ("FIX.4.4", "460", "7"): "7 .. INDEX",
        ("FIX.4.4", "460", "8"): "8 .. LOAN",
        ("FIX.4.4", "460", "9"): "9 .. MONEYMARKET",
        ("FIX.4.4", "460", "10"): "10 .. MORTGAGE",
        ("FIX.4.4", "460", "11"): "11 .. MUNICIPAL",
        ("FIX.4.4", "460", "12"): "12 .. OTHER",
        ("FIX.4.4", "460", "13"): "13 .. FINANCING",
        ("FIX.4.4", "464", "Y"): "Y .. True",
        ("FIX.4.4", "464", "N"): "N .. Fales",
        ("FIX.4.4", "468", "0"): "0 .. RoundToNearest",
        ("FIX.4.4", "468", "1"): "1 .. RoundDown",
        ("FIX.4.4", "468", "2"): "2 .. RoundUp",
        ("FIX.4.4", "477", "1"): "1 .. CREST",
        ("FIX.4.4", "477", "2"): "2 .. NSCC",
        ("FIX.4.4", "477", "3"): "3 .. Euroclear",
        ("FIX.4.4", "477", "4"): "4 .. Clearstream",
        ("FIX.4.4", "477", "5"): "5 .. Cheque",
        ("FIX.4.4", "477", "6"): "6 .. TelegraphicTransfer",
        ("FIX.4.4", "477", "7"): "7 .. FedWire",
        ("FIX.4.4", "477", "8"): "8 .. DirectCredit",
        ("FIX.4.4", "477", "9"): "9 .. ACHCredit",
        ("FIX.4.4", "477", "10"): "10 .. BPAY",
        ("FIX.4.4", "477", "11"): "11 .. HighValueClearingSystemHVACS",
        ("FIX.4.4", "477", "12"): "12 .. ReinvestInFund",
        ("FIX.4.4", "480", "Y"): "Y .. Yes",
        ("FIX.4.4", "480", "N"): "N .. NoExecutionOnly",
        ("FIX.4.4", "480", "M"): "M .. NoWaiverAgreement",
        ("FIX.4.4", "480", "O"): "O .. NoInstitutional",
        ("FIX.4.4", "481", "Y"): "Y .. Passed",
        ("FIX.4.4", "481", "N"): "N .. NotChecked",
        ("FIX.4.4", "481", "1"): "1 .. ExemptBelowLimit",
        ("FIX.4.4", "481", "2"): "2 .. ExemptMoneyType",
        ("FIX.4.4", "481", "3"): "3 .. ExemptAuthorised",
        ("FIX.4.4", "484", "B"): "B .. BidPrice",
        ("FIX.4.4", "484", "C"): "C .. CreationPrice",
        ("FIX.4.4", "484", "D"): "D .. CreationPricePlusAdjustmentPercent",
        ("FIX.4.4", "484", "E"): "E .. CreationPricePlusAdjustmentAmount",
        ("FIX.4.4", "484", "O"): "O .. OfferPrice",
        ("FIX.4.4", "484", "P"): "P .. OfferPriceMinusAdjustmentPercent",
        ("FIX.4.4", "484", "Q"): "Q .. OfferPriceMinusAdjustmentAmount",
        ("FIX.4.4", "484", "S"): "S .. SinglePrice",
        ("FIX.4.4", "492", "1"): "1 .. CREST",
        ("FIX.4.4", "492", "2"): "2 .. NSCC",
        ("FIX.4.4", "492", "3"): "3 .. Euroclear",
        ("FIX.4.4", "492", "4"): "4 .. Clearstream",
        ("FIX.4.4", "492", "5"): "5 .. Cheque",
        ("FIX.4.4", "492", "6"): "6 .. TelegraphicTransfer",
        ("FIX.4.4", "492", "7"): "7 .. FedWire",
        ("FIX.4.4", "492", "8"): "8 .. DebitCard",
        ("FIX.4.4", "492", "9"): "9 .. DirectDebit",
        ("FIX.4.4", "492", "10"): "10 .. DirectCredit",
        ("FIX.4.4", "492", "11"): "11 .. CreditCard",
        ("FIX.4.4", "492", "12"): "12 .. ACHDebit",
        ("FIX.4.4", "492", "13"): "13 .. ACHCredit",
        ("FIX.4.4", "492", "14"): "14 .. BPAY",
        ("FIX.4.4", "492", "15"): "15 .. HighValueClearingSystem",
        ("FIX.4.4", "495", "0"): "0 .. None",
        ("FIX.4.4", "495", "1"): "1 .. MaxiISA",
        ("FIX.4.4", "495", "2"): "2 .. TESSA",
        ("FIX.4.4", "495", "3"): "3 .. MiniCashISA",
        ("FIX.4.4", "495", "4"): "4 .. MiniStocksAndSharesISA",
        ("FIX.4.4", "495", "5"): "5 .. MiniInsuranceISA",
        ("FIX.4.4", "495", "6"): "6 .. CurrentYearPayment",
        ("FIX.4.4", "495", "7"): "7 .. PriorYearPayment",
        ("FIX.4.4", "495", "8"): "8 .. AssetTransfer",
        ("FIX.4.4", "495", "9"): "9 .. EmployeePriorYear",
        ("FIX.4.4", "495", "10"): "10 .. EmployeeCurrentYear",
        ("FIX.4.4", "495", "11"): "11 .. EmployerPriorYear",
        ("FIX.4.4", "495", "12"): "12 .. EmployerCurrentYear",
        ("FIX.4.4", "495", "13"): "13 .. NonFundPrototypeIRA",
        ("FIX.4.4", "495", "14"): "14 .. NonFundQualifiedPlan",
        ("FIX.4.4", "495", "15"): "15 .. DefinedContributionPlan",
        ("FIX.4.4", "495", "16"): "16 .. IRA",
        ("FIX.4.4", "495", "17"): "17 .. IRARollover",
        ("FIX.4.4", "495", "18"): "18 .. KEOGH",
        ("FIX.4.4", "495", "19"): "19 .. ProfitSharingPlan",
        ("FIX.4.4", "495", "20"): "20 .. US401K",
        ("FIX.4.4", "495", "21"): "21 .. SelfDirectedIRA",
        ("FIX.4.4", "495", "22"): "22 .. US403b",
        ("FIX.4.4", "495", "23"): "23 .. US457",
        ("FIX.4.4", "495", "24"): "24 .. RothIRAPrototype",
        ("FIX.4.4", "495", "25"): "25 .. RothIRANonPrototype",
        ("FIX.4.4", "495", "26"): "26 .. RothConversionIRAPrototype",
        ("FIX.4.4", "495", "27"): "27 .. RothConversionIRANonPrototype",
        ("FIX.4.4", "495", "28"): "28 .. EducationIRAPrototype",
        ("FIX.4.4", "495", "29"): "29 .. EducationIRANonPrototype",
        ("FIX.4.4", "497", "Y"): "Y .. Yes",
        ("FIX.4.4", "497", "N"): "N .. No",
        ("FIX.4.4", "506", "A"): "A .. Accepted",
        ("FIX.4.4", "506", "R"): "R .. Rejected",
        ("FIX.4.4", "506", "H"): "H .. Held",
        ("FIX.4.4", "506", "N"): "N .. Reminder",
        ("FIX.4.4", "507", "1"): "1 .. InvalidAccountType",
        ("FIX.4.4", "507", "2"): "2 .. InvalidTaxExemptType",
        ("FIX.4.4", "507", "3"): "3 .. InvalidOwnershipType",
        ("FIX.4.4", "507", "4"): "4 .. NoRegDetails",
        ("FIX.4.4", "507", "5"): "5 .. InvalidRegSeqNo",
        ("FIX.4.4", "507", "6"): "6 .. InvalidRegDetails",
        ("FIX.4.4", "507", "7"): "7 .. InvalidMailingDetails",
        ("FIX.4.4", "507", "8"): "8 .. InvalidMailingInstructions",
        ("FIX.4.4", "507", "9"): "9 .. InvalidInvestorID",
        ("FIX.4.4", "507", "10"): "10 .. InvalidInvestorIDSource",
        ("FIX.4.4", "507", "11"): "11 .. InvalidDateOfBirth",
        ("FIX.4.4", "507", "12"): "12 .. InvalidCountry",
        ("FIX.4.4", "507", "13"): "13 .. InvalidDistribInstns",
        ("FIX.4.4", "507", "14"): "14 .. InvalidPercentage",
        ("FIX.4.4", "507", "15"): "15 .. InvalidPaymentMethod",
        ("FIX.4.4", "507", "16"): "16 .. InvalidAccountName",
        ("FIX.4.4", "507", "17"): "17 .. InvalidAgentCode",
        ("FIX.4.4", "507", "18"): "18 .. InvalidAccountNum",
        ("FIX.4.4", "507", "99"): "99 .. Other",
        ("FIX.4.4", "514", "0"): "0 .. New",
        ("FIX.4.4", "514", "1"): "1 .. Replace",
        ("FIX.4.4", "514", "2"): "2 .. Cancel",
        ("FIX.4.4", "517", "J"): "J .. JointInvestors",
        ("FIX.4.4", "517", "T"): "T .. TenantsInCommon",
        ("FIX.4.4", "517", "2"): "2 .. JointTrustees",
        ("FIX.4.4", "519", "1"): "1 .. CommissionAmount",
        ("FIX.4.4", "519", "2"): "2 .. CommissionPercent",
        ("FIX.4.4", "519", "3"): "3 .. InitialChargeAmount",
        ("FIX.4.4", "519", "4"): "4 .. InitialChargePercent",
        ("FIX.4.4", "519", "5"): "5 .. DiscountAmount",
        ("FIX.4.4", "519", "6"): "6 .. DiscountPercent",
        ("FIX.4.4", "519", "7"): "7 .. DilutionLevyAmount",
        ("FIX.4.4", "519", "8"): "8 .. DilutionLevyPercent",
        ("FIX.4.4", "519", "9"): "9 .. ExitChargeAmount",
        ("FIX.4.4", "519", "10"): "10 .. ExitChargePercent",
        ("FIX.4.4", "519", "11"): "11 .. FundBasedRenewalCommissionPercent",
        ("FIX.4.4", "519", "12"): "12 .. ProjectedFundValue",
        ("FIX.4.4", "519", "13"): "13 .. FundBasedRenewalCommissionOnOrder",
        ("FIX.4.4", "519", "14"): "14 .. FundBasedRenewalCommissionOnFund",
        ("FIX.4.4", "519", "15"): "15 .. NetSettlementAmount",
        ("FIX.4.4", "522", "1"): "1 .. IndividualInvestor",
        ("FIX.4.4", "522", "2"): "2 .. PublicCompany",
        ("FIX.4.4", "522", "3"): "3 .. PrivateCompany",
        ("FIX.4.4", "522", "4"): "4 .. IndividualTrustee",
        ("FIX.4.4", "522", "5"): "5 .. CompanyTrustee",
        ("FIX.4.4", "522", "6"): "6 .. PensionPlan",
        ("FIX.4.4", "522", "7"): "7 .. CustodianUnderGiftsToMinorsAct",
        ("FIX.4.4", "522", "8"): "8 .. Trusts",
        ("FIX.4.4", "522", "9"): "9 .. Fiduciaries",
        ("FIX.4.4", "522", "10"): "10 .. NetworkingSubAccount",
        ("FIX.4.4", "522", "11"): "11 .. NonProfitOrganization",
        ("FIX.4.4", "522", "12"): "12 .. CorporateBody",
        ("FIX.4.4", "522", "13"): "13 .. Nominee",
        ("FIX.4.4", "528", "A"): "A .. Agency",
        ("FIX.4.4", "528", "G"): "G .. Proprietary",
        ("FIX.4.4", "528", "I"): "I .. Individual",
        ("FIX.4.4", "528", "P"): "P .. Principal",
        ("FIX.4.4", "528", "R"): "R .. RisklessPrincipal",
        ("FIX.4.4", "528", "W"): "W .. AgentForOtherMember",
        ("FIX.4.4", "529", "1"): "1 .. ProgramTrade",
        ("FIX.4.4", "529", "2"): "2 .. IndexArbitrage",
        ("FIX.4.4", "529", "3"): "3 .. NonIndexArbitrage",
        ("FIX.4.4", "529", "4"): "4 .. CompetingMarketMaker",
        ("FIX.4.4", "529", "5"): "5 .. ActingAsMarketMakerOrSpecialistInSecurity",
        ("FIX.4.4", "529", "6"): "6 .. ActingAsMarketMakerOrSpecialistInUnderlying",
        ("FIX.4.4", "529", "7"): "7 .. ForeignEntity",
        ("FIX.4.4", "529", "8"): "8 .. ExternalMarketParticipant",
        ("FIX.4.4", "529", "9"): "9 .. ExternalInterConnectedMarketLinkage",
        ("FIX.4.4", "529", "A"): "A .. RisklessArbitrage",
        ("FIX.4.4", "530", "1"): "1 .. CancelOrdersForASecurity",
        ("FIX.4.4", "530", "2"): "2 .. CancelOrdersForAnUnderlyingSecurity",
        ("FIX.4.4", "530", "3"): "3 .. CancelOrdersForAProduct",
        ("FIX.4.4", "530", "4"): "4 .. CancelOrdersForACFICode",
        ("FIX.4.4", "530", "5"): "5 .. CancelOrdersForASecurityType",
        ("FIX.4.4", "530", "6"): "6 .. CancelOrdersForATradingSession",
        ("FIX.4.4", "530", "7"): "7 .. CancelAllOrders",
        ("FIX.4.4", "531", "0"): "0 .. CancelRequestRejected",
        ("FIX.4.4", "531", "1"): "1 .. CancelOrdersForASecurity",
        ("FIX.4.4", "531", "2"): "2 .. CancelOrdersForAnUnderlyingSecurity",
        ("FIX.4.4", "531", "3"): "3 .. CancelOrdersForAProduct",
        ("FIX.4.4", "531", "4"): "4 .. CancelOrdersForACFICode",
        ("FIX.4.4", "531", "5"): "5 .. CancelOrdersForASecurityType",
        ("FIX.4.4", "531", "6"): "6 .. CancelOrdersForATradingSession",
        ("FIX.4.4", "531", "7"): "7 .. CancelAllOrders",
        ("FIX.4.4", "532", "0"): "0 .. MassCancelNotSupported",
        ("FIX.4.4", "532", "1"): "1 .. InvalidOrUnknownSecurity",
        ("FIX.4.4", "532", "2"): "2 .. InvalidOrUnkownUnderlyingSecurity",
        ("FIX.4.4", "532", "3"): "3 .. InvalidOrUnknownProduct",
        ("FIX.4.4", "532", "4"): "4 .. InvalidOrUnknownCFICode",
        ("FIX.4.4", "532", "5"): "5 .. InvalidOrUnknownSecurityType",
        ("FIX.4.4", "532", "6"): "6 .. InvalidOrUnknownTradingSession",
        ("FIX.4.4", "532", "99"): "99 .. Other",
        ("FIX.4.4", "537", "0"): "0 .. Indicative",
        ("FIX.4.4", "537", "1"): "1 .. Tradeable",
        ("FIX.4.4", "537", "2"): "2 .. RestrictedTradeable",
        ("FIX.4.4", "537", "3"): "3 .. Counter",
        ("FIX.4.4", "544", "1"): "1 .. Cash",
        ("FIX.4.4", "544", "2"): "2 .. MarginOpen",
        ("FIX.4.4", "544", "3"): "3 .. MarginClose",
        ("FIX.4.4", "546", "1"): "1 .. LocalMarket",
        ("FIX.4.4", "546", "2"): "2 .. National",
        ("FIX.4.4", "546", "3"): "3 .. Global",
        ("FIX.4.4", "547", "Y"): "Y .. Yes",
        ("FIX.4.4", "547", "N"): "N .. No",
        ("FIX.4.4", "549", "1"): "1 .. CrossAON",
        ("FIX.4.4", "549", "2"): "2 .. CrossIOC",
        ("FIX.4.4", "549", "3"): "3 .. CrossOneSide",
        ("FIX.4.4", "549", "4"): "4 .. CrossSamePrice",
        ("FIX.4.4", "550", "0"): "0 .. None",
        ("FIX.4.4", "550", "1"): "1 .. BuySideIsPrioritized",
        ("FIX.4.4", "550", "2"): "2 .. SellSideIsPrioritized",
        ("FIX.4.4", "552", "1"): "1 .. OneSide",
        ("FIX.4.4", "552", "2"): "2 .. BothSides",
        ("FIX.4.4", "559", "0"): "0 .. Symbol",
        ("FIX.4.4", "559", "1"): "1 .. SecurityTypeAnd",
        ("FIX.4.4", "559", "2"): "2 .. Product",
        ("FIX.4.4", "559", "3"): "3 .. TradingSessionID",
        ("FIX.4.4", "559", "4"): "4 .. AllSecurities",
        ("FIX.4.4", "560", "0"): "0 .. ValidRequest",
        ("FIX.4.4", "560", "1"): "1 .. InvalidOrUnsupportedRequest",
        ("FIX.4.4", "560", "2"): "2 .. NoInstrumentsFound",
        ("FIX.4.4", "560", "3"): "3 .. NotAuthorizedToRetrieveInstrumentData",
        ("FIX.4.4", "560", "4"): "4 .. InstrumentDataTemporarilyUnavailable",
        ("FIX.4.4", "560", "5"): "5 .. RequestForInstrumentDataNotSupported",
        ("FIX.4.4", "563", "0"): "0 .. ReportByMulitlegSecurityOnly",
        ("FIX.4.4", "563", "1"): "1 .. ReportByMultilegSecurityAndInstrumentLegs",
        ("FIX.4.4", "563", "2"): "2 .. ReportByInstrumentLegsOnly",
        ("FIX.4.4", "567", "1"): "1 .. UnknownOrInvalidTradingSessionID",
        ("FIX.4.4", "567", "99"): "99 .. Other",
        ("FIX.4.4", "569", "0"): "0 .. AllTrades",
        ("FIX.4.4", "569", "1"): "1 .. MatchedTradesMatchingCriteria",
        ("FIX.4.4", "569", "2"): "2 .. UnmatchedTradesThatMatchCriteria",
        ("FIX.4.4", "569", "3"): "3 .. UnreportedTradesThatMatchCriteria",
        ("FIX.4.4", "569", "4"): "4 .. AdvisoriesThatMatchCriteria",
        ("FIX.4.4", "570", "Y"): "Y .. PerviouslyReportedToCounterparty",
        ("FIX.4.4", "570", "N"): "N .. NotReportedToCounterparty",
        ("FIX.4.4", "573", "0"): "0 .. Compared",
        ("FIX.4.4", "573", "1"): "1 .. Uncompared",
        ("FIX.4.4", "573", "2"): "2 .. AdvisoryOrAlert",
        ("FIX.4.4", "574", "A1"): "A1 .. ExactMatchPlus4BadgesExecTime",
        ("FIX.4.4", "574", "A2"): "A2 .. ExactMatchPlus4Badges",
        ("FIX.4.4", "574", "A3"): "A3 .. ExactMatchPlus2BadgesExecTime",
        ("FIX.4.4", "574", "A4"): "A4 .. ExactMatchPlus2Badges",
        ("FIX.4.4", "574", "A5"): "A5 .. ExactMatchPlusExecTime",
        ("FIX.4.4", "574", "AQ"): "AQ .. StampedAdvisoriesOrSpecialistAccepts",
        ("FIX.4.4", "574", "S1"): "S1 .. A1ExactMatchSummarizedQuantity",
        ("FIX.4.4", "574", "S2"): "S2 .. A2ExactMatchSummarizedQuantity",
        ("FIX.4.4", "574", "S3"): "S3 .. A3ExactMatchSummarizedQuantity",
        ("FIX.4.4", "574", "S4"): "S4 .. A4ExactMatchSummarizedQuantity",
        ("FIX.4.4", "574", "S5"): "S5 .. A5ExactMatchSummarizedQuantity",
        ("FIX.4.4", "574", "M1"): "M1 .. ExactMatchMinusBadgesTimes",
        ("FIX.4.4", "574", "M2"): "M2 .. SummarizedMatchMinusBadgesTimes",
        ("FIX.4.4", "574", "MT"): "MT .. OCSLockedIn",
        ("FIX.4.4", "574", "M3"): "M3 .. ACTAcceptedTrade",
        ("FIX.4.4", "574", "M4"): "M4 .. ACTDefaultTrade",
        ("FIX.4.4", "574", "M5"): "M5 .. ACTDefaultAfterM2",
        ("FIX.4.4", "574", "M6"): "M6 .. ACTM6Match",
        ("FIX.4.4", "575", "Y"): "Y .. TreatAsOddLot",
        ("FIX.4.4", "575", "N"): "N .. TreatAsRoundLot",
        ("FIX.4.4", "577", "0"): "0 .. ProcessNormally",
        ("FIX.4.4", "577", "1"): "1 .. ExcludeFromAllNetting",
        ("FIX.4.4", "577", "2"): "2 .. BilateralNettingOnly",
        ("FIX.4.4", "577", "3"): "3 .. ExClearing",
        ("FIX.4.4", "577", "4"): "4 .. SpecialTrade",
        ("FIX.4.4", "577", "5"): "5 .. MultilateralNetting",
        ("FIX.4.4", "577", "6"): "6 .. ClearAgainstCentralCounterparty",
        ("FIX.4.4", "577", "7"): "7 .. ExcludeFromCentralCounterparty",
        ("FIX.4.4", "577", "8"): "8 .. ManualMode",
        ("FIX.4.4", "577", "9"): "9 .. AutomaticPostingMode",
        ("FIX.4.4", "577", "10"): "10 .. AutomaticGiveUpMode",
        ("FIX.4.4", "577", "11"): "11 .. QualifiedServiceRepresentativeQSR",
        ("FIX.4.4", "577", "12"): "12 .. CustomerTrade",
        ("FIX.4.4", "577", "13"): "13 .. SelfClearing",
        ("FIX.4.4", "581", "1"): "1 .. CarriedCustomerSide",
        ("FIX.4.4", "581", "2"): "2 .. CarriedNonCustomerSide",
        ("FIX.4.4", "581", "3"): "3 .. HouseTrader",
        ("FIX.4.4", "581", "4"): "4 .. FloorTrader",
        ("FIX.4.4", "581", "6"): "6 .. CarriedNonCustomerSideCrossMargined",
        ("FIX.4.4", "581", "7"): "7 .. HouseTraderCrossMargined",
        ("FIX.4.4", "581", "8"): "8 .. JointBackOfficeAccount",
        ("FIX.4.4", "582", "1"): "1 .. MemberTradingForTheirOwnAccount",
        ("FIX.4.4", "582", "2"): "2 .. ClearingFirmTradingForItsProprietaryAccount",
        ("FIX.4.4", "582", "3"): "3 .. MemberTradingForAnotherMember",
        ("FIX.4.4", "582", "4"): "4 .. AllOther",
        ("FIX.4.4", "585", "1"): "1 .. StatusForOrdersForASecurity",
        ("FIX.4.4", "585", "2"): "2 .. StatusForOrdersForAnUnderlyingSecurity",
        ("FIX.4.4", "585", "3"): "3 .. StatusForOrdersForAProduct",
        ("FIX.4.4", "585", "4"): "4 .. StatusForOrdersForACFICode",
        ("FIX.4.4", "585", "5"): "5 .. StatusForOrdersForASecurityType",
        ("FIX.4.4", "585", "6"): "6 .. StatusForOrdersForATradingSession",
        ("FIX.4.4", "585", "7"): "7 .. StatusForAllOrders",
        ("FIX.4.4", "585", "8"): "8 .. StatusForOrdersForAPartyID",
        ("FIX.4.4", "589", "0"): "0 .. Auto",
        ("FIX.4.4", "589", "1"): "1 .. SpeakWithOrderInitiatorBeforeBooking",
        ("FIX.4.4", "589", "2"): "2 .. Accumulate",
        ("FIX.4.4", "590", "0"): "0 .. EachPartialExecutionIsABookableUnit",
        ("FIX.4.4", "590", "1"): "1 .. AggregatePartialExecutionsOnThisOrder",
        ("FIX.4.4", "590", "2"): "2 .. AggregateExecutionsForThisSymbol",
        ("FIX.4.4", "591", "0"): "0 .. ProRata",
        ("FIX.4.4", "591", "1"): "1 .. DoNotProRata",
        ("FIX.4.4", "626", "1"): "1 .. Calculated",
        ("FIX.4.4", "626", "2"): "2 .. Preliminary",
        ("FIX.4.4", "626", "5"): "5 .. ReadyToBook",
        ("FIX.4.4", "626", "7"): "7 .. WarehouseInstruction",
        ("FIX.4.4", "626", "8"): "8 .. RequestToIntermediary",
        ("FIX.4.4", "635", "B"): "B .. CBOEMember",
        ("FIX.4.4", "635", "C"): "C .. NonMemberAndCustomer",
        ("FIX.4.4", "635", "E"): "E .. EquityMemberAndClearingMember",
        ("FIX.4.4", "635", "F"): "F .. FullAndAssociateMember",
        ("FIX.4.4", "635", "H"): "H .. Firms106HAnd106J",
        ("FIX.4.4", "635", "I"): "I .. GIM",
        ("FIX.4.4", "635", "L"): "L .. Lessee106FEmployees",
        ("FIX.4.4", "635", "M"): "M .. AllOtherOwnershipTypes",
        ("FIX.4.4", "635", "1"): "1 .. FirstYearDelegate",
        ("FIX.4.4", "635", "2"): "2 .. SecondYearDelegate",
        ("FIX.4.4", "635", "3"): "3 .. ThirdYearDelegate",
        ("FIX.4.4", "635", "4"): "4 .. FourthYearDelegate",
        ("FIX.4.4", "635", "5"): "5 .. FifthYearDelegate",
        ("FIX.4.4", "635", "9"): "9 .. SixthYearDelegate",
        ("FIX.4.4", "636", "Y"): "Y .. Working",
        ("FIX.4.4", "636", "N"): "N .. NotWorking",
        ("FIX.4.4", "638", "0"): "0 .. PriorityUnchanged",
        ("FIX.4.4", "638", "1"): "1 .. LostPriorityAsResultOfOrderChange",
        ("FIX.4.4", "650", "Y"): "Y .. LegalConfirm",
        ("FIX.4.4", "650", "N"): "N .. DoesNotConsituteALegalConfirm",
        ("FIX.4.4", "658", "1"): "1 .. UnknownSymbol",
        ("FIX.4.4", "658", "2"): "2 .. Exchange",
        ("FIX.4.4", "658", "3"): "3 .. QuoteRequestExceedsLimit",
        ("FIX.4.4", "658", "4"): "4 .. TooLateToEnter",
        ("FIX.4.4", "658", "5"): "5 .. InvalidPrice",
        ("FIX.4.4", "658", "6"): "6 .. NotAuthorizedToRequestQuote",
        ("FIX.4.4", "658", "7"): "7 .. NoMatchForInquiry",
        ("FIX.4.4", "658", "8"): "8 .. NoMarketForInstrument",
        ("FIX.4.4", "658", "9"): "9 .. NoInventory",
        ("FIX.4.4", "658", "10"): "10 .. Pass",
        ("FIX.4.4", "658", "99"): "99 .. Other",
        ("FIX.4.4", "660", "1"): "1 .. BIC",
        ("FIX.4.4", "660", "2"): "2 .. SIDCode",
        ("FIX.4.4", "660", "3"): "3 .. TFM",
        ("FIX.4.4", "660", "4"): "4 .. OMGEO",
        ("FIX.4.4", "660", "5"): "5 .. DTCCCode",
        ("FIX.4.4", "660", "99"): "99 .. Other",
        ("FIX.4.4", "665", "1"): "1 .. Received",
        ("FIX.4.4", "665", "2"): "2 .. MismatchedAccount",
        ("FIX.4.4", "665", "3"): "3 .. MissingSettlementInstructions",
        ("FIX.4.4", "665", "4"): "4 .. Confirmed",
        ("FIX.4.4", "665", "5"): "5 .. RequestRejected",
        ("FIX.4.4", "666", "0"): "0 .. New",
        ("FIX.4.4", "666", "1"): "1 .. Replace",
        ("FIX.4.4", "666", "2"): "2 .. Cancel",
        ("FIX.4.4", "668", "1"): "1 .. BookEntry",
        ("FIX.4.4", "668", "2"): "2 .. Bearer",
        ("FIX.4.4", "690", "1"): "1 .. ParForPar",
        ("FIX.4.4", "690", "2"): "2 .. ModifiedDuration",
        ("FIX.4.4", "690", "4"): "4 .. Risk",
        ("FIX.4.4", "690", "5"): "5 .. Proceeds",
        ("FIX.4.4", "692", "1"): "1 .. Percent",
        ("FIX.4.4", "692", "2"): "2 .. PerShare",
        ("FIX.4.4", "692", "3"): "3 .. FixedAmount",
        ("FIX.4.4", "692", "4"): "4 .. Discount",
        ("FIX.4.4", "692", "5"): "5 .. Premium",
        ("FIX.4.4", "692", "6"): "6 .. Spread",
        ("FIX.4.4", "692", "7"): "7 .. TEDPrice",
        ("FIX.4.4", "692", "8"): "8 .. TEDYield",
        ("FIX.4.4", "692", "9"): "9 .. YieldSpread",
        ("FIX.4.4", "692", "10"): "10 .. Yield",
        ("FIX.4.4", "694", "1"): "1 .. Hit",
        ("FIX.4.4", "694", "2"): "2 .. Counter",
        ("FIX.4.4", "694", "3"): "3 .. Expired",
        ("FIX.4.4", "694", "4"): "4 .. Cover",
        ("FIX.4.4", "694", "5"): "5 .. DoneAway",
        ("FIX.4.4", "694", "6"): "6 .. Pass",
        ("FIX.4.4", "703", "TQ"): "TQ .. TransactionQuantity",
        ("FIX.4.4", "703", "IAS"): "IAS .. IntraSpreadQty",
        ("FIX.4.4", "703", "IES"): "IES .. InterSpreadQty",
        ("FIX.4.4", "703", "FIN"): "FIN .. EndOfDayQty",
        ("FIX.4.4", "703", "SOD"): "SOD .. StartOfDayQty",
        ("FIX.4.4", "703", "EX"): "EX .. OptionExerciseQty",
        ("FIX.4.4", "703", "AS"): "AS .. OptionAssignment",
        ("FIX.4.4", "703", "TX"): "TX .. TransactionFromExercise",
        ("FIX.4.4", "703", "TA"): "TA .. TransactionFromAssignment",
        ("FIX.4.4", "703", "PIT"): "PIT .. PitTradeQty",
        ("FIX.4.4", "703", "TRF"): "TRF .. TransferTradeQty",
        ("FIX.4.4", "703", "ETR"): "ETR .. ElectronicTradeQty",
        ("FIX.4.4", "703", "ALC"): "ALC .. AllocationTradeQty",
        ("FIX.4.4", "703", "PA"): "PA .. AdjustmentQty",
        ("FIX.4.4", "703", "ASF"): "ASF .. AsOfTradeQty",
        ("FIX.4.4", "703", "DLV"): "DLV .. DeliveryQty",
        ("FIX.4.4", "703", "TOT"): "TOT .. TotalTransactionQty",
        ("FIX.4.4", "703", "XM"): "XM .. CrossMarginQty",
        ("FIX.4.4", "703", "SPL"): "SPL .. IntegralSplit",
        ("FIX.4.4", "706", "0"): "0 .. Submitted",
        ("FIX.4.4", "706", "1"): "1 .. Accepted",
        ("FIX.4.4", "706", "2"): "2 .. Rejected",
        ("FIX.4.4", "707", "FMTM"): "FMTM .. FinalMarkToMarketAmount",
        ("FIX.4.4", "707", "IMTM"): "IMTM .. IncrementalMarkToMarketAmount",
        ("FIX.4.4", "707", "TVAR"): "TVAR .. TradeVariationAmount",
        ("FIX.4.4", "707", "SMTM"): "SMTM .. StartOfDayMarkToMarketAmount",
        ("FIX.4.4", "707", "PREM"): "PREM .. PremiumAmount",
        ("FIX.4.4", "707", "CRES"): "CRES .. CashResidualAmount",
        ("FIX.4.4", "707", "CASH"): "CASH .. CashAmount",
        ("FIX.4.4", "707", "VADJ"): "VADJ .. ValueAdjustedAmount",
        ("FIX.4.4", "709", "1"): "1 .. Exercise",
        ("FIX.4.4", "709", "2"): "2 .. DoNotExercise",
        ("FIX.4.4", "709", "3"): "3 .. PositionAdjustment",
        ("FIX.4.4", "709", "4"): "4 .. PositionChangeSubmission",
        ("FIX.4.4", "709", "5"): "5 .. Pledge",
        ("FIX.4.4", "712", "1"): "1 .. New",
        ("FIX.4.4", "712", "2"): "2 .. Replace",
        ("FIX.4.4", "712", "3"): "3 .. Cancel",
        ("FIX.4.4", "716", "ITD"): "ITD .. Intraday",
        ("FIX.4.4", "716", "RTH"): "RTH .. RegularTradingHours",
        ("FIX.4.4", "716", "ETH"): "ETH .. ElectronicTradingHours",
        ("FIX.4.4", "718", "0"): "0 .. ProcessRequestAsMarginDisposition",
        ("FIX.4.4", "718", "1"): "1 .. DeltaPlus",
        ("FIX.4.4", "718", "2"): "2 .. DeltaMinus",
        ("FIX.4.4", "718", "3"): "3 .. Final",
        ("FIX.4.4", "722", "0"): "0 .. Accepted",
        ("FIX.4.4", "722", "1"): "1 .. AcceptedWithWarnings",
        ("FIX.4.4", "722", "2"): "2 .. Rejected",
        ("FIX.4.4", "722", "3"): "3 .. Completed",
        ("FIX.4.4", "722", "4"): "4 .. CompletedWithWarnings",
        ("FIX.4.4", "723", "0"): "0 .. SuccessfulCompletion",
        ("FIX.4.4", "723", "1"): "1 .. Rejected",
        ("FIX.4.4", "723", "99"): "99 .. Other",
        ("FIX.4.4", "724", "0"): "0 .. Positions",
        ("FIX.4.4", "724", "1"): "1 .. Trades",
        ("FIX.4.4", "724", "2"): "2 .. Exercises",
        ("FIX.4.4", "724", "3"): "3 .. Assignments",
        ("FIX.4.4", "725", "0"): "0 .. Inband",
        ("FIX.4.4", "725", "1"): "1 .. OutOfBand",
        ("FIX.4.4", "728", "0"): "0 .. ValidRequest",
        ("FIX.4.4", "728", "1"): "1 .. InvalidOrUnsupportedRequest",
        ("FIX.4.4", "728", "2"): "2 .. NoPositionsFoundThatMatchCriteria",
        ("FIX.4.4", "728", "3"): "3 .. NotAuthorizedToRequestPositions",
        ("FIX.4.4", "728", "4"): "4 .. RequestForPositionNotSupported",
        ("FIX.4.4", "728", "99"): "99 .. Other",
        ("FIX.4.4", "729", "0"): "0 .. Completed",
        ("FIX.4.4", "729", "1"): "1 .. CompletedWithWarnings",
        ("FIX.4.4", "729", "2"): "2 .. Rejected",
        ("FIX.4.4", "731", "1"): "1 .. Final",
        ("FIX.4.4", "731", "2"): "2 .. Theoretical",
        ("FIX.4.4", "744", "R"): "R .. Random",
        ("FIX.4.4", "744", "P"): "P .. ProRata",
        ("FIX.4.4", "747", "A"): "A .. Automatic",
        ("FIX.4.4", "747", "M"): "M .. Manual",
        ("FIX.4.4", "749", "0"): "0 .. Successful",
        ("FIX.4.4", "749", "1"): "1 .. InvalidOrUnknownInstrument",
        ("FIX.4.4", "749", "2"): "2 .. InvalidTypeOfTradeRequested",
        ("FIX.4.4", "749", "3"): "3 .. InvalidParties",
        ("FIX.4.4", "749", "4"): "4 .. InvalidTransportTypeRequested",
        ("FIX.4.4", "749", "5"): "5 .. InvalidDestinationRequested",
        ("FIX.4.4", "749", "8"): "8 .. TradeRequestTypeNotSupported",
        ("FIX.4.4", "749", "9"): "9 .. NotAuthorized",
        ("FIX.4.4", "749", "99"): "99 .. Other",
        ("FIX.4.4", "750", "0"): "0 .. Accepted",
        ("FIX.4.4", "750", "1"): "1 .. Completed",
        ("FIX.4.4", "750", "2"): "2 .. Rejected",
        ("FIX.4.4", "751", "0"): "0 .. Successful",
        ("FIX.4.4", "751", "1"): "1 .. InvalidPartyOnformation",
        ("FIX.4.4", "751", "2"): "2 .. UnknownInstrument",
        ("FIX.4.4", "751", "3"): "3 .. UnauthorizedToReportTrades",
        ("FIX.4.4", "751", "4"): "4 .. InvalidTradeType",
        ("FIX.4.4", "751", "99"): "99 .. Other",
        ("FIX.4.4", "752", "1"): "1 .. SingleSecurity",
        ("FIX.4.4", "752", "2"): "2 .. IndividualLegOfAMultilegSecurity",
        ("FIX.4.4", "752", "3"): "3 .. MultilegSecurity",
        ("FIX.4.4", "770", "1"): "1 .. ExecutionTime",
        ("FIX.4.4", "770", "2"): "2 .. TimeIn",
        ("FIX.4.4", "770", "3"): "3 .. TimeOut",
        ("FIX.4.4", "770", "4"): "4 .. BrokerReceipt",
        ("FIX.4.4", "770", "5"): "5 .. BrokerExecution",
        ("FIX.4.4", "773", "1"): "1 .. Status",
        ("FIX.4.4", "773", "2"): "2 .. Confirmation",
        ("FIX.4.4", "773", "3"): "3 .. ConfirmationRequestRejected",
        ("FIX.4.4", "774", "1"): "1 .. MismatchedAccount",
        ("FIX.4.4", "774", "2"): "2 .. MissingSettlementInstructions",
        ("FIX.4.4", "774", "99"): "99 .. Other",
        ("FIX.4.4", "775", "0"): "0 .. RegularBooking",
        ("FIX.4.4", "775", "1"): "1 .. CFD",
        ("FIX.4.4", "775", "2"): "2 .. TotalReturnSwap",
        ("FIX.4.4", "780", "0"): "0 .. UseDefaultInstructions",
        ("FIX.4.4", "780", "1"): "1 .. DeriveFromParametersProvided",
        ("FIX.4.4", "780", "2"): "2 .. FullDetailsProvided",
        ("FIX.4.4", "780", "3"): "3 .. SSIDBIDsProvided",
        ("FIX.4.4", "780", "4"): "4 .. PhoneForInstructions",
        ("FIX.4.4", "787", "S"): "S .. Securities",
        ("FIX.4.4", "787", "C"): "C .. Cash",
        ("FIX.4.4", "788", "1"): "1 .. Overnight",
        ("FIX.4.4", "788", "2"): "2 .. Term",
        ("FIX.4.4", "788", "3"): "3 .. Flexible",
        ("FIX.4.4", "788", "4"): "4 .. Open",
        ("FIX.4.4", "792", "0"): "0 .. UnableToProcessRequest",
        ("FIX.4.4", "792", "1"): "1 .. UnknownAccount",
        ("FIX.4.4", "792", "2"): "2 .. NoMatchingSettlementInstructionsFound",
        ("FIX.4.4", "792", "99"): "99 .. Other",
        ("FIX.4.4", "794", "3"): "3 .. SellsideCalculatedUsingPreliminary",
        ("FIX.4.4", "794", "4"): "4 .. SellsideCalculatedWithoutPreliminary",
        ("FIX.4.4", "794", "5"): "5 .. WarehouseRecap",
        ("FIX.4.4", "794", "8"): "8 .. RequestToIntermediary",
        ("FIX.4.4", "796", "1"): "1 .. OriginalDetailsIncomplete",
        ("FIX.4.4", "796", "2"): "2 .. ChangeInUnderlyingOrderDetails",
        ("FIX.4.4", "796", "99"): "99 .. Other",
        ("FIX.4.4", "798", "1"): "1 .. CarriedCustomerSide",
        ("FIX.4.4", "798", "2"): "2 .. CarriedNonCustomerSide",
        ("FIX.4.4", "798", "3"): "3 .. HouseTrader",
        ("FIX.4.4", "798", "4"): "4 .. FloorTrader",
        ("FIX.4.4", "798", "6"): "6 .. CarriedNonCustomerSideCrossMargined",
        ("FIX.4.4", "798", "7"): "7 .. HouseTraderCrossMargined",
        ("FIX.4.4", "798", "8"): "8 .. JointBackOfficeAccount",
        ("FIX.4.4", "803", "1"): "1 .. Firm",
        ("FIX.4.4", "803", "2"): "2 .. Person",
        ("FIX.4.4", "803", "3"): "3 .. System",
        ("FIX.4.4", "803", "4"): "4 .. Application",
        ("FIX.4.4", "803", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.4.4", "803", "6"): "6 .. PostalAddress",
        ("FIX.4.4", "803", "7"): "7 .. PhoneNumber",
        ("FIX.4.4", "803", "8"): "8 .. EmailAddress",
        ("FIX.4.4", "803", "9"): "9 .. ContactName",
        ("FIX.4.4", "803", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.4.4", "803", "11"): "11 .. RegistrationNumber",
        ("FIX.4.4", "803", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.4.4", "803", "13"): "13 .. RegulatoryStatus",
        ("FIX.4.4", "803", "14"): "14 .. RegistrationName",
        ("FIX.4.4", "803", "15"): "15 .. CashAccountNumber",
        ("FIX.4.4", "803", "16"): "16 .. BIC",
        ("FIX.4.4", "803", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.4.4", "803", "18"): "18 .. RegisteredAddress",
        ("FIX.4.4", "803", "19"): "19 .. FundAccountName",
        ("FIX.4.4", "803", "20"): "20 .. TelexNumber",
        ("FIX.4.4", "803", "21"): "21 .. FaxNumber",
        ("FIX.4.4", "803", "22"): "22 .. SecuritiesAccountName",
        ("FIX.4.4", "803", "23"): "23 .. CashAccountName",
        ("FIX.4.4", "803", "24"): "24 .. Department",
        ("FIX.4.4", "803", "25"): "25 .. LocationDesk",
        ("FIX.4.4", "803", "26"): "26 .. PositionAccountType",
        ("FIX.4.4", "808", "1"): "1 .. PendingAccept",
        ("FIX.4.4", "808", "2"): "2 .. PendingRelease",
        ("FIX.4.4", "808", "3"): "3 .. PendingReversal",
        ("FIX.4.4", "808", "4"): "4 .. Accept",
        ("FIX.4.4", "808", "5"): "5 .. BlockLevelReject",
        ("FIX.4.4", "808", "6"): "6 .. AccountLevelReject",
        ("FIX.4.4", "814", "0"): "0 .. NoActionTaken",
        ("FIX.4.4", "814", "1"): "1 .. QueueFlushed",
        ("FIX.4.4", "814", "2"): "2 .. OverlayLast",
        ("FIX.4.4", "814", "3"): "3 .. EndSession",
        ("FIX.4.4", "815", "0"): "0 .. NoActionTaken",
        ("FIX.4.4", "815", "1"): "1 .. QueueFlushed",
        ("FIX.4.4", "815", "2"): "2 .. OverlayLast",
        ("FIX.4.4", "815", "3"): "3 .. EndSession",
        ("FIX.4.4", "819", "0"): "0 .. NoAveragePricing",
        ("FIX.4.4", "819", "1"): "1 .. Trade",
        ("FIX.4.4", "819", "2"): "2 .. LastTrade",
        ("FIX.4.4", "826", "0"): "0 .. AllocationNotRequired",
        ("FIX.4.4", "826", "1"): "1 .. AllocationRequired",
        ("FIX.4.4", "826", "2"): "2 .. UseAllocationProvidedWithTheTrade",
        ("FIX.4.4", "827", "0"): "0 .. ExpireOnTradingSessionClose",
        ("FIX.4.4", "827", "1"): "1 .. ExpireOnTradingSessionOpen",
        ("FIX.4.4", "828", "0"): "0 .. RegularTrade",
        ("FIX.4.4", "828", "1"): "1 .. BlockTrade",
        ("FIX.4.4", "828", "2"): "2 .. EFP",
        ("FIX.4.4", "828", "3"): "3 .. Transfer",
        ("FIX.4.4", "828", "4"): "4 .. LateTrade",
        ("FIX.4.4", "828", "5"): "5 .. TTrade",
        ("FIX.4.4", "828", "6"): "6 .. WeightedAveragePriceTrade",
        ("FIX.4.4", "828", "7"): "7 .. BunchedTrade",
        ("FIX.4.4", "828", "8"): "8 .. LateBunchedTrade",
        ("FIX.4.4", "828", "9"): "9 .. PriorReferencePriceTrade",
        ("FIX.4.4", "828", "10"): "10 .. AfterHoursTrade",
        ("FIX.4.4", "835", "0"): "0 .. Floating",
        ("FIX.4.4", "835", "1"): "1 .. Fixed",
        ("FIX.4.4", "836", "0"): "0 .. Price",
        ("FIX.4.4", "836", "1"): "1 .. BasisPoints",
        ("FIX.4.4", "836", "2"): "2 .. Ticks",
        ("FIX.4.4", "836", "3"): "3 .. PriceTier",
        ("FIX.4.4", "837", "0"): "0 .. OrBetter",
        ("FIX.4.4", "837", "1"): "1 .. Strict",
        ("FIX.4.4", "837", "2"): "2 .. OrWorse",
        ("FIX.4.4", "838", "1"): "1 .. MoreAggressive",
        ("FIX.4.4", "838", "2"): "2 .. MorePassive",
        ("FIX.4.4", "840", "1"): "1 .. Local",
        ("FIX.4.4", "840", "2"): "2 .. National",
        ("FIX.4.4", "840", "3"): "3 .. Global",
        ("FIX.4.4", "840", "4"): "4 .. NationalExcludingLocal",
        ("FIX.4.4", "841", "0"): "0 .. Floating",
        ("FIX.4.4", "841", "1"): "1 .. Fixed",
        ("FIX.4.4", "842", "0"): "0 .. Price",
        ("FIX.4.4", "842", "1"): "1 .. BasisPoints",
        ("FIX.4.4", "842", "2"): "2 .. Ticks",
        ("FIX.4.4", "842", "3"): "3 .. PriceTier",
        ("FIX.4.4", "843", "0"): "0 .. OrBetter",
        ("FIX.4.4", "843", "1"): "1 .. Strict",
        ("FIX.4.4", "843", "2"): "2 .. OrWorse",
        ("FIX.4.4", "844", "1"): "1 .. MoreAggressive",
        ("FIX.4.4", "844", "2"): "2 .. MorePassive",
        ("FIX.4.4", "846", "1"): "1 .. Local",
        ("FIX.4.4", "846", "2"): "2 .. National",
        ("FIX.4.4", "846", "3"): "3 .. Global",
        ("FIX.4.4", "846", "4"): "4 .. NationalExcludingLocal",
        ("FIX.4.4", "847", "1"): "1 .. VWAP",
        ("FIX.4.4", "847", "2"): "2 .. Participate",
        ("FIX.4.4", "847", "3"): "3 .. MininizeMarketImpact",
        ("FIX.4.4", "851", "1"): "1 .. AddedLiquidity",
        ("FIX.4.4", "851", "2"): "2 .. RemovedLiquidity",
        ("FIX.4.4", "851", "3"): "3 .. LiquidityRoutedOut",
        ("FIX.4.4", "852", "Y"): "Y .. ReportTrade",
        ("FIX.4.4", "852", "N"): "N .. DoNotReportTrade",
        ("FIX.4.4", "853", "0"): "0 .. DealerSoldShort",
        ("FIX.4.4", "853", "1"): "1 .. DealerSoldShortExempt",
        ("FIX.4.4", "853", "2"): "2 .. SellingCustomerSoldShort",
        ("FIX.4.4", "853", "3"): "3 .. SellingCustomerSoldShortExempt",
        ("FIX.4.4", "853", "4"): "4 .. QualifiedServiceRepresentative",
        ("FIX.4.4", "853", "5"): "5 .. QSROrAGUContraSideSoldShortExempt",
        ("FIX.4.4", "854", "0"): "0 .. Units",
        ("FIX.4.4", "854", "1"): "1 .. Contracts",
        ("FIX.4.4", "856", "0"): "0 .. Submit",
        ("FIX.4.4", "856", "1"): "1 .. Alleged",
        ("FIX.4.4", "856", "2"): "2 .. Accept",
        ("FIX.4.4", "856", "3"): "3 .. Decline",
        ("FIX.4.4", "856", "4"): "4 .. Addendum",
        ("FIX.4.4", "856", "5"): "5 .. No",
        ("FIX.4.4", "856", "6"): "6 .. TradeReportCancel",
        ("FIX.4.4", "856", "7"): "7 .. LockedIn",
        ("FIX.4.4", "857", "0"): "0 .. NotSpecified",
        ("FIX.4.4", "857", "1"): "1 .. ExplicitListProvided",
        ("FIX.4.4", "865", "1"): "1 .. Put",
        ("FIX.4.4", "865", "2"): "2 .. Call",
        ("FIX.4.4", "865", "3"): "3 .. Tender",
        ("FIX.4.4", "865", "4"): "4 .. SinkingFundCall",
        ("FIX.4.4", "865", "99"): "99 .. Other",
        ("FIX.4.4", "871", "1"): "1 .. Flat",
        ("FIX.4.4", "871", "2"): "2 .. ZeroCoupon",
        ("FIX.4.4", "871", "3"): "3 .. InterestBearing",
        ("FIX.4.4", "871", "4"): "4 .. NoPeriodicPayments",
        ("FIX.4.4", "871", "5"): "5 .. VariableRate",
        ("FIX.4.4", "871", "6"): "6 .. LessFeeForPut",
        ("FIX.4.4", "871", "7"): "7 .. SteppedCoupon",
        ("FIX.4.4", "871", "8"): "8 .. CouponPeriod",
        ("FIX.4.4", "871", "9"): "9 .. When",
        ("FIX.4.4", "871", "10"): "10 .. OriginalIssueDiscount",
        ("FIX.4.4", "871", "11"): "11 .. Callable",
        ("FIX.4.4", "871", "12"): "12 .. EscrowedToMaturity",
        ("FIX.4.4", "871", "13"): "13 .. EscrowedToRedemptionDate",
        ("FIX.4.4", "871", "14"): "14 .. PreRefunded",
        ("FIX.4.4", "871", "15"): "15 .. InDefault",
        ("FIX.4.4", "871", "16"): "16 .. Unrated",
        ("FIX.4.4", "871", "17"): "17 .. Taxable",
        ("FIX.4.4", "871", "18"): "18 .. Indexed",
        ("FIX.4.4", "871", "19"): "19 .. SubjectToAlternativeMinimumTax",
        ("FIX.4.4", "871", "20"): "20 .. OriginalIssueDiscountPrice",
        ("FIX.4.4", "871", "21"): "21 .. CallableBelowMaturityValue",
        ("FIX.4.4", "871", "22"): "22 .. CallableWithoutNotice",
        ("FIX.4.4", "871", "99"): "99 .. Text",
        ("FIX.4.4", "875", "1"): "1 .. Program3a3",
        ("FIX.4.4", "875", "2"): "2 .. Program42",
        ("FIX.4.4", "875", "99"): "99 .. Other",
        ("FIX.4.4", "891", "0"): "0 .. Absolute",
        ("FIX.4.4", "891", "1"): "1 .. PerUnit",
        ("FIX.4.4", "891", "2"): "2 .. Percentage",
        ("FIX.4.4", "893", "Y"): "Y .. LastMessage",
        ("FIX.4.4", "893", "N"): "N .. NotLastMessage",
        ("FIX.4.4", "895", "0"): "0 .. Initial",
        ("FIX.4.4", "895", "1"): "1 .. Scheduled",
        ("FIX.4.4", "895", "2"): "2 .. TimeWarning",
        ("FIX.4.4", "895", "3"): "3 .. MarginDeficiency",
        ("FIX.4.4", "895", "4"): "4 .. MarginExcess",
        ("FIX.4.4", "895", "5"): "5 .. ForwardCollateralDemand",
        ("FIX.4.4", "895", "6"): "6 .. EventOfDefault",
        ("FIX.4.4", "895", "7"): "7 .. AdverseTaxEvent",
        ("FIX.4.4", "896", "0"): "0 .. TradeDate",
        ("FIX.4.4", "896", "1"): "1 .. GCInstrument",
        ("FIX.4.4", "896", "2"): "2 .. CollateralInstrument",
        ("FIX.4.4", "896", "3"): "3 .. SubstitutionEligible",
        ("FIX.4.4", "896", "4"): "4 .. NotAssigned",
        ("FIX.4.4", "896", "5"): "5 .. PartiallyAssigned",
        ("FIX.4.4", "896", "6"): "6 .. FullyAssigned",
        ("FIX.4.4", "896", "7"): "7 .. OutstandingTrades",
        ("FIX.4.4", "903", "0"): "0 .. New",
        ("FIX.4.4", "903", "1"): "1 .. Replace",
        ("FIX.4.4", "903", "2"): "2 .. Cancel",
        ("FIX.4.4", "903", "3"): "3 .. Release",
        ("FIX.4.4", "903", "4"): "4 .. Reverse",
        ("FIX.4.4", "905", "0"): "0 .. Received",
        ("FIX.4.4", "905", "1"): "1 .. Accepted",
        ("FIX.4.4", "905", "2"): "2 .. Declined",
        ("FIX.4.4", "905", "3"): "3 .. Rejected",
        ("FIX.4.4", "906", "0"): "0 .. UnknownDeal",
        ("FIX.4.4", "906", "1"): "1 .. UnknownOrInvalidInstrument",
        ("FIX.4.4", "906", "2"): "2 .. UnauthorizedTransaction",
        ("FIX.4.4", "906", "3"): "3 .. InsufficientCollateral",
        ("FIX.4.4", "906", "4"): "4 .. InvalidTypeOfCollateral",
        ("FIX.4.4", "906", "5"): "5 .. ExcessiveSubstitution",
        ("FIX.4.4", "906", "99"): "99 .. Other",
        ("FIX.4.4", "910", "0"): "0 .. Unassigned",
        ("FIX.4.4", "910", "1"): "1 .. PartiallyAssigned",
        ("FIX.4.4", "910", "2"): "2 .. AssignmentProposed",
        ("FIX.4.4", "910", "3"): "3 .. Assigned",
        ("FIX.4.4", "910", "4"): "4 .. Challenged",
        ("FIX.4.4", "919", "0"): "0 .. VersusPayment",
        ("FIX.4.4", "919", "1"): "1 .. Free",
        ("FIX.4.4", "919", "2"): "2 .. TriParty",
        ("FIX.4.4", "919", "3"): "3 .. HoldInCustody",
        ("FIX.4.4", "924", "1"): "1 .. LogOnUser",
        ("FIX.4.4", "924", "2"): "2 .. LogOffUser",
        ("FIX.4.4", "924", "3"): "3 .. ChangePasswordForUser",
        ("FIX.4.4", "924", "4"): "4 .. RequestIndividualUserStatus",
        ("FIX.4.4", "926", "1"): "1 .. LoggedIn",
        ("FIX.4.4", "926", "2"): "2 .. NotLoggedIn",
        ("FIX.4.4", "926", "3"): "3 .. UserNotRecognised",
        ("FIX.4.4", "926", "4"): "4 .. PasswordIncorrect",
        ("FIX.4.4", "926", "5"): "5 .. PasswordChanged",
        ("FIX.4.4", "926", "6"): "6 .. Other",
        ("FIX.4.4", "928", "1"): "1 .. Connected",
        ("FIX.4.4", "928", "2"): "2 .. NotConnectedUnexpected",
        ("FIX.4.4", "928", "3"): "3 .. NotConnectedExpected",
        ("FIX.4.4", "928", "4"): "4 .. InProcess",
        ("FIX.4.4", "935", "1"): "1 .. Snapshot",
        ("FIX.4.4", "935", "2"): "2 .. Subscribe",
        ("FIX.4.4", "935", "4"): "4 .. StopSubscribing",
        ("FIX.4.4", "935", "8"): "8 .. LevelOfDetail",
        ("FIX.4.4", "937", "1"): "1 .. Full",
        ("FIX.4.4", "937", "2"): "2 .. IncrementalUpdate",
        ("FIX.4.4", "939", "0"): "0 .. Accepted",
        ("FIX.4.4", "939", "1"): "1 .. Rejected",
        ("FIX.4.4", "940", "1"): "1 .. Received",
        ("FIX.4.4", "940", "2"): "2 .. ConfirmRejected",
        ("FIX.4.4", "940", "3"): "3 .. Affirmed",
        ("FIX.4.4", "944", "0"): "0 .. Retain",
        ("FIX.4.4", "944", "1"): "1 .. Add",
        ("FIX.4.4", "944", "2"): "2 .. Remove",
        ("FIX.4.4", "945", "0"): "0 .. Accepted",
        ("FIX.4.4", "945", "1"): "1 .. AcceptedWithWarnings",
        ("FIX.4.4", "945", "2"): "2 .. Completed",
        ("FIX.4.4", "945", "3"): "3 .. CompletedWithWarnings",
        ("FIX.4.4", "945", "4"): "4 .. Rejected",
        ("FIX.4.4", "946", "0"): "0 .. Successful",
        ("FIX.4.4", "946", "1"): "1 .. InvalidOrUnknownInstrument",
        ("FIX.4.4", "946", "2"): "2 .. InvalidOrUnknownCollateralType",
        ("FIX.4.4", "946", "3"): "3 .. InvalidParties",
        ("FIX.4.4", "946", "4"): "4 .. InvalidTransportTypeRequested",
        ("FIX.4.4", "946", "5"): "5 .. InvalidDestinationRequested",
        ("FIX.4.4", "946", "6"): "6 .. NoCollateralFoundForTheTradeSpecified",
        ("FIX.4.4", "946", "7"): "7 .. NoCollateralFoundForTheOrderSpecified",
        ("FIX.4.4", "946", "8"): "8 .. CollateralInquiryTypeNotSupported",
        ("FIX.4.4", "946", "9"): "9 .. UnauthorizedForCollateralInquiry",
        ("FIX.4.4", "946", "99"): "99 .. Other",
        ("FIX.5.0", "4", "B"): "B .. Buy",
        ("FIX.5.0", "4", "S"): "S .. Sell",
        ("FIX.5.0", "4", "T"): "T .. Trade",
        ("FIX.5.0", "4", "X"): "X .. Cross",
        ("FIX.5.0", "5", "N"): "N .. New",
        ("FIX.5.0", "5", "C"): "C .. Cancel",
        ("FIX.5.0", "5", "R"): "R .. Replace",
        ("FIX.5.0", "13", "1"): "1 .. PerUnit",
        ("FIX.5.0", "13", "2"): "2 .. Percent",
        ("FIX.5.0", "13", "3"): "3 .. Absolute",
        ("FIX.5.0", "13", "4"): "4 .. PercentageWaivedCashDiscount",
        ("FIX.5.0", "13", "5"): "5 .. PercentageWaivedEnhancedUnits",
        ("FIX.5.0", "13", "6"): "6 .. PointsPerBondOrContract",
        ("FIX.5.0", "13", "7"): "7 .. BasisPoints",
        ("FIX.5.0", "13", "8"): "8 .. AmountPerContract",
        ("FIX.5.0", "18", "0"): "0 .. StayOnOfferSide",
        ("FIX.5.0", "18", "1"): "1 .. NotHeld",
        ("FIX.5.0", "18", "2"): "2 .. Work",
        ("FIX.5.0", "18", "3"): "3 .. GoAlong",
        ("FIX.5.0", "18", "4"): "4 .. OverTheDay",
        ("FIX.5.0", "18", "5"): "5 .. Held",
        ("FIX.5.0", "18", "6"): "6 .. ParticipateDoNotInitiate",
        ("FIX.5.0", "18", "7"): "7 .. StrictScale",
        ("FIX.5.0", "18", "8"): "8 .. TryToScale",
        ("FIX.5.0", "18", "9"): "9 .. StayOnBidSide",
        ("FIX.5.0", "18", "A"): "A .. NoCross",
        ("FIX.5.0", "18", "B"): "B .. OKToCross",
        ("FIX.5.0", "18", "C"): "C .. CallFirst",
        ("FIX.5.0", "18", "D"): "D .. PercentOfVolume",
        ("FIX.5.0", "18", "E"): "E .. DoNotIncrease",
        ("FIX.5.0", "18", "F"): "F .. DoNotReduce",
        ("FIX.5.0", "18", "G"): "G .. AllOrNone",
        ("FIX.5.0", "18", "H"): "H .. ReinstateOnSystemFailure",
        ("FIX.5.0", "18", "I"): "I .. InstitutionsOnly",
        ("FIX.5.0", "18", "J"): "J .. ReinstateOnTradingHalt",
        ("FIX.5.0", "18", "K"): "K .. CancelOnTradingHalt",
        ("FIX.5.0", "18", "N"): "N .. NonNegotiable",
        ("FIX.5.0", "18", "Q"): "Q .. CancelOnSystemFailure",
        ("FIX.5.0", "18", "S"): "S .. Suspend",
        ("FIX.5.0", "18", "U"): "U .. CustomerDisplayInstruction",
        ("FIX.5.0", "18", "V"): "V .. Netting",
        ("FIX.5.0", "18", "X"): "X .. TradeAlong",
        ("FIX.5.0", "18", "Y"): "Y .. TryToStop",
        ("FIX.5.0", "18", "Z"): "Z .. CancelIfNotBest",
        ("FIX.5.0", "18", "b"): "b .. StrictLimit",
        ("FIX.5.0", "18", "c"): "c .. IgnorePriceValidityChecks",
        ("FIX.5.0", "18", "e"): "e .. WorkToTargetStrategy",
        ("FIX.5.0", "18", "f"): "f .. IntermarketSweep",
        ("FIX.5.0", "18", "g"): "g .. ExternalRoutingAllowed",
        ("FIX.5.0", "18", "h"): "h .. ExternalRoutingNotAllowed",
        ("FIX.5.0", "18", "i"): "i .. ImbalanceOnly",
        ("FIX.5.0", "18", "j"): "j .. SingleExecutionRequestedForBlockTrade",
        ("FIX.5.0", "18", "k"): "k .. BestExecution",
        ("FIX.5.0", "18", "l"): "l .. SuspendOnSystemFailure",
        ("FIX.5.0", "18", "m"): "m .. SuspendOnTradingHalt",
        ("FIX.5.0", "18", "n"): "n .. ReinstateOnConnectionLoss",
        ("FIX.5.0", "18", "o"): "o .. CancelOnConnectionLoss",
        ("FIX.5.0", "18", "p"): "p .. SuspendOnConnectionLoss",
        ("FIX.5.0", "18", "q"): "q .. Release",
        ("FIX.5.0", "18", "r"): "r .. ExecuteAsDeltaNeutral",
        ("FIX.5.0", "18", "s"): "s .. ExecuteAsDurationNeutral",
        ("FIX.5.0", "18", "t"): "t .. ExecuteAsFXNeutral",
        ("FIX.5.0", "18", "u"): "u .. MinGuaranteedFillEligible",
        ("FIX.5.0", "18", "v"): "v .. BypassNonDisplayLiquidity",
        ("FIX.5.0", "18", "w"): "w .. Lock",
        ("FIX.5.0", "18", "x"): "x .. IgnoreNotionalValueChecks",
        ("FIX.5.0", "18", "y"): "y .. TrdAtRefPx",
        ("FIX.5.0", "21", "1"): "1 .. AutomatedExecutionNoIntervention",
        ("FIX.5.0", "21", "2"): "2 .. AutomatedExecutionInterventionOK",
        ("FIX.5.0", "21", "3"): "3 .. ManualOrder",
        ("FIX.5.0", "22", "1"): "1 .. CUSIP",
        ("FIX.5.0", "22", "2"): "2 .. SEDOL",
        ("FIX.5.0", "22", "3"): "3 .. QUIK",
        ("FIX.5.0", "22", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "22", "5"): "5 .. RICCode",
        ("FIX.5.0", "22", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "22", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "22", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "22", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "22", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "22", "B"): "B .. Wertpapier",
        ("FIX.5.0", "22", "C"): "C .. Dutch",
        ("FIX.5.0", "22", "D"): "D .. Valoren",
        ("FIX.5.0", "22", "E"): "E .. Sicovam",
        ("FIX.5.0", "22", "F"): "F .. Belgian",
        ("FIX.5.0", "22", "G"): "G .. Common",
        ("FIX.5.0", "22", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "22", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "22", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "22", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "22", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "22", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "22", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "22", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "22", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "22", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "22", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "22", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "22", "U"): "U .. Synthetic",
        ("FIX.5.0", "22", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "22", "W"): "W .. IndexName",
        ("FIX.5.0", "25", "H"): "H .. High",
        ("FIX.5.0", "25", "L"): "L .. Low",
        ("FIX.5.0", "25", "M"): "M .. Medium",
        ("FIX.5.0", "27", "S"): "S .. Small",
        ("FIX.5.0", "27", "M"): "M .. Medium",
        ("FIX.5.0", "27", "L"): "L .. Large",
        ("FIX.5.0", "27", "U"): "U .. UndisclosedQuantity",
        ("FIX.5.0", "28", "N"): "N .. New",
        ("FIX.5.0", "28", "C"): "C .. Cancel",
        ("FIX.5.0", "28", "R"): "R .. Replace",
        ("FIX.5.0", "29", "1"): "1 .. Agent",
        ("FIX.5.0", "29", "2"): "2 .. CrossAsAgent",
        ("FIX.5.0", "29", "3"): "3 .. CrossAsPrincipal",
        ("FIX.5.0", "29", "4"): "4 .. Principal",
        ("FIX.5.0", "29", "5"): "5 .. RisklessPrincipal",
        ("FIX.5.0", "35", "0"): "0 .. Heartbeat",
        ("FIX.5.0", "35", "1"): "1 .. TestRequest",
        ("FIX.5.0", "35", "2"): "2 .. ResendRequest",
        ("FIX.5.0", "35", "3"): "3 .. Reject",
        ("FIX.5.0", "35", "4"): "4 .. SequenceReset",
        ("FIX.5.0", "35", "5"): "5 .. Logout",
        ("FIX.5.0", "35", "6"): "6 .. IOI",
        ("FIX.5.0", "35", "7"): "7 .. Advertisement",
        ("FIX.5.0", "35", "8"): "8 .. ExecutionReport",
        ("FIX.5.0", "35", "9"): "9 .. OrderCancelReject",
        ("FIX.5.0", "35", "A"): "A .. Logon",
        ("FIX.5.0", "35", "B"): "B .. News",
        ("FIX.5.0", "35", "C"): "C .. Email",
        ("FIX.5.0", "35", "D"): "D .. NewOrderSingle",
        ("FIX.5.0", "35", "E"): "E .. NewOrderList",
        ("FIX.5.0", "35", "F"): "F .. OrderCancelRequest",
        ("FIX.5.0", "35", "G"): "G .. OrderCancelReplaceRequest",
        ("FIX.5.0", "35", "H"): "H .. OrderStatusRequest",
        ("FIX.5.0", "35", "J"): "J .. AllocationInstruction",
        ("FIX.5.0", "35", "K"): "K .. ListCancelRequest",
        ("FIX.5.0", "35", "L"): "L .. ListExecute",
        ("FIX.5.0", "35", "M"): "M .. ListStatusRequest",
        ("FIX.5.0", "35", "N"): "N .. ListStatus",
        ("FIX.5.0", "35", "P"): "P .. AllocationInstructionAck",
        ("FIX.5.0", "35", "Q"): "Q .. DontKnowTrade",
        ("FIX.5.0", "35", "R"): "R .. QuoteRequest",
        ("FIX.5.0", "35", "S"): "S .. Quote",
        ("FIX.5.0", "35", "T"): "T .. SettlementInstructions",
        ("FIX.5.0", "35", "V"): "V .. MarketDataRequest",
        ("FIX.5.0", "35", "W"): "W .. MarketDataSnapshotFullRefresh",
        ("FIX.5.0", "35", "X"): "X .. MarketDataIncrementalRefresh",
        ("FIX.5.0", "35", "Y"): "Y .. MarketDataRequestReject",
        ("FIX.5.0", "35", "Z"): "Z .. QuoteCancel",
        ("FIX.5.0", "35", "a"): "a .. QuoteStatusRequest",
        ("FIX.5.0", "35", "b"): "b .. MassQuoteAck",
        ("FIX.5.0", "35", "c"): "c .. SecurityDefinitionRequest",
        ("FIX.5.0", "35", "d"): "d .. SecurityDefinition",
        ("FIX.5.0", "35", "e"): "e .. SecurityStatusRequest",
        ("FIX.5.0", "35", "f"): "f .. SecurityStatus",
        ("FIX.5.0", "35", "g"): "g .. TradingSessionStatusRequest",
        ("FIX.5.0", "35", "h"): "h .. TradingSessionStatus",
        ("FIX.5.0", "35", "i"): "i .. MassQuote",
        ("FIX.5.0", "35", "j"): "j .. BusinessMessageReject",
        ("FIX.5.0", "35", "k"): "k .. BidRequest",
        ("FIX.5.0", "35", "l"): "l .. BidResponse",
        ("FIX.5.0", "35", "m"): "m .. ListStrikePrice",
        ("FIX.5.0", "35", "n"): "n .. XMLnonFIX",
        ("FIX.5.0", "35", "o"): "o .. RegistrationInstructions",
        ("FIX.5.0", "35", "p"): "p .. RegistrationInstructionsResponse",
        ("FIX.5.0", "35", "q"): "q .. OrderMassCancelRequest",
        ("FIX.5.0", "35", "r"): "r .. OrderMassCancelReport",
        ("FIX.5.0", "35", "s"): "s .. NewOrderCross",
        ("FIX.5.0", "35", "t"): "t .. CrossOrderCancelReplaceRequest",
        ("FIX.5.0", "35", "u"): "u .. CrossOrderCancelRequest",
        ("FIX.5.0", "35", "v"): "v .. SecurityTypeRequest",
        ("FIX.5.0", "35", "w"): "w .. SecurityTypes",
        ("FIX.5.0", "35", "x"): "x .. SecurityListRequest",
        ("FIX.5.0", "35", "y"): "y .. SecurityList",
        ("FIX.5.0", "35", "z"): "z .. DerivativeSecurityListRequest",
        ("FIX.5.0", "35", "AA"): "AA .. DerivativeSecurityList",
        ("FIX.5.0", "35", "AB"): "AB .. NewOrderMultileg",
        ("FIX.5.0", "35", "AC"): "AC .. MultilegOrderCancelReplace",
        ("FIX.5.0", "35", "AD"): "AD .. TradeCaptureReportRequest",
        ("FIX.5.0", "35", "AE"): "AE .. TradeCaptureReport",
        ("FIX.5.0", "35", "AF"): "AF .. OrderMassStatusRequest",
        ("FIX.5.0", "35", "AG"): "AG .. QuoteRequestReject",
        ("FIX.5.0", "35", "AH"): "AH .. RFQRequest",
        ("FIX.5.0", "35", "AI"): "AI .. QuoteStatusReport",
        ("FIX.5.0", "35", "AJ"): "AJ .. QuoteResponse",
        ("FIX.5.0", "35", "AK"): "AK .. Confirmation",
        ("FIX.5.0", "35", "AL"): "AL .. PositionMaintenanceRequest",
        ("FIX.5.0", "35", "AM"): "AM .. PositionMaintenanceReport",
        ("FIX.5.0", "35", "AN"): "AN .. RequestForPositions",
        ("FIX.5.0", "35", "AO"): "AO .. RequestForPositionsAck",
        ("FIX.5.0", "35", "AP"): "AP .. PositionReport",
        ("FIX.5.0", "35", "AQ"): "AQ .. TradeCaptureReportRequestAck",
        ("FIX.5.0", "35", "AR"): "AR .. TradeCaptureReportAck",
        ("FIX.5.0", "35", "AS"): "AS .. AllocationReport",
        ("FIX.5.0", "35", "AT"): "AT .. AllocationReportAck",
        ("FIX.5.0", "35", "AU"): "AU .. ConfirmationAck",
        ("FIX.5.0", "35", "AV"): "AV .. SettlementInstructionRequest",
        ("FIX.5.0", "35", "AW"): "AW .. AssignmentReport",
        ("FIX.5.0", "35", "AX"): "AX .. CollateralRequest",
        ("FIX.5.0", "35", "AY"): "AY .. CollateralAssignment",
        ("FIX.5.0", "35", "AZ"): "AZ .. CollateralResponse",
        ("FIX.5.0", "35", "BA"): "BA .. CollateralReport",
        ("FIX.5.0", "35", "BB"): "BB .. CollateralInquiry",
        ("FIX.5.0", "35", "BC"): "BC .. NetworkCounterpartySystemStatusRequest",
        ("FIX.5.0", "35", "BD"): "BD .. NetworkCounterpartySystemStatusResponse",
        ("FIX.5.0", "35", "BE"): "BE .. UserRequest",
        ("FIX.5.0", "35", "BF"): "BF .. UserResponse",
        ("FIX.5.0", "35", "BG"): "BG .. CollateralInquiryAck",
        ("FIX.5.0", "35", "BH"): "BH .. ConfirmationRequest",
        ("FIX.5.0", "35", "BO"): "BO .. ContraryIntentionReport",
        ("FIX.5.0", "35", "BP"): "BP .. SecurityDefinitionUpdateReport",
        ("FIX.5.0", "35", "BK"): "BK .. SecurityListUpdateReport",
        ("FIX.5.0", "35", "BL"): "BL .. AdjustedPositionReport",
        ("FIX.5.0", "35", "BM"): "BM .. AllocationInstructionAlert",
        ("FIX.5.0", "35", "BN"): "BN .. ExecutionAck",
        ("FIX.5.0", "35", "BJ"): "BJ .. TradingSessionList",
        ("FIX.5.0", "35", "BI"): "BI .. TradingSessionListRequest",
        ("FIX.5.0", "35", "BQ"): "BQ .. SettlementObligationReport",
        ("FIX.5.0", "35", "BR"): "BR .. DerivativeSecurityListUpdateReport",
        ("FIX.5.0", "35", "BS"): "BS .. TradingSessionListUpdateReport",
        ("FIX.5.0", "35", "BT"): "BT .. MarketDefinitionRequest",
        ("FIX.5.0", "35", "BU"): "BU .. MarketDefinition",
        ("FIX.5.0", "35", "BV"): "BV .. MarketDefinitionUpdateReport",
        ("FIX.5.0", "35", "BW"): "BW .. ApplicationMessageRequest",
        ("FIX.5.0", "35", "BX"): "BX .. ApplicationMessageRequestAck",
        ("FIX.5.0", "35", "BY"): "BY .. ApplicationMessageReport",
        ("FIX.5.0", "35", "BZ"): "BZ .. OrderMassActionReport",
        ("FIX.5.0", "35", "CA"): "CA .. OrderMassActionRequest",
        ("FIX.5.0", "35", "CB"): "CB .. UserNotification",
        ("FIX.5.0", "35", "CC"): "CC .. StreamAssignmentRequest",
        ("FIX.5.0", "35", "CD"): "CD .. StreamAssignmentReport",
        ("FIX.5.0", "35", "CE"): "CE .. StreamAssignmentReportACK",
        ("FIX.5.0", "35", "CF"): "CF .. PartyDetailsListRequest",
        ("FIX.5.0", "35", "CG"): "CG .. PartyDetailsListReport",
        ("FIX.5.0", "35", "CH"): "CH .. MarginRequirementInquiry",
        ("FIX.5.0", "35", "CI"): "CI .. MarginRequirementInquiryAck",
        ("FIX.5.0", "35", "CJ"): "CJ .. MarginRequirementReport",
        ("FIX.5.0", "35", "CK"): "CK .. PartyDetailsListUpdateReport",
        ("FIX.5.0", "35", "CL"): "CL .. PartyRiskLimitsRequest",
        ("FIX.5.0", "35", "CM"): "CM .. PartyRiskLimitsReport",
        ("FIX.5.0", "35", "CN"): "CN .. SecurityMassStatusRequest",
        ("FIX.5.0", "35", "CO"): "CO .. SecurityMassStatus",
        ("FIX.5.0", "35", "CQ"): "CQ .. AccountSummaryReport",
        ("FIX.5.0", "35", "CR"): "CR .. PartyRiskLimitsUpdateReport",
        ("FIX.5.0", "35", "CS"): "CS .. PartyRiskLimitsDefinitionRequest",
        ("FIX.5.0", "35", "CT"): "CT .. PartyRiskLimitsDefinitionRequestAck",
        ("FIX.5.0", "35", "CU"): "CU .. PartyEntitlementsRequest",
        ("FIX.5.0", "35", "CV"): "CV .. PartyEntitlementsReport",
        ("FIX.5.0", "35", "CW"): "CW .. QuoteAck",
        ("FIX.5.0", "35", "CX"): "CX .. PartyDetailsDefinitionRequest",
        ("FIX.5.0", "35", "CY"): "CY .. PartyDetailsDefinitionRequestAck",
        ("FIX.5.0", "35", "CZ"): "CZ .. PartyEntitlementsUpdateReport",
        ("FIX.5.0", "35", "DA"): "DA .. PartyEntitlementsDefinitionRequest",
        ("FIX.5.0", "35", "DB"): "DB .. PartyEntitlementsDefinitionRequestAck",
        ("FIX.5.0", "35", "DC"): "DC .. TradeMatchReport",
        ("FIX.5.0", "35", "DD"): "DD .. TradeMatchReportAck",
        ("FIX.5.0", "35", "DE"): "DE .. PartyRiskLimitsReportAck",
        ("FIX.5.0", "35", "DF"): "DF .. PartyRiskLimitCheckRequest",
        ("FIX.5.0", "35", "DG"): "DG .. PartyRiskLimitCheckRequestAck",
        ("FIX.5.0", "35", "DH"): "DH .. PartyActionRequest",
        ("FIX.5.0", "35", "DI"): "DI .. PartyActionReport",
        ("FIX.5.0", "35", "DJ"): "DJ .. MassOrder",
        ("FIX.5.0", "35", "DK"): "DK .. MassOrderAck",
        ("FIX.5.0", "35", "DL"): "DL .. PositionTransferInstruction",
        ("FIX.5.0", "35", "DM"): "DM .. PositionTransferInstructionAck",
        ("FIX.5.0", "35", "DN"): "DN .. PositionTransferReport",
        ("FIX.5.0", "35", "DO"): "DO .. MarketDataStatisticsRequest",
        ("FIX.5.0", "35", "DP"): "DP .. MarketDataStatisticsReport",
        ("FIX.5.0", "35", "DQ"): "DQ .. CollateralReportAck",
        ("FIX.5.0", "35", "DR"): "DR .. MarketDataReport",
        ("FIX.5.0", "35", "DS"): "DS .. CrossRequest",
        ("FIX.5.0", "35", "DT"): "DT .. CrossRequestAck",
        ("FIX.5.0", "35", "DU"): "DU .. AllocationInstructionAlertRequest",
        ("FIX.5.0", "39", "0"): "0 .. New",
        ("FIX.5.0", "39", "1"): "1 .. PartiallyFilled",
        ("FIX.5.0", "39", "2"): "2 .. Filled",
        ("FIX.5.0", "39", "3"): "3 .. DoneForDay",
        ("FIX.5.0", "39", "4"): "4 .. Canceled",
        ("FIX.5.0", "39", "6"): "6 .. PendingCancel",
        ("FIX.5.0", "39", "7"): "7 .. Stopped",
        ("FIX.5.0", "39", "8"): "8 .. Rejected",
        ("FIX.5.0", "39", "9"): "9 .. Suspended",
        ("FIX.5.0", "39", "A"): "A .. PendingNew",
        ("FIX.5.0", "39", "B"): "B .. Calculated",
        ("FIX.5.0", "39", "C"): "C .. Expired",
        ("FIX.5.0", "39", "D"): "D .. AcceptedForBidding",
        ("FIX.5.0", "39", "E"): "E .. PendingReplace",
        ("FIX.5.0", "40", "1"): "1 .. Market",
        ("FIX.5.0", "40", "2"): "2 .. Limit",
        ("FIX.5.0", "40", "3"): "3 .. Stop",
        ("FIX.5.0", "40", "4"): "4 .. StopLimit",
        ("FIX.5.0", "40", "6"): "6 .. WithOrWithout",
        ("FIX.5.0", "40", "8"): "8 .. LimitWithOrWithout",
        ("FIX.5.0", "40", "9"): "9 .. OnBasis",
        ("FIX.5.0", "40", "D"): "D .. PreviouslyQuoted",
        ("FIX.5.0", "40", "E"): "E .. PreviouslyIndicated",
        ("FIX.5.0", "40", "G"): "G .. ForexSwap",
        ("FIX.5.0", "40", "I"): "I .. Funari",
        ("FIX.5.0", "40", "J"): "J .. MarketIfTouched",
        ("FIX.5.0", "40", "K"): "K .. MarketWithLeftOverAsLimit",
        ("FIX.5.0", "40", "L"): "L .. PreviousFundValuationPoint",
        ("FIX.5.0", "40", "M"): "M .. NextFundValuationPoint",
        ("FIX.5.0", "40", "P"): "P .. Pegged",
        ("FIX.5.0", "40", "Q"): "Q .. CounterOrderSelection",
        ("FIX.5.0", "40", "R"): "R .. StopOnBidOrOffer",
        ("FIX.5.0", "40", "S"): "S .. StopLimitOnBidOrOffer",
        ("FIX.5.0", "43", "N"): "N .. OriginalTransmission",
        ("FIX.5.0", "43", "Y"): "Y .. PossibleDuplicate",
        ("FIX.5.0", "54", "1"): "1 .. Buy",
        ("FIX.5.0", "54", "2"): "2 .. Sell",
        ("FIX.5.0", "54", "3"): "3 .. BuyMinus",
        ("FIX.5.0", "54", "4"): "4 .. SellPlus",
        ("FIX.5.0", "54", "5"): "5 .. SellShort",
        ("FIX.5.0", "54", "6"): "6 .. SellShortExempt",
        ("FIX.5.0", "54", "7"): "7 .. Undisclosed",
        ("FIX.5.0", "54", "8"): "8 .. Cross",
        ("FIX.5.0", "54", "9"): "9 .. CrossShort",
        ("FIX.5.0", "54", "A"): "A .. CrossShortExempt",
        ("FIX.5.0", "54", "B"): "B .. AsDefined",
        ("FIX.5.0", "54", "C"): "C .. Opposite",
        ("FIX.5.0", "54", "D"): "D .. Subscribe",
        ("FIX.5.0", "54", "E"): "E .. Redeem",
        ("FIX.5.0", "54", "F"): "F .. Lend",
        ("FIX.5.0", "54", "G"): "G .. Borrow",
        ("FIX.5.0", "54", "H"): "H .. SellUndisclosed",
        ("FIX.5.0", "59", "0"): "0 .. Day",
        ("FIX.5.0", "59", "1"): "1 .. GoodTillCancel",
        ("FIX.5.0", "59", "2"): "2 .. AtTheOpening",
        ("FIX.5.0", "59", "3"): "3 .. ImmediateOrCancel",
        ("FIX.5.0", "59", "4"): "4 .. FillOrKill",
        ("FIX.5.0", "59", "5"): "5 .. GoodTillCrossing",
        ("FIX.5.0", "59", "6"): "6 .. GoodTillDate",
        ("FIX.5.0", "59", "7"): "7 .. AtTheClose",
        ("FIX.5.0", "59", "8"): "8 .. GoodThroughCrossing",
        ("FIX.5.0", "59", "9"): "9 .. AtCrossing",
        ("FIX.5.0", "59", "A"): "A .. GoodForTime",
        ("FIX.5.0", "59", "B"): "B .. GoodForAuction",
        ("FIX.5.0", "61", "0"): "0 .. Normal",
        ("FIX.5.0", "61", "1"): "1 .. Flash",
        ("FIX.5.0", "61", "2"): "2 .. Background",
        ("FIX.5.0", "63", "0"): "0 .. Regular",
        ("FIX.5.0", "63", "1"): "1 .. Cash",
        ("FIX.5.0", "63", "2"): "2 .. NextDay",
        ("FIX.5.0", "63", "3"): "3 .. TPlus2",
        ("FIX.5.0", "63", "4"): "4 .. TPlus3",
        ("FIX.5.0", "63", "5"): "5 .. TPlus4",
        ("FIX.5.0", "63", "6"): "6 .. Future",
        ("FIX.5.0", "63", "7"): "7 .. WhenAndIfIssued",
        ("FIX.5.0", "63", "8"): "8 .. SellersOption",
        ("FIX.5.0", "63", "9"): "9 .. TPlus5",
        ("FIX.5.0", "63", "B"): "B .. BrokenDate",
        ("FIX.5.0", "63", "C"): "C .. FXSpotNextSettlement",
        ("FIX.5.0", "65", "CD"): "CD .. EUCPWithLumpSumInterest",
        ("FIX.5.0", "65", "WI"): "WI .. WhenIssued",
        ("FIX.5.0", "71", "0"): "0 .. New",
        ("FIX.5.0", "71", "1"): "1 .. Replace",
        ("FIX.5.0", "71", "2"): "2 .. Cancel",
        ("FIX.5.0", "71", "6"): "6 .. Reversal",
        ("FIX.5.0", "77", "C"): "C .. Close",
        ("FIX.5.0", "77", "F"): "F .. FIFO",
        ("FIX.5.0", "77", "O"): "O .. Open",
        ("FIX.5.0", "77", "R"): "R .. Rolled",
        ("FIX.5.0", "77", "N"): "N .. CloseButNotifyOnOpen",
        ("FIX.5.0", "77", "D"): "D .. Default",
        ("FIX.5.0", "81", "0"): "0 .. Regular",
        ("FIX.5.0", "81", "1"): "1 .. SoftDollar",
        ("FIX.5.0", "81", "2"): "2 .. StepIn",
        ("FIX.5.0", "81", "3"): "3 .. StepOut",
        ("FIX.5.0", "81", "4"): "4 .. SoftDollarStepIn",
        ("FIX.5.0", "81", "5"): "5 .. SoftDollarStepOut",
        ("FIX.5.0", "81", "6"): "6 .. PlanSponsor",
        ("FIX.5.0", "87", "0"): "0 .. Accepted",
        ("FIX.5.0", "87", "1"): "1 .. BlockLevelReject",
        ("FIX.5.0", "87", "2"): "2 .. AccountLevelReject",
        ("FIX.5.0", "87", "3"): "3 .. Received",
        ("FIX.5.0", "87", "4"): "4 .. Incomplete",
        ("FIX.5.0", "87", "5"): "5 .. RejectedByIntermediary",
        ("FIX.5.0", "87", "6"): "6 .. AllocationPending",
        ("FIX.5.0", "87", "7"): "7 .. Reversed",
        ("FIX.5.0", "87", "8"): "8 .. CancelledByIntermediary",
        ("FIX.5.0", "87", "9"): "9 .. Claimed",
        ("FIX.5.0", "87", "10"): "10 .. Refused",
        ("FIX.5.0", "87", "11"): "11 .. PendingGiveUpApproval",
        ("FIX.5.0", "87", "12"): "12 .. Cancelled",
        ("FIX.5.0", "87", "13"): "13 .. PendingTakeUpApproval",
        ("FIX.5.0", "87", "14"): "14 .. ReversalPending",
        ("FIX.5.0", "88", "0"): "0 .. UnknownAccount",
        ("FIX.5.0", "88", "1"): "1 .. IncorrectQuantity",
        ("FIX.5.0", "88", "2"): "2 .. IncorrectAveragegPrice",
        ("FIX.5.0", "88", "3"): "3 .. UnknownExecutingBrokerMnemonic",
        ("FIX.5.0", "88", "4"): "4 .. CommissionDifference",
        ("FIX.5.0", "88", "5"): "5 .. UnknownOrderID",
        ("FIX.5.0", "88", "6"): "6 .. UnknownListID",
        ("FIX.5.0", "88", "7"): "7 .. OtherSeeText",
        ("FIX.5.0", "88", "8"): "8 .. IncorrectAllocatedQuantity",
        ("FIX.5.0", "88", "9"): "9 .. CalculationDifference",
        ("FIX.5.0", "88", "10"): "10 .. UnknownOrStaleExecID",
        ("FIX.5.0", "88", "11"): "11 .. MismatchedData",
        ("FIX.5.0", "88", "12"): "12 .. UnknownClOrdID",
        ("FIX.5.0", "88", "13"): "13 .. WarehouseRequestRejected",
        ("FIX.5.0", "88", "14"): "14 .. DuplicateOrMissingIndividualAllocId",
        ("FIX.5.0", "88", "15"): "15 .. TradeNotRecognized",
        ("FIX.5.0", "88", "16"): "16 .. DuplicateTrade",
        ("FIX.5.0", "88", "17"): "17 .. IncorrectOrMissingInstrument",
        ("FIX.5.0", "88", "18"): "18 .. IncorrectOrMissingSettlDate",
        ("FIX.5.0", "88", "19"): "19 .. IncorrectOrMissingFundIDOrFundName",
        ("FIX.5.0", "88", "20"): "20 .. IncorrectOrMissingSettlInstructions",
        ("FIX.5.0", "88", "21"): "21 .. IncorrectOrMissingFees",
        ("FIX.5.0", "88", "22"): "22 .. IncorrectOrMissingTax",
        ("FIX.5.0", "88", "23"): "23 .. UnknownOrMissingParty",
        ("FIX.5.0", "88", "24"): "24 .. IncorrectOrMissingSide",
        ("FIX.5.0", "88", "25"): "25 .. IncorrectOrMissingNetMoney",
        ("FIX.5.0", "88", "26"): "26 .. IncorrectOrMissingTradeDate",
        ("FIX.5.0", "88", "27"): "27 .. IncorrectOrMissingSettlCcyInstructions",
        ("FIX.5.0", "88", "28"): "28 .. IncorrectOrMissingProcessCode",
        ("FIX.5.0", "88", "99"): "99 .. Other",
        ("FIX.5.0", "94", "0"): "0 .. New",
        ("FIX.5.0", "94", "1"): "1 .. Reply",
        ("FIX.5.0", "94", "2"): "2 .. AdminReply",
        ("FIX.5.0", "97", "N"): "N .. OriginalTransmission",
        ("FIX.5.0", "97", "Y"): "Y .. PossibleResend",
        ("FIX.5.0", "98", "0"): "0 .. None",
        ("FIX.5.0", "98", "1"): "1 .. PKCS",
        ("FIX.5.0", "98", "2"): "2 .. DES",
        ("FIX.5.0", "98", "3"): "3 .. PKCSDES",
        ("FIX.5.0", "98", "4"): "4 .. PGPDES",
        ("FIX.5.0", "98", "5"): "5 .. PGPDESMD5",
        ("FIX.5.0", "98", "6"): "6 .. PEM",
        ("FIX.5.0", "102", "0"): "0 .. TooLateToCancel",
        ("FIX.5.0", "102", "1"): "1 .. UnknownOrder",
        ("FIX.5.0", "102", "2"): "2 .. BrokerCredit",
        ("FIX.5.0", "102", "3"): "3 .. OrderAlreadyInPendingStatus",
        ("FIX.5.0", "102", "4"): "4 .. UnableToProcessOrderMassCancelRequest",
        ("FIX.5.0", "102", "5"): "5 .. OrigOrdModTime",
        ("FIX.5.0", "102", "6"): "6 .. DuplicateClOrdID",
        ("FIX.5.0", "102", "7"): "7 .. PriceExceedsCurrentPrice",
        ("FIX.5.0", "102", "8"): "8 .. PriceExceedsCurrentPriceBand",
        ("FIX.5.0", "102", "18"): "18 .. InvalidPriceIncrement",
        ("FIX.5.0", "102", "99"): "99 .. Other",
        ("FIX.5.0", "103", "0"): "0 .. BrokerCredit",
        ("FIX.5.0", "103", "1"): "1 .. UnknownSymbol",
        ("FIX.5.0", "103", "2"): "2 .. ExchangeClosed",
        ("FIX.5.0", "103", "3"): "3 .. OrderExceedsLimit",
        ("FIX.5.0", "103", "4"): "4 .. TooLateToEnter",
        ("FIX.5.0", "103", "5"): "5 .. UnknownOrder",
        ("FIX.5.0", "103", "6"): "6 .. DuplicateOrder",
        ("FIX.5.0", "103", "7"): "7 .. DuplicateOfAVerballyCommunicatedOrder",
        ("FIX.5.0", "103", "8"): "8 .. StaleOrder",
        ("FIX.5.0", "103", "9"): "9 .. TradeAlongRequired",
        ("FIX.5.0", "103", "10"): "10 .. InvalidInvestorID",
        ("FIX.5.0", "103", "11"): "11 .. UnsupportedOrderCharacteristic",
        ("FIX.5.0", "103", "12"): "12 .. SurveillanceOption",
        ("FIX.5.0", "103", "13"): "13 .. IncorrectQuantity",
        ("FIX.5.0", "103", "14"): "14 .. IncorrectAllocatedQuantity",
        ("FIX.5.0", "103", "15"): "15 .. UnknownAccount",
        ("FIX.5.0", "103", "16"): "16 .. PriceExceedsCurrentPriceBand",
        ("FIX.5.0", "103", "18"): "18 .. InvalidPriceIncrement",
        ("FIX.5.0", "103", "19"): "19 .. ReferencePriceNotAvailable",
        ("FIX.5.0", "103", "20"): "20 .. NotionalValueExceedsThreshold",
        ("FIX.5.0", "103", "21"): "21 .. AlgorithRiskThresholdBreached",
        ("FIX.5.0", "103", "22"): "22 .. ShortSellNotPermitted",
        ("FIX.5.0", "103", "23"): "23 .. ShortSellSecurityPreBorrowRestriction",
        ("FIX.5.0", "103", "24"): "24 .. ShortSellAccountPreBorrowRestriction",
        ("FIX.5.0", "103", "25"): "25 .. InsufficientCreditLimit",
        ("FIX.5.0", "103", "26"): "26 .. ExceededClipSizeLimit",
        ("FIX.5.0", "103", "27"): "27 .. ExceededMaxNotionalOrderAmt",
        ("FIX.5.0", "103", "28"): "28 .. ExceededDV01PV01Limit",
        ("FIX.5.0", "103", "29"): "29 .. ExceededCS01Limit",
        ("FIX.5.0", "103", "99"): "99 .. Other",
        ("FIX.5.0", "104", "A"): "A .. AllOrNone",
        ("FIX.5.0", "104", "B"): "B .. MarketOnClose",
        ("FIX.5.0", "104", "C"): "C .. AtTheClose",
        ("FIX.5.0", "104", "D"): "D .. VWAP",
        ("FIX.5.0", "104", "E"): "E .. Axe",
        ("FIX.5.0", "104", "F"): "F .. AxeOnBid",
        ("FIX.5.0", "104", "G"): "G .. AxeOnOffer",
        ("FIX.5.0", "104", "H"): "H .. ClientNaturalWorking",
        ("FIX.5.0", "104", "I"): "I .. InTouchWith",
        ("FIX.5.0", "104", "J"): "J .. PositionWanted",
        ("FIX.5.0", "104", "K"): "K .. MarketMaking",
        ("FIX.5.0", "104", "L"): "L .. Limit",
        ("FIX.5.0", "104", "M"): "M .. MoreBehind",
        ("FIX.5.0", "104", "N"): "N .. ClientNaturalBlock",
        ("FIX.5.0", "104", "O"): "O .. AtTheOpen",
        ("FIX.5.0", "104", "P"): "P .. TakingAPosition",
        ("FIX.5.0", "104", "Q"): "Q .. AtTheMarket",
        ("FIX.5.0", "104", "R"): "R .. ReadyToTrade",
        ("FIX.5.0", "104", "S"): "S .. PortfolioShown",
        ("FIX.5.0", "104", "T"): "T .. ThroughTheDay",
        ("FIX.5.0", "104", "U"): "U .. Unwind",
        ("FIX.5.0", "104", "V"): "V .. Versus",
        ("FIX.5.0", "104", "W"): "W .. Indication",
        ("FIX.5.0", "104", "X"): "X .. CrossingOpportunity",
        ("FIX.5.0", "104", "Y"): "Y .. AtTheMidpoint",
        ("FIX.5.0", "104", "Z"): "Z .. PreOpen",
        ("FIX.5.0", "104", "1"): "1 .. QuantityNegotiable",
        ("FIX.5.0", "104", "2"): "2 .. AllowLateBids",
        ("FIX.5.0", "104", "3"): "3 .. ImmediateOrCounter",
        ("FIX.5.0", "104", "4"): "4 .. AutoTrade",
        ("FIX.5.0", "104", "a"): "a .. AutomaticSpot",
        ("FIX.5.0", "104", "b"): "b .. PlatformCalculatedSpot",
        ("FIX.5.0", "104", "c"): "c .. OutsideSpread",
        ("FIX.5.0", "104", "d"): "d .. DeferredSpot",
        ("FIX.5.0", "104", "n"): "n .. NegotiatedSpot",
        ("FIX.5.0", "113", "N"): "N .. SenderReports",
        ("FIX.5.0", "113", "Y"): "Y .. ReceiverReports",
        ("FIX.5.0", "114", "N"): "N .. No",
        ("FIX.5.0", "114", "Y"): "Y .. Yes",
        ("FIX.5.0", "121", "N"): "N .. DoNotExecuteForexAfterSecurityTrade",
        ("FIX.5.0", "121", "Y"): "Y .. ExecuteForexAfterSecurityTrade",
        ("FIX.5.0", "123", "N"): "N .. SequenceReset",
        ("FIX.5.0", "123", "Y"): "Y .. GapFillMessage",
        ("FIX.5.0", "127", "A"): "A .. UnknownSymbol",
        ("FIX.5.0", "127", "B"): "B .. WrongSide",
        ("FIX.5.0", "127", "C"): "C .. QuantityExceedsOrder",
        ("FIX.5.0", "127", "D"): "D .. NoMatchingOrder",
        ("FIX.5.0", "127", "E"): "E .. PriceExceedsLimit",
        ("FIX.5.0", "127", "F"): "F .. CalculationDifference",
        ("FIX.5.0", "127", "G"): "G .. NoMatchingExecutionReport",
        ("FIX.5.0", "127", "Z"): "Z .. Other",
        ("FIX.5.0", "130", "N"): "N .. NotNatural",
        ("FIX.5.0", "130", "Y"): "Y .. Natural",
        ("FIX.5.0", "139", "1"): "1 .. Regulatory",
        ("FIX.5.0", "139", "2"): "2 .. Tax",
        ("FIX.5.0", "139", "4"): "4 .. ExchangeFees",
        ("FIX.5.0", "139", "5"): "5 .. Stamp",
        ("FIX.5.0", "139", "6"): "6 .. Levy",
        ("FIX.5.0", "139", "7"): "7 .. Other",
        ("FIX.5.0", "139", "8"): "8 .. Markup",
        ("FIX.5.0", "139", "9"): "9 .. ConsumptionTax",
        ("FIX.5.0", "139", "10"): "10 .. PerTransaction",
        ("FIX.5.0", "139", "11"): "11 .. Conversion",
        ("FIX.5.0", "139", "12"): "12 .. Agent",
        ("FIX.5.0", "139", "13"): "13 .. TransferFee",
        ("FIX.5.0", "139", "14"): "14 .. SecurityLending",
        ("FIX.5.0", "139", "15"): "15 .. TradeReporting",
        ("FIX.5.0", "139", "16"): "16 .. TaxOnPrincipalAmount",
        ("FIX.5.0", "139", "17"): "17 .. TaxOnAccruedInterestAmount",
        ("FIX.5.0", "139", "18"): "18 .. NewIssuanceFee",
        ("FIX.5.0", "139", "19"): "19 .. ServiceFee",
        ("FIX.5.0", "139", "20"): "20 .. OddLotFee",
        ("FIX.5.0", "139", "21"): "21 .. AuctionFee",
        ("FIX.5.0", "139", "22"): "22 .. ValueAddedTax",
        ("FIX.5.0", "139", "23"): "23 .. SalesTax",
        ("FIX.5.0", "139", "24"): "24 .. ExecutionFee",
        ("FIX.5.0", "139", "25"): "25 .. OrderEntryFee",
        ("FIX.5.0", "139", "26"): "26 .. OrderModificationFee",
        ("FIX.5.0", "139", "27"): "27 .. OrdersCancellationFee",
        ("FIX.5.0", "139", "28"): "28 .. MarketDataAccessFee",
        ("FIX.5.0", "139", "29"): "29 .. MarketDataTerminalFee",
        ("FIX.5.0", "139", "30"): "30 .. MarketDataVolumeFee",
        ("FIX.5.0", "139", "31"): "31 .. ClearingFee",
        ("FIX.5.0", "139", "32"): "32 .. SettlementFee",
        ("FIX.5.0", "139", "33"): "33 .. Rebates",
        ("FIX.5.0", "139", "34"): "34 .. Discounts",
        ("FIX.5.0", "139", "35"): "35 .. Payments",
        ("FIX.5.0", "139", "36"): "36 .. NonMonetaryPayments",
        ("FIX.5.0", "141", "N"): "N .. No",
        ("FIX.5.0", "141", "Y"): "Y .. Yes",
        ("FIX.5.0", "150", "0"): "0 .. New",
        ("FIX.5.0", "150", "3"): "3 .. DoneForDay",
        ("FIX.5.0", "150", "4"): "4 .. Canceled",
        ("FIX.5.0", "150", "5"): "5 .. Replaced",
        ("FIX.5.0", "150", "6"): "6 .. PendingCancel",
        ("FIX.5.0", "150", "7"): "7 .. Stopped",
        ("FIX.5.0", "150", "8"): "8 .. Rejected",
        ("FIX.5.0", "150", "9"): "9 .. Suspended",
        ("FIX.5.0", "150", "A"): "A .. PendingNew",
        ("FIX.5.0", "150", "B"): "B .. Calculated",
        ("FIX.5.0", "150", "C"): "C .. Expired",
        ("FIX.5.0", "150", "D"): "D .. Restated",
        ("FIX.5.0", "150", "E"): "E .. PendingReplace",
        ("FIX.5.0", "150", "F"): "F .. Trade",
        ("FIX.5.0", "150", "G"): "G .. TradeCorrect",
        ("FIX.5.0", "150", "H"): "H .. TradeCancel",
        ("FIX.5.0", "150", "I"): "I .. OrderStatus",
        ("FIX.5.0", "150", "J"): "J .. TradeInAClearingHold",
        ("FIX.5.0", "150", "K"): "K .. TradeHasBeenReleasedToClearing",
        ("FIX.5.0", "150", "L"): "L .. TriggeredOrActivatedBySystem",
        ("FIX.5.0", "150", "M"): "M .. Locked",
        ("FIX.5.0", "150", "N"): "N .. Released",
        ("FIX.5.0", "156", "M"): "M .. Multiply",
        ("FIX.5.0", "156", "D"): "D .. Divide",
        ("FIX.5.0", "160", "0"): "0 .. Default",
        ("FIX.5.0", "160", "1"): "1 .. StandingInstructionsProvided",
        ("FIX.5.0", "160", "2"): "2 .. SpecificAllocationAccountOverriding",
        ("FIX.5.0", "160", "3"): "3 .. SpecificAllocationAccountStanding",
        ("FIX.5.0", "160", "4"): "4 .. SpecificOrderForASingleAccount",
        ("FIX.5.0", "160", "5"): "5 .. RequestReject",
        ("FIX.5.0", "163", "N"): "N .. New",
        ("FIX.5.0", "163", "C"): "C .. Cancel",
        ("FIX.5.0", "163", "R"): "R .. Replace",
        ("FIX.5.0", "163", "T"): "T .. Restate",
        ("FIX.5.0", "165", "1"): "1 .. BrokerCredit",
        ("FIX.5.0", "165", "2"): "2 .. Institution",
        ("FIX.5.0", "165", "3"): "3 .. Investor",
        ("FIX.5.0", "167", "EUSUPRA"): "EUSUPRA .. EuroSupranationalCoupons",
        ("FIX.5.0", "167", "FAC"): "FAC .. FederalAgencyCoupon",
        ("FIX.5.0", "167", "FADN"): "FADN .. FederalAgencyDiscountNote",
        ("FIX.5.0", "167", "PEF"): "PEF .. PrivateExportFunding",
        ("FIX.5.0", "167", "SUPRA"): "SUPRA .. USDSupranationalCoupons",
        ("FIX.5.0", "167", "CORP"): "CORP .. CorporateBond",
        ("FIX.5.0", "167", "CPP"): "CPP .. CorporatePrivatePlacement",
        ("FIX.5.0", "167", "CB"): "CB .. ConvertibleBond",
        ("FIX.5.0", "167", "DUAL"): "DUAL .. DualCurrency",
        ("FIX.5.0", "167", "EUCORP"): "EUCORP .. EuroCorporateBond",
        ("FIX.5.0", "167", "EUFRN"): "EUFRN .. EuroCorporateFloatingRateNotes",
        ("FIX.5.0", "167", "FRN"): "FRN .. USCorporateFloatingRateNotes",
        ("FIX.5.0", "167", "XLINKD"): "XLINKD .. IndexedLinked",
        ("FIX.5.0", "167", "STRUCT"): "STRUCT .. StructuredNotes",
        ("FIX.5.0", "167", "YANK"): "YANK .. YankeeCorporateBond",
        ("FIX.5.0", "167", "FXNDF"): "FXNDF .. NonDeliverableForward",
        ("FIX.5.0", "167", "FXSPOT"): "FXSPOT .. FXSpot",
        ("FIX.5.0", "167", "FXFWD"): "FXFWD .. FXForward",
        ("FIX.5.0", "167", "FXSWAP"): "FXSWAP .. FXSwap",
        ("FIX.5.0", "167", "FXNDS"): "FXNDS .. NonDeliverableSwap",
        ("FIX.5.0", "167", "CAP"): "CAP .. Cap",
        ("FIX.5.0", "167", "CDS"): "CDS .. CreditDefaultSwap",
        ("FIX.5.0", "167", "CLLR"): "CLLR .. Collar",
        ("FIX.5.0", "167", "CMDTYSWAP"): "CMDTYSWAP .. CommoditySwap",
        ("FIX.5.0", "167", "EXOTIC"): "EXOTIC .. Exotic",
        ("FIX.5.0", "167", "OOC"): "OOC .. OptionsOnCombo",
        ("FIX.5.0", "167", "FLR"): "FLR .. Floor",
        ("FIX.5.0", "167", "FRA"): "FRA .. FRA",
        ("FIX.5.0", "167", "FUT"): "FUT .. Future",
        ("FIX.5.0", "167", "FWD"): "FWD .. DerivativeForward",
        ("FIX.5.0", "167", "IRS"): "IRS .. InterestRateSwap",
        ("FIX.5.0", "167", "TRS"): "TRS .. TotalReturnSwap",
        ("FIX.5.0", "167", "LOANLEASE"): "LOANLEASE .. LoanLease",
        ("FIX.5.0", "167", "OOF"): "OOF .. OptionsOnFutures",
        ("FIX.5.0", "167", "OOP"): "OOP .. OptionsOnPhysical",
        ("FIX.5.0", "167", "OPT"): "OPT .. Option",
        ("FIX.5.0", "167", "SPOTFWD"): "SPOTFWD .. SpotForward",
        ("FIX.5.0", "167", "SWAPTION"): "SWAPTION .. SwapOption",
        ("FIX.5.0", "167", "XMISSION"): "XMISSION .. Transmission",
        ("FIX.5.0", "167", "INDEX"): "INDEX .. Index",
        ("FIX.5.0", "167", "BDBSKT"): "BDBSKT .. BondBasket",
        ("FIX.5.0", "167", "CFD"): "CFD .. ContractForDifference",
        ("FIX.5.0", "167", "CRLTNSWAP"): "CRLTNSWAP .. CorrelationSwap",
        ("FIX.5.0", "167", "DVDNDSWAP"): "DVDNDSWAP .. DiviendSwap",
        ("FIX.5.0", "167", "EQBSKT"): "EQBSKT .. EquityBasket",
        ("FIX.5.0", "167", "EQFWD"): "EQFWD .. EquityForward",
        ("FIX.5.0", "167", "RTRNSWAP"): "RTRNSWAP .. ReturnSwap",
        ("FIX.5.0", "167", "VARSWAP"): "VARSWAP .. VarianceSwap",
        ("FIX.5.0", "167", "PRTFLIOSWAP"): "PRTFLIOSWAP .. PortfolioSwaps",
        ("FIX.5.0", "167", "FUTSWAP"): "FUTSWAP .. FuturesOnASwap",
        ("FIX.5.0", "167", "FWDSWAP "): "FWDSWAP  .. ForwardsOnASwap",
        ("FIX.5.0", "167", "FWDFRTAGMT "): "FWDFRTAGMT  .. ForwardFreightAgreement",
        ("FIX.5.0", "167", "SPREADBET"): "SPREADBET .. SpreadBetting",
        ("FIX.5.0", "167", "ETC"): "ETC .. ExchangeTradedCommodity",
        ("FIX.5.0", "167", "CS"): "CS .. CommonStock",
        ("FIX.5.0", "167", "PS"): "PS .. PreferredStock",
        ("FIX.5.0", "167", "DR"): "DR .. DepositoryReceipts",
        ("FIX.5.0", "167", "REPO"): "REPO .. Repurchase",
        ("FIX.5.0", "167", "FORWARD"): "FORWARD .. Forward",
        ("FIX.5.0", "167", "BUYSELL"): "BUYSELL .. BuySellback",
        ("FIX.5.0", "167", "SECLOAN"): "SECLOAN .. SecuritiesLoan",
        ("FIX.5.0", "167", "SECPLEDGE"): "SECPLEDGE .. SecuritiesPledge",
        ("FIX.5.0", "167", "DVPLDG"): "DVPLDG .. DeliveryVersusPledge",
        ("FIX.5.0", "167", "COLLBSKT"): "COLLBSKT .. CollateralBasket",
        ("FIX.5.0", "167", "SFP"): "SFP .. StructuredFinanceProduct",
        ("FIX.5.0", "167", "BRADY"): "BRADY .. BradyBond",
        ("FIX.5.0", "167", "CAN"): "CAN .. CanadianTreasuryNotes",
        ("FIX.5.0", "167", "CTB"): "CTB .. CanadianTreasuryBills",
        ("FIX.5.0", "167", "EUSOV"): "EUSOV .. EuroSovereigns",
        ("FIX.5.0", "167", "PROV"): "PROV .. CanadianProvincialBonds",
        ("FIX.5.0", "167", "TB"): "TB .. TreasuryBill",
        ("FIX.5.0", "167", "TBOND"): "TBOND .. USTreasuryBond",
        ("FIX.5.0", "167", "TINT"): "TINT .. InterestStripFromAnyBondOrNote",
        ("FIX.5.0", "167", "TBILL"): "TBILL .. USTreasuryBill",
        ("FIX.5.0", "167", "TIPS"): "TIPS .. TreasuryInflationProtectedSecurities",
        ("FIX.5.0", "167", "TCAL"): "TCAL .. PrincipalStripOfACallableBondOrNote",
        ("FIX.5.0", "167", "TPRN"): "TPRN .. PrincipalStripFromANonCallableBondOrNote",
        ("FIX.5.0", "167", "TNOTE"): "TNOTE .. USTreasuryNote",
        ("FIX.5.0", "167", "TERM"): "TERM .. TermLoan",
        ("FIX.5.0", "167", "RVLV"): "RVLV .. RevolverLoan",
        ("FIX.5.0", "167", "RVLVTRM"): "RVLVTRM .. Revolver",
        ("FIX.5.0", "167", "BRIDGE"): "BRIDGE .. BridgeLoan",
        ("FIX.5.0", "167", "LOFC"): "LOFC .. LetterOfCredit",
        ("FIX.5.0", "167", "SWING"): "SWING .. SwingLineFacility",
        ("FIX.5.0", "167", "DINP"): "DINP .. DebtorInPossession",
        ("FIX.5.0", "167", "DEFLTED"): "DEFLTED .. Defaulted",
        ("FIX.5.0", "167", "WITHDRN"): "WITHDRN .. Withdrawn",
        ("FIX.5.0", "167", "REPLACD"): "REPLACD .. Replaced",
        ("FIX.5.0", "167", "MATURED"): "MATURED .. Matured",
        ("FIX.5.0", "167", "AMENDED"): "AMENDED .. Amended",
        ("FIX.5.0", "167", "RETIRED"): "RETIRED .. Retired",
        ("FIX.5.0", "167", "BA"): "BA .. BankersAcceptance",
        ("FIX.5.0", "167", "BDN"): "BDN .. BankDepositoryNote",
        ("FIX.5.0", "167", "BN"): "BN .. BankNotes",
        ("FIX.5.0", "167", "BOX"): "BOX .. BillOfExchanges",
        ("FIX.5.0", "167", "CAMM"): "CAMM .. CanadianMoneyMarkets",
        ("FIX.5.0", "167", "CD"): "CD .. CertificateOfDeposit",
        ("FIX.5.0", "167", "CL"): "CL .. CallLoans",
        ("FIX.5.0", "167", "CP"): "CP .. CommercialPaper",
        ("FIX.5.0", "167", "DN"): "DN .. DepositNotes",
        ("FIX.5.0", "167", "EUCD"): "EUCD .. EuroCertificateOfDeposit",
        ("FIX.5.0", "167", "EUCP"): "EUCP .. EuroCommercialPaper",
        ("FIX.5.0", "167", "LQN"): "LQN .. LiquidityNote",
        ("FIX.5.0", "167", "MTN"): "MTN .. MediumTermNotes",
        ("FIX.5.0", "167", "ONITE"): "ONITE .. Overnight",
        ("FIX.5.0", "167", "PN"): "PN .. PromissoryNote",
        ("FIX.5.0", "167", "STN"): "STN .. ShortTermLoanNote",
        ("FIX.5.0", "167", "PZFJ"): "PZFJ .. PlazosFijos",
        ("FIX.5.0", "167", "SLQN"): "SLQN .. SecuredLiquidityNote",
        ("FIX.5.0", "167", "TD"): "TD .. TimeDeposit",
        ("FIX.5.0", "167", "TLQN"): "TLQN .. TermLiquidityNote",
        ("FIX.5.0", "167", "XCN"): "XCN .. ExtendedCommNote",
        ("FIX.5.0", "167", "YCD"): "YCD .. YankeeCertificateOfDeposit",
        ("FIX.5.0", "167", "ABS"): "ABS .. AssetBackedSecurities",
        ("FIX.5.0", "167", "CMB"): "CMB .. CanadianMortgageBonds",
        ("FIX.5.0", "167", "CMBS"): "CMBS .. Corp",
        ("FIX.5.0", "167", "CMO"): "CMO .. CollateralizedMortgageObligation",
        ("FIX.5.0", "167", "IET"): "IET .. IOETTEMortgage",
        ("FIX.5.0", "167", "MBS"): "MBS .. MortgageBackedSecurities",
        ("FIX.5.0", "167", "MIO"): "MIO .. MortgageInterestOnly",
        ("FIX.5.0", "167", "MPO"): "MPO .. MortgagePrincipalOnly",
        ("FIX.5.0", "167", "MPP"): "MPP .. MortgagePrivatePlacement",
        ("FIX.5.0", "167", "MPT"): "MPT .. MiscellaneousPassThrough",
        ("FIX.5.0", "167", "PFAND"): "PFAND .. Pfandbriefe",
        ("FIX.5.0", "167", "TBA"): "TBA .. ToBeAnnounced",
        ("FIX.5.0", "167", "AN"): "AN .. OtherAnticipationNotes",
        ("FIX.5.0", "167", "COFO"): "COFO .. CertificateOfObligation",
        ("FIX.5.0", "167", "COFP"): "COFP .. CertificateOfParticipation",
        ("FIX.5.0", "167", "GO"): "GO .. GeneralObligationBonds",
        ("FIX.5.0", "167", "MT"): "MT .. MandatoryTender",
        ("FIX.5.0", "167", "RAN"): "RAN .. RevenueAnticipationNote",
        ("FIX.5.0", "167", "REV"): "REV .. RevenueBonds",
        ("FIX.5.0", "167", "SPCLA"): "SPCLA .. SpecialAssessment",
        ("FIX.5.0", "167", "SPCLO"): "SPCLO .. SpecialObligation",
        ("FIX.5.0", "167", "SPCLT"): "SPCLT .. SpecialTax",
        ("FIX.5.0", "167", "TAN"): "TAN .. TaxAnticipationNote",
        ("FIX.5.0", "167", "TAXA"): "TAXA .. TaxAllocation",
        ("FIX.5.0", "167", "TECP"): "TECP .. TaxExemptCommercialPaper",
        ("FIX.5.0", "167", "TMCP"): "TMCP .. TaxableMunicipalCP",
        ("FIX.5.0", "167", "TRAN"): "TRAN .. TaxRevenueAnticipationNote",
        ("FIX.5.0", "167", "VRDN"): "VRDN .. VariableRateDemandNote",
        ("FIX.5.0", "167", "WAR"): "WAR .. Warrant",
        ("FIX.5.0", "167", "MF"): "MF .. MutualFund",
        ("FIX.5.0", "167", "MLEG"): "MLEG .. MultilegInstrument",
        ("FIX.5.0", "167", "NONE"): "NONE .. NoSecurityType",
        ("FIX.5.0", "167", "?"): "? .. Wildcard",
        ("FIX.5.0", "167", "CASH"): "CASH .. Cash",
        ("FIX.5.0", "167", "Other"): "Other .. Other",
        ("FIX.5.0", "167", "ETN"): "ETN .. ExchangeTradedNote",
        ("FIX.5.0", "167", "SECDERIV"): "SECDERIV .. SecuritizedDerivative",
        ("FIX.5.0", "169", "0"): "0 .. Other",
        ("FIX.5.0", "169", "1"): "1 .. DTCSID",
        ("FIX.5.0", "169", "2"): "2 .. ThomsonALERT",
        ("FIX.5.0", "169", "3"): "3 .. AGlobalCustodian",
        ("FIX.5.0", "169", "4"): "4 .. AccountNet",
        ("FIX.5.0", "172", "0"): "0 .. Versus",
        ("FIX.5.0", "172", "1"): "1 .. Free",
        ("FIX.5.0", "172", "2"): "2 .. TriParty",
        ("FIX.5.0", "172", "3"): "3 .. HoldInCustody",
        ("FIX.5.0", "197", "0"): "0 .. FXNetting",
        ("FIX.5.0", "197", "1"): "1 .. FXSwap",
        ("FIX.5.0", "201", "0"): "0 .. Put",
        ("FIX.5.0", "201", "1"): "1 .. Call",
        ("FIX.5.0", "201", "2"): "2 .. Other",
        ("FIX.5.0", "201", "3"): "3 .. Chooser",
        ("FIX.5.0", "203", "0"): "0 .. Covered",
        ("FIX.5.0", "203", "1"): "1 .. Uncovered",
        ("FIX.5.0", "208", "N"): "N .. DetailsShouldNotBeCommunicated",
        ("FIX.5.0", "208", "Y"): "Y .. DetailsShouldBeCommunicated",
        ("FIX.5.0", "209", "1"): "1 .. Match",
        ("FIX.5.0", "209", "2"): "2 .. Forward",
        ("FIX.5.0", "209", "3"): "3 .. ForwardAndMatch",
        ("FIX.5.0", "216", "1"): "1 .. TargetFirm",
        ("FIX.5.0", "216", "2"): "2 .. TargetList",
        ("FIX.5.0", "216", "3"): "3 .. BlockFirm",
        ("FIX.5.0", "216", "4"): "4 .. BlockList",
        ("FIX.5.0", "216", "5"): "5 .. TargetPerson",
        ("FIX.5.0", "216", "6"): "6 .. BlockPerson",
        ("FIX.5.0", "221", "EONIA"): "EONIA .. EONIA",
        ("FIX.5.0", "221", "EUREPO"): "EUREPO .. EUREPO",
        ("FIX.5.0", "221", "FutureSWAP"): "FutureSWAP .. FutureSWAP",
        ("FIX.5.0", "221", "LIBID"): "LIBID .. LIBID",
        ("FIX.5.0", "221", "LIBOR"): "LIBOR .. LIBOR",
        ("FIX.5.0", "221", "MuniAAA"): "MuniAAA .. MuniAAA",
        ("FIX.5.0", "221", "OTHER"): "OTHER .. OTHER",
        ("FIX.5.0", "221", "Pfandbriefe"): "Pfandbriefe .. Pfandbriefe",
        ("FIX.5.0", "221", "SONIA"): "SONIA .. SONIA",
        ("FIX.5.0", "221", "SWAP"): "SWAP .. SWAP",
        ("FIX.5.0", "221", "Treasury"): "Treasury .. Treasury",
        ("FIX.5.0", "221", "FEDEFF"): "FEDEFF .. FedFundRateEffective",
        ("FIX.5.0", "221", "FEDOPEN"): "FEDOPEN .. FedOpen",
        ("FIX.5.0", "221", "EURIBOR"): "EURIBOR .. EURIBOR",
        ("FIX.5.0", "233", "AMT"): "AMT .. AlternativeMinimumTax",
        ("FIX.5.0", "233", "AUTOREINV"): "AUTOREINV .. AutoReinvestment",
        ("FIX.5.0", "233", "BANKQUAL"): "BANKQUAL .. BankQualified",
        ("FIX.5.0", "233", "BGNCON"): "BGNCON .. BargainConditions",
        ("FIX.5.0", "233", "COUPON"): "COUPON .. CouponRange",
        ("FIX.5.0", "233", "CURRENCY"): "CURRENCY .. ISOCurrencyCode",
        ("FIX.5.0", "233", "CUSTOMDATE"): "CUSTOMDATE .. CustomStart",
        ("FIX.5.0", "233", "GEOG"): "GEOG .. Geographics",
        ("FIX.5.0", "233", "HAIRCUT"): "HAIRCUT .. ValuationDiscount",
        ("FIX.5.0", "233", "INSURED"): "INSURED .. Insured",
        ("FIX.5.0", "233", "ISSUE"): "ISSUE .. IssueDate",
        ("FIX.5.0", "233", "ISSUER"): "ISSUER .. Issuer",
        ("FIX.5.0", "233", "ISSUESIZE"): "ISSUESIZE .. IssueSizeRange",
        ("FIX.5.0", "233", "LOOKBACK"): "LOOKBACK .. LookbackDays",
        ("FIX.5.0", "233", "LOT"): "LOT .. ExplicitLotIdentifier",
        ("FIX.5.0", "233", "LOTVAR"): "LOTVAR .. LotVariance",
        ("FIX.5.0", "233", "MAT"): "MAT .. MaturityYearAndMonth",
        ("FIX.5.0", "233", "MATURITY"): "MATURITY .. MaturityRange",
        ("FIX.5.0", "233", "MAXSUBS"): "MAXSUBS .. MaximumSubstitutions",
        ("FIX.5.0", "233", "MINDNOM"): "MINDNOM .. MinimumDenomination",
        ("FIX.5.0", "233", "MININCR"): "MININCR .. MinimumIncrement",
        ("FIX.5.0", "233", "MINQTY"): "MINQTY .. MinimumQuantity",
        ("FIX.5.0", "233", "PAYFREQ"): "PAYFREQ .. PaymentFrequency",
        ("FIX.5.0", "233", "PIECES"): "PIECES .. NumberOfPieces",
        ("FIX.5.0", "233", "PMAX"): "PMAX .. PoolsMaximum",
        ("FIX.5.0", "233", "PPL"): "PPL .. PoolsPerLot",
        ("FIX.5.0", "233", "PPM"): "PPM .. PoolsPerMillion",
        ("FIX.5.0", "233", "PPT"): "PPT .. PoolsPerTrade",
        ("FIX.5.0", "233", "PRICE"): "PRICE .. PriceRange",
        ("FIX.5.0", "233", "PRICEFREQ"): "PRICEFREQ .. PricingFrequency",
        ("FIX.5.0", "233", "PROD"): "PROD .. ProductionYear",
        ("FIX.5.0", "233", "PROTECT"): "PROTECT .. CallProtection",
        ("FIX.5.0", "233", "PURPOSE"): "PURPOSE .. Purpose",
        ("FIX.5.0", "233", "PXSOURCE"): "PXSOURCE .. BenchmarkPriceSource",
        ("FIX.5.0", "233", "RATING"): "RATING .. RatingSourceAndRange",
        ("FIX.5.0", "233", "REDEMPTION"): "REDEMPTION .. TypeOfRedemption",
        ("FIX.5.0", "233", "RESTRICTED"): "RESTRICTED .. Restricted",
        ("FIX.5.0", "233", "SECTOR"): "SECTOR .. MarketSector",
        ("FIX.5.0", "233", "SECTYPE"): "SECTYPE .. SecurityTypeIncludedOrExcluded",
        ("FIX.5.0", "233", "STRUCT"): "STRUCT .. Structure",
        ("FIX.5.0", "233", "SUBSFREQ"): "SUBSFREQ .. SubstitutionsFrequency",
        ("FIX.5.0", "233", "SUBSLEFT"): "SUBSLEFT .. SubstitutionsLeft",
        ("FIX.5.0", "233", "TEXT"): "TEXT .. FreeformText",
        ("FIX.5.0", "233", "TRDVAR"): "TRDVAR .. TradeVariance",
        ("FIX.5.0", "233", "WAC"): "WAC .. WeightedAverageCoupon",
        ("FIX.5.0", "233", "WAL"): "WAL .. WeightedAverageLifeCoupon",
        ("FIX.5.0", "233", "WALA"): "WALA .. WeightedAverageLoanAge",
        ("FIX.5.0", "233", "WAM"): "WAM .. WeightedAverageMaturity",
        ("FIX.5.0", "233", "WHOLE"): "WHOLE .. WholePool",
        ("FIX.5.0", "233", "YIELD"): "YIELD .. YieldRange",
        ("FIX.5.0", "233", "ORIGAMT"): "ORIGAMT .. OriginalAmount",
        ("FIX.5.0", "233", "POOLEFFDT"): "POOLEFFDT .. PoolEffectiveDate",
        ("FIX.5.0", "233", "POOLINITFCTR"): "POOLINITFCTR .. PoolInitialFactor",
        ("FIX.5.0", "233", "TRANCHE"): "TRANCHE .. Tranche",
        ("FIX.5.0", "233", "SUBSTITUTION"): "SUBSTITUTION .. Substitution",
        ("FIX.5.0", "233", "MULTEXCHFLLBCK"): "MULTEXCHFLLBCK .. MULTEXCHFLLBCK",
        ("FIX.5.0", "233", "COMPSECFLLBCK"): "COMPSECFLLBCK .. COMPSECFLLBCK",
        ("FIX.5.0", "233", "LOCLJRSDCTN"): "LOCLJRSDCTN .. LOCLJRSDCTN",
        ("FIX.5.0", "233", "RELVJRSDCTN"): "RELVJRSDCTN .. RELVJRSDCTN",
        ("FIX.5.0", "233", "INCURRCVY"): "INCURRCVY .. IncurredRecovery",
        ("FIX.5.0", "233", "ADDTRM"): "ADDTRM .. AdditionalTerm",
        ("FIX.5.0", "233", "MODEQTYDLVY"): "MODEQTYDLVY .. ModifiedEquityDelivery",
        ("FIX.5.0", "233", "NOREFOBLIG"): "NOREFOBLIG .. NoReferenceOblication",
        ("FIX.5.0", "233", "UNKREFOBLIG"): "UNKREFOBLIG .. UnknownReferenceObligation",
        ("FIX.5.0", "233", "ALLGUARANTEES"): "ALLGUARANTEES .. AllGuarantees",
        ("FIX.5.0", "233", "REFPX"): "REFPX .. ReferencePrice",
        ("FIX.5.0", "233", "REFPOLICY"): "REFPOLICY .. ReferencePolicy",
        ("FIX.5.0", "233", "SECRDLIST"): "SECRDLIST .. SecuredList",
        ("FIX.5.0", "233", "AVFICO"): "AVFICO .. AverageFICOScore",
        ("FIX.5.0", "233", "AVSIZE"): "AVSIZE .. AverageLoanSize",
        ("FIX.5.0", "233", "MAXBAL"): "MAXBAL .. MaximumLoanBalance",
        ("FIX.5.0", "233", "POOL"): "POOL .. PoolIdentifier",
        ("FIX.5.0", "233", "ROLLTYPE"): "ROLLTYPE .. TypeOfRollTrade",
        ("FIX.5.0", "233", "REFTRADE"): "REFTRADE .. ReferenceToRollingOrClosingTrade",
        ("FIX.5.0", "233", "REFPRIN"): "REFPRIN .. PrincipalOfRollingOrClosingTrade",
        ("FIX.5.0", "233", "REFINT"): "REFINT .. InterestOfRollingOrClosingTrade",
        ("FIX.5.0", "233", "AVAILQTY"): "AVAILQTY .. AvailableOfferQuantityToBeShownToTheStreet",
        ("FIX.5.0", "233", "BROKERCREDIT"): "BROKERCREDIT .. BrokerCredit",
        ("FIX.5.0", "233", "INTERNALPX"): "INTERNALPX .. OfferPriceToBeShownToInternalBrokers",
        ("FIX.5.0", "233", "INTERNALQTY"): "INTERNALQTY .. OfferQuantityToBeShownToInternalBrokers",
        ("FIX.5.0", "233", "LEAVEQTY"): "LEAVEQTY .. TheMinimumResidualOfferQuantity",
        ("FIX.5.0", "233", "MAXORDQTY"): "MAXORDQTY .. MaximumOrderSize",
        ("FIX.5.0", "233", "ORDRINCR"): "ORDRINCR .. OrderQuantityIncrement",
        ("FIX.5.0", "233", "PRIMARY"): "PRIMARY .. PrimaryOrSecondaryMarketIndicator",
        ("FIX.5.0", "233", "SALESCREDITOVR"): "SALESCREDITOVR .. BrokerSalesCreditOverride",
        ("FIX.5.0", "233", "TRADERCREDIT"): "TRADERCREDIT .. TraderCredit",
        ("FIX.5.0", "233", "DISCOUNT"): "DISCOUNT .. DiscountRate",
        ("FIX.5.0", "233", "YTM"): "YTM .. YieldToMaturity",
        ("FIX.5.0", "233", "ABS"): "ABS .. AbsolutePrepaymentSpeed",
        ("FIX.5.0", "233", "CPP"): "CPP .. ConstantPrepaymentPenalty",
        ("FIX.5.0", "233", "CPR"): "CPR .. ConstantPrepaymentRate",
        ("FIX.5.0", "233", "CPY"): "CPY .. ConstantPrepaymentYield",
        ("FIX.5.0", "233", "HEP"): "HEP .. FinalCPROfHomeEquityPrepaymentCurve",
        ("FIX.5.0", "233", "MHP"): "MHP .. PercentOfManufacturedHousingPrepaymentCurve",
        ("FIX.5.0", "233", "MPR"): "MPR .. MonthlyPrepaymentRate",
        ("FIX.5.0", "233", "PPC"): "PPC .. PercentOfProspectusPrepaymentCurve",
        ("FIX.5.0", "233", "PSA"): "PSA .. PercentOfBMAPrepaymentCurve",
        ("FIX.5.0", "233", "SMM"): "SMM .. SingleMonthlyMortality",
        ("FIX.5.0", "235", "AFTERTAX"): "AFTERTAX .. AfterTaxYield",
        ("FIX.5.0", "235", "ANNUAL"): "ANNUAL .. AnnualYield",
        ("FIX.5.0", "235", "ATISSUE"): "ATISSUE .. YieldAtIssue",
        ("FIX.5.0", "235", "AVGMATURITY"): "AVGMATURITY .. YieldToAverageMaturity",
        ("FIX.5.0", "235", "BOOK"): "BOOK .. BookYield",
        ("FIX.5.0", "235", "CALL"): "CALL .. YieldToNextCall",
        ("FIX.5.0", "235", "CHANGE"): "CHANGE .. YieldChangeSinceClose",
        ("FIX.5.0", "235", "CLOSE"): "CLOSE .. ClosingYield",
        ("FIX.5.0", "235", "COMPOUND"): "COMPOUND .. CompoundYield",
        ("FIX.5.0", "235", "CURRENT"): "CURRENT .. CurrentYield",
        ("FIX.5.0", "235", "GOVTEQUIV"): "GOVTEQUIV .. GvntEquivalentYield",
        ("FIX.5.0", "235", "GROSS"): "GROSS .. TrueGrossYield",
        ("FIX.5.0", "235", "INFLATION"): "INFLATION .. YieldWithInflationAssumption",
        ("FIX.5.0", "235", "INVERSEFLOATER"): "INVERSEFLOATER .. InverseFloaterBondYield",
        ("FIX.5.0", "235", "LASTCLOSE"): "LASTCLOSE .. MostRecentClosingYield",
        ("FIX.5.0", "235", "LASTMONTH"): "LASTMONTH .. ClosingYieldMostRecentMonth",
        ("FIX.5.0", "235", "LASTQUARTER"): "LASTQUARTER .. ClosingYieldMostRecentQuarter",
        ("FIX.5.0", "235", "LASTYEAR"): "LASTYEAR .. ClosingYieldMostRecentYear",
        ("FIX.5.0", "235", "LONGAVGLIFE"): "LONGAVGLIFE .. YieldToLongestAverageLife",
        ("FIX.5.0", "235", "MARK"): "MARK .. MarkToMarketYield",
        ("FIX.5.0", "235", "MATURITY"): "MATURITY .. YieldToMaturity",
        ("FIX.5.0", "235", "NEXTREFUND"): "NEXTREFUND .. YieldToNextRefund",
        ("FIX.5.0", "235", "OPENAVG"): "OPENAVG .. OpenAverageYield",
        ("FIX.5.0", "235", "PREVCLOSE"): "PREVCLOSE .. PreviousCloseYield",
        ("FIX.5.0", "235", "PROCEEDS"): "PROCEEDS .. ProceedsYield",
        ("FIX.5.0", "235", "PUT"): "PUT .. YieldToNextPut",
        ("FIX.5.0", "235", "SEMIANNUAL"): "SEMIANNUAL .. SemiAnnualYield",
        ("FIX.5.0", "235", "SHORTAVGLIFE"): "SHORTAVGLIFE .. YieldToShortestAverageLife",
        ("FIX.5.0", "235", "SIMPLE"): "SIMPLE .. SimpleYield",
        ("FIX.5.0", "235", "TAXEQUIV"): "TAXEQUIV .. TaxEquivalentYield",
        ("FIX.5.0", "235", "TENDER"): "TENDER .. YieldToTenderDate",
        ("FIX.5.0", "235", "TRUE"): "TRUE .. TrueYield",
        ("FIX.5.0", "235", "VALUE1_32"): "VALUE1_32 .. YieldValueOf32nds",
        ("FIX.5.0", "235", "WORST"): "WORST .. YieldToWorst",
        ("FIX.5.0", "258", "N"): "N .. NotTradedFlat",
        ("FIX.5.0", "258", "Y"): "Y .. TradedFlat",
        ("FIX.5.0", "263", "0"): "0 .. Snapshot",
        ("FIX.5.0", "263", "1"): "1 .. SnapshotAndUpdates",
        ("FIX.5.0", "263", "2"): "2 .. DisablePreviousSnapshot",
        ("FIX.5.0", "265", "0"): "0 .. FullRefresh",
        ("FIX.5.0", "265", "1"): "1 .. IncrementalRefresh",
        ("FIX.5.0", "266", "Y"): "Y .. BookEntriesToBeAggregated",
        ("FIX.5.0", "266", "N"): "N .. BookEntriesShouldNotBeAggregated",
        ("FIX.5.0", "269", "0"): "0 .. Bid",
        ("FIX.5.0", "269", "1"): "1 .. Offer",
        ("FIX.5.0", "269", "2"): "2 .. Trade",
        ("FIX.5.0", "269", "3"): "3 .. IndexValue",
        ("FIX.5.0", "269", "4"): "4 .. OpeningPrice",
        ("FIX.5.0", "269", "5"): "5 .. ClosingPrice",
        ("FIX.5.0", "269", "6"): "6 .. SettlementPrice",
        ("FIX.5.0", "269", "7"): "7 .. TradingSessionHighPrice",
        ("FIX.5.0", "269", "8"): "8 .. TradingSessionLowPrice",
        ("FIX.5.0", "269", "9"): "9 .. TradingSessionVWAPPrice",
        ("FIX.5.0", "269", "A"): "A .. Imbalance",
        ("FIX.5.0", "269", "B"): "B .. TradeVolume",
        ("FIX.5.0", "269", "C"): "C .. OpenInterest",
        ("FIX.5.0", "269", "D"): "D .. CompositeUnderlyingPrice",
        ("FIX.5.0", "269", "E"): "E .. SimulatedSellPrice",
        ("FIX.5.0", "269", "F"): "F .. SimulatedBuyPrice",
        ("FIX.5.0", "269", "G"): "G .. MarginRate",
        ("FIX.5.0", "269", "H"): "H .. MidPrice",
        ("FIX.5.0", "269", "J"): "J .. EmptyBook",
        ("FIX.5.0", "269", "K"): "K .. SettleHighPrice",
        ("FIX.5.0", "269", "L"): "L .. SettleLowPrice",
        ("FIX.5.0", "269", "M"): "M .. PriorSettlePrice",
        ("FIX.5.0", "269", "N"): "N .. SessionHighBid",
        ("FIX.5.0", "269", "O"): "O .. SessionLowOffer",
        ("FIX.5.0", "269", "P"): "P .. EarlyPrices",
        ("FIX.5.0", "269", "Q"): "Q .. AuctionClearingPrice",
        ("FIX.5.0", "269", "S"): "S .. SwapValueFactor",
        ("FIX.5.0", "269", "R"): "R .. DailyValueAdjustmentForLongPositions",
        ("FIX.5.0", "269", "T"): "T .. CumulativeValueAdjustmentForLongPositions",
        ("FIX.5.0", "269", "U"): "U .. DailyValueAdjustmentForShortPositions",
        ("FIX.5.0", "269", "V"): "V .. CumulativeValueAdjustmentForShortPositions",
        ("FIX.5.0", "269", "W"): "W .. FixingPrice",
        ("FIX.5.0", "269", "X"): "X .. CashRate",
        ("FIX.5.0", "269", "Y"): "Y .. RecoveryRate",
        ("FIX.5.0", "269", "Z"): "Z .. RecoveryRateForLong",
        ("FIX.5.0", "269", "a"): "a .. RecoveryRateForShort",
        ("FIX.5.0", "269", "b"): "b .. MarketBid",
        ("FIX.5.0", "269", "c"): "c .. MarketOffer",
        ("FIX.5.0", "269", "d"): "d .. ShortSaleMinPrice",
        ("FIX.5.0", "269", "e"): "e .. PreviousClosingPrice",
        ("FIX.5.0", "274", "0"): "0 .. PlusTick",
        ("FIX.5.0", "274", "1"): "1 .. ZeroPlusTick",
        ("FIX.5.0", "274", "2"): "2 .. MinusTick",
        ("FIX.5.0", "274", "3"): "3 .. ZeroMinusTick",
        ("FIX.5.0", "276", "A"): "A .. Open",
        ("FIX.5.0", "276", "B"): "B .. Closed",
        ("FIX.5.0", "276", "C"): "C .. ExchangeBest",
        ("FIX.5.0", "276", "D"): "D .. ConsolidatedBest",
        ("FIX.5.0", "276", "E"): "E .. Locked",
        ("FIX.5.0", "276", "F"): "F .. Crossed",
        ("FIX.5.0", "276", "G"): "G .. Depth",
        ("FIX.5.0", "276", "H"): "H .. FastTrading",
        ("FIX.5.0", "276", "I"): "I .. NonFirm",
        ("FIX.5.0", "276", "L"): "L .. Manual",
        ("FIX.5.0", "276", "J"): "J .. OutrightPrice",
        ("FIX.5.0", "276", "K"): "K .. ImpliedPrice",
        ("FIX.5.0", "276", "M"): "M .. DepthOnOffer",
        ("FIX.5.0", "276", "N"): "N .. DepthOnBid",
        ("FIX.5.0", "276", "O"): "O .. Closing",
        ("FIX.5.0", "276", "P"): "P .. NewsDissemination",
        ("FIX.5.0", "276", "Q"): "Q .. TradingRange",
        ("FIX.5.0", "276", "R"): "R .. OrderInflux",
        ("FIX.5.0", "276", "S"): "S .. DueToRelated",
        ("FIX.5.0", "276", "T"): "T .. NewsPending",
        ("FIX.5.0", "276", "U"): "U .. AdditionalInfo",
        ("FIX.5.0", "276", "V"): "V .. AdditionalInfoDueToRelated",
        ("FIX.5.0", "276", "W"): "W .. Resume",
        ("FIX.5.0", "276", "X"): "X .. ViewOfCommon",
        ("FIX.5.0", "276", "Y"): "Y .. VolumeAlert",
        ("FIX.5.0", "276", "Z"): "Z .. OrderImbalance",
        ("FIX.5.0", "276", "a"): "a .. EquipmentChangeover",
        ("FIX.5.0", "276", "b"): "b .. NoOpen",
        ("FIX.5.0", "276", "c"): "c .. RegularETH",
        ("FIX.5.0", "276", "d"): "d .. AutomaticExecution",
        ("FIX.5.0", "276", "e"): "e .. AutomaticExecutionETH",
        ("FIX.5.0", "276", "f "): "f  .. FastMarketETH",
        ("FIX.5.0", "276", "g"): "g .. InactiveETH",
        ("FIX.5.0", "276", "h"): "h .. Rotation",
        ("FIX.5.0", "276", "i"): "i .. RotationETH",
        ("FIX.5.0", "276", "j"): "j .. Halt",
        ("FIX.5.0", "276", "k"): "k .. HaltETH",
        ("FIX.5.0", "276", "l"): "l .. DueToNewsDissemination",
        ("FIX.5.0", "276", "m"): "m .. DueToNewsPending",
        ("FIX.5.0", "276", "n"): "n .. TradingResume",
        ("FIX.5.0", "276", "o"): "o .. OutOfSequence",
        ("FIX.5.0", "276", "p"): "p .. BidSpecialist",
        ("FIX.5.0", "276", "q"): "q .. OfferSpecialist",
        ("FIX.5.0", "276", "r"): "r .. BidOfferSpecialist",
        ("FIX.5.0", "276", "s"): "s .. EndOfDaySAM",
        ("FIX.5.0", "276", "t"): "t .. ForbiddenSAM",
        ("FIX.5.0", "276", "u"): "u .. FrozenSAM",
        ("FIX.5.0", "276", "v"): "v .. PreOpeningSAM",
        ("FIX.5.0", "276", "w"): "w .. OpeningSAM",
        ("FIX.5.0", "276", "x"): "x .. OpenSAM",
        ("FIX.5.0", "276", "y"): "y .. SurveillanceSAM",
        ("FIX.5.0", "276", "z"): "z .. SuspendedSAM",
        ("FIX.5.0", "276", "0"): "0 .. ReservedSAM",
        ("FIX.5.0", "276", "1"): "1 .. NoActiveSAM",
        ("FIX.5.0", "276", "2"): "2 .. Restricted",
        ("FIX.5.0", "276", "3"): "3 .. RestOfBookVWAP",
        ("FIX.5.0", "276", "4"): "4 .. BetterPricesInConditionalOrders",
        ("FIX.5.0", "276", "5"): "5 .. MedianPrice",
        ("FIX.5.0", "276", "6"): "6 .. FullCurve",
        ("FIX.5.0", "276", "7"): "7 .. FlatCurve",
        ("FIX.5.0", "277", "A"): "A .. Cash",
        ("FIX.5.0", "277", "B"): "B .. AveragePriceTrade",
        ("FIX.5.0", "277", "C"): "C .. CashTrade",
        ("FIX.5.0", "277", "D"): "D .. NextDay",
        ("FIX.5.0", "277", "E"): "E .. Opening",
        ("FIX.5.0", "277", "F"): "F .. IntradayTradeDetail",
        ("FIX.5.0", "277", "G"): "G .. Rule127Trade",
        ("FIX.5.0", "277", "H"): "H .. Rule155Trade",
        ("FIX.5.0", "277", "I"): "I .. SoldLast",
        ("FIX.5.0", "277", "J"): "J .. NextDayTrade",
        ("FIX.5.0", "277", "K"): "K .. Opened",
        ("FIX.5.0", "277", "L"): "L .. Seller",
        ("FIX.5.0", "277", "M"): "M .. Sold",
        ("FIX.5.0", "277", "N"): "N .. StoppedStock",
        ("FIX.5.0", "277", "P"): "P .. ImbalanceMoreBuyers",
        ("FIX.5.0", "277", "Q"): "Q .. ImbalanceMoreSellers",
        ("FIX.5.0", "277", "R"): "R .. OpeningPrice",
        ("FIX.5.0", "277", "S"): "S .. BargainCondition",
        ("FIX.5.0", "277", "T"): "T .. ConvertedPriceIndicator",
        ("FIX.5.0", "277", "U"): "U .. ExchangeLast",
        ("FIX.5.0", "277", "V"): "V .. FinalPriceOfSession",
        ("FIX.5.0", "277", "W"): "W .. ExPit",
        ("FIX.5.0", "277", "X"): "X .. Crossed",
        ("FIX.5.0", "277", "Y"): "Y .. TradesResultingFromManual",
        ("FIX.5.0", "277", "Z"): "Z .. TradesResultingFromIntermarketSweep",
        ("FIX.5.0", "277", "a"): "a .. VolumeOnly",
        ("FIX.5.0", "277", "b"): "b .. DirectPlus",
        ("FIX.5.0", "277", "c"): "c .. Acquisition",
        ("FIX.5.0", "277", "d"): "d .. Bunched",
        ("FIX.5.0", "277", "e"): "e .. Distribution",
        ("FIX.5.0", "277", "f"): "f .. BunchedSale",
        ("FIX.5.0", "277", "g"): "g .. SplitTrade",
        ("FIX.5.0", "277", "h"): "h .. CancelStopped",
        ("FIX.5.0", "277", "i"): "i .. CancelETH",
        ("FIX.5.0", "277", "j"): "j .. CancelStoppedETH",
        ("FIX.5.0", "277", "k"): "k .. OutOfSequenceETH",
        ("FIX.5.0", "277", "l"): "l .. CancelLastETH",
        ("FIX.5.0", "277", "m"): "m .. SoldLastSaleETH",
        ("FIX.5.0", "277", "n"): "n .. CancelLast",
        ("FIX.5.0", "277", "o"): "o .. SoldLastSale",
        ("FIX.5.0", "277", "p"): "p .. CancelOpen",
        ("FIX.5.0", "277", "q"): "q .. CancelOpenETH",
        ("FIX.5.0", "277", "r"): "r .. OpenedSaleETH",
        ("FIX.5.0", "277", "s"): "s .. CancelOnly",
        ("FIX.5.0", "277", "t"): "t .. CancelOnlyETH",
        ("FIX.5.0", "277", "u"): "u .. LateOpenETH",
        ("FIX.5.0", "277", "v"): "v .. AutoExecutionETH",
        ("FIX.5.0", "277", "w"): "w .. Reopen",
        ("FIX.5.0", "277", "x"): "x .. ReopenETH",
        ("FIX.5.0", "277", "y"): "y .. Adjusted",
        ("FIX.5.0", "277", "z"): "z .. AdjustedETH",
        ("FIX.5.0", "277", "AA"): "AA .. Spread",
        ("FIX.5.0", "277", "AB"): "AB .. SpreadETH",
        ("FIX.5.0", "277", "AC"): "AC .. Straddle",
        ("FIX.5.0", "277", "AD"): "AD .. StraddleETH",
        ("FIX.5.0", "277", "AE"): "AE .. Stopped",
        ("FIX.5.0", "277", "AF"): "AF .. StoppedETH",
        ("FIX.5.0", "277", "AG"): "AG .. RegularETH",
        ("FIX.5.0", "277", "AH"): "AH .. Combo",
        ("FIX.5.0", "277", "AI"): "AI .. ComboETH",
        ("FIX.5.0", "277", "AJ"): "AJ .. OfficialClosingPrice",
        ("FIX.5.0", "277", "AK"): "AK .. PriorReferencePrice",
        ("FIX.5.0", "277", "0"): "0 .. Cancel",
        ("FIX.5.0", "277", "AL"): "AL .. StoppedSoldLast",
        ("FIX.5.0", "277", "AM"): "AM .. StoppedOutOfSequence",
        ("FIX.5.0", "277", "AN"): "AN .. OfficalClosingPrice",
        ("FIX.5.0", "277", "AO"): "AO .. CrossedOld",
        ("FIX.5.0", "277", "AP"): "AP .. FastMarket",
        ("FIX.5.0", "277", "AQ"): "AQ .. AutomaticExecution",
        ("FIX.5.0", "277", "AR"): "AR .. FormT",
        ("FIX.5.0", "277", "AS"): "AS .. BasketIndex",
        ("FIX.5.0", "277", "AT"): "AT .. BurstBasket",
        ("FIX.5.0", "277", "AU"): "AU .. TradeThroughExempt",
        ("FIX.5.0", "277", "AV"): "AV .. QuoteSpread",
        ("FIX.5.0", "277", "AW"): "AW .. LastAuctionPrice",
        ("FIX.5.0", "277", "AX"): "AX .. HighPrice",
        ("FIX.5.0", "277", "AY"): "AY .. LowPrice",
        ("FIX.5.0", "277", "AZ"): "AZ .. SystematicInternaliser",
        ("FIX.5.0", "277", "BA"): "BA .. AwayMarket",
        ("FIX.5.0", "277", "BB"): "BB .. MidpointPrice",
        ("FIX.5.0", "277", "BC"): "BC .. TradedBeforeIssueDate",
        ("FIX.5.0", "277", "BD"): "BD .. PreviousClosingPrice",
        ("FIX.5.0", "277", "BE"): "BE .. NationalBestBidOffer",
        ("FIX.5.0", "277", "1"): "1 .. ImpliedTrade",
        ("FIX.5.0", "277", "2"): "2 .. MarketplaceEnteredTrade",
        ("FIX.5.0", "277", "3"): "3 .. MultiAssetClassMultilegTrade",
        ("FIX.5.0", "277", "4"): "4 .. MultilegToMultilegTrade",
        ("FIX.5.0", "277", "5"): "5 .. ShortSaleMinPrice",
        ("FIX.5.0", "277", "6"): "6 .. Benchmark",
        ("FIX.5.0", "279", "0"): "0 .. New",
        ("FIX.5.0", "279", "1"): "1 .. Change",
        ("FIX.5.0", "279", "2"): "2 .. Delete",
        ("FIX.5.0", "279", "3"): "3 .. DeleteThru",
        ("FIX.5.0", "279", "4"): "4 .. DeleteFrom",
        ("FIX.5.0", "279", "5"): "5 .. Overlay",
        ("FIX.5.0", "281", "0"): "0 .. UnknownSymbol",
        ("FIX.5.0", "281", "1"): "1 .. DuplicateMDReqID",
        ("FIX.5.0", "281", "2"): "2 .. InsufficientBandwidth",
        ("FIX.5.0", "281", "3"): "3 .. InsufficientPermissions",
        ("FIX.5.0", "281", "4"): "4 .. UnsupportedSubscriptionRequestType",
        ("FIX.5.0", "281", "5"): "5 .. UnsupportedMarketDepth",
        ("FIX.5.0", "281", "6"): "6 .. UnsupportedMDUpdateType",
        ("FIX.5.0", "281", "7"): "7 .. UnsupportedAggregatedBook",
        ("FIX.5.0", "281", "8"): "8 .. UnsupportedMDEntryType",
        ("FIX.5.0", "281", "9"): "9 .. UnsupportedTradingSessionID",
        ("FIX.5.0", "281", "A"): "A .. UnsupportedScope",
        ("FIX.5.0", "281", "B"): "B .. UnsupportedOpenCloseSettleFlag",
        ("FIX.5.0", "281", "C"): "C .. UnsupportedMDImplicitDelete",
        ("FIX.5.0", "281", "D"): "D .. InsufficientCredit",
        ("FIX.5.0", "285", "0"): "0 .. Cancellation",
        ("FIX.5.0", "285", "1"): "1 .. Error",
        ("FIX.5.0", "286", "0"): "0 .. DailyOpen",
        ("FIX.5.0", "286", "1"): "1 .. SessionOpen",
        ("FIX.5.0", "286", "2"): "2 .. DeliverySettlementEntry",
        ("FIX.5.0", "286", "3"): "3 .. ExpectedEntry",
        ("FIX.5.0", "286", "4"): "4 .. EntryFromPreviousBusinessDay",
        ("FIX.5.0", "286", "5"): "5 .. TheoreticalPriceValue",
        ("FIX.5.0", "291", "1"): "1 .. Bankrupt",
        ("FIX.5.0", "291", "2"): "2 .. PendingDelisting",
        ("FIX.5.0", "291", "3"): "3 .. Restricted",
        ("FIX.5.0", "292", "A"): "A .. ExDividend",
        ("FIX.5.0", "292", "B"): "B .. ExDistribution",
        ("FIX.5.0", "292", "C"): "C .. ExRights",
        ("FIX.5.0", "292", "D"): "D .. New",
        ("FIX.5.0", "292", "E"): "E .. ExInterest",
        ("FIX.5.0", "292", "F"): "F .. CashDividend",
        ("FIX.5.0", "292", "G"): "G .. StockDividend",
        ("FIX.5.0", "292", "H"): "H .. NonIntegerStockSplit",
        ("FIX.5.0", "292", "I"): "I .. ReverseStockSplit",
        ("FIX.5.0", "292", "J"): "J .. StandardIntegerStockSplit",
        ("FIX.5.0", "292", "K"): "K .. PositionConsolidation",
        ("FIX.5.0", "292", "L"): "L .. LiquidationReorganization",
        ("FIX.5.0", "292", "M"): "M .. MergerReorganization",
        ("FIX.5.0", "292", "N"): "N .. RightsOffering",
        ("FIX.5.0", "292", "O"): "O .. ShareholderMeeting",
        ("FIX.5.0", "292", "P"): "P .. Spinoff",
        ("FIX.5.0", "292", "Q"): "Q .. TenderOffer",
        ("FIX.5.0", "292", "R"): "R .. Warrant",
        ("FIX.5.0", "292", "S"): "S .. SpecialAction",
        ("FIX.5.0", "292", "T"): "T .. SymbolConversion",
        ("FIX.5.0", "292", "U"): "U .. CUSIP",
        ("FIX.5.0", "292", "V"): "V .. LeapRollover",
        ("FIX.5.0", "292", "W"): "W .. SuccessionEvent",
        ("FIX.5.0", "297", "0"): "0 .. Accepted",
        ("FIX.5.0", "297", "5"): "5 .. Rejected",
        ("FIX.5.0", "297", "6"): "6 .. RemovedFromMarket",
        ("FIX.5.0", "297", "7"): "7 .. Expired",
        ("FIX.5.0", "297", "8"): "8 .. Query",
        ("FIX.5.0", "297", "9"): "9 .. QuoteNotFound",
        ("FIX.5.0", "297", "10"): "10 .. Pending",
        ("FIX.5.0", "297", "11"): "11 .. Pass",
        ("FIX.5.0", "297", "12"): "12 .. LockedMarketWarning",
        ("FIX.5.0", "297", "13"): "13 .. CrossMarketWarning",
        ("FIX.5.0", "297", "14"): "14 .. CanceledDueToLockMarket",
        ("FIX.5.0", "297", "15"): "15 .. CanceledDueToCrossMarket",
        ("FIX.5.0", "297", "16"): "16 .. Active",
        ("FIX.5.0", "297", "17"): "17 .. Canceled",
        ("FIX.5.0", "297", "18"): "18 .. UnsolicitedQuoteReplenishment",
        ("FIX.5.0", "297", "19"): "19 .. PendingEndTrade",
        ("FIX.5.0", "297", "20"): "20 .. TooLateToEnd",
        ("FIX.5.0", "297", "21"): "21 .. Traded",
        ("FIX.5.0", "297", "22"): "22 .. TradedAndRemoved",
        ("FIX.5.0", "298", "1"): "1 .. CancelForOneOrMoreSecurities",
        ("FIX.5.0", "298", "2"): "2 .. CancelForSecurityType",
        ("FIX.5.0", "298", "3"): "3 .. CancelForUnderlyingSecurity",
        ("FIX.5.0", "298", "4"): "4 .. CancelAllQuotes",
        ("FIX.5.0", "298", "5"): "5 .. CancelSpecifiedSingleQuote",
        ("FIX.5.0", "298", "6"): "6 .. CancelByTypeOfQuote",
        ("FIX.5.0", "298", "7"): "7 .. CancelForSecurityIssuer",
        ("FIX.5.0", "298", "8"): "8 .. CancelForIssuerOfUnderlyingSecurity",
        ("FIX.5.0", "300", "1"): "1 .. UnknownSymbol",
        ("FIX.5.0", "300", "2"): "2 .. Exchange",
        ("FIX.5.0", "300", "3"): "3 .. QuoteRequestExceedsLimit",
        ("FIX.5.0", "300", "4"): "4 .. TooLateToEnter",
        ("FIX.5.0", "300", "5"): "5 .. UnknownQuote",
        ("FIX.5.0", "300", "6"): "6 .. DuplicateQuote",
        ("FIX.5.0", "300", "7"): "7 .. InvalidBid",
        ("FIX.5.0", "300", "8"): "8 .. InvalidPrice",
        ("FIX.5.0", "300", "9"): "9 .. NotAuthorizedToQuoteSecurity",
        ("FIX.5.0", "300", "10"): "10 .. PriceExceedsCurrentPriceBand",
        ("FIX.5.0", "300", "11"): "11 .. QuoteLocked",
        ("FIX.5.0", "300", "12"): "12 .. InvalidOrUnknownSecurityIssuer",
        ("FIX.5.0", "300", "13"): "13 .. InvalidOrUnknownIssuerOfUnderlyingSecurity",
        ("FIX.5.0", "300", "14"): "14 .. NotionalValueExceedsThreshold",
        ("FIX.5.0", "300", "16"): "16 .. ReferencePriceNotAvailable",
        ("FIX.5.0", "300", "17"): "17 .. InsufficientCreditLimit",
        ("FIX.5.0", "300", "18"): "18 .. ExceededClipSizeLimit",
        ("FIX.5.0", "300", "19"): "19 .. ExceededMaxNotionalOrderAmt",
        ("FIX.5.0", "300", "20"): "20 .. ExceededDV01PV01Limit",
        ("FIX.5.0", "300", "21"): "21 .. ExceededCS01Limit",
        ("FIX.5.0", "300", "99"): "99 .. Other",
        ("FIX.5.0", "301", "0"): "0 .. NoAcknowledgement",
        ("FIX.5.0", "301", "1"): "1 .. AcknowledgeOnlyNegativeOrErroneousQuotes",
        ("FIX.5.0", "301", "2"): "2 .. AcknowledgeEachQuoteMessage",
        ("FIX.5.0", "301", "3"): "3 .. SummaryAcknowledgement",
        ("FIX.5.0", "303", "1"): "1 .. Manual",
        ("FIX.5.0", "303", "2"): "2 .. Automatic",
        ("FIX.5.0", "303", "3"): "3 .. ConfirmQuote",
        ("FIX.5.0", "305", "1"): "1 .. CUSIP",
        ("FIX.5.0", "305", "2"): "2 .. SEDOL",
        ("FIX.5.0", "305", "3"): "3 .. QUIK",
        ("FIX.5.0", "305", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "305", "5"): "5 .. RICCode",
        ("FIX.5.0", "305", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "305", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "305", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "305", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "305", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "305", "B"): "B .. Wertpapier",
        ("FIX.5.0", "305", "C"): "C .. Dutch",
        ("FIX.5.0", "305", "D"): "D .. Valoren",
        ("FIX.5.0", "305", "E"): "E .. Sicovam",
        ("FIX.5.0", "305", "F"): "F .. Belgian",
        ("FIX.5.0", "305", "G"): "G .. Common",
        ("FIX.5.0", "305", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "305", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "305", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "305", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "305", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "305", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "305", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "305", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "305", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "305", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "305", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "305", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "305", "U"): "U .. Synthetic",
        ("FIX.5.0", "305", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "305", "W"): "W .. IndexName",
        ("FIX.5.0", "310", "EUSUPRA"): "EUSUPRA .. EuroSupranationalCoupons",
        ("FIX.5.0", "310", "FAC"): "FAC .. FederalAgencyCoupon",
        ("FIX.5.0", "310", "FADN"): "FADN .. FederalAgencyDiscountNote",
        ("FIX.5.0", "310", "PEF"): "PEF .. PrivateExportFunding",
        ("FIX.5.0", "310", "SUPRA"): "SUPRA .. USDSupranationalCoupons",
        ("FIX.5.0", "310", "CORP"): "CORP .. CorporateBond",
        ("FIX.5.0", "310", "CPP"): "CPP .. CorporatePrivatePlacement",
        ("FIX.5.0", "310", "CB"): "CB .. ConvertibleBond",
        ("FIX.5.0", "310", "DUAL"): "DUAL .. DualCurrency",
        ("FIX.5.0", "310", "EUCORP"): "EUCORP .. EuroCorporateBond",
        ("FIX.5.0", "310", "EUFRN"): "EUFRN .. EuroCorporateFloatingRateNotes",
        ("FIX.5.0", "310", "FRN"): "FRN .. USCorporateFloatingRateNotes",
        ("FIX.5.0", "310", "XLINKD"): "XLINKD .. IndexedLinked",
        ("FIX.5.0", "310", "STRUCT"): "STRUCT .. StructuredNotes",
        ("FIX.5.0", "310", "YANK"): "YANK .. YankeeCorporateBond",
        ("FIX.5.0", "310", "FXNDF"): "FXNDF .. NonDeliverableForward",
        ("FIX.5.0", "310", "FXSPOT"): "FXSPOT .. FXSpot",
        ("FIX.5.0", "310", "FXFWD"): "FXFWD .. FXForward",
        ("FIX.5.0", "310", "FXSWAP"): "FXSWAP .. FXSwap",
        ("FIX.5.0", "310", "FXNDS"): "FXNDS .. NonDeliverableSwap",
        ("FIX.5.0", "310", "CAP"): "CAP .. Cap",
        ("FIX.5.0", "310", "CDS"): "CDS .. CreditDefaultSwap",
        ("FIX.5.0", "310", "CLLR"): "CLLR .. Collar",
        ("FIX.5.0", "310", "CMDTYSWAP"): "CMDTYSWAP .. CommoditySwap",
        ("FIX.5.0", "310", "EXOTIC"): "EXOTIC .. Exotic",
        ("FIX.5.0", "310", "OOC"): "OOC .. OptionsOnCombo",
        ("FIX.5.0", "310", "FLR"): "FLR .. Floor",
        ("FIX.5.0", "310", "FRA"): "FRA .. FRA",
        ("FIX.5.0", "310", "FUT"): "FUT .. Future",
        ("FIX.5.0", "310", "FWD"): "FWD .. DerivativeForward",
        ("FIX.5.0", "310", "IRS"): "IRS .. InterestRateSwap",
        ("FIX.5.0", "310", "TRS"): "TRS .. TotalReturnSwap",
        ("FIX.5.0", "310", "LOANLEASE"): "LOANLEASE .. LoanLease",
        ("FIX.5.0", "310", "OOF"): "OOF .. OptionsOnFutures",
        ("FIX.5.0", "310", "OOP"): "OOP .. OptionsOnPhysical",
        ("FIX.5.0", "310", "OPT"): "OPT .. Option",
        ("FIX.5.0", "310", "SPOTFWD"): "SPOTFWD .. SpotForward",
        ("FIX.5.0", "310", "SWAPTION"): "SWAPTION .. SwapOption",
        ("FIX.5.0", "310", "XMISSION"): "XMISSION .. Transmission",
        ("FIX.5.0", "310", "INDEX"): "INDEX .. Index",
        ("FIX.5.0", "310", "BDBSKT"): "BDBSKT .. BondBasket",
        ("FIX.5.0", "310", "CFD"): "CFD .. ContractForDifference",
        ("FIX.5.0", "310", "CRLTNSWAP"): "CRLTNSWAP .. CorrelationSwap",
        ("FIX.5.0", "310", "DVDNDSWAP"): "DVDNDSWAP .. DiviendSwap",
        ("FIX.5.0", "310", "EQBSKT"): "EQBSKT .. EquityBasket",
        ("FIX.5.0", "310", "EQFWD"): "EQFWD .. EquityForward",
        ("FIX.5.0", "310", "RTRNSWAP"): "RTRNSWAP .. ReturnSwap",
        ("FIX.5.0", "310", "VARSWAP"): "VARSWAP .. VarianceSwap",
        ("FIX.5.0", "310", "PRTFLIOSWAP"): "PRTFLIOSWAP .. PortfolioSwaps",
        ("FIX.5.0", "310", "FUTSWAP"): "FUTSWAP .. FuturesOnASwap",
        ("FIX.5.0", "310", "FWDSWAP "): "FWDSWAP  .. ForwardsOnASwap",
        ("FIX.5.0", "310", "FWDFRTAGMT "): "FWDFRTAGMT  .. ForwardFreightAgreement",
        ("FIX.5.0", "310", "SPREADBET"): "SPREADBET .. SpreadBetting",
        ("FIX.5.0", "310", "ETC"): "ETC .. ExchangeTradedCommodity",
        ("FIX.5.0", "310", "CS"): "CS .. CommonStock",
        ("FIX.5.0", "310", "PS"): "PS .. PreferredStock",
        ("FIX.5.0", "310", "DR"): "DR .. DepositoryReceipts",
        ("FIX.5.0", "310", "REPO"): "REPO .. Repurchase",
        ("FIX.5.0", "310", "FORWARD"): "FORWARD .. Forward",
        ("FIX.5.0", "310", "BUYSELL"): "BUYSELL .. BuySellback",
        ("FIX.5.0", "310", "SECLOAN"): "SECLOAN .. SecuritiesLoan",
        ("FIX.5.0", "310", "SECPLEDGE"): "SECPLEDGE .. SecuritiesPledge",
        ("FIX.5.0", "310", "DVPLDG"): "DVPLDG .. DeliveryVersusPledge",
        ("FIX.5.0", "310", "COLLBSKT"): "COLLBSKT .. CollateralBasket",
        ("FIX.5.0", "310", "SFP"): "SFP .. StructuredFinanceProduct",
        ("FIX.5.0", "310", "BRADY"): "BRADY .. BradyBond",
        ("FIX.5.0", "310", "CAN"): "CAN .. CanadianTreasuryNotes",
        ("FIX.5.0", "310", "CTB"): "CTB .. CanadianTreasuryBills",
        ("FIX.5.0", "310", "EUSOV"): "EUSOV .. EuroSovereigns",
        ("FIX.5.0", "310", "PROV"): "PROV .. CanadianProvincialBonds",
        ("FIX.5.0", "310", "TB"): "TB .. TreasuryBill",
        ("FIX.5.0", "310", "TBOND"): "TBOND .. USTreasuryBond",
        ("FIX.5.0", "310", "TINT"): "TINT .. InterestStripFromAnyBondOrNote",
        ("FIX.5.0", "310", "TBILL"): "TBILL .. USTreasuryBill",
        ("FIX.5.0", "310", "TIPS"): "TIPS .. TreasuryInflationProtectedSecurities",
        ("FIX.5.0", "310", "TCAL"): "TCAL .. PrincipalStripOfACallableBondOrNote",
        ("FIX.5.0", "310", "TPRN"): "TPRN .. PrincipalStripFromANonCallableBondOrNote",
        ("FIX.5.0", "310", "TNOTE"): "TNOTE .. USTreasuryNote",
        ("FIX.5.0", "310", "TERM"): "TERM .. TermLoan",
        ("FIX.5.0", "310", "RVLV"): "RVLV .. RevolverLoan",
        ("FIX.5.0", "310", "RVLVTRM"): "RVLVTRM .. Revolver",
        ("FIX.5.0", "310", "BRIDGE"): "BRIDGE .. BridgeLoan",
        ("FIX.5.0", "310", "LOFC"): "LOFC .. LetterOfCredit",
        ("FIX.5.0", "310", "SWING"): "SWING .. SwingLineFacility",
        ("FIX.5.0", "310", "DINP"): "DINP .. DebtorInPossession",
        ("FIX.5.0", "310", "DEFLTED"): "DEFLTED .. Defaulted",
        ("FIX.5.0", "310", "WITHDRN"): "WITHDRN .. Withdrawn",
        ("FIX.5.0", "310", "REPLACD"): "REPLACD .. Replaced",
        ("FIX.5.0", "310", "MATURED"): "MATURED .. Matured",
        ("FIX.5.0", "310", "AMENDED"): "AMENDED .. Amended",
        ("FIX.5.0", "310", "RETIRED"): "RETIRED .. Retired",
        ("FIX.5.0", "310", "BA"): "BA .. BankersAcceptance",
        ("FIX.5.0", "310", "BDN"): "BDN .. BankDepositoryNote",
        ("FIX.5.0", "310", "BN"): "BN .. BankNotes",
        ("FIX.5.0", "310", "BOX"): "BOX .. BillOfExchanges",
        ("FIX.5.0", "310", "CAMM"): "CAMM .. CanadianMoneyMarkets",
        ("FIX.5.0", "310", "CD"): "CD .. CertificateOfDeposit",
        ("FIX.5.0", "310", "CL"): "CL .. CallLoans",
        ("FIX.5.0", "310", "CP"): "CP .. CommercialPaper",
        ("FIX.5.0", "310", "DN"): "DN .. DepositNotes",
        ("FIX.5.0", "310", "EUCD"): "EUCD .. EuroCertificateOfDeposit",
        ("FIX.5.0", "310", "EUCP"): "EUCP .. EuroCommercialPaper",
        ("FIX.5.0", "310", "LQN"): "LQN .. LiquidityNote",
        ("FIX.5.0", "310", "MTN"): "MTN .. MediumTermNotes",
        ("FIX.5.0", "310", "ONITE"): "ONITE .. Overnight",
        ("FIX.5.0", "310", "PN"): "PN .. PromissoryNote",
        ("FIX.5.0", "310", "STN"): "STN .. ShortTermLoanNote",
        ("FIX.5.0", "310", "PZFJ"): "PZFJ .. PlazosFijos",
        ("FIX.5.0", "310", "SLQN"): "SLQN .. SecuredLiquidityNote",
        ("FIX.5.0", "310", "TD"): "TD .. TimeDeposit",
        ("FIX.5.0", "310", "TLQN"): "TLQN .. TermLiquidityNote",
        ("FIX.5.0", "310", "XCN"): "XCN .. ExtendedCommNote",
        ("FIX.5.0", "310", "YCD"): "YCD .. YankeeCertificateOfDeposit",
        ("FIX.5.0", "310", "ABS"): "ABS .. AssetBackedSecurities",
        ("FIX.5.0", "310", "CMB"): "CMB .. CanadianMortgageBonds",
        ("FIX.5.0", "310", "CMBS"): "CMBS .. Corp",
        ("FIX.5.0", "310", "CMO"): "CMO .. CollateralizedMortgageObligation",
        ("FIX.5.0", "310", "IET"): "IET .. IOETTEMortgage",
        ("FIX.5.0", "310", "MBS"): "MBS .. MortgageBackedSecurities",
        ("FIX.5.0", "310", "MIO"): "MIO .. MortgageInterestOnly",
        ("FIX.5.0", "310", "MPO"): "MPO .. MortgagePrincipalOnly",
        ("FIX.5.0", "310", "MPP"): "MPP .. MortgagePrivatePlacement",
        ("FIX.5.0", "310", "MPT"): "MPT .. MiscellaneousPassThrough",
        ("FIX.5.0", "310", "PFAND"): "PFAND .. Pfandbriefe",
        ("FIX.5.0", "310", "TBA"): "TBA .. ToBeAnnounced",
        ("FIX.5.0", "310", "AN"): "AN .. OtherAnticipationNotes",
        ("FIX.5.0", "310", "COFO"): "COFO .. CertificateOfObligation",
        ("FIX.5.0", "310", "COFP"): "COFP .. CertificateOfParticipation",
        ("FIX.5.0", "310", "GO"): "GO .. GeneralObligationBonds",
        ("FIX.5.0", "310", "MT"): "MT .. MandatoryTender",
        ("FIX.5.0", "310", "RAN"): "RAN .. RevenueAnticipationNote",
        ("FIX.5.0", "310", "REV"): "REV .. RevenueBonds",
        ("FIX.5.0", "310", "SPCLA"): "SPCLA .. SpecialAssessment",
        ("FIX.5.0", "310", "SPCLO"): "SPCLO .. SpecialObligation",
        ("FIX.5.0", "310", "SPCLT"): "SPCLT .. SpecialTax",
        ("FIX.5.0", "310", "TAN"): "TAN .. TaxAnticipationNote",
        ("FIX.5.0", "310", "TAXA"): "TAXA .. TaxAllocation",
        ("FIX.5.0", "310", "TECP"): "TECP .. TaxExemptCommercialPaper",
        ("FIX.5.0", "310", "TMCP"): "TMCP .. TaxableMunicipalCP",
        ("FIX.5.0", "310", "TRAN"): "TRAN .. TaxRevenueAnticipationNote",
        ("FIX.5.0", "310", "VRDN"): "VRDN .. VariableRateDemandNote",
        ("FIX.5.0", "310", "WAR"): "WAR .. Warrant",
        ("FIX.5.0", "310", "MF"): "MF .. MutualFund",
        ("FIX.5.0", "310", "MLEG"): "MLEG .. MultilegInstrument",
        ("FIX.5.0", "310", "NONE"): "NONE .. NoSecurityType",
        ("FIX.5.0", "310", "?"): "? .. Wildcard",
        ("FIX.5.0", "310", "CASH"): "CASH .. Cash",
        ("FIX.5.0", "310", "Other"): "Other .. Other",
        ("FIX.5.0", "310", "ETN"): "ETN .. ExchangeTradedNote",
        ("FIX.5.0", "310", "SECDERIV"): "SECDERIV .. SecuritizedDerivative",
        ("FIX.5.0", "312", "CD"): "CD .. EUCPWithLumpSumInterest",
        ("FIX.5.0", "312", "WI"): "WI .. WhenIssued",
        ("FIX.5.0", "315", "0"): "0 .. Put",
        ("FIX.5.0", "315", "1"): "1 .. Call",
        ("FIX.5.0", "315", "2"): "2 .. Other",
        ("FIX.5.0", "315", "3"): "3 .. Chooser",
        ("FIX.5.0", "321", "0"): "0 .. RequestSecurityIdentityAndSpecifications",
        ("FIX.5.0", "321", "1"): "1 .. RequestSecurityIdentityForSpecifications",
        ("FIX.5.0", "321", "4"): "4 .. Symbol",
        ("FIX.5.0", "321", "5"): "5 .. SecurityTypeAndOrCFICode",
        ("FIX.5.0", "321", "6"): "6 .. Product",
        ("FIX.5.0", "321", "7"): "7 .. TradingSessionID",
        ("FIX.5.0", "321", "8"): "8 .. AllSecurities",
        ("FIX.5.0", "321", "9"): "9 .. MarketIDOrMarketID",
        ("FIX.5.0", "323", "1"): "1 .. AcceptAsIs",
        ("FIX.5.0", "323", "2"): "2 .. AcceptWithRevisions",
        ("FIX.5.0", "323", "4"): "4 .. ListOfSecuritiesReturnedPerRequest",
        ("FIX.5.0", "323", "5"): "5 .. RejectSecurityProposal",
        ("FIX.5.0", "323", "6"): "6 .. CannotMatchSelectionCriteria",
        ("FIX.5.0", "325", "N"): "N .. MessageIsBeingSentAsAResultOfAPriorRequest",
        ("FIX.5.0", "325", "Y"): "Y .. MessageIsBeingSentUnsolicited",
        ("FIX.5.0", "326", "1"): "1 .. OpeningDelay",
        ("FIX.5.0", "326", "2"): "2 .. TradingHalt",
        ("FIX.5.0", "326", "3"): "3 .. Resume",
        ("FIX.5.0", "326", "4"): "4 .. NoOpen",
        ("FIX.5.0", "326", "5"): "5 .. PriceIndication",
        ("FIX.5.0", "326", "6"): "6 .. TradingRangeIndication",
        ("FIX.5.0", "326", "7"): "7 .. MarketImbalanceBuy",
        ("FIX.5.0", "326", "8"): "8 .. MarketImbalanceSell",
        ("FIX.5.0", "326", "9"): "9 .. MarketOnCloseImbalanceBuy",
        ("FIX.5.0", "326", "10"): "10 .. MarketOnCloseImbalanceSell",
        ("FIX.5.0", "326", "12"): "12 .. NoMarketImbalance",
        ("FIX.5.0", "326", "13"): "13 .. NoMarketOnCloseImbalance",
        ("FIX.5.0", "326", "14"): "14 .. ITSPreOpening",
        ("FIX.5.0", "326", "15"): "15 .. NewPriceIndication",
        ("FIX.5.0", "326", "16"): "16 .. TradeDisseminationTime",
        ("FIX.5.0", "326", "17"): "17 .. ReadyToTrade",
        ("FIX.5.0", "326", "18"): "18 .. NotAvailableForTrading",
        ("FIX.5.0", "326", "19"): "19 .. NotTradedOnThisMarket",
        ("FIX.5.0", "326", "20"): "20 .. UnknownOrInvalid",
        ("FIX.5.0", "326", "21"): "21 .. PreOpen",
        ("FIX.5.0", "326", "22"): "22 .. OpeningRotation",
        ("FIX.5.0", "326", "23"): "23 .. FastMarket",
        ("FIX.5.0", "326", "24"): "24 .. PreCross",
        ("FIX.5.0", "326", "25"): "25 .. Cross",
        ("FIX.5.0", "326", "26"): "26 .. PostClose",
        ("FIX.5.0", "326", "27"): "27 .. NoCancel",
        ("FIX.5.0", "327", "0"): "0 .. NewsDissemination",
        ("FIX.5.0", "327", "1"): "1 .. OrderInflux",
        ("FIX.5.0", "327", "2"): "2 .. OrderImbalance",
        ("FIX.5.0", "327", "3"): "3 .. AdditionalInformation",
        ("FIX.5.0", "327", "4"): "4 .. NewsPending",
        ("FIX.5.0", "327", "5"): "5 .. EquipmentChangeover",
        ("FIX.5.0", "328", "N"): "N .. HaltWasNotRelatedToAHaltOfTheCommonStock",
        ("FIX.5.0", "328", "Y"): "Y .. HaltWasDueToCommonStockBeingHalted",
        ("FIX.5.0", "329", "N"): "N .. NotRelatedToSecurityHalt",
        ("FIX.5.0", "329", "Y"): "Y .. RelatedToSecurityHalt",
        ("FIX.5.0", "334", "1"): "1 .. Cancel",
        ("FIX.5.0", "334", "2"): "2 .. Error",
        ("FIX.5.0", "334", "3"): "3 .. Correction",
        ("FIX.5.0", "336", "1"): "1 .. Day",
        ("FIX.5.0", "336", "2"): "2 .. HalfDay",
        ("FIX.5.0", "336", "3"): "3 .. Morning",
        ("FIX.5.0", "336", "4"): "4 .. Afternoon",
        ("FIX.5.0", "336", "5"): "5 .. Evening",
        ("FIX.5.0", "336", "6"): "6 .. AfterHours",
        ("FIX.5.0", "336", "7"): "7 .. Holiday",
        ("FIX.5.0", "338", "1"): "1 .. Electronic",
        ("FIX.5.0", "338", "2"): "2 .. OpenOutcry",
        ("FIX.5.0", "338", "3"): "3 .. TwoParty",
        ("FIX.5.0", "338", "4"): "4 .. Voice",
        ("FIX.5.0", "339", "1"): "1 .. Testing",
        ("FIX.5.0", "339", "2"): "2 .. Simulated",
        ("FIX.5.0", "339", "3"): "3 .. Production",
        ("FIX.5.0", "340", "0"): "0 .. Unknown",
        ("FIX.5.0", "340", "1"): "1 .. Halted",
        ("FIX.5.0", "340", "2"): "2 .. Open",
        ("FIX.5.0", "340", "3"): "3 .. Closed",
        ("FIX.5.0", "340", "4"): "4 .. PreOpen",
        ("FIX.5.0", "340", "5"): "5 .. PreClose",
        ("FIX.5.0", "340", "6"): "6 .. RequestRejected",
        ("FIX.5.0", "368", "1"): "1 .. UnknownSymbol",
        ("FIX.5.0", "368", "2"): "2 .. Exchange",
        ("FIX.5.0", "368", "3"): "3 .. QuoteRequestExceedsLimit",
        ("FIX.5.0", "368", "4"): "4 .. TooLateToEnter",
        ("FIX.5.0", "368", "5"): "5 .. UnknownQuote",
        ("FIX.5.0", "368", "6"): "6 .. DuplicateQuote",
        ("FIX.5.0", "368", "7"): "7 .. InvalidBid",
        ("FIX.5.0", "368", "8"): "8 .. InvalidPrice",
        ("FIX.5.0", "368", "9"): "9 .. NotAuthorizedToQuoteSecurity",
        ("FIX.5.0", "368", "10"): "10 .. PriceExceedsCurrentPriceBand",
        ("FIX.5.0", "368", "11"): "11 .. QuoteLocked",
        ("FIX.5.0", "368", "12"): "12 .. InvalidOrUnknownSecurityIssuer",
        ("FIX.5.0", "368", "13"): "13 .. InvalidOrUnknownIssuerOfUnderlyingSecurity",
        ("FIX.5.0", "368", "14"): "14 .. NotionalValueExceedsThreshold",
        ("FIX.5.0", "368", "16"): "16 .. ReferencePriceNotAvailable",
        ("FIX.5.0", "368", "17"): "17 .. InsufficientCreditLimit",
        ("FIX.5.0", "368", "18"): "18 .. ExceededClipSizeLimit",
        ("FIX.5.0", "368", "19"): "19 .. ExceededMaxNotionalOrderAmt",
        ("FIX.5.0", "368", "20"): "20 .. ExceededDV01PV01Limit",
        ("FIX.5.0", "368", "21"): "21 .. ExceededCS01Limit",
        ("FIX.5.0", "368", "99"): "99 .. Other",
        ("FIX.5.0", "372", "0"): "0 .. Heartbeat",
        ("FIX.5.0", "372", "1"): "1 .. TestRequest",
        ("FIX.5.0", "372", "2"): "2 .. ResendRequest",
        ("FIX.5.0", "372", "3"): "3 .. Reject",
        ("FIX.5.0", "372", "4"): "4 .. SequenceReset",
        ("FIX.5.0", "372", "5"): "5 .. Logout",
        ("FIX.5.0", "372", "6"): "6 .. IOI",
        ("FIX.5.0", "372", "7"): "7 .. Advertisement",
        ("FIX.5.0", "372", "8"): "8 .. ExecutionReport",
        ("FIX.5.0", "372", "9"): "9 .. OrderCancelReject",
        ("FIX.5.0", "372", "A"): "A .. Logon",
        ("FIX.5.0", "372", "B"): "B .. News",
        ("FIX.5.0", "372", "C"): "C .. Email",
        ("FIX.5.0", "372", "D"): "D .. NewOrderSingle",
        ("FIX.5.0", "372", "E"): "E .. NewOrderList",
        ("FIX.5.0", "372", "F"): "F .. OrderCancelRequest",
        ("FIX.5.0", "372", "G"): "G .. OrderCancelReplaceRequest",
        ("FIX.5.0", "372", "H"): "H .. OrderStatusRequest",
        ("FIX.5.0", "372", "J"): "J .. AllocationInstruction",
        ("FIX.5.0", "372", "K"): "K .. ListCancelRequest",
        ("FIX.5.0", "372", "L"): "L .. ListExecute",
        ("FIX.5.0", "372", "M"): "M .. ListStatusRequest",
        ("FIX.5.0", "372", "N"): "N .. ListStatus",
        ("FIX.5.0", "372", "P"): "P .. AllocationInstructionAck",
        ("FIX.5.0", "372", "Q"): "Q .. DontKnowTrade",
        ("FIX.5.0", "372", "R"): "R .. QuoteRequest",
        ("FIX.5.0", "372", "S"): "S .. Quote",
        ("FIX.5.0", "372", "T"): "T .. SettlementInstructions",
        ("FIX.5.0", "372", "V"): "V .. MarketDataRequest",
        ("FIX.5.0", "372", "W"): "W .. MarketDataSnapshotFullRefresh",
        ("FIX.5.0", "372", "X"): "X .. MarketDataIncrementalRefresh",
        ("FIX.5.0", "372", "Y"): "Y .. MarketDataRequestReject",
        ("FIX.5.0", "372", "Z"): "Z .. QuoteCancel",
        ("FIX.5.0", "372", "a"): "a .. QuoteStatusRequest",
        ("FIX.5.0", "372", "b"): "b .. MassQuoteAck",
        ("FIX.5.0", "372", "c"): "c .. SecurityDefinitionRequest",
        ("FIX.5.0", "372", "d"): "d .. SecurityDefinition",
        ("FIX.5.0", "372", "e"): "e .. SecurityStatusRequest",
        ("FIX.5.0", "372", "f"): "f .. SecurityStatus",
        ("FIX.5.0", "372", "g"): "g .. TradingSessionStatusRequest",
        ("FIX.5.0", "372", "h"): "h .. TradingSessionStatus",
        ("FIX.5.0", "372", "i"): "i .. MassQuote",
        ("FIX.5.0", "372", "j"): "j .. BusinessMessageReject",
        ("FIX.5.0", "372", "k"): "k .. BidRequest",
        ("FIX.5.0", "372", "l"): "l .. BidResponse",
        ("FIX.5.0", "372", "m"): "m .. ListStrikePrice",
        ("FIX.5.0", "372", "n"): "n .. XMLnonFIX",
        ("FIX.5.0", "372", "o"): "o .. RegistrationInstructions",
        ("FIX.5.0", "372", "p"): "p .. RegistrationInstructionsResponse",
        ("FIX.5.0", "372", "q"): "q .. OrderMassCancelRequest",
        ("FIX.5.0", "372", "r"): "r .. OrderMassCancelReport",
        ("FIX.5.0", "372", "s"): "s .. NewOrderCross",
        ("FIX.5.0", "372", "t"): "t .. CrossOrderCancelReplaceRequest",
        ("FIX.5.0", "372", "u"): "u .. CrossOrderCancelRequest",
        ("FIX.5.0", "372", "v"): "v .. SecurityTypeRequest",
        ("FIX.5.0", "372", "w"): "w .. SecurityTypes",
        ("FIX.5.0", "372", "x"): "x .. SecurityListRequest",
        ("FIX.5.0", "372", "y"): "y .. SecurityList",
        ("FIX.5.0", "372", "z"): "z .. DerivativeSecurityListRequest",
        ("FIX.5.0", "372", "AA"): "AA .. DerivativeSecurityList",
        ("FIX.5.0", "372", "AB"): "AB .. NewOrderMultileg",
        ("FIX.5.0", "372", "AC"): "AC .. MultilegOrderCancelReplace",
        ("FIX.5.0", "372", "AD"): "AD .. TradeCaptureReportRequest",
        ("FIX.5.0", "372", "AE"): "AE .. TradeCaptureReport",
        ("FIX.5.0", "372", "AF"): "AF .. OrderMassStatusRequest",
        ("FIX.5.0", "372", "AG"): "AG .. QuoteRequestReject",
        ("FIX.5.0", "372", "AH"): "AH .. RFQRequest",
        ("FIX.5.0", "372", "AI"): "AI .. QuoteStatusReport",
        ("FIX.5.0", "372", "AJ"): "AJ .. QuoteResponse",
        ("FIX.5.0", "372", "AK"): "AK .. Confirmation",
        ("FIX.5.0", "372", "AL"): "AL .. PositionMaintenanceRequest",
        ("FIX.5.0", "372", "AM"): "AM .. PositionMaintenanceReport",
        ("FIX.5.0", "372", "AN"): "AN .. RequestForPositions",
        ("FIX.5.0", "372", "AO"): "AO .. RequestForPositionsAck",
        ("FIX.5.0", "372", "AP"): "AP .. PositionReport",
        ("FIX.5.0", "372", "AQ"): "AQ .. TradeCaptureReportRequestAck",
        ("FIX.5.0", "372", "AR"): "AR .. TradeCaptureReportAck",
        ("FIX.5.0", "372", "AS"): "AS .. AllocationReport",
        ("FIX.5.0", "372", "AT"): "AT .. AllocationReportAck",
        ("FIX.5.0", "372", "AU"): "AU .. ConfirmationAck",
        ("FIX.5.0", "372", "AV"): "AV .. SettlementInstructionRequest",
        ("FIX.5.0", "372", "AW"): "AW .. AssignmentReport",
        ("FIX.5.0", "372", "AX"): "AX .. CollateralRequest",
        ("FIX.5.0", "372", "AY"): "AY .. CollateralAssignment",
        ("FIX.5.0", "372", "AZ"): "AZ .. CollateralResponse",
        ("FIX.5.0", "372", "BA"): "BA .. CollateralReport",
        ("FIX.5.0", "372", "BB"): "BB .. CollateralInquiry",
        ("FIX.5.0", "372", "BC"): "BC .. NetworkCounterpartySystemStatusRequest",
        ("FIX.5.0", "372", "BD"): "BD .. NetworkCounterpartySystemStatusResponse",
        ("FIX.5.0", "372", "BE"): "BE .. UserRequest",
        ("FIX.5.0", "372", "BF"): "BF .. UserResponse",
        ("FIX.5.0", "372", "BG"): "BG .. CollateralInquiryAck",
        ("FIX.5.0", "372", "BH"): "BH .. ConfirmationRequest",
        ("FIX.5.0", "372", "BO"): "BO .. ContraryIntentionReport",
        ("FIX.5.0", "372", "BP"): "BP .. SecurityDefinitionUpdateReport",
        ("FIX.5.0", "372", "BK"): "BK .. SecurityListUpdateReport",
        ("FIX.5.0", "372", "BL"): "BL .. AdjustedPositionReport",
        ("FIX.5.0", "372", "BM"): "BM .. AllocationInstructionAlert",
        ("FIX.5.0", "372", "BN"): "BN .. ExecutionAck",
        ("FIX.5.0", "372", "BJ"): "BJ .. TradingSessionList",
        ("FIX.5.0", "372", "BI"): "BI .. TradingSessionListRequest",
        ("FIX.5.0", "372", "BQ"): "BQ .. SettlementObligationReport",
        ("FIX.5.0", "372", "BR"): "BR .. DerivativeSecurityListUpdateReport",
        ("FIX.5.0", "372", "BS"): "BS .. TradingSessionListUpdateReport",
        ("FIX.5.0", "372", "BT"): "BT .. MarketDefinitionRequest",
        ("FIX.5.0", "372", "BU"): "BU .. MarketDefinition",
        ("FIX.5.0", "372", "BV"): "BV .. MarketDefinitionUpdateReport",
        ("FIX.5.0", "372", "BW"): "BW .. ApplicationMessageRequest",
        ("FIX.5.0", "372", "BX"): "BX .. ApplicationMessageRequestAck",
        ("FIX.5.0", "372", "BY"): "BY .. ApplicationMessageReport",
        ("FIX.5.0", "372", "BZ"): "BZ .. OrderMassActionReport",
        ("FIX.5.0", "372", "CA"): "CA .. OrderMassActionRequest",
        ("FIX.5.0", "372", "CB"): "CB .. UserNotification",
        ("FIX.5.0", "372", "CC"): "CC .. StreamAssignmentRequest",
        ("FIX.5.0", "372", "CD"): "CD .. StreamAssignmentReport",
        ("FIX.5.0", "372", "CE"): "CE .. StreamAssignmentReportACK",
        ("FIX.5.0", "372", "CF"): "CF .. PartyDetailsListRequest",
        ("FIX.5.0", "372", "CG"): "CG .. PartyDetailsListReport",
        ("FIX.5.0", "372", "CH"): "CH .. MarginRequirementInquiry",
        ("FIX.5.0", "372", "CI"): "CI .. MarginRequirementInquiryAck",
        ("FIX.5.0", "372", "CJ"): "CJ .. MarginRequirementReport",
        ("FIX.5.0", "372", "CK"): "CK .. PartyDetailsListUpdateReport",
        ("FIX.5.0", "372", "CL"): "CL .. PartyRiskLimitsRequest",
        ("FIX.5.0", "372", "CM"): "CM .. PartyRiskLimitsReport",
        ("FIX.5.0", "372", "CN"): "CN .. SecurityMassStatusRequest",
        ("FIX.5.0", "372", "CO"): "CO .. SecurityMassStatus",
        ("FIX.5.0", "372", "CQ"): "CQ .. AccountSummaryReport",
        ("FIX.5.0", "372", "CR"): "CR .. PartyRiskLimitsUpdateReport",
        ("FIX.5.0", "372", "CS"): "CS .. PartyRiskLimitsDefinitionRequest",
        ("FIX.5.0", "372", "CT"): "CT .. PartyRiskLimitsDefinitionRequestAck",
        ("FIX.5.0", "372", "CU"): "CU .. PartyEntitlementsRequest",
        ("FIX.5.0", "372", "CV"): "CV .. PartyEntitlementsReport",
        ("FIX.5.0", "372", "CW"): "CW .. QuoteAck",
        ("FIX.5.0", "372", "CX"): "CX .. PartyDetailsDefinitionRequest",
        ("FIX.5.0", "372", "CY"): "CY .. PartyDetailsDefinitionRequestAck",
        ("FIX.5.0", "372", "CZ"): "CZ .. PartyEntitlementsUpdateReport",
        ("FIX.5.0", "372", "DA"): "DA .. PartyEntitlementsDefinitionRequest",
        ("FIX.5.0", "372", "DB"): "DB .. PartyEntitlementsDefinitionRequestAck",
        ("FIX.5.0", "372", "DC"): "DC .. TradeMatchReport",
        ("FIX.5.0", "372", "DD"): "DD .. TradeMatchReportAck",
        ("FIX.5.0", "372", "DE"): "DE .. PartyRiskLimitsReportAck",
        ("FIX.5.0", "372", "DF"): "DF .. PartyRiskLimitCheckRequest",
        ("FIX.5.0", "372", "DG"): "DG .. PartyRiskLimitCheckRequestAck",
        ("FIX.5.0", "372", "DH"): "DH .. PartyActionRequest",
        ("FIX.5.0", "372", "DI"): "DI .. PartyActionReport",
        ("FIX.5.0", "372", "DJ"): "DJ .. MassOrder",
        ("FIX.5.0", "372", "DK"): "DK .. MassOrderAck",
        ("FIX.5.0", "372", "DL"): "DL .. PositionTransferInstruction",
        ("FIX.5.0", "372", "DM"): "DM .. PositionTransferInstructionAck",
        ("FIX.5.0", "372", "DN"): "DN .. PositionTransferReport",
        ("FIX.5.0", "372", "DO"): "DO .. MarketDataStatisticsRequest",
        ("FIX.5.0", "372", "DP"): "DP .. MarketDataStatisticsReport",
        ("FIX.5.0", "372", "DQ"): "DQ .. CollateralReportAck",
        ("FIX.5.0", "372", "DR"): "DR .. MarketDataReport",
        ("FIX.5.0", "372", "DS"): "DS .. CrossRequest",
        ("FIX.5.0", "372", "DT"): "DT .. CrossRequestAck",
        ("FIX.5.0", "372", "DU"): "DU .. AllocationInstructionAlertRequest",
        ("FIX.5.0", "373", "0"): "0 .. InvalidTagNumber",
        ("FIX.5.0", "373", "1"): "1 .. RequiredTagMissing",
        ("FIX.5.0", "373", "2"): "2 .. TagNotDefinedForThisMessageType",
        ("FIX.5.0", "373", "3"): "3 .. UndefinedTag",
        ("FIX.5.0", "373", "4"): "4 .. TagSpecifiedWithoutAValue",
        ("FIX.5.0", "373", "5"): "5 .. ValueIsIncorrect",
        ("FIX.5.0", "373", "6"): "6 .. IncorrectDataFormatForValue",
        ("FIX.5.0", "373", "7"): "7 .. DecryptionProblem",
        ("FIX.5.0", "373", "8"): "8 .. SignatureProblem",
        ("FIX.5.0", "373", "9"): "9 .. CompIDProblem",
        ("FIX.5.0", "373", "10"): "10 .. SendingTimeAccuracyProblem",
        ("FIX.5.0", "373", "11"): "11 .. InvalidMsgType",
        ("FIX.5.0", "373", "12"): "12 .. XMLValidationError",
        ("FIX.5.0", "373", "13"): "13 .. TagAppearsMoreThanOnce",
        ("FIX.5.0", "373", "14"): "14 .. TagSpecifiedOutOfRequiredOrder",
        ("FIX.5.0", "373", "15"): "15 .. RepeatingGroupFieldsOutOfOrder",
        ("FIX.5.0", "373", "16"): "16 .. IncorrectNumInGroupCountForRepeatingGroup",
        ("FIX.5.0", "373", "17"): "17 .. Non",
        ("FIX.5.0", "373", "18"): "18 .. Invalid",
        ("FIX.5.0", "373", "99"): "99 .. Other",
        ("FIX.5.0", "374", "C"): "C .. Cancel",
        ("FIX.5.0", "374", "N"): "N .. New",
        ("FIX.5.0", "377", "N"): "N .. WasNotSolicited",
        ("FIX.5.0", "377", "Y"): "Y .. WasSolicited",
        ("FIX.5.0", "378", "0"): "0 .. GTCorporateAction",
        ("FIX.5.0", "378", "1"): "1 .. GTRenewal",
        ("FIX.5.0", "378", "2"): "2 .. VerbalChange",
        ("FIX.5.0", "378", "3"): "3 .. RepricingOfOrder",
        ("FIX.5.0", "378", "4"): "4 .. BrokerOption",
        ("FIX.5.0", "378", "5"): "5 .. PartialDeclineOfOrderQty",
        ("FIX.5.0", "378", "6"): "6 .. CancelOnTradingHalt",
        ("FIX.5.0", "378", "7"): "7 .. CancelOnSystemFailure",
        ("FIX.5.0", "378", "8"): "8 .. Market",
        ("FIX.5.0", "378", "9"): "9 .. Canceled",
        ("FIX.5.0", "378", "10"): "10 .. WarehouseRecap",
        ("FIX.5.0", "378", "11"): "11 .. PegRefresh",
        ("FIX.5.0", "378", "12"): "12 .. CancelOnConnectionLoss",
        ("FIX.5.0", "378", "13"): "13 .. CancelOnLogout",
        ("FIX.5.0", "378", "14"): "14 .. AssignTimePriority",
        ("FIX.5.0", "378", "15"): "15 .. CancelledForTradePriceViolation",
        ("FIX.5.0", "378", "16"): "16 .. CancelledForCrossImbalance",
        ("FIX.5.0", "378", "99"): "99 .. Other",
        ("FIX.5.0", "380", "0"): "0 .. Other",
        ("FIX.5.0", "380", "1"): "1 .. UnknownID",
        ("FIX.5.0", "380", "2"): "2 .. UnknownSecurity",
        ("FIX.5.0", "380", "3"): "3 .. UnsupportedMessageType",
        ("FIX.5.0", "380", "4"): "4 .. ApplicationNotAvailable",
        ("FIX.5.0", "380", "5"): "5 .. ConditionallyRequiredFieldMissing",
        ("FIX.5.0", "380", "6"): "6 .. NotAuthorized",
        ("FIX.5.0", "380", "7"): "7 .. DeliverToFirmNotAvailableAtThisTime",
        ("FIX.5.0", "380", "8"): "8 .. ThrottleLimitExceeded",
        ("FIX.5.0", "380", "9"): "9 .. ThrottleLimitExceededSessionDisconnected",
        ("FIX.5.0", "380", "10"): "10 .. ThrottledMessagesRejectedOnRequest",
        ("FIX.5.0", "380", "18"): "18 .. InvalidPriceIncrement",
        ("FIX.5.0", "385", "R"): "R .. Receive",
        ("FIX.5.0", "385", "S"): "S .. Send",
        ("FIX.5.0", "388", "0"): "0 .. RelatedToDisplayedPrice",
        ("FIX.5.0", "388", "1"): "1 .. RelatedToMarketPrice",
        ("FIX.5.0", "388", "2"): "2 .. RelatedToPrimaryPrice",
        ("FIX.5.0", "388", "3"): "3 .. RelatedToLocalPrimaryPrice",
        ("FIX.5.0", "388", "4"): "4 .. RelatedToMidpointPrice",
        ("FIX.5.0", "388", "5"): "5 .. RelatedToLastTradePrice",
        ("FIX.5.0", "388", "6"): "6 .. RelatedToVWAP",
        ("FIX.5.0", "388", "7"): "7 .. AveragePriceGuarantee",
        ("FIX.5.0", "394", "1"): "1 .. NonDisclosed",
        ("FIX.5.0", "394", "2"): "2 .. Disclosed",
        ("FIX.5.0", "394", "3"): "3 .. NoBiddingProcess",
        ("FIX.5.0", "399", "1"): "1 .. Sector",
        ("FIX.5.0", "399", "2"): "2 .. Country",
        ("FIX.5.0", "399", "3"): "3 .. Index",
        ("FIX.5.0", "401", "1"): "1 .. SideValue1",
        ("FIX.5.0", "401", "2"): "2 .. SideValue2",
        ("FIX.5.0", "409", "1"): "1 .. FiveDayMovingAverage",
        ("FIX.5.0", "409", "2"): "2 .. TwentyDayMovingAverage",
        ("FIX.5.0", "409", "3"): "3 .. NormalMarketSize",
        ("FIX.5.0", "409", "4"): "4 .. Other",
        ("FIX.5.0", "411", "N"): "N .. False",
        ("FIX.5.0", "411", "Y"): "Y .. True",
        ("FIX.5.0", "414", "1"): "1 .. BuySideRequests",
        ("FIX.5.0", "414", "2"): "2 .. SellSideSends",
        ("FIX.5.0", "414", "3"): "3 .. RealTimeExecutionReports",
        ("FIX.5.0", "416", "1"): "1 .. Net",
        ("FIX.5.0", "416", "2"): "2 .. Gross",
        ("FIX.5.0", "418", "A"): "A .. Agency",
        ("FIX.5.0", "418", "G"): "G .. VWAPGuarantee",
        ("FIX.5.0", "418", "J"): "J .. GuaranteedClose",
        ("FIX.5.0", "418", "R"): "R .. RiskTrade",
        ("FIX.5.0", "419", "2"): "2 .. ClosingPriceAtMorningSession",
        ("FIX.5.0", "419", "3"): "3 .. ClosingPrice",
        ("FIX.5.0", "419", "4"): "4 .. CurrentPrice",
        ("FIX.5.0", "419", "5"): "5 .. SQ",
        ("FIX.5.0", "419", "6"): "6 .. VWAPThroughADay",
        ("FIX.5.0", "419", "7"): "7 .. VWAPThroughAMorningSession",
        ("FIX.5.0", "419", "8"): "8 .. VWAPThroughAnAfternoonSession",
        ("FIX.5.0", "419", "9"): "9 .. VWAPThroughADayExcept",
        ("FIX.5.0", "419", "A"): "A .. VWAPThroughAMorningSessionExcept",
        ("FIX.5.0", "419", "B"): "B .. VWAPThroughAnAfternoonSessionExcept",
        ("FIX.5.0", "419", "C"): "C .. Strike",
        ("FIX.5.0", "419", "D"): "D .. Open",
        ("FIX.5.0", "419", "Z"): "Z .. Others",
        ("FIX.5.0", "423", "1"): "1 .. Percentage",
        ("FIX.5.0", "423", "2"): "2 .. PerUnit",
        ("FIX.5.0", "423", "3"): "3 .. FixedAmount",
        ("FIX.5.0", "423", "4"): "4 .. Discount",
        ("FIX.5.0", "423", "5"): "5 .. Premium",
        ("FIX.5.0", "423", "6"): "6 .. Spread",
        ("FIX.5.0", "423", "7"): "7 .. TEDPrice",
        ("FIX.5.0", "423", "8"): "8 .. TEDYield",
        ("FIX.5.0", "423", "9"): "9 .. Yield",
        ("FIX.5.0", "423", "10"): "10 .. FixedCabinetTradePrice",
        ("FIX.5.0", "423", "11"): "11 .. VariableCabinetTradePrice",
        ("FIX.5.0", "423", "12"): "12 .. PriceSpread",
        ("FIX.5.0", "423", "13"): "13 .. ProductTicksInHalves",
        ("FIX.5.0", "423", "14"): "14 .. ProductTicksInFourths",
        ("FIX.5.0", "423", "15"): "15 .. ProductTicksInEighths",
        ("FIX.5.0", "423", "16"): "16 .. ProductTicksInSixteenths",
        ("FIX.5.0", "423", "17"): "17 .. ProductTicksInThirtySeconds",
        ("FIX.5.0", "423", "18"): "18 .. ProductTicksInSixtyFourths",
        ("FIX.5.0", "423", "19"): "19 .. ProductTicksInOneTwentyEighths",
        ("FIX.5.0", "423", "20"): "20 .. NormalRateRepresentation",
        ("FIX.5.0", "423", "21"): "21 .. InverseRateRepresentation",
        ("FIX.5.0", "423", "22"): "22 .. BasisPoints",
        ("FIX.5.0", "423", "23"): "23 .. UpfrontPoints",
        ("FIX.5.0", "423", "24"): "24 .. InterestRate",
        ("FIX.5.0", "423", "25"): "25 .. PercentageNotional",
        ("FIX.5.0", "427", "0"): "0 .. BookOutAllTradesOnDayOfExecution",
        ("FIX.5.0", "427", "1"): "1 .. AccumulateUntilFilledOrExpired",
        ("FIX.5.0", "427", "2"): "2 .. AccumulateUntilVerballyNotifiedOtherwise",
        ("FIX.5.0", "429", "1"): "1 .. Ack",
        ("FIX.5.0", "429", "2"): "2 .. Response",
        ("FIX.5.0", "429", "3"): "3 .. Timed",
        ("FIX.5.0", "429", "4"): "4 .. ExecStarted",
        ("FIX.5.0", "429", "5"): "5 .. AllDone",
        ("FIX.5.0", "429", "6"): "6 .. Alert",
        ("FIX.5.0", "430", "1"): "1 .. Net",
        ("FIX.5.0", "430", "2"): "2 .. Gross",
        ("FIX.5.0", "431", "1"): "1 .. InBiddingProcess",
        ("FIX.5.0", "431", "2"): "2 .. ReceivedForExecution",
        ("FIX.5.0", "431", "3"): "3 .. Executing",
        ("FIX.5.0", "431", "4"): "4 .. Cancelling",
        ("FIX.5.0", "431", "5"): "5 .. Alert",
        ("FIX.5.0", "431", "6"): "6 .. AllDone",
        ("FIX.5.0", "431", "7"): "7 .. Reject",
        ("FIX.5.0", "433", "1"): "1 .. Immediate",
        ("FIX.5.0", "433", "2"): "2 .. WaitForInstruction",
        ("FIX.5.0", "433", "3"): "3 .. SellDriven",
        ("FIX.5.0", "433", "4"): "4 .. BuyDrivenCashTopUp",
        ("FIX.5.0", "433", "5"): "5 .. BuyDrivenCashWithdraw",
        ("FIX.5.0", "434", "1"): "1 .. OrderCancelRequest",
        ("FIX.5.0", "434", "2"): "2 .. OrderCancelReplaceRequest",
        ("FIX.5.0", "442", "1"): "1 .. SingleSecurity",
        ("FIX.5.0", "442", "2"): "2 .. IndividualLegOfAMultiLegSecurity",
        ("FIX.5.0", "442", "3"): "3 .. MultiLegSecurity",
        ("FIX.5.0", "447", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "447", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "447", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "447", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "447", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "447", "J"): "J .. TaxID",
        ("FIX.5.0", "447", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "447", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "447", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "447", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "447", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "447", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "447", "B"): "B .. BIC",
        ("FIX.5.0", "447", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "447", "D"): "D .. Proprietary",
        ("FIX.5.0", "447", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "447", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "447", "G"): "G .. MIC",
        ("FIX.5.0", "447", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "447", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "447", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "447", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "447", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "447", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "447", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "447", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "452", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "452", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "452", "3"): "3 .. ClientID",
        ("FIX.5.0", "452", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "452", "5"): "5 .. InvestorID",
        ("FIX.5.0", "452", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "452", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "452", "8"): "8 .. Locate",
        ("FIX.5.0", "452", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "452", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "452", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "452", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "452", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "452", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "452", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "452", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "452", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "452", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "452", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "452", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "452", "22"): "22 .. Exchange",
        ("FIX.5.0", "452", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "452", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "452", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "452", "27"): "27 .. Buyer",
        ("FIX.5.0", "452", "28"): "28 .. Custodian",
        ("FIX.5.0", "452", "29"): "29 .. Intermediary",
        ("FIX.5.0", "452", "30"): "30 .. Agent",
        ("FIX.5.0", "452", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "452", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "452", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "452", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "452", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "452", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "452", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "452", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "452", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "452", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "452", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "452", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "452", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "452", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "452", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "452", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "452", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "452", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "452", "49"): "49 .. AssetManager",
        ("FIX.5.0", "452", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "452", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "452", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "452", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "452", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "452", "55"): "55 .. SessionID",
        ("FIX.5.0", "452", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "452", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "452", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "452", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "452", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "452", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "452", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "452", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "452", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "452", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "452", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "452", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "452", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "452", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "452", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "452", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "452", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "452", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "452", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "452", "75"): "75 .. LocationID",
        ("FIX.5.0", "452", "76"): "76 .. DeskID",
        ("FIX.5.0", "452", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "452", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "452", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "452", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "452", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "452", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "452", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "452", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "452", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "452", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "452", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "452", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "452", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "452", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "452", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "452", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "452", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "452", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "452", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "452", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "452", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "452", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "452", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "452", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "452", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "452", "102"): "102 .. DataRepository",
        ("FIX.5.0", "452", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "452", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "452", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "452", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "452", "107"): "107 .. Correspondent",
        ("FIX.5.0", "452", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "452", "110"): "110 .. Borrower",
        ("FIX.5.0", "452", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "452", "112"): "112 .. Guarantor",
        ("FIX.5.0", "452", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "452", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "452", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "452", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "452", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "452", "118"): "118 .. Operator",
        ("FIX.5.0", "452", "119"): "119 .. CSD",
        ("FIX.5.0", "452", "120"): "120 .. ICSD",
        ("FIX.5.0", "452", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "452", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "452", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "456", "1"): "1 .. CUSIP",
        ("FIX.5.0", "456", "2"): "2 .. SEDOL",
        ("FIX.5.0", "456", "3"): "3 .. QUIK",
        ("FIX.5.0", "456", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "456", "5"): "5 .. RICCode",
        ("FIX.5.0", "456", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "456", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "456", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "456", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "456", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "456", "B"): "B .. Wertpapier",
        ("FIX.5.0", "456", "C"): "C .. Dutch",
        ("FIX.5.0", "456", "D"): "D .. Valoren",
        ("FIX.5.0", "456", "E"): "E .. Sicovam",
        ("FIX.5.0", "456", "F"): "F .. Belgian",
        ("FIX.5.0", "456", "G"): "G .. Common",
        ("FIX.5.0", "456", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "456", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "456", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "456", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "456", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "456", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "456", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "456", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "456", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "456", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "456", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "456", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "456", "U"): "U .. Synthetic",
        ("FIX.5.0", "456", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "456", "W"): "W .. IndexName",
        ("FIX.5.0", "459", "1"): "1 .. CUSIP",
        ("FIX.5.0", "459", "2"): "2 .. SEDOL",
        ("FIX.5.0", "459", "3"): "3 .. QUIK",
        ("FIX.5.0", "459", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "459", "5"): "5 .. RICCode",
        ("FIX.5.0", "459", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "459", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "459", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "459", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "459", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "459", "B"): "B .. Wertpapier",
        ("FIX.5.0", "459", "C"): "C .. Dutch",
        ("FIX.5.0", "459", "D"): "D .. Valoren",
        ("FIX.5.0", "459", "E"): "E .. Sicovam",
        ("FIX.5.0", "459", "F"): "F .. Belgian",
        ("FIX.5.0", "459", "G"): "G .. Common",
        ("FIX.5.0", "459", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "459", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "459", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "459", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "459", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "459", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "459", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "459", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "459", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "459", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "459", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "459", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "459", "U"): "U .. Synthetic",
        ("FIX.5.0", "459", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "459", "W"): "W .. IndexName",
        ("FIX.5.0", "460", "1"): "1 .. AGENCY",
        ("FIX.5.0", "460", "2"): "2 .. COMMODITY",
        ("FIX.5.0", "460", "3"): "3 .. CORPORATE",
        ("FIX.5.0", "460", "4"): "4 .. CURRENCY",
        ("FIX.5.0", "460", "5"): "5 .. EQUITY",
        ("FIX.5.0", "460", "6"): "6 .. GOVERNMENT",
        ("FIX.5.0", "460", "7"): "7 .. INDEX",
        ("FIX.5.0", "460", "8"): "8 .. LOAN",
        ("FIX.5.0", "460", "9"): "9 .. MONEYMARKET",
        ("FIX.5.0", "460", "10"): "10 .. MORTGAGE",
        ("FIX.5.0", "460", "11"): "11 .. MUNICIPAL",
        ("FIX.5.0", "460", "12"): "12 .. OTHER",
        ("FIX.5.0", "460", "13"): "13 .. FINANCING",
        ("FIX.5.0", "462", "1"): "1 .. AGENCY",
        ("FIX.5.0", "462", "2"): "2 .. COMMODITY",
        ("FIX.5.0", "462", "3"): "3 .. CORPORATE",
        ("FIX.5.0", "462", "4"): "4 .. CURRENCY",
        ("FIX.5.0", "462", "5"): "5 .. EQUITY",
        ("FIX.5.0", "462", "6"): "6 .. GOVERNMENT",
        ("FIX.5.0", "462", "7"): "7 .. INDEX",
        ("FIX.5.0", "462", "8"): "8 .. LOAN",
        ("FIX.5.0", "462", "9"): "9 .. MONEYMARKET",
        ("FIX.5.0", "462", "10"): "10 .. MORTGAGE",
        ("FIX.5.0", "462", "11"): "11 .. MUNICIPAL",
        ("FIX.5.0", "462", "12"): "12 .. OTHER",
        ("FIX.5.0", "462", "13"): "13 .. FINANCING",
        ("FIX.5.0", "464", "N"): "N .. False",
        ("FIX.5.0", "464", "Y"): "Y .. True",
        ("FIX.5.0", "468", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "468", "1"): "1 .. RoundDown",
        ("FIX.5.0", "468", "2"): "2 .. RoundUp",
        ("FIX.5.0", "477", "1"): "1 .. CREST",
        ("FIX.5.0", "477", "2"): "2 .. NSCC",
        ("FIX.5.0", "477", "3"): "3 .. Euroclear",
        ("FIX.5.0", "477", "4"): "4 .. Clearstream",
        ("FIX.5.0", "477", "5"): "5 .. Cheque",
        ("FIX.5.0", "477", "6"): "6 .. TelegraphicTransfer",
        ("FIX.5.0", "477", "7"): "7 .. FedWire",
        ("FIX.5.0", "477", "8"): "8 .. DirectCredit",
        ("FIX.5.0", "477", "9"): "9 .. ACHCredit",
        ("FIX.5.0", "477", "10"): "10 .. BPAY",
        ("FIX.5.0", "477", "11"): "11 .. HighValueClearingSystemHVACS",
        ("FIX.5.0", "477", "12"): "12 .. ReinvestInFund",
        ("FIX.5.0", "480", "Y"): "Y .. Yes",
        ("FIX.5.0", "480", "N"): "N .. NoExecutionOnly",
        ("FIX.5.0", "480", "M"): "M .. NoWaiverAgreement",
        ("FIX.5.0", "480", "O"): "O .. NoInstitutional",
        ("FIX.5.0", "481", "Y"): "Y .. Passed",
        ("FIX.5.0", "481", "N"): "N .. NotChecked",
        ("FIX.5.0", "481", "1"): "1 .. ExemptBelowLimit",
        ("FIX.5.0", "481", "2"): "2 .. ExemptMoneyType",
        ("FIX.5.0", "481", "3"): "3 .. ExemptAuthorised",
        ("FIX.5.0", "484", "B"): "B .. BidPrice",
        ("FIX.5.0", "484", "C"): "C .. CreationPrice",
        ("FIX.5.0", "484", "D"): "D .. CreationPricePlusAdjustmentPercent",
        ("FIX.5.0", "484", "E"): "E .. CreationPricePlusAdjustmentAmount",
        ("FIX.5.0", "484", "O"): "O .. OfferPrice",
        ("FIX.5.0", "484", "P"): "P .. OfferPriceMinusAdjustmentPercent",
        ("FIX.5.0", "484", "Q"): "Q .. OfferPriceMinusAdjustmentAmount",
        ("FIX.5.0", "484", "S"): "S .. SinglePrice",
        ("FIX.5.0", "487", "0"): "0 .. New",
        ("FIX.5.0", "487", "1"): "1 .. Cancel",
        ("FIX.5.0", "487", "2"): "2 .. Replace",
        ("FIX.5.0", "487", "3"): "3 .. Release",
        ("FIX.5.0", "487", "4"): "4 .. Reverse",
        ("FIX.5.0", "487", "5"): "5 .. CancelDueToBackOutOfTrade",
        ("FIX.5.0", "492", "1"): "1 .. CREST",
        ("FIX.5.0", "492", "2"): "2 .. NSCC",
        ("FIX.5.0", "492", "3"): "3 .. Euroclear",
        ("FIX.5.0", "492", "4"): "4 .. Clearstream",
        ("FIX.5.0", "492", "5"): "5 .. Cheque",
        ("FIX.5.0", "492", "6"): "6 .. TelegraphicTransfer",
        ("FIX.5.0", "492", "7"): "7 .. FedWire",
        ("FIX.5.0", "492", "8"): "8 .. DebitCard",
        ("FIX.5.0", "492", "9"): "9 .. DirectDebit",
        ("FIX.5.0", "492", "10"): "10 .. DirectCredit",
        ("FIX.5.0", "492", "11"): "11 .. CreditCard",
        ("FIX.5.0", "492", "12"): "12 .. ACHDebit",
        ("FIX.5.0", "492", "13"): "13 .. ACHCredit",
        ("FIX.5.0", "492", "14"): "14 .. BPAY",
        ("FIX.5.0", "492", "15"): "15 .. HighValueClearingSystem",
        ("FIX.5.0", "492", "16"): "16 .. CHIPS",
        ("FIX.5.0", "492", "17"): "17 .. SWIFT",
        ("FIX.5.0", "492", "18"): "18 .. CHAPS",
        ("FIX.5.0", "492", "19"): "19 .. SIC",
        ("FIX.5.0", "492", "20"): "20 .. euroSIC",
        ("FIX.5.0", "495", "0"): "0 .. None",
        ("FIX.5.0", "495", "1"): "1 .. MaxiISA",
        ("FIX.5.0", "495", "2"): "2 .. TESSA",
        ("FIX.5.0", "495", "3"): "3 .. MiniCashISA",
        ("FIX.5.0", "495", "4"): "4 .. MiniStocksAndSharesISA",
        ("FIX.5.0", "495", "5"): "5 .. MiniInsuranceISA",
        ("FIX.5.0", "495", "6"): "6 .. CurrentYearPayment",
        ("FIX.5.0", "495", "7"): "7 .. PriorYearPayment",
        ("FIX.5.0", "495", "8"): "8 .. AssetTransfer",
        ("FIX.5.0", "495", "9"): "9 .. EmployeePriorYear",
        ("FIX.5.0", "495", "10"): "10 .. EmployeeCurrentYear",
        ("FIX.5.0", "495", "11"): "11 .. EmployerPriorYear",
        ("FIX.5.0", "495", "12"): "12 .. EmployerCurrentYear",
        ("FIX.5.0", "495", "13"): "13 .. NonFundPrototypeIRA",
        ("FIX.5.0", "495", "14"): "14 .. NonFundQualifiedPlan",
        ("FIX.5.0", "495", "15"): "15 .. DefinedContributionPlan",
        ("FIX.5.0", "495", "16"): "16 .. IRA",
        ("FIX.5.0", "495", "17"): "17 .. IRARollover",
        ("FIX.5.0", "495", "18"): "18 .. KEOGH",
        ("FIX.5.0", "495", "19"): "19 .. ProfitSharingPlan",
        ("FIX.5.0", "495", "20"): "20 .. US401K",
        ("FIX.5.0", "495", "21"): "21 .. SelfDirectedIRA",
        ("FIX.5.0", "495", "22"): "22 .. US403b",
        ("FIX.5.0", "495", "23"): "23 .. US457",
        ("FIX.5.0", "495", "24"): "24 .. RothIRAPrototype",
        ("FIX.5.0", "495", "25"): "25 .. RothIRANonPrototype",
        ("FIX.5.0", "495", "26"): "26 .. RothConversionIRAPrototype",
        ("FIX.5.0", "495", "27"): "27 .. RothConversionIRANonPrototype",
        ("FIX.5.0", "495", "28"): "28 .. EducationIRAPrototype",
        ("FIX.5.0", "495", "29"): "29 .. EducationIRANonPrototype",
        ("FIX.5.0", "495", "999"): "999 .. Other",
        ("FIX.5.0", "497", "N"): "N .. No",
        ("FIX.5.0", "497", "Y"): "Y .. Yes",
        ("FIX.5.0", "506", "A"): "A .. Accepted",
        ("FIX.5.0", "506", "R"): "R .. Rejected",
        ("FIX.5.0", "506", "H"): "H .. Held",
        ("FIX.5.0", "506", "N"): "N .. Reminder",
        ("FIX.5.0", "507", "1"): "1 .. InvalidAccountType",
        ("FIX.5.0", "507", "2"): "2 .. InvalidTaxExemptType",
        ("FIX.5.0", "507", "3"): "3 .. InvalidOwnershipType",
        ("FIX.5.0", "507", "4"): "4 .. NoRegDetails",
        ("FIX.5.0", "507", "5"): "5 .. InvalidRegSeqNo",
        ("FIX.5.0", "507", "6"): "6 .. InvalidRegDetails",
        ("FIX.5.0", "507", "7"): "7 .. InvalidMailingDetails",
        ("FIX.5.0", "507", "8"): "8 .. InvalidMailingInstructions",
        ("FIX.5.0", "507", "9"): "9 .. InvalidInvestorID",
        ("FIX.5.0", "507", "10"): "10 .. InvalidInvestorIDSource",
        ("FIX.5.0", "507", "11"): "11 .. InvalidDateOfBirth",
        ("FIX.5.0", "507", "12"): "12 .. InvalidCountry",
        ("FIX.5.0", "507", "13"): "13 .. InvalidDistribInstns",
        ("FIX.5.0", "507", "14"): "14 .. InvalidPercentage",
        ("FIX.5.0", "507", "15"): "15 .. InvalidPaymentMethod",
        ("FIX.5.0", "507", "16"): "16 .. InvalidAccountName",
        ("FIX.5.0", "507", "17"): "17 .. InvalidAgentCode",
        ("FIX.5.0", "507", "18"): "18 .. InvalidAccountNum",
        ("FIX.5.0", "507", "99"): "99 .. Other",
        ("FIX.5.0", "514", "0"): "0 .. New",
        ("FIX.5.0", "514", "2"): "2 .. Cancel",
        ("FIX.5.0", "514", "1"): "1 .. Replace",
        ("FIX.5.0", "517", "J"): "J .. JointInvestors",
        ("FIX.5.0", "517", "T"): "T .. TenantsInCommon",
        ("FIX.5.0", "517", "2"): "2 .. JointTrustees",
        ("FIX.5.0", "519", "1"): "1 .. CommissionAmount",
        ("FIX.5.0", "519", "2"): "2 .. CommissionPercent",
        ("FIX.5.0", "519", "3"): "3 .. InitialChargeAmount",
        ("FIX.5.0", "519", "4"): "4 .. InitialChargePercent",
        ("FIX.5.0", "519", "5"): "5 .. DiscountAmount",
        ("FIX.5.0", "519", "6"): "6 .. DiscountPercent",
        ("FIX.5.0", "519", "7"): "7 .. DilutionLevyAmount",
        ("FIX.5.0", "519", "8"): "8 .. DilutionLevyPercent",
        ("FIX.5.0", "519", "9"): "9 .. ExitChargeAmount",
        ("FIX.5.0", "519", "10"): "10 .. ExitChargePercent",
        ("FIX.5.0", "519", "11"): "11 .. FundBasedRenewalCommissionPercent",
        ("FIX.5.0", "519", "12"): "12 .. ProjectedFundValue",
        ("FIX.5.0", "519", "13"): "13 .. FundBasedRenewalCommissionOnOrder",
        ("FIX.5.0", "519", "14"): "14 .. FundBasedRenewalCommissionOnFund",
        ("FIX.5.0", "519", "15"): "15 .. NetSettlementAmount",
        ("FIX.5.0", "522", "1"): "1 .. IndividualInvestor",
        ("FIX.5.0", "522", "2"): "2 .. PublicCompany",
        ("FIX.5.0", "522", "3"): "3 .. PrivateCompany",
        ("FIX.5.0", "522", "4"): "4 .. IndividualTrustee",
        ("FIX.5.0", "522", "5"): "5 .. CompanyTrustee",
        ("FIX.5.0", "522", "6"): "6 .. PensionPlan",
        ("FIX.5.0", "522", "7"): "7 .. CustodianUnderGiftsToMinorsAct",
        ("FIX.5.0", "522", "8"): "8 .. Trusts",
        ("FIX.5.0", "522", "9"): "9 .. Fiduciaries",
        ("FIX.5.0", "522", "10"): "10 .. NetworkingSubAccount",
        ("FIX.5.0", "522", "11"): "11 .. NonProfitOrganization",
        ("FIX.5.0", "522", "12"): "12 .. CorporateBody",
        ("FIX.5.0", "522", "13"): "13 .. Nominee",
        ("FIX.5.0", "522", "14"): "14 .. InstitutionalCustomer",
        ("FIX.5.0", "522", "15"): "15 .. Combined",
        ("FIX.5.0", "522", "16"): "16 .. MemberFirmEmployee",
        ("FIX.5.0", "522", "17"): "17 .. MarketMakingAccount",
        ("FIX.5.0", "522", "18"): "18 .. ProprietaryAccount",
        ("FIX.5.0", "522", "19"): "19 .. NonbrokerDealer",
        ("FIX.5.0", "522", "20"): "20 .. UnknownBeneficialOwnerType",
        ("FIX.5.0", "522", "21"): "21 .. FirmsErrorAccount",
        ("FIX.5.0", "525", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "525", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "525", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "525", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "525", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "525", "J"): "J .. TaxID",
        ("FIX.5.0", "525", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "525", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "525", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "525", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "525", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "525", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "525", "B"): "B .. BIC",
        ("FIX.5.0", "525", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "525", "D"): "D .. Proprietary",
        ("FIX.5.0", "525", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "525", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "525", "G"): "G .. MIC",
        ("FIX.5.0", "525", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "525", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "525", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "525", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "525", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "525", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "525", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "525", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "528", "A"): "A .. Agency",
        ("FIX.5.0", "528", "G"): "G .. Proprietary",
        ("FIX.5.0", "528", "I"): "I .. Individual",
        ("FIX.5.0", "528", "P"): "P .. Principal",
        ("FIX.5.0", "528", "R"): "R .. RisklessPrincipal",
        ("FIX.5.0", "528", "W"): "W .. AgentForOtherMember",
        ("FIX.5.0", "528", "M"): "M .. MixedCapacity",
        ("FIX.5.0", "529", "1"): "1 .. ProgramTrade",
        ("FIX.5.0", "529", "2"): "2 .. IndexArbitrage",
        ("FIX.5.0", "529", "3"): "3 .. NonIndexArbitrage",
        ("FIX.5.0", "529", "4"): "4 .. CompetingMarketMaker",
        ("FIX.5.0", "529", "5"): "5 .. ActingAsMarketMakerOrSpecialistInSecurity",
        ("FIX.5.0", "529", "6"): "6 .. ActingAsMarketMakerOrSpecialistInUnderlying",
        ("FIX.5.0", "529", "7"): "7 .. ForeignEntity",
        ("FIX.5.0", "529", "8"): "8 .. ExternalMarketParticipant",
        ("FIX.5.0", "529", "9"): "9 .. ExternalInterConnectedMarketLinkage",
        ("FIX.5.0", "529", "A"): "A .. RisklessArbitrage",
        ("FIX.5.0", "529", "B"): "B .. IssuerHolding",
        ("FIX.5.0", "529", "C"): "C .. IssuePriceStabilization",
        ("FIX.5.0", "529", "D"): "D .. NonAlgorithmic",
        ("FIX.5.0", "529", "E"): "E .. Algorithmic",
        ("FIX.5.0", "529", "F"): "F .. Cross",
        ("FIX.5.0", "529", "G"): "G .. InsiderAccount",
        ("FIX.5.0", "529", "H"): "H .. SignificantShareholder",
        ("FIX.5.0", "529", "I"): "I .. NormalCourseIssuerBid",
        ("FIX.5.0", "530", "1"): "1 .. CancelOrdersForASecurity",
        ("FIX.5.0", "530", "2"): "2 .. CancelOrdersForAnUnderlyingSecurity",
        ("FIX.5.0", "530", "3"): "3 .. CancelOrdersForAProduct",
        ("FIX.5.0", "530", "4"): "4 .. CancelOrdersForACFICode",
        ("FIX.5.0", "530", "5"): "5 .. CancelOrdersForASecurityType",
        ("FIX.5.0", "530", "6"): "6 .. CancelOrdersForATradingSession",
        ("FIX.5.0", "530", "7"): "7 .. CancelAllOrders",
        ("FIX.5.0", "530", "8"): "8 .. CancelOrdersForAMarket",
        ("FIX.5.0", "530", "9"): "9 .. CancelOrdersForAMarketSegment",
        ("FIX.5.0", "530", "A"): "A .. CancelOrdersForASecurityGroup",
        ("FIX.5.0", "530", "B"): "B .. CancelOrdersForSecurityIssuer",
        ("FIX.5.0", "530", "C"): "C .. CancelForIssuerOfUnderlyingSecurity",
        ("FIX.5.0", "531", "0"): "0 .. CancelRequestRejected",
        ("FIX.5.0", "531", "1"): "1 .. CancelOrdersForASecurity",
        ("FIX.5.0", "531", "2"): "2 .. CancelOrdersForAnUnderlyingSecurity",
        ("FIX.5.0", "531", "3"): "3 .. CancelOrdersForAProduct",
        ("FIX.5.0", "531", "4"): "4 .. CancelOrdersForACFICode",
        ("FIX.5.0", "531", "5"): "5 .. CancelOrdersForASecurityType",
        ("FIX.5.0", "531", "6"): "6 .. CancelOrdersForATradingSession",
        ("FIX.5.0", "531", "7"): "7 .. CancelAllOrders",
        ("FIX.5.0", "531", "8"): "8 .. CancelOrdersForAMarket",
        ("FIX.5.0", "531", "9"): "9 .. CancelOrdersForAMarketSegment",
        ("FIX.5.0", "531", "A"): "A .. CancelOrdersForASecurityGroup",
        ("FIX.5.0", "531", "B"): "B .. CancelOrdersForASecuritiesIssuer",
        ("FIX.5.0", "531", "C"): "C .. CancelOrdersForIssuerOfUnderlyingSecurity",
        ("FIX.5.0", "532", "0"): "0 .. MassCancelNotSupported",
        ("FIX.5.0", "532", "1"): "1 .. InvalidOrUnknownSecurity",
        ("FIX.5.0", "532", "2"): "2 .. InvalidOrUnkownUnderlyingSecurity",
        ("FIX.5.0", "532", "3"): "3 .. InvalidOrUnknownProduct",
        ("FIX.5.0", "532", "4"): "4 .. InvalidOrUnknownCFICode",
        ("FIX.5.0", "532", "5"): "5 .. InvalidOrUnknownSecurityType",
        ("FIX.5.0", "532", "6"): "6 .. InvalidOrUnknownTradingSession",
        ("FIX.5.0", "532", "7"): "7 .. InvalidOrUnknownMarket",
        ("FIX.5.0", "532", "8"): "8 .. InvalidOrUnkownMarketSegment",
        ("FIX.5.0", "532", "9"): "9 .. InvalidOrUnknownSecurityGroup",
        ("FIX.5.0", "532", "10"): "10 .. InvalidOrUnknownSecurityIssuer",
        ("FIX.5.0", "532", "11"): "11 .. InvalidOrUnknownIssuerOfUnderlyingSecurity",
        ("FIX.5.0", "532", "99"): "99 .. Other",
        ("FIX.5.0", "537", "0"): "0 .. Indicative",
        ("FIX.5.0", "537", "1"): "1 .. Tradeable",
        ("FIX.5.0", "537", "2"): "2 .. RestrictedTradeable",
        ("FIX.5.0", "537", "3"): "3 .. Counter",
        ("FIX.5.0", "537", "4"): "4 .. InitiallyTradeable",
        ("FIX.5.0", "538", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "538", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "538", "3"): "3 .. ClientID",
        ("FIX.5.0", "538", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "538", "5"): "5 .. InvestorID",
        ("FIX.5.0", "538", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "538", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "538", "8"): "8 .. Locate",
        ("FIX.5.0", "538", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "538", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "538", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "538", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "538", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "538", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "538", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "538", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "538", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "538", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "538", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "538", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "538", "22"): "22 .. Exchange",
        ("FIX.5.0", "538", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "538", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "538", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "538", "27"): "27 .. Buyer",
        ("FIX.5.0", "538", "28"): "28 .. Custodian",
        ("FIX.5.0", "538", "29"): "29 .. Intermediary",
        ("FIX.5.0", "538", "30"): "30 .. Agent",
        ("FIX.5.0", "538", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "538", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "538", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "538", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "538", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "538", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "538", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "538", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "538", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "538", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "538", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "538", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "538", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "538", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "538", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "538", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "538", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "538", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "538", "49"): "49 .. AssetManager",
        ("FIX.5.0", "538", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "538", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "538", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "538", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "538", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "538", "55"): "55 .. SessionID",
        ("FIX.5.0", "538", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "538", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "538", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "538", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "538", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "538", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "538", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "538", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "538", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "538", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "538", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "538", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "538", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "538", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "538", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "538", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "538", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "538", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "538", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "538", "75"): "75 .. LocationID",
        ("FIX.5.0", "538", "76"): "76 .. DeskID",
        ("FIX.5.0", "538", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "538", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "538", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "538", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "538", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "538", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "538", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "538", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "538", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "538", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "538", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "538", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "538", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "538", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "538", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "538", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "538", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "538", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "538", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "538", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "538", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "538", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "538", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "538", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "538", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "538", "102"): "102 .. DataRepository",
        ("FIX.5.0", "538", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "538", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "538", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "538", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "538", "107"): "107 .. Correspondent",
        ("FIX.5.0", "538", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "538", "110"): "110 .. Borrower",
        ("FIX.5.0", "538", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "538", "112"): "112 .. Guarantor",
        ("FIX.5.0", "538", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "538", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "538", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "538", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "538", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "538", "118"): "118 .. Operator",
        ("FIX.5.0", "538", "119"): "119 .. CSD",
        ("FIX.5.0", "538", "120"): "120 .. ICSD",
        ("FIX.5.0", "538", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "538", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "538", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "544", "1"): "1 .. Cash",
        ("FIX.5.0", "544", "2"): "2 .. MarginOpen",
        ("FIX.5.0", "544", "3"): "3 .. MarginClose",
        ("FIX.5.0", "546", "1"): "1 .. LocalMarket",
        ("FIX.5.0", "546", "2"): "2 .. National",
        ("FIX.5.0", "546", "3"): "3 .. Global",
        ("FIX.5.0", "547", "N"): "N .. No",
        ("FIX.5.0", "547", "Y"): "Y .. Yes",
        ("FIX.5.0", "549", "1"): "1 .. CrossAON",
        ("FIX.5.0", "549", "2"): "2 .. CrossIOC",
        ("FIX.5.0", "549", "3"): "3 .. CrossOneSide",
        ("FIX.5.0", "549", "4"): "4 .. CrossSamePrice",
        ("FIX.5.0", "549", "5"): "5 .. BasisCross",
        ("FIX.5.0", "549", "6"): "6 .. ContingentCross",
        ("FIX.5.0", "549", "7"): "7 .. VWAPCross",
        ("FIX.5.0", "549", "8"): "8 .. STSCross",
        ("FIX.5.0", "549", "9"): "9 .. CustomerToCustomer",
        ("FIX.5.0", "550", "0"): "0 .. None",
        ("FIX.5.0", "550", "1"): "1 .. BuySideIsPrioritized",
        ("FIX.5.0", "550", "2"): "2 .. SellSideIsPrioritized",
        ("FIX.5.0", "552", "1"): "1 .. OneSide",
        ("FIX.5.0", "552", "2"): "2 .. BothSides",
        ("FIX.5.0", "559", "0"): "0 .. Symbol",
        ("FIX.5.0", "559", "1"): "1 .. SecurityTypeAnd",
        ("FIX.5.0", "559", "2"): "2 .. Product",
        ("FIX.5.0", "559", "3"): "3 .. TradingSessionID",
        ("FIX.5.0", "559", "4"): "4 .. AllSecurities",
        ("FIX.5.0", "559", "5"): "5 .. MarketIDOrMarketID",
        ("FIX.5.0", "560", "0"): "0 .. ValidRequest",
        ("FIX.5.0", "560", "1"): "1 .. InvalidOrUnsupportedRequest",
        ("FIX.5.0", "560", "2"): "2 .. NoInstrumentsFound",
        ("FIX.5.0", "560", "3"): "3 .. NotAuthorizedToRetrieveInstrumentData",
        ("FIX.5.0", "560", "4"): "4 .. InstrumentDataTemporarilyUnavailable",
        ("FIX.5.0", "560", "5"): "5 .. RequestForInstrumentDataNotSupported",
        ("FIX.5.0", "563", "0"): "0 .. ReportByMulitlegSecurityOnly",
        ("FIX.5.0", "563", "1"): "1 .. ReportByMultilegSecurityAndInstrumentLegs",
        ("FIX.5.0", "563", "2"): "2 .. ReportByInstrumentLegsOnly",
        ("FIX.5.0", "564", "C"): "C .. Close",
        ("FIX.5.0", "564", "F"): "F .. FIFO",
        ("FIX.5.0", "564", "O"): "O .. Open",
        ("FIX.5.0", "564", "R"): "R .. Rolled",
        ("FIX.5.0", "564", "N"): "N .. CloseButNotifyOnOpen",
        ("FIX.5.0", "564", "D"): "D .. Default",
        ("FIX.5.0", "565", "0"): "0 .. Covered",
        ("FIX.5.0", "565", "1"): "1 .. Uncovered",
        ("FIX.5.0", "567", "1"): "1 .. UnknownOrInvalidTradingSessionID",
        ("FIX.5.0", "567", "99"): "99 .. Other",
        ("FIX.5.0", "569", "0"): "0 .. AllTrades",
        ("FIX.5.0", "569", "1"): "1 .. MatchedTradesMatchingCriteria",
        ("FIX.5.0", "569", "2"): "2 .. UnmatchedTradesThatMatchCriteria",
        ("FIX.5.0", "569", "3"): "3 .. UnreportedTradesThatMatchCriteria",
        ("FIX.5.0", "569", "4"): "4 .. AdvisoriesThatMatchCriteria",
        ("FIX.5.0", "570", "N"): "N .. NotReportedToCounterparty",
        ("FIX.5.0", "570", "Y"): "Y .. PreviouslyReportedToCounterparty",
        ("FIX.5.0", "573", "0"): "0 .. Compared",
        ("FIX.5.0", "573", "1"): "1 .. Uncompared",
        ("FIX.5.0", "573", "2"): "2 .. AdvisoryOrAlert",
        ("FIX.5.0", "574", "1"): "1 .. OnePartyTradeReport",
        ("FIX.5.0", "574", "2"): "2 .. TwoPartyTradeReport",
        ("FIX.5.0", "574", "3"): "3 .. ConfirmedTradeReport",
        ("FIX.5.0", "574", "4"): "4 .. AutoMatch",
        ("FIX.5.0", "574", "5"): "5 .. CrossAuction",
        ("FIX.5.0", "574", "6"): "6 .. CounterOrderSelection",
        ("FIX.5.0", "574", "7"): "7 .. CallAuction",
        ("FIX.5.0", "574", "8"): "8 .. Issuing",
        ("FIX.5.0", "574", "9"): "9 .. SystematicInternaliser",
        ("FIX.5.0", "574", "10"): "10 .. AutoMatchLastLook",
        ("FIX.5.0", "574", "11"): "11 .. CrossAuctionLastLook",
        ("FIX.5.0", "574", "M3"): "M3 .. ACTAcceptedTrade",
        ("FIX.5.0", "574", "M4"): "M4 .. ACTDefaultTrade",
        ("FIX.5.0", "574", "M5"): "M5 .. ACTDefaultAfterM2",
        ("FIX.5.0", "574", "M6"): "M6 .. ACTM6Match",
        ("FIX.5.0", "574", "A1"): "A1 .. ExactMatchPlus4BadgesExecTime",
        ("FIX.5.0", "574", "A2"): "A2 .. ExactMatchPlus4Badges",
        ("FIX.5.0", "574", "A3"): "A3 .. ExactMatchPlus2BadgesExecTime",
        ("FIX.5.0", "574", "A4"): "A4 .. ExactMatchPlus2Badges",
        ("FIX.5.0", "574", "A5"): "A5 .. ExactMatchPlusExecTime",
        ("FIX.5.0", "574", "AQ"): "AQ .. StampedAdvisoriesOrSpecialistAccepts",
        ("FIX.5.0", "574", "S1"): "S1 .. A1ExactMatchSummarizedQuantity",
        ("FIX.5.0", "574", "S2"): "S2 .. A2ExactMatchSummarizedQuantity",
        ("FIX.5.0", "574", "S3"): "S3 .. A3ExactMatchSummarizedQuantity",
        ("FIX.5.0", "574", "S4"): "S4 .. A4ExactMatchSummarizedQuantity",
        ("FIX.5.0", "574", "S5"): "S5 .. A5ExactMatchSummarizedQuantity",
        ("FIX.5.0", "574", "M1"): "M1 .. ExactMatchMinusBadgesTimes",
        ("FIX.5.0", "574", "M2"): "M2 .. SummarizedMatchMinusBadgesTimes",
        ("FIX.5.0", "574", "MT"): "MT .. OCSLockedIn",
        ("FIX.5.0", "575", "N"): "N .. TreatAsRoundLot",
        ("FIX.5.0", "575", "Y"): "Y .. TreatAsOddLot",
        ("FIX.5.0", "577", "0"): "0 .. ProcessNormally",
        ("FIX.5.0", "577", "1"): "1 .. ExcludeFromAllNetting",
        ("FIX.5.0", "577", "2"): "2 .. BilateralNettingOnly",
        ("FIX.5.0", "577", "3"): "3 .. ExClearing",
        ("FIX.5.0", "577", "4"): "4 .. SpecialTrade",
        ("FIX.5.0", "577", "5"): "5 .. MultilateralNetting",
        ("FIX.5.0", "577", "6"): "6 .. ClearAgainstCentralCounterparty",
        ("FIX.5.0", "577", "7"): "7 .. ExcludeFromCentralCounterparty",
        ("FIX.5.0", "577", "8"): "8 .. ManualMode",
        ("FIX.5.0", "577", "9"): "9 .. AutomaticPostingMode",
        ("FIX.5.0", "577", "10"): "10 .. AutomaticGiveUpMode",
        ("FIX.5.0", "577", "11"): "11 .. QualifiedServiceRepresentativeQSR",
        ("FIX.5.0", "577", "12"): "12 .. CustomerTrade",
        ("FIX.5.0", "577", "13"): "13 .. SelfClearing",
        ("FIX.5.0", "577", "14"): "14 .. BuyIn",
        ("FIX.5.0", "581", "1"): "1 .. CarriedCustomerSide",
        ("FIX.5.0", "581", "2"): "2 .. CarriedNonCustomerSide",
        ("FIX.5.0", "581", "3"): "3 .. HouseTrader",
        ("FIX.5.0", "581", "4"): "4 .. FloorTrader",
        ("FIX.5.0", "581", "6"): "6 .. CarriedNonCustomerSideCrossMargined",
        ("FIX.5.0", "581", "7"): "7 .. HouseTraderCrossMargined",
        ("FIX.5.0", "581", "8"): "8 .. JointBackOfficeAccount",
        ("FIX.5.0", "581", "9"): "9 .. EquitiesSpecialist",
        ("FIX.5.0", "581", "10"): "10 .. OptionsMarketMaker",
        ("FIX.5.0", "581", "11"): "11 .. OptionsFirmAccount",
        ("FIX.5.0", "582", "1"): "1 .. MemberTradingForTheirOwnAccount",
        ("FIX.5.0", "582", "2"): "2 .. ClearingFirmTradingForItsProprietaryAccount",
        ("FIX.5.0", "582", "3"): "3 .. MemberTradingForAnotherMember",
        ("FIX.5.0", "582", "4"): "4 .. AllOther",
        ("FIX.5.0", "582", "5"): "5 .. RetailCustomer",
        ("FIX.5.0", "585", "1"): "1 .. StatusForOrdersForASecurity",
        ("FIX.5.0", "585", "2"): "2 .. StatusForOrdersForAnUnderlyingSecurity",
        ("FIX.5.0", "585", "3"): "3 .. StatusForOrdersForAProduct",
        ("FIX.5.0", "585", "4"): "4 .. StatusForOrdersForACFICode",
        ("FIX.5.0", "585", "5"): "5 .. StatusForOrdersForASecurityType",
        ("FIX.5.0", "585", "6"): "6 .. StatusForOrdersForATradingSession",
        ("FIX.5.0", "585", "7"): "7 .. StatusForAllOrders",
        ("FIX.5.0", "585", "8"): "8 .. StatusForOrdersForAPartyID",
        ("FIX.5.0", "585", "9"): "9 .. StatusForSecurityIssuer",
        ("FIX.5.0", "585", "10"): "10 .. StatusForIssuerOfUnderlyingSecurity",
        ("FIX.5.0", "587", "0"): "0 .. Regular",
        ("FIX.5.0", "587", "1"): "1 .. Cash",
        ("FIX.5.0", "587", "2"): "2 .. NextDay",
        ("FIX.5.0", "587", "3"): "3 .. TPlus2",
        ("FIX.5.0", "587", "4"): "4 .. TPlus3",
        ("FIX.5.0", "587", "5"): "5 .. TPlus4",
        ("FIX.5.0", "587", "6"): "6 .. Future",
        ("FIX.5.0", "587", "7"): "7 .. WhenAndIfIssued",
        ("FIX.5.0", "587", "8"): "8 .. SellersOption",
        ("FIX.5.0", "587", "9"): "9 .. TPlus5",
        ("FIX.5.0", "587", "B"): "B .. BrokenDate",
        ("FIX.5.0", "587", "C"): "C .. FXSpotNextSettlement",
        ("FIX.5.0", "589", "0"): "0 .. Auto",
        ("FIX.5.0", "589", "1"): "1 .. SpeakWithOrderInitiatorBeforeBooking",
        ("FIX.5.0", "589", "2"): "2 .. Accumulate",
        ("FIX.5.0", "590", "0"): "0 .. EachPartialExecutionIsABookableUnit",
        ("FIX.5.0", "590", "1"): "1 .. AggregatePartialExecutionsOnThisOrder",
        ("FIX.5.0", "590", "2"): "2 .. AggregateExecutionsForThisSymbol",
        ("FIX.5.0", "591", "0"): "0 .. ProRata",
        ("FIX.5.0", "591", "1"): "1 .. DoNotProRata",
        ("FIX.5.0", "601", "CD"): "CD .. EUCPWithLumpSumInterest",
        ("FIX.5.0", "601", "WI"): "WI .. WhenIssued",
        ("FIX.5.0", "603", "1"): "1 .. CUSIP",
        ("FIX.5.0", "603", "2"): "2 .. SEDOL",
        ("FIX.5.0", "603", "3"): "3 .. QUIK",
        ("FIX.5.0", "603", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "603", "5"): "5 .. RICCode",
        ("FIX.5.0", "603", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "603", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "603", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "603", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "603", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "603", "B"): "B .. Wertpapier",
        ("FIX.5.0", "603", "C"): "C .. Dutch",
        ("FIX.5.0", "603", "D"): "D .. Valoren",
        ("FIX.5.0", "603", "E"): "E .. Sicovam",
        ("FIX.5.0", "603", "F"): "F .. Belgian",
        ("FIX.5.0", "603", "G"): "G .. Common",
        ("FIX.5.0", "603", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "603", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "603", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "603", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "603", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "603", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "603", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "603", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "603", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "603", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "603", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "603", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "603", "U"): "U .. Synthetic",
        ("FIX.5.0", "603", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "603", "W"): "W .. IndexName",
        ("FIX.5.0", "606", "1"): "1 .. CUSIP",
        ("FIX.5.0", "606", "2"): "2 .. SEDOL",
        ("FIX.5.0", "606", "3"): "3 .. QUIK",
        ("FIX.5.0", "606", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "606", "5"): "5 .. RICCode",
        ("FIX.5.0", "606", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "606", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "606", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "606", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "606", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "606", "B"): "B .. Wertpapier",
        ("FIX.5.0", "606", "C"): "C .. Dutch",
        ("FIX.5.0", "606", "D"): "D .. Valoren",
        ("FIX.5.0", "606", "E"): "E .. Sicovam",
        ("FIX.5.0", "606", "F"): "F .. Belgian",
        ("FIX.5.0", "606", "G"): "G .. Common",
        ("FIX.5.0", "606", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "606", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "606", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "606", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "606", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "606", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "606", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "606", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "606", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "606", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "606", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "606", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "606", "U"): "U .. Synthetic",
        ("FIX.5.0", "606", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "606", "W"): "W .. IndexName",
        ("FIX.5.0", "607", "1"): "1 .. AGENCY",
        ("FIX.5.0", "607", "2"): "2 .. COMMODITY",
        ("FIX.5.0", "607", "3"): "3 .. CORPORATE",
        ("FIX.5.0", "607", "4"): "4 .. CURRENCY",
        ("FIX.5.0", "607", "5"): "5 .. EQUITY",
        ("FIX.5.0", "607", "6"): "6 .. GOVERNMENT",
        ("FIX.5.0", "607", "7"): "7 .. INDEX",
        ("FIX.5.0", "607", "8"): "8 .. LOAN",
        ("FIX.5.0", "607", "9"): "9 .. MONEYMARKET",
        ("FIX.5.0", "607", "10"): "10 .. MORTGAGE",
        ("FIX.5.0", "607", "11"): "11 .. MUNICIPAL",
        ("FIX.5.0", "607", "12"): "12 .. OTHER",
        ("FIX.5.0", "607", "13"): "13 .. FINANCING",
        ("FIX.5.0", "609", "EUSUPRA"): "EUSUPRA .. EuroSupranationalCoupons",
        ("FIX.5.0", "609", "FAC"): "FAC .. FederalAgencyCoupon",
        ("FIX.5.0", "609", "FADN"): "FADN .. FederalAgencyDiscountNote",
        ("FIX.5.0", "609", "PEF"): "PEF .. PrivateExportFunding",
        ("FIX.5.0", "609", "SUPRA"): "SUPRA .. USDSupranationalCoupons",
        ("FIX.5.0", "609", "CORP"): "CORP .. CorporateBond",
        ("FIX.5.0", "609", "CPP"): "CPP .. CorporatePrivatePlacement",
        ("FIX.5.0", "609", "CB"): "CB .. ConvertibleBond",
        ("FIX.5.0", "609", "DUAL"): "DUAL .. DualCurrency",
        ("FIX.5.0", "609", "EUCORP"): "EUCORP .. EuroCorporateBond",
        ("FIX.5.0", "609", "EUFRN"): "EUFRN .. EuroCorporateFloatingRateNotes",
        ("FIX.5.0", "609", "FRN"): "FRN .. USCorporateFloatingRateNotes",
        ("FIX.5.0", "609", "XLINKD"): "XLINKD .. IndexedLinked",
        ("FIX.5.0", "609", "STRUCT"): "STRUCT .. StructuredNotes",
        ("FIX.5.0", "609", "YANK"): "YANK .. YankeeCorporateBond",
        ("FIX.5.0", "609", "FXNDF"): "FXNDF .. NonDeliverableForward",
        ("FIX.5.0", "609", "FXSPOT"): "FXSPOT .. FXSpot",
        ("FIX.5.0", "609", "FXFWD"): "FXFWD .. FXForward",
        ("FIX.5.0", "609", "FXSWAP"): "FXSWAP .. FXSwap",
        ("FIX.5.0", "609", "FXNDS"): "FXNDS .. NonDeliverableSwap",
        ("FIX.5.0", "609", "CAP"): "CAP .. Cap",
        ("FIX.5.0", "609", "CDS"): "CDS .. CreditDefaultSwap",
        ("FIX.5.0", "609", "CLLR"): "CLLR .. Collar",
        ("FIX.5.0", "609", "CMDTYSWAP"): "CMDTYSWAP .. CommoditySwap",
        ("FIX.5.0", "609", "EXOTIC"): "EXOTIC .. Exotic",
        ("FIX.5.0", "609", "OOC"): "OOC .. OptionsOnCombo",
        ("FIX.5.0", "609", "FLR"): "FLR .. Floor",
        ("FIX.5.0", "609", "FRA"): "FRA .. FRA",
        ("FIX.5.0", "609", "FUT"): "FUT .. Future",
        ("FIX.5.0", "609", "FWD"): "FWD .. DerivativeForward",
        ("FIX.5.0", "609", "IRS"): "IRS .. InterestRateSwap",
        ("FIX.5.0", "609", "TRS"): "TRS .. TotalReturnSwap",
        ("FIX.5.0", "609", "LOANLEASE"): "LOANLEASE .. LoanLease",
        ("FIX.5.0", "609", "OOF"): "OOF .. OptionsOnFutures",
        ("FIX.5.0", "609", "OOP"): "OOP .. OptionsOnPhysical",
        ("FIX.5.0", "609", "OPT"): "OPT .. Option",
        ("FIX.5.0", "609", "SPOTFWD"): "SPOTFWD .. SpotForward",
        ("FIX.5.0", "609", "SWAPTION"): "SWAPTION .. SwapOption",
        ("FIX.5.0", "609", "XMISSION"): "XMISSION .. Transmission",
        ("FIX.5.0", "609", "INDEX"): "INDEX .. Index",
        ("FIX.5.0", "609", "BDBSKT"): "BDBSKT .. BondBasket",
        ("FIX.5.0", "609", "CFD"): "CFD .. ContractForDifference",
        ("FIX.5.0", "609", "CRLTNSWAP"): "CRLTNSWAP .. CorrelationSwap",
        ("FIX.5.0", "609", "DVDNDSWAP"): "DVDNDSWAP .. DiviendSwap",
        ("FIX.5.0", "609", "EQBSKT"): "EQBSKT .. EquityBasket",
        ("FIX.5.0", "609", "EQFWD"): "EQFWD .. EquityForward",
        ("FIX.5.0", "609", "RTRNSWAP"): "RTRNSWAP .. ReturnSwap",
        ("FIX.5.0", "609", "VARSWAP"): "VARSWAP .. VarianceSwap",
        ("FIX.5.0", "609", "PRTFLIOSWAP"): "PRTFLIOSWAP .. PortfolioSwaps",
        ("FIX.5.0", "609", "FUTSWAP"): "FUTSWAP .. FuturesOnASwap",
        ("FIX.5.0", "609", "FWDSWAP "): "FWDSWAP  .. ForwardsOnASwap",
        ("FIX.5.0", "609", "FWDFRTAGMT "): "FWDFRTAGMT  .. ForwardFreightAgreement",
        ("FIX.5.0", "609", "SPREADBET"): "SPREADBET .. SpreadBetting",
        ("FIX.5.0", "609", "ETC"): "ETC .. ExchangeTradedCommodity",
        ("FIX.5.0", "609", "CS"): "CS .. CommonStock",
        ("FIX.5.0", "609", "PS"): "PS .. PreferredStock",
        ("FIX.5.0", "609", "DR"): "DR .. DepositoryReceipts",
        ("FIX.5.0", "609", "REPO"): "REPO .. Repurchase",
        ("FIX.5.0", "609", "FORWARD"): "FORWARD .. Forward",
        ("FIX.5.0", "609", "BUYSELL"): "BUYSELL .. BuySellback",
        ("FIX.5.0", "609", "SECLOAN"): "SECLOAN .. SecuritiesLoan",
        ("FIX.5.0", "609", "SECPLEDGE"): "SECPLEDGE .. SecuritiesPledge",
        ("FIX.5.0", "609", "DVPLDG"): "DVPLDG .. DeliveryVersusPledge",
        ("FIX.5.0", "609", "COLLBSKT"): "COLLBSKT .. CollateralBasket",
        ("FIX.5.0", "609", "SFP"): "SFP .. StructuredFinanceProduct",
        ("FIX.5.0", "609", "BRADY"): "BRADY .. BradyBond",
        ("FIX.5.0", "609", "CAN"): "CAN .. CanadianTreasuryNotes",
        ("FIX.5.0", "609", "CTB"): "CTB .. CanadianTreasuryBills",
        ("FIX.5.0", "609", "EUSOV"): "EUSOV .. EuroSovereigns",
        ("FIX.5.0", "609", "PROV"): "PROV .. CanadianProvincialBonds",
        ("FIX.5.0", "609", "TB"): "TB .. TreasuryBill",
        ("FIX.5.0", "609", "TBOND"): "TBOND .. USTreasuryBond",
        ("FIX.5.0", "609", "TINT"): "TINT .. InterestStripFromAnyBondOrNote",
        ("FIX.5.0", "609", "TBILL"): "TBILL .. USTreasuryBill",
        ("FIX.5.0", "609", "TIPS"): "TIPS .. TreasuryInflationProtectedSecurities",
        ("FIX.5.0", "609", "TCAL"): "TCAL .. PrincipalStripOfACallableBondOrNote",
        ("FIX.5.0", "609", "TPRN"): "TPRN .. PrincipalStripFromANonCallableBondOrNote",
        ("FIX.5.0", "609", "TNOTE"): "TNOTE .. USTreasuryNote",
        ("FIX.5.0", "609", "TERM"): "TERM .. TermLoan",
        ("FIX.5.0", "609", "RVLV"): "RVLV .. RevolverLoan",
        ("FIX.5.0", "609", "RVLVTRM"): "RVLVTRM .. Revolver",
        ("FIX.5.0", "609", "BRIDGE"): "BRIDGE .. BridgeLoan",
        ("FIX.5.0", "609", "LOFC"): "LOFC .. LetterOfCredit",
        ("FIX.5.0", "609", "SWING"): "SWING .. SwingLineFacility",
        ("FIX.5.0", "609", "DINP"): "DINP .. DebtorInPossession",
        ("FIX.5.0", "609", "DEFLTED"): "DEFLTED .. Defaulted",
        ("FIX.5.0", "609", "WITHDRN"): "WITHDRN .. Withdrawn",
        ("FIX.5.0", "609", "REPLACD"): "REPLACD .. Replaced",
        ("FIX.5.0", "609", "MATURED"): "MATURED .. Matured",
        ("FIX.5.0", "609", "AMENDED"): "AMENDED .. Amended",
        ("FIX.5.0", "609", "RETIRED"): "RETIRED .. Retired",
        ("FIX.5.0", "609", "BA"): "BA .. BankersAcceptance",
        ("FIX.5.0", "609", "BDN"): "BDN .. BankDepositoryNote",
        ("FIX.5.0", "609", "BN"): "BN .. BankNotes",
        ("FIX.5.0", "609", "BOX"): "BOX .. BillOfExchanges",
        ("FIX.5.0", "609", "CAMM"): "CAMM .. CanadianMoneyMarkets",
        ("FIX.5.0", "609", "CD"): "CD .. CertificateOfDeposit",
        ("FIX.5.0", "609", "CL"): "CL .. CallLoans",
        ("FIX.5.0", "609", "CP"): "CP .. CommercialPaper",
        ("FIX.5.0", "609", "DN"): "DN .. DepositNotes",
        ("FIX.5.0", "609", "EUCD"): "EUCD .. EuroCertificateOfDeposit",
        ("FIX.5.0", "609", "EUCP"): "EUCP .. EuroCommercialPaper",
        ("FIX.5.0", "609", "LQN"): "LQN .. LiquidityNote",
        ("FIX.5.0", "609", "MTN"): "MTN .. MediumTermNotes",
        ("FIX.5.0", "609", "ONITE"): "ONITE .. Overnight",
        ("FIX.5.0", "609", "PN"): "PN .. PromissoryNote",
        ("FIX.5.0", "609", "STN"): "STN .. ShortTermLoanNote",
        ("FIX.5.0", "609", "PZFJ"): "PZFJ .. PlazosFijos",
        ("FIX.5.0", "609", "SLQN"): "SLQN .. SecuredLiquidityNote",
        ("FIX.5.0", "609", "TD"): "TD .. TimeDeposit",
        ("FIX.5.0", "609", "TLQN"): "TLQN .. TermLiquidityNote",
        ("FIX.5.0", "609", "XCN"): "XCN .. ExtendedCommNote",
        ("FIX.5.0", "609", "YCD"): "YCD .. YankeeCertificateOfDeposit",
        ("FIX.5.0", "609", "ABS"): "ABS .. AssetBackedSecurities",
        ("FIX.5.0", "609", "CMB"): "CMB .. CanadianMortgageBonds",
        ("FIX.5.0", "609", "CMBS"): "CMBS .. Corp",
        ("FIX.5.0", "609", "CMO"): "CMO .. CollateralizedMortgageObligation",
        ("FIX.5.0", "609", "IET"): "IET .. IOETTEMortgage",
        ("FIX.5.0", "609", "MBS"): "MBS .. MortgageBackedSecurities",
        ("FIX.5.0", "609", "MIO"): "MIO .. MortgageInterestOnly",
        ("FIX.5.0", "609", "MPO"): "MPO .. MortgagePrincipalOnly",
        ("FIX.5.0", "609", "MPP"): "MPP .. MortgagePrivatePlacement",
        ("FIX.5.0", "609", "MPT"): "MPT .. MiscellaneousPassThrough",
        ("FIX.5.0", "609", "PFAND"): "PFAND .. Pfandbriefe",
        ("FIX.5.0", "609", "TBA"): "TBA .. ToBeAnnounced",
        ("FIX.5.0", "609", "AN"): "AN .. OtherAnticipationNotes",
        ("FIX.5.0", "609", "COFO"): "COFO .. CertificateOfObligation",
        ("FIX.5.0", "609", "COFP"): "COFP .. CertificateOfParticipation",
        ("FIX.5.0", "609", "GO"): "GO .. GeneralObligationBonds",
        ("FIX.5.0", "609", "MT"): "MT .. MandatoryTender",
        ("FIX.5.0", "609", "RAN"): "RAN .. RevenueAnticipationNote",
        ("FIX.5.0", "609", "REV"): "REV .. RevenueBonds",
        ("FIX.5.0", "609", "SPCLA"): "SPCLA .. SpecialAssessment",
        ("FIX.5.0", "609", "SPCLO"): "SPCLO .. SpecialObligation",
        ("FIX.5.0", "609", "SPCLT"): "SPCLT .. SpecialTax",
        ("FIX.5.0", "609", "TAN"): "TAN .. TaxAnticipationNote",
        ("FIX.5.0", "609", "TAXA"): "TAXA .. TaxAllocation",
        ("FIX.5.0", "609", "TECP"): "TECP .. TaxExemptCommercialPaper",
        ("FIX.5.0", "609", "TMCP"): "TMCP .. TaxableMunicipalCP",
        ("FIX.5.0", "609", "TRAN"): "TRAN .. TaxRevenueAnticipationNote",
        ("FIX.5.0", "609", "VRDN"): "VRDN .. VariableRateDemandNote",
        ("FIX.5.0", "609", "WAR"): "WAR .. Warrant",
        ("FIX.5.0", "609", "MF"): "MF .. MutualFund",
        ("FIX.5.0", "609", "MLEG"): "MLEG .. MultilegInstrument",
        ("FIX.5.0", "609", "NONE"): "NONE .. NoSecurityType",
        ("FIX.5.0", "609", "?"): "? .. Wildcard",
        ("FIX.5.0", "609", "CASH"): "CASH .. Cash",
        ("FIX.5.0", "609", "Other"): "Other .. Other",
        ("FIX.5.0", "609", "ETN"): "ETN .. ExchangeTradedNote",
        ("FIX.5.0", "609", "SECDERIV"): "SECDERIV .. SecuritizedDerivative",
        ("FIX.5.0", "624", "1"): "1 .. Buy",
        ("FIX.5.0", "624", "2"): "2 .. Sell",
        ("FIX.5.0", "624", "3"): "3 .. BuyMinus",
        ("FIX.5.0", "624", "4"): "4 .. SellPlus",
        ("FIX.5.0", "624", "5"): "5 .. SellShort",
        ("FIX.5.0", "624", "6"): "6 .. SellShortExempt",
        ("FIX.5.0", "624", "7"): "7 .. Undisclosed",
        ("FIX.5.0", "624", "8"): "8 .. Cross",
        ("FIX.5.0", "624", "9"): "9 .. CrossShort",
        ("FIX.5.0", "624", "A"): "A .. CrossShortExempt",
        ("FIX.5.0", "624", "B"): "B .. AsDefined",
        ("FIX.5.0", "624", "C"): "C .. Opposite",
        ("FIX.5.0", "624", "D"): "D .. Subscribe",
        ("FIX.5.0", "624", "E"): "E .. Redeem",
        ("FIX.5.0", "624", "F"): "F .. Lend",
        ("FIX.5.0", "624", "G"): "G .. Borrow",
        ("FIX.5.0", "624", "H"): "H .. SellUndisclosed",
        ("FIX.5.0", "625", "1"): "1 .. PreTrading",
        ("FIX.5.0", "625", "2"): "2 .. OpeningOrOpeningAuction",
        ("FIX.5.0", "625", "3"): "3 .. Continuous",
        ("FIX.5.0", "625", "4"): "4 .. ClosingOrClosingAuction",
        ("FIX.5.0", "625", "5"): "5 .. PostTrading",
        ("FIX.5.0", "625", "6"): "6 .. ScheduledIntradayAuction",
        ("FIX.5.0", "625", "7"): "7 .. Quiescent",
        ("FIX.5.0", "625", "8"): "8 .. AnyAuction",
        ("FIX.5.0", "625", "9"): "9 .. UnscheduledIntradayAuction",
        ("FIX.5.0", "625", "10"): "10 .. OutOfMainSessionTrading",
        ("FIX.5.0", "625", "11"): "11 .. PrivateAuction",
        ("FIX.5.0", "625", "12"): "12 .. PublicAuction",
        ("FIX.5.0", "625", "13"): "13 .. GroupAuction",
        ("FIX.5.0", "626", "1"): "1 .. Calculated",
        ("FIX.5.0", "626", "2"): "2 .. Preliminary",
        ("FIX.5.0", "626", "5"): "5 .. ReadyToBook",
        ("FIX.5.0", "626", "7"): "7 .. WarehouseInstruction",
        ("FIX.5.0", "626", "8"): "8 .. RequestToIntermediary",
        ("FIX.5.0", "626", "12"): "12 .. IncompleteGroup",
        ("FIX.5.0", "626", "13"): "13 .. CompleteGroup",
        ("FIX.5.0", "626", "15"): "15 .. ReopenGroup",
        ("FIX.5.0", "626", "16"): "16 .. CancelGroup",
        ("FIX.5.0", "626", "17"): "17 .. Giveup",
        ("FIX.5.0", "626", "18"): "18 .. Takeup",
        ("FIX.5.0", "626", "19"): "19 .. RefuseTakeup",
        ("FIX.5.0", "626", "20"): "20 .. InitiateReversal",
        ("FIX.5.0", "626", "21"): "21 .. Reverse",
        ("FIX.5.0", "626", "22"): "22 .. RefuseReversal",
        ("FIX.5.0", "626", "23"): "23 .. SubAllocationGiveup",
        ("FIX.5.0", "626", "24"): "24 .. ApproveGiveup",
        ("FIX.5.0", "626", "25"): "25 .. ApproveTakeup",
        ("FIX.5.0", "626", "26"): "26 .. NotionalValueAveragePxGroupAlloc",
        ("FIX.5.0", "635", "1"): "1 .. FirstYearDelegate",
        ("FIX.5.0", "635", "2"): "2 .. SecondYearDelegate",
        ("FIX.5.0", "635", "3"): "3 .. ThirdYearDelegate",
        ("FIX.5.0", "635", "4"): "4 .. FourthYearDelegate",
        ("FIX.5.0", "635", "5"): "5 .. FifthYearDelegate",
        ("FIX.5.0", "635", "9"): "9 .. SixthYearDelegate",
        ("FIX.5.0", "635", "B"): "B .. CBOEMember",
        ("FIX.5.0", "635", "C"): "C .. NonMemberAndCustomer",
        ("FIX.5.0", "635", "E"): "E .. EquityMemberAndClearingMember",
        ("FIX.5.0", "635", "F"): "F .. FullAndAssociateMember",
        ("FIX.5.0", "635", "H"): "H .. Firms106HAnd106J",
        ("FIX.5.0", "635", "I"): "I .. GIM",
        ("FIX.5.0", "635", "L"): "L .. Lessee106FEmployees",
        ("FIX.5.0", "635", "M"): "M .. AllOtherOwnershipTypes",
        ("FIX.5.0", "636", "N"): "N .. NotWorking",
        ("FIX.5.0", "636", "Y"): "Y .. Working",
        ("FIX.5.0", "638", "0"): "0 .. PriorityUnchanged",
        ("FIX.5.0", "638", "1"): "1 .. LostPriorityAsResultOfOrderChange",
        ("FIX.5.0", "650", "N"): "N .. DoesNotConsituteALegalConfirm",
        ("FIX.5.0", "650", "Y"): "Y .. LegalConfirm",
        ("FIX.5.0", "658", "1"): "1 .. UnknownSymbol",
        ("FIX.5.0", "658", "2"): "2 .. Exchange",
        ("FIX.5.0", "658", "3"): "3 .. QuoteRequestExceedsLimit",
        ("FIX.5.0", "658", "4"): "4 .. TooLateToEnter",
        ("FIX.5.0", "658", "5"): "5 .. InvalidPrice",
        ("FIX.5.0", "658", "6"): "6 .. NotAuthorizedToRequestQuote",
        ("FIX.5.0", "658", "7"): "7 .. NoMatchForInquiry",
        ("FIX.5.0", "658", "8"): "8 .. NoMarketForInstrument",
        ("FIX.5.0", "658", "9"): "9 .. NoInventory",
        ("FIX.5.0", "658", "10"): "10 .. Pass",
        ("FIX.5.0", "658", "11"): "11 .. InsufficientCredit",
        ("FIX.5.0", "658", "12"): "12 .. ExceededClipSizeLimit",
        ("FIX.5.0", "658", "13"): "13 .. ExceededMaxNotionalOrderAmt",
        ("FIX.5.0", "658", "14"): "14 .. ExceededDV01PV01Limit",
        ("FIX.5.0", "658", "15"): "15 .. ExceededCS01Limit",
        ("FIX.5.0", "658", "99"): "99 .. Other",
        ("FIX.5.0", "660", "1"): "1 .. BIC",
        ("FIX.5.0", "660", "2"): "2 .. SIDCode",
        ("FIX.5.0", "660", "3"): "3 .. TFM",
        ("FIX.5.0", "660", "4"): "4 .. OMGEO",
        ("FIX.5.0", "660", "5"): "5 .. DTCCCode",
        ("FIX.5.0", "660", "99"): "99 .. Other",
        ("FIX.5.0", "661", "1"): "1 .. BIC",
        ("FIX.5.0", "661", "2"): "2 .. SIDCode",
        ("FIX.5.0", "661", "3"): "3 .. TFM",
        ("FIX.5.0", "661", "4"): "4 .. OMGEO",
        ("FIX.5.0", "661", "5"): "5 .. DTCCCode",
        ("FIX.5.0", "661", "99"): "99 .. Other",
        ("FIX.5.0", "663", "1"): "1 .. Percentage",
        ("FIX.5.0", "663", "2"): "2 .. PerUnit",
        ("FIX.5.0", "663", "3"): "3 .. FixedAmount",
        ("FIX.5.0", "663", "4"): "4 .. Discount",
        ("FIX.5.0", "663", "5"): "5 .. Premium",
        ("FIX.5.0", "663", "6"): "6 .. Spread",
        ("FIX.5.0", "663", "7"): "7 .. TEDPrice",
        ("FIX.5.0", "663", "8"): "8 .. TEDYield",
        ("FIX.5.0", "663", "9"): "9 .. Yield",
        ("FIX.5.0", "663", "10"): "10 .. FixedCabinetTradePrice",
        ("FIX.5.0", "663", "11"): "11 .. VariableCabinetTradePrice",
        ("FIX.5.0", "663", "12"): "12 .. PriceSpread",
        ("FIX.5.0", "663", "13"): "13 .. ProductTicksInHalves",
        ("FIX.5.0", "663", "14"): "14 .. ProductTicksInFourths",
        ("FIX.5.0", "663", "15"): "15 .. ProductTicksInEighths",
        ("FIX.5.0", "663", "16"): "16 .. ProductTicksInSixteenths",
        ("FIX.5.0", "663", "17"): "17 .. ProductTicksInThirtySeconds",
        ("FIX.5.0", "663", "18"): "18 .. ProductTicksInSixtyFourths",
        ("FIX.5.0", "663", "19"): "19 .. ProductTicksInOneTwentyEighths",
        ("FIX.5.0", "663", "20"): "20 .. NormalRateRepresentation",
        ("FIX.5.0", "663", "21"): "21 .. InverseRateRepresentation",
        ("FIX.5.0", "663", "22"): "22 .. BasisPoints",
        ("FIX.5.0", "663", "23"): "23 .. UpfrontPoints",
        ("FIX.5.0", "663", "24"): "24 .. InterestRate",
        ("FIX.5.0", "663", "25"): "25 .. PercentageNotional",
        ("FIX.5.0", "665", "1"): "1 .. Received",
        ("FIX.5.0", "665", "2"): "2 .. MismatchedAccount",
        ("FIX.5.0", "665", "3"): "3 .. MissingSettlementInstructions",
        ("FIX.5.0", "665", "4"): "4 .. Confirmed",
        ("FIX.5.0", "665", "5"): "5 .. RequestRejected",
        ("FIX.5.0", "666", "0"): "0 .. New",
        ("FIX.5.0", "666", "1"): "1 .. Replace",
        ("FIX.5.0", "666", "2"): "2 .. Cancel",
        ("FIX.5.0", "668", "1"): "1 .. BookEntry",
        ("FIX.5.0", "668", "2"): "2 .. Bearer",
        ("FIX.5.0", "677", "EONIA"): "EONIA .. EONIA",
        ("FIX.5.0", "677", "EUREPO"): "EUREPO .. EUREPO",
        ("FIX.5.0", "677", "FutureSWAP"): "FutureSWAP .. FutureSWAP",
        ("FIX.5.0", "677", "LIBID"): "LIBID .. LIBID",
        ("FIX.5.0", "677", "LIBOR"): "LIBOR .. LIBOR",
        ("FIX.5.0", "677", "MuniAAA"): "MuniAAA .. MuniAAA",
        ("FIX.5.0", "677", "OTHER"): "OTHER .. OTHER",
        ("FIX.5.0", "677", "Pfandbriefe"): "Pfandbriefe .. Pfandbriefe",
        ("FIX.5.0", "677", "SONIA"): "SONIA .. SONIA",
        ("FIX.5.0", "677", "SWAP"): "SWAP .. SWAP",
        ("FIX.5.0", "677", "Treasury"): "Treasury .. Treasury",
        ("FIX.5.0", "677", "FEDEFF"): "FEDEFF .. FedFundRateEffective",
        ("FIX.5.0", "677", "FEDOPEN"): "FEDOPEN .. FedOpen",
        ("FIX.5.0", "677", "EURIBOR"): "EURIBOR .. EURIBOR",
        ("FIX.5.0", "680", "1"): "1 .. Percentage",
        ("FIX.5.0", "680", "2"): "2 .. PerUnit",
        ("FIX.5.0", "680", "3"): "3 .. FixedAmount",
        ("FIX.5.0", "680", "4"): "4 .. Discount",
        ("FIX.5.0", "680", "5"): "5 .. Premium",
        ("FIX.5.0", "680", "6"): "6 .. Spread",
        ("FIX.5.0", "680", "7"): "7 .. TEDPrice",
        ("FIX.5.0", "680", "8"): "8 .. TEDYield",
        ("FIX.5.0", "680", "9"): "9 .. Yield",
        ("FIX.5.0", "680", "10"): "10 .. FixedCabinetTradePrice",
        ("FIX.5.0", "680", "11"): "11 .. VariableCabinetTradePrice",
        ("FIX.5.0", "680", "12"): "12 .. PriceSpread",
        ("FIX.5.0", "680", "13"): "13 .. ProductTicksInHalves",
        ("FIX.5.0", "680", "14"): "14 .. ProductTicksInFourths",
        ("FIX.5.0", "680", "15"): "15 .. ProductTicksInEighths",
        ("FIX.5.0", "680", "16"): "16 .. ProductTicksInSixteenths",
        ("FIX.5.0", "680", "17"): "17 .. ProductTicksInThirtySeconds",
        ("FIX.5.0", "680", "18"): "18 .. ProductTicksInSixtyFourths",
        ("FIX.5.0", "680", "19"): "19 .. ProductTicksInOneTwentyEighths",
        ("FIX.5.0", "680", "20"): "20 .. NormalRateRepresentation",
        ("FIX.5.0", "680", "21"): "21 .. InverseRateRepresentation",
        ("FIX.5.0", "680", "22"): "22 .. BasisPoints",
        ("FIX.5.0", "680", "23"): "23 .. UpfrontPoints",
        ("FIX.5.0", "680", "24"): "24 .. InterestRate",
        ("FIX.5.0", "680", "25"): "25 .. PercentageNotional",
        ("FIX.5.0", "682", "S"): "S .. Small",
        ("FIX.5.0", "682", "M"): "M .. Medium",
        ("FIX.5.0", "682", "L"): "L .. Large",
        ("FIX.5.0", "682", "U"): "U .. UndisclosedQuantity",
        ("FIX.5.0", "686", "1"): "1 .. Percentage",
        ("FIX.5.0", "686", "2"): "2 .. PerUnit",
        ("FIX.5.0", "686", "3"): "3 .. FixedAmount",
        ("FIX.5.0", "686", "4"): "4 .. Discount",
        ("FIX.5.0", "686", "5"): "5 .. Premium",
        ("FIX.5.0", "686", "6"): "6 .. Spread",
        ("FIX.5.0", "686", "7"): "7 .. TEDPrice",
        ("FIX.5.0", "686", "8"): "8 .. TEDYield",
        ("FIX.5.0", "686", "9"): "9 .. Yield",
        ("FIX.5.0", "686", "10"): "10 .. FixedCabinetTradePrice",
        ("FIX.5.0", "686", "11"): "11 .. VariableCabinetTradePrice",
        ("FIX.5.0", "686", "12"): "12 .. PriceSpread",
        ("FIX.5.0", "686", "13"): "13 .. ProductTicksInHalves",
        ("FIX.5.0", "686", "14"): "14 .. ProductTicksInFourths",
        ("FIX.5.0", "686", "15"): "15 .. ProductTicksInEighths",
        ("FIX.5.0", "686", "16"): "16 .. ProductTicksInSixteenths",
        ("FIX.5.0", "686", "17"): "17 .. ProductTicksInThirtySeconds",
        ("FIX.5.0", "686", "18"): "18 .. ProductTicksInSixtyFourths",
        ("FIX.5.0", "686", "19"): "19 .. ProductTicksInOneTwentyEighths",
        ("FIX.5.0", "686", "20"): "20 .. NormalRateRepresentation",
        ("FIX.5.0", "686", "21"): "21 .. InverseRateRepresentation",
        ("FIX.5.0", "686", "22"): "22 .. BasisPoints",
        ("FIX.5.0", "686", "23"): "23 .. UpfrontPoints",
        ("FIX.5.0", "686", "24"): "24 .. InterestRate",
        ("FIX.5.0", "686", "25"): "25 .. PercentageNotional",
        ("FIX.5.0", "688", "AMT"): "AMT .. AlternativeMinimumTax",
        ("FIX.5.0", "688", "AUTOREINV"): "AUTOREINV .. AutoReinvestment",
        ("FIX.5.0", "688", "BANKQUAL"): "BANKQUAL .. BankQualified",
        ("FIX.5.0", "688", "BGNCON"): "BGNCON .. BargainConditions",
        ("FIX.5.0", "688", "COUPON"): "COUPON .. CouponRange",
        ("FIX.5.0", "688", "CURRENCY"): "CURRENCY .. ISOCurrencyCode",
        ("FIX.5.0", "688", "CUSTOMDATE"): "CUSTOMDATE .. CustomStart",
        ("FIX.5.0", "688", "GEOG"): "GEOG .. Geographics",
        ("FIX.5.0", "688", "HAIRCUT"): "HAIRCUT .. ValuationDiscount",
        ("FIX.5.0", "688", "INSURED"): "INSURED .. Insured",
        ("FIX.5.0", "688", "ISSUE"): "ISSUE .. IssueDate",
        ("FIX.5.0", "688", "ISSUER"): "ISSUER .. Issuer",
        ("FIX.5.0", "688", "ISSUESIZE"): "ISSUESIZE .. IssueSizeRange",
        ("FIX.5.0", "688", "LOOKBACK"): "LOOKBACK .. LookbackDays",
        ("FIX.5.0", "688", "LOT"): "LOT .. ExplicitLotIdentifier",
        ("FIX.5.0", "688", "LOTVAR"): "LOTVAR .. LotVariance",
        ("FIX.5.0", "688", "MAT"): "MAT .. MaturityYearAndMonth",
        ("FIX.5.0", "688", "MATURITY"): "MATURITY .. MaturityRange",
        ("FIX.5.0", "688", "MAXSUBS"): "MAXSUBS .. MaximumSubstitutions",
        ("FIX.5.0", "688", "MINDNOM"): "MINDNOM .. MinimumDenomination",
        ("FIX.5.0", "688", "MININCR"): "MININCR .. MinimumIncrement",
        ("FIX.5.0", "688", "MINQTY"): "MINQTY .. MinimumQuantity",
        ("FIX.5.0", "688", "PAYFREQ"): "PAYFREQ .. PaymentFrequency",
        ("FIX.5.0", "688", "PIECES"): "PIECES .. NumberOfPieces",
        ("FIX.5.0", "688", "PMAX"): "PMAX .. PoolsMaximum",
        ("FIX.5.0", "688", "PPL"): "PPL .. PoolsPerLot",
        ("FIX.5.0", "688", "PPM"): "PPM .. PoolsPerMillion",
        ("FIX.5.0", "688", "PPT"): "PPT .. PoolsPerTrade",
        ("FIX.5.0", "688", "PRICE"): "PRICE .. PriceRange",
        ("FIX.5.0", "688", "PRICEFREQ"): "PRICEFREQ .. PricingFrequency",
        ("FIX.5.0", "688", "PROD"): "PROD .. ProductionYear",
        ("FIX.5.0", "688", "PROTECT"): "PROTECT .. CallProtection",
        ("FIX.5.0", "688", "PURPOSE"): "PURPOSE .. Purpose",
        ("FIX.5.0", "688", "PXSOURCE"): "PXSOURCE .. BenchmarkPriceSource",
        ("FIX.5.0", "688", "RATING"): "RATING .. RatingSourceAndRange",
        ("FIX.5.0", "688", "REDEMPTION"): "REDEMPTION .. TypeOfRedemption",
        ("FIX.5.0", "688", "RESTRICTED"): "RESTRICTED .. Restricted",
        ("FIX.5.0", "688", "SECTOR"): "SECTOR .. MarketSector",
        ("FIX.5.0", "688", "SECTYPE"): "SECTYPE .. SecurityTypeIncludedOrExcluded",
        ("FIX.5.0", "688", "STRUCT"): "STRUCT .. Structure",
        ("FIX.5.0", "688", "SUBSFREQ"): "SUBSFREQ .. SubstitutionsFrequency",
        ("FIX.5.0", "688", "SUBSLEFT"): "SUBSLEFT .. SubstitutionsLeft",
        ("FIX.5.0", "688", "TEXT"): "TEXT .. FreeformText",
        ("FIX.5.0", "688", "TRDVAR"): "TRDVAR .. TradeVariance",
        ("FIX.5.0", "688", "WAC"): "WAC .. WeightedAverageCoupon",
        ("FIX.5.0", "688", "WAL"): "WAL .. WeightedAverageLifeCoupon",
        ("FIX.5.0", "688", "WALA"): "WALA .. WeightedAverageLoanAge",
        ("FIX.5.0", "688", "WAM"): "WAM .. WeightedAverageMaturity",
        ("FIX.5.0", "688", "WHOLE"): "WHOLE .. WholePool",
        ("FIX.5.0", "688", "YIELD"): "YIELD .. YieldRange",
        ("FIX.5.0", "688", "ORIGAMT"): "ORIGAMT .. OriginalAmount",
        ("FIX.5.0", "688", "POOLEFFDT"): "POOLEFFDT .. PoolEffectiveDate",
        ("FIX.5.0", "688", "POOLINITFCTR"): "POOLINITFCTR .. PoolInitialFactor",
        ("FIX.5.0", "688", "TRANCHE"): "TRANCHE .. Tranche",
        ("FIX.5.0", "688", "SUBSTITUTION"): "SUBSTITUTION .. Substitution",
        ("FIX.5.0", "688", "MULTEXCHFLLBCK"): "MULTEXCHFLLBCK .. MULTEXCHFLLBCK",
        ("FIX.5.0", "688", "COMPSECFLLBCK"): "COMPSECFLLBCK .. COMPSECFLLBCK",
        ("FIX.5.0", "688", "LOCLJRSDCTN"): "LOCLJRSDCTN .. LOCLJRSDCTN",
        ("FIX.5.0", "688", "RELVJRSDCTN"): "RELVJRSDCTN .. RELVJRSDCTN",
        ("FIX.5.0", "688", "INCURRCVY"): "INCURRCVY .. IncurredRecovery",
        ("FIX.5.0", "688", "ADDTRM"): "ADDTRM .. AdditionalTerm",
        ("FIX.5.0", "688", "MODEQTYDLVY"): "MODEQTYDLVY .. ModifiedEquityDelivery",
        ("FIX.5.0", "688", "NOREFOBLIG"): "NOREFOBLIG .. NoReferenceOblication",
        ("FIX.5.0", "688", "UNKREFOBLIG"): "UNKREFOBLIG .. UnknownReferenceObligation",
        ("FIX.5.0", "688", "ALLGUARANTEES"): "ALLGUARANTEES .. AllGuarantees",
        ("FIX.5.0", "688", "REFPX"): "REFPX .. ReferencePrice",
        ("FIX.5.0", "688", "REFPOLICY"): "REFPOLICY .. ReferencePolicy",
        ("FIX.5.0", "688", "SECRDLIST"): "SECRDLIST .. SecuredList",
        ("FIX.5.0", "688", "AVFICO"): "AVFICO .. AverageFICOScore",
        ("FIX.5.0", "688", "AVSIZE"): "AVSIZE .. AverageLoanSize",
        ("FIX.5.0", "688", "MAXBAL"): "MAXBAL .. MaximumLoanBalance",
        ("FIX.5.0", "688", "POOL"): "POOL .. PoolIdentifier",
        ("FIX.5.0", "688", "ROLLTYPE"): "ROLLTYPE .. TypeOfRollTrade",
        ("FIX.5.0", "688", "REFTRADE"): "REFTRADE .. ReferenceToRollingOrClosingTrade",
        ("FIX.5.0", "688", "REFPRIN"): "REFPRIN .. PrincipalOfRollingOrClosingTrade",
        ("FIX.5.0", "688", "REFINT"): "REFINT .. InterestOfRollingOrClosingTrade",
        ("FIX.5.0", "688", "AVAILQTY"): "AVAILQTY .. AvailableOfferQuantityToBeShownToTheStreet",
        ("FIX.5.0", "688", "BROKERCREDIT"): "BROKERCREDIT .. BrokerCredit",
        ("FIX.5.0", "688", "INTERNALPX"): "INTERNALPX .. OfferPriceToBeShownToInternalBrokers",
        ("FIX.5.0", "688", "INTERNALQTY"): "INTERNALQTY .. OfferQuantityToBeShownToInternalBrokers",
        ("FIX.5.0", "688", "LEAVEQTY"): "LEAVEQTY .. TheMinimumResidualOfferQuantity",
        ("FIX.5.0", "688", "MAXORDQTY"): "MAXORDQTY .. MaximumOrderSize",
        ("FIX.5.0", "688", "ORDRINCR"): "ORDRINCR .. OrderQuantityIncrement",
        ("FIX.5.0", "688", "PRIMARY"): "PRIMARY .. PrimaryOrSecondaryMarketIndicator",
        ("FIX.5.0", "688", "SALESCREDITOVR"): "SALESCREDITOVR .. BrokerSalesCreditOverride",
        ("FIX.5.0", "688", "TRADERCREDIT"): "TRADERCREDIT .. TraderCredit",
        ("FIX.5.0", "688", "DISCOUNT"): "DISCOUNT .. DiscountRate",
        ("FIX.5.0", "688", "YTM"): "YTM .. YieldToMaturity",
        ("FIX.5.0", "688", "ABS"): "ABS .. AbsolutePrepaymentSpeed",
        ("FIX.5.0", "688", "CPP"): "CPP .. ConstantPrepaymentPenalty",
        ("FIX.5.0", "688", "CPR"): "CPR .. ConstantPrepaymentRate",
        ("FIX.5.0", "688", "CPY"): "CPY .. ConstantPrepaymentYield",
        ("FIX.5.0", "688", "HEP"): "HEP .. FinalCPROfHomeEquityPrepaymentCurve",
        ("FIX.5.0", "688", "MHP"): "MHP .. PercentOfManufacturedHousingPrepaymentCurve",
        ("FIX.5.0", "688", "MPR"): "MPR .. MonthlyPrepaymentRate",
        ("FIX.5.0", "688", "PPC"): "PPC .. PercentOfProspectusPrepaymentCurve",
        ("FIX.5.0", "688", "PSA"): "PSA .. PercentOfBMAPrepaymentCurve",
        ("FIX.5.0", "688", "SMM"): "SMM .. SingleMonthlyMortality",
        ("FIX.5.0", "690", "1"): "1 .. ParForPar",
        ("FIX.5.0", "690", "2"): "2 .. ModifiedDuration",
        ("FIX.5.0", "690", "4"): "4 .. Risk",
        ("FIX.5.0", "690", "5"): "5 .. Proceeds",
        ("FIX.5.0", "692", "1"): "1 .. Percent",
        ("FIX.5.0", "692", "2"): "2 .. PerShare",
        ("FIX.5.0", "692", "3"): "3 .. FixedAmount",
        ("FIX.5.0", "692", "4"): "4 .. Discount",
        ("FIX.5.0", "692", "5"): "5 .. Premium",
        ("FIX.5.0", "692", "6"): "6 .. Spread",
        ("FIX.5.0", "692", "7"): "7 .. TEDPrice",
        ("FIX.5.0", "692", "8"): "8 .. TEDYield",
        ("FIX.5.0", "692", "9"): "9 .. YieldSpread",
        ("FIX.5.0", "692", "10"): "10 .. Yield",
        ("FIX.5.0", "692", "12"): "12 .. PriceSpread",
        ("FIX.5.0", "692", "13"): "13 .. ProductTicksInHalves",
        ("FIX.5.0", "692", "14"): "14 .. ProductTicksInFourths",
        ("FIX.5.0", "692", "15"): "15 .. ProductTicksInEighths",
        ("FIX.5.0", "692", "16"): "16 .. ProductTicksInSixteenths",
        ("FIX.5.0", "692", "17"): "17 .. ProductTicksInThirtySeconds",
        ("FIX.5.0", "692", "18"): "18 .. ProductTicksInSixtyFourths",
        ("FIX.5.0", "692", "19"): "19 .. ProductTicksInOneTwentyEighths",
        ("FIX.5.0", "692", "20"): "20 .. NormalRateRepresentation",
        ("FIX.5.0", "692", "21"): "21 .. InverseRateRepresentation",
        ("FIX.5.0", "692", "22"): "22 .. BasisPoints",
        ("FIX.5.0", "692", "23"): "23 .. UpFrontPoints",
        ("FIX.5.0", "692", "24"): "24 .. InterestRate",
        ("FIX.5.0", "692", "25"): "25 .. PercentageOfNotional",
        ("FIX.5.0", "694", "1"): "1 .. Hit",
        ("FIX.5.0", "694", "2"): "2 .. Counter",
        ("FIX.5.0", "694", "3"): "3 .. Expired",
        ("FIX.5.0", "694", "4"): "4 .. Cover",
        ("FIX.5.0", "694", "5"): "5 .. DoneAway",
        ("FIX.5.0", "694", "6"): "6 .. Pass",
        ("FIX.5.0", "694", "7"): "7 .. EndTrade",
        ("FIX.5.0", "694", "8"): "8 .. TimedOut",
        ("FIX.5.0", "694", "9"): "9 .. Tied",
        ("FIX.5.0", "694", "10"): "10 .. TiedCover",
        ("FIX.5.0", "695", "A"): "A .. AllOrNone",
        ("FIX.5.0", "695", "B"): "B .. MarketOnClose",
        ("FIX.5.0", "695", "C"): "C .. AtTheClose",
        ("FIX.5.0", "695", "D"): "D .. VWAP",
        ("FIX.5.0", "695", "E"): "E .. Axe",
        ("FIX.5.0", "695", "F"): "F .. AxeOnBid",
        ("FIX.5.0", "695", "G"): "G .. AxeOnOffer",
        ("FIX.5.0", "695", "H"): "H .. ClientNaturalWorking",
        ("FIX.5.0", "695", "I"): "I .. InTouchWith",
        ("FIX.5.0", "695", "J"): "J .. PositionWanted",
        ("FIX.5.0", "695", "K"): "K .. MarketMaking",
        ("FIX.5.0", "695", "L"): "L .. Limit",
        ("FIX.5.0", "695", "M"): "M .. MoreBehind",
        ("FIX.5.0", "695", "N"): "N .. ClientNaturalBlock",
        ("FIX.5.0", "695", "O"): "O .. AtTheOpen",
        ("FIX.5.0", "695", "P"): "P .. TakingAPosition",
        ("FIX.5.0", "695", "Q"): "Q .. AtTheMarket",
        ("FIX.5.0", "695", "R"): "R .. ReadyToTrade",
        ("FIX.5.0", "695", "S"): "S .. PortfolioShown",
        ("FIX.5.0", "695", "T"): "T .. ThroughTheDay",
        ("FIX.5.0", "695", "U"): "U .. Unwind",
        ("FIX.5.0", "695", "V"): "V .. Versus",
        ("FIX.5.0", "695", "W"): "W .. Indication",
        ("FIX.5.0", "695", "X"): "X .. CrossingOpportunity",
        ("FIX.5.0", "695", "Y"): "Y .. AtTheMidpoint",
        ("FIX.5.0", "695", "Z"): "Z .. PreOpen",
        ("FIX.5.0", "695", "1"): "1 .. QuantityNegotiable",
        ("FIX.5.0", "695", "2"): "2 .. AllowLateBids",
        ("FIX.5.0", "695", "3"): "3 .. ImmediateOrCounter",
        ("FIX.5.0", "695", "4"): "4 .. AutoTrade",
        ("FIX.5.0", "695", "a"): "a .. AutomaticSpot",
        ("FIX.5.0", "695", "b"): "b .. PlatformCalculatedSpot",
        ("FIX.5.0", "695", "c"): "c .. OutsideSpread",
        ("FIX.5.0", "695", "d"): "d .. DeferredSpot",
        ("FIX.5.0", "695", "n"): "n .. NegotiatedSpot",
        ("FIX.5.0", "698", "1"): "1 .. Percentage",
        ("FIX.5.0", "698", "2"): "2 .. PerUnit",
        ("FIX.5.0", "698", "3"): "3 .. FixedAmount",
        ("FIX.5.0", "698", "4"): "4 .. Discount",
        ("FIX.5.0", "698", "5"): "5 .. Premium",
        ("FIX.5.0", "698", "6"): "6 .. Spread",
        ("FIX.5.0", "698", "7"): "7 .. TEDPrice",
        ("FIX.5.0", "698", "8"): "8 .. TEDYield",
        ("FIX.5.0", "698", "9"): "9 .. Yield",
        ("FIX.5.0", "698", "10"): "10 .. FixedCabinetTradePrice",
        ("FIX.5.0", "698", "11"): "11 .. VariableCabinetTradePrice",
        ("FIX.5.0", "698", "12"): "12 .. PriceSpread",
        ("FIX.5.0", "698", "13"): "13 .. ProductTicksInHalves",
        ("FIX.5.0", "698", "14"): "14 .. ProductTicksInFourths",
        ("FIX.5.0", "698", "15"): "15 .. ProductTicksInEighths",
        ("FIX.5.0", "698", "16"): "16 .. ProductTicksInSixteenths",
        ("FIX.5.0", "698", "17"): "17 .. ProductTicksInThirtySeconds",
        ("FIX.5.0", "698", "18"): "18 .. ProductTicksInSixtyFourths",
        ("FIX.5.0", "698", "19"): "19 .. ProductTicksInOneTwentyEighths",
        ("FIX.5.0", "698", "20"): "20 .. NormalRateRepresentation",
        ("FIX.5.0", "698", "21"): "21 .. InverseRateRepresentation",
        ("FIX.5.0", "698", "22"): "22 .. BasisPoints",
        ("FIX.5.0", "698", "23"): "23 .. UpfrontPoints",
        ("FIX.5.0", "698", "24"): "24 .. InterestRate",
        ("FIX.5.0", "698", "25"): "25 .. PercentageNotional",
        ("FIX.5.0", "703", "ALC"): "ALC .. AllocationTradeQty",
        ("FIX.5.0", "703", "AS"): "AS .. OptionAssignment",
        ("FIX.5.0", "703", "ASF"): "ASF .. AsOfTradeQty",
        ("FIX.5.0", "703", "DLV"): "DLV .. DeliveryQty",
        ("FIX.5.0", "703", "ETR"): "ETR .. ElectronicTradeQty",
        ("FIX.5.0", "703", "EX"): "EX .. OptionExerciseQty",
        ("FIX.5.0", "703", "FIN"): "FIN .. EndOfDayQty",
        ("FIX.5.0", "703", "IAS"): "IAS .. IntraSpreadQty",
        ("FIX.5.0", "703", "IES"): "IES .. InterSpreadQty",
        ("FIX.5.0", "703", "PA"): "PA .. AdjustmentQty",
        ("FIX.5.0", "703", "PIT"): "PIT .. PitTradeQty",
        ("FIX.5.0", "703", "SOD"): "SOD .. StartOfDayQty",
        ("FIX.5.0", "703", "SPL"): "SPL .. IntegralSplit",
        ("FIX.5.0", "703", "TA"): "TA .. TransactionFromAssignment",
        ("FIX.5.0", "703", "TOT"): "TOT .. TotalTransactionQty",
        ("FIX.5.0", "703", "TQ"): "TQ .. TransactionQuantity",
        ("FIX.5.0", "703", "TRF"): "TRF .. TransferTradeQty",
        ("FIX.5.0", "703", "TX"): "TX .. TransactionFromExercise",
        ("FIX.5.0", "703", "XM"): "XM .. CrossMarginQty",
        ("FIX.5.0", "703", "RCV"): "RCV .. ReceiveQuantity",
        ("FIX.5.0", "703", "CAA"): "CAA .. CorporateActionAdjustment",
        ("FIX.5.0", "703", "DN"): "DN .. DeliveryNoticeQty",
        ("FIX.5.0", "703", "EP"): "EP .. ExchangeForPhysicalQty",
        ("FIX.5.0", "703", "PNTN"): "PNTN .. PrivatelyNegotiatedTradeQty",
        ("FIX.5.0", "703", "DLT"): "DLT .. NetDeltaQty",
        ("FIX.5.0", "703", "CEA"): "CEA .. CreditEventAdjustment",
        ("FIX.5.0", "703", "SEA"): "SEA .. SuccessionEventAdjustment",
        ("FIX.5.0", "703", "NET"): "NET .. NetQty",
        ("FIX.5.0", "703", "GRS"): "GRS .. GrossQty",
        ("FIX.5.0", "703", "ITD"): "ITD .. IntradayQty",
        ("FIX.5.0", "703", "NDAS"): "NDAS .. GrossLongNonDeltaAdjustedSwaptionPosition",
        ("FIX.5.0", "703", "DAS"): "DAS .. LongDeltaAdjustedPairedSwaptionPosition",
        ("FIX.5.0", "703", "EXP"): "EXP .. ExpiringQuantity",
        ("FIX.5.0", "703", "UNEX"): "UNEX .. QuantityNotExercised",
        ("FIX.5.0", "703", "REQ"): "REQ .. RequestedExerciseQuantity",
        ("FIX.5.0", "703", "CFE"): "CFE .. CashFuturesEquivalentQuantity",
        ("FIX.5.0", "706", "0"): "0 .. Submitted",
        ("FIX.5.0", "706", "1"): "1 .. Accepted",
        ("FIX.5.0", "706", "2"): "2 .. Rejected",
        ("FIX.5.0", "707", "CASH"): "CASH .. CashAmount",
        ("FIX.5.0", "707", "CRES"): "CRES .. CashResidualAmount",
        ("FIX.5.0", "707", "FMTM"): "FMTM .. FinalMarkToMarketAmount",
        ("FIX.5.0", "707", "IMTM"): "IMTM .. IncrementalMarkToMarketAmount",
        ("FIX.5.0", "707", "PREM"): "PREM .. PremiumAmount",
        ("FIX.5.0", "707", "SMTM"): "SMTM .. StartOfDayMarkToMarketAmount",
        ("FIX.5.0", "707", "TVAR"): "TVAR .. TradeVariationAmount",
        ("FIX.5.0", "707", "VADJ"): "VADJ .. ValueAdjustedAmount",
        ("FIX.5.0", "707", "SETL"): "SETL .. SettlementValue",
        ("FIX.5.0", "707", "ICPN"): "ICPN .. InitialTradeCouponAmount",
        ("FIX.5.0", "707", "ACPN"): "ACPN .. AccruedCouponAmount",
        ("FIX.5.0", "707", "CPN"): "CPN .. CouponAmount",
        ("FIX.5.0", "707", "IACPN"): "IACPN .. IncrementalAccruedCoupon",
        ("FIX.5.0", "707", "CMTM"): "CMTM .. CollateralizedMarkToMarket",
        ("FIX.5.0", "707", "ICMTM"): "ICMTM .. IncrementalCollateralizedMarkToMarket",
        ("FIX.5.0", "707", "DLV"): "DLV .. CompensationAmount",
        ("FIX.5.0", "707", "BANK"): "BANK .. TotalBankedAmount",
        ("FIX.5.0", "707", "COLAT"): "COLAT .. TotalCollateralizedAmount",
        ("FIX.5.0", "707", "LSNV"): "LSNV .. LongPairedSwapNotionalValue",
        ("FIX.5.0", "707", "SSNV"): "SSNV .. ShortPairedSwapNotionalValue",
        ("FIX.5.0", "707", "SACPN"): "SACPN .. StartOfDayAccruedCoupon",
        ("FIX.5.0", "707", "NPV"): "NPV .. NetPresentValue",
        ("FIX.5.0", "707", "SNPV"): "SNPV .. StartOfDayNetPresentValue",
        ("FIX.5.0", "707", "NCF"): "NCF .. NetCashFlow",
        ("FIX.5.0", "707", "PVFEES"): "PVFEES .. PresentValueOfFees",
        ("FIX.5.0", "707", "PV01"): "PV01 .. PresentValueOneBasisPoints",
        ("FIX.5.0", "707", "5YREN"): "5YREN .. FiveYearEquivalentNotional",
        ("FIX.5.0", "707", "UMTM"): "UMTM .. UndiscountedMarkToMarket",
        ("FIX.5.0", "707", "MTD"): "MTD .. MarkToModel",
        ("FIX.5.0", "707", "VMTM"): "VMTM .. MarkToMarketVariance",
        ("FIX.5.0", "707", "VMTD"): "VMTD .. MarkToModelVariance",
        ("FIX.5.0", "707", "UPFRNT"): "UPFRNT .. UpfrontPayment",
        ("FIX.5.0", "709", "1"): "1 .. Exercise",
        ("FIX.5.0", "709", "2"): "2 .. DoNotExercise",
        ("FIX.5.0", "709", "3"): "3 .. PositionAdjustment",
        ("FIX.5.0", "709", "4"): "4 .. PositionChangeSubmission",
        ("FIX.5.0", "709", "5"): "5 .. Pledge",
        ("FIX.5.0", "709", "6"): "6 .. LargeTraderSubmission",
        ("FIX.5.0", "709", "7"): "7 .. LargePositionsReportingSubmission",
        ("FIX.5.0", "709", "8"): "8 .. LongHoldings",
        ("FIX.5.0", "709", "9"): "9 .. InternalTransfer",
        ("FIX.5.0", "709", "10"): "10 .. TransferOfFirm",
        ("FIX.5.0", "709", "11"): "11 .. ExternalTransfer",
        ("FIX.5.0", "709", "12"): "12 .. CorporateAction",
        ("FIX.5.0", "709", "13"): "13 .. Notification",
        ("FIX.5.0", "709", "14"): "14 .. PositionCreation",
        ("FIX.5.0", "709", "15"): "15 .. Closeout",
        ("FIX.5.0", "709", "16"): "16 .. Reopen",
        ("FIX.5.0", "712", "1"): "1 .. New",
        ("FIX.5.0", "712", "2"): "2 .. Replace",
        ("FIX.5.0", "712", "3"): "3 .. Cancel",
        ("FIX.5.0", "712", "4"): "4 .. Reverse",
        ("FIX.5.0", "716", "ITD"): "ITD .. Intraday",
        ("FIX.5.0", "716", "RTH"): "RTH .. RegularTradingHours",
        ("FIX.5.0", "716", "ETH"): "ETH .. ElectronicTradingHours",
        ("FIX.5.0", "716", "EOD"): "EOD .. EndOfDay",
        ("FIX.5.0", "718", "0"): "0 .. ProcessRequestAsMarginDisposition",
        ("FIX.5.0", "718", "1"): "1 .. DeltaPlus",
        ("FIX.5.0", "718", "2"): "2 .. DeltaMinus",
        ("FIX.5.0", "718", "3"): "3 .. Final",
        ("FIX.5.0", "718", "4"): "4 .. CustomerSpecificPosition",
        ("FIX.5.0", "722", "0"): "0 .. Accepted",
        ("FIX.5.0", "722", "1"): "1 .. AcceptedWithWarnings",
        ("FIX.5.0", "722", "2"): "2 .. Rejected",
        ("FIX.5.0", "722", "3"): "3 .. Completed",
        ("FIX.5.0", "722", "4"): "4 .. CompletedWithWarnings",
        ("FIX.5.0", "723", "0"): "0 .. SuccessfulCompletion",
        ("FIX.5.0", "723", "1"): "1 .. Rejected",
        ("FIX.5.0", "723", "99"): "99 .. Other",
        ("FIX.5.0", "724", "0"): "0 .. Positions",
        ("FIX.5.0", "724", "1"): "1 .. Trades",
        ("FIX.5.0", "724", "2"): "2 .. Exercises",
        ("FIX.5.0", "724", "3"): "3 .. Assignments",
        ("FIX.5.0", "724", "4"): "4 .. SettlementActivity",
        ("FIX.5.0", "724", "5"): "5 .. BackoutMessage",
        ("FIX.5.0", "724", "6"): "6 .. DeltaPositions",
        ("FIX.5.0", "724", "7"): "7 .. NetPosition",
        ("FIX.5.0", "724", "8"): "8 .. LargePositionsReporting",
        ("FIX.5.0", "724", "9"): "9 .. ExercisePositionReportingSubmission",
        ("FIX.5.0", "724", "10"): "10 .. PositionLimitReportingSubmissing",
        ("FIX.5.0", "725", "0"): "0 .. Inband",
        ("FIX.5.0", "725", "1"): "1 .. OutOfBand",
        ("FIX.5.0", "728", "0"): "0 .. ValidRequest",
        ("FIX.5.0", "728", "1"): "1 .. InvalidOrUnsupportedRequest",
        ("FIX.5.0", "728", "2"): "2 .. NoPositionsFoundThatMatchCriteria",
        ("FIX.5.0", "728", "3"): "3 .. NotAuthorizedToRequestPositions",
        ("FIX.5.0", "728", "4"): "4 .. RequestForPositionNotSupported",
        ("FIX.5.0", "728", "99"): "99 .. Other",
        ("FIX.5.0", "729", "0"): "0 .. Completed",
        ("FIX.5.0", "729", "1"): "1 .. CompletedWithWarnings",
        ("FIX.5.0", "729", "2"): "2 .. Rejected",
        ("FIX.5.0", "731", "1"): "1 .. Final",
        ("FIX.5.0", "731", "2"): "2 .. Theoretical",
        ("FIX.5.0", "733", "1"): "1 .. Final",
        ("FIX.5.0", "733", "2"): "2 .. Theoretical",
        ("FIX.5.0", "744", "P"): "P .. ProRata",
        ("FIX.5.0", "744", "R"): "R .. Random",
        ("FIX.5.0", "747", "A"): "A .. Automatic",
        ("FIX.5.0", "747", "M"): "M .. Manual",
        ("FIX.5.0", "749", "0"): "0 .. Successful",
        ("FIX.5.0", "749", "1"): "1 .. InvalidOrUnknownInstrument",
        ("FIX.5.0", "749", "2"): "2 .. InvalidTypeOfTradeRequested",
        ("FIX.5.0", "749", "3"): "3 .. InvalidParties",
        ("FIX.5.0", "749", "4"): "4 .. InvalidTransportTypeRequested",
        ("FIX.5.0", "749", "5"): "5 .. InvalidDestinationRequested",
        ("FIX.5.0", "749", "8"): "8 .. TradeRequestTypeNotSupported",
        ("FIX.5.0", "749", "9"): "9 .. NotAuthorized",
        ("FIX.5.0", "749", "99"): "99 .. Other",
        ("FIX.5.0", "750", "0"): "0 .. Accepted",
        ("FIX.5.0", "750", "1"): "1 .. Completed",
        ("FIX.5.0", "750", "2"): "2 .. Rejected",
        ("FIX.5.0", "751", "0"): "0 .. Successful",
        ("FIX.5.0", "751", "1"): "1 .. InvalidPartyInformation",
        ("FIX.5.0", "751", "2"): "2 .. UnknownInstrument",
        ("FIX.5.0", "751", "3"): "3 .. UnauthorizedToReportTrades",
        ("FIX.5.0", "751", "4"): "4 .. InvalidTradeType",
        ("FIX.5.0", "751", "5"): "5 .. PriceExceedsCurrentPriceBand",
        ("FIX.5.0", "751", "6"): "6 .. ReferencePriceNotAvailable",
        ("FIX.5.0", "751", "7"): "7 .. NotionalValueExceedsThreshold",
        ("FIX.5.0", "751", "99"): "99 .. Other",
        ("FIX.5.0", "752", "1"): "1 .. SingleSecurity",
        ("FIX.5.0", "752", "2"): "2 .. IndividualLegOfAMultilegSecurity",
        ("FIX.5.0", "752", "3"): "3 .. MultilegSecurity",
        ("FIX.5.0", "758", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "758", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "758", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "758", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "758", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "758", "J"): "J .. TaxID",
        ("FIX.5.0", "758", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "758", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "758", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "758", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "758", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "758", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "758", "B"): "B .. BIC",
        ("FIX.5.0", "758", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "758", "D"): "D .. Proprietary",
        ("FIX.5.0", "758", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "758", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "758", "G"): "G .. MIC",
        ("FIX.5.0", "758", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "758", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "758", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "758", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "758", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "758", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "758", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "758", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "759", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "759", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "759", "3"): "3 .. ClientID",
        ("FIX.5.0", "759", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "759", "5"): "5 .. InvestorID",
        ("FIX.5.0", "759", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "759", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "759", "8"): "8 .. Locate",
        ("FIX.5.0", "759", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "759", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "759", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "759", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "759", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "759", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "759", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "759", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "759", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "759", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "759", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "759", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "759", "22"): "22 .. Exchange",
        ("FIX.5.0", "759", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "759", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "759", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "759", "27"): "27 .. Buyer",
        ("FIX.5.0", "759", "28"): "28 .. Custodian",
        ("FIX.5.0", "759", "29"): "29 .. Intermediary",
        ("FIX.5.0", "759", "30"): "30 .. Agent",
        ("FIX.5.0", "759", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "759", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "759", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "759", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "759", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "759", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "759", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "759", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "759", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "759", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "759", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "759", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "759", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "759", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "759", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "759", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "759", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "759", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "759", "49"): "49 .. AssetManager",
        ("FIX.5.0", "759", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "759", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "759", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "759", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "759", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "759", "55"): "55 .. SessionID",
        ("FIX.5.0", "759", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "759", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "759", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "759", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "759", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "759", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "759", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "759", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "759", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "759", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "759", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "759", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "759", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "759", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "759", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "759", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "759", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "759", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "759", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "759", "75"): "75 .. LocationID",
        ("FIX.5.0", "759", "76"): "76 .. DeskID",
        ("FIX.5.0", "759", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "759", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "759", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "759", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "759", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "759", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "759", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "759", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "759", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "759", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "759", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "759", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "759", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "759", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "759", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "759", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "759", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "759", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "759", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "759", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "759", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "759", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "759", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "759", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "759", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "759", "102"): "102 .. DataRepository",
        ("FIX.5.0", "759", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "759", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "759", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "759", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "759", "107"): "107 .. Correspondent",
        ("FIX.5.0", "759", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "759", "110"): "110 .. Borrower",
        ("FIX.5.0", "759", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "759", "112"): "112 .. Guarantor",
        ("FIX.5.0", "759", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "759", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "759", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "759", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "759", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "759", "118"): "118 .. Operator",
        ("FIX.5.0", "759", "119"): "119 .. CSD",
        ("FIX.5.0", "759", "120"): "120 .. ICSD",
        ("FIX.5.0", "759", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "759", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "759", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "761", "1"): "1 .. CUSIP",
        ("FIX.5.0", "761", "2"): "2 .. SEDOL",
        ("FIX.5.0", "761", "3"): "3 .. QUIK",
        ("FIX.5.0", "761", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "761", "5"): "5 .. RICCode",
        ("FIX.5.0", "761", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "761", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "761", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "761", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "761", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "761", "B"): "B .. Wertpapier",
        ("FIX.5.0", "761", "C"): "C .. Dutch",
        ("FIX.5.0", "761", "D"): "D .. Valoren",
        ("FIX.5.0", "761", "E"): "E .. Sicovam",
        ("FIX.5.0", "761", "F"): "F .. Belgian",
        ("FIX.5.0", "761", "G"): "G .. Common",
        ("FIX.5.0", "761", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "761", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "761", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "761", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "761", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "761", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "761", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "761", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "761", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "761", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "761", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "761", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "761", "U"): "U .. Synthetic",
        ("FIX.5.0", "761", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "761", "W"): "W .. IndexName",
        ("FIX.5.0", "770", "1"): "1 .. ExecutionTime",
        ("FIX.5.0", "770", "2"): "2 .. TimeIn",
        ("FIX.5.0", "770", "3"): "3 .. TimeOut",
        ("FIX.5.0", "770", "4"): "4 .. BrokerReceipt",
        ("FIX.5.0", "770", "5"): "5 .. BrokerExecution",
        ("FIX.5.0", "770", "6"): "6 .. DeskReceipt",
        ("FIX.5.0", "770", "7"): "7 .. SubmissionToClearing",
        ("FIX.5.0", "770", "8"): "8 .. TimePriority",
        ("FIX.5.0", "770", "9"): "9 .. OrderbookEntryTime",
        ("FIX.5.0", "770", "10"): "10 .. OrderSubmissionTime",
        ("FIX.5.0", "770", "11"): "11 .. PubliclyReported",
        ("FIX.5.0", "770", "12"): "12 .. PublicReportUpdated",
        ("FIX.5.0", "770", "13"): "13 .. NonPubliclyReported",
        ("FIX.5.0", "770", "14"): "14 .. NonPublicReportUpdated",
        ("FIX.5.0", "770", "15"): "15 .. SubmittedForConfirmation",
        ("FIX.5.0", "770", "16"): "16 .. UpdatedForConfirmation",
        ("FIX.5.0", "770", "17"): "17 .. Confirmed",
        ("FIX.5.0", "770", "18"): "18 .. UpdatedForClearing",
        ("FIX.5.0", "770", "19"): "19 .. Cleared",
        ("FIX.5.0", "770", "20"): "20 .. AllocationsSubmitted",
        ("FIX.5.0", "770", "21"): "21 .. AllocationsUpdated",
        ("FIX.5.0", "770", "22"): "22 .. AllocationsCompleted",
        ("FIX.5.0", "770", "23"): "23 .. SubmittedToRepository",
        ("FIX.5.0", "770", "24"): "24 .. PostTrdContntnEvnt",
        ("FIX.5.0", "770", "25"): "25 .. PostTradeValuation",
        ("FIX.5.0", "770", "26"): "26 .. PreviousTimePriority",
        ("FIX.5.0", "773", "1"): "1 .. Status",
        ("FIX.5.0", "773", "2"): "2 .. Confirmation",
        ("FIX.5.0", "773", "3"): "3 .. ConfirmationRequestRejected",
        ("FIX.5.0", "774", "1"): "1 .. MismatchedAccount",
        ("FIX.5.0", "774", "2"): "2 .. MissingSettlementInstructions",
        ("FIX.5.0", "774", "3"): "3 .. UnknownOrMissingIndividualAllocId",
        ("FIX.5.0", "774", "4"): "4 .. TransactionNotRecognized",
        ("FIX.5.0", "774", "5"): "5 .. DuplicateTransaction",
        ("FIX.5.0", "774", "6"): "6 .. IncorrectOrMissingInstrument",
        ("FIX.5.0", "774", "7"): "7 .. IncorrectOrMissingPrice",
        ("FIX.5.0", "774", "8"): "8 .. IncorrectOrMissingCommission",
        ("FIX.5.0", "774", "9"): "9 .. IncorrectOrMissingSettlDate",
        ("FIX.5.0", "774", "10"): "10 .. IncorrectOrMissingFundIDOrFundName",
        ("FIX.5.0", "774", "11"): "11 .. IncorrectOrMissingQuantity",
        ("FIX.5.0", "774", "12"): "12 .. IncorrectOrMissingFees",
        ("FIX.5.0", "774", "13"): "13 .. IncorrectOrMissingTax",
        ("FIX.5.0", "774", "14"): "14 .. IncorrectOrMissingParty",
        ("FIX.5.0", "774", "15"): "15 .. IncorrectOrMissingSide",
        ("FIX.5.0", "774", "16"): "16 .. IncorrectOrMissingNetMoney",
        ("FIX.5.0", "774", "17"): "17 .. IncorrectOrMissingTradeDate",
        ("FIX.5.0", "774", "18"): "18 .. IncorrectOrMissingSettlCcyInstructions",
        ("FIX.5.0", "774", "19"): "19 .. IncorrectOrMissingCapacity",
        ("FIX.5.0", "774", "99"): "99 .. Other",
        ("FIX.5.0", "775", "0"): "0 .. RegularBooking",
        ("FIX.5.0", "775", "1"): "1 .. CFD",
        ("FIX.5.0", "775", "2"): "2 .. TotalReturnSwap",
        ("FIX.5.0", "776", "0"): "0 .. UnknownAccount",
        ("FIX.5.0", "776", "1"): "1 .. IncorrectQuantity",
        ("FIX.5.0", "776", "2"): "2 .. IncorrectAveragegPrice",
        ("FIX.5.0", "776", "3"): "3 .. UnknownExecutingBrokerMnemonic",
        ("FIX.5.0", "776", "4"): "4 .. CommissionDifference",
        ("FIX.5.0", "776", "5"): "5 .. UnknownOrderID",
        ("FIX.5.0", "776", "6"): "6 .. UnknownListID",
        ("FIX.5.0", "776", "7"): "7 .. OtherSeeText",
        ("FIX.5.0", "776", "8"): "8 .. IncorrectAllocatedQuantity",
        ("FIX.5.0", "776", "9"): "9 .. CalculationDifference",
        ("FIX.5.0", "776", "10"): "10 .. UnknownOrStaleExecID",
        ("FIX.5.0", "776", "11"): "11 .. MismatchedData",
        ("FIX.5.0", "776", "12"): "12 .. UnknownClOrdID",
        ("FIX.5.0", "776", "13"): "13 .. WarehouseRequestRejected",
        ("FIX.5.0", "776", "14"): "14 .. DuplicateOrMissingIndividualAllocId",
        ("FIX.5.0", "776", "15"): "15 .. TradeNotRecognized",
        ("FIX.5.0", "776", "16"): "16 .. DuplicateTrade",
        ("FIX.5.0", "776", "17"): "17 .. IncorrectOrMissingInstrument",
        ("FIX.5.0", "776", "18"): "18 .. IncorrectOrMissingSettlDate",
        ("FIX.5.0", "776", "19"): "19 .. IncorrectOrMissingFundIDOrFundName",
        ("FIX.5.0", "776", "20"): "20 .. IncorrectOrMissingSettlInstructions",
        ("FIX.5.0", "776", "21"): "21 .. IncorrectOrMissingFees",
        ("FIX.5.0", "776", "22"): "22 .. IncorrectOrMissingTax",
        ("FIX.5.0", "776", "23"): "23 .. UnknownOrMissingParty",
        ("FIX.5.0", "776", "24"): "24 .. IncorrectOrMissingSide",
        ("FIX.5.0", "776", "25"): "25 .. IncorrectOrMissingNetMoney",
        ("FIX.5.0", "776", "26"): "26 .. IncorrectOrMissingTradeDate",
        ("FIX.5.0", "776", "27"): "27 .. IncorrectOrMissingSettlCcyInstructions",
        ("FIX.5.0", "776", "28"): "28 .. IncorrectOrMissingProcessCode",
        ("FIX.5.0", "776", "99"): "99 .. Other",
        ("FIX.5.0", "780", "0"): "0 .. UseDefaultInstructions",
        ("FIX.5.0", "780", "1"): "1 .. DeriveFromParametersProvided",
        ("FIX.5.0", "780", "2"): "2 .. FullDetailsProvided",
        ("FIX.5.0", "780", "3"): "3 .. SSIDBIDsProvided",
        ("FIX.5.0", "780", "4"): "4 .. PhoneForInstructions",
        ("FIX.5.0", "783", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "783", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "783", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "783", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "783", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "783", "J"): "J .. TaxID",
        ("FIX.5.0", "783", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "783", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "783", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "783", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "783", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "783", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "783", "B"): "B .. BIC",
        ("FIX.5.0", "783", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "783", "D"): "D .. Proprietary",
        ("FIX.5.0", "783", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "783", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "783", "G"): "G .. MIC",
        ("FIX.5.0", "783", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "783", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "783", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "783", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "783", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "783", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "783", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "783", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "784", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "784", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "784", "3"): "3 .. ClientID",
        ("FIX.5.0", "784", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "784", "5"): "5 .. InvestorID",
        ("FIX.5.0", "784", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "784", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "784", "8"): "8 .. Locate",
        ("FIX.5.0", "784", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "784", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "784", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "784", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "784", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "784", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "784", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "784", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "784", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "784", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "784", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "784", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "784", "22"): "22 .. Exchange",
        ("FIX.5.0", "784", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "784", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "784", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "784", "27"): "27 .. Buyer",
        ("FIX.5.0", "784", "28"): "28 .. Custodian",
        ("FIX.5.0", "784", "29"): "29 .. Intermediary",
        ("FIX.5.0", "784", "30"): "30 .. Agent",
        ("FIX.5.0", "784", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "784", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "784", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "784", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "784", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "784", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "784", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "784", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "784", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "784", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "784", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "784", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "784", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "784", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "784", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "784", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "784", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "784", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "784", "49"): "49 .. AssetManager",
        ("FIX.5.0", "784", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "784", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "784", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "784", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "784", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "784", "55"): "55 .. SessionID",
        ("FIX.5.0", "784", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "784", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "784", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "784", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "784", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "784", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "784", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "784", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "784", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "784", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "784", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "784", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "784", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "784", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "784", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "784", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "784", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "784", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "784", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "784", "75"): "75 .. LocationID",
        ("FIX.5.0", "784", "76"): "76 .. DeskID",
        ("FIX.5.0", "784", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "784", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "784", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "784", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "784", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "784", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "784", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "784", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "784", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "784", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "784", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "784", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "784", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "784", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "784", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "784", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "784", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "784", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "784", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "784", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "784", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "784", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "784", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "784", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "784", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "784", "102"): "102 .. DataRepository",
        ("FIX.5.0", "784", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "784", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "784", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "784", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "784", "107"): "107 .. Correspondent",
        ("FIX.5.0", "784", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "784", "110"): "110 .. Borrower",
        ("FIX.5.0", "784", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "784", "112"): "112 .. Guarantor",
        ("FIX.5.0", "784", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "784", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "784", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "784", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "784", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "784", "118"): "118 .. Operator",
        ("FIX.5.0", "784", "119"): "119 .. CSD",
        ("FIX.5.0", "784", "120"): "120 .. ICSD",
        ("FIX.5.0", "784", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "784", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "784", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "786", "1"): "1 .. Firm",
        ("FIX.5.0", "786", "2"): "2 .. Person",
        ("FIX.5.0", "786", "3"): "3 .. System",
        ("FIX.5.0", "786", "4"): "4 .. Application",
        ("FIX.5.0", "786", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "786", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "786", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "786", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "786", "9"): "9 .. ContactName",
        ("FIX.5.0", "786", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "786", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "786", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "786", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "786", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "786", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "786", "16"): "16 .. BIC",
        ("FIX.5.0", "786", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "786", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "786", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "786", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "786", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "786", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "786", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "786", "24"): "24 .. Department",
        ("FIX.5.0", "786", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "786", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "786", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "786", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "786", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "786", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "786", "31"): "31 .. Location",
        ("FIX.5.0", "786", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "786", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "786", "34"): "34 .. AddressCity",
        ("FIX.5.0", "786", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "786", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "786", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "786", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "786", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "786", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "786", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "786", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "786", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "786", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "786", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "786", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "786", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "786", "48"): "48 .. USPerson",
        ("FIX.5.0", "786", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "786", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "786", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "786", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "786", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "786", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "786", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "786", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "786", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "786", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "786", "59"): "59 .. FCMCode",
        ("FIX.5.0", "786", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "786", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "786", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "786", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "786", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "786", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "786", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "786", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "786", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "786", "69"): "69 .. EndUser",
        ("FIX.5.0", "786", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "786", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "786", "72"): "72 .. Domicile",
        ("FIX.5.0", "786", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "786", "74"): "74 .. Payer",
        ("FIX.5.0", "786", "75"): "75 .. Receiver",
        ("FIX.5.0", "786", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "786", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "786", "78"): "78 .. FirstName",
        ("FIX.5.0", "786", "79"): "79 .. Surname",
        ("FIX.5.0", "786", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "786", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "786", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "786", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "786", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "787", "C"): "C .. Cash",
        ("FIX.5.0", "787", "S"): "S .. Securities",
        ("FIX.5.0", "788", "1"): "1 .. Overnight",
        ("FIX.5.0", "788", "2"): "2 .. Term",
        ("FIX.5.0", "788", "3"): "3 .. Flexible",
        ("FIX.5.0", "788", "4"): "4 .. Open",
        ("FIX.5.0", "792", "0"): "0 .. UnableToProcessRequest",
        ("FIX.5.0", "792", "1"): "1 .. UnknownAccount",
        ("FIX.5.0", "792", "2"): "2 .. NoMatchingSettlementInstructionsFound",
        ("FIX.5.0", "792", "99"): "99 .. Other",
        ("FIX.5.0", "794", "2"): "2 .. PreliminaryRequestToIntermediary",
        ("FIX.5.0", "794", "3"): "3 .. SellsideCalculatedUsingPreliminary",
        ("FIX.5.0", "794", "4"): "4 .. SellsideCalculatedWithoutPreliminary",
        ("FIX.5.0", "794", "5"): "5 .. WarehouseRecap",
        ("FIX.5.0", "794", "8"): "8 .. RequestToIntermediary",
        ("FIX.5.0", "794", "15"): "15 .. Giveup",
        ("FIX.5.0", "794", "16"): "16 .. Takeup",
        ("FIX.5.0", "794", "17"): "17 .. Reversal",
        ("FIX.5.0", "794", "18"): "18 .. Alleged",
        ("FIX.5.0", "794", "19"): "19 .. SubAllocationGiveup",
        ("FIX.5.0", "796", "1"): "1 .. OriginalDetailsIncomplete",
        ("FIX.5.0", "796", "2"): "2 .. ChangeInUnderlyingOrderDetails",
        ("FIX.5.0", "796", "3"): "3 .. CancelledByGiveupFirm",
        ("FIX.5.0", "796", "99"): "99 .. Other",
        ("FIX.5.0", "798", "1"): "1 .. CarriedCustomerSide",
        ("FIX.5.0", "798", "2"): "2 .. CarriedNonCustomerSide",
        ("FIX.5.0", "798", "3"): "3 .. HouseTrader",
        ("FIX.5.0", "798", "4"): "4 .. FloorTrader",
        ("FIX.5.0", "798", "6"): "6 .. CarriedNonCustomerSideCrossMargined",
        ("FIX.5.0", "798", "7"): "7 .. HouseTraderCrossMargined",
        ("FIX.5.0", "798", "8"): "8 .. JointBackOfficeAccount",
        ("FIX.5.0", "803", "1"): "1 .. Firm",
        ("FIX.5.0", "803", "2"): "2 .. Person",
        ("FIX.5.0", "803", "3"): "3 .. System",
        ("FIX.5.0", "803", "4"): "4 .. Application",
        ("FIX.5.0", "803", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "803", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "803", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "803", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "803", "9"): "9 .. ContactName",
        ("FIX.5.0", "803", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "803", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "803", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "803", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "803", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "803", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "803", "16"): "16 .. BIC",
        ("FIX.5.0", "803", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "803", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "803", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "803", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "803", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "803", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "803", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "803", "24"): "24 .. Department",
        ("FIX.5.0", "803", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "803", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "803", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "803", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "803", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "803", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "803", "31"): "31 .. Location",
        ("FIX.5.0", "803", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "803", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "803", "34"): "34 .. AddressCity",
        ("FIX.5.0", "803", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "803", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "803", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "803", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "803", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "803", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "803", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "803", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "803", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "803", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "803", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "803", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "803", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "803", "48"): "48 .. USPerson",
        ("FIX.5.0", "803", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "803", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "803", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "803", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "803", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "803", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "803", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "803", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "803", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "803", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "803", "59"): "59 .. FCMCode",
        ("FIX.5.0", "803", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "803", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "803", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "803", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "803", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "803", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "803", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "803", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "803", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "803", "69"): "69 .. EndUser",
        ("FIX.5.0", "803", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "803", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "803", "72"): "72 .. Domicile",
        ("FIX.5.0", "803", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "803", "74"): "74 .. Payer",
        ("FIX.5.0", "803", "75"): "75 .. Receiver",
        ("FIX.5.0", "803", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "803", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "803", "78"): "78 .. FirstName",
        ("FIX.5.0", "803", "79"): "79 .. Surname",
        ("FIX.5.0", "803", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "803", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "803", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "803", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "803", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "805", "1"): "1 .. Firm",
        ("FIX.5.0", "805", "2"): "2 .. Person",
        ("FIX.5.0", "805", "3"): "3 .. System",
        ("FIX.5.0", "805", "4"): "4 .. Application",
        ("FIX.5.0", "805", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "805", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "805", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "805", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "805", "9"): "9 .. ContactName",
        ("FIX.5.0", "805", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "805", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "805", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "805", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "805", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "805", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "805", "16"): "16 .. BIC",
        ("FIX.5.0", "805", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "805", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "805", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "805", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "805", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "805", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "805", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "805", "24"): "24 .. Department",
        ("FIX.5.0", "805", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "805", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "805", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "805", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "805", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "805", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "805", "31"): "31 .. Location",
        ("FIX.5.0", "805", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "805", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "805", "34"): "34 .. AddressCity",
        ("FIX.5.0", "805", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "805", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "805", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "805", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "805", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "805", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "805", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "805", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "805", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "805", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "805", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "805", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "805", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "805", "48"): "48 .. USPerson",
        ("FIX.5.0", "805", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "805", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "805", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "805", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "805", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "805", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "805", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "805", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "805", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "805", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "805", "59"): "59 .. FCMCode",
        ("FIX.5.0", "805", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "805", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "805", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "805", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "805", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "805", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "805", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "805", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "805", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "805", "69"): "69 .. EndUser",
        ("FIX.5.0", "805", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "805", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "805", "72"): "72 .. Domicile",
        ("FIX.5.0", "805", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "805", "74"): "74 .. Payer",
        ("FIX.5.0", "805", "75"): "75 .. Receiver",
        ("FIX.5.0", "805", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "805", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "805", "78"): "78 .. FirstName",
        ("FIX.5.0", "805", "79"): "79 .. Surname",
        ("FIX.5.0", "805", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "805", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "805", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "805", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "805", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "807", "1"): "1 .. Firm",
        ("FIX.5.0", "807", "2"): "2 .. Person",
        ("FIX.5.0", "807", "3"): "3 .. System",
        ("FIX.5.0", "807", "4"): "4 .. Application",
        ("FIX.5.0", "807", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "807", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "807", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "807", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "807", "9"): "9 .. ContactName",
        ("FIX.5.0", "807", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "807", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "807", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "807", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "807", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "807", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "807", "16"): "16 .. BIC",
        ("FIX.5.0", "807", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "807", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "807", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "807", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "807", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "807", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "807", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "807", "24"): "24 .. Department",
        ("FIX.5.0", "807", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "807", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "807", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "807", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "807", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "807", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "807", "31"): "31 .. Location",
        ("FIX.5.0", "807", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "807", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "807", "34"): "34 .. AddressCity",
        ("FIX.5.0", "807", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "807", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "807", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "807", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "807", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "807", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "807", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "807", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "807", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "807", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "807", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "807", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "807", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "807", "48"): "48 .. USPerson",
        ("FIX.5.0", "807", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "807", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "807", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "807", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "807", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "807", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "807", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "807", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "807", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "807", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "807", "59"): "59 .. FCMCode",
        ("FIX.5.0", "807", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "807", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "807", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "807", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "807", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "807", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "807", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "807", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "807", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "807", "69"): "69 .. EndUser",
        ("FIX.5.0", "807", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "807", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "807", "72"): "72 .. Domicile",
        ("FIX.5.0", "807", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "807", "74"): "74 .. Payer",
        ("FIX.5.0", "807", "75"): "75 .. Receiver",
        ("FIX.5.0", "807", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "807", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "807", "78"): "78 .. FirstName",
        ("FIX.5.0", "807", "79"): "79 .. Surname",
        ("FIX.5.0", "807", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "807", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "807", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "807", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "807", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "808", "1"): "1 .. PendingAccept",
        ("FIX.5.0", "808", "2"): "2 .. PendingRelease",
        ("FIX.5.0", "808", "3"): "3 .. PendingReversal",
        ("FIX.5.0", "808", "4"): "4 .. Accept",
        ("FIX.5.0", "808", "5"): "5 .. BlockLevelReject",
        ("FIX.5.0", "808", "6"): "6 .. AccountLevelReject",
        ("FIX.5.0", "814", "0"): "0 .. NoActionTaken",
        ("FIX.5.0", "814", "1"): "1 .. QueueFlushed",
        ("FIX.5.0", "814", "2"): "2 .. OverlayLast",
        ("FIX.5.0", "814", "3"): "3 .. EndSession",
        ("FIX.5.0", "815", "0"): "0 .. NoActionTaken",
        ("FIX.5.0", "815", "1"): "1 .. QueueFlushed",
        ("FIX.5.0", "815", "2"): "2 .. OverlayLast",
        ("FIX.5.0", "815", "3"): "3 .. EndSession",
        ("FIX.5.0", "819", "0"): "0 .. NoAveragePricing",
        ("FIX.5.0", "819", "1"): "1 .. Trade",
        ("FIX.5.0", "819", "2"): "2 .. LastTrade",
        ("FIX.5.0", "819", "3"): "3 .. NotionalValueAveragePxGroupTrade",
        ("FIX.5.0", "819", "4"): "4 .. AveragePricedTrade",
        ("FIX.5.0", "826", "0"): "0 .. AllocationNotRequired",
        ("FIX.5.0", "826", "1"): "1 .. AllocationRequired",
        ("FIX.5.0", "826", "2"): "2 .. UseAllocationProvidedWithTheTrade",
        ("FIX.5.0", "826", "3"): "3 .. AllocationGiveUpExecutor",
        ("FIX.5.0", "826", "4"): "4 .. AllocationFromExecutor",
        ("FIX.5.0", "826", "5"): "5 .. AllocationToClaimAccount",
        ("FIX.5.0", "826", "6"): "6 .. TradeSplit",
        ("FIX.5.0", "827", "0"): "0 .. ExpireOnTradingSessionClose",
        ("FIX.5.0", "827", "1"): "1 .. ExpireOnTradingSessionOpen",
        ("FIX.5.0", "827", "2"): "2 .. SpecifiedExpiration",
        ("FIX.5.0", "828", "0"): "0 .. RegularTrade",
        ("FIX.5.0", "828", "1"): "1 .. BlockTrade",
        ("FIX.5.0", "828", "2"): "2 .. EFP",
        ("FIX.5.0", "828", "3"): "3 .. Transfer",
        ("FIX.5.0", "828", "4"): "4 .. LateTrade",
        ("FIX.5.0", "828", "5"): "5 .. TTrade",
        ("FIX.5.0", "828", "6"): "6 .. WeightedAveragePriceTrade",
        ("FIX.5.0", "828", "7"): "7 .. BunchedTrade",
        ("FIX.5.0", "828", "8"): "8 .. LateBunchedTrade",
        ("FIX.5.0", "828", "9"): "9 .. PriorReferencePriceTrade",
        ("FIX.5.0", "828", "10"): "10 .. AfterHoursTrade",
        ("FIX.5.0", "828", "11"): "11 .. ExchangeForRisk",
        ("FIX.5.0", "828", "12"): "12 .. ExchangeForSwap",
        ("FIX.5.0", "828", "13"): "13 .. ExchangeOfFuturesFor",
        ("FIX.5.0", "828", "14"): "14 .. ExchangeOfOptionsForOptions",
        ("FIX.5.0", "828", "15"): "15 .. TradingAtSettlement",
        ("FIX.5.0", "828", "16"): "16 .. AllOrNone",
        ("FIX.5.0", "828", "17"): "17 .. FuturesLargeOrderExecution",
        ("FIX.5.0", "828", "18"): "18 .. ExchangeOfFuturesForFutures",
        ("FIX.5.0", "828", "19"): "19 .. OptionInterimTrade",
        ("FIX.5.0", "828", "20"): "20 .. OptionCabinetTrade",
        ("FIX.5.0", "828", "22"): "22 .. PrivatelyNegotiatedTrades",
        ("FIX.5.0", "828", "23"): "23 .. SubstitutionOfFuturesForForwards",
        ("FIX.5.0", "828", "48"): "48 .. NonStandardSettlement",
        ("FIX.5.0", "828", "49"): "49 .. DerivativeRelatedTransaction",
        ("FIX.5.0", "828", "50"): "50 .. PortfolioTrade",
        ("FIX.5.0", "828", "51"): "51 .. VolumeWeightedAverageTrade",
        ("FIX.5.0", "828", "52"): "52 .. ExchangeGrantedTrade",
        ("FIX.5.0", "828", "53"): "53 .. RepurchaseAgreement",
        ("FIX.5.0", "828", "54"): "54 .. OTC",
        ("FIX.5.0", "828", "55"): "55 .. ExchangeBasisFacility",
        ("FIX.5.0", "828", "56"): "56 .. OpeningTrade",
        ("FIX.5.0", "828", "57"): "57 .. NettedTrade",
        ("FIX.5.0", "828", "58"): "58 .. BlockSwapTrade",
        ("FIX.5.0", "828", "59"): "59 .. CreditEventTrade",
        ("FIX.5.0", "828", "60"): "60 .. SuccessionEventTrade",
        ("FIX.5.0", "828", "61"): "61 .. GiveUpGiveInTrade",
        ("FIX.5.0", "828", "62"): "62 .. DarkTrade",
        ("FIX.5.0", "828", "63"): "63 .. TechnicalTrade",
        ("FIX.5.0", "828", "64"): "64 .. Benchmark",
        ("FIX.5.0", "828", "65"): "65 .. PackageTrade",
        ("FIX.5.0", "828", "24"): "24 .. ErrorTrade",
        ("FIX.5.0", "828", "25"): "25 .. SpecialCumDividend",
        ("FIX.5.0", "828", "26"): "26 .. SpecialExDividend",
        ("FIX.5.0", "828", "27"): "27 .. SpecialCumCoupon",
        ("FIX.5.0", "828", "28"): "28 .. SpecialExCoupon",
        ("FIX.5.0", "828", "29"): "29 .. CashSettlement",
        ("FIX.5.0", "828", "30"): "30 .. SpecialPrice",
        ("FIX.5.0", "828", "31"): "31 .. GuaranteedDelivery",
        ("FIX.5.0", "828", "32"): "32 .. SpecialCumRights",
        ("FIX.5.0", "828", "33"): "33 .. SpecialExRights",
        ("FIX.5.0", "828", "34"): "34 .. SpecialCumCapitalRepayments",
        ("FIX.5.0", "828", "35"): "35 .. SpecialExCapitalRepayments",
        ("FIX.5.0", "828", "36"): "36 .. SpecialCumBonus",
        ("FIX.5.0", "828", "37"): "37 .. SpecialExBonus",
        ("FIX.5.0", "828", "38"): "38 .. LargeTrade",
        ("FIX.5.0", "828", "39"): "39 .. WorkedPrincipalTrade",
        ("FIX.5.0", "828", "40"): "40 .. BlockTrades",
        ("FIX.5.0", "828", "41"): "41 .. NameChange",
        ("FIX.5.0", "828", "42"): "42 .. PortfolioTransfer",
        ("FIX.5.0", "828", "43"): "43 .. ProrogationBuy",
        ("FIX.5.0", "828", "44"): "44 .. ProrogationSell",
        ("FIX.5.0", "828", "45"): "45 .. OptionExercise",
        ("FIX.5.0", "828", "46"): "46 .. DeltaNeutralTransaction",
        ("FIX.5.0", "828", "47"): "47 .. FinancingTransaction",
        ("FIX.5.0", "829", "0"): "0 .. CMTA",
        ("FIX.5.0", "829", "1"): "1 .. InternalTransferOrAdjustment",
        ("FIX.5.0", "829", "2"): "2 .. ExternalTransferOrTransferOfAccount",
        ("FIX.5.0", "829", "3"): "3 .. RejectForSubmittingSide",
        ("FIX.5.0", "829", "4"): "4 .. AdvisoryForContraSide",
        ("FIX.5.0", "829", "5"): "5 .. OffsetDueToAnAllocation",
        ("FIX.5.0", "829", "6"): "6 .. OnsetDueToAnAllocation",
        ("FIX.5.0", "829", "7"): "7 .. DifferentialSpread",
        ("FIX.5.0", "829", "8"): "8 .. ImpliedSpreadLegExecutedAgainstAnOutright",
        ("FIX.5.0", "829", "9"): "9 .. TransactionFromExercise",
        ("FIX.5.0", "829", "10"): "10 .. TransactionFromAssignment",
        ("FIX.5.0", "829", "11"): "11 .. ACATS",
        ("FIX.5.0", "829", "33"): "33 .. OffHoursTrade",
        ("FIX.5.0", "829", "34"): "34 .. OnHoursTrade",
        ("FIX.5.0", "829", "35"): "35 .. OTCQuote",
        ("FIX.5.0", "829", "36"): "36 .. ConvertedSWAP",
        ("FIX.5.0", "829", "40"): "40 .. WashTrade",
        ("FIX.5.0", "829", "41"): "41 .. TradeAtSettlement",
        ("FIX.5.0", "829", "42"): "42 .. AuctionTrade",
        ("FIX.5.0", "829", "43"): "43 .. TradeAtMarker",
        ("FIX.5.0", "829", "44"): "44 .. CreditDefault",
        ("FIX.5.0", "829", "45"): "45 .. CreditRestructuring",
        ("FIX.5.0", "829", "46"): "46 .. Merger",
        ("FIX.5.0", "829", "47"): "47 .. SpinOff",
        ("FIX.5.0", "829", "48"): "48 .. MultilateralCompression",
        ("FIX.5.0", "829", "50"): "50 .. Balancing",
        ("FIX.5.0", "829", "51"): "51 .. BasisTradeIndexClose",
        ("FIX.5.0", "829", "14"): "14 .. AI",
        ("FIX.5.0", "829", "15"): "15 .. B",
        ("FIX.5.0", "829", "16"): "16 .. K",
        ("FIX.5.0", "829", "17"): "17 .. LC",
        ("FIX.5.0", "829", "18"): "18 .. M",
        ("FIX.5.0", "829", "19"): "19 .. N",
        ("FIX.5.0", "829", "20"): "20 .. NM",
        ("FIX.5.0", "829", "21"): "21 .. NR",
        ("FIX.5.0", "829", "22"): "22 .. P",
        ("FIX.5.0", "829", "23"): "23 .. PA",
        ("FIX.5.0", "829", "24"): "24 .. PC",
        ("FIX.5.0", "829", "25"): "25 .. PN",
        ("FIX.5.0", "829", "26"): "26 .. R",
        ("FIX.5.0", "829", "27"): "27 .. RO",
        ("FIX.5.0", "829", "28"): "28 .. RT",
        ("FIX.5.0", "829", "29"): "29 .. SW",
        ("FIX.5.0", "829", "30"): "30 .. T",
        ("FIX.5.0", "829", "31"): "31 .. WN",
        ("FIX.5.0", "829", "32"): "32 .. WT",
        ("FIX.5.0", "829", "37"): "37 .. CrossedTrade",
        ("FIX.5.0", "829", "38"): "38 .. InterimProtectedTrade",
        ("FIX.5.0", "829", "39"): "39 .. LargeInScale",
        ("FIX.5.0", "835", "0"): "0 .. Floating",
        ("FIX.5.0", "835", "1"): "1 .. Fixed",
        ("FIX.5.0", "836", "0"): "0 .. Price",
        ("FIX.5.0", "836", "1"): "1 .. BasisPoints",
        ("FIX.5.0", "836", "2"): "2 .. Ticks",
        ("FIX.5.0", "836", "3"): "3 .. PriceTier",
        ("FIX.5.0", "836", "4"): "4 .. Percentage",
        ("FIX.5.0", "837", "0"): "0 .. OrBetter",
        ("FIX.5.0", "837", "1"): "1 .. Strict",
        ("FIX.5.0", "837", "2"): "2 .. OrWorse",
        ("FIX.5.0", "838", "1"): "1 .. MoreAggressive",
        ("FIX.5.0", "838", "2"): "2 .. MorePassive",
        ("FIX.5.0", "840", "1"): "1 .. Local",
        ("FIX.5.0", "840", "2"): "2 .. National",
        ("FIX.5.0", "840", "3"): "3 .. Global",
        ("FIX.5.0", "840", "4"): "4 .. NationalExcludingLocal",
        ("FIX.5.0", "841", "0"): "0 .. Floating",
        ("FIX.5.0", "841", "1"): "1 .. Fixed",
        ("FIX.5.0", "842", "0"): "0 .. Price",
        ("FIX.5.0", "842", "1"): "1 .. BasisPoints",
        ("FIX.5.0", "842", "2"): "2 .. Ticks",
        ("FIX.5.0", "842", "3"): "3 .. PriceTier",
        ("FIX.5.0", "843", "0"): "0 .. OrBetter",
        ("FIX.5.0", "843", "1"): "1 .. Strict",
        ("FIX.5.0", "843", "2"): "2 .. OrWorse",
        ("FIX.5.0", "844", "1"): "1 .. MoreAggressive",
        ("FIX.5.0", "844", "2"): "2 .. MorePassive",
        ("FIX.5.0", "846", "1"): "1 .. Local",
        ("FIX.5.0", "846", "2"): "2 .. National",
        ("FIX.5.0", "846", "3"): "3 .. Global",
        ("FIX.5.0", "846", "4"): "4 .. NationalExcludingLocal",
        ("FIX.5.0", "847", "1"): "1 .. VWAP",
        ("FIX.5.0", "847", "2"): "2 .. Participate",
        ("FIX.5.0", "847", "3"): "3 .. MininizeMarketImpact",
        ("FIX.5.0", "851", "1"): "1 .. AddedLiquidity",
        ("FIX.5.0", "851", "2"): "2 .. RemovedLiquidity",
        ("FIX.5.0", "851", "3"): "3 .. LiquidityRoutedOut",
        ("FIX.5.0", "851", "4"): "4 .. Auction",
        ("FIX.5.0", "851", "5"): "5 .. TriggeredStopOrder",
        ("FIX.5.0", "851", "6"): "6 .. TriggeredContingencyOrder",
        ("FIX.5.0", "851", "7"): "7 .. TriggeredMarketOrder",
        ("FIX.5.0", "852", "N"): "N .. DoNotReportTrade",
        ("FIX.5.0", "852", "Y"): "Y .. ReportTrade",
        ("FIX.5.0", "853", "0"): "0 .. DealerSoldShort",
        ("FIX.5.0", "853", "1"): "1 .. DealerSoldShortExempt",
        ("FIX.5.0", "853", "2"): "2 .. SellingCustomerSoldShort",
        ("FIX.5.0", "853", "3"): "3 .. SellingCustomerSoldShortExempt",
        ("FIX.5.0", "853", "4"): "4 .. QualifiedServiceRepresentative",
        ("FIX.5.0", "853", "5"): "5 .. QSROrAGUContraSideSoldShortExempt",
        ("FIX.5.0", "854", "0"): "0 .. Units",
        ("FIX.5.0", "854", "1"): "1 .. Contracts",
        ("FIX.5.0", "854", "2"): "2 .. UnitsOfMeasurePerTimeUnit",
        ("FIX.5.0", "855", "0"): "0 .. RegularTrade",
        ("FIX.5.0", "855", "1"): "1 .. BlockTrade",
        ("FIX.5.0", "855", "2"): "2 .. EFP",
        ("FIX.5.0", "855", "3"): "3 .. Transfer",
        ("FIX.5.0", "855", "4"): "4 .. LateTrade",
        ("FIX.5.0", "855", "5"): "5 .. TTrade",
        ("FIX.5.0", "855", "6"): "6 .. WeightedAveragePriceTrade",
        ("FIX.5.0", "855", "7"): "7 .. BunchedTrade",
        ("FIX.5.0", "855", "8"): "8 .. LateBunchedTrade",
        ("FIX.5.0", "855", "9"): "9 .. PriorReferencePriceTrade",
        ("FIX.5.0", "855", "10"): "10 .. AfterHoursTrade",
        ("FIX.5.0", "855", "11"): "11 .. ExchangeForRisk",
        ("FIX.5.0", "855", "12"): "12 .. ExchangeForSwap",
        ("FIX.5.0", "855", "13"): "13 .. ExchangeOfFuturesFor",
        ("FIX.5.0", "855", "14"): "14 .. ExchangeOfOptionsForOptions",
        ("FIX.5.0", "855", "15"): "15 .. TradingAtSettlement",
        ("FIX.5.0", "855", "16"): "16 .. AllOrNone",
        ("FIX.5.0", "855", "17"): "17 .. FuturesLargeOrderExecution",
        ("FIX.5.0", "855", "18"): "18 .. ExchangeOfFuturesForFutures",
        ("FIX.5.0", "855", "19"): "19 .. OptionInterimTrade",
        ("FIX.5.0", "855", "20"): "20 .. OptionCabinetTrade",
        ("FIX.5.0", "855", "22"): "22 .. PrivatelyNegotiatedTrades",
        ("FIX.5.0", "855", "23"): "23 .. SubstitutionOfFuturesForForwards",
        ("FIX.5.0", "855", "48"): "48 .. NonStandardSettlement",
        ("FIX.5.0", "855", "49"): "49 .. DerivativeRelatedTransaction",
        ("FIX.5.0", "855", "50"): "50 .. PortfolioTrade",
        ("FIX.5.0", "855", "51"): "51 .. VolumeWeightedAverageTrade",
        ("FIX.5.0", "855", "52"): "52 .. ExchangeGrantedTrade",
        ("FIX.5.0", "855", "53"): "53 .. RepurchaseAgreement",
        ("FIX.5.0", "855", "54"): "54 .. OTC",
        ("FIX.5.0", "855", "55"): "55 .. ExchangeBasisFacility",
        ("FIX.5.0", "855", "56"): "56 .. OpeningTrade",
        ("FIX.5.0", "855", "57"): "57 .. NettedTrade",
        ("FIX.5.0", "855", "58"): "58 .. BlockSwapTrade",
        ("FIX.5.0", "855", "59"): "59 .. CreditEventTrade",
        ("FIX.5.0", "855", "60"): "60 .. SuccessionEventTrade",
        ("FIX.5.0", "855", "61"): "61 .. GiveUpGiveInTrade",
        ("FIX.5.0", "855", "62"): "62 .. DarkTrade",
        ("FIX.5.0", "855", "63"): "63 .. TechnicalTrade",
        ("FIX.5.0", "855", "64"): "64 .. Benchmark",
        ("FIX.5.0", "855", "65"): "65 .. PackageTrade",
        ("FIX.5.0", "855", "24"): "24 .. ErrorTrade",
        ("FIX.5.0", "855", "25"): "25 .. SpecialCumDividend",
        ("FIX.5.0", "855", "26"): "26 .. SpecialExDividend",
        ("FIX.5.0", "855", "27"): "27 .. SpecialCumCoupon",
        ("FIX.5.0", "855", "28"): "28 .. SpecialExCoupon",
        ("FIX.5.0", "855", "29"): "29 .. CashSettlement",
        ("FIX.5.0", "855", "30"): "30 .. SpecialPrice",
        ("FIX.5.0", "855", "31"): "31 .. GuaranteedDelivery",
        ("FIX.5.0", "855", "32"): "32 .. SpecialCumRights",
        ("FIX.5.0", "855", "33"): "33 .. SpecialExRights",
        ("FIX.5.0", "855", "34"): "34 .. SpecialCumCapitalRepayments",
        ("FIX.5.0", "855", "35"): "35 .. SpecialExCapitalRepayments",
        ("FIX.5.0", "855", "36"): "36 .. SpecialCumBonus",
        ("FIX.5.0", "855", "37"): "37 .. SpecialExBonus",
        ("FIX.5.0", "855", "38"): "38 .. LargeTrade",
        ("FIX.5.0", "855", "39"): "39 .. WorkedPrincipalTrade",
        ("FIX.5.0", "855", "40"): "40 .. BlockTrades",
        ("FIX.5.0", "855", "41"): "41 .. NameChange",
        ("FIX.5.0", "855", "42"): "42 .. PortfolioTransfer",
        ("FIX.5.0", "855", "43"): "43 .. ProrogationBuy",
        ("FIX.5.0", "855", "44"): "44 .. ProrogationSell",
        ("FIX.5.0", "855", "45"): "45 .. OptionExercise",
        ("FIX.5.0", "855", "46"): "46 .. DeltaNeutralTransaction",
        ("FIX.5.0", "855", "47"): "47 .. FinancingTransaction",
        ("FIX.5.0", "856", "0"): "0 .. Submit",
        ("FIX.5.0", "856", "1"): "1 .. Alleged",
        ("FIX.5.0", "856", "2"): "2 .. Accept",
        ("FIX.5.0", "856", "3"): "3 .. Decline",
        ("FIX.5.0", "856", "4"): "4 .. Addendum",
        ("FIX.5.0", "856", "5"): "5 .. No",
        ("FIX.5.0", "856", "6"): "6 .. TradeReportCancel",
        ("FIX.5.0", "856", "7"): "7 .. LockedIn",
        ("FIX.5.0", "856", "8"): "8 .. Defaulted",
        ("FIX.5.0", "856", "9"): "9 .. InvalidCMTA",
        ("FIX.5.0", "856", "10"): "10 .. Pended",
        ("FIX.5.0", "856", "11"): "11 .. AllegedNew",
        ("FIX.5.0", "856", "12"): "12 .. AllegedAddendum",
        ("FIX.5.0", "856", "13"): "13 .. AllegedNo",
        ("FIX.5.0", "856", "14"): "14 .. AllegedTradeReportCancel",
        ("FIX.5.0", "856", "15"): "15 .. AllegedTradeBreak",
        ("FIX.5.0", "856", "16"): "16 .. Verify",
        ("FIX.5.0", "856", "17"): "17 .. Dispute",
        ("FIX.5.0", "857", "0"): "0 .. NotSpecified",
        ("FIX.5.0", "857", "1"): "1 .. ExplicitListProvided",
        ("FIX.5.0", "865", "1"): "1 .. Put",
        ("FIX.5.0", "865", "2"): "2 .. Call",
        ("FIX.5.0", "865", "3"): "3 .. Tender",
        ("FIX.5.0", "865", "4"): "4 .. SinkingFundCall",
        ("FIX.5.0", "865", "5"): "5 .. Activation",
        ("FIX.5.0", "865", "6"): "6 .. Inactiviation",
        ("FIX.5.0", "865", "7"): "7 .. LastEligibleTradeDate",
        ("FIX.5.0", "865", "8"): "8 .. SwapStartDate",
        ("FIX.5.0", "865", "9"): "9 .. SwapEndDate",
        ("FIX.5.0", "865", "10"): "10 .. SwapRollDate",
        ("FIX.5.0", "865", "11"): "11 .. SwapNextStartDate",
        ("FIX.5.0", "865", "12"): "12 .. SwapNextRollDate",
        ("FIX.5.0", "865", "13"): "13 .. FirstDeliveryDate",
        ("FIX.5.0", "865", "14"): "14 .. LastDeliveryDate",
        ("FIX.5.0", "865", "15"): "15 .. InitialInventoryDueDate",
        ("FIX.5.0", "865", "16"): "16 .. FinalInventoryDueDate",
        ("FIX.5.0", "865", "17"): "17 .. FirstIntentDate",
        ("FIX.5.0", "865", "18"): "18 .. LastIntentDate",
        ("FIX.5.0", "865", "19"): "19 .. PositionRemovalDate",
        ("FIX.5.0", "865", "20"): "20 .. MinimumNotice",
        ("FIX.5.0", "865", "21"): "21 .. DeliveryStartTime",
        ("FIX.5.0", "865", "22"): "22 .. DeliveryEndTime",
        ("FIX.5.0", "865", "23"): "23 .. FirstNoticeDate",
        ("FIX.5.0", "865", "24"): "24 .. LastNoticeDate",
        ("FIX.5.0", "865", "25"): "25 .. FirstExerciseDate",
        ("FIX.5.0", "865", "26"): "26 .. RedemptionDate",
        ("FIX.5.0", "865", "27"): "27 .. TrdCntntnEfctvDt",
        ("FIX.5.0", "865", "99"): "99 .. Other",
        ("FIX.5.0", "871", "1"): "1 .. Flat",
        ("FIX.5.0", "871", "2"): "2 .. ZeroCoupon",
        ("FIX.5.0", "871", "3"): "3 .. InterestBearing",
        ("FIX.5.0", "871", "4"): "4 .. NoPeriodicPayments",
        ("FIX.5.0", "871", "5"): "5 .. VariableRate",
        ("FIX.5.0", "871", "6"): "6 .. LessFeeForPut",
        ("FIX.5.0", "871", "7"): "7 .. SteppedCoupon",
        ("FIX.5.0", "871", "8"): "8 .. CouponPeriod",
        ("FIX.5.0", "871", "9"): "9 .. When",
        ("FIX.5.0", "871", "10"): "10 .. OriginalIssueDiscount",
        ("FIX.5.0", "871", "11"): "11 .. Callable",
        ("FIX.5.0", "871", "12"): "12 .. EscrowedToMaturity",
        ("FIX.5.0", "871", "13"): "13 .. EscrowedToRedemptionDate",
        ("FIX.5.0", "871", "14"): "14 .. PreRefunded",
        ("FIX.5.0", "871", "15"): "15 .. InDefault",
        ("FIX.5.0", "871", "16"): "16 .. Unrated",
        ("FIX.5.0", "871", "17"): "17 .. Taxable",
        ("FIX.5.0", "871", "18"): "18 .. Indexed",
        ("FIX.5.0", "871", "19"): "19 .. SubjectToAlternativeMinimumTax",
        ("FIX.5.0", "871", "20"): "20 .. OriginalIssueDiscountPrice",
        ("FIX.5.0", "871", "21"): "21 .. CallableBelowMaturityValue",
        ("FIX.5.0", "871", "22"): "22 .. CallableWithoutNotice",
        ("FIX.5.0", "871", "23"): "23 .. PriceTickRulesForSecurity",
        ("FIX.5.0", "871", "24"): "24 .. TradeTypeEligibilityDetailsForSecurity",
        ("FIX.5.0", "871", "25"): "25 .. InstrumentDenominator",
        ("FIX.5.0", "871", "26"): "26 .. InstrumentNumerator",
        ("FIX.5.0", "871", "27"): "27 .. InstrumentPricePrecision",
        ("FIX.5.0", "871", "28"): "28 .. InstrumentStrikePrice",
        ("FIX.5.0", "871", "29"): "29 .. TradeableIndicator",
        ("FIX.5.0", "871", "30"): "30 .. InstrumentEligibleAnonOrders",
        ("FIX.5.0", "871", "31"): "31 .. MinGuaranteedFillVolume",
        ("FIX.5.0", "871", "32"): "32 .. MinGuaranteedFillStatus",
        ("FIX.5.0", "871", "33"): "33 .. TradeAtSettlementEligibility",
        ("FIX.5.0", "871", "34"): "34 .. TestInstrument",
        ("FIX.5.0", "871", "35"): "35 .. DummyInstrument",
        ("FIX.5.0", "871", "36"): "36 .. NegativeSettlementPriceEligibility",
        ("FIX.5.0", "871", "37"): "37 .. NegativeStrikePriceEligibility",
        ("FIX.5.0", "871", "38"): "38 .. USStdContractInd",
        ("FIX.5.0", "871", "39"): "39 .. AdmittedToTradingOnTradingVenue",
        ("FIX.5.0", "871", "40"): "40 .. AverageDailyNotionalAmount",
        ("FIX.5.0", "871", "41"): "41 .. AverageDailyNumberTrades",
        ("FIX.5.0", "871", "99"): "99 .. Text",
        ("FIX.5.0", "875", "1"): "1 .. Program3a3",
        ("FIX.5.0", "875", "2"): "2 .. Program42",
        ("FIX.5.0", "875", "3"): "3 .. Program3a2",
        ("FIX.5.0", "875", "4"): "4 .. Program3a3And3c7",
        ("FIX.5.0", "875", "5"): "5 .. Program3a4",
        ("FIX.5.0", "875", "6"): "6 .. Program3a5",
        ("FIX.5.0", "875", "7"): "7 .. Program3a7",
        ("FIX.5.0", "875", "8"): "8 .. Program3c7",
        ("FIX.5.0", "875", "99"): "99 .. Other",
        ("FIX.5.0", "877", "1"): "1 .. Program3a3",
        ("FIX.5.0", "877", "2"): "2 .. Program42",
        ("FIX.5.0", "877", "3"): "3 .. Program3a2",
        ("FIX.5.0", "877", "4"): "4 .. Program3a3And3c7",
        ("FIX.5.0", "877", "5"): "5 .. Program3a4",
        ("FIX.5.0", "877", "6"): "6 .. Program3a5",
        ("FIX.5.0", "877", "7"): "7 .. Program3a7",
        ("FIX.5.0", "877", "8"): "8 .. Program3c7",
        ("FIX.5.0", "877", "99"): "99 .. Other",
        ("FIX.5.0", "888", "AMT"): "AMT .. AlternativeMinimumTax",
        ("FIX.5.0", "888", "AUTOREINV"): "AUTOREINV .. AutoReinvestment",
        ("FIX.5.0", "888", "BANKQUAL"): "BANKQUAL .. BankQualified",
        ("FIX.5.0", "888", "BGNCON"): "BGNCON .. BargainConditions",
        ("FIX.5.0", "888", "COUPON"): "COUPON .. CouponRange",
        ("FIX.5.0", "888", "CURRENCY"): "CURRENCY .. ISOCurrencyCode",
        ("FIX.5.0", "888", "CUSTOMDATE"): "CUSTOMDATE .. CustomStart",
        ("FIX.5.0", "888", "GEOG"): "GEOG .. Geographics",
        ("FIX.5.0", "888", "HAIRCUT"): "HAIRCUT .. ValuationDiscount",
        ("FIX.5.0", "888", "INSURED"): "INSURED .. Insured",
        ("FIX.5.0", "888", "ISSUE"): "ISSUE .. IssueDate",
        ("FIX.5.0", "888", "ISSUER"): "ISSUER .. Issuer",
        ("FIX.5.0", "888", "ISSUESIZE"): "ISSUESIZE .. IssueSizeRange",
        ("FIX.5.0", "888", "LOOKBACK"): "LOOKBACK .. LookbackDays",
        ("FIX.5.0", "888", "LOT"): "LOT .. ExplicitLotIdentifier",
        ("FIX.5.0", "888", "LOTVAR"): "LOTVAR .. LotVariance",
        ("FIX.5.0", "888", "MAT"): "MAT .. MaturityYearAndMonth",
        ("FIX.5.0", "888", "MATURITY"): "MATURITY .. MaturityRange",
        ("FIX.5.0", "888", "MAXSUBS"): "MAXSUBS .. MaximumSubstitutions",
        ("FIX.5.0", "888", "MINDNOM"): "MINDNOM .. MinimumDenomination",
        ("FIX.5.0", "888", "MININCR"): "MININCR .. MinimumIncrement",
        ("FIX.5.0", "888", "MINQTY"): "MINQTY .. MinimumQuantity",
        ("FIX.5.0", "888", "PAYFREQ"): "PAYFREQ .. PaymentFrequency",
        ("FIX.5.0", "888", "PIECES"): "PIECES .. NumberOfPieces",
        ("FIX.5.0", "888", "PMAX"): "PMAX .. PoolsMaximum",
        ("FIX.5.0", "888", "PPL"): "PPL .. PoolsPerLot",
        ("FIX.5.0", "888", "PPM"): "PPM .. PoolsPerMillion",
        ("FIX.5.0", "888", "PPT"): "PPT .. PoolsPerTrade",
        ("FIX.5.0", "888", "PRICE"): "PRICE .. PriceRange",
        ("FIX.5.0", "888", "PRICEFREQ"): "PRICEFREQ .. PricingFrequency",
        ("FIX.5.0", "888", "PROD"): "PROD .. ProductionYear",
        ("FIX.5.0", "888", "PROTECT"): "PROTECT .. CallProtection",
        ("FIX.5.0", "888", "PURPOSE"): "PURPOSE .. Purpose",
        ("FIX.5.0", "888", "PXSOURCE"): "PXSOURCE .. BenchmarkPriceSource",
        ("FIX.5.0", "888", "RATING"): "RATING .. RatingSourceAndRange",
        ("FIX.5.0", "888", "REDEMPTION"): "REDEMPTION .. TypeOfRedemption",
        ("FIX.5.0", "888", "RESTRICTED"): "RESTRICTED .. Restricted",
        ("FIX.5.0", "888", "SECTOR"): "SECTOR .. MarketSector",
        ("FIX.5.0", "888", "SECTYPE"): "SECTYPE .. SecurityTypeIncludedOrExcluded",
        ("FIX.5.0", "888", "STRUCT"): "STRUCT .. Structure",
        ("FIX.5.0", "888", "SUBSFREQ"): "SUBSFREQ .. SubstitutionsFrequency",
        ("FIX.5.0", "888", "SUBSLEFT"): "SUBSLEFT .. SubstitutionsLeft",
        ("FIX.5.0", "888", "TEXT"): "TEXT .. FreeformText",
        ("FIX.5.0", "888", "TRDVAR"): "TRDVAR .. TradeVariance",
        ("FIX.5.0", "888", "WAC"): "WAC .. WeightedAverageCoupon",
        ("FIX.5.0", "888", "WAL"): "WAL .. WeightedAverageLifeCoupon",
        ("FIX.5.0", "888", "WALA"): "WALA .. WeightedAverageLoanAge",
        ("FIX.5.0", "888", "WAM"): "WAM .. WeightedAverageMaturity",
        ("FIX.5.0", "888", "WHOLE"): "WHOLE .. WholePool",
        ("FIX.5.0", "888", "YIELD"): "YIELD .. YieldRange",
        ("FIX.5.0", "888", "ORIGAMT"): "ORIGAMT .. OriginalAmount",
        ("FIX.5.0", "888", "POOLEFFDT"): "POOLEFFDT .. PoolEffectiveDate",
        ("FIX.5.0", "888", "POOLINITFCTR"): "POOLINITFCTR .. PoolInitialFactor",
        ("FIX.5.0", "888", "TRANCHE"): "TRANCHE .. Tranche",
        ("FIX.5.0", "888", "SUBSTITUTION"): "SUBSTITUTION .. Substitution",
        ("FIX.5.0", "888", "MULTEXCHFLLBCK"): "MULTEXCHFLLBCK .. MULTEXCHFLLBCK",
        ("FIX.5.0", "888", "COMPSECFLLBCK"): "COMPSECFLLBCK .. COMPSECFLLBCK",
        ("FIX.5.0", "888", "LOCLJRSDCTN"): "LOCLJRSDCTN .. LOCLJRSDCTN",
        ("FIX.5.0", "888", "RELVJRSDCTN"): "RELVJRSDCTN .. RELVJRSDCTN",
        ("FIX.5.0", "888", "INCURRCVY"): "INCURRCVY .. IncurredRecovery",
        ("FIX.5.0", "888", "ADDTRM"): "ADDTRM .. AdditionalTerm",
        ("FIX.5.0", "888", "MODEQTYDLVY"): "MODEQTYDLVY .. ModifiedEquityDelivery",
        ("FIX.5.0", "888", "NOREFOBLIG"): "NOREFOBLIG .. NoReferenceOblication",
        ("FIX.5.0", "888", "UNKREFOBLIG"): "UNKREFOBLIG .. UnknownReferenceObligation",
        ("FIX.5.0", "888", "ALLGUARANTEES"): "ALLGUARANTEES .. AllGuarantees",
        ("FIX.5.0", "888", "REFPX"): "REFPX .. ReferencePrice",
        ("FIX.5.0", "888", "REFPOLICY"): "REFPOLICY .. ReferencePolicy",
        ("FIX.5.0", "888", "SECRDLIST"): "SECRDLIST .. SecuredList",
        ("FIX.5.0", "888", "AVFICO"): "AVFICO .. AverageFICOScore",
        ("FIX.5.0", "888", "AVSIZE"): "AVSIZE .. AverageLoanSize",
        ("FIX.5.0", "888", "MAXBAL"): "MAXBAL .. MaximumLoanBalance",
        ("FIX.5.0", "888", "POOL"): "POOL .. PoolIdentifier",
        ("FIX.5.0", "888", "ROLLTYPE"): "ROLLTYPE .. TypeOfRollTrade",
        ("FIX.5.0", "888", "REFTRADE"): "REFTRADE .. ReferenceToRollingOrClosingTrade",
        ("FIX.5.0", "888", "REFPRIN"): "REFPRIN .. PrincipalOfRollingOrClosingTrade",
        ("FIX.5.0", "888", "REFINT"): "REFINT .. InterestOfRollingOrClosingTrade",
        ("FIX.5.0", "888", "AVAILQTY"): "AVAILQTY .. AvailableOfferQuantityToBeShownToTheStreet",
        ("FIX.5.0", "888", "BROKERCREDIT"): "BROKERCREDIT .. BrokerCredit",
        ("FIX.5.0", "888", "INTERNALPX"): "INTERNALPX .. OfferPriceToBeShownToInternalBrokers",
        ("FIX.5.0", "888", "INTERNALQTY"): "INTERNALQTY .. OfferQuantityToBeShownToInternalBrokers",
        ("FIX.5.0", "888", "LEAVEQTY"): "LEAVEQTY .. TheMinimumResidualOfferQuantity",
        ("FIX.5.0", "888", "MAXORDQTY"): "MAXORDQTY .. MaximumOrderSize",
        ("FIX.5.0", "888", "ORDRINCR"): "ORDRINCR .. OrderQuantityIncrement",
        ("FIX.5.0", "888", "PRIMARY"): "PRIMARY .. PrimaryOrSecondaryMarketIndicator",
        ("FIX.5.0", "888", "SALESCREDITOVR"): "SALESCREDITOVR .. BrokerSalesCreditOverride",
        ("FIX.5.0", "888", "TRADERCREDIT"): "TRADERCREDIT .. TraderCredit",
        ("FIX.5.0", "888", "DISCOUNT"): "DISCOUNT .. DiscountRate",
        ("FIX.5.0", "888", "YTM"): "YTM .. YieldToMaturity",
        ("FIX.5.0", "888", "ABS"): "ABS .. AbsolutePrepaymentSpeed",
        ("FIX.5.0", "888", "CPP"): "CPP .. ConstantPrepaymentPenalty",
        ("FIX.5.0", "888", "CPR"): "CPR .. ConstantPrepaymentRate",
        ("FIX.5.0", "888", "CPY"): "CPY .. ConstantPrepaymentYield",
        ("FIX.5.0", "888", "HEP"): "HEP .. FinalCPROfHomeEquityPrepaymentCurve",
        ("FIX.5.0", "888", "MHP"): "MHP .. PercentOfManufacturedHousingPrepaymentCurve",
        ("FIX.5.0", "888", "MPR"): "MPR .. MonthlyPrepaymentRate",
        ("FIX.5.0", "888", "PPC"): "PPC .. PercentOfProspectusPrepaymentCurve",
        ("FIX.5.0", "888", "PSA"): "PSA .. PercentOfBMAPrepaymentCurve",
        ("FIX.5.0", "888", "SMM"): "SMM .. SingleMonthlyMortality",
        ("FIX.5.0", "891", "0"): "0 .. Absolute",
        ("FIX.5.0", "891", "1"): "1 .. PerUnit",
        ("FIX.5.0", "891", "2"): "2 .. Percentage",
        ("FIX.5.0", "893", "N"): "N .. NotLastMessage",
        ("FIX.5.0", "893", "Y"): "Y .. LastMessage",
        ("FIX.5.0", "895", "0"): "0 .. Initial",
        ("FIX.5.0", "895", "1"): "1 .. Scheduled",
        ("FIX.5.0", "895", "2"): "2 .. TimeWarning",
        ("FIX.5.0", "895", "3"): "3 .. MarginDeficiency",
        ("FIX.5.0", "895", "4"): "4 .. MarginExcess",
        ("FIX.5.0", "895", "5"): "5 .. ForwardCollateralDemand",
        ("FIX.5.0", "895", "6"): "6 .. EventOfDefault",
        ("FIX.5.0", "895", "7"): "7 .. AdverseTaxEvent",
        ("FIX.5.0", "895", "8"): "8 .. TransferDeposit",
        ("FIX.5.0", "895", "9"): "9 .. TransferWithdrawal",
        ("FIX.5.0", "895", "10"): "10 .. Pledge",
        ("FIX.5.0", "896", "0"): "0 .. TradeDate",
        ("FIX.5.0", "896", "1"): "1 .. GCInstrument",
        ("FIX.5.0", "896", "2"): "2 .. CollateralInstrument",
        ("FIX.5.0", "896", "3"): "3 .. SubstitutionEligible",
        ("FIX.5.0", "896", "4"): "4 .. NotAssigned",
        ("FIX.5.0", "896", "5"): "5 .. PartiallyAssigned",
        ("FIX.5.0", "896", "6"): "6 .. FullyAssigned",
        ("FIX.5.0", "896", "7"): "7 .. OutstandingTrades",
        ("FIX.5.0", "903", "0"): "0 .. New",
        ("FIX.5.0", "903", "1"): "1 .. Replace",
        ("FIX.5.0", "903", "2"): "2 .. Cancel",
        ("FIX.5.0", "903", "3"): "3 .. Release",
        ("FIX.5.0", "903", "4"): "4 .. Reverse",
        ("FIX.5.0", "905", "0"): "0 .. Received",
        ("FIX.5.0", "905", "1"): "1 .. Accepted",
        ("FIX.5.0", "905", "2"): "2 .. Declined",
        ("FIX.5.0", "905", "3"): "3 .. Rejected",
        ("FIX.5.0", "905", "4"): "4 .. TransactionPending",
        ("FIX.5.0", "905", "5"): "5 .. TransactionCompletedWithWarning",
        ("FIX.5.0", "906", "0"): "0 .. UnknownDeal",
        ("FIX.5.0", "906", "1"): "1 .. UnknownOrInvalidInstrument",
        ("FIX.5.0", "906", "2"): "2 .. UnauthorizedTransaction",
        ("FIX.5.0", "906", "3"): "3 .. InsufficientCollateral",
        ("FIX.5.0", "906", "4"): "4 .. InvalidTypeOfCollateral",
        ("FIX.5.0", "906", "5"): "5 .. ExcessiveSubstitution",
        ("FIX.5.0", "906", "99"): "99 .. Other",
        ("FIX.5.0", "910", "0"): "0 .. Unassigned",
        ("FIX.5.0", "910", "1"): "1 .. PartiallyAssigned",
        ("FIX.5.0", "910", "2"): "2 .. AssignmentProposed",
        ("FIX.5.0", "910", "3"): "3 .. Assigned",
        ("FIX.5.0", "910", "4"): "4 .. Challenged",
        ("FIX.5.0", "912", "N"): "N .. NotLastMessage",
        ("FIX.5.0", "912", "Y"): "Y .. LastMessage",
        ("FIX.5.0", "919", "0"): "0 .. VersusPayment",
        ("FIX.5.0", "919", "1"): "1 .. Free",
        ("FIX.5.0", "919", "2"): "2 .. TriParty",
        ("FIX.5.0", "919", "3"): "3 .. HoldInCustody",
        ("FIX.5.0", "924", "1"): "1 .. LogOnUser",
        ("FIX.5.0", "924", "2"): "2 .. LogOffUser",
        ("FIX.5.0", "924", "3"): "3 .. ChangePasswordForUser",
        ("FIX.5.0", "924", "4"): "4 .. RequestIndividualUserStatus",
        ("FIX.5.0", "924", "5"): "5 .. RequestThrottleLimit",
        ("FIX.5.0", "926", "1"): "1 .. LoggedIn",
        ("FIX.5.0", "926", "2"): "2 .. NotLoggedIn",
        ("FIX.5.0", "926", "3"): "3 .. UserNotRecognised",
        ("FIX.5.0", "926", "4"): "4 .. PasswordIncorrect",
        ("FIX.5.0", "926", "5"): "5 .. PasswordChanged",
        ("FIX.5.0", "926", "6"): "6 .. Other",
        ("FIX.5.0", "926", "7"): "7 .. ForcedUserLogoutByExchange",
        ("FIX.5.0", "926", "8"): "8 .. SessionShutdownWarning",
        ("FIX.5.0", "926", "9"): "9 .. ThrottleParametersChanged",
        ("FIX.5.0", "928", "1"): "1 .. Connected",
        ("FIX.5.0", "928", "2"): "2 .. NotConnectedUnexpected",
        ("FIX.5.0", "928", "3"): "3 .. NotConnectedExpected",
        ("FIX.5.0", "928", "4"): "4 .. InProcess",
        ("FIX.5.0", "935", "1"): "1 .. Snapshot",
        ("FIX.5.0", "935", "2"): "2 .. Subscribe",
        ("FIX.5.0", "935", "4"): "4 .. StopSubscribing",
        ("FIX.5.0", "935", "8"): "8 .. LevelOfDetail",
        ("FIX.5.0", "937", "1"): "1 .. Full",
        ("FIX.5.0", "937", "2"): "2 .. IncrementalUpdate",
        ("FIX.5.0", "939", "0"): "0 .. Accepted",
        ("FIX.5.0", "939", "1"): "1 .. Rejected",
        ("FIX.5.0", "939", "2"): "2 .. Cancelled",
        ("FIX.5.0", "939", "3"): "3 .. AcceptedWithErrors",
        ("FIX.5.0", "939", "4"): "4 .. PendingNew",
        ("FIX.5.0", "939", "5"): "5 .. PendingCancel",
        ("FIX.5.0", "939", "6"): "6 .. PendingReplace",
        ("FIX.5.0", "939", "7"): "7 .. Terminated",
        ("FIX.5.0", "939", "8"): "8 .. PendingVerification",
        ("FIX.5.0", "939", "9"): "9 .. DeemedVerified",
        ("FIX.5.0", "939", "10"): "10 .. Verified",
        ("FIX.5.0", "939", "11"): "11 .. Disputed",
        ("FIX.5.0", "940", "1"): "1 .. Received",
        ("FIX.5.0", "940", "2"): "2 .. ConfirmRejected",
        ("FIX.5.0", "940", "3"): "3 .. Affirmed",
        ("FIX.5.0", "944", "0"): "0 .. Retain",
        ("FIX.5.0", "944", "1"): "1 .. Add",
        ("FIX.5.0", "944", "2"): "2 .. Remove",
        ("FIX.5.0", "945", "0"): "0 .. Accepted",
        ("FIX.5.0", "945", "1"): "1 .. AcceptedWithWarnings",
        ("FIX.5.0", "945", "2"): "2 .. Completed",
        ("FIX.5.0", "945", "3"): "3 .. CompletedWithWarnings",
        ("FIX.5.0", "945", "4"): "4 .. Rejected",
        ("FIX.5.0", "946", "0"): "0 .. Successful",
        ("FIX.5.0", "946", "1"): "1 .. InvalidOrUnknownInstrument",
        ("FIX.5.0", "946", "2"): "2 .. InvalidOrUnknownCollateralType",
        ("FIX.5.0", "946", "3"): "3 .. InvalidParties",
        ("FIX.5.0", "946", "4"): "4 .. InvalidTransportTypeRequested",
        ("FIX.5.0", "946", "5"): "5 .. InvalidDestinationRequested",
        ("FIX.5.0", "946", "6"): "6 .. NoCollateralFoundForTheTradeSpecified",
        ("FIX.5.0", "946", "7"): "7 .. NoCollateralFoundForTheOrderSpecified",
        ("FIX.5.0", "946", "8"): "8 .. CollateralInquiryTypeNotSupported",
        ("FIX.5.0", "946", "9"): "9 .. UnauthorizedForCollateralInquiry",
        ("FIX.5.0", "946", "99"): "99 .. Other",
        ("FIX.5.0", "950", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "950", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "950", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "950", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "950", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "950", "J"): "J .. TaxID",
        ("FIX.5.0", "950", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "950", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "950", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "950", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "950", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "950", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "950", "B"): "B .. BIC",
        ("FIX.5.0", "950", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "950", "D"): "D .. Proprietary",
        ("FIX.5.0", "950", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "950", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "950", "G"): "G .. MIC",
        ("FIX.5.0", "950", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "950", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "950", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "950", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "950", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "950", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "950", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "950", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "951", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "951", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "951", "3"): "3 .. ClientID",
        ("FIX.5.0", "951", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "951", "5"): "5 .. InvestorID",
        ("FIX.5.0", "951", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "951", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "951", "8"): "8 .. Locate",
        ("FIX.5.0", "951", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "951", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "951", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "951", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "951", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "951", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "951", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "951", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "951", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "951", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "951", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "951", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "951", "22"): "22 .. Exchange",
        ("FIX.5.0", "951", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "951", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "951", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "951", "27"): "27 .. Buyer",
        ("FIX.5.0", "951", "28"): "28 .. Custodian",
        ("FIX.5.0", "951", "29"): "29 .. Intermediary",
        ("FIX.5.0", "951", "30"): "30 .. Agent",
        ("FIX.5.0", "951", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "951", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "951", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "951", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "951", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "951", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "951", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "951", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "951", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "951", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "951", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "951", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "951", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "951", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "951", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "951", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "951", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "951", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "951", "49"): "49 .. AssetManager",
        ("FIX.5.0", "951", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "951", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "951", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "951", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "951", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "951", "55"): "55 .. SessionID",
        ("FIX.5.0", "951", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "951", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "951", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "951", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "951", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "951", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "951", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "951", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "951", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "951", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "951", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "951", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "951", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "951", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "951", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "951", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "951", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "951", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "951", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "951", "75"): "75 .. LocationID",
        ("FIX.5.0", "951", "76"): "76 .. DeskID",
        ("FIX.5.0", "951", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "951", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "951", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "951", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "951", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "951", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "951", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "951", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "951", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "951", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "951", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "951", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "951", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "951", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "951", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "951", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "951", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "951", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "951", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "951", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "951", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "951", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "951", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "951", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "951", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "951", "102"): "102 .. DataRepository",
        ("FIX.5.0", "951", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "951", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "951", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "951", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "951", "107"): "107 .. Correspondent",
        ("FIX.5.0", "951", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "951", "110"): "110 .. Borrower",
        ("FIX.5.0", "951", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "951", "112"): "112 .. Guarantor",
        ("FIX.5.0", "951", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "951", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "951", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "951", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "951", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "951", "118"): "118 .. Operator",
        ("FIX.5.0", "951", "119"): "119 .. CSD",
        ("FIX.5.0", "951", "120"): "120 .. ICSD",
        ("FIX.5.0", "951", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "951", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "951", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "954", "1"): "1 .. Firm",
        ("FIX.5.0", "954", "2"): "2 .. Person",
        ("FIX.5.0", "954", "3"): "3 .. System",
        ("FIX.5.0", "954", "4"): "4 .. Application",
        ("FIX.5.0", "954", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "954", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "954", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "954", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "954", "9"): "9 .. ContactName",
        ("FIX.5.0", "954", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "954", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "954", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "954", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "954", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "954", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "954", "16"): "16 .. BIC",
        ("FIX.5.0", "954", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "954", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "954", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "954", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "954", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "954", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "954", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "954", "24"): "24 .. Department",
        ("FIX.5.0", "954", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "954", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "954", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "954", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "954", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "954", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "954", "31"): "31 .. Location",
        ("FIX.5.0", "954", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "954", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "954", "34"): "34 .. AddressCity",
        ("FIX.5.0", "954", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "954", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "954", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "954", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "954", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "954", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "954", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "954", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "954", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "954", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "954", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "954", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "954", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "954", "48"): "48 .. USPerson",
        ("FIX.5.0", "954", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "954", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "954", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "954", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "954", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "954", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "954", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "954", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "954", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "954", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "954", "59"): "59 .. FCMCode",
        ("FIX.5.0", "954", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "954", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "954", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "954", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "954", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "954", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "954", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "954", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "954", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "954", "69"): "69 .. EndUser",
        ("FIX.5.0", "954", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "954", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "954", "72"): "72 .. Domicile",
        ("FIX.5.0", "954", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "954", "74"): "74 .. Payer",
        ("FIX.5.0", "954", "75"): "75 .. Receiver",
        ("FIX.5.0", "954", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "954", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "954", "78"): "78 .. FirstName",
        ("FIX.5.0", "954", "79"): "79 .. Surname",
        ("FIX.5.0", "954", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "954", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "954", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "954", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "954", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "959", "1"): "1 .. Int",
        ("FIX.5.0", "959", "2"): "2 .. Length",
        ("FIX.5.0", "959", "3"): "3 .. NumInGroup",
        ("FIX.5.0", "959", "4"): "4 .. SeqNum",
        ("FIX.5.0", "959", "5"): "5 .. TagNum",
        ("FIX.5.0", "959", "6"): "6 .. Float",
        ("FIX.5.0", "959", "7"): "7 .. Qty",
        ("FIX.5.0", "959", "8"): "8 .. Price",
        ("FIX.5.0", "959", "9"): "9 .. PriceOffset",
        ("FIX.5.0", "959", "10"): "10 .. Amt",
        ("FIX.5.0", "959", "11"): "11 .. Percentage",
        ("FIX.5.0", "959", "12"): "12 .. Char",
        ("FIX.5.0", "959", "13"): "13 .. Boolean",
        ("FIX.5.0", "959", "14"): "14 .. String",
        ("FIX.5.0", "959", "15"): "15 .. MultipleCharValue",
        ("FIX.5.0", "959", "16"): "16 .. Currency",
        ("FIX.5.0", "959", "17"): "17 .. Exchange",
        ("FIX.5.0", "959", "18"): "18 .. MonthYear",
        ("FIX.5.0", "959", "19"): "19 .. UTCTimestamp",
        ("FIX.5.0", "959", "20"): "20 .. UTCTimeOnly",
        ("FIX.5.0", "959", "21"): "21 .. LocalMktDate",
        ("FIX.5.0", "959", "22"): "22 .. UTCDateOnly",
        ("FIX.5.0", "959", "23"): "23 .. Data",
        ("FIX.5.0", "959", "24"): "24 .. MultipleStringValue",
        ("FIX.5.0", "959", "25"): "25 .. Country",
        ("FIX.5.0", "959", "26"): "26 .. Language",
        ("FIX.5.0", "959", "27"): "27 .. TZTimeOnly",
        ("FIX.5.0", "959", "28"): "28 .. TZTimestamp",
        ("FIX.5.0", "959", "29"): "29 .. Tenor",
        ("FIX.5.0", "965", "1"): "1 .. Active",
        ("FIX.5.0", "965", "2"): "2 .. Inactive",
        ("FIX.5.0", "965", "3"): "3 .. ActiveClosingOrdersOnly",
        ("FIX.5.0", "965", "4"): "4 .. Expired",
        ("FIX.5.0", "965", "5"): "5 .. Delisted",
        ("FIX.5.0", "965", "6"): "6 .. KnockedOut",
        ("FIX.5.0", "965", "7"): "7 .. KnockOutRevoked",
        ("FIX.5.0", "965", "8"): "8 .. PendingExpiry",
        ("FIX.5.0", "965", "9"): "9 .. Suspended",
        ("FIX.5.0", "965", "10"): "10 .. Published",
        ("FIX.5.0", "965", "11"): "11 .. PendingDeletion",
        ("FIX.5.0", "974", "FIXED"): "FIXED .. FIXED",
        ("FIX.5.0", "974", "DIFF"): "DIFF .. DIFF",
        ("FIX.5.0", "975", "2"): "2 .. TPlus1",
        ("FIX.5.0", "975", "4"): "4 .. TPlus3",
        ("FIX.5.0", "975", "5"): "5 .. TPlus4",
        ("FIX.5.0", "980", "A"): "A .. Add",
        ("FIX.5.0", "980", "D"): "D .. Delete",
        ("FIX.5.0", "980", "M"): "M .. Modify",
        ("FIX.5.0", "982", "1"): "1 .. AutoExercise",
        ("FIX.5.0", "982", "2"): "2 .. NonAutoExercise",
        ("FIX.5.0", "982", "3"): "3 .. FinalWillBeExercised",
        ("FIX.5.0", "982", "4"): "4 .. ContraryIntention",
        ("FIX.5.0", "982", "5"): "5 .. Difference",
        ("FIX.5.0", "992", "1"): "1 .. SubAllocate",
        ("FIX.5.0", "992", "2"): "2 .. ThirdPartyAllocation",
        ("FIX.5.0", "996", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "996", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "996", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "996", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "996", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "996", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "996", "thm"): "thm .. therms",
        ("FIX.5.0", "996", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "996", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "996", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "996", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "996", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "996", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "996", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "996", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "996", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "996", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "996", "day"): "day .. Day",
        ("FIX.5.0", "996", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "996", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "996", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "996", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "996", "g"): "g .. Grams",
        ("FIX.5.0", "996", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "996", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "996", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "996", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "996", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "996", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "996", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "996", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "996", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "996", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "996", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "996", "L"): "L .. liters",
        ("FIX.5.0", "996", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "996", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "996", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "996", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "996", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "996", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "996", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "996", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "996", "t"): "t .. MetricTons",
        ("FIX.5.0", "996", "tn"): "tn .. Tons",
        ("FIX.5.0", "997", "H"): "H .. Hour",
        ("FIX.5.0", "997", "Min"): "Min .. Minute",
        ("FIX.5.0", "997", "S"): "S .. Second",
        ("FIX.5.0", "997", "D"): "D .. Day",
        ("FIX.5.0", "997", "Wk"): "Wk .. Week",
        ("FIX.5.0", "997", "Mo"): "Mo .. Month",
        ("FIX.5.0", "997", "Yr"): "Yr .. Year",
        ("FIX.5.0", "997", "Q"): "Q .. Quarter",
        ("FIX.5.0", "998", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "998", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "998", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "998", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "998", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "998", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "998", "thm"): "thm .. therms",
        ("FIX.5.0", "998", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "998", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "998", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "998", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "998", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "998", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "998", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "998", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "998", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "998", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "998", "day"): "day .. Day",
        ("FIX.5.0", "998", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "998", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "998", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "998", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "998", "g"): "g .. Grams",
        ("FIX.5.0", "998", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "998", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "998", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "998", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "998", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "998", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "998", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "998", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "998", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "998", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "998", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "998", "L"): "L .. liters",
        ("FIX.5.0", "998", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "998", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "998", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "998", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "998", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "998", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "998", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "998", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "998", "t"): "t .. MetricTons",
        ("FIX.5.0", "998", "tn"): "tn .. Tons",
        ("FIX.5.0", "999", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "999", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "999", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "999", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "999", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "999", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "999", "thm"): "thm .. therms",
        ("FIX.5.0", "999", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "999", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "999", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "999", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "999", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "999", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "999", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "999", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "999", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "999", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "999", "day"): "day .. Day",
        ("FIX.5.0", "999", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "999", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "999", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "999", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "999", "g"): "g .. Grams",
        ("FIX.5.0", "999", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "999", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "999", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "999", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "999", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "999", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "999", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "999", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "999", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "999", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "999", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "999", "L"): "L .. liters",
        ("FIX.5.0", "999", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "999", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "999", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "999", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "999", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "999", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "999", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "999", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "999", "t"): "t .. MetricTons",
        ("FIX.5.0", "999", "tn"): "tn .. Tons",
        ("FIX.5.0", "1000", "H"): "H .. Hour",
        ("FIX.5.0", "1000", "Min"): "Min .. Minute",
        ("FIX.5.0", "1000", "S"): "S .. Second",
        ("FIX.5.0", "1000", "D"): "D .. Day",
        ("FIX.5.0", "1000", "Wk"): "Wk .. Week",
        ("FIX.5.0", "1000", "Mo"): "Mo .. Month",
        ("FIX.5.0", "1000", "Yr"): "Yr .. Year",
        ("FIX.5.0", "1000", "Q"): "Q .. Quarter",
        ("FIX.5.0", "1001", "H"): "H .. Hour",
        ("FIX.5.0", "1001", "Min"): "Min .. Minute",
        ("FIX.5.0", "1001", "S"): "S .. Second",
        ("FIX.5.0", "1001", "D"): "D .. Day",
        ("FIX.5.0", "1001", "Wk"): "Wk .. Week",
        ("FIX.5.0", "1001", "Mo"): "Mo .. Month",
        ("FIX.5.0", "1001", "Yr"): "Yr .. Year",
        ("FIX.5.0", "1001", "Q"): "Q .. Quarter",
        ("FIX.5.0", "1002", "1"): "1 .. Automatic",
        ("FIX.5.0", "1002", "2"): "2 .. Guarantor",
        ("FIX.5.0", "1002", "3"): "3 .. Manual",
        ("FIX.5.0", "1002", "4"): "4 .. BrokerAssigned",
        ("FIX.5.0", "1008", "0"): "0 .. CMTA",
        ("FIX.5.0", "1008", "1"): "1 .. InternalTransferOrAdjustment",
        ("FIX.5.0", "1008", "2"): "2 .. ExternalTransferOrTransferOfAccount",
        ("FIX.5.0", "1008", "3"): "3 .. RejectForSubmittingSide",
        ("FIX.5.0", "1008", "4"): "4 .. AdvisoryForContraSide",
        ("FIX.5.0", "1008", "5"): "5 .. OffsetDueToAnAllocation",
        ("FIX.5.0", "1008", "6"): "6 .. OnsetDueToAnAllocation",
        ("FIX.5.0", "1008", "7"): "7 .. DifferentialSpread",
        ("FIX.5.0", "1008", "8"): "8 .. ImpliedSpreadLegExecutedAgainstAnOutright",
        ("FIX.5.0", "1008", "9"): "9 .. TransactionFromExercise",
        ("FIX.5.0", "1008", "10"): "10 .. TransactionFromAssignment",
        ("FIX.5.0", "1008", "11"): "11 .. ACATS",
        ("FIX.5.0", "1008", "33"): "33 .. OffHoursTrade",
        ("FIX.5.0", "1008", "34"): "34 .. OnHoursTrade",
        ("FIX.5.0", "1008", "35"): "35 .. OTCQuote",
        ("FIX.5.0", "1008", "36"): "36 .. ConvertedSWAP",
        ("FIX.5.0", "1008", "40"): "40 .. WashTrade",
        ("FIX.5.0", "1008", "41"): "41 .. TradeAtSettlement",
        ("FIX.5.0", "1008", "42"): "42 .. AuctionTrade",
        ("FIX.5.0", "1008", "43"): "43 .. TradeAtMarker",
        ("FIX.5.0", "1008", "44"): "44 .. CreditDefault",
        ("FIX.5.0", "1008", "45"): "45 .. CreditRestructuring",
        ("FIX.5.0", "1008", "46"): "46 .. Merger",
        ("FIX.5.0", "1008", "47"): "47 .. SpinOff",
        ("FIX.5.0", "1008", "48"): "48 .. MultilateralCompression",
        ("FIX.5.0", "1008", "50"): "50 .. Balancing",
        ("FIX.5.0", "1008", "51"): "51 .. BasisTradeIndexClose",
        ("FIX.5.0", "1008", "14"): "14 .. AI",
        ("FIX.5.0", "1008", "15"): "15 .. B",
        ("FIX.5.0", "1008", "16"): "16 .. K",
        ("FIX.5.0", "1008", "17"): "17 .. LC",
        ("FIX.5.0", "1008", "18"): "18 .. M",
        ("FIX.5.0", "1008", "19"): "19 .. N",
        ("FIX.5.0", "1008", "20"): "20 .. NM",
        ("FIX.5.0", "1008", "21"): "21 .. NR",
        ("FIX.5.0", "1008", "22"): "22 .. P",
        ("FIX.5.0", "1008", "23"): "23 .. PA",
        ("FIX.5.0", "1008", "24"): "24 .. PC",
        ("FIX.5.0", "1008", "25"): "25 .. PN",
        ("FIX.5.0", "1008", "26"): "26 .. R",
        ("FIX.5.0", "1008", "27"): "27 .. RO",
        ("FIX.5.0", "1008", "28"): "28 .. RT",
        ("FIX.5.0", "1008", "29"): "29 .. SW",
        ("FIX.5.0", "1008", "30"): "30 .. T",
        ("FIX.5.0", "1008", "31"): "31 .. WN",
        ("FIX.5.0", "1008", "32"): "32 .. WT",
        ("FIX.5.0", "1008", "37"): "37 .. CrossedTrade",
        ("FIX.5.0", "1008", "38"): "38 .. InterimProtectedTrade",
        ("FIX.5.0", "1008", "39"): "39 .. LargeInScale",
        ("FIX.5.0", "1013", "1"): "1 .. ExecutionTime",
        ("FIX.5.0", "1013", "2"): "2 .. TimeIn",
        ("FIX.5.0", "1013", "3"): "3 .. TimeOut",
        ("FIX.5.0", "1013", "4"): "4 .. BrokerReceipt",
        ("FIX.5.0", "1013", "5"): "5 .. BrokerExecution",
        ("FIX.5.0", "1013", "6"): "6 .. DeskReceipt",
        ("FIX.5.0", "1013", "7"): "7 .. SubmissionToClearing",
        ("FIX.5.0", "1013", "8"): "8 .. TimePriority",
        ("FIX.5.0", "1013", "9"): "9 .. OrderbookEntryTime",
        ("FIX.5.0", "1013", "10"): "10 .. OrderSubmissionTime",
        ("FIX.5.0", "1013", "11"): "11 .. PubliclyReported",
        ("FIX.5.0", "1013", "12"): "12 .. PublicReportUpdated",
        ("FIX.5.0", "1013", "13"): "13 .. NonPubliclyReported",
        ("FIX.5.0", "1013", "14"): "14 .. NonPublicReportUpdated",
        ("FIX.5.0", "1013", "15"): "15 .. SubmittedForConfirmation",
        ("FIX.5.0", "1013", "16"): "16 .. UpdatedForConfirmation",
        ("FIX.5.0", "1013", "17"): "17 .. Confirmed",
        ("FIX.5.0", "1013", "18"): "18 .. UpdatedForClearing",
        ("FIX.5.0", "1013", "19"): "19 .. Cleared",
        ("FIX.5.0", "1013", "20"): "20 .. AllocationsSubmitted",
        ("FIX.5.0", "1013", "21"): "21 .. AllocationsUpdated",
        ("FIX.5.0", "1013", "22"): "22 .. AllocationsCompleted",
        ("FIX.5.0", "1013", "23"): "23 .. SubmittedToRepository",
        ("FIX.5.0", "1013", "24"): "24 .. PostTrdContntnEvnt",
        ("FIX.5.0", "1013", "25"): "25 .. PostTradeValuation",
        ("FIX.5.0", "1013", "26"): "26 .. PreviousTimePriority",
        ("FIX.5.0", "1015", "0"): "0 .. False",
        ("FIX.5.0", "1015", "1"): "1 .. True",
        ("FIX.5.0", "1021", "1"): "1 .. TopOfBook",
        ("FIX.5.0", "1021", "2"): "2 .. PriceDepth",
        ("FIX.5.0", "1021", "3"): "3 .. OrderDepth",
        ("FIX.5.0", "1024", "0"): "0 .. Book",
        ("FIX.5.0", "1024", "1"): "1 .. OffBook",
        ("FIX.5.0", "1024", "2"): "2 .. Cross",
        ("FIX.5.0", "1024", "3"): "3 .. QuoteDrivenMarket",
        ("FIX.5.0", "1024", "4"): "4 .. DarkOrderBook",
        ("FIX.5.0", "1024", "5"): "5 .. AuctionDrivenMarket",
        ("FIX.5.0", "1024", "6"): "6 .. QuoteNegotiation",
        ("FIX.5.0", "1024", "7"): "7 .. VoiceNegotiation",
        ("FIX.5.0", "1024", "8"): "8 .. HybridMarket",
        ("FIX.5.0", "1031", "A"): "A .. PhoneSimple",
        ("FIX.5.0", "1031", "B"): "B .. PhoneComplex",
        ("FIX.5.0", "1031", "C"): "C .. FCMProvidedScreen",
        ("FIX.5.0", "1031", "D"): "D .. OtherProvidedScreen",
        ("FIX.5.0", "1031", "E"): "E .. ClientProvidedPlatformControlledByFCM",
        ("FIX.5.0", "1031", "F"): "F .. ClientProvidedPlatformDirectToExchange",
        ("FIX.5.0", "1031", "H"): "H .. AlgoEngine",
        ("FIX.5.0", "1031", "J"): "J .. PriceAtExecution",
        ("FIX.5.0", "1031", "W"): "W .. DeskElectronic",
        ("FIX.5.0", "1031", "X"): "X .. DeskPit",
        ("FIX.5.0", "1031", "Y"): "Y .. ClientElectronic",
        ("FIX.5.0", "1031", "Z"): "Z .. ClientPit",
        ("FIX.5.0", "1031", "ADD"): "ADD .. AddOnOrder",
        ("FIX.5.0", "1031", "AON"): "AON .. AllOrNone",
        ("FIX.5.0", "1031", "CND"): "CND .. ConditionalOrder",
        ("FIX.5.0", "1031", "CNH"): "CNH .. CashNotHeld",
        ("FIX.5.0", "1031", "CSH"): "CSH .. DeliveryInstructionsCash",
        ("FIX.5.0", "1031", "DIR"): "DIR .. DirectedOrder",
        ("FIX.5.0", "1031", "DLO"): "DLO .. DiscretionaryLimitOrder",
        ("FIX.5.0", "1031", "E.W"): "E.W .. ExchangeForPhysicalTransaction",
        ("FIX.5.0", "1031", "FOK"): "FOK .. FillOrKill",
        ("FIX.5.0", "1031", "IDX"): "IDX .. IntraDayCross",
        ("FIX.5.0", "1031", "IO"): "IO .. ImbalanceOnly",
        ("FIX.5.0", "1031", "IOC"): "IOC .. ImmediateOrCancel",
        ("FIX.5.0", "1031", "ISO"): "ISO .. IntermarketSweepOrder",
        ("FIX.5.0", "1031", "LOO"): "LOO .. LimitOnOpen",
        ("FIX.5.0", "1031", "LOC"): "LOC .. LimitOnClose",
        ("FIX.5.0", "1031", "MAO"): "MAO .. MarketAtOpen",
        ("FIX.5.0", "1031", "MAC"): "MAC .. MarketAtClose",
        ("FIX.5.0", "1031", "MOO"): "MOO .. MarketOnOpen",
        ("FIX.5.0", "1031", "MOC"): "MOC .. MarketOnClose",
        ("FIX.5.0", "1031", "MPT"): "MPT .. MergerRelatedTransferPosition",
        ("FIX.5.0", "1031", "MQT"): "MQT .. MinimumQuantity",
        ("FIX.5.0", "1031", "MTL"): "MTL .. MarketToLimit",
        ("FIX.5.0", "1031", "ND"): "ND .. DeliveryInstructionsNextDay",
        ("FIX.5.0", "1031", "NH"): "NH .. NotHeld",
        ("FIX.5.0", "1031", "OPT"): "OPT .. OptionsRelatedTransaction",
        ("FIX.5.0", "1031", "OVD"): "OVD .. OverTheDay",
        ("FIX.5.0", "1031", "PEG"): "PEG .. Pegged",
        ("FIX.5.0", "1031", "RSV"): "RSV .. ReserveSizeOrder",
        ("FIX.5.0", "1031", "S.W"): "S.W .. StopStockTransaction",
        ("FIX.5.0", "1031", "SCL"): "SCL .. Scale",
        ("FIX.5.0", "1031", "SLR"): "SLR .. DeliveryInstructionsSellersOption",
        ("FIX.5.0", "1031", "TMO"): "TMO .. TimeOrder",
        ("FIX.5.0", "1031", "TS"): "TS .. TrailingStop",
        ("FIX.5.0", "1031", "WRK"): "WRK .. Work",
        ("FIX.5.0", "1031", "F0"): "F0 .. StayOnOfferside",
        ("FIX.5.0", "1031", "F3"): "F3 .. GoAlong",
        ("FIX.5.0", "1031", "F6"): "F6 .. ParticipateDoNotInitiate",
        ("FIX.5.0", "1031", "F7"): "F7 .. StrictScale",
        ("FIX.5.0", "1031", "F8"): "F8 .. TryToScale",
        ("FIX.5.0", "1031", "F9"): "F9 .. StayOnBidside",
        ("FIX.5.0", "1031", "FA"): "FA .. NoCross",
        ("FIX.5.0", "1031", "FB"): "FB .. OKToCross",
        ("FIX.5.0", "1031", "FC"): "FC .. CallFirst",
        ("FIX.5.0", "1031", "FD"): "FD .. PercentOfVolume",
        ("FIX.5.0", "1031", "FH"): "FH .. ReinstateOnSystemFailure",
        ("FIX.5.0", "1031", "FI"): "FI .. InstitutionOnly",
        ("FIX.5.0", "1031", "FJ"): "FJ .. ReinstateOnTradingHalt",
        ("FIX.5.0", "1031", "FK"): "FK .. CancelOnTradingHalf",
        ("FIX.5.0", "1031", "FL"): "FL .. LastPeg",
        ("FIX.5.0", "1031", "FM"): "FM .. MidPricePeg",
        ("FIX.5.0", "1031", "FN"): "FN .. NonNegotiable",
        ("FIX.5.0", "1031", "FO"): "FO .. OpeningPeg",
        ("FIX.5.0", "1031", "FP"): "FP .. MarketPeg",
        ("FIX.5.0", "1031", "FQ"): "FQ .. CancelOnSystemFailure",
        ("FIX.5.0", "1031", "FR"): "FR .. PrimaryPeg",
        ("FIX.5.0", "1031", "FS"): "FS .. Suspend",
        ("FIX.5.0", "1031", "FT"): "FT .. FixedPegToLocalBBO",
        ("FIX.5.0", "1031", "FW"): "FW .. PegToVWAP",
        ("FIX.5.0", "1031", "FX"): "FX .. TradeAlong",
        ("FIX.5.0", "1031", "FY"): "FY .. TryToStop",
        ("FIX.5.0", "1031", "FZ"): "FZ .. CancelIfNotBest",
        ("FIX.5.0", "1031", "Fb"): "Fb .. StrictLimit",
        ("FIX.5.0", "1031", "Fc"): "Fc .. IgnorePriceValidityChecks",
        ("FIX.5.0", "1031", "Fd"): "Fd .. PegToLimitPrice",
        ("FIX.5.0", "1031", "Fe"): "Fe .. WorkToTargetStrategy",
        ("FIX.5.0", "1031", "G"): "G .. GOrderAndFCMAPIorFIX",
        ("FIX.5.0", "1032", "1"): "1 .. FINRAOATS",
        ("FIX.5.0", "1032", "2"): "2 .. FIAExecutionSourceCode",
        ("FIX.5.0", "1033", "A"): "A .. Agency",
        ("FIX.5.0", "1033", "AR"): "AR .. Arbitrage",
        ("FIX.5.0", "1033", "B"): "B .. BlockTrading",
        ("FIX.5.0", "1033", "C"): "C .. ConvertibleDesk",
        ("FIX.5.0", "1033", "CR"): "CR .. CentralRiskBooks",
        ("FIX.5.0", "1033", "D"): "D .. Derivatives",
        ("FIX.5.0", "1033", "EC"): "EC .. EquityCapitalMarkets",
        ("FIX.5.0", "1033", "IN"): "IN .. International",
        ("FIX.5.0", "1033", "IS"): "IS .. Institutional",
        ("FIX.5.0", "1033", "O"): "O .. Other",
        ("FIX.5.0", "1033", "PF"): "PF .. PreferredTrading",
        ("FIX.5.0", "1033", "PR"): "PR .. Proprietary",
        ("FIX.5.0", "1033", "PT"): "PT .. ProgramTrading",
        ("FIX.5.0", "1033", "S"): "S .. Sales",
        ("FIX.5.0", "1033", "SW"): "SW .. Swaps",
        ("FIX.5.0", "1033", "T"): "T .. TradingDeskSystem",
        ("FIX.5.0", "1033", "TR"): "TR .. Treasury",
        ("FIX.5.0", "1034", "1"): "1 .. FINRAOATS",
        ("FIX.5.0", "1035", "A"): "A .. PhoneSimple",
        ("FIX.5.0", "1035", "B"): "B .. PhoneComplex",
        ("FIX.5.0", "1035", "C"): "C .. FCMProvidedScreen",
        ("FIX.5.0", "1035", "D"): "D .. OtherProvidedScreen",
        ("FIX.5.0", "1035", "E"): "E .. ClientProvidedPlatformControlledByFCM",
        ("FIX.5.0", "1035", "F"): "F .. ClientProvidedPlatformDirectToExchange",
        ("FIX.5.0", "1035", "H"): "H .. AlgoEngine",
        ("FIX.5.0", "1035", "J"): "J .. PriceAtExecution",
        ("FIX.5.0", "1035", "W"): "W .. DeskElectronic",
        ("FIX.5.0", "1035", "X"): "X .. DeskPit",
        ("FIX.5.0", "1035", "Y"): "Y .. ClientElectronic",
        ("FIX.5.0", "1035", "Z"): "Z .. ClientPit",
        ("FIX.5.0", "1035", "ADD"): "ADD .. AddOnOrder",
        ("FIX.5.0", "1035", "AON"): "AON .. AllOrNone",
        ("FIX.5.0", "1035", "CND"): "CND .. ConditionalOrder",
        ("FIX.5.0", "1035", "CNH"): "CNH .. CashNotHeld",
        ("FIX.5.0", "1035", "CSH"): "CSH .. DeliveryInstructionsCash",
        ("FIX.5.0", "1035", "DIR"): "DIR .. DirectedOrder",
        ("FIX.5.0", "1035", "DLO"): "DLO .. DiscretionaryLimitOrder",
        ("FIX.5.0", "1035", "E.W"): "E.W .. ExchangeForPhysicalTransaction",
        ("FIX.5.0", "1035", "FOK"): "FOK .. FillOrKill",
        ("FIX.5.0", "1035", "IDX"): "IDX .. IntraDayCross",
        ("FIX.5.0", "1035", "IO"): "IO .. ImbalanceOnly",
        ("FIX.5.0", "1035", "IOC"): "IOC .. ImmediateOrCancel",
        ("FIX.5.0", "1035", "ISO"): "ISO .. IntermarketSweepOrder",
        ("FIX.5.0", "1035", "LOO"): "LOO .. LimitOnOpen",
        ("FIX.5.0", "1035", "LOC"): "LOC .. LimitOnClose",
        ("FIX.5.0", "1035", "MAO"): "MAO .. MarketAtOpen",
        ("FIX.5.0", "1035", "MAC"): "MAC .. MarketAtClose",
        ("FIX.5.0", "1035", "MOO"): "MOO .. MarketOnOpen",
        ("FIX.5.0", "1035", "MOC"): "MOC .. MarketOnClose",
        ("FIX.5.0", "1035", "MPT"): "MPT .. MergerRelatedTransferPosition",
        ("FIX.5.0", "1035", "MQT"): "MQT .. MinimumQuantity",
        ("FIX.5.0", "1035", "MTL"): "MTL .. MarketToLimit",
        ("FIX.5.0", "1035", "ND"): "ND .. DeliveryInstructionsNextDay",
        ("FIX.5.0", "1035", "NH"): "NH .. NotHeld",
        ("FIX.5.0", "1035", "OPT"): "OPT .. OptionsRelatedTransaction",
        ("FIX.5.0", "1035", "OVD"): "OVD .. OverTheDay",
        ("FIX.5.0", "1035", "PEG"): "PEG .. Pegged",
        ("FIX.5.0", "1035", "RSV"): "RSV .. ReserveSizeOrder",
        ("FIX.5.0", "1035", "S.W"): "S.W .. StopStockTransaction",
        ("FIX.5.0", "1035", "SCL"): "SCL .. Scale",
        ("FIX.5.0", "1035", "SLR"): "SLR .. DeliveryInstructionsSellersOption",
        ("FIX.5.0", "1035", "TMO"): "TMO .. TimeOrder",
        ("FIX.5.0", "1035", "TS"): "TS .. TrailingStop",
        ("FIX.5.0", "1035", "WRK"): "WRK .. Work",
        ("FIX.5.0", "1035", "F0"): "F0 .. StayOnOfferside",
        ("FIX.5.0", "1035", "F3"): "F3 .. GoAlong",
        ("FIX.5.0", "1035", "F6"): "F6 .. ParticipateDoNotInitiate",
        ("FIX.5.0", "1035", "F7"): "F7 .. StrictScale",
        ("FIX.5.0", "1035", "F8"): "F8 .. TryToScale",
        ("FIX.5.0", "1035", "F9"): "F9 .. StayOnBidside",
        ("FIX.5.0", "1035", "FA"): "FA .. NoCross",
        ("FIX.5.0", "1035", "FB"): "FB .. OKToCross",
        ("FIX.5.0", "1035", "FC"): "FC .. CallFirst",
        ("FIX.5.0", "1035", "FD"): "FD .. PercentOfVolume",
        ("FIX.5.0", "1035", "FH"): "FH .. ReinstateOnSystemFailure",
        ("FIX.5.0", "1035", "FI"): "FI .. InstitutionOnly",
        ("FIX.5.0", "1035", "FJ"): "FJ .. ReinstateOnTradingHalt",
        ("FIX.5.0", "1035", "FK"): "FK .. CancelOnTradingHalf",
        ("FIX.5.0", "1035", "FL"): "FL .. LastPeg",
        ("FIX.5.0", "1035", "FM"): "FM .. MidPricePeg",
        ("FIX.5.0", "1035", "FN"): "FN .. NonNegotiable",
        ("FIX.5.0", "1035", "FO"): "FO .. OpeningPeg",
        ("FIX.5.0", "1035", "FP"): "FP .. MarketPeg",
        ("FIX.5.0", "1035", "FQ"): "FQ .. CancelOnSystemFailure",
        ("FIX.5.0", "1035", "FR"): "FR .. PrimaryPeg",
        ("FIX.5.0", "1035", "FS"): "FS .. Suspend",
        ("FIX.5.0", "1035", "FT"): "FT .. FixedPegToLocalBBO",
        ("FIX.5.0", "1035", "FW"): "FW .. PegToVWAP",
        ("FIX.5.0", "1035", "FX"): "FX .. TradeAlong",
        ("FIX.5.0", "1035", "FY"): "FY .. TryToStop",
        ("FIX.5.0", "1035", "FZ"): "FZ .. CancelIfNotBest",
        ("FIX.5.0", "1035", "Fb"): "Fb .. StrictLimit",
        ("FIX.5.0", "1035", "Fc"): "Fc .. IgnorePriceValidityChecks",
        ("FIX.5.0", "1035", "Fd"): "Fd .. PegToLimitPrice",
        ("FIX.5.0", "1035", "Fe"): "Fe .. WorkToTargetStrategy",
        ("FIX.5.0", "1035", "G"): "G .. GOrderAndFCMAPIorFIX",
        ("FIX.5.0", "1036", "0"): "0 .. Received",
        ("FIX.5.0", "1036", "1"): "1 .. Accepted",
        ("FIX.5.0", "1036", "2"): "2 .. Don",
        ("FIX.5.0", "1039", "C"): "C .. CashSettlementRequired",
        ("FIX.5.0", "1039", "P"): "P .. PhysicalSettlementRequired",
        ("FIX.5.0", "1039", "E"): "E .. Election",
        ("FIX.5.0", "1043", "0"): "0 .. SpecificDeposit",
        ("FIX.5.0", "1043", "1"): "1 .. General",
        ("FIX.5.0", "1046", "D"): "D .. Divide",
        ("FIX.5.0", "1046", "M"): "M .. Multiply",
        ("FIX.5.0", "1047", "O"): "O .. Open",
        ("FIX.5.0", "1047", "C"): "C .. Close",
        ("FIX.5.0", "1047", "R"): "R .. Rolled",
        ("FIX.5.0", "1047", "F"): "F .. FIFO",
        ("FIX.5.0", "1048", "A"): "A .. Agent",
        ("FIX.5.0", "1048", "P"): "P .. Principal",
        ("FIX.5.0", "1048", "R"): "R .. RisklessPrincipal",
        ("FIX.5.0", "1049", "P"): "P .. ProRata",
        ("FIX.5.0", "1049", "R"): "R .. Random",
        ("FIX.5.0", "1050", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1050", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1050", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1050", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1050", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1050", "J"): "J .. TaxID",
        ("FIX.5.0", "1050", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1050", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1050", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1050", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1050", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1050", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1050", "B"): "B .. BIC",
        ("FIX.5.0", "1050", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1050", "D"): "D .. Proprietary",
        ("FIX.5.0", "1050", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1050", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1050", "G"): "G .. MIC",
        ("FIX.5.0", "1050", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1050", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1050", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1050", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1050", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1050", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1050", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1050", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1051", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1051", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1051", "3"): "3 .. ClientID",
        ("FIX.5.0", "1051", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1051", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1051", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1051", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1051", "8"): "8 .. Locate",
        ("FIX.5.0", "1051", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1051", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1051", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1051", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1051", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1051", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1051", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1051", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1051", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1051", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1051", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1051", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1051", "22"): "22 .. Exchange",
        ("FIX.5.0", "1051", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1051", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1051", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1051", "27"): "27 .. Buyer",
        ("FIX.5.0", "1051", "28"): "28 .. Custodian",
        ("FIX.5.0", "1051", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1051", "30"): "30 .. Agent",
        ("FIX.5.0", "1051", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1051", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1051", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1051", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1051", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1051", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1051", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1051", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1051", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1051", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1051", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1051", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1051", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1051", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1051", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1051", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1051", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1051", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1051", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1051", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1051", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1051", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1051", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1051", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1051", "55"): "55 .. SessionID",
        ("FIX.5.0", "1051", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1051", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1051", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1051", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1051", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1051", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1051", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1051", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1051", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1051", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1051", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1051", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1051", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1051", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1051", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1051", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1051", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1051", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1051", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1051", "75"): "75 .. LocationID",
        ("FIX.5.0", "1051", "76"): "76 .. DeskID",
        ("FIX.5.0", "1051", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1051", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1051", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1051", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1051", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1051", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1051", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1051", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1051", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1051", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1051", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1051", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1051", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1051", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1051", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1051", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1051", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1051", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1051", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1051", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1051", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1051", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1051", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1051", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1051", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1051", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1051", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1051", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1051", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1051", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1051", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1051", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1051", "110"): "110 .. Borrower",
        ("FIX.5.0", "1051", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1051", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1051", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1051", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1051", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1051", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1051", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1051", "118"): "118 .. Operator",
        ("FIX.5.0", "1051", "119"): "119 .. CSD",
        ("FIX.5.0", "1051", "120"): "120 .. ICSD",
        ("FIX.5.0", "1051", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1051", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1051", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1054", "1"): "1 .. Firm",
        ("FIX.5.0", "1054", "2"): "2 .. Person",
        ("FIX.5.0", "1054", "3"): "3 .. System",
        ("FIX.5.0", "1054", "4"): "4 .. Application",
        ("FIX.5.0", "1054", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1054", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1054", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1054", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1054", "9"): "9 .. ContactName",
        ("FIX.5.0", "1054", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1054", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1054", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1054", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1054", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1054", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1054", "16"): "16 .. BIC",
        ("FIX.5.0", "1054", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1054", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1054", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1054", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1054", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1054", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1054", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1054", "24"): "24 .. Department",
        ("FIX.5.0", "1054", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1054", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1054", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1054", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1054", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1054", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1054", "31"): "31 .. Location",
        ("FIX.5.0", "1054", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1054", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1054", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1054", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1054", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1054", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1054", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1054", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1054", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1054", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1054", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1054", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1054", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1054", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1054", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1054", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1054", "48"): "48 .. USPerson",
        ("FIX.5.0", "1054", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1054", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1054", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1054", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1054", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1054", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1054", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1054", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1054", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1054", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1054", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1054", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1054", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1054", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1054", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1054", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1054", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1054", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1054", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1054", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1054", "69"): "69 .. EndUser",
        ("FIX.5.0", "1054", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1054", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1054", "72"): "72 .. Domicile",
        ("FIX.5.0", "1054", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1054", "74"): "74 .. Payer",
        ("FIX.5.0", "1054", "75"): "75 .. Receiver",
        ("FIX.5.0", "1054", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1054", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1054", "78"): "78 .. FirstName",
        ("FIX.5.0", "1054", "79"): "79 .. Surname",
        ("FIX.5.0", "1054", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1054", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1054", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1054", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1054", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1057", "Y"): "Y .. OrderInitiatorIsAggressor",
        ("FIX.5.0", "1057", "N"): "N .. OrderInitiatorIsPassive",
        ("FIX.5.0", "1060", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1060", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1060", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1060", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1060", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1060", "J"): "J .. TaxID",
        ("FIX.5.0", "1060", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1060", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1060", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1060", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1060", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1060", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1060", "B"): "B .. BIC",
        ("FIX.5.0", "1060", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1060", "D"): "D .. Proprietary",
        ("FIX.5.0", "1060", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1060", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1060", "G"): "G .. MIC",
        ("FIX.5.0", "1060", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1060", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1060", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1060", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1060", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1060", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1060", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1060", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1061", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1061", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1061", "3"): "3 .. ClientID",
        ("FIX.5.0", "1061", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1061", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1061", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1061", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1061", "8"): "8 .. Locate",
        ("FIX.5.0", "1061", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1061", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1061", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1061", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1061", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1061", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1061", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1061", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1061", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1061", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1061", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1061", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1061", "22"): "22 .. Exchange",
        ("FIX.5.0", "1061", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1061", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1061", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1061", "27"): "27 .. Buyer",
        ("FIX.5.0", "1061", "28"): "28 .. Custodian",
        ("FIX.5.0", "1061", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1061", "30"): "30 .. Agent",
        ("FIX.5.0", "1061", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1061", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1061", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1061", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1061", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1061", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1061", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1061", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1061", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1061", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1061", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1061", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1061", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1061", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1061", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1061", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1061", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1061", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1061", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1061", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1061", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1061", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1061", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1061", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1061", "55"): "55 .. SessionID",
        ("FIX.5.0", "1061", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1061", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1061", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1061", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1061", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1061", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1061", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1061", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1061", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1061", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1061", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1061", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1061", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1061", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1061", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1061", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1061", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1061", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1061", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1061", "75"): "75 .. LocationID",
        ("FIX.5.0", "1061", "76"): "76 .. DeskID",
        ("FIX.5.0", "1061", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1061", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1061", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1061", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1061", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1061", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1061", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1061", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1061", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1061", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1061", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1061", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1061", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1061", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1061", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1061", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1061", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1061", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1061", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1061", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1061", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1061", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1061", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1061", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1061", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1061", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1061", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1061", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1061", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1061", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1061", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1061", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1061", "110"): "110 .. Borrower",
        ("FIX.5.0", "1061", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1061", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1061", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1061", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1061", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1061", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1061", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1061", "118"): "118 .. Operator",
        ("FIX.5.0", "1061", "119"): "119 .. CSD",
        ("FIX.5.0", "1061", "120"): "120 .. ICSD",
        ("FIX.5.0", "1061", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1061", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1061", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1064", "1"): "1 .. Firm",
        ("FIX.5.0", "1064", "2"): "2 .. Person",
        ("FIX.5.0", "1064", "3"): "3 .. System",
        ("FIX.5.0", "1064", "4"): "4 .. Application",
        ("FIX.5.0", "1064", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1064", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1064", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1064", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1064", "9"): "9 .. ContactName",
        ("FIX.5.0", "1064", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1064", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1064", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1064", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1064", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1064", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1064", "16"): "16 .. BIC",
        ("FIX.5.0", "1064", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1064", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1064", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1064", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1064", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1064", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1064", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1064", "24"): "24 .. Department",
        ("FIX.5.0", "1064", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1064", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1064", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1064", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1064", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1064", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1064", "31"): "31 .. Location",
        ("FIX.5.0", "1064", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1064", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1064", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1064", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1064", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1064", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1064", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1064", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1064", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1064", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1064", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1064", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1064", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1064", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1064", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1064", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1064", "48"): "48 .. USPerson",
        ("FIX.5.0", "1064", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1064", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1064", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1064", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1064", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1064", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1064", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1064", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1064", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1064", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1064", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1064", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1064", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1064", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1064", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1064", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1064", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1064", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1064", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1064", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1064", "69"): "69 .. EndUser",
        ("FIX.5.0", "1064", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1064", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1064", "72"): "72 .. Domicile",
        ("FIX.5.0", "1064", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1064", "74"): "74 .. Payer",
        ("FIX.5.0", "1064", "75"): "75 .. Receiver",
        ("FIX.5.0", "1064", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1064", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1064", "78"): "78 .. FirstName",
        ("FIX.5.0", "1064", "79"): "79 .. Surname",
        ("FIX.5.0", "1064", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1064", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1064", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1064", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1064", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1070", "0"): "0 .. Indicative",
        ("FIX.5.0", "1070", "1"): "1 .. Tradeable",
        ("FIX.5.0", "1070", "2"): "2 .. RestrictedTradeable",
        ("FIX.5.0", "1070", "3"): "3 .. Counter",
        ("FIX.5.0", "1070", "4"): "4 .. IndicativeAndTradeable",
        ("FIX.5.0", "1081", "0"): "0 .. SecondaryOrderID",
        ("FIX.5.0", "1081", "1"): "1 .. OrderID",
        ("FIX.5.0", "1081", "2"): "2 .. MDEntryID",
        ("FIX.5.0", "1081", "3"): "3 .. QuoteEntryID",
        ("FIX.5.0", "1081", "4"): "4 .. OriginalOrderID",
        ("FIX.5.0", "1081", "5"): "5 .. QuoteID",
        ("FIX.5.0", "1081", "6"): "6 .. QuoteReqID",
        ("FIX.5.0", "1083", "1"): "1 .. Immediate",
        ("FIX.5.0", "1083", "2"): "2 .. Exhaust",
        ("FIX.5.0", "1084", "1"): "1 .. Initial",
        ("FIX.5.0", "1084", "2"): "2 .. New",
        ("FIX.5.0", "1084", "3"): "3 .. Random",
        ("FIX.5.0", "1084", "4"): "4 .. Undisclosed",
        ("FIX.5.0", "1092", "0"): "0 .. None",
        ("FIX.5.0", "1092", "1"): "1 .. Local",
        ("FIX.5.0", "1092", "2"): "2 .. National",
        ("FIX.5.0", "1092", "3"): "3 .. Global",
        ("FIX.5.0", "1093", "1"): "1 .. OddLot",
        ("FIX.5.0", "1093", "2"): "2 .. RoundLot",
        ("FIX.5.0", "1093", "3"): "3 .. BlockLot",
        ("FIX.5.0", "1093", "4"): "4 .. RoundLotBasedUpon",
        ("FIX.5.0", "1094", "1"): "1 .. LastPeg",
        ("FIX.5.0", "1094", "2"): "2 .. MidPricePeg",
        ("FIX.5.0", "1094", "3"): "3 .. OpeningPeg",
        ("FIX.5.0", "1094", "4"): "4 .. MarketPeg",
        ("FIX.5.0", "1094", "5"): "5 .. PrimaryPeg",
        ("FIX.5.0", "1094", "7"): "7 .. PegToVWAP",
        ("FIX.5.0", "1094", "8"): "8 .. TrailingStopPeg",
        ("FIX.5.0", "1094", "9"): "9 .. PegToLimitPrice",
        ("FIX.5.0", "1094", "10"): "10 .. ShortSaleMinPricePeg",
        ("FIX.5.0", "1096", "1"): "1 .. CUSIP",
        ("FIX.5.0", "1096", "2"): "2 .. SEDOL",
        ("FIX.5.0", "1096", "3"): "3 .. QUIK",
        ("FIX.5.0", "1096", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "1096", "5"): "5 .. RICCode",
        ("FIX.5.0", "1096", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "1096", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "1096", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "1096", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "1096", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "1096", "B"): "B .. Wertpapier",
        ("FIX.5.0", "1096", "C"): "C .. Dutch",
        ("FIX.5.0", "1096", "D"): "D .. Valoren",
        ("FIX.5.0", "1096", "E"): "E .. Sicovam",
        ("FIX.5.0", "1096", "F"): "F .. Belgian",
        ("FIX.5.0", "1096", "G"): "G .. Common",
        ("FIX.5.0", "1096", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "1096", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "1096", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "1096", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "1096", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "1096", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "1096", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "1096", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "1096", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "1096", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "1096", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "1096", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "1096", "U"): "U .. Synthetic",
        ("FIX.5.0", "1096", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "1096", "W"): "W .. IndexName",
        ("FIX.5.0", "1100", "1"): "1 .. PartialExecution",
        ("FIX.5.0", "1100", "2"): "2 .. SpecifiedTradingSession",
        ("FIX.5.0", "1100", "3"): "3 .. NextAuction",
        ("FIX.5.0", "1100", "4"): "4 .. PriceMovement",
        ("FIX.5.0", "1100", "5"): "5 .. OnOrderEntryOrModification",
        ("FIX.5.0", "1101", "1"): "1 .. Activate",
        ("FIX.5.0", "1101", "2"): "2 .. Modify",
        ("FIX.5.0", "1101", "3"): "3 .. Cancel",
        ("FIX.5.0", "1105", "1"): "1 .. CUSIP",
        ("FIX.5.0", "1105", "2"): "2 .. SEDOL",
        ("FIX.5.0", "1105", "3"): "3 .. QUIK",
        ("FIX.5.0", "1105", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "1105", "5"): "5 .. RICCode",
        ("FIX.5.0", "1105", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "1105", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "1105", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "1105", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "1105", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "1105", "B"): "B .. Wertpapier",
        ("FIX.5.0", "1105", "C"): "C .. Dutch",
        ("FIX.5.0", "1105", "D"): "D .. Valoren",
        ("FIX.5.0", "1105", "E"): "E .. Sicovam",
        ("FIX.5.0", "1105", "F"): "F .. Belgian",
        ("FIX.5.0", "1105", "G"): "G .. Common",
        ("FIX.5.0", "1105", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "1105", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "1105", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "1105", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "1105", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "1105", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "1105", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "1105", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "1105", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "1105", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "1105", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "1105", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "1105", "U"): "U .. Synthetic",
        ("FIX.5.0", "1105", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "1105", "W"): "W .. IndexName",
        ("FIX.5.0", "1107", "1"): "1 .. BestOffer",
        ("FIX.5.0", "1107", "2"): "2 .. LastTrade",
        ("FIX.5.0", "1107", "3"): "3 .. BestBid",
        ("FIX.5.0", "1107", "4"): "4 .. BestBidOrLastTrade",
        ("FIX.5.0", "1107", "5"): "5 .. BestOfferOrLastTrade",
        ("FIX.5.0", "1107", "6"): "6 .. BestMid",
        ("FIX.5.0", "1108", "0"): "0 .. None",
        ("FIX.5.0", "1108", "1"): "1 .. Local",
        ("FIX.5.0", "1108", "2"): "2 .. National",
        ("FIX.5.0", "1108", "3"): "3 .. Global",
        ("FIX.5.0", "1109", "U"): "U .. Up",
        ("FIX.5.0", "1109", "D"): "D .. Down",
        ("FIX.5.0", "1111", "1"): "1 .. Market",
        ("FIX.5.0", "1111", "2"): "2 .. Limit",
        ("FIX.5.0", "1115", "1"): "1 .. Order",
        ("FIX.5.0", "1115", "2"): "2 .. Quote",
        ("FIX.5.0", "1115", "3"): "3 .. PrivatelyNegotiatedTrade",
        ("FIX.5.0", "1115", "4"): "4 .. MultilegOrder",
        ("FIX.5.0", "1115", "5"): "5 .. LinkedOrder",
        ("FIX.5.0", "1115", "6"): "6 .. QuoteRequest",
        ("FIX.5.0", "1115", "7"): "7 .. ImpliedOrder",
        ("FIX.5.0", "1115", "8"): "8 .. CrossOrder",
        ("FIX.5.0", "1115", "9"): "9 .. StreamingPrice",
        ("FIX.5.0", "1115", "A"): "A .. InternalCrossOrder",
        ("FIX.5.0", "1118", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1118", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1118", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1118", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1118", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1118", "J"): "J .. TaxID",
        ("FIX.5.0", "1118", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1118", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1118", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1118", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1118", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1118", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1118", "B"): "B .. BIC",
        ("FIX.5.0", "1118", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1118", "D"): "D .. Proprietary",
        ("FIX.5.0", "1118", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1118", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1118", "G"): "G .. MIC",
        ("FIX.5.0", "1118", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1118", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1118", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1118", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1118", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1118", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1118", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1118", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1119", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1119", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1119", "3"): "3 .. ClientID",
        ("FIX.5.0", "1119", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1119", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1119", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1119", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1119", "8"): "8 .. Locate",
        ("FIX.5.0", "1119", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1119", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1119", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1119", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1119", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1119", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1119", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1119", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1119", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1119", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1119", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1119", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1119", "22"): "22 .. Exchange",
        ("FIX.5.0", "1119", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1119", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1119", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1119", "27"): "27 .. Buyer",
        ("FIX.5.0", "1119", "28"): "28 .. Custodian",
        ("FIX.5.0", "1119", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1119", "30"): "30 .. Agent",
        ("FIX.5.0", "1119", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1119", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1119", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1119", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1119", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1119", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1119", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1119", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1119", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1119", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1119", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1119", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1119", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1119", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1119", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1119", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1119", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1119", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1119", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1119", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1119", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1119", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1119", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1119", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1119", "55"): "55 .. SessionID",
        ("FIX.5.0", "1119", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1119", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1119", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1119", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1119", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1119", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1119", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1119", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1119", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1119", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1119", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1119", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1119", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1119", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1119", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1119", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1119", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1119", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1119", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1119", "75"): "75 .. LocationID",
        ("FIX.5.0", "1119", "76"): "76 .. DeskID",
        ("FIX.5.0", "1119", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1119", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1119", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1119", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1119", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1119", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1119", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1119", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1119", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1119", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1119", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1119", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1119", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1119", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1119", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1119", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1119", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1119", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1119", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1119", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1119", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1119", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1119", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1119", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1119", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1119", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1119", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1119", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1119", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1119", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1119", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1119", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1119", "110"): "110 .. Borrower",
        ("FIX.5.0", "1119", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1119", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1119", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1119", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1119", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1119", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1119", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1119", "118"): "118 .. Operator",
        ("FIX.5.0", "1119", "119"): "119 .. CSD",
        ("FIX.5.0", "1119", "120"): "120 .. ICSD",
        ("FIX.5.0", "1119", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1119", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1119", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1122", "1"): "1 .. Firm",
        ("FIX.5.0", "1122", "2"): "2 .. Person",
        ("FIX.5.0", "1122", "3"): "3 .. System",
        ("FIX.5.0", "1122", "4"): "4 .. Application",
        ("FIX.5.0", "1122", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1122", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1122", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1122", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1122", "9"): "9 .. ContactName",
        ("FIX.5.0", "1122", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1122", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1122", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1122", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1122", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1122", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1122", "16"): "16 .. BIC",
        ("FIX.5.0", "1122", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1122", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1122", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1122", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1122", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1122", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1122", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1122", "24"): "24 .. Department",
        ("FIX.5.0", "1122", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1122", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1122", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1122", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1122", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1122", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1122", "31"): "31 .. Location",
        ("FIX.5.0", "1122", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1122", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1122", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1122", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1122", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1122", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1122", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1122", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1122", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1122", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1122", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1122", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1122", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1122", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1122", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1122", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1122", "48"): "48 .. USPerson",
        ("FIX.5.0", "1122", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1122", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1122", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1122", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1122", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1122", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1122", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1122", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1122", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1122", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1122", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1122", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1122", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1122", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1122", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1122", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1122", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1122", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1122", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1122", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1122", "69"): "69 .. EndUser",
        ("FIX.5.0", "1122", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1122", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1122", "72"): "72 .. Domicile",
        ("FIX.5.0", "1122", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1122", "74"): "74 .. Payer",
        ("FIX.5.0", "1122", "75"): "75 .. Receiver",
        ("FIX.5.0", "1122", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1122", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1122", "78"): "78 .. FirstName",
        ("FIX.5.0", "1122", "79"): "79 .. Surname",
        ("FIX.5.0", "1122", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1122", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1122", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1122", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1122", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1123", "0"): "0 .. TradeConfirmation",
        ("FIX.5.0", "1123", "1"): "1 .. TwoPartyReport",
        ("FIX.5.0", "1123", "2"): "2 .. OnePartyReportForMatching",
        ("FIX.5.0", "1123", "3"): "3 .. OnePartyReportForPassThrough",
        ("FIX.5.0", "1123", "4"): "4 .. AutomatedFloorOrderRouting",
        ("FIX.5.0", "1123", "5"): "5 .. TwoPartyReportForClaim",
        ("FIX.5.0", "1123", "6"): "6 .. OnePartyReport",
        ("FIX.5.0", "1123", "7"): "7 .. ThirdPtyRptForPassThrough",
        ("FIX.5.0", "1123", "8"): "8 .. OnePartyReportAutoMatch",
        ("FIX.5.0", "1124", "0"): "0 .. TradeConfirmation",
        ("FIX.5.0", "1124", "1"): "1 .. TwoPartyReport",
        ("FIX.5.0", "1124", "2"): "2 .. OnePartyReportForMatching",
        ("FIX.5.0", "1124", "3"): "3 .. OnePartyReportForPassThrough",
        ("FIX.5.0", "1124", "4"): "4 .. AutomatedFloorOrderRouting",
        ("FIX.5.0", "1124", "5"): "5 .. TwoPartyReportForClaim",
        ("FIX.5.0", "1124", "6"): "6 .. OnePartyReport",
        ("FIX.5.0", "1124", "7"): "7 .. ThirdPtyRptForPassThrough",
        ("FIX.5.0", "1124", "8"): "8 .. OnePartyReportAutoMatch",
        ("FIX.5.0", "1128", "0"): "0 .. FIX27",
        ("FIX.5.0", "1128", "1"): "1 .. FIX30",
        ("FIX.5.0", "1128", "2"): "2 .. FIX40",
        ("FIX.5.0", "1128", "3"): "3 .. FIX41",
        ("FIX.5.0", "1128", "4"): "4 .. FIX42",
        ("FIX.5.0", "1128", "5"): "5 .. FIX43",
        ("FIX.5.0", "1128", "6"): "6 .. FIX44",
        ("FIX.5.0", "1128", "7"): "7 .. FIX50",
        ("FIX.5.0", "1128", "8"): "8 .. FIX50SP1",
        ("FIX.5.0", "1128", "9"): "9 .. FIX50SP2",
        ("FIX.5.0", "1130", "0"): "0 .. FIX27",
        ("FIX.5.0", "1130", "1"): "1 .. FIX30",
        ("FIX.5.0", "1130", "2"): "2 .. FIX40",
        ("FIX.5.0", "1130", "3"): "3 .. FIX41",
        ("FIX.5.0", "1130", "4"): "4 .. FIX42",
        ("FIX.5.0", "1130", "5"): "5 .. FIX43",
        ("FIX.5.0", "1130", "6"): "6 .. FIX44",
        ("FIX.5.0", "1130", "7"): "7 .. FIX50",
        ("FIX.5.0", "1130", "8"): "8 .. FIX50SP1",
        ("FIX.5.0", "1130", "9"): "9 .. FIX50SP2",
        ("FIX.5.0", "1133", "B"): "B .. BIC",
        ("FIX.5.0", "1133", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1133", "D"): "D .. Proprietary",
        ("FIX.5.0", "1133", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1133", "G"): "G .. MIC",
        ("FIX.5.0", "1137", "0"): "0 .. FIX27",
        ("FIX.5.0", "1137", "1"): "1 .. FIX30",
        ("FIX.5.0", "1137", "2"): "2 .. FIX40",
        ("FIX.5.0", "1137", "3"): "3 .. FIX41",
        ("FIX.5.0", "1137", "4"): "4 .. FIX42",
        ("FIX.5.0", "1137", "5"): "5 .. FIX43",
        ("FIX.5.0", "1137", "6"): "6 .. FIX44",
        ("FIX.5.0", "1137", "7"): "7 .. FIX50",
        ("FIX.5.0", "1137", "8"): "8 .. FIX50SP1",
        ("FIX.5.0", "1137", "9"): "9 .. FIX50SP2",
        ("FIX.5.0", "1144", "0"): "0 .. NotImplied",
        ("FIX.5.0", "1144", "1"): "1 .. ImpliedIn",
        ("FIX.5.0", "1144", "2"): "2 .. ImpliedOut",
        ("FIX.5.0", "1144", "3"): "3 .. BothImpliedInAndImpliedOut",
        ("FIX.5.0", "1159", "1"): "1 .. Preliminary",
        ("FIX.5.0", "1159", "2"): "2 .. Final",
        ("FIX.5.0", "1162", "C"): "C .. Cancel",
        ("FIX.5.0", "1162", "N"): "N .. New",
        ("FIX.5.0", "1162", "R"): "R .. Replace",
        ("FIX.5.0", "1162", "T"): "T .. Restate",
        ("FIX.5.0", "1164", "1"): "1 .. InstructionsOfBroker",
        ("FIX.5.0", "1164", "2"): "2 .. InstructionsForInstitution",
        ("FIX.5.0", "1164", "3"): "3 .. Investor",
        ("FIX.5.0", "1164", "4"): "4 .. BuyersSettlementInstructions",
        ("FIX.5.0", "1164", "5"): "5 .. SellersSettlementInstructions",
        ("FIX.5.0", "1167", "0"): "0 .. Accepted",
        ("FIX.5.0", "1167", "5"): "5 .. Rejected",
        ("FIX.5.0", "1167", "6"): "6 .. RemovedFromMarket",
        ("FIX.5.0", "1167", "7"): "7 .. Expired",
        ("FIX.5.0", "1167", "12"): "12 .. LockedMarketWarning",
        ("FIX.5.0", "1167", "13"): "13 .. CrossMarketWarning",
        ("FIX.5.0", "1167", "14"): "14 .. CanceledDueToLockMarket",
        ("FIX.5.0", "1167", "15"): "15 .. CanceledDueToCrossMarket",
        ("FIX.5.0", "1167", "16"): "16 .. Active",
        ("FIX.5.0", "1171", "Y"): "Y .. PrivateQuote",
        ("FIX.5.0", "1171", "N"): "N .. PublicQuote",
        ("FIX.5.0", "1172", "1"): "1 .. AllMarketParticipants",
        ("FIX.5.0", "1172", "2"): "2 .. SpecifiedMarketParticipants",
        ("FIX.5.0", "1172", "3"): "3 .. AllMarketMakers",
        ("FIX.5.0", "1172", "4"): "4 .. PrimaryMarketMaker",
        ("FIX.5.0", "1174", "1"): "1 .. OrderImbalance",
        ("FIX.5.0", "1174", "2"): "2 .. TradingResumes",
        ("FIX.5.0", "1174", "3"): "3 .. PriceVolatilityInterruption",
        ("FIX.5.0", "1174", "4"): "4 .. ChangeOfTradingSession",
        ("FIX.5.0", "1174", "5"): "5 .. ChangeOfTradingSubsession",
        ("FIX.5.0", "1174", "6"): "6 .. ChangeOfSecurityTradingStatus",
        ("FIX.5.0", "1174", "7"): "7 .. ChangeOfBookType",
        ("FIX.5.0", "1174", "8"): "8 .. ChangeOfMarketDepth",
        ("FIX.5.0", "1174", "9"): "9 .. CorporateAction",
        ("FIX.5.0", "1176", "1"): "1 .. ExchangeLast",
        ("FIX.5.0", "1176", "2"): "2 .. High",
        ("FIX.5.0", "1176", "3"): "3 .. AveragePrice",
        ("FIX.5.0", "1176", "4"): "4 .. Turnover",
        ("FIX.5.0", "1178", "1"): "1 .. Customer",
        ("FIX.5.0", "1178", "2"): "2 .. CustomerProfessional",
        ("FIX.5.0", "1178", "3"): "3 .. DoNotTradeThrough",
        ("FIX.5.0", "1191", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "1191", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "1191", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "1191", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "1191", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "1191", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "1191", "thm"): "thm .. therms",
        ("FIX.5.0", "1191", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "1191", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "1191", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "1191", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "1191", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "1191", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "1191", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "1191", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "1191", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "1191", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "1191", "day"): "day .. Day",
        ("FIX.5.0", "1191", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "1191", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "1191", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "1191", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "1191", "g"): "g .. Grams",
        ("FIX.5.0", "1191", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "1191", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "1191", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "1191", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "1191", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "1191", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "1191", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "1191", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "1191", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "1191", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "1191", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "1191", "L"): "L .. liters",
        ("FIX.5.0", "1191", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "1191", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "1191", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "1191", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "1191", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "1191", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "1191", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "1191", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "1191", "t"): "t .. MetricTons",
        ("FIX.5.0", "1191", "tn"): "tn .. Tons",
        ("FIX.5.0", "1193", "C"): "C .. CashSettlementRequired",
        ("FIX.5.0", "1193", "P"): "P .. PhysicalSettlementRequired",
        ("FIX.5.0", "1193", "E"): "E .. Election",
        ("FIX.5.0", "1194", "0"): "0 .. European",
        ("FIX.5.0", "1194", "1"): "1 .. American",
        ("FIX.5.0", "1194", "2"): "2 .. Bermuda",
        ("FIX.5.0", "1194", "99"): "99 .. Other",
        ("FIX.5.0", "1196", "STD"): "STD .. Standard",
        ("FIX.5.0", "1196", "INX"): "INX .. Index",
        ("FIX.5.0", "1196", "INT"): "INT .. InterestRateIndex",
        ("FIX.5.0", "1196", "PCTPAR"): "PCTPAR .. PercentOfPar",
        ("FIX.5.0", "1197", "EQTY"): "EQTY .. PremiumStyle",
        ("FIX.5.0", "1197", "FUT"): "FUT .. FuturesStyleMarkToMarket",
        ("FIX.5.0", "1197", "FUTDA"): "FUTDA .. FuturesStyleWithAnAttachedCashAdjustment",
        ("FIX.5.0", "1197", "CDS"): "CDS .. CDSStyleCollateralization",
        ("FIX.5.0", "1197", "CDSD"): "CDSD .. CDSInDeliveryUseRecoveryRateToCalculate",
        ("FIX.5.0", "1198", "0"): "0 .. PreListedOnly",
        ("FIX.5.0", "1198", "1"): "1 .. UserRequested",
        ("FIX.5.0", "1209", "0"): "0 .. RegularTrading",
        ("FIX.5.0", "1209", "1"): "1 .. VariableCabinet",
        ("FIX.5.0", "1209", "2"): "2 .. FixedCabinet",
        ("FIX.5.0", "1209", "3"): "3 .. TradedAsASpreadLeg",
        ("FIX.5.0", "1209", "5"): "5 .. TradedAsSpread",
        ("FIX.5.0", "1210", "1"): "1 .. Flat",
        ("FIX.5.0", "1210", "2"): "2 .. ZeroCoupon",
        ("FIX.5.0", "1210", "3"): "3 .. InterestBearing",
        ("FIX.5.0", "1210", "4"): "4 .. NoPeriodicPayments",
        ("FIX.5.0", "1210", "5"): "5 .. VariableRate",
        ("FIX.5.0", "1210", "6"): "6 .. LessFeeForPut",
        ("FIX.5.0", "1210", "7"): "7 .. SteppedCoupon",
        ("FIX.5.0", "1210", "8"): "8 .. CouponPeriod",
        ("FIX.5.0", "1210", "9"): "9 .. When",
        ("FIX.5.0", "1210", "10"): "10 .. OriginalIssueDiscount",
        ("FIX.5.0", "1210", "11"): "11 .. Callable",
        ("FIX.5.0", "1210", "12"): "12 .. EscrowedToMaturity",
        ("FIX.5.0", "1210", "13"): "13 .. EscrowedToRedemptionDate",
        ("FIX.5.0", "1210", "14"): "14 .. PreRefunded",
        ("FIX.5.0", "1210", "15"): "15 .. InDefault",
        ("FIX.5.0", "1210", "16"): "16 .. Unrated",
        ("FIX.5.0", "1210", "17"): "17 .. Taxable",
        ("FIX.5.0", "1210", "18"): "18 .. Indexed",
        ("FIX.5.0", "1210", "19"): "19 .. SubjectToAlternativeMinimumTax",
        ("FIX.5.0", "1210", "20"): "20 .. OriginalIssueDiscountPrice",
        ("FIX.5.0", "1210", "21"): "21 .. CallableBelowMaturityValue",
        ("FIX.5.0", "1210", "22"): "22 .. CallableWithoutNotice",
        ("FIX.5.0", "1210", "23"): "23 .. PriceTickRulesForSecurity",
        ("FIX.5.0", "1210", "24"): "24 .. TradeTypeEligibilityDetailsForSecurity",
        ("FIX.5.0", "1210", "25"): "25 .. InstrumentDenominator",
        ("FIX.5.0", "1210", "26"): "26 .. InstrumentNumerator",
        ("FIX.5.0", "1210", "27"): "27 .. InstrumentPricePrecision",
        ("FIX.5.0", "1210", "28"): "28 .. InstrumentStrikePrice",
        ("FIX.5.0", "1210", "29"): "29 .. TradeableIndicator",
        ("FIX.5.0", "1210", "30"): "30 .. InstrumentEligibleAnonOrders",
        ("FIX.5.0", "1210", "31"): "31 .. MinGuaranteedFillVolume",
        ("FIX.5.0", "1210", "32"): "32 .. MinGuaranteedFillStatus",
        ("FIX.5.0", "1210", "33"): "33 .. TradeAtSettlementEligibility",
        ("FIX.5.0", "1210", "34"): "34 .. TestInstrument",
        ("FIX.5.0", "1210", "35"): "35 .. DummyInstrument",
        ("FIX.5.0", "1210", "36"): "36 .. NegativeSettlementPriceEligibility",
        ("FIX.5.0", "1210", "37"): "37 .. NegativeStrikePriceEligibility",
        ("FIX.5.0", "1210", "38"): "38 .. USStdContractInd",
        ("FIX.5.0", "1210", "39"): "39 .. AdmittedToTradingOnTradingVenue",
        ("FIX.5.0", "1210", "40"): "40 .. AverageDailyNotionalAmount",
        ("FIX.5.0", "1210", "41"): "41 .. AverageDailyNumberTrades",
        ("FIX.5.0", "1210", "99"): "99 .. Text",
        ("FIX.5.0", "1215", "CD"): "CD .. EUCPWithLumpSumInterest",
        ("FIX.5.0", "1215", "WI"): "WI .. WhenIssued",
        ("FIX.5.0", "1217", "1"): "1 .. CUSIP",
        ("FIX.5.0", "1217", "2"): "2 .. SEDOL",
        ("FIX.5.0", "1217", "3"): "3 .. QUIK",
        ("FIX.5.0", "1217", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "1217", "5"): "5 .. RICCode",
        ("FIX.5.0", "1217", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "1217", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "1217", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "1217", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "1217", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "1217", "B"): "B .. Wertpapier",
        ("FIX.5.0", "1217", "C"): "C .. Dutch",
        ("FIX.5.0", "1217", "D"): "D .. Valoren",
        ("FIX.5.0", "1217", "E"): "E .. Sicovam",
        ("FIX.5.0", "1217", "F"): "F .. Belgian",
        ("FIX.5.0", "1217", "G"): "G .. Common",
        ("FIX.5.0", "1217", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "1217", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "1217", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "1217", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "1217", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "1217", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "1217", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "1217", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "1217", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "1217", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "1217", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "1217", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "1217", "U"): "U .. Synthetic",
        ("FIX.5.0", "1217", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "1217", "W"): "W .. IndexName",
        ("FIX.5.0", "1220", "1"): "1 .. CUSIP",
        ("FIX.5.0", "1220", "2"): "2 .. SEDOL",
        ("FIX.5.0", "1220", "3"): "3 .. QUIK",
        ("FIX.5.0", "1220", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "1220", "5"): "5 .. RICCode",
        ("FIX.5.0", "1220", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "1220", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "1220", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "1220", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "1220", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "1220", "B"): "B .. Wertpapier",
        ("FIX.5.0", "1220", "C"): "C .. Dutch",
        ("FIX.5.0", "1220", "D"): "D .. Valoren",
        ("FIX.5.0", "1220", "E"): "E .. Sicovam",
        ("FIX.5.0", "1220", "F"): "F .. Belgian",
        ("FIX.5.0", "1220", "G"): "G .. Common",
        ("FIX.5.0", "1220", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "1220", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "1220", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "1220", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "1220", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "1220", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "1220", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "1220", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "1220", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "1220", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "1220", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "1220", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "1220", "U"): "U .. Synthetic",
        ("FIX.5.0", "1220", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "1220", "W"): "W .. IndexName",
        ("FIX.5.0", "1238", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "1238", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "1238", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "1238", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "1238", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "1238", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "1238", "thm"): "thm .. therms",
        ("FIX.5.0", "1238", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "1238", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "1238", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "1238", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "1238", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "1238", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "1238", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "1238", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "1238", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "1238", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "1238", "day"): "day .. Day",
        ("FIX.5.0", "1238", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "1238", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "1238", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "1238", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "1238", "g"): "g .. Grams",
        ("FIX.5.0", "1238", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "1238", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "1238", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "1238", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "1238", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "1238", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "1238", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "1238", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "1238", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "1238", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "1238", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "1238", "L"): "L .. liters",
        ("FIX.5.0", "1238", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "1238", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "1238", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "1238", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "1238", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "1238", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "1238", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "1238", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "1238", "t"): "t .. MetricTons",
        ("FIX.5.0", "1238", "tn"): "tn .. Tons",
        ("FIX.5.0", "1246", "1"): "1 .. AGENCY",
        ("FIX.5.0", "1246", "2"): "2 .. COMMODITY",
        ("FIX.5.0", "1246", "3"): "3 .. CORPORATE",
        ("FIX.5.0", "1246", "4"): "4 .. CURRENCY",
        ("FIX.5.0", "1246", "5"): "5 .. EQUITY",
        ("FIX.5.0", "1246", "6"): "6 .. GOVERNMENT",
        ("FIX.5.0", "1246", "7"): "7 .. INDEX",
        ("FIX.5.0", "1246", "8"): "8 .. LOAN",
        ("FIX.5.0", "1246", "9"): "9 .. MONEYMARKET",
        ("FIX.5.0", "1246", "10"): "10 .. MORTGAGE",
        ("FIX.5.0", "1246", "11"): "11 .. MUNICIPAL",
        ("FIX.5.0", "1246", "12"): "12 .. OTHER",
        ("FIX.5.0", "1246", "13"): "13 .. FINANCING",
        ("FIX.5.0", "1249", "EUSUPRA"): "EUSUPRA .. EuroSupranationalCoupons",
        ("FIX.5.0", "1249", "FAC"): "FAC .. FederalAgencyCoupon",
        ("FIX.5.0", "1249", "FADN"): "FADN .. FederalAgencyDiscountNote",
        ("FIX.5.0", "1249", "PEF"): "PEF .. PrivateExportFunding",
        ("FIX.5.0", "1249", "SUPRA"): "SUPRA .. USDSupranationalCoupons",
        ("FIX.5.0", "1249", "CORP"): "CORP .. CorporateBond",
        ("FIX.5.0", "1249", "CPP"): "CPP .. CorporatePrivatePlacement",
        ("FIX.5.0", "1249", "CB"): "CB .. ConvertibleBond",
        ("FIX.5.0", "1249", "DUAL"): "DUAL .. DualCurrency",
        ("FIX.5.0", "1249", "EUCORP"): "EUCORP .. EuroCorporateBond",
        ("FIX.5.0", "1249", "EUFRN"): "EUFRN .. EuroCorporateFloatingRateNotes",
        ("FIX.5.0", "1249", "FRN"): "FRN .. USCorporateFloatingRateNotes",
        ("FIX.5.0", "1249", "XLINKD"): "XLINKD .. IndexedLinked",
        ("FIX.5.0", "1249", "STRUCT"): "STRUCT .. StructuredNotes",
        ("FIX.5.0", "1249", "YANK"): "YANK .. YankeeCorporateBond",
        ("FIX.5.0", "1249", "FXNDF"): "FXNDF .. NonDeliverableForward",
        ("FIX.5.0", "1249", "FXSPOT"): "FXSPOT .. FXSpot",
        ("FIX.5.0", "1249", "FXFWD"): "FXFWD .. FXForward",
        ("FIX.5.0", "1249", "FXSWAP"): "FXSWAP .. FXSwap",
        ("FIX.5.0", "1249", "FXNDS"): "FXNDS .. NonDeliverableSwap",
        ("FIX.5.0", "1249", "CAP"): "CAP .. Cap",
        ("FIX.5.0", "1249", "CDS"): "CDS .. CreditDefaultSwap",
        ("FIX.5.0", "1249", "CLLR"): "CLLR .. Collar",
        ("FIX.5.0", "1249", "CMDTYSWAP"): "CMDTYSWAP .. CommoditySwap",
        ("FIX.5.0", "1249", "EXOTIC"): "EXOTIC .. Exotic",
        ("FIX.5.0", "1249", "OOC"): "OOC .. OptionsOnCombo",
        ("FIX.5.0", "1249", "FLR"): "FLR .. Floor",
        ("FIX.5.0", "1249", "FRA"): "FRA .. FRA",
        ("FIX.5.0", "1249", "FUT"): "FUT .. Future",
        ("FIX.5.0", "1249", "FWD"): "FWD .. DerivativeForward",
        ("FIX.5.0", "1249", "IRS"): "IRS .. InterestRateSwap",
        ("FIX.5.0", "1249", "TRS"): "TRS .. TotalReturnSwap",
        ("FIX.5.0", "1249", "LOANLEASE"): "LOANLEASE .. LoanLease",
        ("FIX.5.0", "1249", "OOF"): "OOF .. OptionsOnFutures",
        ("FIX.5.0", "1249", "OOP"): "OOP .. OptionsOnPhysical",
        ("FIX.5.0", "1249", "OPT"): "OPT .. Option",
        ("FIX.5.0", "1249", "SPOTFWD"): "SPOTFWD .. SpotForward",
        ("FIX.5.0", "1249", "SWAPTION"): "SWAPTION .. SwapOption",
        ("FIX.5.0", "1249", "XMISSION"): "XMISSION .. Transmission",
        ("FIX.5.0", "1249", "INDEX"): "INDEX .. Index",
        ("FIX.5.0", "1249", "BDBSKT"): "BDBSKT .. BondBasket",
        ("FIX.5.0", "1249", "CFD"): "CFD .. ContractForDifference",
        ("FIX.5.0", "1249", "CRLTNSWAP"): "CRLTNSWAP .. CorrelationSwap",
        ("FIX.5.0", "1249", "DVDNDSWAP"): "DVDNDSWAP .. DiviendSwap",
        ("FIX.5.0", "1249", "EQBSKT"): "EQBSKT .. EquityBasket",
        ("FIX.5.0", "1249", "EQFWD"): "EQFWD .. EquityForward",
        ("FIX.5.0", "1249", "RTRNSWAP"): "RTRNSWAP .. ReturnSwap",
        ("FIX.5.0", "1249", "VARSWAP"): "VARSWAP .. VarianceSwap",
        ("FIX.5.0", "1249", "PRTFLIOSWAP"): "PRTFLIOSWAP .. PortfolioSwaps",
        ("FIX.5.0", "1249", "FUTSWAP"): "FUTSWAP .. FuturesOnASwap",
        ("FIX.5.0", "1249", "FWDSWAP "): "FWDSWAP  .. ForwardsOnASwap",
        ("FIX.5.0", "1249", "FWDFRTAGMT "): "FWDFRTAGMT  .. ForwardFreightAgreement",
        ("FIX.5.0", "1249", "SPREADBET"): "SPREADBET .. SpreadBetting",
        ("FIX.5.0", "1249", "ETC"): "ETC .. ExchangeTradedCommodity",
        ("FIX.5.0", "1249", "CS"): "CS .. CommonStock",
        ("FIX.5.0", "1249", "PS"): "PS .. PreferredStock",
        ("FIX.5.0", "1249", "DR"): "DR .. DepositoryReceipts",
        ("FIX.5.0", "1249", "REPO"): "REPO .. Repurchase",
        ("FIX.5.0", "1249", "FORWARD"): "FORWARD .. Forward",
        ("FIX.5.0", "1249", "BUYSELL"): "BUYSELL .. BuySellback",
        ("FIX.5.0", "1249", "SECLOAN"): "SECLOAN .. SecuritiesLoan",
        ("FIX.5.0", "1249", "SECPLEDGE"): "SECPLEDGE .. SecuritiesPledge",
        ("FIX.5.0", "1249", "DVPLDG"): "DVPLDG .. DeliveryVersusPledge",
        ("FIX.5.0", "1249", "COLLBSKT"): "COLLBSKT .. CollateralBasket",
        ("FIX.5.0", "1249", "SFP"): "SFP .. StructuredFinanceProduct",
        ("FIX.5.0", "1249", "BRADY"): "BRADY .. BradyBond",
        ("FIX.5.0", "1249", "CAN"): "CAN .. CanadianTreasuryNotes",
        ("FIX.5.0", "1249", "CTB"): "CTB .. CanadianTreasuryBills",
        ("FIX.5.0", "1249", "EUSOV"): "EUSOV .. EuroSovereigns",
        ("FIX.5.0", "1249", "PROV"): "PROV .. CanadianProvincialBonds",
        ("FIX.5.0", "1249", "TB"): "TB .. TreasuryBill",
        ("FIX.5.0", "1249", "TBOND"): "TBOND .. USTreasuryBond",
        ("FIX.5.0", "1249", "TINT"): "TINT .. InterestStripFromAnyBondOrNote",
        ("FIX.5.0", "1249", "TBILL"): "TBILL .. USTreasuryBill",
        ("FIX.5.0", "1249", "TIPS"): "TIPS .. TreasuryInflationProtectedSecurities",
        ("FIX.5.0", "1249", "TCAL"): "TCAL .. PrincipalStripOfACallableBondOrNote",
        ("FIX.5.0", "1249", "TPRN"): "TPRN .. PrincipalStripFromANonCallableBondOrNote",
        ("FIX.5.0", "1249", "TNOTE"): "TNOTE .. USTreasuryNote",
        ("FIX.5.0", "1249", "TERM"): "TERM .. TermLoan",
        ("FIX.5.0", "1249", "RVLV"): "RVLV .. RevolverLoan",
        ("FIX.5.0", "1249", "RVLVTRM"): "RVLVTRM .. Revolver",
        ("FIX.5.0", "1249", "BRIDGE"): "BRIDGE .. BridgeLoan",
        ("FIX.5.0", "1249", "LOFC"): "LOFC .. LetterOfCredit",
        ("FIX.5.0", "1249", "SWING"): "SWING .. SwingLineFacility",
        ("FIX.5.0", "1249", "DINP"): "DINP .. DebtorInPossession",
        ("FIX.5.0", "1249", "DEFLTED"): "DEFLTED .. Defaulted",
        ("FIX.5.0", "1249", "WITHDRN"): "WITHDRN .. Withdrawn",
        ("FIX.5.0", "1249", "REPLACD"): "REPLACD .. Replaced",
        ("FIX.5.0", "1249", "MATURED"): "MATURED .. Matured",
        ("FIX.5.0", "1249", "AMENDED"): "AMENDED .. Amended",
        ("FIX.5.0", "1249", "RETIRED"): "RETIRED .. Retired",
        ("FIX.5.0", "1249", "BA"): "BA .. BankersAcceptance",
        ("FIX.5.0", "1249", "BDN"): "BDN .. BankDepositoryNote",
        ("FIX.5.0", "1249", "BN"): "BN .. BankNotes",
        ("FIX.5.0", "1249", "BOX"): "BOX .. BillOfExchanges",
        ("FIX.5.0", "1249", "CAMM"): "CAMM .. CanadianMoneyMarkets",
        ("FIX.5.0", "1249", "CD"): "CD .. CertificateOfDeposit",
        ("FIX.5.0", "1249", "CL"): "CL .. CallLoans",
        ("FIX.5.0", "1249", "CP"): "CP .. CommercialPaper",
        ("FIX.5.0", "1249", "DN"): "DN .. DepositNotes",
        ("FIX.5.0", "1249", "EUCD"): "EUCD .. EuroCertificateOfDeposit",
        ("FIX.5.0", "1249", "EUCP"): "EUCP .. EuroCommercialPaper",
        ("FIX.5.0", "1249", "LQN"): "LQN .. LiquidityNote",
        ("FIX.5.0", "1249", "MTN"): "MTN .. MediumTermNotes",
        ("FIX.5.0", "1249", "ONITE"): "ONITE .. Overnight",
        ("FIX.5.0", "1249", "PN"): "PN .. PromissoryNote",
        ("FIX.5.0", "1249", "STN"): "STN .. ShortTermLoanNote",
        ("FIX.5.0", "1249", "PZFJ"): "PZFJ .. PlazosFijos",
        ("FIX.5.0", "1249", "SLQN"): "SLQN .. SecuredLiquidityNote",
        ("FIX.5.0", "1249", "TD"): "TD .. TimeDeposit",
        ("FIX.5.0", "1249", "TLQN"): "TLQN .. TermLiquidityNote",
        ("FIX.5.0", "1249", "XCN"): "XCN .. ExtendedCommNote",
        ("FIX.5.0", "1249", "YCD"): "YCD .. YankeeCertificateOfDeposit",
        ("FIX.5.0", "1249", "ABS"): "ABS .. AssetBackedSecurities",
        ("FIX.5.0", "1249", "CMB"): "CMB .. CanadianMortgageBonds",
        ("FIX.5.0", "1249", "CMBS"): "CMBS .. Corp",
        ("FIX.5.0", "1249", "CMO"): "CMO .. CollateralizedMortgageObligation",
        ("FIX.5.0", "1249", "IET"): "IET .. IOETTEMortgage",
        ("FIX.5.0", "1249", "MBS"): "MBS .. MortgageBackedSecurities",
        ("FIX.5.0", "1249", "MIO"): "MIO .. MortgageInterestOnly",
        ("FIX.5.0", "1249", "MPO"): "MPO .. MortgagePrincipalOnly",
        ("FIX.5.0", "1249", "MPP"): "MPP .. MortgagePrivatePlacement",
        ("FIX.5.0", "1249", "MPT"): "MPT .. MiscellaneousPassThrough",
        ("FIX.5.0", "1249", "PFAND"): "PFAND .. Pfandbriefe",
        ("FIX.5.0", "1249", "TBA"): "TBA .. ToBeAnnounced",
        ("FIX.5.0", "1249", "AN"): "AN .. OtherAnticipationNotes",
        ("FIX.5.0", "1249", "COFO"): "COFO .. CertificateOfObligation",
        ("FIX.5.0", "1249", "COFP"): "COFP .. CertificateOfParticipation",
        ("FIX.5.0", "1249", "GO"): "GO .. GeneralObligationBonds",
        ("FIX.5.0", "1249", "MT"): "MT .. MandatoryTender",
        ("FIX.5.0", "1249", "RAN"): "RAN .. RevenueAnticipationNote",
        ("FIX.5.0", "1249", "REV"): "REV .. RevenueBonds",
        ("FIX.5.0", "1249", "SPCLA"): "SPCLA .. SpecialAssessment",
        ("FIX.5.0", "1249", "SPCLO"): "SPCLO .. SpecialObligation",
        ("FIX.5.0", "1249", "SPCLT"): "SPCLT .. SpecialTax",
        ("FIX.5.0", "1249", "TAN"): "TAN .. TaxAnticipationNote",
        ("FIX.5.0", "1249", "TAXA"): "TAXA .. TaxAllocation",
        ("FIX.5.0", "1249", "TECP"): "TECP .. TaxExemptCommercialPaper",
        ("FIX.5.0", "1249", "TMCP"): "TMCP .. TaxableMunicipalCP",
        ("FIX.5.0", "1249", "TRAN"): "TRAN .. TaxRevenueAnticipationNote",
        ("FIX.5.0", "1249", "VRDN"): "VRDN .. VariableRateDemandNote",
        ("FIX.5.0", "1249", "WAR"): "WAR .. Warrant",
        ("FIX.5.0", "1249", "MF"): "MF .. MutualFund",
        ("FIX.5.0", "1249", "MLEG"): "MLEG .. MultilegInstrument",
        ("FIX.5.0", "1249", "NONE"): "NONE .. NoSecurityType",
        ("FIX.5.0", "1249", "?"): "? .. Wildcard",
        ("FIX.5.0", "1249", "CASH"): "CASH .. Cash",
        ("FIX.5.0", "1249", "Other"): "Other .. Other",
        ("FIX.5.0", "1249", "ETN"): "ETN .. ExchangeTradedNote",
        ("FIX.5.0", "1249", "SECDERIV"): "SECDERIV .. SecuritizedDerivative",
        ("FIX.5.0", "1255", "P"): "P .. ProRata",
        ("FIX.5.0", "1255", "R"): "R .. Random",
        ("FIX.5.0", "1256", "1"): "1 .. Active",
        ("FIX.5.0", "1256", "2"): "2 .. Inactive",
        ("FIX.5.0", "1256", "3"): "3 .. ActiveClosingOrdersOnly",
        ("FIX.5.0", "1256", "4"): "4 .. Expired",
        ("FIX.5.0", "1256", "5"): "5 .. Delisted",
        ("FIX.5.0", "1256", "6"): "6 .. KnockedOut",
        ("FIX.5.0", "1256", "7"): "7 .. KnockOutRevoked",
        ("FIX.5.0", "1256", "8"): "8 .. PendingExpiry",
        ("FIX.5.0", "1256", "9"): "9 .. Suspended",
        ("FIX.5.0", "1256", "10"): "10 .. Published",
        ("FIX.5.0", "1256", "11"): "11 .. PendingDeletion",
        ("FIX.5.0", "1269", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "1269", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "1269", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "1269", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "1269", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "1269", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "1269", "thm"): "thm .. therms",
        ("FIX.5.0", "1269", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "1269", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "1269", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "1269", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "1269", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "1269", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "1269", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "1269", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "1269", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "1269", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "1269", "day"): "day .. Day",
        ("FIX.5.0", "1269", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "1269", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "1269", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "1269", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "1269", "g"): "g .. Grams",
        ("FIX.5.0", "1269", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "1269", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "1269", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "1269", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "1269", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "1269", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "1269", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "1269", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "1269", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "1269", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "1269", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "1269", "L"): "L .. liters",
        ("FIX.5.0", "1269", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "1269", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "1269", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "1269", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "1269", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "1269", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "1269", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "1269", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "1269", "t"): "t .. MetricTons",
        ("FIX.5.0", "1269", "tn"): "tn .. Tons",
        ("FIX.5.0", "1271", "H"): "H .. Hour",
        ("FIX.5.0", "1271", "Min"): "Min .. Minute",
        ("FIX.5.0", "1271", "S"): "S .. Second",
        ("FIX.5.0", "1271", "D"): "D .. Day",
        ("FIX.5.0", "1271", "Wk"): "Wk .. Week",
        ("FIX.5.0", "1271", "Mo"): "Mo .. Month",
        ("FIX.5.0", "1271", "Yr"): "Yr .. Year",
        ("FIX.5.0", "1271", "Q"): "Q .. Quarter",
        ("FIX.5.0", "1287", "1"): "1 .. Put",
        ("FIX.5.0", "1287", "2"): "2 .. Call",
        ("FIX.5.0", "1287", "3"): "3 .. Tender",
        ("FIX.5.0", "1287", "4"): "4 .. SinkingFundCall",
        ("FIX.5.0", "1287", "5"): "5 .. Activation",
        ("FIX.5.0", "1287", "6"): "6 .. Inactiviation",
        ("FIX.5.0", "1287", "7"): "7 .. LastEligibleTradeDate",
        ("FIX.5.0", "1287", "8"): "8 .. SwapStartDate",
        ("FIX.5.0", "1287", "9"): "9 .. SwapEndDate",
        ("FIX.5.0", "1287", "10"): "10 .. SwapRollDate",
        ("FIX.5.0", "1287", "11"): "11 .. SwapNextStartDate",
        ("FIX.5.0", "1287", "12"): "12 .. SwapNextRollDate",
        ("FIX.5.0", "1287", "13"): "13 .. FirstDeliveryDate",
        ("FIX.5.0", "1287", "14"): "14 .. LastDeliveryDate",
        ("FIX.5.0", "1287", "15"): "15 .. InitialInventoryDueDate",
        ("FIX.5.0", "1287", "16"): "16 .. FinalInventoryDueDate",
        ("FIX.5.0", "1287", "17"): "17 .. FirstIntentDate",
        ("FIX.5.0", "1287", "18"): "18 .. LastIntentDate",
        ("FIX.5.0", "1287", "19"): "19 .. PositionRemovalDate",
        ("FIX.5.0", "1287", "20"): "20 .. MinimumNotice",
        ("FIX.5.0", "1287", "21"): "21 .. DeliveryStartTime",
        ("FIX.5.0", "1287", "22"): "22 .. DeliveryEndTime",
        ("FIX.5.0", "1287", "23"): "23 .. FirstNoticeDate",
        ("FIX.5.0", "1287", "24"): "24 .. LastNoticeDate",
        ("FIX.5.0", "1287", "25"): "25 .. FirstExerciseDate",
        ("FIX.5.0", "1287", "26"): "26 .. RedemptionDate",
        ("FIX.5.0", "1287", "27"): "27 .. TrdCntntnEfctvDt",
        ("FIX.5.0", "1287", "99"): "99 .. Other",
        ("FIX.5.0", "1294", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1294", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1294", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1294", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1294", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1294", "J"): "J .. TaxID",
        ("FIX.5.0", "1294", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1294", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1294", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1294", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1294", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1294", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1294", "B"): "B .. BIC",
        ("FIX.5.0", "1294", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1294", "D"): "D .. Proprietary",
        ("FIX.5.0", "1294", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1294", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1294", "G"): "G .. MIC",
        ("FIX.5.0", "1294", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1294", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1294", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1294", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1294", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1294", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1294", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1294", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1295", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1295", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1295", "3"): "3 .. ClientID",
        ("FIX.5.0", "1295", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1295", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1295", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1295", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1295", "8"): "8 .. Locate",
        ("FIX.5.0", "1295", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1295", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1295", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1295", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1295", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1295", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1295", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1295", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1295", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1295", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1295", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1295", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1295", "22"): "22 .. Exchange",
        ("FIX.5.0", "1295", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1295", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1295", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1295", "27"): "27 .. Buyer",
        ("FIX.5.0", "1295", "28"): "28 .. Custodian",
        ("FIX.5.0", "1295", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1295", "30"): "30 .. Agent",
        ("FIX.5.0", "1295", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1295", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1295", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1295", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1295", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1295", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1295", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1295", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1295", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1295", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1295", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1295", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1295", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1295", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1295", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1295", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1295", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1295", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1295", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1295", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1295", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1295", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1295", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1295", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1295", "55"): "55 .. SessionID",
        ("FIX.5.0", "1295", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1295", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1295", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1295", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1295", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1295", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1295", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1295", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1295", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1295", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1295", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1295", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1295", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1295", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1295", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1295", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1295", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1295", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1295", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1295", "75"): "75 .. LocationID",
        ("FIX.5.0", "1295", "76"): "76 .. DeskID",
        ("FIX.5.0", "1295", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1295", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1295", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1295", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1295", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1295", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1295", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1295", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1295", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1295", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1295", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1295", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1295", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1295", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1295", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1295", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1295", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1295", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1295", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1295", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1295", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1295", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1295", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1295", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1295", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1295", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1295", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1295", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1295", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1295", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1295", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1295", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1295", "110"): "110 .. Borrower",
        ("FIX.5.0", "1295", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1295", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1295", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1295", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1295", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1295", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1295", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1295", "118"): "118 .. Operator",
        ("FIX.5.0", "1295", "119"): "119 .. CSD",
        ("FIX.5.0", "1295", "120"): "120 .. ICSD",
        ("FIX.5.0", "1295", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1295", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1295", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1298", "1"): "1 .. Firm",
        ("FIX.5.0", "1298", "2"): "2 .. Person",
        ("FIX.5.0", "1298", "3"): "3 .. System",
        ("FIX.5.0", "1298", "4"): "4 .. Application",
        ("FIX.5.0", "1298", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1298", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1298", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1298", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1298", "9"): "9 .. ContactName",
        ("FIX.5.0", "1298", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1298", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1298", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1298", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1298", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1298", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1298", "16"): "16 .. BIC",
        ("FIX.5.0", "1298", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1298", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1298", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1298", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1298", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1298", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1298", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1298", "24"): "24 .. Department",
        ("FIX.5.0", "1298", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1298", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1298", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1298", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1298", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1298", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1298", "31"): "31 .. Location",
        ("FIX.5.0", "1298", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1298", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1298", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1298", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1298", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1298", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1298", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1298", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1298", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1298", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1298", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1298", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1298", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1298", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1298", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1298", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1298", "48"): "48 .. USPerson",
        ("FIX.5.0", "1298", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1298", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1298", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1298", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1298", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1298", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1298", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1298", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1298", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1298", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1298", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1298", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1298", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1298", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1298", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1298", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1298", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1298", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1298", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1298", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1298", "69"): "69 .. EndUser",
        ("FIX.5.0", "1298", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1298", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1298", "72"): "72 .. Domicile",
        ("FIX.5.0", "1298", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1298", "74"): "74 .. Payer",
        ("FIX.5.0", "1298", "75"): "75 .. Receiver",
        ("FIX.5.0", "1298", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1298", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1298", "78"): "78 .. FirstName",
        ("FIX.5.0", "1298", "79"): "79 .. Surname",
        ("FIX.5.0", "1298", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1298", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1298", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1298", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1298", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1299", "0"): "0 .. European",
        ("FIX.5.0", "1299", "1"): "1 .. American",
        ("FIX.5.0", "1299", "2"): "2 .. Bermuda",
        ("FIX.5.0", "1299", "99"): "99 .. Other",
        ("FIX.5.0", "1302", "0"): "0 .. Months",
        ("FIX.5.0", "1302", "1"): "1 .. Days",
        ("FIX.5.0", "1302", "2"): "2 .. Weeks",
        ("FIX.5.0", "1302", "3"): "3 .. Years",
        ("FIX.5.0", "1303", "0"): "0 .. YearMonthOnly",
        ("FIX.5.0", "1303", "1"): "1 .. YearMonthDay",
        ("FIX.5.0", "1303", "2"): "2 .. YearMonthWeek",
        ("FIX.5.0", "1304", "0"): "0 .. European",
        ("FIX.5.0", "1304", "1"): "1 .. American",
        ("FIX.5.0", "1304", "2"): "2 .. Bermuda",
        ("FIX.5.0", "1304", "99"): "99 .. Other",
        ("FIX.5.0", "1305", "0"): "0 .. Price",
        ("FIX.5.0", "1305", "1"): "1 .. Ticks",
        ("FIX.5.0", "1305", "2"): "2 .. Percentage",
        ("FIX.5.0", "1306", "0"): "0 .. Price",
        ("FIX.5.0", "1306", "1"): "1 .. Ticks",
        ("FIX.5.0", "1306", "2"): "2 .. Percentage",
        ("FIX.5.0", "1308", "0"): "0 .. StayOnOfferSide",
        ("FIX.5.0", "1308", "1"): "1 .. NotHeld",
        ("FIX.5.0", "1308", "2"): "2 .. Work",
        ("FIX.5.0", "1308", "3"): "3 .. GoAlong",
        ("FIX.5.0", "1308", "4"): "4 .. OverTheDay",
        ("FIX.5.0", "1308", "5"): "5 .. Held",
        ("FIX.5.0", "1308", "6"): "6 .. ParticipateDoNotInitiate",
        ("FIX.5.0", "1308", "7"): "7 .. StrictScale",
        ("FIX.5.0", "1308", "8"): "8 .. TryToScale",
        ("FIX.5.0", "1308", "9"): "9 .. StayOnBidSide",
        ("FIX.5.0", "1308", "A"): "A .. NoCross",
        ("FIX.5.0", "1308", "B"): "B .. OKToCross",
        ("FIX.5.0", "1308", "C"): "C .. CallFirst",
        ("FIX.5.0", "1308", "D"): "D .. PercentOfVolume",
        ("FIX.5.0", "1308", "E"): "E .. DoNotIncrease",
        ("FIX.5.0", "1308", "F"): "F .. DoNotReduce",
        ("FIX.5.0", "1308", "G"): "G .. AllOrNone",
        ("FIX.5.0", "1308", "H"): "H .. ReinstateOnSystemFailure",
        ("FIX.5.0", "1308", "I"): "I .. InstitutionsOnly",
        ("FIX.5.0", "1308", "J"): "J .. ReinstateOnTradingHalt",
        ("FIX.5.0", "1308", "K"): "K .. CancelOnTradingHalt",
        ("FIX.5.0", "1308", "N"): "N .. NonNegotiable",
        ("FIX.5.0", "1308", "Q"): "Q .. CancelOnSystemFailure",
        ("FIX.5.0", "1308", "S"): "S .. Suspend",
        ("FIX.5.0", "1308", "U"): "U .. CustomerDisplayInstruction",
        ("FIX.5.0", "1308", "V"): "V .. Netting",
        ("FIX.5.0", "1308", "X"): "X .. TradeAlong",
        ("FIX.5.0", "1308", "Y"): "Y .. TryToStop",
        ("FIX.5.0", "1308", "Z"): "Z .. CancelIfNotBest",
        ("FIX.5.0", "1308", "b"): "b .. StrictLimit",
        ("FIX.5.0", "1308", "c"): "c .. IgnorePriceValidityChecks",
        ("FIX.5.0", "1308", "e"): "e .. WorkToTargetStrategy",
        ("FIX.5.0", "1308", "f"): "f .. IntermarketSweep",
        ("FIX.5.0", "1308", "g"): "g .. ExternalRoutingAllowed",
        ("FIX.5.0", "1308", "h"): "h .. ExternalRoutingNotAllowed",
        ("FIX.5.0", "1308", "i"): "i .. ImbalanceOnly",
        ("FIX.5.0", "1308", "j"): "j .. SingleExecutionRequestedForBlockTrade",
        ("FIX.5.0", "1308", "k"): "k .. BestExecution",
        ("FIX.5.0", "1308", "l"): "l .. SuspendOnSystemFailure",
        ("FIX.5.0", "1308", "m"): "m .. SuspendOnTradingHalt",
        ("FIX.5.0", "1308", "n"): "n .. ReinstateOnConnectionLoss",
        ("FIX.5.0", "1308", "o"): "o .. CancelOnConnectionLoss",
        ("FIX.5.0", "1308", "p"): "p .. SuspendOnConnectionLoss",
        ("FIX.5.0", "1308", "q"): "q .. Release",
        ("FIX.5.0", "1308", "r"): "r .. ExecuteAsDeltaNeutral",
        ("FIX.5.0", "1308", "s"): "s .. ExecuteAsDurationNeutral",
        ("FIX.5.0", "1308", "t"): "t .. ExecuteAsFXNeutral",
        ("FIX.5.0", "1308", "u"): "u .. MinGuaranteedFillEligible",
        ("FIX.5.0", "1308", "v"): "v .. BypassNonDisplayLiquidity",
        ("FIX.5.0", "1308", "w"): "w .. Lock",
        ("FIX.5.0", "1308", "x"): "x .. IgnoreNotionalValueChecks",
        ("FIX.5.0", "1308", "y"): "y .. TrdAtRefPx",
        ("FIX.5.0", "1313", "1"): "1 .. Flat",
        ("FIX.5.0", "1313", "2"): "2 .. ZeroCoupon",
        ("FIX.5.0", "1313", "3"): "3 .. InterestBearing",
        ("FIX.5.0", "1313", "4"): "4 .. NoPeriodicPayments",
        ("FIX.5.0", "1313", "5"): "5 .. VariableRate",
        ("FIX.5.0", "1313", "6"): "6 .. LessFeeForPut",
        ("FIX.5.0", "1313", "7"): "7 .. SteppedCoupon",
        ("FIX.5.0", "1313", "8"): "8 .. CouponPeriod",
        ("FIX.5.0", "1313", "9"): "9 .. When",
        ("FIX.5.0", "1313", "10"): "10 .. OriginalIssueDiscount",
        ("FIX.5.0", "1313", "11"): "11 .. Callable",
        ("FIX.5.0", "1313", "12"): "12 .. EscrowedToMaturity",
        ("FIX.5.0", "1313", "13"): "13 .. EscrowedToRedemptionDate",
        ("FIX.5.0", "1313", "14"): "14 .. PreRefunded",
        ("FIX.5.0", "1313", "15"): "15 .. InDefault",
        ("FIX.5.0", "1313", "16"): "16 .. Unrated",
        ("FIX.5.0", "1313", "17"): "17 .. Taxable",
        ("FIX.5.0", "1313", "18"): "18 .. Indexed",
        ("FIX.5.0", "1313", "19"): "19 .. SubjectToAlternativeMinimumTax",
        ("FIX.5.0", "1313", "20"): "20 .. OriginalIssueDiscountPrice",
        ("FIX.5.0", "1313", "21"): "21 .. CallableBelowMaturityValue",
        ("FIX.5.0", "1313", "22"): "22 .. CallableWithoutNotice",
        ("FIX.5.0", "1313", "23"): "23 .. PriceTickRulesForSecurity",
        ("FIX.5.0", "1313", "24"): "24 .. TradeTypeEligibilityDetailsForSecurity",
        ("FIX.5.0", "1313", "25"): "25 .. InstrumentDenominator",
        ("FIX.5.0", "1313", "26"): "26 .. InstrumentNumerator",
        ("FIX.5.0", "1313", "27"): "27 .. InstrumentPricePrecision",
        ("FIX.5.0", "1313", "28"): "28 .. InstrumentStrikePrice",
        ("FIX.5.0", "1313", "29"): "29 .. TradeableIndicator",
        ("FIX.5.0", "1313", "30"): "30 .. InstrumentEligibleAnonOrders",
        ("FIX.5.0", "1313", "31"): "31 .. MinGuaranteedFillVolume",
        ("FIX.5.0", "1313", "32"): "32 .. MinGuaranteedFillStatus",
        ("FIX.5.0", "1313", "33"): "33 .. TradeAtSettlementEligibility",
        ("FIX.5.0", "1313", "34"): "34 .. TestInstrument",
        ("FIX.5.0", "1313", "35"): "35 .. DummyInstrument",
        ("FIX.5.0", "1313", "36"): "36 .. NegativeSettlementPriceEligibility",
        ("FIX.5.0", "1313", "37"): "37 .. NegativeStrikePriceEligibility",
        ("FIX.5.0", "1313", "38"): "38 .. USStdContractInd",
        ("FIX.5.0", "1313", "39"): "39 .. AdmittedToTradingOnTradingVenue",
        ("FIX.5.0", "1313", "40"): "40 .. AverageDailyNotionalAmount",
        ("FIX.5.0", "1313", "41"): "41 .. AverageDailyNumberTrades",
        ("FIX.5.0", "1313", "99"): "99 .. Text",
        ("FIX.5.0", "1315", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "1315", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "1315", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "1315", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "1315", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "1315", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "1315", "thm"): "thm .. therms",
        ("FIX.5.0", "1315", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "1315", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "1315", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "1315", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "1315", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "1315", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "1315", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "1315", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "1315", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "1315", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "1315", "day"): "day .. Day",
        ("FIX.5.0", "1315", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "1315", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "1315", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "1315", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "1315", "g"): "g .. Grams",
        ("FIX.5.0", "1315", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "1315", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "1315", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "1315", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "1315", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "1315", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "1315", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "1315", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "1315", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "1315", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "1315", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "1315", "L"): "L .. liters",
        ("FIX.5.0", "1315", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "1315", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "1315", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "1315", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "1315", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "1315", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "1315", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "1315", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "1315", "t"): "t .. MetricTons",
        ("FIX.5.0", "1315", "tn"): "tn .. Tons",
        ("FIX.5.0", "1317", "C"): "C .. CashSettlementRequired",
        ("FIX.5.0", "1317", "P"): "P .. PhysicalSettlementRequired",
        ("FIX.5.0", "1317", "E"): "E .. Election",
        ("FIX.5.0", "1318", "STD"): "STD .. Standard",
        ("FIX.5.0", "1318", "INX"): "INX .. Index",
        ("FIX.5.0", "1318", "INT"): "INT .. InterestRateIndex",
        ("FIX.5.0", "1318", "PCTPAR"): "PCTPAR .. PercentOfPar",
        ("FIX.5.0", "1319", "EQTY"): "EQTY .. PremiumStyle",
        ("FIX.5.0", "1319", "FUT"): "FUT .. FuturesStyleMarkToMarket",
        ("FIX.5.0", "1319", "FUTDA"): "FUTDA .. FuturesStyleWithAnAttachedCashAdjustment",
        ("FIX.5.0", "1319", "CDS"): "CDS .. CDSStyleCollateralization",
        ("FIX.5.0", "1319", "CDSD"): "CDSD .. CDSInDeliveryUseRecoveryRateToCalculate",
        ("FIX.5.0", "1320", "0"): "0 .. PreListedOnly",
        ("FIX.5.0", "1320", "1"): "1 .. UserRequested",
        ("FIX.5.0", "1323", "0"): "0 .. Put",
        ("FIX.5.0", "1323", "1"): "1 .. Call",
        ("FIX.5.0", "1323", "2"): "2 .. Other",
        ("FIX.5.0", "1323", "3"): "3 .. Chooser",
        ("FIX.5.0", "1324", "A"): "A .. Add",
        ("FIX.5.0", "1324", "D"): "D .. Delete",
        ("FIX.5.0", "1324", "M"): "M .. Modify",
        ("FIX.5.0", "1324", "S"): "S .. Snapshot",
        ("FIX.5.0", "1327", "A"): "A .. Add",
        ("FIX.5.0", "1327", "D"): "D .. Delete",
        ("FIX.5.0", "1327", "M"): "M .. Modify",
        ("FIX.5.0", "1347", "0"): "0 .. Retransmission",
        ("FIX.5.0", "1347", "1"): "1 .. Subscription",
        ("FIX.5.0", "1347", "2"): "2 .. RequestLastSeqNum",
        ("FIX.5.0", "1347", "3"): "3 .. RequestApplications",
        ("FIX.5.0", "1347", "4"): "4 .. Unsubscribe",
        ("FIX.5.0", "1347", "5"): "5 .. CancelRetransmission",
        ("FIX.5.0", "1347", "6"): "6 .. CancelRetransmissionUnsubscribe",
        ("FIX.5.0", "1348", "0"): "0 .. RequestSuccessfullyProcessed",
        ("FIX.5.0", "1348", "1"): "1 .. ApplicationDoesNotExist",
        ("FIX.5.0", "1348", "2"): "2 .. MessagesNotAvailable",
        ("FIX.5.0", "1354", "0"): "0 .. ApplicationDoesNotExist",
        ("FIX.5.0", "1354", "1"): "1 .. MessagesRequestedAreNotAvailable",
        ("FIX.5.0", "1354", "2"): "2 .. UserNotAuthorizedForApplication",
        ("FIX.5.0", "1358", "0"): "0 .. Put",
        ("FIX.5.0", "1358", "1"): "1 .. Call",
        ("FIX.5.0", "1358", "2"): "2 .. Other",
        ("FIX.5.0", "1358", "3"): "3 .. Chooser",
        ("FIX.5.0", "1368", "0"): "0 .. TradingResumes",
        ("FIX.5.0", "1368", "1"): "1 .. ChangeOfTradingSession",
        ("FIX.5.0", "1368", "2"): "2 .. ChangeOfTradingSubsession",
        ("FIX.5.0", "1368", "3"): "3 .. ChangeOfTradingStatus",
        ("FIX.5.0", "1373", "1"): "1 .. SuspendOrders",
        ("FIX.5.0", "1373", "2"): "2 .. ReleaseOrdersFromSuspension",
        ("FIX.5.0", "1373", "3"): "3 .. CancelOrders",
        ("FIX.5.0", "1374", "1"): "1 .. AllOrdersForASecurity",
        ("FIX.5.0", "1374", "2"): "2 .. AllOrdersForAnUnderlyingSecurity",
        ("FIX.5.0", "1374", "3"): "3 .. AllOrdersForAProduct",
        ("FIX.5.0", "1374", "4"): "4 .. AllOrdersForACFICode",
        ("FIX.5.0", "1374", "5"): "5 .. AllOrdersForASecurityType",
        ("FIX.5.0", "1374", "6"): "6 .. AllOrdersForATradingSession",
        ("FIX.5.0", "1374", "7"): "7 .. AllOrders",
        ("FIX.5.0", "1374", "8"): "8 .. AllOrdersForAMarket",
        ("FIX.5.0", "1374", "9"): "9 .. AllOrdersForAMarketSegment",
        ("FIX.5.0", "1374", "10"): "10 .. AllOrdersForASecurityGroup",
        ("FIX.5.0", "1374", "11"): "11 .. CancelForSecurityIssuer",
        ("FIX.5.0", "1374", "12"): "12 .. CancelForIssuerOfUnderlyingSecurity",
        ("FIX.5.0", "1375", "0"): "0 .. Rejected",
        ("FIX.5.0", "1375", "1"): "1 .. Accepted",
        ("FIX.5.0", "1375", "2"): "2 .. Completed",
        ("FIX.5.0", "1376", "0"): "0 .. MassActionNotSupported",
        ("FIX.5.0", "1376", "1"): "1 .. InvalidOrUnknownSecurity",
        ("FIX.5.0", "1376", "2"): "2 .. InvalidOrUnknownUnderlyingSecurity",
        ("FIX.5.0", "1376", "3"): "3 .. InvalidOrUnknownProduct",
        ("FIX.5.0", "1376", "4"): "4 .. InvalidOrUnknownCFICode",
        ("FIX.5.0", "1376", "5"): "5 .. InvalidOrUnknownSecurityType",
        ("FIX.5.0", "1376", "6"): "6 .. InvalidOrUnknownTradingSession",
        ("FIX.5.0", "1376", "7"): "7 .. InvalidOrUnknownMarket",
        ("FIX.5.0", "1376", "8"): "8 .. InvalidOrUnknownMarketSegment",
        ("FIX.5.0", "1376", "9"): "9 .. InvalidOrUnknownSecurityGroup",
        ("FIX.5.0", "1376", "10"): "10 .. InvalidOrUnknownSecurityIssuer",
        ("FIX.5.0", "1376", "11"): "11 .. InvalidOrUnknownIssuerOfUnderlyingSecurity",
        ("FIX.5.0", "1376", "99"): "99 .. Other",
        ("FIX.5.0", "1377", "0"): "0 .. PredefinedMultilegSecurity",
        ("FIX.5.0", "1377", "1"): "1 .. UserDefinedMultilegSecurity",
        ("FIX.5.0", "1377", "2"): "2 .. UserDefined",
        ("FIX.5.0", "1378", "0"): "0 .. NetPrice",
        ("FIX.5.0", "1378", "1"): "1 .. ReversedNetPrice",
        ("FIX.5.0", "1378", "2"): "2 .. YieldDifference",
        ("FIX.5.0", "1378", "3"): "3 .. Individual",
        ("FIX.5.0", "1378", "4"): "4 .. ContractWeightedAveragePrice",
        ("FIX.5.0", "1378", "5"): "5 .. MultipliedPrice",
        ("FIX.5.0", "1384", "0"): "0 .. StayOnOfferSide",
        ("FIX.5.0", "1384", "1"): "1 .. NotHeld",
        ("FIX.5.0", "1384", "2"): "2 .. Work",
        ("FIX.5.0", "1384", "3"): "3 .. GoAlong",
        ("FIX.5.0", "1384", "4"): "4 .. OverTheDay",
        ("FIX.5.0", "1384", "5"): "5 .. Held",
        ("FIX.5.0", "1384", "6"): "6 .. ParticipateDoNotInitiate",
        ("FIX.5.0", "1384", "7"): "7 .. StrictScale",
        ("FIX.5.0", "1384", "8"): "8 .. TryToScale",
        ("FIX.5.0", "1384", "9"): "9 .. StayOnBidSide",
        ("FIX.5.0", "1384", "A"): "A .. NoCross",
        ("FIX.5.0", "1384", "B"): "B .. OKToCross",
        ("FIX.5.0", "1384", "C"): "C .. CallFirst",
        ("FIX.5.0", "1384", "D"): "D .. PercentOfVolume",
        ("FIX.5.0", "1384", "E"): "E .. DoNotIncrease",
        ("FIX.5.0", "1384", "F"): "F .. DoNotReduce",
        ("FIX.5.0", "1384", "G"): "G .. AllOrNone",
        ("FIX.5.0", "1384", "H"): "H .. ReinstateOnSystemFailure",
        ("FIX.5.0", "1384", "I"): "I .. InstitutionsOnly",
        ("FIX.5.0", "1384", "J"): "J .. ReinstateOnTradingHalt",
        ("FIX.5.0", "1384", "K"): "K .. CancelOnTradingHalt",
        ("FIX.5.0", "1384", "N"): "N .. NonNegotiable",
        ("FIX.5.0", "1384", "Q"): "Q .. CancelOnSystemFailure",
        ("FIX.5.0", "1384", "S"): "S .. Suspend",
        ("FIX.5.0", "1384", "U"): "U .. CustomerDisplayInstruction",
        ("FIX.5.0", "1384", "V"): "V .. Netting",
        ("FIX.5.0", "1384", "X"): "X .. TradeAlong",
        ("FIX.5.0", "1384", "Y"): "Y .. TryToStop",
        ("FIX.5.0", "1384", "Z"): "Z .. CancelIfNotBest",
        ("FIX.5.0", "1384", "b"): "b .. StrictLimit",
        ("FIX.5.0", "1384", "c"): "c .. IgnorePriceValidityChecks",
        ("FIX.5.0", "1384", "e"): "e .. WorkToTargetStrategy",
        ("FIX.5.0", "1384", "f"): "f .. IntermarketSweep",
        ("FIX.5.0", "1384", "g"): "g .. ExternalRoutingAllowed",
        ("FIX.5.0", "1384", "h"): "h .. ExternalRoutingNotAllowed",
        ("FIX.5.0", "1384", "i"): "i .. ImbalanceOnly",
        ("FIX.5.0", "1384", "j"): "j .. SingleExecutionRequestedForBlockTrade",
        ("FIX.5.0", "1384", "k"): "k .. BestExecution",
        ("FIX.5.0", "1384", "l"): "l .. SuspendOnSystemFailure",
        ("FIX.5.0", "1384", "m"): "m .. SuspendOnTradingHalt",
        ("FIX.5.0", "1384", "n"): "n .. ReinstateOnConnectionLoss",
        ("FIX.5.0", "1384", "o"): "o .. CancelOnConnectionLoss",
        ("FIX.5.0", "1384", "p"): "p .. SuspendOnConnectionLoss",
        ("FIX.5.0", "1384", "q"): "q .. Release",
        ("FIX.5.0", "1384", "r"): "r .. ExecuteAsDeltaNeutral",
        ("FIX.5.0", "1384", "s"): "s .. ExecuteAsDurationNeutral",
        ("FIX.5.0", "1384", "t"): "t .. ExecuteAsFXNeutral",
        ("FIX.5.0", "1384", "u"): "u .. MinGuaranteedFillEligible",
        ("FIX.5.0", "1384", "v"): "v .. BypassNonDisplayLiquidity",
        ("FIX.5.0", "1384", "w"): "w .. Lock",
        ("FIX.5.0", "1384", "x"): "x .. IgnoreNotionalValueChecks",
        ("FIX.5.0", "1384", "y"): "y .. TrdAtRefPx",
        ("FIX.5.0", "1385", "1"): "1 .. OneCancelsTheOther",
        ("FIX.5.0", "1385", "2"): "2 .. OneTriggersTheOther",
        ("FIX.5.0", "1385", "3"): "3 .. OneUpdatesTheOtherAbsolute",
        ("FIX.5.0", "1385", "4"): "4 .. OneUpdatesTheOtherProportional",
        ("FIX.5.0", "1385", "5"): "5 .. BidAndOffer",
        ("FIX.5.0", "1385", "6"): "6 .. BidAndOfferOCO",
        ("FIX.5.0", "1386", "0"): "0 .. BrokerCredit",
        ("FIX.5.0", "1386", "2"): "2 .. ExchangeClosed",
        ("FIX.5.0", "1386", "4"): "4 .. TooLateToEnter",
        ("FIX.5.0", "1386", "5"): "5 .. UnknownOrder",
        ("FIX.5.0", "1386", "6"): "6 .. DuplicateOrder",
        ("FIX.5.0", "1386", "11"): "11 .. UnsupportedOrderCharacteristic",
        ("FIX.5.0", "1386", "99"): "99 .. Other",
        ("FIX.5.0", "1388", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1388", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1388", "3"): "3 .. ClientID",
        ("FIX.5.0", "1388", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1388", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1388", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1388", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1388", "8"): "8 .. Locate",
        ("FIX.5.0", "1388", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1388", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1388", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1388", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1388", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1388", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1388", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1388", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1388", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1388", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1388", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1388", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1388", "22"): "22 .. Exchange",
        ("FIX.5.0", "1388", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1388", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1388", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1388", "27"): "27 .. Buyer",
        ("FIX.5.0", "1388", "28"): "28 .. Custodian",
        ("FIX.5.0", "1388", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1388", "30"): "30 .. Agent",
        ("FIX.5.0", "1388", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1388", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1388", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1388", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1388", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1388", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1388", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1388", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1388", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1388", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1388", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1388", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1388", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1388", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1388", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1388", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1388", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1388", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1388", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1388", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1388", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1388", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1388", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1388", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1388", "55"): "55 .. SessionID",
        ("FIX.5.0", "1388", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1388", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1388", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1388", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1388", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1388", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1388", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1388", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1388", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1388", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1388", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1388", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1388", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1388", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1388", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1388", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1388", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1388", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1388", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1388", "75"): "75 .. LocationID",
        ("FIX.5.0", "1388", "76"): "76 .. DeskID",
        ("FIX.5.0", "1388", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1388", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1388", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1388", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1388", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1388", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1388", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1388", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1388", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1388", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1388", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1388", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1388", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1388", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1388", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1388", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1388", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1388", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1388", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1388", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1388", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1388", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1388", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1388", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1388", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1388", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1388", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1388", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1388", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1388", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1388", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1388", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1388", "110"): "110 .. Borrower",
        ("FIX.5.0", "1388", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1388", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1388", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1388", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1388", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1388", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1388", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1388", "118"): "118 .. Operator",
        ("FIX.5.0", "1388", "119"): "119 .. CSD",
        ("FIX.5.0", "1388", "120"): "120 .. ICSD",
        ("FIX.5.0", "1388", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1388", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1388", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1390", "0"): "0 .. DoNotPublishTrade",
        ("FIX.5.0", "1390", "1"): "1 .. PublishTrade",
        ("FIX.5.0", "1390", "2"): "2 .. DeferredPublication",
        ("FIX.5.0", "1390", "3"): "3 .. Published",
        ("FIX.5.0", "1395", "A"): "A .. Add",
        ("FIX.5.0", "1395", "D"): "D .. Delete",
        ("FIX.5.0", "1395", "M"): "M .. Modify",
        ("FIX.5.0", "1409", "0"): "0 .. SessionActive",
        ("FIX.5.0", "1409", "1"): "1 .. SessionPasswordChanged",
        ("FIX.5.0", "1409", "2"): "2 .. SessionPasswordDueToExpire",
        ("FIX.5.0", "1409", "3"): "3 .. NewSessionPasswordDoesNotComplyWithPolicy",
        ("FIX.5.0", "1409", "4"): "4 .. SessionLogoutComplete",
        ("FIX.5.0", "1409", "5"): "5 .. InvalidUsernameOrPassword",
        ("FIX.5.0", "1409", "6"): "6 .. AccountLocked",
        ("FIX.5.0", "1409", "7"): "7 .. LogonsAreNotAllowedAtThisTime",
        ("FIX.5.0", "1409", "8"): "8 .. PasswordExpired",
        ("FIX.5.0", "1409", "9"): "9 .. ReceivedMsgSeqNumTooLow",
        ("FIX.5.0", "1409", "10"): "10 .. ReceivedNextExpectedMsgSeqNumTooHigh",
        ("FIX.5.0", "1411", "1"): "1 .. Firm",
        ("FIX.5.0", "1411", "2"): "2 .. Person",
        ("FIX.5.0", "1411", "3"): "3 .. System",
        ("FIX.5.0", "1411", "4"): "4 .. Application",
        ("FIX.5.0", "1411", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1411", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1411", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1411", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1411", "9"): "9 .. ContactName",
        ("FIX.5.0", "1411", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1411", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1411", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1411", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1411", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1411", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1411", "16"): "16 .. BIC",
        ("FIX.5.0", "1411", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1411", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1411", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1411", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1411", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1411", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1411", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1411", "24"): "24 .. Department",
        ("FIX.5.0", "1411", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1411", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1411", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1411", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1411", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1411", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1411", "31"): "31 .. Location",
        ("FIX.5.0", "1411", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1411", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1411", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1411", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1411", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1411", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1411", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1411", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1411", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1411", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1411", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1411", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1411", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1411", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1411", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1411", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1411", "48"): "48 .. USPerson",
        ("FIX.5.0", "1411", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1411", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1411", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1411", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1411", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1411", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1411", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1411", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1411", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1411", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1411", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1411", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1411", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1411", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1411", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1411", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1411", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1411", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1411", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1411", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1411", "69"): "69 .. EndUser",
        ("FIX.5.0", "1411", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1411", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1411", "72"): "72 .. Domicile",
        ("FIX.5.0", "1411", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1411", "74"): "74 .. Payer",
        ("FIX.5.0", "1411", "75"): "75 .. Receiver",
        ("FIX.5.0", "1411", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1411", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1411", "78"): "78 .. FirstName",
        ("FIX.5.0", "1411", "79"): "79 .. Surname",
        ("FIX.5.0", "1411", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1411", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1411", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1411", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1411", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1416", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1416", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1416", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1416", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1416", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1416", "J"): "J .. TaxID",
        ("FIX.5.0", "1416", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1416", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1416", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1416", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1416", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1416", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1416", "B"): "B .. BIC",
        ("FIX.5.0", "1416", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1416", "D"): "D .. Proprietary",
        ("FIX.5.0", "1416", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1416", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1416", "G"): "G .. MIC",
        ("FIX.5.0", "1416", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1416", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1416", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1416", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1416", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1416", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1416", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1416", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1417", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1417", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1417", "3"): "3 .. ClientID",
        ("FIX.5.0", "1417", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1417", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1417", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1417", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1417", "8"): "8 .. Locate",
        ("FIX.5.0", "1417", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1417", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1417", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1417", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1417", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1417", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1417", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1417", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1417", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1417", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1417", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1417", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1417", "22"): "22 .. Exchange",
        ("FIX.5.0", "1417", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1417", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1417", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1417", "27"): "27 .. Buyer",
        ("FIX.5.0", "1417", "28"): "28 .. Custodian",
        ("FIX.5.0", "1417", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1417", "30"): "30 .. Agent",
        ("FIX.5.0", "1417", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1417", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1417", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1417", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1417", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1417", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1417", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1417", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1417", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1417", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1417", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1417", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1417", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1417", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1417", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1417", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1417", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1417", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1417", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1417", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1417", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1417", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1417", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1417", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1417", "55"): "55 .. SessionID",
        ("FIX.5.0", "1417", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1417", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1417", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1417", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1417", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1417", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1417", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1417", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1417", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1417", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1417", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1417", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1417", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1417", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1417", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1417", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1417", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1417", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1417", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1417", "75"): "75 .. LocationID",
        ("FIX.5.0", "1417", "76"): "76 .. DeskID",
        ("FIX.5.0", "1417", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1417", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1417", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1417", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1417", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1417", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1417", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1417", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1417", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1417", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1417", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1417", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1417", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1417", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1417", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1417", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1417", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1417", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1417", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1417", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1417", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1417", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1417", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1417", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1417", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1417", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1417", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1417", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1417", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1417", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1417", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1417", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1417", "110"): "110 .. Borrower",
        ("FIX.5.0", "1417", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1417", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1417", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1417", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1417", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1417", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1417", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1417", "118"): "118 .. Operator",
        ("FIX.5.0", "1417", "119"): "119 .. CSD",
        ("FIX.5.0", "1417", "120"): "120 .. ICSD",
        ("FIX.5.0", "1417", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1417", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1417", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1419", "0"): "0 .. European",
        ("FIX.5.0", "1419", "1"): "1 .. American",
        ("FIX.5.0", "1419", "2"): "2 .. Bermuda",
        ("FIX.5.0", "1419", "99"): "99 .. Other",
        ("FIX.5.0", "1420", "0"): "0 .. European",
        ("FIX.5.0", "1420", "1"): "1 .. American",
        ("FIX.5.0", "1420", "2"): "2 .. Bermuda",
        ("FIX.5.0", "1420", "99"): "99 .. Other",
        ("FIX.5.0", "1421", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "1421", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "1421", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "1421", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "1421", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "1421", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "1421", "thm"): "thm .. therms",
        ("FIX.5.0", "1421", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "1421", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "1421", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "1421", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "1421", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "1421", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "1421", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "1421", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "1421", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "1421", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "1421", "day"): "day .. Day",
        ("FIX.5.0", "1421", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "1421", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "1421", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "1421", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "1421", "g"): "g .. Grams",
        ("FIX.5.0", "1421", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "1421", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "1421", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "1421", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "1421", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "1421", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "1421", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "1421", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "1421", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "1421", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "1421", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "1421", "L"): "L .. liters",
        ("FIX.5.0", "1421", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "1421", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "1421", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "1421", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "1421", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "1421", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "1421", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "1421", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "1421", "t"): "t .. MetricTons",
        ("FIX.5.0", "1421", "tn"): "tn .. Tons",
        ("FIX.5.0", "1424", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "1424", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "1424", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "1424", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "1424", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "1424", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "1424", "thm"): "thm .. therms",
        ("FIX.5.0", "1424", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "1424", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "1424", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "1424", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "1424", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "1424", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "1424", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "1424", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "1424", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "1424", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "1424", "day"): "day .. Day",
        ("FIX.5.0", "1424", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "1424", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "1424", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "1424", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "1424", "g"): "g .. Grams",
        ("FIX.5.0", "1424", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "1424", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "1424", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "1424", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "1424", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "1424", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "1424", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "1424", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "1424", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "1424", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "1424", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "1424", "L"): "L .. liters",
        ("FIX.5.0", "1424", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "1424", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "1424", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "1424", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "1424", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "1424", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "1424", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "1424", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "1424", "t"): "t .. MetricTons",
        ("FIX.5.0", "1424", "tn"): "tn .. Tons",
        ("FIX.5.0", "1426", "0"): "0 .. ApplSeqNumReset",
        ("FIX.5.0", "1426", "1"): "1 .. LastMessageSent",
        ("FIX.5.0", "1426", "2"): "2 .. ApplicationAlive",
        ("FIX.5.0", "1426", "3"): "3 .. ResendComplete",
        ("FIX.5.0", "1429", "0"): "0 .. Seconds",
        ("FIX.5.0", "1429", "1"): "1 .. TenthsOfASecond",
        ("FIX.5.0", "1429", "2"): "2 .. HundredthsOfASecond",
        ("FIX.5.0", "1429", "3"): "3 .. Milliseconds",
        ("FIX.5.0", "1429", "4"): "4 .. Microseconds",
        ("FIX.5.0", "1429", "5"): "5 .. Nanoseconds",
        ("FIX.5.0", "1429", "10"): "10 .. Minutes",
        ("FIX.5.0", "1429", "11"): "11 .. Hours",
        ("FIX.5.0", "1429", "12"): "12 .. Days",
        ("FIX.5.0", "1429", "13"): "13 .. Weeks",
        ("FIX.5.0", "1429", "14"): "14 .. Months",
        ("FIX.5.0", "1429", "15"): "15 .. Years",
        ("FIX.5.0", "1430", "E"): "E .. Electronic",
        ("FIX.5.0", "1430", "P"): "P .. Pit",
        ("FIX.5.0", "1430", "X"): "X .. ExPit",
        ("FIX.5.0", "1430", "C"): "C .. ClearingHouse",
        ("FIX.5.0", "1430", "R"): "R .. RegisteredMarket",
        ("FIX.5.0", "1430", "O"): "O .. OffMarket",
        ("FIX.5.0", "1430", "B"): "B .. CentralLimitOrderBook",
        ("FIX.5.0", "1430", "Q"): "Q .. QuoteDrivenMarket",
        ("FIX.5.0", "1430", "D"): "D .. DarkOrderBook",
        ("FIX.5.0", "1430", "A"): "A .. AuctionDrivenMarket",
        ("FIX.5.0", "1430", "N"): "N .. QuoteNegotiation",
        ("FIX.5.0", "1430", "V"): "V .. VoiceNegotiation",
        ("FIX.5.0", "1430", "H"): "H .. HybridMarket",
        ("FIX.5.0", "1431", "0"): "0 .. GTCFromPreviousDay",
        ("FIX.5.0", "1431", "1"): "1 .. PartialFillRemaining",
        ("FIX.5.0", "1431", "2"): "2 .. OrderChanged",
        ("FIX.5.0", "1432", "1"): "1 .. MemberTradingForTheirOwnAccount",
        ("FIX.5.0", "1432", "2"): "2 .. ClearingFirmTradingForItsProprietaryAccount",
        ("FIX.5.0", "1432", "3"): "3 .. MemberTradingForAnotherMember",
        ("FIX.5.0", "1432", "4"): "4 .. AllOther",
        ("FIX.5.0", "1434", "0"): "0 .. UtilityProvidedStandardModel",
        ("FIX.5.0", "1434", "1"): "1 .. ProprietaryModel",
        ("FIX.5.0", "1435", "0"): "0 .. Shares",
        ("FIX.5.0", "1435", "1"): "1 .. Hours",
        ("FIX.5.0", "1435", "2"): "2 .. Days",
        ("FIX.5.0", "1436", "0"): "0 .. Shares",
        ("FIX.5.0", "1436", "1"): "1 .. Hours",
        ("FIX.5.0", "1436", "2"): "2 .. Days",
        ("FIX.5.0", "1437", "0"): "0 .. Shares",
        ("FIX.5.0", "1437", "1"): "1 .. Hours",
        ("FIX.5.0", "1437", "2"): "2 .. Days",
        ("FIX.5.0", "1438", "0"): "0 .. Shares",
        ("FIX.5.0", "1438", "1"): "1 .. Hours",
        ("FIX.5.0", "1438", "2"): "2 .. Days",
        ("FIX.5.0", "1439", "0"): "0 .. NERCEasternOffPeak",
        ("FIX.5.0", "1439", "1"): "1 .. NERCWesternOffPeak",
        ("FIX.5.0", "1439", "2"): "2 .. NERCCalendarAllDaysInMonth",
        ("FIX.5.0", "1439", "3"): "3 .. NERCEasternPeak",
        ("FIX.5.0", "1439", "4"): "4 .. NERCWesternPeak",
        ("FIX.5.0", "1439", "5"): "5 .. AllTimes",
        ("FIX.5.0", "1439", "6"): "6 .. OnPeak",
        ("FIX.5.0", "1439", "7"): "7 .. OffPeak",
        ("FIX.5.0", "1439", "8"): "8 .. Base",
        ("FIX.5.0", "1439", "9"): "9 .. Block",
        ("FIX.5.0", "1439", "99"): "99 .. Other",
        ("FIX.5.0", "1440", "0"): "0 .. NERCEasternOffPeak",
        ("FIX.5.0", "1440", "1"): "1 .. NERCWesternOffPeak",
        ("FIX.5.0", "1440", "2"): "2 .. NERCCalendarAllDaysInMonth",
        ("FIX.5.0", "1440", "3"): "3 .. NERCEasternPeak",
        ("FIX.5.0", "1440", "4"): "4 .. NERCWesternPeak",
        ("FIX.5.0", "1440", "5"): "5 .. AllTimes",
        ("FIX.5.0", "1440", "6"): "6 .. OnPeak",
        ("FIX.5.0", "1440", "7"): "7 .. OffPeak",
        ("FIX.5.0", "1440", "8"): "8 .. Base",
        ("FIX.5.0", "1440", "9"): "9 .. Block",
        ("FIX.5.0", "1440", "99"): "99 .. Other",
        ("FIX.5.0", "1441", "0"): "0 .. NERCEasternOffPeak",
        ("FIX.5.0", "1441", "1"): "1 .. NERCWesternOffPeak",
        ("FIX.5.0", "1441", "2"): "2 .. NERCCalendarAllDaysInMonth",
        ("FIX.5.0", "1441", "3"): "3 .. NERCEasternPeak",
        ("FIX.5.0", "1441", "4"): "4 .. NERCWesternPeak",
        ("FIX.5.0", "1441", "5"): "5 .. AllTimes",
        ("FIX.5.0", "1441", "6"): "6 .. OnPeak",
        ("FIX.5.0", "1441", "7"): "7 .. OffPeak",
        ("FIX.5.0", "1441", "8"): "8 .. Base",
        ("FIX.5.0", "1441", "9"): "9 .. Block",
        ("FIX.5.0", "1441", "99"): "99 .. Other",
        ("FIX.5.0", "1442", "0"): "0 .. NERCEasternOffPeak",
        ("FIX.5.0", "1442", "1"): "1 .. NERCWesternOffPeak",
        ("FIX.5.0", "1442", "2"): "2 .. NERCCalendarAllDaysInMonth",
        ("FIX.5.0", "1442", "3"): "3 .. NERCEasternPeak",
        ("FIX.5.0", "1442", "4"): "4 .. NERCWesternPeak",
        ("FIX.5.0", "1442", "5"): "5 .. AllTimes",
        ("FIX.5.0", "1442", "6"): "6 .. OnPeak",
        ("FIX.5.0", "1442", "7"): "7 .. OffPeak",
        ("FIX.5.0", "1442", "8"): "8 .. Base",
        ("FIX.5.0", "1442", "9"): "9 .. Block",
        ("FIX.5.0", "1442", "99"): "99 .. Other",
        ("FIX.5.0", "1443", "1"): "1 .. AddedLiquidity",
        ("FIX.5.0", "1443", "2"): "2 .. RemovedLiquidity",
        ("FIX.5.0", "1443", "3"): "3 .. LiquidityRoutedOut",
        ("FIX.5.0", "1443", "4"): "4 .. Auction",
        ("FIX.5.0", "1443", "5"): "5 .. TriggeredStopOrder",
        ("FIX.5.0", "1443", "6"): "6 .. TriggeredContingencyOrder",
        ("FIX.5.0", "1443", "7"): "7 .. TriggeredMarketOrder",
        ("FIX.5.0", "1444", "1"): "1 .. AddedLiquidity",
        ("FIX.5.0", "1444", "2"): "2 .. RemovedLiquidity",
        ("FIX.5.0", "1444", "3"): "3 .. LiquidityRoutedOut",
        ("FIX.5.0", "1444", "4"): "4 .. Auction",
        ("FIX.5.0", "1444", "5"): "5 .. TriggeredStopOrder",
        ("FIX.5.0", "1444", "6"): "6 .. TriggeredContingencyOrder",
        ("FIX.5.0", "1444", "7"): "7 .. TriggeredMarketOrder",
        ("FIX.5.0", "1446", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "1446", "1"): "1 .. Reuters",
        ("FIX.5.0", "1446", "2"): "2 .. Telerate",
        ("FIX.5.0", "1446", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "1446", "99"): "99 .. Other",
        ("FIX.5.0", "1447", "0"): "0 .. Primary",
        ("FIX.5.0", "1447", "1"): "1 .. Secondary",
        ("FIX.5.0", "1449", "FR"): "FR .. FullRestructuring",
        ("FIX.5.0", "1449", "MR"): "MR .. ModifiedRestructuring",
        ("FIX.5.0", "1449", "MM"): "MM .. ModifiedModRestructuring",
        ("FIX.5.0", "1449", "XR"): "XR .. NoRestructuringSpecified",
        ("FIX.5.0", "1450", "SD"): "SD .. SeniorSecured",
        ("FIX.5.0", "1450", "SR"): "SR .. Senior",
        ("FIX.5.0", "1450", "SB"): "SB .. Subordinated",
        ("FIX.5.0", "1450", "JR"): "JR .. Junior",
        ("FIX.5.0", "1450", "MZ"): "MZ .. Mezzanine",
        ("FIX.5.0", "1450", "SN"): "SN .. SeniorNonPreferred",
        ("FIX.5.0", "1453", "FR"): "FR .. FullRestructuring",
        ("FIX.5.0", "1453", "MR"): "MR .. ModifiedRestructuring",
        ("FIX.5.0", "1453", "MM"): "MM .. ModifiedModRestructuring",
        ("FIX.5.0", "1453", "XR"): "XR .. NoRestructuringSpecified",
        ("FIX.5.0", "1454", "SD"): "SD .. SeniorSecured",
        ("FIX.5.0", "1454", "SR"): "SR .. Senior",
        ("FIX.5.0", "1454", "SB"): "SB .. Subordinated",
        ("FIX.5.0", "1454", "JR"): "JR .. Junior",
        ("FIX.5.0", "1454", "MZ"): "MZ .. Mezzanine",
        ("FIX.5.0", "1454", "SN"): "SN .. SeniorNonPreferred",
        ("FIX.5.0", "1463", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1463", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1463", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1463", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1463", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1463", "J"): "J .. TaxID",
        ("FIX.5.0", "1463", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1463", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1463", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1463", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1463", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1463", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1463", "B"): "B .. BIC",
        ("FIX.5.0", "1463", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1463", "D"): "D .. Proprietary",
        ("FIX.5.0", "1463", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1463", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1463", "G"): "G .. MIC",
        ("FIX.5.0", "1463", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1463", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1463", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1463", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1463", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1463", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1463", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1463", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1464", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1464", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1464", "3"): "3 .. ClientID",
        ("FIX.5.0", "1464", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1464", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1464", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1464", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1464", "8"): "8 .. Locate",
        ("FIX.5.0", "1464", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1464", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1464", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1464", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1464", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1464", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1464", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1464", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1464", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1464", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1464", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1464", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1464", "22"): "22 .. Exchange",
        ("FIX.5.0", "1464", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1464", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1464", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1464", "27"): "27 .. Buyer",
        ("FIX.5.0", "1464", "28"): "28 .. Custodian",
        ("FIX.5.0", "1464", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1464", "30"): "30 .. Agent",
        ("FIX.5.0", "1464", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1464", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1464", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1464", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1464", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1464", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1464", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1464", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1464", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1464", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1464", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1464", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1464", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1464", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1464", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1464", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1464", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1464", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1464", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1464", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1464", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1464", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1464", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1464", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1464", "55"): "55 .. SessionID",
        ("FIX.5.0", "1464", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1464", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1464", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1464", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1464", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1464", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1464", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1464", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1464", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1464", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1464", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1464", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1464", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1464", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1464", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1464", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1464", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1464", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1464", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1464", "75"): "75 .. LocationID",
        ("FIX.5.0", "1464", "76"): "76 .. DeskID",
        ("FIX.5.0", "1464", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1464", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1464", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1464", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1464", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1464", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1464", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1464", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1464", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1464", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1464", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1464", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1464", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1464", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1464", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1464", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1464", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1464", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1464", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1464", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1464", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1464", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1464", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1464", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1464", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1464", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1464", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1464", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1464", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1464", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1464", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1464", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1464", "110"): "110 .. Borrower",
        ("FIX.5.0", "1464", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1464", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1464", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1464", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1464", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1464", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1464", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1464", "118"): "118 .. Operator",
        ("FIX.5.0", "1464", "119"): "119 .. CSD",
        ("FIX.5.0", "1464", "120"): "120 .. ICSD",
        ("FIX.5.0", "1464", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1464", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1464", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1470", "1"): "1 .. IndustryClassification",
        ("FIX.5.0", "1470", "2"): "2 .. TradingList",
        ("FIX.5.0", "1470", "3"): "3 .. Market",
        ("FIX.5.0", "1470", "4"): "4 .. NewspaperList",
        ("FIX.5.0", "1471", "1"): "1 .. ICB",
        ("FIX.5.0", "1471", "2"): "2 .. NAICS",
        ("FIX.5.0", "1471", "3"): "3 .. GICS",
        ("FIX.5.0", "1473", "0"): "0 .. CompanyNews",
        ("FIX.5.0", "1473", "1"): "1 .. MarketplaceNews",
        ("FIX.5.0", "1473", "2"): "2 .. FinancialMarketNews",
        ("FIX.5.0", "1473", "3"): "3 .. TechnicalNews",
        ("FIX.5.0", "1473", "99"): "99 .. OtherNews",
        ("FIX.5.0", "1477", "0"): "0 .. Replacement",
        ("FIX.5.0", "1477", "1"): "1 .. OtherLanguage",
        ("FIX.5.0", "1477", "2"): "2 .. Complimentary",
        ("FIX.5.0", "1477", "3"): "3 .. Withdrawal",
        ("FIX.5.0", "1478", "1"): "1 .. FixedStrike",
        ("FIX.5.0", "1478", "2"): "2 .. StrikeSetAtExpiration",
        ("FIX.5.0", "1478", "3"): "3 .. StrikeSetToAverageAcrossLife",
        ("FIX.5.0", "1478", "4"): "4 .. StrikeSetToOptimalValue",
        ("FIX.5.0", "1479", "1"): "1 .. LessThan",
        ("FIX.5.0", "1479", "2"): "2 .. LessThanOrEqual",
        ("FIX.5.0", "1479", "3"): "3 .. Equal",
        ("FIX.5.0", "1479", "4"): "4 .. GreaterThanOrEqual",
        ("FIX.5.0", "1479", "5"): "5 .. GreaterThan",
        ("FIX.5.0", "1481", "1"): "1 .. Regular",
        ("FIX.5.0", "1481", "2"): "2 .. SpecialReference",
        ("FIX.5.0", "1481", "3"): "3 .. OptimalValue",
        ("FIX.5.0", "1481", "4"): "4 .. AverageValue",
        ("FIX.5.0", "1482", "1"): "1 .. Vanilla",
        ("FIX.5.0", "1482", "2"): "2 .. Capped",
        ("FIX.5.0", "1482", "3"): "3 .. Binary",
        ("FIX.5.0", "1482", "4"): "4 .. Asian",
        ("FIX.5.0", "1482", "5"): "5 .. Barrier",
        ("FIX.5.0", "1482", "6"): "6 .. DigitalBarrier",
        ("FIX.5.0", "1482", "7"): "7 .. Lookback",
        ("FIX.5.0", "1482", "8"): "8 .. OtherPathDependent",
        ("FIX.5.0", "1482", "99"): "99 .. Other",
        ("FIX.5.0", "1484", "1"): "1 .. Capped",
        ("FIX.5.0", "1484", "2"): "2 .. Trigger",
        ("FIX.5.0", "1484", "3"): "3 .. KnockInUp",
        ("FIX.5.0", "1484", "4"): "4 .. KnockInDown",
        ("FIX.5.0", "1484", "5"): "5 .. KnockOutUp",
        ("FIX.5.0", "1484", "6"): "6 .. KnockOutDown",
        ("FIX.5.0", "1484", "7"): "7 .. Underlying",
        ("FIX.5.0", "1484", "8"): "8 .. ResetBarrier",
        ("FIX.5.0", "1484", "9"): "9 .. RollingBarrier",
        ("FIX.5.0", "1484", "10"): "10 .. OneTouch",
        ("FIX.5.0", "1484", "11"): "11 .. NoTouch",
        ("FIX.5.0", "1484", "12"): "12 .. DblOneTouch",
        ("FIX.5.0", "1484", "13"): "13 .. DblNoTouch",
        ("FIX.5.0", "1484", "14"): "14 .. FXComposite",
        ("FIX.5.0", "1484", "15"): "15 .. FXQuanto",
        ("FIX.5.0", "1484", "16"): "16 .. FXCrssCcy",
        ("FIX.5.0", "1484", "17"): "17 .. StrkSpread",
        ("FIX.5.0", "1484", "18"): "18 .. ClndrSpread",
        ("FIX.5.0", "1484", "19"): "19 .. PxObsvtn",
        ("FIX.5.0", "1484", "20"): "20 .. PassThrough",
        ("FIX.5.0", "1484", "21"): "21 .. StrkSched",
        ("FIX.5.0", "1484", "22"): "22 .. EquityValuation",
        ("FIX.5.0", "1484", "23"): "23 .. DividendValuation",
        ("FIX.5.0", "1487", "1"): "1 .. LessThanComplexEventPrice",
        ("FIX.5.0", "1487", "2"): "2 .. LessThanOrEqualToComplexEventPrice",
        ("FIX.5.0", "1487", "3"): "3 .. EqualToComplexEventPrice",
        ("FIX.5.0", "1487", "4"): "4 .. GreaterThanOrEqualToComplexEventPrice",
        ("FIX.5.0", "1487", "5"): "5 .. GreaterThanComplexEventPrice",
        ("FIX.5.0", "1489", "1"): "1 .. Expiration",
        ("FIX.5.0", "1489", "2"): "2 .. Immediate",
        ("FIX.5.0", "1489", "3"): "3 .. SpecifiedDate",
        ("FIX.5.0", "1489", "4"): "4 .. Close",
        ("FIX.5.0", "1489", "5"): "5 .. Open",
        ("FIX.5.0", "1489", "6"): "6 .. OfficialSettlPrice",
        ("FIX.5.0", "1489", "7"): "7 .. DerivativesClose",
        ("FIX.5.0", "1489", "8"): "8 .. AsSpecifiedMasterConfirmation",
        ("FIX.5.0", "1490", "1"): "1 .. And",
        ("FIX.5.0", "1490", "2"): "2 .. Or",
        ("FIX.5.0", "1498", "1"): "1 .. StreamAssignmentForNewCustomer",
        ("FIX.5.0", "1498", "2"): "2 .. StreamAssignmentForExistingCustomer",
        ("FIX.5.0", "1502", "0"): "0 .. UnknownClient",
        ("FIX.5.0", "1502", "1"): "1 .. ExceedsMaximumSize",
        ("FIX.5.0", "1502", "2"): "2 .. UnknownOrInvalidCurrencyPair",
        ("FIX.5.0", "1502", "3"): "3 .. NoAvailableStream",
        ("FIX.5.0", "1502", "99"): "99 .. Other",
        ("FIX.5.0", "1503", "0"): "0 .. AssignmentAccepted",
        ("FIX.5.0", "1503", "1"): "1 .. AssignmentRejected",
        ("FIX.5.0", "1509", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1509", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1509", "3"): "3 .. ClientID",
        ("FIX.5.0", "1509", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1509", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1509", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1509", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1509", "8"): "8 .. Locate",
        ("FIX.5.0", "1509", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1509", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1509", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1509", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1509", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1509", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1509", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1509", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1509", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1509", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1509", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1509", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1509", "22"): "22 .. Exchange",
        ("FIX.5.0", "1509", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1509", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1509", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1509", "27"): "27 .. Buyer",
        ("FIX.5.0", "1509", "28"): "28 .. Custodian",
        ("FIX.5.0", "1509", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1509", "30"): "30 .. Agent",
        ("FIX.5.0", "1509", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1509", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1509", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1509", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1509", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1509", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1509", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1509", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1509", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1509", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1509", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1509", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1509", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1509", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1509", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1509", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1509", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1509", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1509", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1509", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1509", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1509", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1509", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1509", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1509", "55"): "55 .. SessionID",
        ("FIX.5.0", "1509", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1509", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1509", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1509", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1509", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1509", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1509", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1509", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1509", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1509", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1509", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1509", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1509", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1509", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1509", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1509", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1509", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1509", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1509", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1509", "75"): "75 .. LocationID",
        ("FIX.5.0", "1509", "76"): "76 .. DeskID",
        ("FIX.5.0", "1509", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1509", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1509", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1509", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1509", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1509", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1509", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1509", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1509", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1509", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1509", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1509", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1509", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1509", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1509", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1509", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1509", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1509", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1509", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1509", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1509", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1509", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1509", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1509", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1509", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1509", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1509", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1509", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1509", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1509", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1509", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1509", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1509", "110"): "110 .. Borrower",
        ("FIX.5.0", "1509", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1509", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1509", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1509", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1509", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1509", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1509", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1509", "118"): "118 .. Operator",
        ("FIX.5.0", "1509", "119"): "119 .. CSD",
        ("FIX.5.0", "1509", "120"): "120 .. ICSD",
        ("FIX.5.0", "1509", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1509", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1509", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1511", "0"): "0 .. ValidRequest",
        ("FIX.5.0", "1511", "1"): "1 .. InvalidOrUnsupportedRequest",
        ("FIX.5.0", "1511", "2"): "2 .. NoDataFound",
        ("FIX.5.0", "1511", "3"): "3 .. NotAuthorized",
        ("FIX.5.0", "1511", "4"): "4 .. DataTemporarilyUnavailable",
        ("FIX.5.0", "1511", "5"): "5 .. RequestForDataNotSupported",
        ("FIX.5.0", "1511", "99"): "99 .. Other",
        ("FIX.5.0", "1515", "0"): "0 .. IsAlso",
        ("FIX.5.0", "1515", "1"): "1 .. ClearsFor",
        ("FIX.5.0", "1515", "2"): "2 .. ClearsThrough",
        ("FIX.5.0", "1515", "3"): "3 .. TradesFor",
        ("FIX.5.0", "1515", "4"): "4 .. TradesThrough",
        ("FIX.5.0", "1515", "5"): "5 .. Sponsors",
        ("FIX.5.0", "1515", "6"): "6 .. SponsoredThrough",
        ("FIX.5.0", "1515", "7"): "7 .. ProvidesGuaranteeFor",
        ("FIX.5.0", "1515", "8"): "8 .. IsGuaranteedBy",
        ("FIX.5.0", "1515", "9"): "9 .. MemberOf",
        ("FIX.5.0", "1515", "10"): "10 .. HasMembers",
        ("FIX.5.0", "1515", "11"): "11 .. ProvidesMarketplaceFor",
        ("FIX.5.0", "1515", "12"): "12 .. ParticipantOfMarketplace",
        ("FIX.5.0", "1515", "13"): "13 .. CarriesPositionsFor",
        ("FIX.5.0", "1515", "14"): "14 .. PostsTradesTo",
        ("FIX.5.0", "1515", "15"): "15 .. EntersTradesFor",
        ("FIX.5.0", "1515", "16"): "16 .. EntersTradesThrough",
        ("FIX.5.0", "1515", "17"): "17 .. ProvidesQuotesTo",
        ("FIX.5.0", "1515", "18"): "18 .. RequestsQuotesFrom",
        ("FIX.5.0", "1515", "19"): "19 .. InvestsFor",
        ("FIX.5.0", "1515", "20"): "20 .. InvestsThrough",
        ("FIX.5.0", "1515", "21"): "21 .. BrokersTradesFor",
        ("FIX.5.0", "1515", "22"): "22 .. BrokersTradesThrough",
        ("FIX.5.0", "1515", "23"): "23 .. ProvidesTradingServicesFor",
        ("FIX.5.0", "1515", "24"): "24 .. UsesTradingServicesOf",
        ("FIX.5.0", "1515", "25"): "25 .. ApprovesOf",
        ("FIX.5.0", "1515", "26"): "26 .. ApprovedBy",
        ("FIX.5.0", "1515", "27"): "27 .. ParentFirmFor",
        ("FIX.5.0", "1515", "28"): "28 .. SubsidiaryOf",
        ("FIX.5.0", "1515", "29"): "29 .. RegulatoryOwnerOf",
        ("FIX.5.0", "1515", "30"): "30 .. OwnedByRegulatory",
        ("FIX.5.0", "1515", "31"): "31 .. Controls",
        ("FIX.5.0", "1515", "32"): "32 .. IsControlledBy",
        ("FIX.5.0", "1515", "33"): "33 .. LegalOwnerOf",
        ("FIX.5.0", "1515", "34"): "34 .. OwnedByLegal",
        ("FIX.5.0", "1515", "35"): "35 .. BeneficialOwnerOf",
        ("FIX.5.0", "1515", "36"): "36 .. OwnedByBeneficial",
        ("FIX.5.0", "1515", "37"): "37 .. SettlesFor",
        ("FIX.5.0", "1515", "38"): "38 .. SettlesThrough",
        ("FIX.5.0", "1518", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1518", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1518", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1518", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1518", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1518", "J"): "J .. TaxID",
        ("FIX.5.0", "1518", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1518", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1518", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1518", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1518", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1518", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1518", "B"): "B .. BIC",
        ("FIX.5.0", "1518", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1518", "D"): "D .. Proprietary",
        ("FIX.5.0", "1518", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1518", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1518", "G"): "G .. MIC",
        ("FIX.5.0", "1518", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1518", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1518", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1518", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1518", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1518", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1518", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1518", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1521", "1"): "1 .. Firm",
        ("FIX.5.0", "1521", "2"): "2 .. Person",
        ("FIX.5.0", "1521", "3"): "3 .. System",
        ("FIX.5.0", "1521", "4"): "4 .. Application",
        ("FIX.5.0", "1521", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1521", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1521", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1521", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1521", "9"): "9 .. ContactName",
        ("FIX.5.0", "1521", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1521", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1521", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1521", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1521", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1521", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1521", "16"): "16 .. BIC",
        ("FIX.5.0", "1521", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1521", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1521", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1521", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1521", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1521", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1521", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1521", "24"): "24 .. Department",
        ("FIX.5.0", "1521", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1521", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1521", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1521", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1521", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1521", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1521", "31"): "31 .. Location",
        ("FIX.5.0", "1521", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1521", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1521", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1521", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1521", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1521", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1521", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1521", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1521", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1521", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1521", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1521", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1521", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1521", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1521", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1521", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1521", "48"): "48 .. USPerson",
        ("FIX.5.0", "1521", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1521", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1521", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1521", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1521", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1521", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1521", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1521", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1521", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1521", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1521", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1521", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1521", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1521", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1521", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1521", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1521", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1521", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1521", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1521", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1521", "69"): "69 .. EndUser",
        ("FIX.5.0", "1521", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1521", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1521", "72"): "72 .. Domicile",
        ("FIX.5.0", "1521", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1521", "74"): "74 .. Payer",
        ("FIX.5.0", "1521", "75"): "75 .. Receiver",
        ("FIX.5.0", "1521", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1521", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1521", "78"): "78 .. FirstName",
        ("FIX.5.0", "1521", "79"): "79 .. Surname",
        ("FIX.5.0", "1521", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1521", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1521", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1521", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1521", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1523", "0"): "0 .. Accepted",
        ("FIX.5.0", "1523", "1"): "1 .. Rejected",
        ("FIX.5.0", "1523", "2"): "2 .. Received",
        ("FIX.5.0", "1530", "0"): "0 .. CreditLimit",
        ("FIX.5.0", "1530", "1"): "1 .. GrossLimit",
        ("FIX.5.0", "1530", "2"): "2 .. NetLimit",
        ("FIX.5.0", "1530", "3"): "3 .. Exposure",
        ("FIX.5.0", "1530", "4"): "4 .. LongLimit",
        ("FIX.5.0", "1530", "5"): "5 .. ShortLimit",
        ("FIX.5.0", "1530", "6"): "6 .. CashMargin",
        ("FIX.5.0", "1530", "7"): "7 .. AdditionalMargin",
        ("FIX.5.0", "1530", "8"): "8 .. TotalMargin",
        ("FIX.5.0", "1530", "9"): "9 .. LimitConsumed",
        ("FIX.5.0", "1530", "10"): "10 .. ClipSize",
        ("FIX.5.0", "1530", "11"): "11 .. MaxNotionalOrderSize",
        ("FIX.5.0", "1530", "12"): "12 .. DV01PV01Limit",
        ("FIX.5.0", "1530", "13"): "13 .. CS01Limit",
        ("FIX.5.0", "1530", "14"): "14 .. VolumeLimitPerTimePeriod",
        ("FIX.5.0", "1530", "15"): "15 .. VolFilledPctOrdVolTmPeriod",
        ("FIX.5.0", "1530", "16"): "16 .. NotlFilledPctNotlTmPeriod",
        ("FIX.5.0", "1530", "17"): "17 .. TransactionExecutionLimitPerTimePeriod",
        ("FIX.5.0", "1535", "1"): "1 .. Include",
        ("FIX.5.0", "1535", "2"): "2 .. Exclude",
        ("FIX.5.0", "1539", "1"): "1 .. CUSIP",
        ("FIX.5.0", "1539", "2"): "2 .. SEDOL",
        ("FIX.5.0", "1539", "3"): "3 .. QUIK",
        ("FIX.5.0", "1539", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "1539", "5"): "5 .. RICCode",
        ("FIX.5.0", "1539", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "1539", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "1539", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "1539", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "1539", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "1539", "B"): "B .. Wertpapier",
        ("FIX.5.0", "1539", "C"): "C .. Dutch",
        ("FIX.5.0", "1539", "D"): "D .. Valoren",
        ("FIX.5.0", "1539", "E"): "E .. Sicovam",
        ("FIX.5.0", "1539", "F"): "F .. Belgian",
        ("FIX.5.0", "1539", "G"): "G .. Common",
        ("FIX.5.0", "1539", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "1539", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "1539", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "1539", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "1539", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "1539", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "1539", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "1539", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "1539", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "1539", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "1539", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "1539", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "1539", "U"): "U .. Synthetic",
        ("FIX.5.0", "1539", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "1539", "W"): "W .. IndexName",
        ("FIX.5.0", "1543", "1"): "1 .. AGENCY",
        ("FIX.5.0", "1543", "2"): "2 .. COMMODITY",
        ("FIX.5.0", "1543", "3"): "3 .. CORPORATE",
        ("FIX.5.0", "1543", "4"): "4 .. CURRENCY",
        ("FIX.5.0", "1543", "5"): "5 .. EQUITY",
        ("FIX.5.0", "1543", "6"): "6 .. GOVERNMENT",
        ("FIX.5.0", "1543", "7"): "7 .. INDEX",
        ("FIX.5.0", "1543", "8"): "8 .. LOAN",
        ("FIX.5.0", "1543", "9"): "9 .. MONEYMARKET",
        ("FIX.5.0", "1543", "10"): "10 .. MORTGAGE",
        ("FIX.5.0", "1543", "11"): "11 .. MUNICIPAL",
        ("FIX.5.0", "1543", "12"): "12 .. OTHER",
        ("FIX.5.0", "1543", "13"): "13 .. FINANCING",
        ("FIX.5.0", "1547", "EUSUPRA"): "EUSUPRA .. EuroSupranationalCoupons",
        ("FIX.5.0", "1547", "FAC"): "FAC .. FederalAgencyCoupon",
        ("FIX.5.0", "1547", "FADN"): "FADN .. FederalAgencyDiscountNote",
        ("FIX.5.0", "1547", "PEF"): "PEF .. PrivateExportFunding",
        ("FIX.5.0", "1547", "SUPRA"): "SUPRA .. USDSupranationalCoupons",
        ("FIX.5.0", "1547", "CORP"): "CORP .. CorporateBond",
        ("FIX.5.0", "1547", "CPP"): "CPP .. CorporatePrivatePlacement",
        ("FIX.5.0", "1547", "CB"): "CB .. ConvertibleBond",
        ("FIX.5.0", "1547", "DUAL"): "DUAL .. DualCurrency",
        ("FIX.5.0", "1547", "EUCORP"): "EUCORP .. EuroCorporateBond",
        ("FIX.5.0", "1547", "EUFRN"): "EUFRN .. EuroCorporateFloatingRateNotes",
        ("FIX.5.0", "1547", "FRN"): "FRN .. USCorporateFloatingRateNotes",
        ("FIX.5.0", "1547", "XLINKD"): "XLINKD .. IndexedLinked",
        ("FIX.5.0", "1547", "STRUCT"): "STRUCT .. StructuredNotes",
        ("FIX.5.0", "1547", "YANK"): "YANK .. YankeeCorporateBond",
        ("FIX.5.0", "1547", "FXNDF"): "FXNDF .. NonDeliverableForward",
        ("FIX.5.0", "1547", "FXSPOT"): "FXSPOT .. FXSpot",
        ("FIX.5.0", "1547", "FXFWD"): "FXFWD .. FXForward",
        ("FIX.5.0", "1547", "FXSWAP"): "FXSWAP .. FXSwap",
        ("FIX.5.0", "1547", "FXNDS"): "FXNDS .. NonDeliverableSwap",
        ("FIX.5.0", "1547", "CAP"): "CAP .. Cap",
        ("FIX.5.0", "1547", "CDS"): "CDS .. CreditDefaultSwap",
        ("FIX.5.0", "1547", "CLLR"): "CLLR .. Collar",
        ("FIX.5.0", "1547", "CMDTYSWAP"): "CMDTYSWAP .. CommoditySwap",
        ("FIX.5.0", "1547", "EXOTIC"): "EXOTIC .. Exotic",
        ("FIX.5.0", "1547", "OOC"): "OOC .. OptionsOnCombo",
        ("FIX.5.0", "1547", "FLR"): "FLR .. Floor",
        ("FIX.5.0", "1547", "FRA"): "FRA .. FRA",
        ("FIX.5.0", "1547", "FUT"): "FUT .. Future",
        ("FIX.5.0", "1547", "FWD"): "FWD .. DerivativeForward",
        ("FIX.5.0", "1547", "IRS"): "IRS .. InterestRateSwap",
        ("FIX.5.0", "1547", "TRS"): "TRS .. TotalReturnSwap",
        ("FIX.5.0", "1547", "LOANLEASE"): "LOANLEASE .. LoanLease",
        ("FIX.5.0", "1547", "OOF"): "OOF .. OptionsOnFutures",
        ("FIX.5.0", "1547", "OOP"): "OOP .. OptionsOnPhysical",
        ("FIX.5.0", "1547", "OPT"): "OPT .. Option",
        ("FIX.5.0", "1547", "SPOTFWD"): "SPOTFWD .. SpotForward",
        ("FIX.5.0", "1547", "SWAPTION"): "SWAPTION .. SwapOption",
        ("FIX.5.0", "1547", "XMISSION"): "XMISSION .. Transmission",
        ("FIX.5.0", "1547", "INDEX"): "INDEX .. Index",
        ("FIX.5.0", "1547", "BDBSKT"): "BDBSKT .. BondBasket",
        ("FIX.5.0", "1547", "CFD"): "CFD .. ContractForDifference",
        ("FIX.5.0", "1547", "CRLTNSWAP"): "CRLTNSWAP .. CorrelationSwap",
        ("FIX.5.0", "1547", "DVDNDSWAP"): "DVDNDSWAP .. DiviendSwap",
        ("FIX.5.0", "1547", "EQBSKT"): "EQBSKT .. EquityBasket",
        ("FIX.5.0", "1547", "EQFWD"): "EQFWD .. EquityForward",
        ("FIX.5.0", "1547", "RTRNSWAP"): "RTRNSWAP .. ReturnSwap",
        ("FIX.5.0", "1547", "VARSWAP"): "VARSWAP .. VarianceSwap",
        ("FIX.5.0", "1547", "PRTFLIOSWAP"): "PRTFLIOSWAP .. PortfolioSwaps",
        ("FIX.5.0", "1547", "FUTSWAP"): "FUTSWAP .. FuturesOnASwap",
        ("FIX.5.0", "1547", "FWDSWAP "): "FWDSWAP  .. ForwardsOnASwap",
        ("FIX.5.0", "1547", "FWDFRTAGMT "): "FWDFRTAGMT  .. ForwardFreightAgreement",
        ("FIX.5.0", "1547", "SPREADBET"): "SPREADBET .. SpreadBetting",
        ("FIX.5.0", "1547", "ETC"): "ETC .. ExchangeTradedCommodity",
        ("FIX.5.0", "1547", "CS"): "CS .. CommonStock",
        ("FIX.5.0", "1547", "PS"): "PS .. PreferredStock",
        ("FIX.5.0", "1547", "DR"): "DR .. DepositoryReceipts",
        ("FIX.5.0", "1547", "REPO"): "REPO .. Repurchase",
        ("FIX.5.0", "1547", "FORWARD"): "FORWARD .. Forward",
        ("FIX.5.0", "1547", "BUYSELL"): "BUYSELL .. BuySellback",
        ("FIX.5.0", "1547", "SECLOAN"): "SECLOAN .. SecuritiesLoan",
        ("FIX.5.0", "1547", "SECPLEDGE"): "SECPLEDGE .. SecuritiesPledge",
        ("FIX.5.0", "1547", "DVPLDG"): "DVPLDG .. DeliveryVersusPledge",
        ("FIX.5.0", "1547", "COLLBSKT"): "COLLBSKT .. CollateralBasket",
        ("FIX.5.0", "1547", "SFP"): "SFP .. StructuredFinanceProduct",
        ("FIX.5.0", "1547", "BRADY"): "BRADY .. BradyBond",
        ("FIX.5.0", "1547", "CAN"): "CAN .. CanadianTreasuryNotes",
        ("FIX.5.0", "1547", "CTB"): "CTB .. CanadianTreasuryBills",
        ("FIX.5.0", "1547", "EUSOV"): "EUSOV .. EuroSovereigns",
        ("FIX.5.0", "1547", "PROV"): "PROV .. CanadianProvincialBonds",
        ("FIX.5.0", "1547", "TB"): "TB .. TreasuryBill",
        ("FIX.5.0", "1547", "TBOND"): "TBOND .. USTreasuryBond",
        ("FIX.5.0", "1547", "TINT"): "TINT .. InterestStripFromAnyBondOrNote",
        ("FIX.5.0", "1547", "TBILL"): "TBILL .. USTreasuryBill",
        ("FIX.5.0", "1547", "TIPS"): "TIPS .. TreasuryInflationProtectedSecurities",
        ("FIX.5.0", "1547", "TCAL"): "TCAL .. PrincipalStripOfACallableBondOrNote",
        ("FIX.5.0", "1547", "TPRN"): "TPRN .. PrincipalStripFromANonCallableBondOrNote",
        ("FIX.5.0", "1547", "TNOTE"): "TNOTE .. USTreasuryNote",
        ("FIX.5.0", "1547", "TERM"): "TERM .. TermLoan",
        ("FIX.5.0", "1547", "RVLV"): "RVLV .. RevolverLoan",
        ("FIX.5.0", "1547", "RVLVTRM"): "RVLVTRM .. Revolver",
        ("FIX.5.0", "1547", "BRIDGE"): "BRIDGE .. BridgeLoan",
        ("FIX.5.0", "1547", "LOFC"): "LOFC .. LetterOfCredit",
        ("FIX.5.0", "1547", "SWING"): "SWING .. SwingLineFacility",
        ("FIX.5.0", "1547", "DINP"): "DINP .. DebtorInPossession",
        ("FIX.5.0", "1547", "DEFLTED"): "DEFLTED .. Defaulted",
        ("FIX.5.0", "1547", "WITHDRN"): "WITHDRN .. Withdrawn",
        ("FIX.5.0", "1547", "REPLACD"): "REPLACD .. Replaced",
        ("FIX.5.0", "1547", "MATURED"): "MATURED .. Matured",
        ("FIX.5.0", "1547", "AMENDED"): "AMENDED .. Amended",
        ("FIX.5.0", "1547", "RETIRED"): "RETIRED .. Retired",
        ("FIX.5.0", "1547", "BA"): "BA .. BankersAcceptance",
        ("FIX.5.0", "1547", "BDN"): "BDN .. BankDepositoryNote",
        ("FIX.5.0", "1547", "BN"): "BN .. BankNotes",
        ("FIX.5.0", "1547", "BOX"): "BOX .. BillOfExchanges",
        ("FIX.5.0", "1547", "CAMM"): "CAMM .. CanadianMoneyMarkets",
        ("FIX.5.0", "1547", "CD"): "CD .. CertificateOfDeposit",
        ("FIX.5.0", "1547", "CL"): "CL .. CallLoans",
        ("FIX.5.0", "1547", "CP"): "CP .. CommercialPaper",
        ("FIX.5.0", "1547", "DN"): "DN .. DepositNotes",
        ("FIX.5.0", "1547", "EUCD"): "EUCD .. EuroCertificateOfDeposit",
        ("FIX.5.0", "1547", "EUCP"): "EUCP .. EuroCommercialPaper",
        ("FIX.5.0", "1547", "LQN"): "LQN .. LiquidityNote",
        ("FIX.5.0", "1547", "MTN"): "MTN .. MediumTermNotes",
        ("FIX.5.0", "1547", "ONITE"): "ONITE .. Overnight",
        ("FIX.5.0", "1547", "PN"): "PN .. PromissoryNote",
        ("FIX.5.0", "1547", "STN"): "STN .. ShortTermLoanNote",
        ("FIX.5.0", "1547", "PZFJ"): "PZFJ .. PlazosFijos",
        ("FIX.5.0", "1547", "SLQN"): "SLQN .. SecuredLiquidityNote",
        ("FIX.5.0", "1547", "TD"): "TD .. TimeDeposit",
        ("FIX.5.0", "1547", "TLQN"): "TLQN .. TermLiquidityNote",
        ("FIX.5.0", "1547", "XCN"): "XCN .. ExtendedCommNote",
        ("FIX.5.0", "1547", "YCD"): "YCD .. YankeeCertificateOfDeposit",
        ("FIX.5.0", "1547", "ABS"): "ABS .. AssetBackedSecurities",
        ("FIX.5.0", "1547", "CMB"): "CMB .. CanadianMortgageBonds",
        ("FIX.5.0", "1547", "CMBS"): "CMBS .. Corp",
        ("FIX.5.0", "1547", "CMO"): "CMO .. CollateralizedMortgageObligation",
        ("FIX.5.0", "1547", "IET"): "IET .. IOETTEMortgage",
        ("FIX.5.0", "1547", "MBS"): "MBS .. MortgageBackedSecurities",
        ("FIX.5.0", "1547", "MIO"): "MIO .. MortgageInterestOnly",
        ("FIX.5.0", "1547", "MPO"): "MPO .. MortgagePrincipalOnly",
        ("FIX.5.0", "1547", "MPP"): "MPP .. MortgagePrivatePlacement",
        ("FIX.5.0", "1547", "MPT"): "MPT .. MiscellaneousPassThrough",
        ("FIX.5.0", "1547", "PFAND"): "PFAND .. Pfandbriefe",
        ("FIX.5.0", "1547", "TBA"): "TBA .. ToBeAnnounced",
        ("FIX.5.0", "1547", "AN"): "AN .. OtherAnticipationNotes",
        ("FIX.5.0", "1547", "COFO"): "COFO .. CertificateOfObligation",
        ("FIX.5.0", "1547", "COFP"): "COFP .. CertificateOfParticipation",
        ("FIX.5.0", "1547", "GO"): "GO .. GeneralObligationBonds",
        ("FIX.5.0", "1547", "MT"): "MT .. MandatoryTender",
        ("FIX.5.0", "1547", "RAN"): "RAN .. RevenueAnticipationNote",
        ("FIX.5.0", "1547", "REV"): "REV .. RevenueBonds",
        ("FIX.5.0", "1547", "SPCLA"): "SPCLA .. SpecialAssessment",
        ("FIX.5.0", "1547", "SPCLO"): "SPCLO .. SpecialObligation",
        ("FIX.5.0", "1547", "SPCLT"): "SPCLT .. SpecialTax",
        ("FIX.5.0", "1547", "TAN"): "TAN .. TaxAnticipationNote",
        ("FIX.5.0", "1547", "TAXA"): "TAXA .. TaxAllocation",
        ("FIX.5.0", "1547", "TECP"): "TECP .. TaxExemptCommercialPaper",
        ("FIX.5.0", "1547", "TMCP"): "TMCP .. TaxableMunicipalCP",
        ("FIX.5.0", "1547", "TRAN"): "TRAN .. TaxRevenueAnticipationNote",
        ("FIX.5.0", "1547", "VRDN"): "VRDN .. VariableRateDemandNote",
        ("FIX.5.0", "1547", "WAR"): "WAR .. Warrant",
        ("FIX.5.0", "1547", "MF"): "MF .. MutualFund",
        ("FIX.5.0", "1547", "MLEG"): "MLEG .. MultilegInstrument",
        ("FIX.5.0", "1547", "NONE"): "NONE .. NoSecurityType",
        ("FIX.5.0", "1547", "?"): "? .. Wildcard",
        ("FIX.5.0", "1547", "CASH"): "CASH .. Cash",
        ("FIX.5.0", "1547", "Other"): "Other .. Other",
        ("FIX.5.0", "1547", "ETN"): "ETN .. ExchangeTradedNote",
        ("FIX.5.0", "1547", "SECDERIV"): "SECDERIV .. SecuritizedDerivative",
        ("FIX.5.0", "1553", "0"): "0 .. Put",
        ("FIX.5.0", "1553", "1"): "1 .. Call",
        ("FIX.5.0", "1553", "2"): "2 .. Other",
        ("FIX.5.0", "1553", "3"): "3 .. Chooser",
        ("FIX.5.0", "1557", "0"): "0 .. Regular",
        ("FIX.5.0", "1557", "1"): "1 .. Cash",
        ("FIX.5.0", "1557", "2"): "2 .. NextDay",
        ("FIX.5.0", "1557", "3"): "3 .. TPlus2",
        ("FIX.5.0", "1557", "4"): "4 .. TPlus3",
        ("FIX.5.0", "1557", "5"): "5 .. TPlus4",
        ("FIX.5.0", "1557", "6"): "6 .. Future",
        ("FIX.5.0", "1557", "7"): "7 .. WhenAndIfIssued",
        ("FIX.5.0", "1557", "8"): "8 .. SellersOption",
        ("FIX.5.0", "1557", "9"): "9 .. TPlus5",
        ("FIX.5.0", "1557", "B"): "B .. BrokenDate",
        ("FIX.5.0", "1557", "C"): "C .. FXSpotNextSettlement",
        ("FIX.5.0", "1564", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1564", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1564", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1564", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1564", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1564", "J"): "J .. TaxID",
        ("FIX.5.0", "1564", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1564", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1564", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1564", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1564", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1564", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1564", "B"): "B .. BIC",
        ("FIX.5.0", "1564", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1564", "D"): "D .. Proprietary",
        ("FIX.5.0", "1564", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1564", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1564", "G"): "G .. MIC",
        ("FIX.5.0", "1564", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1564", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1564", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1564", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1564", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1564", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1564", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1564", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1565", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1565", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1565", "3"): "3 .. ClientID",
        ("FIX.5.0", "1565", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1565", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1565", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1565", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1565", "8"): "8 .. Locate",
        ("FIX.5.0", "1565", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1565", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1565", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1565", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1565", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1565", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1565", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1565", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1565", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1565", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1565", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1565", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1565", "22"): "22 .. Exchange",
        ("FIX.5.0", "1565", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1565", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1565", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1565", "27"): "27 .. Buyer",
        ("FIX.5.0", "1565", "28"): "28 .. Custodian",
        ("FIX.5.0", "1565", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1565", "30"): "30 .. Agent",
        ("FIX.5.0", "1565", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1565", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1565", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1565", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1565", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1565", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1565", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1565", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1565", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1565", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1565", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1565", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1565", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1565", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1565", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1565", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1565", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1565", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1565", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1565", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1565", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1565", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1565", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1565", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1565", "55"): "55 .. SessionID",
        ("FIX.5.0", "1565", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1565", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1565", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1565", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1565", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1565", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1565", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1565", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1565", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1565", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1565", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1565", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1565", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1565", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1565", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1565", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1565", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1565", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1565", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1565", "75"): "75 .. LocationID",
        ("FIX.5.0", "1565", "76"): "76 .. DeskID",
        ("FIX.5.0", "1565", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1565", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1565", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1565", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1565", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1565", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1565", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1565", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1565", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1565", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1565", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1565", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1565", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1565", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1565", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1565", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1565", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1565", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1565", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1565", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1565", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1565", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1565", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1565", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1565", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1565", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1565", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1565", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1565", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1565", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1565", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1565", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1565", "110"): "110 .. Borrower",
        ("FIX.5.0", "1565", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1565", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1565", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1565", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1565", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1565", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1565", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1565", "118"): "118 .. Operator",
        ("FIX.5.0", "1565", "119"): "119 .. CSD",
        ("FIX.5.0", "1565", "120"): "120 .. ICSD",
        ("FIX.5.0", "1565", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1565", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1565", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1568", "1"): "1 .. Firm",
        ("FIX.5.0", "1568", "2"): "2 .. Person",
        ("FIX.5.0", "1568", "3"): "3 .. System",
        ("FIX.5.0", "1568", "4"): "4 .. Application",
        ("FIX.5.0", "1568", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1568", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1568", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1568", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1568", "9"): "9 .. ContactName",
        ("FIX.5.0", "1568", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1568", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1568", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1568", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1568", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1568", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1568", "16"): "16 .. BIC",
        ("FIX.5.0", "1568", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1568", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1568", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1568", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1568", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1568", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1568", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1568", "24"): "24 .. Department",
        ("FIX.5.0", "1568", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1568", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1568", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1568", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1568", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1568", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1568", "31"): "31 .. Location",
        ("FIX.5.0", "1568", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1568", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1568", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1568", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1568", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1568", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1568", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1568", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1568", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1568", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1568", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1568", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1568", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1568", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1568", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1568", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1568", "48"): "48 .. USPerson",
        ("FIX.5.0", "1568", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1568", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1568", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1568", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1568", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1568", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1568", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1568", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1568", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1568", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1568", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1568", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1568", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1568", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1568", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1568", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1568", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1568", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1568", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1568", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1568", "69"): "69 .. EndUser",
        ("FIX.5.0", "1568", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1568", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1568", "72"): "72 .. Domicile",
        ("FIX.5.0", "1568", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1568", "74"): "74 .. Payer",
        ("FIX.5.0", "1568", "75"): "75 .. Receiver",
        ("FIX.5.0", "1568", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1568", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1568", "78"): "78 .. FirstName",
        ("FIX.5.0", "1568", "79"): "79 .. Surname",
        ("FIX.5.0", "1568", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1568", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1568", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1568", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1568", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1571", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1571", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1571", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1571", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1571", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1571", "J"): "J .. TaxID",
        ("FIX.5.0", "1571", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1571", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1571", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1571", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1571", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1571", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1571", "B"): "B .. BIC",
        ("FIX.5.0", "1571", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1571", "D"): "D .. Proprietary",
        ("FIX.5.0", "1571", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1571", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1571", "G"): "G .. MIC",
        ("FIX.5.0", "1571", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1571", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1571", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1571", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1571", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1571", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1571", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1571", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1574", "1"): "1 .. Firm",
        ("FIX.5.0", "1574", "2"): "2 .. Person",
        ("FIX.5.0", "1574", "3"): "3 .. System",
        ("FIX.5.0", "1574", "4"): "4 .. Application",
        ("FIX.5.0", "1574", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1574", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1574", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1574", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1574", "9"): "9 .. ContactName",
        ("FIX.5.0", "1574", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1574", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1574", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1574", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1574", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1574", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1574", "16"): "16 .. BIC",
        ("FIX.5.0", "1574", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1574", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1574", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1574", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1574", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1574", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1574", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1574", "24"): "24 .. Department",
        ("FIX.5.0", "1574", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1574", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1574", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1574", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1574", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1574", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1574", "31"): "31 .. Location",
        ("FIX.5.0", "1574", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1574", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1574", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1574", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1574", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1574", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1574", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1574", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1574", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1574", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1574", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1574", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1574", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1574", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1574", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1574", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1574", "48"): "48 .. USPerson",
        ("FIX.5.0", "1574", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1574", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1574", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1574", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1574", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1574", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1574", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1574", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1574", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1574", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1574", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1574", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1574", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1574", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1574", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1574", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1574", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1574", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1574", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1574", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1574", "69"): "69 .. EndUser",
        ("FIX.5.0", "1574", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1574", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1574", "72"): "72 .. Domicile",
        ("FIX.5.0", "1574", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1574", "74"): "74 .. Payer",
        ("FIX.5.0", "1574", "75"): "75 .. Receiver",
        ("FIX.5.0", "1574", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1574", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1574", "78"): "78 .. FirstName",
        ("FIX.5.0", "1574", "79"): "79 .. Surname",
        ("FIX.5.0", "1574", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1574", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1574", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1574", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1574", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1575", "AMTZ"): "AMTZ .. Amortizing",
        ("FIX.5.0", "1575", "CNST"): "CNST .. ConstantNotionalSchedule",
        ("FIX.5.0", "1575", "ACRT"): "ACRT .. AccretingNotionalSchedule",
        ("FIX.5.0", "1575", "CUST"): "CUST .. CustomNotionalSchedule",
        ("FIX.5.0", "1583", "0"): "0 .. Fee",
        ("FIX.5.0", "1583", "1"): "1 .. CreditControls",
        ("FIX.5.0", "1583", "2"): "2 .. Margin",
        ("FIX.5.0", "1583", "3"): "3 .. EntitlementOrEligibility",
        ("FIX.5.0", "1583", "4"): "4 .. MarketData",
        ("FIX.5.0", "1583", "5"): "5 .. AccountSelection",
        ("FIX.5.0", "1583", "6"): "6 .. DeliveryProcess",
        ("FIX.5.0", "1583", "7"): "7 .. Sector",
        ("FIX.5.0", "1585", "0"): "0 .. OptionsSettlement",
        ("FIX.5.0", "1585", "1"): "1 .. PendingErosionAdjustment",
        ("FIX.5.0", "1585", "2"): "2 .. FinalErosionAdjustment",
        ("FIX.5.0", "1585", "3"): "3 .. TearUpCouponAmount",
        ("FIX.5.0", "1585", "4"): "4 .. PriceAlignmentInterest",
        ("FIX.5.0", "1585", "5"): "5 .. DeliveryInvoiceCharges",
        ("FIX.5.0", "1585", "6"): "6 .. DeliveryStorageCharges",
        ("FIX.5.0", "1588", "CASH"): "CASH .. CashAmount",
        ("FIX.5.0", "1588", "CRES"): "CRES .. CashResidualAmount",
        ("FIX.5.0", "1588", "FMTM"): "FMTM .. FinalMarkToMarketAmount",
        ("FIX.5.0", "1588", "IMTM"): "IMTM .. IncrementalMarkToMarketAmount",
        ("FIX.5.0", "1588", "PREM"): "PREM .. PremiumAmount",
        ("FIX.5.0", "1588", "SMTM"): "SMTM .. StartOfDayMarkToMarketAmount",
        ("FIX.5.0", "1588", "TVAR"): "TVAR .. TradeVariationAmount",
        ("FIX.5.0", "1588", "VADJ"): "VADJ .. ValueAdjustedAmount",
        ("FIX.5.0", "1588", "SETL"): "SETL .. SettlementValue",
        ("FIX.5.0", "1588", "ICPN"): "ICPN .. InitialTradeCouponAmount",
        ("FIX.5.0", "1588", "ACPN"): "ACPN .. AccruedCouponAmount",
        ("FIX.5.0", "1588", "CPN"): "CPN .. CouponAmount",
        ("FIX.5.0", "1588", "IACPN"): "IACPN .. IncrementalAccruedCoupon",
        ("FIX.5.0", "1588", "CMTM"): "CMTM .. CollateralizedMarkToMarket",
        ("FIX.5.0", "1588", "ICMTM"): "ICMTM .. IncrementalCollateralizedMarkToMarket",
        ("FIX.5.0", "1588", "DLV"): "DLV .. CompensationAmount",
        ("FIX.5.0", "1588", "BANK"): "BANK .. TotalBankedAmount",
        ("FIX.5.0", "1588", "COLAT"): "COLAT .. TotalCollateralizedAmount",
        ("FIX.5.0", "1588", "LSNV"): "LSNV .. LongPairedSwapNotionalValue",
        ("FIX.5.0", "1588", "SSNV"): "SSNV .. ShortPairedSwapNotionalValue",
        ("FIX.5.0", "1588", "SACPN"): "SACPN .. StartOfDayAccruedCoupon",
        ("FIX.5.0", "1588", "NPV"): "NPV .. NetPresentValue",
        ("FIX.5.0", "1588", "SNPV"): "SNPV .. StartOfDayNetPresentValue",
        ("FIX.5.0", "1588", "NCF"): "NCF .. NetCashFlow",
        ("FIX.5.0", "1588", "PVFEES"): "PVFEES .. PresentValueOfFees",
        ("FIX.5.0", "1588", "PV01"): "PV01 .. PresentValueOneBasisPoints",
        ("FIX.5.0", "1588", "5YREN"): "5YREN .. FiveYearEquivalentNotional",
        ("FIX.5.0", "1588", "UMTM"): "UMTM .. UndiscountedMarkToMarket",
        ("FIX.5.0", "1588", "MTD"): "MTD .. MarkToModel",
        ("FIX.5.0", "1588", "VMTM"): "VMTM .. MarkToMarketVariance",
        ("FIX.5.0", "1588", "VMTD"): "VMTD .. MarkToModelVariance",
        ("FIX.5.0", "1588", "UPFRNT"): "UPFRNT .. UpfrontPayment",
        ("FIX.5.0", "1590", "0"): "0 .. OptionsSettlement",
        ("FIX.5.0", "1590", "1"): "1 .. PendingErosionAdjustment",
        ("FIX.5.0", "1590", "2"): "2 .. FinalErosionAdjustment",
        ("FIX.5.0", "1590", "3"): "3 .. TearUpCouponAmount",
        ("FIX.5.0", "1590", "4"): "4 .. PriceAlignmentInterest",
        ("FIX.5.0", "1590", "5"): "5 .. DeliveryInvoiceCharges",
        ("FIX.5.0", "1590", "6"): "6 .. DeliveryStorageCharges",
        ("FIX.5.0", "1591", "0"): "0 .. Units",
        ("FIX.5.0", "1591", "1"): "1 .. Contracts",
        ("FIX.5.0", "1591", "2"): "2 .. UnitsOfMeasurePerTimeUnit",
        ("FIX.5.0", "1598", "0"): "0 .. TradeClearingAtExecutionPrice",
        ("FIX.5.0", "1598", "1"): "1 .. TradeClearingAtAlternateClearingPrice",
        ("FIX.5.0", "1607", "1"): "1 .. InvalidInstrumentRequested",
        ("FIX.5.0", "1607", "2"): "2 .. InstrumentAlreadyExists",
        ("FIX.5.0", "1607", "3"): "3 .. RequestTypeNotSupported",
        ("FIX.5.0", "1607", "4"): "4 .. SystemUnavailableForInstrumentCreation",
        ("FIX.5.0", "1607", "5"): "5 .. IneligibleInstrumentGroup",
        ("FIX.5.0", "1607", "6"): "6 .. InstrumentIDUnavailable",
        ("FIX.5.0", "1607", "7"): "7 .. InvalidOrMissingDataOnOptionLeg",
        ("FIX.5.0", "1607", "8"): "8 .. InvalidOrMissingDataOnFutureLeg",
        ("FIX.5.0", "1607", "10"): "10 .. InvalidOrMissingDataOnFXLeg",
        ("FIX.5.0", "1607", "11"): "11 .. InvalidLegPriceSpecified",
        ("FIX.5.0", "1607", "12"): "12 .. InvalidInstrumentStructureSpecified",
        ("FIX.5.0", "1609", "0"): "0 .. ThrottleLimitNotExceededNotQueued",
        ("FIX.5.0", "1609", "1"): "1 .. QueuedDueToThrottleLimitExceeded",
        ("FIX.5.0", "1611", "0"): "0 .. QueueInbound",
        ("FIX.5.0", "1611", "1"): "1 .. QueueOutbound",
        ("FIX.5.0", "1611", "2"): "2 .. Reject",
        ("FIX.5.0", "1611", "3"): "3 .. Disconnect",
        ("FIX.5.0", "1611", "4"): "4 .. Warning",
        ("FIX.5.0", "1612", "0"): "0 .. InboundRate",
        ("FIX.5.0", "1612", "1"): "1 .. OutstandingRequests",
        ("FIX.5.0", "1615", "0"): "0 .. Seconds",
        ("FIX.5.0", "1615", "1"): "1 .. TenthsOfASecond",
        ("FIX.5.0", "1615", "2"): "2 .. HundredthsOfASecond",
        ("FIX.5.0", "1615", "3"): "3 .. Milliseconds",
        ("FIX.5.0", "1615", "4"): "4 .. Microseconds",
        ("FIX.5.0", "1615", "5"): "5 .. Nanoseconds",
        ("FIX.5.0", "1615", "10"): "10 .. Minutes",
        ("FIX.5.0", "1615", "11"): "11 .. Hours",
        ("FIX.5.0", "1615", "12"): "12 .. Days",
        ("FIX.5.0", "1615", "13"): "13 .. Weeks",
        ("FIX.5.0", "1615", "14"): "14 .. Months",
        ("FIX.5.0", "1615", "15"): "15 .. Years",
        ("FIX.5.0", "1617", "1"): "1 .. Assignment",
        ("FIX.5.0", "1617", "2"): "2 .. Rejected",
        ("FIX.5.0", "1617", "3"): "3 .. Terminate",
        ("FIX.5.0", "1619", "0"): "0 .. Heartbeat",
        ("FIX.5.0", "1619", "1"): "1 .. TestRequest",
        ("FIX.5.0", "1619", "2"): "2 .. ResendRequest",
        ("FIX.5.0", "1619", "3"): "3 .. Reject",
        ("FIX.5.0", "1619", "4"): "4 .. SequenceReset",
        ("FIX.5.0", "1619", "5"): "5 .. Logout",
        ("FIX.5.0", "1619", "6"): "6 .. IOI",
        ("FIX.5.0", "1619", "7"): "7 .. Advertisement",
        ("FIX.5.0", "1619", "8"): "8 .. ExecutionReport",
        ("FIX.5.0", "1619", "9"): "9 .. OrderCancelReject",
        ("FIX.5.0", "1619", "A"): "A .. Logon",
        ("FIX.5.0", "1619", "B"): "B .. News",
        ("FIX.5.0", "1619", "C"): "C .. Email",
        ("FIX.5.0", "1619", "D"): "D .. NewOrderSingle",
        ("FIX.5.0", "1619", "E"): "E .. NewOrderList",
        ("FIX.5.0", "1619", "F"): "F .. OrderCancelRequest",
        ("FIX.5.0", "1619", "G"): "G .. OrderCancelReplaceRequest",
        ("FIX.5.0", "1619", "H"): "H .. OrderStatusRequest",
        ("FIX.5.0", "1619", "J"): "J .. AllocationInstruction",
        ("FIX.5.0", "1619", "K"): "K .. ListCancelRequest",
        ("FIX.5.0", "1619", "L"): "L .. ListExecute",
        ("FIX.5.0", "1619", "M"): "M .. ListStatusRequest",
        ("FIX.5.0", "1619", "N"): "N .. ListStatus",
        ("FIX.5.0", "1619", "P"): "P .. AllocationInstructionAck",
        ("FIX.5.0", "1619", "Q"): "Q .. DontKnowTrade",
        ("FIX.5.0", "1619", "R"): "R .. QuoteRequest",
        ("FIX.5.0", "1619", "S"): "S .. Quote",
        ("FIX.5.0", "1619", "T"): "T .. SettlementInstructions",
        ("FIX.5.0", "1619", "V"): "V .. MarketDataRequest",
        ("FIX.5.0", "1619", "W"): "W .. MarketDataSnapshotFullRefresh",
        ("FIX.5.0", "1619", "X"): "X .. MarketDataIncrementalRefresh",
        ("FIX.5.0", "1619", "Y"): "Y .. MarketDataRequestReject",
        ("FIX.5.0", "1619", "Z"): "Z .. QuoteCancel",
        ("FIX.5.0", "1619", "a"): "a .. QuoteStatusRequest",
        ("FIX.5.0", "1619", "b"): "b .. MassQuoteAck",
        ("FIX.5.0", "1619", "c"): "c .. SecurityDefinitionRequest",
        ("FIX.5.0", "1619", "d"): "d .. SecurityDefinition",
        ("FIX.5.0", "1619", "e"): "e .. SecurityStatusRequest",
        ("FIX.5.0", "1619", "f"): "f .. SecurityStatus",
        ("FIX.5.0", "1619", "g"): "g .. TradingSessionStatusRequest",
        ("FIX.5.0", "1619", "h"): "h .. TradingSessionStatus",
        ("FIX.5.0", "1619", "i"): "i .. MassQuote",
        ("FIX.5.0", "1619", "j"): "j .. BusinessMessageReject",
        ("FIX.5.0", "1619", "k"): "k .. BidRequest",
        ("FIX.5.0", "1619", "l"): "l .. BidResponse",
        ("FIX.5.0", "1619", "m"): "m .. ListStrikePrice",
        ("FIX.5.0", "1619", "n"): "n .. XMLnonFIX",
        ("FIX.5.0", "1619", "o"): "o .. RegistrationInstructions",
        ("FIX.5.0", "1619", "p"): "p .. RegistrationInstructionsResponse",
        ("FIX.5.0", "1619", "q"): "q .. OrderMassCancelRequest",
        ("FIX.5.0", "1619", "r"): "r .. OrderMassCancelReport",
        ("FIX.5.0", "1619", "s"): "s .. NewOrderCross",
        ("FIX.5.0", "1619", "t"): "t .. CrossOrderCancelReplaceRequest",
        ("FIX.5.0", "1619", "u"): "u .. CrossOrderCancelRequest",
        ("FIX.5.0", "1619", "v"): "v .. SecurityTypeRequest",
        ("FIX.5.0", "1619", "w"): "w .. SecurityTypes",
        ("FIX.5.0", "1619", "x"): "x .. SecurityListRequest",
        ("FIX.5.0", "1619", "y"): "y .. SecurityList",
        ("FIX.5.0", "1619", "z"): "z .. DerivativeSecurityListRequest",
        ("FIX.5.0", "1619", "AA"): "AA .. DerivativeSecurityList",
        ("FIX.5.0", "1619", "AB"): "AB .. NewOrderMultileg",
        ("FIX.5.0", "1619", "AC"): "AC .. MultilegOrderCancelReplace",
        ("FIX.5.0", "1619", "AD"): "AD .. TradeCaptureReportRequest",
        ("FIX.5.0", "1619", "AE"): "AE .. TradeCaptureReport",
        ("FIX.5.0", "1619", "AF"): "AF .. OrderMassStatusRequest",
        ("FIX.5.0", "1619", "AG"): "AG .. QuoteRequestReject",
        ("FIX.5.0", "1619", "AH"): "AH .. RFQRequest",
        ("FIX.5.0", "1619", "AI"): "AI .. QuoteStatusReport",
        ("FIX.5.0", "1619", "AJ"): "AJ .. QuoteResponse",
        ("FIX.5.0", "1619", "AK"): "AK .. Confirmation",
        ("FIX.5.0", "1619", "AL"): "AL .. PositionMaintenanceRequest",
        ("FIX.5.0", "1619", "AM"): "AM .. PositionMaintenanceReport",
        ("FIX.5.0", "1619", "AN"): "AN .. RequestForPositions",
        ("FIX.5.0", "1619", "AO"): "AO .. RequestForPositionsAck",
        ("FIX.5.0", "1619", "AP"): "AP .. PositionReport",
        ("FIX.5.0", "1619", "AQ"): "AQ .. TradeCaptureReportRequestAck",
        ("FIX.5.0", "1619", "AR"): "AR .. TradeCaptureReportAck",
        ("FIX.5.0", "1619", "AS"): "AS .. AllocationReport",
        ("FIX.5.0", "1619", "AT"): "AT .. AllocationReportAck",
        ("FIX.5.0", "1619", "AU"): "AU .. ConfirmationAck",
        ("FIX.5.0", "1619", "AV"): "AV .. SettlementInstructionRequest",
        ("FIX.5.0", "1619", "AW"): "AW .. AssignmentReport",
        ("FIX.5.0", "1619", "AX"): "AX .. CollateralRequest",
        ("FIX.5.0", "1619", "AY"): "AY .. CollateralAssignment",
        ("FIX.5.0", "1619", "AZ"): "AZ .. CollateralResponse",
        ("FIX.5.0", "1619", "BA"): "BA .. CollateralReport",
        ("FIX.5.0", "1619", "BB"): "BB .. CollateralInquiry",
        ("FIX.5.0", "1619", "BC"): "BC .. NetworkCounterpartySystemStatusRequest",
        ("FIX.5.0", "1619", "BD"): "BD .. NetworkCounterpartySystemStatusResponse",
        ("FIX.5.0", "1619", "BE"): "BE .. UserRequest",
        ("FIX.5.0", "1619", "BF"): "BF .. UserResponse",
        ("FIX.5.0", "1619", "BG"): "BG .. CollateralInquiryAck",
        ("FIX.5.0", "1619", "BH"): "BH .. ConfirmationRequest",
        ("FIX.5.0", "1619", "BO"): "BO .. ContraryIntentionReport",
        ("FIX.5.0", "1619", "BP"): "BP .. SecurityDefinitionUpdateReport",
        ("FIX.5.0", "1619", "BK"): "BK .. SecurityListUpdateReport",
        ("FIX.5.0", "1619", "BL"): "BL .. AdjustedPositionReport",
        ("FIX.5.0", "1619", "BM"): "BM .. AllocationInstructionAlert",
        ("FIX.5.0", "1619", "BN"): "BN .. ExecutionAck",
        ("FIX.5.0", "1619", "BJ"): "BJ .. TradingSessionList",
        ("FIX.5.0", "1619", "BI"): "BI .. TradingSessionListRequest",
        ("FIX.5.0", "1619", "BQ"): "BQ .. SettlementObligationReport",
        ("FIX.5.0", "1619", "BR"): "BR .. DerivativeSecurityListUpdateReport",
        ("FIX.5.0", "1619", "BS"): "BS .. TradingSessionListUpdateReport",
        ("FIX.5.0", "1619", "BT"): "BT .. MarketDefinitionRequest",
        ("FIX.5.0", "1619", "BU"): "BU .. MarketDefinition",
        ("FIX.5.0", "1619", "BV"): "BV .. MarketDefinitionUpdateReport",
        ("FIX.5.0", "1619", "BW"): "BW .. ApplicationMessageRequest",
        ("FIX.5.0", "1619", "BX"): "BX .. ApplicationMessageRequestAck",
        ("FIX.5.0", "1619", "BY"): "BY .. ApplicationMessageReport",
        ("FIX.5.0", "1619", "BZ"): "BZ .. OrderMassActionReport",
        ("FIX.5.0", "1619", "CA"): "CA .. OrderMassActionRequest",
        ("FIX.5.0", "1619", "CB"): "CB .. UserNotification",
        ("FIX.5.0", "1619", "CC"): "CC .. StreamAssignmentRequest",
        ("FIX.5.0", "1619", "CD"): "CD .. StreamAssignmentReport",
        ("FIX.5.0", "1619", "CE"): "CE .. StreamAssignmentReportACK",
        ("FIX.5.0", "1619", "CF"): "CF .. PartyDetailsListRequest",
        ("FIX.5.0", "1619", "CG"): "CG .. PartyDetailsListReport",
        ("FIX.5.0", "1619", "CH"): "CH .. MarginRequirementInquiry",
        ("FIX.5.0", "1619", "CI"): "CI .. MarginRequirementInquiryAck",
        ("FIX.5.0", "1619", "CJ"): "CJ .. MarginRequirementReport",
        ("FIX.5.0", "1619", "CK"): "CK .. PartyDetailsListUpdateReport",
        ("FIX.5.0", "1619", "CL"): "CL .. PartyRiskLimitsRequest",
        ("FIX.5.0", "1619", "CM"): "CM .. PartyRiskLimitsReport",
        ("FIX.5.0", "1619", "CN"): "CN .. SecurityMassStatusRequest",
        ("FIX.5.0", "1619", "CO"): "CO .. SecurityMassStatus",
        ("FIX.5.0", "1619", "CQ"): "CQ .. AccountSummaryReport",
        ("FIX.5.0", "1619", "CR"): "CR .. PartyRiskLimitsUpdateReport",
        ("FIX.5.0", "1619", "CS"): "CS .. PartyRiskLimitsDefinitionRequest",
        ("FIX.5.0", "1619", "CT"): "CT .. PartyRiskLimitsDefinitionRequestAck",
        ("FIX.5.0", "1619", "CU"): "CU .. PartyEntitlementsRequest",
        ("FIX.5.0", "1619", "CV"): "CV .. PartyEntitlementsReport",
        ("FIX.5.0", "1619", "CW"): "CW .. QuoteAck",
        ("FIX.5.0", "1619", "CX"): "CX .. PartyDetailsDefinitionRequest",
        ("FIX.5.0", "1619", "CY"): "CY .. PartyDetailsDefinitionRequestAck",
        ("FIX.5.0", "1619", "CZ"): "CZ .. PartyEntitlementsUpdateReport",
        ("FIX.5.0", "1619", "DA"): "DA .. PartyEntitlementsDefinitionRequest",
        ("FIX.5.0", "1619", "DB"): "DB .. PartyEntitlementsDefinitionRequestAck",
        ("FIX.5.0", "1619", "DC"): "DC .. TradeMatchReport",
        ("FIX.5.0", "1619", "DD"): "DD .. TradeMatchReportAck",
        ("FIX.5.0", "1619", "DE"): "DE .. PartyRiskLimitsReportAck",
        ("FIX.5.0", "1619", "DF"): "DF .. PartyRiskLimitCheckRequest",
        ("FIX.5.0", "1619", "DG"): "DG .. PartyRiskLimitCheckRequestAck",
        ("FIX.5.0", "1619", "DH"): "DH .. PartyActionRequest",
        ("FIX.5.0", "1619", "DI"): "DI .. PartyActionReport",
        ("FIX.5.0", "1619", "DJ"): "DJ .. MassOrder",
        ("FIX.5.0", "1619", "DK"): "DK .. MassOrderAck",
        ("FIX.5.0", "1619", "DL"): "DL .. PositionTransferInstruction",
        ("FIX.5.0", "1619", "DM"): "DM .. PositionTransferInstructionAck",
        ("FIX.5.0", "1619", "DN"): "DN .. PositionTransferReport",
        ("FIX.5.0", "1619", "DO"): "DO .. MarketDataStatisticsRequest",
        ("FIX.5.0", "1619", "DP"): "DP .. MarketDataStatisticsReport",
        ("FIX.5.0", "1619", "DQ"): "DQ .. CollateralReportAck",
        ("FIX.5.0", "1619", "DR"): "DR .. MarketDataReport",
        ("FIX.5.0", "1619", "DS"): "DS .. CrossRequest",
        ("FIX.5.0", "1619", "DT"): "DT .. CrossRequestAck",
        ("FIX.5.0", "1619", "DU"): "DU .. AllocationInstructionAlertRequest",
        ("FIX.5.0", "1625", "1"): "1 .. Match",
        ("FIX.5.0", "1625", "2"): "2 .. DoNotMatch",
        ("FIX.5.0", "1628", "0"): "0 .. ThisOrder",
        ("FIX.5.0", "1628", "1"): "1 .. OtherOrder",
        ("FIX.5.0", "1628", "2"): "2 .. AllOtherOrdersForGivenSecurity",
        ("FIX.5.0", "1628", "3"): "3 .. AllOtherOrdersForGivenSecurityAndPrice",
        ("FIX.5.0", "1628", "4"): "4 .. AllOtherOrdersForGivenSecurityAndSide",
        ("FIX.5.0", "1628", "5"): "5 .. AllOtherOrdersForGivenSecurityPriceAndSide",
        ("FIX.5.0", "1631", "0"): "0 .. CreditLimit",
        ("FIX.5.0", "1631", "1"): "1 .. GrossPositionLimit",
        ("FIX.5.0", "1631", "2"): "2 .. NetPositionLimit",
        ("FIX.5.0", "1631", "3"): "3 .. RiskExposureLimit",
        ("FIX.5.0", "1631", "4"): "4 .. LongPositionLimit",
        ("FIX.5.0", "1631", "5"): "5 .. ShortPositionLimit",
        ("FIX.5.0", "1637", "0"): "0 .. Summary",
        ("FIX.5.0", "1637", "1"): "1 .. Detail",
        ("FIX.5.0", "1637", "2"): "2 .. ExcessDeficit",
        ("FIX.5.0", "1637", "3"): "3 .. NetPosition",
        ("FIX.5.0", "1638", "0"): "0 .. Summary",
        ("FIX.5.0", "1638", "1"): "1 .. Detail",
        ("FIX.5.0", "1638", "2"): "2 .. ExcessDeficit",
        ("FIX.5.0", "1640", "0"): "0 .. Accepted",
        ("FIX.5.0", "1640", "1"): "1 .. AcceptedWithWarnings",
        ("FIX.5.0", "1640", "2"): "2 .. Completed",
        ("FIX.5.0", "1640", "3"): "3 .. CompletedWithWarnings",
        ("FIX.5.0", "1640", "4"): "4 .. Rejected",
        ("FIX.5.0", "1641", "0"): "0 .. Successful",
        ("FIX.5.0", "1641", "1"): "1 .. InvalidOrUnknownInstrument",
        ("FIX.5.0", "1641", "2"): "2 .. InvalidOrUnknownMarginClass",
        ("FIX.5.0", "1641", "3"): "3 .. InvalidParties",
        ("FIX.5.0", "1641", "4"): "4 .. InvalidTransportTypeReq",
        ("FIX.5.0", "1641", "5"): "5 .. InvalidDestinationReq",
        ("FIX.5.0", "1641", "6"): "6 .. NoMarginReqFound",
        ("FIX.5.0", "1641", "7"): "7 .. MarginReqInquiryQualifierNotSupported",
        ("FIX.5.0", "1641", "8"): "8 .. UnauthorizedForMarginReqInquiry",
        ("FIX.5.0", "1641", "99"): "99 .. Other",
        ("FIX.5.0", "1644", "1"): "1 .. AdditionalMargin",
        ("FIX.5.0", "1644", "2"): "2 .. AdjustedMargin",
        ("FIX.5.0", "1644", "3"): "3 .. UnadjustedMargin",
        ("FIX.5.0", "1644", "4"): "4 .. BinaryAddOnAmount",
        ("FIX.5.0", "1644", "5"): "5 .. CashBalanceAmount",
        ("FIX.5.0", "1644", "6"): "6 .. ConcentrationMargin",
        ("FIX.5.0", "1644", "7"): "7 .. CoreMargin",
        ("FIX.5.0", "1644", "8"): "8 .. DeliveryMargin",
        ("FIX.5.0", "1644", "9"): "9 .. DiscretionaryMargin",
        ("FIX.5.0", "1644", "10"): "10 .. FuturesSpreadMargin",
        ("FIX.5.0", "1644", "11"): "11 .. InitialMargin",
        ("FIX.5.0", "1644", "12"): "12 .. LiquidatingMargin",
        ("FIX.5.0", "1644", "13"): "13 .. MarginCallAmount",
        ("FIX.5.0", "1644", "14"): "14 .. MarginDeficitAmount",
        ("FIX.5.0", "1644", "15"): "15 .. MarginExcessAmount",
        ("FIX.5.0", "1644", "16"): "16 .. OptionPremiumAmount",
        ("FIX.5.0", "1644", "17"): "17 .. PremiumMargin",
        ("FIX.5.0", "1644", "18"): "18 .. ReserveMargin",
        ("FIX.5.0", "1644", "19"): "19 .. SecurityCollateralAmount",
        ("FIX.5.0", "1644", "20"): "20 .. StressTestAddOnAmount",
        ("FIX.5.0", "1644", "21"): "21 .. SuperMargin",
        ("FIX.5.0", "1644", "22"): "22 .. TotalMargin",
        ("FIX.5.0", "1644", "23"): "23 .. VariationMargin",
        ("FIX.5.0", "1644", "24"): "24 .. SecondaryVariationMargin",
        ("FIX.5.0", "1644", "25"): "25 .. RolledUpMarginDeficit",
        ("FIX.5.0", "1644", "26"): "26 .. SpreadResponseMargin",
        ("FIX.5.0", "1644", "27"): "27 .. SystemicRiskMargin",
        ("FIX.5.0", "1644", "28"): "28 .. CurveRiskMargin",
        ("FIX.5.0", "1644", "29"): "29 .. IndexSpreadRiskMargin",
        ("FIX.5.0", "1644", "30"): "30 .. SectorRiskMargin",
        ("FIX.5.0", "1644", "31"): "31 .. JumpToDefaultRiskMargin",
        ("FIX.5.0", "1644", "32"): "32 .. BasisRiskMargin",
        ("FIX.5.0", "1644", "33"): "33 .. InterestRateRiskMargin",
        ("FIX.5.0", "1644", "34"): "34 .. JumpToHealthRiskMargin",
        ("FIX.5.0", "1644", "35"): "35 .. OtherRiskMargin",
        ("FIX.5.0", "1648", "1"): "1 .. HedgesForInstrument",
        ("FIX.5.0", "1648", "2"): "2 .. Underlier",
        ("FIX.5.0", "1648", "3"): "3 .. EquityEquivalent",
        ("FIX.5.0", "1648", "4"): "4 .. NearestExchangeTradedContract",
        ("FIX.5.0", "1648", "5"): "5 .. RetailEquivalent",
        ("FIX.5.0", "1648", "6"): "6 .. Leg",
        ("FIX.5.0", "1651", "1"): "1 .. CUSIP",
        ("FIX.5.0", "1651", "2"): "2 .. SEDOL",
        ("FIX.5.0", "1651", "3"): "3 .. QUIK",
        ("FIX.5.0", "1651", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "1651", "5"): "5 .. RICCode",
        ("FIX.5.0", "1651", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "1651", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "1651", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "1651", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "1651", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "1651", "B"): "B .. Wertpapier",
        ("FIX.5.0", "1651", "C"): "C .. Dutch",
        ("FIX.5.0", "1651", "D"): "D .. Valoren",
        ("FIX.5.0", "1651", "E"): "E .. Sicovam",
        ("FIX.5.0", "1651", "F"): "F .. Belgian",
        ("FIX.5.0", "1651", "G"): "G .. Common",
        ("FIX.5.0", "1651", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "1651", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "1651", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "1651", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "1651", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "1651", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "1651", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "1651", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "1651", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "1651", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "1651", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "1651", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "1651", "U"): "U .. Synthetic",
        ("FIX.5.0", "1651", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "1651", "W"): "W .. IndexName",
        ("FIX.5.0", "1655", "0"): "0 .. NoParticipation",
        ("FIX.5.0", "1655", "1"): "1 .. BuyParticipation",
        ("FIX.5.0", "1655", "2"): "2 .. SellParticipation",
        ("FIX.5.0", "1655", "3"): "3 .. BothBuyAndSellParticipation",
        ("FIX.5.0", "1659", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1659", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1659", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1659", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1659", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1659", "J"): "J .. TaxID",
        ("FIX.5.0", "1659", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1659", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1659", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1659", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1659", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1659", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1659", "B"): "B .. BIC",
        ("FIX.5.0", "1659", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1659", "D"): "D .. Proprietary",
        ("FIX.5.0", "1659", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1659", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1659", "G"): "G .. MIC",
        ("FIX.5.0", "1659", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1659", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1659", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1659", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1659", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1659", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1659", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1659", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1660", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1660", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1660", "3"): "3 .. ClientID",
        ("FIX.5.0", "1660", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1660", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1660", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1660", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1660", "8"): "8 .. Locate",
        ("FIX.5.0", "1660", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1660", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1660", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1660", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1660", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1660", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1660", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1660", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1660", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1660", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1660", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1660", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1660", "22"): "22 .. Exchange",
        ("FIX.5.0", "1660", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1660", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1660", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1660", "27"): "27 .. Buyer",
        ("FIX.5.0", "1660", "28"): "28 .. Custodian",
        ("FIX.5.0", "1660", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1660", "30"): "30 .. Agent",
        ("FIX.5.0", "1660", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1660", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1660", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1660", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1660", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1660", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1660", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1660", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1660", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1660", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1660", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1660", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1660", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1660", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1660", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1660", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1660", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1660", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1660", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1660", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1660", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1660", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1660", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1660", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1660", "55"): "55 .. SessionID",
        ("FIX.5.0", "1660", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1660", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1660", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1660", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1660", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1660", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1660", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1660", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1660", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1660", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1660", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1660", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1660", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1660", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1660", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1660", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1660", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1660", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1660", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1660", "75"): "75 .. LocationID",
        ("FIX.5.0", "1660", "76"): "76 .. DeskID",
        ("FIX.5.0", "1660", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1660", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1660", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1660", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1660", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1660", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1660", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1660", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1660", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1660", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1660", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1660", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1660", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1660", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1660", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1660", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1660", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1660", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1660", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1660", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1660", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1660", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1660", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1660", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1660", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1660", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1660", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1660", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1660", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1660", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1660", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1660", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1660", "110"): "110 .. Borrower",
        ("FIX.5.0", "1660", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1660", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1660", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1660", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1660", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1660", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1660", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1660", "118"): "118 .. Operator",
        ("FIX.5.0", "1660", "119"): "119 .. CSD",
        ("FIX.5.0", "1660", "120"): "120 .. ICSD",
        ("FIX.5.0", "1660", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1660", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1660", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1663", "1"): "1 .. Firm",
        ("FIX.5.0", "1663", "2"): "2 .. Person",
        ("FIX.5.0", "1663", "3"): "3 .. System",
        ("FIX.5.0", "1663", "4"): "4 .. Application",
        ("FIX.5.0", "1663", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1663", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1663", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1663", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1663", "9"): "9 .. ContactName",
        ("FIX.5.0", "1663", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1663", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1663", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1663", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1663", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1663", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1663", "16"): "16 .. BIC",
        ("FIX.5.0", "1663", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1663", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1663", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1663", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1663", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1663", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1663", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1663", "24"): "24 .. Department",
        ("FIX.5.0", "1663", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1663", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1663", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1663", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1663", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1663", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1663", "31"): "31 .. Location",
        ("FIX.5.0", "1663", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1663", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1663", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1663", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1663", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1663", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1663", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1663", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1663", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1663", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1663", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1663", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1663", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1663", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1663", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1663", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1663", "48"): "48 .. USPerson",
        ("FIX.5.0", "1663", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1663", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1663", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1663", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1663", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1663", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1663", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1663", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1663", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1663", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1663", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1663", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1663", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1663", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1663", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1663", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1663", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1663", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1663", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1663", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1663", "69"): "69 .. EndUser",
        ("FIX.5.0", "1663", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1663", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1663", "72"): "72 .. Domicile",
        ("FIX.5.0", "1663", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1663", "74"): "74 .. Payer",
        ("FIX.5.0", "1663", "75"): "75 .. Receiver",
        ("FIX.5.0", "1663", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1663", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1663", "78"): "78 .. FirstName",
        ("FIX.5.0", "1663", "79"): "79 .. Surname",
        ("FIX.5.0", "1663", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1663", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1663", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1663", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1663", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1672", "0"): "0 .. Active",
        ("FIX.5.0", "1672", "1"): "1 .. Suspended",
        ("FIX.5.0", "1672", "2"): "2 .. Halted",
        ("FIX.5.0", "1674", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "1674", "18"): "18 .. Current",
        ("FIX.5.0", "1674", "19"): "19 .. New",
        ("FIX.5.0", "1674", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "1674", "0"): "0 .. Agency",
        ("FIX.5.0", "1674", "1"): "1 .. Principal",
        ("FIX.5.0", "1674", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "1674", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "1674", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "1674", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "1674", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "1674", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "1674", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "1674", "27"): "27 .. Supervisor",
        ("FIX.5.0", "1674", "22"): "22 .. Algorithm",
        ("FIX.5.0", "1674", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "1674", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "1674", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "1674", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "1674", "7"): "7 .. Bank",
        ("FIX.5.0", "1674", "8"): "8 .. Hub",
        ("FIX.5.0", "1674", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "1674", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "1674", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "1674", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "1674", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "1674", "21"): "21 .. Specialist",
        ("FIX.5.0", "1674", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "1675", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "1675", "18"): "18 .. Current",
        ("FIX.5.0", "1675", "19"): "19 .. New",
        ("FIX.5.0", "1675", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "1675", "0"): "0 .. Agency",
        ("FIX.5.0", "1675", "1"): "1 .. Principal",
        ("FIX.5.0", "1675", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "1675", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "1675", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "1675", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "1675", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "1675", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "1675", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "1675", "27"): "27 .. Supervisor",
        ("FIX.5.0", "1675", "22"): "22 .. Algorithm",
        ("FIX.5.0", "1675", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "1675", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "1675", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "1675", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "1675", "7"): "7 .. Bank",
        ("FIX.5.0", "1675", "8"): "8 .. Hub",
        ("FIX.5.0", "1675", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "1675", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "1675", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "1675", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "1675", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "1675", "21"): "21 .. Specialist",
        ("FIX.5.0", "1675", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "1679", "1"): "1 .. OpeningDelay",
        ("FIX.5.0", "1679", "2"): "2 .. TradingHalt",
        ("FIX.5.0", "1679", "3"): "3 .. Resume",
        ("FIX.5.0", "1679", "4"): "4 .. NoOpen",
        ("FIX.5.0", "1679", "5"): "5 .. PriceIndication",
        ("FIX.5.0", "1679", "6"): "6 .. TradingRangeIndication",
        ("FIX.5.0", "1679", "7"): "7 .. MarketImbalanceBuy",
        ("FIX.5.0", "1679", "8"): "8 .. MarketImbalanceSell",
        ("FIX.5.0", "1679", "9"): "9 .. MarketOnCloseImbalanceBuy",
        ("FIX.5.0", "1679", "10"): "10 .. MarketOnCloseImbalanceSell",
        ("FIX.5.0", "1679", "12"): "12 .. NoMarketImbalance",
        ("FIX.5.0", "1679", "13"): "13 .. NoMarketOnCloseImbalance",
        ("FIX.5.0", "1679", "14"): "14 .. ITSPreOpening",
        ("FIX.5.0", "1679", "15"): "15 .. NewPriceIndication",
        ("FIX.5.0", "1679", "16"): "16 .. TradeDisseminationTime",
        ("FIX.5.0", "1679", "17"): "17 .. ReadyToTrade",
        ("FIX.5.0", "1679", "18"): "18 .. NotAvailableForTrading",
        ("FIX.5.0", "1679", "19"): "19 .. NotTradedOnThisMarket",
        ("FIX.5.0", "1679", "20"): "20 .. UnknownOrInvalid",
        ("FIX.5.0", "1679", "21"): "21 .. PreOpen",
        ("FIX.5.0", "1679", "22"): "22 .. OpeningRotation",
        ("FIX.5.0", "1679", "23"): "23 .. FastMarket",
        ("FIX.5.0", "1679", "24"): "24 .. PreCross",
        ("FIX.5.0", "1679", "25"): "25 .. Cross",
        ("FIX.5.0", "1679", "26"): "26 .. PostClose",
        ("FIX.5.0", "1679", "27"): "27 .. NoCancel",
        ("FIX.5.0", "1680", "1"): "1 .. OrderImbalance",
        ("FIX.5.0", "1680", "2"): "2 .. TradingResumes",
        ("FIX.5.0", "1680", "3"): "3 .. PriceVolatilityInterruption",
        ("FIX.5.0", "1680", "4"): "4 .. ChangeOfTradingSession",
        ("FIX.5.0", "1680", "5"): "5 .. ChangeOfTradingSubsession",
        ("FIX.5.0", "1680", "6"): "6 .. ChangeOfSecurityTradingStatus",
        ("FIX.5.0", "1680", "7"): "7 .. ChangeOfBookType",
        ("FIX.5.0", "1680", "8"): "8 .. ChangeOfMarketDepth",
        ("FIX.5.0", "1680", "9"): "9 .. CorporateAction",
        ("FIX.5.0", "1681", "0"): "0 .. NewsDissemination",
        ("FIX.5.0", "1681", "1"): "1 .. OrderInflux",
        ("FIX.5.0", "1681", "2"): "2 .. OrderImbalance",
        ("FIX.5.0", "1681", "3"): "3 .. AdditionalInformation",
        ("FIX.5.0", "1681", "4"): "4 .. NewsPending",
        ("FIX.5.0", "1681", "5"): "5 .. EquipmentChangeover",
        ("FIX.5.0", "1682", "1"): "1 .. OpeningDelay",
        ("FIX.5.0", "1682", "2"): "2 .. TradingHalt",
        ("FIX.5.0", "1682", "3"): "3 .. Resume",
        ("FIX.5.0", "1682", "4"): "4 .. NoOpen",
        ("FIX.5.0", "1682", "5"): "5 .. PriceIndication",
        ("FIX.5.0", "1682", "6"): "6 .. TradingRangeIndication",
        ("FIX.5.0", "1682", "7"): "7 .. MarketImbalanceBuy",
        ("FIX.5.0", "1682", "8"): "8 .. MarketImbalanceSell",
        ("FIX.5.0", "1682", "9"): "9 .. MarketOnCloseImbalanceBuy",
        ("FIX.5.0", "1682", "10"): "10 .. MarketOnCloseImbalanceSell",
        ("FIX.5.0", "1682", "12"): "12 .. NoMarketImbalance",
        ("FIX.5.0", "1682", "13"): "13 .. NoMarketOnCloseImbalance",
        ("FIX.5.0", "1682", "14"): "14 .. ITSPreOpening",
        ("FIX.5.0", "1682", "15"): "15 .. NewPriceIndication",
        ("FIX.5.0", "1682", "16"): "16 .. TradeDisseminationTime",
        ("FIX.5.0", "1682", "17"): "17 .. ReadyToTrade",
        ("FIX.5.0", "1682", "18"): "18 .. NotAvailableForTrading",
        ("FIX.5.0", "1682", "19"): "19 .. NotTradedOnThisMarket",
        ("FIX.5.0", "1682", "20"): "20 .. UnknownOrInvalid",
        ("FIX.5.0", "1682", "21"): "21 .. PreOpen",
        ("FIX.5.0", "1682", "22"): "22 .. OpeningRotation",
        ("FIX.5.0", "1682", "23"): "23 .. FastMarket",
        ("FIX.5.0", "1682", "24"): "24 .. PreCross",
        ("FIX.5.0", "1682", "25"): "25 .. Cross",
        ("FIX.5.0", "1682", "26"): "26 .. PostClose",
        ("FIX.5.0", "1682", "27"): "27 .. NoCancel",
        ("FIX.5.0", "1684", "0"): "0 .. NewsDissemination",
        ("FIX.5.0", "1684", "1"): "1 .. OrderInflux",
        ("FIX.5.0", "1684", "2"): "2 .. OrderImbalance",
        ("FIX.5.0", "1684", "3"): "3 .. AdditionalInformation",
        ("FIX.5.0", "1684", "4"): "4 .. NewsPending",
        ("FIX.5.0", "1684", "5"): "5 .. EquipmentChangeover",
        ("FIX.5.0", "1685", "0"): "0 .. RejectIfThrottleLimitExceeded",
        ("FIX.5.0", "1685", "1"): "1 .. QueueIfThrottleLimitExceeded",
        ("FIX.5.0", "1686", "0"): "0 .. OutstandingRequestsUnchanged",
        ("FIX.5.0", "1686", "1"): "1 .. OutstandingRequestsDecreased",
        ("FIX.5.0", "1687", "0"): "0 .. NoRestrictions",
        ("FIX.5.0", "1687", "1"): "1 .. SecurityNotShortable",
        ("FIX.5.0", "1687", "2"): "2 .. SecurityNotShortableAtOrBelowBestBid",
        ("FIX.5.0", "1687", "3"): "3 .. SecurityNotShortableWithoutPreBorrow",
        ("FIX.5.0", "1688", "0"): "0 .. ExemptionReasonUnknown",
        ("FIX.5.0", "1688", "1"): "1 .. IncomingSSE",
        ("FIX.5.0", "1688", "2"): "2 .. AboveNationalBestBid",
        ("FIX.5.0", "1688", "3"): "3 .. DelayedDelivery",
        ("FIX.5.0", "1688", "4"): "4 .. OddLot",
        ("FIX.5.0", "1688", "5"): "5 .. DomesticArbitrage",
        ("FIX.5.0", "1688", "6"): "6 .. InternationalArbitrage",
        ("FIX.5.0", "1688", "7"): "7 .. UnderwriterOrSyndicateDistribution",
        ("FIX.5.0", "1688", "8"): "8 .. RisklessPrincipal",
        ("FIX.5.0", "1688", "9"): "9 .. VWAP",
        ("FIX.5.0", "1689", "0"): "0 .. ExemptionReasonUnknown",
        ("FIX.5.0", "1689", "1"): "1 .. IncomingSSE",
        ("FIX.5.0", "1689", "2"): "2 .. AboveNationalBestBid",
        ("FIX.5.0", "1689", "3"): "3 .. DelayedDelivery",
        ("FIX.5.0", "1689", "4"): "4 .. OddLot",
        ("FIX.5.0", "1689", "5"): "5 .. DomesticArbitrage",
        ("FIX.5.0", "1689", "6"): "6 .. InternationalArbitrage",
        ("FIX.5.0", "1689", "7"): "7 .. UnderwriterOrSyndicateDistribution",
        ("FIX.5.0", "1689", "8"): "8 .. RisklessPrincipal",
        ("FIX.5.0", "1689", "9"): "9 .. VWAP",
        ("FIX.5.0", "1690", "0"): "0 .. ExemptionReasonUnknown",
        ("FIX.5.0", "1690", "1"): "1 .. IncomingSSE",
        ("FIX.5.0", "1690", "2"): "2 .. AboveNationalBestBid",
        ("FIX.5.0", "1690", "3"): "3 .. DelayedDelivery",
        ("FIX.5.0", "1690", "4"): "4 .. OddLot",
        ("FIX.5.0", "1690", "5"): "5 .. DomesticArbitrage",
        ("FIX.5.0", "1690", "6"): "6 .. InternationalArbitrage",
        ("FIX.5.0", "1690", "7"): "7 .. UnderwriterOrSyndicateDistribution",
        ("FIX.5.0", "1690", "8"): "8 .. RisklessPrincipal",
        ("FIX.5.0", "1690", "9"): "9 .. VWAP",
        ("FIX.5.0", "1692", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "1692", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "1692", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "1692", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "1692", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "1692", "J"): "J .. TaxID",
        ("FIX.5.0", "1692", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "1692", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "1692", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "1692", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "1692", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "1692", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "1692", "B"): "B .. BIC",
        ("FIX.5.0", "1692", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "1692", "D"): "D .. Proprietary",
        ("FIX.5.0", "1692", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "1692", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "1692", "G"): "G .. MIC",
        ("FIX.5.0", "1692", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "1692", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "1692", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "1692", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "1692", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "1692", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "1692", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "1692", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "1693", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "1693", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "1693", "3"): "3 .. ClientID",
        ("FIX.5.0", "1693", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "1693", "5"): "5 .. InvestorID",
        ("FIX.5.0", "1693", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "1693", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "1693", "8"): "8 .. Locate",
        ("FIX.5.0", "1693", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "1693", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "1693", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "1693", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "1693", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "1693", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "1693", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "1693", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "1693", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "1693", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "1693", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "1693", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "1693", "22"): "22 .. Exchange",
        ("FIX.5.0", "1693", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "1693", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "1693", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "1693", "27"): "27 .. Buyer",
        ("FIX.5.0", "1693", "28"): "28 .. Custodian",
        ("FIX.5.0", "1693", "29"): "29 .. Intermediary",
        ("FIX.5.0", "1693", "30"): "30 .. Agent",
        ("FIX.5.0", "1693", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "1693", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "1693", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "1693", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "1693", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "1693", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "1693", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "1693", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "1693", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "1693", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "1693", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "1693", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "1693", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "1693", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "1693", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "1693", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "1693", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "1693", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "1693", "49"): "49 .. AssetManager",
        ("FIX.5.0", "1693", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "1693", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "1693", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "1693", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "1693", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "1693", "55"): "55 .. SessionID",
        ("FIX.5.0", "1693", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "1693", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "1693", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "1693", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "1693", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "1693", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "1693", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "1693", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "1693", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "1693", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "1693", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "1693", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "1693", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "1693", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "1693", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "1693", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "1693", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "1693", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "1693", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "1693", "75"): "75 .. LocationID",
        ("FIX.5.0", "1693", "76"): "76 .. DeskID",
        ("FIX.5.0", "1693", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "1693", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "1693", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "1693", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "1693", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "1693", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "1693", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "1693", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "1693", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "1693", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "1693", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "1693", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "1693", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "1693", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "1693", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "1693", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "1693", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "1693", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "1693", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "1693", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "1693", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "1693", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "1693", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "1693", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "1693", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "1693", "102"): "102 .. DataRepository",
        ("FIX.5.0", "1693", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "1693", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "1693", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "1693", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "1693", "107"): "107 .. Correspondent",
        ("FIX.5.0", "1693", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "1693", "110"): "110 .. Borrower",
        ("FIX.5.0", "1693", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "1693", "112"): "112 .. Guarantor",
        ("FIX.5.0", "1693", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "1693", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "1693", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "1693", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "1693", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "1693", "118"): "118 .. Operator",
        ("FIX.5.0", "1693", "119"): "119 .. CSD",
        ("FIX.5.0", "1693", "120"): "120 .. ICSD",
        ("FIX.5.0", "1693", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "1693", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "1693", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "1696", "1"): "1 .. Firm",
        ("FIX.5.0", "1696", "2"): "2 .. Person",
        ("FIX.5.0", "1696", "3"): "3 .. System",
        ("FIX.5.0", "1696", "4"): "4 .. Application",
        ("FIX.5.0", "1696", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "1696", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "1696", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "1696", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "1696", "9"): "9 .. ContactName",
        ("FIX.5.0", "1696", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "1696", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "1696", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "1696", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "1696", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "1696", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "1696", "16"): "16 .. BIC",
        ("FIX.5.0", "1696", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "1696", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "1696", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "1696", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "1696", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "1696", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "1696", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "1696", "24"): "24 .. Department",
        ("FIX.5.0", "1696", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "1696", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "1696", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "1696", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "1696", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "1696", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "1696", "31"): "31 .. Location",
        ("FIX.5.0", "1696", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "1696", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "1696", "34"): "34 .. AddressCity",
        ("FIX.5.0", "1696", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "1696", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "1696", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "1696", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "1696", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "1696", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "1696", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "1696", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "1696", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "1696", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "1696", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "1696", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "1696", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "1696", "48"): "48 .. USPerson",
        ("FIX.5.0", "1696", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "1696", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "1696", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "1696", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "1696", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "1696", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "1696", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "1696", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "1696", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "1696", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "1696", "59"): "59 .. FCMCode",
        ("FIX.5.0", "1696", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "1696", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "1696", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "1696", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "1696", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "1696", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "1696", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "1696", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "1696", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "1696", "69"): "69 .. EndUser",
        ("FIX.5.0", "1696", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "1696", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "1696", "72"): "72 .. Domicile",
        ("FIX.5.0", "1696", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "1696", "74"): "74 .. Payer",
        ("FIX.5.0", "1696", "75"): "75 .. Receiver",
        ("FIX.5.0", "1696", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "1696", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "1696", "78"): "78 .. FirstName",
        ("FIX.5.0", "1696", "79"): "79 .. Surname",
        ("FIX.5.0", "1696", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "1696", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "1696", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "1696", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "1696", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "1698", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "1698", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "1698", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "1698", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "1698", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "1698", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "1698", "thm"): "thm .. therms",
        ("FIX.5.0", "1698", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "1698", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "1698", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "1698", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "1698", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "1698", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "1698", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "1698", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "1698", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "1698", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "1698", "day"): "day .. Day",
        ("FIX.5.0", "1698", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "1698", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "1698", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "1698", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "1698", "g"): "g .. Grams",
        ("FIX.5.0", "1698", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "1698", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "1698", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "1698", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "1698", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "1698", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "1698", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "1698", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "1698", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "1698", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "1698", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "1698", "L"): "L .. liters",
        ("FIX.5.0", "1698", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "1698", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "1698", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "1698", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "1698", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "1698", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "1698", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "1698", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "1698", "t"): "t .. MetricTons",
        ("FIX.5.0", "1698", "tn"): "tn .. Tons",
        ("FIX.5.0", "1724", "1"): "1 .. OrderReceivedFromCustomer",
        ("FIX.5.0", "1724", "2"): "2 .. OrderReceivedFromWithinFirm",
        ("FIX.5.0", "1724", "3"): "3 .. OrderReceivedFromAnotherBrokerDealer",
        ("FIX.5.0", "1724", "4"): "4 .. OrderReceivedFromCustomerOrWithFirm",
        ("FIX.5.0", "1724", "5"): "5 .. OrderReceivedFromDirectAccessCustomer",
        ("FIX.5.0", "1735", "0"): "0 .. Rollup",
        ("FIX.5.0", "1735", "1"): "1 .. DoNotRollUp",
        ("FIX.5.0", "1738", "0"): "0 .. Completed",
        ("FIX.5.0", "1738", "1"): "1 .. Refused",
        ("FIX.5.0", "1738", "2"): "2 .. Cancelled",
        ("FIX.5.0", "1739", "0"): "0 .. Bond",
        ("FIX.5.0", "1739", "1"): "1 .. ConvertBond",
        ("FIX.5.0", "1739", "2"): "2 .. Mortgage",
        ("FIX.5.0", "1739", "3"): "3 .. Loan",
        ("FIX.5.0", "1740", "0"): "0 .. PercentPar",
        ("FIX.5.0", "1740", "1"): "1 .. DealSpread",
        ("FIX.5.0", "1740", "2"): "2 .. UpfrontPnts",
        ("FIX.5.0", "1740", "3"): "3 .. UpfrontAmt",
        ("FIX.5.0", "1740", "4"): "4 .. ParUpfrontAmt",
        ("FIX.5.0", "1740", "5"): "5 .. SpreadUpfrontAmt",
        ("FIX.5.0", "1740", "6"): "6 .. UpfrontPntsAmt",
        ("FIX.5.0", "1741", "1"): "1 .. Percentage",
        ("FIX.5.0", "1741", "3"): "3 .. FixedAmount",
        ("FIX.5.0", "1744", "0"): "0 .. NoApplRecoveryNeeded",
        ("FIX.5.0", "1744", "1"): "1 .. ApplRecoveryNeeded",
        ("FIX.5.0", "1760", "1"): "1 .. Definitions",
        ("FIX.5.0", "1760", "2"): "2 .. Utilization",
        ("FIX.5.0", "1760", "3"): "3 .. DefinitionsAndUtilizations",
        ("FIX.5.0", "1761", "0"): "0 .. Successful",
        ("FIX.5.0", "1761", "1"): "1 .. InvalidParty",
        ("FIX.5.0", "1761", "2"): "2 .. InvalidRelatedParty",
        ("FIX.5.0", "1761", "3"): "3 .. InvalidRiskLimitType",
        ("FIX.5.0", "1761", "4"): "4 .. InvalidRiskLimitID",
        ("FIX.5.0", "1761", "5"): "5 .. InvalidRiskLimitAmount",
        ("FIX.5.0", "1761", "6"): "6 .. InvalidRiskWarningLevelAction",
        ("FIX.5.0", "1761", "7"): "7 .. InvalidRiskInstrumentScope",
        ("FIX.5.0", "1761", "8"): "8 .. RiskLimitActionsNotSupported",
        ("FIX.5.0", "1761", "9"): "9 .. WarningLevelsNotSupported",
        ("FIX.5.0", "1761", "10"): "10 .. WarningLevelActionsNotSupported",
        ("FIX.5.0", "1761", "11"): "11 .. RiskInstrumentScopeNotSupported",
        ("FIX.5.0", "1761", "12"): "12 .. RiskLimitNotApprovedForParty",
        ("FIX.5.0", "1761", "13"): "13 .. RiskLimitAlreadyDefinedForParty",
        ("FIX.5.0", "1761", "14"): "14 .. InstrumentNotApprovedForParty",
        ("FIX.5.0", "1761", "98"): "98 .. NotAuthorized",
        ("FIX.5.0", "1761", "99"): "99 .. Other",
        ("FIX.5.0", "1762", "0"): "0 .. Accepted",
        ("FIX.5.0", "1762", "1"): "1 .. AcceptedWithChanges",
        ("FIX.5.0", "1762", "2"): "2 .. Rejected",
        ("FIX.5.0", "1762", "3"): "3 .. AcceptancePending",
        ("FIX.5.0", "1763", "0"): "0 .. Accepted",
        ("FIX.5.0", "1763", "1"): "1 .. AcceptedWithChanges",
        ("FIX.5.0", "1763", "2"): "2 .. Rejected",
        ("FIX.5.0", "1764", "0"): "0 .. Successful",
        ("FIX.5.0", "1764", "1"): "1 .. InvalidParty",
        ("FIX.5.0", "1764", "2"): "2 .. InvalidRelatedParty",
        ("FIX.5.0", "1764", "3"): "3 .. InvalidRiskLimitType",
        ("FIX.5.0", "1764", "4"): "4 .. InvalidRiskLimitID",
        ("FIX.5.0", "1764", "5"): "5 .. InvalidRiskLimitAmount",
        ("FIX.5.0", "1764", "6"): "6 .. InvalidRiskWarningLevelAction",
        ("FIX.5.0", "1764", "7"): "7 .. InvalidRiskInstrumentScope",
        ("FIX.5.0", "1764", "8"): "8 .. RiskLimitActionsNotSupported",
        ("FIX.5.0", "1764", "9"): "9 .. WarningLevelsNotSupported",
        ("FIX.5.0", "1764", "10"): "10 .. WarningLevelActionsNotSupported",
        ("FIX.5.0", "1764", "11"): "11 .. RiskInstrumentScopeNotSupported",
        ("FIX.5.0", "1764", "12"): "12 .. RiskLimitNotApprovedForParty",
        ("FIX.5.0", "1764", "13"): "13 .. RiskLimitAlreadyDefinedForParty",
        ("FIX.5.0", "1764", "14"): "14 .. InstrumentNotApprovedForParty",
        ("FIX.5.0", "1764", "98"): "98 .. NotAuthorized",
        ("FIX.5.0", "1764", "99"): "99 .. Other",
        ("FIX.5.0", "1767", "0"): "0 .. QueueInbound",
        ("FIX.5.0", "1767", "1"): "1 .. QueueOutbound",
        ("FIX.5.0", "1767", "2"): "2 .. Reject",
        ("FIX.5.0", "1767", "3"): "3 .. Disconnect",
        ("FIX.5.0", "1767", "4"): "4 .. Warning",
        ("FIX.5.0", "1767", "5"): "5 .. PingCreditCheckWithRevalidation",
        ("FIX.5.0", "1767", "6"): "6 .. PingCreditCheckNoRevalidation",
        ("FIX.5.0", "1767", "7"): "7 .. PushCreditCheckWithRevalidation",
        ("FIX.5.0", "1767", "8"): "8 .. PushCreditCheckNoRevalidation",
        ("FIX.5.0", "1767", "9"): "9 .. Suspend",
        ("FIX.5.0", "1767", "10"): "10 .. HaltTrading",
        ("FIX.5.0", "1769", "0"): "0 .. QueueInbound",
        ("FIX.5.0", "1769", "1"): "1 .. QueueOutbound",
        ("FIX.5.0", "1769", "2"): "2 .. Reject",
        ("FIX.5.0", "1769", "3"): "3 .. Disconnect",
        ("FIX.5.0", "1769", "4"): "4 .. Warning",
        ("FIX.5.0", "1769", "5"): "5 .. PingCreditCheckWithRevalidation",
        ("FIX.5.0", "1769", "6"): "6 .. PingCreditCheckNoRevalidation",
        ("FIX.5.0", "1769", "7"): "7 .. PushCreditCheckWithRevalidation",
        ("FIX.5.0", "1769", "8"): "8 .. PushCreditCheckNoRevalidation",
        ("FIX.5.0", "1769", "9"): "9 .. Suspend",
        ("FIX.5.0", "1769", "10"): "10 .. HaltTrading",
        ("FIX.5.0", "1775", "0"): "0 .. Trade",
        ("FIX.5.0", "1775", "1"): "1 .. MakeMarkets",
        ("FIX.5.0", "1775", "2"): "2 .. HoldPositions",
        ("FIX.5.0", "1775", "3"): "3 .. PerformGiveUps",
        ("FIX.5.0", "1775", "4"): "4 .. SubmitIOIs",
        ("FIX.5.0", "1775", "5"): "5 .. SubscribeMarketData",
        ("FIX.5.0", "1775", "6"): "6 .. ShortWithPreBorrow",
        ("FIX.5.0", "1775", "7"): "7 .. SubmitQuoteRequests",
        ("FIX.5.0", "1775", "8"): "8 .. RespondToQuoteRequests",
        ("FIX.5.0", "1779", "29"): "29 .. Tenor",
        ("FIX.5.0", "1779", "32"): "32 .. Pattern",
        ("FIX.5.0", "1779", "33"): "33 .. Reserved100Plus",
        ("FIX.5.0", "1779", "34"): "34 .. Reserved1000Plus",
        ("FIX.5.0", "1779", "35"): "35 .. Reserved4000Plus",
        ("FIX.5.0", "1779", "14"): "14 .. String",
        ("FIX.5.0", "1779", "15"): "15 .. MultipleCharValue",
        ("FIX.5.0", "1779", "16"): "16 .. Currency",
        ("FIX.5.0", "1779", "17"): "17 .. Exchange",
        ("FIX.5.0", "1779", "18"): "18 .. MonthYear",
        ("FIX.5.0", "1779", "19"): "19 .. UTCTimestamp",
        ("FIX.5.0", "1779", "20"): "20 .. UTCTimeOnly",
        ("FIX.5.0", "1779", "21"): "21 .. LocalMktDate",
        ("FIX.5.0", "1779", "22"): "22 .. UTCDateOnly",
        ("FIX.5.0", "1779", "23"): "23 .. data",
        ("FIX.5.0", "1779", "24"): "24 .. MultipleStringValue",
        ("FIX.5.0", "1779", "25"): "25 .. Country",
        ("FIX.5.0", "1779", "26"): "26 .. Language",
        ("FIX.5.0", "1779", "27"): "27 .. TZTimeOnly",
        ("FIX.5.0", "1779", "28"): "28 .. TZTimestamp",
        ("FIX.5.0", "1779", "31"): "31 .. XMLData",
        ("FIX.5.0", "1779", "12"): "12 .. char",
        ("FIX.5.0", "1779", "13"): "13 .. Boolean",
        ("FIX.5.0", "1779", "6"): "6 .. float",
        ("FIX.5.0", "1779", "7"): "7 .. Qty",
        ("FIX.5.0", "1779", "8"): "8 .. Price",
        ("FIX.5.0", "1779", "9"): "9 .. PriceOffset",
        ("FIX.5.0", "1779", "10"): "10 .. Amt",
        ("FIX.5.0", "1779", "11"): "11 .. Percentage",
        ("FIX.5.0", "1779", "1"): "1 .. int",
        ("FIX.5.0", "1779", "2"): "2 .. Length",
        ("FIX.5.0", "1779", "3"): "3 .. NumInGroup",
        ("FIX.5.0", "1779", "4"): "4 .. SeqNum",
        ("FIX.5.0", "1779", "5"): "5 .. TagNum",
        ("FIX.5.0", "1779", "30"): "30 .. DayOfMonth",
        ("FIX.5.0", "1785", "0"): "0 .. Automatic",
        ("FIX.5.0", "1785", "1"): "1 .. Manual",
        ("FIX.5.0", "1786", "0"): "0 .. NumberOfUnits",
        ("FIX.5.0", "1786", "1"): "1 .. NumberOfRoundLots",
        ("FIX.5.0", "1796", "1"): "1 .. Added",
        ("FIX.5.0", "1796", "2"): "2 .. Modified",
        ("FIX.5.0", "1796", "3"): "3 .. Deleted",
        ("FIX.5.0", "1796", "4"): "4 .. PartiallyFilled",
        ("FIX.5.0", "1796", "5"): "5 .. Filled",
        ("FIX.5.0", "1796", "6"): "6 .. Suspended",
        ("FIX.5.0", "1796", "7"): "7 .. Released",
        ("FIX.5.0", "1796", "8"): "8 .. Restated",
        ("FIX.5.0", "1796", "9"): "9 .. Locked",
        ("FIX.5.0", "1796", "10"): "10 .. Triggered",
        ("FIX.5.0", "1796", "11"): "11 .. Activated",
        ("FIX.5.0", "1798", "1"): "1 .. AddOrderRequest",
        ("FIX.5.0", "1798", "2"): "2 .. ModifyOrderRequest",
        ("FIX.5.0", "1798", "3"): "3 .. DeleteOrderRequest",
        ("FIX.5.0", "1798", "4"): "4 .. OrderEnteredOOB",
        ("FIX.5.0", "1798", "5"): "5 .. OrderModifiedOOB",
        ("FIX.5.0", "1798", "6"): "6 .. OrderDeletedOOB",
        ("FIX.5.0", "1798", "7"): "7 .. OrderActivatedOrTriggered",
        ("FIX.5.0", "1798", "8"): "8 .. OrderExpired",
        ("FIX.5.0", "1798", "9"): "9 .. ReserveOrderRefreshed",
        ("FIX.5.0", "1798", "10"): "10 .. AwayMarketBetter",
        ("FIX.5.0", "1798", "11"): "11 .. CorporateAction",
        ("FIX.5.0", "1798", "12"): "12 .. StartOfDay",
        ("FIX.5.0", "1798", "13"): "13 .. EndOfDay",
        ("FIX.5.0", "1801", "1"): "1 .. AddedLiquidity",
        ("FIX.5.0", "1801", "2"): "2 .. RemovedLiquidity",
        ("FIX.5.0", "1801", "3"): "3 .. LiquidityRoutedOut",
        ("FIX.5.0", "1801", "4"): "4 .. Auction",
        ("FIX.5.0", "1801", "5"): "5 .. TriggeredStopOrder",
        ("FIX.5.0", "1801", "6"): "6 .. TriggeredContingencyOrder",
        ("FIX.5.0", "1801", "7"): "7 .. TriggeredMarketOrder",
        ("FIX.5.0", "1803", "0"): "0 .. None",
        ("FIX.5.0", "1803", "1"): "1 .. BlockOrderAuction",
        ("FIX.5.0", "1803", "2"): "2 .. DirectedOrderAuction",
        ("FIX.5.0", "1803", "3"): "3 .. ExposureOrderAuction",
        ("FIX.5.0", "1803", "4"): "4 .. FlashOrderAuction",
        ("FIX.5.0", "1803", "5"): "5 .. FacilitationOrderAuction",
        ("FIX.5.0", "1803", "6"): "6 .. SolicitationOrderAuction",
        ("FIX.5.0", "1803", "7"): "7 .. PriceImprovementMechanism",
        ("FIX.5.0", "1803", "8"): "8 .. DirectedOrderPriceImprovementMechanism",
        ("FIX.5.0", "1805", "0"): "0 .. AutomatedAuctionPermitted",
        ("FIX.5.0", "1805", "1"): "1 .. AutomatedAuctionNotPermitted",
        ("FIX.5.0", "1807", "0"): "0 .. NotLocked",
        ("FIX.5.0", "1807", "1"): "1 .. AwayMarketNetter",
        ("FIX.5.0", "1807", "2"): "2 .. ThreeTickLocked",
        ("FIX.5.0", "1807", "3"): "3 .. LockedByMarketMaker",
        ("FIX.5.0", "1807", "4"): "4 .. DirectedOrderLock",
        ("FIX.5.0", "1807", "5"): "5 .. MultilegLock",
        ("FIX.5.0", "1807", "6"): "6 .. MarketOrderLock",
        ("FIX.5.0", "1807", "7"): "7 .. PreAssignmentLock",
        ("FIX.5.0", "1810", "1"): "1 .. ISO",
        ("FIX.5.0", "1810", "2"): "2 .. NoAwayMarketBetterCheck",
        ("FIX.5.0", "1813", "1"): "1 .. Volume",
        ("FIX.5.0", "1813", "2"): "2 .. Price",
        ("FIX.5.0", "1813", "3"): "3 .. Side",
        ("FIX.5.0", "1813", "4"): "4 .. AON",
        ("FIX.5.0", "1813", "5"): "5 .. General",
        ("FIX.5.0", "1813", "6"): "6 .. ClearingAccount",
        ("FIX.5.0", "1813", "7"): "7 .. CMTAAccount",
        ("FIX.5.0", "1814", "0"): "0 .. No",
        ("FIX.5.0", "1814", "1"): "1 .. Yes",
        ("FIX.5.0", "1814", "2"): "2 .. UseDefaultSetting",
        ("FIX.5.0", "1815", "1"): "1 .. Customer",
        ("FIX.5.0", "1815", "2"): "2 .. CustomerProfessional",
        ("FIX.5.0", "1815", "3"): "3 .. BrokerDealer",
        ("FIX.5.0", "1815", "4"): "4 .. CustomerBrokerDealer",
        ("FIX.5.0", "1815", "5"): "5 .. Principal",
        ("FIX.5.0", "1815", "6"): "6 .. MarketMaker",
        ("FIX.5.0", "1815", "7"): "7 .. AwayMarketMaker",
        ("FIX.5.0", "1815", "8"): "8 .. SystematicInternaliser",
        ("FIX.5.0", "1816", "1"): "1 .. Customer",
        ("FIX.5.0", "1816", "2"): "2 .. Firm",
        ("FIX.5.0", "1816", "3"): "3 .. MarketMaker",
        ("FIX.5.0", "1817", "1"): "1 .. Customer",
        ("FIX.5.0", "1817", "2"): "2 .. Firm",
        ("FIX.5.0", "1817", "3"): "3 .. MarketMaker",
        ("FIX.5.0", "1818", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "1818", "18"): "18 .. Current",
        ("FIX.5.0", "1818", "19"): "19 .. New",
        ("FIX.5.0", "1818", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "1818", "0"): "0 .. Agency",
        ("FIX.5.0", "1818", "1"): "1 .. Principal",
        ("FIX.5.0", "1818", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "1818", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "1818", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "1818", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "1818", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "1818", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "1818", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "1818", "27"): "27 .. Supervisor",
        ("FIX.5.0", "1818", "22"): "22 .. Algorithm",
        ("FIX.5.0", "1818", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "1818", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "1818", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "1818", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "1818", "7"): "7 .. Bank",
        ("FIX.5.0", "1818", "8"): "8 .. Hub",
        ("FIX.5.0", "1818", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "1818", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "1818", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "1818", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "1818", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "1818", "21"): "21 .. Specialist",
        ("FIX.5.0", "1818", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "1821", "1"): "1 .. NBBid",
        ("FIX.5.0", "1821", "2"): "2 .. NBOffer",
        ("FIX.5.0", "1822", "1"): "1 .. Once",
        ("FIX.5.0", "1822", "2"): "2 .. Multiple",
        ("FIX.5.0", "1823", "0"): "0 .. NotTriggered",
        ("FIX.5.0", "1823", "1"): "1 .. Triggered",
        ("FIX.5.0", "1823", "2"): "2 .. StopOrderTriggered",
        ("FIX.5.0", "1823", "3"): "3 .. OCOOrderTriggered",
        ("FIX.5.0", "1823", "4"): "4 .. OTOOrderTriggered",
        ("FIX.5.0", "1823", "5"): "5 .. OUOOrderTriggered",
        ("FIX.5.0", "1827", "H"): "H .. Hour",
        ("FIX.5.0", "1827", "Min"): "Min .. Minute",
        ("FIX.5.0", "1827", "S"): "S .. Second",
        ("FIX.5.0", "1827", "D"): "D .. Day",
        ("FIX.5.0", "1827", "Wk"): "Wk .. Week",
        ("FIX.5.0", "1827", "Mo"): "Mo .. Month",
        ("FIX.5.0", "1827", "Yr"): "Yr .. Year",
        ("FIX.5.0", "1832", "0"): "0 .. NotCleared",
        ("FIX.5.0", "1832", "1"): "1 .. Cleared",
        ("FIX.5.0", "1832", "2"): "2 .. Submitted",
        ("FIX.5.0", "1832", "3"): "3 .. Rejected",
        ("FIX.5.0", "1833", "0"): "0 .. TwoComponentIntercommoditySpread",
        ("FIX.5.0", "1833", "1"): "1 .. IndexOrBasket",
        ("FIX.5.0", "1833", "2"): "2 .. TwoComponentLocationBasis",
        ("FIX.5.0", "1833", "99"): "99 .. Other",
        ("FIX.5.0", "1834", "0"): "0 .. Principal",
        ("FIX.5.0", "1834", "1"): "1 .. Agent",
        ("FIX.5.0", "1834", "2"): "2 .. Customer",
        ("FIX.5.0", "1834", "3"): "3 .. Counterparty",
        ("FIX.5.0", "1836", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "1836", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "1836", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "1836", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "1836", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "1836", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "1836", "thm"): "thm .. therms",
        ("FIX.5.0", "1836", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "1836", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "1836", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "1836", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "1836", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "1836", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "1836", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "1836", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "1836", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "1836", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "1836", "day"): "day .. Day",
        ("FIX.5.0", "1836", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "1836", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "1836", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "1836", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "1836", "g"): "g .. Grams",
        ("FIX.5.0", "1836", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "1836", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "1836", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "1836", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "1836", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "1836", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "1836", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "1836", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "1836", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "1836", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "1836", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "1836", "L"): "L .. liters",
        ("FIX.5.0", "1836", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "1836", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "1836", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "1836", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "1836", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "1836", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "1836", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "1836", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "1836", "t"): "t .. MetricTons",
        ("FIX.5.0", "1836", "tn"): "tn .. Tons",
        ("FIX.5.0", "1839", "0"): "0 .. SpecialCumDividend",
        ("FIX.5.0", "1839", "1"): "1 .. SpecialCumRights",
        ("FIX.5.0", "1839", "2"): "2 .. SpecialExDividend",
        ("FIX.5.0", "1839", "3"): "3 .. SpecialExRights",
        ("FIX.5.0", "1839", "4"): "4 .. SpecialCumCoupon",
        ("FIX.5.0", "1839", "5"): "5 .. SpecialCumCapitalRepayments",
        ("FIX.5.0", "1839", "6"): "6 .. SpecialExCoupon",
        ("FIX.5.0", "1839", "7"): "7 .. SpecialExCapitalRepayments",
        ("FIX.5.0", "1839", "8"): "8 .. CashSettlement",
        ("FIX.5.0", "1839", "9"): "9 .. SpecialCumBonus",
        ("FIX.5.0", "1839", "10"): "10 .. SpecialPrice",
        ("FIX.5.0", "1839", "11"): "11 .. SpecialExBonus",
        ("FIX.5.0", "1839", "12"): "12 .. GuaranteedDelivery",
        ("FIX.5.0", "1839", "13"): "13 .. SpecialDividend",
        ("FIX.5.0", "1839", "14"): "14 .. PriceImprovement",
        ("FIX.5.0", "1839", "15"): "15 .. NonPriceFormingTrade",
        ("FIX.5.0", "1839", "16"): "16 .. TradeExemptedFromTradingObligation",
        ("FIX.5.0", "1839", "17"): "17 .. PricePending",
        ("FIX.5.0", "1839", "18"): "18 .. PriceNotApplicable",
        ("FIX.5.0", "1840", "0"): "0 .. PendingClear",
        ("FIX.5.0", "1840", "1"): "1 .. Claimed",
        ("FIX.5.0", "1840", "2"): "2 .. Cleared",
        ("FIX.5.0", "1840", "3"): "3 .. Rejected",
        ("FIX.5.0", "1842", "0"): "0 .. ClearedQuantity",
        ("FIX.5.0", "1842", "1"): "1 .. LongSideClaimedQuantity",
        ("FIX.5.0", "1842", "2"): "2 .. ShortSideClaimedQuantity",
        ("FIX.5.0", "1842", "3"): "3 .. LongSideRejectedQuantity",
        ("FIX.5.0", "1842", "4"): "4 .. ShortSideRejectedQuantity",
        ("FIX.5.0", "1842", "5"): "5 .. PendingQuantity",
        ("FIX.5.0", "1842", "6"): "6 .. TransactionQuantity",
        ("FIX.5.0", "1842", "7"): "7 .. RemainingQuantity",
        ("FIX.5.0", "1842", "8"): "8 .. PreviousRemainingQuantity",
        ("FIX.5.0", "1845", "CASH"): "CASH .. CashAmount",
        ("FIX.5.0", "1845", "CRES"): "CRES .. CashResidualAmount",
        ("FIX.5.0", "1845", "FMTM"): "FMTM .. FinalMarkToMarketAmount",
        ("FIX.5.0", "1845", "IMTM"): "IMTM .. IncrementalMarkToMarketAmount",
        ("FIX.5.0", "1845", "PREM"): "PREM .. PremiumAmount",
        ("FIX.5.0", "1845", "SMTM"): "SMTM .. StartOfDayMarkToMarketAmount",
        ("FIX.5.0", "1845", "TVAR"): "TVAR .. TradeVariationAmount",
        ("FIX.5.0", "1845", "VADJ"): "VADJ .. ValueAdjustedAmount",
        ("FIX.5.0", "1845", "SETL"): "SETL .. SettlementValue",
        ("FIX.5.0", "1845", "ICPN"): "ICPN .. InitialTradeCouponAmount",
        ("FIX.5.0", "1845", "ACPN"): "ACPN .. AccruedCouponAmount",
        ("FIX.5.0", "1845", "CPN"): "CPN .. CouponAmount",
        ("FIX.5.0", "1845", "IACPN"): "IACPN .. IncrementalAccruedCoupon",
        ("FIX.5.0", "1845", "CMTM"): "CMTM .. CollateralizedMarkToMarket",
        ("FIX.5.0", "1845", "ICMTM"): "ICMTM .. IncrementalCollateralizedMarkToMarket",
        ("FIX.5.0", "1845", "DLV"): "DLV .. CompensationAmount",
        ("FIX.5.0", "1845", "BANK"): "BANK .. TotalBankedAmount",
        ("FIX.5.0", "1845", "COLAT"): "COLAT .. TotalCollateralizedAmount",
        ("FIX.5.0", "1845", "LSNV"): "LSNV .. LongPairedSwapNotionalValue",
        ("FIX.5.0", "1845", "SSNV"): "SSNV .. ShortPairedSwapNotionalValue",
        ("FIX.5.0", "1845", "SACPN"): "SACPN .. StartOfDayAccruedCoupon",
        ("FIX.5.0", "1845", "NPV"): "NPV .. NetPresentValue",
        ("FIX.5.0", "1845", "SNPV"): "SNPV .. StartOfDayNetPresentValue",
        ("FIX.5.0", "1845", "NCF"): "NCF .. NetCashFlow",
        ("FIX.5.0", "1845", "PVFEES"): "PVFEES .. PresentValueOfFees",
        ("FIX.5.0", "1845", "PV01"): "PV01 .. PresentValueOneBasisPoints",
        ("FIX.5.0", "1845", "5YREN"): "5YREN .. FiveYearEquivalentNotional",
        ("FIX.5.0", "1845", "UMTM"): "UMTM .. UndiscountedMarkToMarket",
        ("FIX.5.0", "1845", "MTD"): "MTD .. MarkToModel",
        ("FIX.5.0", "1845", "VMTM"): "VMTM .. MarkToMarketVariance",
        ("FIX.5.0", "1845", "VMTD"): "VMTD .. MarkToModelVariance",
        ("FIX.5.0", "1845", "UPFRNT"): "UPFRNT .. UpfrontPayment",
        ("FIX.5.0", "1848", "0"): "0 .. Add",
        ("FIX.5.0", "1848", "1"): "1 .. DoNotAdd",
        ("FIX.5.0", "1849", "0"): "0 .. Offset",
        ("FIX.5.0", "1849", "1"): "1 .. Onset",
        ("FIX.5.0", "1850", "0"): "0 .. OptionsSettlement",
        ("FIX.5.0", "1850", "1"): "1 .. PendingErosionAdjustment",
        ("FIX.5.0", "1850", "2"): "2 .. FinalErosionAdjustment",
        ("FIX.5.0", "1850", "3"): "3 .. TearUpCouponAmount",
        ("FIX.5.0", "1850", "4"): "4 .. PriceAlignmentInterest",
        ("FIX.5.0", "1850", "5"): "5 .. DeliveryInvoiceCharges",
        ("FIX.5.0", "1850", "6"): "6 .. DeliveryStorageCharges",
        ("FIX.5.0", "1853", "0"): "0 .. NoAvgPricing",
        ("FIX.5.0", "1853", "1"): "1 .. TradeIsPartAvgPriceGrp",
        ("FIX.5.0", "1853", "2"): "2 .. LastTradeIsPartAvgPriceGrp",
        ("FIX.5.0", "1857", "0"): "0 .. NonFIXSource",
        ("FIX.5.0", "1857", "1"): "1 .. TradeID",
        ("FIX.5.0", "1857", "2"): "2 .. SecondaryTradeID",
        ("FIX.5.0", "1857", "3"): "3 .. TradeReportID",
        ("FIX.5.0", "1857", "4"): "4 .. FirmTradeID",
        ("FIX.5.0", "1857", "5"): "5 .. SecondaryFirmTradeID",
        ("FIX.5.0", "1857", "6"): "6 .. RegulatoryTradeID",
        ("FIX.5.0", "1863", "1"): "1 .. PosMaintRptID",
        ("FIX.5.0", "1863", "2"): "2 .. TransferID",
        ("FIX.5.0", "1863", "3"): "3 .. PositionEntityID",
        ("FIX.5.0", "1865", "0"): "0 .. ReceivedNotYetProcessed",
        ("FIX.5.0", "1865", "1"): "1 .. Accepted",
        ("FIX.5.0", "1865", "2"): "2 .. Rejected",
        ("FIX.5.0", "1869", "1"): "1 .. PriceCheck",
        ("FIX.5.0", "1869", "2"): "2 .. NotionalValueCheck",
        ("FIX.5.0", "1869", "3"): "3 .. QuantityCheck",
        ("FIX.5.0", "1870", "0"): "0 .. DoNotCheck",
        ("FIX.5.0", "1870", "1"): "1 .. Check",
        ("FIX.5.0", "1870", "2"): "2 .. BestEffort",
        ("FIX.5.0", "1877", "0"): "0 .. Successful",
        ("FIX.5.0", "1877", "1"): "1 .. InvalidParty",
        ("FIX.5.0", "1877", "2"): "2 .. InvalidRelatedParty",
        ("FIX.5.0", "1877", "3"): "3 .. InvalidPartyStatus",
        ("FIX.5.0", "1877", "98"): "98 .. NotAuthorized",
        ("FIX.5.0", "1877", "99"): "99 .. Other",
        ("FIX.5.0", "1878", "0"): "0 .. Accepted",
        ("FIX.5.0", "1878", "1"): "1 .. AcceptedWithChanges",
        ("FIX.5.0", "1878", "2"): "2 .. Rejected",
        ("FIX.5.0", "1878", "3"): "3 .. AcceptancePending",
        ("FIX.5.0", "1879", "0"): "0 .. Accepted",
        ("FIX.5.0", "1879", "1"): "1 .. AcceptedWithChanges",
        ("FIX.5.0", "1879", "2"): "2 .. Rejected",
        ("FIX.5.0", "1880", "0"): "0 .. Successful",
        ("FIX.5.0", "1880", "1"): "1 .. InvalidParty",
        ("FIX.5.0", "1880", "2"): "2 .. InvalidRelatedParty",
        ("FIX.5.0", "1880", "3"): "3 .. InvalidPartyStatus",
        ("FIX.5.0", "1880", "98"): "98 .. NotAuthorized",
        ("FIX.5.0", "1880", "99"): "99 .. Other",
        ("FIX.5.0", "1881", "0"): "0 .. Successful",
        ("FIX.5.0", "1881", "1"): "1 .. InvalidParty",
        ("FIX.5.0", "1881", "2"): "2 .. InvalidRelatedParty",
        ("FIX.5.0", "1881", "3"): "3 .. InvalidEntitlementType",
        ("FIX.5.0", "1881", "4"): "4 .. InvalidEntitlementID",
        ("FIX.5.0", "1881", "5"): "5 .. InvalidEntitlementAttribute",
        ("FIX.5.0", "1881", "6"): "6 .. InvalidInstrumentScope",
        ("FIX.5.0", "1881", "7"): "7 .. InvalidMarketSegmentScope",
        ("FIX.5.0", "1881", "8"): "8 .. InvalidStartDate",
        ("FIX.5.0", "1881", "9"): "9 .. InvalidEndDate",
        ("FIX.5.0", "1881", "10"): "10 .. InstrumentScopeNotSupported",
        ("FIX.5.0", "1881", "11"): "11 .. MarketSegmentScopeNotSupported",
        ("FIX.5.0", "1881", "12"): "12 .. EntitlementNotApprovedForParty",
        ("FIX.5.0", "1881", "13"): "13 .. EntitlementAlreadyDefinedForParty",
        ("FIX.5.0", "1881", "14"): "14 .. InstrumentNotApprovedForParty",
        ("FIX.5.0", "1881", "98"): "98 .. NotAuthorized",
        ("FIX.5.0", "1881", "99"): "99 .. Other",
        ("FIX.5.0", "1882", "0"): "0 .. Accepted",
        ("FIX.5.0", "1882", "1"): "1 .. AcceptedWithChanges",
        ("FIX.5.0", "1882", "2"): "2 .. Rejected",
        ("FIX.5.0", "1882", "3"): "3 .. AcceptancePending",
        ("FIX.5.0", "1883", "0"): "0 .. Accepted",
        ("FIX.5.0", "1883", "1"): "1 .. AcceptedWithChanges",
        ("FIX.5.0", "1883", "2"): "2 .. Rejected",
        ("FIX.5.0", "1883", "3"): "3 .. Pending",
        ("FIX.5.0", "1883", "4"): "4 .. Requested",
        ("FIX.5.0", "1883", "5"): "5 .. Deferred",
        ("FIX.5.0", "1884", "0"): "0 .. Successful",
        ("FIX.5.0", "1884", "1"): "1 .. InvalidParty",
        ("FIX.5.0", "1884", "2"): "2 .. InvalidRelatedParty",
        ("FIX.5.0", "1884", "3"): "3 .. InvalidEntitlementType",
        ("FIX.5.0", "1884", "4"): "4 .. InvalidEntitlementID",
        ("FIX.5.0", "1884", "5"): "5 .. InvalidEntitlementAttribute",
        ("FIX.5.0", "1884", "6"): "6 .. InvalidInstrumentScope",
        ("FIX.5.0", "1884", "7"): "7 .. InvalidMarketSegmentScope",
        ("FIX.5.0", "1884", "8"): "8 .. InvalidStartDate",
        ("FIX.5.0", "1884", "9"): "9 .. InvalidEndDate",
        ("FIX.5.0", "1884", "10"): "10 .. InstrumentScopeNotSupported",
        ("FIX.5.0", "1884", "11"): "11 .. MarketSegmentScopeNotSupported",
        ("FIX.5.0", "1884", "12"): "12 .. EntitlementNotApprovedForParty",
        ("FIX.5.0", "1884", "13"): "13 .. EntitlementAlreadyDefinedForParty",
        ("FIX.5.0", "1884", "14"): "14 .. InstrumentNotApprovedForParty",
        ("FIX.5.0", "1884", "98"): "98 .. NotAuthorized",
        ("FIX.5.0", "1884", "99"): "99 .. Other",
        ("FIX.5.0", "1886", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "1886", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "1886", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "1886", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "1886", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "1886", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "1886", "thm"): "thm .. therms",
        ("FIX.5.0", "1886", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "1886", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "1886", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "1886", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "1886", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "1886", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "1886", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "1886", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "1886", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "1886", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "1886", "day"): "day .. Day",
        ("FIX.5.0", "1886", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "1886", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "1886", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "1886", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "1886", "g"): "g .. Grams",
        ("FIX.5.0", "1886", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "1886", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "1886", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "1886", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "1886", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "1886", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "1886", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "1886", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "1886", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "1886", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "1886", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "1886", "L"): "L .. liters",
        ("FIX.5.0", "1886", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "1886", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "1886", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "1886", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "1886", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "1886", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "1886", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "1886", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "1886", "t"): "t .. MetricTons",
        ("FIX.5.0", "1886", "tn"): "tn .. Tons",
        ("FIX.5.0", "1896", "0"): "0 .. ReceivedNotProcessed",
        ("FIX.5.0", "1896", "1"): "1 .. Accepted",
        ("FIX.5.0", "1896", "2"): "2 .. Rejected",
        ("FIX.5.0", "1897", "0"): "0 .. Successful",
        ("FIX.5.0", "1897", "1"): "1 .. InvalidPartyInformation",
        ("FIX.5.0", "1897", "2"): "2 .. UnknownInstrument",
        ("FIX.5.0", "1897", "3"): "3 .. Unauthorized",
        ("FIX.5.0", "1897", "4"): "4 .. InvalidTradeType",
        ("FIX.5.0", "1897", "99"): "99 .. Other",
        ("FIX.5.0", "1899", "E"): "E .. Electronic",
        ("FIX.5.0", "1899", "P"): "P .. Pit",
        ("FIX.5.0", "1899", "X"): "X .. ExPit",
        ("FIX.5.0", "1899", "C"): "C .. ClearingHouse",
        ("FIX.5.0", "1899", "R"): "R .. RegisteredMarket",
        ("FIX.5.0", "1899", "O"): "O .. OffMarket",
        ("FIX.5.0", "1899", "B"): "B .. CentralLimitOrderBook",
        ("FIX.5.0", "1899", "Q"): "Q .. QuoteDrivenMarket",
        ("FIX.5.0", "1899", "D"): "D .. DarkOrderBook",
        ("FIX.5.0", "1899", "A"): "A .. AuctionDrivenMarket",
        ("FIX.5.0", "1899", "N"): "N .. QuoteNegotiation",
        ("FIX.5.0", "1899", "V"): "V .. VoiceNegotiation",
        ("FIX.5.0", "1899", "H"): "H .. HybridMarket",
        ("FIX.5.0", "1904", "0"): "0 .. InitialBlockTrade",
        ("FIX.5.0", "1904", "1"): "1 .. Allocation",
        ("FIX.5.0", "1904", "2"): "2 .. Clearing",
        ("FIX.5.0", "1904", "3"): "3 .. Compression",
        ("FIX.5.0", "1904", "4"): "4 .. Novation",
        ("FIX.5.0", "1904", "5"): "5 .. Termination",
        ("FIX.5.0", "1904", "6"): "6 .. PostTrdVal",
        ("FIX.5.0", "1906", "0"): "0 .. Current",
        ("FIX.5.0", "1906", "1"): "1 .. Previous",
        ("FIX.5.0", "1906", "2"): "2 .. Block",
        ("FIX.5.0", "1906", "3"): "3 .. Related",
        ("FIX.5.0", "1906", "4"): "4 .. ClearedBlockTrade",
        ("FIX.5.0", "1906", "5"): "5 .. TradingVenueTransactionIdentifier",
        ("FIX.5.0", "1911", "0"): "0 .. InitialBlockTrade",
        ("FIX.5.0", "1911", "1"): "1 .. Allocation",
        ("FIX.5.0", "1911", "2"): "2 .. Clearing",
        ("FIX.5.0", "1911", "3"): "3 .. Compression",
        ("FIX.5.0", "1911", "4"): "4 .. Novation",
        ("FIX.5.0", "1911", "5"): "5 .. Termination",
        ("FIX.5.0", "1911", "6"): "6 .. PostTrdVal",
        ("FIX.5.0", "1912", "0"): "0 .. Current",
        ("FIX.5.0", "1912", "1"): "1 .. Previous",
        ("FIX.5.0", "1912", "2"): "2 .. Block",
        ("FIX.5.0", "1912", "3"): "3 .. Related",
        ("FIX.5.0", "1912", "4"): "4 .. ClearedBlockTrade",
        ("FIX.5.0", "1912", "5"): "5 .. TradingVenueTransactionIdentifier",
        ("FIX.5.0", "1916", "0"): "0 .. Seconds",
        ("FIX.5.0", "1916", "1"): "1 .. TenthsOfASecond",
        ("FIX.5.0", "1916", "2"): "2 .. HundredthsOfASecond",
        ("FIX.5.0", "1916", "3"): "3 .. Milliseconds",
        ("FIX.5.0", "1916", "4"): "4 .. Microseconds",
        ("FIX.5.0", "1916", "5"): "5 .. Nanoseconds",
        ("FIX.5.0", "1916", "10"): "10 .. Minutes",
        ("FIX.5.0", "1916", "11"): "11 .. Hours",
        ("FIX.5.0", "1916", "12"): "12 .. Days",
        ("FIX.5.0", "1916", "13"): "13 .. Weeks",
        ("FIX.5.0", "1916", "14"): "14 .. Months",
        ("FIX.5.0", "1916", "15"): "15 .. Years",
        ("FIX.5.0", "1923", "0"): "0 .. Amount",
        ("FIX.5.0", "1923", "1"): "1 .. Percentage",
        ("FIX.5.0", "1924", "0"): "0 .. DoNotIntendToClear",
        ("FIX.5.0", "1924", "1"): "1 .. IntendToClear",
        ("FIX.5.0", "1925", "0"): "0 .. ProcessNormally",
        ("FIX.5.0", "1925", "1"): "1 .. ExcludeFromAllNetting",
        ("FIX.5.0", "1925", "2"): "2 .. BilateralNettingOnly",
        ("FIX.5.0", "1925", "3"): "3 .. ExClearing",
        ("FIX.5.0", "1925", "4"): "4 .. SpecialTrade",
        ("FIX.5.0", "1925", "5"): "5 .. MultilateralNetting",
        ("FIX.5.0", "1925", "6"): "6 .. ClearAgainstCentralCounterparty",
        ("FIX.5.0", "1925", "7"): "7 .. ExcludeFromCentralCounterparty",
        ("FIX.5.0", "1925", "8"): "8 .. ManualMode",
        ("FIX.5.0", "1925", "9"): "9 .. AutomaticPostingMode",
        ("FIX.5.0", "1925", "10"): "10 .. AutomaticGiveUpMode",
        ("FIX.5.0", "1925", "11"): "11 .. QualifiedServiceRepresentativeQSR",
        ("FIX.5.0", "1925", "12"): "12 .. CustomerTrade",
        ("FIX.5.0", "1925", "13"): "13 .. SelfClearing",
        ("FIX.5.0", "1925", "14"): "14 .. BuyIn",
        ("FIX.5.0", "1927", "0"): "0 .. NonElectronic",
        ("FIX.5.0", "1927", "1"): "1 .. Electronic",
        ("FIX.5.0", "1927", "2"): "2 .. Unconfirmed",
        ("FIX.5.0", "1931", "0"): "0 .. NonElectronic",
        ("FIX.5.0", "1931", "1"): "1 .. Electronic",
        ("FIX.5.0", "1932", "0"): "0 .. NoException",
        ("FIX.5.0", "1932", "1"): "1 .. Exception",
        ("FIX.5.0", "1932", "2"): "2 .. EndUserException",
        ("FIX.5.0", "1932", "3"): "3 .. InterAffiliateException",
        ("FIX.5.0", "1932", "4"): "4 .. TreasuryAffiliateException",
        ("FIX.5.0", "1932", "5"): "5 .. CooperativeException",
        ("FIX.5.0", "1933", "PAY"): "PAY .. Pay",
        ("FIX.5.0", "1933", "RCV"): "RCV .. Rcv",
        ("FIX.5.0", "1933", "NA"): "NA .. NA",
        ("FIX.5.0", "1934", "0"): "0 .. RT",
        ("FIX.5.0", "1934", "1"): "1 .. PET",
        ("FIX.5.0", "1934", "2"): "2 .. Snapshot",
        ("FIX.5.0", "1934", "3"): "3 .. Confirmation",
        ("FIX.5.0", "1934", "4"): "4 .. RTPET",
        ("FIX.5.0", "1934", "5"): "5 .. PETConfirmation",
        ("FIX.5.0", "1934", "6"): "6 .. RTPETConfirmation",
        ("FIX.5.0", "1934", "7"): "7 .. PostTrade",
        ("FIX.5.0", "1934", "8"): "8 .. Verification",
        ("FIX.5.0", "1934", "9"): "9 .. PstTrdEvnt",
        ("FIX.5.0", "1934", "10"): "10 .. PstTrdEvntRTReportable",
        ("FIX.5.0", "1934", "11"): "11 .. LMTF",
        ("FIX.5.0", "1934", "12"): "12 .. DATF",
        ("FIX.5.0", "1934", "13"): "13 .. VOLO",
        ("FIX.5.0", "1934", "14"): "14 .. FWAF",
        ("FIX.5.0", "1934", "15"): "15 .. IDAF",
        ("FIX.5.0", "1934", "16"): "16 .. VOLW",
        ("FIX.5.0", "1934", "17"): "17 .. FULF",
        ("FIX.5.0", "1934", "18"): "18 .. FULA",
        ("FIX.5.0", "1934", "19"): "19 .. FULV",
        ("FIX.5.0", "1934", "20"): "20 .. FULJ",
        ("FIX.5.0", "1934", "21"): "21 .. COAF",
        ("FIX.5.0", "1936", "0"): "0 .. Uncollateralized",
        ("FIX.5.0", "1936", "1"): "1 .. PartiallyCollateralized",
        ("FIX.5.0", "1936", "2"): "2 .. OneWayCollaterallization",
        ("FIX.5.0", "1936", "3"): "3 .. FullyCollateralized",
        ("FIX.5.0", "1937", "0"): "0 .. Novation",
        ("FIX.5.0", "1937", "1"): "1 .. PartialNovation",
        ("FIX.5.0", "1937", "2"): "2 .. TradeUnwind",
        ("FIX.5.0", "1937", "3"): "3 .. PartialTradeUnwind",
        ("FIX.5.0", "1937", "4"): "4 .. Exercise",
        ("FIX.5.0", "1937", "5"): "5 .. Netting",
        ("FIX.5.0", "1937", "6"): "6 .. FullNetting",
        ("FIX.5.0", "1937", "7"): "7 .. PartialNetting",
        ("FIX.5.0", "1937", "8"): "8 .. Amendment",
        ("FIX.5.0", "1937", "9"): "9 .. Increase",
        ("FIX.5.0", "1937", "10"): "10 .. CreditEvent",
        ("FIX.5.0", "1937", "11"): "11 .. StrategicRestructuring",
        ("FIX.5.0", "1937", "12"): "12 .. SuccessionEventReorganization",
        ("FIX.5.0", "1937", "13"): "13 .. SuccessionEventRenaming",
        ("FIX.5.0", "1937", "14"): "14 .. Porting",
        ("FIX.5.0", "1937", "15"): "15 .. Withdrawl",
        ("FIX.5.0", "1937", "16"): "16 .. Void",
        ("FIX.5.0", "1937", "17"): "17 .. AccountTransfer",
        ("FIX.5.0", "1937", "18"): "18 .. GiveUp",
        ("FIX.5.0", "1937", "19"): "19 .. TakeUp",
        ("FIX.5.0", "1937", "20"): "20 .. AveragePricing",
        ("FIX.5.0", "1937", "21"): "21 .. Reversal",
        ("FIX.5.0", "1937", "22"): "22 .. AllocTrdPosting",
        ("FIX.5.0", "1937", "23"): "23 .. Cascade",
        ("FIX.5.0", "1937", "24"): "24 .. Delivery",
        ("FIX.5.0", "1937", "25"): "25 .. OptionAsgn",
        ("FIX.5.0", "1937", "26"): "26 .. Expiration",
        ("FIX.5.0", "1937", "27"): "27 .. Maturity",
        ("FIX.5.0", "1937", "28"): "28 .. EqualPosAdj",
        ("FIX.5.0", "1937", "29"): "29 .. UnequalPosAdj",
        ("FIX.5.0", "1937", "30"): "30 .. Correction",
        ("FIX.5.0", "1937", "99"): "99 .. Other",
        ("FIX.5.0", "1938", "1"): "1 .. InterestRate",
        ("FIX.5.0", "1938", "2"): "2 .. Currency",
        ("FIX.5.0", "1938", "3"): "3 .. Credit",
        ("FIX.5.0", "1938", "4"): "4 .. Equity",
        ("FIX.5.0", "1938", "5"): "5 .. Commodity",
        ("FIX.5.0", "1938", "6"): "6 .. Other",
        ("FIX.5.0", "1938", "7"): "7 .. Cash",
        ("FIX.5.0", "1938", "8"): "8 .. Debt",
        ("FIX.5.0", "1938", "9"): "9 .. Fund",
        ("FIX.5.0", "1938", "10"): "10 .. LoanFacility",
        ("FIX.5.0", "1939", "13"): "13 .. Metals",
        ("FIX.5.0", "1939", "14"): "14 .. Bullion",
        ("FIX.5.0", "1939", "15"): "15 .. Energy",
        ("FIX.5.0", "1939", "16"): "16 .. CommodityIndex",
        ("FIX.5.0", "1939", "17"): "17 .. Agricultural",
        ("FIX.5.0", "1939", "18"): "18 .. Environmental",
        ("FIX.5.0", "1939", "19"): "19 .. Freight",
        ("FIX.5.0", "1939", "41"): "41 .. Fertilizer",
        ("FIX.5.0", "1939", "42"): "42 .. IndustrialProduct",
        ("FIX.5.0", "1939", "43"): "43 .. Inflation",
        ("FIX.5.0", "1939", "44"): "44 .. Paper",
        ("FIX.5.0", "1939", "45"): "45 .. Polypropylene",
        ("FIX.5.0", "1939", "46"): "46 .. OfficialEconomicStatistics",
        ("FIX.5.0", "1939", "4"): "4 .. SingleName",
        ("FIX.5.0", "1939", "5"): "5 .. CreditIndex",
        ("FIX.5.0", "1939", "6"): "6 .. IndexTranche",
        ("FIX.5.0", "1939", "7"): "7 .. CreditBasket",
        ("FIX.5.0", "1939", "3"): "3 .. Basket",
        ("FIX.5.0", "1939", "38"): "38 .. FXCrossRates",
        ("FIX.5.0", "1939", "39"): "39 .. FXEmergingMarkets",
        ("FIX.5.0", "1939", "40"): "40 .. FXMajors",
        ("FIX.5.0", "1939", "20"): "20 .. Government",
        ("FIX.5.0", "1939", "21"): "21 .. Agency",
        ("FIX.5.0", "1939", "22"): "22 .. Corporate",
        ("FIX.5.0", "1939", "23"): "23 .. Financing",
        ("FIX.5.0", "1939", "24"): "24 .. MoneyMarket",
        ("FIX.5.0", "1939", "25"): "25 .. Mortgage",
        ("FIX.5.0", "1939", "26"): "26 .. Municipal",
        ("FIX.5.0", "1939", "9"): "9 .. Common",
        ("FIX.5.0", "1939", "10"): "10 .. Preferred",
        ("FIX.5.0", "1939", "11"): "11 .. EquityIndex",
        ("FIX.5.0", "1939", "12"): "12 .. EquityBasket",
        ("FIX.5.0", "1939", "34"): "34 .. DividendIndex",
        ("FIX.5.0", "1939", "35"): "35 .. StockDividend",
        ("FIX.5.0", "1939", "36"): "36 .. ExchangeTradedFund",
        ("FIX.5.0", "1939", "37"): "37 .. VolatilityIndex",
        ("FIX.5.0", "1939", "27"): "27 .. MutualFund",
        ("FIX.5.0", "1939", "28"): "28 .. CollectiveInvestmentVehicle",
        ("FIX.5.0", "1939", "29"): "29 .. InvestmentProgram",
        ("FIX.5.0", "1939", "30"): "30 .. SpecializedAccountProgram",
        ("FIX.5.0", "1939", "1"): "1 .. SingleCurrency",
        ("FIX.5.0", "1939", "2"): "2 .. CrossCurrency",
        ("FIX.5.0", "1939", "31"): "31 .. TermLoan",
        ("FIX.5.0", "1939", "32"): "32 .. BridgeLoan",
        ("FIX.5.0", "1939", "33"): "33 .. LetterOfCredit",
        ("FIX.5.0", "1939", "8"): "8 .. Exotic",
        ("FIX.5.0", "1939", "47"): "47 .. Other",
        ("FIX.5.0", "1939", "48"): "48 .. Other",
        ("FIX.5.0", "1941", "BS"): "BS .. BasisSwap",
        ("FIX.5.0", "1941", "IX"): "IX .. IndexSwap",
        ("FIX.5.0", "1941", "BB"): "BB .. BroadBasedSecuritySwap",
        ("FIX.5.0", "1941", "SK"): "SK .. BasketSwap",
        ("FIX.5.0", "1946", "0"): "0 .. Zero",
        ("FIX.5.0", "1946", "1"): "1 .. FixedRate",
        ("FIX.5.0", "1946", "2"): "2 .. FloatingRate",
        ("FIX.5.0", "1946", "3"): "3 .. Structured",
        ("FIX.5.0", "1949", "D"): "D .. Day",
        ("FIX.5.0", "1949", "Wk"): "Wk .. Week",
        ("FIX.5.0", "1949", "Mo"): "Mo .. Month",
        ("FIX.5.0", "1949", "Yr"): "Yr .. Year",
        ("FIX.5.0", "1949", "H"): "H .. Hour",
        ("FIX.5.0", "1949", "Min"): "Min .. Minute",
        ("FIX.5.0", "1949", "S"): "S .. Second",
        ("FIX.5.0", "1949", "T"): "T .. Term",
        ("FIX.5.0", "1950", "0"): "0 .. OneOne",
        ("FIX.5.0", "1950", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "1950", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "1950", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "1950", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "1950", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "1950", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "1950", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "1950", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "1950", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "1950", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "1950", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "1950", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "1950", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "1950", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "1950", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "1950", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "1950", "17"): "17 .. Act364",
        ("FIX.5.0", "1952", "1"): "1 .. CUSIP",
        ("FIX.5.0", "1952", "2"): "2 .. SEDOL",
        ("FIX.5.0", "1952", "3"): "3 .. QUIK",
        ("FIX.5.0", "1952", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "1952", "5"): "5 .. RICCode",
        ("FIX.5.0", "1952", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "1952", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "1952", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "1952", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "1952", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "1952", "B"): "B .. Wertpapier",
        ("FIX.5.0", "1952", "C"): "C .. Dutch",
        ("FIX.5.0", "1952", "D"): "D .. Valoren",
        ("FIX.5.0", "1952", "E"): "E .. Sicovam",
        ("FIX.5.0", "1952", "F"): "F .. Belgian",
        ("FIX.5.0", "1952", "G"): "G .. Common",
        ("FIX.5.0", "1952", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "1952", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "1952", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "1952", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "1952", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "1952", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "1952", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "1952", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "1952", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "1952", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "1952", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "1952", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "1952", "U"): "U .. Synthetic",
        ("FIX.5.0", "1952", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "1952", "W"): "W .. IndexName",
        ("FIX.5.0", "1954", "0"): "0 .. Unknown",
        ("FIX.5.0", "1954", "1"): "1 .. FirstLien",
        ("FIX.5.0", "1954", "2"): "2 .. SecondLien",
        ("FIX.5.0", "1954", "3"): "3 .. ThirdLien",
        ("FIX.5.0", "1955", "0"): "0 .. BridgeLoan",
        ("FIX.5.0", "1955", "1"): "1 .. LetterOfCredit",
        ("FIX.5.0", "1955", "2"): "2 .. RevolvingLoan",
        ("FIX.5.0", "1955", "3"): "3 .. SwinglineFunding",
        ("FIX.5.0", "1955", "4"): "4 .. TermLoan",
        ("FIX.5.0", "1955", "5"): "5 .. TradeClaim",
        ("FIX.5.0", "1956", "1"): "1 .. Asian",
        ("FIX.5.0", "1956", "2"): "2 .. AustralianNewZealand",
        ("FIX.5.0", "1956", "3"): "3 .. EuropeanEmergingMarkets",
        ("FIX.5.0", "1956", "4"): "4 .. Japanese",
        ("FIX.5.0", "1956", "5"): "5 .. NorthAmericanHighYield",
        ("FIX.5.0", "1956", "6"): "6 .. NorthAmericanInsurance",
        ("FIX.5.0", "1956", "7"): "7 .. NorthAmericanInvestmentGrade",
        ("FIX.5.0", "1956", "8"): "8 .. Singaporean",
        ("FIX.5.0", "1956", "9"): "9 .. WesternEuropean",
        ("FIX.5.0", "1956", "10"): "10 .. WesternEuropeanInsurance",
        ("FIX.5.0", "1974", "0"): "0 .. InitialBlockTrade",
        ("FIX.5.0", "1974", "1"): "1 .. Allocation",
        ("FIX.5.0", "1974", "2"): "2 .. Clearing",
        ("FIX.5.0", "1974", "3"): "3 .. Compression",
        ("FIX.5.0", "1974", "4"): "4 .. Novation",
        ("FIX.5.0", "1974", "5"): "5 .. Termination",
        ("FIX.5.0", "1974", "6"): "6 .. PostTrdVal",
        ("FIX.5.0", "1975", "0"): "0 .. Current",
        ("FIX.5.0", "1975", "1"): "1 .. Previous",
        ("FIX.5.0", "1975", "2"): "2 .. Block",
        ("FIX.5.0", "1975", "3"): "3 .. Related",
        ("FIX.5.0", "1975", "4"): "4 .. ClearedBlockTrade",
        ("FIX.5.0", "1975", "5"): "5 .. TradingVenueTransactionIdentifier",
        ("FIX.5.0", "1977", "1"): "1 .. InterestRate",
        ("FIX.5.0", "1977", "2"): "2 .. Currency",
        ("FIX.5.0", "1977", "3"): "3 .. Credit",
        ("FIX.5.0", "1977", "4"): "4 .. Equity",
        ("FIX.5.0", "1977", "5"): "5 .. Commodity",
        ("FIX.5.0", "1977", "6"): "6 .. Other",
        ("FIX.5.0", "1977", "7"): "7 .. Cash",
        ("FIX.5.0", "1977", "8"): "8 .. Debt",
        ("FIX.5.0", "1977", "9"): "9 .. Fund",
        ("FIX.5.0", "1977", "10"): "10 .. LoanFacility",
        ("FIX.5.0", "1978", "13"): "13 .. Metals",
        ("FIX.5.0", "1978", "14"): "14 .. Bullion",
        ("FIX.5.0", "1978", "15"): "15 .. Energy",
        ("FIX.5.0", "1978", "16"): "16 .. CommodityIndex",
        ("FIX.5.0", "1978", "17"): "17 .. Agricultural",
        ("FIX.5.0", "1978", "18"): "18 .. Environmental",
        ("FIX.5.0", "1978", "19"): "19 .. Freight",
        ("FIX.5.0", "1978", "41"): "41 .. Fertilizer",
        ("FIX.5.0", "1978", "42"): "42 .. IndustrialProduct",
        ("FIX.5.0", "1978", "43"): "43 .. Inflation",
        ("FIX.5.0", "1978", "44"): "44 .. Paper",
        ("FIX.5.0", "1978", "45"): "45 .. Polypropylene",
        ("FIX.5.0", "1978", "46"): "46 .. OfficialEconomicStatistics",
        ("FIX.5.0", "1978", "4"): "4 .. SingleName",
        ("FIX.5.0", "1978", "5"): "5 .. CreditIndex",
        ("FIX.5.0", "1978", "6"): "6 .. IndexTranche",
        ("FIX.5.0", "1978", "7"): "7 .. CreditBasket",
        ("FIX.5.0", "1978", "3"): "3 .. Basket",
        ("FIX.5.0", "1978", "38"): "38 .. FXCrossRates",
        ("FIX.5.0", "1978", "39"): "39 .. FXEmergingMarkets",
        ("FIX.5.0", "1978", "40"): "40 .. FXMajors",
        ("FIX.5.0", "1978", "20"): "20 .. Government",
        ("FIX.5.0", "1978", "21"): "21 .. Agency",
        ("FIX.5.0", "1978", "22"): "22 .. Corporate",
        ("FIX.5.0", "1978", "23"): "23 .. Financing",
        ("FIX.5.0", "1978", "24"): "24 .. MoneyMarket",
        ("FIX.5.0", "1978", "25"): "25 .. Mortgage",
        ("FIX.5.0", "1978", "26"): "26 .. Municipal",
        ("FIX.5.0", "1978", "9"): "9 .. Common",
        ("FIX.5.0", "1978", "10"): "10 .. Preferred",
        ("FIX.5.0", "1978", "11"): "11 .. EquityIndex",
        ("FIX.5.0", "1978", "12"): "12 .. EquityBasket",
        ("FIX.5.0", "1978", "34"): "34 .. DividendIndex",
        ("FIX.5.0", "1978", "35"): "35 .. StockDividend",
        ("FIX.5.0", "1978", "36"): "36 .. ExchangeTradedFund",
        ("FIX.5.0", "1978", "37"): "37 .. VolatilityIndex",
        ("FIX.5.0", "1978", "27"): "27 .. MutualFund",
        ("FIX.5.0", "1978", "28"): "28 .. CollectiveInvestmentVehicle",
        ("FIX.5.0", "1978", "29"): "29 .. InvestmentProgram",
        ("FIX.5.0", "1978", "30"): "30 .. SpecializedAccountProgram",
        ("FIX.5.0", "1978", "1"): "1 .. SingleCurrency",
        ("FIX.5.0", "1978", "2"): "2 .. CrossCurrency",
        ("FIX.5.0", "1978", "31"): "31 .. TermLoan",
        ("FIX.5.0", "1978", "32"): "32 .. BridgeLoan",
        ("FIX.5.0", "1978", "33"): "33 .. LetterOfCredit",
        ("FIX.5.0", "1978", "8"): "8 .. Exotic",
        ("FIX.5.0", "1978", "47"): "47 .. Other",
        ("FIX.5.0", "1978", "48"): "48 .. Other",
        ("FIX.5.0", "1980", "0"): "0 .. BlockToBeAllocated",
        ("FIX.5.0", "1980", "1"): "1 .. BlockNotToBeAllocated",
        ("FIX.5.0", "1980", "2"): "2 .. AllocatedTrade",
        ("FIX.5.0", "1982", "1"): "1 .. Put",
        ("FIX.5.0", "1982", "2"): "2 .. Call",
        ("FIX.5.0", "1982", "3"): "3 .. Tender",
        ("FIX.5.0", "1982", "4"): "4 .. SinkingFundCall",
        ("FIX.5.0", "1982", "5"): "5 .. Activation",
        ("FIX.5.0", "1982", "6"): "6 .. Inactiviation",
        ("FIX.5.0", "1982", "7"): "7 .. LastEligibleTradeDate",
        ("FIX.5.0", "1982", "8"): "8 .. SwapStartDate",
        ("FIX.5.0", "1982", "9"): "9 .. SwapEndDate",
        ("FIX.5.0", "1982", "10"): "10 .. SwapRollDate",
        ("FIX.5.0", "1982", "11"): "11 .. SwapNextStartDate",
        ("FIX.5.0", "1982", "12"): "12 .. SwapNextRollDate",
        ("FIX.5.0", "1982", "13"): "13 .. FirstDeliveryDate",
        ("FIX.5.0", "1982", "14"): "14 .. LastDeliveryDate",
        ("FIX.5.0", "1982", "15"): "15 .. InitialInventoryDueDate",
        ("FIX.5.0", "1982", "16"): "16 .. FinalInventoryDueDate",
        ("FIX.5.0", "1982", "17"): "17 .. FirstIntentDate",
        ("FIX.5.0", "1982", "18"): "18 .. LastIntentDate",
        ("FIX.5.0", "1982", "19"): "19 .. PositionRemovalDate",
        ("FIX.5.0", "1982", "20"): "20 .. MinimumNotice",
        ("FIX.5.0", "1982", "21"): "21 .. DeliveryStartTime",
        ("FIX.5.0", "1982", "22"): "22 .. DeliveryEndTime",
        ("FIX.5.0", "1982", "23"): "23 .. FirstNoticeDate",
        ("FIX.5.0", "1982", "24"): "24 .. LastNoticeDate",
        ("FIX.5.0", "1982", "25"): "25 .. FirstExerciseDate",
        ("FIX.5.0", "1982", "26"): "26 .. RedemptionDate",
        ("FIX.5.0", "1982", "27"): "27 .. TrdCntntnEfctvDt",
        ("FIX.5.0", "1982", "99"): "99 .. Other",
        ("FIX.5.0", "1989", "0"): "0 .. Zero",
        ("FIX.5.0", "1989", "1"): "1 .. FixedRate",
        ("FIX.5.0", "1989", "2"): "2 .. FloatingRate",
        ("FIX.5.0", "1989", "3"): "3 .. Structured",
        ("FIX.5.0", "1992", "D"): "D .. Day",
        ("FIX.5.0", "1992", "Wk"): "Wk .. Week",
        ("FIX.5.0", "1992", "Mo"): "Mo .. Month",
        ("FIX.5.0", "1992", "Yr"): "Yr .. Year",
        ("FIX.5.0", "1992", "H"): "H .. Hour",
        ("FIX.5.0", "1992", "Min"): "Min .. Minute",
        ("FIX.5.0", "1992", "S"): "S .. Second",
        ("FIX.5.0", "1992", "T"): "T .. Term",
        ("FIX.5.0", "1993", "0"): "0 .. OneOne",
        ("FIX.5.0", "1993", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "1993", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "1993", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "1993", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "1993", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "1993", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "1993", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "1993", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "1993", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "1993", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "1993", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "1993", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "1993", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "1993", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "1993", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "1993", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "1993", "17"): "17 .. Act364",
        ("FIX.5.0", "1995", "1"): "1 .. CUSIP",
        ("FIX.5.0", "1995", "2"): "2 .. SEDOL",
        ("FIX.5.0", "1995", "3"): "3 .. QUIK",
        ("FIX.5.0", "1995", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "1995", "5"): "5 .. RICCode",
        ("FIX.5.0", "1995", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "1995", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "1995", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "1995", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "1995", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "1995", "B"): "B .. Wertpapier",
        ("FIX.5.0", "1995", "C"): "C .. Dutch",
        ("FIX.5.0", "1995", "D"): "D .. Valoren",
        ("FIX.5.0", "1995", "E"): "E .. Sicovam",
        ("FIX.5.0", "1995", "F"): "F .. Belgian",
        ("FIX.5.0", "1995", "G"): "G .. Common",
        ("FIX.5.0", "1995", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "1995", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "1995", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "1995", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "1995", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "1995", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "1995", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "1995", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "1995", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "1995", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "1995", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "1995", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "1995", "U"): "U .. Synthetic",
        ("FIX.5.0", "1995", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "1995", "W"): "W .. IndexName",
        ("FIX.5.0", "1997", "1"): "1 .. CUSIP",
        ("FIX.5.0", "1997", "2"): "2 .. SEDOL",
        ("FIX.5.0", "1997", "3"): "3 .. QUIK",
        ("FIX.5.0", "1997", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "1997", "5"): "5 .. RICCode",
        ("FIX.5.0", "1997", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "1997", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "1997", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "1997", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "1997", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "1997", "B"): "B .. Wertpapier",
        ("FIX.5.0", "1997", "C"): "C .. Dutch",
        ("FIX.5.0", "1997", "D"): "D .. Valoren",
        ("FIX.5.0", "1997", "E"): "E .. Sicovam",
        ("FIX.5.0", "1997", "F"): "F .. Belgian",
        ("FIX.5.0", "1997", "G"): "G .. Common",
        ("FIX.5.0", "1997", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "1997", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "1997", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "1997", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "1997", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "1997", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "1997", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "1997", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "1997", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "1997", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "1997", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "1997", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "1997", "U"): "U .. Synthetic",
        ("FIX.5.0", "1997", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "1997", "W"): "W .. IndexName",
        ("FIX.5.0", "1998", "0"): "0 .. Unknown",
        ("FIX.5.0", "1998", "1"): "1 .. FirstLien",
        ("FIX.5.0", "1998", "2"): "2 .. SecondLien",
        ("FIX.5.0", "1998", "3"): "3 .. ThirdLien",
        ("FIX.5.0", "1999", "0"): "0 .. BridgeLoan",
        ("FIX.5.0", "1999", "1"): "1 .. LetterOfCredit",
        ("FIX.5.0", "1999", "2"): "2 .. RevolvingLoan",
        ("FIX.5.0", "1999", "3"): "3 .. SwinglineFunding",
        ("FIX.5.0", "1999", "4"): "4 .. TermLoan",
        ("FIX.5.0", "1999", "5"): "5 .. TradeClaim",
        ("FIX.5.0", "2000", "1"): "1 .. Asian",
        ("FIX.5.0", "2000", "2"): "2 .. AustralianNewZealand",
        ("FIX.5.0", "2000", "3"): "3 .. EuropeanEmergingMarkets",
        ("FIX.5.0", "2000", "4"): "4 .. Japanese",
        ("FIX.5.0", "2000", "5"): "5 .. NorthAmericanHighYield",
        ("FIX.5.0", "2000", "6"): "6 .. NorthAmericanInsurance",
        ("FIX.5.0", "2000", "7"): "7 .. NorthAmericanInvestmentGrade",
        ("FIX.5.0", "2000", "8"): "8 .. Singaporean",
        ("FIX.5.0", "2000", "9"): "9 .. WesternEuropean",
        ("FIX.5.0", "2000", "10"): "10 .. WesternEuropeanInsurance",
        ("FIX.5.0", "2010", "P"): "P .. ProRata",
        ("FIX.5.0", "2010", "R"): "R .. Random",
        ("FIX.5.0", "2011", "1"): "1 .. Active",
        ("FIX.5.0", "2011", "2"): "2 .. Inactive",
        ("FIX.5.0", "2011", "3"): "3 .. ActiveClosingOrdersOnly",
        ("FIX.5.0", "2011", "4"): "4 .. Expired",
        ("FIX.5.0", "2011", "5"): "5 .. Delisted",
        ("FIX.5.0", "2011", "6"): "6 .. KnockedOut",
        ("FIX.5.0", "2011", "7"): "7 .. KnockOutRevoked",
        ("FIX.5.0", "2011", "8"): "8 .. PendingExpiry",
        ("FIX.5.0", "2011", "9"): "9 .. Suspended",
        ("FIX.5.0", "2011", "10"): "10 .. Published",
        ("FIX.5.0", "2011", "11"): "11 .. PendingDeletion",
        ("FIX.5.0", "2012", "0"): "0 .. Bond",
        ("FIX.5.0", "2012", "1"): "1 .. ConvertibleBond",
        ("FIX.5.0", "2012", "2"): "2 .. Mortgage",
        ("FIX.5.0", "2012", "3"): "3 .. Loan",
        ("FIX.5.0", "2013", "1"): "1 .. InterestRate",
        ("FIX.5.0", "2013", "2"): "2 .. Currency",
        ("FIX.5.0", "2013", "3"): "3 .. Credit",
        ("FIX.5.0", "2013", "4"): "4 .. Equity",
        ("FIX.5.0", "2013", "5"): "5 .. Commodity",
        ("FIX.5.0", "2013", "6"): "6 .. Other",
        ("FIX.5.0", "2013", "7"): "7 .. Cash",
        ("FIX.5.0", "2013", "8"): "8 .. Debt",
        ("FIX.5.0", "2013", "9"): "9 .. Fund",
        ("FIX.5.0", "2013", "10"): "10 .. LoanFacility",
        ("FIX.5.0", "2014", "13"): "13 .. Metals",
        ("FIX.5.0", "2014", "14"): "14 .. Bullion",
        ("FIX.5.0", "2014", "15"): "15 .. Energy",
        ("FIX.5.0", "2014", "16"): "16 .. CommodityIndex",
        ("FIX.5.0", "2014", "17"): "17 .. Agricultural",
        ("FIX.5.0", "2014", "18"): "18 .. Environmental",
        ("FIX.5.0", "2014", "19"): "19 .. Freight",
        ("FIX.5.0", "2014", "41"): "41 .. Fertilizer",
        ("FIX.5.0", "2014", "42"): "42 .. IndustrialProduct",
        ("FIX.5.0", "2014", "43"): "43 .. Inflation",
        ("FIX.5.0", "2014", "44"): "44 .. Paper",
        ("FIX.5.0", "2014", "45"): "45 .. Polypropylene",
        ("FIX.5.0", "2014", "46"): "46 .. OfficialEconomicStatistics",
        ("FIX.5.0", "2014", "4"): "4 .. SingleName",
        ("FIX.5.0", "2014", "5"): "5 .. CreditIndex",
        ("FIX.5.0", "2014", "6"): "6 .. IndexTranche",
        ("FIX.5.0", "2014", "7"): "7 .. CreditBasket",
        ("FIX.5.0", "2014", "3"): "3 .. Basket",
        ("FIX.5.0", "2014", "38"): "38 .. FXCrossRates",
        ("FIX.5.0", "2014", "39"): "39 .. FXEmergingMarkets",
        ("FIX.5.0", "2014", "40"): "40 .. FXMajors",
        ("FIX.5.0", "2014", "20"): "20 .. Government",
        ("FIX.5.0", "2014", "21"): "21 .. Agency",
        ("FIX.5.0", "2014", "22"): "22 .. Corporate",
        ("FIX.5.0", "2014", "23"): "23 .. Financing",
        ("FIX.5.0", "2014", "24"): "24 .. MoneyMarket",
        ("FIX.5.0", "2014", "25"): "25 .. Mortgage",
        ("FIX.5.0", "2014", "26"): "26 .. Municipal",
        ("FIX.5.0", "2014", "9"): "9 .. Common",
        ("FIX.5.0", "2014", "10"): "10 .. Preferred",
        ("FIX.5.0", "2014", "11"): "11 .. EquityIndex",
        ("FIX.5.0", "2014", "12"): "12 .. EquityBasket",
        ("FIX.5.0", "2014", "34"): "34 .. DividendIndex",
        ("FIX.5.0", "2014", "35"): "35 .. StockDividend",
        ("FIX.5.0", "2014", "36"): "36 .. ExchangeTradedFund",
        ("FIX.5.0", "2014", "37"): "37 .. VolatilityIndex",
        ("FIX.5.0", "2014", "27"): "27 .. MutualFund",
        ("FIX.5.0", "2014", "28"): "28 .. CollectiveInvestmentVehicle",
        ("FIX.5.0", "2014", "29"): "29 .. InvestmentProgram",
        ("FIX.5.0", "2014", "30"): "30 .. SpecializedAccountProgram",
        ("FIX.5.0", "2014", "1"): "1 .. SingleCurrency",
        ("FIX.5.0", "2014", "2"): "2 .. CrossCurrency",
        ("FIX.5.0", "2014", "31"): "31 .. TermLoan",
        ("FIX.5.0", "2014", "32"): "32 .. BridgeLoan",
        ("FIX.5.0", "2014", "33"): "33 .. LetterOfCredit",
        ("FIX.5.0", "2014", "8"): "8 .. Exotic",
        ("FIX.5.0", "2014", "47"): "47 .. Other",
        ("FIX.5.0", "2014", "48"): "48 .. Other",
        ("FIX.5.0", "2016", "BS"): "BS .. BasisSwap",
        ("FIX.5.0", "2016", "IX"): "IX .. IndexSwap",
        ("FIX.5.0", "2016", "BB"): "BB .. BroadBasedSecuritySwap",
        ("FIX.5.0", "2016", "SK"): "SK .. BasketSwap",
        ("FIX.5.0", "2023", "1"): "1 .. FixedStrike",
        ("FIX.5.0", "2023", "2"): "2 .. StrikeSetAtExpiration",
        ("FIX.5.0", "2023", "3"): "3 .. StrikeSetToAverageAcrossLife",
        ("FIX.5.0", "2023", "4"): "4 .. StrikeSetToOptimalValue",
        ("FIX.5.0", "2024", "1"): "1 .. LessThan",
        ("FIX.5.0", "2024", "2"): "2 .. LessThanOrEqual",
        ("FIX.5.0", "2024", "3"): "3 .. Equal",
        ("FIX.5.0", "2024", "4"): "4 .. GreaterThanOrEqual",
        ("FIX.5.0", "2024", "5"): "5 .. GreaterThan",
        ("FIX.5.0", "2028", "1"): "1 .. Vanilla",
        ("FIX.5.0", "2028", "2"): "2 .. Capped",
        ("FIX.5.0", "2028", "3"): "3 .. Binary",
        ("FIX.5.0", "2028", "4"): "4 .. Asian",
        ("FIX.5.0", "2028", "5"): "5 .. Barrier",
        ("FIX.5.0", "2028", "6"): "6 .. DigitalBarrier",
        ("FIX.5.0", "2028", "7"): "7 .. Lookback",
        ("FIX.5.0", "2028", "8"): "8 .. OtherPathDependent",
        ("FIX.5.0", "2028", "99"): "99 .. Other",
        ("FIX.5.0", "2030", "STD"): "STD .. Standard",
        ("FIX.5.0", "2030", "INX"): "INX .. Index",
        ("FIX.5.0", "2030", "INT"): "INT .. InterestRateIndex",
        ("FIX.5.0", "2030", "PCTPAR"): "PCTPAR .. PercentOfPar",
        ("FIX.5.0", "2031", "EQTY"): "EQTY .. PremiumStyle",
        ("FIX.5.0", "2031", "FUT"): "FUT .. FuturesStyleMarkToMarket",
        ("FIX.5.0", "2031", "FUTDA"): "FUTDA .. FuturesStyleWithAnAttachedCashAdjustment",
        ("FIX.5.0", "2031", "CDS"): "CDS .. CDSStyleCollateralization",
        ("FIX.5.0", "2031", "CDSD"): "CDSD .. CDSInDeliveryUseRecoveryRateToCalculate",
        ("FIX.5.0", "2032", "0"): "0 .. PreListedOnly",
        ("FIX.5.0", "2032", "1"): "1 .. UserRequested",
        ("FIX.5.0", "2043", "0"): "0 .. NoRestrictions",
        ("FIX.5.0", "2043", "1"): "1 .. SecurityNotShortable",
        ("FIX.5.0", "2043", "2"): "2 .. SecurityNotShortableAtOrBelowBestBid",
        ("FIX.5.0", "2043", "3"): "3 .. SecurityNotShortableWithoutPreBorrow",
        ("FIX.5.0", "2046", "1"): "1 .. Capped",
        ("FIX.5.0", "2046", "2"): "2 .. Trigger",
        ("FIX.5.0", "2046", "3"): "3 .. KnockInUp",
        ("FIX.5.0", "2046", "4"): "4 .. KnockInDown",
        ("FIX.5.0", "2046", "5"): "5 .. KnockOutUp",
        ("FIX.5.0", "2046", "6"): "6 .. KnockOutDown",
        ("FIX.5.0", "2046", "7"): "7 .. Underlying",
        ("FIX.5.0", "2046", "8"): "8 .. ResetBarrier",
        ("FIX.5.0", "2046", "9"): "9 .. RollingBarrier",
        ("FIX.5.0", "2046", "10"): "10 .. OneTouch",
        ("FIX.5.0", "2046", "11"): "11 .. NoTouch",
        ("FIX.5.0", "2046", "12"): "12 .. DblOneTouch",
        ("FIX.5.0", "2046", "13"): "13 .. DblNoTouch",
        ("FIX.5.0", "2046", "14"): "14 .. FXComposite",
        ("FIX.5.0", "2046", "15"): "15 .. FXQuanto",
        ("FIX.5.0", "2046", "16"): "16 .. FXCrssCcy",
        ("FIX.5.0", "2046", "17"): "17 .. StrkSpread",
        ("FIX.5.0", "2046", "18"): "18 .. ClndrSpread",
        ("FIX.5.0", "2046", "19"): "19 .. PxObsvtn",
        ("FIX.5.0", "2046", "20"): "20 .. PassThrough",
        ("FIX.5.0", "2046", "21"): "21 .. StrkSched",
        ("FIX.5.0", "2046", "22"): "22 .. EquityValuation",
        ("FIX.5.0", "2046", "23"): "23 .. DividendValuation",
        ("FIX.5.0", "2049", "1"): "1 .. LessThanComplexEventPrice",
        ("FIX.5.0", "2049", "2"): "2 .. LessThanOrEqualToComplexEventPrice",
        ("FIX.5.0", "2049", "3"): "3 .. EqualToComplexEventPrice",
        ("FIX.5.0", "2049", "4"): "4 .. GreaterThanOrEqualToComplexEventPrice",
        ("FIX.5.0", "2049", "5"): "5 .. GreaterThanComplexEventPrice",
        ("FIX.5.0", "2051", "1"): "1 .. Expiration",
        ("FIX.5.0", "2051", "2"): "2 .. Immediate",
        ("FIX.5.0", "2051", "3"): "3 .. SpecifiedDate",
        ("FIX.5.0", "2051", "4"): "4 .. Close",
        ("FIX.5.0", "2051", "5"): "5 .. Open",
        ("FIX.5.0", "2051", "6"): "6 .. OfficialSettlPrice",
        ("FIX.5.0", "2051", "7"): "7 .. DerivativesClose",
        ("FIX.5.0", "2051", "8"): "8 .. AsSpecifiedMasterConfirmation",
        ("FIX.5.0", "2052", "1"): "1 .. And",
        ("FIX.5.0", "2052", "2"): "2 .. Or",
        ("FIX.5.0", "2060", "1"): "1 .. Put",
        ("FIX.5.0", "2060", "2"): "2 .. Call",
        ("FIX.5.0", "2060", "3"): "3 .. Tender",
        ("FIX.5.0", "2060", "4"): "4 .. SinkingFundCall",
        ("FIX.5.0", "2060", "5"): "5 .. Activation",
        ("FIX.5.0", "2060", "6"): "6 .. Inactiviation",
        ("FIX.5.0", "2060", "7"): "7 .. LastEligibleTradeDate",
        ("FIX.5.0", "2060", "8"): "8 .. SwapStartDate",
        ("FIX.5.0", "2060", "9"): "9 .. SwapEndDate",
        ("FIX.5.0", "2060", "10"): "10 .. SwapRollDate",
        ("FIX.5.0", "2060", "11"): "11 .. SwapNextStartDate",
        ("FIX.5.0", "2060", "12"): "12 .. SwapNextRollDate",
        ("FIX.5.0", "2060", "13"): "13 .. FirstDeliveryDate",
        ("FIX.5.0", "2060", "14"): "14 .. LastDeliveryDate",
        ("FIX.5.0", "2060", "15"): "15 .. InitialInventoryDueDate",
        ("FIX.5.0", "2060", "16"): "16 .. FinalInventoryDueDate",
        ("FIX.5.0", "2060", "17"): "17 .. FirstIntentDate",
        ("FIX.5.0", "2060", "18"): "18 .. LastIntentDate",
        ("FIX.5.0", "2060", "19"): "19 .. PositionRemovalDate",
        ("FIX.5.0", "2060", "20"): "20 .. MinimumNotice",
        ("FIX.5.0", "2060", "21"): "21 .. DeliveryStartTime",
        ("FIX.5.0", "2060", "22"): "22 .. DeliveryEndTime",
        ("FIX.5.0", "2060", "23"): "23 .. FirstNoticeDate",
        ("FIX.5.0", "2060", "24"): "24 .. LastNoticeDate",
        ("FIX.5.0", "2060", "25"): "25 .. FirstExerciseDate",
        ("FIX.5.0", "2060", "26"): "26 .. RedemptionDate",
        ("FIX.5.0", "2060", "27"): "27 .. TrdCntntnEfctvDt",
        ("FIX.5.0", "2060", "99"): "99 .. Other",
        ("FIX.5.0", "2063", "H"): "H .. Hour",
        ("FIX.5.0", "2063", "Min"): "Min .. Minute",
        ("FIX.5.0", "2063", "S"): "S .. Second",
        ("FIX.5.0", "2063", "D"): "D .. Day",
        ("FIX.5.0", "2063", "Wk"): "Wk .. Week",
        ("FIX.5.0", "2063", "Mo"): "Mo .. Month",
        ("FIX.5.0", "2063", "Yr"): "Yr .. Year",
        ("FIX.5.0", "2067", "1"): "1 .. InterestRate",
        ("FIX.5.0", "2067", "2"): "2 .. Currency",
        ("FIX.5.0", "2067", "3"): "3 .. Credit",
        ("FIX.5.0", "2067", "4"): "4 .. Equity",
        ("FIX.5.0", "2067", "5"): "5 .. Commodity",
        ("FIX.5.0", "2067", "6"): "6 .. Other",
        ("FIX.5.0", "2067", "7"): "7 .. Cash",
        ("FIX.5.0", "2067", "8"): "8 .. Debt",
        ("FIX.5.0", "2067", "9"): "9 .. Fund",
        ("FIX.5.0", "2067", "10"): "10 .. LoanFacility",
        ("FIX.5.0", "2068", "13"): "13 .. Metals",
        ("FIX.5.0", "2068", "14"): "14 .. Bullion",
        ("FIX.5.0", "2068", "15"): "15 .. Energy",
        ("FIX.5.0", "2068", "16"): "16 .. CommodityIndex",
        ("FIX.5.0", "2068", "17"): "17 .. Agricultural",
        ("FIX.5.0", "2068", "18"): "18 .. Environmental",
        ("FIX.5.0", "2068", "19"): "19 .. Freight",
        ("FIX.5.0", "2068", "41"): "41 .. Fertilizer",
        ("FIX.5.0", "2068", "42"): "42 .. IndustrialProduct",
        ("FIX.5.0", "2068", "43"): "43 .. Inflation",
        ("FIX.5.0", "2068", "44"): "44 .. Paper",
        ("FIX.5.0", "2068", "45"): "45 .. Polypropylene",
        ("FIX.5.0", "2068", "46"): "46 .. OfficialEconomicStatistics",
        ("FIX.5.0", "2068", "4"): "4 .. SingleName",
        ("FIX.5.0", "2068", "5"): "5 .. CreditIndex",
        ("FIX.5.0", "2068", "6"): "6 .. IndexTranche",
        ("FIX.5.0", "2068", "7"): "7 .. CreditBasket",
        ("FIX.5.0", "2068", "3"): "3 .. Basket",
        ("FIX.5.0", "2068", "38"): "38 .. FXCrossRates",
        ("FIX.5.0", "2068", "39"): "39 .. FXEmergingMarkets",
        ("FIX.5.0", "2068", "40"): "40 .. FXMajors",
        ("FIX.5.0", "2068", "20"): "20 .. Government",
        ("FIX.5.0", "2068", "21"): "21 .. Agency",
        ("FIX.5.0", "2068", "22"): "22 .. Corporate",
        ("FIX.5.0", "2068", "23"): "23 .. Financing",
        ("FIX.5.0", "2068", "24"): "24 .. MoneyMarket",
        ("FIX.5.0", "2068", "25"): "25 .. Mortgage",
        ("FIX.5.0", "2068", "26"): "26 .. Municipal",
        ("FIX.5.0", "2068", "9"): "9 .. Common",
        ("FIX.5.0", "2068", "10"): "10 .. Preferred",
        ("FIX.5.0", "2068", "11"): "11 .. EquityIndex",
        ("FIX.5.0", "2068", "12"): "12 .. EquityBasket",
        ("FIX.5.0", "2068", "34"): "34 .. DividendIndex",
        ("FIX.5.0", "2068", "35"): "35 .. StockDividend",
        ("FIX.5.0", "2068", "36"): "36 .. ExchangeTradedFund",
        ("FIX.5.0", "2068", "37"): "37 .. VolatilityIndex",
        ("FIX.5.0", "2068", "27"): "27 .. MutualFund",
        ("FIX.5.0", "2068", "28"): "28 .. CollectiveInvestmentVehicle",
        ("FIX.5.0", "2068", "29"): "29 .. InvestmentProgram",
        ("FIX.5.0", "2068", "30"): "30 .. SpecializedAccountProgram",
        ("FIX.5.0", "2068", "1"): "1 .. SingleCurrency",
        ("FIX.5.0", "2068", "2"): "2 .. CrossCurrency",
        ("FIX.5.0", "2068", "31"): "31 .. TermLoan",
        ("FIX.5.0", "2068", "32"): "32 .. BridgeLoan",
        ("FIX.5.0", "2068", "33"): "33 .. LetterOfCredit",
        ("FIX.5.0", "2068", "8"): "8 .. Exotic",
        ("FIX.5.0", "2068", "47"): "47 .. Other",
        ("FIX.5.0", "2068", "48"): "48 .. Other",
        ("FIX.5.0", "2070", "BS"): "BS .. BasisSwap",
        ("FIX.5.0", "2070", "IX"): "IX .. IndexSwap",
        ("FIX.5.0", "2070", "BB"): "BB .. BroadBasedSecuritySwap",
        ("FIX.5.0", "2070", "SK"): "SK .. BasketSwap",
        ("FIX.5.0", "2077", "1"): "1 .. InterestRate",
        ("FIX.5.0", "2077", "2"): "2 .. Currency",
        ("FIX.5.0", "2077", "3"): "3 .. Credit",
        ("FIX.5.0", "2077", "4"): "4 .. Equity",
        ("FIX.5.0", "2077", "5"): "5 .. Commodity",
        ("FIX.5.0", "2077", "6"): "6 .. Other",
        ("FIX.5.0", "2077", "7"): "7 .. Cash",
        ("FIX.5.0", "2077", "8"): "8 .. Debt",
        ("FIX.5.0", "2077", "9"): "9 .. Fund",
        ("FIX.5.0", "2077", "10"): "10 .. LoanFacility",
        ("FIX.5.0", "2078", "13"): "13 .. Metals",
        ("FIX.5.0", "2078", "14"): "14 .. Bullion",
        ("FIX.5.0", "2078", "15"): "15 .. Energy",
        ("FIX.5.0", "2078", "16"): "16 .. CommodityIndex",
        ("FIX.5.0", "2078", "17"): "17 .. Agricultural",
        ("FIX.5.0", "2078", "18"): "18 .. Environmental",
        ("FIX.5.0", "2078", "19"): "19 .. Freight",
        ("FIX.5.0", "2078", "41"): "41 .. Fertilizer",
        ("FIX.5.0", "2078", "42"): "42 .. IndustrialProduct",
        ("FIX.5.0", "2078", "43"): "43 .. Inflation",
        ("FIX.5.0", "2078", "44"): "44 .. Paper",
        ("FIX.5.0", "2078", "45"): "45 .. Polypropylene",
        ("FIX.5.0", "2078", "46"): "46 .. OfficialEconomicStatistics",
        ("FIX.5.0", "2078", "4"): "4 .. SingleName",
        ("FIX.5.0", "2078", "5"): "5 .. CreditIndex",
        ("FIX.5.0", "2078", "6"): "6 .. IndexTranche",
        ("FIX.5.0", "2078", "7"): "7 .. CreditBasket",
        ("FIX.5.0", "2078", "3"): "3 .. Basket",
        ("FIX.5.0", "2078", "38"): "38 .. FXCrossRates",
        ("FIX.5.0", "2078", "39"): "39 .. FXEmergingMarkets",
        ("FIX.5.0", "2078", "40"): "40 .. FXMajors",
        ("FIX.5.0", "2078", "20"): "20 .. Government",
        ("FIX.5.0", "2078", "21"): "21 .. Agency",
        ("FIX.5.0", "2078", "22"): "22 .. Corporate",
        ("FIX.5.0", "2078", "23"): "23 .. Financing",
        ("FIX.5.0", "2078", "24"): "24 .. MoneyMarket",
        ("FIX.5.0", "2078", "25"): "25 .. Mortgage",
        ("FIX.5.0", "2078", "26"): "26 .. Municipal",
        ("FIX.5.0", "2078", "9"): "9 .. Common",
        ("FIX.5.0", "2078", "10"): "10 .. Preferred",
        ("FIX.5.0", "2078", "11"): "11 .. EquityIndex",
        ("FIX.5.0", "2078", "12"): "12 .. EquityBasket",
        ("FIX.5.0", "2078", "34"): "34 .. DividendIndex",
        ("FIX.5.0", "2078", "35"): "35 .. StockDividend",
        ("FIX.5.0", "2078", "36"): "36 .. ExchangeTradedFund",
        ("FIX.5.0", "2078", "37"): "37 .. VolatilityIndex",
        ("FIX.5.0", "2078", "27"): "27 .. MutualFund",
        ("FIX.5.0", "2078", "28"): "28 .. CollectiveInvestmentVehicle",
        ("FIX.5.0", "2078", "29"): "29 .. InvestmentProgram",
        ("FIX.5.0", "2078", "30"): "30 .. SpecializedAccountProgram",
        ("FIX.5.0", "2078", "1"): "1 .. SingleCurrency",
        ("FIX.5.0", "2078", "2"): "2 .. CrossCurrency",
        ("FIX.5.0", "2078", "31"): "31 .. TermLoan",
        ("FIX.5.0", "2078", "32"): "32 .. BridgeLoan",
        ("FIX.5.0", "2078", "33"): "33 .. LetterOfCredit",
        ("FIX.5.0", "2078", "8"): "8 .. Exotic",
        ("FIX.5.0", "2078", "47"): "47 .. Other",
        ("FIX.5.0", "2078", "48"): "48 .. Other",
        ("FIX.5.0", "2081", "1"): "1 .. InterestRate",
        ("FIX.5.0", "2081", "2"): "2 .. Currency",
        ("FIX.5.0", "2081", "3"): "3 .. Credit",
        ("FIX.5.0", "2081", "4"): "4 .. Equity",
        ("FIX.5.0", "2081", "5"): "5 .. Commodity",
        ("FIX.5.0", "2081", "6"): "6 .. Other",
        ("FIX.5.0", "2081", "7"): "7 .. Cash",
        ("FIX.5.0", "2081", "8"): "8 .. Debt",
        ("FIX.5.0", "2081", "9"): "9 .. Fund",
        ("FIX.5.0", "2081", "10"): "10 .. LoanFacility",
        ("FIX.5.0", "2082", "13"): "13 .. Metals",
        ("FIX.5.0", "2082", "14"): "14 .. Bullion",
        ("FIX.5.0", "2082", "15"): "15 .. Energy",
        ("FIX.5.0", "2082", "16"): "16 .. CommodityIndex",
        ("FIX.5.0", "2082", "17"): "17 .. Agricultural",
        ("FIX.5.0", "2082", "18"): "18 .. Environmental",
        ("FIX.5.0", "2082", "19"): "19 .. Freight",
        ("FIX.5.0", "2082", "41"): "41 .. Fertilizer",
        ("FIX.5.0", "2082", "42"): "42 .. IndustrialProduct",
        ("FIX.5.0", "2082", "43"): "43 .. Inflation",
        ("FIX.5.0", "2082", "44"): "44 .. Paper",
        ("FIX.5.0", "2082", "45"): "45 .. Polypropylene",
        ("FIX.5.0", "2082", "46"): "46 .. OfficialEconomicStatistics",
        ("FIX.5.0", "2082", "4"): "4 .. SingleName",
        ("FIX.5.0", "2082", "5"): "5 .. CreditIndex",
        ("FIX.5.0", "2082", "6"): "6 .. IndexTranche",
        ("FIX.5.0", "2082", "7"): "7 .. CreditBasket",
        ("FIX.5.0", "2082", "3"): "3 .. Basket",
        ("FIX.5.0", "2082", "38"): "38 .. FXCrossRates",
        ("FIX.5.0", "2082", "39"): "39 .. FXEmergingMarkets",
        ("FIX.5.0", "2082", "40"): "40 .. FXMajors",
        ("FIX.5.0", "2082", "20"): "20 .. Government",
        ("FIX.5.0", "2082", "21"): "21 .. Agency",
        ("FIX.5.0", "2082", "22"): "22 .. Corporate",
        ("FIX.5.0", "2082", "23"): "23 .. Financing",
        ("FIX.5.0", "2082", "24"): "24 .. MoneyMarket",
        ("FIX.5.0", "2082", "25"): "25 .. Mortgage",
        ("FIX.5.0", "2082", "26"): "26 .. Municipal",
        ("FIX.5.0", "2082", "9"): "9 .. Common",
        ("FIX.5.0", "2082", "10"): "10 .. Preferred",
        ("FIX.5.0", "2082", "11"): "11 .. EquityIndex",
        ("FIX.5.0", "2082", "12"): "12 .. EquityBasket",
        ("FIX.5.0", "2082", "34"): "34 .. DividendIndex",
        ("FIX.5.0", "2082", "35"): "35 .. StockDividend",
        ("FIX.5.0", "2082", "36"): "36 .. ExchangeTradedFund",
        ("FIX.5.0", "2082", "37"): "37 .. VolatilityIndex",
        ("FIX.5.0", "2082", "27"): "27 .. MutualFund",
        ("FIX.5.0", "2082", "28"): "28 .. CollectiveInvestmentVehicle",
        ("FIX.5.0", "2082", "29"): "29 .. InvestmentProgram",
        ("FIX.5.0", "2082", "30"): "30 .. SpecializedAccountProgram",
        ("FIX.5.0", "2082", "1"): "1 .. SingleCurrency",
        ("FIX.5.0", "2082", "2"): "2 .. CrossCurrency",
        ("FIX.5.0", "2082", "31"): "31 .. TermLoan",
        ("FIX.5.0", "2082", "32"): "32 .. BridgeLoan",
        ("FIX.5.0", "2082", "33"): "33 .. LetterOfCredit",
        ("FIX.5.0", "2082", "8"): "8 .. Exotic",
        ("FIX.5.0", "2082", "47"): "47 .. Other",
        ("FIX.5.0", "2082", "48"): "48 .. Other",
        ("FIX.5.0", "2089", "D"): "D .. Divide",
        ("FIX.5.0", "2089", "M"): "M .. Multiply",
        ("FIX.5.0", "2091", "D"): "D .. Divide",
        ("FIX.5.0", "2091", "M"): "M .. Multiply",
        ("FIX.5.0", "2095", "D"): "D .. Divide",
        ("FIX.5.0", "2095", "M"): "M .. Multiply",
        ("FIX.5.0", "2098", "D"): "D .. Divide",
        ("FIX.5.0", "2098", "M"): "M .. Multiply",
        ("FIX.5.0", "2109", "0"): "0 .. Base64",
        ("FIX.5.0", "2109", "1"): "1 .. RawBinary",
        ("FIX.5.0", "2115", "0"): "0 .. AutoSpot",
        ("FIX.5.0", "2115", "1"): "1 .. NegotiatedSpot",
        ("FIX.5.0", "2115", "2"): "2 .. PhoneSpot",
        ("FIX.5.0", "2117", "1"): "1 .. Buy",
        ("FIX.5.0", "2117", "2"): "2 .. Sell",
        ("FIX.5.0", "2118", "1"): "1 .. Buy",
        ("FIX.5.0", "2118", "2"): "2 .. Sell",
        ("FIX.5.0", "2121", "0"): "0 .. Close",
        ("FIX.5.0", "2121", "1"): "1 .. Open",
        ("FIX.5.0", "2121", "2"): "2 .. OfficialSettl",
        ("FIX.5.0", "2121", "3"): "3 .. ValuationTime",
        ("FIX.5.0", "2121", "4"): "4 .. ExcahgneSettlTime",
        ("FIX.5.0", "2121", "5"): "5 .. DerivativesClose",
        ("FIX.5.0", "2121", "6"): "6 .. AsSpecified",
        ("FIX.5.0", "2126", "0"): "0 .. Currency1PerCurrency2",
        ("FIX.5.0", "2126", "1"): "1 .. Currency2PerCurrency1",
        ("FIX.5.0", "2129", "0"): "0 .. ExercisingParty",
        ("FIX.5.0", "2129", "1"): "1 .. NonExercisingParty",
        ("FIX.5.0", "2129", "2"): "2 .. MasterAgreeent",
        ("FIX.5.0", "2129", "3"): "3 .. Supplement",
        ("FIX.5.0", "2134", "0"): "0 .. SellerNotifies",
        ("FIX.5.0", "2134", "1"): "1 .. BuyerNotifies",
        ("FIX.5.0", "2134", "2"): "2 .. SellerOrBuyerNotifies",
        ("FIX.5.0", "2141", "STD"): "STD .. Straddle",
        ("FIX.5.0", "2141", "STG"): "STG .. Strangle",
        ("FIX.5.0", "2141", "BF"): "BF .. Butterfly",
        ("FIX.5.0", "2141", "CNDR"): "CNDR .. Condor",
        ("FIX.5.0", "2141", "CISN"): "CISN .. CallableInversibleSnowball",
        ("FIX.5.0", "2141", "OTHER"): "OTHER .. Other",
        ("FIX.5.0", "2143", "1"): "1 .. Negotiation",
        ("FIX.5.0", "2143", "2"): "2 .. Cancellation",
        ("FIX.5.0", "2144", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "2144", "1"): "1 .. RoundDown",
        ("FIX.5.0", "2144", "2"): "2 .. RoundUp",
        ("FIX.5.0", "2147", "P"): "P .. ProRata",
        ("FIX.5.0", "2147", "R"): "R .. Random",
        ("FIX.5.0", "2148", "1"): "1 .. Active",
        ("FIX.5.0", "2148", "2"): "2 .. Inactive",
        ("FIX.5.0", "2148", "3"): "3 .. ActiveClosingOrdersOnly",
        ("FIX.5.0", "2148", "4"): "4 .. Expired",
        ("FIX.5.0", "2148", "5"): "5 .. Delisted",
        ("FIX.5.0", "2148", "6"): "6 .. KnockedOut",
        ("FIX.5.0", "2148", "7"): "7 .. KnockOutRevoked",
        ("FIX.5.0", "2148", "8"): "8 .. PendingExpiry",
        ("FIX.5.0", "2148", "9"): "9 .. Suspended",
        ("FIX.5.0", "2148", "10"): "10 .. Published",
        ("FIX.5.0", "2148", "11"): "11 .. PendingDeletion",
        ("FIX.5.0", "2149", "FR"): "FR .. FullRestructuring",
        ("FIX.5.0", "2149", "MR"): "MR .. ModifiedRestructuring",
        ("FIX.5.0", "2149", "MM"): "MM .. ModifiedModRestructuring",
        ("FIX.5.0", "2149", "XR"): "XR .. NoRestructuringSpecified",
        ("FIX.5.0", "2150", "SD"): "SD .. SeniorSecured",
        ("FIX.5.0", "2150", "SR"): "SR .. Senior",
        ("FIX.5.0", "2150", "SB"): "SB .. Subordinated",
        ("FIX.5.0", "2150", "JR"): "JR .. Junior",
        ("FIX.5.0", "2150", "MZ"): "MZ .. Mezzanine",
        ("FIX.5.0", "2150", "SN"): "SN .. SeniorNonPreferred",
        ("FIX.5.0", "2155", "0"): "0 .. Bond",
        ("FIX.5.0", "2155", "1"): "1 .. ConvertBond",
        ("FIX.5.0", "2155", "2"): "2 .. Mortgage",
        ("FIX.5.0", "2155", "3"): "3 .. Loan",
        ("FIX.5.0", "2156", "AMTZ"): "AMTZ .. Amortizing",
        ("FIX.5.0", "2156", "CNST"): "CNST .. ConstantNotionalSchedule",
        ("FIX.5.0", "2156", "ACRT"): "ACRT .. AccretingNotionalSchedule",
        ("FIX.5.0", "2156", "CUST"): "CUST .. CustomNotionalSchedule",
        ("FIX.5.0", "2161", "0"): "0 .. Zero",
        ("FIX.5.0", "2161", "1"): "1 .. FixedRate",
        ("FIX.5.0", "2161", "2"): "2 .. FloatingRate",
        ("FIX.5.0", "2161", "3"): "3 .. Structured",
        ("FIX.5.0", "2164", "D"): "D .. Day",
        ("FIX.5.0", "2164", "Wk"): "Wk .. Week",
        ("FIX.5.0", "2164", "Mo"): "Mo .. Month",
        ("FIX.5.0", "2164", "Yr"): "Yr .. Year",
        ("FIX.5.0", "2164", "H"): "H .. Hour",
        ("FIX.5.0", "2164", "Min"): "Min .. Minute",
        ("FIX.5.0", "2164", "S"): "S .. Second",
        ("FIX.5.0", "2164", "T"): "T .. Term",
        ("FIX.5.0", "2165", "0"): "0 .. OneOne",
        ("FIX.5.0", "2165", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "2165", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "2165", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "2165", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "2165", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "2165", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "2165", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "2165", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "2165", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "2165", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "2165", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "2165", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "2165", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "2165", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "2165", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "2165", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "2165", "17"): "17 .. Act364",
        ("FIX.5.0", "2167", "1"): "1 .. CUSIP",
        ("FIX.5.0", "2167", "2"): "2 .. SEDOL",
        ("FIX.5.0", "2167", "3"): "3 .. QUIK",
        ("FIX.5.0", "2167", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "2167", "5"): "5 .. RICCode",
        ("FIX.5.0", "2167", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "2167", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "2167", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "2167", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "2167", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "2167", "B"): "B .. Wertpapier",
        ("FIX.5.0", "2167", "C"): "C .. Dutch",
        ("FIX.5.0", "2167", "D"): "D .. Valoren",
        ("FIX.5.0", "2167", "E"): "E .. Sicovam",
        ("FIX.5.0", "2167", "F"): "F .. Belgian",
        ("FIX.5.0", "2167", "G"): "G .. Common",
        ("FIX.5.0", "2167", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "2167", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "2167", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "2167", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "2167", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "2167", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "2167", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "2167", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "2167", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "2167", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "2167", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "2167", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "2167", "U"): "U .. Synthetic",
        ("FIX.5.0", "2167", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "2167", "W"): "W .. IndexName",
        ("FIX.5.0", "2169", "0"): "0 .. Unknown",
        ("FIX.5.0", "2169", "1"): "1 .. FirstLien",
        ("FIX.5.0", "2169", "2"): "2 .. SecondLien",
        ("FIX.5.0", "2169", "3"): "3 .. ThirdLien",
        ("FIX.5.0", "2170", "0"): "0 .. BridgeLoan",
        ("FIX.5.0", "2170", "1"): "1 .. LetterOfCredit",
        ("FIX.5.0", "2170", "2"): "2 .. RevolvingLoan",
        ("FIX.5.0", "2170", "3"): "3 .. SwinglineFunding",
        ("FIX.5.0", "2170", "4"): "4 .. TermLoan",
        ("FIX.5.0", "2170", "5"): "5 .. TradeClaim",
        ("FIX.5.0", "2171", "1"): "1 .. Asian",
        ("FIX.5.0", "2171", "2"): "2 .. AustralianNewZealand",
        ("FIX.5.0", "2171", "3"): "3 .. EuropeanEmergingMarkets",
        ("FIX.5.0", "2171", "4"): "4 .. Japanese",
        ("FIX.5.0", "2171", "5"): "5 .. NorthAmericanHighYield",
        ("FIX.5.0", "2171", "6"): "6 .. NorthAmericanInsurance",
        ("FIX.5.0", "2171", "7"): "7 .. NorthAmericanInvestmentGrade",
        ("FIX.5.0", "2171", "8"): "8 .. Singaporean",
        ("FIX.5.0", "2171", "9"): "9 .. WesternEuropean",
        ("FIX.5.0", "2171", "10"): "10 .. WesternEuropeanInsurance",
        ("FIX.5.0", "2183", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "2183", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "2183", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "2183", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "2183", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "2183", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "2183", "thm"): "thm .. therms",
        ("FIX.5.0", "2183", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "2183", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "2183", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "2183", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "2183", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "2183", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "2183", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "2183", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "2183", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "2183", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "2183", "day"): "day .. Day",
        ("FIX.5.0", "2183", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "2183", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "2183", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "2183", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "2183", "g"): "g .. Grams",
        ("FIX.5.0", "2183", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "2183", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "2183", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "2183", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "2183", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "2183", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "2183", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "2183", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "2183", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "2183", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "2183", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "2183", "L"): "L .. liters",
        ("FIX.5.0", "2183", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "2183", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "2183", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "2183", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "2183", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "2183", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "2183", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "2183", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "2183", "t"): "t .. MetricTons",
        ("FIX.5.0", "2183", "tn"): "tn .. Tons",
        ("FIX.5.0", "2186", "1"): "1 .. FixedStrike",
        ("FIX.5.0", "2186", "2"): "2 .. StrikeSetAtExpiration",
        ("FIX.5.0", "2186", "3"): "3 .. StrikeSetToAverageAcrossLife",
        ("FIX.5.0", "2186", "4"): "4 .. StrikeSetToOptimalValue",
        ("FIX.5.0", "2187", "1"): "1 .. LessThan",
        ("FIX.5.0", "2187", "2"): "2 .. LessThanOrEqual",
        ("FIX.5.0", "2187", "3"): "3 .. Equal",
        ("FIX.5.0", "2187", "4"): "4 .. GreaterThanOrEqual",
        ("FIX.5.0", "2187", "5"): "5 .. GreaterThan",
        ("FIX.5.0", "2189", "1"): "1 .. Regular",
        ("FIX.5.0", "2189", "2"): "2 .. SpecialReference",
        ("FIX.5.0", "2189", "3"): "3 .. OptimalValue",
        ("FIX.5.0", "2189", "4"): "4 .. AverageValue",
        ("FIX.5.0", "2192", "C"): "C .. CashSettlementRequired",
        ("FIX.5.0", "2192", "P"): "P .. PhysicalSettlementRequired",
        ("FIX.5.0", "2192", "E"): "E .. Election",
        ("FIX.5.0", "2193", "1"): "1 .. Vanilla",
        ("FIX.5.0", "2193", "2"): "2 .. Capped",
        ("FIX.5.0", "2193", "3"): "3 .. Binary",
        ("FIX.5.0", "2193", "4"): "4 .. Asian",
        ("FIX.5.0", "2193", "5"): "5 .. Barrier",
        ("FIX.5.0", "2193", "6"): "6 .. DigitalBarrier",
        ("FIX.5.0", "2193", "7"): "7 .. Lookback",
        ("FIX.5.0", "2193", "8"): "8 .. OtherPathDependent",
        ("FIX.5.0", "2193", "99"): "99 .. Other",
        ("FIX.5.0", "2195", "STD"): "STD .. Standard",
        ("FIX.5.0", "2195", "INX"): "INX .. Index",
        ("FIX.5.0", "2195", "INT"): "INT .. InterestRateIndex",
        ("FIX.5.0", "2195", "PCTPAR"): "PCTPAR .. PercentOfPar",
        ("FIX.5.0", "2196", "EQTY"): "EQTY .. PremiumStyle",
        ("FIX.5.0", "2196", "FUT"): "FUT .. FuturesStyleMarkToMarket",
        ("FIX.5.0", "2196", "FUTDA"): "FUTDA .. FuturesStyleWithAnAttachedCashAdjustment",
        ("FIX.5.0", "2196", "CDS"): "CDS .. CDSStyleCollateralization",
        ("FIX.5.0", "2196", "CDSD"): "CDSD .. CDSInDeliveryUseRecoveryRateToCalculate",
        ("FIX.5.0", "2199", "0"): "0 .. PreListedOnly",
        ("FIX.5.0", "2199", "1"): "1 .. UserRequested",
        ("FIX.5.0", "2207", "1"): "1 .. Program3a3",
        ("FIX.5.0", "2207", "2"): "2 .. Program42",
        ("FIX.5.0", "2207", "3"): "3 .. Program3a2",
        ("FIX.5.0", "2207", "4"): "4 .. Program3a3And3c7",
        ("FIX.5.0", "2207", "5"): "5 .. Program3a4",
        ("FIX.5.0", "2207", "6"): "6 .. Program3a5",
        ("FIX.5.0", "2207", "7"): "7 .. Program3a7",
        ("FIX.5.0", "2207", "8"): "8 .. Program3c7",
        ("FIX.5.0", "2207", "99"): "99 .. Other",
        ("FIX.5.0", "2209", "0"): "0 .. NoRestrictions",
        ("FIX.5.0", "2209", "1"): "1 .. SecurityNotShortable",
        ("FIX.5.0", "2209", "2"): "2 .. SecurityNotShortableAtOrBelowBestBid",
        ("FIX.5.0", "2209", "3"): "3 .. SecurityNotShortableWithoutPreBorrow",
        ("FIX.5.0", "2210", "1"): "1 .. Financials",
        ("FIX.5.0", "2210", "2"): "2 .. Commodities",
        ("FIX.5.0", "2210", "3"): "3 .. AlternativeInvestments",
        ("FIX.5.0", "2211", "STD"): "STD .. Straddle",
        ("FIX.5.0", "2211", "STG"): "STG .. Strangle",
        ("FIX.5.0", "2211", "BF"): "BF .. Butterfly",
        ("FIX.5.0", "2211", "CNDR"): "CNDR .. Condor",
        ("FIX.5.0", "2211", "CISN"): "CISN .. CallableInversibleSnowball",
        ("FIX.5.0", "2211", "OTHER"): "OTHER .. Other",
        ("FIX.5.0", "2213", "1"): "1 .. Negotiation",
        ("FIX.5.0", "2213", "2"): "2 .. Cancellation",
        ("FIX.5.0", "2214", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "2214", "1"): "1 .. RoundDown",
        ("FIX.5.0", "2214", "2"): "2 .. RoundUp",
        ("FIX.5.0", "2219", "1"): "1 .. Capped",
        ("FIX.5.0", "2219", "2"): "2 .. Trigger",
        ("FIX.5.0", "2219", "3"): "3 .. KnockInUp",
        ("FIX.5.0", "2219", "4"): "4 .. KnockInDown",
        ("FIX.5.0", "2219", "5"): "5 .. KnockOutUp",
        ("FIX.5.0", "2219", "6"): "6 .. KnockOutDown",
        ("FIX.5.0", "2219", "7"): "7 .. Underlying",
        ("FIX.5.0", "2219", "8"): "8 .. ResetBarrier",
        ("FIX.5.0", "2219", "9"): "9 .. RollingBarrier",
        ("FIX.5.0", "2219", "10"): "10 .. OneTouch",
        ("FIX.5.0", "2219", "11"): "11 .. NoTouch",
        ("FIX.5.0", "2219", "12"): "12 .. DblOneTouch",
        ("FIX.5.0", "2219", "13"): "13 .. DblNoTouch",
        ("FIX.5.0", "2219", "14"): "14 .. FXComposite",
        ("FIX.5.0", "2219", "15"): "15 .. FXQuanto",
        ("FIX.5.0", "2219", "16"): "16 .. FXCrssCcy",
        ("FIX.5.0", "2219", "17"): "17 .. StrkSpread",
        ("FIX.5.0", "2219", "18"): "18 .. ClndrSpread",
        ("FIX.5.0", "2219", "19"): "19 .. PxObsvtn",
        ("FIX.5.0", "2219", "20"): "20 .. PassThrough",
        ("FIX.5.0", "2219", "21"): "21 .. StrkSched",
        ("FIX.5.0", "2219", "22"): "22 .. EquityValuation",
        ("FIX.5.0", "2219", "23"): "23 .. DividendValuation",
        ("FIX.5.0", "2220", "1"): "1 .. Buy",
        ("FIX.5.0", "2220", "2"): "2 .. Sell",
        ("FIX.5.0", "2221", "1"): "1 .. Buy",
        ("FIX.5.0", "2221", "2"): "2 .. Sell",
        ("FIX.5.0", "2225", "0"): "0 .. Close",
        ("FIX.5.0", "2225", "1"): "1 .. Open",
        ("FIX.5.0", "2225", "2"): "2 .. OfficialSettl",
        ("FIX.5.0", "2225", "3"): "3 .. ValuationTime",
        ("FIX.5.0", "2225", "4"): "4 .. ExcahgneSettlTime",
        ("FIX.5.0", "2225", "5"): "5 .. DerivativesClose",
        ("FIX.5.0", "2225", "6"): "6 .. AsSpecified",
        ("FIX.5.0", "2229", "1"): "1 .. LessThanComplexEventPrice",
        ("FIX.5.0", "2229", "2"): "2 .. LessThanOrEqualToComplexEventPrice",
        ("FIX.5.0", "2229", "3"): "3 .. EqualToComplexEventPrice",
        ("FIX.5.0", "2229", "4"): "4 .. GreaterThanOrEqualToComplexEventPrice",
        ("FIX.5.0", "2229", "5"): "5 .. GreaterThanComplexEventPrice",
        ("FIX.5.0", "2231", "1"): "1 .. Expiration",
        ("FIX.5.0", "2231", "2"): "2 .. Immediate",
        ("FIX.5.0", "2231", "3"): "3 .. SpecifiedDate",
        ("FIX.5.0", "2231", "4"): "4 .. Close",
        ("FIX.5.0", "2231", "5"): "5 .. Open",
        ("FIX.5.0", "2231", "6"): "6 .. OfficialSettlPrice",
        ("FIX.5.0", "2231", "7"): "7 .. DerivativesClose",
        ("FIX.5.0", "2231", "8"): "8 .. AsSpecifiedMasterConfirmation",
        ("FIX.5.0", "2232", "1"): "1 .. And",
        ("FIX.5.0", "2232", "2"): "2 .. Or",
        ("FIX.5.0", "2235", "0"): "0 .. Currency1PerCurrency2",
        ("FIX.5.0", "2235", "1"): "1 .. Currency2PerCurrency1",
        ("FIX.5.0", "2238", "0"): "0 .. ExercisingParty",
        ("FIX.5.0", "2238", "1"): "1 .. NonExercisingParty",
        ("FIX.5.0", "2238", "2"): "2 .. MasterAgreeent",
        ("FIX.5.0", "2238", "3"): "3 .. Supplement",
        ("FIX.5.0", "2243", "0"): "0 .. SellerNotifies",
        ("FIX.5.0", "2243", "1"): "1 .. BuyerNotifies",
        ("FIX.5.0", "2243", "2"): "2 .. SellerOrBuyerNotifies",
        ("FIX.5.0", "2256", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "2256", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "2256", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "2256", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "2256", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "2256", "J"): "J .. TaxID",
        ("FIX.5.0", "2256", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "2256", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "2256", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "2256", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "2256", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "2256", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "2256", "B"): "B .. BIC",
        ("FIX.5.0", "2256", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "2256", "D"): "D .. Proprietary",
        ("FIX.5.0", "2256", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "2256", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "2256", "G"): "G .. MIC",
        ("FIX.5.0", "2256", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "2256", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "2256", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "2256", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "2256", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "2256", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "2256", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "2256", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "2257", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "2257", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "2257", "3"): "3 .. ClientID",
        ("FIX.5.0", "2257", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "2257", "5"): "5 .. InvestorID",
        ("FIX.5.0", "2257", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "2257", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "2257", "8"): "8 .. Locate",
        ("FIX.5.0", "2257", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "2257", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "2257", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "2257", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "2257", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "2257", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "2257", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "2257", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "2257", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "2257", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "2257", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "2257", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "2257", "22"): "22 .. Exchange",
        ("FIX.5.0", "2257", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "2257", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "2257", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "2257", "27"): "27 .. Buyer",
        ("FIX.5.0", "2257", "28"): "28 .. Custodian",
        ("FIX.5.0", "2257", "29"): "29 .. Intermediary",
        ("FIX.5.0", "2257", "30"): "30 .. Agent",
        ("FIX.5.0", "2257", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "2257", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "2257", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "2257", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "2257", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "2257", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "2257", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "2257", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "2257", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "2257", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "2257", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "2257", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "2257", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "2257", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "2257", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "2257", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "2257", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "2257", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "2257", "49"): "49 .. AssetManager",
        ("FIX.5.0", "2257", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "2257", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "2257", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "2257", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "2257", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "2257", "55"): "55 .. SessionID",
        ("FIX.5.0", "2257", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "2257", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "2257", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "2257", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "2257", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "2257", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "2257", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "2257", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "2257", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "2257", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "2257", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "2257", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "2257", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "2257", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "2257", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "2257", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "2257", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "2257", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "2257", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "2257", "75"): "75 .. LocationID",
        ("FIX.5.0", "2257", "76"): "76 .. DeskID",
        ("FIX.5.0", "2257", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "2257", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "2257", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "2257", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "2257", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "2257", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "2257", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "2257", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "2257", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "2257", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "2257", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "2257", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "2257", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "2257", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "2257", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "2257", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "2257", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "2257", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "2257", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "2257", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "2257", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "2257", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "2257", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "2257", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "2257", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "2257", "102"): "102 .. DataRepository",
        ("FIX.5.0", "2257", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "2257", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "2257", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "2257", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "2257", "107"): "107 .. Correspondent",
        ("FIX.5.0", "2257", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "2257", "110"): "110 .. Borrower",
        ("FIX.5.0", "2257", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "2257", "112"): "112 .. Guarantor",
        ("FIX.5.0", "2257", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "2257", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "2257", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "2257", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "2257", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "2257", "118"): "118 .. Operator",
        ("FIX.5.0", "2257", "119"): "119 .. CSD",
        ("FIX.5.0", "2257", "120"): "120 .. ICSD",
        ("FIX.5.0", "2257", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "2257", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "2257", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "2260", "1"): "1 .. Firm",
        ("FIX.5.0", "2260", "2"): "2 .. Person",
        ("FIX.5.0", "2260", "3"): "3 .. System",
        ("FIX.5.0", "2260", "4"): "4 .. Application",
        ("FIX.5.0", "2260", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "2260", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "2260", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "2260", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "2260", "9"): "9 .. ContactName",
        ("FIX.5.0", "2260", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "2260", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "2260", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "2260", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "2260", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "2260", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "2260", "16"): "16 .. BIC",
        ("FIX.5.0", "2260", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "2260", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "2260", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "2260", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "2260", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "2260", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "2260", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "2260", "24"): "24 .. Department",
        ("FIX.5.0", "2260", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "2260", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "2260", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "2260", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "2260", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "2260", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "2260", "31"): "31 .. Location",
        ("FIX.5.0", "2260", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "2260", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "2260", "34"): "34 .. AddressCity",
        ("FIX.5.0", "2260", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "2260", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "2260", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "2260", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "2260", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "2260", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "2260", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "2260", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "2260", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "2260", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "2260", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "2260", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "2260", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "2260", "48"): "48 .. USPerson",
        ("FIX.5.0", "2260", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "2260", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "2260", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "2260", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "2260", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "2260", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "2260", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "2260", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "2260", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "2260", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "2260", "59"): "59 .. FCMCode",
        ("FIX.5.0", "2260", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "2260", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "2260", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "2260", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "2260", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "2260", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "2260", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "2260", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "2260", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "2260", "69"): "69 .. EndUser",
        ("FIX.5.0", "2260", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "2260", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "2260", "72"): "72 .. Domicile",
        ("FIX.5.0", "2260", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "2260", "74"): "74 .. Payer",
        ("FIX.5.0", "2260", "75"): "75 .. Receiver",
        ("FIX.5.0", "2260", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "2260", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "2260", "78"): "78 .. FirstName",
        ("FIX.5.0", "2260", "79"): "79 .. Surname",
        ("FIX.5.0", "2260", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "2260", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "2260", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "2260", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "2260", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "2261", "1"): "1 .. Buy",
        ("FIX.5.0", "2261", "2"): "2 .. Sell",
        ("FIX.5.0", "2262", "1"): "1 .. Buy",
        ("FIX.5.0", "2262", "2"): "2 .. Sell",
        ("FIX.5.0", "2265", "0"): "0 .. Close",
        ("FIX.5.0", "2265", "1"): "1 .. Open",
        ("FIX.5.0", "2265", "2"): "2 .. OfficialSettl",
        ("FIX.5.0", "2265", "3"): "3 .. ValuationTime",
        ("FIX.5.0", "2265", "4"): "4 .. ExcahgneSettlTime",
        ("FIX.5.0", "2265", "5"): "5 .. DerivativesClose",
        ("FIX.5.0", "2265", "6"): "6 .. AsSpecified",
        ("FIX.5.0", "2270", "0"): "0 .. Currency1PerCurrency2",
        ("FIX.5.0", "2270", "1"): "1 .. Currency2PerCurrency1",
        ("FIX.5.0", "2273", "0"): "0 .. ExercisingParty",
        ("FIX.5.0", "2273", "1"): "1 .. NonExercisingParty",
        ("FIX.5.0", "2273", "2"): "2 .. MasterAgreeent",
        ("FIX.5.0", "2273", "3"): "3 .. Supplement",
        ("FIX.5.0", "2278", "0"): "0 .. SellerNotifies",
        ("FIX.5.0", "2278", "1"): "1 .. BuyerNotifies",
        ("FIX.5.0", "2278", "2"): "2 .. SellerOrBuyerNotifies",
        ("FIX.5.0", "2289", "AMTZ"): "AMTZ .. Amortizing",
        ("FIX.5.0", "2289", "CNST"): "CNST .. ConstantNotionalSchedule",
        ("FIX.5.0", "2289", "ACRT"): "ACRT .. AccretingNotionalSchedule",
        ("FIX.5.0", "2289", "CUST"): "CUST .. CustomNotionalSchedule",
        ("FIX.5.0", "2290", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "2290", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "2290", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "2290", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "2290", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "2290", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "2290", "thm"): "thm .. therms",
        ("FIX.5.0", "2290", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "2290", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "2290", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "2290", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "2290", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "2290", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "2290", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "2290", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "2290", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "2290", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "2290", "day"): "day .. Day",
        ("FIX.5.0", "2290", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "2290", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "2290", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "2290", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "2290", "g"): "g .. Grams",
        ("FIX.5.0", "2290", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "2290", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "2290", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "2290", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "2290", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "2290", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "2290", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "2290", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "2290", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "2290", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "2290", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "2290", "L"): "L .. liters",
        ("FIX.5.0", "2290", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "2290", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "2290", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "2290", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "2290", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "2290", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "2290", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "2290", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "2290", "t"): "t .. MetricTons",
        ("FIX.5.0", "2290", "tn"): "tn .. Tons",
        ("FIX.5.0", "2295", "STD"): "STD .. Straddle",
        ("FIX.5.0", "2295", "STG"): "STG .. Strangle",
        ("FIX.5.0", "2295", "BF"): "BF .. Butterfly",
        ("FIX.5.0", "2295", "CNDR"): "CNDR .. Condor",
        ("FIX.5.0", "2295", "CISN"): "CISN .. CallableInversibleSnowball",
        ("FIX.5.0", "2295", "OTHER"): "OTHER .. Other",
        ("FIX.5.0", "2297", "1"): "1 .. Negotiation",
        ("FIX.5.0", "2297", "2"): "2 .. Cancellation",
        ("FIX.5.0", "2298", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "2298", "1"): "1 .. RoundDown",
        ("FIX.5.0", "2298", "2"): "2 .. RoundUp",
        ("FIX.5.0", "2316", "0"): "0 .. Accepted",
        ("FIX.5.0", "2316", "1"): "1 .. Rejected",
        ("FIX.5.0", "2317", "0"): "0 .. UnkRiskLmtRprtID",
        ("FIX.5.0", "2317", "1"): "1 .. UnkPty",
        ("FIX.5.0", "2317", "99"): "99 .. Other",
        ("FIX.5.0", "2320", "0"): "0 .. New",
        ("FIX.5.0", "2320", "1"): "1 .. Cancel",
        ("FIX.5.0", "2320", "2"): "2 .. Replace",
        ("FIX.5.0", "2321", "0"): "0 .. Submit",
        ("FIX.5.0", "2321", "1"): "1 .. LimitConsumed",
        ("FIX.5.0", "2323", "0"): "0 .. AllOrNone",
        ("FIX.5.0", "2323", "1"): "1 .. Partial",
        ("FIX.5.0", "2325", "0"): "0 .. Approved",
        ("FIX.5.0", "2325", "1"): "1 .. PartiallyApproved",
        ("FIX.5.0", "2325", "2"): "2 .. Rejected",
        ("FIX.5.0", "2325", "3"): "3 .. ApprovalPending",
        ("FIX.5.0", "2325", "4"): "4 .. Cancelled",
        ("FIX.5.0", "2326", "0"): "0 .. Successful",
        ("FIX.5.0", "2326", "1"): "1 .. InvalidParty",
        ("FIX.5.0", "2326", "2"): "2 .. ReqExceedsCreditLimit",
        ("FIX.5.0", "2326", "3"): "3 .. ReqExceedsClipSizeLimit",
        ("FIX.5.0", "2326", "4"): "4 .. ReqExceedsMaxNotional",
        ("FIX.5.0", "2326", "99"): "99 .. Other",
        ("FIX.5.0", "2329", "0"): "0 .. Suspend",
        ("FIX.5.0", "2329", "1"): "1 .. HaltTrading",
        ("FIX.5.0", "2329", "2"): "2 .. Reinstate",
        ("FIX.5.0", "2332", "0"): "0 .. Accepted",
        ("FIX.5.0", "2332", "1"): "1 .. Completed",
        ("FIX.5.0", "2332", "2"): "2 .. Rejected",
        ("FIX.5.0", "2333", "0"): "0 .. InvalidParty",
        ("FIX.5.0", "2333", "1"): "1 .. UnkReqParty",
        ("FIX.5.0", "2333", "98"): "98 .. NotAuthorized",
        ("FIX.5.0", "2333", "99"): "99 .. Other",
        ("FIX.5.0", "2335", "0"): "0 .. RiskLimitRequestID",
        ("FIX.5.0", "2335", "1"): "1 .. RiskLimitCheckID",
        ("FIX.5.0", "2335", "3"): "3 .. OutOfBandID",
        ("FIX.5.0", "2337", "H"): "H .. Hour",
        ("FIX.5.0", "2337", "Min"): "Min .. Minute",
        ("FIX.5.0", "2337", "S"): "S .. Second",
        ("FIX.5.0", "2337", "D"): "D .. Day",
        ("FIX.5.0", "2337", "Wk"): "Wk .. Week",
        ("FIX.5.0", "2337", "Mo"): "Mo .. Month",
        ("FIX.5.0", "2337", "Yr"): "Yr .. Year",
        ("FIX.5.0", "2337", "Q"): "Q .. Quarter",
        ("FIX.5.0", "2338", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2338", "18"): "18 .. Current",
        ("FIX.5.0", "2338", "19"): "19 .. New",
        ("FIX.5.0", "2338", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2338", "0"): "0 .. Agency",
        ("FIX.5.0", "2338", "1"): "1 .. Principal",
        ("FIX.5.0", "2338", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2338", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2338", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2338", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2338", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2338", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2338", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2338", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2338", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2338", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2338", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2338", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2338", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2338", "7"): "7 .. Bank",
        ("FIX.5.0", "2338", "8"): "8 .. Hub",
        ("FIX.5.0", "2338", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2338", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2338", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2338", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2338", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2338", "21"): "21 .. Specialist",
        ("FIX.5.0", "2338", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2339", "0"): "0 .. None",
        ("FIX.5.0", "2339", "1"): "1 .. PlusOneModel",
        ("FIX.5.0", "2339", "2"): "2 .. PingModel",
        ("FIX.5.0", "2339", "3"): "3 .. PushModel",
        ("FIX.5.0", "2343", "0"): "0 .. Accepted",
        ("FIX.5.0", "2343", "1"): "1 .. Rejected",
        ("FIX.5.0", "2343", "2"): "2 .. ClaimRequired",
        ("FIX.5.0", "2343", "3"): "3 .. PreDefinedLimitCheckSucceeded",
        ("FIX.5.0", "2343", "4"): "4 .. PreDefinedLimitCheckFailed",
        ("FIX.5.0", "2343", "5"): "5 .. PreDefinedAutoAcceptRuleInvoked",
        ("FIX.5.0", "2343", "6"): "6 .. PreDefinedAutoRejectRuleInvoked",
        ("FIX.5.0", "2343", "7"): "7 .. AcceptedByClearingFirm",
        ("FIX.5.0", "2343", "8"): "8 .. RejectedByClearingFirm",
        ("FIX.5.0", "2343", "9"): "9 .. Pending",
        ("FIX.5.0", "2343", "10"): "10 .. AcceptedByCreditHub",
        ("FIX.5.0", "2343", "11"): "11 .. RejectedByCreditHub",
        ("FIX.5.0", "2343", "12"): "12 .. PendingCreditHubCheck",
        ("FIX.5.0", "2343", "13"): "13 .. AcceptedByExecVenue",
        ("FIX.5.0", "2343", "14"): "14 .. RejectedByExecVenue",
        ("FIX.5.0", "2344", "0"): "0 .. Accepted",
        ("FIX.5.0", "2344", "1"): "1 .. Rejected",
        ("FIX.5.0", "2344", "2"): "2 .. ClaimRequired",
        ("FIX.5.0", "2344", "3"): "3 .. PreDefinedLimitCheckSucceeded",
        ("FIX.5.0", "2344", "4"): "4 .. PreDefinedLimitCheckFailed",
        ("FIX.5.0", "2344", "5"): "5 .. PreDefinedAutoAcceptRuleInvoked",
        ("FIX.5.0", "2344", "6"): "6 .. PreDefinedAutoRejectRuleInvoked",
        ("FIX.5.0", "2344", "7"): "7 .. AcceptedByClearingFirm",
        ("FIX.5.0", "2344", "8"): "8 .. RejectedByClearingFirm",
        ("FIX.5.0", "2344", "9"): "9 .. Pending",
        ("FIX.5.0", "2344", "10"): "10 .. AcceptedByCreditHub",
        ("FIX.5.0", "2344", "11"): "11 .. RejectedByCreditHub",
        ("FIX.5.0", "2344", "12"): "12 .. PendingCreditHubCheck",
        ("FIX.5.0", "2344", "13"): "13 .. AcceptedByExecVenue",
        ("FIX.5.0", "2344", "14"): "14 .. RejectedByExecVenue",
        ("FIX.5.0", "2347", "0"): "0 .. None",
        ("FIX.5.0", "2347", "1"): "1 .. SEFRequiredTransaction",
        ("FIX.5.0", "2347", "2"): "2 .. SEFPermittedTransaction",
        ("FIX.5.0", "2348", "1"): "1 .. Financials",
        ("FIX.5.0", "2348", "2"): "2 .. Commodities",
        ("FIX.5.0", "2348", "3"): "3 .. AlternativeInvestments",
        ("FIX.5.0", "2355", "0"): "0 .. Disabled",
        ("FIX.5.0", "2355", "1"): "1 .. Enabled",
        ("FIX.5.0", "2356", "0"): "0 .. NoRemunerationPaid",
        ("FIX.5.0", "2356", "1"): "1 .. RemunerationPaid",
        ("FIX.5.0", "2364", "1"): "1 .. New",
        ("FIX.5.0", "2364", "2"): "2 .. Replace",
        ("FIX.5.0", "2364", "3"): "3 .. Cancel",
        ("FIX.5.0", "2364", "4"): "4 .. Reverse",
        ("FIX.5.0", "2366", "M"): "M .. Multiply",
        ("FIX.5.0", "2366", "D"): "D .. Divide",
        ("FIX.5.0", "2375", "I"): "I .. ISINOrAltInstrmtID",
        ("FIX.5.0", "2375", "E"): "E .. InterimTaxonomy",
        ("FIX.5.0", "2376", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2376", "18"): "18 .. Current",
        ("FIX.5.0", "2376", "19"): "19 .. New",
        ("FIX.5.0", "2376", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2376", "0"): "0 .. Agency",
        ("FIX.5.0", "2376", "1"): "1 .. Principal",
        ("FIX.5.0", "2376", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2376", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2376", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2376", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2376", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2376", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2376", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2376", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2376", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2376", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2376", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2376", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2376", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2376", "7"): "7 .. Bank",
        ("FIX.5.0", "2376", "8"): "8 .. Hub",
        ("FIX.5.0", "2376", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2376", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2376", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2376", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2376", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2376", "21"): "21 .. Specialist",
        ("FIX.5.0", "2376", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2377", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2377", "18"): "18 .. Current",
        ("FIX.5.0", "2377", "19"): "19 .. New",
        ("FIX.5.0", "2377", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2377", "0"): "0 .. Agency",
        ("FIX.5.0", "2377", "1"): "1 .. Principal",
        ("FIX.5.0", "2377", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2377", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2377", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2377", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2377", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2377", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2377", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2377", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2377", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2377", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2377", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2377", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2377", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2377", "7"): "7 .. Bank",
        ("FIX.5.0", "2377", "8"): "8 .. Hub",
        ("FIX.5.0", "2377", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2377", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2377", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2377", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2377", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2377", "21"): "21 .. Specialist",
        ("FIX.5.0", "2377", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2378", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2378", "18"): "18 .. Current",
        ("FIX.5.0", "2378", "19"): "19 .. New",
        ("FIX.5.0", "2378", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2378", "0"): "0 .. Agency",
        ("FIX.5.0", "2378", "1"): "1 .. Principal",
        ("FIX.5.0", "2378", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2378", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2378", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2378", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2378", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2378", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2378", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2378", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2378", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2378", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2378", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2378", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2378", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2378", "7"): "7 .. Bank",
        ("FIX.5.0", "2378", "8"): "8 .. Hub",
        ("FIX.5.0", "2378", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2378", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2378", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2378", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2378", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2378", "21"): "21 .. Specialist",
        ("FIX.5.0", "2378", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2379", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2379", "18"): "18 .. Current",
        ("FIX.5.0", "2379", "19"): "19 .. New",
        ("FIX.5.0", "2379", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2379", "0"): "0 .. Agency",
        ("FIX.5.0", "2379", "1"): "1 .. Principal",
        ("FIX.5.0", "2379", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2379", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2379", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2379", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2379", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2379", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2379", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2379", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2379", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2379", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2379", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2379", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2379", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2379", "7"): "7 .. Bank",
        ("FIX.5.0", "2379", "8"): "8 .. Hub",
        ("FIX.5.0", "2379", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2379", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2379", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2379", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2379", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2379", "21"): "21 .. Specialist",
        ("FIX.5.0", "2379", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2380", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2380", "18"): "18 .. Current",
        ("FIX.5.0", "2380", "19"): "19 .. New",
        ("FIX.5.0", "2380", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2380", "0"): "0 .. Agency",
        ("FIX.5.0", "2380", "1"): "1 .. Principal",
        ("FIX.5.0", "2380", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2380", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2380", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2380", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2380", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2380", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2380", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2380", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2380", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2380", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2380", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2380", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2380", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2380", "7"): "7 .. Bank",
        ("FIX.5.0", "2380", "8"): "8 .. Hub",
        ("FIX.5.0", "2380", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2380", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2380", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2380", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2380", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2380", "21"): "21 .. Specialist",
        ("FIX.5.0", "2380", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2381", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2381", "18"): "18 .. Current",
        ("FIX.5.0", "2381", "19"): "19 .. New",
        ("FIX.5.0", "2381", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2381", "0"): "0 .. Agency",
        ("FIX.5.0", "2381", "1"): "1 .. Principal",
        ("FIX.5.0", "2381", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2381", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2381", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2381", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2381", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2381", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2381", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2381", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2381", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2381", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2381", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2381", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2381", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2381", "7"): "7 .. Bank",
        ("FIX.5.0", "2381", "8"): "8 .. Hub",
        ("FIX.5.0", "2381", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2381", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2381", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2381", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2381", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2381", "21"): "21 .. Specialist",
        ("FIX.5.0", "2381", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2382", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2382", "18"): "18 .. Current",
        ("FIX.5.0", "2382", "19"): "19 .. New",
        ("FIX.5.0", "2382", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2382", "0"): "0 .. Agency",
        ("FIX.5.0", "2382", "1"): "1 .. Principal",
        ("FIX.5.0", "2382", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2382", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2382", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2382", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2382", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2382", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2382", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2382", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2382", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2382", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2382", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2382", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2382", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2382", "7"): "7 .. Bank",
        ("FIX.5.0", "2382", "8"): "8 .. Hub",
        ("FIX.5.0", "2382", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2382", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2382", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2382", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2382", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2382", "21"): "21 .. Specialist",
        ("FIX.5.0", "2382", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2383", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2383", "18"): "18 .. Current",
        ("FIX.5.0", "2383", "19"): "19 .. New",
        ("FIX.5.0", "2383", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2383", "0"): "0 .. Agency",
        ("FIX.5.0", "2383", "1"): "1 .. Principal",
        ("FIX.5.0", "2383", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2383", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2383", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2383", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2383", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2383", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2383", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2383", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2383", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2383", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2383", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2383", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2383", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2383", "7"): "7 .. Bank",
        ("FIX.5.0", "2383", "8"): "8 .. Hub",
        ("FIX.5.0", "2383", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2383", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2383", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2383", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2383", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2383", "21"): "21 .. Specialist",
        ("FIX.5.0", "2383", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2384", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2384", "18"): "18 .. Current",
        ("FIX.5.0", "2384", "19"): "19 .. New",
        ("FIX.5.0", "2384", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2384", "0"): "0 .. Agency",
        ("FIX.5.0", "2384", "1"): "1 .. Principal",
        ("FIX.5.0", "2384", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2384", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2384", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2384", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2384", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2384", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2384", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2384", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2384", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2384", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2384", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2384", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2384", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2384", "7"): "7 .. Bank",
        ("FIX.5.0", "2384", "8"): "8 .. Hub",
        ("FIX.5.0", "2384", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2384", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2384", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2384", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2384", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2384", "21"): "21 .. Specialist",
        ("FIX.5.0", "2384", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2385", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2385", "18"): "18 .. Current",
        ("FIX.5.0", "2385", "19"): "19 .. New",
        ("FIX.5.0", "2385", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2385", "0"): "0 .. Agency",
        ("FIX.5.0", "2385", "1"): "1 .. Principal",
        ("FIX.5.0", "2385", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2385", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2385", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2385", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2385", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2385", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2385", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2385", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2385", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2385", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2385", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2385", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2385", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2385", "7"): "7 .. Bank",
        ("FIX.5.0", "2385", "8"): "8 .. Hub",
        ("FIX.5.0", "2385", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2385", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2385", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2385", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2385", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2385", "21"): "21 .. Specialist",
        ("FIX.5.0", "2385", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2386", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2386", "18"): "18 .. Current",
        ("FIX.5.0", "2386", "19"): "19 .. New",
        ("FIX.5.0", "2386", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2386", "0"): "0 .. Agency",
        ("FIX.5.0", "2386", "1"): "1 .. Principal",
        ("FIX.5.0", "2386", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2386", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2386", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2386", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2386", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2386", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2386", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2386", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2386", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2386", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2386", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2386", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2386", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2386", "7"): "7 .. Bank",
        ("FIX.5.0", "2386", "8"): "8 .. Hub",
        ("FIX.5.0", "2386", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2386", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2386", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2386", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2386", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2386", "21"): "21 .. Specialist",
        ("FIX.5.0", "2386", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2387", "0"): "0 .. DoesNotApply",
        ("FIX.5.0", "2387", "1"): "1 .. ContingentTrade",
        ("FIX.5.0", "2387", "2"): "2 .. NonContingentTrade",
        ("FIX.5.0", "2388", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2388", "18"): "18 .. Current",
        ("FIX.5.0", "2388", "19"): "19 .. New",
        ("FIX.5.0", "2388", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2388", "0"): "0 .. Agency",
        ("FIX.5.0", "2388", "1"): "1 .. Principal",
        ("FIX.5.0", "2388", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2388", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2388", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2388", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2388", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2388", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2388", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2388", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2388", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2388", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2388", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2388", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2388", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2388", "7"): "7 .. Bank",
        ("FIX.5.0", "2388", "8"): "8 .. Hub",
        ("FIX.5.0", "2388", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2388", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2388", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2388", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2388", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2388", "21"): "21 .. Specialist",
        ("FIX.5.0", "2388", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2389", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2389", "18"): "18 .. Current",
        ("FIX.5.0", "2389", "19"): "19 .. New",
        ("FIX.5.0", "2389", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2389", "0"): "0 .. Agency",
        ("FIX.5.0", "2389", "1"): "1 .. Principal",
        ("FIX.5.0", "2389", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2389", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2389", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2389", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2389", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2389", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2389", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2389", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2389", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2389", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2389", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2389", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2389", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2389", "7"): "7 .. Bank",
        ("FIX.5.0", "2389", "8"): "8 .. Hub",
        ("FIX.5.0", "2389", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2389", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2389", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2389", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2389", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2389", "21"): "21 .. Specialist",
        ("FIX.5.0", "2389", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2391", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "2391", "18"): "18 .. Current",
        ("FIX.5.0", "2391", "19"): "19 .. New",
        ("FIX.5.0", "2391", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "2391", "0"): "0 .. Agency",
        ("FIX.5.0", "2391", "1"): "1 .. Principal",
        ("FIX.5.0", "2391", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "2391", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "2391", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "2391", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "2391", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "2391", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "2391", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "2391", "27"): "27 .. Supervisor",
        ("FIX.5.0", "2391", "22"): "22 .. Algorithm",
        ("FIX.5.0", "2391", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "2391", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "2391", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "2391", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "2391", "7"): "7 .. Bank",
        ("FIX.5.0", "2391", "8"): "8 .. Hub",
        ("FIX.5.0", "2391", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "2391", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "2391", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "2391", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "2391", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "2391", "21"): "21 .. Specialist",
        ("FIX.5.0", "2391", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "2393", "0"): "0 .. RiskLimitRequestID",
        ("FIX.5.0", "2393", "1"): "1 .. RiskLimitCheckID",
        ("FIX.5.0", "2393", "3"): "3 .. OutOfBandID",
        ("FIX.5.0", "2396", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "2396", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "2396", "3"): "3 .. ClientID",
        ("FIX.5.0", "2396", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "2396", "5"): "5 .. InvestorID",
        ("FIX.5.0", "2396", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "2396", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "2396", "8"): "8 .. Locate",
        ("FIX.5.0", "2396", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "2396", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "2396", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "2396", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "2396", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "2396", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "2396", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "2396", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "2396", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "2396", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "2396", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "2396", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "2396", "22"): "22 .. Exchange",
        ("FIX.5.0", "2396", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "2396", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "2396", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "2396", "27"): "27 .. Buyer",
        ("FIX.5.0", "2396", "28"): "28 .. Custodian",
        ("FIX.5.0", "2396", "29"): "29 .. Intermediary",
        ("FIX.5.0", "2396", "30"): "30 .. Agent",
        ("FIX.5.0", "2396", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "2396", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "2396", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "2396", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "2396", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "2396", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "2396", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "2396", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "2396", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "2396", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "2396", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "2396", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "2396", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "2396", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "2396", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "2396", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "2396", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "2396", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "2396", "49"): "49 .. AssetManager",
        ("FIX.5.0", "2396", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "2396", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "2396", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "2396", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "2396", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "2396", "55"): "55 .. SessionID",
        ("FIX.5.0", "2396", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "2396", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "2396", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "2396", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "2396", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "2396", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "2396", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "2396", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "2396", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "2396", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "2396", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "2396", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "2396", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "2396", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "2396", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "2396", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "2396", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "2396", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "2396", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "2396", "75"): "75 .. LocationID",
        ("FIX.5.0", "2396", "76"): "76 .. DeskID",
        ("FIX.5.0", "2396", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "2396", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "2396", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "2396", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "2396", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "2396", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "2396", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "2396", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "2396", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "2396", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "2396", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "2396", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "2396", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "2396", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "2396", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "2396", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "2396", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "2396", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "2396", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "2396", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "2396", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "2396", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "2396", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "2396", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "2396", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "2396", "102"): "102 .. DataRepository",
        ("FIX.5.0", "2396", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "2396", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "2396", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "2396", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "2396", "107"): "107 .. Correspondent",
        ("FIX.5.0", "2396", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "2396", "110"): "110 .. Borrower",
        ("FIX.5.0", "2396", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "2396", "112"): "112 .. Guarantor",
        ("FIX.5.0", "2396", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "2396", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "2396", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "2396", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "2396", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "2396", "118"): "118 .. Operator",
        ("FIX.5.0", "2396", "119"): "119 .. CSD",
        ("FIX.5.0", "2396", "120"): "120 .. ICSD",
        ("FIX.5.0", "2396", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "2396", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "2396", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "2397", "1"): "1 .. ClearingMember",
        ("FIX.5.0", "2397", "2"): "2 .. Client",
        ("FIX.5.0", "2398", "1"): "1 .. ClearingMember",
        ("FIX.5.0", "2398", "2"): "2 .. Client",
        ("FIX.5.0", "2399", "1"): "1 .. ClearingMember",
        ("FIX.5.0", "2399", "2"): "2 .. Client",
        ("FIX.5.0", "2402", "1"): "1 .. OrderEntry",
        ("FIX.5.0", "2402", "2"): "2 .. HItLift",
        ("FIX.5.0", "2402", "3"): "3 .. ViewIndicativePx",
        ("FIX.5.0", "2402", "4"): "4 .. ViewExecutablePx",
        ("FIX.5.0", "2402", "5"): "5 .. SingleQuote",
        ("FIX.5.0", "2402", "6"): "6 .. StreamingQuotes",
        ("FIX.5.0", "2402", "7"): "7 .. SingleBroker",
        ("FIX.5.0", "2402", "8"): "8 .. MultiBrokers",
        ("FIX.5.0", "2403", "1"): "1 .. QuoteEntry",
        ("FIX.5.0", "2403", "2"): "2 .. QuoteModification",
        ("FIX.5.0", "2405", "0"): "0 .. Unspecified",
        ("FIX.5.0", "2405", "1"): "1 .. Manual",
        ("FIX.5.0", "2405", "2"): "2 .. Automated",
        ("FIX.5.0", "2405", "3"): "3 .. VoiceBrokered",
        ("FIX.5.0", "2414", "1"): "1 .. CUSIP",
        ("FIX.5.0", "2414", "2"): "2 .. SEDOL",
        ("FIX.5.0", "2414", "3"): "3 .. QUIK",
        ("FIX.5.0", "2414", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "2414", "5"): "5 .. RICCode",
        ("FIX.5.0", "2414", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "2414", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "2414", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "2414", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "2414", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "2414", "B"): "B .. Wertpapier",
        ("FIX.5.0", "2414", "C"): "C .. Dutch",
        ("FIX.5.0", "2414", "D"): "D .. Valoren",
        ("FIX.5.0", "2414", "E"): "E .. Sicovam",
        ("FIX.5.0", "2414", "F"): "F .. Belgian",
        ("FIX.5.0", "2414", "G"): "G .. Common",
        ("FIX.5.0", "2414", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "2414", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "2414", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "2414", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "2414", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "2414", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "2414", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "2414", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "2414", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "2414", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "2414", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "2414", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "2414", "U"): "U .. Synthetic",
        ("FIX.5.0", "2414", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "2414", "W"): "W .. IndexName",
        ("FIX.5.0", "2425", "1"): "1 .. Accepted",
        ("FIX.5.0", "2425", "2"): "2 .. AcceptedWithAdditionalEvents",
        ("FIX.5.0", "2425", "3"): "3 .. Rejected",
        ("FIX.5.0", "2426", "0"): "0 .. Successful",
        ("FIX.5.0", "2426", "1"): "1 .. ResponseLevelNotSupported",
        ("FIX.5.0", "2426", "2"): "2 .. InvalidMarket",
        ("FIX.5.0", "2426", "3"): "3 .. InvalidMarketSegment",
        ("FIX.5.0", "2426", "99"): "99 .. Other",
        ("FIX.5.0", "2427", "0"): "0 .. NoAck",
        ("FIX.5.0", "2427", "1"): "1 .. MinimumAck",
        ("FIX.5.0", "2427", "2"): "2 .. AckEach",
        ("FIX.5.0", "2427", "3"): "3 .. SummaryAck",
        ("FIX.5.0", "2429", "1"): "1 .. Add",
        ("FIX.5.0", "2429", "2"): "2 .. Modify",
        ("FIX.5.0", "2429", "3"): "3 .. Delete",
        ("FIX.5.0", "2429", "4"): "4 .. Suspend",
        ("FIX.5.0", "2429", "5"): "5 .. Release",
        ("FIX.5.0", "2431", "1"): "1 .. OrdAddedOnRequest",
        ("FIX.5.0", "2431", "2"): "2 .. OrdReplacedOnRequest",
        ("FIX.5.0", "2431", "3"): "3 .. OrdCxldOnRequest",
        ("FIX.5.0", "2431", "4"): "4 .. UnsolicitedOrdCxl",
        ("FIX.5.0", "2431", "5"): "5 .. NonRestingOrdAddedOnRequest",
        ("FIX.5.0", "2431", "6"): "6 .. OrdReplacedWithNonRestingOrdOnRequest",
        ("FIX.5.0", "2431", "7"): "7 .. TriggerOrdReplacedOnRequest",
        ("FIX.5.0", "2431", "8"): "8 .. SuspendedOrdReplacedOnRequest",
        ("FIX.5.0", "2431", "9"): "9 .. SuspendedOrdCxldOnRequest",
        ("FIX.5.0", "2431", "10"): "10 .. OrdCxlPending",
        ("FIX.5.0", "2431", "11"): "11 .. PendingCxlExecuted",
        ("FIX.5.0", "2431", "12"): "12 .. RestingOrdTriggered",
        ("FIX.5.0", "2431", "13"): "13 .. SuspendedOrdActivated",
        ("FIX.5.0", "2431", "14"): "14 .. ActiveOrdSuspended",
        ("FIX.5.0", "2431", "15"): "15 .. OrdExpired",
        ("FIX.5.0", "2435", "1"): "1 .. Firm",
        ("FIX.5.0", "2435", "2"): "2 .. Person",
        ("FIX.5.0", "2435", "3"): "3 .. System",
        ("FIX.5.0", "2435", "4"): "4 .. Application",
        ("FIX.5.0", "2435", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "2435", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "2435", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "2435", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "2435", "9"): "9 .. ContactName",
        ("FIX.5.0", "2435", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "2435", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "2435", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "2435", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "2435", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "2435", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "2435", "16"): "16 .. BIC",
        ("FIX.5.0", "2435", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "2435", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "2435", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "2435", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "2435", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "2435", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "2435", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "2435", "24"): "24 .. Department",
        ("FIX.5.0", "2435", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "2435", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "2435", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "2435", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "2435", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "2435", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "2435", "31"): "31 .. Location",
        ("FIX.5.0", "2435", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "2435", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "2435", "34"): "34 .. AddressCity",
        ("FIX.5.0", "2435", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "2435", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "2435", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "2435", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "2435", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "2435", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "2435", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "2435", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "2435", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "2435", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "2435", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "2435", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "2435", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "2435", "48"): "48 .. USPerson",
        ("FIX.5.0", "2435", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "2435", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "2435", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "2435", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "2435", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "2435", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "2435", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "2435", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "2435", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "2435", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "2435", "59"): "59 .. FCMCode",
        ("FIX.5.0", "2435", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "2435", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "2435", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "2435", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "2435", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "2435", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "2435", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "2435", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "2435", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "2435", "69"): "69 .. EndUser",
        ("FIX.5.0", "2435", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "2435", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "2435", "72"): "72 .. Domicile",
        ("FIX.5.0", "2435", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "2435", "74"): "74 .. Payer",
        ("FIX.5.0", "2435", "75"): "75 .. Receiver",
        ("FIX.5.0", "2435", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "2435", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "2435", "78"): "78 .. FirstName",
        ("FIX.5.0", "2435", "79"): "79 .. Surname",
        ("FIX.5.0", "2435", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "2435", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "2435", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "2435", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "2435", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "2439", "0"): "0 .. New",
        ("FIX.5.0", "2439", "1"): "1 .. Replace",
        ("FIX.5.0", "2439", "2"): "2 .. Cancel",
        ("FIX.5.0", "2440", "0"): "0 .. RequestTransfer",
        ("FIX.5.0", "2440", "1"): "1 .. AcceptTransfer",
        ("FIX.5.0", "2440", "2"): "2 .. DeclineTransfer",
        ("FIX.5.0", "2441", "0"): "0 .. InterFirmTransfer",
        ("FIX.5.0", "2441", "1"): "1 .. IntraFirmTransfer",
        ("FIX.5.0", "2441", "2"): "2 .. CMTA",
        ("FIX.5.0", "2442", "0"): "0 .. Received",
        ("FIX.5.0", "2442", "1"): "1 .. RejectedByIntermediary",
        ("FIX.5.0", "2442", "2"): "2 .. AcceptPending",
        ("FIX.5.0", "2442", "3"): "3 .. Accepted",
        ("FIX.5.0", "2442", "4"): "4 .. Declined",
        ("FIX.5.0", "2442", "5"): "5 .. Cancelled",
        ("FIX.5.0", "2443", "0"): "0 .. Success",
        ("FIX.5.0", "2443", "1"): "1 .. InvalidParty",
        ("FIX.5.0", "2443", "2"): "2 .. UnknownInstrument",
        ("FIX.5.0", "2443", "3"): "3 .. UnauthorizedToSubmitXfer",
        ("FIX.5.0", "2443", "4"): "4 .. UnknownPosition",
        ("FIX.5.0", "2443", "99"): "99 .. Other",
        ("FIX.5.0", "2444", "0"): "0 .. Submit",
        ("FIX.5.0", "2444", "1"): "1 .. Alleged",
        ("FIX.5.0", "2446", "1"): "1 .. Buy",
        ("FIX.5.0", "2446", "2"): "2 .. Sell",
        ("FIX.5.0", "2446", "3"): "3 .. BuyMinus",
        ("FIX.5.0", "2446", "4"): "4 .. SellPlus",
        ("FIX.5.0", "2446", "5"): "5 .. SellShort",
        ("FIX.5.0", "2446", "6"): "6 .. SellShortExempt",
        ("FIX.5.0", "2446", "7"): "7 .. Undisclosed",
        ("FIX.5.0", "2446", "8"): "8 .. Cross",
        ("FIX.5.0", "2446", "9"): "9 .. CrossShort",
        ("FIX.5.0", "2446", "A"): "A .. CrossShortExempt",
        ("FIX.5.0", "2446", "B"): "B .. AsDefined",
        ("FIX.5.0", "2446", "C"): "C .. Opposite",
        ("FIX.5.0", "2446", "D"): "D .. Subscribe",
        ("FIX.5.0", "2446", "E"): "E .. Redeem",
        ("FIX.5.0", "2446", "F"): "F .. Lend",
        ("FIX.5.0", "2446", "G"): "G .. Borrow",
        ("FIX.5.0", "2446", "H"): "H .. SellUndisclosed",
        ("FIX.5.0", "2456", "1"): "1 .. Count",
        ("FIX.5.0", "2456", "2"): "2 .. AverageVolume",
        ("FIX.5.0", "2456", "3"): "3 .. TotalVolume",
        ("FIX.5.0", "2456", "4"): "4 .. Distribution",
        ("FIX.5.0", "2456", "5"): "5 .. Ratio",
        ("FIX.5.0", "2456", "6"): "6 .. Liquidity",
        ("FIX.5.0", "2456", "7"): "7 .. VWAP",
        ("FIX.5.0", "2456", "8"): "8 .. Volatility",
        ("FIX.5.0", "2456", "9"): "9 .. Duration",
        ("FIX.5.0", "2456", "10"): "10 .. Tick",
        ("FIX.5.0", "2456", "11"): "11 .. AverageValue",
        ("FIX.5.0", "2456", "12"): "12 .. TotalValue",
        ("FIX.5.0", "2456", "13"): "13 .. High",
        ("FIX.5.0", "2456", "14"): "14 .. Low",
        ("FIX.5.0", "2456", "15"): "15 .. Midpoint",
        ("FIX.5.0", "2456", "16"): "16 .. First",
        ("FIX.5.0", "2456", "17"): "17 .. Last",
        ("FIX.5.0", "2456", "18"): "18 .. Final",
        ("FIX.5.0", "2456", "19"): "19 .. ExchangeBest",
        ("FIX.5.0", "2456", "20"): "20 .. ExchangeBestWithVolume",
        ("FIX.5.0", "2456", "21"): "21 .. ConsolidatedBest",
        ("FIX.5.0", "2456", "22"): "22 .. ConsolidatedBestWithVolume",
        ("FIX.5.0", "2456", "23"): "23 .. TWAP",
        ("FIX.5.0", "2456", "24"): "24 .. AverageDuration",
        ("FIX.5.0", "2456", "25"): "25 .. AveragePrice",
        ("FIX.5.0", "2456", "26"): "26 .. TotalFees",
        ("FIX.5.0", "2456", "27"): "27 .. TotalBenefits",
        ("FIX.5.0", "2456", "28"): "28 .. MedianValue",
        ("FIX.5.0", "2456", "29"): "29 .. AverageLiquidity",
        ("FIX.5.0", "2456", "30"): "30 .. MedianDuration",
        ("FIX.5.0", "2457", "1"): "1 .. BidPrices",
        ("FIX.5.0", "2457", "2"): "2 .. OfferPrices",
        ("FIX.5.0", "2457", "3"): "3 .. BidDepth",
        ("FIX.5.0", "2457", "4"): "4 .. OfferDepth",
        ("FIX.5.0", "2457", "5"): "5 .. Orders",
        ("FIX.5.0", "2457", "6"): "6 .. Quotes",
        ("FIX.5.0", "2457", "7"): "7 .. OrdersAndQuotes",
        ("FIX.5.0", "2457", "8"): "8 .. Trades",
        ("FIX.5.0", "2457", "9"): "9 .. TradePrices",
        ("FIX.5.0", "2457", "10"): "10 .. AuctionPrices",
        ("FIX.5.0", "2457", "11"): "11 .. OpeningPrices",
        ("FIX.5.0", "2457", "12"): "12 .. ClosingPrices",
        ("FIX.5.0", "2457", "13"): "13 .. SettlementPrices",
        ("FIX.5.0", "2457", "14"): "14 .. UnderlyingPrices",
        ("FIX.5.0", "2457", "15"): "15 .. OpenInterest",
        ("FIX.5.0", "2457", "16"): "16 .. IndexValues",
        ("FIX.5.0", "2457", "17"): "17 .. MarginRates",
        ("FIX.5.0", "2457", "18"): "18 .. Outages",
        ("FIX.5.0", "2457", "19"): "19 .. ScheduledAuctions",
        ("FIX.5.0", "2457", "20"): "20 .. ReferencePrices",
        ("FIX.5.0", "2457", "21"): "21 .. TradeValue",
        ("FIX.5.0", "2457", "22"): "22 .. MarketDataFeeItems",
        ("FIX.5.0", "2457", "23"): "23 .. Rebates",
        ("FIX.5.0", "2457", "24"): "24 .. Discounts",
        ("FIX.5.0", "2457", "25"): "25 .. Payments",
        ("FIX.5.0", "2457", "26"): "26 .. Taxes",
        ("FIX.5.0", "2457", "27"): "27 .. Levies",
        ("FIX.5.0", "2457", "28"): "28 .. Benefits",
        ("FIX.5.0", "2457", "29"): "29 .. Fees",
        ("FIX.5.0", "2457", "30"): "30 .. OrdersRFQs",
        ("FIX.5.0", "2457", "31"): "31 .. MarketMakers",
        ("FIX.5.0", "2457", "32"): "32 .. TradingInterruptions",
        ("FIX.5.0", "2457", "33"): "33 .. TradingSuspensions",
        ("FIX.5.0", "2457", "34"): "34 .. NoQuotes",
        ("FIX.5.0", "2457", "35"): "35 .. RequestForQuotes",
        ("FIX.5.0", "2457", "36"): "36 .. TradeVolume",
        ("FIX.5.0", "2458", "1"): "1 .. Visible",
        ("FIX.5.0", "2458", "2"): "2 .. Hidden",
        ("FIX.5.0", "2458", "3"): "3 .. Indicative",
        ("FIX.5.0", "2458", "4"): "4 .. Tradeable",
        ("FIX.5.0", "2458", "5"): "5 .. Passive",
        ("FIX.5.0", "2458", "6"): "6 .. MarketConsensus",
        ("FIX.5.0", "2458", "7"): "7 .. Power",
        ("FIX.5.0", "2458", "8"): "8 .. HardwareError",
        ("FIX.5.0", "2458", "9"): "9 .. SoftwareError",
        ("FIX.5.0", "2458", "10"): "10 .. NetworkError",
        ("FIX.5.0", "2458", "11"): "11 .. Failed",
        ("FIX.5.0", "2458", "12"): "12 .. Executed",
        ("FIX.5.0", "2458", "13"): "13 .. Entered",
        ("FIX.5.0", "2458", "14"): "14 .. Modified",
        ("FIX.5.0", "2458", "15"): "15 .. Cancelled",
        ("FIX.5.0", "2458", "16"): "16 .. MarketDataAccess",
        ("FIX.5.0", "2458", "17"): "17 .. TerminalAccess",
        ("FIX.5.0", "2458", "18"): "18 .. Volume",
        ("FIX.5.0", "2458", "19"): "19 .. Cleared",
        ("FIX.5.0", "2458", "20"): "20 .. Settled",
        ("FIX.5.0", "2458", "21"): "21 .. Other",
        ("FIX.5.0", "2458", "22"): "22 .. Monetary",
        ("FIX.5.0", "2458", "23"): "23 .. NonMonetary",
        ("FIX.5.0", "2458", "24"): "24 .. Gross",
        ("FIX.5.0", "2458", "25"): "25 .. LargeInScale",
        ("FIX.5.0", "2458", "26"): "26 .. NeitherHiddenNorLargeInScale",
        ("FIX.5.0", "2458", "27"): "27 .. CorporateAction",
        ("FIX.5.0", "2458", "28"): "28 .. VenueDecision",
        ("FIX.5.0", "2458", "29"): "29 .. MinimumTimePeriod",
        ("FIX.5.0", "2458", "30"): "30 .. Open",
        ("FIX.5.0", "2458", "31"): "31 .. NotExecuted",
        ("FIX.5.0", "2458", "32"): "32 .. Aggressive",
        ("FIX.5.0", "2458", "33"): "33 .. Directed",
        ("FIX.5.0", "2459", "1"): "1 .. EntryRate",
        ("FIX.5.0", "2459", "2"): "2 .. ModificationRate",
        ("FIX.5.0", "2459", "3"): "3 .. CancelRate",
        ("FIX.5.0", "2459", "4"): "4 .. DownwardMove",
        ("FIX.5.0", "2459", "5"): "5 .. UpwardMove",
        ("FIX.5.0", "2461", "0"): "0 .. Seconds",
        ("FIX.5.0", "2461", "1"): "1 .. TenthsOfASecond",
        ("FIX.5.0", "2461", "2"): "2 .. HundredthsOfASecond",
        ("FIX.5.0", "2461", "3"): "3 .. Milliseconds",
        ("FIX.5.0", "2461", "4"): "4 .. Microseconds",
        ("FIX.5.0", "2461", "5"): "5 .. Nanoseconds",
        ("FIX.5.0", "2461", "10"): "10 .. Minutes",
        ("FIX.5.0", "2461", "11"): "11 .. Hours",
        ("FIX.5.0", "2461", "12"): "12 .. Days",
        ("FIX.5.0", "2461", "13"): "13 .. Weeks",
        ("FIX.5.0", "2461", "14"): "14 .. Months",
        ("FIX.5.0", "2461", "15"): "15 .. Years",
        ("FIX.5.0", "2463", "0"): "0 .. Seconds",
        ("FIX.5.0", "2463", "1"): "1 .. TenthsOfASecond",
        ("FIX.5.0", "2463", "2"): "2 .. HundredthsOfASecond",
        ("FIX.5.0", "2463", "3"): "3 .. Milliseconds",
        ("FIX.5.0", "2463", "4"): "4 .. Microseconds",
        ("FIX.5.0", "2463", "5"): "5 .. Nanoseconds",
        ("FIX.5.0", "2463", "10"): "10 .. Minutes",
        ("FIX.5.0", "2463", "11"): "11 .. Hours",
        ("FIX.5.0", "2463", "12"): "12 .. Days",
        ("FIX.5.0", "2463", "13"): "13 .. Weeks",
        ("FIX.5.0", "2463", "14"): "14 .. Months",
        ("FIX.5.0", "2463", "15"): "15 .. Years",
        ("FIX.5.0", "2464", "1"): "1 .. SlidingWindow",
        ("FIX.5.0", "2464", "2"): "2 .. SlidingWindowPeak",
        ("FIX.5.0", "2464", "3"): "3 .. FixedDateRange",
        ("FIX.5.0", "2464", "4"): "4 .. FixedTimeRange",
        ("FIX.5.0", "2464", "5"): "5 .. CurrentTimeUnit",
        ("FIX.5.0", "2464", "6"): "6 .. PreviousTimeUnit",
        ("FIX.5.0", "2464", "7"): "7 .. MaximumRange",
        ("FIX.5.0", "2464", "8"): "8 .. MaximumRangeUpToPreviousTimeUnit",
        ("FIX.5.0", "2465", "H"): "H .. Hour",
        ("FIX.5.0", "2465", "Min"): "Min .. Minute",
        ("FIX.5.0", "2465", "S"): "S .. Second",
        ("FIX.5.0", "2465", "D"): "D .. Day",
        ("FIX.5.0", "2465", "Wk"): "Wk .. Week",
        ("FIX.5.0", "2465", "Mo"): "Mo .. Month",
        ("FIX.5.0", "2465", "Yr"): "Yr .. Year",
        ("FIX.5.0", "2465", "Q"): "Q .. Quarter",
        ("FIX.5.0", "2467", "0"): "0 .. Seconds",
        ("FIX.5.0", "2467", "1"): "1 .. TenthsOfASecond",
        ("FIX.5.0", "2467", "2"): "2 .. HundredthsOfASecond",
        ("FIX.5.0", "2467", "3"): "3 .. Milliseconds",
        ("FIX.5.0", "2467", "4"): "4 .. Microseconds",
        ("FIX.5.0", "2467", "5"): "5 .. Nanoseconds",
        ("FIX.5.0", "2467", "10"): "10 .. Minutes",
        ("FIX.5.0", "2467", "11"): "11 .. Hours",
        ("FIX.5.0", "2467", "12"): "12 .. Days",
        ("FIX.5.0", "2467", "13"): "13 .. Weeks",
        ("FIX.5.0", "2467", "14"): "14 .. Months",
        ("FIX.5.0", "2467", "15"): "15 .. Years",
        ("FIX.5.0", "2472", "1"): "1 .. BuyersToSellers",
        ("FIX.5.0", "2472", "2"): "2 .. UpticksToDownticks",
        ("FIX.5.0", "2472", "3"): "3 .. MarketMakerToNonMarketMaker",
        ("FIX.5.0", "2472", "4"): "4 .. AutomatedToNonAutomated",
        ("FIX.5.0", "2472", "5"): "5 .. OrdersToTrades",
        ("FIX.5.0", "2472", "6"): "6 .. QuotesToTrades",
        ("FIX.5.0", "2472", "7"): "7 .. OrdersAndQuotesToTrades",
        ("FIX.5.0", "2472", "8"): "8 .. FailedToTotalTradedValue",
        ("FIX.5.0", "2472", "9"): "9 .. BenefitsToTotalTradedValue",
        ("FIX.5.0", "2472", "10"): "10 .. FeesToTotalTradedValue",
        ("FIX.5.0", "2472", "11"): "11 .. TradeVolumeToTotalTradedVolume",
        ("FIX.5.0", "2472", "12"): "12 .. OrdersToTotalNumberOrders",
        ("FIX.5.0", "2473", "0"): "0 .. Successful",
        ("FIX.5.0", "2473", "1"): "1 .. InvalidOrUnknownMarket",
        ("FIX.5.0", "2473", "2"): "2 .. InvalidOrUnknownMarketSegment",
        ("FIX.5.0", "2473", "3"): "3 .. InvalidOrUnknownSecurityList",
        ("FIX.5.0", "2473", "4"): "4 .. InvalidOrUnknownInstruments",
        ("FIX.5.0", "2473", "5"): "5 .. InvalidParties",
        ("FIX.5.0", "2473", "6"): "6 .. TradeDateOutOfSupportedRange",
        ("FIX.5.0", "2473", "7"): "7 .. UnsupportedStatisticType",
        ("FIX.5.0", "2473", "8"): "8 .. UnsupportedScopeOrSubScope",
        ("FIX.5.0", "2473", "9"): "9 .. UnsupportedScopeType",
        ("FIX.5.0", "2473", "10"): "10 .. MarketDepthNotSupported",
        ("FIX.5.0", "2473", "11"): "11 .. FrequencyNotSupported",
        ("FIX.5.0", "2473", "12"): "12 .. UnsupportedStatisticInterval",
        ("FIX.5.0", "2473", "13"): "13 .. UnsupportedStatisticDateRange",
        ("FIX.5.0", "2473", "14"): "14 .. UnsupportedStatisticTimeRange",
        ("FIX.5.0", "2473", "15"): "15 .. UnsupportedRatioType",
        ("FIX.5.0", "2473", "16"): "16 .. InvalidOrUnknownTradeInputSource",
        ("FIX.5.0", "2473", "17"): "17 .. InvalidOrUnknownTradingSession",
        ("FIX.5.0", "2473", "18"): "18 .. UnauthorizedForStatisticRequest",
        ("FIX.5.0", "2473", "99"): "99 .. Other",
        ("FIX.5.0", "2477", "1"): "1 .. Active",
        ("FIX.5.0", "2477", "2"): "2 .. Inactive",
        ("FIX.5.0", "2479", "1"): "1 .. Absolute",
        ("FIX.5.0", "2479", "2"): "2 .. Percentage",
        ("FIX.5.0", "2480", "0"): "0 .. Seconds",
        ("FIX.5.0", "2480", "1"): "1 .. TenthsOfASecond",
        ("FIX.5.0", "2480", "2"): "2 .. HundredthsOfASecond",
        ("FIX.5.0", "2480", "3"): "3 .. Milliseconds",
        ("FIX.5.0", "2480", "4"): "4 .. Microseconds",
        ("FIX.5.0", "2480", "5"): "5 .. Nanoseconds",
        ("FIX.5.0", "2480", "10"): "10 .. Minutes",
        ("FIX.5.0", "2480", "11"): "11 .. Hours",
        ("FIX.5.0", "2480", "12"): "12 .. Days",
        ("FIX.5.0", "2480", "13"): "13 .. Weeks",
        ("FIX.5.0", "2480", "14"): "14 .. Months",
        ("FIX.5.0", "2480", "15"): "15 .. Years",
        ("FIX.5.0", "2483", "0"): "0 .. Accepted",
        ("FIX.5.0", "2483", "1"): "1 .. Rejected",
        ("FIX.5.0", "2483", "2"): "2 .. ClaimRequired",
        ("FIX.5.0", "2483", "3"): "3 .. PreDefinedLimitCheckSucceeded",
        ("FIX.5.0", "2483", "4"): "4 .. PreDefinedLimitCheckFailed",
        ("FIX.5.0", "2483", "5"): "5 .. PreDefinedAutoAcceptRuleInvoked",
        ("FIX.5.0", "2483", "6"): "6 .. PreDefinedAutoRejectRuleInvoked",
        ("FIX.5.0", "2483", "7"): "7 .. AcceptedByClearingFirm",
        ("FIX.5.0", "2483", "8"): "8 .. RejectedByClearingFirm",
        ("FIX.5.0", "2483", "9"): "9 .. Pending",
        ("FIX.5.0", "2483", "10"): "10 .. AcceptedByCreditHub",
        ("FIX.5.0", "2483", "11"): "11 .. RejectedByCreditHub",
        ("FIX.5.0", "2483", "12"): "12 .. PendingCreditHubCheck",
        ("FIX.5.0", "2483", "13"): "13 .. AcceptedByExecVenue",
        ("FIX.5.0", "2483", "14"): "14 .. RejectedByExecVenue",
        ("FIX.5.0", "2487", "0"): "0 .. UnknownTrade",
        ("FIX.5.0", "2487", "1"): "1 .. UnknownInstrument",
        ("FIX.5.0", "2487", "2"): "2 .. UnknownCounterparty",
        ("FIX.5.0", "2487", "3"): "3 .. UnknownPosition",
        ("FIX.5.0", "2487", "4"): "4 .. UnacceptableCollateral",
        ("FIX.5.0", "2487", "99"): "99 .. Other",
        ("FIX.5.0", "2488", "0"): "0 .. Accepted",
        ("FIX.5.0", "2488", "1"): "1 .. Received",
        ("FIX.5.0", "2488", "2"): "2 .. Rejected",
        ("FIX.5.0", "2491", "1"): "1 .. Financials",
        ("FIX.5.0", "2491", "2"): "2 .. Commodities",
        ("FIX.5.0", "2491", "3"): "3 .. AlternativeInvestments",
        ("FIX.5.0", "2504", "0"): "0 .. VersusPayment",
        ("FIX.5.0", "2504", "1"): "1 .. Free",
        ("FIX.5.0", "2504", "2"): "2 .. TriParty",
        ("FIX.5.0", "2504", "3"): "3 .. HoldInCustody",
        ("FIX.5.0", "2514", "1"): "1 .. Overnight",
        ("FIX.5.0", "2514", "2"): "2 .. Term",
        ("FIX.5.0", "2514", "3"): "3 .. Flexible",
        ("FIX.5.0", "2514", "4"): "4 .. Open",
        ("FIX.5.0", "2523", "0"): "0 .. NoCross",
        ("FIX.5.0", "2523", "1"): "1 .. CrossRejected",
        ("FIX.5.0", "2523", "2"): "2 .. CrossAccepted",
        ("FIX.5.0", "2524", "0"): "0 .. NotReported",
        ("FIX.5.0", "2524", "1"): "1 .. OnBook",
        ("FIX.5.0", "2524", "2"): "2 .. SISeller",
        ("FIX.5.0", "2524", "3"): "3 .. SIBuyer",
        ("FIX.5.0", "2524", "4"): "4 .. NonSISeller",
        ("FIX.5.0", "2524", "5"): "5 .. SubDelegationByFirm",
        ("FIX.5.0", "2524", "6"): "6 .. Reportable",
        ("FIX.5.0", "2524", "7"): "7 .. NonSIBuyer",
        ("FIX.5.0", "2524", "8"): "8 .. OffBook",
        ("FIX.5.0", "2524", "9"): "9 .. NotReportable",
        ("FIX.5.0", "2530", "1"): "1 .. ASWSpread",
        ("FIX.5.0", "2530", "2"): "2 .. OIS",
        ("FIX.5.0", "2530", "3"): "3 .. ZSpread",
        ("FIX.5.0", "2530", "4"): "4 .. DiscountMargin",
        ("FIX.5.0", "2530", "5"): "5 .. ISpread",
        ("FIX.5.0", "2530", "6"): "6 .. OAS",
        ("FIX.5.0", "2530", "7"): "7 .. GSpread",
        ("FIX.5.0", "2530", "8"): "8 .. CDSBasis",
        ("FIX.5.0", "2530", "9"): "9 .. CDSInterpolatedBasis",
        ("FIX.5.0", "2532", "1"): "1 .. Bid",
        ("FIX.5.0", "2532", "2"): "2 .. Mid",
        ("FIX.5.0", "2532", "3"): "3 .. Offer",
        ("FIX.5.0", "2535", "1"): "1 .. StartInstrumentRefData",
        ("FIX.5.0", "2535", "2"): "2 .. EndInstrumentRefData",
        ("FIX.5.0", "2535", "3"): "3 .. StartOffMarketTrades",
        ("FIX.5.0", "2535", "4"): "4 .. EndOffMarketTrades",
        ("FIX.5.0", "2535", "5"): "5 .. StartOrderBookTrades",
        ("FIX.5.0", "2535", "6"): "6 .. EndOrderBookTrades",
        ("FIX.5.0", "2535", "7"): "7 .. StartOpenInterest",
        ("FIX.5.0", "2535", "8"): "8 .. EndOpenInterest",
        ("FIX.5.0", "2535", "9"): "9 .. StartSettlementPrices",
        ("FIX.5.0", "2535", "10"): "10 .. EndSettlementPrices",
        ("FIX.5.0", "2535", "11"): "11 .. StartStatsRefData",
        ("FIX.5.0", "2535", "12"): "12 .. EndStatsRefData",
        ("FIX.5.0", "2535", "13"): "13 .. StartStatistics",
        ("FIX.5.0", "2535", "14"): "14 .. EndStatistics",
        ("FIX.5.0", "2542", "1"): "1 .. Active",
        ("FIX.5.0", "2542", "2"): "2 .. Inactive",
        ("FIX.5.0", "2542", "3"): "3 .. Published",
        ("FIX.5.0", "2543", "1"): "1 .. Pool",
        ("FIX.5.0", "2543", "2"): "2 .. Retail",
        ("FIX.5.0", "2543", "3"): "3 .. Wholesale",
        ("FIX.5.0", "2544", "1"): "1 .. InterProductSpread",
        ("FIX.5.0", "2547", "1"): "1 .. MarketSegmentPoolMember",
        ("FIX.5.0", "2547", "2"): "2 .. RetailSegment",
        ("FIX.5.0", "2547", "3"): "3 .. WholesaleSegment",
        ("FIX.5.0", "2559", "N"): "N .. No",
        ("FIX.5.0", "2559", "Y"): "Y .. Yes",
        ("FIX.5.0", "2564", "0"): "0 .. Seconds",
        ("FIX.5.0", "2564", "1"): "1 .. TenthsOfASecond",
        ("FIX.5.0", "2564", "2"): "2 .. HundredthsOfASecond",
        ("FIX.5.0", "2564", "3"): "3 .. Milliseconds",
        ("FIX.5.0", "2564", "4"): "4 .. Microseconds",
        ("FIX.5.0", "2564", "5"): "5 .. Nanoseconds",
        ("FIX.5.0", "2564", "10"): "10 .. Minutes",
        ("FIX.5.0", "2564", "11"): "11 .. Hours",
        ("FIX.5.0", "2564", "12"): "12 .. Days",
        ("FIX.5.0", "2564", "13"): "13 .. Weeks",
        ("FIX.5.0", "2564", "14"): "14 .. Months",
        ("FIX.5.0", "2564", "15"): "15 .. Years",
        ("FIX.5.0", "2566", "0"): "0 .. Seconds",
        ("FIX.5.0", "2566", "1"): "1 .. TenthsOfASecond",
        ("FIX.5.0", "2566", "2"): "2 .. HundredthsOfASecond",
        ("FIX.5.0", "2566", "3"): "3 .. Milliseconds",
        ("FIX.5.0", "2566", "4"): "4 .. Microseconds",
        ("FIX.5.0", "2566", "5"): "5 .. Nanoseconds",
        ("FIX.5.0", "2566", "10"): "10 .. Minutes",
        ("FIX.5.0", "2566", "11"): "11 .. Hours",
        ("FIX.5.0", "2566", "12"): "12 .. Days",
        ("FIX.5.0", "2566", "13"): "13 .. Weeks",
        ("FIX.5.0", "2566", "14"): "14 .. Months",
        ("FIX.5.0", "2566", "15"): "15 .. Years",
        ("FIX.5.0", "2570", "0"): "0 .. NoPriority",
        ("FIX.5.0", "2570", "1"): "1 .. UnconditionalPriority",
        ("FIX.5.0", "2579", "1"): "1 .. Shares",
        ("FIX.5.0", "2579", "2"): "2 .. Derivatives",
        ("FIX.5.0", "2579", "3"): "3 .. PaymentVsPayment",
        ("FIX.5.0", "2579", "4"): "4 .. Notional",
        ("FIX.5.0", "2579", "5"): "5 .. Cascade",
        ("FIX.5.0", "2579", "6"): "6 .. Repurchase",
        ("FIX.5.0", "2579", "99"): "99 .. Other",
        ("FIX.5.0", "2581", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "2581", "1"): "1 .. None",
        ("FIX.5.0", "2581", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "2581", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "2581", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "2581", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "2581", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "2581", "7"): "7 .. NearestDay",
        ("FIX.5.0", "2592", "0"): "0 .. Automatic",
        ("FIX.5.0", "2592", "1"): "1 .. Manual",
        ("FIX.5.0", "2594", "0"): "0 .. AggregatedOrder",
        ("FIX.5.0", "2594", "1"): "1 .. PendingAllocation",
        ("FIX.5.0", "2594", "2"): "2 .. LiquidityProvisionActivityOrder",
        ("FIX.5.0", "2594", "3"): "3 .. RiskReductionOrder",
        ("FIX.5.0", "2594", "4"): "4 .. AlgorithmicOrder",
        ("FIX.5.0", "2594", "5"): "5 .. SystemicInternaliserOrder",
        ("FIX.5.0", "2594", "6"): "6 .. AllExecutionsSubmittedToAPA",
        ("FIX.5.0", "2594", "7"): "7 .. OrderExecutionInstructedByClient",
        ("FIX.5.0", "2594", "8"): "8 .. LargeInScale",
        ("FIX.5.0", "2594", "9"): "9 .. Hidden",
        ("FIX.5.0", "2599", "0"): "0 .. Close",
        ("FIX.5.0", "2599", "1"): "1 .. HedgeElection",
        ("FIX.5.0", "2601", "0"): "0 .. Bid",
        ("FIX.5.0", "2601", "1"): "1 .. Mid",
        ("FIX.5.0", "2601", "2"): "2 .. Offer",
        ("FIX.5.0", "2602", "0"): "0 .. CalculationAgent",
        ("FIX.5.0", "2602", "1"): "1 .. OptionsExchange",
        ("FIX.5.0", "2605", "0"): "0 .. Bid",
        ("FIX.5.0", "2605", "1"): "1 .. Mid",
        ("FIX.5.0", "2605", "2"): "2 .. Offer",
        ("FIX.5.0", "2606", "0"): "0 .. CalculationAgent",
        ("FIX.5.0", "2606", "1"): "1 .. OptionsExchange",
        ("FIX.5.0", "2610", "0"): "0 .. Close",
        ("FIX.5.0", "2610", "1"): "1 .. HedgeElection",
        ("FIX.5.0", "2613", "0"): "0 .. Close",
        ("FIX.5.0", "2613", "1"): "1 .. HedgeElection",
        ("FIX.5.0", "2617", "0"): "0 .. Execution",
        ("FIX.5.0", "2617", "1"): "1 .. PortfolioRebalancing",
        ("FIX.5.0", "2617", "2"): "2 .. Standard",
        ("FIX.5.0", "2621", "1"): "1 .. CUSIP",
        ("FIX.5.0", "2621", "2"): "2 .. SEDOL",
        ("FIX.5.0", "2621", "3"): "3 .. QUIK",
        ("FIX.5.0", "2621", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "2621", "5"): "5 .. RICCode",
        ("FIX.5.0", "2621", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "2621", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "2621", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "2621", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "2621", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "2621", "B"): "B .. Wertpapier",
        ("FIX.5.0", "2621", "C"): "C .. Dutch",
        ("FIX.5.0", "2621", "D"): "D .. Valoren",
        ("FIX.5.0", "2621", "E"): "E .. Sicovam",
        ("FIX.5.0", "2621", "F"): "F .. Belgian",
        ("FIX.5.0", "2621", "G"): "G .. Common",
        ("FIX.5.0", "2621", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "2621", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "2621", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "2621", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "2621", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "2621", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "2621", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "2621", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "2621", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "2621", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "2621", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "2621", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "2621", "U"): "U .. Synthetic",
        ("FIX.5.0", "2621", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "2621", "W"): "W .. IndexName",
        ("FIX.5.0", "2623", "0"): "0 .. Bid",
        ("FIX.5.0", "2623", "1"): "1 .. Mid",
        ("FIX.5.0", "2623", "2"): "2 .. Offer",
        ("FIX.5.0", "2624", "0"): "0 .. CalculationAgent",
        ("FIX.5.0", "2624", "1"): "1 .. OptionsExchange",
        ("FIX.5.0", "2632", "0"): "0 .. MarketValuation",
        ("FIX.5.0", "2632", "1"): "1 .. PortfolioValue",
        ("FIX.5.0", "2632", "2"): "2 .. ValueConfirmed",
        ("FIX.5.0", "2632", "3"): "3 .. CollateralCreditValue",
        ("FIX.5.0", "2632", "4"): "4 .. AdditionalCollateralValue",
        ("FIX.5.0", "2641", "0"): "0 .. Unspecified",
        ("FIX.5.0", "2641", "1"): "1 .. Acceptance",
        ("FIX.5.0", "2641", "2"): "2 .. Broker",
        ("FIX.5.0", "2641", "3"): "3 .. ClearingBroker",
        ("FIX.5.0", "2641", "4"): "4 .. Retail",
        ("FIX.5.0", "2641", "5"): "5 .. SalesCommission",
        ("FIX.5.0", "2641", "6"): "6 .. LocalCommission",
        ("FIX.5.0", "2641", "7"): "7 .. ResearchPayment",
        ("FIX.5.0", "2642", "1"): "1 .. PerUnit",
        ("FIX.5.0", "2642", "2"): "2 .. Percent",
        ("FIX.5.0", "2642", "3"): "3 .. Absolute",
        ("FIX.5.0", "2642", "4"): "4 .. PercentageWaivedCashDiscount",
        ("FIX.5.0", "2642", "5"): "5 .. PercentageWaivedEnhancedUnits",
        ("FIX.5.0", "2642", "6"): "6 .. PointsPerBondOrContract",
        ("FIX.5.0", "2642", "7"): "7 .. BasisPoints",
        ("FIX.5.0", "2642", "8"): "8 .. AmountPerContract",
        ("FIX.5.0", "2644", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "2644", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "2644", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "2644", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "2644", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "2644", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "2644", "thm"): "thm .. therms",
        ("FIX.5.0", "2644", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "2644", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "2644", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "2644", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "2644", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "2644", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "2644", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "2644", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "2644", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "2644", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "2644", "day"): "day .. Day",
        ("FIX.5.0", "2644", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "2644", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "2644", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "2644", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "2644", "g"): "g .. Grams",
        ("FIX.5.0", "2644", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "2644", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "2644", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "2644", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "2644", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "2644", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "2644", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "2644", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "2644", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "2644", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "2644", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "2644", "L"): "L .. liters",
        ("FIX.5.0", "2644", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "2644", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "2644", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "2644", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "2644", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "2644", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "2644", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "2644", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "2644", "t"): "t .. MetricTons",
        ("FIX.5.0", "2644", "tn"): "tn .. Tons",
        ("FIX.5.0", "2655", "0"): "0 .. Unspecified",
        ("FIX.5.0", "2655", "1"): "1 .. Acceptance",
        ("FIX.5.0", "2655", "2"): "2 .. Broker",
        ("FIX.5.0", "2655", "3"): "3 .. ClearingBroker",
        ("FIX.5.0", "2655", "4"): "4 .. Retail",
        ("FIX.5.0", "2655", "5"): "5 .. SalesCommission",
        ("FIX.5.0", "2655", "6"): "6 .. LocalCommission",
        ("FIX.5.0", "2655", "7"): "7 .. ResearchPayment",
        ("FIX.5.0", "2656", "1"): "1 .. PerUnit",
        ("FIX.5.0", "2656", "2"): "2 .. Percent",
        ("FIX.5.0", "2656", "3"): "3 .. Absolute",
        ("FIX.5.0", "2656", "4"): "4 .. PercentageWaivedCashDiscount",
        ("FIX.5.0", "2656", "5"): "5 .. PercentageWaivedEnhancedUnits",
        ("FIX.5.0", "2656", "6"): "6 .. PointsPerBondOrContract",
        ("FIX.5.0", "2656", "7"): "7 .. BasisPoints",
        ("FIX.5.0", "2656", "8"): "8 .. AmountPerContract",
        ("FIX.5.0", "2658", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "2658", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "2658", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "2658", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "2658", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "2658", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "2658", "thm"): "thm .. therms",
        ("FIX.5.0", "2658", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "2658", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "2658", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "2658", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "2658", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "2658", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "2658", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "2658", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "2658", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "2658", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "2658", "day"): "day .. Day",
        ("FIX.5.0", "2658", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "2658", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "2658", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "2658", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "2658", "g"): "g .. Grams",
        ("FIX.5.0", "2658", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "2658", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "2658", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "2658", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "2658", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "2658", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "2658", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "2658", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "2658", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "2658", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "2658", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "2658", "L"): "L .. liters",
        ("FIX.5.0", "2658", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "2658", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "2658", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "2658", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "2658", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "2658", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "2658", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "2658", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "2658", "t"): "t .. MetricTons",
        ("FIX.5.0", "2658", "tn"): "tn .. Tons",
        ("FIX.5.0", "2667", "0"): "0 .. NonAlgorithmicTrade",
        ("FIX.5.0", "2667", "1"): "1 .. AlgorithmicTrade",
        ("FIX.5.0", "2669", "0"): "0 .. PreTradeTransparencyWaiver",
        ("FIX.5.0", "2669", "1"): "1 .. PostTradeDeferral",
        ("FIX.5.0", "2669", "2"): "2 .. ExemptFromPublication",
        ("FIX.5.0", "2670", "0"): "0 .. NoBookOrderDueToAverageSpreadPrice",
        ("FIX.5.0", "2670", "1"): "1 .. NoBookOrderDueToRefPrice",
        ("FIX.5.0", "2670", "2"): "2 .. NoBookOrderDueToOtherConditions",
        ("FIX.5.0", "2670", "3"): "3 .. NoPublicPriceDueToRefPrice",
        ("FIX.5.0", "2670", "4"): "4 .. NoPublicPriceDueToIlliquid",
        ("FIX.5.0", "2670", "5"): "5 .. NoPublicPriceDueToOrderSize",
        ("FIX.5.0", "2670", "6"): "6 .. DeferralDueToLargeInScale",
        ("FIX.5.0", "2670", "7"): "7 .. DeferralDueToIlliquid",
        ("FIX.5.0", "2670", "8"): "8 .. DeferralDueToSizeSpecific",
        ("FIX.5.0", "2670", "9"): "9 .. NoPublicPriceDueToLargeInScale",
        ("FIX.5.0", "2670", "10"): "10 .. NoPublicPriceSizeDueToOrderHidden",
        ("FIX.5.0", "2670", "11"): "11 .. ExemptedDueToSecuritiesFinancingTransaction",
        ("FIX.5.0", "2670", "12"): "12 .. ExemptedDueToESCBPolicyTransaction",
        ("FIX.5.0", "2671", "0"): "0 .. NotReported",
        ("FIX.5.0", "2671", "1"): "1 .. OnBook",
        ("FIX.5.0", "2671", "2"): "2 .. SISeller",
        ("FIX.5.0", "2671", "3"): "3 .. SIBuyer",
        ("FIX.5.0", "2671", "4"): "4 .. NonSISeller",
        ("FIX.5.0", "2671", "5"): "5 .. SubDelegationByFirm",
        ("FIX.5.0", "2671", "6"): "6 .. Reportable",
        ("FIX.5.0", "2671", "7"): "7 .. NonSIBuyer",
        ("FIX.5.0", "2671", "8"): "8 .. OffBook",
        ("FIX.5.0", "2671", "9"): "9 .. NotReportable",
        ("FIX.5.0", "2675", "0"): "0 .. None",
        ("FIX.5.0", "2675", "1"): "1 .. TradingRiskControl",
        ("FIX.5.0", "2675", "2"): "2 .. ClearingRiskControl",
        ("FIX.5.0", "2675", "3"): "3 .. MarketMakerProtection",
        ("FIX.5.0", "2675", "4"): "4 .. StopTrading",
        ("FIX.5.0", "2675", "5"): "5 .. EmergencyAction",
        ("FIX.5.0", "2675", "6"): "6 .. SessionLossLogout",
        ("FIX.5.0", "2675", "7"): "7 .. DuplicateLogin",
        ("FIX.5.0", "2675", "8"): "8 .. ProductNotTraded",
        ("FIX.5.0", "2675", "9"): "9 .. InstrumentNotTraded",
        ("FIX.5.0", "2675", "10"): "10 .. CompleInstrumentDeleted",
        ("FIX.5.0", "2675", "11"): "11 .. CircuitBreakerActivated",
        ("FIX.5.0", "2675", "99"): "99 .. Other",
        ("FIX.5.0", "2677", "0"): "0 .. OrderSuspended",
        ("FIX.5.0", "2677", "1"): "1 .. InstrumentSuspended",
        ("FIX.5.0", "2679", "0"): "0 .. NoChange",
        ("FIX.5.0", "2679", "1"): "1 .. ExecutingPartyChange",
        ("FIX.5.0", "2679", "2"): "2 .. EnteringPartyChange",
        ("FIX.5.0", "2679", "3"): "3 .. SpecifiedPartyChange",
        ("FIX.5.0", "2681", "0"): "0 .. StandardITM",
        ("FIX.5.0", "2681", "1"): "1 .. ATMITM",
        ("FIX.5.0", "2681", "2"): "2 .. ATMCallITM",
        ("FIX.5.0", "2681", "3"): "3 .. ATMPutITM",
        ("FIX.5.0", "2682", "0"): "0 .. StandardITM",
        ("FIX.5.0", "2682", "1"): "1 .. ATMITM",
        ("FIX.5.0", "2682", "2"): "2 .. ATMCallITM",
        ("FIX.5.0", "2682", "3"): "3 .. ATMPutITM",
        ("FIX.5.0", "2683", "0"): "0 .. StandardITM",
        ("FIX.5.0", "2683", "1"): "1 .. ATMITM",
        ("FIX.5.0", "2683", "2"): "2 .. ATMCallITM",
        ("FIX.5.0", "2683", "3"): "3 .. ATMPutITM",
        ("FIX.5.0", "2684", "0"): "0 .. StandardITM",
        ("FIX.5.0", "2684", "1"): "1 .. ATMITM",
        ("FIX.5.0", "2684", "2"): "2 .. ATMCallITM",
        ("FIX.5.0", "2684", "3"): "3 .. ATMPutITM",
        ("FIX.5.0", "2694", "0"): "0 .. MarketValuation",
        ("FIX.5.0", "2694", "1"): "1 .. PortfolioValue",
        ("FIX.5.0", "2694", "2"): "2 .. ValueConfirmed",
        ("FIX.5.0", "2694", "3"): "3 .. CollateralCreditValue",
        ("FIX.5.0", "2694", "4"): "4 .. AdditionalCollateralValue",
        ("FIX.5.0", "2697", "D"): "D .. Divide",
        ("FIX.5.0", "2697", "M"): "M .. Multiply",
        ("FIX.5.0", "2704", "0"): "0 .. NoRestriction",
        ("FIX.5.0", "2704", "1"): "1 .. TradedOnlyOnTradingVenue",
        ("FIX.5.0", "2704", "2"): "2 .. TradedOnlyOnSI",
        ("FIX.5.0", "2704", "3"): "3 .. TradedOnTradingVenueOrSI",
        ("FIX.5.0", "2705", "0"): "0 .. Normal",
        ("FIX.5.0", "2705", "1"): "1 .. Stressed",
        ("FIX.5.0", "2705", "2"): "2 .. Exceptional",
        ("FIX.5.0", "2707", "0"): "0 .. QuoteAboveStandardMarketSize",
        ("FIX.5.0", "2707", "1"): "1 .. QuoteAboveSpecificInstrumentSize",
        ("FIX.5.0", "2707", "2"): "2 .. QuoteApplicableForLiquidtyProvisionActivity",
        ("FIX.5.0", "2710", "0"): "0 .. AccruedInterestIsFactored",
        ("FIX.5.0", "2710", "1"): "1 .. TaxIsFactored",
        ("FIX.5.0", "2710", "2"): "2 .. BondAmortizationIsFactored",
        ("FIX.5.0", "2711", "1"): "1 .. Range1",
        ("FIX.5.0", "2711", "2"): "2 .. Range2",
        ("FIX.5.0", "2711", "3"): "3 .. Range3",
        ("FIX.5.0", "2712", "0"): "0 .. Contributes",
        ("FIX.5.0", "2712", "1"): "1 .. DoesNotContribute",
        ("FIX.5.0", "2723", "D"): "D .. Day",
        ("FIX.5.0", "2723", "Wk"): "Wk .. Week",
        ("FIX.5.0", "2723", "Mo"): "Mo .. Month",
        ("FIX.5.0", "2723", "Yr"): "Yr .. Year",
        ("FIX.5.0", "2725", "0"): "0 .. ResearchPaymentAccount",
        ("FIX.5.0", "2725", "1"): "1 .. CommissionSharingAgreement",
        ("FIX.5.0", "2725", "2"): "2 .. OtherTypeResearchPayment",
        ("FIX.5.0", "2726", "0"): "0 .. ResearchPaymentAccount",
        ("FIX.5.0", "2726", "1"): "1 .. CommissionSharingAgreement",
        ("FIX.5.0", "2726", "2"): "2 .. OtherTypeResearchPayment",
        ("FIX.5.0", "2730", "D"): "D .. Day",
        ("FIX.5.0", "2730", "Wk"): "Wk .. Week",
        ("FIX.5.0", "2730", "Mo"): "Mo .. Month",
        ("FIX.5.0", "2730", "Yr"): "Yr .. Year",
        ("FIX.5.0", "2732", "1"): "1 .. CUSIP",
        ("FIX.5.0", "2732", "2"): "2 .. SEDOL",
        ("FIX.5.0", "2732", "3"): "3 .. QUIK",
        ("FIX.5.0", "2732", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "2732", "5"): "5 .. RICCode",
        ("FIX.5.0", "2732", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "2732", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "2732", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "2732", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "2732", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "2732", "B"): "B .. Wertpapier",
        ("FIX.5.0", "2732", "C"): "C .. Dutch",
        ("FIX.5.0", "2732", "D"): "D .. Valoren",
        ("FIX.5.0", "2732", "E"): "E .. Sicovam",
        ("FIX.5.0", "2732", "F"): "F .. Belgian",
        ("FIX.5.0", "2732", "G"): "G .. Common",
        ("FIX.5.0", "2732", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "2732", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "2732", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "2732", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "2732", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "2732", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "2732", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "2732", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "2732", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "2732", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "2732", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "2732", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "2732", "U"): "U .. Synthetic",
        ("FIX.5.0", "2732", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "2732", "W"): "W .. IndexName",
        ("FIX.5.0", "2736", "0"): "0 .. ArgusMcCloskey",
        ("FIX.5.0", "2736", "1"): "1 .. Baltic",
        ("FIX.5.0", "2736", "2"): "2 .. Exchange",
        ("FIX.5.0", "2736", "3"): "3 .. GlobalCoal",
        ("FIX.5.0", "2736", "4"): "4 .. IHSMcCloskey",
        ("FIX.5.0", "2736", "5"): "5 .. Platts",
        ("FIX.5.0", "2736", "99"): "99 .. Other",
        ("FIX.5.0", "2748", "0"): "0 .. AdmitToTradeRequestDate",
        ("FIX.5.0", "2748", "1"): "1 .. AdmitToTradeApprovalDate",
        ("FIX.5.0", "2748", "2"): "2 .. AdmitToTradeOrFirstTradeDate",
        ("FIX.5.0", "2748", "3"): "3 .. TerminationDate",
        ("FIX.5.0", "2753", "1"): "1 .. Dividend",
        ("FIX.5.0", "2753", "2"): "2 .. Variance",
        ("FIX.5.0", "2753", "3"): "3 .. Volatility",
        ("FIX.5.0", "2753", "4"): "4 .. TotalReturn",
        ("FIX.5.0", "2753", "5"): "5 .. ContractForDifference",
        ("FIX.5.0", "2753", "6"): "6 .. CreditDefault",
        ("FIX.5.0", "2753", "7"): "7 .. SpreadBet",
        ("FIX.5.0", "2753", "8"): "8 .. Price",
        ("FIX.5.0", "2753", "9"): "9 .. ForwardPriceUnderlyingInstrument",
        ("FIX.5.0", "2753", "99"): "99 .. Other",
        ("FIX.5.0", "2755", "1"): "1 .. Dividend",
        ("FIX.5.0", "2755", "2"): "2 .. Variance",
        ("FIX.5.0", "2755", "3"): "3 .. Volatility",
        ("FIX.5.0", "2755", "4"): "4 .. TotalReturn",
        ("FIX.5.0", "2755", "5"): "5 .. ContractForDifference",
        ("FIX.5.0", "2755", "6"): "6 .. CreditDefault",
        ("FIX.5.0", "2755", "7"): "7 .. SpreadBet",
        ("FIX.5.0", "2755", "8"): "8 .. Price",
        ("FIX.5.0", "2755", "9"): "9 .. ForwardPriceUnderlyingInstrument",
        ("FIX.5.0", "2755", "99"): "99 .. Other",
        ("FIX.5.0", "2757", "1"): "1 .. Dividend",
        ("FIX.5.0", "2757", "2"): "2 .. Variance",
        ("FIX.5.0", "2757", "3"): "3 .. Volatility",
        ("FIX.5.0", "2757", "4"): "4 .. TotalReturn",
        ("FIX.5.0", "2757", "5"): "5 .. ContractForDifference",
        ("FIX.5.0", "2757", "6"): "6 .. CreditDefault",
        ("FIX.5.0", "2757", "7"): "7 .. SpreadBet",
        ("FIX.5.0", "2757", "8"): "8 .. Price",
        ("FIX.5.0", "2757", "9"): "9 .. ForwardPriceUnderlyingInstrument",
        ("FIX.5.0", "2757", "99"): "99 .. Other",
        ("FIX.5.0", "2763", "0"): "0 .. TimeWeightedAveragePrice",
        ("FIX.5.0", "2763", "1"): "1 .. VolumeWeightedAveragePrice",
        ("FIX.5.0", "2763", "2"): "2 .. PercentOfVolumeAvveragePrice",
        ("FIX.5.0", "2763", "3"): "3 .. LimitOrderAveragePrice",
        ("FIX.5.0", "2767", "0"): "0 .. Added",
        ("FIX.5.0", "2767", "1"): "1 .. Canceled",
        ("FIX.5.0", "2767", "2"): "2 .. Replaced",
        ("FIX.5.0", "40002", "1"): "1 .. CUSIP",
        ("FIX.5.0", "40002", "2"): "2 .. SEDOL",
        ("FIX.5.0", "40002", "3"): "3 .. QUIK",
        ("FIX.5.0", "40002", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "40002", "5"): "5 .. RICCode",
        ("FIX.5.0", "40002", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "40002", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "40002", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "40002", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "40002", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "40002", "B"): "B .. Wertpapier",
        ("FIX.5.0", "40002", "C"): "C .. Dutch",
        ("FIX.5.0", "40002", "D"): "D .. Valoren",
        ("FIX.5.0", "40002", "E"): "E .. Sicovam",
        ("FIX.5.0", "40002", "F"): "F .. Belgian",
        ("FIX.5.0", "40002", "G"): "G .. Common",
        ("FIX.5.0", "40002", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "40002", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "40002", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "40002", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "40002", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "40002", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "40002", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "40002", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "40002", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "40002", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "40002", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "40002", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "40002", "U"): "U .. Synthetic",
        ("FIX.5.0", "40002", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "40002", "W"): "W .. IndexName",
        ("FIX.5.0", "40010", "SD"): "SD .. SeniorSecured",
        ("FIX.5.0", "40010", "SR"): "SR .. Senior",
        ("FIX.5.0", "40010", "SB"): "SB .. Subordinated",
        ("FIX.5.0", "40010", "JR"): "JR .. Junior",
        ("FIX.5.0", "40010", "MZ"): "MZ .. Mezzanine",
        ("FIX.5.0", "40010", "SN"): "SN .. SeniorNonPreferred",
        ("FIX.5.0", "40011", "0"): "0 .. Zero",
        ("FIX.5.0", "40011", "1"): "1 .. FixedRate",
        ("FIX.5.0", "40011", "2"): "2 .. FloatingRate",
        ("FIX.5.0", "40011", "3"): "3 .. Structured",
        ("FIX.5.0", "40017", "D"): "D .. Day",
        ("FIX.5.0", "40017", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40017", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40017", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40017", "H"): "H .. Hour",
        ("FIX.5.0", "40017", "Min"): "Min .. Minute",
        ("FIX.5.0", "40017", "S"): "S .. Second",
        ("FIX.5.0", "40017", "T"): "T .. Term",
        ("FIX.5.0", "40018", "0"): "0 .. OneOne",
        ("FIX.5.0", "40018", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "40018", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "40018", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "40018", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "40018", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "40018", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "40018", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "40018", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "40018", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "40018", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "40018", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "40018", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "40018", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "40018", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "40018", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "40018", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "40018", "17"): "17 .. Act364",
        ("FIX.5.0", "40027", "0"): "0 .. Bid",
        ("FIX.5.0", "40027", "1"): "1 .. Mid",
        ("FIX.5.0", "40027", "2"): "2 .. Offer",
        ("FIX.5.0", "40038", "0"): "0 .. Market",
        ("FIX.5.0", "40038", "1"): "1 .. Highest",
        ("FIX.5.0", "40038", "2"): "2 .. AverageMarket",
        ("FIX.5.0", "40038", "3"): "3 .. AverageHighest",
        ("FIX.5.0", "40038", "4"): "4 .. BlendedMarket",
        ("FIX.5.0", "40038", "5"): "5 .. BlendedHighest",
        ("FIX.5.0", "40038", "6"): "6 .. AverageBlendedMarket",
        ("FIX.5.0", "40038", "7"): "7 .. AverageBlendedHighest",
        ("FIX.5.0", "40050", "0"): "0 .. PaymentCashSettlement",
        ("FIX.5.0", "40050", "1"): "1 .. PhysicalDelivery",
        ("FIX.5.0", "40052", "1"): "1 .. Buy",
        ("FIX.5.0", "40052", "2"): "2 .. Sell",
        ("FIX.5.0", "40053", "1"): "1 .. Buy",
        ("FIX.5.0", "40053", "2"): "2 .. Sell",
        ("FIX.5.0", "40058", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40058", "1"): "1 .. None",
        ("FIX.5.0", "40058", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40058", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40058", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40058", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40058", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40058", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40062", "D"): "D .. Day",
        ("FIX.5.0", "40062", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40062", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40062", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40063", "0"): "0 .. Business",
        ("FIX.5.0", "40063", "1"): "1 .. Calendar",
        ("FIX.5.0", "40063", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40063", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40063", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40063", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40066", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40066", "1"): "1 .. None",
        ("FIX.5.0", "40066", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40066", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40066", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40066", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40066", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40066", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40070", "D"): "D .. Day",
        ("FIX.5.0", "40070", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40070", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40070", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40071", "0"): "0 .. Business",
        ("FIX.5.0", "40071", "1"): "1 .. Calendar",
        ("FIX.5.0", "40071", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40071", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40071", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40071", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40073", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40073", "1"): "1 .. None",
        ("FIX.5.0", "40073", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40073", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40073", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40073", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40073", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40073", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40076", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40076", "1"): "1 .. None",
        ("FIX.5.0", "40076", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40076", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40076", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40076", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40076", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40076", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40083", "D"): "D .. Day",
        ("FIX.5.0", "40083", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40083", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40083", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40083", "H"): "H .. Hour",
        ("FIX.5.0", "40083", "Min"): "Min .. Minute",
        ("FIX.5.0", "40083", "S"): "S .. Second",
        ("FIX.5.0", "40083", "T"): "T .. Term",
        ("FIX.5.0", "40084", "1"): "1 .. FirstDay",
        ("FIX.5.0", "40084", "2"): "2 .. SecondDay",
        ("FIX.5.0", "40084", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "40084", "4"): "4 .. FourthDay",
        ("FIX.5.0", "40084", "5"): "5 .. FifthDay",
        ("FIX.5.0", "40084", "6"): "6 .. SixthDay",
        ("FIX.5.0", "40084", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "40084", "8"): "8 .. EighthDay",
        ("FIX.5.0", "40084", "9"): "9 .. NinthDay",
        ("FIX.5.0", "40084", "10"): "10 .. TenthDay",
        ("FIX.5.0", "40084", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "40084", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "40084", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "40084", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "40084", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "40084", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "40084", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "40084", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "40084", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "40084", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "40084", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "40084", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "40084", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "40084", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "40084", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "40084", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "40084", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "40084", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "40084", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "40084", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "40084", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "40084", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "40084", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "40084", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "40084", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "40084", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "40084", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "40084", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "40084", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "40084", "MON"): "MON .. MON",
        ("FIX.5.0", "40084", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "40084", "WED"): "WED .. WED",
        ("FIX.5.0", "40084", "THU"): "THU .. THU",
        ("FIX.5.0", "40084", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "40084", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "40084", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "40087", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40087", "1"): "1 .. Reuters",
        ("FIX.5.0", "40087", "2"): "2 .. Telerate",
        ("FIX.5.0", "40087", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "40087", "99"): "99 .. Other",
        ("FIX.5.0", "40089", "0"): "0 .. ExercisingParty",
        ("FIX.5.0", "40089", "1"): "1 .. NonExercisingParty",
        ("FIX.5.0", "40089", "2"): "2 .. MasterAgreeent",
        ("FIX.5.0", "40089", "3"): "3 .. Supplement",
        ("FIX.5.0", "40091", "0"): "0 .. MandatoryEarlyTermination",
        ("FIX.5.0", "40091", "1"): "1 .. OptionalEarlyTermination",
        ("FIX.5.0", "40091", "2"): "2 .. Cancelable",
        ("FIX.5.0", "40091", "3"): "3 .. Extendable",
        ("FIX.5.0", "40091", "4"): "4 .. MutualEarlyTermination",
        ("FIX.5.0", "40093", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40093", "1"): "1 .. None",
        ("FIX.5.0", "40093", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40093", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40093", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40093", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40093", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40093", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40097", "D"): "D .. Day",
        ("FIX.5.0", "40097", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40097", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40097", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40098", "0"): "0 .. ExercisingParty",
        ("FIX.5.0", "40098", "1"): "1 .. NonExercisingParty",
        ("FIX.5.0", "40098", "2"): "2 .. MasterAgreeent",
        ("FIX.5.0", "40098", "3"): "3 .. Supplement",
        ("FIX.5.0", "40099", "1"): "1 .. Buy",
        ("FIX.5.0", "40099", "2"): "2 .. Sell",
        ("FIX.5.0", "40100", "1"): "1 .. Buy",
        ("FIX.5.0", "40100", "2"): "2 .. Sell",
        ("FIX.5.0", "40101", "0"): "0 .. European",
        ("FIX.5.0", "40101", "1"): "1 .. American",
        ("FIX.5.0", "40101", "2"): "2 .. Bermuda",
        ("FIX.5.0", "40101", "99"): "99 .. Other",
        ("FIX.5.0", "40108", "0"): "0 .. CashPrice",
        ("FIX.5.0", "40108", "1"): "1 .. CashPriceAlternate",
        ("FIX.5.0", "40108", "2"): "2 .. ParYieldCurveAdjusted",
        ("FIX.5.0", "40108", "3"): "3 .. ZeroCouponYieldCurveAdjusted",
        ("FIX.5.0", "40108", "4"): "4 .. ParYieldCurveUnadjusted",
        ("FIX.5.0", "40108", "5"): "5 .. CrossCurrency",
        ("FIX.5.0", "40108", "6"): "6 .. CollateralizedPrice",
        ("FIX.5.0", "40111", "0"): "0 .. Bid",
        ("FIX.5.0", "40111", "1"): "1 .. Mid",
        ("FIX.5.0", "40111", "2"): "2 .. Offer",
        ("FIX.5.0", "40111", "3"): "3 .. ExercisingPartyPays",
        ("FIX.5.0", "40112", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40112", "1"): "1 .. Reuters",
        ("FIX.5.0", "40112", "2"): "2 .. Telerate",
        ("FIX.5.0", "40112", "99"): "99 .. Other",
        ("FIX.5.0", "40116", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40116", "1"): "1 .. None",
        ("FIX.5.0", "40116", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40116", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40116", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40116", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40116", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40116", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40120", "D"): "D .. Day",
        ("FIX.5.0", "40120", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40120", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40120", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40121", "0"): "0 .. Business",
        ("FIX.5.0", "40121", "1"): "1 .. Calendar",
        ("FIX.5.0", "40121", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40121", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40121", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40121", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40123", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40123", "1"): "1 .. None",
        ("FIX.5.0", "40123", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40123", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40123", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40123", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40123", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40123", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40126", "D"): "D .. Day",
        ("FIX.5.0", "40126", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40126", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40126", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40128", "D"): "D .. Day",
        ("FIX.5.0", "40128", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40128", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40128", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40128", "H"): "H .. Hour",
        ("FIX.5.0", "40128", "Min"): "Min .. Minute",
        ("FIX.5.0", "40128", "S"): "S .. Second",
        ("FIX.5.0", "40128", "T"): "T .. Term",
        ("FIX.5.0", "40132", "D"): "D .. Day",
        ("FIX.5.0", "40132", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40132", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40132", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40133", "0"): "0 .. Business",
        ("FIX.5.0", "40133", "1"): "1 .. Calendar",
        ("FIX.5.0", "40133", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40133", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40133", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40133", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40144", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "40144", "1"): "1 .. Adjusted",
        ("FIX.5.0", "40146", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40146", "1"): "1 .. None",
        ("FIX.5.0", "40146", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40146", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40146", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40146", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40146", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40146", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40150", "D"): "D .. Day",
        ("FIX.5.0", "40150", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40150", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40150", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40151", "0"): "0 .. Business",
        ("FIX.5.0", "40151", "1"): "1 .. Calendar",
        ("FIX.5.0", "40151", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40151", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40151", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40151", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40156", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40156", "1"): "1 .. None",
        ("FIX.5.0", "40156", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40156", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40156", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40156", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40156", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40156", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40160", "D"): "D .. Day",
        ("FIX.5.0", "40160", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40160", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40160", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40161", "0"): "0 .. Business",
        ("FIX.5.0", "40161", "1"): "1 .. Calendar",
        ("FIX.5.0", "40161", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40161", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40161", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40161", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40163", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40163", "1"): "1 .. None",
        ("FIX.5.0", "40163", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40163", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40163", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40163", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40163", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40163", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40167", "D"): "D .. Day",
        ("FIX.5.0", "40167", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40167", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40167", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40168", "0"): "0 .. Business",
        ("FIX.5.0", "40168", "1"): "1 .. Calendar",
        ("FIX.5.0", "40168", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40168", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40168", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40168", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40173", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "40173", "1"): "1 .. Adjusted",
        ("FIX.5.0", "40176", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "40176", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "40176", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "40176", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "40176", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "40176", "J"): "J .. TaxID",
        ("FIX.5.0", "40176", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "40176", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "40176", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "40176", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "40176", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "40176", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "40176", "B"): "B .. BIC",
        ("FIX.5.0", "40176", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "40176", "D"): "D .. Proprietary",
        ("FIX.5.0", "40176", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "40176", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "40176", "G"): "G .. MIC",
        ("FIX.5.0", "40176", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "40176", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "40176", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "40176", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "40176", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "40176", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "40176", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "40176", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "40177", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "40177", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "40177", "3"): "3 .. ClientID",
        ("FIX.5.0", "40177", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "40177", "5"): "5 .. InvestorID",
        ("FIX.5.0", "40177", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "40177", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "40177", "8"): "8 .. Locate",
        ("FIX.5.0", "40177", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "40177", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "40177", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "40177", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "40177", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "40177", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "40177", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "40177", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "40177", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "40177", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "40177", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "40177", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "40177", "22"): "22 .. Exchange",
        ("FIX.5.0", "40177", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "40177", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "40177", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "40177", "27"): "27 .. Buyer",
        ("FIX.5.0", "40177", "28"): "28 .. Custodian",
        ("FIX.5.0", "40177", "29"): "29 .. Intermediary",
        ("FIX.5.0", "40177", "30"): "30 .. Agent",
        ("FIX.5.0", "40177", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "40177", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "40177", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "40177", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "40177", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "40177", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "40177", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "40177", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "40177", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "40177", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "40177", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "40177", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "40177", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "40177", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "40177", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "40177", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "40177", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "40177", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "40177", "49"): "49 .. AssetManager",
        ("FIX.5.0", "40177", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "40177", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "40177", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "40177", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "40177", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "40177", "55"): "55 .. SessionID",
        ("FIX.5.0", "40177", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "40177", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "40177", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "40177", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "40177", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "40177", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "40177", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "40177", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "40177", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "40177", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "40177", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "40177", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "40177", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "40177", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "40177", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "40177", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "40177", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "40177", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "40177", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "40177", "75"): "75 .. LocationID",
        ("FIX.5.0", "40177", "76"): "76 .. DeskID",
        ("FIX.5.0", "40177", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "40177", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "40177", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "40177", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "40177", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "40177", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "40177", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "40177", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "40177", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "40177", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "40177", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "40177", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "40177", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "40177", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "40177", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "40177", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "40177", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "40177", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "40177", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "40177", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "40177", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "40177", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "40177", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "40177", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "40177", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "40177", "102"): "102 .. DataRepository",
        ("FIX.5.0", "40177", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "40177", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "40177", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "40177", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "40177", "107"): "107 .. Correspondent",
        ("FIX.5.0", "40177", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "40177", "110"): "110 .. Borrower",
        ("FIX.5.0", "40177", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "40177", "112"): "112 .. Guarantor",
        ("FIX.5.0", "40177", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "40177", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "40177", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "40177", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "40177", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "40177", "118"): "118 .. Operator",
        ("FIX.5.0", "40177", "119"): "119 .. CSD",
        ("FIX.5.0", "40177", "120"): "120 .. ICSD",
        ("FIX.5.0", "40177", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "40177", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "40177", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "40180", "1"): "1 .. Firm",
        ("FIX.5.0", "40180", "2"): "2 .. Person",
        ("FIX.5.0", "40180", "3"): "3 .. System",
        ("FIX.5.0", "40180", "4"): "4 .. Application",
        ("FIX.5.0", "40180", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "40180", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "40180", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "40180", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "40180", "9"): "9 .. ContactName",
        ("FIX.5.0", "40180", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "40180", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "40180", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "40180", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "40180", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "40180", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "40180", "16"): "16 .. BIC",
        ("FIX.5.0", "40180", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "40180", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "40180", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "40180", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "40180", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "40180", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "40180", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "40180", "24"): "24 .. Department",
        ("FIX.5.0", "40180", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "40180", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "40180", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "40180", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "40180", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "40180", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "40180", "31"): "31 .. Location",
        ("FIX.5.0", "40180", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "40180", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "40180", "34"): "34 .. AddressCity",
        ("FIX.5.0", "40180", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "40180", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "40180", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "40180", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "40180", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "40180", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "40180", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "40180", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "40180", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "40180", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "40180", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "40180", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "40180", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "40180", "48"): "48 .. USPerson",
        ("FIX.5.0", "40180", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "40180", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "40180", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "40180", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "40180", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "40180", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "40180", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "40180", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "40180", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "40180", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "40180", "59"): "59 .. FCMCode",
        ("FIX.5.0", "40180", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "40180", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "40180", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "40180", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "40180", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "40180", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "40180", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "40180", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "40180", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "40180", "69"): "69 .. EndUser",
        ("FIX.5.0", "40180", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "40180", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "40180", "72"): "72 .. Domicile",
        ("FIX.5.0", "40180", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "40180", "74"): "74 .. Payer",
        ("FIX.5.0", "40180", "75"): "75 .. Receiver",
        ("FIX.5.0", "40180", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "40180", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "40180", "78"): "78 .. FirstName",
        ("FIX.5.0", "40180", "79"): "79 .. Surname",
        ("FIX.5.0", "40180", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "40180", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "40180", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "40180", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "40180", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "40196", "D"): "D .. Day",
        ("FIX.5.0", "40196", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40196", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40196", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40197", "0"): "0 .. Business",
        ("FIX.5.0", "40197", "1"): "1 .. Calendar",
        ("FIX.5.0", "40197", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40197", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40197", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40197", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40200", "H"): "H .. RestructuringMultipleHoldingObligations",
        ("FIX.5.0", "40200", "E"): "E .. RestructuringMultipleCreditEventNotices",
        ("FIX.5.0", "40200", "C"): "C .. FloatingRateInterestShortfall",
        ("FIX.5.0", "40213", "0"): "0 .. Brokerage",
        ("FIX.5.0", "40213", "1"): "1 .. UpfrontFee",
        ("FIX.5.0", "40213", "2"): "2 .. IndependentAmountCollateral",
        ("FIX.5.0", "40213", "3"): "3 .. PrincipalExchange",
        ("FIX.5.0", "40213", "4"): "4 .. NovationTermination",
        ("FIX.5.0", "40213", "5"): "5 .. EarlyTerminationProvision",
        ("FIX.5.0", "40213", "6"): "6 .. CancelableProvision",
        ("FIX.5.0", "40213", "7"): "7 .. ExtendibleProvision",
        ("FIX.5.0", "40213", "8"): "8 .. CapRateProvision",
        ("FIX.5.0", "40213", "9"): "9 .. FloorRateProvision",
        ("FIX.5.0", "40213", "10"): "10 .. OptionPremium",
        ("FIX.5.0", "40213", "11"): "11 .. SettlementPayment",
        ("FIX.5.0", "40213", "12"): "12 .. CashSettl",
        ("FIX.5.0", "40213", "99"): "99 .. Other",
        ("FIX.5.0", "40214", "1"): "1 .. Buy",
        ("FIX.5.0", "40214", "2"): "2 .. Sell",
        ("FIX.5.0", "40215", "1"): "1 .. Buy",
        ("FIX.5.0", "40215", "2"): "2 .. Sell",
        ("FIX.5.0", "40220", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40220", "1"): "1 .. None",
        ("FIX.5.0", "40220", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40220", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40220", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40220", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40220", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40220", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40227", "0"): "0 .. Standard",
        ("FIX.5.0", "40227", "1"): "1 .. Net",
        ("FIX.5.0", "40227", "2"): "2 .. StandardfNet",
        ("FIX.5.0", "40235", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "40235", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "40235", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "40235", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "40235", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "40235", "J"): "J .. TaxID",
        ("FIX.5.0", "40235", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "40235", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "40235", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "40235", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "40235", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "40235", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "40235", "B"): "B .. BIC",
        ("FIX.5.0", "40235", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "40235", "D"): "D .. Proprietary",
        ("FIX.5.0", "40235", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "40235", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "40235", "G"): "G .. MIC",
        ("FIX.5.0", "40235", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "40235", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "40235", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "40235", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "40235", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "40235", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "40235", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "40235", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "40236", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "40236", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "40236", "3"): "3 .. ClientID",
        ("FIX.5.0", "40236", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "40236", "5"): "5 .. InvestorID",
        ("FIX.5.0", "40236", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "40236", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "40236", "8"): "8 .. Locate",
        ("FIX.5.0", "40236", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "40236", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "40236", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "40236", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "40236", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "40236", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "40236", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "40236", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "40236", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "40236", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "40236", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "40236", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "40236", "22"): "22 .. Exchange",
        ("FIX.5.0", "40236", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "40236", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "40236", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "40236", "27"): "27 .. Buyer",
        ("FIX.5.0", "40236", "28"): "28 .. Custodian",
        ("FIX.5.0", "40236", "29"): "29 .. Intermediary",
        ("FIX.5.0", "40236", "30"): "30 .. Agent",
        ("FIX.5.0", "40236", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "40236", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "40236", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "40236", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "40236", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "40236", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "40236", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "40236", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "40236", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "40236", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "40236", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "40236", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "40236", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "40236", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "40236", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "40236", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "40236", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "40236", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "40236", "49"): "49 .. AssetManager",
        ("FIX.5.0", "40236", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "40236", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "40236", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "40236", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "40236", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "40236", "55"): "55 .. SessionID",
        ("FIX.5.0", "40236", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "40236", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "40236", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "40236", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "40236", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "40236", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "40236", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "40236", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "40236", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "40236", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "40236", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "40236", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "40236", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "40236", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "40236", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "40236", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "40236", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "40236", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "40236", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "40236", "75"): "75 .. LocationID",
        ("FIX.5.0", "40236", "76"): "76 .. DeskID",
        ("FIX.5.0", "40236", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "40236", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "40236", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "40236", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "40236", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "40236", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "40236", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "40236", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "40236", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "40236", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "40236", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "40236", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "40236", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "40236", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "40236", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "40236", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "40236", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "40236", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "40236", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "40236", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "40236", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "40236", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "40236", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "40236", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "40236", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "40236", "102"): "102 .. DataRepository",
        ("FIX.5.0", "40236", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "40236", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "40236", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "40236", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "40236", "107"): "107 .. Correspondent",
        ("FIX.5.0", "40236", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "40236", "110"): "110 .. Borrower",
        ("FIX.5.0", "40236", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "40236", "112"): "112 .. Guarantor",
        ("FIX.5.0", "40236", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "40236", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "40236", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "40236", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "40236", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "40236", "118"): "118 .. Operator",
        ("FIX.5.0", "40236", "119"): "119 .. CSD",
        ("FIX.5.0", "40236", "120"): "120 .. ICSD",
        ("FIX.5.0", "40236", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "40236", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "40236", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "40237", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "40237", "18"): "18 .. Current",
        ("FIX.5.0", "40237", "19"): "19 .. New",
        ("FIX.5.0", "40237", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "40237", "0"): "0 .. Agency",
        ("FIX.5.0", "40237", "1"): "1 .. Principal",
        ("FIX.5.0", "40237", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "40237", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "40237", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "40237", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "40237", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "40237", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "40237", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "40237", "27"): "27 .. Supervisor",
        ("FIX.5.0", "40237", "22"): "22 .. Algorithm",
        ("FIX.5.0", "40237", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "40237", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "40237", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "40237", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "40237", "7"): "7 .. Bank",
        ("FIX.5.0", "40237", "8"): "8 .. Hub",
        ("FIX.5.0", "40237", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "40237", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "40237", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "40237", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "40237", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "40237", "21"): "21 .. Specialist",
        ("FIX.5.0", "40237", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "40240", "1"): "1 .. Firm",
        ("FIX.5.0", "40240", "2"): "2 .. Person",
        ("FIX.5.0", "40240", "3"): "3 .. System",
        ("FIX.5.0", "40240", "4"): "4 .. Application",
        ("FIX.5.0", "40240", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "40240", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "40240", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "40240", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "40240", "9"): "9 .. ContactName",
        ("FIX.5.0", "40240", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "40240", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "40240", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "40240", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "40240", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "40240", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "40240", "16"): "16 .. BIC",
        ("FIX.5.0", "40240", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "40240", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "40240", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "40240", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "40240", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "40240", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "40240", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "40240", "24"): "24 .. Department",
        ("FIX.5.0", "40240", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "40240", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "40240", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "40240", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "40240", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "40240", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "40240", "31"): "31 .. Location",
        ("FIX.5.0", "40240", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "40240", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "40240", "34"): "34 .. AddressCity",
        ("FIX.5.0", "40240", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "40240", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "40240", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "40240", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "40240", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "40240", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "40240", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "40240", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "40240", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "40240", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "40240", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "40240", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "40240", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "40240", "48"): "48 .. USPerson",
        ("FIX.5.0", "40240", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "40240", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "40240", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "40240", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "40240", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "40240", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "40240", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "40240", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "40240", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "40240", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "40240", "59"): "59 .. FCMCode",
        ("FIX.5.0", "40240", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "40240", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "40240", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "40240", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "40240", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "40240", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "40240", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "40240", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "40240", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "40240", "69"): "69 .. EndUser",
        ("FIX.5.0", "40240", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "40240", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "40240", "72"): "72 .. Domicile",
        ("FIX.5.0", "40240", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "40240", "74"): "74 .. Payer",
        ("FIX.5.0", "40240", "75"): "75 .. Receiver",
        ("FIX.5.0", "40240", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "40240", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "40240", "78"): "78 .. FirstName",
        ("FIX.5.0", "40240", "79"): "79 .. Surname",
        ("FIX.5.0", "40240", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "40240", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "40240", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "40240", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "40240", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "40242", "0"): "0 .. PaymentCashSettlement",
        ("FIX.5.0", "40242", "1"): "1 .. PhysicalDelivery",
        ("FIX.5.0", "40244", "1"): "1 .. Buy",
        ("FIX.5.0", "40244", "2"): "2 .. Sell",
        ("FIX.5.0", "40245", "1"): "1 .. Buy",
        ("FIX.5.0", "40245", "2"): "2 .. Sell",
        ("FIX.5.0", "40250", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40250", "1"): "1 .. None",
        ("FIX.5.0", "40250", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40250", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40250", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40250", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40250", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40250", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40254", "D"): "D .. Day",
        ("FIX.5.0", "40254", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40254", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40254", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40255", "0"): "0 .. Business",
        ("FIX.5.0", "40255", "1"): "1 .. Calendar",
        ("FIX.5.0", "40255", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40255", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40255", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40255", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40258", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40258", "1"): "1 .. None",
        ("FIX.5.0", "40258", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40258", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40258", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40258", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40258", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40258", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40262", "D"): "D .. Day",
        ("FIX.5.0", "40262", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40262", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40262", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40263", "0"): "0 .. Business",
        ("FIX.5.0", "40263", "1"): "1 .. Calendar",
        ("FIX.5.0", "40263", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40263", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40263", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40263", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40265", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40265", "1"): "1 .. None",
        ("FIX.5.0", "40265", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40265", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40265", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40265", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40265", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40265", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40268", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40268", "1"): "1 .. None",
        ("FIX.5.0", "40268", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40268", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40268", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40268", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40268", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40268", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40275", "D"): "D .. Day",
        ("FIX.5.0", "40275", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40275", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40275", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40275", "H"): "H .. Hour",
        ("FIX.5.0", "40275", "Min"): "Min .. Minute",
        ("FIX.5.0", "40275", "S"): "S .. Second",
        ("FIX.5.0", "40275", "T"): "T .. Term",
        ("FIX.5.0", "40276", "1"): "1 .. FirstDay",
        ("FIX.5.0", "40276", "2"): "2 .. SecondDay",
        ("FIX.5.0", "40276", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "40276", "4"): "4 .. FourthDay",
        ("FIX.5.0", "40276", "5"): "5 .. FifthDay",
        ("FIX.5.0", "40276", "6"): "6 .. SixthDay",
        ("FIX.5.0", "40276", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "40276", "8"): "8 .. EighthDay",
        ("FIX.5.0", "40276", "9"): "9 .. NinthDay",
        ("FIX.5.0", "40276", "10"): "10 .. TenthDay",
        ("FIX.5.0", "40276", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "40276", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "40276", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "40276", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "40276", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "40276", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "40276", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "40276", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "40276", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "40276", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "40276", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "40276", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "40276", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "40276", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "40276", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "40276", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "40276", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "40276", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "40276", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "40276", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "40276", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "40276", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "40276", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "40276", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "40276", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "40276", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "40276", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "40276", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "40276", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "40276", "MON"): "MON .. MON",
        ("FIX.5.0", "40276", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "40276", "WED"): "WED .. WED",
        ("FIX.5.0", "40276", "THU"): "THU .. THU",
        ("FIX.5.0", "40276", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "40276", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "40276", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "40279", "0"): "0 .. Periodic",
        ("FIX.5.0", "40279", "1"): "1 .. Initial",
        ("FIX.5.0", "40279", "2"): "2 .. Single",
        ("FIX.5.0", "40279", "3"): "3 .. Dividend",
        ("FIX.5.0", "40279", "4"): "4 .. Interest",
        ("FIX.5.0", "40279", "5"): "5 .. DividendReturn",
        ("FIX.5.0", "40279", "6"): "6 .. PriceReturn",
        ("FIX.5.0", "40279", "7"): "7 .. TotalReturn",
        ("FIX.5.0", "40279", "8"): "8 .. Variance",
        ("FIX.5.0", "40279", "9"): "9 .. Correlation",
        ("FIX.5.0", "40283", "0"): "0 .. OneOne",
        ("FIX.5.0", "40283", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "40283", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "40283", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "40283", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "40283", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "40283", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "40283", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "40283", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "40283", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "40283", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "40283", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "40283", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "40283", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "40283", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "40283", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "40283", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "40283", "17"): "17 .. Act364",
        ("FIX.5.0", "40285", "0"): "0 .. Standard",
        ("FIX.5.0", "40285", "1"): "1 .. FRA",
        ("FIX.5.0", "40287", "0"): "0 .. OneOne",
        ("FIX.5.0", "40287", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "40287", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "40287", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "40287", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "40287", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "40287", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "40287", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "40287", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "40287", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "40287", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "40287", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "40287", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "40287", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "40287", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "40287", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "40287", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "40287", "17"): "17 .. Act364",
        ("FIX.5.0", "40288", "0"): "0 .. None",
        ("FIX.5.0", "40288", "1"): "1 .. Flat",
        ("FIX.5.0", "40288", "2"): "2 .. Straight",
        ("FIX.5.0", "40288", "3"): "3 .. SpreadExclusive",
        ("FIX.5.0", "40292", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40292", "1"): "1 .. None",
        ("FIX.5.0", "40292", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40292", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40292", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40292", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40292", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40292", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40295", "D"): "D .. Day",
        ("FIX.5.0", "40295", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40295", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40295", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40295", "T"): "T .. Term",
        ("FIX.5.0", "40296", "1"): "1 .. FirstDay",
        ("FIX.5.0", "40296", "2"): "2 .. SecondDay",
        ("FIX.5.0", "40296", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "40296", "4"): "4 .. FourthDay",
        ("FIX.5.0", "40296", "5"): "5 .. FifthDay",
        ("FIX.5.0", "40296", "6"): "6 .. SixthDay",
        ("FIX.5.0", "40296", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "40296", "8"): "8 .. EighthDay",
        ("FIX.5.0", "40296", "9"): "9 .. NinthDay",
        ("FIX.5.0", "40296", "10"): "10 .. TenthDay",
        ("FIX.5.0", "40296", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "40296", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "40296", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "40296", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "40296", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "40296", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "40296", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "40296", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "40296", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "40296", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "40296", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "40296", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "40296", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "40296", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "40296", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "40296", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "40296", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "40296", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "40296", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "40296", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "40296", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "40296", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "40296", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "40296", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "40296", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "40296", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "40296", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "40296", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "40296", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "40296", "MON"): "MON .. MON",
        ("FIX.5.0", "40296", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "40296", "WED"): "WED .. WED",
        ("FIX.5.0", "40296", "THU"): "THU .. THU",
        ("FIX.5.0", "40296", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "40296", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "40296", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "40301", "D"): "D .. Day",
        ("FIX.5.0", "40301", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40301", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40301", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40302", "0"): "0 .. Business",
        ("FIX.5.0", "40302", "1"): "1 .. Calendar",
        ("FIX.5.0", "40302", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40302", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40302", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40302", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40304", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40304", "1"): "1 .. None",
        ("FIX.5.0", "40304", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40304", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40304", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40304", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40304", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40304", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40307", "D"): "D .. Day",
        ("FIX.5.0", "40307", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40307", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40307", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40307", "H"): "H .. Hour",
        ("FIX.5.0", "40307", "Min"): "Min .. Minute",
        ("FIX.5.0", "40307", "S"): "S .. Second",
        ("FIX.5.0", "40307", "T"): "T .. Term",
        ("FIX.5.0", "40308", "MON"): "MON .. Monday",
        ("FIX.5.0", "40308", "TUE"): "TUE .. Tuesday",
        ("FIX.5.0", "40308", "WED"): "WED .. Wednesday",
        ("FIX.5.0", "40308", "THU"): "THU .. Thursday",
        ("FIX.5.0", "40308", "FRI"): "FRI .. Friday",
        ("FIX.5.0", "40308", "SAT"): "SAT .. Saturday",
        ("FIX.5.0", "40308", "SUN"): "SUN .. Sunday",
        ("FIX.5.0", "40310", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40310", "1"): "1 .. None",
        ("FIX.5.0", "40310", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40310", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40310", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40310", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40310", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40310", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40313", "D"): "D .. Day",
        ("FIX.5.0", "40313", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40313", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40313", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40314", "0"): "0 .. Business",
        ("FIX.5.0", "40314", "1"): "1 .. Calendar",
        ("FIX.5.0", "40314", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40314", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40314", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40314", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40317", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40317", "1"): "1 .. None",
        ("FIX.5.0", "40317", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40317", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40317", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40317", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40317", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40317", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40320", "D"): "D .. Day",
        ("FIX.5.0", "40320", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40320", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40320", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40321", "0"): "0 .. Business",
        ("FIX.5.0", "40321", "1"): "1 .. Calendar",
        ("FIX.5.0", "40321", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40321", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40321", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40321", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40324", "D"): "D .. Day",
        ("FIX.5.0", "40324", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40324", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40324", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40325", "0"): "0 .. Business",
        ("FIX.5.0", "40325", "1"): "1 .. Calendar",
        ("FIX.5.0", "40325", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40325", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40325", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40325", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40332", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40332", "1"): "1 .. Reuters",
        ("FIX.5.0", "40332", "2"): "2 .. Telerate",
        ("FIX.5.0", "40332", "99"): "99 .. Other",
        ("FIX.5.0", "40333", "D"): "D .. Day",
        ("FIX.5.0", "40333", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40333", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40333", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40337", "0"): "0 .. Short",
        ("FIX.5.0", "40337", "1"): "1 .. Long",
        ("FIX.5.0", "40338", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40338", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40340", "1"): "1 .. Buyer",
        ("FIX.5.0", "40340", "2"): "2 .. Seller",
        ("FIX.5.0", "40341", "1"): "1 .. Buyer",
        ("FIX.5.0", "40341", "2"): "2 .. Seller",
        ("FIX.5.0", "40343", "1"): "1 .. Buyer",
        ("FIX.5.0", "40343", "2"): "2 .. Seller",
        ("FIX.5.0", "40344", "1"): "1 .. Buyer",
        ("FIX.5.0", "40344", "2"): "2 .. Seller",
        ("FIX.5.0", "40346", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "40346", "1"): "1 .. RoundDown",
        ("FIX.5.0", "40346", "2"): "2 .. RoundUp",
        ("FIX.5.0", "40348", "0"): "0 .. Unweighted",
        ("FIX.5.0", "40348", "1"): "1 .. Weighted",
        ("FIX.5.0", "40349", "0"): "0 .. ZeroInterestRateMethod",
        ("FIX.5.0", "40349", "1"): "1 .. NegativeInterestRateMethod",
        ("FIX.5.0", "40351", "D"): "D .. Day",
        ("FIX.5.0", "40351", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40351", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40351", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40352", "0"): "0 .. Business",
        ("FIX.5.0", "40352", "1"): "1 .. Calendar",
        ("FIX.5.0", "40352", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40352", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40352", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40352", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40353", "0"): "0 .. None",
        ("FIX.5.0", "40353", "1"): "1 .. LinearZeroYield",
        ("FIX.5.0", "40354", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40354", "1"): "1 .. Reuters",
        ("FIX.5.0", "40354", "2"): "2 .. Telerate",
        ("FIX.5.0", "40354", "99"): "99 .. Other",
        ("FIX.5.0", "40358", "0"): "0 .. None",
        ("FIX.5.0", "40358", "1"): "1 .. ISDA",
        ("FIX.5.0", "40358", "2"): "2 .. AFMA",
        ("FIX.5.0", "40360", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40360", "1"): "1 .. None",
        ("FIX.5.0", "40360", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40360", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40360", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40360", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40360", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40360", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40364", "D"): "D .. Day",
        ("FIX.5.0", "40364", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40364", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40364", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40365", "0"): "0 .. Business",
        ("FIX.5.0", "40365", "1"): "1 .. Calendar",
        ("FIX.5.0", "40365", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40365", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40365", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40365", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40366", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40366", "1"): "1 .. Reuters",
        ("FIX.5.0", "40366", "2"): "2 .. Telerate",
        ("FIX.5.0", "40366", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "40366", "99"): "99 .. Other",
        ("FIX.5.0", "40369", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "40369", "1"): "1 .. Adjusted",
        ("FIX.5.0", "40371", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40371", "1"): "1 .. Reuters",
        ("FIX.5.0", "40371", "2"): "2 .. Telerate",
        ("FIX.5.0", "40371", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "40371", "99"): "99 .. Other",
        ("FIX.5.0", "40373", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40373", "1"): "1 .. Reuters",
        ("FIX.5.0", "40373", "2"): "2 .. Telerate",
        ("FIX.5.0", "40373", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "40373", "99"): "99 .. Other",
        ("FIX.5.0", "40375", "0"): "0 .. Notional",
        ("FIX.5.0", "40375", "1"): "1 .. CashFlow",
        ("FIX.5.0", "40375", "2"): "2 .. FXLinkedNotional",
        ("FIX.5.0", "40375", "3"): "3 .. FixedRate",
        ("FIX.5.0", "40375", "4"): "4 .. FutureValueNotional",
        ("FIX.5.0", "40375", "5"): "5 .. KnownAmount",
        ("FIX.5.0", "40375", "6"): "6 .. FloatingRateMultiplier",
        ("FIX.5.0", "40375", "7"): "7 .. Spread",
        ("FIX.5.0", "40375", "8"): "8 .. CapRate",
        ("FIX.5.0", "40375", "9"): "9 .. FloorRate",
        ("FIX.5.0", "40375", "10"): "10 .. NonDeliverableSettlPaymentDates",
        ("FIX.5.0", "40375", "11"): "11 .. NonDeliverableSettlCalculationDates",
        ("FIX.5.0", "40375", "12"): "12 .. NonDeliverableFXFixingDates",
        ("FIX.5.0", "40375", "13"): "13 .. SettlPeriodNotnl",
        ("FIX.5.0", "40375", "14"): "14 .. SettlPeriodPx",
        ("FIX.5.0", "40375", "15"): "15 .. CalcPeriod",
        ("FIX.5.0", "40375", "16"): "16 .. DividendAccrualRateMultiplier",
        ("FIX.5.0", "40375", "17"): "17 .. DividendAccrualRateSpread",
        ("FIX.5.0", "40375", "18"): "18 .. DividendAccrualCapRate",
        ("FIX.5.0", "40375", "19"): "19 .. DividendAccrualFloorRate",
        ("FIX.5.0", "40375", "20"): "20 .. CompoundingRateMultiplier",
        ("FIX.5.0", "40375", "21"): "21 .. CompoundingRateSpread",
        ("FIX.5.0", "40375", "22"): "22 .. CompoundingCapRate",
        ("FIX.5.0", "40375", "23"): "23 .. CompoundingFloorRate",
        ("FIX.5.0", "40376", "0"): "0 .. Initial",
        ("FIX.5.0", "40376", "1"): "1 .. Final",
        ("FIX.5.0", "40376", "2"): "2 .. CompoundingInitial",
        ("FIX.5.0", "40376", "3"): "3 .. CompoundingFinal",
        ("FIX.5.0", "40379", "1"): "1 .. Buy",
        ("FIX.5.0", "40379", "2"): "2 .. Sell",
        ("FIX.5.0", "40380", "1"): "1 .. Buy",
        ("FIX.5.0", "40380", "2"): "2 .. Sell",
        ("FIX.5.0", "40386", "0"): "0 .. Short",
        ("FIX.5.0", "40386", "1"): "1 .. Long",
        ("FIX.5.0", "40387", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40387", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40391", "D"): "D .. Day",
        ("FIX.5.0", "40391", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40391", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40391", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40391", "H"): "H .. Hour",
        ("FIX.5.0", "40391", "Min"): "Min .. Minute",
        ("FIX.5.0", "40391", "S"): "S .. Second",
        ("FIX.5.0", "40391", "T"): "T .. Term",
        ("FIX.5.0", "40395", "0"): "0 .. Initial",
        ("FIX.5.0", "40395", "1"): "1 .. Previous",
        ("FIX.5.0", "40399", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40399", "1"): "1 .. None",
        ("FIX.5.0", "40399", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40399", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40399", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40399", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40399", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40399", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40402", "D"): "D .. Day",
        ("FIX.5.0", "40402", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40402", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40402", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40403", "0"): "0 .. Business",
        ("FIX.5.0", "40403", "1"): "1 .. Calendar",
        ("FIX.5.0", "40403", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40403", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40403", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40403", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40408", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40408", "1"): "1 .. None",
        ("FIX.5.0", "40408", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40408", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40408", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40408", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40408", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40408", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40411", "D"): "D .. Day",
        ("FIX.5.0", "40411", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40411", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40411", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40412", "0"): "0 .. Business",
        ("FIX.5.0", "40412", "1"): "1 .. Calendar",
        ("FIX.5.0", "40412", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40412", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40412", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40412", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40415", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40415", "1"): "1 .. Reuters",
        ("FIX.5.0", "40415", "2"): "2 .. Telerate",
        ("FIX.5.0", "40415", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "40415", "99"): "99 .. Other",
        ("FIX.5.0", "40416", "0"): "0 .. Primary",
        ("FIX.5.0", "40416", "1"): "1 .. Secondary",
        ("FIX.5.0", "40419", "0"): "0 .. Initial",
        ("FIX.5.0", "40419", "1"): "1 .. Final",
        ("FIX.5.0", "40419", "2"): "2 .. CompoundingInitial",
        ("FIX.5.0", "40419", "3"): "3 .. CompoundingFinal",
        ("FIX.5.0", "40420", "0"): "0 .. Short",
        ("FIX.5.0", "40420", "1"): "1 .. Long",
        ("FIX.5.0", "40425", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40425", "1"): "1 .. Reuters",
        ("FIX.5.0", "40425", "2"): "2 .. Telerate",
        ("FIX.5.0", "40425", "99"): "99 .. Other",
        ("FIX.5.0", "40427", "D"): "D .. Day",
        ("FIX.5.0", "40427", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40427", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40427", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40430", "0"): "0 .. Short",
        ("FIX.5.0", "40430", "1"): "1 .. Long",
        ("FIX.5.0", "40431", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40431", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40433", "1"): "1 .. Buyer",
        ("FIX.5.0", "40433", "2"): "2 .. Seller",
        ("FIX.5.0", "40434", "1"): "1 .. Buyer",
        ("FIX.5.0", "40434", "2"): "2 .. Seller",
        ("FIX.5.0", "40436", "1"): "1 .. Buyer",
        ("FIX.5.0", "40436", "2"): "2 .. Seller",
        ("FIX.5.0", "40437", "1"): "1 .. Buyer",
        ("FIX.5.0", "40437", "2"): "2 .. Seller",
        ("FIX.5.0", "40439", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40439", "1"): "1 .. Reuters",
        ("FIX.5.0", "40439", "2"): "2 .. Telerate",
        ("FIX.5.0", "40439", "99"): "99 .. Other",
        ("FIX.5.0", "40441", "D"): "D .. Day",
        ("FIX.5.0", "40441", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40441", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40441", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40444", "0"): "0 .. Short",
        ("FIX.5.0", "40444", "1"): "1 .. Long",
        ("FIX.5.0", "40445", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40445", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40449", "0"): "0 .. MandatoryEarlyTermination",
        ("FIX.5.0", "40449", "1"): "1 .. OptionalEarlyTermination",
        ("FIX.5.0", "40449", "2"): "2 .. Cancelable",
        ("FIX.5.0", "40449", "3"): "3 .. Extendable",
        ("FIX.5.0", "40449", "4"): "4 .. MutualEarlyTermination",
        ("FIX.5.0", "40451", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40451", "1"): "1 .. None",
        ("FIX.5.0", "40451", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40451", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40451", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40451", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40451", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40451", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40455", "D"): "D .. Day",
        ("FIX.5.0", "40455", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40455", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40455", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40456", "0"): "0 .. ExercisingParty",
        ("FIX.5.0", "40456", "1"): "1 .. NonExercisingParty",
        ("FIX.5.0", "40456", "2"): "2 .. MasterAgreeent",
        ("FIX.5.0", "40456", "3"): "3 .. Supplement",
        ("FIX.5.0", "40457", "1"): "1 .. Buy",
        ("FIX.5.0", "40457", "2"): "2 .. Sell",
        ("FIX.5.0", "40458", "1"): "1 .. Buy",
        ("FIX.5.0", "40458", "2"): "2 .. Sell",
        ("FIX.5.0", "40459", "0"): "0 .. European",
        ("FIX.5.0", "40459", "1"): "1 .. American",
        ("FIX.5.0", "40459", "2"): "2 .. Bermuda",
        ("FIX.5.0", "40459", "99"): "99 .. Other",
        ("FIX.5.0", "40466", "0"): "0 .. CashPrice",
        ("FIX.5.0", "40466", "1"): "1 .. CashPriceAlternate",
        ("FIX.5.0", "40466", "2"): "2 .. ParYieldCurveAdjusted",
        ("FIX.5.0", "40466", "3"): "3 .. ZeroCouponYieldCurveAdjusted",
        ("FIX.5.0", "40466", "4"): "4 .. ParYieldCurveUnadjusted",
        ("FIX.5.0", "40466", "5"): "5 .. CrossCurrency",
        ("FIX.5.0", "40466", "6"): "6 .. CollateralizedPrice",
        ("FIX.5.0", "40469", "0"): "0 .. Bid",
        ("FIX.5.0", "40469", "1"): "1 .. Mid",
        ("FIX.5.0", "40469", "2"): "2 .. Offer",
        ("FIX.5.0", "40469", "3"): "3 .. ExercisingPartyPays",
        ("FIX.5.0", "40470", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40470", "1"): "1 .. Reuters",
        ("FIX.5.0", "40470", "2"): "2 .. Telerate",
        ("FIX.5.0", "40470", "99"): "99 .. Other",
        ("FIX.5.0", "40475", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "40475", "1"): "1 .. Adjusted",
        ("FIX.5.0", "40476", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40476", "1"): "1 .. None",
        ("FIX.5.0", "40476", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40476", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40476", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40476", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40476", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40476", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40479", "D"): "D .. Day",
        ("FIX.5.0", "40479", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40479", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40479", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40481", "D"): "D .. Day",
        ("FIX.5.0", "40481", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40481", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40481", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40481", "H"): "H .. Hour",
        ("FIX.5.0", "40481", "Min"): "Min .. Minute",
        ("FIX.5.0", "40481", "S"): "S .. Second",
        ("FIX.5.0", "40481", "T"): "T .. Term",
        ("FIX.5.0", "40485", "D"): "D .. Day",
        ("FIX.5.0", "40485", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40485", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40485", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40486", "0"): "0 .. Business",
        ("FIX.5.0", "40486", "1"): "1 .. Calendar",
        ("FIX.5.0", "40486", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40486", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40486", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40486", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40497", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "40497", "1"): "1 .. Adjusted",
        ("FIX.5.0", "40499", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40499", "1"): "1 .. None",
        ("FIX.5.0", "40499", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40499", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40499", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40499", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40499", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40499", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40503", "D"): "D .. Day",
        ("FIX.5.0", "40503", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40503", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40503", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40504", "0"): "0 .. Business",
        ("FIX.5.0", "40504", "1"): "1 .. Calendar",
        ("FIX.5.0", "40504", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40504", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40504", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40504", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40509", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40509", "1"): "1 .. None",
        ("FIX.5.0", "40509", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40509", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40509", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40509", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40509", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40509", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40513", "D"): "D .. Day",
        ("FIX.5.0", "40513", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40513", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40513", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40514", "0"): "0 .. Business",
        ("FIX.5.0", "40514", "1"): "1 .. Calendar",
        ("FIX.5.0", "40514", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40514", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40514", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40514", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40516", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40516", "1"): "1 .. None",
        ("FIX.5.0", "40516", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40516", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40516", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40516", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40516", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40516", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40520", "D"): "D .. Day",
        ("FIX.5.0", "40520", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40520", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40520", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40521", "0"): "0 .. Business",
        ("FIX.5.0", "40521", "1"): "1 .. Calendar",
        ("FIX.5.0", "40521", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40521", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40521", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40521", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40526", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40526", "1"): "1 .. None",
        ("FIX.5.0", "40526", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40526", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40526", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40526", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40526", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40526", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40530", "D"): "D .. Day",
        ("FIX.5.0", "40530", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40530", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40530", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40531", "0"): "0 .. Business",
        ("FIX.5.0", "40531", "1"): "1 .. Calendar",
        ("FIX.5.0", "40531", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40531", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40531", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40531", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40535", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "40535", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "40535", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "40535", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "40535", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "40535", "J"): "J .. TaxID",
        ("FIX.5.0", "40535", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "40535", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "40535", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "40535", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "40535", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "40535", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "40535", "B"): "B .. BIC",
        ("FIX.5.0", "40535", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "40535", "D"): "D .. Proprietary",
        ("FIX.5.0", "40535", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "40535", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "40535", "G"): "G .. MIC",
        ("FIX.5.0", "40535", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "40535", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "40535", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "40535", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "40535", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "40535", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "40535", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "40535", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "40536", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "40536", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "40536", "3"): "3 .. ClientID",
        ("FIX.5.0", "40536", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "40536", "5"): "5 .. InvestorID",
        ("FIX.5.0", "40536", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "40536", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "40536", "8"): "8 .. Locate",
        ("FIX.5.0", "40536", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "40536", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "40536", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "40536", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "40536", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "40536", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "40536", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "40536", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "40536", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "40536", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "40536", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "40536", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "40536", "22"): "22 .. Exchange",
        ("FIX.5.0", "40536", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "40536", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "40536", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "40536", "27"): "27 .. Buyer",
        ("FIX.5.0", "40536", "28"): "28 .. Custodian",
        ("FIX.5.0", "40536", "29"): "29 .. Intermediary",
        ("FIX.5.0", "40536", "30"): "30 .. Agent",
        ("FIX.5.0", "40536", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "40536", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "40536", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "40536", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "40536", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "40536", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "40536", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "40536", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "40536", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "40536", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "40536", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "40536", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "40536", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "40536", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "40536", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "40536", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "40536", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "40536", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "40536", "49"): "49 .. AssetManager",
        ("FIX.5.0", "40536", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "40536", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "40536", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "40536", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "40536", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "40536", "55"): "55 .. SessionID",
        ("FIX.5.0", "40536", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "40536", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "40536", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "40536", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "40536", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "40536", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "40536", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "40536", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "40536", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "40536", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "40536", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "40536", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "40536", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "40536", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "40536", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "40536", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "40536", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "40536", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "40536", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "40536", "75"): "75 .. LocationID",
        ("FIX.5.0", "40536", "76"): "76 .. DeskID",
        ("FIX.5.0", "40536", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "40536", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "40536", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "40536", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "40536", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "40536", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "40536", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "40536", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "40536", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "40536", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "40536", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "40536", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "40536", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "40536", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "40536", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "40536", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "40536", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "40536", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "40536", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "40536", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "40536", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "40536", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "40536", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "40536", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "40536", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "40536", "102"): "102 .. DataRepository",
        ("FIX.5.0", "40536", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "40536", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "40536", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "40536", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "40536", "107"): "107 .. Correspondent",
        ("FIX.5.0", "40536", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "40536", "110"): "110 .. Borrower",
        ("FIX.5.0", "40536", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "40536", "112"): "112 .. Guarantor",
        ("FIX.5.0", "40536", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "40536", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "40536", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "40536", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "40536", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "40536", "118"): "118 .. Operator",
        ("FIX.5.0", "40536", "119"): "119 .. CSD",
        ("FIX.5.0", "40536", "120"): "120 .. ICSD",
        ("FIX.5.0", "40536", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "40536", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "40536", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "40539", "1"): "1 .. Firm",
        ("FIX.5.0", "40539", "2"): "2 .. Person",
        ("FIX.5.0", "40539", "3"): "3 .. System",
        ("FIX.5.0", "40539", "4"): "4 .. Application",
        ("FIX.5.0", "40539", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "40539", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "40539", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "40539", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "40539", "9"): "9 .. ContactName",
        ("FIX.5.0", "40539", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "40539", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "40539", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "40539", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "40539", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "40539", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "40539", "16"): "16 .. BIC",
        ("FIX.5.0", "40539", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "40539", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "40539", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "40539", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "40539", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "40539", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "40539", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "40539", "24"): "24 .. Department",
        ("FIX.5.0", "40539", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "40539", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "40539", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "40539", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "40539", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "40539", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "40539", "31"): "31 .. Location",
        ("FIX.5.0", "40539", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "40539", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "40539", "34"): "34 .. AddressCity",
        ("FIX.5.0", "40539", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "40539", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "40539", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "40539", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "40539", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "40539", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "40539", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "40539", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "40539", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "40539", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "40539", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "40539", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "40539", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "40539", "48"): "48 .. USPerson",
        ("FIX.5.0", "40539", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "40539", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "40539", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "40539", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "40539", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "40539", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "40539", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "40539", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "40539", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "40539", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "40539", "59"): "59 .. FCMCode",
        ("FIX.5.0", "40539", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "40539", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "40539", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "40539", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "40539", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "40539", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "40539", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "40539", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "40539", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "40539", "69"): "69 .. EndUser",
        ("FIX.5.0", "40539", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "40539", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "40539", "72"): "72 .. Domicile",
        ("FIX.5.0", "40539", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "40539", "74"): "74 .. Payer",
        ("FIX.5.0", "40539", "75"): "75 .. Receiver",
        ("FIX.5.0", "40539", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "40539", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "40539", "78"): "78 .. FirstName",
        ("FIX.5.0", "40539", "79"): "79 .. Surname",
        ("FIX.5.0", "40539", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "40539", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "40539", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "40539", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "40539", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "40541", "0"): "0 .. PaymentCashSettlement",
        ("FIX.5.0", "40541", "1"): "1 .. PhysicalDelivery",
        ("FIX.5.0", "40543", "1"): "1 .. Buy",
        ("FIX.5.0", "40543", "2"): "2 .. Sell",
        ("FIX.5.0", "40544", "1"): "1 .. Buy",
        ("FIX.5.0", "40544", "2"): "2 .. Sell",
        ("FIX.5.0", "40549", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40549", "1"): "1 .. None",
        ("FIX.5.0", "40549", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40549", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40549", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40549", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40549", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40549", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40553", "D"): "D .. Day",
        ("FIX.5.0", "40553", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40553", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40553", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40554", "0"): "0 .. Business",
        ("FIX.5.0", "40554", "1"): "1 .. Calendar",
        ("FIX.5.0", "40554", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40554", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40554", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40554", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40556", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40556", "1"): "1 .. None",
        ("FIX.5.0", "40556", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40556", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40556", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40556", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40556", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40556", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40559", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40559", "1"): "1 .. None",
        ("FIX.5.0", "40559", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40559", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40559", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40559", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40559", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40559", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40566", "D"): "D .. Day",
        ("FIX.5.0", "40566", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40566", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40566", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40566", "H"): "H .. Hour",
        ("FIX.5.0", "40566", "Min"): "Min .. Minute",
        ("FIX.5.0", "40566", "S"): "S .. Second",
        ("FIX.5.0", "40566", "T"): "T .. Term",
        ("FIX.5.0", "40567", "1"): "1 .. FirstDay",
        ("FIX.5.0", "40567", "2"): "2 .. SecondDay",
        ("FIX.5.0", "40567", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "40567", "4"): "4 .. FourthDay",
        ("FIX.5.0", "40567", "5"): "5 .. FifthDay",
        ("FIX.5.0", "40567", "6"): "6 .. SixthDay",
        ("FIX.5.0", "40567", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "40567", "8"): "8 .. EighthDay",
        ("FIX.5.0", "40567", "9"): "9 .. NinthDay",
        ("FIX.5.0", "40567", "10"): "10 .. TenthDay",
        ("FIX.5.0", "40567", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "40567", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "40567", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "40567", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "40567", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "40567", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "40567", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "40567", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "40567", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "40567", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "40567", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "40567", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "40567", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "40567", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "40567", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "40567", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "40567", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "40567", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "40567", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "40567", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "40567", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "40567", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "40567", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "40567", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "40567", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "40567", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "40567", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "40567", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "40567", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "40567", "MON"): "MON .. MON",
        ("FIX.5.0", "40567", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "40567", "WED"): "WED .. WED",
        ("FIX.5.0", "40567", "THU"): "THU .. THU",
        ("FIX.5.0", "40567", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "40567", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "40567", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "40568", "0"): "0 .. Periodic",
        ("FIX.5.0", "40568", "1"): "1 .. Initial",
        ("FIX.5.0", "40568", "2"): "2 .. Single",
        ("FIX.5.0", "40568", "3"): "3 .. Dividend",
        ("FIX.5.0", "40568", "4"): "4 .. Interest",
        ("FIX.5.0", "40568", "5"): "5 .. DividendReturn",
        ("FIX.5.0", "40568", "6"): "6 .. PriceReturn",
        ("FIX.5.0", "40568", "7"): "7 .. TotalReturn",
        ("FIX.5.0", "40568", "8"): "8 .. Variance",
        ("FIX.5.0", "40568", "9"): "9 .. Correlation",
        ("FIX.5.0", "40572", "0"): "0 .. OneOne",
        ("FIX.5.0", "40572", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "40572", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "40572", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "40572", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "40572", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "40572", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "40572", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "40572", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "40572", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "40572", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "40572", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "40572", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "40572", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "40572", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "40572", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "40572", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "40572", "17"): "17 .. Act364",
        ("FIX.5.0", "40574", "0"): "0 .. Standard",
        ("FIX.5.0", "40574", "1"): "1 .. FRA",
        ("FIX.5.0", "40576", "0"): "0 .. OneOne",
        ("FIX.5.0", "40576", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "40576", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "40576", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "40576", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "40576", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "40576", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "40576", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "40576", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "40576", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "40576", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "40576", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "40576", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "40576", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "40576", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "40576", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "40576", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "40576", "17"): "17 .. Act364",
        ("FIX.5.0", "40577", "0"): "0 .. None",
        ("FIX.5.0", "40577", "1"): "1 .. Flat",
        ("FIX.5.0", "40577", "2"): "2 .. Straight",
        ("FIX.5.0", "40577", "3"): "3 .. SpreadExclusive",
        ("FIX.5.0", "40581", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40581", "1"): "1 .. None",
        ("FIX.5.0", "40581", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40581", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40581", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40581", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40581", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40581", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40584", "D"): "D .. Day",
        ("FIX.5.0", "40584", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40584", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40584", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40584", "T"): "T .. Term",
        ("FIX.5.0", "40585", "1"): "1 .. FirstDay",
        ("FIX.5.0", "40585", "2"): "2 .. SecondDay",
        ("FIX.5.0", "40585", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "40585", "4"): "4 .. FourthDay",
        ("FIX.5.0", "40585", "5"): "5 .. FifthDay",
        ("FIX.5.0", "40585", "6"): "6 .. SixthDay",
        ("FIX.5.0", "40585", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "40585", "8"): "8 .. EighthDay",
        ("FIX.5.0", "40585", "9"): "9 .. NinthDay",
        ("FIX.5.0", "40585", "10"): "10 .. TenthDay",
        ("FIX.5.0", "40585", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "40585", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "40585", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "40585", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "40585", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "40585", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "40585", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "40585", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "40585", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "40585", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "40585", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "40585", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "40585", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "40585", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "40585", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "40585", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "40585", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "40585", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "40585", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "40585", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "40585", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "40585", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "40585", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "40585", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "40585", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "40585", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "40585", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "40585", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "40585", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "40585", "MON"): "MON .. MON",
        ("FIX.5.0", "40585", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "40585", "WED"): "WED .. WED",
        ("FIX.5.0", "40585", "THU"): "THU .. THU",
        ("FIX.5.0", "40585", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "40585", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "40585", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "40590", "D"): "D .. Day",
        ("FIX.5.0", "40590", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40590", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40590", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40591", "0"): "0 .. Business",
        ("FIX.5.0", "40591", "1"): "1 .. Calendar",
        ("FIX.5.0", "40591", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40591", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40591", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40591", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40593", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40593", "1"): "1 .. None",
        ("FIX.5.0", "40593", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40593", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40593", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40593", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40593", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40593", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40596", "D"): "D .. Day",
        ("FIX.5.0", "40596", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40596", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40596", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40596", "H"): "H .. Hour",
        ("FIX.5.0", "40596", "Min"): "Min .. Minute",
        ("FIX.5.0", "40596", "S"): "S .. Second",
        ("FIX.5.0", "40596", "T"): "T .. Term",
        ("FIX.5.0", "40597", "MON"): "MON .. Monday",
        ("FIX.5.0", "40597", "TUE"): "TUE .. Tuesday",
        ("FIX.5.0", "40597", "WED"): "WED .. Wednesday",
        ("FIX.5.0", "40597", "THU"): "THU .. Thursday",
        ("FIX.5.0", "40597", "FRI"): "FRI .. Friday",
        ("FIX.5.0", "40597", "SAT"): "SAT .. Saturday",
        ("FIX.5.0", "40597", "SUN"): "SUN .. Sunday",
        ("FIX.5.0", "40599", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40599", "1"): "1 .. None",
        ("FIX.5.0", "40599", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40599", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40599", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40599", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40599", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40599", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40602", "D"): "D .. Day",
        ("FIX.5.0", "40602", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40602", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40602", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40603", "0"): "0 .. Business",
        ("FIX.5.0", "40603", "1"): "1 .. Calendar",
        ("FIX.5.0", "40603", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40603", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40603", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40603", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40606", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40606", "1"): "1 .. None",
        ("FIX.5.0", "40606", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40606", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40606", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40606", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40606", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40606", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40609", "D"): "D .. Day",
        ("FIX.5.0", "40609", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40609", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40609", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40610", "0"): "0 .. Business",
        ("FIX.5.0", "40610", "1"): "1 .. Calendar",
        ("FIX.5.0", "40610", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40610", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40610", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40610", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40613", "D"): "D .. Day",
        ("FIX.5.0", "40613", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40613", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40613", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40614", "0"): "0 .. Business",
        ("FIX.5.0", "40614", "1"): "1 .. Calendar",
        ("FIX.5.0", "40614", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40614", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40614", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40614", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40621", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40621", "1"): "1 .. Reuters",
        ("FIX.5.0", "40621", "2"): "2 .. Telerate",
        ("FIX.5.0", "40621", "99"): "99 .. Other",
        ("FIX.5.0", "40622", "D"): "D .. Day",
        ("FIX.5.0", "40622", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40622", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40622", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40626", "0"): "0 .. Short",
        ("FIX.5.0", "40626", "1"): "1 .. Long",
        ("FIX.5.0", "40627", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40627", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40629", "1"): "1 .. Buyer",
        ("FIX.5.0", "40629", "2"): "2 .. Seller",
        ("FIX.5.0", "40630", "1"): "1 .. Buyer",
        ("FIX.5.0", "40630", "2"): "2 .. Seller",
        ("FIX.5.0", "40632", "1"): "1 .. Buyer",
        ("FIX.5.0", "40632", "2"): "2 .. Seller",
        ("FIX.5.0", "40633", "1"): "1 .. Buyer",
        ("FIX.5.0", "40633", "2"): "2 .. Seller",
        ("FIX.5.0", "40635", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "40635", "1"): "1 .. RoundDown",
        ("FIX.5.0", "40635", "2"): "2 .. RoundUp",
        ("FIX.5.0", "40637", "0"): "0 .. Unweighted",
        ("FIX.5.0", "40637", "1"): "1 .. Weighted",
        ("FIX.5.0", "40638", "0"): "0 .. ZeroInterestRateMethod",
        ("FIX.5.0", "40638", "1"): "1 .. NegativeInterestRateMethod",
        ("FIX.5.0", "40640", "D"): "D .. Day",
        ("FIX.5.0", "40640", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40640", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40640", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40641", "0"): "0 .. Business",
        ("FIX.5.0", "40641", "1"): "1 .. Calendar",
        ("FIX.5.0", "40641", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40641", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40641", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40641", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40642", "0"): "0 .. None",
        ("FIX.5.0", "40642", "1"): "1 .. LinearZeroYield",
        ("FIX.5.0", "40643", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40643", "1"): "1 .. Reuters",
        ("FIX.5.0", "40643", "2"): "2 .. Telerate",
        ("FIX.5.0", "40643", "99"): "99 .. Other",
        ("FIX.5.0", "40647", "0"): "0 .. None",
        ("FIX.5.0", "40647", "1"): "1 .. ISDA",
        ("FIX.5.0", "40647", "2"): "2 .. AFMA",
        ("FIX.5.0", "40649", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40649", "1"): "1 .. None",
        ("FIX.5.0", "40649", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40649", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40649", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40649", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40649", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40649", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40653", "D"): "D .. Day",
        ("FIX.5.0", "40653", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40653", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40653", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40654", "0"): "0 .. Business",
        ("FIX.5.0", "40654", "1"): "1 .. Calendar",
        ("FIX.5.0", "40654", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40654", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40654", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40654", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40658", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "40658", "1"): "1 .. Adjusted",
        ("FIX.5.0", "40661", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40661", "1"): "1 .. Reuters",
        ("FIX.5.0", "40661", "2"): "2 .. Telerate",
        ("FIX.5.0", "40661", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "40661", "99"): "99 .. Other",
        ("FIX.5.0", "40663", "0"): "0 .. ExercisingParty",
        ("FIX.5.0", "40663", "1"): "1 .. NonExercisingParty",
        ("FIX.5.0", "40663", "2"): "2 .. MasterAgreeent",
        ("FIX.5.0", "40663", "3"): "3 .. Supplement",
        ("FIX.5.0", "40665", "0"): "0 .. Notional",
        ("FIX.5.0", "40665", "1"): "1 .. CashFlow",
        ("FIX.5.0", "40665", "2"): "2 .. FXLinkedNotional",
        ("FIX.5.0", "40665", "3"): "3 .. FixedRate",
        ("FIX.5.0", "40665", "4"): "4 .. FutureValueNotional",
        ("FIX.5.0", "40665", "5"): "5 .. KnownAmount",
        ("FIX.5.0", "40665", "6"): "6 .. FloatingRateMultiplier",
        ("FIX.5.0", "40665", "7"): "7 .. Spread",
        ("FIX.5.0", "40665", "8"): "8 .. CapRate",
        ("FIX.5.0", "40665", "9"): "9 .. FloorRate",
        ("FIX.5.0", "40665", "10"): "10 .. NonDeliverableSettlPaymentDates",
        ("FIX.5.0", "40665", "11"): "11 .. NonDeliverableSettlCalculationDates",
        ("FIX.5.0", "40665", "12"): "12 .. NonDeliverableFXFixingDates",
        ("FIX.5.0", "40665", "13"): "13 .. SettlPeriodNotnl",
        ("FIX.5.0", "40665", "14"): "14 .. SettlPeriodPx",
        ("FIX.5.0", "40665", "15"): "15 .. CalcPeriod",
        ("FIX.5.0", "40665", "16"): "16 .. DividendAccrualRateMultiplier",
        ("FIX.5.0", "40665", "17"): "17 .. DividendAccrualRateSpread",
        ("FIX.5.0", "40665", "18"): "18 .. DividendAccrualCapRate",
        ("FIX.5.0", "40665", "19"): "19 .. DividendAccrualFloorRate",
        ("FIX.5.0", "40665", "20"): "20 .. CompoundingRateMultiplier",
        ("FIX.5.0", "40665", "21"): "21 .. CompoundingRateSpread",
        ("FIX.5.0", "40665", "22"): "22 .. CompoundingCapRate",
        ("FIX.5.0", "40665", "23"): "23 .. CompoundingFloorRate",
        ("FIX.5.0", "40666", "0"): "0 .. Initial",
        ("FIX.5.0", "40666", "1"): "1 .. Final",
        ("FIX.5.0", "40666", "2"): "2 .. CompoundingInitial",
        ("FIX.5.0", "40666", "3"): "3 .. CompoundingFinal",
        ("FIX.5.0", "40669", "1"): "1 .. Buy",
        ("FIX.5.0", "40669", "2"): "2 .. Sell",
        ("FIX.5.0", "40670", "1"): "1 .. Buy",
        ("FIX.5.0", "40670", "2"): "2 .. Sell",
        ("FIX.5.0", "40676", "0"): "0 .. Short",
        ("FIX.5.0", "40676", "1"): "1 .. Long",
        ("FIX.5.0", "40677", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40677", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40681", "D"): "D .. Day",
        ("FIX.5.0", "40681", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40681", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40681", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40681", "H"): "H .. Hour",
        ("FIX.5.0", "40681", "Min"): "Min .. Minute",
        ("FIX.5.0", "40681", "S"): "S .. Second",
        ("FIX.5.0", "40681", "T"): "T .. Term",
        ("FIX.5.0", "40685", "0"): "0 .. Initial",
        ("FIX.5.0", "40685", "1"): "1 .. Previous",
        ("FIX.5.0", "40689", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40689", "1"): "1 .. None",
        ("FIX.5.0", "40689", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40689", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40689", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40689", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40689", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40689", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40692", "D"): "D .. Day",
        ("FIX.5.0", "40692", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40692", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40692", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40693", "0"): "0 .. Business",
        ("FIX.5.0", "40693", "1"): "1 .. Calendar",
        ("FIX.5.0", "40693", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40693", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40693", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40693", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40698", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40698", "1"): "1 .. None",
        ("FIX.5.0", "40698", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40698", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40698", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40698", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40698", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40698", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40701", "D"): "D .. Day",
        ("FIX.5.0", "40701", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40701", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40701", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40702", "0"): "0 .. Business",
        ("FIX.5.0", "40702", "1"): "1 .. Calendar",
        ("FIX.5.0", "40702", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40702", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40702", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40702", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40705", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40705", "1"): "1 .. Reuters",
        ("FIX.5.0", "40705", "2"): "2 .. Telerate",
        ("FIX.5.0", "40705", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "40705", "99"): "99 .. Other",
        ("FIX.5.0", "40706", "0"): "0 .. Primary",
        ("FIX.5.0", "40706", "1"): "1 .. Secondary",
        ("FIX.5.0", "40709", "0"): "0 .. Initial",
        ("FIX.5.0", "40709", "1"): "1 .. Final",
        ("FIX.5.0", "40709", "2"): "2 .. CompoundingInitial",
        ("FIX.5.0", "40709", "3"): "3 .. CompoundingFinal",
        ("FIX.5.0", "40710", "0"): "0 .. Short",
        ("FIX.5.0", "40710", "1"): "1 .. Long",
        ("FIX.5.0", "40715", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40715", "1"): "1 .. Reuters",
        ("FIX.5.0", "40715", "2"): "2 .. Telerate",
        ("FIX.5.0", "40715", "99"): "99 .. Other",
        ("FIX.5.0", "40717", "D"): "D .. Day",
        ("FIX.5.0", "40717", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40717", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40717", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40720", "0"): "0 .. Short",
        ("FIX.5.0", "40720", "1"): "1 .. Long",
        ("FIX.5.0", "40721", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40721", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40723", "1"): "1 .. Buyer",
        ("FIX.5.0", "40723", "2"): "2 .. Seller",
        ("FIX.5.0", "40724", "1"): "1 .. Buyer",
        ("FIX.5.0", "40724", "2"): "2 .. Seller",
        ("FIX.5.0", "40726", "1"): "1 .. Buyer",
        ("FIX.5.0", "40726", "2"): "2 .. Seller",
        ("FIX.5.0", "40727", "1"): "1 .. Buyer",
        ("FIX.5.0", "40727", "2"): "2 .. Seller",
        ("FIX.5.0", "40729", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40729", "1"): "1 .. Reuters",
        ("FIX.5.0", "40729", "2"): "2 .. Telerate",
        ("FIX.5.0", "40729", "99"): "99 .. Other",
        ("FIX.5.0", "40731", "D"): "D .. Day",
        ("FIX.5.0", "40731", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40731", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40731", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40734", "0"): "0 .. Short",
        ("FIX.5.0", "40734", "1"): "1 .. Long",
        ("FIX.5.0", "40735", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40735", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40738", "0"): "0 .. Periodic",
        ("FIX.5.0", "40738", "1"): "1 .. Initial",
        ("FIX.5.0", "40738", "2"): "2 .. Single",
        ("FIX.5.0", "40738", "3"): "3 .. Dividend",
        ("FIX.5.0", "40738", "4"): "4 .. Interest",
        ("FIX.5.0", "40738", "5"): "5 .. DividendReturn",
        ("FIX.5.0", "40738", "6"): "6 .. PriceReturn",
        ("FIX.5.0", "40738", "7"): "7 .. TotalReturn",
        ("FIX.5.0", "40738", "8"): "8 .. Variance",
        ("FIX.5.0", "40738", "9"): "9 .. Correlation",
        ("FIX.5.0", "40742", "0"): "0 .. OneOne",
        ("FIX.5.0", "40742", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "40742", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "40742", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "40742", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "40742", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "40742", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "40742", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "40742", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "40742", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "40742", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "40742", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "40742", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "40742", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "40742", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "40742", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "40742", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "40742", "17"): "17 .. Act364",
        ("FIX.5.0", "40744", "0"): "0 .. Standard",
        ("FIX.5.0", "40744", "1"): "1 .. FRA",
        ("FIX.5.0", "40746", "0"): "0 .. OneOne",
        ("FIX.5.0", "40746", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "40746", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "40746", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "40746", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "40746", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "40746", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "40746", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "40746", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "40746", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "40746", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "40746", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "40746", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "40746", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "40746", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "40746", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "40746", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "40746", "17"): "17 .. Act364",
        ("FIX.5.0", "40747", "0"): "0 .. None",
        ("FIX.5.0", "40747", "1"): "1 .. Flat",
        ("FIX.5.0", "40747", "2"): "2 .. Straight",
        ("FIX.5.0", "40747", "3"): "3 .. SpreadExclusive",
        ("FIX.5.0", "40751", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40751", "1"): "1 .. None",
        ("FIX.5.0", "40751", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40751", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40751", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40751", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40751", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40751", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40754", "D"): "D .. Day",
        ("FIX.5.0", "40754", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40754", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40754", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40754", "T"): "T .. Term",
        ("FIX.5.0", "40755", "1"): "1 .. FirstDay",
        ("FIX.5.0", "40755", "2"): "2 .. SecondDay",
        ("FIX.5.0", "40755", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "40755", "4"): "4 .. FourthDay",
        ("FIX.5.0", "40755", "5"): "5 .. FifthDay",
        ("FIX.5.0", "40755", "6"): "6 .. SixthDay",
        ("FIX.5.0", "40755", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "40755", "8"): "8 .. EighthDay",
        ("FIX.5.0", "40755", "9"): "9 .. NinthDay",
        ("FIX.5.0", "40755", "10"): "10 .. TenthDay",
        ("FIX.5.0", "40755", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "40755", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "40755", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "40755", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "40755", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "40755", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "40755", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "40755", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "40755", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "40755", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "40755", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "40755", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "40755", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "40755", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "40755", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "40755", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "40755", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "40755", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "40755", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "40755", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "40755", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "40755", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "40755", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "40755", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "40755", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "40755", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "40755", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "40755", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "40755", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "40755", "MON"): "MON .. MON",
        ("FIX.5.0", "40755", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "40755", "WED"): "WED .. WED",
        ("FIX.5.0", "40755", "THU"): "THU .. THU",
        ("FIX.5.0", "40755", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "40755", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "40755", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "40760", "D"): "D .. Day",
        ("FIX.5.0", "40760", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40760", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40760", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40762", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40762", "1"): "1 .. None",
        ("FIX.5.0", "40762", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40762", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40762", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40762", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40762", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40762", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40765", "D"): "D .. Day",
        ("FIX.5.0", "40765", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40765", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40765", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40765", "H"): "H .. Hour",
        ("FIX.5.0", "40765", "Min"): "Min .. Minute",
        ("FIX.5.0", "40765", "S"): "S .. Second",
        ("FIX.5.0", "40765", "T"): "T .. Term",
        ("FIX.5.0", "40766", "MON"): "MON .. Monday",
        ("FIX.5.0", "40766", "TUE"): "TUE .. Tuesday",
        ("FIX.5.0", "40766", "WED"): "WED .. Wednesday",
        ("FIX.5.0", "40766", "THU"): "THU .. Thursday",
        ("FIX.5.0", "40766", "FRI"): "FRI .. Friday",
        ("FIX.5.0", "40766", "SAT"): "SAT .. Saturday",
        ("FIX.5.0", "40766", "SUN"): "SUN .. Sunday",
        ("FIX.5.0", "40768", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40768", "1"): "1 .. None",
        ("FIX.5.0", "40768", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40768", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40768", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40768", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40768", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40768", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40771", "D"): "D .. Day",
        ("FIX.5.0", "40771", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40771", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40771", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40772", "0"): "0 .. Business",
        ("FIX.5.0", "40772", "1"): "1 .. Calendar",
        ("FIX.5.0", "40772", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40772", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40772", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40772", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40775", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40775", "1"): "1 .. None",
        ("FIX.5.0", "40775", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40775", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40775", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40775", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40775", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40775", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40778", "D"): "D .. Day",
        ("FIX.5.0", "40778", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40778", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40778", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40779", "0"): "0 .. Business",
        ("FIX.5.0", "40779", "1"): "1 .. Calendar",
        ("FIX.5.0", "40779", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40779", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40779", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40779", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40782", "D"): "D .. Day",
        ("FIX.5.0", "40782", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40782", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40782", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40783", "0"): "0 .. Business",
        ("FIX.5.0", "40783", "1"): "1 .. Calendar",
        ("FIX.5.0", "40783", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40783", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40783", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40783", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40790", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40790", "1"): "1 .. Reuters",
        ("FIX.5.0", "40790", "2"): "2 .. Telerate",
        ("FIX.5.0", "40790", "99"): "99 .. Other",
        ("FIX.5.0", "40791", "D"): "D .. Day",
        ("FIX.5.0", "40791", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40791", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40791", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40795", "0"): "0 .. Short",
        ("FIX.5.0", "40795", "1"): "1 .. Long",
        ("FIX.5.0", "40796", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40796", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40798", "1"): "1 .. Buyer",
        ("FIX.5.0", "40798", "2"): "2 .. Seller",
        ("FIX.5.0", "40799", "1"): "1 .. Buyer",
        ("FIX.5.0", "40799", "2"): "2 .. Seller",
        ("FIX.5.0", "40801", "1"): "1 .. Buyer",
        ("FIX.5.0", "40801", "2"): "2 .. Seller",
        ("FIX.5.0", "40802", "1"): "1 .. Buyer",
        ("FIX.5.0", "40802", "2"): "2 .. Seller",
        ("FIX.5.0", "40804", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "40804", "1"): "1 .. RoundDown",
        ("FIX.5.0", "40804", "2"): "2 .. RoundUp",
        ("FIX.5.0", "40806", "0"): "0 .. Unweighted",
        ("FIX.5.0", "40806", "1"): "1 .. Weighted",
        ("FIX.5.0", "40807", "0"): "0 .. ZeroInterestRateMethod",
        ("FIX.5.0", "40807", "1"): "1 .. NegativeInterestRateMethod",
        ("FIX.5.0", "40809", "D"): "D .. Day",
        ("FIX.5.0", "40809", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40809", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40809", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40810", "0"): "0 .. Business",
        ("FIX.5.0", "40810", "1"): "1 .. Calendar",
        ("FIX.5.0", "40810", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40810", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40810", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40810", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40811", "0"): "0 .. None",
        ("FIX.5.0", "40811", "1"): "1 .. LinearZeroYield",
        ("FIX.5.0", "40812", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40812", "1"): "1 .. Reuters",
        ("FIX.5.0", "40812", "2"): "2 .. Telerate",
        ("FIX.5.0", "40812", "99"): "99 .. Other",
        ("FIX.5.0", "40816", "0"): "0 .. None",
        ("FIX.5.0", "40816", "1"): "1 .. ISDA",
        ("FIX.5.0", "40816", "2"): "2 .. AFMA",
        ("FIX.5.0", "40818", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40818", "1"): "1 .. None",
        ("FIX.5.0", "40818", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40818", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40818", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40818", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40818", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40818", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40822", "D"): "D .. Day",
        ("FIX.5.0", "40822", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40822", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40822", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40823", "0"): "0 .. Business",
        ("FIX.5.0", "40823", "1"): "1 .. Calendar",
        ("FIX.5.0", "40823", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40823", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40823", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40823", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40827", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "40827", "1"): "1 .. Adjusted",
        ("FIX.5.0", "40829", "0"): "0 .. Notional",
        ("FIX.5.0", "40829", "1"): "1 .. CashFlow",
        ("FIX.5.0", "40829", "2"): "2 .. FXLinkedNotional",
        ("FIX.5.0", "40829", "3"): "3 .. FixedRate",
        ("FIX.5.0", "40829", "4"): "4 .. FutureValueNotional",
        ("FIX.5.0", "40829", "5"): "5 .. KnownAmount",
        ("FIX.5.0", "40829", "6"): "6 .. FloatingRateMultiplier",
        ("FIX.5.0", "40829", "7"): "7 .. Spread",
        ("FIX.5.0", "40829", "8"): "8 .. CapRate",
        ("FIX.5.0", "40829", "9"): "9 .. FloorRate",
        ("FIX.5.0", "40829", "10"): "10 .. NonDeliverableSettlPaymentDates",
        ("FIX.5.0", "40829", "11"): "11 .. NonDeliverableSettlCalculationDates",
        ("FIX.5.0", "40829", "12"): "12 .. NonDeliverableFXFixingDates",
        ("FIX.5.0", "40829", "13"): "13 .. SettlPeriodNotnl",
        ("FIX.5.0", "40829", "14"): "14 .. SettlPeriodPx",
        ("FIX.5.0", "40829", "15"): "15 .. CalcPeriod",
        ("FIX.5.0", "40829", "16"): "16 .. DividendAccrualRateMultiplier",
        ("FIX.5.0", "40829", "17"): "17 .. DividendAccrualRateSpread",
        ("FIX.5.0", "40829", "18"): "18 .. DividendAccrualCapRate",
        ("FIX.5.0", "40829", "19"): "19 .. DividendAccrualFloorRate",
        ("FIX.5.0", "40829", "20"): "20 .. CompoundingRateMultiplier",
        ("FIX.5.0", "40829", "21"): "21 .. CompoundingRateSpread",
        ("FIX.5.0", "40829", "22"): "22 .. CompoundingCapRate",
        ("FIX.5.0", "40829", "23"): "23 .. CompoundingFloorRate",
        ("FIX.5.0", "40830", "0"): "0 .. Initial",
        ("FIX.5.0", "40830", "1"): "1 .. Final",
        ("FIX.5.0", "40830", "2"): "2 .. CompoundingInitial",
        ("FIX.5.0", "40830", "3"): "3 .. CompoundingFinal",
        ("FIX.5.0", "40833", "1"): "1 .. Buy",
        ("FIX.5.0", "40833", "2"): "2 .. Sell",
        ("FIX.5.0", "40834", "1"): "1 .. Buy",
        ("FIX.5.0", "40834", "2"): "2 .. Sell",
        ("FIX.5.0", "40840", "0"): "0 .. Short",
        ("FIX.5.0", "40840", "1"): "1 .. Long",
        ("FIX.5.0", "40841", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40841", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40845", "D"): "D .. Day",
        ("FIX.5.0", "40845", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40845", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40845", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40845", "H"): "H .. Hour",
        ("FIX.5.0", "40845", "Min"): "Min .. Minute",
        ("FIX.5.0", "40845", "S"): "S .. Second",
        ("FIX.5.0", "40845", "T"): "T .. Term",
        ("FIX.5.0", "40849", "0"): "0 .. Initial",
        ("FIX.5.0", "40849", "1"): "1 .. Previous",
        ("FIX.5.0", "40853", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40853", "1"): "1 .. None",
        ("FIX.5.0", "40853", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40853", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40853", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40853", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40853", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40853", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40856", "D"): "D .. Day",
        ("FIX.5.0", "40856", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40856", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40856", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40857", "0"): "0 .. Business",
        ("FIX.5.0", "40857", "1"): "1 .. Calendar",
        ("FIX.5.0", "40857", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40857", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40857", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40857", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40862", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40862", "1"): "1 .. None",
        ("FIX.5.0", "40862", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40862", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40862", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40862", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40862", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40862", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40865", "D"): "D .. Day",
        ("FIX.5.0", "40865", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40865", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40865", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40866", "0"): "0 .. Business",
        ("FIX.5.0", "40866", "1"): "1 .. Calendar",
        ("FIX.5.0", "40866", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40866", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40866", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40866", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40869", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40869", "1"): "1 .. Reuters",
        ("FIX.5.0", "40869", "2"): "2 .. Telerate",
        ("FIX.5.0", "40869", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "40869", "99"): "99 .. Other",
        ("FIX.5.0", "40870", "0"): "0 .. Primary",
        ("FIX.5.0", "40870", "1"): "1 .. Secondary",
        ("FIX.5.0", "40873", "0"): "0 .. Initial",
        ("FIX.5.0", "40873", "1"): "1 .. Final",
        ("FIX.5.0", "40873", "2"): "2 .. CompoundingInitial",
        ("FIX.5.0", "40873", "3"): "3 .. CompoundingFinal",
        ("FIX.5.0", "40874", "0"): "0 .. Short",
        ("FIX.5.0", "40874", "1"): "1 .. Long",
        ("FIX.5.0", "40879", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40879", "1"): "1 .. Reuters",
        ("FIX.5.0", "40879", "2"): "2 .. Telerate",
        ("FIX.5.0", "40879", "99"): "99 .. Other",
        ("FIX.5.0", "40881", "D"): "D .. Day",
        ("FIX.5.0", "40881", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40881", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40881", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40884", "0"): "0 .. Short",
        ("FIX.5.0", "40884", "1"): "1 .. Long",
        ("FIX.5.0", "40885", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40885", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40887", "1"): "1 .. Buyer",
        ("FIX.5.0", "40887", "2"): "2 .. Seller",
        ("FIX.5.0", "40888", "1"): "1 .. Buyer",
        ("FIX.5.0", "40888", "2"): "2 .. Seller",
        ("FIX.5.0", "40890", "1"): "1 .. Buyer",
        ("FIX.5.0", "40890", "2"): "2 .. Seller",
        ("FIX.5.0", "40891", "1"): "1 .. Buyer",
        ("FIX.5.0", "40891", "2"): "2 .. Seller",
        ("FIX.5.0", "40893", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40893", "1"): "1 .. Reuters",
        ("FIX.5.0", "40893", "2"): "2 .. Telerate",
        ("FIX.5.0", "40893", "99"): "99 .. Other",
        ("FIX.5.0", "40895", "D"): "D .. Day",
        ("FIX.5.0", "40895", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40895", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40895", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40898", "0"): "0 .. Short",
        ("FIX.5.0", "40898", "1"): "1 .. Long",
        ("FIX.5.0", "40899", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "40899", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "40904", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "40904", "1"): "1 .. Reuters",
        ("FIX.5.0", "40904", "2"): "2 .. Telerate",
        ("FIX.5.0", "40904", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "40904", "99"): "99 .. Other",
        ("FIX.5.0", "40906", "0"): "0 .. ExercisingParty",
        ("FIX.5.0", "40906", "1"): "1 .. NonExercisingParty",
        ("FIX.5.0", "40906", "2"): "2 .. MasterAgreeent",
        ("FIX.5.0", "40906", "3"): "3 .. Supplement",
        ("FIX.5.0", "40908", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40908", "1"): "1 .. None",
        ("FIX.5.0", "40908", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40908", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40908", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40908", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40908", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40908", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40912", "D"): "D .. Day",
        ("FIX.5.0", "40912", "Wk"): "Wk .. Week",
        ("FIX.5.0", "40912", "Mo"): "Mo .. Month",
        ("FIX.5.0", "40912", "Yr"): "Yr .. Year",
        ("FIX.5.0", "40913", "0"): "0 .. Business",
        ("FIX.5.0", "40913", "1"): "1 .. Calendar",
        ("FIX.5.0", "40913", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40913", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40913", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40913", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40918", "23"): "23 .. FirmOrLegalEntity",
        ("FIX.5.0", "40918", "18"): "18 .. Current",
        ("FIX.5.0", "40918", "19"): "19 .. New",
        ("FIX.5.0", "40918", "24"): "24 .. NaturalPerson",
        ("FIX.5.0", "40918", "0"): "0 .. Agency",
        ("FIX.5.0", "40918", "1"): "1 .. Principal",
        ("FIX.5.0", "40918", "2"): "2 .. RisklessPrincipal",
        ("FIX.5.0", "40918", "9"): "9 .. PrimaryTrdRepository",
        ("FIX.5.0", "40918", "10"): "10 .. OrigTrdRepository",
        ("FIX.5.0", "40918", "11"): "11 .. AddtnlIntlTrdRepository",
        ("FIX.5.0", "40918", "12"): "12 .. AddtnlDomesticTrdRepository",
        ("FIX.5.0", "40918", "25"): "25 .. RegularTrader",
        ("FIX.5.0", "40918", "26"): "26 .. HeadTrader",
        ("FIX.5.0", "40918", "27"): "27 .. Supervisor",
        ("FIX.5.0", "40918", "22"): "22 .. Algorithm",
        ("FIX.5.0", "40918", "13"): "13 .. RelatedExchange",
        ("FIX.5.0", "40918", "14"): "14 .. OptionsExchange",
        ("FIX.5.0", "40918", "15"): "15 .. SpecifiedExchange",
        ("FIX.5.0", "40918", "16"): "16 .. ConstituentExchange",
        ("FIX.5.0", "40918", "7"): "7 .. Bank",
        ("FIX.5.0", "40918", "8"): "8 .. Hub",
        ("FIX.5.0", "40918", "3"): "3 .. GeneralClearingMember",
        ("FIX.5.0", "40918", "4"): "4 .. IndividualClearingMember",
        ("FIX.5.0", "40918", "5"): "5 .. PreferredMarketMaker",
        ("FIX.5.0", "40918", "6"): "6 .. DirectedMarketMaker",
        ("FIX.5.0", "40918", "20"): "20 .. DesignatedSponsor",
        ("FIX.5.0", "40918", "21"): "21 .. Specialist",
        ("FIX.5.0", "40918", "17"): "17 .. ExemptFromTradeReporting",
        ("FIX.5.0", "40919", "1"): "1 .. Percentage",
        ("FIX.5.0", "40919", "2"): "2 .. PerUnit",
        ("FIX.5.0", "40919", "3"): "3 .. FixedAmount",
        ("FIX.5.0", "40919", "4"): "4 .. Discount",
        ("FIX.5.0", "40919", "5"): "5 .. Premium",
        ("FIX.5.0", "40919", "6"): "6 .. Spread",
        ("FIX.5.0", "40919", "7"): "7 .. TEDPrice",
        ("FIX.5.0", "40919", "8"): "8 .. TEDYield",
        ("FIX.5.0", "40919", "9"): "9 .. Yield",
        ("FIX.5.0", "40919", "10"): "10 .. FixedCabinetTradePrice",
        ("FIX.5.0", "40919", "11"): "11 .. VariableCabinetTradePrice",
        ("FIX.5.0", "40919", "12"): "12 .. PriceSpread",
        ("FIX.5.0", "40919", "13"): "13 .. ProductTicksInHalves",
        ("FIX.5.0", "40919", "14"): "14 .. ProductTicksInFourths",
        ("FIX.5.0", "40919", "15"): "15 .. ProductTicksInEighths",
        ("FIX.5.0", "40919", "16"): "16 .. ProductTicksInSixteenths",
        ("FIX.5.0", "40919", "17"): "17 .. ProductTicksInThirtySeconds",
        ("FIX.5.0", "40919", "18"): "18 .. ProductTicksInSixtyFourths",
        ("FIX.5.0", "40919", "19"): "19 .. ProductTicksInOneTwentyEighths",
        ("FIX.5.0", "40919", "20"): "20 .. NormalRateRepresentation",
        ("FIX.5.0", "40919", "21"): "21 .. InverseRateRepresentation",
        ("FIX.5.0", "40919", "22"): "22 .. BasisPoints",
        ("FIX.5.0", "40919", "23"): "23 .. UpfrontPoints",
        ("FIX.5.0", "40919", "24"): "24 .. InterestRate",
        ("FIX.5.0", "40919", "25"): "25 .. PercentageNotional",
        ("FIX.5.0", "40920", "0"): "0 .. Business",
        ("FIX.5.0", "40920", "1"): "1 .. Calendar",
        ("FIX.5.0", "40920", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "40920", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "40920", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "40920", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "40921", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40921", "1"): "1 .. None",
        ("FIX.5.0", "40921", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40921", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40921", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40921", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40921", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40921", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40922", "1"): "1 .. FirstDay",
        ("FIX.5.0", "40922", "2"): "2 .. SecondDay",
        ("FIX.5.0", "40922", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "40922", "4"): "4 .. FourthDay",
        ("FIX.5.0", "40922", "5"): "5 .. FifthDay",
        ("FIX.5.0", "40922", "6"): "6 .. SixthDay",
        ("FIX.5.0", "40922", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "40922", "8"): "8 .. EighthDay",
        ("FIX.5.0", "40922", "9"): "9 .. NinthDay",
        ("FIX.5.0", "40922", "10"): "10 .. TenthDay",
        ("FIX.5.0", "40922", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "40922", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "40922", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "40922", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "40922", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "40922", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "40922", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "40922", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "40922", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "40922", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "40922", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "40922", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "40922", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "40922", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "40922", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "40922", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "40922", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "40922", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "40922", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "40922", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "40922", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "40922", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "40922", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "40922", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "40922", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "40922", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "40922", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "40922", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "40922", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "40922", "MON"): "MON .. MON",
        ("FIX.5.0", "40922", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "40922", "WED"): "WED .. WED",
        ("FIX.5.0", "40922", "THU"): "THU .. THU",
        ("FIX.5.0", "40922", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "40922", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "40922", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "40925", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40925", "1"): "1 .. None",
        ("FIX.5.0", "40925", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40925", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40925", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40925", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40925", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40925", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40926", "1"): "1 .. FirstDay",
        ("FIX.5.0", "40926", "2"): "2 .. SecondDay",
        ("FIX.5.0", "40926", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "40926", "4"): "4 .. FourthDay",
        ("FIX.5.0", "40926", "5"): "5 .. FifthDay",
        ("FIX.5.0", "40926", "6"): "6 .. SixthDay",
        ("FIX.5.0", "40926", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "40926", "8"): "8 .. EighthDay",
        ("FIX.5.0", "40926", "9"): "9 .. NinthDay",
        ("FIX.5.0", "40926", "10"): "10 .. TenthDay",
        ("FIX.5.0", "40926", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "40926", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "40926", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "40926", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "40926", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "40926", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "40926", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "40926", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "40926", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "40926", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "40926", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "40926", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "40926", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "40926", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "40926", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "40926", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "40926", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "40926", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "40926", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "40926", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "40926", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "40926", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "40926", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "40926", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "40926", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "40926", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "40926", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "40926", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "40926", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "40926", "MON"): "MON .. MON",
        ("FIX.5.0", "40926", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "40926", "WED"): "WED .. WED",
        ("FIX.5.0", "40926", "THU"): "THU .. THU",
        ("FIX.5.0", "40926", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "40926", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "40926", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "40964", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "40964", "1"): "1 .. None",
        ("FIX.5.0", "40964", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "40964", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "40964", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "40964", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "40964", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "40964", "7"): "7 .. NearestDay",
        ("FIX.5.0", "40965", "1"): "1 .. FirstDay",
        ("FIX.5.0", "40965", "2"): "2 .. SecondDay",
        ("FIX.5.0", "40965", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "40965", "4"): "4 .. FourthDay",
        ("FIX.5.0", "40965", "5"): "5 .. FifthDay",
        ("FIX.5.0", "40965", "6"): "6 .. SixthDay",
        ("FIX.5.0", "40965", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "40965", "8"): "8 .. EighthDay",
        ("FIX.5.0", "40965", "9"): "9 .. NinthDay",
        ("FIX.5.0", "40965", "10"): "10 .. TenthDay",
        ("FIX.5.0", "40965", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "40965", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "40965", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "40965", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "40965", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "40965", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "40965", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "40965", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "40965", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "40965", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "40965", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "40965", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "40965", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "40965", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "40965", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "40965", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "40965", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "40965", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "40965", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "40965", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "40965", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "40965", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "40965", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "40965", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "40965", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "40965", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "40965", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "40965", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "40965", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "40965", "MON"): "MON .. MON",
        ("FIX.5.0", "40965", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "40965", "WED"): "WED .. WED",
        ("FIX.5.0", "40965", "THU"): "THU .. THU",
        ("FIX.5.0", "40965", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "40965", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "40965", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "40993", "0"): "0 .. Initial",
        ("FIX.5.0", "40993", "1"): "1 .. Intermediate",
        ("FIX.5.0", "40993", "2"): "2 .. Final",
        ("FIX.5.0", "40993", "3"): "3 .. Prepaid",
        ("FIX.5.0", "40993", "4"): "4 .. Postpaid",
        ("FIX.5.0", "40993", "5"): "5 .. Variable",
        ("FIX.5.0", "40993", "6"): "6 .. Fixed",
        ("FIX.5.0", "40993", "7"): "7 .. Swap",
        ("FIX.5.0", "40993", "8"): "8 .. Conditional",
        ("FIX.5.0", "41002", "D"): "D .. Day",
        ("FIX.5.0", "41002", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41002", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41002", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41003", "0"): "0 .. Business",
        ("FIX.5.0", "41003", "1"): "1 .. Calendar",
        ("FIX.5.0", "41003", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41003", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41003", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41003", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41006", "H"): "H .. RestructuringMultipleHoldingObligations",
        ("FIX.5.0", "41006", "E"): "E .. RestructuringMultipleCreditEventNotices",
        ("FIX.5.0", "41006", "C"): "C .. FloatingRateInterestShortfall",
        ("FIX.5.0", "41011", "0"): "0 .. AsianOut",
        ("FIX.5.0", "41011", "1"): "1 .. AsianIn",
        ("FIX.5.0", "41011", "2"): "2 .. BarrierCap",
        ("FIX.5.0", "41011", "3"): "3 .. BarrierFloor",
        ("FIX.5.0", "41011", "4"): "4 .. KnockOut",
        ("FIX.5.0", "41011", "5"): "5 .. KnockIn",
        ("FIX.5.0", "41014", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "41014", "1"): "1 .. Reuters",
        ("FIX.5.0", "41014", "2"): "2 .. Telerate",
        ("FIX.5.0", "41014", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "41014", "99"): "99 .. Other",
        ("FIX.5.0", "41015", "0"): "0 .. Primary",
        ("FIX.5.0", "41015", "1"): "1 .. Secondary",
        ("FIX.5.0", "41023", "D"): "D .. Day",
        ("FIX.5.0", "41023", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41023", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41023", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41024", "0"): "0 .. Business",
        ("FIX.5.0", "41024", "1"): "1 .. Calendar",
        ("FIX.5.0", "41024", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41024", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41024", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41024", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41025", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41025", "1"): "1 .. None",
        ("FIX.5.0", "41025", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41025", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41025", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41025", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41025", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41025", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41035", "D"): "D .. Day",
        ("FIX.5.0", "41035", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41035", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41035", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41036", "1"): "1 .. FirstDay",
        ("FIX.5.0", "41036", "2"): "2 .. SecondDay",
        ("FIX.5.0", "41036", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "41036", "4"): "4 .. FourthDay",
        ("FIX.5.0", "41036", "5"): "5 .. FifthDay",
        ("FIX.5.0", "41036", "6"): "6 .. SixthDay",
        ("FIX.5.0", "41036", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "41036", "8"): "8 .. EighthDay",
        ("FIX.5.0", "41036", "9"): "9 .. NinthDay",
        ("FIX.5.0", "41036", "10"): "10 .. TenthDay",
        ("FIX.5.0", "41036", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "41036", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "41036", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "41036", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "41036", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "41036", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "41036", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "41036", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "41036", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "41036", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "41036", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "41036", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "41036", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "41036", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "41036", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "41036", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "41036", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "41036", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "41036", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "41036", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "41036", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "41036", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "41036", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "41036", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "41036", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "41036", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "41036", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "41036", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "41036", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "41036", "MON"): "MON .. MON",
        ("FIX.5.0", "41036", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "41036", "WED"): "WED .. WED",
        ("FIX.5.0", "41036", "THU"): "THU .. THU",
        ("FIX.5.0", "41036", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "41036", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "41036", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "41038", "0"): "0 .. Notional",
        ("FIX.5.0", "41038", "1"): "1 .. Delivery",
        ("FIX.5.0", "41038", "2"): "2 .. PhysicalSettlPeriods",
        ("FIX.5.0", "41041", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41041", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41041", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41041", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41041", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41041", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41041", "thm"): "thm .. therms",
        ("FIX.5.0", "41041", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41041", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41041", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41041", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41041", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41041", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41041", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41041", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41041", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41041", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41041", "day"): "day .. Day",
        ("FIX.5.0", "41041", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41041", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41041", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41041", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41041", "g"): "g .. Grams",
        ("FIX.5.0", "41041", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41041", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41041", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41041", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41041", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41041", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41041", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41041", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41041", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41041", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41041", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41041", "L"): "L .. liters",
        ("FIX.5.0", "41041", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41041", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41041", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41041", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41041", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41041", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41041", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41041", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41041", "t"): "t .. MetricTons",
        ("FIX.5.0", "41041", "tn"): "tn .. Tons",
        ("FIX.5.0", "41042", "0"): "0 .. Term",
        ("FIX.5.0", "41042", "1"): "1 .. PerBusinessDay",
        ("FIX.5.0", "41042", "2"): "2 .. PerCalculationPeriod",
        ("FIX.5.0", "41042", "3"): "3 .. PerSettlPeriod",
        ("FIX.5.0", "41042", "4"): "4 .. PerCalendarDay",
        ("FIX.5.0", "41042", "5"): "5 .. PerHour",
        ("FIX.5.0", "41042", "6"): "6 .. PerMonth",
        ("FIX.5.0", "41045", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41045", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41045", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41045", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41045", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41045", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41045", "thm"): "thm .. therms",
        ("FIX.5.0", "41045", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41045", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41045", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41045", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41045", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41045", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41045", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41045", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41045", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41045", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41045", "day"): "day .. Day",
        ("FIX.5.0", "41045", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41045", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41045", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41045", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41045", "g"): "g .. Grams",
        ("FIX.5.0", "41045", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41045", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41045", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41045", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41045", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41045", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41045", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41045", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41045", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41045", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41045", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41045", "L"): "L .. liters",
        ("FIX.5.0", "41045", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41045", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41045", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41045", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41045", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41045", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41045", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41045", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41045", "t"): "t .. MetricTons",
        ("FIX.5.0", "41045", "tn"): "tn .. Tons",
        ("FIX.5.0", "41046", "0"): "0 .. Absolute",
        ("FIX.5.0", "41046", "1"): "1 .. Percentage",
        ("FIX.5.0", "41049", "0"): "0 .. AllTimes",
        ("FIX.5.0", "41049", "1"): "1 .. OnPeak",
        ("FIX.5.0", "41049", "2"): "2 .. OffPeak",
        ("FIX.5.0", "41049", "3"): "3 .. Base",
        ("FIX.5.0", "41049", "4"): "4 .. BlockHours",
        ("FIX.5.0", "41049", "5"): "5 .. Other",
        ("FIX.5.0", "41050", "0"): "0 .. DoNotIncludeHolidays",
        ("FIX.5.0", "41050", "1"): "1 .. IncludeHolidays",
        ("FIX.5.0", "41052", "1"): "1 .. Monday",
        ("FIX.5.0", "41052", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41052", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41052", "4"): "4 .. Thursday",
        ("FIX.5.0", "41052", "5"): "5 .. Friday",
        ("FIX.5.0", "41052", "6"): "6 .. Saturday",
        ("FIX.5.0", "41052", "7"): "7 .. Sunday",
        ("FIX.5.0", "41052", "8"): "8 .. AllWeekdays",
        ("FIX.5.0", "41052", "9"): "9 .. AllDays",
        ("FIX.5.0", "41052", "10"): "10 .. AllWeekends",
        ("FIX.5.0", "41057", "0"): "0 .. Hour",
        ("FIX.5.0", "41057", "1"): "1 .. Timestamp",
        ("FIX.5.0", "41058", "0"): "0 .. Periodic",
        ("FIX.5.0", "41058", "1"): "1 .. Initial",
        ("FIX.5.0", "41058", "2"): "2 .. Single",
        ("FIX.5.0", "41063", "1"): "1 .. Firm",
        ("FIX.5.0", "41063", "2"): "2 .. NonFirm",
        ("FIX.5.0", "41063", "3"): "3 .. ForceMajeure",
        ("FIX.5.0", "41063", "4"): "4 .. SystemFirm",
        ("FIX.5.0", "41063", "5"): "5 .. UnitFirm",
        ("FIX.5.0", "41065", "0"): "0 .. Buyer",
        ("FIX.5.0", "41065", "1"): "1 .. Seller",
        ("FIX.5.0", "41069", "0"): "0 .. Transfers",
        ("FIX.5.0", "41069", "1"): "1 .. DoesNotTransfer",
        ("FIX.5.0", "41073", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41073", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41073", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41073", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41073", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41073", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41073", "thm"): "thm .. therms",
        ("FIX.5.0", "41073", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41073", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41073", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41073", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41073", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41073", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41073", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41073", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41073", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41073", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41073", "day"): "day .. Day",
        ("FIX.5.0", "41073", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41073", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41073", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41073", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41073", "g"): "g .. Grams",
        ("FIX.5.0", "41073", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41073", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41073", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41073", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41073", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41073", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41073", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41073", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41073", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41073", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41073", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41073", "L"): "L .. liters",
        ("FIX.5.0", "41073", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41073", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41073", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41073", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41073", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41073", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41073", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41073", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41073", "t"): "t .. MetricTons",
        ("FIX.5.0", "41073", "tn"): "tn .. Tons",
        ("FIX.5.0", "41074", "0"): "0 .. Absolute",
        ("FIX.5.0", "41074", "1"): "1 .. Percentage",
        ("FIX.5.0", "41075", "1"): "1 .. Buyer",
        ("FIX.5.0", "41075", "2"): "2 .. Seller",
        ("FIX.5.0", "41080", "0"): "0 .. Buyer",
        ("FIX.5.0", "41080", "1"): "1 .. Seller",
        ("FIX.5.0", "41087", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41087", "1"): "1 .. Applicable",
        ("FIX.5.0", "41087", "2"): "2 .. AsInMasterAgreement",
        ("FIX.5.0", "41087", "3"): "3 .. AsInConfirmation",
        ("FIX.5.0", "41088", "0"): "0 .. MasterAgreement",
        ("FIX.5.0", "41088", "1"): "1 .. Confirmation",
        ("FIX.5.0", "41097", "0"): "0 .. Basket",
        ("FIX.5.0", "41097", "1"): "1 .. Bond",
        ("FIX.5.0", "41097", "2"): "2 .. Cash",
        ("FIX.5.0", "41097", "3"): "3 .. Commodity",
        ("FIX.5.0", "41097", "4"): "4 .. ConvertibleBond",
        ("FIX.5.0", "41097", "5"): "5 .. Equity",
        ("FIX.5.0", "41097", "6"): "6 .. ExchangeTradedFund",
        ("FIX.5.0", "41097", "7"): "7 .. Future",
        ("FIX.5.0", "41097", "8"): "8 .. Index",
        ("FIX.5.0", "41097", "9"): "9 .. Loan",
        ("FIX.5.0", "41097", "10"): "10 .. Mortgage",
        ("FIX.5.0", "41097", "11"): "11 .. MutualFund",
        ("FIX.5.0", "41099", "1"): "1 .. CUSIP",
        ("FIX.5.0", "41099", "2"): "2 .. SEDOL",
        ("FIX.5.0", "41099", "3"): "3 .. QUIK",
        ("FIX.5.0", "41099", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "41099", "5"): "5 .. RICCode",
        ("FIX.5.0", "41099", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "41099", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "41099", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "41099", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "41099", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "41099", "B"): "B .. Wertpapier",
        ("FIX.5.0", "41099", "C"): "C .. Dutch",
        ("FIX.5.0", "41099", "D"): "D .. Valoren",
        ("FIX.5.0", "41099", "E"): "E .. Sicovam",
        ("FIX.5.0", "41099", "F"): "F .. Belgian",
        ("FIX.5.0", "41099", "G"): "G .. Common",
        ("FIX.5.0", "41099", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "41099", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "41099", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "41099", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "41099", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "41099", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "41099", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "41099", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "41099", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "41099", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "41099", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "41099", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "41099", "U"): "U .. Synthetic",
        ("FIX.5.0", "41099", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "41099", "W"): "W .. IndexName",
        ("FIX.5.0", "41111", "0"): "0 .. NotRequired",
        ("FIX.5.0", "41111", "1"): "1 .. NonElectronic",
        ("FIX.5.0", "41111", "2"): "2 .. Electronic",
        ("FIX.5.0", "41111", "3"): "3 .. Unknown",
        ("FIX.5.0", "41118", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41118", "1"): "1 .. None",
        ("FIX.5.0", "41118", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41118", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41118", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41118", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41118", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41118", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41119", "0"): "0 .. Business",
        ("FIX.5.0", "41119", "1"): "1 .. Calendar",
        ("FIX.5.0", "41119", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41119", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41119", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41119", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41121", "D"): "D .. Day",
        ("FIX.5.0", "41121", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41121", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41121", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41123", "D"): "D .. Day",
        ("FIX.5.0", "41123", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41123", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41123", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41127", "D"): "D .. Day",
        ("FIX.5.0", "41127", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41127", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41127", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41128", "0"): "0 .. Business",
        ("FIX.5.0", "41128", "1"): "1 .. Calendar",
        ("FIX.5.0", "41128", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41128", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41128", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41128", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41139", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41139", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41142", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41142", "1"): "1 .. None",
        ("FIX.5.0", "41142", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41142", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41142", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41142", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41142", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41142", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41145", "D"): "D .. Day",
        ("FIX.5.0", "41145", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41145", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41145", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41147", "D"): "D .. Day",
        ("FIX.5.0", "41147", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41147", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41147", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41148", "1"): "1 .. FirstDay",
        ("FIX.5.0", "41148", "2"): "2 .. SecondDay",
        ("FIX.5.0", "41148", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "41148", "4"): "4 .. FourthDay",
        ("FIX.5.0", "41148", "5"): "5 .. FifthDay",
        ("FIX.5.0", "41148", "6"): "6 .. SixthDay",
        ("FIX.5.0", "41148", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "41148", "8"): "8 .. EighthDay",
        ("FIX.5.0", "41148", "9"): "9 .. NinthDay",
        ("FIX.5.0", "41148", "10"): "10 .. TenthDay",
        ("FIX.5.0", "41148", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "41148", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "41148", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "41148", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "41148", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "41148", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "41148", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "41148", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "41148", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "41148", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "41148", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "41148", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "41148", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "41148", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "41148", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "41148", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "41148", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "41148", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "41148", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "41148", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "41148", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "41148", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "41148", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "41148", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "41148", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "41148", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "41148", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "41148", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "41148", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "41148", "MON"): "MON .. MON",
        ("FIX.5.0", "41148", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "41148", "WED"): "WED .. WED",
        ("FIX.5.0", "41148", "THU"): "THU .. THU",
        ("FIX.5.0", "41148", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "41148", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "41148", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "41149", "0"): "0 .. Business",
        ("FIX.5.0", "41149", "1"): "1 .. Calendar",
        ("FIX.5.0", "41149", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41149", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41149", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41149", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41154", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41154", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41155", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41155", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41155", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41155", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41155", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41155", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41155", "thm"): "thm .. therms",
        ("FIX.5.0", "41155", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41155", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41155", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41155", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41155", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41155", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41155", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41155", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41155", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41155", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41155", "day"): "day .. Day",
        ("FIX.5.0", "41155", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41155", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41155", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41155", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41155", "g"): "g .. Grams",
        ("FIX.5.0", "41155", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41155", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41155", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41155", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41155", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41155", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41155", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41155", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41155", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41155", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41155", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41155", "L"): "L .. liters",
        ("FIX.5.0", "41155", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41155", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41155", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41155", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41155", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41155", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41155", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41155", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41155", "t"): "t .. MetricTons",
        ("FIX.5.0", "41155", "tn"): "tn .. Tons",
        ("FIX.5.0", "41158", "D"): "D .. Day",
        ("FIX.5.0", "41158", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41158", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41158", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41159", "0"): "0 .. Business",
        ("FIX.5.0", "41159", "1"): "1 .. Calendar",
        ("FIX.5.0", "41159", "2"): "2 .. Commodity",
        ("FIX.5.0", "41159", "3"): "3 .. Currency",
        ("FIX.5.0", "41159", "4"): "4 .. Exchange",
        ("FIX.5.0", "41159", "5"): "5 .. Scheduled",
        ("FIX.5.0", "41160", "0"): "0 .. Prepaid",
        ("FIX.5.0", "41160", "1"): "1 .. Postpaid",
        ("FIX.5.0", "41160", "2"): "2 .. Variable",
        ("FIX.5.0", "41160", "3"): "3 .. Fixed",
        ("FIX.5.0", "41162", "0"): "0 .. EveryDay",
        ("FIX.5.0", "41162", "1"): "1 .. Monday",
        ("FIX.5.0", "41162", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41162", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41162", "4"): "4 .. Thursday",
        ("FIX.5.0", "41162", "5"): "5 .. Friday",
        ("FIX.5.0", "41162", "6"): "6 .. Saturday",
        ("FIX.5.0", "41162", "7"): "7 .. Sunday",
        ("FIX.5.0", "41167", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41167", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41167", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41167", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41167", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41167", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41167", "thm"): "thm .. therms",
        ("FIX.5.0", "41167", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41167", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41167", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41167", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41167", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41167", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41167", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41167", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41167", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41167", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41167", "day"): "day .. Day",
        ("FIX.5.0", "41167", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41167", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41167", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41167", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41167", "g"): "g .. Grams",
        ("FIX.5.0", "41167", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41167", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41167", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41167", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41167", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41167", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41167", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41167", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41167", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41167", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41167", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41167", "L"): "L .. liters",
        ("FIX.5.0", "41167", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41167", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41167", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41167", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41167", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41167", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41167", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41167", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41167", "t"): "t .. MetricTons",
        ("FIX.5.0", "41167", "tn"): "tn .. Tons",
        ("FIX.5.0", "41169", "0"): "0 .. Absolute",
        ("FIX.5.0", "41169", "1"): "1 .. Percentage",
        ("FIX.5.0", "41172", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41172", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41172", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41172", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41172", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41172", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41172", "thm"): "thm .. therms",
        ("FIX.5.0", "41172", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41172", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41172", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41172", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41172", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41172", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41172", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41172", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41172", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41172", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41172", "day"): "day .. Day",
        ("FIX.5.0", "41172", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41172", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41172", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41172", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41172", "g"): "g .. Grams",
        ("FIX.5.0", "41172", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41172", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41172", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41172", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41172", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41172", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41172", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41172", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41172", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41172", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41172", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41172", "L"): "L .. liters",
        ("FIX.5.0", "41172", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41172", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41172", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41172", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41172", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41172", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41172", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41172", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41172", "t"): "t .. MetricTons",
        ("FIX.5.0", "41172", "tn"): "tn .. Tons",
        ("FIX.5.0", "41173", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41173", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41173", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41173", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41173", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41173", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41173", "thm"): "thm .. therms",
        ("FIX.5.0", "41173", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41173", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41173", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41173", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41173", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41173", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41173", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41173", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41173", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41173", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41173", "day"): "day .. Day",
        ("FIX.5.0", "41173", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41173", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41173", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41173", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41173", "g"): "g .. Grams",
        ("FIX.5.0", "41173", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41173", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41173", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41173", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41173", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41173", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41173", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41173", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41173", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41173", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41173", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41173", "L"): "L .. liters",
        ("FIX.5.0", "41173", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41173", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41173", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41173", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41173", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41173", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41173", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41173", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41173", "t"): "t .. MetricTons",
        ("FIX.5.0", "41173", "tn"): "tn .. Tons",
        ("FIX.5.0", "41174", "0"): "0 .. All",
        ("FIX.5.0", "41174", "1"): "1 .. First",
        ("FIX.5.0", "41174", "2"): "2 .. Last",
        ("FIX.5.0", "41174", "3"): "3 .. Penultimate",
        ("FIX.5.0", "41177", "D"): "D .. Day",
        ("FIX.5.0", "41177", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41177", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41177", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41179", "D"): "D .. Day",
        ("FIX.5.0", "41179", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41179", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41179", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41187", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41187", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41187", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41187", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41187", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41187", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41187", "thm"): "thm .. therms",
        ("FIX.5.0", "41187", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41187", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41187", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41187", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41187", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41187", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41187", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41187", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41187", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41187", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41187", "day"): "day .. Day",
        ("FIX.5.0", "41187", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41187", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41187", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41187", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41187", "g"): "g .. Grams",
        ("FIX.5.0", "41187", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41187", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41187", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41187", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41187", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41187", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41187", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41187", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41187", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41187", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41187", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41187", "L"): "L .. liters",
        ("FIX.5.0", "41187", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41187", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41187", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41187", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41187", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41187", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41187", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41187", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41187", "t"): "t .. MetricTons",
        ("FIX.5.0", "41187", "tn"): "tn .. Tons",
        ("FIX.5.0", "41195", "D"): "D .. Day",
        ("FIX.5.0", "41195", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41195", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41195", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41198", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41198", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41198", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41198", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41198", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41198", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41198", "thm"): "thm .. therms",
        ("FIX.5.0", "41198", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41198", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41198", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41198", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41198", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41198", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41198", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41198", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41198", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41198", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41198", "day"): "day .. Day",
        ("FIX.5.0", "41198", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41198", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41198", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41198", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41198", "g"): "g .. Grams",
        ("FIX.5.0", "41198", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41198", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41198", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41198", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41198", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41198", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41198", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41198", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41198", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41198", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41198", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41198", "L"): "L .. liters",
        ("FIX.5.0", "41198", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41198", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41198", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41198", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41198", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41198", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41198", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41198", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41198", "t"): "t .. MetricTons",
        ("FIX.5.0", "41198", "tn"): "tn .. Tons",
        ("FIX.5.0", "41199", "0"): "0 .. Average",
        ("FIX.5.0", "41199", "1"): "1 .. Maximum",
        ("FIX.5.0", "41199", "2"): "2 .. Minimum",
        ("FIX.5.0", "41199", "3"): "3 .. Cumulative",
        ("FIX.5.0", "41201", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41201", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41201", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41201", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41201", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41201", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41201", "thm"): "thm .. therms",
        ("FIX.5.0", "41201", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41201", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41201", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41201", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41201", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41201", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41201", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41201", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41201", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41201", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41201", "day"): "day .. Day",
        ("FIX.5.0", "41201", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41201", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41201", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41201", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41201", "g"): "g .. Grams",
        ("FIX.5.0", "41201", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41201", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41201", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41201", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41201", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41201", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41201", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41201", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41201", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41201", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41201", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41201", "L"): "L .. liters",
        ("FIX.5.0", "41201", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41201", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41201", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41201", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41201", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41201", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41201", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41201", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41201", "t"): "t .. MetricTons",
        ("FIX.5.0", "41201", "tn"): "tn .. Tons",
        ("FIX.5.0", "41204", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41204", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41204", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41204", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41204", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41204", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41204", "thm"): "thm .. therms",
        ("FIX.5.0", "41204", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41204", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41204", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41204", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41204", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41204", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41204", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41204", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41204", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41204", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41204", "day"): "day .. Day",
        ("FIX.5.0", "41204", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41204", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41204", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41204", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41204", "g"): "g .. Grams",
        ("FIX.5.0", "41204", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41204", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41204", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41204", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41204", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41204", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41204", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41204", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41204", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41204", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41204", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41204", "L"): "L .. liters",
        ("FIX.5.0", "41204", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41204", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41204", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41204", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41204", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41204", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41204", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41204", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41204", "t"): "t .. MetricTons",
        ("FIX.5.0", "41204", "tn"): "tn .. Tons",
        ("FIX.5.0", "41206", "0"): "0 .. Absolute",
        ("FIX.5.0", "41206", "1"): "1 .. Percentage",
        ("FIX.5.0", "41210", "D"): "D .. Day",
        ("FIX.5.0", "41210", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41210", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41210", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41212", "D"): "D .. Day",
        ("FIX.5.0", "41212", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41212", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41212", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41213", "0"): "0 .. Business",
        ("FIX.5.0", "41213", "1"): "1 .. Calendar",
        ("FIX.5.0", "41213", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41213", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41213", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41213", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41214", "0"): "0 .. All",
        ("FIX.5.0", "41214", "1"): "1 .. First",
        ("FIX.5.0", "41214", "2"): "2 .. Last",
        ("FIX.5.0", "41214", "3"): "3 .. Penultimate",
        ("FIX.5.0", "41217", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41217", "1"): "1 .. None",
        ("FIX.5.0", "41217", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41217", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41217", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41217", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41217", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41217", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41222", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41222", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41226", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41226", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41228", "0"): "0 .. EveryDay",
        ("FIX.5.0", "41228", "1"): "1 .. Monday",
        ("FIX.5.0", "41228", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41228", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41228", "4"): "4 .. Thursday",
        ("FIX.5.0", "41228", "5"): "5 .. Friday",
        ("FIX.5.0", "41228", "6"): "6 .. Saturday",
        ("FIX.5.0", "41228", "7"): "7 .. Sunday",
        ("FIX.5.0", "41233", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41233", "1"): "1 .. None",
        ("FIX.5.0", "41233", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41233", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41233", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41233", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41233", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41233", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41237", "D"): "D .. Day",
        ("FIX.5.0", "41243", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41243", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41248", "D"): "D .. Day",
        ("FIX.5.0", "41248", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41248", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41248", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41254", "1"): "1 .. CUSIP",
        ("FIX.5.0", "41254", "2"): "2 .. SEDOL",
        ("FIX.5.0", "41254", "3"): "3 .. QUIK",
        ("FIX.5.0", "41254", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "41254", "5"): "5 .. RICCode",
        ("FIX.5.0", "41254", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "41254", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "41254", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "41254", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "41254", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "41254", "B"): "B .. Wertpapier",
        ("FIX.5.0", "41254", "C"): "C .. Dutch",
        ("FIX.5.0", "41254", "D"): "D .. Valoren",
        ("FIX.5.0", "41254", "E"): "E .. Sicovam",
        ("FIX.5.0", "41254", "F"): "F .. Belgian",
        ("FIX.5.0", "41254", "G"): "G .. Common",
        ("FIX.5.0", "41254", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "41254", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "41254", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "41254", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "41254", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "41254", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "41254", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "41254", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "41254", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "41254", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "41254", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "41254", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "41254", "U"): "U .. Synthetic",
        ("FIX.5.0", "41254", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "41254", "W"): "W .. IndexName",
        ("FIX.5.0", "41258", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41258", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41258", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41258", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41258", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41258", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41258", "thm"): "thm .. therms",
        ("FIX.5.0", "41258", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41258", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41258", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41258", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41258", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41258", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41258", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41258", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41258", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41258", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41258", "day"): "day .. Day",
        ("FIX.5.0", "41258", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41258", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41258", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41258", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41258", "g"): "g .. Grams",
        ("FIX.5.0", "41258", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41258", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41258", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41258", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41258", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41258", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41258", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41258", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41258", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41258", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41258", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41258", "L"): "L .. liters",
        ("FIX.5.0", "41258", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41258", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41258", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41258", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41258", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41258", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41258", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41258", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41258", "t"): "t .. MetricTons",
        ("FIX.5.0", "41258", "tn"): "tn .. Tons",
        ("FIX.5.0", "41267", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41267", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41269", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41269", "1"): "1 .. None",
        ("FIX.5.0", "41269", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41269", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41269", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41269", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41269", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41269", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41273", "D"): "D .. Day",
        ("FIX.5.0", "41274", "0"): "0 .. Business",
        ("FIX.5.0", "41274", "1"): "1 .. Calendar",
        ("FIX.5.0", "41274", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41274", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41274", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41274", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41282", "0"): "0 .. City",
        ("FIX.5.0", "41282", "1"): "1 .. Airport",
        ("FIX.5.0", "41282", "2"): "2 .. WeatherStation",
        ("FIX.5.0", "41282", "3"): "3 .. WeatherIndex",
        ("FIX.5.0", "41284", "1"): "1 .. Monday",
        ("FIX.5.0", "41284", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41284", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41284", "4"): "4 .. Thursday",
        ("FIX.5.0", "41284", "5"): "5 .. Friday",
        ("FIX.5.0", "41284", "6"): "6 .. Saturday",
        ("FIX.5.0", "41284", "7"): "7 .. Sunday",
        ("FIX.5.0", "41284", "8"): "8 .. AllWeekdays",
        ("FIX.5.0", "41284", "9"): "9 .. AllDays",
        ("FIX.5.0", "41284", "10"): "10 .. AllWeekends",
        ("FIX.5.0", "41292", "0"): "0 .. AllTimes",
        ("FIX.5.0", "41292", "1"): "1 .. OnPeak",
        ("FIX.5.0", "41292", "2"): "2 .. OffPeak",
        ("FIX.5.0", "41292", "3"): "3 .. Base",
        ("FIX.5.0", "41292", "4"): "4 .. BlockHours",
        ("FIX.5.0", "41292", "5"): "5 .. Other",
        ("FIX.5.0", "41294", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41294", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41294", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41294", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41294", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41294", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41294", "thm"): "thm .. therms",
        ("FIX.5.0", "41294", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41294", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41294", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41294", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41294", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41294", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41294", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41294", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41294", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41294", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41294", "day"): "day .. Day",
        ("FIX.5.0", "41294", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41294", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41294", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41294", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41294", "g"): "g .. Grams",
        ("FIX.5.0", "41294", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41294", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41294", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41294", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41294", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41294", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41294", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41294", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41294", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41294", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41294", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41294", "L"): "L .. liters",
        ("FIX.5.0", "41294", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41294", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41294", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41294", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41294", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41294", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41294", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41294", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41294", "t"): "t .. MetricTons",
        ("FIX.5.0", "41294", "tn"): "tn .. Tons",
        ("FIX.5.0", "41296", "D"): "D .. Day",
        ("FIX.5.0", "41296", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41296", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41296", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41298", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41298", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41298", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41298", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41298", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41298", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41298", "thm"): "thm .. therms",
        ("FIX.5.0", "41298", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41298", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41298", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41298", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41298", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41298", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41298", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41298", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41298", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41298", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41298", "day"): "day .. Day",
        ("FIX.5.0", "41298", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41298", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41298", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41298", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41298", "g"): "g .. Grams",
        ("FIX.5.0", "41298", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41298", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41298", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41298", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41298", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41298", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41298", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41298", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41298", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41298", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41298", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41298", "L"): "L .. liters",
        ("FIX.5.0", "41298", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41298", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41298", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41298", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41298", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41298", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41298", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41298", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41298", "t"): "t .. MetricTons",
        ("FIX.5.0", "41298", "tn"): "tn .. Tons",
        ("FIX.5.0", "41300", "0"): "0 .. DoNotIncludeHolidays",
        ("FIX.5.0", "41300", "1"): "1 .. IncludeHolidays",
        ("FIX.5.0", "41307", "H"): "H .. Hour",
        ("FIX.5.0", "41307", "Min"): "Min .. Minute",
        ("FIX.5.0", "41307", "S"): "S .. Second",
        ("FIX.5.0", "41307", "D"): "D .. Day",
        ("FIX.5.0", "41307", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41307", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41307", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41307", "Q"): "Q .. Quarter",
        ("FIX.5.0", "41308", "0"): "0 .. Term",
        ("FIX.5.0", "41308", "1"): "1 .. PerBusinessDay",
        ("FIX.5.0", "41308", "2"): "2 .. PerCalculationPeriod",
        ("FIX.5.0", "41308", "3"): "3 .. PerSettlPeriod",
        ("FIX.5.0", "41308", "4"): "4 .. PerCalendarDay",
        ("FIX.5.0", "41308", "5"): "5 .. PerHour",
        ("FIX.5.0", "41308", "6"): "6 .. PerMonth",
        ("FIX.5.0", "41309", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41309", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41309", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41309", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41309", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41309", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41309", "thm"): "thm .. therms",
        ("FIX.5.0", "41309", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41309", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41309", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41309", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41309", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41309", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41309", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41309", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41309", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41309", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41309", "day"): "day .. Day",
        ("FIX.5.0", "41309", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41309", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41309", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41309", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41309", "g"): "g .. Grams",
        ("FIX.5.0", "41309", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41309", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41309", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41309", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41309", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41309", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41309", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41309", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41309", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41309", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41309", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41309", "L"): "L .. liters",
        ("FIX.5.0", "41309", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41309", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41309", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41309", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41309", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41309", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41309", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41309", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41309", "t"): "t .. MetricTons",
        ("FIX.5.0", "41309", "tn"): "tn .. Tons",
        ("FIX.5.0", "41311", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41311", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41311", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41311", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41311", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41311", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41311", "thm"): "thm .. therms",
        ("FIX.5.0", "41311", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41311", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41311", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41311", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41311", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41311", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41311", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41311", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41311", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41311", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41311", "day"): "day .. Day",
        ("FIX.5.0", "41311", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41311", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41311", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41311", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41311", "g"): "g .. Grams",
        ("FIX.5.0", "41311", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41311", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41311", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41311", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41311", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41311", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41311", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41311", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41311", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41311", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41311", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41311", "L"): "L .. liters",
        ("FIX.5.0", "41311", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41311", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41311", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41311", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41311", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41311", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41311", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41311", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41311", "t"): "t .. MetricTons",
        ("FIX.5.0", "41311", "tn"): "tn .. Tons",
        ("FIX.5.0", "41318", "1"): "1 .. CUSIP",
        ("FIX.5.0", "41318", "2"): "2 .. SEDOL",
        ("FIX.5.0", "41318", "3"): "3 .. QUIK",
        ("FIX.5.0", "41318", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "41318", "5"): "5 .. RICCode",
        ("FIX.5.0", "41318", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "41318", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "41318", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "41318", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "41318", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "41318", "B"): "B .. Wertpapier",
        ("FIX.5.0", "41318", "C"): "C .. Dutch",
        ("FIX.5.0", "41318", "D"): "D .. Valoren",
        ("FIX.5.0", "41318", "E"): "E .. Sicovam",
        ("FIX.5.0", "41318", "F"): "F .. Belgian",
        ("FIX.5.0", "41318", "G"): "G .. Common",
        ("FIX.5.0", "41318", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "41318", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "41318", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "41318", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "41318", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "41318", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "41318", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "41318", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "41318", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "41318", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "41318", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "41318", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "41318", "U"): "U .. Synthetic",
        ("FIX.5.0", "41318", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "41318", "W"): "W .. IndexName",
        ("FIX.5.0", "41326", "SD"): "SD .. SeniorSecured",
        ("FIX.5.0", "41326", "SR"): "SR .. Senior",
        ("FIX.5.0", "41326", "SB"): "SB .. Subordinated",
        ("FIX.5.0", "41326", "JR"): "JR .. Junior",
        ("FIX.5.0", "41326", "MZ"): "MZ .. Mezzanine",
        ("FIX.5.0", "41326", "SN"): "SN .. SeniorNonPreferred",
        ("FIX.5.0", "41327", "0"): "0 .. Zero",
        ("FIX.5.0", "41327", "1"): "1 .. FixedRate",
        ("FIX.5.0", "41327", "2"): "2 .. FloatingRate",
        ("FIX.5.0", "41327", "3"): "3 .. Structured",
        ("FIX.5.0", "41333", "D"): "D .. Day",
        ("FIX.5.0", "41333", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41333", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41333", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41333", "H"): "H .. Hour",
        ("FIX.5.0", "41333", "Min"): "Min .. Minute",
        ("FIX.5.0", "41333", "S"): "S .. Second",
        ("FIX.5.0", "41333", "T"): "T .. Term",
        ("FIX.5.0", "41334", "0"): "0 .. OneOne",
        ("FIX.5.0", "41334", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "41334", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "41334", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "41334", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "41334", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "41334", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "41334", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "41334", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "41334", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "41334", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "41334", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "41334", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "41334", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "41334", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "41334", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "41334", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "41334", "17"): "17 .. Act364",
        ("FIX.5.0", "41351", "0"): "0 .. Bid",
        ("FIX.5.0", "41351", "1"): "1 .. Mid",
        ("FIX.5.0", "41351", "2"): "2 .. Offer",
        ("FIX.5.0", "41361", "0"): "0 .. Market",
        ("FIX.5.0", "41361", "1"): "1 .. Highest",
        ("FIX.5.0", "41361", "2"): "2 .. AverageMarket",
        ("FIX.5.0", "41361", "3"): "3 .. AverageHighest",
        ("FIX.5.0", "41361", "4"): "4 .. BlendedMarket",
        ("FIX.5.0", "41361", "5"): "5 .. BlendedHighest",
        ("FIX.5.0", "41361", "6"): "6 .. AverageBlendedMarket",
        ("FIX.5.0", "41361", "7"): "7 .. AverageBlendedHighest",
        ("FIX.5.0", "41371", "D"): "D .. Day",
        ("FIX.5.0", "41371", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41371", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41371", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41372", "0"): "0 .. Business",
        ("FIX.5.0", "41372", "1"): "1 .. Calendar",
        ("FIX.5.0", "41372", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41372", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41372", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41372", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41375", "H"): "H .. RestructuringMultipleHoldingObligations",
        ("FIX.5.0", "41375", "E"): "E .. RestructuringMultipleCreditEventNotices",
        ("FIX.5.0", "41375", "C"): "C .. FloatingRateInterestShortfall",
        ("FIX.5.0", "41380", "0"): "0 .. AsianOut",
        ("FIX.5.0", "41380", "1"): "1 .. AsianIn",
        ("FIX.5.0", "41380", "2"): "2 .. BarrierCap",
        ("FIX.5.0", "41380", "3"): "3 .. BarrierFloor",
        ("FIX.5.0", "41380", "4"): "4 .. KnockOut",
        ("FIX.5.0", "41380", "5"): "5 .. KnockIn",
        ("FIX.5.0", "41383", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "41383", "1"): "1 .. Reuters",
        ("FIX.5.0", "41383", "2"): "2 .. Telerate",
        ("FIX.5.0", "41383", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "41383", "99"): "99 .. Other",
        ("FIX.5.0", "41384", "0"): "0 .. Primary",
        ("FIX.5.0", "41384", "1"): "1 .. Secondary",
        ("FIX.5.0", "41392", "D"): "D .. Day",
        ("FIX.5.0", "41392", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41392", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41392", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41393", "0"): "0 .. Business",
        ("FIX.5.0", "41393", "1"): "1 .. Calendar",
        ("FIX.5.0", "41393", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41393", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41393", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41393", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41394", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41394", "1"): "1 .. None",
        ("FIX.5.0", "41394", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41394", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41394", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41394", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41394", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41394", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41404", "D"): "D .. Day",
        ("FIX.5.0", "41404", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41404", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41404", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41405", "1"): "1 .. FirstDay",
        ("FIX.5.0", "41405", "2"): "2 .. SecondDay",
        ("FIX.5.0", "41405", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "41405", "4"): "4 .. FourthDay",
        ("FIX.5.0", "41405", "5"): "5 .. FifthDay",
        ("FIX.5.0", "41405", "6"): "6 .. SixthDay",
        ("FIX.5.0", "41405", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "41405", "8"): "8 .. EighthDay",
        ("FIX.5.0", "41405", "9"): "9 .. NinthDay",
        ("FIX.5.0", "41405", "10"): "10 .. TenthDay",
        ("FIX.5.0", "41405", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "41405", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "41405", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "41405", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "41405", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "41405", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "41405", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "41405", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "41405", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "41405", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "41405", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "41405", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "41405", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "41405", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "41405", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "41405", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "41405", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "41405", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "41405", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "41405", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "41405", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "41405", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "41405", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "41405", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "41405", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "41405", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "41405", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "41405", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "41405", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "41405", "MON"): "MON .. MON",
        ("FIX.5.0", "41405", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "41405", "WED"): "WED .. WED",
        ("FIX.5.0", "41405", "THU"): "THU .. THU",
        ("FIX.5.0", "41405", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "41405", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "41405", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "41409", "0"): "0 .. Notional",
        ("FIX.5.0", "41409", "1"): "1 .. Delivery",
        ("FIX.5.0", "41409", "2"): "2 .. PhysicalSettlPeriods",
        ("FIX.5.0", "41412", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41412", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41412", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41412", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41412", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41412", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41412", "thm"): "thm .. therms",
        ("FIX.5.0", "41412", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41412", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41412", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41412", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41412", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41412", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41412", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41412", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41412", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41412", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41412", "day"): "day .. Day",
        ("FIX.5.0", "41412", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41412", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41412", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41412", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41412", "g"): "g .. Grams",
        ("FIX.5.0", "41412", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41412", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41412", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41412", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41412", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41412", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41412", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41412", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41412", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41412", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41412", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41412", "L"): "L .. liters",
        ("FIX.5.0", "41412", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41412", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41412", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41412", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41412", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41412", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41412", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41412", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41412", "t"): "t .. MetricTons",
        ("FIX.5.0", "41412", "tn"): "tn .. Tons",
        ("FIX.5.0", "41413", "0"): "0 .. Term",
        ("FIX.5.0", "41413", "1"): "1 .. PerBusinessDay",
        ("FIX.5.0", "41413", "2"): "2 .. PerCalculationPeriod",
        ("FIX.5.0", "41413", "3"): "3 .. PerSettlPeriod",
        ("FIX.5.0", "41413", "4"): "4 .. PerCalendarDay",
        ("FIX.5.0", "41413", "5"): "5 .. PerHour",
        ("FIX.5.0", "41413", "6"): "6 .. PerMonth",
        ("FIX.5.0", "41416", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41416", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41416", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41416", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41416", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41416", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41416", "thm"): "thm .. therms",
        ("FIX.5.0", "41416", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41416", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41416", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41416", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41416", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41416", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41416", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41416", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41416", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41416", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41416", "day"): "day .. Day",
        ("FIX.5.0", "41416", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41416", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41416", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41416", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41416", "g"): "g .. Grams",
        ("FIX.5.0", "41416", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41416", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41416", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41416", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41416", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41416", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41416", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41416", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41416", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41416", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41416", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41416", "L"): "L .. liters",
        ("FIX.5.0", "41416", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41416", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41416", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41416", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41416", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41416", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41416", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41416", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41416", "t"): "t .. MetricTons",
        ("FIX.5.0", "41416", "tn"): "tn .. Tons",
        ("FIX.5.0", "41417", "0"): "0 .. Absolute",
        ("FIX.5.0", "41417", "1"): "1 .. Percentage",
        ("FIX.5.0", "41420", "0"): "0 .. AllTimes",
        ("FIX.5.0", "41420", "1"): "1 .. OnPeak",
        ("FIX.5.0", "41420", "2"): "2 .. OffPeak",
        ("FIX.5.0", "41420", "3"): "3 .. Base",
        ("FIX.5.0", "41420", "4"): "4 .. BlockHours",
        ("FIX.5.0", "41420", "5"): "5 .. Other",
        ("FIX.5.0", "41421", "0"): "0 .. DoNotIncludeHolidays",
        ("FIX.5.0", "41421", "1"): "1 .. IncludeHolidays",
        ("FIX.5.0", "41423", "1"): "1 .. Monday",
        ("FIX.5.0", "41423", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41423", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41423", "4"): "4 .. Thursday",
        ("FIX.5.0", "41423", "5"): "5 .. Friday",
        ("FIX.5.0", "41423", "6"): "6 .. Saturday",
        ("FIX.5.0", "41423", "7"): "7 .. Sunday",
        ("FIX.5.0", "41423", "8"): "8 .. AllWeekdays",
        ("FIX.5.0", "41423", "9"): "9 .. AllDays",
        ("FIX.5.0", "41423", "10"): "10 .. AllWeekends",
        ("FIX.5.0", "41428", "0"): "0 .. Hour",
        ("FIX.5.0", "41428", "1"): "1 .. Timestamp",
        ("FIX.5.0", "41429", "0"): "0 .. Periodic",
        ("FIX.5.0", "41429", "1"): "1 .. Initial",
        ("FIX.5.0", "41429", "2"): "2 .. Single",
        ("FIX.5.0", "41434", "1"): "1 .. Firm",
        ("FIX.5.0", "41434", "2"): "2 .. NonFirm",
        ("FIX.5.0", "41434", "3"): "3 .. ForceMajeure",
        ("FIX.5.0", "41434", "4"): "4 .. SystemFirm",
        ("FIX.5.0", "41434", "5"): "5 .. UnitFirm",
        ("FIX.5.0", "41436", "0"): "0 .. Buyer",
        ("FIX.5.0", "41436", "1"): "1 .. Seller",
        ("FIX.5.0", "41440", "0"): "0 .. Transfers",
        ("FIX.5.0", "41440", "1"): "1 .. DoesNotTransfer",
        ("FIX.5.0", "41444", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41444", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41444", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41444", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41444", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41444", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41444", "thm"): "thm .. therms",
        ("FIX.5.0", "41444", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41444", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41444", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41444", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41444", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41444", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41444", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41444", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41444", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41444", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41444", "day"): "day .. Day",
        ("FIX.5.0", "41444", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41444", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41444", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41444", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41444", "g"): "g .. Grams",
        ("FIX.5.0", "41444", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41444", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41444", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41444", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41444", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41444", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41444", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41444", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41444", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41444", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41444", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41444", "L"): "L .. liters",
        ("FIX.5.0", "41444", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41444", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41444", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41444", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41444", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41444", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41444", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41444", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41444", "t"): "t .. MetricTons",
        ("FIX.5.0", "41444", "tn"): "tn .. Tons",
        ("FIX.5.0", "41445", "0"): "0 .. Absolute",
        ("FIX.5.0", "41445", "1"): "1 .. Percentage",
        ("FIX.5.0", "41446", "1"): "1 .. Buyer",
        ("FIX.5.0", "41446", "2"): "2 .. Seller",
        ("FIX.5.0", "41451", "0"): "0 .. Buyer",
        ("FIX.5.0", "41451", "1"): "1 .. Seller",
        ("FIX.5.0", "41462", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41462", "1"): "1 .. Applicable",
        ("FIX.5.0", "41462", "2"): "2 .. AsInMasterAgreement",
        ("FIX.5.0", "41462", "3"): "3 .. AsInConfirmation",
        ("FIX.5.0", "41463", "0"): "0 .. MasterAgreement",
        ("FIX.5.0", "41463", "1"): "1 .. Confirmation",
        ("FIX.5.0", "41472", "0"): "0 .. Basket",
        ("FIX.5.0", "41472", "1"): "1 .. Bond",
        ("FIX.5.0", "41472", "2"): "2 .. Cash",
        ("FIX.5.0", "41472", "3"): "3 .. Commodity",
        ("FIX.5.0", "41472", "4"): "4 .. ConvertibleBond",
        ("FIX.5.0", "41472", "5"): "5 .. Equity",
        ("FIX.5.0", "41472", "6"): "6 .. ExchangeTradedFund",
        ("FIX.5.0", "41472", "7"): "7 .. Future",
        ("FIX.5.0", "41472", "8"): "8 .. Index",
        ("FIX.5.0", "41472", "9"): "9 .. Loan",
        ("FIX.5.0", "41472", "10"): "10 .. Mortgage",
        ("FIX.5.0", "41472", "11"): "11 .. MutualFund",
        ("FIX.5.0", "41474", "1"): "1 .. CUSIP",
        ("FIX.5.0", "41474", "2"): "2 .. SEDOL",
        ("FIX.5.0", "41474", "3"): "3 .. QUIK",
        ("FIX.5.0", "41474", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "41474", "5"): "5 .. RICCode",
        ("FIX.5.0", "41474", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "41474", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "41474", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "41474", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "41474", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "41474", "B"): "B .. Wertpapier",
        ("FIX.5.0", "41474", "C"): "C .. Dutch",
        ("FIX.5.0", "41474", "D"): "D .. Valoren",
        ("FIX.5.0", "41474", "E"): "E .. Sicovam",
        ("FIX.5.0", "41474", "F"): "F .. Belgian",
        ("FIX.5.0", "41474", "G"): "G .. Common",
        ("FIX.5.0", "41474", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "41474", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "41474", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "41474", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "41474", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "41474", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "41474", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "41474", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "41474", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "41474", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "41474", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "41474", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "41474", "U"): "U .. Synthetic",
        ("FIX.5.0", "41474", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "41474", "W"): "W .. IndexName",
        ("FIX.5.0", "41486", "0"): "0 .. NotRequired",
        ("FIX.5.0", "41486", "1"): "1 .. NonElectronic",
        ("FIX.5.0", "41486", "2"): "2 .. Electronic",
        ("FIX.5.0", "41486", "3"): "3 .. Unknown",
        ("FIX.5.0", "41493", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41493", "1"): "1 .. None",
        ("FIX.5.0", "41493", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41493", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41493", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41493", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41493", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41493", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41494", "0"): "0 .. Business",
        ("FIX.5.0", "41494", "1"): "1 .. Calendar",
        ("FIX.5.0", "41494", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41494", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41494", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41494", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41496", "D"): "D .. Day",
        ("FIX.5.0", "41496", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41496", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41496", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41498", "D"): "D .. Day",
        ("FIX.5.0", "41498", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41498", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41498", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41502", "D"): "D .. Day",
        ("FIX.5.0", "41502", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41502", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41502", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41503", "0"): "0 .. Business",
        ("FIX.5.0", "41503", "1"): "1 .. Calendar",
        ("FIX.5.0", "41503", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41503", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41503", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41503", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41514", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41514", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41517", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41517", "1"): "1 .. None",
        ("FIX.5.0", "41517", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41517", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41517", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41517", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41517", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41517", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41520", "D"): "D .. Day",
        ("FIX.5.0", "41520", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41520", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41520", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41522", "D"): "D .. Day",
        ("FIX.5.0", "41522", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41522", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41522", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41523", "1"): "1 .. FirstDay",
        ("FIX.5.0", "41523", "2"): "2 .. SecondDay",
        ("FIX.5.0", "41523", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "41523", "4"): "4 .. FourthDay",
        ("FIX.5.0", "41523", "5"): "5 .. FifthDay",
        ("FIX.5.0", "41523", "6"): "6 .. SixthDay",
        ("FIX.5.0", "41523", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "41523", "8"): "8 .. EighthDay",
        ("FIX.5.0", "41523", "9"): "9 .. NinthDay",
        ("FIX.5.0", "41523", "10"): "10 .. TenthDay",
        ("FIX.5.0", "41523", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "41523", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "41523", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "41523", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "41523", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "41523", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "41523", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "41523", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "41523", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "41523", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "41523", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "41523", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "41523", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "41523", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "41523", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "41523", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "41523", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "41523", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "41523", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "41523", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "41523", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "41523", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "41523", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "41523", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "41523", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "41523", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "41523", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "41523", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "41523", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "41523", "MON"): "MON .. MON",
        ("FIX.5.0", "41523", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "41523", "WED"): "WED .. WED",
        ("FIX.5.0", "41523", "THU"): "THU .. THU",
        ("FIX.5.0", "41523", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "41523", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "41523", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "41524", "0"): "0 .. Business",
        ("FIX.5.0", "41524", "1"): "1 .. Calendar",
        ("FIX.5.0", "41524", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41524", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41524", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41524", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41529", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41529", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41531", "0"): "0 .. EveryDay",
        ("FIX.5.0", "41531", "1"): "1 .. Monday",
        ("FIX.5.0", "41531", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41531", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41531", "4"): "4 .. Thursday",
        ("FIX.5.0", "41531", "5"): "5 .. Friday",
        ("FIX.5.0", "41531", "6"): "6 .. Saturday",
        ("FIX.5.0", "41531", "7"): "7 .. Sunday",
        ("FIX.5.0", "41536", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41536", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41536", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41536", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41536", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41536", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41536", "thm"): "thm .. therms",
        ("FIX.5.0", "41536", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41536", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41536", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41536", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41536", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41536", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41536", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41536", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41536", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41536", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41536", "day"): "day .. Day",
        ("FIX.5.0", "41536", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41536", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41536", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41536", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41536", "g"): "g .. Grams",
        ("FIX.5.0", "41536", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41536", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41536", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41536", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41536", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41536", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41536", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41536", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41536", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41536", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41536", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41536", "L"): "L .. liters",
        ("FIX.5.0", "41536", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41536", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41536", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41536", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41536", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41536", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41536", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41536", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41536", "t"): "t .. MetricTons",
        ("FIX.5.0", "41536", "tn"): "tn .. Tons",
        ("FIX.5.0", "41538", "0"): "0 .. Absolute",
        ("FIX.5.0", "41538", "1"): "1 .. Percentage",
        ("FIX.5.0", "41541", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41541", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41541", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41541", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41541", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41541", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41541", "thm"): "thm .. therms",
        ("FIX.5.0", "41541", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41541", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41541", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41541", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41541", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41541", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41541", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41541", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41541", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41541", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41541", "day"): "day .. Day",
        ("FIX.5.0", "41541", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41541", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41541", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41541", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41541", "g"): "g .. Grams",
        ("FIX.5.0", "41541", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41541", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41541", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41541", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41541", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41541", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41541", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41541", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41541", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41541", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41541", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41541", "L"): "L .. liters",
        ("FIX.5.0", "41541", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41541", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41541", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41541", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41541", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41541", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41541", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41541", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41541", "t"): "t .. MetricTons",
        ("FIX.5.0", "41541", "tn"): "tn .. Tons",
        ("FIX.5.0", "41542", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41542", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41542", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41542", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41542", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41542", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41542", "thm"): "thm .. therms",
        ("FIX.5.0", "41542", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41542", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41542", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41542", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41542", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41542", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41542", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41542", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41542", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41542", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41542", "day"): "day .. Day",
        ("FIX.5.0", "41542", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41542", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41542", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41542", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41542", "g"): "g .. Grams",
        ("FIX.5.0", "41542", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41542", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41542", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41542", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41542", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41542", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41542", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41542", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41542", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41542", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41542", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41542", "L"): "L .. liters",
        ("FIX.5.0", "41542", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41542", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41542", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41542", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41542", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41542", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41542", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41542", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41542", "t"): "t .. MetricTons",
        ("FIX.5.0", "41542", "tn"): "tn .. Tons",
        ("FIX.5.0", "41543", "0"): "0 .. All",
        ("FIX.5.0", "41543", "1"): "1 .. First",
        ("FIX.5.0", "41543", "2"): "2 .. Last",
        ("FIX.5.0", "41543", "3"): "3 .. Penultimate",
        ("FIX.5.0", "41546", "D"): "D .. Day",
        ("FIX.5.0", "41546", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41546", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41546", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41548", "D"): "D .. Day",
        ("FIX.5.0", "41548", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41548", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41548", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41556", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41556", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41556", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41556", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41556", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41556", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41556", "thm"): "thm .. therms",
        ("FIX.5.0", "41556", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41556", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41556", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41556", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41556", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41556", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41556", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41556", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41556", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41556", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41556", "day"): "day .. Day",
        ("FIX.5.0", "41556", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41556", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41556", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41556", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41556", "g"): "g .. Grams",
        ("FIX.5.0", "41556", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41556", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41556", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41556", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41556", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41556", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41556", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41556", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41556", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41556", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41556", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41556", "L"): "L .. liters",
        ("FIX.5.0", "41556", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41556", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41556", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41556", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41556", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41556", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41556", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41556", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41556", "t"): "t .. MetricTons",
        ("FIX.5.0", "41556", "tn"): "tn .. Tons",
        ("FIX.5.0", "41563", "D"): "D .. Day",
        ("FIX.5.0", "41563", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41563", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41563", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41567", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41567", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41567", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41567", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41567", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41567", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41567", "thm"): "thm .. therms",
        ("FIX.5.0", "41567", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41567", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41567", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41567", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41567", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41567", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41567", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41567", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41567", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41567", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41567", "day"): "day .. Day",
        ("FIX.5.0", "41567", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41567", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41567", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41567", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41567", "g"): "g .. Grams",
        ("FIX.5.0", "41567", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41567", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41567", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41567", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41567", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41567", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41567", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41567", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41567", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41567", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41567", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41567", "L"): "L .. liters",
        ("FIX.5.0", "41567", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41567", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41567", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41567", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41567", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41567", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41567", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41567", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41567", "t"): "t .. MetricTons",
        ("FIX.5.0", "41567", "tn"): "tn .. Tons",
        ("FIX.5.0", "41568", "0"): "0 .. Average",
        ("FIX.5.0", "41568", "1"): "1 .. Maximum",
        ("FIX.5.0", "41568", "2"): "2 .. Minimum",
        ("FIX.5.0", "41568", "3"): "3 .. Cumulative",
        ("FIX.5.0", "41570", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41570", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41570", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41570", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41570", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41570", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41570", "thm"): "thm .. therms",
        ("FIX.5.0", "41570", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41570", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41570", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41570", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41570", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41570", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41570", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41570", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41570", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41570", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41570", "day"): "day .. Day",
        ("FIX.5.0", "41570", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41570", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41570", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41570", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41570", "g"): "g .. Grams",
        ("FIX.5.0", "41570", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41570", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41570", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41570", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41570", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41570", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41570", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41570", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41570", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41570", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41570", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41570", "L"): "L .. liters",
        ("FIX.5.0", "41570", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41570", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41570", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41570", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41570", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41570", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41570", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41570", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41570", "t"): "t .. MetricTons",
        ("FIX.5.0", "41570", "tn"): "tn .. Tons",
        ("FIX.5.0", "41573", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41573", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41573", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41573", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41573", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41573", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41573", "thm"): "thm .. therms",
        ("FIX.5.0", "41573", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41573", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41573", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41573", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41573", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41573", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41573", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41573", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41573", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41573", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41573", "day"): "day .. Day",
        ("FIX.5.0", "41573", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41573", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41573", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41573", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41573", "g"): "g .. Grams",
        ("FIX.5.0", "41573", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41573", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41573", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41573", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41573", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41573", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41573", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41573", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41573", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41573", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41573", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41573", "L"): "L .. liters",
        ("FIX.5.0", "41573", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41573", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41573", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41573", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41573", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41573", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41573", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41573", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41573", "t"): "t .. MetricTons",
        ("FIX.5.0", "41573", "tn"): "tn .. Tons",
        ("FIX.5.0", "41575", "0"): "0 .. Absolute",
        ("FIX.5.0", "41575", "1"): "1 .. Percentage",
        ("FIX.5.0", "41579", "D"): "D .. Day",
        ("FIX.5.0", "41579", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41579", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41579", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41581", "D"): "D .. Day",
        ("FIX.5.0", "41581", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41581", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41581", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41582", "0"): "0 .. Business",
        ("FIX.5.0", "41582", "1"): "1 .. Calendar",
        ("FIX.5.0", "41582", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41582", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41582", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41582", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41583", "0"): "0 .. All",
        ("FIX.5.0", "41583", "1"): "1 .. First",
        ("FIX.5.0", "41583", "2"): "2 .. Last",
        ("FIX.5.0", "41583", "3"): "3 .. Penultimate",
        ("FIX.5.0", "41586", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41586", "1"): "1 .. None",
        ("FIX.5.0", "41586", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41586", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41586", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41586", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41586", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41586", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41588", "0"): "0 .. Hour",
        ("FIX.5.0", "41588", "1"): "1 .. Timestamp",
        ("FIX.5.0", "41591", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41591", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41595", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41595", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41597", "0"): "0 .. EveryDay",
        ("FIX.5.0", "41597", "1"): "1 .. Monday",
        ("FIX.5.0", "41597", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41597", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41597", "4"): "4 .. Thursday",
        ("FIX.5.0", "41597", "5"): "5 .. Friday",
        ("FIX.5.0", "41597", "6"): "6 .. Saturday",
        ("FIX.5.0", "41597", "7"): "7 .. Sunday",
        ("FIX.5.0", "41610", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41610", "1"): "1 .. None",
        ("FIX.5.0", "41610", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41610", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41610", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41610", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41610", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41610", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41630", "D"): "D .. Day",
        ("FIX.5.0", "41630", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41630", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41630", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41631", "0"): "0 .. Business",
        ("FIX.5.0", "41631", "1"): "1 .. Calendar",
        ("FIX.5.0", "41631", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41631", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41631", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41631", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41634", "H"): "H .. RestructuringMultipleHoldingObligations",
        ("FIX.5.0", "41634", "E"): "E .. RestructuringMultipleCreditEventNotices",
        ("FIX.5.0", "41634", "C"): "C .. FloatingRateInterestShortfall",
        ("FIX.5.0", "41640", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41640", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41645", "D"): "D .. Day",
        ("FIX.5.0", "41645", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41645", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41645", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41651", "1"): "1 .. CUSIP",
        ("FIX.5.0", "41651", "2"): "2 .. SEDOL",
        ("FIX.5.0", "41651", "3"): "3 .. QUIK",
        ("FIX.5.0", "41651", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "41651", "5"): "5 .. RICCode",
        ("FIX.5.0", "41651", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "41651", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "41651", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "41651", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "41651", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "41651", "B"): "B .. Wertpapier",
        ("FIX.5.0", "41651", "C"): "C .. Dutch",
        ("FIX.5.0", "41651", "D"): "D .. Valoren",
        ("FIX.5.0", "41651", "E"): "E .. Sicovam",
        ("FIX.5.0", "41651", "F"): "F .. Belgian",
        ("FIX.5.0", "41651", "G"): "G .. Common",
        ("FIX.5.0", "41651", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "41651", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "41651", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "41651", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "41651", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "41651", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "41651", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "41651", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "41651", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "41651", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "41651", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "41651", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "41651", "U"): "U .. Synthetic",
        ("FIX.5.0", "41651", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "41651", "W"): "W .. IndexName",
        ("FIX.5.0", "41655", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41655", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41655", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41655", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41655", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41655", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41655", "thm"): "thm .. therms",
        ("FIX.5.0", "41655", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41655", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41655", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41655", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41655", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41655", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41655", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41655", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41655", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41655", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41655", "day"): "day .. Day",
        ("FIX.5.0", "41655", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41655", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41655", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41655", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41655", "g"): "g .. Grams",
        ("FIX.5.0", "41655", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41655", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41655", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41655", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41655", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41655", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41655", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41655", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41655", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41655", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41655", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41655", "L"): "L .. liters",
        ("FIX.5.0", "41655", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41655", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41655", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41655", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41655", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41655", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41655", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41655", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41655", "t"): "t .. MetricTons",
        ("FIX.5.0", "41655", "tn"): "tn .. Tons",
        ("FIX.5.0", "41664", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41664", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41666", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41666", "1"): "1 .. None",
        ("FIX.5.0", "41666", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41666", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41666", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41666", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41666", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41666", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41670", "D"): "D .. Day",
        ("FIX.5.0", "41671", "0"): "0 .. Business",
        ("FIX.5.0", "41671", "1"): "1 .. Calendar",
        ("FIX.5.0", "41671", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41671", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41671", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41671", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41679", "0"): "0 .. City",
        ("FIX.5.0", "41679", "1"): "1 .. Airport",
        ("FIX.5.0", "41679", "2"): "2 .. WeatherStation",
        ("FIX.5.0", "41679", "3"): "3 .. WeatherIndex",
        ("FIX.5.0", "41681", "1"): "1 .. Monday",
        ("FIX.5.0", "41681", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41681", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41681", "4"): "4 .. Thursday",
        ("FIX.5.0", "41681", "5"): "5 .. Friday",
        ("FIX.5.0", "41681", "6"): "6 .. Saturday",
        ("FIX.5.0", "41681", "7"): "7 .. Sunday",
        ("FIX.5.0", "41681", "8"): "8 .. AllWeekdays",
        ("FIX.5.0", "41681", "9"): "9 .. AllDays",
        ("FIX.5.0", "41681", "10"): "10 .. AllWeekends",
        ("FIX.5.0", "41689", "0"): "0 .. AllTimes",
        ("FIX.5.0", "41689", "1"): "1 .. OnPeak",
        ("FIX.5.0", "41689", "2"): "2 .. OffPeak",
        ("FIX.5.0", "41689", "3"): "3 .. Base",
        ("FIX.5.0", "41689", "4"): "4 .. BlockHours",
        ("FIX.5.0", "41689", "5"): "5 .. Other",
        ("FIX.5.0", "41691", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41691", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41691", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41691", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41691", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41691", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41691", "thm"): "thm .. therms",
        ("FIX.5.0", "41691", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41691", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41691", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41691", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41691", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41691", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41691", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41691", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41691", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41691", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41691", "day"): "day .. Day",
        ("FIX.5.0", "41691", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41691", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41691", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41691", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41691", "g"): "g .. Grams",
        ("FIX.5.0", "41691", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41691", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41691", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41691", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41691", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41691", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41691", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41691", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41691", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41691", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41691", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41691", "L"): "L .. liters",
        ("FIX.5.0", "41691", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41691", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41691", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41691", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41691", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41691", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41691", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41691", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41691", "t"): "t .. MetricTons",
        ("FIX.5.0", "41691", "tn"): "tn .. Tons",
        ("FIX.5.0", "41693", "D"): "D .. Day",
        ("FIX.5.0", "41693", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41693", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41693", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41695", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41695", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41695", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41695", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41695", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41695", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41695", "thm"): "thm .. therms",
        ("FIX.5.0", "41695", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41695", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41695", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41695", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41695", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41695", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41695", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41695", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41695", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41695", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41695", "day"): "day .. Day",
        ("FIX.5.0", "41695", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41695", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41695", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41695", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41695", "g"): "g .. Grams",
        ("FIX.5.0", "41695", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41695", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41695", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41695", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41695", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41695", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41695", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41695", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41695", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41695", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41695", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41695", "L"): "L .. liters",
        ("FIX.5.0", "41695", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41695", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41695", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41695", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41695", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41695", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41695", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41695", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41695", "t"): "t .. MetricTons",
        ("FIX.5.0", "41695", "tn"): "tn .. Tons",
        ("FIX.5.0", "41697", "0"): "0 .. DoNotIncludeHolidays",
        ("FIX.5.0", "41697", "1"): "1 .. IncludeHolidays",
        ("FIX.5.0", "41701", "1"): "1 .. CUSIP",
        ("FIX.5.0", "41701", "2"): "2 .. SEDOL",
        ("FIX.5.0", "41701", "3"): "3 .. QUIK",
        ("FIX.5.0", "41701", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "41701", "5"): "5 .. RICCode",
        ("FIX.5.0", "41701", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "41701", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "41701", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "41701", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "41701", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "41701", "B"): "B .. Wertpapier",
        ("FIX.5.0", "41701", "C"): "C .. Dutch",
        ("FIX.5.0", "41701", "D"): "D .. Valoren",
        ("FIX.5.0", "41701", "E"): "E .. Sicovam",
        ("FIX.5.0", "41701", "F"): "F .. Belgian",
        ("FIX.5.0", "41701", "G"): "G .. Common",
        ("FIX.5.0", "41701", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "41701", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "41701", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "41701", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "41701", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "41701", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "41701", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "41701", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "41701", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "41701", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "41701", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "41701", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "41701", "U"): "U .. Synthetic",
        ("FIX.5.0", "41701", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "41701", "W"): "W .. IndexName",
        ("FIX.5.0", "41704", "H"): "H .. Hour",
        ("FIX.5.0", "41704", "Min"): "Min .. Minute",
        ("FIX.5.0", "41704", "S"): "S .. Second",
        ("FIX.5.0", "41704", "D"): "D .. Day",
        ("FIX.5.0", "41704", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41704", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41704", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41704", "Q"): "Q .. Quarter",
        ("FIX.5.0", "41705", "0"): "0 .. Term",
        ("FIX.5.0", "41705", "1"): "1 .. PerBusinessDay",
        ("FIX.5.0", "41705", "2"): "2 .. PerCalculationPeriod",
        ("FIX.5.0", "41705", "3"): "3 .. PerSettlPeriod",
        ("FIX.5.0", "41705", "4"): "4 .. PerCalendarDay",
        ("FIX.5.0", "41705", "5"): "5 .. PerHour",
        ("FIX.5.0", "41705", "6"): "6 .. PerMonth",
        ("FIX.5.0", "41706", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41706", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41706", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41706", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41706", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41706", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41706", "thm"): "thm .. therms",
        ("FIX.5.0", "41706", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41706", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41706", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41706", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41706", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41706", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41706", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41706", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41706", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41706", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41706", "day"): "day .. Day",
        ("FIX.5.0", "41706", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41706", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41706", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41706", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41706", "g"): "g .. Grams",
        ("FIX.5.0", "41706", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41706", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41706", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41706", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41706", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41706", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41706", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41706", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41706", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41706", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41706", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41706", "L"): "L .. liters",
        ("FIX.5.0", "41706", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41706", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41706", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41706", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41706", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41706", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41706", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41706", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41706", "t"): "t .. MetricTons",
        ("FIX.5.0", "41706", "tn"): "tn .. Tons",
        ("FIX.5.0", "41708", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41708", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41708", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41708", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41708", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41708", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41708", "thm"): "thm .. therms",
        ("FIX.5.0", "41708", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41708", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41708", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41708", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41708", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41708", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41708", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41708", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41708", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41708", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41708", "day"): "day .. Day",
        ("FIX.5.0", "41708", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41708", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41708", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41708", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41708", "g"): "g .. Grams",
        ("FIX.5.0", "41708", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41708", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41708", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41708", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41708", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41708", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41708", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41708", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41708", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41708", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41708", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41708", "L"): "L .. liters",
        ("FIX.5.0", "41708", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41708", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41708", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41708", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41708", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41708", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41708", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41708", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41708", "t"): "t .. MetricTons",
        ("FIX.5.0", "41708", "tn"): "tn .. Tons",
        ("FIX.5.0", "41721", "D"): "D .. Day",
        ("FIX.5.0", "41721", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41721", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41721", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41722", "0"): "0 .. Business",
        ("FIX.5.0", "41722", "1"): "1 .. Calendar",
        ("FIX.5.0", "41722", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41722", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41722", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41722", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41725", "H"): "H .. RestructuringMultipleHoldingObligations",
        ("FIX.5.0", "41725", "E"): "E .. RestructuringMultipleCreditEventNotices",
        ("FIX.5.0", "41725", "C"): "C .. FloatingRateInterestShortfall",
        ("FIX.5.0", "41730", "0"): "0 .. AsianOut",
        ("FIX.5.0", "41730", "1"): "1 .. AsianIn",
        ("FIX.5.0", "41730", "2"): "2 .. BarrierCap",
        ("FIX.5.0", "41730", "3"): "3 .. BarrierFloor",
        ("FIX.5.0", "41730", "4"): "4 .. KnockOut",
        ("FIX.5.0", "41730", "5"): "5 .. KnockIn",
        ("FIX.5.0", "41733", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "41733", "1"): "1 .. Reuters",
        ("FIX.5.0", "41733", "2"): "2 .. Telerate",
        ("FIX.5.0", "41733", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "41733", "99"): "99 .. Other",
        ("FIX.5.0", "41734", "0"): "0 .. Primary",
        ("FIX.5.0", "41734", "1"): "1 .. Secondary",
        ("FIX.5.0", "41742", "D"): "D .. Day",
        ("FIX.5.0", "41742", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41742", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41742", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41743", "0"): "0 .. Business",
        ("FIX.5.0", "41743", "1"): "1 .. Calendar",
        ("FIX.5.0", "41743", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41743", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41743", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41743", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41744", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41744", "1"): "1 .. None",
        ("FIX.5.0", "41744", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41744", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41744", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41744", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41744", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41744", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41754", "D"): "D .. Day",
        ("FIX.5.0", "41754", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41754", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41754", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41755", "1"): "1 .. FirstDay",
        ("FIX.5.0", "41755", "2"): "2 .. SecondDay",
        ("FIX.5.0", "41755", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "41755", "4"): "4 .. FourthDay",
        ("FIX.5.0", "41755", "5"): "5 .. FifthDay",
        ("FIX.5.0", "41755", "6"): "6 .. SixthDay",
        ("FIX.5.0", "41755", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "41755", "8"): "8 .. EighthDay",
        ("FIX.5.0", "41755", "9"): "9 .. NinthDay",
        ("FIX.5.0", "41755", "10"): "10 .. TenthDay",
        ("FIX.5.0", "41755", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "41755", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "41755", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "41755", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "41755", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "41755", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "41755", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "41755", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "41755", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "41755", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "41755", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "41755", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "41755", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "41755", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "41755", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "41755", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "41755", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "41755", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "41755", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "41755", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "41755", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "41755", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "41755", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "41755", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "41755", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "41755", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "41755", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "41755", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "41755", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "41755", "MON"): "MON .. MON",
        ("FIX.5.0", "41755", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "41755", "WED"): "WED .. WED",
        ("FIX.5.0", "41755", "THU"): "THU .. THU",
        ("FIX.5.0", "41755", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "41755", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "41755", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "41757", "0"): "0 .. Notional",
        ("FIX.5.0", "41757", "1"): "1 .. Delivery",
        ("FIX.5.0", "41757", "2"): "2 .. PhysicalSettlPeriods",
        ("FIX.5.0", "41760", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41760", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41760", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41760", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41760", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41760", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41760", "thm"): "thm .. therms",
        ("FIX.5.0", "41760", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41760", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41760", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41760", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41760", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41760", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41760", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41760", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41760", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41760", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41760", "day"): "day .. Day",
        ("FIX.5.0", "41760", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41760", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41760", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41760", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41760", "g"): "g .. Grams",
        ("FIX.5.0", "41760", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41760", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41760", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41760", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41760", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41760", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41760", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41760", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41760", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41760", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41760", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41760", "L"): "L .. liters",
        ("FIX.5.0", "41760", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41760", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41760", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41760", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41760", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41760", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41760", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41760", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41760", "t"): "t .. MetricTons",
        ("FIX.5.0", "41760", "tn"): "tn .. Tons",
        ("FIX.5.0", "41761", "0"): "0 .. Term",
        ("FIX.5.0", "41761", "1"): "1 .. PerBusinessDay",
        ("FIX.5.0", "41761", "2"): "2 .. PerCalculationPeriod",
        ("FIX.5.0", "41761", "3"): "3 .. PerSettlPeriod",
        ("FIX.5.0", "41761", "4"): "4 .. PerCalendarDay",
        ("FIX.5.0", "41761", "5"): "5 .. PerHour",
        ("FIX.5.0", "41761", "6"): "6 .. PerMonth",
        ("FIX.5.0", "41764", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41764", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41764", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41764", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41764", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41764", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41764", "thm"): "thm .. therms",
        ("FIX.5.0", "41764", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41764", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41764", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41764", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41764", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41764", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41764", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41764", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41764", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41764", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41764", "day"): "day .. Day",
        ("FIX.5.0", "41764", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41764", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41764", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41764", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41764", "g"): "g .. Grams",
        ("FIX.5.0", "41764", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41764", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41764", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41764", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41764", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41764", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41764", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41764", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41764", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41764", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41764", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41764", "L"): "L .. liters",
        ("FIX.5.0", "41764", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41764", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41764", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41764", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41764", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41764", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41764", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41764", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41764", "t"): "t .. MetricTons",
        ("FIX.5.0", "41764", "tn"): "tn .. Tons",
        ("FIX.5.0", "41765", "0"): "0 .. Absolute",
        ("FIX.5.0", "41765", "1"): "1 .. Percentage",
        ("FIX.5.0", "41768", "0"): "0 .. AllTimes",
        ("FIX.5.0", "41768", "1"): "1 .. OnPeak",
        ("FIX.5.0", "41768", "2"): "2 .. OffPeak",
        ("FIX.5.0", "41768", "3"): "3 .. Base",
        ("FIX.5.0", "41768", "4"): "4 .. BlockHours",
        ("FIX.5.0", "41768", "5"): "5 .. Other",
        ("FIX.5.0", "41769", "0"): "0 .. DoNotIncludeHolidays",
        ("FIX.5.0", "41769", "1"): "1 .. IncludeHolidays",
        ("FIX.5.0", "41771", "1"): "1 .. Monday",
        ("FIX.5.0", "41771", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41771", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41771", "4"): "4 .. Thursday",
        ("FIX.5.0", "41771", "5"): "5 .. Friday",
        ("FIX.5.0", "41771", "6"): "6 .. Saturday",
        ("FIX.5.0", "41771", "7"): "7 .. Sunday",
        ("FIX.5.0", "41771", "8"): "8 .. AllWeekdays",
        ("FIX.5.0", "41771", "9"): "9 .. AllDays",
        ("FIX.5.0", "41771", "10"): "10 .. AllWeekends",
        ("FIX.5.0", "41776", "0"): "0 .. Hour",
        ("FIX.5.0", "41776", "1"): "1 .. Timestamp",
        ("FIX.5.0", "41777", "0"): "0 .. Periodic",
        ("FIX.5.0", "41777", "1"): "1 .. Initial",
        ("FIX.5.0", "41777", "2"): "2 .. Single",
        ("FIX.5.0", "41782", "1"): "1 .. Firm",
        ("FIX.5.0", "41782", "2"): "2 .. NonFirm",
        ("FIX.5.0", "41782", "3"): "3 .. ForceMajeure",
        ("FIX.5.0", "41782", "4"): "4 .. SystemFirm",
        ("FIX.5.0", "41782", "5"): "5 .. UnitFirm",
        ("FIX.5.0", "41784", "0"): "0 .. Buyer",
        ("FIX.5.0", "41784", "1"): "1 .. Seller",
        ("FIX.5.0", "41788", "0"): "0 .. Transfers",
        ("FIX.5.0", "41788", "1"): "1 .. DoesNotTransfer",
        ("FIX.5.0", "41792", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41792", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41792", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41792", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41792", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41792", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41792", "thm"): "thm .. therms",
        ("FIX.5.0", "41792", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41792", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41792", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41792", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41792", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41792", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41792", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41792", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41792", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41792", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41792", "day"): "day .. Day",
        ("FIX.5.0", "41792", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41792", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41792", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41792", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41792", "g"): "g .. Grams",
        ("FIX.5.0", "41792", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41792", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41792", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41792", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41792", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41792", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41792", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41792", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41792", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41792", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41792", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41792", "L"): "L .. liters",
        ("FIX.5.0", "41792", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41792", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41792", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41792", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41792", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41792", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41792", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41792", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41792", "t"): "t .. MetricTons",
        ("FIX.5.0", "41792", "tn"): "tn .. Tons",
        ("FIX.5.0", "41793", "0"): "0 .. Absolute",
        ("FIX.5.0", "41793", "1"): "1 .. Percentage",
        ("FIX.5.0", "41794", "1"): "1 .. Buyer",
        ("FIX.5.0", "41794", "2"): "2 .. Seller",
        ("FIX.5.0", "41799", "0"): "0 .. Buyer",
        ("FIX.5.0", "41799", "1"): "1 .. Seller",
        ("FIX.5.0", "41815", "0"): "0 .. NotRequired",
        ("FIX.5.0", "41815", "1"): "1 .. NonElectronic",
        ("FIX.5.0", "41815", "2"): "2 .. Electronic",
        ("FIX.5.0", "41815", "3"): "3 .. Unknown",
        ("FIX.5.0", "41822", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41822", "1"): "1 .. None",
        ("FIX.5.0", "41822", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41822", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41822", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41822", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41822", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41822", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41823", "0"): "0 .. Business",
        ("FIX.5.0", "41823", "1"): "1 .. Calendar",
        ("FIX.5.0", "41823", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41823", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41823", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41823", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41825", "D"): "D .. Day",
        ("FIX.5.0", "41825", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41825", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41825", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41827", "D"): "D .. Day",
        ("FIX.5.0", "41827", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41827", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41827", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41831", "D"): "D .. Day",
        ("FIX.5.0", "41831", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41831", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41831", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41832", "0"): "0 .. Business",
        ("FIX.5.0", "41832", "1"): "1 .. Calendar",
        ("FIX.5.0", "41832", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41832", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41832", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41832", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41843", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41843", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41846", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41846", "1"): "1 .. None",
        ("FIX.5.0", "41846", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41846", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41846", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41846", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41846", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41846", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41849", "D"): "D .. Day",
        ("FIX.5.0", "41849", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41849", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41849", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41851", "D"): "D .. Day",
        ("FIX.5.0", "41851", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41851", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41851", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41852", "1"): "1 .. FirstDay",
        ("FIX.5.0", "41852", "2"): "2 .. SecondDay",
        ("FIX.5.0", "41852", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "41852", "4"): "4 .. FourthDay",
        ("FIX.5.0", "41852", "5"): "5 .. FifthDay",
        ("FIX.5.0", "41852", "6"): "6 .. SixthDay",
        ("FIX.5.0", "41852", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "41852", "8"): "8 .. EighthDay",
        ("FIX.5.0", "41852", "9"): "9 .. NinthDay",
        ("FIX.5.0", "41852", "10"): "10 .. TenthDay",
        ("FIX.5.0", "41852", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "41852", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "41852", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "41852", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "41852", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "41852", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "41852", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "41852", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "41852", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "41852", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "41852", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "41852", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "41852", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "41852", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "41852", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "41852", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "41852", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "41852", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "41852", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "41852", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "41852", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "41852", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "41852", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "41852", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "41852", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "41852", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "41852", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "41852", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "41852", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "41852", "MON"): "MON .. MON",
        ("FIX.5.0", "41852", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "41852", "WED"): "WED .. WED",
        ("FIX.5.0", "41852", "THU"): "THU .. THU",
        ("FIX.5.0", "41852", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "41852", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "41852", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "41853", "0"): "0 .. Business",
        ("FIX.5.0", "41853", "1"): "1 .. Calendar",
        ("FIX.5.0", "41853", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41853", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41853", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41853", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41858", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41858", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41859", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41859", "1"): "1 .. Applicable",
        ("FIX.5.0", "41859", "2"): "2 .. AsInMasterAgreement",
        ("FIX.5.0", "41859", "3"): "3 .. AsInConfirmation",
        ("FIX.5.0", "41860", "0"): "0 .. MasterAgreement",
        ("FIX.5.0", "41860", "1"): "1 .. Confirmation",
        ("FIX.5.0", "41869", "0"): "0 .. Basket",
        ("FIX.5.0", "41869", "1"): "1 .. Bond",
        ("FIX.5.0", "41869", "2"): "2 .. Cash",
        ("FIX.5.0", "41869", "3"): "3 .. Commodity",
        ("FIX.5.0", "41869", "4"): "4 .. ConvertibleBond",
        ("FIX.5.0", "41869", "5"): "5 .. Equity",
        ("FIX.5.0", "41869", "6"): "6 .. ExchangeTradedFund",
        ("FIX.5.0", "41869", "7"): "7 .. Future",
        ("FIX.5.0", "41869", "8"): "8 .. Index",
        ("FIX.5.0", "41869", "9"): "9 .. Loan",
        ("FIX.5.0", "41869", "10"): "10 .. Mortgage",
        ("FIX.5.0", "41869", "11"): "11 .. MutualFund",
        ("FIX.5.0", "41871", "1"): "1 .. CUSIP",
        ("FIX.5.0", "41871", "2"): "2 .. SEDOL",
        ("FIX.5.0", "41871", "3"): "3 .. QUIK",
        ("FIX.5.0", "41871", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "41871", "5"): "5 .. RICCode",
        ("FIX.5.0", "41871", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "41871", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "41871", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "41871", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "41871", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "41871", "B"): "B .. Wertpapier",
        ("FIX.5.0", "41871", "C"): "C .. Dutch",
        ("FIX.5.0", "41871", "D"): "D .. Valoren",
        ("FIX.5.0", "41871", "E"): "E .. Sicovam",
        ("FIX.5.0", "41871", "F"): "F .. Belgian",
        ("FIX.5.0", "41871", "G"): "G .. Common",
        ("FIX.5.0", "41871", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "41871", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "41871", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "41871", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "41871", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "41871", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "41871", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "41871", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "41871", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "41871", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "41871", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "41871", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "41871", "U"): "U .. Synthetic",
        ("FIX.5.0", "41871", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "41871", "W"): "W .. IndexName",
        ("FIX.5.0", "41879", "0"): "0 .. EveryDay",
        ("FIX.5.0", "41879", "1"): "1 .. Monday",
        ("FIX.5.0", "41879", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41879", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41879", "4"): "4 .. Thursday",
        ("FIX.5.0", "41879", "5"): "5 .. Friday",
        ("FIX.5.0", "41879", "6"): "6 .. Saturday",
        ("FIX.5.0", "41879", "7"): "7 .. Sunday",
        ("FIX.5.0", "41884", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41884", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41884", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41884", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41884", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41884", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41884", "thm"): "thm .. therms",
        ("FIX.5.0", "41884", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41884", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41884", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41884", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41884", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41884", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41884", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41884", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41884", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41884", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41884", "day"): "day .. Day",
        ("FIX.5.0", "41884", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41884", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41884", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41884", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41884", "g"): "g .. Grams",
        ("FIX.5.0", "41884", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41884", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41884", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41884", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41884", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41884", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41884", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41884", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41884", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41884", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41884", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41884", "L"): "L .. liters",
        ("FIX.5.0", "41884", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41884", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41884", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41884", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41884", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41884", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41884", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41884", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41884", "t"): "t .. MetricTons",
        ("FIX.5.0", "41884", "tn"): "tn .. Tons",
        ("FIX.5.0", "41886", "0"): "0 .. Absolute",
        ("FIX.5.0", "41886", "1"): "1 .. Percentage",
        ("FIX.5.0", "41889", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41889", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41889", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41889", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41889", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41889", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41889", "thm"): "thm .. therms",
        ("FIX.5.0", "41889", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41889", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41889", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41889", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41889", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41889", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41889", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41889", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41889", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41889", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41889", "day"): "day .. Day",
        ("FIX.5.0", "41889", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41889", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41889", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41889", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41889", "g"): "g .. Grams",
        ("FIX.5.0", "41889", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41889", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41889", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41889", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41889", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41889", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41889", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41889", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41889", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41889", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41889", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41889", "L"): "L .. liters",
        ("FIX.5.0", "41889", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41889", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41889", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41889", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41889", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41889", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41889", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41889", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41889", "t"): "t .. MetricTons",
        ("FIX.5.0", "41889", "tn"): "tn .. Tons",
        ("FIX.5.0", "41890", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41890", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41890", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41890", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41890", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41890", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41890", "thm"): "thm .. therms",
        ("FIX.5.0", "41890", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41890", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41890", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41890", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41890", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41890", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41890", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41890", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41890", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41890", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41890", "day"): "day .. Day",
        ("FIX.5.0", "41890", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41890", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41890", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41890", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41890", "g"): "g .. Grams",
        ("FIX.5.0", "41890", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41890", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41890", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41890", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41890", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41890", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41890", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41890", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41890", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41890", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41890", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41890", "L"): "L .. liters",
        ("FIX.5.0", "41890", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41890", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41890", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41890", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41890", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41890", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41890", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41890", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41890", "t"): "t .. MetricTons",
        ("FIX.5.0", "41890", "tn"): "tn .. Tons",
        ("FIX.5.0", "41891", "0"): "0 .. All",
        ("FIX.5.0", "41891", "1"): "1 .. First",
        ("FIX.5.0", "41891", "2"): "2 .. Last",
        ("FIX.5.0", "41891", "3"): "3 .. Penultimate",
        ("FIX.5.0", "41894", "D"): "D .. Day",
        ("FIX.5.0", "41894", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41894", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41894", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41896", "D"): "D .. Day",
        ("FIX.5.0", "41896", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41896", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41896", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41904", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41904", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41904", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41904", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41904", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41904", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41904", "thm"): "thm .. therms",
        ("FIX.5.0", "41904", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41904", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41904", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41904", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41904", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41904", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41904", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41904", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41904", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41904", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41904", "day"): "day .. Day",
        ("FIX.5.0", "41904", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41904", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41904", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41904", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41904", "g"): "g .. Grams",
        ("FIX.5.0", "41904", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41904", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41904", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41904", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41904", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41904", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41904", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41904", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41904", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41904", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41904", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41904", "L"): "L .. liters",
        ("FIX.5.0", "41904", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41904", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41904", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41904", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41904", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41904", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41904", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41904", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41904", "t"): "t .. MetricTons",
        ("FIX.5.0", "41904", "tn"): "tn .. Tons",
        ("FIX.5.0", "41911", "D"): "D .. Day",
        ("FIX.5.0", "41911", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41911", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41911", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41915", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41915", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41915", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41915", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41915", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41915", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41915", "thm"): "thm .. therms",
        ("FIX.5.0", "41915", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41915", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41915", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41915", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41915", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41915", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41915", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41915", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41915", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41915", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41915", "day"): "day .. Day",
        ("FIX.5.0", "41915", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41915", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41915", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41915", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41915", "g"): "g .. Grams",
        ("FIX.5.0", "41915", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41915", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41915", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41915", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41915", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41915", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41915", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41915", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41915", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41915", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41915", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41915", "L"): "L .. liters",
        ("FIX.5.0", "41915", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41915", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41915", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41915", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41915", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41915", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41915", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41915", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41915", "t"): "t .. MetricTons",
        ("FIX.5.0", "41915", "tn"): "tn .. Tons",
        ("FIX.5.0", "41916", "0"): "0 .. Average",
        ("FIX.5.0", "41916", "1"): "1 .. Maximum",
        ("FIX.5.0", "41916", "2"): "2 .. Minimum",
        ("FIX.5.0", "41916", "3"): "3 .. Cumulative",
        ("FIX.5.0", "41918", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41918", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41918", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41918", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41918", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41918", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41918", "thm"): "thm .. therms",
        ("FIX.5.0", "41918", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41918", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41918", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41918", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41918", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41918", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41918", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41918", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41918", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41918", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41918", "day"): "day .. Day",
        ("FIX.5.0", "41918", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41918", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41918", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41918", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41918", "g"): "g .. Grams",
        ("FIX.5.0", "41918", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41918", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41918", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41918", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41918", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41918", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41918", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41918", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41918", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41918", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41918", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41918", "L"): "L .. liters",
        ("FIX.5.0", "41918", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41918", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41918", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41918", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41918", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41918", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41918", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41918", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41918", "t"): "t .. MetricTons",
        ("FIX.5.0", "41918", "tn"): "tn .. Tons",
        ("FIX.5.0", "41921", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41921", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41921", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41921", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41921", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41921", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41921", "thm"): "thm .. therms",
        ("FIX.5.0", "41921", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41921", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41921", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41921", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41921", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41921", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41921", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41921", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41921", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41921", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41921", "day"): "day .. Day",
        ("FIX.5.0", "41921", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41921", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41921", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41921", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41921", "g"): "g .. Grams",
        ("FIX.5.0", "41921", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41921", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41921", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41921", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41921", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41921", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41921", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41921", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41921", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41921", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41921", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41921", "L"): "L .. liters",
        ("FIX.5.0", "41921", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41921", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41921", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41921", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41921", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41921", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41921", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41921", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41921", "t"): "t .. MetricTons",
        ("FIX.5.0", "41921", "tn"): "tn .. Tons",
        ("FIX.5.0", "41923", "0"): "0 .. Absolute",
        ("FIX.5.0", "41923", "1"): "1 .. Percentage",
        ("FIX.5.0", "41927", "D"): "D .. Day",
        ("FIX.5.0", "41927", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41927", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41927", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41929", "D"): "D .. Day",
        ("FIX.5.0", "41929", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41929", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41929", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41930", "0"): "0 .. Business",
        ("FIX.5.0", "41930", "1"): "1 .. Calendar",
        ("FIX.5.0", "41930", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41930", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41930", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41930", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41931", "0"): "0 .. All",
        ("FIX.5.0", "41931", "1"): "1 .. First",
        ("FIX.5.0", "41931", "2"): "2 .. Last",
        ("FIX.5.0", "41931", "3"): "3 .. Penultimate",
        ("FIX.5.0", "41934", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41934", "1"): "1 .. None",
        ("FIX.5.0", "41934", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41934", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41934", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41934", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41934", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41934", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41935", "0"): "0 .. Hour",
        ("FIX.5.0", "41935", "1"): "1 .. Timestamp",
        ("FIX.5.0", "41936", "0"): "0 .. Hour",
        ("FIX.5.0", "41936", "1"): "1 .. Timestamp",
        ("FIX.5.0", "41939", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41939", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41943", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41943", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41945", "0"): "0 .. EveryDay",
        ("FIX.5.0", "41945", "1"): "1 .. Monday",
        ("FIX.5.0", "41945", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41945", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41945", "4"): "4 .. Thursday",
        ("FIX.5.0", "41945", "5"): "5 .. Friday",
        ("FIX.5.0", "41945", "6"): "6 .. Saturday",
        ("FIX.5.0", "41945", "7"): "7 .. Sunday",
        ("FIX.5.0", "41950", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41950", "1"): "1 .. None",
        ("FIX.5.0", "41950", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41950", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41950", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41950", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41950", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41950", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41956", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "41956", "1"): "1 .. Adjusted",
        ("FIX.5.0", "41961", "D"): "D .. Day",
        ("FIX.5.0", "41961", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41961", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41961", "Yr"): "Yr .. Year",
        ("FIX.5.0", "41967", "1"): "1 .. CUSIP",
        ("FIX.5.0", "41967", "2"): "2 .. SEDOL",
        ("FIX.5.0", "41967", "3"): "3 .. QUIK",
        ("FIX.5.0", "41967", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "41967", "5"): "5 .. RICCode",
        ("FIX.5.0", "41967", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "41967", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "41967", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "41967", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "41967", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "41967", "B"): "B .. Wertpapier",
        ("FIX.5.0", "41967", "C"): "C .. Dutch",
        ("FIX.5.0", "41967", "D"): "D .. Valoren",
        ("FIX.5.0", "41967", "E"): "E .. Sicovam",
        ("FIX.5.0", "41967", "F"): "F .. Belgian",
        ("FIX.5.0", "41967", "G"): "G .. Common",
        ("FIX.5.0", "41967", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "41967", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "41967", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "41967", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "41967", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "41967", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "41967", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "41967", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "41967", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "41967", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "41967", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "41967", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "41967", "U"): "U .. Synthetic",
        ("FIX.5.0", "41967", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "41967", "W"): "W .. IndexName",
        ("FIX.5.0", "41971", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "41971", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "41971", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "41971", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "41971", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "41971", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "41971", "thm"): "thm .. therms",
        ("FIX.5.0", "41971", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "41971", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "41971", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "41971", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "41971", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "41971", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "41971", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "41971", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "41971", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "41971", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "41971", "day"): "day .. Day",
        ("FIX.5.0", "41971", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "41971", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "41971", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "41971", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "41971", "g"): "g .. Grams",
        ("FIX.5.0", "41971", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "41971", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "41971", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "41971", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "41971", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "41971", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "41971", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "41971", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "41971", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "41971", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "41971", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "41971", "L"): "L .. liters",
        ("FIX.5.0", "41971", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "41971", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "41971", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "41971", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "41971", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "41971", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "41971", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "41971", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "41971", "t"): "t .. MetricTons",
        ("FIX.5.0", "41971", "tn"): "tn .. Tons",
        ("FIX.5.0", "41980", "Wk"): "Wk .. Week",
        ("FIX.5.0", "41980", "Mo"): "Mo .. Month",
        ("FIX.5.0", "41982", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "41982", "1"): "1 .. None",
        ("FIX.5.0", "41982", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "41982", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "41982", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "41982", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "41982", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "41982", "7"): "7 .. NearestDay",
        ("FIX.5.0", "41986", "D"): "D .. Day",
        ("FIX.5.0", "41987", "0"): "0 .. Business",
        ("FIX.5.0", "41987", "1"): "1 .. Calendar",
        ("FIX.5.0", "41987", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "41987", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "41987", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "41987", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "41995", "0"): "0 .. City",
        ("FIX.5.0", "41995", "1"): "1 .. Airport",
        ("FIX.5.0", "41995", "2"): "2 .. WeatherStation",
        ("FIX.5.0", "41995", "3"): "3 .. WeatherIndex",
        ("FIX.5.0", "41997", "1"): "1 .. Monday",
        ("FIX.5.0", "41997", "2"): "2 .. Tuesday",
        ("FIX.5.0", "41997", "3"): "3 .. Wednesday",
        ("FIX.5.0", "41997", "4"): "4 .. Thursday",
        ("FIX.5.0", "41997", "5"): "5 .. Friday",
        ("FIX.5.0", "41997", "6"): "6 .. Saturday",
        ("FIX.5.0", "41997", "7"): "7 .. Sunday",
        ("FIX.5.0", "41997", "8"): "8 .. AllWeekdays",
        ("FIX.5.0", "41997", "9"): "9 .. AllDays",
        ("FIX.5.0", "41997", "10"): "10 .. AllWeekends",
        ("FIX.5.0", "42005", "0"): "0 .. AllTimes",
        ("FIX.5.0", "42005", "1"): "1 .. OnPeak",
        ("FIX.5.0", "42005", "2"): "2 .. OffPeak",
        ("FIX.5.0", "42005", "3"): "3 .. Base",
        ("FIX.5.0", "42005", "4"): "4 .. BlockHours",
        ("FIX.5.0", "42005", "5"): "5 .. Other",
        ("FIX.5.0", "42007", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "42007", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "42007", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "42007", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "42007", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "42007", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "42007", "thm"): "thm .. therms",
        ("FIX.5.0", "42007", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "42007", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "42007", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "42007", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "42007", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "42007", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "42007", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "42007", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "42007", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "42007", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "42007", "day"): "day .. Day",
        ("FIX.5.0", "42007", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "42007", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "42007", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "42007", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "42007", "g"): "g .. Grams",
        ("FIX.5.0", "42007", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "42007", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "42007", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "42007", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "42007", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "42007", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "42007", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "42007", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "42007", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "42007", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "42007", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "42007", "L"): "L .. liters",
        ("FIX.5.0", "42007", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "42007", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "42007", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "42007", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "42007", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "42007", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "42007", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "42007", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "42007", "t"): "t .. MetricTons",
        ("FIX.5.0", "42007", "tn"): "tn .. Tons",
        ("FIX.5.0", "42009", "D"): "D .. Day",
        ("FIX.5.0", "42009", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42009", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42009", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42011", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "42011", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "42011", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "42011", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "42011", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "42011", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "42011", "thm"): "thm .. therms",
        ("FIX.5.0", "42011", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "42011", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "42011", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "42011", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "42011", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "42011", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "42011", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "42011", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "42011", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "42011", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "42011", "day"): "day .. Day",
        ("FIX.5.0", "42011", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "42011", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "42011", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "42011", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "42011", "g"): "g .. Grams",
        ("FIX.5.0", "42011", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "42011", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "42011", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "42011", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "42011", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "42011", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "42011", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "42011", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "42011", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "42011", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "42011", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "42011", "L"): "L .. liters",
        ("FIX.5.0", "42011", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "42011", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "42011", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "42011", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "42011", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "42011", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "42011", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "42011", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "42011", "t"): "t .. MetricTons",
        ("FIX.5.0", "42011", "tn"): "tn .. Tons",
        ("FIX.5.0", "42013", "0"): "0 .. DoNotIncludeHolidays",
        ("FIX.5.0", "42013", "1"): "1 .. IncludeHolidays",
        ("FIX.5.0", "42020", "H"): "H .. Hour",
        ("FIX.5.0", "42020", "Min"): "Min .. Minute",
        ("FIX.5.0", "42020", "S"): "S .. Second",
        ("FIX.5.0", "42020", "D"): "D .. Day",
        ("FIX.5.0", "42020", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42020", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42020", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42020", "Q"): "Q .. Quarter",
        ("FIX.5.0", "42021", "0"): "0 .. Term",
        ("FIX.5.0", "42021", "1"): "1 .. PerBusinessDay",
        ("FIX.5.0", "42021", "2"): "2 .. PerCalculationPeriod",
        ("FIX.5.0", "42021", "3"): "3 .. PerSettlPeriod",
        ("FIX.5.0", "42021", "4"): "4 .. PerCalendarDay",
        ("FIX.5.0", "42021", "5"): "5 .. PerHour",
        ("FIX.5.0", "42021", "6"): "6 .. PerMonth",
        ("FIX.5.0", "42022", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "42022", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "42022", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "42022", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "42022", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "42022", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "42022", "thm"): "thm .. therms",
        ("FIX.5.0", "42022", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "42022", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "42022", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "42022", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "42022", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "42022", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "42022", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "42022", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "42022", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "42022", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "42022", "day"): "day .. Day",
        ("FIX.5.0", "42022", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "42022", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "42022", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "42022", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "42022", "g"): "g .. Grams",
        ("FIX.5.0", "42022", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "42022", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "42022", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "42022", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "42022", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "42022", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "42022", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "42022", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "42022", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "42022", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "42022", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "42022", "L"): "L .. liters",
        ("FIX.5.0", "42022", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "42022", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "42022", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "42022", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "42022", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "42022", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "42022", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "42022", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "42022", "t"): "t .. MetricTons",
        ("FIX.5.0", "42022", "tn"): "tn .. Tons",
        ("FIX.5.0", "42024", "Bcf"): "Bcf .. BillionCubicFeet",
        ("FIX.5.0", "42024", "CBM"): "CBM .. CubicMeters",
        ("FIX.5.0", "42024", "GJ"): "GJ .. gigajoules",
        ("FIX.5.0", "42024", "kWh"): "kWh .. KilowattHours",
        ("FIX.5.0", "42024", "MMBtu"): "MMBtu .. OneMillionBTU",
        ("FIX.5.0", "42024", "MWh"): "MWh .. MegawattHours",
        ("FIX.5.0", "42024", "thm"): "thm .. therms",
        ("FIX.5.0", "42024", "Alw"): "Alw .. Allowances",
        ("FIX.5.0", "42024", "Bbl"): "Bbl .. Barrels",
        ("FIX.5.0", "42024", "BDFT"): "BDFT .. BoardFeet",
        ("FIX.5.0", "42024", "Bu"): "Bu .. Bushels",
        ("FIX.5.0", "42024", "Ccy"): "Ccy .. Currency",
        ("FIX.5.0", "42024", "CDD"): "CDD .. CoolingDegreeDay",
        ("FIX.5.0", "42024", "CER"): "CER .. CertifiedEmissionsReduction",
        ("FIX.5.0", "42024", "CPD"): "CPD .. CriticalPrecipDay",
        ("FIX.5.0", "42024", "CRT"): "CRT .. ClimateReserveTonnes",
        ("FIX.5.0", "42024", "cwt"): "cwt .. Hundredweight",
        ("FIX.5.0", "42024", "day"): "day .. Day",
        ("FIX.5.0", "42024", "dt"): "dt .. DryMetricTons",
        ("FIX.5.0", "42024", "EnvAllwnc"): "EnvAllwnc .. EnvAllwncCert",
        ("FIX.5.0", "42024", "EnvCrd"): "EnvCrd .. EnvironmentalCredit",
        ("FIX.5.0", "42024", "EnvOfst"): "EnvOfst .. EnvironmentalOffset",
        ("FIX.5.0", "42024", "g"): "g .. Grams",
        ("FIX.5.0", "42024", "Gal"): "Gal .. Gallons",
        ("FIX.5.0", "42024", "GT"): "GT .. GrossTons",
        ("FIX.5.0", "42024", "HDD"): "HDD .. HeatingDegreeDay",
        ("FIX.5.0", "42024", "IPNT"): "IPNT .. IndexPoint",
        ("FIX.5.0", "42024", "kg"): "kg .. Kilograms",
        ("FIX.5.0", "42024", "kL"): "kL .. kiloliters",
        ("FIX.5.0", "42024", "kW-a"): "kW-a .. KilowattYear",
        ("FIX.5.0", "42024", "kW-d"): "kW-d .. KilowattDay",
        ("FIX.5.0", "42024", "kW-h"): "kW-h .. KilowattHour",
        ("FIX.5.0", "42024", "kW-M"): "kW-M .. KilowattMonth",
        ("FIX.5.0", "42024", "kW-min "): "kW-min  .. KilowattMinute",
        ("FIX.5.0", "42024", "L"): "L .. liters",
        ("FIX.5.0", "42024", "lbs"): "lbs .. Pounds",
        ("FIX.5.0", "42024", "MW-a"): "MW-a .. MegawattYear",
        ("FIX.5.0", "42024", "MW-d"): "MW-d .. MegawattDay",
        ("FIX.5.0", "42024", "MW-h"): "MW-h .. MegawattHour",
        ("FIX.5.0", "42024", "MW-M"): "MW-M .. MegawattMonth",
        ("FIX.5.0", "42024", "MW-min"): "MW-min .. MegawattMinute",
        ("FIX.5.0", "42024", "oz_tr"): "oz_tr .. TroyOunces",
        ("FIX.5.0", "42024", "PRINC"): "PRINC .. PrincipalWithRelationToDebtInstrument",
        ("FIX.5.0", "42024", "t"): "t .. MetricTons",
        ("FIX.5.0", "42024", "tn"): "tn .. Tons",
        ("FIX.5.0", "42027", "SD"): "SD .. SeniorSecured",
        ("FIX.5.0", "42027", "SR"): "SR .. Senior",
        ("FIX.5.0", "42027", "SB"): "SB .. Subordinated",
        ("FIX.5.0", "42027", "JR"): "JR .. Junior",
        ("FIX.5.0", "42027", "MZ"): "MZ .. Mezzanine",
        ("FIX.5.0", "42027", "SN"): "SN .. SeniorNonPreferred",
        ("FIX.5.0", "42028", "0"): "0 .. Zero",
        ("FIX.5.0", "42028", "1"): "1 .. FixedRate",
        ("FIX.5.0", "42028", "2"): "2 .. FloatingRate",
        ("FIX.5.0", "42028", "3"): "3 .. Structured",
        ("FIX.5.0", "42034", "D"): "D .. Day",
        ("FIX.5.0", "42034", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42034", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42034", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42034", "H"): "H .. Hour",
        ("FIX.5.0", "42034", "Min"): "Min .. Minute",
        ("FIX.5.0", "42034", "S"): "S .. Second",
        ("FIX.5.0", "42034", "T"): "T .. Term",
        ("FIX.5.0", "42035", "0"): "0 .. OneOne",
        ("FIX.5.0", "42035", "1"): "1 .. ThrityThreeSixtyUS",
        ("FIX.5.0", "42035", "2"): "2 .. ThirtyThreeSixtySIA",
        ("FIX.5.0", "42035", "3"): "3 .. ThirtyThreeSixtyM",
        ("FIX.5.0", "42035", "4"): "4 .. ThirtyEThreeSixty",
        ("FIX.5.0", "42035", "5"): "5 .. ThirtyEThreeSixtyISDA",
        ("FIX.5.0", "42035", "6"): "6 .. ActThreeSixty",
        ("FIX.5.0", "42035", "7"): "7 .. ActThreeSixtyFiveFixed",
        ("FIX.5.0", "42035", "8"): "8 .. ActActAFB",
        ("FIX.5.0", "42035", "9"): "9 .. ActActICMA",
        ("FIX.5.0", "42035", "10"): "10 .. ActActISMAUltimo",
        ("FIX.5.0", "42035", "11"): "11 .. ActActISDA",
        ("FIX.5.0", "42035", "12"): "12 .. BusTwoFiftyTwo",
        ("FIX.5.0", "42035", "13"): "13 .. ThirtyEPlusThreeSixty",
        ("FIX.5.0", "42035", "14"): "14 .. ActThreeSixtyFiveL",
        ("FIX.5.0", "42035", "15"): "15 .. NLThreeSixtyFive",
        ("FIX.5.0", "42035", "16"): "16 .. NLThreeSixty",
        ("FIX.5.0", "42035", "17"): "17 .. Act364",
        ("FIX.5.0", "42048", "0"): "0 .. Bid",
        ("FIX.5.0", "42048", "1"): "1 .. Mid",
        ("FIX.5.0", "42048", "2"): "2 .. Offer",
        ("FIX.5.0", "42058", "0"): "0 .. Market",
        ("FIX.5.0", "42058", "1"): "1 .. Highest",
        ("FIX.5.0", "42058", "2"): "2 .. AverageMarket",
        ("FIX.5.0", "42058", "3"): "3 .. AverageHighest",
        ("FIX.5.0", "42058", "4"): "4 .. BlendedMarket",
        ("FIX.5.0", "42058", "5"): "5 .. BlendedHighest",
        ("FIX.5.0", "42058", "6"): "6 .. AverageBlendedMarket",
        ("FIX.5.0", "42058", "7"): "7 .. AverageBlendedHighest",
        ("FIX.5.0", "42082", "D"): "D .. Day",
        ("FIX.5.0", "42082", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42082", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42082", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42083", "0"): "0 .. Business",
        ("FIX.5.0", "42083", "1"): "1 .. Calendar",
        ("FIX.5.0", "42083", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42083", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42083", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42083", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42086", "H"): "H .. RestructuringMultipleHoldingObligations",
        ("FIX.5.0", "42086", "E"): "E .. RestructuringMultipleCreditEventNotices",
        ("FIX.5.0", "42086", "C"): "C .. FloatingRateInterestShortfall",
        ("FIX.5.0", "42092", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42092", "1"): "1 .. None",
        ("FIX.5.0", "42092", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42092", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42092", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42092", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42092", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42092", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42095", "D"): "D .. Day",
        ("FIX.5.0", "42095", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42095", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42095", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42096", "0"): "0 .. Business",
        ("FIX.5.0", "42096", "1"): "1 .. Calendar",
        ("FIX.5.0", "42096", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42096", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42096", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42096", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42101", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42101", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42102", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "42102", "1"): "1 .. Reuters",
        ("FIX.5.0", "42102", "2"): "2 .. Telerate",
        ("FIX.5.0", "42102", "99"): "99 .. Other",
        ("FIX.5.0", "42106", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42106", "1"): "1 .. None",
        ("FIX.5.0", "42106", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42106", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42106", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42106", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42106", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42106", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42109", "D"): "D .. Day",
        ("FIX.5.0", "42109", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42109", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42109", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42110", "0"): "0 .. Business",
        ("FIX.5.0", "42110", "1"): "1 .. Calendar",
        ("FIX.5.0", "42110", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42110", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42110", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42110", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42114", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42114", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42115", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42115", "1"): "1 .. None",
        ("FIX.5.0", "42115", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42115", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42115", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42115", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42115", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42115", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42117", "D"): "D .. Day",
        ("FIX.5.0", "42117", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42117", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42117", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42119", "D"): "D .. Day",
        ("FIX.5.0", "42119", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42119", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42119", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42119", "H"): "H .. Hour",
        ("FIX.5.0", "42119", "Min"): "Min .. Minute",
        ("FIX.5.0", "42119", "S"): "S .. Second",
        ("FIX.5.0", "42119", "T"): "T .. Term",
        ("FIX.5.0", "42123", "D"): "D .. Day",
        ("FIX.5.0", "42123", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42123", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42123", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42124", "0"): "0 .. Business",
        ("FIX.5.0", "42124", "1"): "1 .. Calendar",
        ("FIX.5.0", "42124", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42124", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42124", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42124", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42134", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42134", "1"): "1 .. None",
        ("FIX.5.0", "42134", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42134", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42134", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42134", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42134", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42134", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42137", "D"): "D .. Day",
        ("FIX.5.0", "42137", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42137", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42137", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42138", "0"): "0 .. Business",
        ("FIX.5.0", "42138", "1"): "1 .. Calendar",
        ("FIX.5.0", "42138", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42138", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42138", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42138", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42143", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42143", "1"): "1 .. None",
        ("FIX.5.0", "42143", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42143", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42143", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42143", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42143", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42143", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42146", "D"): "D .. Day",
        ("FIX.5.0", "42146", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42146", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42146", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42147", "0"): "0 .. Business",
        ("FIX.5.0", "42147", "1"): "1 .. Calendar",
        ("FIX.5.0", "42147", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42147", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42147", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42147", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42150", "0"): "0 .. MandatoryEarlyTermination",
        ("FIX.5.0", "42150", "1"): "1 .. OptionalEarlyTermination",
        ("FIX.5.0", "42150", "2"): "2 .. Cancelable",
        ("FIX.5.0", "42150", "3"): "3 .. Extendable",
        ("FIX.5.0", "42150", "4"): "4 .. MutualEarlyTermination",
        ("FIX.5.0", "42152", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42152", "1"): "1 .. None",
        ("FIX.5.0", "42152", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42152", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42152", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42152", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42152", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42152", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42155", "D"): "D .. Day",
        ("FIX.5.0", "42155", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42155", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42155", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42156", "0"): "0 .. ExercisingParty",
        ("FIX.5.0", "42156", "1"): "1 .. NonExercisingParty",
        ("FIX.5.0", "42156", "2"): "2 .. MasterAgreeent",
        ("FIX.5.0", "42156", "3"): "3 .. Supplement",
        ("FIX.5.0", "42157", "1"): "1 .. Buy",
        ("FIX.5.0", "42157", "2"): "2 .. Sell",
        ("FIX.5.0", "42158", "1"): "1 .. Buy",
        ("FIX.5.0", "42158", "2"): "2 .. Sell",
        ("FIX.5.0", "42159", "0"): "0 .. European",
        ("FIX.5.0", "42159", "1"): "1 .. American",
        ("FIX.5.0", "42159", "2"): "2 .. Bermuda",
        ("FIX.5.0", "42159", "99"): "99 .. Other",
        ("FIX.5.0", "42166", "0"): "0 .. CashPrice",
        ("FIX.5.0", "42166", "1"): "1 .. CashPriceAlternate",
        ("FIX.5.0", "42166", "2"): "2 .. ParYieldCurveAdjusted",
        ("FIX.5.0", "42166", "3"): "3 .. ZeroCouponYieldCurveAdjusted",
        ("FIX.5.0", "42166", "4"): "4 .. ParYieldCurveUnadjusted",
        ("FIX.5.0", "42166", "5"): "5 .. CrossCurrency",
        ("FIX.5.0", "42166", "6"): "6 .. CollateralizedPrice",
        ("FIX.5.0", "42169", "0"): "0 .. Bid",
        ("FIX.5.0", "42169", "1"): "1 .. Mid",
        ("FIX.5.0", "42169", "2"): "2 .. Offer",
        ("FIX.5.0", "42169", "3"): "3 .. ExercisingPartyPays",
        ("FIX.5.0", "42175", "6"): "6 .. UKNationalInsuranceOrPensionNumber",
        ("FIX.5.0", "42175", "7"): "7 .. USSocialSecurityNumber",
        ("FIX.5.0", "42175", "8"): "8 .. USEmployerOrTaxIDNumber",
        ("FIX.5.0", "42175", "9"): "9 .. AustralianBusinessNumber",
        ("FIX.5.0", "42175", "A"): "A .. AustralianTaxFileNumber",
        ("FIX.5.0", "42175", "J"): "J .. TaxID",
        ("FIX.5.0", "42175", "1"): "1 .. KoreanInvestorID",
        ("FIX.5.0", "42175", "2"): "2 .. TaiwaneseForeignInvestorID",
        ("FIX.5.0", "42175", "3"): "3 .. TaiwaneseTradingAcct",
        ("FIX.5.0", "42175", "4"): "4 .. MalaysianCentralDepository",
        ("FIX.5.0", "42175", "5"): "5 .. ChineseInvestorID",
        ("FIX.5.0", "42175", "I"): "I .. ISITCAcronym",
        ("FIX.5.0", "42175", "B"): "B .. BIC",
        ("FIX.5.0", "42175", "C"): "C .. GeneralIdentifier",
        ("FIX.5.0", "42175", "D"): "D .. Proprietary",
        ("FIX.5.0", "42175", "E"): "E .. ISOCountryCode",
        ("FIX.5.0", "42175", "F"): "F .. SettlementEntityLocation",
        ("FIX.5.0", "42175", "G"): "G .. MIC",
        ("FIX.5.0", "42175", "H"): "H .. CSDParticipant",
        ("FIX.5.0", "42175", "K"): "K .. AustralianCompanyNumber",
        ("FIX.5.0", "42175", "L"): "L .. AustralianRegisteredBodyNumber",
        ("FIX.5.0", "42175", "M"): "M .. CFTCReportingFirmIdentifier",
        ("FIX.5.0", "42175", "N"): "N .. LegalEntityIdentifier",
        ("FIX.5.0", "42175", "O"): "O .. InterimIdentifier",
        ("FIX.5.0", "42175", "P"): "P .. ShortCodeIdentifier",
        ("FIX.5.0", "42175", "Q"): "Q .. NationalIDNaturalPerson",
        ("FIX.5.0", "42176", "1"): "1 .. ExecutingFirm",
        ("FIX.5.0", "42176", "2"): "2 .. BrokerOfCredit",
        ("FIX.5.0", "42176", "3"): "3 .. ClientID",
        ("FIX.5.0", "42176", "4"): "4 .. ClearingFirm",
        ("FIX.5.0", "42176", "5"): "5 .. InvestorID",
        ("FIX.5.0", "42176", "6"): "6 .. IntroducingFirm",
        ("FIX.5.0", "42176", "7"): "7 .. EnteringFirm",
        ("FIX.5.0", "42176", "8"): "8 .. Locate",
        ("FIX.5.0", "42176", "9"): "9 .. FundManagerClientID",
        ("FIX.5.0", "42176", "10"): "10 .. SettlementLocation",
        ("FIX.5.0", "42176", "11"): "11 .. OrderOriginationTrader",
        ("FIX.5.0", "42176", "12"): "12 .. ExecutingTrader",
        ("FIX.5.0", "42176", "13"): "13 .. OrderOriginationFirm",
        ("FIX.5.0", "42176", "15"): "15 .. CorrespondantClearingFirm",
        ("FIX.5.0", "42176", "16"): "16 .. ExecutingSystem",
        ("FIX.5.0", "42176", "17"): "17 .. ContraFirm",
        ("FIX.5.0", "42176", "18"): "18 .. ContraClearingFirm",
        ("FIX.5.0", "42176", "19"): "19 .. SponsoringFirm",
        ("FIX.5.0", "42176", "20"): "20 .. UnderlyingContraFirm",
        ("FIX.5.0", "42176", "21"): "21 .. ClearingOrganization",
        ("FIX.5.0", "42176", "22"): "22 .. Exchange",
        ("FIX.5.0", "42176", "24"): "24 .. CustomerAccount",
        ("FIX.5.0", "42176", "25"): "25 .. CorrespondentClearingOrganization",
        ("FIX.5.0", "42176", "26"): "26 .. CorrespondentBroker",
        ("FIX.5.0", "42176", "27"): "27 .. Buyer",
        ("FIX.5.0", "42176", "28"): "28 .. Custodian",
        ("FIX.5.0", "42176", "29"): "29 .. Intermediary",
        ("FIX.5.0", "42176", "30"): "30 .. Agent",
        ("FIX.5.0", "42176", "31"): "31 .. SubCustodian",
        ("FIX.5.0", "42176", "32"): "32 .. Beneficiary",
        ("FIX.5.0", "42176", "33"): "33 .. InterestedParty",
        ("FIX.5.0", "42176", "34"): "34 .. RegulatoryBody",
        ("FIX.5.0", "42176", "35"): "35 .. LiquidityProvider",
        ("FIX.5.0", "42176", "36"): "36 .. EnteringTrader",
        ("FIX.5.0", "42176", "37"): "37 .. ContraTrader",
        ("FIX.5.0", "42176", "38"): "38 .. PositionAccount",
        ("FIX.5.0", "42176", "39"): "39 .. ContraInvestorID",
        ("FIX.5.0", "42176", "40"): "40 .. TransferToFirm",
        ("FIX.5.0", "42176", "41"): "41 .. ContraPositionAccount",
        ("FIX.5.0", "42176", "42"): "42 .. ContraExchange",
        ("FIX.5.0", "42176", "43"): "43 .. InternalCarryAccount",
        ("FIX.5.0", "42176", "44"): "44 .. OrderEntryOperatorID",
        ("FIX.5.0", "42176", "45"): "45 .. SecondaryAccountNumber",
        ("FIX.5.0", "42176", "46"): "46 .. ForeignFirm",
        ("FIX.5.0", "42176", "47"): "47 .. ThirdPartyAllocationFirm",
        ("FIX.5.0", "42176", "48"): "48 .. ClaimingAccount",
        ("FIX.5.0", "42176", "49"): "49 .. AssetManager",
        ("FIX.5.0", "42176", "50"): "50 .. PledgorAccount",
        ("FIX.5.0", "42176", "51"): "51 .. PledgeeAccount",
        ("FIX.5.0", "42176", "52"): "52 .. LargeTraderReportableAccount",
        ("FIX.5.0", "42176", "53"): "53 .. TraderMnemonic",
        ("FIX.5.0", "42176", "54"): "54 .. SenderLocation",
        ("FIX.5.0", "42176", "55"): "55 .. SessionID",
        ("FIX.5.0", "42176", "56"): "56 .. AcceptableCounterparty",
        ("FIX.5.0", "42176", "57"): "57 .. UnacceptableCounterparty",
        ("FIX.5.0", "42176", "58"): "58 .. EnteringUnit",
        ("FIX.5.0", "42176", "59"): "59 .. ExecutingUnit",
        ("FIX.5.0", "42176", "60"): "60 .. IntroducingBroker",
        ("FIX.5.0", "42176", "61"): "61 .. QuoteOriginator",
        ("FIX.5.0", "42176", "62"): "62 .. ReportOriginator",
        ("FIX.5.0", "42176", "63"): "63 .. SystematicInternaliser",
        ("FIX.5.0", "42176", "64"): "64 .. MultilateralTradingFacility",
        ("FIX.5.0", "42176", "65"): "65 .. RegulatedMarket",
        ("FIX.5.0", "42176", "66"): "66 .. MarketMaker",
        ("FIX.5.0", "42176", "67"): "67 .. InvestmentFirm",
        ("FIX.5.0", "42176", "68"): "68 .. HostCompetentAuthority",
        ("FIX.5.0", "42176", "69"): "69 .. HomeCompetentAuthority",
        ("FIX.5.0", "42176", "70"): "70 .. CompetentAuthorityLiquidity",
        ("FIX.5.0", "42176", "71"): "71 .. CompetentAuthorityTransactionVenue",
        ("FIX.5.0", "42176", "72"): "72 .. ReportingIntermediary",
        ("FIX.5.0", "42176", "73"): "73 .. ExecutionVenue",
        ("FIX.5.0", "42176", "74"): "74 .. MarketDataEntryOriginator",
        ("FIX.5.0", "42176", "75"): "75 .. LocationID",
        ("FIX.5.0", "42176", "76"): "76 .. DeskID",
        ("FIX.5.0", "42176", "77"): "77 .. MarketDataMarket",
        ("FIX.5.0", "42176", "78"): "78 .. AllocationEntity",
        ("FIX.5.0", "42176", "79"): "79 .. PrimeBroker",
        ("FIX.5.0", "42176", "80"): "80 .. StepOutFirm",
        ("FIX.5.0", "42176", "81"): "81 .. BrokerClearingID",
        ("FIX.5.0", "42176", "82"): "82 .. CentralRegistrationDepository",
        ("FIX.5.0", "42176", "83"): "83 .. ClearingAccount",
        ("FIX.5.0", "42176", "84"): "84 .. AcceptableSettlingCounterparty",
        ("FIX.5.0", "42176", "85"): "85 .. UnacceptableSettlingCounterparty",
        ("FIX.5.0", "42176", "86"): "86 .. CLSMemberBank",
        ("FIX.5.0", "42176", "87"): "87 .. InConcertGroup",
        ("FIX.5.0", "42176", "88"): "88 .. InConcertControllingEntity",
        ("FIX.5.0", "42176", "89"): "89 .. LargePositionsReportingAccount",
        ("FIX.5.0", "42176", "90"): "90 .. SettlementFirm",
        ("FIX.5.0", "42176", "91"): "91 .. SettlementAccount",
        ("FIX.5.0", "42176", "92"): "92 .. ReportingMarketCenter",
        ("FIX.5.0", "42176", "93"): "93 .. RelatedReportingMarketCenter",
        ("FIX.5.0", "42176", "94"): "94 .. AwayMarket",
        ("FIX.5.0", "42176", "95"): "95 .. GiveupTradingFirm",
        ("FIX.5.0", "42176", "96"): "96 .. TakeupTradingFirm",
        ("FIX.5.0", "42176", "97"): "97 .. GiveupClearingFirm",
        ("FIX.5.0", "42176", "98"): "98 .. TakeupClearingFirm",
        ("FIX.5.0", "42176", "99"): "99 .. OriginatingMarket",
        ("FIX.5.0", "42176", "100"): "100 .. MarginAccount",
        ("FIX.5.0", "42176", "101"): "101 .. CollateralAssetAccount",
        ("FIX.5.0", "42176", "102"): "102 .. DataRepository",
        ("FIX.5.0", "42176", "103"): "103 .. CalculationAgent",
        ("FIX.5.0", "42176", "104"): "104 .. ExerciseNoticeSender",
        ("FIX.5.0", "42176", "105"): "105 .. ExerciseNoticeReceiver",
        ("FIX.5.0", "42176", "106"): "106 .. RateReferenceBank",
        ("FIX.5.0", "42176", "107"): "107 .. Correspondent",
        ("FIX.5.0", "42176", "109"): "109 .. BeneficiaryBank",
        ("FIX.5.0", "42176", "110"): "110 .. Borrower",
        ("FIX.5.0", "42176", "111"): "111 .. PrimaryObligator",
        ("FIX.5.0", "42176", "112"): "112 .. Guarantor",
        ("FIX.5.0", "42176", "113"): "113 .. ExcludedReferenceEntity",
        ("FIX.5.0", "42176", "114"): "114 .. DeterminingParty",
        ("FIX.5.0", "42176", "115"): "115 .. HedgingParty",
        ("FIX.5.0", "42176", "116"): "116 .. ReportingEntity",
        ("FIX.5.0", "42176", "117"): "117 .. SalesPerson",
        ("FIX.5.0", "42176", "118"): "118 .. Operator",
        ("FIX.5.0", "42176", "119"): "119 .. CSD",
        ("FIX.5.0", "42176", "120"): "120 .. ICSD",
        ("FIX.5.0", "42176", "121"): "121 .. TradingSubAccount",
        ("FIX.5.0", "42176", "122"): "122 .. InvestmentDecisionMaker",
        ("FIX.5.0", "42176", "123"): "123 .. PublishingIntermediary",
        ("FIX.5.0", "42179", "1"): "1 .. Firm",
        ("FIX.5.0", "42179", "2"): "2 .. Person",
        ("FIX.5.0", "42179", "3"): "3 .. System",
        ("FIX.5.0", "42179", "4"): "4 .. Application",
        ("FIX.5.0", "42179", "5"): "5 .. FullLegalNameOfFirm",
        ("FIX.5.0", "42179", "6"): "6 .. PostalAddress",
        ("FIX.5.0", "42179", "7"): "7 .. PhoneNumber",
        ("FIX.5.0", "42179", "8"): "8 .. EmailAddress",
        ("FIX.5.0", "42179", "9"): "9 .. ContactName",
        ("FIX.5.0", "42179", "10"): "10 .. SecuritiesAccountNumber",
        ("FIX.5.0", "42179", "11"): "11 .. RegistrationNumber",
        ("FIX.5.0", "42179", "12"): "12 .. RegisteredAddressForConfirmation",
        ("FIX.5.0", "42179", "13"): "13 .. RegulatoryStatus",
        ("FIX.5.0", "42179", "14"): "14 .. RegistrationName",
        ("FIX.5.0", "42179", "15"): "15 .. CashAccountNumber",
        ("FIX.5.0", "42179", "16"): "16 .. BIC",
        ("FIX.5.0", "42179", "17"): "17 .. CSDParticipantMemberCode",
        ("FIX.5.0", "42179", "18"): "18 .. RegisteredAddress",
        ("FIX.5.0", "42179", "19"): "19 .. FundAccountName",
        ("FIX.5.0", "42179", "20"): "20 .. TelexNumber",
        ("FIX.5.0", "42179", "21"): "21 .. FaxNumber",
        ("FIX.5.0", "42179", "22"): "22 .. SecuritiesAccountName",
        ("FIX.5.0", "42179", "23"): "23 .. CashAccountName",
        ("FIX.5.0", "42179", "24"): "24 .. Department",
        ("FIX.5.0", "42179", "25"): "25 .. LocationDesk",
        ("FIX.5.0", "42179", "26"): "26 .. PositionAccountType",
        ("FIX.5.0", "42179", "27"): "27 .. SecurityLocateID",
        ("FIX.5.0", "42179", "28"): "28 .. MarketMaker",
        ("FIX.5.0", "42179", "29"): "29 .. EligibleCounterparty",
        ("FIX.5.0", "42179", "30"): "30 .. ProfessionalClient",
        ("FIX.5.0", "42179", "31"): "31 .. Location",
        ("FIX.5.0", "42179", "32"): "32 .. ExecutionVenue",
        ("FIX.5.0", "42179", "33"): "33 .. CurrencyDeliveryIdentifier",
        ("FIX.5.0", "42179", "34"): "34 .. AddressCity",
        ("FIX.5.0", "42179", "35"): "35 .. AddressStateOrProvince",
        ("FIX.5.0", "42179", "36"): "36 .. AddressPostalCode",
        ("FIX.5.0", "42179", "37"): "37 .. AddressStreet",
        ("FIX.5.0", "42179", "38"): "38 .. AddressISOCountryCode",
        ("FIX.5.0", "42179", "39"): "39 .. ISOCountryCode",
        ("FIX.5.0", "42179", "40"): "40 .. MarketSegment",
        ("FIX.5.0", "42179", "41"): "41 .. CustomerAccountType",
        ("FIX.5.0", "42179", "42"): "42 .. OmnibusAccount",
        ("FIX.5.0", "42179", "43"): "43 .. FundsSegregationType",
        ("FIX.5.0", "42179", "44"): "44 .. GuaranteeFund",
        ("FIX.5.0", "42179", "45"): "45 .. SwapDealer",
        ("FIX.5.0", "42179", "46"): "46 .. MajorParticipant",
        ("FIX.5.0", "42179", "47"): "47 .. FinancialEntity",
        ("FIX.5.0", "42179", "48"): "48 .. USPerson",
        ("FIX.5.0", "42179", "49"): "49 .. ReportingEntityIndicator",
        ("FIX.5.0", "42179", "50"): "50 .. ElectedClearingRequirementException",
        ("FIX.5.0", "42179", "51"): "51 .. BusinessCenter",
        ("FIX.5.0", "42179", "52"): "52 .. ReferenceText",
        ("FIX.5.0", "42179", "53"): "53 .. ShortMarkingExemptAccount",
        ("FIX.5.0", "42179", "54"): "54 .. ParentFirmIdentifier",
        ("FIX.5.0", "42179", "55"): "55 .. ParentFirmName",
        ("FIX.5.0", "42179", "56"): "56 .. DealIdentifier",
        ("FIX.5.0", "42179", "57"): "57 .. SystemTradeID",
        ("FIX.5.0", "42179", "58"): "58 .. SystemTradeSubID",
        ("FIX.5.0", "42179", "59"): "59 .. FCMCode",
        ("FIX.5.0", "42179", "60"): "60 .. DlvryTrmlCode",
        ("FIX.5.0", "42179", "61"): "61 .. VolntyRptEntity",
        ("FIX.5.0", "42179", "62"): "62 .. RptObligJursdctn",
        ("FIX.5.0", "42179", "63"): "63 .. VolntyRptJursdctn",
        ("FIX.5.0", "42179", "64"): "64 .. CompanyActivities",
        ("FIX.5.0", "42179", "65"): "65 .. EEAreaDomiciled",
        ("FIX.5.0", "42179", "66"): "66 .. ContractLinked",
        ("FIX.5.0", "42179", "67"): "67 .. ContractAbove",
        ("FIX.5.0", "42179", "68"): "68 .. VolntyRptPty",
        ("FIX.5.0", "42179", "69"): "69 .. EndUser",
        ("FIX.5.0", "42179", "70"): "70 .. LocationOrJurisdiction",
        ("FIX.5.0", "42179", "71"): "71 .. DerivativesDealer",
        ("FIX.5.0", "42179", "72"): "72 .. Domicile",
        ("FIX.5.0", "42179", "73"): "73 .. ExemptFromRecognition",
        ("FIX.5.0", "42179", "74"): "74 .. Payer",
        ("FIX.5.0", "42179", "75"): "75 .. Receiver",
        ("FIX.5.0", "42179", "76"): "76 .. SystematicInternaliser",
        ("FIX.5.0", "42179", "77"): "77 .. PublishingEntityIndicator",
        ("FIX.5.0", "42179", "78"): "78 .. FirstName",
        ("FIX.5.0", "42179", "79"): "79 .. Surname",
        ("FIX.5.0", "42179", "80"): "80 .. DateOfBirth",
        ("FIX.5.0", "42179", "81"): "81 .. OrderTransmittingFirm",
        ("FIX.5.0", "42179", "82"): "82 .. OrderTransmittingFirmBuyer",
        ("FIX.5.0", "42179", "83"): "83 .. OrderTransmitterSeller",
        ("FIX.5.0", "42179", "84"): "84 .. LegalEntityIdentifier",
        ("FIX.5.0", "42192", "0"): "0 .. Proprietary",
        ("FIX.5.0", "42192", "1"): "1 .. EIC",
        ("FIX.5.0", "42194", "0"): "0 .. Proprietary",
        ("FIX.5.0", "42194", "1"): "1 .. EIC",
        ("FIX.5.0", "42196", "0"): "0 .. Proprietary",
        ("FIX.5.0", "42196", "1"): "1 .. EIC",
        ("FIX.5.0", "42208", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42208", "1"): "1 .. None",
        ("FIX.5.0", "42208", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42208", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42208", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42208", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42208", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42208", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42211", "D"): "D .. Day",
        ("FIX.5.0", "42211", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42211", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42211", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42212", "0"): "0 .. Business",
        ("FIX.5.0", "42212", "1"): "1 .. Calendar",
        ("FIX.5.0", "42212", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42212", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42212", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42212", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42217", "0"): "0 .. Close",
        ("FIX.5.0", "42217", "1"): "1 .. Hedge",
        ("FIX.5.0", "42220", "D"): "D .. Day",
        ("FIX.5.0", "42220", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42220", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42220", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42223", "0"): "0 .. Short",
        ("FIX.5.0", "42223", "1"): "1 .. Long",
        ("FIX.5.0", "42224", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "42224", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "42226", "1"): "1 .. Buyer",
        ("FIX.5.0", "42226", "2"): "2 .. Seller",
        ("FIX.5.0", "42227", "1"): "1 .. Buyer",
        ("FIX.5.0", "42227", "2"): "2 .. Seller",
        ("FIX.5.0", "42229", "1"): "1 .. Buyer",
        ("FIX.5.0", "42229", "2"): "2 .. Seller",
        ("FIX.5.0", "42230", "1"): "1 .. Buyer",
        ("FIX.5.0", "42230", "2"): "2 .. Seller",
        ("FIX.5.0", "42232", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "42232", "1"): "1 .. RoundDown",
        ("FIX.5.0", "42232", "2"): "2 .. RoundUp",
        ("FIX.5.0", "42234", "0"): "0 .. Unweighted",
        ("FIX.5.0", "42234", "1"): "1 .. Weighted",
        ("FIX.5.0", "42235", "0"): "0 .. ZeroInterestRateMethod",
        ("FIX.5.0", "42235", "1"): "1 .. NegativeInterestRateMethod",
        ("FIX.5.0", "42240", "D"): "D .. Day",
        ("FIX.5.0", "42240", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42240", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42240", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42241", "0"): "0 .. Business",
        ("FIX.5.0", "42241", "1"): "1 .. Calendar",
        ("FIX.5.0", "42241", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42241", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42241", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42241", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42243", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42243", "1"): "1 .. None",
        ("FIX.5.0", "42243", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42243", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42243", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42243", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42243", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42243", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42246", "0"): "0 .. ExDate",
        ("FIX.5.0", "42246", "1"): "1 .. RecordDate",
        ("FIX.5.0", "42247", "0"): "0 .. RecordAmount",
        ("FIX.5.0", "42247", "1"): "1 .. ExAmount",
        ("FIX.5.0", "42247", "2"): "2 .. PaidAmount",
        ("FIX.5.0", "42247", "3"): "3 .. PerMasterConfirm",
        ("FIX.5.0", "42249", "1"): "1 .. Buyer",
        ("FIX.5.0", "42249", "2"): "2 .. Seller",
        ("FIX.5.0", "42250", "0"): "0 .. RecordAmount",
        ("FIX.5.0", "42250", "1"): "1 .. ExAmount",
        ("FIX.5.0", "42250", "2"): "2 .. PaidAmount",
        ("FIX.5.0", "42250", "3"): "3 .. PerMasterConfirm",
        ("FIX.5.0", "42254", "0"): "0 .. None",
        ("FIX.5.0", "42254", "1"): "1 .. Flat",
        ("FIX.5.0", "42254", "2"): "2 .. Straight",
        ("FIX.5.0", "42254", "3"): "3 .. SpreadExclusive",
        ("FIX.5.0", "42258", "0"): "0 .. PotentialAdjustment",
        ("FIX.5.0", "42258", "1"): "1 .. CashEquivalent",
        ("FIX.5.0", "42259", "0"): "0 .. EquityAmountReceiver",
        ("FIX.5.0", "42259", "1"): "1 .. CalculationAgent",
        ("FIX.5.0", "42267", "D"): "D .. Day",
        ("FIX.5.0", "42267", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42267", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42267", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42268", "0"): "0 .. Business",
        ("FIX.5.0", "42268", "1"): "1 .. Calendar",
        ("FIX.5.0", "42268", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42268", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42268", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42268", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42270", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42270", "1"): "1 .. None",
        ("FIX.5.0", "42270", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42270", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42270", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42270", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42270", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42270", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42280", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42280", "1"): "1 .. None",
        ("FIX.5.0", "42280", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42280", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42280", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42280", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42280", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42280", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42284", "D"): "D .. Day",
        ("FIX.5.0", "42284", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42284", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42284", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42285", "0"): "0 .. Business",
        ("FIX.5.0", "42285", "1"): "1 .. Calendar",
        ("FIX.5.0", "42285", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42285", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42285", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42285", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42290", "D"): "D .. Day",
        ("FIX.5.0", "42290", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42290", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42290", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42291", "0"): "0 .. Business",
        ("FIX.5.0", "42291", "1"): "1 .. Calendar",
        ("FIX.5.0", "42291", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42291", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42291", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42291", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42300", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42300", "1"): "1 .. None",
        ("FIX.5.0", "42300", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42300", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42300", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42300", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42300", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42300", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42303", "D"): "D .. Day",
        ("FIX.5.0", "42303", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42303", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42303", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42304", "0"): "0 .. Business",
        ("FIX.5.0", "42304", "1"): "1 .. Calendar",
        ("FIX.5.0", "42304", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42304", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42304", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42304", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42309", "0"): "0 .. Close",
        ("FIX.5.0", "42309", "1"): "1 .. Hedge",
        ("FIX.5.0", "42314", "D"): "D .. Day",
        ("FIX.5.0", "42314", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42314", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42314", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42317", "0"): "0 .. Short",
        ("FIX.5.0", "42317", "1"): "1 .. Long",
        ("FIX.5.0", "42318", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "42318", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "42320", "1"): "1 .. Buyer",
        ("FIX.5.0", "42320", "2"): "2 .. Seller",
        ("FIX.5.0", "42321", "1"): "1 .. Buyer",
        ("FIX.5.0", "42321", "2"): "2 .. Seller",
        ("FIX.5.0", "42323", "1"): "1 .. Buyer",
        ("FIX.5.0", "42323", "2"): "2 .. Seller",
        ("FIX.5.0", "42324", "1"): "1 .. Buyer",
        ("FIX.5.0", "42324", "2"): "2 .. Seller",
        ("FIX.5.0", "42326", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "42326", "1"): "1 .. RoundDown",
        ("FIX.5.0", "42326", "2"): "2 .. RoundUp",
        ("FIX.5.0", "42328", "0"): "0 .. Unweighted",
        ("FIX.5.0", "42328", "1"): "1 .. Weighted",
        ("FIX.5.0", "42329", "0"): "0 .. ZeroInterestRateMethod",
        ("FIX.5.0", "42329", "1"): "1 .. NegativeInterestRateMethod",
        ("FIX.5.0", "42332", "D"): "D .. Day",
        ("FIX.5.0", "42332", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42332", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42332", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42333", "0"): "0 .. Business",
        ("FIX.5.0", "42333", "1"): "1 .. Calendar",
        ("FIX.5.0", "42333", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42333", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42333", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42333", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42335", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42335", "1"): "1 .. None",
        ("FIX.5.0", "42335", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42335", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42335", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42335", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42335", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42335", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42338", "0"): "0 .. ExDate",
        ("FIX.5.0", "42338", "1"): "1 .. RecordDate",
        ("FIX.5.0", "42339", "0"): "0 .. RecordAmount",
        ("FIX.5.0", "42339", "1"): "1 .. ExAmount",
        ("FIX.5.0", "42339", "2"): "2 .. PaidAmount",
        ("FIX.5.0", "42339", "3"): "3 .. PerMasterConfirm",
        ("FIX.5.0", "42341", "1"): "1 .. Buyer",
        ("FIX.5.0", "42341", "2"): "2 .. Seller",
        ("FIX.5.0", "42342", "0"): "0 .. RecordAmount",
        ("FIX.5.0", "42342", "1"): "1 .. ExAmount",
        ("FIX.5.0", "42342", "2"): "2 .. PaidAmount",
        ("FIX.5.0", "42342", "3"): "3 .. PerMasterConfirm",
        ("FIX.5.0", "42346", "0"): "0 .. None",
        ("FIX.5.0", "42346", "1"): "1 .. Flat",
        ("FIX.5.0", "42346", "2"): "2 .. Straight",
        ("FIX.5.0", "42346", "3"): "3 .. SpreadExclusive",
        ("FIX.5.0", "42350", "0"): "0 .. PotentialAdjustment",
        ("FIX.5.0", "42350", "1"): "1 .. CashEquivalent",
        ("FIX.5.0", "42351", "0"): "0 .. EquityAmountReceiver",
        ("FIX.5.0", "42351", "1"): "1 .. CalculationAgent",
        ("FIX.5.0", "42359", "D"): "D .. Day",
        ("FIX.5.0", "42359", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42359", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42359", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42360", "0"): "0 .. Business",
        ("FIX.5.0", "42360", "1"): "1 .. Calendar",
        ("FIX.5.0", "42360", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42360", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42360", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42360", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42362", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42362", "1"): "1 .. None",
        ("FIX.5.0", "42362", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42362", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42362", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42362", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42362", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42362", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42372", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42372", "1"): "1 .. None",
        ("FIX.5.0", "42372", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42372", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42372", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42372", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42372", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42372", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42376", "D"): "D .. Day",
        ("FIX.5.0", "42376", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42376", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42376", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42377", "0"): "0 .. Business",
        ("FIX.5.0", "42377", "1"): "1 .. Calendar",
        ("FIX.5.0", "42377", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42377", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42377", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42377", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42382", "D"): "D .. Day",
        ("FIX.5.0", "42382", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42382", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42382", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42383", "0"): "0 .. Business",
        ("FIX.5.0", "42383", "1"): "1 .. Calendar",
        ("FIX.5.0", "42383", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42383", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42383", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42383", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42391", "1"): "1 .. Buy",
        ("FIX.5.0", "42391", "2"): "2 .. Sell",
        ("FIX.5.0", "42397", "0"): "0 .. Bid",
        ("FIX.5.0", "42397", "1"): "1 .. Mid",
        ("FIX.5.0", "42397", "2"): "2 .. Offer",
        ("FIX.5.0", "42398", "0"): "0 .. None",
        ("FIX.5.0", "42398", "1"): "1 .. LinearZeroYield",
        ("FIX.5.0", "42402", "0"): "0 .. None",
        ("FIX.5.0", "42402", "1"): "1 .. LinearZeroYield",
        ("FIX.5.0", "42403", "0"): "0 .. Initial",
        ("FIX.5.0", "42403", "1"): "1 .. InitialAndFinal",
        ("FIX.5.0", "42403", "2"): "2 .. Final",
        ("FIX.5.0", "42403", "3"): "3 .. AnyPeriod",
        ("FIX.5.0", "42407", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42407", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42408", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42408", "1"): "1 .. None",
        ("FIX.5.0", "42408", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42408", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42408", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42408", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42408", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42408", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42411", "D"): "D .. Day",
        ("FIX.5.0", "42411", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42411", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42411", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42412", "0"): "0 .. Business",
        ("FIX.5.0", "42412", "1"): "1 .. Calendar",
        ("FIX.5.0", "42412", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42412", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42412", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42412", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42415", "D"): "D .. Day",
        ("FIX.5.0", "42415", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42415", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42415", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42415", "H"): "H .. Hour",
        ("FIX.5.0", "42415", "Min"): "Min .. Minute",
        ("FIX.5.0", "42415", "S"): "S .. Second",
        ("FIX.5.0", "42415", "T"): "T .. Term",
        ("FIX.5.0", "42416", "1"): "1 .. FirstDay",
        ("FIX.5.0", "42416", "2"): "2 .. SecondDay",
        ("FIX.5.0", "42416", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "42416", "4"): "4 .. FourthDay",
        ("FIX.5.0", "42416", "5"): "5 .. FifthDay",
        ("FIX.5.0", "42416", "6"): "6 .. SixthDay",
        ("FIX.5.0", "42416", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "42416", "8"): "8 .. EighthDay",
        ("FIX.5.0", "42416", "9"): "9 .. NinthDay",
        ("FIX.5.0", "42416", "10"): "10 .. TenthDay",
        ("FIX.5.0", "42416", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "42416", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "42416", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "42416", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "42416", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "42416", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "42416", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "42416", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "42416", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "42416", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "42416", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "42416", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "42416", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "42416", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "42416", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "42416", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "42416", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "42416", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "42416", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "42416", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "42416", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "42416", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "42416", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "42416", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "42416", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "42416", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "42416", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "42416", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "42416", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "42416", "MON"): "MON .. MON",
        ("FIX.5.0", "42416", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "42416", "WED"): "WED .. WED",
        ("FIX.5.0", "42416", "THU"): "THU .. THU",
        ("FIX.5.0", "42416", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "42416", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "42416", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "42424", "D"): "D .. Day",
        ("FIX.5.0", "42424", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42424", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42424", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42425", "0"): "0 .. Business",
        ("FIX.5.0", "42425", "1"): "1 .. Calendar",
        ("FIX.5.0", "42425", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42425", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42425", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42425", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42429", "D"): "D .. Day",
        ("FIX.5.0", "42429", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42429", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42429", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42432", "0"): "0 .. Short",
        ("FIX.5.0", "42432", "1"): "1 .. Long",
        ("FIX.5.0", "42433", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "42433", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "42435", "1"): "1 .. Buyer",
        ("FIX.5.0", "42435", "2"): "2 .. Seller",
        ("FIX.5.0", "42436", "1"): "1 .. Buyer",
        ("FIX.5.0", "42436", "2"): "2 .. Seller",
        ("FIX.5.0", "42438", "1"): "1 .. Buyer",
        ("FIX.5.0", "42438", "2"): "2 .. Seller",
        ("FIX.5.0", "42439", "1"): "1 .. Buyer",
        ("FIX.5.0", "42439", "2"): "2 .. Seller",
        ("FIX.5.0", "42441", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "42441", "1"): "1 .. RoundDown",
        ("FIX.5.0", "42441", "2"): "2 .. RoundUp",
        ("FIX.5.0", "42443", "0"): "0 .. Unweighted",
        ("FIX.5.0", "42443", "1"): "1 .. Weighted",
        ("FIX.5.0", "42444", "0"): "0 .. ZeroInterestRateMethod",
        ("FIX.5.0", "42444", "1"): "1 .. NegativeInterestRateMethod",
        ("FIX.5.0", "42448", "D"): "D .. Day",
        ("FIX.5.0", "42448", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42448", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42448", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42449", "0"): "0 .. Business",
        ("FIX.5.0", "42449", "1"): "1 .. Calendar",
        ("FIX.5.0", "42449", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42449", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42449", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42449", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42456", "D"): "D .. Day",
        ("FIX.5.0", "42456", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42456", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42456", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42457", "0"): "0 .. Business",
        ("FIX.5.0", "42457", "1"): "1 .. Calendar",
        ("FIX.5.0", "42457", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42457", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42457", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42457", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42461", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42461", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42464", "0"): "0 .. Business",
        ("FIX.5.0", "42464", "1"): "1 .. Calendar",
        ("FIX.5.0", "42464", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42464", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42464", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42464", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42473", "0"): "0 .. Volatility",
        ("FIX.5.0", "42473", "1"): "1 .. Variance",
        ("FIX.5.0", "42478", "0"): "0 .. Previous",
        ("FIX.5.0", "42478", "1"): "1 .. Last",
        ("FIX.5.0", "42478", "2"): "2 .. Both",
        ("FIX.5.0", "42489", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42489", "1"): "1 .. None",
        ("FIX.5.0", "42489", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42489", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42489", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42489", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42489", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42489", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42492", "D"): "D .. Day",
        ("FIX.5.0", "42492", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42492", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42492", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42493", "0"): "0 .. Business",
        ("FIX.5.0", "42493", "1"): "1 .. Calendar",
        ("FIX.5.0", "42493", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42493", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42493", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42493", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42498", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42498", "1"): "1 .. None",
        ("FIX.5.0", "42498", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42498", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42498", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42498", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42498", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42498", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42501", "D"): "D .. Day",
        ("FIX.5.0", "42501", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42501", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42501", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42502", "0"): "0 .. Business",
        ("FIX.5.0", "42502", "1"): "1 .. Calendar",
        ("FIX.5.0", "42502", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42502", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42502", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42502", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42506", "0"): "0 .. FlatFee",
        ("FIX.5.0", "42506", "1"): "1 .. AmortizedFee",
        ("FIX.5.0", "42506", "2"): "2 .. FundingFee",
        ("FIX.5.0", "42506", "3"): "3 .. FlatAndFundingFee",
        ("FIX.5.0", "42506", "4"): "4 .. AmortizedAndFundingFee",
        ("FIX.5.0", "42509", "0"): "0 .. PriceValuation",
        ("FIX.5.0", "42509", "1"): "1 .. DividendValuation",
        ("FIX.5.0", "42512", "D"): "D .. Day",
        ("FIX.5.0", "42512", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42512", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42512", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42513", "0"): "0 .. Business",
        ("FIX.5.0", "42513", "1"): "1 .. Calendar",
        ("FIX.5.0", "42513", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42513", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42513", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42513", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42517", "D"): "D .. Day",
        ("FIX.5.0", "42517", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42517", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42517", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42518", "0"): "0 .. Business",
        ("FIX.5.0", "42518", "1"): "1 .. Calendar",
        ("FIX.5.0", "42518", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42518", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42518", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42518", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42523", "D"): "D .. Day",
        ("FIX.5.0", "42523", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42523", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42523", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42524", "0"): "0 .. Business",
        ("FIX.5.0", "42524", "1"): "1 .. Calendar",
        ("FIX.5.0", "42524", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42524", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42524", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42524", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42527", "D"): "D .. Day",
        ("FIX.5.0", "42527", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42527", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42527", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42527", "H"): "H .. Hour",
        ("FIX.5.0", "42527", "Min"): "Min .. Minute",
        ("FIX.5.0", "42527", "S"): "S .. Second",
        ("FIX.5.0", "42527", "T"): "T .. Term",
        ("FIX.5.0", "42528", "1"): "1 .. FirstDay",
        ("FIX.5.0", "42528", "2"): "2 .. SecondDay",
        ("FIX.5.0", "42528", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "42528", "4"): "4 .. FourthDay",
        ("FIX.5.0", "42528", "5"): "5 .. FifthDay",
        ("FIX.5.0", "42528", "6"): "6 .. SixthDay",
        ("FIX.5.0", "42528", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "42528", "8"): "8 .. EighthDay",
        ("FIX.5.0", "42528", "9"): "9 .. NinthDay",
        ("FIX.5.0", "42528", "10"): "10 .. TenthDay",
        ("FIX.5.0", "42528", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "42528", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "42528", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "42528", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "42528", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "42528", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "42528", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "42528", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "42528", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "42528", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "42528", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "42528", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "42528", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "42528", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "42528", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "42528", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "42528", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "42528", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "42528", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "42528", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "42528", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "42528", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "42528", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "42528", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "42528", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "42528", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "42528", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "42528", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "42528", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "42528", "MON"): "MON .. MON",
        ("FIX.5.0", "42528", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "42528", "WED"): "WED .. WED",
        ("FIX.5.0", "42528", "THU"): "THU .. THU",
        ("FIX.5.0", "42528", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "42528", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "42528", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "42529", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42529", "1"): "1 .. None",
        ("FIX.5.0", "42529", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42529", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42529", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42529", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42529", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42529", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42533", "M"): "M .. Multiply",
        ("FIX.5.0", "42533", "D"): "D .. Divide",
        ("FIX.5.0", "42535", "0"): "0 .. Initial",
        ("FIX.5.0", "42535", "1"): "1 .. Interim",
        ("FIX.5.0", "42535", "2"): "2 .. Final",
        ("FIX.5.0", "42536", "1"): "1 .. PerUnit",
        ("FIX.5.0", "42536", "2"): "2 .. Percent",
        ("FIX.5.0", "42536", "3"): "3 .. Absolute",
        ("FIX.5.0", "42536", "4"): "4 .. PercentageWaivedCashDiscount",
        ("FIX.5.0", "42536", "5"): "5 .. PercentageWaivedEnhancedUnits",
        ("FIX.5.0", "42536", "6"): "6 .. PointsPerBondOrContract",
        ("FIX.5.0", "42536", "7"): "7 .. BasisPoints",
        ("FIX.5.0", "42536", "8"): "8 .. AmountPerContract",
        ("FIX.5.0", "42544", "0"): "0 .. Bid",
        ("FIX.5.0", "42544", "1"): "1 .. Mid",
        ("FIX.5.0", "42544", "2"): "2 .. Offer",
        ("FIX.5.0", "42547", "0"): "0 .. Open",
        ("FIX.5.0", "42547", "1"): "1 .. OfficialSettlPx",
        ("FIX.5.0", "42547", "2"): "2 .. Xetra",
        ("FIX.5.0", "42547", "3"): "3 .. Close",
        ("FIX.5.0", "42547", "4"): "4 .. DerivativesClose",
        ("FIX.5.0", "42547", "5"): "5 .. High",
        ("FIX.5.0", "42547", "6"): "6 .. Low",
        ("FIX.5.0", "42547", "7"): "7 .. AsSpecifiedInMasterConfirmation",
        ("FIX.5.0", "42555", "0"): "0 .. Open",
        ("FIX.5.0", "42555", "1"): "1 .. OfficialSettlPx",
        ("FIX.5.0", "42555", "2"): "2 .. Xetra",
        ("FIX.5.0", "42555", "3"): "3 .. Close",
        ("FIX.5.0", "42555", "4"): "4 .. DerivativesClose",
        ("FIX.5.0", "42555", "5"): "5 .. High",
        ("FIX.5.0", "42555", "6"): "6 .. Low",
        ("FIX.5.0", "42555", "7"): "7 .. AsSpecifiedInMasterConfirmation",
        ("FIX.5.0", "42558", "0"): "0 .. None",
        ("FIX.5.0", "42558", "1"): "1 .. FuturesPrice",
        ("FIX.5.0", "42558", "2"): "2 .. OptionsPrice",
        ("FIX.5.0", "42559", "0"): "0 .. Close",
        ("FIX.5.0", "42559", "1"): "1 .. HedgeElection",
        ("FIX.5.0", "42561", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "42561", "1"): "1 .. Reuters",
        ("FIX.5.0", "42561", "2"): "2 .. Telerate",
        ("FIX.5.0", "42561", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "42561", "99"): "99 .. Other",
        ("FIX.5.0", "42565", "0"): "0 .. Gross",
        ("FIX.5.0", "42565", "1"): "1 .. Net",
        ("FIX.5.0", "42565", "2"): "2 .. Accrued",
        ("FIX.5.0", "42565", "3"): "3 .. CleanNet",
        ("FIX.5.0", "42568", "0"): "0 .. AbsoluteTerms",
        ("FIX.5.0", "42568", "1"): "1 .. PercentageOfNotional",
        ("FIX.5.0", "42573", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42573", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42575", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42575", "1"): "1 .. None",
        ("FIX.5.0", "42575", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42575", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42575", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42575", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42575", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42575", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42578", "D"): "D .. Day",
        ("FIX.5.0", "42578", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42578", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42578", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42579", "0"): "0 .. Business",
        ("FIX.5.0", "42579", "1"): "1 .. Calendar",
        ("FIX.5.0", "42579", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42579", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42579", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42579", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42586", "0"): "0 .. Execution",
        ("FIX.5.0", "42586", "1"): "1 .. PortfolioRebalancing",
        ("FIX.5.0", "42586", "2"): "2 .. Standard",
        ("FIX.5.0", "42590", "1"): "1 .. Buy",
        ("FIX.5.0", "42590", "2"): "2 .. Sell",
        ("FIX.5.0", "42596", "0"): "0 .. Bid",
        ("FIX.5.0", "42596", "1"): "1 .. Mid",
        ("FIX.5.0", "42596", "2"): "2 .. Offer",
        ("FIX.5.0", "42597", "0"): "0 .. None",
        ("FIX.5.0", "42597", "1"): "1 .. LinearZeroYield",
        ("FIX.5.0", "42603", "0"): "0 .. None",
        ("FIX.5.0", "42603", "1"): "1 .. LinearZeroYield",
        ("FIX.5.0", "42604", "0"): "0 .. Initial",
        ("FIX.5.0", "42604", "1"): "1 .. InitialAndFinal",
        ("FIX.5.0", "42604", "2"): "2 .. Final",
        ("FIX.5.0", "42604", "3"): "3 .. AnyPeriod",
        ("FIX.5.0", "42608", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42608", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42609", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42609", "1"): "1 .. None",
        ("FIX.5.0", "42609", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42609", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42609", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42609", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42609", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42609", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42612", "D"): "D .. Day",
        ("FIX.5.0", "42612", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42612", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42612", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42613", "0"): "0 .. Business",
        ("FIX.5.0", "42613", "1"): "1 .. Calendar",
        ("FIX.5.0", "42613", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42613", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42613", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42613", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42616", "D"): "D .. Day",
        ("FIX.5.0", "42616", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42616", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42616", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42616", "H"): "H .. Hour",
        ("FIX.5.0", "42616", "Min"): "Min .. Minute",
        ("FIX.5.0", "42616", "S"): "S .. Second",
        ("FIX.5.0", "42616", "T"): "T .. Term",
        ("FIX.5.0", "42617", "1"): "1 .. FirstDay",
        ("FIX.5.0", "42617", "2"): "2 .. SecondDay",
        ("FIX.5.0", "42617", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "42617", "4"): "4 .. FourthDay",
        ("FIX.5.0", "42617", "5"): "5 .. FifthDay",
        ("FIX.5.0", "42617", "6"): "6 .. SixthDay",
        ("FIX.5.0", "42617", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "42617", "8"): "8 .. EighthDay",
        ("FIX.5.0", "42617", "9"): "9 .. NinthDay",
        ("FIX.5.0", "42617", "10"): "10 .. TenthDay",
        ("FIX.5.0", "42617", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "42617", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "42617", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "42617", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "42617", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "42617", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "42617", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "42617", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "42617", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "42617", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "42617", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "42617", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "42617", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "42617", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "42617", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "42617", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "42617", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "42617", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "42617", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "42617", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "42617", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "42617", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "42617", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "42617", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "42617", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "42617", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "42617", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "42617", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "42617", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "42617", "MON"): "MON .. MON",
        ("FIX.5.0", "42617", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "42617", "WED"): "WED .. WED",
        ("FIX.5.0", "42617", "THU"): "THU .. THU",
        ("FIX.5.0", "42617", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "42617", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "42617", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "42625", "D"): "D .. Day",
        ("FIX.5.0", "42625", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42625", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42625", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42626", "0"): "0 .. Business",
        ("FIX.5.0", "42626", "1"): "1 .. Calendar",
        ("FIX.5.0", "42626", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42626", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42626", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42626", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42630", "D"): "D .. Day",
        ("FIX.5.0", "42630", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42630", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42630", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42633", "0"): "0 .. Short",
        ("FIX.5.0", "42633", "1"): "1 .. Long",
        ("FIX.5.0", "42634", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "42634", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "42636", "1"): "1 .. Buyer",
        ("FIX.5.0", "42636", "2"): "2 .. Seller",
        ("FIX.5.0", "42637", "1"): "1 .. Buyer",
        ("FIX.5.0", "42637", "2"): "2 .. Seller",
        ("FIX.5.0", "42639", "1"): "1 .. Buyer",
        ("FIX.5.0", "42639", "2"): "2 .. Seller",
        ("FIX.5.0", "42640", "1"): "1 .. Buyer",
        ("FIX.5.0", "42640", "2"): "2 .. Seller",
        ("FIX.5.0", "42642", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "42642", "1"): "1 .. RoundDown",
        ("FIX.5.0", "42642", "2"): "2 .. RoundUp",
        ("FIX.5.0", "42644", "0"): "0 .. Unweighted",
        ("FIX.5.0", "42644", "1"): "1 .. Weighted",
        ("FIX.5.0", "42645", "0"): "0 .. ZeroInterestRateMethod",
        ("FIX.5.0", "42645", "1"): "1 .. NegativeInterestRateMethod",
        ("FIX.5.0", "42649", "D"): "D .. Day",
        ("FIX.5.0", "42649", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42649", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42649", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42650", "0"): "0 .. Business",
        ("FIX.5.0", "42650", "1"): "1 .. Calendar",
        ("FIX.5.0", "42650", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42650", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42650", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42650", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42657", "D"): "D .. Day",
        ("FIX.5.0", "42657", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42657", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42657", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42658", "0"): "0 .. Business",
        ("FIX.5.0", "42658", "1"): "1 .. Calendar",
        ("FIX.5.0", "42658", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42658", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42658", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42658", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42662", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42662", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42665", "0"): "0 .. Business",
        ("FIX.5.0", "42665", "1"): "1 .. Calendar",
        ("FIX.5.0", "42665", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42665", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42665", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42665", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42674", "0"): "0 .. Volatility",
        ("FIX.5.0", "42674", "1"): "1 .. Variance",
        ("FIX.5.0", "42679", "0"): "0 .. Previous",
        ("FIX.5.0", "42679", "1"): "1 .. Last",
        ("FIX.5.0", "42679", "2"): "2 .. Both",
        ("FIX.5.0", "42690", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42690", "1"): "1 .. None",
        ("FIX.5.0", "42690", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42690", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42690", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42690", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42690", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42690", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42693", "D"): "D .. Day",
        ("FIX.5.0", "42693", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42693", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42693", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42694", "0"): "0 .. Business",
        ("FIX.5.0", "42694", "1"): "1 .. Calendar",
        ("FIX.5.0", "42694", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42694", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42694", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42694", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42699", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42699", "1"): "1 .. None",
        ("FIX.5.0", "42699", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42699", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42699", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42699", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42699", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42699", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42702", "D"): "D .. Day",
        ("FIX.5.0", "42702", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42702", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42702", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42703", "0"): "0 .. Business",
        ("FIX.5.0", "42703", "1"): "1 .. Calendar",
        ("FIX.5.0", "42703", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42703", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42703", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42703", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42707", "0"): "0 .. FlatFee",
        ("FIX.5.0", "42707", "1"): "1 .. AmortizedFee",
        ("FIX.5.0", "42707", "2"): "2 .. FundingFee",
        ("FIX.5.0", "42707", "3"): "3 .. FlatAndFundingFee",
        ("FIX.5.0", "42707", "4"): "4 .. AmortizedAndFundingFee",
        ("FIX.5.0", "42710", "0"): "0 .. PriceValuation",
        ("FIX.5.0", "42710", "1"): "1 .. DividendValuation",
        ("FIX.5.0", "42713", "D"): "D .. Day",
        ("FIX.5.0", "42713", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42713", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42713", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42714", "0"): "0 .. Business",
        ("FIX.5.0", "42714", "1"): "1 .. Calendar",
        ("FIX.5.0", "42714", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42714", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42714", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42714", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42718", "D"): "D .. Day",
        ("FIX.5.0", "42718", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42718", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42718", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42719", "0"): "0 .. Business",
        ("FIX.5.0", "42719", "1"): "1 .. Calendar",
        ("FIX.5.0", "42719", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42719", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42719", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42719", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42724", "D"): "D .. Day",
        ("FIX.5.0", "42724", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42724", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42724", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42725", "0"): "0 .. Business",
        ("FIX.5.0", "42725", "1"): "1 .. Calendar",
        ("FIX.5.0", "42725", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42725", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42725", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42725", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42728", "D"): "D .. Day",
        ("FIX.5.0", "42728", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42728", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42728", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42728", "H"): "H .. Hour",
        ("FIX.5.0", "42728", "Min"): "Min .. Minute",
        ("FIX.5.0", "42728", "S"): "S .. Second",
        ("FIX.5.0", "42728", "T"): "T .. Term",
        ("FIX.5.0", "42729", "1"): "1 .. FirstDay",
        ("FIX.5.0", "42729", "2"): "2 .. SecondDay",
        ("FIX.5.0", "42729", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "42729", "4"): "4 .. FourthDay",
        ("FIX.5.0", "42729", "5"): "5 .. FifthDay",
        ("FIX.5.0", "42729", "6"): "6 .. SixthDay",
        ("FIX.5.0", "42729", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "42729", "8"): "8 .. EighthDay",
        ("FIX.5.0", "42729", "9"): "9 .. NinthDay",
        ("FIX.5.0", "42729", "10"): "10 .. TenthDay",
        ("FIX.5.0", "42729", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "42729", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "42729", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "42729", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "42729", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "42729", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "42729", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "42729", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "42729", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "42729", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "42729", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "42729", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "42729", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "42729", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "42729", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "42729", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "42729", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "42729", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "42729", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "42729", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "42729", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "42729", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "42729", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "42729", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "42729", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "42729", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "42729", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "42729", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "42729", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "42729", "MON"): "MON .. MON",
        ("FIX.5.0", "42729", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "42729", "WED"): "WED .. WED",
        ("FIX.5.0", "42729", "THU"): "THU .. THU",
        ("FIX.5.0", "42729", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "42729", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "42729", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "42730", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42730", "1"): "1 .. None",
        ("FIX.5.0", "42730", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42730", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42730", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42730", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42730", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42730", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42734", "M"): "M .. Multiply",
        ("FIX.5.0", "42734", "D"): "D .. Divide",
        ("FIX.5.0", "42736", "0"): "0 .. Initial",
        ("FIX.5.0", "42736", "1"): "1 .. Interim",
        ("FIX.5.0", "42736", "2"): "2 .. Final",
        ("FIX.5.0", "42737", "1"): "1 .. PerUnit",
        ("FIX.5.0", "42737", "2"): "2 .. Percent",
        ("FIX.5.0", "42737", "3"): "3 .. Absolute",
        ("FIX.5.0", "42737", "4"): "4 .. PercentageWaivedCashDiscount",
        ("FIX.5.0", "42737", "5"): "5 .. PercentageWaivedEnhancedUnits",
        ("FIX.5.0", "42737", "6"): "6 .. PointsPerBondOrContract",
        ("FIX.5.0", "42737", "7"): "7 .. BasisPoints",
        ("FIX.5.0", "42737", "8"): "8 .. AmountPerContract",
        ("FIX.5.0", "42745", "0"): "0 .. Bid",
        ("FIX.5.0", "42745", "1"): "1 .. Mid",
        ("FIX.5.0", "42745", "2"): "2 .. Offer",
        ("FIX.5.0", "42748", "0"): "0 .. Open",
        ("FIX.5.0", "42748", "1"): "1 .. OfficialSettlPx",
        ("FIX.5.0", "42748", "2"): "2 .. Xetra",
        ("FIX.5.0", "42748", "3"): "3 .. Close",
        ("FIX.5.0", "42748", "4"): "4 .. DerivativesClose",
        ("FIX.5.0", "42748", "5"): "5 .. High",
        ("FIX.5.0", "42748", "6"): "6 .. Low",
        ("FIX.5.0", "42748", "7"): "7 .. AsSpecifiedInMasterConfirmation",
        ("FIX.5.0", "42756", "0"): "0 .. Open",
        ("FIX.5.0", "42756", "1"): "1 .. OfficialSettlPx",
        ("FIX.5.0", "42756", "2"): "2 .. Xetra",
        ("FIX.5.0", "42756", "3"): "3 .. Close",
        ("FIX.5.0", "42756", "4"): "4 .. DerivativesClose",
        ("FIX.5.0", "42756", "5"): "5 .. High",
        ("FIX.5.0", "42756", "6"): "6 .. Low",
        ("FIX.5.0", "42756", "7"): "7 .. AsSpecifiedInMasterConfirmation",
        ("FIX.5.0", "42759", "0"): "0 .. None",
        ("FIX.5.0", "42759", "1"): "1 .. FuturesPrice",
        ("FIX.5.0", "42759", "2"): "2 .. OptionsPrice",
        ("FIX.5.0", "42760", "0"): "0 .. Close",
        ("FIX.5.0", "42760", "1"): "1 .. HedgeElection",
        ("FIX.5.0", "42762", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "42762", "1"): "1 .. Reuters",
        ("FIX.5.0", "42762", "2"): "2 .. Telerate",
        ("FIX.5.0", "42762", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "42762", "99"): "99 .. Other",
        ("FIX.5.0", "42766", "0"): "0 .. Gross",
        ("FIX.5.0", "42766", "1"): "1 .. Net",
        ("FIX.5.0", "42766", "2"): "2 .. Accrued",
        ("FIX.5.0", "42766", "3"): "3 .. CleanNet",
        ("FIX.5.0", "42769", "0"): "0 .. AbsoluteTerms",
        ("FIX.5.0", "42769", "1"): "1 .. PercentageOfNotional",
        ("FIX.5.0", "42774", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42774", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42778", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42778", "1"): "1 .. None",
        ("FIX.5.0", "42778", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42778", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42778", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42778", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42778", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42778", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42781", "D"): "D .. Day",
        ("FIX.5.0", "42781", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42781", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42781", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42782", "0"): "0 .. Business",
        ("FIX.5.0", "42782", "1"): "1 .. Calendar",
        ("FIX.5.0", "42782", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42782", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42782", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42782", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42787", "0"): "0 .. Execution",
        ("FIX.5.0", "42787", "1"): "1 .. PortfolioRebalancing",
        ("FIX.5.0", "42787", "2"): "2 .. Standard",
        ("FIX.5.0", "42791", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42791", "1"): "1 .. None",
        ("FIX.5.0", "42791", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42791", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42791", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42791", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42791", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42791", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42794", "D"): "D .. Day",
        ("FIX.5.0", "42794", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42794", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42794", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42795", "0"): "0 .. Business",
        ("FIX.5.0", "42795", "1"): "1 .. Calendar",
        ("FIX.5.0", "42795", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42795", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42795", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42795", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42798", "0"): "0 .. Close",
        ("FIX.5.0", "42798", "1"): "1 .. Hedge",
        ("FIX.5.0", "42803", "D"): "D .. Day",
        ("FIX.5.0", "42803", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42803", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42803", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42806", "0"): "0 .. Short",
        ("FIX.5.0", "42806", "1"): "1 .. Long",
        ("FIX.5.0", "42807", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "42807", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "42809", "1"): "1 .. Buyer",
        ("FIX.5.0", "42809", "2"): "2 .. Seller",
        ("FIX.5.0", "42810", "1"): "1 .. Buyer",
        ("FIX.5.0", "42810", "2"): "2 .. Seller",
        ("FIX.5.0", "42812", "1"): "1 .. Buyer",
        ("FIX.5.0", "42812", "2"): "2 .. Seller",
        ("FIX.5.0", "42813", "1"): "1 .. Buyer",
        ("FIX.5.0", "42813", "2"): "2 .. Seller",
        ("FIX.5.0", "42815", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "42815", "1"): "1 .. RoundDown",
        ("FIX.5.0", "42815", "2"): "2 .. RoundUp",
        ("FIX.5.0", "42817", "0"): "0 .. Unweighted",
        ("FIX.5.0", "42817", "1"): "1 .. Weighted",
        ("FIX.5.0", "42818", "0"): "0 .. ZeroInterestRateMethod",
        ("FIX.5.0", "42818", "1"): "1 .. NegativeInterestRateMethod",
        ("FIX.5.0", "42821", "D"): "D .. Day",
        ("FIX.5.0", "42821", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42821", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42821", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42822", "0"): "0 .. Business",
        ("FIX.5.0", "42822", "1"): "1 .. Calendar",
        ("FIX.5.0", "42822", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42822", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42822", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42822", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42824", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42824", "1"): "1 .. None",
        ("FIX.5.0", "42824", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42824", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42824", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42824", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42824", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42824", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42827", "0"): "0 .. ExDate",
        ("FIX.5.0", "42827", "1"): "1 .. RecordDate",
        ("FIX.5.0", "42828", "0"): "0 .. RecordAmount",
        ("FIX.5.0", "42828", "1"): "1 .. ExAmount",
        ("FIX.5.0", "42828", "2"): "2 .. PaidAmount",
        ("FIX.5.0", "42828", "3"): "3 .. PerMasterConfirm",
        ("FIX.5.0", "42830", "1"): "1 .. Buyer",
        ("FIX.5.0", "42830", "2"): "2 .. Seller",
        ("FIX.5.0", "42831", "0"): "0 .. RecordAmount",
        ("FIX.5.0", "42831", "1"): "1 .. ExAmount",
        ("FIX.5.0", "42831", "2"): "2 .. PaidAmount",
        ("FIX.5.0", "42831", "3"): "3 .. PerMasterConfirm",
        ("FIX.5.0", "42835", "0"): "0 .. None",
        ("FIX.5.0", "42835", "1"): "1 .. Flat",
        ("FIX.5.0", "42835", "2"): "2 .. Straight",
        ("FIX.5.0", "42835", "3"): "3 .. SpreadExclusive",
        ("FIX.5.0", "42839", "0"): "0 .. PotentialAdjustment",
        ("FIX.5.0", "42839", "1"): "1 .. CashEquivalent",
        ("FIX.5.0", "42840", "0"): "0 .. EquityAmountReceiver",
        ("FIX.5.0", "42840", "1"): "1 .. CalculationAgent",
        ("FIX.5.0", "42848", "D"): "D .. Day",
        ("FIX.5.0", "42848", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42848", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42848", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42849", "0"): "0 .. Business",
        ("FIX.5.0", "42849", "1"): "1 .. Calendar",
        ("FIX.5.0", "42849", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42849", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42849", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42849", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42851", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42851", "1"): "1 .. None",
        ("FIX.5.0", "42851", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42851", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42851", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42851", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42851", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42851", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42868", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42868", "1"): "1 .. None",
        ("FIX.5.0", "42868", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42868", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42868", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42868", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42868", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42868", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42872", "D"): "D .. Day",
        ("FIX.5.0", "42872", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42872", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42872", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42873", "0"): "0 .. Business",
        ("FIX.5.0", "42873", "1"): "1 .. Calendar",
        ("FIX.5.0", "42873", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42873", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42873", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42873", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42878", "D"): "D .. Day",
        ("FIX.5.0", "42878", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42878", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42878", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42879", "0"): "0 .. Business",
        ("FIX.5.0", "42879", "1"): "1 .. Calendar",
        ("FIX.5.0", "42879", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42879", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42879", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42879", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42887", "1"): "1 .. Buy",
        ("FIX.5.0", "42887", "2"): "2 .. Sell",
        ("FIX.5.0", "42893", "0"): "0 .. Bid",
        ("FIX.5.0", "42893", "1"): "1 .. Mid",
        ("FIX.5.0", "42893", "2"): "2 .. Offer",
        ("FIX.5.0", "42894", "0"): "0 .. None",
        ("FIX.5.0", "42894", "1"): "1 .. LinearZeroYield",
        ("FIX.5.0", "42898", "0"): "0 .. None",
        ("FIX.5.0", "42898", "1"): "1 .. LinearZeroYield",
        ("FIX.5.0", "42899", "0"): "0 .. Initial",
        ("FIX.5.0", "42899", "1"): "1 .. InitialAndFinal",
        ("FIX.5.0", "42899", "2"): "2 .. Final",
        ("FIX.5.0", "42899", "3"): "3 .. AnyPeriod",
        ("FIX.5.0", "42903", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42903", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42904", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42904", "1"): "1 .. None",
        ("FIX.5.0", "42904", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42904", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42904", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42904", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42904", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42904", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42907", "D"): "D .. Day",
        ("FIX.5.0", "42907", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42907", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42907", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42908", "0"): "0 .. Business",
        ("FIX.5.0", "42908", "1"): "1 .. Calendar",
        ("FIX.5.0", "42908", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42908", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42908", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42908", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42911", "D"): "D .. Day",
        ("FIX.5.0", "42911", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42911", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42911", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42911", "H"): "H .. Hour",
        ("FIX.5.0", "42911", "Min"): "Min .. Minute",
        ("FIX.5.0", "42911", "S"): "S .. Second",
        ("FIX.5.0", "42911", "T"): "T .. Term",
        ("FIX.5.0", "42912", "1"): "1 .. FirstDay",
        ("FIX.5.0", "42912", "2"): "2 .. SecondDay",
        ("FIX.5.0", "42912", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "42912", "4"): "4 .. FourthDay",
        ("FIX.5.0", "42912", "5"): "5 .. FifthDay",
        ("FIX.5.0", "42912", "6"): "6 .. SixthDay",
        ("FIX.5.0", "42912", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "42912", "8"): "8 .. EighthDay",
        ("FIX.5.0", "42912", "9"): "9 .. NinthDay",
        ("FIX.5.0", "42912", "10"): "10 .. TenthDay",
        ("FIX.5.0", "42912", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "42912", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "42912", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "42912", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "42912", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "42912", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "42912", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "42912", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "42912", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "42912", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "42912", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "42912", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "42912", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "42912", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "42912", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "42912", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "42912", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "42912", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "42912", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "42912", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "42912", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "42912", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "42912", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "42912", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "42912", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "42912", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "42912", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "42912", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "42912", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "42912", "MON"): "MON .. MON",
        ("FIX.5.0", "42912", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "42912", "WED"): "WED .. WED",
        ("FIX.5.0", "42912", "THU"): "THU .. THU",
        ("FIX.5.0", "42912", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "42912", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "42912", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "42920", "D"): "D .. Day",
        ("FIX.5.0", "42920", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42920", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42920", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42921", "0"): "0 .. Business",
        ("FIX.5.0", "42921", "1"): "1 .. Calendar",
        ("FIX.5.0", "42921", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42921", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42921", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42921", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42925", "D"): "D .. Day",
        ("FIX.5.0", "42925", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42925", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42925", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42928", "0"): "0 .. Short",
        ("FIX.5.0", "42928", "1"): "1 .. Long",
        ("FIX.5.0", "42929", "0"): "0 .. BondEquivalentYield",
        ("FIX.5.0", "42929", "1"): "1 .. MoneyMarketYield",
        ("FIX.5.0", "42931", "1"): "1 .. Buyer",
        ("FIX.5.0", "42931", "2"): "2 .. Seller",
        ("FIX.5.0", "42932", "1"): "1 .. Buyer",
        ("FIX.5.0", "42932", "2"): "2 .. Seller",
        ("FIX.5.0", "42934", "1"): "1 .. Buyer",
        ("FIX.5.0", "42934", "2"): "2 .. Seller",
        ("FIX.5.0", "42935", "1"): "1 .. Buyer",
        ("FIX.5.0", "42935", "2"): "2 .. Seller",
        ("FIX.5.0", "42937", "0"): "0 .. RoundToNearest",
        ("FIX.5.0", "42937", "1"): "1 .. RoundDown",
        ("FIX.5.0", "42937", "2"): "2 .. RoundUp",
        ("FIX.5.0", "42939", "0"): "0 .. Unweighted",
        ("FIX.5.0", "42939", "1"): "1 .. Weighted",
        ("FIX.5.0", "42940", "0"): "0 .. ZeroInterestRateMethod",
        ("FIX.5.0", "42940", "1"): "1 .. NegativeInterestRateMethod",
        ("FIX.5.0", "42944", "D"): "D .. Day",
        ("FIX.5.0", "42944", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42944", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42944", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42945", "0"): "0 .. Business",
        ("FIX.5.0", "42945", "1"): "1 .. Calendar",
        ("FIX.5.0", "42945", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42945", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42945", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42945", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42952", "D"): "D .. Day",
        ("FIX.5.0", "42952", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42952", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42952", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42953", "0"): "0 .. Business",
        ("FIX.5.0", "42953", "1"): "1 .. Calendar",
        ("FIX.5.0", "42953", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42953", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42953", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42953", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42957", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "42957", "1"): "1 .. Adjusted",
        ("FIX.5.0", "42960", "0"): "0 .. Business",
        ("FIX.5.0", "42960", "1"): "1 .. Calendar",
        ("FIX.5.0", "42960", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42960", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42960", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42960", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42969", "0"): "0 .. Volatility",
        ("FIX.5.0", "42969", "1"): "1 .. Variance",
        ("FIX.5.0", "42974", "0"): "0 .. Previous",
        ("FIX.5.0", "42974", "1"): "1 .. Last",
        ("FIX.5.0", "42974", "2"): "2 .. Both",
        ("FIX.5.0", "42985", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42985", "1"): "1 .. None",
        ("FIX.5.0", "42985", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42985", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42985", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42985", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42985", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42985", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42988", "D"): "D .. Day",
        ("FIX.5.0", "42988", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42988", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42988", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42989", "0"): "0 .. Business",
        ("FIX.5.0", "42989", "1"): "1 .. Calendar",
        ("FIX.5.0", "42989", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42989", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42989", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42989", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "42994", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "42994", "1"): "1 .. None",
        ("FIX.5.0", "42994", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "42994", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "42994", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "42994", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "42994", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "42994", "7"): "7 .. NearestDay",
        ("FIX.5.0", "42997", "D"): "D .. Day",
        ("FIX.5.0", "42997", "Wk"): "Wk .. Week",
        ("FIX.5.0", "42997", "Mo"): "Mo .. Month",
        ("FIX.5.0", "42997", "Yr"): "Yr .. Year",
        ("FIX.5.0", "42998", "0"): "0 .. Business",
        ("FIX.5.0", "42998", "1"): "1 .. Calendar",
        ("FIX.5.0", "42998", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "42998", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "42998", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "42998", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "43002", "0"): "0 .. FlatFee",
        ("FIX.5.0", "43002", "1"): "1 .. AmortizedFee",
        ("FIX.5.0", "43002", "2"): "2 .. FundingFee",
        ("FIX.5.0", "43002", "3"): "3 .. FlatAndFundingFee",
        ("FIX.5.0", "43002", "4"): "4 .. AmortizedAndFundingFee",
        ("FIX.5.0", "43009", "0"): "0 .. PriceValuation",
        ("FIX.5.0", "43009", "1"): "1 .. DividendValuation",
        ("FIX.5.0", "43012", "D"): "D .. Day",
        ("FIX.5.0", "43012", "Wk"): "Wk .. Week",
        ("FIX.5.0", "43012", "Mo"): "Mo .. Month",
        ("FIX.5.0", "43012", "Yr"): "Yr .. Year",
        ("FIX.5.0", "43013", "0"): "0 .. Business",
        ("FIX.5.0", "43013", "1"): "1 .. Calendar",
        ("FIX.5.0", "43013", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "43013", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "43013", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "43013", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "43017", "D"): "D .. Day",
        ("FIX.5.0", "43017", "Wk"): "Wk .. Week",
        ("FIX.5.0", "43017", "Mo"): "Mo .. Month",
        ("FIX.5.0", "43017", "Yr"): "Yr .. Year",
        ("FIX.5.0", "43018", "0"): "0 .. Business",
        ("FIX.5.0", "43018", "1"): "1 .. Calendar",
        ("FIX.5.0", "43018", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "43018", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "43018", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "43018", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "43023", "D"): "D .. Day",
        ("FIX.5.0", "43023", "Wk"): "Wk .. Week",
        ("FIX.5.0", "43023", "Mo"): "Mo .. Month",
        ("FIX.5.0", "43023", "Yr"): "Yr .. Year",
        ("FIX.5.0", "43024", "0"): "0 .. Business",
        ("FIX.5.0", "43024", "1"): "1 .. Calendar",
        ("FIX.5.0", "43024", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "43024", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "43024", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "43024", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "43027", "D"): "D .. Day",
        ("FIX.5.0", "43027", "Wk"): "Wk .. Week",
        ("FIX.5.0", "43027", "Mo"): "Mo .. Month",
        ("FIX.5.0", "43027", "Yr"): "Yr .. Year",
        ("FIX.5.0", "43027", "H"): "H .. Hour",
        ("FIX.5.0", "43027", "Min"): "Min .. Minute",
        ("FIX.5.0", "43027", "S"): "S .. Second",
        ("FIX.5.0", "43027", "T"): "T .. Term",
        ("FIX.5.0", "43028", "1"): "1 .. FirstDay",
        ("FIX.5.0", "43028", "2"): "2 .. SecondDay",
        ("FIX.5.0", "43028", "3"): "3 .. ThirdDay",
        ("FIX.5.0", "43028", "4"): "4 .. FourthDay",
        ("FIX.5.0", "43028", "5"): "5 .. FifthDay",
        ("FIX.5.0", "43028", "6"): "6 .. SixthDay",
        ("FIX.5.0", "43028", "7"): "7 .. SeventhDay",
        ("FIX.5.0", "43028", "8"): "8 .. EighthDay",
        ("FIX.5.0", "43028", "9"): "9 .. NinthDay",
        ("FIX.5.0", "43028", "10"): "10 .. TenthDay",
        ("FIX.5.0", "43028", "11"): "11 .. EleventhDay",
        ("FIX.5.0", "43028", "12"): "12 .. TwelvthDay",
        ("FIX.5.0", "43028", "13"): "13 .. ThirteenthDay",
        ("FIX.5.0", "43028", "14"): "14 .. ForteenthDay",
        ("FIX.5.0", "43028", "15"): "15 .. FifteenthDay",
        ("FIX.5.0", "43028", "16"): "16 .. SixteenthDay",
        ("FIX.5.0", "43028", "17"): "17 .. SeventeenthDay",
        ("FIX.5.0", "43028", "18"): "18 .. EighteenthDay",
        ("FIX.5.0", "43028", "19"): "19 .. NineteenthDay",
        ("FIX.5.0", "43028", "20"): "20 .. TwentiethDay",
        ("FIX.5.0", "43028", "21"): "21 .. TwentyFirstDay",
        ("FIX.5.0", "43028", "22"): "22 .. TwentySecondDay",
        ("FIX.5.0", "43028", "23"): "23 .. TwentyThirdDay",
        ("FIX.5.0", "43028", "24"): "24 .. TwentyFourthDay",
        ("FIX.5.0", "43028", "25"): "25 .. TwentyFifthDay",
        ("FIX.5.0", "43028", "26"): "26 .. TwentySixthDay",
        ("FIX.5.0", "43028", "27"): "27 .. TwentySeventhDay",
        ("FIX.5.0", "43028", "28"): "28 .. TwentyEigthDa28y",
        ("FIX.5.0", "43028", "29"): "29 .. TwentyNinthDay",
        ("FIX.5.0", "43028", "30"): "30 .. ThirtiethDay",
        ("FIX.5.0", "43028", "EOM"): "EOM .. EOM",
        ("FIX.5.0", "43028", "FRN"): "FRN .. FRN",
        ("FIX.5.0", "43028", "IMM"): "IMM .. IMM",
        ("FIX.5.0", "43028", "IMMCAD"): "IMMCAD .. IMMCAD",
        ("FIX.5.0", "43028", "IMMAUD"): "IMMAUD .. IMMAUD",
        ("FIX.5.0", "43028", "IMMNZD"): "IMMNZD .. IMMNZD",
        ("FIX.5.0", "43028", "SFE"): "SFE .. SFE",
        ("FIX.5.0", "43028", "NONE"): "NONE .. NONE",
        ("FIX.5.0", "43028", "TBILL"): "TBILL .. TBILL",
        ("FIX.5.0", "43028", "MON"): "MON .. MON",
        ("FIX.5.0", "43028", "TUE"): "TUE .. TUE",
        ("FIX.5.0", "43028", "WED"): "WED .. WED",
        ("FIX.5.0", "43028", "THU"): "THU .. THU",
        ("FIX.5.0", "43028", "FRI"): "FRI .. FRI",
        ("FIX.5.0", "43028", "SAT"): "SAT .. SAT",
        ("FIX.5.0", "43028", "SUN"): "SUN .. SUN",
        ("FIX.5.0", "43029", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "43029", "1"): "1 .. None",
        ("FIX.5.0", "43029", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "43029", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "43029", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "43029", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "43029", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "43029", "7"): "7 .. NearestDay",
        ("FIX.5.0", "43033", "M"): "M .. Multiply",
        ("FIX.5.0", "43033", "D"): "D .. Divide",
        ("FIX.5.0", "43035", "0"): "0 .. Initial",
        ("FIX.5.0", "43035", "1"): "1 .. Interim",
        ("FIX.5.0", "43035", "2"): "2 .. Final",
        ("FIX.5.0", "43036", "1"): "1 .. PerUnit",
        ("FIX.5.0", "43036", "2"): "2 .. Percent",
        ("FIX.5.0", "43036", "3"): "3 .. Absolute",
        ("FIX.5.0", "43036", "4"): "4 .. PercentageWaivedCashDiscount",
        ("FIX.5.0", "43036", "5"): "5 .. PercentageWaivedEnhancedUnits",
        ("FIX.5.0", "43036", "6"): "6 .. PointsPerBondOrContract",
        ("FIX.5.0", "43036", "7"): "7 .. BasisPoints",
        ("FIX.5.0", "43036", "8"): "8 .. AmountPerContract",
        ("FIX.5.0", "43044", "0"): "0 .. Bid",
        ("FIX.5.0", "43044", "1"): "1 .. Mid",
        ("FIX.5.0", "43044", "2"): "2 .. Offer",
        ("FIX.5.0", "43047", "0"): "0 .. Open",
        ("FIX.5.0", "43047", "1"): "1 .. OfficialSettlPx",
        ("FIX.5.0", "43047", "2"): "2 .. Xetra",
        ("FIX.5.0", "43047", "3"): "3 .. Close",
        ("FIX.5.0", "43047", "4"): "4 .. DerivativesClose",
        ("FIX.5.0", "43047", "5"): "5 .. High",
        ("FIX.5.0", "43047", "6"): "6 .. Low",
        ("FIX.5.0", "43047", "7"): "7 .. AsSpecifiedInMasterConfirmation",
        ("FIX.5.0", "43055", "0"): "0 .. Open",
        ("FIX.5.0", "43055", "1"): "1 .. OfficialSettlPx",
        ("FIX.5.0", "43055", "2"): "2 .. Xetra",
        ("FIX.5.0", "43055", "3"): "3 .. Close",
        ("FIX.5.0", "43055", "4"): "4 .. DerivativesClose",
        ("FIX.5.0", "43055", "5"): "5 .. High",
        ("FIX.5.0", "43055", "6"): "6 .. Low",
        ("FIX.5.0", "43055", "7"): "7 .. AsSpecifiedInMasterConfirmation",
        ("FIX.5.0", "43058", "0"): "0 .. None",
        ("FIX.5.0", "43058", "1"): "1 .. FuturesPrice",
        ("FIX.5.0", "43058", "2"): "2 .. OptionsPrice",
        ("FIX.5.0", "43059", "0"): "0 .. Close",
        ("FIX.5.0", "43059", "1"): "1 .. HedgeElection",
        ("FIX.5.0", "43061", "0"): "0 .. Bloomberg",
        ("FIX.5.0", "43061", "1"): "1 .. Reuters",
        ("FIX.5.0", "43061", "2"): "2 .. Telerate",
        ("FIX.5.0", "43061", "3"): "3 .. ISDARateOption",
        ("FIX.5.0", "43061", "99"): "99 .. Other",
        ("FIX.5.0", "43065", "0"): "0 .. Gross",
        ("FIX.5.0", "43065", "1"): "1 .. Net",
        ("FIX.5.0", "43065", "2"): "2 .. Accrued",
        ("FIX.5.0", "43065", "3"): "3 .. CleanNet",
        ("FIX.5.0", "43068", "0"): "0 .. AbsoluteTerms",
        ("FIX.5.0", "43068", "1"): "1 .. PercentageOfNotional",
        ("FIX.5.0", "43073", "0"): "0 .. Unadjusted",
        ("FIX.5.0", "43073", "1"): "1 .. Adjusted",
        ("FIX.5.0", "43077", "0"): "0 .. NotApplicable",
        ("FIX.5.0", "43077", "1"): "1 .. None",
        ("FIX.5.0", "43077", "2"): "2 .. FollowingDay",
        ("FIX.5.0", "43077", "3"): "3 .. FloatingRateNote",
        ("FIX.5.0", "43077", "4"): "4 .. ModifiedFollowingDay",
        ("FIX.5.0", "43077", "5"): "5 .. PrecedingDay",
        ("FIX.5.0", "43077", "6"): "6 .. ModifiedPrecedingDay",
        ("FIX.5.0", "43077", "7"): "7 .. NearestDay",
        ("FIX.5.0", "43080", "D"): "D .. Day",
        ("FIX.5.0", "43080", "Wk"): "Wk .. Week",
        ("FIX.5.0", "43080", "Mo"): "Mo .. Month",
        ("FIX.5.0", "43080", "Yr"): "Yr .. Year",
        ("FIX.5.0", "43081", "0"): "0 .. Business",
        ("FIX.5.0", "43081", "1"): "1 .. Calendar",
        ("FIX.5.0", "43081", "2"): "2 .. CommodityBusiness",
        ("FIX.5.0", "43081", "3"): "3 .. CurrencyBusiness",
        ("FIX.5.0", "43081", "4"): "4 .. ExchangeBusiness",
        ("FIX.5.0", "43081", "5"): "5 .. ScheduledTradingDay",
        ("FIX.5.0", "43086", "0"): "0 .. Execution",
        ("FIX.5.0", "43086", "1"): "1 .. PortfolioRebalancing",
        ("FIX.5.0", "43086", "2"): "2 .. Standard",
        ("FIX.5.0", "43089", "1"): "1 .. CUSIP",
        ("FIX.5.0", "43089", "2"): "2 .. SEDOL",
        ("FIX.5.0", "43089", "3"): "3 .. QUIK",
        ("FIX.5.0", "43089", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "43089", "5"): "5 .. RICCode",
        ("FIX.5.0", "43089", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "43089", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "43089", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "43089", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "43089", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "43089", "B"): "B .. Wertpapier",
        ("FIX.5.0", "43089", "C"): "C .. Dutch",
        ("FIX.5.0", "43089", "D"): "D .. Valoren",
        ("FIX.5.0", "43089", "E"): "E .. Sicovam",
        ("FIX.5.0", "43089", "F"): "F .. Belgian",
        ("FIX.5.0", "43089", "G"): "G .. Common",
        ("FIX.5.0", "43089", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "43089", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "43089", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "43089", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "43089", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "43089", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "43089", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "43089", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "43089", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "43089", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "43089", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "43089", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "43089", "U"): "U .. Synthetic",
        ("FIX.5.0", "43089", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "43089", "W"): "W .. IndexName",
        ("FIX.5.0", "43091", "1"): "1 .. CUSIP",
        ("FIX.5.0", "43091", "2"): "2 .. SEDOL",
        ("FIX.5.0", "43091", "3"): "3 .. QUIK",
        ("FIX.5.0", "43091", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "43091", "5"): "5 .. RICCode",
        ("FIX.5.0", "43091", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "43091", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "43091", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "43091", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "43091", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "43091", "B"): "B .. Wertpapier",
        ("FIX.5.0", "43091", "C"): "C .. Dutch",
        ("FIX.5.0", "43091", "D"): "D .. Valoren",
        ("FIX.5.0", "43091", "E"): "E .. Sicovam",
        ("FIX.5.0", "43091", "F"): "F .. Belgian",
        ("FIX.5.0", "43091", "G"): "G .. Common",
        ("FIX.5.0", "43091", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "43091", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "43091", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "43091", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "43091", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "43091", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "43091", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "43091", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "43091", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "43091", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "43091", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "43091", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "43091", "U"): "U .. Synthetic",
        ("FIX.5.0", "43091", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "43091", "W"): "W .. IndexName",
        ("FIX.5.0", "43093", "1"): "1 .. CUSIP",
        ("FIX.5.0", "43093", "2"): "2 .. SEDOL",
        ("FIX.5.0", "43093", "3"): "3 .. QUIK",
        ("FIX.5.0", "43093", "4"): "4 .. ISINNumber",
        ("FIX.5.0", "43093", "5"): "5 .. RICCode",
        ("FIX.5.0", "43093", "6"): "6 .. ISOCurrencyCode",
        ("FIX.5.0", "43093", "7"): "7 .. ISOCountryCode",
        ("FIX.5.0", "43093", "8"): "8 .. ExchangeSymbol",
        ("FIX.5.0", "43093", "9"): "9 .. ConsolidatedTapeAssociation",
        ("FIX.5.0", "43093", "A"): "A .. BloombergSymbol",
        ("FIX.5.0", "43093", "B"): "B .. Wertpapier",
        ("FIX.5.0", "43093", "C"): "C .. Dutch",
        ("FIX.5.0", "43093", "D"): "D .. Valoren",
        ("FIX.5.0", "43093", "E"): "E .. Sicovam",
        ("FIX.5.0", "43093", "F"): "F .. Belgian",
        ("FIX.5.0", "43093", "G"): "G .. Common",
        ("FIX.5.0", "43093", "H"): "H .. ClearingHouse",
        ("FIX.5.0", "43093", "I"): "I .. ISDAFpMLSpecification",
        ("FIX.5.0", "43093", "J"): "J .. OptionPriceReportingAuthority",
        ("FIX.5.0", "43093", "K"): "K .. ISDAFpMLURL",
        ("FIX.5.0", "43093", "L"): "L .. LetterOfCredit",
        ("FIX.5.0", "43093", "M"): "M .. MarketplaceAssignedIdentifier",
        ("FIX.5.0", "43093", "N"): "N .. MarkitREDEntityCLIP",
        ("FIX.5.0", "43093", "P"): "P .. MarkitREDPairCLIP",
        ("FIX.5.0", "43093", "Q"): "Q .. CFTCCommodityCode",
        ("FIX.5.0", "43093", "R"): "R .. ISDACommodityReferencePrice",
        ("FIX.5.0", "43093", "S"): "S .. FinancialInstrumentGlobalIdentifier",
        ("FIX.5.0", "43093", "T"): "T .. LegalEntityIdentifier",
        ("FIX.5.0", "43093", "U"): "U .. Synthetic",
        ("FIX.5.0", "43093", "V"): "V .. FidessaInstrumentMnemonic",
        ("FIX.5.0", "43093", "W"): "W .. IndexName",
        ("FIX.5.0", "50002", "0"): "0 .. Update",
        ("FIX.5.0", "50002", "1"): "1 .. Snapshot",
    }

    __MARKET_MAP = {
        ("RIC","0"): "0 .. None",
        ("RIC","A"): "A .. American Stock Exchange",
        ("RIC","AM"): "AM .. Amman Stock Exchange",
        ("RIC","AS"): "AS .. AEX Stock Exchange",
        ("RIC","AX"): "AX .. Australian Stock Exchange",
        ("RIC","B"): "B .. Belfox",
        ("RIC","B"): "B .. Boston Stock Exchange",
        ("RIC","BC"): "BC .. Barcelona Stock Exchange - Floor Trading",
        ("RIC","BE"): "BE .. Berlin Stock Exchange",
        ("RIC","BH"): "BH .. Bahrain Stock Exchange",
        ("RIC","BI"): "BI .. Bilbao Stock Exchange",
        ("RIC","BK"): "BK .. Thailand Stock Exchange",
        ("RIC","BM"): "BM .. Bremen Stock Exchange",
        ("RIC","BN"): "BN .. Berne Stock Exchange",
        ("RIC","BO"): "BO .. Bombay Stock Exchange",
        ("RIC","BR"): "BR .. Brussels Stock Exchange",
        ("RIC","BT"): "BT .. Botswana Share Market",
        ("RIC","BY"): "BY .. Beirut Stock Exchange",
        ("RIC","C"): "C .. Cincinnati Stock Exchange",
        ("RIC","CH"): "CH .. Channel Islands",
        ("RIC","CI"): "CI .. Abidjan Stock Exchange",
        ("RIC","CL"): "CL .. Calcutta Stock Exchange",
        ("RIC","CM"): "CM .. Colombo Stock Exchange",
        ("RIC","CO"): "CO .. Copenhagen Stock Exchange",
        ("RIC","D"): "D .. Dusseldorf Stock Exchange",
        ("RIC","DL"): "DL .. Dehli Stock Exchange",
        ("RIC","DU"): "DU .. Dubai Financial Market",
        ("RIC","E"): "E .. AEX Options and Futures Exchange",
        ("RIC","EB"): "EB .. Electronic Stock Exchange of Venezuela",
        ("RIC","F"): "F .. Frankfurt Stock Exchange",
        ("RIC","FU"): "FU .. Fukuoka Stock Exchange",
        ("RIC","GH"): "GH .. Ghana Stock Exchange",
        ("RIC","H"): "H .. Hamburg Stock Exchange",
        ("RIC","HA"): "HA .. Hannover Stock Exchange",
        ("RIC","HE"): "HE .. Helsinki Stock Exchange",
        ("RIC","HK"): "HK .. Hong Kong Stock Exchange",
        ("RIC","I"): "I .. Irish Stock Exchange",
        ("RIC","IC"): "IC .. Iceland Stock Exchange",
        ("RIC","IN"): "IN .. Interbolsa (Portugal)",
        ("RIC","IS"): "IS .. Istanbul Stock Exchange",
        ("RIC","J"): "J .. Johannesburg Stock Exchange",
        ("RIC","JK"): "JK .. Jakarta Stock Exchange",
        ("RIC","K"): "K .. Toronto Options Exchange",
        ("RIC","KA"): "KA .. Karachi Stock Exchange",
        ("RIC","KL"): "KL .. Kuala Lumpur Stock Exchange",
        ("RIC","KQ"): "KQ .. KOSDAQ (Korea)",
        ("RIC","KS"): "KS .. Korea Stock Exchange",
        ("RIC","KW"): "KW .. Kuwait Stock Exchange",
        ("RIC","KY"): "KY .. Kyoto Stock Exchange",
        ("RIC","KZ"): "KZ .. Kazakhstan Stock Exchange",
        ("RIC","L"): "L .. London Stock Exchange",
        ("RIC","LA"): "LA .. Latin American Market In Spain (LATIBEX)",
        ("RIC","LG"): "LG .. Lagos Stock Exchange",
        ("RIC","LM"): "LM .. Lima Stock Exchange",
        ("RIC","LN"): "LN .. Le Nouveau Marche",
        ("RIC","LS"): "LS .. Lisbon Stock Exchange (Portugal)",
        ("RIC","LU"): "LU .. Luxembourg Stock Exchange",
        ("RIC","LZ"): "LZ .. Lusaka Stock Exchange",
        ("RIC","M"): "M .. Montreal Exchange",
        ("RIC","MA"): "MA .. Madrid Stock Exchange - Floor Trading",
        ("RIC","MC"): "MC .. Madrid Stock Exchange - CATS Feed",
        ("RIC","MD"): "MD .. Madras Stock Exchange",
        ("RIC","MI"): "MI .. Milan Stock Exchange",
        ("RIC","ML"): "ML .. Medellin Stock Excahnge",
        ("RIC","MM"): "MM .. Moscow Inter Bank Currency Exchange",
        ("RIC","MO"): "MO .. Moscow Stock Exchange",
        ("RIC","MT"): "MT .. Malta Stock Exchange",
        ("RIC","MU"): "MU .. Munich Stock Exchange",
        ("RIC","MW"): "MW .. Chicago Stock Exchange",
        ("RIC","MX"): "MX .. Mexican Stock Exchange",
        ("RIC","MZ"): "MZ .. Mauritius Stock Exchange",
        ("RIC","N"): "N .. New York Stock Exchange",
        ("RIC","NG"): "NG .. Nagoya Stock Exchange",
        ("RIC","NM"): "NM .. Namibia Stock Exchange",
        ("RIC","NR"): "NR .. Nairobi Stock Exchange",
        ("RIC","NS"): "NS .. National Stock Exchange of India",
        ("RIC","NW"): "NW .. NewEx (Austria)",
        ("RIC","NZ"): "NZ .. New Zealand Stock Exchange",
        ("RIC","O"): "O .. NASDAQ",
        ("RIC","OB"): "OB .. NASDAQ Dealers - Bulletin Board",
        ("RIC","OD"): "OD .. Occidente Stock Exchange",
        ("RIC","OJ"): "OJ .. NASDAQ Japan",
        ("RIC","OL"): "OL .. Oslo Stock Exchange",
        ("RIC","OM"): "OM .. Muscat Stock Exchange",
        ("RIC","OS"): "OS .. Osaka Stock Exchange",
        ("RIC","P"): "P .. Pacific Stock Exchange",
        ("RIC","PA"): "PA .. Paris Stock Exchange",
        ("RIC","PE"): "PE .. St. Petersburg Stock Exchange",
        ("RIC","PFT"): "PFT .. Ukraine PFTS",
        ("RIC","PH"): "PH .. Philadelphia Stock Exchange",
        ("RIC","PNK"): "PNK .. Pink Sheets (National Quotation Bureau)",
        ("RIC","PR"): "PR .. Prague Stock Exchange",
        ("RIC","PS"): "PS .. Philippine Stock Exchange",
        ("RIC","Q"): "Q .. Japanese Securities Dealers Association (JASDAQ)",
        ("RIC","QA"): "QA .. Doha Securities Market",
        ("RIC","RI"): "RI .. Riga Stock Exchange",
        ("RIC","RQ"): "RQ .. RASDAQ (Romania)",
        ("RIC","RTS"): "RTS .. Russian Trading System",
        ("RIC","S"): "S .. SWX Swiss Exchange",
        ("RIC","SA"): "SA .. Sao Paulo Stock Exchange",
        ("RIC","SBI"): "SBI .. SBI Stock Exchange (Sweden)",
        ("RIC","SE"): "SE .. Saudi Stock Exchange",
        ("RIC","SG"): "SG .. Stuttgart Stock Exchange",
        ("RIC","SI"): "SI .. Singapore Stock Exchange",
        ("RIC","SN"): "SN .. Santiago Stock Exchange",
        ("RIC","SO"): "SO .. Rio de Janeiro OTC Stock Exchange (SOMA)",
        ("RIC","SP"): "SP .. Sapporo Stock Exchange",
        ("RIC","SS"): "SS .. Shanghai Stock Exchange",
        ("RIC","ST"): "ST .. Stockholm Stock Exchange",
        ("RIC","SU"): "SU .. Surabaya Stock Exchange",
        ("RIC","SZ"): "SZ .. Shenzhen Stock Exchange",
        ("RIC","T"): "T .. Tokyo Stock Exchange",
        ("RIC","TA"): "TA .. Tel Aviv Stock Exchange",
        ("RIC","TH"): "TH .. Third Market",
        ("RIC","TL"): "TL .. Tallinn Stock Exchange",
        ("RIC","TN"): "TN .. Tunis Stock Exchange",
        ("RIC","TO"): "TO .. Toronto Stock Exchange",
        ("RIC","TP"): "TP .. Tradepoint Stock Exchange",
        ("RIC","TW"): "TW .. Taiwan Stock Exchange",
        ("RIC","TWO"): "TWO .. Taiwan OTC Securities Exchange",
        ("RIC","V"): "V .. Canadian Ventures Exchange",
        ("RIC","VA"): "VA .. Valencia Stock Exchange",
        ("RIC","VL"): "VL .. Vilnus Stock Exchange",
        ("RIC","VX"): "VX .. virt-x",
        ("RIC","W"): "W .. Chicago Board Options Exchange",
        ("RIC","X"): "X .. Philadelphia Stock Exchange Options",
        ("RIC","Y"): "Y .. International Securities Exchange (ISE)",
        ("RIC","Z"): "Z .. Eurex Switzerland (SFF)",
        ("RIC","ZI"): "ZI .. Zimbabwe Stock Exchange",
        ("RIC","d"): "d .. Eurex Germany (DTB)",
        ("RIC","p"): "p .. MONEP Paris Stock Options",
        ("RIC","1"): "1 .. American Stock Exchange Options",
        ("RIC","2"): "2 .. Chicago Mercantile Exchange (CME)",
        ("RIC","3"): "3 .. London International Financial Futures Exchange (LIFFE)",
        ("RIC","4"): "4 .. POSIT",
        ("RIC","5"): "5 .. London Traded Options Market",
        ("RIC","6"): "6 .. Montreal Exchange Options (MOE)",
        ("RIC","8"): "8 .. Pacific Stock Exchange Options (PAO)",
        ("RIC","9"): "9 .. Vancouver Options Exchange (VAO)",
        ("RIC","10"): "10 .. NYSE BBSS (broker booth system)",
        ("RIC","11"): "11 .. Non-exchange-based Over The Counter Market",
        ("RIC","12"): "12 .. New York Mercantile Exchange (NYMEX)",
        ("RIC","13"): "13 .. NYFIX Millennium",
        ("RIC","14"): "14 .. Jiway",
        ("RIC","15"): "15 .. International Securities Market Association(ISMA)",
        ("RIC","16"): "16 .. MEFF Renta Variable",
        ("RIC","17"): "17 .. Stockholm Options Market",
        ("RIC","30"): "30 .. TradeWeb",
        ("RIC","31"): "31 .. Bloomberg TradeBook",
        ("RIC","32"): "32 .. BondBook",
        ("RIC","33"): "33 .. MarketAxess",
        ("RIC","34"): "34 .. MuniCenter",
        ("RIC","35"): "35 .. BondClick",
        ("RIC","36"): "36 .. BondHub",
        ("RIC","37"): "37 .. LIMITrader",
        ("RIC","38"): "38 .. Visible Markets",
        ("RIC","39"): "39 .. Archipelago ECN",
        ("RIC","40"): "40 .. ATTAIN ECN",
        ("RIC","41"): "41 .. BRUT ECN",
        ("RIC","42"): "42 .. GlobeNet ECN",
        ("RIC","43"): "43 .. Instinet ECN",
        ("RIC","44"): "44 .. Island ECN",
        ("RIC","45"): "45 .. MarketXT ECN",
        ("RIC","46"): "46 .. NexTrade ECN",
        ("RIC","47"): "47 .. REDIBook ECN",
        ("RIC","48"): "48 .. Intercontinental Exchange",
        ("RIC","49"): "49 .. NQLX",
        ("RIC","50"): "50 .. OneChicago",
        ("MIC","3579"): "3579 .. SSY FUTURES LTD -  FREIGHT SCREEN",
        ("MIC","360T"): "360T .. 360T",
        ("MIC","A2XX"): "A2XX .. A2X",
        ("MIC","AACA"): "AACA .. CREDIT AGRICOLE CIB - SYSTEMATIC INTERNALISER",
        ("MIC","AAPA"): "AAPA .. ATHENS EXCHANGE - APA",
        ("MIC","AATS"): "AATS .. ASSENT ATS",
        ("MIC","ABNA"): "ABNA .. ABN AMRO BANK NV - SYSTEMATIC INTERNALISER",
        ("MIC","ABSI"): "ABSI .. ALM. BRAND BANK - SYSTEMATIC INTERNALISER",
        ("MIC","ABUL"): "ABUL .. BULGARIAN STOCK EXCHANGE - ALTERNATIVE MARKET",
        ("MIC","ACEX"): "ACEX .. ACE DERIVATIVES & COMMODITY EXCHANGE LTD",
        ("MIC","ACKF"): "ACKF .. KCG ACKNOWLEDGE FI",
        ("MIC","ADVT"): "ADVT .. ADVISE TECHNOLOGIES - APA TRANSPARENCY REPORTING",
        ("MIC","AFDL"): "AFDL .. ABIDE FINANCIAL DRSP LIMITED APA",
        ("MIC","AFET"): "AFET .. AGRICULTURAL FUTURES EXCHANGE OF THAILAND",
        ("MIC","AFSA"): "AFSA .. AFS - OTF",
        ("MIC","AFSI"): "AFSI .. AFS - OTF - INTEREST RATE DERIVATIVES",
        ("MIC","AFSL"): "AFSL .. AFS - OTF - STRUCTURED PRODUCTS",
        ("MIC","AFSO"): "AFSO .. AFS - OTF - BONDS",
        ("MIC","AFSX"): "AFSX .. AFS - OTF - FX FORWARDS",
        ("MIC","AIMX"): "AIMX .. LONDON STOCK EXCHANGE - AIM MTF",
        ("MIC","AIXK"): "AIXK .. ASTANA INTERNATIONAL EXCHANGE LTD",
        ("MIC","AKIS"): "AKIS .. AKIS - BANCA AKROS - SYSTEMATIC INTERNALISER",
        ("MIC","ALDP"): "ALDP .. NYSE ALTERNEXT DARK",
        ("MIC","ALLT"): "ALLT .. ALLT-OTF",
        ("MIC","ALSI"): "ALSI .. AKTIESELSKABET ARBEJDERNES LANDSBANK - SYSTEMATIC INTERNALISER",
        ("MIC","ALTX"): "ALTX .. JSE ALTERNATE EXCHANGE",
        ("MIC","ALXB"): "ALXB .. EURONEXT GROWTH BRUSSELS",
        ("MIC","ALXL"): "ALXL .. EURONEXT GROWTH LISBON",
        ("MIC","ALXP"): "ALXP .. EURONEXT GROWTH PARIS",
        ("MIC","AMPX"): "AMPX .. ASSET MATCH PRIVATE EXCHANGE",
        ("MIC","AMTS"): "AMTS .. MTS NETHERLANDS",
        ("MIC","AMXO"): "AMXO .. NYSE AMEX OPTIONS",
        ("MIC","ANLP"): "ANLP .. AUSTRALIA AND NEW ZEALAND BANKING GROUP LIMITED",
        ("MIC","ANTS"): "ANTS .. ABBEY NATIONAL TREASURY SERVICES PLC - SYSTEMATIC INTERNALISER",
        ("MIC","ANZL"): "ANZL .. AUSTRALIA AND NEW ZEALAND BANKING GROUP LIMITED SYSTEMATIC INTERNALISER",
        ("MIC","APAW"): "APAW .. WIENER BOERSE AG - APA",
        ("MIC","APCL"): "APCL .. ASIA PACIFIC CLEAR",
        ("MIC","APEX"): "APEX .. ASIA PACIFIC EXCHANGE",
        ("MIC","APXL"): "APXL .. SYDNEY STOCK EXCHANGE LIMITED",
        ("MIC","AQEA"): "AQEA .. AQUIS SA AUCTION ON DEMAND (AOD)",
        ("MIC","AQEU"): "AQEU .. AQUIS SA",
        ("MIC","AQUA"): "AQUA .. AQUA EQUITIES L.P.",
        ("MIC","AQXA"): "AQXA .. AQUIS AUCTION ON DEMAND (AOD)",
        ("MIC","AQXE"): "AQXE .. AQUIS EXCHANGE",
        ("MIC","ARAX"): "ARAX .. ARAX COMMODITIES LTD",
        ("MIC","ARCD"): "ARCD .. ARCA DARK",
        ("MIC","ARCO"): "ARCO .. NYSE ARCA OPTIONS",
        ("MIC","ARCX"): "ARCX .. NYSE ARCA",
        ("MIC","AREX"): "AREX .. AREX - AUTOMATED RECEIVABLES EXCHANGE",
        ("MIC","ARIA"): "ARIA .. ARIAN TRADING FACILITY",
        ("MIC","ASEX"): "ASEX .. ATHENS STOCK EXCHANGE",
        ("MIC","ASTR"): "ASTR .. CLEARCORP DEALING SYSTEMS INDIA LIMITED - ASTROID",
        ("MIC","ASXB"): "ASXB .. ASX BOOKBUILD",
        ("MIC","ASXC"): "ASXC .. ASX - CENTRE POINT",
        ("MIC","ASXP"): "ASXP .. ASX - PUREMATCH",
        ("MIC","ASXT"): "ASXT .. ASX TRADEMATCH",
        ("MIC","ASXV"): "ASXV .. ASX - VOLUMEMATCH",
        ("MIC","ATDF"): "ATDF .. AUTOMATED TRADING DESK FINANCIAL SERVICES, LLC",
        ("MIC","ATFX"): "ATFX .. ATFUND MTF",
        ("MIC","ATLB"): "ATLB .. BRAEMAR ATLANTIC SECURITIES LTD",
        ("MIC","AURB"): "AURB .. AUREL",
        ("MIC","AURO"): "AURO .. AUREL - OTF",
        ("MIC","AUTB"): "AUTB .. AUTILLA - BASE METALS",
        ("MIC","AUTO"): "AUTO .. AUTOBAHN FX",
        ("MIC","AUTP"): "AUTP .. AUTILLA - PRECIOUS METALS",
        ("MIC","AUTX"): "AUTX .. AUTILLA",
        ("MIC","AWBX"): "AWBX .. AUSTRALIAN WHEAT BOARD",
        ("MIC","AWEX"): "AWEX .. AUSTRALIAN WOOL EXCHANGE",
        ("MIC","AXSI"): "AXSI .. ALEXANDRIA PANKKIIRILIIKE OYJ - SYSTEMATIC INTERNALISER",
        ("MIC","BAAD"): "BAAD .. BAADER BANK - SYSTEMATIC INTERNALISER",
        ("MIC","BACE"): "BACE .. BOLSA DE CEREALES DE BUENOS AIRES",
        ("MIC","BAEP"): "BAEP .. BANK OF AMERICA MERRILL LYNCH EQUITY SWAP - EXTERNAL PRICE SOURCE",
        ("MIC","BAIP"): "BAIP .. BANK OF AMERICA MERRILL LYNCH EQUITY SWAP - INTERNAL PRICE SOURCE",
        ("MIC","BAJM"): "BAJM .. BARBADOS STOCK EXCHANGE - JUNIOR MARKET",
        ("MIC","BALT"): "BALT .. THE BALTIC EXCHANGE",
        ("MIC","BAML"): "BAML .. BANK OF AMERICA - MERRILL LYNCH INSTINCT X ATS",
        ("MIC","BAMX"): "BAMX .. BANK OF AMERICA - MERRILL LYNCH GCX",
        ("MIC","BANA"): "BANA .. BANK OF AMERICA, NATIONAL ASSOCIATION",
        ("MIC","BAPA"): "BAPA .. BLOOMBERG - APA",
        ("MIC","BAPE"): "BAPE .. BLOOMBERG DATA REPORTING SERVICES B.V. - APA",
        ("MIC","BAPX"): "BAPX .. BALTPOOL",
        ("MIC","BARD"): "BARD .. BARCLAYS FX - TRADING",
        ("MIC","BARK"): "BARK .. CBOE EUROPE - REGULATED MARKET DARK BOOK",
        ("MIC","BARL"): "BARL .. BARCLAYS LIQUID MARKETS",
        ("MIC","BARO"): "BARO .. CBOE EUROPE - REGULATED MARKET OFF BOOK",
        ("MIC","BART"): "BART .. CBOE EUROPE - REGULATED MARKET INTEGRATED BOOK",
        ("MIC","BARU"): "BARU .. CBOE EUROPE - REGULATED MARKET DARK BOOK (NL)",
        ("MIC","BARX"): "BARX .. BARCLAYS ATS",
        ("MIC","BASE"): "BASE .. BANK OF AMERICA MERRILL LYNCH - SWAP EXECUTIONS",
        ("MIC","BASI"): "BASI .. BANK OF AMERICA, NATIONAL ASSOCIATION - LONDON BRANCH OFFICE - SYSTEMATIC INTERNALISER",
        ("MIC","BASX"): "BASX .. BANK OF AMERICA MERRILL LYNCH - SIMPLEX BLAST X",
        ("MIC","BATD"): "BATD .. CBOE EUROPE -BXE DARK ORDER BOOK",
        ("MIC","BATE"): "BATE .. CBOE EUROPE -BXE ORDER BOOKS",
        ("MIC","BATF"): "BATF .. CBOE EUROPE - BXE OFF-BOOK",
        ("MIC","BATO"): "BATO .. CBOE BZX OPTIONS EXCHANGE",
        ("MIC","BATP"): "BATP .. CBOE EUROPE - BXE PERIODIC",
        ("MIC","BATS"): "BATS .. CBOE BZX U.S. EQUITIES EXCHANGE",
        ("MIC","BATY"): "BATY .. CBOE BYX U.S. EQUITIES EXCHANGE",
        ("MIC","BBIE"): "BBIE .. BARCLAYS BANK IRELAND PLC",
        ("MIC","BBIS"): "BBIS .. BARCLAYS BANK IRELAND PLC - SYSTEMATIC INTERNALISER",
        ("MIC","BBSF"): "BBSF .. BLOOMBERG SEF LLC",
        ("MIC","BBSI"): "BBSI .. BARCLAYS BANK PLC - SYSTEMATIC INTERNALISER",
        ("MIC","BBVA"): "BBVA .. BANCO BILBAO VIZCAYA ARGENTARIA S.A - SYSTEMATIC INTERNALISER",
        ("MIC","BCDX"): "BCDX .. BARCLAYS DIRECT EX ATS",
        ("MIC","BCEE"): "BCEE .. BANQUE ET CAISSE D'EPARGNE DE L'ETAT, LUXEMBOURG - BCEE - SYSTEMATIC INTERNALISER",
        ("MIC","BCFS"): "BCFS .. BOLSA DE COMERCIO DE SANTA FE",
        ("MIC","BCMM"): "BCMM .. BOLSA DE CEREAIS E MERCADORIAS DE MARINGA",
        ("MIC","BCRM"): "BCRM .. CBOE EUROPE REGULATED MARKETS",
        ("MIC","BCSE"): "BCSE .. BELARUS CURRENCY AND STOCK EXCHANGE",
        ("MIC","BCSI"): "BCSI .. BARCLAYS CAPITAL SECURITIES LIMITED - SYSTEMATIC INTERNALISER",
        ("MIC","BCSL"): "BCSL .. BARCLAYS CAPITAL SECURITIES LIMITED",
        ("MIC","BCXE"): "BCXE .. CBOE EUROPE EQUITIES",
        ("MIC","BDEA"): "BDEA .. BAYERNLB - IDEAL FX - SYSTEMATIC INTERNALISER",
        ("MIC","BEAM"): "BEAM .. MTS ASSOCIATED MARKETS",
        ("MIC","BEEX"): "BEEX .. BOND ELECTRONIC EXCHANGE",
        ("MIC","BEIS"): "BEIS .. BANCA ETICA - SYSTEMATIC INTERNALISER",
        ("MIC","BELB"): "BELB .. BELFIUS BANK NV/SA",
        ("MIC","BELF"): "BELF .. BELFIUS BANK NV/SA FIXED INCOME - SYSTEMATIC INTERNALISER",
        ("MIC","BERA"): "BERA .. BOERSE BERLIN - REGULIERTER MARKT",
        ("MIC","BERB"): "BERB .. BOERSE BERLIN - FREIVERKEHR",
        ("MIC","BERC"): "BERC .. BOERSE BERLIN - BERLIN SECOND REGULATED MARKET",
        ("MIC","BETA"): "BETA .. BETA MARKET",
        ("MIC","BETP"): "BETP .. BLOOMBERG TRADEBOOK JAPAN LIMITED",
        ("MIC","BEUD"): "BEUD .. CBOE EUROPE - BXE DARK ORDER BOOK (NL)",
        ("MIC","BEUE"): "BEUE .. CBOE EUROPE - BXE ORDER BOOKS (NL)",
        ("MIC","BEUF"): "BEUF .. CBOE EUROPE - BXE OFF-BOOK (NL)",
        ("MIC","BEUO"): "BEUO .. CBOE EUROPE - REGULATED MARKET OFF BOOK (NL)",
        ("MIC","BEUP"): "BEUP .. CBOE EUROPE - BXE PERIODIC (NL)",
        ("MIC","BEUT"): "BEUT .. CBOE EUROPE - REGULATED MARKET INTEGRATED BOOK (NL)",
        ("MIC","BFEX"): "BFEX .. BAHRAIN FINANCIAL EXCHANGE",
        ("MIC","BGCB"): "BGCB .. BGC BROKERS LP - TRAYPORT",
        ("MIC","BGCD"): "BGCD .. BGC DERIVATIVE MARKETS L.P.",
        ("MIC","BGCF"): "BGCF .. BGC FINANCIAL INC",
        ("MIC","BGCI"): "BGCI .. BGC BROKERS LP",
        ("MIC","BGCO"): "BGCO .. BGC BROKERS LP - OTF",
        ("MIC","BGFI"): "BGFI .. BERENBERG FIXED INCOME - SYSTEMATIC INTERNALISER",
        ("MIC","BGFX"): "BGFX .. BERENBERG FX - SYSTEMATIC INTERNALISER",
        ("MIC","BGSI"): "BGSI .. BERENBERG - SYSTEMATIC - INTERNALISER",
        ("MIC","BHSF"): "BHSF .. BATS HOTSPOT SEF LLC",
        ("MIC","BHWA"): "BHWA .. BANK HANDLOWY W WARSZAWIE S.A. - SYSTEMATIC INTERNALISER",
        ("MIC","BIDS"): "BIDS .. BIDS TRADING L.P.",
        ("MIC","BILT"): "BILT .. OFF-EXCHANGE TRANSACTIONS - LISTED AND UNLISTED INSTRUMENTS",
        ("MIC","BINV"): "BINV .. BOND INVEST - SYSTEMATIC INTERNALISER",
        ("MIC","BISI"): "BISI .. GOLDMAN SACHS INTERNATIONAL BANK - SYSTEMATIC INTERNALISER",
        ("MIC","BIVA"): "BIVA .. BOLSA INSTITUCIONAL DE VALORES",
        ("MIC","BKBF"): "BKBF .. BNY MELLON S.A./N.V. - BRUSSELS BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","BKBR"): "BKBR .. BNY MELLON S.A./N.V. - BRUSSELS BRANCH",
        ("MIC","BKLF"): "BKLF .. BNY MELLON - SYSTEMATIC INTERNALISER",
        ("MIC","BKLN"): "BKLN .. BNY MELLON INTERNATIONAL - LONDON BRANCH",
        ("MIC","BKSK"): "BKSK .. BKS BANK AG - SYSTEMATIC INTERNALISER",
        ("MIC","BLBB"): "BLBB .. BAYERNLB",
        ("MIC","BLBF"): "BLBF .. BANJA LUKA STOCK EXCHANGE - FREE MARKET",
        ("MIC","BLBS"): "BLBS .. BAYERNLB - BONDS MARKET - SYSTEMATIC INTERNALISER",
        ("MIC","BLEV"): "BLEV .. BLOCK EVENT",
        ("MIC","BLIQ"): "BLIQ .. BAYERNLB - LIQP BONDS MARKET",
        ("MIC","BLKX"): "BLKX .. INSTINET BLOCKCROSS ATS",
        ("MIC","BLNK"): "BLNK .. BLINK MTF",
        ("MIC","BLOX"): "BLOX .. BLOCKMATCH DARK",
        ("MIC","BLPX"): "BLPX .. BELGIAN POWER EXCHANGE",
        ("MIC","BLTD"): "BLTD .. BLOOMBERG TRADEBOOK LLC",
        ("MIC","BLTX"): "BLTX .. BALTEX - FREIGHT DERIVATIVES MARKET",
        ("MIC","BLXA"): "BLXA .. INSTINET BLX (AU)",
        ("MIC","BMCL"): "BMCL .. BME CLEARING S.A.",
        ("MIC","BMEA"): "BMEA .. BME - APA",
        ("MIC","BMEX"): "BMEX .. BME - BOLSAS Y MERCADOS ESPANOLES",
        ("MIC","BMLB"): "BMLB .. BANK OF MONTREAL - LONDON BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","BMLX"): "BMLX .. BANK OF AMERICA MERRILL LYNCH INTERNATIONAL LIMITED",
        ("MIC","BMSI"): "BMSI .. BANK OF AMERICA MERRILL LYNCH INTERNATIONAL LIMITED - SYSTEMATIC INTERNALISER",
        ("MIC","BMTF"): "BMTF .. BLOOMBERG TRADING FACILITY LIMITED",
        ("MIC","BMTS"): "BMTS .. MTS BELGIUM",
        ("MIC","BNDD"): "BNDD .. TRADEWEB DIRECT LLC",
        ("MIC","BNPA"): "BNPA .. BNP PARIBAS ARBITRAGE SNC - SYSTEMATIC INTERNALISER",
        ("MIC","BNPF"): "BNPF .. BNP PARIBAS FORTIS SA/NV- SYSTEMATIC INTERNALISER",
        ("MIC","BNPS"): "BNPS .. BNP PARIBAS SA - SYSTEMATIC INTERNALISER",
        ("MIC","BNSX"): "BNSX .. THE BANK OF NOVA SCOTIA - SYSTEMATIC INTERNALISER",
        ("MIC","BNTW"): "BNTW .. BLOCKMATCH MTF NT",
        ("MIC","BNYC"): "BNYC .. CONVERGEX",
        ("MIC","BOAL"): "BOAL .. BANK OF AMERICA, NATIONAL ASSOCIATION - LONDON BRANCH OFFICE",
        ("MIC","BOAT"): "BOAT .. CINNOBER BOAT",
        ("MIC","BOFS"): "BOFS .. BANK OF SCOTLAND PLC - SYSTEMATIC INTERNALISER",
        ("MIC","BOSC"): "BOSC .. BONDSCAPE",
        ("MIC","BOSD"): "BOSD .. NASDAQ OMX BX DARK",
        ("MIC","BOSP"): "BOSP .. WARSAW STOCK EXCHANGE/BONDS/CATALYST/BONDSPOT/MTF",
        ("MIC","BOTC"): "BOTC .. OFF EXCHANGE IDENTIFIER FOR OTC TRADES REPORTED TO CBOE EUROPE",
        ("MIC","BOTE"): "BOTE .. BOTSWANA STOCK EXCHANGE - EXCHANGE TRADED FUNDS (ETF)",
        ("MIC","BOTV"): "BOTV .. BOTSWANA STOCK EXCHANGE - VENTURE CAPITAL",
        ("MIC","BOVA"): "BOVA .. BOLSA DE CORREDORES - BOLSA DE VALORES",
        ("MIC","BOVM"): "BOVM .. BOLSA DE VALORES MINAS-ESPIRITO SANTO-BRASILIA",
        ("MIC","BPAS"): "BPAS .. BANCA PASSADORE - SYSTEMATIC INTERNALISER",
        ("MIC","BPKO"): "BPKO .. PKO BANK POLSKI S.A. - SYSTEMATIC INTERNALISER",
        ("MIC","BPLC"): "BPLC .. BARCLAYS BANK PLC",
        ("MIC","BPOL"): "BPOL .. BLOOMBERG BPOOL",
        ("MIC","BPSX"): "BPSX .. BNP PARIBAS SECURITIES SERVICES - SYSTEMATIC INTERNALISER",
        ("MIC","BRDE"): "BRDE .. BRD - GROUPE SOCIETE GENERALE S.A.",
        ("MIC","BRDL"): "BRDL .. BRD - GROUPE SOCIETE GENERALE S.A. - LIQUIDITY PROVIDER",
        ("MIC","BRFQ"): "BRFQ .. BLOCKMATCH MTF RFQ",
        ("MIC","BRIX"): "BRIX .. BRAZILIAN ENERGY EXCHANGE",
        ("MIC","BRNX"): "BRNX .. BERNSTEIN CROSS (BERN-X)",
        ("MIC","BSEX"): "BSEX .. BAKU STOCK EXCHANGE",
        ("MIC","BSFX"): "BSFX .. BAYERNLB - FX HANDEL - SYSTEMATIC INTERNALISER",
        ("MIC","BSME"): "BSME .. BSE SME",
        ("MIC","BTEC"): "BTEC .. ICAP ELECTRONIC BROKING (US)",
        ("MIC","BTEE"): "BTEE .. BROKERTEC EUROPE LIMITED - ALL MARKETS",
        ("MIC","BTFE"): "BTFE .. BLOOMBERG TRADING FACILITY B.V.",
        ("MIC","BTNL"): "BTNL .. BITNOMIAL",
        ("MIC","BULK"): "BULK .. NASDAQ COMMODITIES - BULK COMMODITY",
        ("MIC","BVCA"): "BVCA .. CARACAS STOCK EXCHANGE",
        ("MIC","BVMF"): "BVMF .. BM&FBOVESPA S.A. - BOLSA DE VALORES, MERCADORIAS E FUTUROS",
        ("MIC","BVUK"): "BVUK .. BONDVISION UK",
        ("MIC","BVUR"): "BVUR .. BOLSA ELECTRONICA DE VALORES DEL URUGUAY",
        ("MIC","BVUS"): "BVUS .. BONDVISION US",
        ("MIC","BYXD"): "BYXD .. CBOE BYX U.S. EQUITIES EXCHANGE DARK",
        ("MIC","BZXD"): "BZXD .. CBOE BZX U.S. EQUITIES EXCHANGE DARK",
        ("MIC","C2OX"): "C2OX .. C2 OPTIONS EXCHANGE INC.",
        ("MIC","CABK"): "CABK .. CAIXABANK - SYSTEMATIC INTERNALISER",
        ("MIC","CAES"): "CAES .. CREDIT SUISSE AES CROSSFINDER",
        ("MIC","CAND"): "CAND .. CANDEAL.CA INC",
        ("MIC","CANX"): "CANX .. CANNEX FINANCIAL EXCHANGE LTS",
        ("MIC","CAPA"): "CAPA .. CBOE EUROPE - APA (NL)",
        ("MIC","CAPI"): "CAPI .. CAPI - OTF",
        ("MIC","CASI"): "CASI .. CARNEGIE INVESTMENT BANK AB - SYSTEMATIC INTERNALISER",
        ("MIC","CATS"): "CATS .. CATS",
        ("MIC","CAVE"): "CAVE .. CAVEAT EMPTOR",
        ("MIC","CBAL"): "CBAL .. COMMONWEALTH BANK OF AUSTRALIA - LONDON BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","CBKA"): "CBKA .. COMMERZBANK AG",
        ("MIC","CBKD"): "CBKD .. COMMERZBANK EQUITY DERIVATIVES - SYSTEMATIC INTERNALISER",
        ("MIC","CBKE"): "CBKE .. COMMERZBANK EQUITY - SYSTEMATIC INTERNALISER",
        ("MIC","CBKF"): "CBKF .. COMMERZBANK COMMANDER - SYSTEMATIC INTERNALISER",
        ("MIC","CBKG"): "CBKG .. COMMERZBANK FIXED INCOME - SYSTEMATIC INTERNALISER",
        ("MIC","CBLC"): "CBLC .. CITIBLOC",
        ("MIC","CBNL"): "CBNL .. CITIBANK N.A. - LONDON BRANCH",
        ("MIC","CBSX"): "CBSX .. CBOE STOCK EXCHANGE",
        ("MIC","CBTS"): "CBTS .. CME SWAPS MARKETS (CBOT)",
        ("MIC","CCEU"): "CCEU .. CITADEL CONNECT EUROPE - SYSTEMATIC INTERNALISER",
        ("MIC","CCFX"): "CCFX .. CHINA FINANCIAL FUTURES EXCHANGE",
        ("MIC","CCLX"): "CCLX .. FINESTI S.A.",
        ("MIC","CCML"): "CCML .. CONTINENTAL CAPITAL MARKETS LIMITED - OTF",
        ("MIC","CCMS"): "CCMS .. CONTINENTAL CAPITAL MARKETS S.A.",
        ("MIC","CCO2"): "CCO2 .. CANTORCO2E.COM LIMITED",
        ("MIC","CCRM"): "CCRM .. CBOE EUROPE REGULATED MARKETS (NL)",
        ("MIC","CCXE"): "CCXE .. CBOE EUROPE EQUITIES - EUROPEAN EQUITIES (NL)",
        ("MIC","CDED"): "CDED .. CITADEL SECURITIES",
        ("MIC","CDSL"): "CDSL .. CLEARCORP DEALING SYSTEMS (INDIA) LTD.",
        ("MIC","CECS"): "CECS .. CME SWAPS MARKETS (COMEX)",
        ("MIC","CEPL"): "CEPL .. CITIBANK EUROPE PLC",
        ("MIC","CETI"): "CETI .. CETIP S.A. - MERCADOS ORGANIZADOS",
        ("MIC","CEUD"): "CEUD .. CBOE EUROPE - CXE DARK ORDER BOOK (NL)",
        ("MIC","CEUE"): "CEUE .. CBOE EUROPE - CXE ORDER BOOKS (NL)",
        ("MIC","CEUO"): "CEUO .. CBOE EUROPE - CXE OFF-BOOK (NL)",
        ("MIC","CFAU"): "CFAU .. CROSSFINDER AUSTRALIA",
        ("MIC","CFBC"): "CFBC .. CHINA FOREIGN EXCHANGE TRADE SYSTEM - SHANGHAI - HONG KONG BOND CONNECT",
        ("MIC","CFHK"): "CFHK .. CROSSFINDER HONG KONG",
        ("MIC","CFJP"): "CFJP .. CROSSFINDER JAPAN",
        ("MIC","CFTW"): "CFTW .. CROSSFINDER TAIWAN",
        ("MIC","CGCM"): "CGCM .. CASSA DI COMPENSAZIONE E GARANZIA SPA - COLLATERALIZED MONEY MARKET GUARANTEE SERVICE",
        ("MIC","CGDB"): "CGDB .. CASSA DI COMPENSAZIONE E GARANZIA SPA - BONDS CCP SERVICE",
        ("MIC","CGEB"): "CGEB .. CASSA DI COMPENSAZIONE E GARANZIA SPA - EURO BONDS CCP SERVICE",
        ("MIC","CGEC"): "CGEC .. CGEC - CITIGROUP GLOBAL MARKETS  EUROPE AG - CENTRAL RISK DESK",
        ("MIC","CGEE"): "CGEE .. CITI MATCH - DE",
        ("MIC","CGET"): "CGET .. CITIGROUP GLOBAL MARKETS  EUROPE AG - EMEA TRADING",
        ("MIC","CGGD"): "CGGD .. CASSA DI COMPENSAZIONE E GARANZIA SPA - CCP AGRICULTURAL COMMODITY DERIVATIVES",
        ("MIC","CGIT"): "CGIT .. CASSA DI COMPENSAZIONE E GARANZIA SPA",
        ("MIC","CGMA"): "CGMA .. CITI MATCH AUSTRALIA",
        ("MIC","CGMC"): "CGMC .. CITIGROUP GLOBAL MARKETS  LIMITED - CENTRAL RISK DESK",
        ("MIC","CGMD"): "CGMD .. CITIGROUP GLOBAL MARKETS EUROPE AG - SYSTEMATIC INTERNALISER",
        ("MIC","CGME"): "CGME .. CITI MATCH - GB",
        ("MIC","CGMH"): "CGMH .. CITI MATCH - HK",
        ("MIC","CGMI"): "CGMI .. CITIGROUP GLOBAL MARKETS",
        ("MIC","CGML"): "CGML .. CITIGROUP GLOBAL MARKETS LIMITED",
        ("MIC","CGMT"): "CGMT .. CITIGROUP GLOBAL MARKETS  LIMITED - EMEA TRADING",
        ("MIC","CGMX"): "CGMX .. CITI MEXICO RPI (RETAIL PRICE IMPROVEMENT)",
        ("MIC","CGND"): "CGND .. CASSA DI COMPENSAZIONE E GARANZIA SPA - CCP ENERGY DERIVATIVES",
        ("MIC","CGQD"): "CGQD .. CASSA DI COMPENSAZIONE E GARANZIA SPA - CCP EQUITY DERIVATIVES",
        ("MIC","CGQT"): "CGQT .. CASSA DI COMPENSAZIONE E GARANZIA SPA - EQUITY CCP SERVICE",
        ("MIC","CGTR"): "CGTR .. CASSA DI COMPENSAZIONE E GARANZIA SPA - TRIPARTY REPO CCP SERVICE",
        ("MIC","CHEV"): "CHEV .. CA CHEUVREUX",
        ("MIC","CHIA"): "CHIA .. CHI-X AUSTRALIA",
        ("MIC","CHIC"): "CHIC .. CHI-X CANADA ATS",
        ("MIC","CHID"): "CHID .. CBOE EUROPE - CXE DARK ORDER BOOK",
        ("MIC","CHIJ"): "CHIJ .. CHI-X JAPAN",
        ("MIC","CHIO"): "CHIO .. CBOE EUROPE - CXE OFF-BOOK",
        ("MIC","CHIS"): "CHIS .. CHI-X JAPAN SELECT",
        ("MIC","CHIV"): "CHIV .. CHI-X JAPAN VWAP CROSSING",
        ("MIC","CHIX"): "CHIX .. CBOE EUROPE - CXE ORDER BOOKS",
        ("MIC","CIBC"): "CIBC .. CANADIAN IMPERIAL BANK OF COMMERCE - SYSTEMATIC INTERNALISER",
        ("MIC","CIBP"): "CIBP .. CIBC WORLD MARKETS PLC - SYSTEMATIC INTERNALISER",
        ("MIC","CICX"): "CICX .. CITI CROSS",
        ("MIC","CIMB"): "CIMB .. CIMD S.V. S.A. - OTF FOR BONDS",
        ("MIC","CIMD"): "CIMD .. CIMD S.V. S.A. - OTF",
        ("MIC","CIME"): "CIME .. CIMD S.V. S.A. - OTF FOR ENERGY DERIVATIVES",
        ("MIC","CIMV"): "CIMV .. CIMD S.V. S.A. - OTF FOR DERIVATIVES",
        ("MIC","CIOI"): "CIOI .. CIOI",
        ("MIC","CITD"): "CITD .. CITI DARK",
        ("MIC","CITX"): "CITX .. CITI MATCH - JP",
        ("MIC","CLAU"): "CLAU .. CLSA AUSTRALIA - DARK",
        ("MIC","CLHK"): "CLHK .. CLSA HONG KONG - DARK",
        ("MIC","CLJP"): "CLJP .. CLSA JAPAN - DARK",
        ("MIC","CLMX"): "CLMX .. CLIMEX",
        ("MIC","CLPH"): "CLPH .. CLSA PHILIPPINES - DARK",
        ("MIC","CLTD"): "CLTD .. CLEARTRADE EXCHANGE",
        ("MIC","CMCI"): "CMCI .. CREDIT INDUSTRIEL ET COMMERCIAL - SYSTEMATIC INTERNALISER",
        ("MIC","CMEC"): "CMEC .. CME CLEARING EUROPE",
        ("MIC","CMEE"): "CMEE .. CME  EUROPE",
        ("MIC","CMES"): "CMES .. CME SWAPS MARKETS (CME)",
        ("MIC","CMET"): "CMET .. CLEAR MARKETS JAPAN, INC.",
        ("MIC","CMMT"): "CMMT .. CLEAR MARKETS EUROPE LIMITED",
        ("MIC","CMSF"): "CMSF .. CLEAR MARKETS NORTH AMERICA, INC.",
        ("MIC","CNSI"): "CNSI .. CARNEGIE AS - SYSTEMATIC INTERNALISER",
        ("MIC","COAL"): "COAL .. LA COTE ALPHA",
        ("MIC","COMG"): "COMG .. COMMERG LTD.",
        ("MIC","COMM"): "COMM .. COMMERG",
        ("MIC","CORE"): "CORE .. ATD - CITIGROUP AGENCY OPTION AND EQUITIES ROUTING ENGINE",
        ("MIC","COTC"): "COTC .. BMO CAPITAL MARKETS - CAD OTC TRADES",
        ("MIC","CPTX"): "CPTX .. CAPTIN",
        ("MIC","CRED"): "CRED .. CREDIT SUISSE (US)",
        ("MIC","CRYD"): "CRYD .. CRYEX - FX AND DIGITAL CURRENCIES",
        ("MIC","CRYP"): "CRYP .. CRYPTO FACILITIES",
        ("MIC","CRYX"): "CRYX .. CRYEX",
        ("MIC","CSAG"): "CSAG .. CREDIT SUISSE AG - SYSTEMATIC INTERNALISER",
        ("MIC","CSAU"): "CSAU .. CREDIT SUISSE EQUITIES (AUSTRALIA) LIMITED",
        ("MIC","CSBX"): "CSBX .. CREDIT SUISSE AES EUROPE BENCHMARK CROSS",
        ("MIC","CSCF"): "CSCF .. CREDIT SUISSE AES CROSSFINDER EUROPE",
        ("MIC","CSEC"): "CSEC .. CITADEL SECURITIES FI - SYSTEMATIC INTERNALISER",
        ("MIC","CSEU"): "CSEU .. CREDIT SUISSE (EUROPE)",
        ("MIC","CSHK"): "CSHK .. CREDIT SUISSE SECURITIES (HONG KONG) LIMITED",
        ("MIC","CSIN"): "CSIN .. CREDIT SUISSE INTERNATIONAL",
        ("MIC","CSJP"): "CSJP .. CREDIT SUISSE EQUITIES (JAPAN) LIMITED",
        ("MIC","CSLB"): "CSLB .. CREDIT SUISSE AG - LONDON BRANCH",
        ("MIC","CSLP"): "CSLP .. CREDIT SUISSE LIGHT POOL",
        ("MIC","CSOT"): "CSOT .. CREDIT SUISSE AG - OHS",
        ("MIC","CSSI"): "CSSI .. CREDIT SUISSE INTERNATIONAL - SYSTEMATIC INTERNALISER",
        ("MIC","CSSX"): "CSSX .. CHINA STAINLESS STEEL EXCHANGE",
        ("MIC","CSTO"): "CSTO .. NASDAQ CLEARING AB",
        ("MIC","CSZH"): "CSZH .. CREDIT SUISSE (CH)",
        ("MIC","CURX"): "CURX .. CUREX FX",
        ("MIC","CXAC"): "CXAC .. CHI-X MARKET AUSTRALIA - LIMIT VENUE",
        ("MIC","CXAF"): "CXAF .. CHI-X AUSTRALIA MARKET PEG (FARPOINT) VENUE",
        ("MIC","CXAM"): "CXAM .. CHI-X MOC",
        ("MIC","CXAN"): "CXAN .. CHI-X AUSTRALIA PRIMARY PEG (NEARPOINT) VENUE",
        ("MIC","CXAP"): "CXAP .. CHI-X AUSTRALIA MID-POINT VENUE",
        ("MIC","CXAQ"): "CXAQ .. CHI-X AUSTRALIA -QUOTED MANAGED FUNDS",
        ("MIC","CXAR"): "CXAR .. CHI-X AUSTRALIA - TRANSFERABLE CUSTODY RECEIPT MARKET",
        ("MIC","CXAV"): "CXAV .. CHI-X VWAP",
        ("MIC","CXAW"): "CXAW .. CHI-X AUSTRALIA - WARRANTS",
        ("MIC","CXOT"): "CXOT .. CREDITEX BROKERAGE LLP - OTF",
        ("MIC","CXRT"): "CXRT .. CREDITEX BROKERAGE LLP - MTF",
        ("MIC","DAIW"): "DAIW .. DAIWA CAPITAL MARKETS EUROPE LIMITED",
        ("MIC","DAPA"): "DAPA .. DEUTSCHE BOERSE AG - APA SERVICE",
        ("MIC","DASI"): "DASI .. DANSKE BANK A/S - SYSTEMATIC INTERNALISER",
        ("MIC","DAVY"): "DAVY .. J AND E DAVY - IRISH GOVERNMENT BOND - SYSTEMATIC INTERNALISER",
        ("MIC","DBAG"): "DBAG .. DEUTSCHE BANK AG",
        ("MIC","DBCR"): "DBCR .. DEUTSCHE BANK - CENTRAL RISK BOOK",
        ("MIC","DBCX"): "DBCX .. DEUTSCHE BANK - CLOSE CROSS",
        ("MIC","DBDC"): "DBDC .. DEUTSCHE BANK - DIRECT CAPITAL ACCESS",
        ("MIC","DBES"): "DBES .. DEUTSCHE BANK AG - SYSTEMATIC INTERNALISER",
        ("MIC","DBHK"): "DBHK .. DEUTSCHE BANK HONG KONG ATS",
        ("MIC","DBIX"): "DBIX .. DEUTSCHE BANK INTERNALISATION",
        ("MIC","DBMO"): "DBMO .. DEUTSCHE BANK - MANUAL OTC",
        ("MIC","DBOX"): "DBOX .. DEUTSCHE BANK OFF EXCHANGE TRADING",
        ("MIC","DBSE"): "DBSE .. DEUTSCHE BANK - SUPERX EU",
        ("MIC","DBSX"): "DBSX .. DEUTSCHE BANK SUPER X",
        ("MIC","DCSE"): "DCSE .. NASDAQ COPENHAGEN A/S - NORDIC@MID",
        ("MIC","DCSX"): "DCSX .. DUTCH CARIBBEAN SECURITIES EXCHANGE",
        ("MIC","DDTX"): "DDTX .. DEAL DONE TRADING - SYSTEMATIC INTERNALISER",
        ("MIC","DEAL"): "DEAL .. DCX (DERIVATIVES CURRENCY EXCHANGE)",
        ("MIC","DEKA"): "DEKA .. DEKABANK DEUTSCHE GIROZENTRALE",
        ("MIC","DGCX"): "DGCX .. DUBAI GOLD & COMMODITIES EXCHANGE DMCC",
        ("MIC","DHEL"): "DHEL .. NASDAQ HELSINKI LTD - NORDIC@MID",
        ("MIC","DICE"): "DICE .. NASDAQ ICELAND HF. - NORDIC@MID",
        ("MIC","DIFX"): "DIFX .. NASDAQ DUBAI",
        ("MIC","DKED"): "DKED .. NASDAQ STOCKHOLM AB - DANISH EQ DERIVATIVES",
        ("MIC","DKFI"): "DKFI .. NASDAQ STOCKHOLM AB - DANISH FI DERIVATIVES",
        ("MIC","DKTC"): "DKTC .. DANSK OTC",
        ("MIC","DNDK"): "DNDK .. FIRST NORTH DENMARK - NORDIC@MID",
        ("MIC","DNFI"): "DNFI .. FIRST NORTH FINLAND - NORDIC@MID",
        ("MIC","DNIS"): "DNIS .. FIRST NORTH ICELAND - NORDIC@MID",
        ("MIC","DNSE"): "DNSE .. FIRST NORTH SWEDEN - NORDIC@MID",
        ("MIC","DOSE"): "DOSE .. FIRST NORTH SWEDEN - NORWAY NORDIC@MID",
        ("MIC","DOTS"): "DOTS .. SWISS DOTS BY CATS",
        ("MIC","DOWG"): "DOWG .. DOWGATE",
        ("MIC","DRCT"): "DRCT .. DAIWA DRECT",
        ("MIC","DSMD"): "DSMD .. QATAR EXCHANGE",
        ("MIC","DSTO"): "DSTO .. NASDAQ STOCKHOLM AB - NORDIC@MID",
        ("MIC","DUMX"): "DUMX .. DUBAI MERCANTILE EXCHANGE",
        ("MIC","DUSA"): "DUSA .. BOERSE DUESSELDORF - REGULIERTER MARKT",
        ("MIC","DUSB"): "DUSB .. BOERSE DUESSELDORF - FREIVERKEHR",
        ("MIC","DUSC"): "DUSC .. BOERSE DUESSELDORF - QUOTRIX - REGULIERTER MARKT",
        ("MIC","DUSD"): "DUSD .. BOERSE DUESSELDORF - QUOTRIX MTF",
        ("MIC","DVFX"): "DVFX .. DIGITAL VEGA",
        ("MIC","DWSF"): "DWSF .. DW SEF LLC",
        ("MIC","DZBK"): "DZBK .. DZ BANK - SYSTEMATIC INTERNALISER",
        ("MIC","EBON"): "EBON .. NASDAQ STOCKHOLM AB - EUR FI DERIVATIVES",
        ("MIC","EBRA"): "EBRA .. BRATISLAVA STOCK EXCHANGE-MTF",
        ("MIC","EBSC"): "EBSC .. EBS MARKET- CLOB - FOR THE TRADING OF SPOT FX, PRECIOUS METALS AND OTHER FX PRODUCTS",
        ("MIC","EBSD"): "EBSD .. EBS MTF - RFQ - FOR THE TRADING OF FX PRODUCTS",
        ("MIC","EBSI"): "EBSI .. EBS MTF - RFQ - FOR ASSET MANAGERS TRADING FX PRODUCTS",
        ("MIC","EBSM"): "EBSM .. EBS MTF - CLOB - FOR THE TRADING OF FX PRODUCTS",
        ("MIC","EBSS"): "EBSS .. EBS SERVICE COMPANY LIMITED - ALL MARKETS",
        ("MIC","ECAG"): "ECAG .. EUREX CLEARING AG",
        ("MIC","ECAL"): "ECAL .. EUREX CLEARING ASIA PTE. LTD.",
        ("MIC","ECHO"): "ECHO .. LONDON STOCK EXCHANGE - APA",
        ("MIC","EDDP"): "EDDP .. CBOE EDGX U.S. EQUITIES EXCHANGE DARK",
        ("MIC","EDGA"): "EDGA .. CBOE EDGA U.S. EQUITIES EXCHANGE",
        ("MIC","EDGD"): "EDGD .. CBOE EDGA U.S. EQUITIES EXCHANGE DARK",
        ("MIC","EDGO"): "EDGO .. CBOE EDGX OPTIONS EXCHANGE",
        ("MIC","EDGX"): "EDGX .. CBOE EDGX U.S. EQUITIES EXCHANGE",
        ("MIC","EDRF"): "EDRF .. EDMOND DE ROTHSCHILD  (FRANCE) - SYSTEMATIC INTERNALISER",
        ("MIC","EEAL"): "EEAL .. EUREX EXCHANGE ASIA PTE. LTD.",
        ("MIC","EESE"): "EESE .. EAST EUROPEAN STOCK EXCHANGE",
        ("MIC","EGMT"): "EGMT .. EG MARKET TECHNOLOGIES",
        ("MIC","EGSI"): "EGSI .. ERSTE GROUP BANK AG",
        ("MIC","ELEU"): "ELEU .. NASDAQ COMMODITIES - EUR POWER/ENERGY",
        ("MIC","ELIX"): "ELIX .. ELIXIUM",
        ("MIC","ELNO"): "ELNO .. NASDAQ COMMODITIES - NOK POWER/ENERGY",
        ("MIC","ELSE"): "ELSE .. NASDAQ COMMODITIES - SEK POWER/ENERGY",
        ("MIC","ELUK"): "ELUK .. NASDAQ COMMODITIES - GBP POWER/ENERGY",
        ("MIC","EMBX"): "EMBX .. EMERGING MARKETS BOND EXCHANGE LIMITED",
        ("MIC","EMDR"): "EMDR .. E-MID - E-MIDER MARKET",
        ("MIC","EMIB"): "EMIB .. E-MID - BANCA D'ITALIA SHARES TRADING MARKET",
        ("MIC","EMID"): "EMID .. E-MID",
        ("MIC","EMIR"): "EMIR .. E-MID REPO",
        ("MIC","EMLD"): "EMLD .. MIAX EMERALD, LLC",
        ("MIC","EMTF"): "EMTF .. EURO MTF",
        ("MIC","EMTS"): "EMTS .. EBM",
        ("MIC","ENAX"): "ENAX .. ATHENS EXCHANGE ALTERNATIVE MARKET",
        ("MIC","ENCL"): "ENCL .. ENCLEAR",
        ("MIC","ENSY"): "ENSY .. EURONEXT SYNAPSE",
        ("MIC","ENXB"): "ENXB .. EURONEXT - EASY NEXT",
        ("MIC","ENXL"): "ENXL .. EURONEXT ACCESS LISBON",
        ("MIC","EOTC"): "EOTC .. E-OTC",
        ("MIC","EPEX"): "EPEX .. EPEX SPOT SE",
        ("MIC","EQLD"): "EQLD .. EQUILEND EUROPE LIMITED",
        ("MIC","EQTA"): "EQTA .. BOERSE BERLIN EQUIDUCT TRADING - REGULIERTER MARKT",
        ("MIC","EQTB"): "EQTB .. BOERSE BERLIN EQUIDUCT TRADING - BERLIN SECOND REGULATED MARKET",
        ("MIC","EQTC"): "EQTC .. BOERSE BERLIN EQUIDUCT TRADING - FREIVERKEHR",
        ("MIC","EQTD"): "EQTD .. BOERSE BERLIN EQUIDUCT TRADING - OTC",
        ("MIC","ERIS"): "ERIS .. ERIS EXCHANGE",
        ("MIC","ERST"): "ERST .. ERSTE BEFEKTETESI ZRT - SYSTEMATIC INTERNALISER",
        ("MIC","ESPD"): "ESPD .. NASDAQ OMX ESPEED",
        ("MIC","ESTO"): "ESTO .. NASDAQ STOCKHOLM AB - NORWAY ETF",
        ("MIC","ETFP"): "ETFP .. ELECTRONIC ETF, ETC/ETN AND OPEN-END FUNDS MARKET",
        ("MIC","ETLX"): "ETLX .. EUROTLX",
        ("MIC","ETPA"): "ETPA .. ENERGY TRADING PLATFORM AMSTERDAM",
        ("MIC","ETSC"): "ETSC .. ETS EURASIAN TRADING SYSTEM COMMODITY EXCHANGE",
        ("MIC","EUCH"): "EUCH .. EUREX ZURICH",
        ("MIC","EURM"): "EURM .. EUREX REPO MARKET",
        ("MIC","EUSC"): "EUSC .. EUREX CH SECLEND MARKET",
        ("MIC","EUSP"): "EUSP .. EUREX OTC SPOT MARKET",
        ("MIC","EUWA"): "EUWA .. EUWAX AG - SYSTEMATIC INTERNALISER",
        ("MIC","EUWB"): "EUWB .. NASDAQ STOCKHOLM AB - EUR WB EQ DERIVATIVES",
        ("MIC","EUWX"): "EUWX .. EUWAX",
        ("MIC","EWSM"): "EWSM .. EUROPEAN WHOLESALE SECURITIES MARKET",
        ("MIC","EXAA"): "EXAA .. WIENER BOERSE AG, AUSTRIAN ENERGY EXCHANGE",
        ("MIC","EXBO"): "EXBO .. EXANE BNP PARIBAS - BID-OFFER CROSSING",
        ("MIC","EXCP"): "EXCP .. EXANE BNP PARIBAS - CLOSING PRICE",
        ("MIC","EXDC"): "EXDC .. EXANE BNP PARIBAS - DIRECT CAPITAL ACCESS",
        ("MIC","EXEU"): "EXEU .. EXANE BNP PARIBAS",
        ("MIC","EXLP"): "EXLP .. EXANE BNP PARIBAS - LIQUIDITY PROVISION",
        ("MIC","EXMP"): "EXMP .. EXANE BNP PARIBAS - MID POINT",
        ("MIC","EXOR"): "EXOR .. EXANE BNP PARIBAS - CHILD ORDER CROSSING",
        ("MIC","EXOT"): "EXOT .. EXOTIX CAPITAL  - OTF",
        ("MIC","EXSB"): "EXSB .. EXANE DERIVATIVES - SYSTEMATIC INTERNALISER CORPORATE BONDS",
        ("MIC","EXSD"): "EXSD .. EXANE DERIVATIVES - SYSTEMATIC INTERNALISER OTC DERIVATIVES",
        ("MIC","EXSE"): "EXSE .. EXANE BNP PARIBAS - SYSTEMATIC INTERNALISER",
        ("MIC","EXSF"): "EXSF .. EXANE DERIVATIVES - SYSTEMATIC INTERNALISER OTHER BONDS",
        ("MIC","EXSH"): "EXSH .. EXANE DERIVATIVES - SYSTEMATIC INTERNALISER SHARES",
        ("MIC","EXSI"): "EXSI .. EXANE BNP PARIBAS - GB - SYSTEMATIC INTERNALISER",
        ("MIC","EXSP"): "EXSP .. EXANE DERIVATIVES - SYSTEMATIC INTERNALISER STRUCTURED PRODUCTS",
        ("MIC","EXSY"): "EXSY .. EXANE DERIVATIVES - SYSTEMATIC INTERNALISER",
        ("MIC","EXTR"): "EXTR .. ENERGY EXCHANGE ISTANBUL",
        ("MIC","EXVP"): "EXVP .. EXANE BNP PARIBAS - VOLUME PROFILE CROSSING",
        ("MIC","EXYY"): "EXYY .. EXANE DERIVATIVES - SYSTEMATIC INTERNALISER CONVERTIBLES",
        ("MIC","FAIR"): "FAIR .. CANTOR SPREADFAIR",
        ("MIC","FAST"): "FAST .. FASTMATCH",
        ("MIC","FBSI"): "FBSI .. FINECO BANK - SYSTEMATIC INTERNALISER",
        ("MIC","FCBT"): "FCBT .. CHICAGO BOARD OF TRADE (FLOOR)",
        ("MIC","FCME"): "FCME .. CHICAGO MERCANTILE EXCHANGE (FLOOR)",
        ("MIC","FGEX"): "FGEX .. KAASUPORSSI - FINNISH GAS EXCHANGE",
        ("MIC","FICO"): "FICO .. FINANCIALCONTENT",
        ("MIC","FICX"): "FICX .. FINANCIAL INFORMATION CONTRIBUTORS EXCHANGE",
        ("MIC","FIED"): "FIED .. NASDAQ STOCKHOLM AB - FINNISH EQ DERIVATIVES",
        ("MIC","FINC"): "FINC .. FINRA/NASDAQ TRF CHICAGO (TRADE REPORTING FACILITY)",
        ("MIC","FINN"): "FINN .. FINRA/NASDAQ TRF CARTERET (TRADE REPORTING FACILITY)",
        ("MIC","FINO"): "FINO .. FINRA ORF (TRADE REPORTING FACILITY)",
        ("MIC","FINR"): "FINR .. FINRA",
        ("MIC","FINY"): "FINY .. FINRA/NYSE TRF (TRADE REPORTING FACILITY)",
        ("MIC","FISH"): "FISH .. FISH POOL ASA",
        ("MIC","FISU"): "FISU .. FREIGHT INVESTOR SERVICES LIMITED",
        ("MIC","FMTS"): "FMTS .. MTS FRANCE SAS",
        ("MIC","FNCS"): "FNCS .. FENICS - US TREASURIES",
        ("MIC","FNDK"): "FNDK .. FIRST NORTH DENMARK",
        ("MIC","FNEE"): "FNEE .. FIRST NORTH ESTONIA",
        ("MIC","FNFI"): "FNFI .. FIRST NORTH FINLAND",
        ("MIC","FNIS"): "FNIS .. FIRST NORTH ICELAND",
        ("MIC","FNLT"): "FNLT .. FIRST NORTH LITHUANIA",
        ("MIC","FNLV"): "FNLV .. FIRST NORTH LATVIA",
        ("MIC","FNSE"): "FNSE .. FIRST NORTH SWEDEN",
        ("MIC","FRAA"): "FRAA .. BOERSE FRANKFURT - REGULIERTER MARKT",
        ("MIC","FRAB"): "FRAB .. BOERSE FRANKFURT - FREIVERKEHR",
        ("MIC","FREI"): "FREI .. NASDAQ COMMODITIES - FREIGHT COMMODITY",
        ("MIC","FRTE"): "FRTE .. FORTE  - OTF",
        ("MIC","FSEF"): "FSEF .. FTSEF LLC",
        ("MIC","FSHX"): "FSHX .. FISHEX",
        ("MIC","FTFM"): "FTFM .. 42 FINANCIAL SERVICES - MTF",
        ("MIC","FTFS"): "FTFS .. 42 FINANCIAL SERVICES",
        ("MIC","FXAL"): "FXAL .. FXALL",
        ("MIC","FXCL"): "FXCL .. CLEARCORP DEALING SYSTEMS INDIA LIMITED - FX-CLEAR",
        ("MIC","FXCM"): "FXCM .. FXCM",
        ("MIC","FXGB"): "FXGB .. FXCM - MTF",
        ("MIC","FXOP"): "FXOP .. TRADITION-NEX OTF",
        ("MIC","FXSW"): "FXSW .. CLEARCORP DEALING SYSTEMS INDIA LIMITED - FX-SWAP",
        ("MIC","G1XX"): "G1XX .. G1 EXECUTION SERVICES",
        ("MIC","GBOT"): "GBOT .. BOURSE AFRICA LIMITED",
        ("MIC","GEMX"): "GEMX .. GEMMA (GILT EDGED MARKET MAKERS'ASSOCIATION)",
        ("MIC","GETB"): "GETB .. LITHUANIAN NATURAL GAS EXCHANGE",
        ("MIC","GFBM"): "GFBM .. GFI BROKERS - MTF",
        ("MIC","GFBO"): "GFBO .. GFI BROKERS - OTF",
        ("MIC","GFIB"): "GFIB .. GFI BROKERS",
        ("MIC","GFIC"): "GFIC .. GFI SECURITIES LTD",
        ("MIC","GFIF"): "GFIF .. GFI FOREXMATCH",
        ("MIC","GFIN"): "GFIN .. GFI ENERGYMATCH",
        ("MIC","GFIR"): "GFIR .. GFI RATESMATCH",
        ("MIC","GFKS"): "GFKS .. GARANTUM FONDKOMMISSION AB - SYSTEMATIC INTERNALISER",
        ("MIC","GFSM"): "GFSM .. GFI SECURITIES LTD - MTF",
        ("MIC","GFSO"): "GFSO .. GFI SECURITIES LTD - OTF",
        ("MIC","GLBX"): "GLBX .. CME GLOBEX",
        ("MIC","GLLC"): "GLLC .. GATE US LLC",
        ("MIC","GLMX"): "GLMX .. GLMX",
        ("MIC","GLPS"): "GLPS .. ESSEX RADEZ, LLC",
        ("MIC","GLPX"): "GLPX .. ACS EXECUTION SERVICES, LLC",
        ("MIC","GMEG"): "GMEG .. GMEX EXCHANGE",
        ("MIC","GMGD"): "GMGD .. GMG DUBAI LIMITED",
        ("MIC","GMGL"): "GMGL .. GMG BROKERS LIMITED",
        ("MIC","GMNI"): "GMNI .. ISE GEMINI EXCHANGE",
        ("MIC","GMTF"): "GMTF .. GALAXY",
        ("MIC","GMTS"): "GMTS .. MTS GERMANY",
        ("MIC","GOTC"): "GOTC .. GLOBAL OTC",
        ("MIC","GOVX"): "GOVX .. GOVEX",
        ("MIC","GREE"): "GREE .. THE GREEN EXCHANGE",
        ("MIC","GRIF"): "GRIF .. GRIFFIN MARKETS LIMITED",
        ("MIC","GRIO"): "GRIO .. GRIFFIN MARKETS LIMITED - OTF",
        ("MIC","GRSE"): "GRSE .. THE GREEN STOCK EXCHANGE - ACB IMPACT MARKETS",
        ("MIC","GSAL"): "GSAL .. GOLDMAN SACHS (ASIA) LLC",
        ("MIC","GSBX"): "GSBX .. GOLDMAN SACHS INTERNATIONAL - SIGMA BCN",
        ("MIC","GSCI"): "GSCI .. THE GUYANA ASSOCIATION OF SECURITIES COMPANIES AND INTERMEDIARIES INC.",
        ("MIC","GSCO"): "GSCO .. GOLDMAN SACHS AND CO.",
        ("MIC","GSEF"): "GSEF .. GFI SWAPS EXCHANGE, LLC",
        ("MIC","GSIB"): "GSIB .. GOLDMAN SACHS INTERNATIONAL BANK",
        ("MIC","GSIL"): "GSIL .. GOLDMAN SACHS INTERNATIONAL",
        ("MIC","GSSI"): "GSSI .. GOLDMAN SACHS INTERNATIONAL - SYSTEMATIC INTERNALISER",
        ("MIC","GSXH"): "GSXH .. GSX HONG KONG",
        ("MIC","GSXL"): "GSXL .. THE GIBRALTAR STOCK EXCHANGE",
        ("MIC","GTCO"): "GTCO .. KCG AMERICAS LLC",
        ("MIC","GTSX"): "GTSX .. GTSX",
        ("MIC","GTXE"): "GTXE .. GTX ECN",
        ("MIC","GTXS"): "GTXS .. GTX SEF, LLC",
        ("MIC","GXGF"): "GXGF .. GXG MTF FIRST QUOTE",
        ("MIC","GXGM"): "GXGM .. GXG MTF",
        ("MIC","GXGR"): "GXGR .. GXG MARKETS A/S",
        ("MIC","GXMA"): "GXMA .. GX MARKETCENTER",
        ("MIC","HAMA"): "HAMA .. BOERSE HAMBURG - REGULIERTER MARKT",
        ("MIC","HAMB"): "HAMB .. BOERSE HAMBURG - FREIVERKEHR",
        ("MIC","HAML"): "HAML .. BOERSE HAMBURG - LANG AND SCHWARZ EXCHANGE",
        ("MIC","HAMM"): "HAMM .. BOERSE HAMBURG - LANG AND SCHWARZ EXCHANGE - REGULIERTER MARKT",
        ("MIC","HAMN"): "HAMN .. BOERSE HAMBURG - LANG AND SCHWARZ EXCHANGE - FREIVERKEHR",
        ("MIC","HANA"): "HANA .. BOERSE HANNOVER - REGULIERTER MARKT",
        ("MIC","HANB"): "HANB .. BOERSE HANNOVER - FREIVERKEHR",
        ("MIC","HBFR"): "HBFR .. HSBC FRANCE - SYSTEMATIC INTERNALISER",
        ("MIC","HBPL"): "HBPL .. HSBC BANK POLSKA S.A. - SYSTEMATIC INTERNALISER",
        ("MIC","HCHC"): "HCHC .. ICE CLEAR NETHERLANDS B.V.",
        ("MIC","HDAT"): "HDAT .. ELECTRONIC SECONDARY SECURITIES MARKET (HDAT)",
        ("MIC","HEGX"): "HEGX .. NADEX",
        ("MIC","HELA"): "HELA .. HELABA - SYSTEMATIC INTERNALISER",
        ("MIC","HEMO"): "HEMO .. HENEX S.A.",
        ("MIC","HMOD"): "HMOD .. HI-MTF ORDER DRIVEN",
        ("MIC","HMTF"): "HMTF .. HI-MTF",
        ("MIC","HOTC"): "HOTC .. HELLENIC EXCHANGE OTC MARKET",
        ("MIC","HPCO"): "HPCO .. HPC SA - SWAPTIONS",
        ("MIC","HPCS"): "HPCS .. HPC SA - COMMODITIES",
        ("MIC","HPCV"): "HPCV .. HPC SA - VOICE OTF",
        ("MIC","HPCX"): "HPCX .. HPC SA",
        ("MIC","HPPO"): "HPPO .. POTAMUS TRADING LLC",
        ("MIC","HRFQ"): "HRFQ .. HI-MTF RFQ",
        ("MIC","HRSI"): "HRSI .. HUDSON RIVER TRADING - SYSTEMATIC INTERNALISER",
        ("MIC","HRTF"): "HRTF .. HUDSON RIVER TRADING (HRT)",
        ("MIC","HRTX"): "HRTX .. HUDSON RIVER TRADING",
        ("MIC","HSBC"): "HSBC .. HSBC - SYSTEMATIC INTERNALISER",
        ("MIC","HSBT"): "HSBT .. HSBC TRINKAUS AND BURKHARDT AG - SYSTEMATIC INTERNALISER",
        ("MIC","HSFX"): "HSFX .. HOTSPOT FX",
        ("MIC","HSTC"): "HSTC .. HANOI STOCK EXCHANGE",
        ("MIC","HSXA"): "HSXA .. HSBC-X HONG KONG",
        ("MIC","HUDX"): "HUDX .. HUNGARIAN DERIVATIVE ENERGY EXCHANGE",
        ("MIC","HUNG"): "HUNG .. MTS HUNGARY",
        ("MIC","HUPX"): "HUPX .. HUNGARIAN POWER EXCHANGE",
        ("MIC","IATS"): "IATS .. INTERACTIVE BROKERS LLC - ATS",
        ("MIC","IBAL"): "IBAL .. ICE BENCHMARK ADMINISTRATION",
        ("MIC","IBEQ"): "IBEQ .. INVEST BANCA - IBIS EQUITY - SYSTEMATIC INTERNALISER",
        ("MIC","IBEX"): "IBEX .. INDEPENDENT BULGARIAN ENERGY EXCHANGE",
        ("MIC","IBGH"): "IBGH .. IBERIAN GAS HUB",
        ("MIC","IBIS"): "IBIS .. INVEST BANCA - SYSTEMATIC INTERNALISER",
        ("MIC","IBUL"): "IBUL .. BULGARIAN STOCK EXCHANGE - INTERNATIONAL INSTRUMENTS",
        ("MIC","ICBX"): "ICBX .. INSTINET CBX (US)",
        ("MIC","ICDX"): "ICDX .. INDONESIA COMMODITY AND DERIVATIVES EXCHANGE",
        ("MIC","ICEL"): "ICEL .. ISLAND ECN LTD, THE",
        ("MIC","ICEO"): "ICEO .. ICAP ENERGY - OTF",
        ("MIC","ICES"): "ICES .. ICE SWAP TRADE LLC",
        ("MIC","ICOT"): "ICOT .. ICAP EU - OTF",
        ("MIC","ICPM"): "ICPM .. ICAP MTF",
        ("MIC","ICRO"): "ICRO .. INSTINET VWAP CROSS",
        ("MIC","ICSU"): "ICSU .. ICAP SEF (US)  LLC.",
        ("MIC","ICXL"): "ICXL .. INDIAN COMMODITY EXCHANGE LTD.",
        ("MIC","IECL"): "IECL .. ICAP ENERGY LTD OTF - COMMODITIES AND ENERGY DERIVATIVES",
        ("MIC","IEPA"): "IEPA .. INTERCONTINENTAL EXCHANGE",
        ("MIC","IEXD"): "IEXD .. INVESTORS EXCHANGE - DARK",
        ("MIC","IEXG"): "IEXG .. INVESTORS EXCHANGE",
        ("MIC","IFCA"): "IFCA .. ICE FUTURES CANADA",
        ("MIC","IFED"): "IFED .. ICE FUTURES U.S. ENERGY DIVISION",
        ("MIC","IFEN"): "IFEN .. ICE FUTURES EUROPE - OIL AND REFINED PRODUCTS DIVISION",
        ("MIC","IFEU"): "IFEU .. ICE FUTURES EUROPE",
        ("MIC","IFLL"): "IFLL .. ICE FUTURES EUROPE - FINANCIAL PRODUCTS DIVISION",
        ("MIC","IFLO"): "IFLO .. ICE FUTURES EUROPE - EQUITY PRODUCTS DIVISION",
        ("MIC","IFLS"): "IFLS .. SWAPXECUTE",
        ("MIC","IFLX"): "IFLX .. ICE FUTURES EUROPE - AGRICULTURAL PRODUCTS DIVISION",
        ("MIC","IFSG"): "IFSG .. ICE FUTURES SINGAPORE",
        ("MIC","IFSM"): "IFSM .. INSTITUTIONAL FINANCIAL SECURITIES MARKET",
        ("MIC","IFUS"): "IFUS .. ICE FUTURES U.S.",
        ("MIC","IFUT"): "IFUT .. ICE FUTURES EUROPE - EUROPEAN UTILITIES DIVISION",
        ("MIC","IGDL"): "IGDL .. ICAP GLOBAL DERIVATIVES LIMITED",
        ("MIC","IINX"): "IINX .. INDIA INTERNATIONAL EXCHANGE (IFSC) LIMITED",
        ("MIC","IKBS"): "IKBS .. IKB DEUTSCHE INDUSTRIEBANK AG - SYSTEMATIC INTERNALISER",
        ("MIC","IMAG"): "IMAG .. ICE MARKETS AGRICULTURE",
        ("MIC","IMBD"): "IMBD .. ICE MARKETS BONDS",
        ("MIC","IMCC"): "IMCC .. CREDITEX SECURITIES CORPORATION",
        ("MIC","IMCE"): "IMCE .. ICAP MTF - CASH EQUITY",
        ("MIC","IMCG"): "IMCG .. CREDITEX LLC",
        ("MIC","IMCO"): "IMCO .. ICE ENDEX OTF FUTURES",
        ("MIC","IMCR"): "IMCR .. ICE MARKETS CREDIT",
        ("MIC","IMCS"): "IMCS .. IMC FINANCIAL MARKETS",
        ("MIC","IMCT"): "IMCT .. IMC - SYSTEMATIC INTERNALISER",
        ("MIC","IMED"): "IMED .. ICAP MTF - EQUITY DERIVATIVES",
        ("MIC","IMEN"): "IMEN .. ICE MARKETS ENERGY",
        ("MIC","IMEQ"): "IMEQ .. ICE MARKETS EQUITY",
        ("MIC","IMET"): "IMET .. ICAP MTF - ETFS",
        ("MIC","IMEX"): "IMEX .. IRAN MERCANTILE EXCHANGE",
        ("MIC","IMFD"): "IMFD .. ICAP MTF - FX DERIVATIVES",
        ("MIC","IMFX"): "IMFX .. ICE MARKETS FOREIGN EXCHANGE",
        ("MIC","IMGB"): "IMGB .. ICAP MTF - GOVERNMENT BONDS EXCLUDING GILTS",
        ("MIC","IMGI"): "IMGI .. ICAP MTF - GILTS",
        ("MIC","IMIR"): "IMIR .. ICE MARKETS RATES",
        ("MIC","IMMH"): "IMMH .. IMMH - BANCA IMI - SYSTEMATIC INTERNALISER",
        ("MIC","IMMM"): "IMMM .. ICAP MTF - MONEY MARKET INSTRUMENTS",
        ("MIC","IMRD"): "IMRD .. ICAP MTF - INTEREST RATE DERIVATIVES",
        ("MIC","IMSB"): "IMSB .. ICAP MTF - CORPORATE BONDS AND SECURITIES DEBT",
        ("MIC","IMTF"): "IMTF .. INTEGRAL MTF",
        ("MIC","IMTS"): "IMTS .. MTS IRELAND",
        ("MIC","INCA"): "INCA .. INSTINET DESK CROSS",
        ("MIC","INCR"): "INCR .. INTELLIGENTCROSS",
        ("MIC","INGB"): "INGB .. ING BANK NV - SYSTEMATIC INTERNALISER",
        ("MIC","INGE"): "INGE .. ING BANK NV - SPRINTERS AND EQUITY - SYSTEMATIC INTERNALISER",
        ("MIC","INGF"): "INGF .. ING BANK NV - FOREIGN EXCHANGE - SYSTEMATIC INTERNALISER",
        ("MIC","INGW"): "INGW .. ING BANK SLASKI SA - SYSTEMATIC INTERNALISER",
        ("MIC","INSE"): "INSE .. NSE IFSC LIMITED",
        ("MIC","INVE"): "INVE .. INVESTEC BANK PLC - SYSTEMATIC INTERNALISER",
        ("MIC","IOCD"): "IOCD .. ICAP SECURITIES OTF - CREDIT DERIVATIVES",
        ("MIC","IOED"): "IOED .. ICAP SECURITIES OTF - EQUITY DERIVATIVES",
        ("MIC","IOFI"): "IOFI .. ICAP SECURITIES OTF - CORPORATE BONDS AND SECURITISED DEBT",
        ("MIC","IOFX"): "IOFX .. ICAP SECURITIES OTF - FX DERIVATIVES",
        ("MIC","IOGB"): "IOGB .. ICAP SECURITIES OTF - GOVERNMENT BONDS EXCLUDING  UK GILTS",
        ("MIC","IOGI"): "IOGI .. ICAP SECURITIES OTF - GOVERNMENT BONDS - UK GILTS",
        ("MIC","IOIR"): "IOIR .. ICAP SECURITIES OTF - INTEREST RATE DERIVATIVES",
        ("MIC","IOMM"): "IOMM .. ICAP SECURITIES OTF - MONEY MARKET INSTRUMENTS",
        ("MIC","IOTC"): "IOTC .. INSTINET EUROPE LIMITED OTC",
        ("MIC","IOTF"): "IOTF .. ICAP SECURITIES - OTF",
        ("MIC","IPSX"): "IPSX .. IPSX",
        ("MIC","IPXP"): "IPXP .. IPSX PRIME",
        ("MIC","ISBA"): "ISBA .. BANCA DI ASTI - SYSTEMATIC INTERNALISER",
        ("MIC","ISBV"): "ISBV .. BIVER BANCA - SYSTEMATIC INTERNALISER",
        ("MIC","ISDA"): "ISDA .. ISDAFIX",
        ("MIC","ISEX"): "ISEX .. INTER-CONNECTED STOCK EXCHANGE OF INDIA LTD",
        ("MIC","ISSI"): "ISSI .. ICBC STANDARD BANK - SYSTEMATIC INTERNALISER",
        ("MIC","ISWA"): "ISWA .. I-SWAP",
        ("MIC","ISWB"): "ISWB .. I-SWAP - TRADE REGISTRATION",
        ("MIC","ISWC"): "ISWC .. I-SWAP ORDER BOOK",
        ("MIC","ISWE"): "ISWE .. ICAP GLOBAL DERIVATIVES LIMITED - ELECTRONIC",
        ("MIC","ISWR"): "ISWR .. I-SWAP TARGETED STREAMING/RFQ",
        ("MIC","ISWV"): "ISWV .. ICAP GLOBAL DERIVATIVES LIMITED - VOICE",
        ("MIC","ITGI"): "ITGI .. ITG - POSIT",
        ("MIC","ITGL"): "ITGL .. ITG - POSIT MTF",
        ("MIC","IVZX"): "IVZX .. INVESCO CANADA PTF TRADES",
        ("MIC","JADX"): "JADX .. JOINT ASIAN DERIVATIVES EXCHANGE",
        ("MIC","JBSI"): "JBSI .. JYSKE BANK - SYSTEMATIC INTERNALISER",
        ("MIC","JEFS"): "JEFS .. JEFFERIES INTERNATIONAL - SYSTEMATIC INTERNALISER",
        ("MIC","JEFX"): "JEFX .. JETX",
        ("MIC","JISI"): "JISI .. JEFFERIES INTERNATIONAL",
        ("MIC","JLQD"): "JLQD .. JUMP LIQUIDITY",
        ("MIC","JLSI"): "JLSI .. JUMP LIQUIDITY - SYSTEMATIC INTERNALISER",
        ("MIC","JPBX"): "JPBX .. JPBX",
        ("MIC","JPCB"): "JPCB .. JPMORGAN CHASE BANK N.A. LONDON BRANCH",
        ("MIC","JPEU"): "JPEU .. J.P. MORGAN AG - SYSTEMATIC INTERNALISER",
        ("MIC","JPMI"): "JPMI .. JP MORGAN - JPMI MARKET",
        ("MIC","JPMX"): "JPMX .. JPMX",
        ("MIC","JPSI"): "JPSI .. J.P. MORGAN SECURITIES PLC",
        ("MIC","JSEB"): "JSEB .. JSE BOND ELECTRONIC TRADING PLATFORM",
        ("MIC","JSES"): "JSES .. JANE STREET EXECUTION SERVICES LLC",
        ("MIC","JSJX"): "JSJX .. JANE STREET JX",
        ("MIC","JSSI"): "JSSI .. JANE STREET FINANCIAL LTD - SYSTEMATIC INTERNALISER",
        ("MIC","JYSI"): "JYSI .. DEN JYSKE SPAREKASSE - SYSTEMATIC INTERNALISER",
        ("MIC","KAIX"): "KAIX .. KAI-X",
        ("MIC","KBCB"): "KBCB .. KBC BANK NV  GROUP MARKETS - SYSTEMATIC INTERNALISER",
        ("MIC","KBLM"): "KBLM .. KYTE BROKING LIMITED",
        ("MIC","KCCP"): "KCCP .. KELER CCP",
        ("MIC","KELR"): "KELR .. KELER",
        ("MIC","KLEU"): "KLEU .. KNIGHT LINK EUROPE",
        ("MIC","KMUX"): "KMUX .. ZURCHER KANTONALBANK - EKMU-X",
        ("MIC","KNCM"): "KNCM .. KNIGHT CAPITAL MARKETS LLC",
        ("MIC","KNEM"): "KNEM .. KNIGHT EQUITY MARKETS LP",
        ("MIC","KNIG"): "KNIG .. KNIGHT",
        ("MIC","KNLI"): "KNLI .. KNIGHT LINK",
        ("MIC","KNMX"): "KNMX .. KNIGHT MATCH ATS",
        ("MIC","KOTF"): "KOTF .. OTFLINK",
        ("MIC","LASF"): "LASF .. LATAM SEF",
        ("MIC","LASP"): "LASP .. LAN AND SPAR BANK A/S",
        ("MIC","LBBW"): "LBBW .. LBBW - LANDESBANK BADEN-WUERTTEMBERG",
        ("MIC","LBCM"): "LBCM .. LLOYDS BANK CORPORATE MARKETS - SYSTEMATIC INTERNALISER",
        ("MIC","LBWL"): "LBWL .. LBBW - LANDESBANK BADEN-WUERTTEMBERG - LIQUIDITY PROVIDER",
        ("MIC","LBWS"): "LBWS .. LBBW - LANDESBANK BADEN-WUERTTEMBERG - SYSTEMATIC INTERNALISER",
        ("MIC","LCHC"): "LCHC .. LCH.CLEARNET",
        ("MIC","LCUR"): "LCUR .. CURRENEX LDFX",
        ("MIC","LEDG"): "LEDG .. LEDGERX",
        ("MIC","LEUE"): "LEUE .. LIQUIDNET EU LIMITED EQUITY MTF",
        ("MIC","LEUF"): "LEUF .. LIQUIDNET EU LIMITED FIXED INCOME MTF",
        ("MIC","LEVL"): "LEVL .. LEVEL ATS",
        ("MIC","LICA"): "LICA .. LIQUIDNET CANADA ATS",
        ("MIC","LIFI"): "LIFI .. LIQUIDNET, INC. FIXED INCOME ATS",
        ("MIC","LIGA"): "LIGA .. LIGA BANK EG - SYSTEMATIC INTERNALISER",
        ("MIC","LIQF"): "LIQF .. LIQUIDNET EUROPE LIMITED",
        ("MIC","LIQH"): "LIQH .. LIQUIDNET H20",
        ("MIC","LIQU"): "LIQU .. LIQUIDNET SYSTEMS",
        ("MIC","LISX"): "LISX .. CBOE  EUROPE - LIS SERVICE",
        ("MIC","LISZ"): "LISZ .. CBOE EUROPE - LIS SERVICE (NL)",
        ("MIC","LIUH"): "LIUH .. LIQUIDNET, INC. H2O ATS",
        ("MIC","LIUS"): "LIUS .. LIQUIDNET, INC.",
        ("MIC","LMAD"): "LMAD .. LMAX - DERIVATIVES",
        ("MIC","LMAE"): "LMAE .. LMAX - EQUITIES",
        ("MIC","LMAF"): "LMAF .. LMAX - FX",
        ("MIC","LMAO"): "LMAO .. LMAX - INDICES/RATES/COMMODITIES",
        ("MIC","LMAX"): "LMAX .. LMAX",
        ("MIC","LMEC"): "LMEC .. LME CLEAR",
        ("MIC","LMNR"): "LMNR .. LUMINOR BANK - SYSTEMATIC INTERNALISER",
        ("MIC","LMNX"): "LMNX .. LUMINEX TRADING & ANALYTICS LLC - ATS",
        ("MIC","LOTC"): "LOTC .. OTC MARKET",
        ("MIC","LOUI"): "LOUI .. LOUIS CAPITAL MARKETS UK LLP",
        ("MIC","LOYD"): "LOYD .. LLOYDS BANK - SYSTEMATIC INTERNALISER",
        ("MIC","LPPM"): "LPPM .. LONDON PLATINUM AND PALLADIUM MARKET",
        ("MIC","LQED"): "LQED .. LIQUIDITYEDGE",
        ("MIC","LQFI"): "LQFI .. CITI LIQUIFI",
        ("MIC","LSSI"): "LSSI .. LANG AND SCHWARZ TRADE CENTER - SYSTEMATIC INTERNALISER",
        ("MIC","LTAA"): "LTAA .. LUMINEX TRADING & ANALYTICS LLC",
        ("MIC","LXJP"): "LXJP .. BARCLAYS LX JAPAN",
        ("MIC","LYNX"): "LYNX .. LYNX ATS",
        ("MIC","MABX"): "MABX .. MERCADO ALTERNATIVO BURSATIL",
        ("MIC","MACB"): "MACB .. MACQUARIE AUSTRALIA BLOCK CROSSING",
        ("MIC","MAEL"): "MAEL .. MARKETAXESS EUROPE LIMITED",
        ("MIC","MAGM"): "MAGM .. MAGMA ATS",
        ("MIC","MAKX"): "MAKX .. MAKOR SECURITIES LONDON LTD",
        ("MIC","MALX"): "MALX .. MALDIVES STOCK EXCHANGE",
        ("MIC","MANL"): "MANL .. MARKETAXESS B.V.",
        ("MIC","MAQH"): "MAQH .. MACQUARIE INTERNAL MARKETS (HONG KONG)",
        ("MIC","MAQI"): "MAQI .. MACQUARIE BANK INTERNATIONAL LIMITED - SYSTEMATIC INTERNALISER",
        ("MIC","MAQJ"): "MAQJ .. MACQUARIE INTERNAL MARKETS (JAPAN)",
        ("MIC","MAQL"): "MAQL .. MACQUARIE CAPITAL EUROPE LIMITED",
        ("MIC","MAQU"): "MAQU .. MACQUARIE BANK LIMITED, LONDON BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","MAQX"): "MAQX .. MACQUARIE INTERNAL MARKETS (AUSTRALIA)",
        ("MIC","MARF"): "MARF .. MERCADO ALTERNATIVO DE RENTA FIJA",
        ("MIC","MASG"): "MASG .. MARKETAXESS SINGAPORE PTE LIMITED",
        ("MIC","MATN"): "MATN .. MATCH NOW",
        ("MIC","MBPL"): "MBPL .. MBANK S.A. - SYSTEMATIC INTERNALISER",
        ("MIC","MBUL"): "MBUL .. MTF SOFIA",
        ("MIC","MCRY"): "MCRY .. ISE MERCURY, LLC",
        ("MIC","MCSE"): "MCSE .. NASDAQ COPENHAGEN A/S - AUCTION ON DEMAND",
        ("MIC","MCUR"): "MCUR .. CURRENEX MTF",
        ("MIC","MCXR"): "MCXR .. CURRENEX MTF - RFQ",
        ("MIC","MCXS"): "MCXS .. CURRENEX MTF - STREAMING",
        ("MIC","MCXX"): "MCXX .. MCX STOCK EXCHANGE LTD",
        ("MIC","MCZK"): "MCZK .. MTS CZECH REPUBLIC",
        ("MIC","MDRV"): "MDRV .. MIBGAS - DERIVATIVES",
        ("MIC","MELO"): "MELO .. NASDAQ MIDPOINT-ELO (M-ELO)",
        ("MIC","MERD"): "MERD .. MERKUR MARKET - DARK POOL",
        ("MIC","MERF"): "MERF .. MERCADO ELECTRONICO DE RENTA FIJA",
        ("MIC","MERK"): "MERK .. MERKUR MARKET",
        ("MIC","MESQ"): "MESQ .. ACE MARKET",
        ("MIC","METZ"): "METZ .. MET ZURICH",
        ("MIC","MFGL"): "MFGL .. MF GLOBAL ENERGY MTF",
        ("MIC","MFOX"): "MFOX .. EURONEXT - MERCADO DE FUTUROS E OPCOES",
        ("MIC","MFXA"): "MFXA .. FX CONNECT - MTF - ALLOCATIONS",
        ("MIC","MFXC"): "MFXC .. FX CONNECT - MTF",
        ("MIC","MFXR"): "MFXR .. FX CONNECT - MTF - RFQ",
        ("MIC","MHBD"): "MHBD .. MIZUHO BANK, LTD. DUESSELDORF BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","MHEL"): "MHEL .. NASDAQ HELSINKI LTD -  AUCTION ON DEMAND",
        ("MIC","MHIP"): "MHIP .. MIZUHO INTERNATIONAL - SYSTEMATIC INTERNALISER",
        ("MIC","MIBG"): "MIBG .. MERCADO ORGANIZADO DEL GAS",
        ("MIC","MICE"): "MICE .. NASDAQ ICELAND HF. - AUCTION ON DEMAND",
        ("MIC","MIHI"): "MIHI .. MIAMI INTERNATIONAL HOLDINGS, INC.",
        ("MIC","MISX"): "MISX .. MOSCOW EXCHANGE",
        ("MIC","MIVX"): "MIVX .. MARKET FOR INVESTMENT VEHICLES",
        ("MIC","MIZX"): "MIZX .. MIZUHO INTERNAL CROSSING",
        ("MIC","MKAP"): "MKAP .. MERITKAPITAL - SYSTEMATIC INTERNALISER",
        ("MIC","MLCO"): "MLCO .. BANK OF AMERICA - MERRILL LYNCH OTC",
        ("MIC","MLER"): "MLER .. BOFAML SECURITIES EUROPE - RFQ - SYSTEMATIC INTERNALISER",
        ("MIC","MLES"): "MLES .. BOFAML SECURITIES EUROPE - SYSTEMATIC INTERNALISER",
        ("MIC","MLEX"): "MLEX .. BOFAML SECURITIES EUROPE",
        ("MIC","MLIB"): "MLIB .. MERRILL LYNCH INTERNATIONAL BANK DESIGNATED ACTIVITY COMPANY",
        ("MIC","MLIX"): "MLIX .. MERRILL LYNCH INTERNATIONAL",
        ("MIC","MLRQ"): "MLRQ .. MERRILL LYNCH INTERNATIONAL - RFQ - SYSTEMATIC INTERNALISER",
        ("MIC","MLSI"): "MLSI .. MERRILL LYNCH INTERNATIONAL - SYSTEMATIC INTERNALISER",
        ("MIC","MLVX"): "MLVX .. BANK OF AMERICA - MERRILL LYNCH VWAP CROSS",
        ("MIC","MLXB"): "MLXB .. EURONEXT ACCESS BRUSSELS",
        ("MIC","MNDK"): "MNDK .. FIRST NORTH DENMARK - AUCTION ON DEMAND",
        ("MIC","MNFI"): "MNFI .. FIRST NORTH FINLAND - AUCTION ON DEMAND",
        ("MIC","MNIS"): "MNIS .. FIRST NORTH ICELAND - AUCTION ON DEMAND",
        ("MIC","MNSE"): "MNSE .. FIRST NORTH SWEDEN - AUCTION ON DEMAND",
        ("MIC","MOCX"): "MOCX .. MOC CROSS",
        ("MIC","MOSE"): "MOSE .. FIRST NORTH SWEDEN - NORWAY AUCTION ON DEMAND",
        ("MIC","MOTX"): "MOTX .. ELECTRONIC BOND MARKET",
        ("MIC","MPRL"): "MPRL .. MIAX PEARL, LLC",
        ("MIC","MSAL"): "MSAL .. MORGAN STANLEY AUSTRALIA SECURITIES LIMITED",
        ("MIC","MSCO"): "MSCO .. MORGAN STANLEY AND CO. LLC",
        ("MIC","MSIP"): "MSIP .. MORGAN STANLEY AND CO. INTERNATIONAL PLC",
        ("MIC","MSMS"): "MSMS .. MORGAN STANLEY MUFG SECURITIES CO., LTD",
        ("MIC","MSPL"): "MSPL .. MS POOL",
        ("MIC","MSRP"): "MSRP .. MS RPOOL",
        ("MIC","MSSI"): "MSSI .. MORGAN STANLEY AND CO. INTERNATIONAL PLC - SYSTEMATIC INTERNALISER",
        ("MIC","MSTO"): "MSTO .. NASDAQ STOCKHOLM AB - AUCTION ON DEMAND",
        ("MIC","MSTX"): "MSTX .. MS TRAJECTORY CROSS",
        ("MIC","MTAA"): "MTAA .. ELECTRONIC SHARE MARKET",
        ("MIC","MTAH"): "MTAH .. BORSA ITALIANA EQUITY MTF",
        ("MIC","MTCH"): "MTCH .. BONDMATCH",
        ("MIC","MTSA"): "MTSA .. MTS AUSTRIA",
        ("MIC","MTSB"): "MTSB .. MTS BONDS.COM",
        ("MIC","MTSC"): "MTSC .. MTS ITALIA",
        ("MIC","MTSD"): "MTSD .. MTS DENMARK",
        ("MIC","MTSF"): "MTSF .. MTS FINLAND",
        ("MIC","MTSG"): "MTSG .. MTS GREECE",
        ("MIC","MTSO"): "MTSO .. MTS S.P.A.",
        ("MIC","MTUS"): "MTUS .. MTS MARKETS INTERNATIONAL INC.",
        ("MIC","MTXA"): "MTXA .. MARKETAXESS CANADA COMPANY",
        ("MIC","MTXC"): "MTXC .. MARKETAXESS CORPORATION SINGLE-NAME CDS CENTRAL LIMIT ORDER",
        ("MIC","MTXM"): "MTXM .. MARKETAXESS CORPORATION MID-X TRADING SYSTEM",
        ("MIC","MTXS"): "MTXS .. MARKETAXESS SEF CORPORATION",
        ("MIC","MTXX"): "MTXX .. MARKETAXESS CORPORATION",
        ("MIC","MUBE"): "MUBE .. MUFG BANK (EUROPE) N.V. - SYSTEMATIC INTERNALISER",
        ("MIC","MUBL"): "MUBL .. MUFG BANK, LTD. - LONDON BRANCH -SYSTEMATIC INTERNALISER",
        ("MIC","MUBM"): "MUBM .. MUFG BANK, LTD. - MILANO BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","MUBP"): "MUBP .. MUFG BANK, LTD. - PARIS BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","MUFP"): "MUFP .. MARIANA UFP LLP",
        ("MIC","MUNA"): "MUNA .. BOERSE MUENCHEN - REGULIERTER MARKT",
        ("MIC","MUNB"): "MUNB .. BOERSE MUENCHEN - FREIVERKEHR",
        ("MIC","MUNC"): "MUNC .. BOERSE MUENCHEN - GETTEX - REGULIERTER MARKT",
        ("MIC","MUND"): "MUND .. BOERSE MUENCHEN - GETTEX - FREIVERKEHR",
        ("MIC","MUSE"): "MUSE .. MUFG SECURITIES EMEA PLC - SYSTEMATIC INTERNALISER",
        ("MIC","MUTI"): "MUTI .. MITSUBISHI UFJ TRUST INTERNATIONAL LIMITED",
        ("MIC","MVCX"): "MVCX .. MERCADO DE VALORES DE CORDOBA S.A.",
        ("MIC","MYTR"): "MYTR .. MYTREASURY",
        ("MIC","N2EX"): "N2EX .. N2EX",
        ("MIC","NABA"): "NABA .. NATIONAL AUSTRALIA BANK - FX DERIVATIVES AND FIXED INCOME SECURITIES - SYSTEMATIC INTERNALISER",
        ("MIC","NABE"): "NABE .. NAB EUROPE LIMITED - SYSTEMATIC INTERNALISER",
        ("MIC","NABL"): "NABL .. NATIONAL AUSTRALIA BANK - SYSTEMATIC INTERNALISER",
        ("MIC","NABU"): "NABU .. NAB EUROPE LIMITED - SYSTEMATIC INTERNALISER - FIXED INCOME SECURITIES",
        ("MIC","NAMX"): "NAMX .. NATIONAL MERCANTILE EXCHANGE",
        ("MIC","NAPA"): "NAPA .. NASDAQ STOCKHOLM AB - APA SERVICE",
        ("MIC","NASD"): "NASD .. NSDQ DARK",
        ("MIC","NASX"): "NASX .. NASD OTC MARKET",
        ("MIC","NATX"): "NATX .. NATIXIS - SYSTEMATIC INTERNALISER",
        ("MIC","NBLX"): "NBLX .. NOBLE EXCHANGE",
        ("MIC","NBOT"): "NBOT .. NATIONAL BOARD OF TRADE LIMITED",
        ("MIC","NCEL"): "NCEL .. PAKISTAN MERCANTILE EXCHANGE",
        ("MIC","NCML"): "NCML .. SMBC NIKKO CAPITAL MARKETS LIMITED",
        ("MIC","NDCM"): "NDCM .. ICE ENDEX UK OCM GAS SPOT",
        ("MIC","NDEX"): "NDEX .. ICE ENDEX FUTURES",
        ("MIC","NDXS"): "NDXS .. ICE ENDEX EUROPEAN GAS SPOT",
        ("MIC","NEOC"): "NEOC .. AEQUITAS NEO EXCHANGE - CONNECT MARKET",
        ("MIC","NEOD"): "NEOD .. AEQUITAS NEO EXCHANGE - DARK MARKET",
        ("MIC","NEOE"): "NEOE .. AEQUITAS NEO EXCHANGE",
        ("MIC","NEON"): "NEON .. AEQUITAS NEO EXCHANGE - NEO LIT MARKET BY PRICE",
        ("MIC","NESI"): "NESI .. NOMURA EUROPE - SYSTEMATIC INTERNALISER",
        ("MIC","NEXD"): "NEXD .. NEX EXCHANGE MAIN BOARD (NON-EQUITY)",
        ("MIC","NEXF"): "NEXF .. NEX EXCHANGE GROWTH (NON-EQUITY)",
        ("MIC","NEXG"): "NEXG .. NEX EXCHANGE GROWTH (EQUITY)",
        ("MIC","NEXL"): "NEXL .. NEX EXCHANGE MAIN BOARD (EQUITY)",
        ("MIC","NEXN"): "NEXN .. NEX EXCHANGE TRADING (NON-EQUITY)",
        ("MIC","NEXO"): "NEXO .. NOREXECO ASA",
        ("MIC","NEXS"): "NEXS .. NEX SEF",
        ("MIC","NEXT"): "NEXT .. NEX EXCHANGE TRADING (EQUITY)",
        ("MIC","NEXX"): "NEXX .. NEX EXCHANGE",
        ("MIC","NEXY"): "NEXY .. EBS MTF - RFQ - FOR CORPORATES TRADING FX PRODUCTS",
        ("MIC","NFSA"): "NFSA .. FIDELITY CROSSSTREAM",
        ("MIC","NFSC"): "NFSC .. NATIONAL FINANCIAL SERVICES, LLC",
        ("MIC","NFSD"): "NFSD .. FIDELITY DARK",
        ("MIC","NGXC"): "NGXC .. NATURAL GAS EXCHANGE",
        ("MIC","NIBR"): "NIBR .. NORWEGIAN INTER BANK OFFERED RATE",
        ("MIC","NILX"): "NILX .. NILE STOCK EXCHANGE",
        ("MIC","NLBX"): "NLBX .. NORDDEUTSCHE LANDESBANK - GIROZENTRALE - SYSTEMATIC INTERNALISER",
        ("MIC","NLPX"): "NLPX .. APX POWER NL",
        ("MIC","NMCE"): "NMCE .. NATIONAL MULTI-COMMODITY EXCHANGE OF INDIA",
        ("MIC","NMRA"): "NMRA .. NOMURA SECURITIES INTERNATIONAL",
        ("MIC","NMRJ"): "NMRJ .. NOMURA SECURITIES CO LTD",
        ("MIC","NMTF"): "NMTF .. NORDIC MTF",
        ("MIC","NOCO"): "NOCO .. NASDAQ STOCKHOLM AB - COMMODITIES",
        ("MIC","NODX"): "NODX .. NODAL EXCHANGE",
        ("MIC","NOED"): "NOED .. NASDAQ STOCKHOLM AB - NORWEGIAN EQ DERIVATIVES",
        ("MIC","NOFF"): "NOFF .. NOMURA OTC TRADES",
        ("MIC","NOFI"): "NOFI .. NASDAQ STOCKHOLM AB - NORWEGIAN FI DERIVATIVES",
        ("MIC","NOPS"): "NOPS .. NORD POOL SPOT AS",
        ("MIC","NORD"): "NORD .. HSH NORDBANK - SYSTEMATIC INTERNALISER",
        ("MIC","NORX"): "NORX .. NASDAQ OMX COMMODITIES",
        ("MIC","NOSC"): "NOSC .. NOS CLEARING ASA",
        ("MIC","NOSI"): "NOSI .. NOMURA - SYSTEMATIC INTERNALISER",
        ("MIC","NOTC"): "NOTC .. NORWEGIAN OVER THE COUNTER MARKET",
        ("MIC","NPEX"): "NPEX .. NPEX",
        ("MIC","NPGA"): "NPGA .. GASPOINT NORDIC A/S",
        ("MIC","NSXB"): "NSXB .. BENDIGO STOCK EXCHANGE LIMITED",
        ("MIC","NWMS"): "NWMS .. ROYAL BANK OF SCOTLAND - SYSTEMATIC INTERNALISER",
        ("MIC","NWNV"): "NWNV .. NATWEST MARKETS NV",
        ("MIC","NXEU"): "NXEU .. NX",
        ("MIC","NXJP"): "NXJP .. NX JAPAN",
        ("MIC","NXSE"): "NXSE .. NX SELECT JAPAN",
        ("MIC","NXUS"): "NXUS .. NX ATS - CROSSING PLATFORM",
        ("MIC","NXVW"): "NXVW .. NX VWAP",
        ("MIC","NYFX"): "NYFX .. MILLENNIUM",
        ("MIC","NYMS"): "NYMS .. CME SWAPS MARKETS (NYMEX)",
        ("MIC","NYPC"): "NYPC .. NEW YORK PORTFOLIO CLEARING",
        ("MIC","NYSD"): "NYSD .. NYSE DARK",
        ("MIC","NYSI"): "NYSI .. NYKREDIT BANK - SYSTEMATIC INTERNALISER",
        ("MIC","NZFX"): "NZFX .. NEW ZEALAND FUTURES & OPTIONS",
        ("MIC","OAPA"): "OAPA .. OSLO BORS - APA",
        ("MIC","OBKL"): "OBKL .. OBERBANK AG - SYSTEMATIC INTERNALISER",
        ("MIC","OCSI"): "OCSI .. SOVA CAPITAL - SYSTEMATIC INTERNALISER",
        ("MIC","ODDO"): "ODDO .. ODDO BHF",
        ("MIC","ODOC"): "ODOC .. ODDO CONTREPARTIE",
        ("MIC","OLLC"): "OLLC .. OTCEX LLC",
        ("MIC","OMEL"): "OMEL .. OMI POLO ESPANOL S.A. (OMIE)",
        ("MIC","OMGA"): "OMGA .. OMEGA ATS",
        ("MIC","OMIC"): "OMIC .. THE IBERIAN ENERGY CLEARING HOUSE",
        ("MIC","OMIP"): "OMIP .. OPERADOR DE MERCADO IBERICO DE ENERGIA - PORTUGAL",
        ("MIC","ONSE"): "ONSE .. FIRST NORTH SWEDEN - NORWAY",
        ("MIC","OOTC"): "OOTC .. OTHER OTC",
        ("MIC","OPCO"): "OPCO .. OP CORPORATE BANK PLC - SYSTEMATIC INTERNALISER",
        ("MIC","OPEX"): "OPEX .. PEX-PRIVATE EXCHANGE",
        ("MIC","OPRA"): "OPRA .. OPTIONS PRICE REPORTING AUTHORITY",
        ("MIC","OSLC"): "OSLC .. SIX X-CLEAR AG",
        ("MIC","OTCB"): "OTCB .. OTCQB MARKETPLACE",
        ("MIC","OTCE"): "OTCE .. OTCEX",
        ("MIC","OTCM"): "OTCM .. OTC MARKETS",
        ("MIC","OTCQ"): "OTCQ .. OTCQX MARKETPLACE",
        ("MIC","OTCX"): "OTCX .. OTC EXCHANGE OF INDIA",
        ("MIC","OTPB"): "OTPB .. OTP BANK NYRT - SYSTEMATIC INTERNALISER",
        ("MIC","OTXB"): "OTXB .. BERNER KANTONALBANK OTC-X",
        ("MIC","PARX"): "PARX .. PARFX",
        ("MIC","PAVE"): "PAVE .. ALTERNATIVE PLATFORM FOR SPANISH SECURITIES",
        ("MIC","PDEX"): "PDEX .. PHILIPPINE DEALING AND EXCHANGE CORP",
        ("MIC","PDQX"): "PDQX .. CODA MARKETS",
        ("MIC","PEEL"): "PEEL .. PEEL HUNT LLP UK",
        ("MIC","PFTQ"): "PFTQ .. PFTS QUOTE DRIVEN",
        ("MIC","PFTS"): "PFTS .. PFTS STOCK EXCHANGE",
        ("MIC","PIEU"): "PIEU .. ARITAS FINANCIAL LTD",
        ("MIC","PINC"): "PINC .. OTC PINK CURRENT",
        ("MIC","PINI"): "PINI .. OTC PINK NO INFORMATION",
        ("MIC","PINL"): "PINL .. OTC PINK LIMITED",
        ("MIC","PINX"): "PINX .. OTC PINK MARKETPLACE",
        ("MIC","PIPE"): "PIPE .. ARITAS SECURITIES LLC",
        ("MIC","PIRM"): "PIRM .. PIRUM",
        ("MIC","PLDX"): "PLDX .. PLUS DERIVATIVES EXCHANGE",
        ("MIC","PLPD"): "PLPD .. WARSAW STOCK EXCHANGE/COMMODITIES/POLISH POWER EXCHANGE/COMMODITY DERIVATIVES",
        ("MIC","PLPX"): "PLPX .. WARSAW STOCK EXCHANGE/COMMODITIES/POLISH POWER EXCHANGE/ENERGY MARKET",
        ("MIC","PNED"): "PNED .. NASDAQ STOCKHOLM AB - PAN-NORDIC EQ DERIVATIVES",
        ("MIC","PORT"): "PORT .. MTS PORTUGAL",
        ("MIC","POTC"): "POTC .. PACIFIC OTC",
        ("MIC","PPEX"): "PPEX .. PROPERTY PARTNER EXCHANGE",
        ("MIC","PROS"): "PROS .. PROSPECTS",
        ("MIC","PRSE"): "PRSE .. PRAGMA ATS",
        ("MIC","PSGM"): "PSGM .. OTC GREY MARKET",
        ("MIC","PTPG"): "PTPG .. POLISH TRADING POINT",
        ("MIC","PURE"): "PURE .. PURE TRADING",
        ("MIC","PVMF"): "PVMF .. PVM OTF",
        ("MIC","PXIL"): "PXIL .. POWER EXCHANGE INDIA LTD.",
        ("MIC","QWIX"): "QWIX .. Q-WIXX PLATFORM",
        ("MIC","R5FX"): "R5FX .. R5FX LIMITED",
        ("MIC","RABO"): "RABO .. RABOBANK - SYSTEMATIC INTERNALISER",
        ("MIC","RBCB"): "RBCB .. RBC INVESTOR SERVICES BANK S.A.",
        ("MIC","RBCC"): "RBCC .. RBC - LONDON BRANCH",
        ("MIC","RBCE"): "RBCE .. RBC EUROPE LIMITED",
        ("MIC","RBCM"): "RBCM .. RBC - ROYAL BANK OF CANADA",
        ("MIC","RBCT"): "RBCT .. RBC INVESTOR SERVICES TRUST",
        ("MIC","RBHU"): "RBHU .. RAIFFEISEN BANK (HUNGARY) - SYSTEMATIC INTERNALISER",
        ("MIC","RBIV"): "RBIV .. RAIFFEISEN BANK INTERNATIONAL AG - SYSTEMATIC INTERNALISER",
        ("MIC","RBSI"): "RBSI .. RBC INVESTOR SERVICES BANK S.A. - SYSTEMATIC INTERNALISER",
        ("MIC","RBSX"): "RBSX .. RBS CROSS",
        ("MIC","RENC"): "RENC .. RENAISSANCE SECURITIES (CYPRUS) LIMITED",
        ("MIC","RESE"): "RESE .. EBS MTF - RESET",
        ("MIC","REST"): "REST .. NEX SEF MTF - RESET - RISK MITIGATION SERVICES",
        ("MIC","RICD"): "RICD .. RIVERCROSS DARK",
        ("MIC","RICX"): "RICX .. RIVERCROSS",
        ("MIC","RLBO"): "RLBO .. RAIFFEISENLANDESBANK OBEROSTERREICH - SYSTEMATIC INTERNALISER",
        ("MIC","RMTS"): "RMTS .. MTS ISRAEL",
        ("MIC","ROCO"): "ROCO .. TAIPEI EXCHANGE",
        ("MIC","ROFX"): "ROFX .. ROSARIO FUTURE EXCHANGE",
        ("MIC","ROSR"): "ROSR .. SIX REPO AG",
        ("MIC","ROTC"): "ROTC .. RWANDA OTC MARKET",
        ("MIC","RPDX"): "RPDX .. MOSCOW ENERGY EXCHANGE",
        ("MIC","RPWC"): "RPWC .. WARSAW STOCK EXCHANGE/BONDS/CATALYST/BONDSPOT/REGULATED MARKET",
        ("MIC","RSEX"): "RSEX .. RWANDA STOCK EXCHANGE",
        ("MIC","RTSI"): "RTSI .. RBC INVESTOR SERVICES TRUST - SYSTEMATIC INTERNALISER",
        ("MIC","RTSL"): "RTSL .. REUTERS TRANSACTION SERVICES LIMITED",
        ("MIC","RTSX"): "RTSX .. MOSCOW EXCHANGE-DERIVATIVES AND CLASSICA MARKET",
        ("MIC","RUSX"): "RUSX .. NON-PROFIT PARTNERSHIP FOR THE DEVELOPMENT OF FINANCIAL MARKET RTS",
        ("MIC","S3FM"): "S3FM .. SOCIETY3 FUNDERSMART",
        ("MIC","SANT"): "SANT .. BANCO SANTANDER S.A. - SYSTEMATIC INTERNALISER",
        ("MIC","SB1M"): "SB1M .. SPAREBANK 1 MARKETS - SYSTEMATIC INTERNALISER",
        ("MIC","SBAR"): "SBAR .. BOLSA DE BARCELONA RENTA FIJA",
        ("MIC","SBEX"): "SBEX .. SCOTIABANK - SYSTEMATIC INTERNALISER",
        ("MIC","SBIJ"): "SBIJ .. SBI JAPANNEXT-J-MARKET",
        ("MIC","SBIL"): "SBIL .. BOLSA DE BILBAO RENTA FIJA",
        ("MIC","SBIU"): "SBIU .. SBI JAPANNEXT - U - MARKET",
        ("MIC","SBIV"): "SBIV .. SBI JAPANNEXT - VWAP CROSSING",
        ("MIC","SBSI"): "SBSI .. SYDBANK A/S",
        ("MIC","SCXA"): "SCXA .. SEED DIGITAL SECURITIES MARKET",
        ("MIC","SCXF"): "SCXF .. SEED FUTURES",
        ("MIC","SCXM"): "SCXM .. SEED DIGITAL COMMODITIES MARKET",
        ("MIC","SCXO"): "SCXO .. SEED CX",
        ("MIC","SCXS"): "SCXS .. SEED SEF",
        ("MIC","SEBS"): "SEBS .. SEB - SYSTEMATIC INTERNALISER",
        ("MIC","SECF"): "SECF .. SECFINEX",
        ("MIC","SEDX"): "SEDX .. SECURITISED DERIVATIVES MARKET",
        ("MIC","SEED"): "SEED .. NASDAQ STOCKHOLM AB - SWEDISH EQ DERIVATIVES",
        ("MIC","SELC"): "SELC .. SISTEMA ESPECIAL DE LIQUIDACAO E CUSTODIA DE TITULOS PUBLICOS",
        ("MIC","SEMX"): "SEMX .. SEMOPX",
        ("MIC","SEND"): "SEND .. SEND - SISTEMA ELECTRONICO DE NEGOCIACION DE DEUDA",
        ("MIC","SEPE"): "SEPE .. STOCK EXCHANGE PERSPECTIVA",
        ("MIC","SFCL"): "SFCL .. SWISSCANTO FUNDS CENTRE LIMITED",
        ("MIC","SGEX"): "SGEX .. SHANGHAI GOLD EXCHANGE",
        ("MIC","SGMA"): "SGMA .. GOLDMAN SACH MTF",
        ("MIC","SGMT"): "SGMT .. SIGMA X2",
        ("MIC","SGMX"): "SGMX .. SIGMA X MTF",
        ("MIC","SGMY"): "SGMY .. SIGMA X MTF - AUCTION BOOK",
        ("MIC","SGOE"): "SGOE .. SG OPTION EUROPE - SYSTEMATIC INTERNALISER",
        ("MIC","SHAD"): "SHAD .. D.E. SHAW DARK",
        ("MIC","SHAR"): "SHAR .. ASSET MATCH",
        ("MIC","SHAW"): "SHAW .. D.E. SHAW",
        ("MIC","SHSC"): "SHSC .. STOCK EXCHANGE OF HONG KONG LIMITED - SHANGHAI - HONG KONG STOCK CONNECT",
        ("MIC","SIAB"): "SIAB .. ALIOR BANK - SYSTEMATIC INTERNALISER",
        ("MIC","SIBC"): "SIBC .. SIB CYPRUS LTD - SYSTEMATIC INTERNALISER",
        ("MIC","SICS"): "SICS .. CREDIT SUISSE SECURITIES (EUROPE) LIMITED - SYSTEMATIC INTERNALISER",
        ("MIC","SIGA"): "SIGA .. SIGMA X AUSTRALIA",
        ("MIC","SIGH"): "SIGH .. SIGMA X HONG KONG",
        ("MIC","SIGJ"): "SIGJ .. SIGMA X JAPAN",
        ("MIC","SIMV"): "SIMV .. SIM VENTURE SECURITIES EXCHANGE",
        ("MIC","SISI"): "SISI .. SUSQUEHANNA INTERNATIONAL SECURITIES LIMITED - SYSTEMATIC INTERNALISER",
        ("MIC","SKSI"): "SKSI .. SPAREKASSEN KRONJYLLAND - SYSTEMATIC INTERNALISER",
        ("MIC","SLHB"): "SLHB .. SALZBURGER LANDES-HYPOTHEKENBANK",
        ("MIC","SLKK"): "SLKK .. MTS SLOVAKIA",
        ("MIC","SMBB"): "SMBB .. SUMITOMO MITSUI BANKING CORPORATION - BRUSSELS BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","SMBD"): "SMBD .. SUMITOMO MITSUI BANKING CORPORATION -  DUESSELDORF BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","SMBE"): "SMBE .. SUMITOMO MITSUI BANKING CORPORATION EUROPE LIMITED",
        ("MIC","SMBG"): "SMBG .. SUMITOMO MITSUI BANKING CORPORATION - LONDON BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","SMBP"): "SMBP .. SUMITOMO MITSUI BANKING CORPORATION EUROPE LIMITED - PARIS BRANCH",
        ("MIC","SMEX"): "SMEX .. SINGAPORE MERCANTILE EXCHANGE PTE LTD",
        ("MIC","SMTS"): "SMTS .. MTS SPAIN",
        ("MIC","SNSI"): "SNSI .. SPAR NORD BANK - SYSTEMATIC INTERNALISER",
        ("MIC","SOHO"): "SOHO .. TWO SIGMA SECURITIES, LLC",
        ("MIC","SPEC"): "SPEC .. MAREX SPECTRON INTERNATIONAL LIMITED OTF",
        ("MIC","SPIM"): "SPIM .. ST. PETERSBURG INTERNATIONAL MERCANTILE EXCHANGE",
        ("MIC","SPRZ"): "SPRZ .. SPREADZERO",
        ("MIC","SPXE"): "SPXE .. SPX",
        ("MIC","SQUA"): "SQUA .. SQUARE GLOBAL - OTF",
        ("MIC","SSBI"): "SSBI .. STATE STREET BANK INTERNATIONAL GMBH",
        ("MIC","SSBM"): "SSBM .. STATE STREET BANK INTERNATIONAL FX - SYSTEMATIC INTERNALISER",
        ("MIC","SSBT"): "SSBT .. STATE STREET BANK AND TRUST COMPANY",
        ("MIC","SSEX"): "SSEX .. SOCIAL STOCK EXCHANGE",
        ("MIC","SSFX"): "SSFX .. STATE STREET BANK AND TRUST FX - SYSTEMATIC INTERNALISER",
        ("MIC","SSOB"): "SSOB .. BONDVISION EUROPE MTF",
        ("MIC","SSTX"): "SSTX .. E-EXCHANGE",
        ("MIC","SSWM"): "SSWM .. SSW MARKET MAKING - SYSTEMATIC INTERNALISER",
        ("MIC","STAN"): "STAN .. STANDARD CHARTERED - SYSTEMATIC INTERNALISER",
        ("MIC","STEE"): "STEE .. NASDAQ COMMODITIES - STEEL COMMODITY",
        ("MIC","STFL"): "STFL .. STIFEL NICOLAUS EUROPE LIMITED -SYSTEMATIC INTERNALISER",
        ("MIC","STOX"): "STOX .. STOXX LIMITED",
        ("MIC","STSI"): "STSI .. SUN TRADING INTERNATIONAL - SYSTEMATIC INTERNALISER",
        ("MIC","STUA"): "STUA .. BOERSE STUTTGART - REGULIERTER MARKT",
        ("MIC","STUB"): "STUB .. BOERSE STUTTGART - FREIVERKEHR",
        ("MIC","STUC"): "STUC .. BOERSE STUTTGART - REGULIERTER MARKT - TECHNICAL PLATFORM 2",
        ("MIC","STUD"): "STUD .. BOERSE STUTTGART - FREIVERKEHR - TECHNICAL PLATFORM 2",
        ("MIC","SUNB"): "SUNB .. SUNRISE BROKERS",
        ("MIC","SUNO"): "SUNO .. SUNRISE - OTF",
        ("MIC","SUNT"): "SUNT .. SUN TRADING LLC",
        ("MIC","SVES"): "SVES .. SVENSKA HANDELSBANKEN AB - SYSTEMATIC INTERNALISER",
        ("MIC","SVEX"): "SVEX .. SVENSKA HANDELSBANKEN AB - SVEX",
        ("MIC","SWAP"): "SWAP .. SWAPSTREAM",
        ("MIC","SWBI"): "SWBI .. SWEDBANK - SYSTEMATIC INTERNALISER",
        ("MIC","SWEE"): "SWEE .. SWEDBANK ESTONIA - SYSTEMATIC INTERNALISER",
        ("MIC","SWLT"): "SWLT .. SWEDBANK LITHUANIA - SYSTEMATIC INTERNALISER",
        ("MIC","SWLV"): "SWLV .. SWEDBANK LATVIA - SYSTEMATIC INTERNALISER",
        ("MIC","SZSC"): "SZSC .. STOCK EXCHANGE OF HONG KONG LIMITED - SHENZHEN - HONG KONG STOCK CONNECT",
        ("MIC","TBSP"): "TBSP .. WARSAW STOCK EXCHANGE/BONDS/BONDSPOT/TREASURY BOND MARKET",
        ("MIC","TCDS"): "TCDS .. TRADITION OTF",
        ("MIC","TCML"): "TCML .. TIDE CM",
        ("MIC","TDBL"): "TDBL .. TORONTO DOMINION BANK - LONDON BRANCH",
        ("MIC","TDON"): "TDON .. TRADITION PARIS - TSAF",
        ("MIC","TECO"): "TECO .. TULLETT PREBON EUROPE - OTF - COMMODITY AND ENERGY DERIVATIVES",
        ("MIC","TEEG"): "TEEG .. TULLETT PREBON SECURITIES - OTF - EUROPEAN GOVERNMENT BONDS",
        ("MIC","TEFD"): "TEFD .. TULLETT PREBON EUROPE -  MTF - FX DERIVATIVES",
        ("MIC","TEFX"): "TEFX .. TULLETT PREBON EUROPE - OTF - FX DERIVATIVES",
        ("MIC","TEGB"): "TEGB .. TULLETT PREBON EUROPE - OTF - GOVERNMENT BONDS EXCLUDING UK GILTS",
        ("MIC","TEGI"): "TEGI .. TULLETT PREBON EUROPE - OTF - GOVERNMENT BONDS - UK GILTS",
        ("MIC","TEIR"): "TEIR .. TULLETT PREBON EUROPE - OTF - INTEREST RATE DERIVATIVES",
        ("MIC","TEMB"): "TEMB .. TULLETT PREBON EUROPE -  MTF - GOVERNMENT BONDS EXCLUDING UK GILTS",
        ("MIC","TEMF"): "TEMF .. TULLETT PREBON EUROPE -  MTF - CORPORATE BONDS AND SECURITISED DEBT",
        ("MIC","TEMI"): "TEMI .. TULLETT PREBON EUROPE -  MTF - MONEY MARKETS",
        ("MIC","TEMM"): "TEMM .. TULLETT PREBON EUROPE - OTF - MONEY MARKET INSTRUMENTS",
        ("MIC","TEMR"): "TEMR .. TULLETT PREBON EUROPE -  MTF - REPOS",
        ("MIC","TERA"): "TERA .. TERAEXCHANGE",
        ("MIC","TERE"): "TERE .. TULLETT PREBON EUROPE - OTF - REPOS",
        ("MIC","TFEX"): "TFEX .. THAILAND FUTURES EXCHANGE",
        ("MIC","TFSA"): "TFSA .. TFS GREEN AUSTRALIAN GREEN MARKETS",
        ("MIC","TFSD"): "TFSD .. T.F.S. DERIVATIVES HK LIMITED",
        ("MIC","TFSU"): "TFSU .. TFS GREEN UNITED STATES GREEN MARKETS",
        ("MIC","TGAT"): "TGAT .. TRADEGATE EXCHANGE",
        ("MIC","TGSI"): "TGSI .. TRADEGATE EXCHANGE - SYSTEMATIC INTERNALISER",
        ("MIC","THEM"): "THEM .. THEMIS TRADING LLC",
        ("MIC","THRE"): "THRE .. THOMSON REUTERS (SEF) LLC",
        ("MIC","TIRD"): "TIRD .. TULLETT PREBON EUROPE - MTF - INTEREST RATE DERIVATIVES",
        ("MIC","TMID"): "TMID .. TRUMID ATS",
        ("MIC","TNLA"): "TNLA .. EURONEXT - TRADED BUT NOT LISTED AMSTERDAM",
        ("MIC","TNLB"): "TNLB .. EURONEXT - TRADING FACILITY BRUSSELS",
        ("MIC","TOCP"): "TOCP .. TORA CROSSPOINT",
        ("MIC","TOMG"): "TOMG .. TULLETT PREBON EUROPE - OTF - MADRID - GOVERNMENT BONDS - EXCLUDING GILTS",
        ("MIC","TPEL"): "TPEL .. TULLETT PREBON (EUROPE) LIMITED",
        ("MIC","TPEO"): "TPEO .. TULLETT PREBON EUROPE - OTF",
        ("MIC","TPES"): "TPES .. TULLETT PREBON EUROPE - OTF - MADRID",
        ("MIC","TPEU"): "TPEU .. TULLETT PREBON EU OTF",
        ("MIC","TPIC"): "TPIC .. TP ICAP (EUROPE) - MTF",
        ("MIC","TPIE"): "TPIE .. THE PROPERTY INVESTMENT EXCHANGE",
        ("MIC","TPIO"): "TPIO .. TP ICAP (EUROPE) - MTF - ORDERBOOK",
        ("MIC","TPIR"): "TPIR .. TP ICAP (EUROPE) - MTF - REGISTRATION",
        ("MIC","TPIS"): "TPIS .. TULLETT PREBON - INSTITUTIONAL SERVICES - OTF",
        ("MIC","TPSE"): "TPSE .. TP SEF, INC.",
        ("MIC","TPSL"): "TPSL .. TULLETT PREBON SECURITIES - MTF",
        ("MIC","TPSO"): "TPSO .. TULLETT PREBON SECURITIES - OTF",
        ("MIC","TQEA"): "TQEA .. TURQUOISE EUROPE - LIT AUCTIONS",
        ("MIC","TQEB"): "TQEB .. TURQUOISE EUROPE - NYLON CASH ORDER BOOK",
        ("MIC","TQEM"): "TQEM .. TURQUOISE EUROPE - DARK",
        ("MIC","TQEX"): "TQEX .. TURQUOISE EUROPE",
        ("MIC","TRAL"): "TRAL .. REUTERS TRANSACTION SERVICES LIMITED - FXALL RFQ",
        ("MIC","TRAS"): "TRAS .. TRADITION ASIA LIMITED",
        ("MIC","TRAX"): "TRAX .. TRAX APA",
        ("MIC","TRCK"): "TRCK .. TRACK ECN",
        ("MIC","TRDE"): "TRDE .. TRADITION",
        ("MIC","TRDX"): "TRDX .. TRAD-X",
        ("MIC","TREA"): "TREA .. TRADEWEB EUROPE LIMITED - APA",
        ("MIC","TREO"): "TREO .. TRADEWEB EUROPE LIMITED - OTF",
        ("MIC","TREU"): "TREU .. TRADEWEB EUROPE LIMITED",
        ("MIC","TRFW"): "TRFW .. REUTERS TRANSACTION SERVICES LIMITED - FORWARDS MATCHING",
        ("MIC","TRNL"): "TRNL .. TRAX NL APA",
        ("MIC","TRPX"): "TRPX .. TROP-X",
        ("MIC","TRQA"): "TRQA .. TURQUOISE LIT AUCTIONS",
        ("MIC","TRQB"): "TRQB .. TURQUOISE NYLON CASH ORDER BOOK",
        ("MIC","TRQC"): "TRQC .. TURQUOISE CFD",
        ("MIC","TRQM"): "TRQM .. TURQUOISE DARK",
        ("MIC","TRQS"): "TRQS .. TURQUOISE SWAPMATCH",
        ("MIC","TRQX"): "TRQX .. TURQUOISE",
        ("MIC","TRSI"): "TRSI .. TOWER RESEARCH CAPITAL EUROPE LTD",
        ("MIC","TRU1"): "TRU1 .. TRUEEX LLC - DESIGNATED CONTRACT MARKET (DMC)",
        ("MIC","TRU2"): "TRU2 .. TRUEEX LLC - SEF (SWAP EXECUTION FACILITY)",
        ("MIC","TRUX"): "TRUX .. TRUEEX LLC",
        ("MIC","TRWB"): "TRWB .. TRADEWEB LLC",
        ("MIC","TSAD"): "TSAD .. TRADITION SECURITIES AND DERIVATIVES INC.",
        ("MIC","TSAF"): "TSAF .. TSAF OTC - OTF",
        ("MIC","TSBX"): "TSBX .. DBOT ATS, LLC",
        ("MIC","TSCB"): "TSCB .. TULLETT PREBON SECURITIES - OTF - CORPORATE BONDS",
        ("MIC","TSCD"): "TSCD .. TULLETT PREBON SECURITIES - OTF - CREDIT DERIVATIVES",
        ("MIC","TSED"): "TSED .. TULLETT PREBON SECURITIES - OTF - EQUITY DERIVATIVES",
        ("MIC","TSEF"): "TSEF .. TRADITION SEF",
        ("MIC","TSFI"): "TSFI .. TULLETT PREBON SECURITIES - OTF - CORPORATE BONDS AND SECURITISED DEBT",
        ("MIC","TSGB"): "TSGB .. TULLETT PREBON SECURITIES - OTF - GOVERNMENT BONDS EXCLUDING UK GILTS",
        ("MIC","TSGI"): "TSGI .. TULLETT PREBON SECURITIES - OTF - GOVERNMENT BONDS - UK GILTS",
        ("MIC","TSIG"): "TSIG .. TRADITION SINGAPORE PTE. LTD.",
        ("MIC","TSMM"): "TSMM .. TULLETT PREBON SECURITIES - OTF - MONEY MARKET INSTRUMENTS",
        ("MIC","TSMR"): "TSMR .. TULLETT PREBON SECURITIES - MTF - REPOS",
        ("MIC","TSRE"): "TSRE .. TULLETT PREBON SECURITIES - OTF - REPOS",
        ("MIC","TWEA"): "TWEA .. TRADEWEB EU BV - APA",
        ("MIC","TWEM"): "TWEM .. TRADEWEB EU BV - MTF",
        ("MIC","TWEO"): "TWEO .. TRADEWEB EU BV - OTF",
        ("MIC","TWEU"): "TWEU .. TRADEWEB EU BV",
        ("MIC","TWGP"): "TWGP .. TRADEWEB EUROPE LIMITED - GLOBAL TRADING PLATFORM",
        ("MIC","TWHK"): "TWHK .. TRADEWEB EUROPE LIMITED - HONG KONG",
        ("MIC","TWJP"): "TWJP .. TRADEWEB JAPAN KK - PTS",
        ("MIC","TWJT"): "TWJT .. TRADEWEB JAPAN KK - ETP",
        ("MIC","TWSF"): "TWSF .. TW SEF LLC",
        ("MIC","TWSG"): "TWSG .. TRADEWEB EUROPE LIMITED - SINGAPORE",
        ("MIC","UBIM"): "UBIM .. UBI BANCA - SYSTEMATIC INTERNALISER",
        ("MIC","UBIN"): "UBIN .. UNION BANK OF INDIA UK LTD",
        ("MIC","UBIS"): "UBIS .. UNIPOL BANCA S.P.A.",
        ("MIC","UBSA"): "UBSA .. UBS ATS",
        ("MIC","UBSB"): "UBSB .. UBS AG LONDON BRANCH EMEA TRADING",
        ("MIC","UBSC"): "UBSC .. UBS PIN-FX",
        ("MIC","UBSD"): "UBSD .. UBS EUROPE SE",
        ("MIC","UBSE"): "UBSE .. UBS PIN (EMEA)",
        ("MIC","UBSF"): "UBSF .. UBS FX",
        ("MIC","UBSG"): "UBSG .. UBS TRADING",
        ("MIC","UBSI"): "UBSI .. UBS LIMITED EMEA TRADING - SYSTEMATIC INTERNALISER",
        ("MIC","UBSL"): "UBSL .. UBS LIMITED EMEA TRADING",
        ("MIC","UBSP"): "UBSP .. UBS PIN (UBS PRICE IMPROVEMENT NETWORK)",
        ("MIC","UBSS"): "UBSS .. UBS SECURITIES LLC",
        ("MIC","UBSX"): "UBSX .. UBS CROSS",
        ("MIC","UBSY"): "UBSY .. UBS AG LONDON BRANCH - SYSTEMATIC INTERNALISER",
        ("MIC","UCDE"): "UCDE .. UNICREDIT BANK AG - SYSTEMATIC INTERNALISER",
        ("MIC","UFEX"): "UFEX .. UFEX",
        ("MIC","UKEX"): "UKEX .. UKRAINIAN EXCHANGE",
        ("MIC","UKGD"): "UKGD .. UK GILTS MARKET",
        ("MIC","UKPX"): "UKPX .. APX POWER UK",
        ("MIC","ULTX"): "ULTX .. ALT XCHANGE (U)",
        ("MIC","USWB"): "USWB .. NASDAQ STOCKHOLM AB - USD WB EQ DERIVATIVES",
        ("MIC","UTSL"): "UTSL .. UEDA TRADITION SECURITIES LTD.",
        ("MIC","VABD"): "VABD .. BONDPOINT",
        ("MIC","VAGL"): "VAGL .. VIRTUAL AUCTION GLOBAL LIMITED",
        ("MIC","VCMO"): "VCMO .. VANTAGE CAPITAL MARKETS LLP - OTF",
        ("MIC","VEGA"): "VEGA .. VEGA-CHI",
        ("MIC","VERT"): "VERT .. VERTICAL",
        ("MIC","VFCM"): "VFCM .. VIRTU FINANCIAL CAPITAL MARKETS LLC",
        ("MIC","VFIL"): "VFIL .. VIRTU FINANCIAL IRELAND LIMITED",
        ("MIC","VFSI"): "VFSI .. VIRTU FINANCIAL IRELAND LIMITED - SYSTEMATIC INTERNALISER",
        ("MIC","VFXO"): "VFXO .. VIRTU FINANCIAL IRELAND LIMITED - OTC",
        ("MIC","VIRT"): "VIRT .. VIRTU FINANCIAL BD",
        ("MIC","VLEX"): "VLEX .. VONTOBEL LIQUIDITY EXTENDER",
        ("MIC","VMFX"): "VMFX .. THE FAROESE SECURITIES MARKET",
        ("MIC","VMTS"): "VMTS .. MTS SLOVENIA",
        ("MIC","VONT"): "VONT .. BANK VONTOBEL EUROPE AG - SYSTEMATIC INTERNALISER",
        ("MIC","VPXB"): "VPXB .. EURONEXT - VENTES PUBLIQUES BRUSSELS",
        ("MIC","VTBC"): "VTBC .. VTB CAPITAL PLC - SYSTEMATIC INTERNALISER",
        ("MIC","VTEX"): "VTEX .. VORTEX",
        ("MIC","VTLS"): "VTLS .. TRADELINK",
        ("MIC","VTPS"): "VTPS .. TRADEPLUS",
        ("MIC","VWDX"): "VWDX .. VWD TRANSACTIONSOLUTIONS AG",
        ("MIC","WBAH"): "WBAH .. WIENER BOERSE AG AMTLICHER HANDEL (OFFICIAL MARKET)",
        ("MIC","WBCL"): "WBCL .. WARSAW STOCK EXCHANGE/BONDS/CATALYST/LISTING",
        ("MIC","WBDM"): "WBDM .. WIENER BOERSE AG DRITTER MARKT (THIRD MARKET)",
        ("MIC","WBGF"): "WBGF .. WIENER BOERSE AG GEREGELTER FREIVERKEHR (SECOND REGULATED MARKET)",
        ("MIC","WBKP"): "WBKP .. BANK ZACHODNI WBK S.A. - SYSTEMATIC INTERNALISER",
        ("MIC","WBON"): "WBON .. WARSAW STOCK EXCHANGE/ BONDS/CATALYST/MAIN MARKET",
        ("MIC","WDER"): "WDER .. WARSAW STOCK EXCHANGE/FINANCIAL DERIVATIVES",
        ("MIC","WEED"): "WEED .. WEEDEN AND CO MARKETS",
        ("MIC","WELN"): "WELN .. WEL - SYSTEMATIC INTERNALISER",
        ("MIC","WELS"): "WELS .. WELLS FARGO SECURITIES, LLC - CASH EQUITIES",
        ("MIC","WETP"): "WETP .. WARSAW STOCK EXCHANGE/ ETPS",
        ("MIC","WGAS"): "WGAS .. WARSAW STOCK EXCHANGE/COMMODITIES/POLISH POWER EXCHANGE/GAS",
        ("MIC","WIND"): "WIND .. WARSAW STOCK EXCHANGE/INDICES",
        ("MIC","WINS"): "WINS .. WINTERFLOOD SECURITIES LIMITED - ELECTRONIC PLATFORM",
        ("MIC","WINX"): "WINX .. WINTERFLOOD SECURITIES LIMITED - MANUAL TRADING",
        ("MIC","WIPO"): "WIPO .. WARSAW STOCK EXCHANGE - IPO BOOK BUILDING",
        ("MIC","WMTF"): "WMTF .. WARSAW STOCK EXCHANGE/BONDS/CATALYST/MTF",
        ("MIC","WOOD"): "WOOD .. WOOD & COMPANY FINANCIAL SERVICES, A.S. - SYSTEMATIC INTERNALISER",
        ("MIC","WOPO"): "WOPO .. WARSAW STOCK EXCHANGE - SPO BOOK BUILDING",
        ("MIC","WQXL"): "WQXL .. EURONEXT - MARKET WITHOUT QUOTATIONS LISBON",
        ("MIC","WSAG"): "WSAG .. WALL STREET ACCESS",
        ("MIC","WSIL"): "WSIL .. WELLS FARGO SECURITIES INTERNATIONAL",
        ("MIC","WSIN"): "WSIN .. WESTPAC BANKING CORPORATION - SYSTEMATIC INTERNALISER",
        ("MIC","XABC"): "XABC .. ABG SUNDAL COLLIER AB - SYSTEMATIC INTERNALISER",
        ("MIC","XABG"): "XABG .. ABG SUNDAL COLLIER ASA - SYSTEMATIC INTERNALISER",
        ("MIC","XADE"): "XADE .. ATHENS EXCHANGE S.A. DERIVATIVES MARKET",
        ("MIC","XADF"): "XADF .. FINRA ALTERNATIVE DISPLAY FACILITY (ADF)",
        ("MIC","XADS"): "XADS .. ABU DHABI SECURITIES EXCHANGE",
        ("MIC","XAFR"): "XAFR .. ALTERNATIVA FRANCE",
        ("MIC","XAFX"): "XAFX .. AFRICAN STOCK EXCHANGE",
        ("MIC","XAIM"): "XAIM .. AIM ITALIA - MERCATO ALTERNATIVO DEL CAPITALE",
        ("MIC","XALG"): "XALG .. ALGIERS STOCK EXCHANGE",
        ("MIC","XALT"): "XALT .. ALTEX-ATS",
        ("MIC","XAMM"): "XAMM .. AMMAN STOCK EXCHANGE",
        ("MIC","XAMS"): "XAMS .. EURONEXT - EURONEXT AMSTERDAM",
        ("MIC","XAPA"): "XAPA .. EURONEXT - APA",
        ("MIC","XAQS"): "XAQS .. AUTOMATED EQUITY FINANCE MARKETS",
        ("MIC","XARM"): "XARM .. NASDAQ OMX ARMENIA",
        ("MIC","XASE"): "XASE .. NYSE MKT LLC",
        ("MIC","XASX"): "XASX .. ASX - ALL MARKETS",
        ("MIC","XATH"): "XATH .. ATHENS EXCHANGE S.A. CASH MARKET",
        ("MIC","XATL"): "XATL .. ATLANTIC SECURITIES MARKET",
        ("MIC","XATS"): "XATS .. ALPHA EXCHANGE",
        ("MIC","XBAA"): "XBAA .. BAHAMAS INTERNATIONAL SECURITIES EXCHANGE",
        ("MIC","XBAB"): "XBAB .. BARBADOS STOCK EXCHANGE",
        ("MIC","XBAH"): "XBAH .. BAHRAIN BOURSE",
        ("MIC","XBAN"): "XBAN .. BANGALORE STOCK EXCHANGE LTD",
        ("MIC","XBAR"): "XBAR .. BOLSA DE BARCELONA",
        ("MIC","XBBJ"): "XBBJ .. JAKARTA FUTURES EXCHANGE (BURSA BERJANGKA JAKARTA)",
        ("MIC","XBBK"): "XBBK .. PERIMETER FINANCIAL CORP. - BLOCKBOOK ATS",
        ("MIC","XBCC"): "XBCC .. BOLSA DE COMERCIO DE CORDOBA",
        ("MIC","XBCL"): "XBCL .. LA BOLSA ELECTRONICA DE CHILE",
        ("MIC","XBCM"): "XBCM .. BOLSA DE COMERCIO DE MENDOZA S.A.",
        ("MIC","XBCV"): "XBCV .. BOLSA CENTROAMERICANA DE VALORES S.A.",
        ("MIC","XBCX"): "XBCX .. MERCADO DE VALORES DE MENDOZA S.A.",
        ("MIC","XBDA"): "XBDA .. BERMUDA STOCK EXCHANGE LTD",
        ("MIC","XBDV"): "XBDV .. BOLSA DE DIVIDA E VALORES DE ANGOLA (BODIVA) - ANGOLA SECURITIES EXCHANGE",
        ("MIC","XBEL"): "XBEL .. BELGRADE STOCK EXCHANGE",
        ("MIC","XBER"): "XBER .. BOERSE BERLIN",
        ("MIC","XBES"): "XBES .. JSE CASH BOND MARKET",
        ("MIC","XBEY"): "XBEY .. BOURSE DE BEYROUTH - BEIRUT STOCK EXCHANGE",
        ("MIC","XBIL"): "XBIL .. BOLSA DE VALORES DE BILBAO",
        ("MIC","XBKF"): "XBKF .. STOCK EXCHANGE OF THAILAND - FOREIGN BOARD",
        ("MIC","XBKK"): "XBKK .. STOCK EXCHANGE OF THAILAND",
        ("MIC","XBLB"): "XBLB .. BANJA LUKA STOCK EXCHANGE",
        ("MIC","XBLN"): "XBLN .. BLUENEXT",
        ("MIC","XBNV"): "XBNV .. BOLSA NACIONAL DE VALORES, S.A.",
        ("MIC","XBOG"): "XBOG .. BOLSA DE VALORES DE COLOMBIA",
        ("MIC","XBOL"): "XBOL .. BOLSA BOLIVIANA DE VALORES S.A.",
        ("MIC","XBOM"): "XBOM .. BSE LTD",
        ("MIC","XBOS"): "XBOS .. NASDAQ OMX BX",
        ("MIC","XBOT"): "XBOT .. BOTSWANA STOCK EXCHANGE",
        ("MIC","XBOX"): "XBOX .. BOX OPTIONS EXCHANGE",
        ("MIC","XBRA"): "XBRA .. BRATISLAVA STOCK EXCHANGE",
        ("MIC","XBRD"): "XBRD .. EURONEXT - EURONEXT BRUSSELS - DERIVATIVES",
        ("MIC","XBRM"): "XBRM .. ROMANIAN  COMMODITIES EXCHANGE",
        ("MIC","XBRN"): "XBRN .. BX SWISS AG",
        ("MIC","XBRT"): "XBRT .. BRUT ECN",
        ("MIC","XBRU"): "XBRU .. EURONEXT - EURONEXT BRUSSELS",
        ("MIC","XBRV"): "XBRV .. BOURSE REGIONALE DES VALEURS MOBILIERES",
        ("MIC","XBSD"): "XBSD .. DERIVATIVES REGULATED MARKET - BVB",
        ("MIC","XBSE"): "XBSE .. SPOT REGULATED MARKET - BVB",
        ("MIC","XBTR"): "XBTR .. SIX SWISS BILATERAL TRADING PLATFORM FOR STRUCTURED OTC PRODUCTS",
        ("MIC","XBUD"): "XBUD .. BUDAPEST STOCK EXCHANGE",
        ("MIC","XBUE"): "XBUE .. BOLSA DE COMERCIO DE BUENOS AIRES",
        ("MIC","XBUL"): "XBUL .. BULGARIAN STOCK EXCHANGE",
        ("MIC","XBVC"): "XBVC .. CAPE VERDE STOCK EXCHANGE",
        ("MIC","XBVM"): "XBVM .. MOZAMBIQUE STOCK  EXCHANGE",
        ("MIC","XBVR"): "XBVR .. BOLSA DE VALORES DE LA REPUBLICA DOMINICANA SA.",
        ("MIC","XBXO"): "XBXO .. NASDAQ OMX BX OPTIONS",
        ("MIC","XCAI"): "XCAI .. EGYPTIAN EXCHANGE",
        ("MIC","XCAL"): "XCAL .. CALCUTTA STOCK EXCHANGE",
        ("MIC","XCAN"): "XCAN .. CAN-ATS",
        ("MIC","XCAS"): "XCAS .. CASABLANCA STOCK EXCHANGE",
        ("MIC","XCAY"): "XCAY .. CAYMAN ISLANDS STOCK EXCHANGE",
        ("MIC","XCBF"): "XCBF .. CBOE FUTURES EXCHANGE",
        ("MIC","XCBO"): "XCBO .. CBOE GLOBAL MARKETS INC.",
        ("MIC","XCBT"): "XCBT .. CHICAGO BOARD OF TRADE",
        ("MIC","XCDE"): "XCDE .. BAXTER FINANCIAL SERVICES",
        ("MIC","XCEC"): "XCEC .. COMMODITIES EXCHANGE CENTER",
        ("MIC","XCET"): "XCET .. UZBEK COMMODITY EXCHANGE",
        ("MIC","XCFE"): "XCFE .. CHINA FOREIGN EXCHANGE TRADE SYSTEM",
        ("MIC","XCFF"): "XCFF .. CANTOR FINANCIAL FUTURES EXCHANGE",
        ("MIC","XCGS"): "XCGS .. CHINESE GOLD & SILVER EXCHANGE SOCIETY",
        ("MIC","XCHG"): "XCHG .. CHITTAGONG STOCK EXCHANGE LTD.",
        ("MIC","XCHI"): "XCHI .. CHICAGO STOCK EXCHANGE, INC",
        ("MIC","XCIE"): "XCIE .. THE INTERNATIONAL STOCK EXCHANGE",
        ("MIC","XCIS"): "XCIS .. NATIONAL STOCK EXCHANGE, INC.",
        ("MIC","XCME"): "XCME .. CHICAGO MERCANTILE EXCHANGE",
        ("MIC","XCNF"): "XCNF .. BOLSA DE COMERCIO CONFEDERADA S.A.",
        ("MIC","XCNQ"): "XCNQ .. CANADIAN NATIONAL STOCK EXCHANGE",
        ("MIC","XCOL"): "XCOL .. COLOMBO STOCK EXCHANGE",
        ("MIC","XCOR"): "XCOR .. ICMA",
        ("MIC","XCRO"): "XCRO .. CROATIAN POWER EXCHANGE",
        ("MIC","XCSE"): "XCSE .. NASDAQ COPENHAGEN A/S",
        ("MIC","XCSX"): "XCSX .. CAMBODIA SECURITIES EXCHANGE",
        ("MIC","XCUE"): "XCUE .. UZBEKISTAN REPUBLICAN CURRENCY EXCHANGE",
        ("MIC","XCUR"): "XCUR .. CURRENEX",
        ("MIC","XCX2"): "XCX2 .. CX2",
        ("MIC","XCXD"): "XCXD .. NASDAQ CXD",
        ("MIC","XCYO"): "XCYO .. CYPRUS STOCK EXCHANGE - OTC",
        ("MIC","XCYS"): "XCYS .. CYPRUS STOCK EXCHANGE",
        ("MIC","XDAR"): "XDAR .. DAR ES  SALAAM STOCK EXCHANGE",
        ("MIC","XDBC"): "XDBC .. DEUTSCHE BOERSE AG - CUSTOMIZED INDICES",
        ("MIC","XDBV"): "XDBV .. DEUTSCHE BOERSE AG - VOLATILITY INDICES",
        ("MIC","XDBX"): "XDBX .. DEUTSCHE BOERSE AG - INDICES",
        ("MIC","XDCE"): "XDCE .. DALIAN COMMODITY EXCHANGE",
        ("MIC","XDES"): "XDES .. DELHI STOCK EXCHANGE",
        ("MIC","XDFB"): "XDFB .. JOINT-STOCK COMPANY \"STOCK EXCHANGE INNEX\"",
        ("MIC","XDFM"): "XDFM .. DUBAI FINANCIAL MARKET",
        ("MIC","XDHA"): "XDHA .. DHAKA STOCK EXCHANGE LTD",
        ("MIC","XDMI"): "XDMI .. ITALIAN DERIVATIVES MARKET",
        ("MIC","XDNB"): "XDNB .. DNB BANK ASA - SYSTEMATIC INTERNALISER",
        ("MIC","XDPA"): "XDPA .. CADE - MERCADO DE DEUDA PUBLICA ANOTADA",
        ("MIC","XDRF"): "XDRF .. AIAF - MERCADO DE RENTA FIJA",
        ("MIC","XDSE"): "XDSE .. DAMASCUS SECURITIES EXCHANGE",
        ("MIC","XDSM"): "XDSM .. BORSA ISTANBUL - DEBT SECURITIES MARKET",
        ("MIC","XDSX"): "XDSX .. DOUALA STOCK EXCHANGE",
        ("MIC","XDUB"): "XDUB .. IRISH STOCK EXCHANGE - ALL MARKET",
        ("MIC","XDUS"): "XDUS .. BOERSE DUESSELDORF",
        ("MIC","XEBI"): "XEBI .. ENERGY BROKING IRELAND GAS TRADING PLATFORM",
        ("MIC","XECB"): "XECB .. ECB EXCHANGE RATES",
        ("MIC","XECC"): "XECC .. EUROPEAN COMMODITY CLEARING AG",
        ("MIC","XECM"): "XECM .. MTF - CYPRUS EXCHANGE",
        ("MIC","XECS"): "XECS .. EASTERN CARIBBEAN SECURITIES EXCHANGE",
        ("MIC","XEDA"): "XEDA .. ELECTRICITY DAY-AHEAD MARKET",
        ("MIC","XEEE"): "XEEE .. EUROPEAN ENERGY EXCHANGE",
        ("MIC","XEEO"): "XEEO .. EUROPEAN ENERGY EXCHANGE - NON-MTF MARKET",
        ("MIC","XEER"): "XEER .. EUROPEAN ENERGY EXCHANGE - REGULATED MARKET",
        ("MIC","XEHQ"): "XEHQ .. EUREX REPO - HQLA MARKET",
        ("MIC","XEID"): "XEID .. ELECTRICITY INTRA-DAY MARKET",
        ("MIC","XELX"): "XELX .. ELX",
        ("MIC","XEMD"): "XEMD .. MERCADO MEXICANO DE DERIVADOS",
        ("MIC","XEMS"): "XEMS .. EMS EXCHANGE",
        ("MIC","XEQT"): "XEQT .. BOERSE BERLIN EQUIDUCT TRADING",
        ("MIC","XEQY"): "XEQY .. BORSA ISTANBUL - EQUITY MARKET",
        ("MIC","XERE"): "XERE .. EUREX REPO - FUNDING AND FINANCING PRODUCTS",
        ("MIC","XERT"): "XERT .. EUREX REPO - TRIPARTY",
        ("MIC","XESM"): "XESM .. IRISH STOCK EXCHANGE - ENTREPRISE SECURITIES MARKET",
        ("MIC","XETA"): "XETA .. XETRA - REGULIERTER MARKT",
        ("MIC","XETB"): "XETB .. XETRA - FREIVERKEHR",
        ("MIC","XETR"): "XETR .. XETRA",
        ("MIC","XEUB"): "XEUB .. EUREX BONDS",
        ("MIC","XEUC"): "XEUC .. EURONEXT COM, COMMODITIES FUTURES AND OPTIONS",
        ("MIC","XEUE"): "XEUE .. EURONEXT EQF, EQUITIES AND INDICES DERIVATIVES",
        ("MIC","XEUI"): "XEUI .. EURONEXT IRF, INTEREST RATE FUTURE AND OPTIONS",
        ("MIC","XEUM"): "XEUM .. EUREX REPO SECLEND MARKET",
        ("MIC","XEUP"): "XEUP .. EUREX REPO GMBH",
        ("MIC","XEUR"): "XEUR .. EUREX DEUTSCHLAND",
        ("MIC","XEYE"): "XEYE .. IRISH STOCK EXCHANGE  - GLOBAL EXCHANGE MARKET",
        ("MIC","XFCI"): "XFCI .. FINANCIALCONTENT - INDEXES",
        ("MIC","XFDA"): "XFDA .. FINANCIALCONTENT - DIGITAL ASSET TRADE REPORTING FACILITY",
        ("MIC","XFEX"): "XFEX .. FINANCIAL AND ENERGY EXCHANGE GROUP",
        ("MIC","XFKA"): "XFKA .. FUKUOKA STOCK EXCHANGE",
        ("MIC","XFNO"): "XFNO .. BORSA ISTANBUL - FUTURES AND OPTIONS MARKET",
        ("MIC","XFRA"): "XFRA .. DEUTSCHE BOERSE AG",
        ("MIC","XGAS"): "XGAS .. CENTRAL EASTERN EUROPEAN GAS EXCHANGE LTD",
        ("MIC","XGAT"): "XGAT .. TRADEGATE EXCHANGE - FREIVERKEHR",
        ("MIC","XGCL"): "XGCL .. GLOBAL COAL LIMITED",
        ("MIC","XGCX"): "XGCX .. GLOBAL COMMODITIES EXCHANGE",
        ("MIC","XGDX"): "XGDX .. GLOBAL DERIVATIVES EXCHANGE",
        ("MIC","XGEM"): "XGEM .. HONG KONG GROWTH ENTERPRISES MARKET",
        ("MIC","XGHA"): "XGHA .. GHANA STOCK EXCHANGE",
        ("MIC","XGME"): "XGME .. GESTORE MERCATO ELETTRICO - ITALIAN POWER EXCHANGE",
        ("MIC","XGMX"): "XGMX .. GLOBALCLEAR MERCANTILE EXCHANGE",
        ("MIC","XGRM"): "XGRM .. TRADEGATE EXCHANGE - REGULIERTER MARKT",
        ("MIC","XGSE"): "XGSE .. GEORGIA STOCK EXCHANGE",
        ("MIC","XGSX"): "XGSX .. GLOBAL SECURITIES EXCHANGE",
        ("MIC","XGTG"): "XGTG .. BOLSA DE VALORES NACIONAL SA",
        ("MIC","XGUA"): "XGUA .. BOLSA DE VALORES DE GUAYAQUIL",
        ("MIC","XHAM"): "XHAM .. HANSEATISCHE WERTPAPIERBOERSE HAMBURG",
        ("MIC","XHAN"): "XHAN .. NIEDERSAECHSISCHE BOERSE ZU HANNOVER",
        ("MIC","XHEL"): "XHEL .. NASDAQ HELSINKI LTD",
        ("MIC","XHKF"): "XHKF .. HONG KONG FUTURES EXCHANGE LTD.",
        ("MIC","XHKG"): "XHKG .. HONG KONG EXCHANGES AND CLEARING LTD",
        ("MIC","XHNF"): "XHNF .. HANOI STOCK EXCHANGE - DERIVATIVES",
        ("MIC","XHNX"): "XHNX .. HANOI STOCK EXCHANGE (UNLISTED PUBLIC COMPANY TRADING PLATFORM)",
        ("MIC","XIBE"): "XIBE .. BAKU INTERBANK CURRENCY EXCHANGE",
        ("MIC","XICB"): "XICB .. SIX CORPORATE BONDS AG",
        ("MIC","XICE"): "XICE .. NASDAQ ICELAND HF.",
        ("MIC","XICX"): "XICX .. INSTINET CANADA CROSS",
        ("MIC","XIDX"): "XIDX .. INDONESIA STOCK EXCHANGE",
        ("MIC","XIEL"): "XIEL .. INSTINET EUROPE LIMITED",
        ("MIC","XIHK"): "XIHK .. INSTINET PACIFIC LTD",
        ("MIC","XIMA"): "XIMA .. INTERNATIONAL MARTIME EXCHANGE",
        ("MIC","XIMC"): "XIMC .. MULTI COMMODITY EXCHANGE OF INDIA LTD.",
        ("MIC","XIMM"): "XIMM .. INTERNATIONAL MONETARY MARKET",
        ("MIC","XINE"): "XINE .. SHANGHAI INTERNATIONAL ENERGY EXCHANGE",
        ("MIC","XINS"): "XINS .. INSTINET",
        ("MIC","XINV"): "XINV .. INVESTRO",
        ("MIC","XIOM"): "XIOM .. INDEX AND OPTIONS MARKET",
        ("MIC","XIPO"): "XIPO .. HELEX ELECTRONIC BOOK BUILDING",
        ("MIC","XIQS"): "XIQS .. IRAQ STOCK EXCHANGE",
        ("MIC","XISA"): "XISA .. INTERNATIONAL SECURITIES EXCHANGE, LLC -  ALTERNATIVE MARKETS",
        ("MIC","XISE"): "XISE .. INTERNATIONAL SECURITIES EXCHANGE, LLC - EQUITIES",
        ("MIC","XIST"): "XIST .. BORSA ISTANBUL",
        ("MIC","XISX"): "XISX .. INTERNATIONAL SECURITIES EXCHANGE, LLC",
        ("MIC","XJAM"): "XJAM .. JAMAICA STOCK EXCHANGE",
        ("MIC","XJAS"): "XJAS .. TOKYO STOCK EXCHANGE JASDAQ",
        ("MIC","XJNB"): "XJNB .. JAKARTA NEGOTIATED BOARD",
        ("MIC","XJPX"): "XJPX .. JAPAN EXCHANGE GROUP",
        ("MIC","XJSE"): "XJSE .. JOHANNESBURG STOCK EXCHANGE",
        ("MIC","XKAC"): "XKAC .. OSAKA DOJIMA COMMODITY EXCHANGE",
        ("MIC","XKAR"): "XKAR .. THE PAKISTAN STOCK EXCHANGE LIMITED",
        ("MIC","XKAZ"): "XKAZ .. KAZAKHSTAN STOCK EXCHANGE",
        ("MIC","XKBT"): "XKBT .. KANSAS CITY BOARD OF TRADE",
        ("MIC","XKCE"): "XKCE .. KHOREZM INTERREGION COMMODITY EXCHANGE",
        ("MIC","XKCM"): "XKCM .. KOREA EXCHANGE COMMODITY MARKET",
        ("MIC","XKEM"): "XKEM .. KOREA EXCHANGE EMISSIONS MARKET",
        ("MIC","XKFB"): "XKFB .. KOREA FREEBOARD MARKET",
        ("MIC","XKFE"): "XKFE .. KOREA EXCHANGE (FUTURES MARKET)",
        ("MIC","XKHA"): "XKHA .. KHARTOUM STOCK EXCHANGE",
        ("MIC","XKHR"): "XKHR .. KHARKOV COMMODITY EXCHANGE",
        ("MIC","XKIE"): "XKIE .. KIEV UNIVERSAL EXCHANGE",
        ("MIC","XKIS"): "XKIS .. KIEV INTERNATIONAL STOCK EXCHANGE",
        ("MIC","XKLS"): "XKLS .. BURSA MALAYSIA",
        ("MIC","XKON"): "XKON .. KOREA NEW EXCHANGE",
        ("MIC","XKOS"): "XKOS .. KOREA EXCHANGE (KOSDAQ)",
        ("MIC","XKRX"): "XKRX .. KOREA EXCHANGE (STOCK MARKET)",
        ("MIC","XKSE"): "XKSE .. KYRGYZ STOCK EXCHANGE",
        ("MIC","XKUW"): "XKUW .. KUWAIT STOCK EXCHANGE",
        ("MIC","XLAO"): "XLAO .. LAO SECURITIES EXCHANGE",
        ("MIC","XLAT"): "XLAT .. LATIBEX",
        ("MIC","XLBM"): "XLBM .. LONDON BULLION MARKET",
        ("MIC","XLCH"): "XLCH .. LCH.CLEARNET LTD",
        ("MIC","XLDN"): "XLDN .. EURONEXT - EURONEXT LONDON",
        ("MIC","XLDX"): "XLDX .. LONDON DERIVATIVES EXCHANGE",
        ("MIC","XLFX"): "XLFX .. LABUAN INTERNATIONAL FINANCIAL EXCHANGE",
        ("MIC","XLGT"): "XLGT .. LGT BANK AG - SYSTEMATIC INTERNALISER",
        ("MIC","XLIM"): "XLIM .. BOLSA DE VALORES DE LIMA",
        ("MIC","XLIS"): "XLIS .. EURONEXT - EURONEXT LISBON",
        ("MIC","XLIT"): "XLIT .. AB NASDAQ VILNIUS",
        ("MIC","XLJM"): "XLJM .. SI ENTER",
        ("MIC","XLJU"): "XLJU .. LJUBLJANA STOCK EXCHANGE (OFFICIAL MARKET)",
        ("MIC","XLLB"): "XLLB .. LIECHTENSTEINISCHE LANDESBANK AG - SYSTEMATIC INTERNALISER",
        ("MIC","XLME"): "XLME .. LONDON METAL EXCHANGE",
        ("MIC","XLOD"): "XLOD .. LONDON STOCK EXCHANGE - DERIVATIVES MARKET",
        ("MIC","XLOM"): "XLOM .. LONDON STOCK EXCHANGE - MTF",
        ("MIC","XLON"): "XLON .. LONDON STOCK EXCHANGE",
        ("MIC","XLSM"): "XLSM .. LIBYAN STOCK MARKET",
        ("MIC","XLUS"): "XLUS .. LUSAKA STOCK EXCHANGE",
        ("MIC","XLUX"): "XLUX .. LUXEMBOURG STOCK EXCHANGE",
        ("MIC","XMAB"): "XMAB .. MERCADO ABIERTO ELECTRONICO S.A.",
        ("MIC","XMAD"): "XMAD .. BOLSA DE MADRID",
        ("MIC","XMAE"): "XMAE .. MACEDONIAN STOCK EXCHANGE",
        ("MIC","XMAI"): "XMAI .. MARKET FOR ALTERNATIVE INVESTMENT",
        ("MIC","XMAL"): "XMAL .. MALTA STOCK EXCHANGE",
        ("MIC","XMAN"): "XMAN .. BOLSA DE VALORES DE NICARAGUA",
        ("MIC","XMAT"): "XMAT .. EURONEXT PARIS MATIF",
        ("MIC","XMAU"): "XMAU .. STOCK EXCHANGE OF MAURITIUS LTD",
        ("MIC","XMDG"): "XMDG .. MARCHE INTERBANCAIRE DES DEVISES M.I.D.",
        ("MIC","XMDS"): "XMDS .. MADRAS STOCK EXCHANGE",
        ("MIC","XMER"): "XMER .. MERCHANTS' EXCHANGE",
        ("MIC","XMEV"): "XMEV .. MERCADO DE VALORES DE BUENOS AIRES S.A.",
        ("MIC","XMEX"): "XMEX .. BOLSA MEXICANA DE VALORES (MEXICAN STOCK EXCHANGE)",
        ("MIC","XMGE"): "XMGE .. MINNEAPOLIS GRAIN EXCHANGE",
        ("MIC","XMIL"): "XMIL .. BORSA ITALIANA S.P.A.",
        ("MIC","XMIO"): "XMIO .. MIAMI INTERNATIONAL SECURITIES EXCHANGE",
        ("MIC","XMLI"): "XMLI .. EURONEXT ACCESS PARIS",
        ("MIC","XMNT"): "XMNT .. BOLSA DE VALORES DE MONTEVIDEO",
        ("MIC","XMNX"): "XMNX .. MONTENEGRO STOCK EXCHANGE",
        ("MIC","XMOC"): "XMOC .. MONTREAL CLIMATE EXCHANGE",
        ("MIC","XMOD"): "XMOD .. THE MONTREAL EXCHANGE / BOURSE DE MONTREAL",
        ("MIC","XMOL"): "XMOL .. MOLDOVA STOCK EXCHANGE",
        ("MIC","XMON"): "XMON .. EURONEXT PARIS MONEP",
        ("MIC","XMOS"): "XMOS .. MOSCOW STOCK EXCHANGE",
        ("MIC","XMOT"): "XMOT .. EXTRAMOT",
        ("MIC","XMPW"): "XMPW .. MEFF POWER DERIVATIVES",
        ("MIC","XMRV"): "XMRV .. MEFF FINANCIAL DERIVATIVES",
        ("MIC","XMSM"): "XMSM .. IRISH STOCK EXCHANGE - MAIN SECURITIES MARKET",
        ("MIC","XMSW"): "XMSW .. MALAWI STOCK EXCHANGE",
        ("MIC","XMTB"): "XMTB .. MERCADO A TERMINO DE BUENOS AIRES S.A.",
        ("MIC","XMTS"): "XMTS .. EUROMTS LTD",
        ("MIC","XMUN"): "XMUN .. BOERSE MUENCHEN",
        ("MIC","XMUS"): "XMUS .. MUSCAT SECURITIES MARKET",
        ("MIC","XMVL"): "XMVL .. MERCADO DE VALORES DEL LITORAL S.A.",
        ("MIC","XNAF"): "XNAF .. SISTEMA ESPANOL DE NEGOCIACION DE ACTIVOS FINANCIEROS",
        ("MIC","XNAI"): "XNAI .. NAIROBI STOCK EXCHANGE",
        ("MIC","XNAM"): "XNAM .. NAMIBIAN STOCK EXCHANGE",
        ("MIC","XNAS"): "XNAS .. NASDAQ - ALL MARKETS",
        ("MIC","XNCD"): "XNCD .. NATIONAL COMMODITY & DERIVATIVES EXCHANGE LTD",
        ("MIC","XNCM"): "XNCM .. NASDAQ CAPITAL MARKET",
        ("MIC","XNCO"): "XNCO .. WARSAW STOCK EXCHANGE/ EQUITIES/NEW CONNECT-MTF",
        ("MIC","XNDQ"): "XNDQ .. NASDAQ OPTIONS MARKET",
        ("MIC","XNDX"): "XNDX .. NORDIC DERIVATIVES EXCHANGE",
        ("MIC","XNEC"): "XNEC .. NATIONAL STOCK EXCHANGE OF AUSTRALIA LIMITED",
        ("MIC","XNEP"): "XNEP .. NEPAL STOCK EXCHANGE",
        ("MIC","XNGM"): "XNGM .. NORDIC GROWTH MARKET",
        ("MIC","XNGO"): "XNGO .. NAGOYA STOCK EXCHANGE",
        ("MIC","XNGS"): "XNGS .. NASDAQ/NGS (GLOBAL SELECT MARKET)",
        ("MIC","XNIM"): "XNIM .. NASDAQ INTERMARKET",
        ("MIC","XNKS"): "XNKS .. CENTRAL JAPAN COMMODITIES EXCHANGE",
        ("MIC","XNLI"): "XNLI .. NYSE LIFFE",
        ("MIC","XNMR"): "XNMR .. NORDIC MTF REPORTING",
        ("MIC","XNMS"): "XNMS .. NASDAQ/NMS (GLOBAL MARKET)",
        ("MIC","XNOR"): "XNOR .. NORDEA - SYSTEMATIC INTERNALISER",
        ("MIC","XNSA"): "XNSA .. THE NIGERIAN STOCK EXCHANGE",
        ("MIC","XNSE"): "XNSE .. NATIONAL STOCK EXCHANGE OF INDIA",
        ("MIC","XNXC"): "XNXC .. NXCHANGE",
        ("MIC","XNYE"): "XNYE .. NEW YORK MERCANTILE EXCHANGE - OTC MARKETS",
        ("MIC","XNYL"): "XNYL .. NEW YORK MERCANTILE EXCHANGE - ENERGY MARKETS",
        ("MIC","XNYM"): "XNYM .. NEW YORK MERCANTILE EXCHANGE",
        ("MIC","XNYS"): "XNYS .. NEW YORK STOCK EXCHANGE, INC.",
        ("MIC","XNZE"): "XNZE .. NEW ZEALAND EXCHANGE LTD",
        ("MIC","XOAA"): "XOAA .. OSLO BORS ASA - OSLO AXESS LIT X AUCTIONS",
        ("MIC","XOAD"): "XOAD .. OSLO AXESS NORTH SEA - DARK POOL",
        ("MIC","XOAM"): "XOAM .. NORDIC ALTERNATIVE BOND MARKET",
        ("MIC","XOAS"): "XOAS .. OSLO AXESS",
        ("MIC","XOCH"): "XOCH .. ONECHICAGO, LLC",
        ("MIC","XODE"): "XODE .. ODESSA COMMODITY EXCHANGE",
        ("MIC","XOFF"): "XOFF .. OFF-EXCHANGE TRANSACTIONS - LISTED INSTRUMENTS",
        ("MIC","XOPV"): "XOPV .. OTC PUBLICATION VENUE",
        ("MIC","XOSA"): "XOSA .. OSLO BORS - LIT X AUCTIONS",
        ("MIC","XOSC"): "XOSC .. OSLO CONNECT",
        ("MIC","XOSD"): "XOSD .. OSLO BORS NORTH SEA - DARK POOL",
        ("MIC","XOSE"): "XOSE .. OSAKA EXCHANGE",
        ("MIC","XOSJ"): "XOSJ .. OSAKA EXCHANGE J-NET",
        ("MIC","XOSL"): "XOSL .. OSLO BORS ASA",
        ("MIC","XOTC"): "XOTC .. OTCBB",
        ("MIC","XPAC"): "XPAC .. POSIT AUCTION",
        ("MIC","XPAE"): "XPAE .. PALESTINE SECURITIES EXCHANGE",
        ("MIC","XPAR"): "XPAR .. EURONEXT - EURONEXT PARIS",
        ("MIC","XPBT"): "XPBT .. NASDAQ OMX FUTURES EXCHANGE",
        ("MIC","XPET"): "XPET .. SAINT PETERSBURG EXCHANGE",
        ("MIC","XPHL"): "XPHL .. NASDAQ OMX PHLX",
        ("MIC","XPHO"): "XPHO .. PHILADELPHIA OPTIONS EXCHANGE",
        ("MIC","XPHS"): "XPHS .. PHILIPPINE STOCK EXCHANGE, INC.",
        ("MIC","XPHX"): "XPHX .. PEEL HUNT CROSSING",
        ("MIC","XPIC"): "XPIC .. SAINT-PETERSBURG CURRENCY EXCHANGE",
        ("MIC","XPMS"): "XPMS .. BORSA ISTANBUL - PRECIOUS METALS AND DIAMONDS MARKETS",
        ("MIC","XPOM"): "XPOM .. PORT MORESBY STOCK EXCHANGE",
        ("MIC","XPOR"): "XPOR .. PORTAL",
        ("MIC","XPOS"): "XPOS .. POSIT DARK",
        ("MIC","XPOT"): "XPOT .. POWERNEXT - OTF",
        ("MIC","XPOW"): "XPOW .. POWERNEXT",
        ("MIC","XPRA"): "XPRA .. PRAGUE STOCK EXCHANGE",
        ("MIC","XPRI"): "XPRI .. PRIDNEPROVSK COMMODITY EXCHANGE",
        ("MIC","XPRM"): "XPRM .. PRAGUE STOCK EXCHANGE - MTF",
        ("MIC","XPSF"): "XPSF .. POWERNEXT - GAS SPOT AND FUTURES",
        ("MIC","XPST"): "XPST .. POSIT - ASIA PACIFIC",
        ("MIC","XPSX"): "XPSX .. NASDAQ OMX PSX",
        ("MIC","XPTY"): "XPTY .. BOLSA DE VALORES DE PANAMA, S.A.",
        ("MIC","XPXE"): "XPXE .. POWER EXCHANGE CENTRAL EUROPE",
        ("MIC","XQMH"): "XQMH .. SIX SWISS EXCHANGE - STRUCTURED PRODUCTS",
        ("MIC","XQTX"): "XQTX .. BOERSE DUESSELDORF - QUOTRIX",
        ("MIC","XQUI"): "XQUI .. BOLSA DE VALORES DE QUITO",
        ("MIC","XRAS"): "XRAS .. RASDAQ",
        ("MIC","XRBM"): "XRBM .. RINGGIT BOND MARKET",
        ("MIC","XRCB"): "XRCB .. RAIFFEISEN CENTROBANK AG - SYSTEMATIC INTERNALISER",
        ("MIC","XREP"): "XREP .. SIX REPO AG - CH REPO MARKET",
        ("MIC","XRFQ"): "XRFQ .. POSIT RFQ",
        ("MIC","XRIS"): "XRIS .. NASDAQ RIGA AS",
        ("MIC","XRMO"): "XRMO .. RM-SYSTEM CZECH STOCK EXCHANGE (MTF)",
        ("MIC","XRMZ"): "XRMZ .. RM-SYSTEM CZECH STOCK EXCHANGE",
        ("MIC","XROS"): "XROS .. BOLSA DE COMERCIO ROSARIO",
        ("MIC","XROT"): "XROT .. SIX REPO AG - OTC SPOT MARKET",
        ("MIC","XROX"): "XROX .. MERCADO DE VALORES DE ROSARIO S.A.",
        ("MIC","XRPM"): "XRPM .. ROMANIAN POWER MARKET",
        ("MIC","XRSP"): "XRSP .. PEEL HUNT RETAIL",
        ("MIC","XRUS"): "XRUS .. INTERNET DIRECT-ACCESS EXCHANGE",
        ("MIC","XSAF"): "XSAF .. JSE EQUITY DERIVATIVES MARKET",
        ("MIC","XSAM"): "XSAM .. SAMARA CURRENCY INTERBANK EXCHANGE",
        ("MIC","XSAP"): "XSAP .. SAPPORO SECURITIES EXCHANGE",
        ("MIC","XSAT"): "XSAT .. SPOTLIGHT STOCK MARKET",
        ("MIC","XSAU"): "XSAU .. SAUDI STOCK EXCHANGE",
        ("MIC","XSBI"): "XSBI .. SBI JAPANNEXT - X - MARKET",
        ("MIC","XSBT"): "XSBT .. SINGAPORE EXCHANGE BOND TRADING PTE. LTD",
        ("MIC","XSC1"): "XSC1 .. BOERSE FRANKFURT WARRANTS TECHNICAL 1",
        ("MIC","XSC2"): "XSC2 .. BOERSE FRANKFURT WARRANTS TECHNICAL 2",
        ("MIC","XSC3"): "XSC3 .. BOERSE FRANKFURT WARRANTS TECHNICAL 3",
        ("MIC","XSCA"): "XSCA .. SINGAPORE CATALIST MARKET",
        ("MIC","XSCE"): "XSCE .. SINGAPORE COMMODITY EXCHANGE",
        ("MIC","XSCO"): "XSCO .. BOERSE FRANKFURT WARRANTS TECHNICAL",
        ("MIC","XSCU"): "XSCU .. STOXX LIMITED - CUSTOMIZED INDICES",
        ("MIC","XSEC"): "XSEC .. SHENZHEN STOCK EXCHANGE - SHENZHEN - HONG KONG STOCK CONNECT",
        ("MIC","XSEF"): "XSEF .. SWAPEX, LLC",
        ("MIC","XSES"): "XSES .. SINGAPORE EXCHANGE",
        ("MIC","XSFA"): "XSFA .. JSE COMMODITY DERIVATIVES MARKET",
        ("MIC","XSFE"): "XSFE .. ASX - TRADE24",
        ("MIC","XSGA"): "XSGA .. SOCIETE GENERALE - SYSTEMATIC INTERNALISER",
        ("MIC","XSGE"): "XSGE .. SHANGHAI FUTURES EXCHANGE",
        ("MIC","XSGO"): "XSGO .. SANTIAGO STOCK EXCHANGE",
        ("MIC","XSHE"): "XSHE .. SHENZHEN STOCK EXCHANGE",
        ("MIC","XSHG"): "XSHG .. SHANGHAI STOCK EXCHANGE",
        ("MIC","XSIB"): "XSIB .. SIBERIAN EXCHANGE",
        ("MIC","XSIM"): "XSIM .. SINGAPORE EXCHANGE DERIVATIVES CLEARING LIMITED",
        ("MIC","XSLS"): "XSLS .. SIX SWISS EXCHANGE - SLS",
        ("MIC","XSMP"): "XSMP .. EURONEXT BLOCK",
        ("MIC","XSOP"): "XSOP .. BSP REGIONAL ENERGY EXCHANGE - SOUTH POOL",
        ("MIC","XSPM"): "XSPM .. EURONEXT STRUCTURED PRODUCTS MTF",
        ("MIC","XSPS"): "XSPS .. SOUTH PACIFIC STOCK EXCHANGE",
        ("MIC","XSSC"): "XSSC .. SHANGHAI STOCK EXCHANGE - SHANGHAI - HONG KONG STOCK CONNECT",
        ("MIC","XSSE"): "XSSE .. SARAJEVO STOCK EXCHANGE",
        ("MIC","XSTC"): "XSTC .. HOCHIMINH STOCK EXCHANGE",
        ("MIC","XSTE"): "XSTE .. REPUBLICAN STOCK EXCHANGE",
        ("MIC","XSTF"): "XSTF .. BOERSE STUTTGART - TECHNICAL PLATFORM 2",
        ("MIC","XSTM"): "XSTM .. FIDELITY CROSSSTREAM ATS",
        ("MIC","XSTO"): "XSTO .. NASDAQ STOCKHOLM AB",
        ("MIC","XSTU"): "XSTU .. BOERSE STUTTGART",
        ("MIC","XSTV"): "XSTV .. STOXX LIMITED - VOLATILITY INDICES",
        ("MIC","XSTX"): "XSTX .. STOXX LIMITED - INDICES",
        ("MIC","XSVA"): "XSVA .. EL SALVADOR STOCK EXCHANGE",
        ("MIC","XSWA"): "XSWA .. SWAZILAND STOCK EXCHANGE",
        ("MIC","XSWB"): "XSWB .. SWX SWISS BLOCK",
        ("MIC","XSWM"): "XSWM .. SIX SWISS EXCHANGE - SIX SWISS EXCHANGE AT MIDPOINT",
        ("MIC","XSWX"): "XSWX .. SIX SWISS EXCHANGE",
        ("MIC","XTAE"): "XTAE .. TEL AVIV STOCK EXCHANGE",
        ("MIC","XTAF"): "XTAF .. TAIWAN FUTURES EXCHANGE",
        ("MIC","XTAI"): "XTAI .. TAIWAN STOCK EXCHANGE",
        ("MIC","XTAL"): "XTAL .. NASDAQ TALLINN AS",
        ("MIC","XTAM"): "XTAM .. TOKYO STOCK EXCHANGE-TOKYO PRO MARKET",
        ("MIC","XTEH"): "XTEH .. TEHRAN STOCK EXCHANGE",
        ("MIC","XTFF"): "XTFF .. TOKYO FINANCIAL  EXCHANGE",
        ("MIC","XTIR"): "XTIR .. TIRANA STOCK EXCHANGE",
        ("MIC","XTK1"): "XTK1 .. TOKYO STOCK EXCHANGE - TOSTNET-1",
        ("MIC","XTK2"): "XTK2 .. TOKYO STOCK EXCHANGE - TOSTNET-2",
        ("MIC","XTK3"): "XTK3 .. TOKYO STOCK EXCHANGE - TOSTNET-3",
        ("MIC","XTKS"): "XTKS .. TOKYO STOCK EXCHANGE",
        ("MIC","XTKT"): "XTKT .. TOKYO COMMODITY EXCHANGE",
        ("MIC","XTND"): "XTND .. XTEND",
        ("MIC","XTNX"): "XTNX .. TSX VENTURE EXCHANGE - NEX",
        ("MIC","XTRN"): "XTRN .. TRINIDAD AND TOBAGO STOCK EXCHANGE",
        ("MIC","XTRZ"): "XTRZ .. ZAGREB MONEY AND SHORT TERM SECURITIES MARKET INC",
        ("MIC","XTSE"): "XTSE .. TORONTO STOCK EXCHANGE",
        ("MIC","XTSX"): "XTSX .. TSX VENTURE EXCHANGE",
        ("MIC","XTUC"): "XTUC .. NUEVA BOLSA DE COMERCIO DE TUCUMAN S.A.",
        ("MIC","XTUN"): "XTUN .. BOURSE DE TUNIS",
        ("MIC","XTXM"): "XTXM .. XTX MARKETS",
        ("MIC","XUAX"): "XUAX .. UKRAINIAN STOCK EXCHANGE",
        ("MIC","XUBS"): "XUBS .. UBS MTF",
        ("MIC","XUGA"): "XUGA .. UGANDA SECURITIES EXCHANGE",
        ("MIC","XUKR"): "XUKR .. UKRAINIAN UNIVERSAL COMMODITY EXCHANGE",
        ("MIC","XULA"): "XULA .. MONGOLIAN STOCK EXCHANGE",
        ("MIC","XUMP"): "XUMP .. UBS MTF - PERIODIC AUCTION",
        ("MIC","XUNI"): "XUNI .. UNIVERSAL BROKER'S EXCHANGE 'TASHKENT'",
        ("MIC","XUSE"): "XUSE .. UNITED STOCK EXCHANGE",
        ("MIC","XVAL"): "XVAL .. BOLSA DE VALENCIA",
        ("MIC","XVES"): "XVES .. VESTIMA",
        ("MIC","XVIE"): "XVIE .. WIENER BOERSE AG, WERTPAPIERBOERSE (SECURITIES EXCHANGE)",
        ("MIC","XVPA"): "XVPA .. BOLSA DE VALORES Y PRODUCTOS DE ASUNCION SA",
        ("MIC","XVPB"): "XVPB .. VP BANK AG - SYSTEMATIC INTERNALISER",
        ("MIC","XVTX"): "XVTX .. SIX SWISS EXCHANGE - BLUE CHIPS SEGMENT",
        ("MIC","XWAR"): "XWAR .. WARSAW STOCK EXCHANGE/EQUITIES/MAIN MARKET",
        ("MIC","XWBO"): "XWBO .. WIENER BOERSE AG",
        ("MIC","XWEE"): "XWEE .. WEEDEN ATS",
        ("MIC","XXSC"): "XXSC .. FRANKFURT CEF SC",
        ("MIC","XXXX"): "XXXX .. NO MARKET (E.G. UNLISTED)",
        ("MIC","XYIE"): "XYIE .. YIELDBROKER PTY LTD",
        ("MIC","XZAG"): "XZAG .. ZAGREB STOCK EXCHANGE",
        ("MIC","XZAM"): "XZAM .. THE ZAGREB STOCK EXCHANGE MTF",
        ("MIC","XZAP"): "XZAP .. PROGRESS MARKET",
        ("MIC","XZCE"): "XZCE .. ZHENGZHOU COMMODITY EXCHANGE",
        ("MIC","XZIM"): "XZIM .. ZIMBABWE STOCK EXCHANGE",
        ("MIC","YLDX"): "YLDX .. JSE INTEREST RATE DERIVATIVES MARKET",
        ("MIC","ZAPA"): "ZAPA .. ZAGREB STOCK EXCHANGE - APA",
        ("MIC","ZARX"): "ZARX .. ZAR X",
        ("MIC","ZFXM"): "ZFXM .. JSE CURRENCY DERIVATIVES MARKET",
        ("MIC","ZKBX"): "ZKBX .. ZURCHER KANTONALBANK SECURITIES EXCHANGE",
        ("MIC","ZOBX"): "ZOBX .. ZOBEX"
    }


# vim:ft=python
